/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_assemble.h"
#include "thumb_assemble.h"
#include "arm_assemble_one.h"
#include "arm_opcodes.h"
#include "arm_instructions.h"
#include "../../utils/util_error.h"
/* }}} */
/*!
 * 
 * Assemble an arm instruction (t_arm_ins to raw arm instruction)
 *
 * \param i_ins The instruction to assemble 
 *
 * \param data A buffer (must be at least 4 bytes long in which the instruction
 * is assembled 
 *
 * \return void 
*/
/* ArmAssembleOne {{{ */
void ArmAssembleOne(t_ins * i_ins, char * data)
{
  if (INS_TYPE(i_ins) != IT_DATA)
  {
    /* Get the opcode from the opcode table and set it */
    *((t_uint32 *)data) = arm_opcode_table[ARMINSOPCODE(i_ins)].opcode;

    /* Set the other bits (registers, immediates,...) */
    switch (INS_TYPE(i_ins))
    {
      case IT_BRANCH   : 
	ArmAssembleBranch(i_ins, (t_uint32*)data); 
	break;
      case IT_DATAPROC : 
	ArmAssembleDataProc(i_ins, (t_uint32*)data); 
	break;
      case IT_LOAD     : 
	ArmAssembleLoad(i_ins, (t_uint32*)data); 
	break;
      case IT_STORE    : 
	ArmAssembleStore(i_ins, (t_uint32*)data); 
	break;
      case IT_LOAD_MULTIPLE: 
      case IT_STORE_MULTIPLE: 
	ArmAssembleMultipleTransfer(i_ins, (t_uint32 *) data); 
	break;
      case IT_SWI:
	ArmAssembleSWI(i_ins, (t_uint32 *) data); 
	break;
      case IT_MUL:
	ArmAssembleMUL(i_ins, (t_uint32 *) data); 
	break;
           case IT_STATUS:
	ArmAssembleStatus(i_ins, (t_uint32 *) data);
	break;
      case IT_SWAP:
	ArmAssembleSwap(i_ins, (t_uint32 *) data);
	break;

      /* Floating point instructions (as used on the XScale) */
      case IT_FLT_STORE:
      case IT_FLT_LOAD:
	ArmAssembleFLTMEM(i_ins, (t_uint32 *) data);
	break;
      case IT_FLT_STATUS:
	ArmAssembleFLTStatus(i_ins, (t_uint32 *) data);
	break;
      case IT_FLT_ALU:
      case IT_FLT_INT:
	ArmAssembleFLTCPDO(i_ins, (t_uint32 *) data);
	break;

      /* Constant-/Addressproducers */
      case IT_CONSTS:
	/* Constant-/Addressproducers should already be transformed to normal
	 * Arm instructions */
	FATAL(("Attempt to assemble an address producer or a constant producer assembled: @I. This is impossible: these instructions should be converted to normal arm instructions by the layout code.",i_ins));
	break;

      default:
	/* Print the instruction we cannot assemble. Will probably fail,
	 * because it is likely that the print instruction is also not
	 * implemented. */
	FATAL(("Instruction type not implemented! Implement @I\n", i_ins));
    }
  } 
  else 
  {
    /* The value of a data pseudoinstruction is stored in the immediate field
     * of t_arm_ins */
    *((t_uint32 *) data)=ARMINSIMMEDIATE(i_ins);
  }
}
/* }}} */
/*!
 *  
 * Assemble the entire section. The assembled instructions are written to the
 * temporary buffer of the section (allocated by DiabloAction).
 *
 * \todo Should only have on parameter (section) and the relocation code at the
 * end should probably move to DiabloAction
 *
 * \param obj The object in which the section resides (remove me)
 *
 * \param sec The section to assemble
 *
 * \return void 
*/
/* ArmAssembleSection {{{ */
void ArmAssembleSection(t_object * obj, t_section * sec)
{
  t_uint32 address=0;
  int nins = 0;
  t_ins * i_ins;
  char * data = SECTION_TMP_BUF(sec);

  i_ins = sec->data; 
  address=G_T_UINT32(SECTION_GET_CADDRESS(sec));

  while(i_ins != NULL)
  {
    /* we have to reassemble it */
    /* Use thumb assemble one to assemble thumb instruction */
    if (ARMINSFLAGS(i_ins) & FL_THUMB) {
    	ThumbAssembleOne(i_ins, data);
	if ((ARMINSOPCODE(i_ins) == ARM_BL) || (ARMINSOPCODE(i_ins) == ARM_BLX))
		data += 2;
	data += 2;
    }
    else {
    /* Use arm assemble one to assemble the instruction */
    	ArmAssembleOne(i_ins, data);
    	nins++;
    	data += 4;
    	address+=4;
    }

    /* TODO: Move me to DiabloAction */
    while(RELOCATABLE_REFED_BY(i_ins))
      RelocSetToRelocatable(RELOCATABLE_REFED_BY(i_ins)->rel,T_RELOCATABLE(sec));

    while(RELOCATABLE_REFERS_TO(i_ins))
      RelocSetFrom(RELOCATABLE_REFERS_TO(i_ins)->rel,T_RELOCATABLE(sec));
    /* END TODO */
    
    i_ins = INS_NEXT(i_ins); 
  }
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
