/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_assemble.h"
#include "arm_opcodes.h"
#include "arm_registers.h"
#include "arm_print.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../utils/util_bitmanipulation.h" 
/* }}} */
/* Defines (utility functions) {{{ */
#define ASM_COND(_cc)          (_cc << 28)
#define ASM_IMM24(_imm)        ((_imm >> 2) & 0x00ffffff)
#define ASM_HAS_IMMEDIATE      (0x1 << 25)
#define ASM_LOADSTORE_WITH_REG (0x1 << 25)
#define ASM_UPDATES_CC         (0x1 << 20)
#define ASM_REGA(_reg)         (_reg << 12)
#define ASM_REGB(_reg)         (_reg << 16)
#define ASM_REGC(_reg)         (_reg)
#define ASM_REGS(_reg)         (_reg << 8)
#define ASM_SHREG_TYPE(_sh)    (((_sh - 4) << 5) | 0x00000010)
#define ASM_SHIMM_TYPE(_sh)    (_sh << 5)
#define ASM_SHIFT_IMM(imm)     (imm << 7)              /* amount to shift if there is a flexible   */
#define ASM_IMM_SHAMOUNT(_sh)  ((_sh / 2) << 8)        /* amount to shift if there is an immediate */
#define ASM_PREINDEX           (0x1 << 24)
#define ASM_DIRUP              (0x1 << 23)
#define ASM_WRITEBACK          (0x1 << 21)
/* }}} */
/*! 
 * \par ins The shift to assemble 
 * \return 32 bit word with flexible operand in the right place
 * 
 * Takes an instruction, and encodes the flexible operand in a 32 bit word,
 * so one can 'or' this with an instruction opcode
 *
 * \todo add some checks
 */
/* ArmEncodeShift {{{ */
t_uint32 ArmEncodeShift(t_ins * ins)
{
  t_uint32 ret = 0;
  switch (ARMINSSHIFTTYPE(ins))
    {
    case ARM_SHIFT_TYPE_LSL_IMM :
    case ARM_SHIFT_TYPE_LSR_IMM :
    case ARM_SHIFT_TYPE_ASR_IMM :
    case ARM_SHIFT_TYPE_ROR_IMM :
      /* shifting is done with an immediate */
      ret |= ASM_SHIMM_TYPE(ARMINSSHIFTTYPE(ins)) | ASM_SHIFT_IMM(ARMINSSHIFTLENGTH(ins));
      break;
    case ARM_SHIFT_TYPE_LSL_REG :
    case ARM_SHIFT_TYPE_LSR_REG :
    case ARM_SHIFT_TYPE_ASR_REG :
    case ARM_SHIFT_TYPE_ROR_REG :
      /* shifting is done with a register */
      if ((ARMINSREGS(ins) == ARM_REG_NONE))
        FATAL(("Assembly : Shift with register, but no register specified!"));
      ret |= ASM_REGS(ARMINSREGS(ins)) | ASM_SHREG_TYPE(ARMINSSHIFTTYPE(ins));
      break;
    case ARM_SHIFT_TYPE_RRX : ret |= ASM_SHIMM_TYPE(ARM_SHIFT_TYPE_ROR_IMM);
      break;
    case ARM_SHIFT_TYPE_NONE : ret |= ASM_SHIMM_TYPE(ARM_SHIFT_TYPE_LSL_IMM);
      break;
    default : FATAL(("Assembly : Unknown shifttype"));
    }
  return ret;
}
/* }}} */
/*! 
 * \par ins the instruction to assemble 
 * \par instr the 32 bit placeholder 
 * \return TRUE if immediate could be encoded, FALSE otherwise
 * 
 * Tries to encode an immediate in an instruction, return TRUE if succes
 */
/* ArmEncodeImmediate {{{ */
t_bool ArmEncodeImmediate(t_ins * ins, t_uint32 * instr)
{
  t_uint32 i;
  t_uint32 value = (t_uint32) ARMINSIMMEDIATE(ins);
  for (i = 0; i < 32; i+=2)
    {
      if ((BitRotateLeft(((t_uint32) value),i) & 0xffffff00) == 0)
	/* result can be stored in 8 bits */
	{
	  *instr |= BitRotateLeft(value,i); /* the rotated imm is the one to insert */
	  /* the amount to rotate to get the original imm back is i : TODO check this */
	  *instr |= ASM_IMM_SHAMOUNT(i);
	  return TRUE;
	}
    }
  return FALSE;
}
/* }}} */
/* Arm Integer  Assemble Functions {{{ */
/*!
 * \par ins the instruction to assemble (t_arm_ins*)
 * \par instr the 32 bit placeholder (t_uint32*)
 * 
 * Assembles B and BL
 */
/* ArmAssembleBranch {{{ */
void ArmAssembleBranch(t_ins * ins, t_uint32 * instr)
{
  *instr= arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
  if (INS_TYPE(ins) != IT_BRANCH)
    FATAL(("Wrong assembly function (BRANCH) for instruction of type %s", arm_opcode_table[ARMINSOPCODE(ins)].opcode));
  switch (ARMINSOPCODE(ins))
  {
  	case ARM_B:
  	case ARM_BL:
	  *instr |= ASM_COND(ARMINSCONDITION(ins));
	  *instr |= ASM_IMM24(ARMINSIMMEDIATE(ins));
	  break;
	case ARM_BX:
	case ARM_BLX:
	  *instr|= ASM_COND(ARMINSCONDITION(ins));
	  *instr|=(ARMINSREGB(ins));
	  break;
	default:
	  FATAL(("Unknown branch instruction!"));
  }

}
/* }}} */
/*!
 * Assembles Arm swap (SWP) instructions
 *
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 * \return void 
*/
/* ArmAssembleSwap {{{ */
void ArmAssembleSwap(t_ins * ins, t_uint32 * instr)
{
   *instr= arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
   *instr |= ASM_COND(ARMINSCONDITION(ins));
   *instr|=(ARMINSREGA(ins)<<12);
   *instr|=(ARMINSREGC(ins));
   *instr|=(ARMINSREGB(ins)<<16);
}
/* }}} */
/*! 
 * \par ins the instruction to assemble (t_arm_ins*)
 * \par instr the 32 bit placeholder (t_uint32*)
 * 
 * Assembles Mov and Mvn, i
 */
/* ArmAssembleDataProc {{{ */
void ArmAssembleDataProc(t_ins * ins, t_uint32 * instr)
{
  t_uint32 tmp = 0;
  if (INS_TYPE(ins) != IT_DATAPROC)
    FATAL(("Wrong assembly function (DATAPROC) for instruction %s", arm_opcode_table[ARMINSOPCODE(ins)].desc));

  *instr= arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */

  *instr |= ASM_COND(ARMINSCONDITION(ins)); /* set conditionbits */

  if (ARMINSFLAGS(ins) & FL_S)	/* set updates-flag */
    *instr |= ASM_UPDATES_CC;

  if (ARMINSFLAGS(ins) & FL_IMMED)
    {
      /* encode the immediate */
      if (ArmEncodeImmediate(ins,&tmp) == FALSE)
	FATAL(("Immediate %8x cannot be assembled in an ARM-instruction!", ARMINSIMMEDIATE(ins)));
      /* TODO : find a solution if this happens! */
      else
	{
	  *instr |= tmp;
	  *instr |= ASM_HAS_IMMEDIATE;
	}
    }
  else
    {
      /* encode the flexible operand */
      *instr |= ASM_REGC(ARMINSREGC(ins));
      *instr |= ArmEncodeShift(ins);
    }

  /* fill in ARMINSREGA */
  if (ARMINSREGA(ins) != ARM_REG_NONE)
    *instr |= ASM_REGA(ARMINSREGA(ins));
  else
    *instr |= ASM_REGA(0);

  /* fill in ARMINSREGB */
  if (ARMINSREGB(ins) != ARM_REG_NONE)
    *instr |= ASM_REGB(ARMINSREGB(ins));
  else
    *instr |= ASM_REGB(0);
}
/* }}} */
/*!
 * Assembles Arm System Calls (SWI)
 *
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 * \return void 
*/
/* ArmAssembleSWI {{{ */
void ArmAssembleSWI(t_ins * ins, t_uint32 * instr)
{
   *instr = arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
   *instr|= ASM_COND(ARMINSCONDITION(ins)); /* set conditionbits */
   *instr|=ARMINSIMMEDIATE(ins);
}
/* }}} */
/*!
 * Assembles Multiplies (MUL, MULL, ...)
 *
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 * \return void 
*/
/* ArmAssembleMUL {{{ */
void  ArmAssembleMUL(t_ins * ins, t_uint32 * instr)
{
   *instr = arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
   *instr |= ASM_COND(ARMINSCONDITION(ins)); /* set conditionbits */
   if (ARMINSFLAGS(ins) & FL_S) *instr |=(1 << 20); 
   *instr|=(ARMINSREGA(ins)<<16);
   *instr|=(ARMINSREGB(ins));
   *instr|=(ARMINSREGC(ins)<<8);
   *instr|=(ARMINSREGS(ins)<<12);
}
/* }}} */
/*!
 * Assembles the instructions that query the ARM Status registers (SPSR and
 * ...): MSR, ... 
 * 
 * \todo MRS is not implemented
 * \todo What was the name of the other status register? 
 *
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 * \return void 
*/
/* ArmAssembleStatus {{{ */
void ArmAssembleStatus(t_ins * ins, t_uint32 * instr)
{
  t_bool set=FALSE;
  *instr = arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
  *instr |= ASM_COND(ARMINSCONDITION(ins)); /* set conditionbits */
  if (ARMINSFLAGS(ins) & FL_SPSR) *instr |=(1 << 22); 
  if (ARMINSOPCODE(ins)==ARM_MSR) 
  {
    if (ARMINSFLAGS(ins) & FL_S)  {
      *instr |=(1 << 19);
      set=TRUE;
    }
    
    if (!set) FATAL(("MSR is NOOP, @I",ins));

    if (ARMINSFLAGS(ins) & FL_IMMED)
      {
	FATAL(("Implement constant ARM_STATUS"));
      }
    else
      {
	*instr|=ARMINSREGC(ins);
      }
    
  }
  else
    {
      *instr|=(ARMINSREGA(ins)<<12);
    }
}
/* }}} */
/*!
 * Assembles the normal loads (LDR, LDRH, ...) 
 *
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 * \return void 
*/
/* ArmAssembleLoad {{{ */
void ArmAssembleLoad(t_ins * ins, t_uint32 * instr)
{
  *instr = arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
  *instr |= ASM_COND(ARMINSCONDITION(ins)); /* set conditionbits */

  *instr |= ASM_REGA(ARMINSREGA(ins));
  *instr |= ASM_REGB(ARMINSREGB(ins));

  if (ARMINSFLAGS(ins) & FL_PREINDEX)
    *instr |= ASM_PREINDEX;
  if (ARMINSFLAGS(ins) & FL_DIRUP)
    *instr |= ASM_DIRUP;
  if (ARMINSFLAGS(ins) & FL_WRITEBACK)
    {
      if (ARMINSFLAGS(ins) & FL_PREINDEX)
	*instr |= ASM_WRITEBACK;
      else
	FATAL(("T-variant of loads not implemented yet! Fixme!"));
    }

  if (ARMINSOPCODE(ins) == ARM_LDR || ARMINSOPCODE(ins) == ARM_LDRB)
    {
      if (ARMINSFLAGS(ins) & FL_IMMED)
	{
	  /* encode the immediate */
	  if (ARMINSIMMEDIATE(ins) > 4095) FATAL(("Only 12bit immediates can be assembled in a LDR/LDRB instruction! %d, @I",ARMINSIMMEDIATE(ins),ins));
	  *instr |= ARMINSIMMEDIATE(ins);
	} 
      else 
	{
	  /* encode the flexible operand */
	  *instr |= ASM_REGC(ARMINSREGC(ins));
	  *instr |= ArmEncodeShift(ins);
	  *instr |= ASM_LOADSTORE_WITH_REG;
	}
    } 
  else if (ARMINSOPCODE(ins) == ARM_LDRSH  || ARMINSOPCODE(ins) == ARM_LDRH || ARMINSOPCODE(ins)==ARM_LDRSB)
    {
      if (ARMINSFLAGS(ins) & FL_IMMED)
	{
	  if (ARMINSIMMEDIATE(ins) > 256) FATAL(("Only 8bit immediates can be assembled in a LDR/LDRB instruction!\nIns was @I",ins));
	  *instr |= ((ARMINSIMMEDIATE(ins)<<4) & 0xf00) | (ARMINSIMMEDIATE(ins) & 0xf);
	  *instr |= 1<<22;
	}
      else
	{
	  if (ARMINSSHIFTTYPE(ins) != ARM_SHIFT_TYPE_NONE)
	    FATAL(("flexible operand not supported for ARM_LDR(S)(H/B)\nInstruction was: @I\nShifttype was %d\n",ins,ARMINSSHIFTTYPE(ins)));
	  
	  *instr |= ASM_REGC(ARMINSREGC(ins));
	  *instr |= ArmEncodeShift(ins);
	}
    }
  else
    {
      FATAL(("Only LDR and LDRB implemented for assembly"));
    }
}
/* }}} */
/*!
 *  Assembles the normal stores (STR, STRH, ...)
 *
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 * \return void 
*/
/* ArmAssembleStore {{{ */
void ArmAssembleStore(t_ins * ins, t_uint32 * instr)
{
  *instr = arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
  *instr |= ASM_COND(ARMINSCONDITION(ins)); /* set conditionbits */

  *instr |= ASM_REGA(ARMINSREGA(ins));
  *instr |= ASM_REGB(ARMINSREGB(ins));

  if (ARMINSFLAGS(ins) & FL_PREINDEX)
    *instr |= ASM_PREINDEX;
  if (ARMINSFLAGS(ins) & FL_DIRUP)
    *instr |= ASM_DIRUP;
  if (ARMINSFLAGS(ins) & FL_WRITEBACK)
    {
      if (ARMINSFLAGS(ins) & FL_PREINDEX)
	*instr |= ASM_WRITEBACK;
      else
	FATAL(("T-variant of stores not implemented yet! Fixme!"));
    }

  if (ARMINSOPCODE(ins) == ARM_STR || ARMINSOPCODE(ins) == ARM_STRB)
    {
      if (ARMINSFLAGS(ins) & FL_IMMED)
	{
	  /* encode the immediate */
	  if (ARMINSIMMEDIATE(ins) > 4095) FATAL(("Only 12bit immediates can be assembled in a LDR/LDRB instruction!"));
	  *instr |= ARMINSIMMEDIATE(ins);
	} 
      else 
	{
	  /* encode the flexible operand */
	  *instr |= ASM_REGC(ARMINSREGC(ins));
	  *instr |= ArmEncodeShift(ins);
	  *instr |= ASM_LOADSTORE_WITH_REG;
	}
    } 
  else if (ARMINSOPCODE(ins) == ARM_STRH)
    {
      if (ARMINSFLAGS(ins) & FL_IMMED)
	{
	  if (ARMINSIMMEDIATE(ins) > 255) FATAL(("Only 8bit immediates can be assembled in a LDR/LDRB instruction!"));
	  *instr |= ((ARMINSIMMEDIATE(ins)<<4) & 0xf00) | (ARMINSIMMEDIATE(ins) & 0xf);
	  *instr |= 1<<22;
	}
      else
	{
	  if (ARMINSSHIFTTYPE(ins) != ARM_SHIFT_TYPE_NONE)
	    FATAL(("flexible operand not supported for ARM_STRH\n"));
	  
	  *instr |= ASM_REGC(ARMINSREGC(ins));
	  *instr |= ArmEncodeShift(ins);
	}
    }
  else 
    {
      FATAL(("Only STR and STRB implemented for assembly"));
    }
}
/* }}} */
/*!
 * Assembles Multiple loads/stores (LDM, STM) 
 *
 * \todo Remove dependence on immediates (use multiple...)
 * 
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 * \return void 
*/
/* ArmAssembleMultipleTransfer {{{ */
void ArmAssembleMultipleTransfer(t_ins * ins, t_uint32 * instr)
{
  *instr = arm_opcode_table[ARMINSOPCODE(ins)].opcode;
  *instr |= ASM_COND(ARMINSCONDITION(ins));
   
  if (ARMINSFLAGS(ins) & FL_PREINDEX)
    *instr |= ASM_PREINDEX;
  if (ARMINSFLAGS(ins) & FL_DIRUP)
    *instr |= ASM_DIRUP;
  if (ARMINSFLAGS(ins) & FL_WRITEBACK)
    *instr |= ASM_WRITEBACK;

  *instr |= ASM_REGB(ARMINSREGB(ins));
  *instr |= ARMINSIMMEDIATE(ins) & 0xffff;
}
/* }}} */
/* }}} */
/* Arm Floating Assemble Functions {{{ */
/*!
 * Assembles the instructions that get/set the floating point status register
 * (WFS and RFS). 
 *
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 * \return void 
*/
/* ArmAssembleFLTStatus {{{ */
void ArmAssembleFLTStatus(t_ins * ins, t_uint32 * instr)
{
  *instr = arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
  *instr |= ASM_COND(ARMINSCONDITION(ins)); /* set conditionbits */

  if (ARMINSOPCODE(ins)==ARM_WFS)
  {
    *instr|=(ARMINSREGB(ins)<<12);
  }
  else if (ARMINSOPCODE(ins)==ARM_RFS)
  {
    *instr|=(ARMINSREGA(ins)<<12);
  }
  else
  {
    FATAL(("Unknown ARM_FLT_Status!"));
  }
}
/* }}} */
/*!
 * Assembles all but the memory and status register floating point
 * instructions.
 *
 * \todo change the name
 * \todo rounding mode does not get set in all cases, simple to fix
 *
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 * \return void 
*/
/* ArmAssembleFLTCPDO {{{ */
void ArmAssembleFLTCPDO(t_ins * ins,t_uint32 * instr)
{
  *instr = arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
  *instr |= ASM_COND(ARMINSCONDITION(ins)); /* set conditionbits */
   switch(ARMINSOPCODE(ins))
   {
     case ARM_MVF:
     case ARM_MNF:
     case ARM_SQT:
     case ARM_ABS:
       *instr|=((ARMINSREGA(ins) - ARM_REG_F0 )<<12);
      
       if (ARMINSFLAGS(ins)&FL_IMMED)
       {
	 if (ARMINSFLAGS(ins)&FL_FLT_IMMED)
	 {
	   *instr|=0x8;
	   *instr|= (ARMINSIMMEDIATE(ins) & 0x7);
	 }
	 else
	 {
	   FATAL(("IMPLEMENT!"));
	 }
       }
       else
	 *instr|=ARMINSREGB(ins) - ARM_REG_F0;
       if (ARMINSFLAGS(ins)& FL_FLT_DOUBLE_EXTENDED)
	 *instr|=1<<19;
       else if (ARMINSFLAGS(ins)& FL_FLT_DOUBLE)
	 *instr|=1<<7;
       else if (!(ARMINSFLAGS(ins)& FL_FLT_SINGLE))
        FATAL(("No size for float"));
       if (ARMINSFLAGS(ins) & FL_FLT_RZ) *instr|= (0x3 << 5);
       if (ARMINSFLAGS(ins) & FL_FLT_RMI) *instr|= (0x1 << 6);
       if (ARMINSFLAGS(ins) & FL_FLT_RPI) *instr|= (0x1 << 5);
       break;
     case ARM_FLT:
        *instr|=((ARMINSREGA(ins)- ARM_REG_F0)<<16);
	*instr|=ARMINSREGB(ins)<<12;
	if (ARMINSFLAGS(ins)& FL_FLT_DOUBLE_EXTENDED)
	  *instr|=1<<19;
	else if (ARMINSFLAGS(ins)& FL_FLT_DOUBLE)
	  *instr|=1<<7;
	else if (!(ARMINSFLAGS(ins)& FL_FLT_SINGLE))
	  FATAL(("No size for float"));
	break;
     case ARM_ADF:
     case ARM_SUF:
     case ARM_RSF:
     case ARM_MUF:
     case ARM_DVF:
     case ARM_RDF:
     case ARM_FDV:
     case ARM_FRD:
     case ARM_FML:
	*instr|=((ARMINSREGB(ins)- ARM_REG_F0)<<16);
	*instr|=((ARMINSREGA(ins)- ARM_REG_F0)<<12);
	if (ARMINSREGC(ins)!=ARM_REG_NONE)
	{
	  *instr|=((ARMINSREGC(ins)- ARM_REG_F0));
	}
	else
	{
	  if ((ARMINSFLAGS(ins)&FL_IMMED) && (ARMINSFLAGS(ins)&FL_FLT_IMMED))
	  {
	    *instr|=0x8;
	    *instr|= (ARMINSIMMEDIATE(ins) & 0x7);
	  }
	  else
	    FATAL(("implement"));
	}
	if (ARMINSFLAGS(ins)& FL_FLT_DOUBLE_EXTENDED)
	  *instr|=1<<19;
	else if (ARMINSFLAGS(ins)& FL_FLT_DOUBLE)
	  *instr|=1<<7;
	else if (!(ARMINSFLAGS(ins)& FL_FLT_SINGLE))
	  FATAL(("No size for float"));
	break;
     case ARM_CMFE:
     case ARM_CMF:
     case ARM_CNF:
     case ARM_CNFE:
       *instr|=((ARMINSREGB(ins)- ARM_REG_F0)<<16);
       if (ARMINSFLAGS(ins)&FL_IMMED)
       {
	 if (ARMINSFLAGS(ins)&FL_FLT_IMMED)
	 {
	   *instr|=0x8;
	   *instr|= (ARMINSIMMEDIATE(ins) & 0x7);
	 }
	 else
	 {
	   FATAL(("IMPLEMENT!"));
	 }
       }
       else
	 *instr|=((ARMINSREGC(ins)- ARM_REG_F0));
       break;
     case ARM_FIX:
       *instr|=ARMINSREGA(ins)<<12;
       *instr|=(ARMINSREGB(ins)- ARM_REG_F0);
       if (ARMINSFLAGS(ins) & FL_FLT_RZ) *instr|= (0x3 << 5);
       if (ARMINSFLAGS(ins) & FL_FLT_RMI) *instr|= (0x1 << 6);
       if (ARMINSFLAGS(ins) & FL_FLT_RPI) *instr|= (0x1 << 5);
       break;
     default:
       FATAL(("Implement @I",ins));

   }

}
/* }}} */
/*!
 * Assembles the Arm floating point loads and stores
 *
 * \param ins the instruction to assemble (t_arm_ins*)
 * \param instr the 32 bit placeholder (t_uint32*)
 *
 *
 * \todo Check if all dataformats for loads and stores are implemented correct
 * 
 * \return void 
*/
/* ArmAssembleFLTMEM {{{ */
void ArmAssembleFLTMEM(t_ins * ins, t_uint32 * instr)
{
  *instr = arm_opcode_table[ARMINSOPCODE(ins)].opcode;  /* Set the opcode */
  *instr |= ASM_COND(ARMINSCONDITION(ins)); /* set conditionbits */
  if ((ARMINSOPCODE(ins)==ARM_SFM)||(ARMINSOPCODE(ins)==ARM_LFM))
  {
    t_reg tel;
    t_reg min=ARM_REG_NONE;
    t_reg max=ARM_REG_NONE;
    for (tel=0; tel<32; tel++)
    {
      if (RegsetIn(ARMINSMULTIPLE(ins),tel))
      {
	if ((tel<ARM_REG_F0) || (tel>ARM_REG_F7))
	{
	  FATAL(("Illegal register in SFM"));
	}
	else
	{
	  if ((max!=ARM_REG_NONE)&&(max+1!=tel))
	  {
	    FATAL(("Wrong regs for SFM!"));
	  }
	  else if (min==ARM_REG_NONE)
	  {
	    min=max=tel;
	  }
	  else
	  {
	    max=tel;
	  }
	}
	
      }
    }

    if (max-min>3) FATAL(("Too much registers in SFM"));

    *instr|=(ARMINSREGB(ins)<<16);
    *instr|=(min-ARM_REG_F0)<<12;
    if ((max-min)==0)
      *instr|=1<<15;
    else if ((max-min)==1)
      *instr|=1<<22;
    else if ((max-min)==2)
    {
      *instr|=1<<15;
      *instr|=1<<22;
    }
    *instr|=ARMINSIMMEDIATE(ins);
    if (ARMINSFLAGS(ins)&FL_PREINDEX) *instr|=1 << 24;
    if (ARMINSFLAGS(ins)&FL_DIRUP) *instr|=1 << 23;
    if (ARMINSFLAGS(ins)&FL_WRITEBACK) *instr|=1 << 21;
  }
  else
  {
    if (ARMINSFLAGS(ins)&FL_PREINDEX) *instr|=1 << 24;
    if (ARMINSFLAGS(ins)&FL_DIRUP) *instr|=1 << 23;
    if (ARMINSFLAGS(ins)&FL_WRITEBACK) *instr|=1 << 21;
    *instr|=(ARMINSREGB(ins)<<16);
    *instr|=(ARMINSREGA(ins)-ARM_REG_F0)<<12;
    if (ARMINSIMMEDIATE(ins)&0x3) FATAL(("Non aligned floating load/store"));
    *instr|=ARMINSIMMEDIATE(ins)>>2;
    if (ARMINSFLAGS(ins)&FL_FLT_PACKED) *instr|=(1<<22) | (1<<15);
    else if (ARMINSFLAGS(ins)&FL_FLT_DOUBLE_EXTENDED) *instr|=(1<<22);
    else if (ARMINSFLAGS(ins)&FL_FLT_DOUBLE) *instr|=(1<<15);
  }
}
/* }}} */
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
