/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "../../config.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_procstate.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_iterators.h"
#include "../../utils/util_error.h"
#include "../../utils/util_options.h"
#include "../../utils/util_bitmanipulation.h"
#include "../../simple_types.h"
#include "../../diablo_options.h"
#include "arm_emulation.h"
#include "arm_instructions.h"
#include "arm_registers.h"
#include "arm_description.h"
#include "arm_utils.h"
#include "arm_constant_propagation.h"
/* }}} */
/* \todo this should be set in an include file somewhere */
extern t_uint32 global_optimization_phase;
/*!
 * Resets an instruction's S bit when it turns out the instruction doesn't
 * affect the status bits.
 *
 * \param ins The instruction to optimize
 * \param before The register state before executing the instruction
 * \param after The register state after executing the instruction
 *
 * \return t_bool FALSE if the optimization killed an instruction 
*/
/* ArmInsOptKillSBit {{{ */
t_bool ArmInsOptKillSBit(t_ins * ins, t_procstate * before, t_procstate * after)
{
  if (ARMINSFLAGS(ins) & FL_S)

    if (ConditionsAreEqualAndKnown(before,after))
    {
      if (ARMINSOPCODE(ins)==ARM_CMP || ARMINSOPCODE(ins)==ARM_CMN || ARMINSOPCODE(ins)==ARM_TST || ARMINSOPCODE(ins)==ARM_TEQ)
      {
	InsKillDelayed(ins);
	return FALSE;
      }
      else
      {
	ARMINSFLAGS(ins) &= ~FL_S;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	return TRUE;
      }
    }
  return TRUE;
}
/* }}} */
/*!
 * Kills the instruction if it is idempotent.
 *     
 *
 * \warning This should only be called after ArmInsOptKillSBit! 
 *
 * \param ins The instruction to optimize
 * \param before The register state before executing the instruction
 * \param after The register state after executing the instruction
 *
 * \return t_bool FALSE if the optimization killed an instruction 
*/
/* ArmInsOptKillIdempotent {{{ */
t_bool ArmInsOptKillIdempotent(t_ins * ins, t_procstate * before, t_procstate * after)
{
  t_bool doesnothing;
  t_reg r;

  doesnothing = TRUE;

  REGSET_FOREACH_REG(INS_REGS_DEF(ins),r)
  {
    if (!RegIsEqualAndKnown(r,before,after))
    {
      doesnothing = FALSE;
      break;
    }
  }
  /* If the S bit was set, the instruction effectively changes the status bits, otherwise ArmInsOptKillSBit
   * would have reset the S bit. */
  doesnothing = doesnothing && !(ARMINSFLAGS(ins) & FL_S);

  /* only remove the instruction if it doesn't have a side effect (like writing something to memory) */
  if (doesnothing && !InsHasSideEffect(ins))
  {
    InsKillDelayed(ins);
    return FALSE;
  }
  return TRUE;
}
/* }}} */
/*!
 * turns indirect jumps and calls of which the destination is known into direct
 * jumps or calls, and changes the control flow graph accordingly.
 *
 * \param ins The instruction to optimize
 * \param before The register state before executing the instruction
 * \param after The register state after executing the instruction
 *
 * \return t_bool always TRUE (this optimization kills no instructions) 
*/
/* ArmInsOptMakeIndirectJumpsDirect {{{ */
t_bool ArmInsOptMakeIndirectJumpsDirect(t_ins * ins, t_procstate * before, t_procstate * after)
{
  register_content value;

  /* MOV pc,rXX with rXX == constant value */
  if ((ARMINSREGA(ins)==ARM_REG_R15) && (ARMINSOPCODE(ins)==ARM_MOV) && (ARMINSREGC(ins)!=ARM_REG_NONE) && (!ProcStateGetReg(before,ARMINSREGC(ins),&value)))
  {
    t_reloc * rel;
    t_cfg_edge * i_edge, *tmp;
    t_bbl * link = NULL;
    t_uint32 found=0;

    if (!ProcStateGetTag(before,ARMINSREGC(ins),&rel))
    {
      if (REL_TO_RELOCATABLE(rel)->type==RT_BBL)
      {
	BBL_FOREACH_SUCC_EDGE_SAFE(INS_BBL(ins),i_edge,tmp)
	{
	  if ((EDGE_CAT(i_edge)==ET_IPJUMP) || (EDGE_CAT(i_edge)==ET_CALL))
	  {
	    VERIFY_FASSERT(!found,("Multiple edges!"));
	    found = EDGE_CAT(i_edge);
	    if (EDGE_CAT(i_edge)==ET_CALL)
	    {
	      link = CFG_EDGE_TAIL(i_edge->corr_edge);
	    }
	    if (i_edge->corr_edge==tmp) tmp = T_CFG_EDGE(EDGE_NEXT(tmp));
	    if (i_edge->corr_edge->procstate)
	      ProcStateFree(i_edge->corr_edge->procstate);
	    CfgEdgeKill(i_edge->corr_edge);
	    if (i_edge->procstate)
	      ProcStateFree(i_edge->procstate);
	    CfgEdgeKill(i_edge);
	  }
	}

	VERIFY_FASSERT(found,("No edge found"));
	DiabloPrint(stdout,"Change @I to ",ins);

	if (found == ET_IPJUMP)
	{
	  t_cfg_edge * edge;
	  ARMINSOPCODE(ins) = ARM_B;
	  INS_TYPE(ins) = IT_BRANCH;
	  ARMINSREGA(ins) = ARMINSREGB(ins) = ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = value.i;
	  edge = CfgAppendNode(INS_BBL(ins)->cfg,INS_BBL(ins),T_BBL(REL_TO_RELOCATABLE(rel)),(BBL_FUN(INS_BBL(ins))==BBL_FUN(REL_TO_RELOCATABLE(rel)))?ET_JUMP:ET_IPJUMP);
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	  INS_ATTRIB(ins) |= IF_CHANGED;

	  /* if the newly added edge is interprocedural,
	   * allocate a procstate for it (the cleanup
	   * loop at the end of ConstantPropagation expects
	   * an allocated procstate for all interprocedural
	   * edges) */
	  if (EDGE_CAT(edge) == ET_IPJUMP)
	  {
	    edge->procstate = ProcStateNew();
	    ArmEdgePropagator(edge,ins);
	    /* set all registers to BOT, just to make sure 
	     * we can't do anything wrong should the procstate
	     * information on the edges be used afterwards (shouldn't 
	     * happen though...) */
	    ProcStateSetAllBot(edge->procstate,AllRegs());
	  }

	  DiabloPrint(stdout,"@I\n",ins);fflush(stdout);
	}
	else
	{
	  t_cfg_edge * edge;
	  
	  ARMINSOPCODE(ins) = ARM_BL;
	  ARMINSREGA(ins) = ARMINSREGB(ins) = ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = value.i;
	  edge = CfgEdgeNewCall(INS_BBL(ins)->cfg,INS_BBL(ins),T_BBL(REL_TO_RELOCATABLE(rel)),link,FUN_BBL_LAST(BBL_FUN(REL_TO_RELOCATABLE(rel))));
	  INS_ATTRIB(ins) |= IF_CHANGED | IF_PCDEF;
	  INS_TYPE(ins) = IT_BRANCH;
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	  edge->procstate = ProcStateNew();
	  edge->corr_edge->procstate = ProcStateNew();
	  ArmEdgePropagator(edge,ins);
	  ArmEdgePropagator(edge->corr_edge,ins);
	  DiabloPrint(stdout,"@I\n",ins);
	  DiabloPrint(stdout,"Hupsla :@E added!\n",edge);
	}
      }
      else
      {
	FATAL(("Constant move detected... Weird: @I -> @R",ins,rel));
      }
    }
    else
    {
      DiabloPrint(stdout,"InsOpt: found constant move into pc but not tagged! @I: value %d\n",ins,value.i);
    }
  }

  /* always return true, because the instruction is never deleted */
  return TRUE;
}
/* }}} */
/*!
 * Optimizes switches of which the destination is known 
 *
 * \param ins The instruction to optimize
 * \param before The register state before executing the instruction
 * \param after The register state after executing the instruction
 *
 * \todo Verify this for other kinds of switches (gcc-style LDR)
 *
 * \return t_bool FALSE if the optimization killed an instruction 
*/
/* ArmInsOptSwitchOptimizer {{{ */
t_bool ArmInsOptSwitchOptimizer(t_ins * ins, t_procstate * before, t_procstate * after)
{
  t_cfg_edge * i_edge;
  t_bbl * block;
  t_bool flag = FALSE;

  if (ins != BBL_INS_LAST(INS_BBL(ins))) return TRUE;	/* switch instructions are always at the end of a block */

  /* check for outgoing switch edges */
  block = INS_BBL(ins);
  BBL_FOREACH_SUCC_EDGE(block,i_edge)
    if (EDGE_CAT(i_edge) == ET_SWITCH)
      flag = TRUE;

  if (flag)
  {
    if (ARMINSREGA(ins) != ARM_REG_R15)
    {
      /* the switch instruction has been killed earlier on because it was guaranteed never to be executed
       * => this means we should always take the default case (fallthrough edge), and all switch edges
       * may be deleted */
      t_cfg_edge * tmp;

      DiabloPrint(stdout,"Switch at @G always takes fallthrough case\n",INS_CADDRESS(ins));

      /* remove all switch edges */
      BBL_FOREACH_SUCC_EDGE_SAFE(block,i_edge,tmp)
      {
	if (EDGE_CAT(i_edge) == ET_SWITCH)
	  {
	    if (i_edge->procstate)
	      ProcStateFree(i_edge->procstate); 
	    CfgEdgeKill(i_edge);
	  }
      }

      /* for gcc-style switches: remove the data block containing the jump table
       * for ads-style switches: cut the branch instructions for the switch cases
       * loose from the default case. the branch instructions will be cleaned up 
       * later by dead code removal */
      if (BBL_NEXT(block) && BBL_NEXT(BBL_NEXT(block)) && IS_DATABBL(BBL_NEXT(BBL_NEXT(block))))
      {
	/* gcc-style switch table */
	t_bbl * jumptable = BBL_NEXT(BBL_NEXT(block));
	t_ins * ins;
	BBL_FOREACH_INS(jumptable,ins)
	  InsKillDelayed(ins);
	while (RELOCATABLE_REFED_BY(jumptable)) 
	  RelocTableRemoveReloc(block->cfg->sec->obj->reloc_table,RELOCATABLE_REFED_BY(jumptable)->rel);
	BblKill(jumptable);
      }
      else
      {
	/* ads-style switch table */
	t_bbl * ft_block;

	DiabloPrint(stdout,"switch_opt 1: @I\n", ins);

	BBL_FOREACH_SUCC_EDGE(block,i_edge)
	  if (EDGE_CAT(i_edge) == ET_FALLTHROUGH)
	    break;
	VERIFY_FASSERT(i_edge,("Switch statement at @G has no fallthrough block!",INS_CADDRESS(ins)));

	ft_block = CFG_EDGE_TAIL(i_edge);
	BBL_FOREACH_SUCC_EDGE(ft_block,i_edge)
	  if (EDGE_CAT(i_edge) == ET_FALLTHROUGH)
	    break;
	VERIFY_FASSERT(i_edge,("Switch statement at @G has no case blocks!",INS_CADDRESS(ins)));

	if (i_edge->procstate)
	  ProcStateFree(i_edge->procstate);

	CfgEdgeKill(i_edge);

	/* now also remove the EF_FROM_SWITCH_TABLE from the jump edge out of the fallthrough block */
	BBL_FOREACH_SUCC_EDGE(ft_block,i_edge)
	  if (EDGE_CAT(i_edge) == ET_JUMP)
	    break;
	VERIFY_FASSERT(i_edge,("Switch statement at @G has no jump after fallthrough block!",INS_CADDRESS(ins)));

	CFG_EDGE_FLAGS(i_edge) &= (~EF_FROM_SWITCH_TABLE);
      }
    }
    else if (ARMINSOPCODE(ins) == ARM_ADD)
    {
      /* ads-style switch */
      /* if the value of the switch discriminator is known, we can optimize this */
      register_content content;
      t_reg discriminator = ARMINSREGC(ins);

      /* Sanity check */
      VERIFY_FASSERT(ARMINSOPCODE(ins) == ARM_ADD,("Switch statement with really odd pc-definer: @I. Implement!",ins));
      VERIFY_FASSERT(ARMINSREGB(ins) == ARM_REG_R15,("Switch statement with really odd pc-definer: @I. Implement!",ins));
      VERIFY_FASSERT(ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_LSL_IMM,("Switch statement with really odd pc-definer: @I. Implement!",ins));
      VERIFY_FASSERT(ARMINSSHIFTLENGTH(ins) == 2,("Switch statement with really odd pc-definer: @I. Implement!",ins));

      if (!ProcStateGetReg(before,discriminator,&content))
      {
	/* it's a constant, optimize the switch statement */
	t_cfg_edge * tmp;
	t_bbl * switchblock;
	t_address switchaddr;

	DiabloPrint(stdout,"Switch at @G has constant discriminator %d\n",INS_CADDRESS(ins),content.i);

	/* find the taken switch block */
	switchaddr = GENERIC_ADD_UINT32(INS_CADDRESS(ins),8+4*content.i);
	BBL_FOREACH_SUCC_EDGE(block,i_edge)
	  if (EDGE_CAT(i_edge) == ET_SWITCH)
	    if (BBL_CADDRESS(CFG_EDGE_TAIL(i_edge)) == switchaddr)
	      break;
	VERIFY_FASSERT(i_edge,("Switch at @G doesn't have a case %d!",INS_CADDRESS(ins),content.i));

	switchblock = CFG_EDGE_TAIL(i_edge);

	/* cut the fallthrough edge from the switch block
	 * => all underlying blocks are cut loose so they will be removed by dead block removal */
	BBL_FOREACH_SUCC_EDGE_SAFE(switchblock,i_edge,tmp)
	  if (EDGE_CAT(i_edge) == ET_FALLTHROUGH)
	    {
	      if (i_edge->procstate)
		ProcStateFree(i_edge->procstate);
	      CfgEdgeKill(i_edge);
	    }

	/* remove the EF_FROM_SWITCH_TABLE from the jump edge out of the switch block */
	BBL_FOREACH_SUCC_EDGE(switchblock,i_edge)
	  if (EDGE_CAT(i_edge) == ET_JUMP)
	    break;
	VERIFY_FASSERT(i_edge,("Switch case at @G does not have a jump edge!",switchaddr));

	CFG_EDGE_FLAGS(i_edge) &= (~EF_FROM_SWITCH_TABLE);

	/* cut all edges from the first block of the switch statement to the switch cases */
	BBL_FOREACH_SUCC_EDGE_SAFE(block,i_edge,tmp)
	  {
	    if (i_edge->procstate)
	      ProcStateFree(i_edge->procstate);
	    CfgEdgeKill(i_edge);
	  }

	/* Add a new fallthrough edge from the first block of the switch statement to the taken switch case */
	CfgAppendNode(block->cfg,block,switchblock,ET_FALLTHROUGH);

	/* Kill the last instruction of the first block of the switch statement (it is no longer needed) */
	InsKillDelayed(ins);
	return FALSE;
      }
      else if (ARMINSCONDITION(ins) == ARM_CONDITION_AL)
      {
	/* switch discriminator is not constant, but the switch instruction
	 * has been unconditionalized. this means that the value of the 
	 * switch statement was always within bounds, so the fallthrough case
	 * will never be taken. we should remove the fallthrough edge into the
	 * fallthrough case accordingly. */
	t_bbl * ft_case = BBL_NEXT(INS_BBL(ins));
	t_cfg_edge * edge;

	/* we need an extra check: even if the fallthrough case has 
	 * already been removed, we will reach this point again for 
	 * every subsequent invocation of ConstantPropagation.
	 * Therefore, we need to check if the fallthrough case
	 * still exists: this is true if there is no incoming switch edge */
	BBL_FOREACH_PRED_EDGE(ft_case, edge)
	  if (EDGE_CAT(edge) == ET_SWITCH)
	    break;
	if (!edge)
	{
	  DiabloPrint(stdout,"switch_opt 3: @I\n", ins);
	  /* it suffices to kill the fallthrough edge, other optimizations
	   * will clean up the rest of the mess */
	  if (BBL_PRED_FIRST(ft_case)->procstate)
	    ProcStateFree(BBL_PRED_FIRST(ft_case)->procstate);
	  CfgEdgeKill(BBL_PRED_FIRST(ft_case));
	}

      }
    }
    else if (ARMINSOPCODE(ins) == ARM_LDR)
    {
      /* gcc-style switches */
      if (ARMINSCONDITION(ins) == ARM_CONDITION_AL)
      {
	/* switch discriminator is not constant, but the switch instruction
	 * has been unconditionalized. this means that the value of the 
	 * switch statement was always within bounds, so the fallthrough case
	 * will never be taken. we should remove the fallthrough edge into the
	 * fallthrough case accordingly. */
	t_cfg_edge * edge;

	/* we need an extra check: even if the fallthrough case has 
	 * already been removed, we will reach this point again for 
	 * every subsequent invocation of ConstantPropagation.
	 * Therefore, we need to check if the fallthrough case
	 * still exists: this is true if there is no incoming switch edge */
	BBL_FOREACH_SUCC_EDGE(INS_BBL(ins), edge)
	  if (EDGE_CAT(edge) == ET_FALLTHROUGH)
	    break;

	if (edge)
	{
	  DiabloPrint(stdout,"switch_opt 3: @I\n", ins);
	  /* it suffices to kill the fallthrough edge, other optimizations
	   * will clean up the rest of the mess */
	  if (edge->procstate)
	    ProcStateFree(edge->procstate);
	  CfgEdgeKill(edge);
	}
      }
    }
  }
  return TRUE;
}
/* }}} */
/*!
 * Tries to produce the constant result of a function via a MOV/MVN rXX,#const
 *
 * \param ins The instruction to optimize
 * \param before The register state before executing the instruction
 * \param after The register state after executing the instruction
 *
 * 
 * \return t_bool always TRUE (this optimization kills no instructions)
*/
/*ArmInsOptEncodeConstantResult {{{ */
t_bool ArmInsOptEncodeConstantResult(t_ins * ins, t_procstate * before, t_procstate * after)
{
  t_reg regA = ARMINSREGA(ins);
  register_content value;
  t_uint32 valueA = 0;
  t_lattice_level level;
  t_reloc * reloc;

  if ((ArmInsDefinedRegCount(ins) != 1) || ArmInstructionIsNOOP(ins)) /* don't try to encode more than one constant in an instruction */
    return TRUE;

  if (regA == ARM_REG_NONE) /* if the defined register isn't stored in the Reg A slot, we can't do anything in this optimization */
    if (ARMINSOPCODE(ins) != ARM_LDM) /* only exception: ARM_LDM (see further) */
      return TRUE;
  
  if (ARMINSOPCODE(ins)==ARM_LDM)
    {
      REGSET_FOREACH_REG(ARMINSIMMEDIATE(ins),regA)
	break;
    }
  
  level = ProcStateGetReg(after,regA,&value);
  
  if (!level)
    valueA = G_T_UINT32(value.i);
  else
    return TRUE;	/* regA should contain a constant value for this optimization to be possible */
  
  if (ProcStateGetTag(after,regA,&reloc) != TOP)	/* the constant may not be an address, otherwise this 
							   optimization is not correct! */
    return TRUE;
  
  switch (INS_TYPE(ins)) 
  {
    case IT_DATAPROC:
      if (!(ARMINSFLAGS(ins) & FL_S))
      {
	if (ArmIsEncodableConstantForOpcode(valueA,ARM_MOV) && !((ARMINSOPCODE(ins) == ARM_MOV) && (ARMINSFLAGS(ins) & FL_IMMED)))
	{
	  INS_TYPE(ins) = IT_DATAPROC;
	  ARMINSOPCODE(ins) = ARM_MOV;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = valueA;
	  ARMINSREGS(ins) = ARM_REG_NONE;
	  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	  ARMINSSHIFTLENGTH(ins) = 0;
	  ARMINSFLAGS(ins) = 0x0 | FL_IMMED; 
	  /* flag the changes for the assembler */
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	}
	else if (ArmIsEncodableConstantForOpcode(~valueA,ARM_MVN) && !((ARMINSOPCODE(ins) == ARM_MVN) && (ARMINSFLAGS(ins) & FL_IMMED)))
	{
	  INS_TYPE(ins) = IT_DATAPROC;
	  ARMINSOPCODE(ins) = ARM_MVN;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = ~valueA;
	  ARMINSREGS(ins) = ARM_REG_NONE;
	  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	  ARMINSSHIFTLENGTH(ins) = 0;
	  ARMINSFLAGS(ins) = 0x0 | FL_IMMED; 
	  /* flag the changes for the assembler */
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	}
      }
      break;

    case IT_MUL:
    case IT_LOAD:
      if (!(ARMINSFLAGS(ins) & FL_S))
      {
	if (ArmIsEncodableConstantForOpcode(valueA,ARM_MOV))
	{
	  INS_TYPE(ins) = IT_DATAPROC;
	  ARMINSOPCODE(ins) = ARM_MOV;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = valueA;
	  ARMINSREGS(ins) = ARM_REG_NONE;
	  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	  ARMINSSHIFTLENGTH(ins) = 0;
	  ARMINSFLAGS(ins) = 0x0 | FL_IMMED; 
	  /* flag the changes for the assembler */
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	}
	else if (ArmIsEncodableConstantForOpcode(~valueA,ARM_MVN))
	{
	  INS_TYPE(ins) = IT_DATAPROC;
	  ARMINSOPCODE(ins) = ARM_MVN;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = ~valueA;
	  ARMINSREGS(ins) = ARM_REG_NONE;
	  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	  ARMINSSHIFTLENGTH(ins) = 0;
	  ARMINSFLAGS(ins) = 0x0 | FL_IMMED; 
	  /* flag the changes for the assembler */
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	}

      }
      break;

    case IT_STORE_MULTIPLE:
      break;		/* we can't do anything for STMs */
    case IT_LOAD_MULTIPLE:
      /* I DON'T SEE WHY WE SHOULD COMPUTE THE VALUE BEFORE THIS INSTRUCTION 
      level = ProcStateGetReg(before,regA,&value);
      if (!level)
	valueA = G_T_UINT32(value.i);
      */
      if (!level && !(ARMINSFLAGS(ins) & FL_S))
      {
	if (ArmIsEncodableConstantForOpcode(valueA,ARM_MOV))
	{
	  INS_TYPE(ins) = IT_DATAPROC;
	  ARMINSOPCODE(ins) = ARM_MOV;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = valueA;
	  ARMINSREGS(ins) = ARM_REG_NONE;
	  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	  ARMINSSHIFTLENGTH(ins) = 0;
	  ARMINSFLAGS(ins) = 0x0 | FL_IMMED; 
	  /* flag the changes for the assembler */
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	}
	else if (ArmIsEncodableConstantForOpcode(~valueA,ARM_MVN))
	{
	  INS_TYPE(ins) = IT_DATAPROC;
	  ARMINSOPCODE(ins) = ARM_MVN;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = ~valueA;
	  ARMINSREGS(ins) = ARM_REG_NONE;
	  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	  ARMINSSHIFTLENGTH(ins) = 0;
	  ARMINSFLAGS(ins) = 0x0 | FL_IMMED; 
	  /* flag the changes for the assembler */
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	}

      }
      break;


    case IT_BRANCH:
      /* do nothing, branches always have their immediates encoded in them from the start */


    case IT_SWI:
    case IT_STORE:
    case IT_STATUS:
    case IT_SWAP:
    case IT_DATA:
    case IT_CONSTS:
      break;


    default:
      printf("Instruction can't be changed for now. Implement!\n");
  }

  return TRUE;
}
/* }}} */
/*!
 * use MOV or MVN or ADD or SUB to produce a constant value or address from a known register 
 *
 * \param ins The instruction to optimize
 * \param before The register state before executing the instruction
 * \param after The register state after executing the instruction
 *
*/
/* ArmInsOptProduceConstantFromOtherRegister {{{ */
t_bool ArmInsOptProduceConstantFromOtherRegister(t_ins * ins, t_procstate * before, t_procstate * after,int count)
{
#if 1
  register_content prod_value;
  t_reloc * prod_reloc;
  t_lattice_level prod_value_level;
  t_lattice_level prod_reloc_level;
  t_reg r,regdef;
  t_regset allowedregs;
  t_bool missed_opportunity;
  t_ins * i_ins;

  
  if (global_optimization_phase == 0)
    allowedregs = InsRegsLiveBefore(ins);
  else
    allowedregs = arm_description.all_registers;

  /* there must only be one defined register and the S bit must be cleared */
  if ((ARMINSFLAGS(ins) & FL_S) ||  (ArmInsDefinedRegCount(ins) != 1))
    return TRUE;
  /* the instruction mustn't have a side effect */
  if ((INS_TYPE(ins)==IT_STATUS) || (INS_TYPE(ins) == IT_STORE) ||(INS_TYPE(ins) == IT_SWAP) || (ARMINSOPCODE(ins) == ARM_STM))
    return TRUE;
  /* if the instruction already is a MOV or MVN, we can't do much about it */
  if ((ARMINSOPCODE(ins) == ARM_MOV) || (ARMINSOPCODE(ins) == ARM_MVN))
    return TRUE;

  
  /* find the defined register */
  REGSET_FOREACH_REG(INS_REGS_DEF(ins),regdef) break;

  prod_value_level = ProcStateGetReg(after,regdef,&prod_value);
  prod_reloc_level = ProcStateGetTag(after,regdef,&prod_reloc);

  /* the register must contain a value, and the associated relocation mustn't be BOT (TOP is okay) */
  if (prod_value_level || (prod_reloc_level == BOT))
    return TRUE;

  /* to be on the safe side, we only try this optimization with relocations pointing
   * TO_BBL or TO_SUBSECTION */
  if ((!prod_reloc_level) && ((REL_TO_RELOCATABLE(prod_reloc)->type != RT_BBL) && (REL_TO_RELOCATABLE(prod_reloc)->type != RT_SUBSECTION)))
    return TRUE;

  /* try to find a live register that already contains this value and produce it
   * from there */

  if (global_options.rely_on_calling_conventions)
    {
      t_regset meaningfull = RegsetNew();
      RegsetDup(meaningfull,BBL_REGS_DEFINED_IN(INS_BBL(ins)));
      BBL_FOREACH_INS(INS_BBL(ins),i_ins)
	{
	  if (i_ins==ins) break;
	  RegsetSetUnion(meaningfull,INS_REGS_DEF(i_ins));
	}
      
      RegsetSetIntersect(allowedregs,meaningfull);
    }
  missed_opportunity = FALSE;

  REGSET_FOREACH_REG(allowedregs,r)
  {
    register_content other_value;
    t_reloc * other_reloc;
    t_lattice_level other_value_level;
    t_lattice_level other_reloc_level;

    /*    if (r == regdef) continue; */

    other_value_level = ProcStateGetReg(before,r,&other_value);
    other_reloc_level = ProcStateGetTag(before,r,&other_reloc);

    /* the register needs to contain a known value of the same type (const or address) of the produced register */
    if (other_value_level || (other_reloc_level != prod_reloc_level)) continue;

    /* the value in the register must of course be equal to the one in the produced register */
    if (!GENERIC_EQ(prod_value.i,other_value.i)) continue;

    /* last check: if we are looking at addresses, do they point to the same block? */
    if (!prod_reloc_level && (REL_TO_RELOCATABLE(prod_reloc) != REL_TO_RELOCATABLE(other_reloc)))
      {
	missed_opportunity = TRUE;
	continue;
      }

    /* we're certain now: change the instruction to a MOV */

    /*    DiabloPrintArch(stdout,INS_BBL(ins)->fun->cfg->sec->obj->description,"TAKEN %d MOV OPP @I\n @P\n",count,ins,before); */

    ARMINSOPCODE(ins) = ARM_MOV;
    ARMINSREGB(ins) = ARM_REG_NONE;
    ARMINSREGC(ins) = r;
    ARMINSIMMEDIATE(ins) = 0;
    ARMINSFLAGS(ins) = 0;
    INS_ATTRIB(ins) |= IF_CHANGED;
    INS_TYPE(ins) = IT_DATAPROC;
    InsFreeReferedRelocsDelayed(ins);
    ARMINSREGS(ins) = ARM_REG_NONE;
    ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
    INS_REGS_USE(ins) = ArmUsedRegisters(ins);
    INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

    return TRUE;
  }

  if (missed_opportunity)
    {
      /*      DiabloPrintArch(stdout,INS_BBL(ins)->fun->cfg->sec->obj->description,"MISSED MOV OPP @I\n @P\n",ins,before); */
    }

  /* for constants (NOT addresses) we can try another option: find a register that contains the exact opposite
   * of the value we want to use, and produce it using a MVN */
  if (prod_reloc_level == TOP)
  {
    REGSET_FOREACH_REG(allowedregs,r)
    {
      register_content other_value;
      t_reloc * other_reloc;
      t_lattice_level other_value_level;
      t_lattice_level other_reloc_level;

      /*if (r == regdef) continue;*/

      other_value_level = ProcStateGetReg(before,r,&other_value);
      other_reloc_level = ProcStateGetTag(before,r,&other_reloc);

      /* the register needs to contain a known value that is not an address */
      if (other_value_level || (prod_reloc_level != TOP)) continue;

      /* the value in the register must be the exact opposite of the one in the produced register */
      if (!GENERIC_EQ(prod_value.i,GENERIC_NOT(other_value.i))) continue;

      /* we're certain now: change the instruction to a MVN */
      ARMINSOPCODE(ins) = ARM_MVN;
      ARMINSREGB(ins) = ARM_REG_NONE;
      ARMINSREGC(ins) = r;
      ARMINSIMMEDIATE(ins) = 0;
      ARMINSFLAGS(ins) = 0;
      INS_ATTRIB(ins) |= IF_CHANGED;
      INS_TYPE(ins) = IT_DATAPROC;
      InsFreeReferedRelocsDelayed(ins);
      ARMINSREGS(ins) = ARM_REG_NONE;
      ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
      INS_REGS_USE(ins) = ArmUsedRegisters(ins);
      INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

      return TRUE;
    }
  }

  /* try to find a live register that contains a value from which this one can be produced 
   * by an ADD or SUB. Of course, this only make sense if the instruction isn't already of a 
   * similar form. */
  if (!((INS_TYPE(ins) == IT_DATAPROC) && (ARMINSFLAGS(ins) & FL_IMMED)))
  {
    t_reloc * other_reloc;

    missed_opportunity = FALSE;

    REGSET_FOREACH_REG(allowedregs,r)
    {
      register_content other_value;
      t_lattice_level other_value_level;
      t_lattice_level other_reloc_level;
      t_uint32 difference;

      other_value_level = ProcStateGetReg(before,r,&other_value);
      other_reloc_level = ProcStateGetTag(before,r,&other_reloc);

      /* the value of the register must be known and of the same type of the produced register (const or address)
       * for the register to be of any use to us */
      if (other_value_level || (other_reloc_level != prod_reloc_level)) continue;

      /* we can only optimize if both relocations point to the same block */
      if (
	  (prod_reloc_level == TOP) || 
	  (
	   REL_TO_RELOCATABLE(prod_reloc) != REL_TO_RELOCATABLE(other_reloc)
#if 0
	   && !(
		RELOCATABLE_TYPE(REL_TO_RELOCATABLE(prod_reloc))==RT_SUBSECTION && 
		RELOCATABLE_TYPE(REL_TO_RELOCATABLE(other_reloc))==RT_SUBSECTION &&
		SECTION_PARENT_SECTION(T_SECTION(REL_TO_RELOCATABLE(prod_reloc))) == SECTION_PARENT_SECTION(T_SECTION(REL_TO_RELOCATABLE(other_reloc)))
		)
#endif
	   )
	  )
	{
	  /*	  printf("types: %d %d\n",RELOCATABLE_TYPE(REL_TO_RELOCATABLE(prod_reloc)),RELOCATABLE_TYPE(REL_TO_RELOCATABLE(other_reloc)));*/
	  difference = G_T_UINT32(prod_value.i) - G_T_UINT32(other_value.i);
	  if (ArmIsEncodableConstantForOpcode(difference,ARM_ADD))
	    {
	      /*	      DiabloPrint(stdout,"a    1: @R\na    2: @R\n",prod_reloc,other_reloc);*/
	      missed_opportunity = TRUE;
	    }
	  else if (ArmIsEncodableConstantForOpcode(-difference,ARM_SUB))
	    {
	      /*	      DiabloPrint(stdout,"s    1: @R\ns    2: @R\n",prod_reloc,other_reloc);*/
	      missed_opportunity = TRUE;	  
	    }
	  continue;
	}
      

      difference = G_T_UINT32(prod_value.i) - G_T_UINT32(other_value.i);

      if (ArmIsEncodableConstantForOpcode(difference,ARM_ADD))
      {
	/*	DiabloPrintArch(stdout,INS_BBL(ins)->fun->cfg->sec->obj->description,"TAKEN %d ADD OPP @I\n @P\n",count,ins,before); */
	/* encode with an add */
	ARMINSOPCODE(ins) = ARM_ADD;
	ARMINSREGB(ins) = r;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = difference;
	ARMINSFLAGS(ins) = FL_IMMED;
	INS_ATTRIB(ins) |= IF_CHANGED;
	INS_TYPE(ins) = IT_DATAPROC;
	InsFreeReferedRelocsDelayed(ins);
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	missed_opportunity = FALSE;
      }
      else if (ArmIsEncodableConstantForOpcode(-difference,ARM_SUB))
      {
	/*	DiabloPrintArch(stdout,INS_BBL(ins)->fun->cfg->sec->obj->description,"TAKEN %d SUB OPP @I\n @P\n",count,ins,before); */
	/* encode with a sub */
	ARMINSOPCODE(ins) = ARM_SUB;
	ARMINSREGB(ins) = r;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = -difference;
	ARMINSFLAGS(ins) = FL_IMMED;
	INS_ATTRIB(ins) |= IF_CHANGED;
	INS_TYPE(ins) = IT_DATAPROC;
	InsFreeReferedRelocsDelayed(ins);
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	missed_opportunity = FALSE;
      }
      else
      {
	/* no encoding possible, let's continue to the next register */
	missed_opportunity = TRUE;
	continue;
      }

      /* if we get here, the address producer has been changed to an add or sub so we 
       * don't need to continue looking for a good value. we can just return from here.
       */
      return TRUE;
    }

    if (missed_opportunity)
      {
/*	DiabloPrintArch(stdout,INS_BBL(ins)->fun->cfg->sec->obj->description,"MISSED ADD OR SUB OPP @I\n @P\n",ins,before); 
	if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(other_reloc)) == RT_SUBSECTION &&
	    SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(other_reloc)) && SECTION_NAME(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(other_reloc))))
	  DiabloPrint(stdout,"Section: %s\n",SECTION_NAME(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(other_reloc))));
	if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(prod_reloc)) == RT_SUBSECTION &&
	    SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(prod_reloc)) && SECTION_NAME(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(prod_reloc))))
	  DiabloPrint(stdout,"Prod Section: %s\n",SECTION_NAME(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(prod_reloc)))); */
      }
    
  }

#endif
  /* if we get here, no optimization was possible so we just return */
  return TRUE;
}
/* }}} */
/*!
 *  tries to encode operands whose value are known into the immediate operand of the instruction
 *
 * \param ins The instruction to optimize
 * \param before The register state before executing the instruction
 * \param after The register state after executing the instruction
 *
 * \return t_bool always TRUE (this optimization kills no instructions)
*/
/* ArmInsOptEncodeConstantOperands {{{ */
t_bool ArmInsOptEncodeConstantOperands(t_ins * ins, t_procstate * before, t_procstate * after)
{
  t_bool retrying_with_switched_operands = FALSE; 
  t_reg regB = ARMINSREGB(ins);
  t_reg regC = ARMINSREGC(ins);
  t_reg regS = ARMINSREGS(ins);

  register_content value;
  t_uint32 valueB = 0;
  t_lattice_level levelB=BOT;

  t_uint32 valueC = 0;
  t_lattice_level levelC=BOT;

  t_uint32 valueS = 0;
  t_lattice_level levelS=BOT;
  t_reloc * relocB;
  t_reloc * relocC;
  t_reloc * relocS;
  t_lattice_level level_relocB=BOT;
  t_lattice_level level_relocC=BOT;
  t_lattice_level level_relocS=BOT;

  /* don't do anything for NOOPs and instructions that already have an immediate operand */

  if (ArmInstructionIsNOOP(ins)) return TRUE;
  if (regC == ARM_REG_NONE) return TRUE;

retry_with_switched_operands:
  if (regB!=ARM_REG_NONE)
  {
    levelB = ProcStateGetReg(before,regB,&value);
    if (!levelB)
      valueB = G_T_UINT32(value.i);
    level_relocB = ProcStateGetTag(after,regB,&relocB);

  }

  /*if (regC!=ARM_REG_NONE)*/
  {
    levelC = ProcStateGetReg(before,regC,&value);
    if (!levelC)
      valueC = G_T_UINT32(value.i);
    level_relocC = ProcStateGetTag(after,regC,&relocC);
  }

  if (regS!=ARM_REG_NONE)
  {
    levelS = ProcStateGetReg(before,regS,&value);
    level_relocS = ProcStateGetTag(after,regS,&relocS);
  }
  else
  {
    level_relocS = TOP;
  }

  switch (INS_TYPE(ins)) 
  {
    case IT_DATAPROC:
      /* Calculate the second operand */
      if (ARMINSSHIFTTYPE(ins) != ARM_SHIFT_TYPE_NONE)
      {
	levelS = ArmInsExtractShift(ins,before,&value);
	
	if (!levelS)
	  valueS = G_T_UINT32(value.i);
      }
      else
      {
	valueS=valueC;
	levelS=levelC;
	level_relocS=level_relocC;
      }

      /* At this point levelS,valueS,level_relocS hold the second
       * operand used in the data processing instruction (shifts are
       * removed) */

      /* If we know the second operand and the value is encodable and it is not a relocatable value */

      if (!levelS && ArmIsEncodableConstantForOpcode(valueS,ARMINSOPCODE(ins)) && level_relocS==TOP)
      {

	/* If the instruction does not have the S bit or it doesn't
	 * have a shifted flexible .... We can encode it if it does
	 * not have an encode flexible or if it has no FL_S bit*/

	if ( (!(ARMINSFLAGS(ins) & FL_S)) || (!ArmInsHasShiftedFlexible(ins)))
	{
	  /* incorporate the operand in the instruction */
	  ARMINSIMMEDIATE(ins) = valueS;
	  ARMINSFLAGS(ins) |= FL_IMMED;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSREGS(ins) = ARM_REG_NONE;
	  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	  ARMINSSHIFTLENGTH(ins) = 0;
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	}
      }
     break;

    case IT_STORE:
    case IT_LOAD:
      /* Calculate the second operand */
      if (ARMINSSHIFTTYPE(ins) != ARM_SHIFT_TYPE_NONE)
      {
	levelS  = ArmInsExtractShift(ins,before,&value);
	if (!levelS)
	  valueS = G_T_UINT32(value.i);
      }
      else
      {
	valueS=valueC;
	levelS=levelC;
	level_relocS = level_relocC;
      }

      if (!levelS && ArmIsEncodableConstantForOpcode(valueS,ARMINSOPCODE(ins)) && level_relocS==TOP)
      {
	/* incorporate the operand in the instruction */
	/* be influenced by our changing of the second operand */
	ARMINSIMMEDIATE(ins) = valueS;
	ARMINSFLAGS(ins) |= FL_IMMED;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	INS_ATTRIB(ins) |= IF_CHANGED;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	/*		  DiabloPrint(stdout,"9 Changed to: @I\n",ins); */
      } 
      break;

    case IT_MUL:
      if ((!levelC)||(!levelB))
      {
	if (!levelC) 
	{ 
	  if ((ARMINSOPCODE(ins)==ARM_MUL) && (valueC==1))
	  {
/*            DiabloPrint(stdout,">>>>>>>> @I C:%d\n",ins,valueC);*/
	    ArmInstructionMakeMov(ins,ARMINSREGA(ins),ARMINSREGB(ins),0,ARMINSCONDITION(ins));
/*            DiabloPrint(stdout,"IS >>>>>>>> @I C:%d\n",ins,valueC);*/
	  }
	  else if ((ARMINSOPCODE(ins)==ARM_MUL || ARMINSOPCODE(ins)==ARM_UMULL) && (valueC==0))
	  {
/*            DiabloPrint(stdout,">>>>>>>> @I C:%d\n",ins,valueC);*/
	    ArmInstructionMakeMov(ins,ARMINSREGA(ins),ARM_REG_NONE,0,ARMINSCONDITION(ins));
/*            DiabloPrint(stdout,"IS >>>>>>>> @I C:%d\n",ins,valueC);*/
	  }
	  else
	  {
/*            DiabloPrint(stdout,"POT >>>>>>>> @I C:%d\n",ins,valueC);*/
	  }
	}
	else if (!levelB)  
	{
	  if ((ARMINSOPCODE(ins)==ARM_MUL) && (valueB==1))
	  {
/*            DiabloPrint(stdout,">>>>>>>> @I B:%d\n",ins,valueB);*/
	    ArmInstructionMakeMov(ins,ARMINSREGA(ins),ARMINSREGC(ins),0,ARMINSCONDITION(ins));
/*            DiabloPrint(stdout,"IS >>>>>>>> @I B:%d\n",ins,valueB);*/
	  }
	  else if ((ARMINSOPCODE(ins)==ARM_MUL || ARMINSOPCODE(ins)==ARM_UMULL) && (valueB==0))
	  {
/*            DiabloPrint(stdout,">>>>>>>> @I B:%d\n",ins,valueB);*/
	    ArmInstructionMakeMov(ins,ARMINSREGA(ins),ARM_REG_NONE,0,ARMINSCONDITION(ins));
/*            DiabloPrint(stdout,"IS >>>>>>>> @I B:%d\n",ins,valueB);*/
	  }
	  else if ((ARMINSOPCODE(ins)==ARM_MLA) && (valueB==0))
	  {
/*            DiabloPrint(stdout,">>>>>>>> @I B:%d\n",ins,valueB);*/
	    ArmInstructionMakeMov(ins,ARMINSREGA(ins),ARMINSREGS(ins),0,ARMINSCONDITION(ins));
/*            DiabloPrint(stdout,"IS >>>>>>>> @I B:%d\n",ins,valueB);*/
	  }
	  else
	  {
/*            DiabloPrint(stdout,"POT >>>>>>>> @I B:%d\n",ins,valueB);*/
	  }
	}
      }
      break;
    default:
      if ((!levelC)||(!levelB))
      {
	if (!levelC) 
	{
/*          DiabloPrint(stdout,"A >>>>>>>> @I C:%d\n",ins,valueC);*/
	}
	if (!levelB)
	{
/*          DiabloPrint(stdout,"A >>>>>>>> @I B:%d\n",ins,valueB);*/
	}
      }
      /* do nothing because none of the other instructions offer possibilities for optimisation */
      break;
  }

  /* For commutative instructions without immediate operand or shifted operand:
   * 	try to switch regB and regC and retry the previous optimization */
  if (!retrying_with_switched_operands && ArmInsIsCommutative(ins))
  {
    if (!ArmInsHasShiftedFlexible(ins))
    {
      /* regB needs to be constant, and may not contain a relocation */
      if (!levelB && (level_relocB == TOP))
      {
	/* swap regB and regC */
	if (ARMINSREGC(ins)!=ARM_REG_NONE) 
	{
	  t_reg tmp = ARMINSREGC(ins);
	  ARMINSREGC(ins) = ARMINSREGB(ins);
	  ARMINSREGB(ins) = tmp;
	  INS_ATTRIB(ins) |= IF_CHANGED;

	  /*DiabloPrint(stdout,"-~-~- Swapped operands: ins now @I\n",ins); */

	  retrying_with_switched_operands = TRUE;
	  goto retry_with_switched_operands;
	}
      }
    }
  }
  return TRUE;
}
/* }}} */
/*!
 *  tries to encode a constant in two instructions if encoding it in one instructions isn't possible
 *
 * \param ins The instruction to optimize
 * \param before The register state before executing the instruction
 * \param after The register state after executing the instruction
 *
 * \return t_bool always TRUE (this optimization kills no instructions)
*/
/* ArmInsOptEncodeConstantResultInTwoInstructions {{{ */
t_bool ArmInsOptEncodeConstantResultInTwoInstructions(t_ins * ins, t_procstate * before, t_procstate * after)
{
  t_reg regdef;
  register_content content;
  t_uint32 value = 0;
  t_lattice_level level;
  t_reloc * reloc;

  /* don't try to do this in the first optimization phase - it potentially increases program size */
  if (global_optimization_phase == 0)
    return TRUE;

  if ((ArmInsDefinedRegCount(ins) != 1) || ArmInstructionIsNOOP(ins)) /* don't try to encode more than one constant in an instruction */
    return TRUE;
  if (ARMINSFLAGS(ins) & FL_S) /* don't do this for instructions that set the S bit: potentially dangerous */
    return TRUE;

  /* find the defined register */
  REGSET_FOREACH_REG(INS_REGS_DEF(ins),regdef) break;

  /* the instruction mustn't have a side effect */
  if ((INS_TYPE(ins)==IT_STATUS) || (INS_TYPE(ins) == IT_STORE) ||(INS_TYPE(ins) == IT_SWAP) || (ARMINSOPCODE(ins) == ARM_STM))
    return TRUE;

  level = ProcStateGetReg(after,regdef,&content);
  if (!level)
    value = G_T_UINT32(content.i);
  else
    return TRUE;	/* regdef should contain a constant value for this optimization to be possible */
  
  if (ProcStateGetTag(after,regdef,&reloc) != TOP)	/* the constant may not be an address, otherwise this 
							   optimization is not correct! */
    return TRUE;

  /* don't do this for instructions of the form MOV or MVN or DATAPROC rX,rY,#constant */
  if ((ARMINSOPCODE(ins) == ARM_MOV) || (ARMINSOPCODE(ins) == ARM_MVN))
    return TRUE;
  if ((INS_TYPE(ins) == IT_DATAPROC) && (ARMINSFLAGS(ins) & FL_IMMED))
    return TRUE;

  /* try to do it... */
  if (ArmConstantIsProducableInTwoInstructions(value))
  {
    DiabloPrint(stdout,"--- Encoding %x in 2 instructions (orig: @I)\n",value,ins);
    ArmEncodeConstantInTwoInstructions(ins, value);
  }
  return TRUE;
}
/* }}} */


t_bool ArmOptMoveNullIntoPC(t_ins * ins)
{
  if (ARMINSOPCODE(ins) == ARM_MOV && ARMINSREGA(ins)==ARM_REG_R15 && ARMINSREGB(ins)==ARM_REG_NONE &&  ARMINSREGC(ins)==ARM_REG_NONE && ARMINSIMMEDIATE(ins)==0)
    {
      t_bbl * bbl = INS_BBL(ins);
      t_cfg_edge * edge, *safe;

      BBL_FOREACH_SUCC_EDGE_SAFE(bbl,edge,safe)
	{
	  if (edge->corr_edge)
	    {
	      if (edge->corr_edge->procstate)
		ProcStateFree(edge->corr_edge->procstate);
	      CfgEdgeKill(edge->corr_edge);
	    }
	  
	  if (edge->procstate)
	    ProcStateFree(edge->procstate);
	  CfgEdgeKill(edge);
	}

      CfgEdgeNew(bbl->fun->cfg,bbl,bbl,ET_JUMP);
    }


  return TRUE;
}

t_bool ArmInsOptChangeLoadBaseRegister(t_ins * ins, t_procstate * before, t_procstate * after)
{
  t_reg regB = ARMINSREGB(ins);
  t_reg regC = ARMINSREGC(ins);
  t_reg regI;
  register_content contentB;
  register_content contentI;
  t_reloc* relocI;
  t_reloc* relocB;
  t_int32 addressB, addressI;
  t_int32 imm = ARMINSIMMEDIATE(ins);
  t_int32 new_imm;
  
  return FALSE;

  if (regC!=ARM_REG_NONE) return FALSE;
  if (INS_TYPE(ins)!=IT_LOAD && INS_TYPE(ins)!=IT_STORE && ARMINSOPCODE(ins)!=ARM_ADD && ARMINSOPCODE(ins)!=ARM_SUB) return FALSE;
  if (ARMINSSHIFTTYPE(ins)!=ARM_SHIFT_TYPE_NONE) return FALSE;
  if (ProcStateGetReg(before,regB,&contentB)) return FALSE;
  if (ProcStateGetTag(before,regB,&relocB)) return FALSE;
  /*  if (RegsetIn(InsRegsLiveAfter(ins),regB)) return FALSE;*/
  
  addressB = GENERIC_EXTRACT_UINT32(contentB.i);
  
  REGSET_FOREACH_REG(AllRegs(),regI)
    {
      if (regI==regB) continue;
      if (!global_optimization_phase)
	if (!RegsetIn(InsRegsLiveOrUsedOrDefinedBefore(ins),regI)) continue;
      if (ProcStateGetReg(before,regI,&contentI)) continue;
      if (ProcStateGetTag(before,regI,&relocI)) continue;
      if (REL_TO_RELOCATABLE(relocB) != REL_TO_RELOCATABLE(relocI)) continue;
     
      addressI = GENERIC_EXTRACT_UINT32(contentI.i);
      
      
      DiabloPrint(stdout,"INS LOADS FROM CONSTANT ADDRESS: @I\n",ins);fflush(stdout);
      DiabloPrintArch(stdout,BBL_FUN(INS_BBL(ins))->cfg->sec->obj->description,"ADDRESS: @G\n",contentB.i);fflush(stdout);
      DiabloPrintArch(stdout,BBL_FUN(INS_BBL(ins))->cfg->sec->obj->description,"ALSO AVAILABLE: ADDRESS: @G in reg %d\n@iB\n",contentI.i,regI,INS_BBL(ins));fflush(stdout);
      
      ARMINSREGB(ins)=regI;

      if (ARMINSOPCODE(ins)!=ARM_SUB && !ArmIsEncodableConstantForOpcode(addressB-addressI+imm,ARM_ADD))
	{
	  printf("NOPE\n");
	  continue;
	}

      if (ARMINSOPCODE(ins)==ARM_SUB && !ArmIsEncodableConstantForOpcode(-addressB+addressI+imm,ARM_SUB))
	{
	  printf("NOPE\n");
	  continue;
	}

      if (ARMINSOPCODE(ins)==ARM_SUB)
      	new_imm=imm+addressI-addressB;
      else
	new_imm=imm+addressB-addressI;

      ARMINSIMMEDIATE(ins)=new_imm;

      INS_REGS_USE(ins) = ArmUsedRegisters(ins);

      DiabloPrint(stdout,"AFTER:\n@iB\n",INS_BBL(ins));
      return FALSE;

    }
  
  return FALSE;
}

#ifdef DIETER

#define OPT_NONE 0
#define OPT_BASIC (OPT_NONE + 1)
#define OPT_ELIMINATE_PRODUCER (OPT_BASIC + 1)

t_bool ArmOptMemoryAccess(t_ins * ins, t_procstate * before, t_procstate * after)
{
  register_content reg_use_value, other_value;
  t_reloc * reg_use_reloc, * other_reloc;
  t_reg reg_use, other_reg, new_reg;
  t_int32 difference, new_immed;
  t_uint32 new_hotness;
  t_uint32 optimization = OPT_NONE;

  t_uint32 orig_address, new_address;  // Debug
  t_bool flag = FALSE;                 // Debug

  /* Optimize only LDR, LDRB, STR or STRB instructions */
  if(ARMINSOPCODE(ins) == ARM_LDR || ARMINSOPCODE(ins) == ARM_LDRB ||
     ARMINSOPCODE(ins) == ARM_STR || ARMINSOPCODE(ins) == ARM_STRB)
    {
      if (ARMINSFLAGS(ins) & FL_IMMED)  // Ins has immediate offset?
	{
	  reg_use = ARMINSREGB(ins);

	  if(VALUE == ProcStateGetTag(before, reg_use, &reg_use_reloc))
	    {
	      new_hotness = reg_use_reloc->hotness;
	      new_reg = reg_use;
	      ProcStateGetReg(before, reg_use, &reg_use_value);

	      orig_address = G_T_UINT32(reg_use_value.i) +                                        // Debug
		((ARMINSFLAGS(ins) & FL_DIRUP) ? ARMINSIMMEDIATE(ins) : (-ARMINSIMMEDIATE(ins))); // Debug

	      t_regset dead_after = InsRegsLiveAfter(ins);
	      RegsetSetInvers(dead_after);
	      REGSET_FOREACH_REG(arm_description.all_registers, other_reg)
		{
		  if(other_reg != reg_use &&
		     VALUE == ProcStateGetTag(before, other_reg, &other_reloc) &&
		     REL_TO_RELOCATABLE(reg_use_reloc) == REL_TO_RELOCATABLE(other_reloc))
		    {
		      ProcStateGetReg(before, other_reg, &other_value);

		      difference = G_T_UINT32(reg_use_value.i) - G_T_UINT32(other_value.i) +
			((ARMINSFLAGS(ins) & FL_DIRUP) ? ARMINSIMMEDIATE(ins) : (-ARMINSIMMEDIATE(ins)));
		      
		      if(-4095 < difference && difference < 4095) // Difference encodable in 12bits ?
			{
#if 1 // Debug
			  if(!flag)
			    {
			      printf("ADDRESS=0x%X REG=%d VALUE=0x%X HOTNESS=0x%X\n", 
				     INS_CADDRESS(ins), reg_use, reg_use_value, new_hotness);
			      DiabloPrint(stdout, "  @I\n", ins);
			      flag = TRUE;
			    }

			  printf("    REG=%d VALUE=0x%X DIFF=0x%X HOTNESS=0x%X\n", 
				 other_reg, other_value, difference, other_reloc->hotness);
#endif

			  /* ANALYSE OPTIMIZATIONS */

#if 0
			  /* 1. If the producers is in the same BBL as the load/store
			     and the used reg is dead after this ins, we try to eliminate
			     the producer completely. */
			  if(optimization <= OPT_ELIMINATE_PRODUCER &&
			     INS_BBL(T_INS(REL_FROM(reg_use_reloc))) == INS_BBL(ins) &&
			     (RegsetIn(dead_after, ARMINSREGB(ins)) || ARMINSREGA(ins) == ARMINSREGB(ins))
			     )
			    {
			      t_ins * producer = T_INS(REL_FROM(reg_use_reloc));
			      t_ins * temp = INS_NEXT(producer);
			      t_uint32 used_adaptable = 0;
			      t_uint32 used_not_adaptable = 0;

			      while(temp != ins)
				{
				  if(RegsetIn(INS_REGS_USE(temp), ARMINSREGA(producer)))
				    {
				      if(0)
					{
					  used_adaptable++;
					}
				      else
					{
					  used_not_adaptable++;
					}
				    }

				  temp = INS_NEXT(temp);
				}
			      
			      if(!used_not_adaptable &&
				 (other_reloc->hotness > new_hotness ||
				  (other_reloc->hotness == new_hotness && GENERIC_LT(INS_CADDRESS(T_INS(REL_FROM(other_reloc))), INS_CADDRESS(producer)))
				  )
				 )
				{
				  new_hotness = other_reloc->hotness;
				  new_reg = other_reg;
				  new_immed = difference;
				  optimization = OPT_ELIMINATE_PRODUCER;
				}
			    }
#endif

			  /* 2. Basic optimization */
			  if(optimization <= OPT_BASIC &&
			     other_reloc->hotness > new_hotness)
			    {
			      new_hotness = other_reloc->hotness;
			      new_reg = other_reg;
			      new_immed = difference;
			      optimization = OPT_BASIC;
			    }
			}
		    }
		}
	    }

	  if(optimization)
	    {
	      /* Updating regs */
	      ARMINSREGB(ins) = new_reg;
	      ARMINSREGC(ins) = ARM_REG_NONE;
	      ARMINSREGS(ins) = ARM_REG_NONE;
	      INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	      INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	      
	      /* Updating flags */
	      ARMINSFLAGS(ins) |= FL_IMMED;
	      ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	      INS_ATTRIB(ins) |= IF_CHANGED;

	      /* Updating reloc hotness. */
	      ProcStateGetTag(before, new_reg, &other_reloc);
	      reg_use_reloc->hotness += 1 - (1 << 16);
	      other_reloc->hotness += (1 << 16) - 1;

	      if(optimization == OPT_ELIMINATE_PRODUCER)
		{
		  printf("    -- OPT_ELIMINATE_PRODUCER --\n");
		  DiabloPrint(stdout, "       @I\n", T_INS(REL_FROM(reg_use_reloc)));
		  if(!(reg_use_reloc->hotness & 0xFFFF00))
		    {
		      InsKill(T_INS(REL_FROM(reg_use_reloc)));
		      printf("    -- Producer killed --\n");
		    }
		  else
		    printf("    -- Producer NOT killed --\n");
		}
	      else if(optimization == OPT_BASIC)
		{
		  printf("    -- OPT_BASIC --\n");
		}
	      
	      /* Updating immediate */
	      ProcStateGetReg(before, new_reg, &other_value); // DEBUG
	      if(new_immed >= 0)
		{
		  ARMINSIMMEDIATE(ins) = new_immed;
		  ARMINSFLAGS(ins) |= FL_DIRUP;
		  printf("  + REG=%d VALUE=0x%X DIFF=0x%X HOTNESS=0x%X\n", new_reg, other_value, new_immed, other_reloc->hotness);
		}
	      else
		{
		  ARMINSIMMEDIATE(ins) = -new_immed;
		  ARMINSFLAGS(ins) &= ~FL_DIRUP;
		  printf("  - REG=%d VALUE=0x%X DIFF=0x%X HOTNESS=0x%X\n", new_reg, other_value, new_immed, other_reloc->hotness);
		}

#if 1 // Debug
	      DiabloPrint(stdout, "  @I\n", ins);
	      /* Checking... */
	      new_address = G_T_UINT32(other_value.i) +
		((ARMINSFLAGS(ins) & FL_DIRUP) ? ARMINSIMMEDIATE(ins) : (-ARMINSIMMEDIATE(ins)));
	      if(orig_address == new_address)
		{
		  printf("    >>>> OK\n");
		}
	      else
		{
		  printf("         0x%X != 0x%X\n", orig_address, new_address);
		  FATAL(("    >>>> OTHER ADDRESS PRODUCED!!!\n"));
		}
#endif

#if 1 // Debug
	      DiabloPrint(stdout, "        @B\n", INS_BBL(ins));
	      t_ins * temp_ins = BBL_INS_FIRST(INS_BBL(ins));
	      while(temp_ins)
		{
		  if(temp_ins == ins)
		    DiabloPrint(stdout, "        ->@I\n", temp_ins);
		  else
		    DiabloPrint(stdout, "          @I\n", temp_ins);
		  temp_ins = INS_NEXT(temp_ins);
		}
#endif
	      
	      return TRUE;
	    }
   	}
      else // (Scaled) register offset
	{
	  
	}
    }
  else // Other then load/store
    {

    }

  return TRUE;
}
#endif


/* vim: set shiftwidth=2 foldmethod=marker : */
