/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_description.h"
#include "arm_instructions.h" 
#include "arm_registers.h" 
#include "arm_copy_propagation.h" 
#include "arm_ls_fwd.h" 
#include "arm_inline.h" 
#include "arm_peephole.h" 
#include "../../kernel/diablo_opt_misc.h"

#ifdef INSTRUMENTATION
#include "../../atom_internal.h"
#include "arm_atom.h"
#endif
/* }}} */
/*! \todo Document, keep only one definition of register names */
/* names {{{ */
static char * names[] =    {"R0",  "R1",  "R2",  "R3",  "R4",  "R5",  "R6",  "R7",  "R8",  "R9",  "R10",  "R11",  "R12",  
			    "SP",  "RA",  "PC",  "CPSR",  "SPSR",  "cQ",  "cC",  "cV",  "cZ",  "cN", "f0", "f1", "f2", "f3", "f4", "f5", "f6", "f7"};
/* }}} */

/* An array of function names of which we now they behave different from calling conventions */
static char *unbehaved_funs[] = {"_dcmple", "_dcmpge", "_fcmple", "_fcmpge", ""};

/*!\addtogroup ARM_BACKEND */
/*@{ */
/*! 
 * This structure holds all Arm specific optimizations. They can be called
 * using ArchCall("<name of optimization">); 
 */
const t_arch_specific_optimization arm_opts[]={
  /*! BranchElimination for the ARM */
  { "BranchElimination", BranchElimination},
  /*! Branchforwarding for the ARM */
  { "BranchForwarding", BranchForwarding},
  /*! Optimizes multiple loads and stores. 
   *
   * \todo This is Arm specific. Rename, move to an arm file 
   */
  { "OptimizeStackLoadAndStores", OptimizeStackLoadAndStores},
  /*! Inlining for the Arm */
  { "GeneralInlining", ArmGeneralInlining},
  /*! Inlining for the Arm (one call site/small functions) */
  { "InlineTrivial", ArmInlineTrivial},
  /*! Copy propagation for the Arm */
  { "OptCopyPropagation", ArmOptCopyPropagation},
  /*! Load store forwarding. Removes unnecessary loads/stores */
  { "ArmLoadStoreFwd", ArmLoadStoreFwd},
  /*! Merges bbls that are connected by a fallthrough edge 
   *
   * \todo if this is arm specific move it, rename it
   * */
  { "MergeBbls", MergeBbls}, 
  { "ArmPeepholeOptimizations", ArmPeepholeOptimizations},
  { NULL,NULL}
} ;
/*@}*/

#ifdef INSTRUMENTATION
static t_atom_actions arm_atom_actions = {
  NULL,
  NULL,
  ArmAtomAddSave,
  ArmAtomAddRestore,
  ArmAtomAddParams,
  NULL,
  NULL,
  ArmAtomAddCall,
  ArmIsInstType
};
#endif

/*!\addtogroup ARM_BACKEND */
/*@{ */
const t_architecture_description arm_description =
{
  /*! Size of an instruction: encoded minimal size */
#ifndef THUMB_SUPPORT
  32, 
#else
  16,
#endif
  /*! Size of an instruction: encoded maximum size */
  32, 
  /*! Size of an instruction: \todo document */
#ifndef THUMB_SUPPORT
  32, /* The mod size */
#else
  16, /* The mod size */
#endif
  /*! Size of an instruction: the number of instructions in a bundle */
  1, /* bundel size */
  /*! Size of an instruction: size of the bundle template */ 
  0,
  /*! Size of an instruction: disassembled instruction size */
  sizeof(t_arm_ins), 
  /*! \todo document */
  ADDRSIZE32,
  /*! The number of general purpose integer registers */
  16,
  /*! The number of general purpose floating point registers */
  8, 
  /*! The number of predicate registers */
  5,
  /*! The number of branch registers */
  0,  
  /*! The number of special registers */
  2, 
  /*! An array containing the name of each register */
   names,
#if MAX_REG_ITERATOR > 64

#ifndef BIT64ADDRSUPPORT
   {MAX_REG_ITERATOR, {0xffff,0x7fff,0}},
   {MAX_REG_ITERATOR, {0xffff,0x000f,0}},
   {MAX_REG_ITERATOR, {0x0000,0x7f80,0}},
   {MAX_REG_ITERATOR, {0x0000,0x007c,0}},
   {MAX_REG_ITERATOR, {0x2ff0,0x7800,0}}, /* Callee saved */  
   {MAX_REG_ITERATOR, {0xf00f,0x0780,0}}, /* callee may use */
   {MAX_REG_ITERATOR, {0xd00f,0x07ff,0}}, /* callee may change */
   {MAX_REG_ITERATOR, {0x000f,0x0780,0}},/* may contain return value */
   {MAX_REG_ITERATOR, {0x8000,0x0000,0}},/* always live */
   {MAX_REG_ITERATOR, {0}},
   {MAX_REG_ITERATOR, {0}},
   {MAX_REG_ITERATOR, {0x0,0x007f,0}},   /* Dead over call */
#else
   {MAX_REG_ITERATOR, {0x7fffffffULL,0ULL}},
   {MAX_REG_ITERATOR, {0x000fffffULL,0ULL}},
   {MAX_REG_ITERATOR, {0x7f800000ULL,0ULL}},
   {MAX_REG_ITERATOR, {0x007c0000ULL,0ULL}},
   {MAX_REG_ITERATOR, {0x78002ff0ULL,0ULL}}, /* Callee saved */  
   {MAX_REG_ITERATOR, {0x0780f00fULL,0ULL}}, /* callee may use */
   {MAX_REG_ITERATOR, {0x07ffd00fULL,0ULL}}, /* callee may change */
   {MAX_REG_ITERATOR, {0x0780000fULL,0ULL}},/* may contain return value */
   {MAX_REG_ITERATOR, {0x00008000ULL,0ULL}},/* always live */
   {MAX_REG_ITERATOR, {0ULL}},
   {MAX_REG_ITERATOR, {0ULL}},
   {MAX_REG_ITERATOR, {0x007f0000ULL,0ULL}},/*Dead over call */
#endif

#elif MAX_REG_ITERATOR > 32
   0x7fffffffLL,
   0x0000ffffLL,
   0x7f800000LL,
   0x007c0000LL,
   0x78002ff0LL, /* Callee saved */  
   0x0780f00fLL, /* callee may use */
   0x07ffd00fLL, /* callee may change */
   0x0780000fLL, /* may contain return value */
   0x00008000LL, /* always live (i.e. program counter) */
          0x0LL,
          0x0LL,
     0x7f0000LL,/* Dead over call */
#else
   0x7fffffff,
   0x0000ffff,
   0x7f800000,
   0x007c0000,
   0x78002ff0, /* Callee saved */  
   0x0780f00f, /* callee may use */
   0x07ffd00f, /* callee may change */
   0x0780000f, /* may contain return value */
   0x00008000, /* always live (i.e. program counter) */
   0x0,
   0x0,
   0x007f0000,/* Dead over call */
#endif
   /*! The program counter */
   ARM_REG_R15,
   /*! The arm specific optimizations */
   arm_opts,
#ifdef INSTRUMENTATION
   &arm_atom_actions,
#endif
  unbehaved_funs,
  ARMaddJumpInstruction,
  ArmIsCallInstruction,
#if MAX_REG_ITERATOR > 64
#ifndef BIT64ADDRSUPPORT
   {MAX_REG_ITERATOR, {0x4000,0x0000,0}},
#else
   {MAX_REG_ITERATOR, {0x00004000ULL,0ULL}},
#endif
#elif MAX_REG_ITERATOR > 32
   0x00004000LL,
#else
   0x00004000,
#endif
  ArmcomputeLiveRegsBeforeSwi
};
/*@}*/
/* \todo Clean this up and document */
#if MAX_REG_ITERATOR > 64
#ifndef BIT64ADDRSUPPORT
t_regset ARM_ALL_BUT_PC_AND_COND = {MAX_REG_ITERATOR,{0x7fff,0}};
#else
t_regset ARM_ALL_BUT_PC_AND_COND = {MAX_REG_ITERATOR,{0x7fffULL,0ULL}};
#endif
#endif
/* vim: set shiftwidth=2 foldmethod=marker : */
