/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "../../utils/util_error.h"
#include "../../utils/util_bitmanipulation.h"
#include "arm_disassemble_one.h"
#include "arm_instructions.h"
#include "arm_registers.h"
#include "arm_utils.h"
/*}}}*/
/* Condition Codes */
#define ARM_EXTRACT_CONDITION_CODE(i) (((i) & 0xf0000000) >> 28)
/*!
 * Utility function: fills in the fields that describe a shift in t_arm_ins
 *
 * \param ins The arm instruction to fill
 * \param instr The encode arm instruction
 *
 * \return void 
*/
/* Utility Function: ArmShiftDecode{{{ */
static void ArmShiftDecode(t_ins * ins, t_uint32 instr) {

  if (instr & (1 << 4)) {
    /* shiftlengte zit in een register */
    ARMINSREGS(ins) = (instr & 0x00000f00) >> 8;
    ARMINSSHIFTLENGTH(ins) = 0;
    ARMINSSHIFTTYPE(ins) = ((instr & 0x00000060) >> 5)+4;
  } else {
    /* shiftlengte is een immediate */
    ARMINSREGS(ins) = ARM_REG_NONE;
    ARMINSSHIFTTYPE(ins) = (instr & 0x00000060) >> 5;
    ARMINSSHIFTLENGTH(ins) = (instr & 0x00000f80) >> 7;
    if ((ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_ROR_IMM) && (ARMINSSHIFTLENGTH(ins) == 0))
      ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_RRX; /* speciaal geval */
    if ((ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_LSL_IMM) && (ARMINSSHIFTLENGTH(ins) == 0))
      ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE; /* speciaal geval */
  }
}
/*}}}*/
/*!
 * Disassemble Arm Branches (B, BL, BX) 
 *
 * \param ins The arm instruction that gets filled in
 * \param instr The encoded arm instruction
 * \param opc The opcode
 *
 * \return void 
*/
/* ArmDisassembleBranch {{{ */
void ArmDisassembleBranch(t_ins * ins,  t_uint32 instr, t_uint16 opc)
{
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSFLAGS(ins) = 0x0;
  ARMINSREGA(ins) = ARMINSREGC(ins) = ARM_REG_NONE;
  ARMINSOPCODE(ins) = opc;
  INS_TYPE(ins) = IT_BRANCH;

  /* next, prev en Bblock moeten nog ingevuld worden */
  switch (opc) {
  case ARM_B:
  case ARM_BL:
    ARMINSREGB(ins) = ARM_REG_NONE;
    ARMINSIMMEDIATE(ins) = 4 * BitSignExtend(instr & 0x00ffffff,23);
    ARMINSFLAGS(ins) |= FL_IMMED;
    break;
  case ARM_BX:
    ARMINSIMMEDIATE(ins) = 0;
    ARMINSREGB(ins) = instr & 0x0000000f;
    break;
  case ARM_BLX:
    ARMINSREGB(ins) = instr & 0x0000000f;
    ARMINSIMMEDIATE(ins) = 0;
    break;
  default:
    FATAL(("Unknown branch instruction\n"));
  }
  /* patch because a special kind of BLX was wrongly disassembled */
  if (ARMINSCONDITION(ins) == ARM_CONDITION_NV && (opc == ARM_B || opc == ARM_BL))
  {
    ARMINSOPCODE(ins) = opc = ARM_BLX;
    ARMINSCONDITION(ins) = ARM_CONDITION_AL;
    INS_ATTRIB(ins) &= ~IF_CONDITIONAL;
    ARMINSIMMEDIATE(ins) += (instr & 0x01000000)?2:0;
  }
}
/*}}}*/
/*!
 * Disassemble Arm Software Interrupts (SWI)
 *
 * \param ins The arm instruction that gets filled in
 * \param instr The encoded arm instruction
 * \param opc The opcode
 *
 * \return void 
 * ArmDisassembleSWI {{{ */
void ArmDisassembleSWI(t_ins * ins, t_uint32 instr, t_uint16 opc )
{
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSFLAGS(ins) = FL_IMMED;
  ARMINSREGA(ins) = ARMINSREGB(ins) = ARMINSREGC(ins) = ARM_REG_NONE;
  ARMINSOPCODE(ins) = opc;
  INS_TYPE(ins) = IT_SWI;

  ARMINSIMMEDIATE(ins) = instr & 0x00ffffff;
}
/*}}}*/
/*!
 * Disassemble Arm Multiplies 
 *
 * \param ins The arm instruction that gets filled in
 * \param instr The encoded arm instruction
 * \param opc The opcode
 *
 * \return void 
*/
/* ArmDisassembleMultiplication {{{ */
void ArmDisassembleMultiplication(t_ins * ins, t_uint32 instr, t_uint16 opc )
{
  /* Rd/RdHi = RegA */
  /* Rn/RdLo = RegS */
  /* Rs      = RegC */
  /* Rm      = RegB */

  ARMINSOPCODE(ins) = opc;
  INS_TYPE(ins) = IT_MUL;
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = 0;

  ARMINSFLAGS(ins) = 0x0;

  if (instr & (1 << 20)) ARMINSFLAGS(ins) |= FL_S;

  /* speciaal: gebruik bij de MUL-varianten RegS als vierde register */
  /* dus als het lage deel van het resultaat bij 64bits vermenigvuldiging, */
  /* of als Rn bij MLA */
  ARMINSREGA(ins) = (instr & 0x000f0000) >> 16;
  ARMINSREGB(ins) = (instr & 0x0000000f);
  ARMINSREGC(ins) = (instr & 0x00000f00) >> 8;
  ARMINSREGS(ins) = (instr & 0x0000f000) >> 12;
}
/*}}}*/
/*!
 * Disassemble Arm Multiple Loads/Stores (LDM, STM)
 *
 * \param ins The arm instruction that gets filled in
 * \param instr The encoded arm instruction
 * \param opc The opcode
 *
 * \return void 
*/
/* ArmDisassembleMultipleTransfer {{{ */
void ArmDisassembleMultipleTransfer(t_ins * ins, t_uint32 instr, t_uint16 opc) 
{
  /* speciaal: gebruik de onderste helft van Immediate hier als registerlist */
  /* gebruik RegB als basisregister (consistent met single register data transfer) */

  ARMINSOPCODE(ins) = opc;
  switch(opc)
  {
     case ARM_LDM:
	INS_TYPE(ins) = IT_LOAD_MULTIPLE;
	break;
     case ARM_STM:
	INS_TYPE(ins) = IT_STORE_MULTIPLE;
	break;
     default:
	FATAL(("IMPLEMENT"));
  }
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSREGS(ins) = ARM_REG_NONE;
 
  ARMINSIMMEDIATE(ins) = 0 | (instr & 0x0000ffff);

  ARMINSFLAGS(ins) = 0x0;

  if (instr & (1 << 24)) ARMINSFLAGS(ins) |= FL_PREINDEX;
  if (instr & (1 << 23)) ARMINSFLAGS(ins) |= FL_DIRUP;
  if (instr & (1 << 22)) ARMINSFLAGS(ins) |= FL_S;
  if (instr & (1 << 21)) ARMINSFLAGS(ins) |= FL_WRITEBACK;

  ARMINSREGA(ins) = ARMINSREGC(ins) = ARM_REG_NONE;
  ARMINSREGB(ins) = (instr & 0x000f0000) >> 16;
}
/*}}}*/
/*!
 * Disassemble Arm Swap instructions (SWP)
 *
 * \param ins The arm instruction that gets filled in
 * \param instr The encoded arm instruction
 * \param opc The opcode
 *
 * \return void 
*/
/* ArmDisassembleSwap{{{*/
void ArmDisassembleSwap(t_ins * ins, t_uint32 instr, t_uint16 opc) 
{
  ARMINSOPCODE(ins) = opc;
  INS_TYPE(ins) = IT_SWAP;
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSREGS(ins) = ARM_REG_NONE;
 
  ARMINSIMMEDIATE(ins) = 0;

  ARMINSFLAGS(ins) = 0x0;

  ARMINSREGA(ins) = (instr & 0x0000f000) >> 12;
  ARMINSREGB(ins) = (instr & 0x000f0000) >> 16;
  ARMINSREGC(ins) = (instr & 0x0000000f);
}
/*}}}*/
/*!
 * \todo Document 
 *
 * \param ins The arm instruction that gets filled in
 * \param instr The encoded arm instruction
 * \param opc The opcode
 *
 * \return void 
*/
/* ArmDisassembleMRS {{{ */
void ArmDisassembleMRS(t_ins * ins, t_uint32 instr, t_uint16 opc) 
{
  ARMINSOPCODE(ins) = opc;
  INS_TYPE(ins) = IT_STATUS;
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSREGS(ins) = ARM_REG_NONE;
 
  ARMINSFLAGS(ins) = 0x0;

  if (instr & (1 << 22)) ARMINSFLAGS(ins) |= FL_SPSR;

  ARMINSIMMEDIATE(ins) = 0;
  ARMINSREGA(ins) = (instr & 0x0000f000) >> 12;
  ARMINSREGB(ins) = ARMINSREGC(ins) = ARM_REG_NONE;
}
/* }}} */
/*!
 * \todo Document 
 *
 * \param ins The arm instruction that gets filled in
 * \param instr The encoded arm instruction
 * \param opc  The opcode
 *
 * \return void 
*/
/* ArmDisassembleMSR {{{ */
void ArmDisassembleMSR(t_ins * ins, t_uint32 instr, t_uint16 opc ) 
{
   ARMINSOPCODE(ins) = opc;
   INS_TYPE(ins) = IT_STATUS;
   ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
   INS_ATTRIB(ins) = 0x0;
   if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 
   ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
   ARMINSSHIFTLENGTH(ins) = 0;
   ARMINSREGS(ins) = ARM_REG_NONE;

   ARMINSFLAGS(ins) = 0x0;

   /* ARM DDI 01000E p.62-63: 
    * if R == 1 we modify the SPSR
    * else we modify the CPSR */

   if (instr & (1 << 22)) ARMINSFLAGS(ins) |= FL_SPSR;

   if (instr & (1<<16)) ARMINSFLAGS(ins) |= FL_CONTROL; /*FATAL(("Sets control bits to %s %x",(instr & (1 << 25))?"Imm":"Reg",(instr & (1 << 25))?BitRotateRight(instr & 0x000000ff, 2 * ((instr & 0x00000f00) >> 8)):instr & 0x0000000f));*/
   if (instr & (1<<17)) FATAL(("Sets extension bits"));
   if (instr & (1<<18)) FATAL(("Sets status bits"));
   if (instr & (1<<19)) ARMINSFLAGS(ins)|=FL_S; 
   

   ARMINSREGA(ins) = ARMINSREGB(ins) = ARM_REG_NONE;

   if (instr & (1 << 25)) {
      /* operand is een immediate */
      ARMINSREGC(ins) = ARM_REG_NONE;
      ARMINSIMMEDIATE(ins) = BitRotateRight(instr & 0x000000ff, 2 * ((instr & 0x00000f00) >> 8));
      ARMINSFLAGS(ins) |= FL_IMMED; 
   } else {
      /* operand zit in een register */
      ARMINSIMMEDIATE(ins) = 0;
      ARMINSREGC(ins) = instr & 0x0000000f;
   }
}
/*}}}*/
/*!
 * Disassemble normal ARM Loads and Stores (LDR, LDRB, STRH, ...)
 *
 * \param ins The arm instruction that gets filled in
 * \param instr The encoded arm instruction
 * \param opc The opcode
 *
 * \return void 
*/
/* ArmDisassembleDataTransfer {{{ */
void ArmDisassembleDataTransfer(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
  ARMINSOPCODE(ins) = opc;
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 

  if ((ARMINSOPCODE(ins) == ARM_LDR) ||(ARMINSOPCODE(ins) == ARM_LDRB) ||(ARMINSOPCODE(ins) == ARM_LDRH) ||(ARMINSOPCODE(ins) == ARM_LDRSH) || 
      (ARMINSOPCODE(ins) == ARM_LDRSB))
    INS_TYPE(ins) = IT_LOAD;
  else
    INS_TYPE(ins) = IT_STORE;

  ARMINSREGS(ins) = ARM_REG_NONE;

  ARMINSFLAGS(ins) = 0x0;

  switch (opc) {
  case ARM_STR:
  case ARM_LDR:
  case ARM_STRB:
  case ARM_LDRB:
    ARMINSREGA(ins) = (instr & 0x0000f000) >> 12;
    ARMINSREGB(ins) = (instr & 0x000f0000) >> 16;

    if (instr & (1 << 24)) ARMINSFLAGS(ins) |= FL_PREINDEX;
    if (instr & (1 << 23)) ARMINSFLAGS(ins) |= FL_DIRUP;
    if (instr & (1 << 21)) ARMINSFLAGS(ins) |= FL_WRITEBACK;

    if (instr & (1 << 25)) 
      {
	/* offset in offsetregister */
	ARMINSIMMEDIATE(ins) = 0;
	ARMINSREGC(ins) = instr & 0x0000000f;
	ArmShiftDecode(ins,instr);
      } 
    else 
      {
	/* offset gespecifieerd als 12-bit immediate */
	ARMINSIMMEDIATE(ins) = instr & 0x00000fff;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSFLAGS(ins) |= FL_IMMED;
      }
    break;
    
  case ARM_STRH:
  case ARM_LDRH:
    /*    case ARM_STRSH: */
  case ARM_LDRSH:
    /*    case ARM_STRSB: */
  case ARM_LDRSB:
    ARMINSREGA(ins) = (instr & 0x0000f000) >> 12;
    ARMINSREGB(ins) = (instr & 0x000f0000) >> 16;

    if (instr & (1 << 24)) ARMINSFLAGS(ins) |= FL_PREINDEX;
    if (instr & (1 << 23)) ARMINSFLAGS(ins) |= FL_DIRUP;
    if (instr & (1 << 21)) ARMINSFLAGS(ins) |= FL_WRITEBACK;

    if (instr & (1 << 22)) 
      {
	/* offset gespecifieerd als 8-bit immediate */
	ARMINSIMMEDIATE(ins) = ((instr & 0x00000f00) >> 4) | (instr & 0x000000f);
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSFLAGS(ins) |= FL_IMMED;
      } 
    else 
      {
	/* offset in offsetregister */
	ARMINSIMMEDIATE(ins) = 0;
	ARMINSREGC(ins) = instr & 0x0000000f;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
      }
  }

}
/*}}}*/
/*!
 * \todo Document 
 *
 * \param ins
 * \param instr
 * \param opc
 *
 * \return void 
*/
/* ArmDisassembleDataProcessing {{{ */
void ArmDisassembleDataProcessing(t_ins * ins, t_uint32 instr, t_uint16 opc) 
{
  ARMINSOPCODE(ins) = opc;
  INS_TYPE(ins) = IT_DATAPROC;
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 

  ARMINSREGA(ins) = (instr & 0x0000f000) >> 12;
  ARMINSREGB(ins) = (instr & 0x000f0000) >> 16;

  ARMINSFLAGS(ins) = 0x0;

  if (instr & (1 << 20)) ARMINSFLAGS(ins) |= FL_S;

  if (instr & (1 << 25)) {
    /* operand 2 is een immediate */
    ARMINSREGC(ins) = ARM_REG_NONE;
    ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
    ARMINSSHIFTLENGTH(ins) = 0;
    ARMINSREGS(ins) = ARM_REG_NONE;

    ARMINSIMMEDIATE(ins) = BitRotateRight(instr & 0x000000ff, 2 * ((instr & 0x00000f00) >> 8)); 
    ARMINSFLAGS(ins) |= FL_IMMED;
  } else {
    /* operand 2 is een register */
    ARMINSIMMEDIATE(ins) = 0;
    ARMINSREGC(ins) = instr & 0x0000000f;
    ArmShiftDecode(ins,instr);
  }

  /* enkele speciale gevallen waarbij een van de registers genegeerd wordt */
  if ((opc == ARM_MOV) || (opc == ARM_MVN))
    ARMINSREGB(ins) = ARM_REG_NONE;
  if ((opc == ARM_CMP) || (opc == ARM_CMN) || (opc == ARM_TST) || (opc == ARM_TEQ))
    ARMINSREGA(ins) = ARM_REG_NONE;
}
/*}}}*/
/*!
 * \todo Document 
 *
 * \param ins
 * \param instr
 * \param opc 
 *
 * \return void 
*/
/* ArmDisassembleUnsupported {{{ */
void ArmDisassembleUnsupported(t_ins * ins, t_uint32 instr, t_uint16 opc ) 
{
   FATAL(("Unsupported arm instruction. Opcode = %d\n",opc));
}
/* }}} */
/*!
 * \todo Document 
 *
 * \param ins
 * \param instr
 * \param opc 
 *
 * \return void 
*/
/* Data {{{ */
void ArmDisassembleData(t_ins * ins,t_uint32 instr, t_uint16 opc ) {
  /* fills in a arm_ins with the appropriate values if we're dealing with a data word */
  INS_TYPE(ins) = IT_DATA;
  ARMINSOPCODE(ins) = ARM_DATA;
  ARMINSCONDITION(ins) = ARM_CONDITION_AL;
  ARMINSREGA(ins) = ARMINSREGB(ins) = ARMINSREGC(ins) = ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = instr;
  ARMINSFLAGS(ins) = 0;
  INS_ATTRIB(ins) = IF_DATA;
}
/* }}} */
/*!
 * \todo Document 
 *
 * \param ins
 * \param instr
 * \param opc
 *
 * \return void 
*/
/* ArmDisassembleCPDO  {{{ */
void ArmDisassembleCPDO(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
  ARMINSOPCODE(ins) = opc;
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 

  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSFLAGS(ins) = 0x0;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;

 
  INS_TYPE(ins) = IT_FLT_ALU;

  if ((instr & (1<<19)) && (instr & (1<<7)))
    FATAL(("Illegal size for float"));
  else if (instr & (1<<19))
    ARMINSFLAGS(ins)|=FL_FLT_DOUBLE_EXTENDED;
  else if (instr & (1<<7))
    ARMINSFLAGS(ins)|=FL_FLT_DOUBLE;
  else 
    ARMINSFLAGS(ins)|=FL_FLT_SINGLE;


   switch(opc)
   {
     case ARM_MVF:
     case ARM_MNF:
     case ARM_SQT:
     case ARM_ABS:
       ARMINSREGA(ins) = ARM_REG_F0+ ((instr & 0x00007000) >> 12);
       if (instr & 0x8)
       {
	 ARMINSFLAGS(ins)|=FL_FLT_IMMED | FL_IMMED;
	 ARMINSIMMEDIATE(ins)= (instr & 0x00000007);
	 ARMINSREGC(ins) = ARM_REG_NONE;
       }
       else
       {
	 ARMINSREGB(ins) = ARM_REG_F0+(instr & 0x00000007) ;
       }
       ARMINSREGC(ins) = ARM_REG_NONE;
       /* Rounding mode */
       if (instr & (1 << 6))
       {
	 if (instr & (1 << 5)) ARMINSFLAGS(ins) |= FL_FLT_RZ;
	 else ARMINSFLAGS(ins) |= FL_FLT_RMI;
       }
       else if (instr & (1 << 5)) ARMINSFLAGS(ins) |= FL_FLT_RPI;
       break;
     case ARM_ADF:
     case ARM_SUF:
     case ARM_RSF:
     case ARM_MUF:
     case ARM_DVF:
     case ARM_RDF:
     case ARM_FML:
     case ARM_FDV:
     case ARM_FRD:
       ARMINSREGA(ins) = ARM_REG_F0+ ((instr & 0x00007000) >> 12);
       ARMINSREGB(ins) = ARM_REG_F0+ ((instr & 0x00070000) >> 16);
       if (instr &0x8)
       {
	 ARMINSIMMEDIATE(ins) = (instr & 0x00000007);
	 ARMINSFLAGS(ins) |= FL_FLT_IMMED | FL_IMMED;
	 ARMINSREGC(ins) = ARM_REG_NONE;
	 /*FATAL(("Implement Immediate operands!"));*/
       }
       else
       {
	 ARMINSREGC(ins) = ARM_REG_F0+ (instr & 0x7) ;
       }
       break;

     default:
       FATAL(("Unknown CPDO: Opcode = %d\n",opc));
   }


}
/* }}} */
/*!
 * \todo Document 
 *
 * \param ins
 * \param instr
 * \param opc
 *
 * \return void 
*/
/* ArmDisassembleCPRT {{{ */
void ArmDisassembleCPRT(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
  ARMINSOPCODE(ins) = opc;
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 
  
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSFLAGS(ins) = 0x0;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;

 
  switch(opc)
  {
	case ARM_FLT:
	  if ((instr & (1<<19)) && (instr & (1<<7)))
	    FATAL(("Illegal size for float"));
	  else if (instr & (1<<19))
	    ARMINSFLAGS(ins)|=FL_FLT_DOUBLE_EXTENDED;
	  else if (instr & (1<<7))
	    ARMINSFLAGS(ins)|=FL_FLT_DOUBLE;
	  else 
	    ARMINSFLAGS(ins)|=FL_FLT_SINGLE;

	  INS_TYPE(ins)=IT_FLT_INT;
	  ARMINSREGA(ins) = ARM_REG_F0+ ((instr & 0x000f0000) >> 16);
	  ARMINSREGB(ins) = (instr & 0x0000f000) >> 12;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  break;
	case ARM_FIX:
	  if ((instr & (1 << 19)) && (instr & (1 << 7)))
	    FATAL(("Bits should be zero!"));
	  /* Ignoring rounding mode for the moment! */
	  INS_TYPE(ins)=IT_FLT_INT;
	  ARMINSREGA(ins) = (instr & 0x0000f000) >> 12;
	  ARMINSREGB(ins) = ARM_REG_F0+ (instr & 0x0000000f);
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  if (instr & (1 << 6))
	  {
	    if (instr & (1 << 5)) ARMINSFLAGS(ins) |= FL_FLT_RZ;
	    else ARMINSFLAGS(ins) |= FL_FLT_RMI;
	  }
	  else if (instr & (1 << 5)) ARMINSFLAGS(ins) |= FL_FLT_RPI;
	  break;
	case ARM_WFS:
	  INS_TYPE(ins)=IT_FLT_STATUS;
	  ARMINSIMMEDIATE(ins) = 0;
	  ARMINSREGA(ins)=ARM_REG_FPSR;
	  ARMINSREGB(ins)= (instr & 0x0000f000) >> 12;
	  ARMINSREGC(ins)= ARM_REG_NONE;
	  break;
	case ARM_RFS:
	  INS_TYPE(ins)=IT_FLT_STATUS;
	  ARMINSIMMEDIATE(ins) = 0;
	  ARMINSREGA(ins) = (instr & 0x0000f000) >> 12;
	  ARMINSREGB(ins)=ARM_REG_FPSR;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  break;
	case ARM_WFC:
	  FATAL(("implement wfc"));
	  break;
	case ARM_RFC:
	  FATAL(("implement rfc"));
	  break;
	case ARM_CMFE:
	case ARM_CNFE:
	case ARM_CMF:
	case ARM_CNF:

	  ARMINSREGA(ins) = ARM_REG_NONE;
	  ARMINSFLAGS(ins) |= FL_S;
	  INS_TYPE(ins) = IT_FLT_ALU;
	  ARMINSREGB(ins) = ARM_REG_F0+ ((instr & 0x00070000) >> 16);
	  if (instr & 0x8)
	  {
	    ARMINSIMMEDIATE(ins) = (instr & 0x00000007);
	    ARMINSFLAGS(ins) |= FL_FLT_IMMED | FL_IMMED;
	    ARMINSREGC(ins) = ARM_REG_NONE;
	  }
	  else
	    ARMINSREGC(ins) = ARM_REG_F0+ (instr & 0x00000007);
	  break;
	  
	default:
	  FATAL(("Implement unknown CPRT"));
  }
}
/* }}} */
/*!
 * \todo Document 
 *
 * \param ins
 * \param instr
 * \param opc
 *
 * \return void 
*/
/* ArmDisassembleCPDT {{{ */
void ArmDisassembleCPDT(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
  ARMINSOPCODE(ins) = opc;
  ARMINSCONDITION(ins) = ARM_EXTRACT_CONDITION_CODE(instr);
  INS_ATTRIB(ins) = 0x0;
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL) INS_ATTRIB(ins)|=IF_CONDITIONAL; 

  if ((ARMINSOPCODE(ins) == ARM_LFM) ||(ARMINSOPCODE(ins) == ARM_LDF))
    INS_TYPE(ins) = IT_FLT_LOAD;
  else
    INS_TYPE(ins) = IT_FLT_STORE;

  ARMINSREGS(ins) = ARM_REG_NONE;

  ARMINSFLAGS(ins) = 0x0;

  switch (opc) {
    case ARM_STF:
    case ARM_LDF:
      if (instr & (1 << 24)) ARMINSFLAGS(ins) |= FL_PREINDEX;
      if (instr & (1 << 23)) ARMINSFLAGS(ins) |= FL_DIRUP;
      if (instr & (1 << 21)) ARMINSFLAGS(ins) |= FL_WRITEBACK;
      if ((instr & (1<<22)) && (instr & (1<<15)))
	ARMINSFLAGS(ins)|=FL_FLT_PACKED;
      else if (instr & (1<<22))
	ARMINSFLAGS(ins)|=FL_FLT_DOUBLE_EXTENDED;
      else if (instr & (1<<15))
	ARMINSFLAGS(ins)|=FL_FLT_DOUBLE;
      else 
	ARMINSFLAGS(ins)|=FL_FLT_SINGLE;



      ARMINSREGA(ins) = ARM_REG_F0+ ((instr & 0x00007000) >> 12);
      ARMINSREGB(ins) = (instr & 0x000f0000) >> 16;

      /* offset gespecifieerd als 8-bit immediate */
      ARMINSIMMEDIATE(ins) = 4*(instr & 0x000000ff);
      ARMINSREGC(ins) = ARM_REG_NONE;
      ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
      ARMINSSHIFTLENGTH(ins) = 0;
      ARMINSFLAGS(ins) |= FL_IMMED;
      break;

    case ARM_SFM:
    case ARM_LFM:
      {
	int num;
	int tel;
	if (instr & (1 << 24)) ARMINSFLAGS(ins) |= FL_PREINDEX;
	if (instr & (1 << 23)) ARMINSFLAGS(ins) |= FL_DIRUP;
	if (instr & (1 << 21)) ARMINSFLAGS(ins) |= FL_WRITEBACK;
	ARMINSIMMEDIATE(ins) = instr & 0x000000ff;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGA(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSFLAGS(ins) |= FL_IMMED;
	ARMINSMULTIPLE(ins)=RegsetNew(); 

	num=instr &(1<< 22)?2:0;
	num+=instr & (1 << 15)?1:0;
	if (num==0) num=4;
	for (tel=0; tel<num; tel++)
	{
	  if ((((instr&0x7000)>>12)+ARM_REG_F0+tel) > ARM_REG_F7) FATAL(("Implement wrap around in float load/store multiple"));
	  RegsetAddReg(ARMINSMULTIPLE(ins),((instr&0x7000)>>12)+ARM_REG_F0+tel);
	}
	ARMINSREGB(ins) = (instr & 0x000f0000) >> 16;
      }

      break;
  }

}
/*}}}*/
/* vim: set shiftwidth=2 foldmethod=marker : */
