/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_emulation.h"
#include "arm_utils.h"
#include "arm_instructions.h"
#include "arm_registers.h"
#include "../../utils/util_error.h"
#include "../../utils/util_bitmanipulation.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_procstate.h"
#include "../../kernel/diablo_iterators.h"
#include "../../diablo_options.h"
#include <string.h>
#include <strings.h>
/*}}}*/

#define CARRY_FROM(_a,_b,_r) ((_a >> 31) ? ((_b >> 31) | ((~_r) >> 31)) : ((_b >> 31) * ((~_r) >> 31)))
#define BORROWED_FROM(_a,_b,_r) ((_a >> 31) ? ((_b >> 31) & (_r >> 31)) : ((_b >> 31) | (_r >> 31)))

/* Macros for flexible operands */

#define ARM_SIGN_BIT(_x) (((_x) >> 31) & 0x1)


#define ISVALUE(x) ((x)==VALUE)
#define ISBOT(x) ((x)==BOT)
#define ISTOP(x) ((x)==TOP)


/*!
 * This procedure Extracts the Shift operand (i.e. the value that will be used
 * after all shifts are applied) from 1. a instruction and 2. the state in
 * which the instruction is executed and stores this value in shifted_value. It
 * also modifies the condition bits in case of a S-etting instruction 
 *
 * \todo Complete documentation
 *
 * \param ins
 * \param state
 * \param shifted_value
 *
 * \return t_lattice_level 
*/
/* Utility function to emulate the shifter part of an instruction {{{ */
t_lattice_level ArmInsExtractShift(t_ins * ins, t_procstate * state, register_content * shifted_value)
{
  t_uint32 shiftimmediate;
  register_content tmp_value_to_shift;
  register_content tmp_shiftreg_value;
  t_bool carry_set=FALSE;
  t_bool tmp_flag;
  t_uint32 value_to_shift;
  t_uint32 shiftreg_value;

  /* Get the value base register that holds the value to shift, if that fails,
   * we make worst case assumptions for the shifter {{{*/

  if (!ISVALUE(ProcStateGetReg(state,ARMINSREGC(ins),&tmp_value_to_shift)))
    {
      if (ArmInsUpdatesCond(ins))
	{
	  ProcStateSetCondBot(state,ARM_REG_C_CONDITION);
	}
      return BOT;
    }

  /*}}}*/
  
  value_to_shift = G_T_UINT32(tmp_value_to_shift.i);
  shiftimmediate = ARMINSSHIFTLENGTH(ins);
  
  switch (ARMINSSHIFTTYPE(ins)) /* perform the right shift and set carry if necessary */
    {
      /* Shifts over an immediate value {{{ */
    case ARM_SHIFT_TYPE_LSL_IMM:
      if (shiftimmediate > 0) 
	if (BIT_IS_SET(value_to_shift, (32-shiftimmediate))) 
	  carry_set=TRUE;
      GenericSet_uint32(shifted_value->i,value_to_shift << shiftimmediate);
      break;
    case ARM_SHIFT_TYPE_LSR_IMM: /* no carry can be shifted out here */
      GenericSet_uint32(shifted_value->i,(value_to_shift >> shiftimmediate));
      break;
    case ARM_SHIFT_TYPE_ASR_IMM: /* no carry can be shifted out here */
      if ((t_int32) value_to_shift < 0)
	GenericSet_uint32(shifted_value->i,(value_to_shift >> shiftimmediate) | ~(0xFFFFFFFF >> shiftimmediate));
      else 
	GenericSet_uint32(shifted_value->i,value_to_shift >> shiftimmediate);
      break;
    case ARM_SHIFT_TYPE_ROR_IMM:
      if (shiftimmediate == 0)
	{
	  GenericSet_uint32(shifted_value->i,value_to_shift);
	  break;
	}
      if (BIT_IS_SET(value_to_shift, (shiftimmediate-1))) 
	carry_set=TRUE;
      GenericSet_uint32(shifted_value->i,(value_to_shift >> shiftimmediate) | (value_to_shift << (32 - shiftimmediate)));
      break;
      /*}}}*/

      /* Shifts over a register, can be necessary to make worst case assumptions, if the amount to shift is unknown {{{*/
    case ARM_SHIFT_TYPE_LSL_REG:
      if (!ISVALUE(ProcStateGetReg(state,ARMINSREGS(ins),&tmp_shiftreg_value)))
	{
	  if (ArmInsUpdatesCond(ins))
	    {
	      ProcStateSetCondBot(state,ARM_REG_C_CONDITION);
	    }
	  return BOT;
	}
      shiftreg_value = G_T_UINT32(tmp_shiftreg_value.i);
      
      if (ARM_SHIFT_AMOUNT(shiftreg_value) == 0) /* no carry can be shifted out here */
	GenericSet_uint32(shifted_value->i,value_to_shift);
      else if (ARM_SHIFT_AMOUNT(shiftreg_value) > 32) /* no carry can be shifted out here */
	GenericSet_uint32(shifted_value->i,0);
      else
	{
	  if (BIT_IS_SET(value_to_shift, (32 - ARM_SHIFT_AMOUNT(shiftreg_value)))) 
	    carry_set=TRUE;
	  GenericSet_uint32(shifted_value->i,value_to_shift << ARM_SHIFT_AMOUNT(shiftreg_value));
	}
      break;
    case ARM_SHIFT_TYPE_LSR_REG: /* no carry can be shifted out here */
      if (!ISVALUE(ProcStateGetReg(state,ARMINSREGS(ins),&tmp_shiftreg_value)))
	{
	  if (ArmInsUpdatesCond(ins))
	    {
	      ProcStateSetCondBot(state,ARM_REG_C_CONDITION);
	    }
	  return BOT;
	}
      shiftreg_value = G_T_UINT32(tmp_shiftreg_value.i);
      GenericSet_uint32(shifted_value->i,(value_to_shift >> ARM_SHIFT_AMOUNT(shiftreg_value)));
      break;
    case ARM_SHIFT_TYPE_ASR_REG: /* no carry can be shifted out here */
      if (!ISVALUE(ProcStateGetReg(state,ARMINSREGS(ins),&tmp_shiftreg_value)))
	{
	  if (ArmInsUpdatesCond(ins))
	    {
	      ProcStateSetCondBot(state,ARM_REG_C_CONDITION);
	    }
	  return BOT;
	}
      shiftreg_value = G_T_UINT32(tmp_shiftreg_value.i);

      if ((t_int32) value_to_shift < 0)
	GenericSet_uint32(shifted_value->i,(value_to_shift >> ARM_SHIFT_AMOUNT(shiftreg_value)) | ~(0xFFFFFFFF >> ARM_SHIFT_AMOUNT(shiftreg_value)));
      else GenericSet_uint32(shifted_value->i,value_to_shift >> ARM_SHIFT_AMOUNT(shiftreg_value));
      break;
    case ARM_SHIFT_TYPE_ROR_REG:
      if (!ISVALUE(ProcStateGetReg(state,ARMINSREGS(ins),&tmp_shiftreg_value)))
	{
	  if (ArmInsUpdatesCond(ins))
	    {
	      ProcStateSetCondBot(state,ARM_REG_C_CONDITION);
	    }
	  return BOT;
	}
      shiftreg_value = G_T_UINT32(tmp_shiftreg_value.i);

      if (ARM_ROTATE_AMOUNT(shiftreg_value) == 0)
	{
	  GenericSet_uint32(shifted_value->i,value_to_shift);
	  break;
	}
      if (BIT_IS_SET(value_to_shift, (ARM_ROTATE_AMOUNT(shiftreg_value)-1))) 
	carry_set=TRUE;
      GenericSet_uint32(shifted_value->i,(value_to_shift >> ARM_ROTATE_AMOUNT(shiftreg_value)) | (value_to_shift << (32 - ARM_ROTATE_AMOUNT(shiftreg_value)))  /* CHECK THIS */);
      break;
      /*}}}*/

    case ARM_SHIFT_TYPE_RRX:
      if (value_to_shift & 0x1) carry_set=TRUE;
      GenericSet_uint32(shifted_value->i,(BOT==ProcStateGetCond(state,ARM_REG_C_CONDITION,&tmp_flag) || tmp_flag) ? ((LL(0x1) << 31) & (value_to_shift >> 1)) : (value_to_shift >> 1));
      break;
    default: 
      FATAL(("Unknown shifttype while emulating!!"));
      break;
    }

  if (ArmInsUpdatesCond(ins))
  {
    ProcStateSetCond(state,ARM_REG_C_CONDITION,carry_set);
  }

  return VALUE;
}
/*}}}*/



#define ArmInsRegisterBMatters(ins) (ARMINSOPCODE(ins)!=ARM_MVN && ARMINSOPCODE(ins)!=ARM_MOV)
#define ArmInsNeedsCarry(ins)       (ARMINSOPCODE(ins)==ARM_RSC || ARMINSOPCODE(ins)==ARM_SBC || ARMINSOPCODE(ins)==ARM_ADC)
#define ArmInsIsMul(ins)            (ARMINSOPCODE(ins)==ARM_MUL || ARMINSOPCODE(ins)==ARM_MLA  || ARMINSOPCODE(ins)==ARM_UMULL)
#define ArmInsIsLogical(ins)        (ARMINSOPCODE(ins)==ARM_AND|| ARMINSOPCODE(ins)==ARM_TST|| ARMINSOPCODE(ins)==ARM_BIC|| ARMINSOPCODE(ins)==ARM_EOR|| ARMINSOPCODE(ins)==ARM_ORR|| ARMINSOPCODE(ins)==ARM_TEQ|| ARMINSOPCODE(ins)==ARM_MOV|| ARMINSOPCODE(ins)==ARM_MVN)

static void ArmInsEmulDetermineTag(t_ins * ins, t_procstate * state)
{
  t_reg reg;

  if (!RegsetIn(INS_REGS_USE(ins),ARMINSREGA(ins)))
    ProcStateSetTagTop(state,ARMINSREGA(ins));

  REGSET_FOREACH_REG(INS_REGS_USE(ins),reg)
  {
    ProcStateJoinTags(state,ARMINSREGA(ins),reg);
  }
}

/* Emulate specific instructions {{{ */

static t_bool ArmGetRegisterB(t_ins * ins, t_procstate * state, register_content * tmp_src1)
{
  /* Get register B and return FALSE if that fails */
  if (!ISVALUE(ProcStateGetReg(state,ARMINSREGB(ins),tmp_src1)))
  {
    return FALSE;
  }
 
  return TRUE;
}

static t_bool ArmGetRegisterC(t_ins * ins, t_procstate * state, register_content * tmp_src2)
{
  /* If the instruction has an immediate, we need this value */
  if (ArmInsHasImmediate(ins))
    GenericSet_uint32(tmp_src2->i,ARMINSIMMEDIATE(ins));
  /* If it has a shifted value, we need that value */
  else if (ArmInsHasShiftedFlexible(ins))
  {
    /* Get the shifted value is not constant, make worst case assumptions if
     * that fails */
    if (!ISVALUE(ArmInsExtractShift(ins,state,tmp_src2)))
    {
      return FALSE;
    }
  }
  /* If it is none of the above, simply get the value from the register and
   * make worst case assumptions if that fails */
  else if (!ISVALUE(ProcStateGetReg(state,ARMINSREGC(ins),tmp_src2)))
  {
    return FALSE;
  }
  return TRUE;
}

static t_bool ArmGetRegisterS(t_ins * ins, t_procstate * state, register_content * tmp_src3)
{
  /* Get register S and return FALSE if that fails */
  if (!ISVALUE(ProcStateGetReg(state,ARMINSREGS(ins),tmp_src3)))
  {
    return FALSE;
  }
 
  return TRUE;
}


static t_bool ArmGetCarry(t_ins * ins, t_procstate * state, t_bool * carry_flag)
{
  if (!ISVALUE(ProcStateGetCond(state,ARM_REG_C_CONDITION,carry_flag)))
  {
	  return FALSE;
  }
  return TRUE;
}

/*#define LOAD_AGGRESSIVE */
/* Emulate numerical an logical instructions {{{ */
static void ArmInsEmulNumericalAndLogical(t_ins * ins, t_procstate * state)
{
  register_content tmp_src1, tmp_src2, tmp_src3, dest;
  t_uint32 src1,src2;
  t_regset mask = NullRegs;
  t_int64 tmp = 0L; 
  t_int64 carry_set=FALSE; /* Just to keep the compiler happy, value is never used! */
  t_bool carry_flag;
  t_bool can_emulate=TRUE;
  /* Emulate the tags {{{*/
  if (ARMINSREGA(ins)!=ARM_REG_NONE)
  {
#if 0
    t_reloc * rel1;
    t_reloc * rel2;
    t_reloc * tmp_rel;

    if ((ARMINSOPCODE(ins)==ARM_ADD)&&(!ArmInsHasImmediate(ins))&&(!ArmInsHasShiftedFlexible(ins))&&(ISVALUE( ProcStateGetTag(state,ARMINSREGB(ins),&rel1)) &&(ISVALUE( ProcStateGetTag(state,ARMINSREGC(ins),&rel2)))  ))
    {
      if (((rel1->write_type==SREL32)|| (rel2->write_type==SREL32))
      && ((rel1->write_type==ABS32)|| (rel2->write_type==ABS32) ||  (rel1->write_type==ADDRESS_PROC) ||  (rel2->write_type==ADDRESS_PROC)) )
      {
	if (rel2->write_type==SREL32)
	{
	  tmp_rel=rel2;
	  rel2=rel1;
	  rel1=tmp_rel;
	}

	/* rel1 holds an srel value */

	if (REL_TO_RELOCATABLE(rel2)->type==RT_BBL)
	{
	  if (rel1->from_type==RELOC_FROM_INSTRUCTION)
	  {
	    if (INS_BBL(T_INS(rel1->from))==T_BBL(rel2REL_TO_RELOCATABLE()))
	    {
	      /* Cool, made a srel into an abs value, the only remaining value is rel1 */
	      ProcStateSetTag(state,ARMINSREGA(ins),rel1);
	    }
	    else
	      FATAL(("Implement 3"));
	  }
	  else
	    FATAL(("Implement 1"));
	}
	else
	  FATAL(("Implement 2"));
	
      }
      else
      FATAL(("Adding two relocations, @R and @R\n",rel1,rel2));
    }
    else
#endif
	  ArmInsEmulDetermineTag(ins,state);
  }
  /*}}}*/
 
  {
    t_reg reg;
    REGSET_FOREACH_REG(INS_REGS_USE(ins),reg)
    {
      if (ISTOP(ProcStateGetReg(state,reg,&tmp_src1)))
      {
#ifdef LOAD_AGGRESSIVE
	DiabloPrint(stdout,"== Top encountered in reg %d in @I\n",reg,ins);
	ProcStateSetRegsetTop(state,INS_REGS_DEF(ins));
	if (ArmInsUpdatesCond(ins))
	{
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);

	  if ((!ArmInsIsLogical(ins))&& (!ArmInsIsMul(ins)))
	  {
	    RegsetAddReg(mask,ARM_REG_C_CONDITION);
	    RegsetAddReg(mask,ARM_REG_Q_CONDITION);
	    RegsetAddReg(mask,ARM_REG_V_CONDITION);
	  }
	  ProcStateSetCondsetTop(state,mask);
	}
#else
	FATAL(("Top encountered"));
#endif
	return;
      }
    }
  }

  if (ArmInsRegisterBMatters(ins)) 
    can_emulate&=ArmGetRegisterB(ins,state,&tmp_src1);
  can_emulate&=ArmGetRegisterC(ins,state,&tmp_src2);
  if (ArmInsNeedsCarry(ins)) 
    can_emulate&=ArmGetCarry(ins,state,&carry_flag);
  if (ARMINSOPCODE(ins)==ARM_MLA)
    can_emulate&=ArmGetRegisterS(ins,state,&tmp_src3);


  if (!can_emulate)
  {
    ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
    if (ArmInsUpdatesCond(ins))
    {
      RegsetAddReg(mask,ARM_REG_N_CONDITION);
      RegsetAddReg(mask,ARM_REG_Z_CONDITION);

      if ((!ArmInsIsLogical(ins))&& (!ArmInsIsMul(ins)))
      {
	RegsetAddReg(mask,ARM_REG_C_CONDITION);
	RegsetAddReg(mask,ARM_REG_Q_CONDITION);
	RegsetAddReg(mask,ARM_REG_V_CONDITION);
      }
      ProcStateSetCondsetBot(state,mask);
    }
    return;
  }
  
  src1 = G_T_UINT32(tmp_src1.i);
  src2 = G_T_UINT32(tmp_src2.i);

  bzero(&dest,sizeof(register_content));

  /* Perform the operation and set the defined register{{{ */
  switch(ARMINSOPCODE(ins))
  {
    /* Logical instructions */
    case ARM_MVN:
      GenericSet_uint32( dest.i,~src2);
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      break;
    case ARM_MOV:
      GenericSet_uint32( dest.i,src2);
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      break;
    case ARM_TST:
      GenericSet_uint32( dest.i,src1 & src2);
      break;
    case ARM_AND:
      GenericSet_uint32( dest.i,src1 & src2);
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      break;
    case ARM_BIC:
      GenericSet_uint32( dest.i,src1 & ~src2);
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      break;
    case ARM_EOR:
      GenericSet_uint32( dest.i,src1 ^ src2);
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      break;
    case ARM_ORR:
      GenericSet_uint32( dest.i,src1 | src2);
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      break;
    case ARM_TEQ:
      GenericSet_uint32( dest.i,src1 ^ src2);
      break;


    /* Arithmitic */
    case ARM_CMN:
      tmp =  (t_int32) src1;  
      tmp +=  (t_int32) src2;
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      carry_set = CARRY_FROM(src1, src2, (t_uint32) ((t_uint64) tmp));
      break;
    case ARM_CMP:
      tmp =  (t_int32) src1;  
      tmp -=  (t_int32) src2;
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      carry_set = !BORROWED_FROM(src1, src2, (t_uint32) ((t_uint64) tmp));
      break;
    case ARM_RSB:
      tmp =  (t_int32) src2;  
      tmp -=  (t_int32) src1;
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      carry_set = !BORROWED_FROM(src1, src2, (t_uint32) ((t_uint64) tmp));
      break;
    case ARM_SUB:
      tmp =  (t_int32) src1;  
      tmp -=  (t_int32) src2;
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      carry_set = !BORROWED_FROM(src1, src2, (t_uint32) ((t_uint64) tmp));
      break;
    case ARM_RSC:
      tmp =  (t_int32) src2;  
      tmp -=  (t_int32) src1;
      if (!carry_flag)
	tmp--;
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      carry_set = !BORROWED_FROM(src1, src2, (t_uint32) ((t_uint64) tmp));
      break;
    case ARM_SBC:
      tmp =  (t_int32) src1;  
      tmp -=  (t_int32) src2;
      if (!carry_flag)
	tmp--;
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      carry_set = !BORROWED_FROM(src1, src2, (t_uint32) ((t_uint64) tmp));
      break;
    case ARM_ADD:
      tmp =  (t_int32) src1;  
      tmp +=  (t_int32) src2;
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      carry_set = CARRY_FROM(src1, src2, (t_uint32) ((t_uint64) tmp));
      break;
    case ARM_ADC:
      tmp =  (t_int32) src1;  
      tmp +=  (t_int32) src2;
      if (carry_flag)
	tmp++;
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      carry_set = CARRY_FROM(src1, src2, (t_uint32) ((t_uint64) tmp));
      break;

      /* MUL */
    case ARM_MUL:
      tmp  = (t_int32) src1;
      tmp *= (t_int32) src2;
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      break;
    case ARM_MLA:
      tmp  = (t_int32) src1;
      tmp *= (t_int32) src2;
      tmp += (t_int32) G_T_UINT32(tmp_src3.i);
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      break;

    case ARM_UMULL:
      tmp  = (t_int32) src1;
      tmp *= (t_int32) src2;
      /* Lower half goes into RegS */
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGS(ins),dest);
      tmp >>= 32;
      /* Upper half goes into RegA */
      GenericSet_uint32( dest.i,(t_uint32) ((t_uint64) tmp));
      ProcStateSetReg(state,ARMINSREGA(ins),dest);
      break;
    default:
      break;
    }
  /*}}}*/
  
  /* If necessary: Set Conditions {{{*/
  if (ArmInsUpdatesCond(ins))
    {
      if (ARMINSOPCODE(ins)==ARM_UMULL) FATAL(("Implement"));

      if (G_T_UINT32(dest.i)==0)
	ProcStateSetCond(state,ARM_REG_Z_CONDITION,TRUE);
      else
	ProcStateSetCond(state,ARM_REG_Z_CONDITION,FALSE);
      
      if (ARM_SIGN_BIT(G_T_UINT32(dest.i))==1)
	ProcStateSetCond(state,ARM_REG_N_CONDITION,TRUE);
      else
	ProcStateSetCond(state,ARM_REG_N_CONDITION,FALSE);

      if ((!ArmInsIsLogical(ins)) && (!ArmInsIsMul(ins)))
      {
	if (carry_set)
	  ProcStateSetCond(state,ARM_REG_C_CONDITION,TRUE);
	else
	  ProcStateSetCond(state,ARM_REG_C_CONDITION,FALSE);
      
	/* might be undefined however */
	if (((tmp >> 32) & 0x1) != ARM_SIGN_BIT(tmp))
	  ProcStateSetCond(state,ARM_REG_V_CONDITION,TRUE);
	else
	  ProcStateSetCond(state,ARM_REG_V_CONDITION,FALSE);
      }
    }
  /* }}} */
}

/* }}} */

/* Emulate constant producers {{{ */
static void ArmInsEmulConstProd(t_ins * ins, t_procstate * state)
{
  register_content dest;
  bzero(&dest,sizeof(register_content));
  GenericSet_uint32(dest.i,ARMINSIMMEDIATE(ins));
  ProcStateSetReg(state,ARMINSREGA(ins),dest);
  ProcStateSetTagTop(state,ARMINSREGA(ins));
}
/* }}} */

/* Emulate address producers {{{*/

static void ArmInsEmulAddrProd(t_ins * ins, t_procstate * state)
{
  register_content dest;
  bzero(&dest,sizeof(register_content));
  GenericSet_uint32(dest.i,ARMINSIMMEDIATE(ins));
  ProcStateSetTag(state,ARMINSREGA(ins),RELOCATABLE_REFERS_TO(ins)->rel);
  ProcStateSetReg(state,ARMINSREGA(ins),dest);
}
/* }}}*/


/* Emulate loads {{{*/
static void ArmInsEmulLoad(t_ins * ins, t_procstate * state)
{
  t_bool can_emulate=TRUE;
  t_reloc * base_reloc;
  register_content tmp_base_address, tmp_offset, dest;
  t_uint32 base_address=0; /* Just to keep the compiler happy, value is never used */
  t_uint32 offset=0; /* Just to keep the compiler happy, value is never used */
  t_uint32 offset2;
  t_lattice_level base_reloc_unknown;
  t_instruction_flags insflags = ARMINSFLAGS(ins);
  int i;
  t_uint32 num_regs = 0;

  if (ARMINSOPCODE(ins) == ARM_LDM) for (i=0; i<16; i++) if (ARMINSIMMEDIATE(ins) & (1 << i)) num_regs++;
  /* get the base address and reloc from the state */
 
  base_reloc_unknown = ProcStateGetTag(state,ARMINSREGB(ins),&base_reloc);
  can_emulate&=ArmGetRegisterB(ins,state,&tmp_base_address);
  if (ARMINSOPCODE(ins)!=ARM_LDM) can_emulate&=ArmGetRegisterC(ins,state,&tmp_offset);

  bzero(&dest,sizeof(register_content));

  if (can_emulate)
  {
    if (ARMINSOPCODE(ins)!=ARM_LDM) offset = G_T_UINT32(tmp_offset.i);
    base_address = G_T_UINT32(tmp_base_address.i);
    if (ISVALUE(base_reloc_unknown))
    {
      switch(ARMINSOPCODE(ins))
      {
	case ARM_LDR:
	case ARM_LDRB:
	case ARM_LDRSH:
	case ARM_LDRH:
	case ARM_LDRSB:
	  /* Do the load {{{  */
	  if (REL_TO_RELOCATABLE(base_reloc)->type==RT_SUBSECTION)
	  {
	    t_section *sec = T_SECTION(REL_TO_RELOCATABLE(base_reloc));
	    if (sec->type != CODE_SECTION)
	    {
	      /* Bruno: added this check. We should load these values! */
	      if (sec->type == RO_DATA_SECTION) WARNING(("Missing an opportunity for sec %s", sec->name));
	      ProcStateSetRegBot(state,ARMINSREGA(ins));
	      ProcStateSetTagBot(state,ARMINSREGA(ins));
	    }
	    else
	    {
	      /* Bruno: relocation to a code subsection? They should always point to a basic block! */
	      FATAL(("This should never happen"));
	    }
	  }
	  else if (REL_TO_RELOCATABLE(base_reloc)->type==RT_BBL)
	  {
	    t_bbl * bbl;
	    t_address tmp_addr;
	    t_uint32 tmp;

	    if (insflags & FL_PREINDEX) /* only apply offset if preindexed! */
	      tmp = base_address + (insflags & FL_DIRUP ? +1 : -1)*offset;
	    else
	      tmp= base_address;

	    bbl=T_BBL(REL_TO_RELOCATABLE(base_reloc));
	    GenericSet_uint32(tmp_addr,tmp);
	    offset2 = G_T_UINT32(GENERIC_SUB(tmp_addr,BBL_CADDRESS(bbl)));
	    /* Ludo: Only LDR must load from an 4-aligned address (added && ...) */
	    if ( (offset2 & 3) &&  ARMINSOPCODE(ins)==ARM_LDR ) /* end Ludo */
	    {
	      ProcStateSetRegBot(state,ARMINSREGA(ins));
	      ProcStateSetTagBot(state,ARMINSREGA(ins));
	    }
	    /* Ludo: LDRH and LDRSH must load from an 2-aligned address */
	    else if ( (offset2 & 1) && (ARMINSOPCODE(ins)==ARM_LDRH || ARMINSOPCODE(ins)==ARM_LDRSH) ) 
	    {
	      ProcStateSetRegBot(state,ARMINSREGA(ins));
	      ProcStateSetTagBot(state,ARMINSREGA(ins));
	    }
	    /* end Ludo */
	    else if ((offset2 < 0) || (offset2 >= 4*bbl->nins)) 
	    {
	      /* Load is out of block */
#ifdef LOAD_AGGRESSIVE
	      DiabloPrint(stdout,"== Load out of block: @I loads at offset %d in block @B\n",ins,offset2,bbl);
	      ProcStateSetRegTop(state,ARMINSREGA(ins));
	      ProcStateSetTagTop(state,ARMINSREGA(ins));
#else
	      ProcStateSetRegBot(state,ARMINSREGA(ins));
	      ProcStateSetTagBot(state,ARMINSREGA(ins));
#endif
	    } 
	    else 
	    {
	      t_uint32 i;
	      t_ins * i_ins;
	      t_uint32 loaded;
	      i_ins = bbl->ins_first;
	      /* Ludo: changed offset to (offset & 3) because addresses are not garantueed aligned */
	      for (i=0; i < (offset2 & ~(3)); i+=4) i_ins = INS_NEXT(i_ins);
	      /* end Ludo */
	      if (!i_ins) FATAL(("NO INS! (offset = %d), bbl nins=%d",offset2,bbl->nins));
	      loaded = ARMINSIMMEDIATE(i_ins);

	      /* Ludo: now extract the right bytes from the loaded word and
	       * eventually sign-extend.  We silently assume here that the
	       * memory system is litte endian, because the load's behave
	       * different for big endian memories */

	      if (ARMINSOPCODE(ins) != ARM_LDR)
		loaded >>= (offset2 & 3) << 3; 
	      switch ARMINSOPCODE(ins) 
	      {
		case ARM_LDRB:
		  loaded &= 0xff;
		  break;
		case ARM_LDRSB:
		  loaded &= 0xff;
		  /* Sign extend */
		  if (loaded & 0x8) loaded |= 0xffffff00;
		  break;
		case ARM_LDRH:
		  loaded &= 0xffff;
		  break;
		case ARM_LDRSH:
		  loaded &= 0xffff;
		  /* Sign extend */
		  if (loaded & 0x80) loaded |= 0xffff0000;
		  break;
		default: 
		  break;
	      }
	      /* end Ludo */
	      GenericSet_uint32(dest.i,loaded);
	      if (RELOCATABLE_REFERS_TO(i_ins))
		ProcStateSetTag(state,ARMINSREGA(ins),RELOCATABLE_REFERS_TO(i_ins)->rel);
	      else
		ProcStateSetTagTop(state,ARMINSREGA(ins));
	      ProcStateSetReg(state,ARMINSREGA(ins),dest);
	    }
	  }
	  else
	  {
	    ProcStateSetRegBot(state,ARMINSREGA(ins));
	    ProcStateSetTagBot(state,ARMINSREGA(ins));
	  }
	  /*}}}*/
	  break;
	case ARM_LDM:
	  /* load from known address: base address is known {{{*/

	  if (REL_TO_RELOCATABLE(base_reloc)->type==RT_SUBSECTION)
	  {
	    t_section *sec = T_SECTION(REL_TO_RELOCATABLE(base_reloc));
	    if (sec->type != CODE_SECTION)
	    {
	      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
	      ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
	    }
	    else
	    {
	      FATAL(("This cannot happen!"));
	    }
	  }
	  else if (REL_TO_RELOCATABLE(base_reloc)->type==RT_BBL)
	  {
	    /* Emulate load from a bbl {{{*/
	    t_bbl * bbl;
	    int sign = (insflags & FL_DIRUP ? +1 : -1);
	    int startoffset = (insflags & FL_PREINDEX ? 4 : 0);
	    t_address tmp_addr;
	    t_uint32 tmp = base_address + sign*startoffset;
	    bbl=T_BBL(REL_TO_RELOCATABLE(base_reloc));
	    GenericSet_uint32(tmp_addr,tmp);
	    offset = G_T_UINT32(GENERIC_SUB(tmp_addr,BBL_CADDRESS(bbl)));
	    if (offset & 3) 
	    {
	      /* Loading from an unaligned address is unpredictable */
	      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
	      ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
	    }

	    if ((offset+4*sign*num_regs<0) || (offset + 4*sign*num_regs > 4*bbl->nins) || (((t_int32)offset) <0)) 
	    {
	 /*     BblPrint(bbl,0x1f);
	      FATAL(("Went out of block in @I, base address = %x offset in block is %d, numregs=%d and sign = %d\n",ins,base_address,offset,num_regs,sign));
*/
	      /* Load is out of block */
#ifdef LOAD_AGGRESSIVE
	      ProcStateSetRegsetTop(state,INS_REGS_DEF(ins));
	      ProcStateSetTagsetTop(state,INS_REGS_DEF(ins));
#else
	      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
	      ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
#endif
	    } 
	    else 
	    {
	      t_uint32 i;
	      t_ins * i_ins;
	      t_uint32 loaded;

	      i_ins = bbl->ins_first;
	      for (i=0; i < offset; i+=4) i_ins = INS_NEXT(i_ins);

	      for (i=0; i<16; i++)
		if (ARMINSIMMEDIATE(ins) & (1 << i)) 
		{
		  loaded = ARMINSIMMEDIATE(i_ins);
		  GenericSet_uint32(dest.i,loaded);
		  if (RELOCATABLE_REFERS_TO(i_ins))
		    ProcStateSetTag(state,i,RELOCATABLE_REFERS_TO(i_ins)->rel);
		  else
		    ProcStateSetTagTop(state,i);
		  ProcStateSetReg(state,i,dest);
		  if (sign > 0) i_ins = INS_NEXT(i_ins); else i_ins = INS_PREV(i_ins);
		}

	    }
	    /*}}}*/
	  }
	  else
	  {
	    ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
	    ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
	  }
	  /* }}} */
	  break;
	default:
	  FATAL(("Implement new type of load"));
      }

    }
    else if (ISBOT(base_reloc_unknown))
    {
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
      ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
    }
    else
    {
#ifdef LOAD_AGGRESSIVE
      DiabloPrint(stdout, "== @I is a load and the base register has no tag...\n",ins);
      ProcStateSetRegsetTop(state,INS_REGS_DEF(ins));
      ProcStateSetTagsetTop(state,INS_REGS_DEF(ins));
#else
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
      ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
#endif
    }
  }
  /* The address is not known, but it is possible we know the tag. {{{*/
  else
  {
    switch(ARMINSOPCODE(ins))
    {
      case ARM_LDR:
      case ARM_LDRB:
      case ARM_LDRSH:
      case ARM_LDRH:
      case ARM_LDRSB:
	if ((ISVALUE(base_reloc_unknown))&&(REL_TO_RELOCATABLE(base_reloc)->type==RT_BBL))
	{
	  t_ins * i_ins;
	  t_reloc * rel=NULL;
	  t_uint32 refs=0;

	  ProcStateSetTagTop(state,ARMINSREGA(ins));
	  BBL_FOREACH_INS(T_BBL(REL_TO_RELOCATABLE(base_reloc)),i_ins)
	  {
	    if (RELOCATABLE_REFERS_TO(i_ins)) 
	    {
	      if (refs==0)
	      {
		ProcStateSetTag(state,ARMINSREGA(ins),RELOCATABLE_REFERS_TO(i_ins)->rel);
		refs++;
		rel=RELOCATABLE_REFERS_TO(i_ins)->rel;
	      }
	      else if (!(TwoRelocsInSameBlock(RELOCATABLE_REFERS_TO(i_ins)->rel,rel)))
	      {
		ProcStateSetTagBot(state,ARMINSREGA(ins));
		break;
	      }
	    }
	  }
	}
	else if ((ISVALUE(base_reloc_unknown))&&(REL_TO_RELOCATABLE(base_reloc)->type==RT_SUBSECTION)&&(T_SECTION(REL_TO_RELOCATABLE(base_reloc))->type==RO_DATA_SECTION))
	  WARNING(("Missing an opportunity to load everything from @R",base_reloc));
	else ProcStateSetTagBot(state,ARMINSREGA(ins));
	ProcStateSetRegBot(state,ARMINSREGA(ins));
	break;
      default:
	ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
	ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
	break;
    }
  }
  /*}}}*/
  /* If we have writeback, emulate the indexing operation {{{ */
  if (ARMINSOPCODE(ins)==ARM_LDM)
  {

    if (insflags & FL_WRITEBACK)
    {	/* writeback enabled */
      if (!can_emulate)
      {
	ProcStateSetRegBot(state,ARMINSREGB(ins));
      } 
      else 
      {
	if (insflags & FL_DIRUP)
	  GenericSet_uint32(dest.i,base_address + 4 * num_regs);
	else 
	  GenericSet_uint32(dest.i,base_address - 4 * num_regs);
	if (ISVALUE(base_reloc_unknown))
	  ProcStateSetTag(state,ARMINSREGB(ins),base_reloc);
	ProcStateSetReg(state,ARMINSREGB(ins),dest);
      }
    }
  }
  else
    if (!(insflags & FL_PREINDEX) || (insflags & FL_WRITEBACK)) 
    {	/* writeback enabled */
      if ((!can_emulate))
      {
	ProcStateSetRegBot(state,ARMINSREGB(ins));
      } 
      else 
      {
	if (insflags & FL_DIRUP)
	{
	  if (ARMINSOPCODE(ins)==ARM_LDM)
	    GenericSet_uint32(dest.i,base_address + 4 * num_regs);
	  else
	    GenericSet_uint32(dest.i,base_address + offset);
	}
	else
	{
	  if (ARMINSOPCODE(ins)==ARM_LDM)
	    GenericSet_uint32(dest.i,base_address - 4 * num_regs);
	  else
	    GenericSet_uint32(dest.i,base_address - offset);
	}

	if (ISVALUE(base_reloc_unknown))
	  ProcStateSetTag(state,ARMINSREGB(ins),base_reloc);
	ProcStateSetReg(state,ARMINSREGB(ins),dest);
      }
    }
  /*}}} */
}
/*}}}*/

/* Emulate multiple stores {{{*/
static void ArmInsEmulMultipleStore(t_ins * ins, t_procstate * state)
{
  t_reloc * base_reloc;
  register_content tmp_base_address, tmp_offset, dest;
  t_uint32 base_address,offset;
  t_lattice_level base_address_unknown;
  t_lattice_level offset_unknown;
  t_lattice_level base_reloc_unknown;
  t_instruction_flags insflags = ARMINSFLAGS(ins);
  t_uint32 num_regs = 0;
  int i;
  
  for (i=0; i<16; i++)
    if (ARMINSIMMEDIATE(ins) & (1 << i)) 
      num_regs++;
  
  /* get the base address and reloc from the state */
  
  base_reloc_unknown = ProcStateGetTag(state,ARMINSREGB(ins),&base_reloc);
  base_address_unknown = ProcStateGetReg(state,ARMINSREGB(ins),&tmp_base_address);
  base_address = G_T_UINT32(tmp_base_address.i);
  
  /* get the offset from the state */

  if (ArmInsHasImmediate(ins))
    {
      offset_unknown = VALUE;
      GenericSet_uint32(tmp_offset.i,ARMINSIMMEDIATE(ins));
    }
  else if (ArmInsHasShiftedFlexible(ins))
    {
      offset_unknown = ArmInsExtractShift(ins,state,&tmp_offset);
    }
  else
    offset_unknown = ProcStateGetReg(state,ARMINSREGC(ins),&tmp_offset);
  
  offset = G_T_UINT32(tmp_offset.i);

  
  if (ISBOT(base_address_unknown)) 
    {
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
      ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
      return;
    }

  
  if (insflags & FL_WRITEBACK)
    {	/* writeback enabled */
      if ((!ISVALUE(base_address_unknown)) || (!ISVALUE(offset_unknown)))
	{
	  ProcStateSetRegBot(state,ARMINSREGB(ins));
	} 
      else 
	{
	  if (insflags & FL_DIRUP)
	    GenericSet_uint32(dest.i,base_address + 4 * num_regs);
	  else 
	    GenericSet_uint32(dest.i,base_address - 4 * num_regs);
	  if (ISVALUE(base_reloc_unknown))
	    {
	      ProcStateSetTag(state,ARMINSREGB(ins),base_reloc);
	    }
	  ProcStateSetReg(state,ARMINSREGB(ins),dest);
	}
    }
}
/*}}}*/

/* Emulate stores {{{*/
static void ArmInsEmulStore(t_ins * ins, t_procstate * state)
{
  t_reloc * base_reloc;
  register_content tmp_base_address, tmp_offset, dest;
  t_uint32 base_address,offset;
  t_lattice_level base_address_unknown;
  t_lattice_level offset_unknown;
  t_lattice_level base_reloc_unknown;
  t_instruction_flags insflags = ARMINSFLAGS(ins);

  /* get the base address and reloc from the state */
  
  base_reloc_unknown = ProcStateGetTag(state,ARMINSREGB(ins),&base_reloc);
  base_address_unknown = ProcStateGetReg(state,ARMINSREGB(ins),&tmp_base_address);
  base_address = G_T_UINT32(tmp_base_address.i);
  
  /* get the offset from the state {{{ */

  if (ArmInsHasImmediate(ins))
    {
      offset_unknown = VALUE;
      GenericSet_uint32(tmp_offset.i,ARMINSIMMEDIATE(ins));
    }
  else if (ArmInsHasShiftedFlexible(ins))
    {
      offset_unknown = ArmInsExtractShift(ins,state,&tmp_offset);
    }
  else
    offset_unknown = ProcStateGetReg(state,ARMINSREGC(ins),&tmp_offset);
  /* }}} */
  
  offset = G_T_UINT32(tmp_offset.i);
  
  if (ISBOT(base_address_unknown))
    {
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
      return;
    }

  bzero(&dest,sizeof(register_content));
  
  if (!(insflags & FL_PREINDEX) || (insflags & FL_WRITEBACK)) 
    {	/* writeback enabled */
      if ((!ISVALUE(base_address_unknown)) || (!ISVALUE(offset_unknown)))
	{
	  ProcStateSetRegBot(state,ARMINSREGB(ins));
	} 
      else 
	{
	  if (insflags & FL_DIRUP)
	    GenericSet_uint32(dest.i,base_address + offset);
	  else 
	    GenericSet_uint32(dest.i,base_address - offset);
	  if (ISVALUE(base_reloc_unknown))
	    ProcStateSetTag(state,ARMINSREGB(ins),base_reloc);
	  ProcStateSetReg(state,ARMINSREGB(ins),dest);
	}
    }
}/*}}}*/

/* }}} */

/*!
 * \todo Document and move to constant propagator
 *
 * \param reg
 * \param state_a
 * \param state_b
 *
 * \return t_bool 
*/
/* Move me to constant propagator {{{ */
t_bool RegIsEqualAndKnown(t_reg reg, t_procstate * state_a, t_procstate * state_b)
{
  register_content value1;
  register_content value2;

  if (!ISVALUE(ProcStateGetReg(state_a,reg, &value1)))
    return FALSE;

  if (!ISVALUE(ProcStateGetReg(state_b,reg, &value2)))
    return FALSE;

  if (GENERIC_EQ(value1.i,value2.i))
    return TRUE;

  return FALSE;
}
/* }}} */
/*!
 * \todo Document and move to constant propagator
 *
 * \param state_a
 * \param state_b
 *
 * \return t_bool 
*/
/*  Move me to constant propagator {{{ */
t_bool ConditionsAreEqualAndKnown(t_procstate * state_a, t_procstate * state_b)
{
  t_bool value_a1;
  t_bool value_a2;
  t_bool value_a3;
  t_bool value_a4;

  t_bool value_b1;
  t_bool value_b2;
  t_bool value_b3;
  t_bool value_b4;

  if (ProcStateGetCond(state_a,ARM_REG_N_CONDITION,&value_a1))
    return FALSE;
  if (ProcStateGetCond(state_b,ARM_REG_N_CONDITION,&value_b1))
    return FALSE;
  if (value_a1!=value_b1)
    return FALSE;

  if (ProcStateGetCond(state_a,ARM_REG_V_CONDITION,&value_a2))
    return FALSE;
  if (ProcStateGetCond(state_b,ARM_REG_V_CONDITION,&value_b2))
    return FALSE;
  if (value_a2!=value_b2)
    return FALSE;

  if (ProcStateGetCond(state_a,ARM_REG_Z_CONDITION,&value_a3))
    return FALSE;
  if (ProcStateGetCond(state_b,ARM_REG_Z_CONDITION,&value_b3))
    return FALSE;
  if (value_a3!=value_b3)
    return FALSE;
  
  if (ProcStateGetCond(state_a,ARM_REG_C_CONDITION,&value_a4))
    return FALSE;
  if (ProcStateGetCond(state_b,ARM_REG_C_CONDITION,&value_b4))
    return FALSE;
  if (value_a4!=value_b4)
    return FALSE;

 return TRUE;
}

/*}}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param state
 *
 * \return void 
*/
/* The instruction emulator that ignores conditions {{{ */
void ArmInsEmulNoConditions(t_ins * ins, t_procstate * state)
{
  t_regset mask = NullRegs;
  /* excluding the PC is a hack as long as address producers depend on the PC */

  switch (ARMINSOPCODE(ins))
    {
      case ARM_PSEUDO_CALL:
	FATAL(("Pseudo call in emulation!"));
	break;
    case ARM_CONSTANT_PRODUCER:
      ArmInsEmulConstProd(ins, state);
      break;
    case ARM_ADDRESS_PRODUCER:
      ArmInsEmulAddrProd(ins, state);
      break;
    case ARM_UMULL:
    case ARM_MUL:
    case ARM_MLA:
    case ARM_AND:
    case ARM_TST:
    case ARM_BIC:
    case ARM_EOR:
    case ARM_ORR:
    case ARM_TEQ:
    case ARM_MOV:
    case ARM_MVN:
    case ARM_CMN:
    case ARM_CMP:
    case ARM_ADD:
    case ARM_ADC:
    case ARM_SUB:
    case ARM_SBC:
    case ARM_RSC:
    case ARM_RSB:
      ArmInsEmulNumericalAndLogical(ins, state);
      break;
    case ARM_LDR:
    case ARM_LDRB:
    case ARM_LDRSH:
    case ARM_LDRH:
    case ARM_LDRSB:
    case ARM_LDM:
      ArmInsEmulLoad(ins,state);
      break;
    case ARM_STR:
    case ARM_STRB:
    case ARM_STRH:
      ArmInsEmulStore(ins,state);
      break;
    case ARM_STM:
      ArmInsEmulMultipleStore(ins,state);
      break;
    case ARM_B:
    case ARM_BL:
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
      break;
    case ARM_SWI:
    case ARM_MRS:
    case ARM_MSR:
    case ARM_UMLAL:
    case ARM_SMULL:
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
      if (ARMINSFLAGS(ins) & FL_S)
      {
	RegsetAddReg(mask,ARM_REG_N_CONDITION);
	RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	RegsetAddReg(mask,ARM_REG_C_CONDITION);
	RegsetAddReg(mask,ARM_REG_Q_CONDITION);
	RegsetAddReg(mask,ARM_REG_V_CONDITION);
	ProcStateSetCondsetBot(state,mask);
      }
      ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
      break;

    default:
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
      if (ARMINSFLAGS(ins) & FL_S)
        {
          RegsetAddReg(mask,ARM_REG_N_CONDITION);
          RegsetAddReg(mask,ARM_REG_Z_CONDITION);
          RegsetAddReg(mask,ARM_REG_C_CONDITION);
          RegsetAddReg(mask,ARM_REG_Q_CONDITION);
          RegsetAddReg(mask,ARM_REG_V_CONDITION);
          ProcStateSetCondsetBot(state,mask);
        }
      ProcStateSetTagsetBot(state,INS_REGS_DEF(ins));
      break;
    }
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param state
 *
 * \return void 
*/
/* The complete instruction emulator {{{ */
void ArmInstructionEmulator(t_ins *  ins, t_procstate * state)
{
  t_bool execute, flag1,flag2;
  t_bool uncertain = FALSE;
  t_procstate * orig_state = NULL;

  /* step one: if the predicate cannot hold, the instruction will not
     be executed */

  switch(ARMINSCONDITION(ins))
    {
    case ARM_CONDITION_EQ :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_Z_CONDITION,&execute)))
	if (!execute) 
	  goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_NE :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_Z_CONDITION,&execute)))
	if (execute) 
	  goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_CS :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_C_CONDITION,&execute)))
	if (!execute) 
	  goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_CC :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_C_CONDITION,&execute)))
	if (execute) 
	  goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_MI :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_N_CONDITION,&execute)))
	if (!execute) 
	  goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_PL :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_N_CONDITION,&execute)))
	if (execute) 
	  goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_VS : 
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_V_CONDITION,&execute)))
	if (!execute) 
	  goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_VC :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_V_CONDITION,&execute)))
	if (execute) 
	  goto nothing_to_be_done; 
      break;

    case ARM_CONDITION_HI :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_C_CONDITION,&execute)))
	if (!execute) 
	  goto nothing_to_be_done; 
      if (!(uncertain = (uncertain || ProcStateGetCond(state,ARM_REG_Z_CONDITION,&execute))))
	if (execute) 
	  goto nothing_to_be_done; 
      break;

    case ARM_CONDITION_LS :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_C_CONDITION,&execute)))
	if (execute) 
	  if (!(uncertain = (uncertain || ProcStateGetCond(state,ARM_REG_Z_CONDITION,&execute))))
	    if (!execute) 
	      goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_GE :
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_N_CONDITION,&flag1)))
	if (!(uncertain = (uncertain || ProcStateGetCond(state,ARM_REG_V_CONDITION,&flag2))))
	  if (flag1 != flag2)
	    goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_LT : 
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_N_CONDITION,&flag1)))
	if (!(uncertain = (uncertain || ProcStateGetCond(state,ARM_REG_V_CONDITION,&flag2))))
	  if (flag1 == flag2)
	    goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_GT : 
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_Z_CONDITION,&execute)))
	if (execute) 
	  goto nothing_to_be_done;
      if (!(uncertain = (uncertain || ProcStateGetCond(state,ARM_REG_N_CONDITION,&flag1))))
	if (!(uncertain = (uncertain || ProcStateGetCond(state,ARM_REG_V_CONDITION,&flag2))))
	  if (flag1 != flag2)
	    goto nothing_to_be_done; 
      break;
      
    case ARM_CONDITION_LE : 
      if (!(uncertain = ProcStateGetCond(state,ARM_REG_Z_CONDITION,&execute)))
	if (!execute) 
	  if (!(uncertain = (uncertain || ProcStateGetCond(state,ARM_REG_N_CONDITION,&flag1))))
	    if (!(uncertain = (uncertain || ProcStateGetCond(state,ARM_REG_V_CONDITION,&flag2))))
	      if (flag1 == flag2)
		  goto nothing_to_be_done; 
      break;
    case ARM_CONDITION_AL:
      INS_ATTRIB(ins) &= ~IF_ALWAYS_EXECED;
      break;
    case ARM_CONDITION_NV : 
      goto nothing_to_be_done; 
    }

  /* set instruction flags */
  
  /* an instruction can be executed */

  INS_ATTRIB(ins) |= IF_EXECED;

  if (uncertain)
    {
      /* but it will not always be executed */
      INS_ATTRIB(ins) &= ~IF_ALWAYS_EXECED;
      orig_state = ProcStateNew();
      ProcStateDup(orig_state,state);
    }
  
  /* actual emulation */

  ArmInsEmulNoConditions(ins,state);
   
  if (uncertain) /* We don't know what happened, join both the new and the prev state*/
    {

      ProcStateJoin(state, orig_state , AllRegs(), NullRegs);
      ProcStateFree(orig_state);
      orig_state = NULL;
    }

  return;

 nothing_to_be_done:

  /* it might be that an instruction is never executed in one context,
     while it is executed in another one */

  INS_ATTRIB(ins) &= ~IF_ALWAYS_EXECED;
  
  return;
}
/* }}} */

/* vim: set shiftwidth=2 foldmethod=marker: */
