/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_flowgraph.h"
#include "arm_instructions.h"
#include "arm_registers.h"
#include "arm_description.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../kernel/diablo_symbol.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_addresses.h"
#include "../../kernel/diablo_iterators.h"
#include "../../diablo_options.h"
#include <string.h>
#ifdef HAVE_LIBGEN
#include <libgen.h>
#endif
/* }}} */
/* Options {{{ */
#define ARMFG_REMOVE_PC24RELOCS
#define ARMFG_OPTIMIZE_WEAK_CALLS
/*#define ARMFG_ASSUME_SWI_ADS*/
#define KEEP_INFO
/* }}} */
/* Static function declarations {{{ */
static t_bbl * find_close_bbl_by_address(t_bbl * bbl_start, t_address addr);
static t_ins * ArmFindConditionalInstructionThatDefinesRegisterInBbl(t_ins * start, t_regset* reg, t_uint32 * offset,t_arm_condition_code cn);
static t_ins * ArmFindInstructionThatDefinesRegisterInBbl(t_ins * start, t_regset* reg, t_uint32 * offset);
static void CountLoadAddress(t_address address);
/* }}} */

/*!
 * Removes the relocs from branches, as they are not strictly necessary and
 * tend to complicate the rest of the flowgraph construction
 *
 * \param obj A disassembled object
 *
 * \return t_uint32 the number of relocs removed
*/
/* ArmRemovePC24Relocs {{{ */
t_uint32 ArmRemovePC24Relocs(t_object * obj)
{
  t_reloc * rel, *tmp;
  t_uint32 ret=0;
  OBJECT_FOREACH_RELOC_SAFE(obj,rel,tmp)
    if (REL_FROM(rel)->type == RT_INS)
      if (INS_TYPE(REL_FROM(rel)) == IT_BRANCH)
      {
	RelocTableRemoveReloc(obj->reloc_table,rel);
	ret++;
      }
  return ret;
}
/* }}} */
/*!
 * Probably the most important function in the backend: Create a flowgraph from
 * a list of disassembled arm instructions. Works in 3 big steps:
 * - Does leader detection on the list of instructions to identify basic blocks
 * - Converts all position-dependent instructions to pseudo instructions
 * - Draws edges between the basic blocks
 *
 * \todo Needs only one parameter
 * 
 * \param obj
 * \param section The (disassembled) section to flowgraph
 *
 * \return void 
*/
/* ArmFlowgraphSection {{{ */
void ArmFlowgraphSection(t_object * obj, t_section * section) 
{
  t_uint32 ret;

#ifdef ARMFG_REMOVE_PC24RELOCS
  /* Remove the PC24 relocs as they aren't strictly necessary and cause a 
   * great deal of trouble afterwards */
  ANALYSIS_START("Removing PC24 relocs from branches");
  ret=ArmRemovePC24Relocs(obj);
  ANALYSIS_END("Removing PC24 relocs from branches",("%d relocs removed",ret)); 
#endif

#ifdef ARMFG_OPTIMIZE_WEAK_CALLS
  /* Remove the calls created by weak symbols */
  ANALYSIS_START("Patch calls to weak symbols"); 
  ret=ArmPatchCallsToWeakSymbols(obj,section);
  ANALYSIS_END("Patch calls to weak symbols",("%d calls patched",ret)); 
#endif

  /* Find the leaders in the instruction list */
  ANALYSIS_START("Leader detection");
  ret=ArmFindBBLLeaders(obj,section);
  ANALYSIS_END("Leader detection",("%d leaders",ret));

  /* Create the basic blocks (platform independent) */
  ANALYSIS_START("Creating Basic Blocks");
  ret=CfgCreateBasicBlocks(section);
  ANALYSIS_END("Creating Basic Blocks",("%d blocks",ret));

  ANALYSIS_START("Detecting switch tables");
  ArmOptimizeSwitches(section);
  ANALYSIS_SIMPLE_END("Detecting switch tables");
  
  
  /* Create the edges between basic blocks */
  ANALYSIS_START("Creating Basic Block graph");
  ret=ArmAddBasicBlockEdges(section);
  ANALYSIS_END("Creating Basic Block graph",("%d edges added",ret));
}
/* }}} */
/*!
 * When we find an address to function we normally add a hell edge to that
 * function, because we do not know where the address will be used. For switch
 * tables we know where the addresses are used. That is why we detect these
 * tables first and avoid that hell edges are created.
 *
 * \todo Needs only one parameter
 * 
 * \param obj
 * \param sec The section we are creating the flowgraph for, with basic blocks
 * created (but without edges)
 *
 * \return void 
 */
/* ArmOptimizeSwitches {{{ */
void ArmOptimizeSwitches(t_section * sec)
{
  t_uint32 tel;
  t_cfg * cfg=SECTION_CFG(sec);

  t_bbl * bbl, * databbl, *dest, * bbl2;
  t_ins * ins;
  t_ins * data;
  
  CFG_FOREACH_BBL_SAFE(cfg,bbl,bbl2)
  {
    if ((BBL_INS_LAST(bbl)&&(INS_ATTRIB(BBL_INS_LAST(bbl))&IF_SWITCHJUMP)))
    {
      if ((ARMINSOPCODE(BBL_INS_LAST(bbl))==ARM_LDR) 
	  && (ARMINSREGA(BBL_INS_LAST(bbl))==ARM_REG_R15) 
	  && (ARMINSREGB(BBL_INS_LAST(bbl))==ARM_REG_R15))
      {
	/*DiabloPrint(stdout,"@I\n",BBL_INS_LAST(bbl));*/
	if (ARMINSSHIFTLENGTH(BBL_INS_LAST(bbl))!=2) FATAL(("UStrange switch %d!"));
	if (ARMINSSHIFTTYPE(BBL_INS_LAST(bbl))!=ARM_SHIFT_TYPE_LSL_IMM) FATAL(("UStrange switch %d!"));
	BBL_FOREACH_INS_R(bbl,ins)
	{
	  if (ins==BBL_INS_LAST(bbl)) continue;

	  if (ARMINSREGC(ins)==ARMINSREGC(BBL_INS_LAST(bbl))) FATAL(("Strange switch!"));
	  if ((ARMINSOPCODE(ins)==ARM_CMP)&&(ARMINSREGB(ins)==ARMINSREGC(BBL_INS_LAST(bbl)))) {  break; }
	}

	if (!ins) FATAL(("No switch check found for switch!"));
	if (ARMINSREGC(ins)!=ARM_REG_NONE) FATAL(("No valid switch compare!"));
	/*  printf("Valid values : %d\n",ARMINSIMMEDIATE(ins));*/

	databbl=BBL_NEXT(bbl);
	if (!databbl) FATAL(("Databbl not found!"));
	databbl=BBL_NEXT(databbl);
	if (!databbl) FATAL(("Databbl not found!"));

	/*	DiabloPrint(stdout,"@iB\n",databbl);*/

	if (!GENERIC_EQ(GENERIC_ADD_UINT32(INS_CADDRESS(BBL_INS_LAST(bbl)),8),BBL_CADDRESS(databbl)))
	  FATAL(("Datablock <-> Switch mismatch"));

	if (BBL_NINS(databbl)<ARMINSIMMEDIATE(ins))
	  FATAL(("Found only part of switch table!"));

	for (tel=0, data=BBL_INS_FIRST(databbl); tel<=ARMINSIMMEDIATE(ins); tel++, data=INS_NEXT(data))
	{
	  t_reloc_ref * ref=RELOCATABLE_REFERS_TO(data);

	  if (!ref) FATAL(("NO REFS!"));
	  /*DiabloPrint(stdout,"Looking for address %x\n",ARMINSIMMEDIATE(data));*/
	  dest=find_close_bbl_by_address(databbl, ARMINSIMMEDIATE(data));
	  if (T_RELOCATABLE(dest)!=REL_TO_RELOCATABLE(ref->rel)) FATAL(("Relocations corrupt @I <> @R!",data,ref->rel));
	  if (!dest) FATAL(("Destination bbl for switch not found!"));

	  ref->rel->hell=FALSE;

	  {
	    t_cfg_edge * edge = CfgEdgeNew(cfg,bbl,dest,ET_SWITCH);
	    edge->switchvalue=tel;
	    edge->rel = ref->rel;
	  }

	}
      }
      else if ((ARMINSOPCODE(BBL_INS_LAST(bbl))==ARM_ADD) 
	  && (ARMINSREGA(BBL_INS_LAST(bbl))==ARM_REG_R15) 
	  && (ARMINSREGB(BBL_INS_LAST(bbl))==ARM_REG_R15))
      {
	t_arm_ins * prev=T_ARM_INS(INS_PREV(BBL_INS_LAST(bbl)));

	while(prev)
	{
	  if (ARMINSREGA(prev)==ARMINSREGC(BBL_INS_LAST(bbl))) break;
	  if ((ARMINSOPCODE(prev)==ARM_CMP)
	      &&(ARMINSCONDITION(BBL_INS_LAST(bbl))!=ARM_CONDITION_AL)
	      &&((ARMINSREGB(prev)==ARMINSREGC(BBL_INS_LAST(bbl)))
		||(ARMINSREGC(prev)==ARMINSREGC(BBL_INS_LAST(bbl))) )) break;
	  prev=T_ARM_INS(INS_PREV(prev));
	}

	if (!prev) FATAL(("No definition found!"));


	if (  (ARMINSOPCODE(prev)==ARM_CMP)
	    &&(ARMINSSHIFTTYPE(BBL_INS_LAST(bbl))==ARM_SHIFT_TYPE_LSL_IMM)
	    &&(ARMINSSHIFTLENGTH(BBL_INS_LAST(bbl))==2)
	    &&(ARMINSCONDITION(BBL_INS_LAST(bbl))==ARM_CONDITION_LS)
	    &&(ARMINSFLAGS(prev)&FL_IMMED)
	    )
	{

	}
	else if ((ARMINSOPCODE(prev)==ARM_RSB)
	         &&(ARMINSFLAGS(prev)&FL_IMMED)
	         &&(ARMINSSHIFTTYPE(BBL_INS_LAST(bbl))==ARM_SHIFT_TYPE_LSL_IMM)
	         &&(ARMINSSHIFTLENGTH(BBL_INS_LAST(bbl))==2)
	         &&(ARMINSCONDITION(BBL_INS_LAST(bbl))==ARM_CONDITION_AL))
	{
	  int i;
	  int j=0;
	  t_ins * block_end = SecGetInsByAddress(sec,INS_CADDRESS(BBL_INS_LAST(bbl))); 
	  
	  if (!block_end) FATAL(("no block end"));
	  if (!INS_NEXT(block_end)) FATAL(("no next ins"));
	  block_end=INS_NEXT(block_end);
	  i=ARMINSIMMEDIATE(prev)+1;
	  while (i--)
	  {
	    t_bbl * tmp;
	    t_bbl * jumpdest=INS_BBL(INS_NEXT(block_end)); 
	    t_ins * first=INS_COPY(INS_NEXT(block_end));
	    t_ins * iter;

	    /* Black magic block splitting thingie (don't say we didn't warn you --- may cause serious brain damage)  {{{ */
	    if (!(INS_ATTRIB(INS_NEXT(block_end)) & IF_BBL_LEADER))  
	    {
	      t_cfg_edge * edge;
	      /* Allocate a new block, gets appended to the list of basic blocks */
	      tmp = BblNew(cfg);
	      /* Remove it from the list of basic blocks */
	      GRAPH_NODE_LAST(cfg) = T_NODE(BBL_PREV(tmp));
	      BBL_NEXT(GRAPH_NODE_LAST(cfg))=NULL;
	      
	      /* Insert it after the block we want to split */
	      BBL_PREV(tmp) = jumpdest;
	      BBL_NEXT(tmp) = BBL_NEXT(jumpdest);
	      if (BBL_NEXT(tmp)) BBL_PREV(BBL_NEXT(tmp)) = tmp;
	      BBL_NEXT(jumpdest) = tmp;
	      /* Set the address */
	      BBL_OLD_ADDRESS(tmp)=INS_OLD_ADDRESS(INS_NEXT(block_end));
	      BBL_CADDRESS(tmp)   =INS_CADDRESS(INS_NEXT(block_end));

	      /* Remove the instruction and all its successors from the
	       * original basic block and move the list of instructions to the
	       * new basic block  */
	      INS_NEXT(INS_PREV(first))=NULL;
	      BBL_INS_LAST(tmp) = BBL_INS_LAST(jumpdest);
	      BBL_INS_LAST(jumpdest) = INS_PREV(first);
	      INS_PREV(first)=NULL;
	      BBL_INS_FIRST(tmp) =first;


	      /* Set the bbl field in the moved instructions and update the
	       * instruction count */
	      BBL_FOREACH_INS(tmp,iter)
	      {
		INS_BBL(iter) = tmp;
		BBL_NINS(tmp)++;
		BBL_NINS(jumpdest)--;
	      }

	      /* Do the same for the original instructions */ 
	      iter=INS_NEXT(block_end);
	      while(!(INS_ATTRIB(iter) & IF_BBL_LEADER))
	      {
		INS_BBL(iter) = tmp;
		iter=INS_NEXT(iter);
	      }

	      INS_ATTRIB(INS_NEXT(block_end))|=IF_BBL_LEADER;

	      edge=CfgEdgeNew(cfg,bbl,tmp,ET_SWITCH);
	      edge->switchvalue=j;
	      j++;
	    }
	    else
	    {
	      FATAL(("Implement block was already split....\n"));
	    }
	    /* }}} */

	    
	    block_end=INS_NEXT(block_end);
	  }

	 /* FATAL(("Implement adds @I\n",BBL_INS_LAST(bbl)));*/
	}
	else FATAL(("Implement adds @I\n",BBL_INS_LAST(bbl)));
      }
      
      else FATAL(("Unknown switch jump statement @I!",BBL_INS_LAST(bbl)));
    }
  }

}
/* }}} */
/*!
 * Add edges between the newly created basic blocks. Both the lineair (array)
 * representation of the instructions and the basic block representation are
 * available: you can access the (copy) basic block by using INS_BBL of a
 * leader in the lineair representation.
 *
 * Although connecting the blocks isn't that hard, it could be difficult to
 * find out if a given jump instructions is a call or just a jump (without
 * return). Getting this completly safe is probably not what you want: if you
 * implement this completly safe, you'll probably end up with one big
 * function, which will degrade the performance of the rest of diablo or hell
 * nodes after each jump which will degrade the performance even further.
 *
 * On the other hand, if you get this wrong (e.g add a jump edge where there
 * should have been a call edge), then chances are that your flowgraph (and
 * thus the produced binary) is wrong (the successor block of the call block
 * will have no incoming (return) edge, and thus this block is dead). 
 *
 * The solution is to look at code that is produced by your compiler(s).
 * It'll probably have only a few possibilties to produce calls. If you
 * handle them as safe as possible, than you're off the hook. 
 *
 *
 * \todo Needs only one parameter
 * 
 * \param obj
 * \param section The disassembled section, with basic blocks already created
 * we want to flowgraph
 *
 * \return t_uint32 
*/
/* ArmAddBasicBlockEdges {{{ */
t_uint32 ArmAddBasicBlockEdges(t_section * section) 
{
  t_ins *ins, * iter, *stop, * block_end, * block_start, * jump_target;
  t_bbl * block, * ft_block;
  t_bbl * bbbl;
  t_cfg_edge * edge, * j_edge;
  t_cfg * flowgraph=SECTION_CFG(section);
  t_bool flag;
  t_reloc_ref * ref;
  t_uint32 nedges=0;

  /* walk through the original(!) array (we need offsets) and add 
   * all possible successors to every node */

  SECTION_FOREACH_INS(section,ins)
  {
    /* we're only interested in the first and last instructions of a basic block */

    if (!(INS_ATTRIB(ins) & IF_BBL_LEADER)) 
      continue;

    block_start = ins;
    block=INS_BBL(block_start);

    block_end = T_INS(&(T_ARM_INS(ins)[block->nins - 1])); /* Warning: This is not the same as the end of the bbl, since the end is a copy of the original end */

    ft_block = INS_NEXT(block_end)?INS_BBL(INS_NEXT(block_end)):NULL;  /* the FallThrough basic block */

    /* 1. ADD EDGES FOR RELOCATIONS: 
     * 
     * each relocation gets it's own edge. These edges will be automatically
     * deleted when the relocation is removed. There's no need to add an edge
     * for the obvious relocations (such as the relocations that are present
     * for the calculation of the offsets in intermodular jumps), since they
     * are completly described by the edge itself (the reason for the edge is
     * that there is a jump, not that there is a relocation) 
     *
     * {{{ */
    
    for (ref=RELOCATABLE_REFED_BY(block); ref!=NULL; ref=ref->next)
    {

      /* Skip the obvious relocations and reloctions to data */

      if (ref->rel->hell && (INS_TYPE(block_start)!=IT_DATA))
      {
	/* Some extra checks if the relocation is from an instruction */
	if (REL_FROM(ref->rel)->type==RT_INS)
	{
	  t_cfg_edge * found=NULL;
	  t_cfg_edge * i_edge;
	  if (RELOCATABLE_REFERS_TO(REL_FROM(ref->rel))->next)
	  {
	    DiabloPrint(stdout,"1. @R\n",RELOCATABLE_REFERS_TO(REL_FROM(ref->rel))->rel);
	    DiabloPrint(stdout,"2. @R\n",RELOCATABLE_REFERS_TO(REL_FROM(ref->rel))->next->rel);
	    FATAL(("Multiple relocs from instruction?"));
	  }
	  if ((RELOCATABLE_REFERS_TO(REL_FROM(ref->rel))->rel)!=ref->rel) FATAL(("Relocation wrong"));


	  BBL_FOREACH_PRED_EDGE(block,i_edge)
	  {
	    if ((CFG_EDGE_HEAD(i_edge)==flowgraph->hell_node) && (EDGE_CAT(i_edge) == ET_CALL))
	    {
	      found=i_edge;
	      break;
	    }
	  }

	  if (found)
	  {
	    ref->rel->edge=found;
	    found->refcount++;
	    /* also increment the refcount of the corresponding return edge!!! */
	    VERIFY_FASSERT(found->corr_edge,("Call edge @E does not have a corresponding edge!",found));
	    found->corr_edge->refcount++;
	  }
	  else
	  {
	    ref->rel->edge=CfgEdgeNewCall(flowgraph, flowgraph->hell_node, block,NULL,NULL);
	    nedges++;
	  }
	  /* Some comments on this */
	  VERBOSE(1,("Adding INS arrow from hell! (relocation @R, @B)",ref->rel,block));
	}
	else
	{
	  t_cfg_edge * found=NULL;
	  t_cfg_edge * i_edge;
	  /* Add the edge */
	  BBL_FOREACH_PRED_EDGE(block,i_edge)
	  {
	    if ((CFG_EDGE_HEAD(i_edge)==flowgraph->hell_node) && (EDGE_CAT(i_edge) == ET_CALL))
	    {
	      found=i_edge;
	      break;
	    }
	  }

	  if (found)
	  {
	    ref->rel->edge=found;
	    found->refcount++;
	    /* also increment the refcount of the corresponding return edge!!! */
	    VERIFY_FASSERT(found->corr_edge,("Call edge @E does not have a corresponding edge!",found));
	    found->corr_edge->refcount++;
	  }
	  else
	  {

	    ref->rel->edge=CfgEdgeNewCall(flowgraph, flowgraph->hell_node, block,NULL,NULL);
	    nedges++;
	  }
	  VERBOSE(1,("Adding arrow from hell! (relocation @R, @B)",ref->rel,block));
	}
      }
    }
    /* }}} */

    /* 2. ADD NORMAL EDGES */
    /* now determine the successors of the block and their types */

    switch (INS_TYPE(block_end)) 
    {
      /* Branches are straight forward.... */
      case IT_BRANCH:
	{
	  /* If the branch is conditional, add a fall-through edge */
	  if ((ft_block != NULL)&&(!IS_DATABBL(ft_block))&& (ARMINSCONDITION(block_end) != ARM_CONDITION_AL))
	  {
	    CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
	    nedges++;
	  }

	  /* Add arrows for the jump, dependent on the type of jump: BL adds
	   * calls (unless return is really impossible), B adds jumps */

	  switch (ARMINSOPCODE(block_end)) 
	  {
	    case ARM_B:
	      /* get the instruction we're jumping to */
	      jump_target = (t_ins *) (T_ARM_INS(block_end) + (ARMINSIMMEDIATE(block_end) + 8)/4);
	      CfgEdgeNew(flowgraph, block, INS_BBL(jump_target),ET_JUMP);
	      nedges++;
	      break;

	    case ARM_BL:
	      /* get the instruction we're jumping to */
	      jump_target = (t_ins *) (T_ARM_INS(block_end) + (ARMINSIMMEDIATE(block_end) + 8)/4);
	      if ((ft_block != NULL)&&(!IS_DATABBL(ft_block)))
	      {
		CfgEdgeNew(flowgraph, block,ft_block, ET_FUNLINK);
		CfgEdgeNewCall(flowgraph, block,INS_BBL(jump_target),ft_block,NULL);
		nedges++;
	      }
	      else
	      {
		CfgEdgeNew(flowgraph, block, INS_BBL(jump_target),ET_JUMP);
		nedges++;
	      }
	      break;

	    case ARM_BX:
	      if (ARMINSREGB(block_end)==ARM_REG_R14)
	      {
		ArmInstructionMakeMov(block_end,ARM_REG_R15,ARM_REG_R14,0,ARMINSCONDITION(block_end));


		CfgEdgeNew(flowgraph,  block,  flowgraph->exit_hell_node, ET_RETURN);
		nedges++;
	      }
	      else
	      {
		FATAL(("We should not find BX's since thumb is not yet supported"));
	      }
	      break;
	    case ARM_BLX:
	      CfgEdgeNew(flowgraph,  block, ft_block, ET_FUNLINK);
	      nedges++;
	      CfgEdgeNewCall(flowgraph,  block, flowgraph->hell_node,ft_block,NULL);
	      nedges++;
	      break;
	    default:
	      break;
	  }
	  break;
	}

      /* SWI's go to SWI-hell */
      case IT_SWI:
#ifdef ARMFG_ASSUME_SWI_ADS
	if (ft_block != NULL)
	{
	  CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
	  nedges++;
	}
#else
	/* If the SWI is conditional, we add a fall-throug edge */
	if ((ft_block != NULL)&&(!IS_DATABBL(ft_block))&& (ARMINSCONDITION(block_end) != ARM_CONDITION_AL))
	{
	  CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
	  nedges++;
	}
	/* link call and return site of the SWI */
	if (ft_block != NULL)
	{
	  CfgEdgeNew(flowgraph, block,ft_block, ET_SWILINK);
	  nedges++;
	}
	else
	  FATAL(("Could not place SWI link arrow because the end of the section was reached!"));
	/* since we don't know where a SWI jumps to, we'll add hell as a successor */
	CfgEdgeNew(flowgraph, block,flowgraph->swi_hell_node,ET_SWI);
	nedges++;
#endif
	break;

      /* Dataprocessing instructions can be used as jumps on architectures with explicit program counters */
      case IT_DATAPROC:
	if (ARMINSREGA(block_end) == ARM_REG_R15)
	{
	  if ((ft_block != NULL)&&(!IS_DATABBL(ft_block))&& (ARMINSCONDITION(block_end) != ARM_CONDITION_AL))
	  {
	    CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
	    nedges++;
	  }
	  
	  /* A dataprocessing instruction that define R15 (= the Program
	   * counter) => control flow altering instruction.  there are
	   * different possibilities: 
	   *
	   * 1) MOV pc,r14 is (probably) a function return, although THIS IS
	   * NOT SAFE (but a safe enough assumption)
	   *  
	   * 2) Any other operation is either a jump (or even a switch) or a
	   * function call depending on whether r14 holds the return address or
	   * not.
	   */

	  if ((ARMINSOPCODE(block_end) == ARM_MOV) && (ARMINSREGC(block_end) == ARM_REG_R14) && (ARMINSREGB(block_end) == ARM_REG_NONE) && (ARMINSIMMEDIATE(block_end)==0))
	  {
	    /* MOV pc,r14 */
	    /* function return */
	    /*DiabloPrint(stdout,"Return edge at @I\n",block_end);*/
	    CfgEdgeNew(flowgraph,  block,  flowgraph->exit_hell_node, ET_RETURN);
	    nedges++;
	  }
	  else
	  {
	    t_regset search=INS_REGS_USE(block_end);
	    RegsetSubReg(search,ARM_REG_R15);
	    RegsetSubReg(search,16);

	    stop = INS_PREV(block_start);
	    /* Switch detection! */
	    if ((ARMINSOPCODE(block_end)==ARM_ADD)&&(ARMINSSHIFTTYPE(block_end)==ARM_SHIFT_TYPE_LSL_IMM)&&(ARMINSCONDITION(block_end)==ARM_CONDITION_LS))
	    {
	      for (iter = block_end; iter != stop; iter = INS_PREV(iter))
	      {
		if (RegsetIsSubset(search ,INS_REGS_USE(iter)) && (ARMINSOPCODE(iter)==ARM_CMP))
		  break;
	      }
	      
	      if (iter==stop) FATAL(("Bound not found! Switch approx at @I",iter));

	      /* Iter contains the instruction that calculates the bounds check. We will now add switch edges for every switch */

	      CFG_FOREACH_BBL(flowgraph,bbbl)
	      {
		if ((G_T_UINT32(BBL_CADDRESS(bbbl))>G_T_UINT32(BBL_CADDRESS(INS_BBL(ins)))+INS_BBL(ins)->nins*4)
		    && (G_T_UINT32(BBL_CADDRESS(bbbl))<=G_T_UINT32(BBL_CADDRESS(INS_BBL(ins)))+(1+ARMINSIMMEDIATE(iter))*4+INS_BBL(ins)->nins*4))
		{
		  CfgEdgeNew(flowgraph, block,  bbbl,ET_SWITCH)->switchvalue=G_T_UINT32(GENERIC_SUB(BBL_CADDRESS(bbbl),INS_CADDRESS(block_end)))/4 - 2;
		  nedges++;
/*		  CfgEdgeNew(flowgraph, BBL_PREV(bbbl),bbbl,ET_FALLTHROUGH);*/
/*		  nedges++;*/
		}
	      }
	    }
	    else if (!(INS_ATTRIB(block_end)&IF_SWITCHJUMP))/* It's not a switch */
	    {
	      /* Find the instruction that defines R14 (= the return register)  */
	      for (iter = block_end; iter != stop; iter = INS_PREV(iter))
	      {
		if (RegsetIn(INS_REGS_DEF(iter),ARM_REG_R14))
		  break;
	      }

	      if (iter != stop) /* We found the definition of R14 */
	      {
		/* if iter == INS_PREV(block_end) and *iter = MOV pc, r14 => function call
		 * if iter != INS_PREV(block_end) and *iter = ADD pc, constant => function call
		 * if iter defines r14 independently from the pc => jump
		 * else FATAL!
		 */
		if (!RegsetIn(INS_REGS_USE(iter) , ARM_REG_R15)) 
		{
		  /* This is something really odd: r14 is defined, which indicates we're looking at an indirect function
		   * call, but the value of r14 isn't derived from the program counter, implicating that 
		   * function return will not bring us to the next instruction...
		   * Usually we want to come down with a FATAL in this situation but there is one exception:
		   * the indirect call could be an optimised tail call, in which case the definition of r14
		   * should have been a load from the stack.
		   */

		  if (((ARMINSOPCODE(iter) == ARM_LDR) || (ARMINSOPCODE(iter) == ARM_LDM)) && (ARMINSREGB(iter) == ARM_REG_R13))
		  {
		    /* tail call. should be viewed as a jump (will become interprocedural later on) */
		    /* the destination is the hell node, because we can't know the destination until after constant propagation */
		    CfgEdgeNew(flowgraph, block, flowgraph->hell_node, ET_JUMP);
		    nedges++;
		  }
		  else
		  {
		    FATAL(("Check if this happens! This would mean we would return to somewhere else.... Quite dangerous I think.\n Instruction: @I",block_end));
		    /* jump */
		    /* the destination is the hell node, because we can't know the destination until after constant propagation */
		    CfgEdgeNew(flowgraph, block, flowgraph->hell_node, ET_JUMP);
		    nedges++;
		  }
		}
		else if (((t_ins *) iter == INS_PREV(block_end)) && ((ARMINSOPCODE(iter) == ARM_MOV) && (ARMINSREGC(iter) == ARM_REG_R15)))
		{
		  /* function call */
		  if (!IS_DATABBL(ft_block))
		    /* In case code follows this instruction, we can assume it's a funtion call */
		  {
		    CfgEdgeNew(flowgraph,  block, ft_block, ET_FUNLINK);
		    CfgEdgeNewCall(flowgraph,  block, flowgraph->hell_node,ft_block,NULL);
		    nedges++;
		    nedges++;
		  }
		  else
		    /* otherwise it's certainly not a function call */
		  {
		    CfgEdgeNew(flowgraph, block, flowgraph->hell_node, ET_JUMP);
		    nedges++;
		  }
		}
		else if (((ARMINSOPCODE(iter) == ARM_ADD) && (ARMINSREGB(iter) == ARM_REG_R15) && (ARMINSFLAGS(iter) & FL_IMMED)) &&  ((t_ins *) &(T_ARM_INS(iter)[(ARMINSIMMEDIATE(iter)+8)/4]) == INS_NEXT(block_end)))
		{
		  /* yep, it's a function call */
		  CfgEdgeNew(flowgraph,  block, ft_block, ET_FUNLINK);
		  CfgEdgeNewCall(flowgraph,  block, flowgraph->hell_node,ft_block,NULL);
		  nedges++;
		  nedges++;
		}
		else
		{
		  FATAL(("Encountered something that looks like an indirect function call but the link register is wrong! (block %x)",G_T_UINT32(BBL_CADDRESS(INS_BBL(block_end)))));
		}
	      }
	      else /* We did not find the definition for R14 */
	      {

		/* no r14-defining instruction found within the block, this
		 * must be a jump. THIS IS NOT SAFE! WE COULD HAVE MISSED IT!
		 * But we assume a compiler does not produce this kind of code
		 * if it wants to implement a call.  The destination is the
		 * hell node, because we can't know the destination until after
		 * constant propagation */

		CfgEdgeNew(flowgraph, block, flowgraph->hell_node, ET_JUMP);
		nedges++;
	      }
	    }
	  }

	}
	else
	{
	  /* this instruction does NOT alter control flow, it just happens to be the end of a bbl */
	  if (ft_block != NULL) 
	  {
	    CfgEdgeNew(flowgraph, block, ft_block, ET_FALLTHROUGH);
	    nedges++;
	  }
	}
	break;

      /* Loads can be used as jump instructions on architectures with explicit program counters */
      case IT_LOAD:

	if (ARMINSREGA(block_end) == ARM_REG_R15)
	{
	  /* control flow altering instruction.  there are different
	   * possibilities: LDR pc,r13,... is interpreted as a function return
	   * otherwise the operation is either a jump or a function call
	   * depending on whether r14 is designated the return address in the
	   * same basic block or not.
	   */
	  if ((ARMINSCONDITION(block_end) != ARM_CONDITION_AL) && (INS_NEXT(block_end) != NULL))
	  {
	    CfgEdgeNew(flowgraph, block, ft_block, ET_FALLTHROUGH);
	    nedges++;
	  }

	  if (ARMINSREGB(block_end) == ARM_REG_R13)
	  {
	    /* function return */
	    CfgEdgeNew(flowgraph, block, flowgraph->exit_hell_node, ET_RETURN);
	    nedges++;
	  }
	  else if (RegsetIn(INS_REGS_USE(block_end) , ARM_REG_R15)&&(!RegsetIsMutualExclusive(INS_REGS_USE(block_end) , ARM_ALL_BUT_PC_AND_COND) ))
	  {
	    if ((ARMINSOPCODE(block_end) == ARM_LDR)&&(ARMINSSHIFTTYPE(block_end)==ARM_SHIFT_TYPE_LSL_IMM)&&(ARMINSCONDITION(block_end)==ARM_CONDITION_LS))
	    {
	      /* This is a switch of the form:
	       *	ldrls pc,[pc,rBase lsl 2]
	       *	b <fallthrough case>
	       * <$d>
	       *	<address of case 0>
	       *	<address of case 1>
	       *	...
	       *
	       * We take a conservative approach to this kind of switch:
	       * 	- add a jump edge to the hell node --- No longer, do switch detection
	       * 	- because the true code for the different cases has relocations 
	       * 	  pointing to it from the data block, incoming hell edges to
	       * 	  this code will automatically be added
	       * 	- in the previous step, a relocation was added to block_end that 
	       * 	  points to the data block, in order not to lose this block during
	       * 	  dead data removal
	       * 	- during layout, the switch table block is automatically linked to the
	       * 	  fallthrough case block, so the switch construct stays together in the
	       * 	  final layout
	       */
	    }
	    else
	    {
	      FATAL(("Should not get here @I",block_end));
	    }
	  }
	  else
	  {
	    stop = INS_PREV(block_start);
	    for (iter = block_end; iter != stop; iter = INS_PREV(iter))
	    {
	      if (RegsetIn(INS_REGS_DEF(iter) , 14))
		break;
	    }

	    if (iter != stop)
	    {
	      /* if iter == INS_PREV(block_end) and *iter = MOV pc, r14 => function call
	       * if iter != INS_PREV(block_end) and *iter = ADD pc, constant => function call
	       * if iter defines r14 independently from the pc => jump
	       * else FATAL!
	       */
	      if (!RegsetIn(INS_REGS_USE(iter), 15))
	      {
		FATAL(("Check if this happens! This would mean we would return to somewhere else.... Quite dangerous I think."));
		/* jump */
		/* the destination is the hell node, because we can't know the destination until after constant propagation */
		CfgEdgeNew(flowgraph, block, flowgraph->hell_node, ET_JUMP);
		nedges++;
	      }
	      else
	      {
		if (((t_ins *) iter == INS_PREV(block_end)) && ((ARMINSOPCODE(iter) == ARM_MOV) && (ARMINSREGC(iter) == ARM_REG_R15)))
		{
		  /* function call */
		  CfgEdgeNew(flowgraph, block,ft_block, ET_FUNLINK);
		  CfgEdgeNewCall(flowgraph, block,flowgraph->hell_node,ft_block,NULL);
		  nedges++;
		  nedges++;
		}
		else if (((ARMINSOPCODE(iter) == ARM_ADD) && (ARMINSREGB(iter) == ARM_REG_R15) && (ARMINSFLAGS(iter) & FL_IMMED)) && ((t_ins *) &(T_ARM_INS(iter)[(ARMINSIMMEDIATE(iter)+8)/4]) == INS_NEXT(block_end)))
		{
		  /* yep, it's a function call */
		  CfgEdgeNew(flowgraph, block, ft_block, ET_FUNLINK);
		  CfgEdgeNewCall(flowgraph, block, flowgraph->hell_node,ft_block,NULL);
		  nedges++;
		  nedges++;
		}
		else
		{
		  FATAL(("Encountered an indirect function call but the link register is wrong! (block %x)",G_T_UINT32(BBL_CADDRESS(INS_BBL(block_end)))));
		}
	      }			
	    }
	    else
	    {
	      /* no r14-defining instruction found within the block, this must be a jump */
	      /* the destination is the hell node, because we can't know the destination until after constant propagation */
	      CfgEdgeNew(flowgraph, block,  flowgraph->hell_node, ET_JUMP);
	      nedges++;
	    }
	  }

	}
	else
	{
	  /* this instruction does NOT alter control flow, it just happens to be the end of a bbl */
	  if (ft_block != NULL) 
	  { 
	    CfgEdgeNew(flowgraph, block, ft_block, ET_FALLTHROUGH);
	    nedges++;
	  }
	}
	break;

      case IT_LOAD_MULTIPLE:
      case IT_STORE_MULTIPLE:

	if ((ARMINSIMMEDIATE(block_end) & (1 << ARM_REG_R15)) && (ARMINSOPCODE(block_end) == ARM_LDM))
	{
	  /* control flow altering instruction.
	   * there are different possibilities: 
	   *   LDR r13,{...,r15} is interpreted as a function return
	   *   LDR r11,{...,r11,r13,r15} is interpreted as a function return as well, as this restore the fp, sp and pc using the fp 
	   *   any other operation is either a jump or a function call depending on whether r14 is designated the return 
	   *   address in the same basic block or not.
	   */
	  if ((ARMINSCONDITION(block_end) != ARM_CONDITION_AL) && (INS_NEXT(block_end) != NULL))
	  {
	    CfgEdgeNew(flowgraph,  block, ft_block, ET_FALLTHROUGH);
	    nedges++;
	  }

	  if (ARMINSREGB(block_end) == ARM_REG_R13)
	  {
	    /* function return */
	    CfgEdgeNew(flowgraph,  block,  flowgraph->exit_hell_node, ET_RETURN);
	    nedges++;
	  }
	  else if (ARMINSREGB(block_end) == ARM_REG_R11 && (ARMINSIMMEDIATE(block_end) & (1 << ARM_REG_R13)) && (ARMINSIMMEDIATE(block_end) & (1 << ARM_REG_R11)))
	  {
	    /* function return */
	    CfgEdgeNew(flowgraph,  block,  flowgraph->exit_hell_node, ET_RETURN);
	    nedges++;
	  }
	  else
	  {
	    stop = INS_PREV(block_start);
	    for (iter = block_end; iter != stop; iter = INS_PREV(iter))
	    {
	      if (RegsetIn(INS_REGS_DEF(iter) , 14))
		break;
	    }

	    if (iter != stop)
	    {
	      /* if iter == INS_PREV(block_end) and *iter = MOV pc, r14 => function call
	       * if iter != INS_PREV(block_end) and *iter = ADD pc, constant => function call
	       * if iter defines r14 independently from the pc => jump
	       * else FATAL!
	       */
	      if (!RegsetIn(INS_REGS_USE(iter) , 15))
	      {
		FATAL(("Check if this happens! This would mean we would return to somewhere else.... Quite dangerous I think."));
		/* jump */
		/* the destination is the hell node, because we can't know the destination until after constant propagation */
		CfgEdgeNew(flowgraph,  block,  flowgraph->hell_node, ET_JUMP);
		nedges++;
	      }
	      else if (((t_ins *) iter == INS_PREV(block_end))&&((ARMINSOPCODE(iter) == ARM_MOV) && (ARMINSREGC(iter) == ARM_REG_R15)))
	      {
		/* function call */
		CfgEdgeNew(flowgraph,  block, ft_block, ET_FUNLINK);
		CfgEdgeNewCall(flowgraph,  block, flowgraph->hell_node,ft_block,NULL);
		nedges++;
		nedges++;
	      }
	      else if (((ARMINSOPCODE(iter) == ARM_ADD) && (ARMINSREGB(iter) == ARM_REG_R15) && (ARMINSFLAGS(iter) & FL_IMMED))&&((t_ins *) &(T_ARM_INS(iter)[(ARMINSIMMEDIATE(iter)+8)/4]) == INS_NEXT(block_end)))
	      {
		/* yep, it's a function call */
		CfgEdgeNew(flowgraph,  block, ft_block, ET_FUNLINK);
		CfgEdgeNewCall(flowgraph,  block, flowgraph->hell_node,ft_block,NULL);
		nedges++;
		nedges++;
	      }
	      else
	      {
		  FATAL(("Encountered something that looks like an indirect function call but the link register is wrong! (block %x)",G_T_UINT32(BBL_CADDRESS(INS_BBL(block_end)))));
	      }
	    }
	    else
	    {
	      /* no r14-defining instruction found within the block, this must be a jump */
	      /* the destination is the hell node, because we can't know the destination until after constant propagation */
	      CfgEdgeNew(flowgraph,  block,  flowgraph->hell_node, ET_JUMP);
	      nedges++;
	    }
	  }

	}
	else
	{
	  /* this instruction does NOT alter control flow, it just happens to be the end of a bbl */
	  if (ft_block != NULL)
	  {
	    CfgEdgeNew(flowgraph,  block, ft_block, ET_FALLTHROUGH);
	    nedges++;
	  }
	}
	break;

      /* Ignore data */
      case IT_DATA:
	break;

      default:
	/* instructions that end up here do not alter the control flow, they just happen to be */
	/* the end of a basic block */
	if (RegsetIn(INS_REGS_DEF(block_end) , 15)) 
	{
	  /* consistency check */
	  FATAL(("pc-defining instruction @I not recognized as such!",block_end));
	}

	if (INS_NEXT(block_end) != NULL)
	{
	  CfgEdgeNew(flowgraph,  block, ft_block, ET_FALLTHROUGH);
	  nedges++;
	}
    }
  }



  CFG_FOREACH_BBL(flowgraph,block)
  {
    flag = FALSE;
    BBL_FOREACH_PRED_EDGE(block,edge)
      if (EDGE_CAT(edge)==ET_SWITCH)
      {
	flag = TRUE;
	break;
      }

    if (flag)
    {
      BBL_FOREACH_SUCC_EDGE(block,j_edge)
	if (EDGE_CAT(j_edge)==ET_JUMP)
	{
	  CFG_EDGE_FLAGS(j_edge)|=EF_FROM_SWITCH_TABLE;
	}
    }
    else
    {
      BBL_FOREACH_SUCC_EDGE(block,j_edge)
	if (EDGE_CAT(j_edge)==ET_JUMP)
	{
	  CFG_EDGE_FLAGS(j_edge)&=~EF_FROM_SWITCH_TABLE;
	}
    }
  }

  return nedges;
}
/* }}} */
/*!
 * Basic leader-detection algorithm. 
 * ---------------------------------
 *
 * See e.g. the Red Dragon Book ("Compilers: Principles, Techniques and
 * Tools", by Alfred V. Aho, Ravi Sethi, and Jeffrey D. Ullman) for more
 * details. 
 *
 * The aim of this function is to identify all start-addresses of basic
 * blocks (and mark them). There are four reasons to mark an instructions as
 * a basic block leader:
 * 
 * 1. The instruction is the target of a direct or indirect jump. This is not
 * as easy as it seems, especially when the program counter is explicit:
 * every instruction that changes the pc can be considered as a jump
 * instruction.
 *
 * 2. It's the successor of a control-flow altering instruction (if a jump is
 * conditional, than the next instruction will be the first instruction of the
 * fall-through block, else it is either target of an other jump or a dead
 * block. In the latter case, it will be removed by the unreachable block
 * removal routines, so we just mark it as a basic block leader).
 *
 * 3. There is an address produced to this basic block, this is either 
 *
 *    a. done directly, using the program counter (we assume no real
 *    code-address calculations are present (so not address of function + x) 
 *
 *    b. using a relocation (and not necessary detected when scanning the
 *    instructions) 
 *
 * For both cases we need to assume there will be an indirect jump to this
 * address.
 *
 * 4. The start of data blocks in code or instructions following datablocks
 * are considered basic block leaders.
 *
 *
 * This function will find all basic block leaders that are not part of a
 * switch statement. More switch checking will be done in optimize switches.
 * 
 * \todo needs only one parameter
 *
 * \param obj
 * \param code
 *
 * \return t_uint32 
 */
/* ArmFindBBLLeaders {{{ */
t_uint32 ArmFindBBLLeaders(t_object * obj, t_section * code) 
{
  /* We can still assume that all instructions are stored sequentially in the
   * array Ains the last instruction stored can be recognised with INSNEXT ==
   * NULL */

  t_ins * ins;
  t_uint32  nleaders=0;

  /* Initialisation: The first instruction of the program is always a leader.
   * The reasoning is equal to that of reason 2: Either there will be a jump to
   * this address or it's the entry point or it is dead. */

  INS_ATTRIB(code->data) |= IF_BBL_LEADER;

  SECTION_FOREACH_INS(code,ins)
  {
    t_bool pc_def_detected_and_resolved = FALSE, pc_use_detected_and_resolved = FALSE;

    /* 1. DATA (reason 4) {{{ */ 
    if (INS_TYPE(ins) == IT_DATA)
    {
      /* The start of a block of data is considered a basic block leader */
      INS_ATTRIB(ins) |= IF_BBL_LEADER; 

      /* Process all the other data */
      while (ins && INS_TYPE(ins) == IT_DATA) {  
	ins = INS_NEXT(ins); 
      } 

      /* The instruction following the data is a basic block leader */
      if (ins) INS_ATTRIB(ins) |= IF_BBL_LEADER; 
      else
	/* If there are no instructions left, we can exit the outer loop
	 * (all blocks are marked) */
	break; 
    }
    /* }}} */
    /* 2. NORMAL BRANCHES (reason 1 and 2) {{{ */
    if (INS_TYPE(ins) == IT_BRANCH)
    {
      t_ins * target;
      t_address target_address;
      pc_def_detected_and_resolved = TRUE;
      /* The instruction following the branch instruction is a basic block leader */
      if (INS_NEXT(ins) != NULL) 
	INS_ATTRIB((t_arm_ins *) INS_NEXT(ins)) |= IF_BBL_LEADER;


      /* Align address before switching to arm */

      if (ARMINSFLAGS(ins) & FL_THUMB) {
      	ARMINSIMMEDIATE(ins) = 0xfffffffc;
	pc_use_detected_and_resolved = TRUE;
      }
      
      /* The target of the branch will also be a basic block leader */
      
      target_address=GENERIC_ADD_UINT32(INS_CADDRESS(ins),8 + ARMINSIMMEDIATE(ins));
      target=SecGetInsByAddress(code,target_address);

      if (target) INS_ATTRIB(target) |= IF_BBL_LEADER; 
      else 
      {
	/* This used to be a FATAL, but that probably isn't necessary.
	 * Such a branch usually means that this piece of code never can be executed.
	 * We just issue a WARNING and change the branch into something saner: a branch
	 * to itself => if the program ever executes this instruction, it hangs instead of 
	 * dieing on a SEGFAULT */
	INS_ATTRIB(ins) |= IF_BBL_LEADER;
	INS_ATTRIB(ins) |= IF_CHANGED;
	ARMINSIMMEDIATE(ins) = -8;
      }
    }
    /* }}} */
    /* 3. CHANGES TO THE PROGRAM COUNTER WITH A DATAPROCESSING INSTRUCTION (reason 1 and 2)  {{{ */
    else if ((INS_TYPE(ins) == IT_DATAPROC) && (ARMINSREGA(ins)==ARM_REG_R15)) 
    {
      /* we don't know much, but it sure changes the program counter 
	 perhaps we should do a more thorough check on MOV with an immediate third operand? */
      pc_def_detected_and_resolved = TRUE;

      if (INS_NEXT(ins) != NULL) INS_ATTRIB((t_arm_ins *) INS_NEXT(ins)) |= IF_BBL_LEADER;

      if ((ARMINSOPCODE(ins) == ARM_ADD) &&( (ARMINSREGB(ins) ==ARM_REG_R15) || (ARMINSREGC(ins) ==ARM_REG_R15)))
      {
	INS_ATTRIB(ins)|=IF_SWITCHJUMP;
	pc_use_detected_and_resolved = TRUE;
      }
      else if ((ARMINSOPCODE(ins)==ARM_MOV) && (ARMINSREGB(ins)!=ARM_REG_R15) && (ARMINSREGC(ins)!=ARM_REG_R15))
      {
	/* Copies an address in the pc, no problem, as there is already an edge
	 * to this address (relocation or address producer) */
      }
      else if ((ARMINSOPCODE(ins)==ARM_ADD) && (ARMINSFLAGS(ins) & FL_IMMED) && (ARMINSREGB(ins)!=ARM_REG_R15))
      {
	/* Idem: Copies an address in the pc, no problem, as there is already an edge
	 * to this address (relocation or address producer) */
	if (ARMINSIMMEDIATE(ins)==0)
	{
	  ARMINSOPCODE(ins)=ARM_MOV;
	  ARMINSREGC(ins)=ARMINSREGB(ins);
	  ARMINSREGB(ins)=ARM_REG_NONE;
	  ARMINSFLAGS(ins) &=~FL_IMMED;
	}
	/* This would mean we manipulate code addresses. We assume this cannot happen */
	else
	{
	  FATAL(("Cannot handle @I\n",ins));
	}
      }
      else FATAL(("Implement @I\n!",ins));
    }
    /* }}} */
    /* 4. CREATION OF AN ADDRESS BY A DATAPROCESSING INSTRUCTION USING THE PC (reason 3) {{{ */ 
    else if ((INS_TYPE(ins) == IT_DATAPROC) && (ARMINSREGA(ins)!=ARM_REG_R15) && ((ARMINSREGB(ins) ==ARM_REG_R15) || (ARMINSREGC(ins) ==ARM_REG_R15) )) 
    {
      if ((ARMINSOPCODE(ins)==ARM_ADD) && (ARMINSFLAGS(ins) & FL_IMMED))
      {
	t_address target_address;
	t_ins * target;
	target_address=GENERIC_ADD_UINT32(INS_CADDRESS(ins),8 + ARMINSIMMEDIATE(ins));
	target=SecGetInsByAddress(code,target_address);

	if (GENERIC_LT(target_address,SECTION_GET_END_ADDRESS(code)))
	{
	  if ((!(INS_ATTRIB(target) & IF_DATA)))
	  {
#ifndef REALSAFE
	    if (ARMINSREGA(ins)!=14)
#endif
	      INS_ATTRIB(target) |= IF_BBL_LEADER; 
#ifndef REALSAFE
	    else
	      INS_ATTRIB(target) |= IF_BBL_LEADER;
#endif
	  }
	}
	else
	  FATAL(("Data processing operation produced an address that lies out of the section:\n@I",ins));
	pc_use_detected_and_resolved=TRUE;
      }
      else if ((ARMINSOPCODE(ins)==ARM_SUB) && (ARMINSFLAGS(ins) & FL_IMMED))
      {
	
	  t_address target_address;
	  t_ins * target;
	 
	  target_address=GENERIC_ADD_INT32(INS_CADDRESS(ins),8 - ARMINSIMMEDIATE(ins));
	  
	  target=SecGetInsByAddress(code,target_address);
	  if ((GENERIC_LT(target_address,SECTION_GET_END_ADDRESS(code))) && (GENERIC_GE(target_address,SECTION_GET_CADDRESS(code))))
	    {
	      if ((!(INS_ATTRIB(target) & IF_DATA)))
		{
#ifndef REALSAFE
		  if (ARMINSREGA(ins)!=14)
#endif
		    INS_ATTRIB(target) |= IF_BBL_LEADER; 
#ifndef REALSAFE
		  else
		    INS_ATTRIB(target) |= IF_BBL_LEADER;
#endif
		}
	    }
	  else
	    FATAL(("Data processing operation produced an address that lies out of the section:\n@I", ins));
	  pc_use_detected_and_resolved=TRUE;
      }
      else if ((ARMINSOPCODE(ins)==ARM_ADD)) 
      {
	  
	  /* This is a PCrelative offset + PC. Potentially dangerous (since
	   * it takes a lot of analysis to find out what is calculated
	   * here), but it it's probably a REL32 relocation + pc. We'll
	   * check and remove these further on (in the address producer
	   * construction)  */
	  
	  pc_use_detected_and_resolved=TRUE; 
      }
      else if ((ARMINSOPCODE(ins)==ARM_MOV) && (ARMINSREGB(ins) ==ARM_REG_NONE))
      {
	  
	  /* A copy operation: the instruction + 8 (bytes) could be the target of an indirect jump */
	  
	  t_ins * target=INS_NEXT(ins)?INS_NEXT(INS_NEXT(ins)):NULL;
	  if (target) 
	    INS_ATTRIB(target) |= IF_BBL_LEADER; 
	  pc_use_detected_and_resolved=TRUE;
      }
      else FATAL(("Dataprocessing instruction @I calculates something from the pc, but we do not understand what it does.",ins)); 
    }
    /* }}} */
    /* 5. LOADS USED AS JUMP INSTRUCTIONS (reason 1 and 2) {{{ */
    else if ((INS_TYPE(ins) == IT_LOAD) && (ARMINSREGA(ins) == ARM_REG_R15) && !(RegsetIn(INS_REGS_USE(ins),ARM_REG_R15)))
    {
      /* A load from memory into the pc => the next instruction is a
       * leader.  Since the loaded address is either produced by an other
       * instruction (that is thus converted to an address producer which
       * will have an associated relocation) or produced by the linker (and
       * thus must have a relocation associated) we do not need to worry about
       * the target of this type of jump. It will become a jump to hell,
       * and can be resolved if the load comes from constant memory by the
       * constant propagator. */
      
      pc_def_detected_and_resolved = TRUE;
      if (INS_NEXT(ins) != NULL) INS_ATTRIB((t_arm_ins *) INS_NEXT(ins)) |= IF_BBL_LEADER;
    }
    /* }}} */
    /* 6. MULTIPLE LOADS USED AS JUMP INSTRUCTIONS (reason 1 and 2) {{{ */
    else if ((ARMINSOPCODE(ins) == ARM_LDM) && (ARMINSIMMEDIATE(ins) & (1 << ARM_REG_R15))) 
    {
      /* This alters (loads) the pc, so the next instruction is a leader.
       * Can be a return or a stored address. See normal loads for
       * explanation. */
      pc_def_detected_and_resolved = TRUE;
      if (INS_NEXT(ins) != NULL) INS_ATTRIB((t_arm_ins *) INS_NEXT(ins)) |= IF_BBL_LEADER;
    }
    /* }}} */
    /* 7. STORES THAT CREATE ADDRESSES USING THE PC {{{ */
    else if ((ARMINSOPCODE(ins) == ARM_STM) && (ARMINSIMMEDIATE(ins) & (1 << ARM_REG_R15))) 
    {
      /* This copies the pc (puts it somewhere in memory). To be real safe
       * we should assume it can be used to return to this address + 4 or 8
       * dependent on the processor implementation.  The only problem with
       * that is that gcc produced binaries put their address on the stack
       * at the start of each procedure.... So we ignore it for now */
      
      t_ins * saved=INS_NEXT(ins)?INS_NEXT(INS_NEXT(ins)):NULL;
      pc_use_detected_and_resolved = TRUE;
      if (saved)
      {
	/* just to check if it ever really happens outside of the gcc function call code: */
	if (ARMINSREGB(ins) != ARM_REG_R13) FATAL(("STM of pc outside of the stack! Check this:\n@I",ins));
      }
    }
    /* }}} */
    /* 8. SYSTEM CALLS {{{ */
    else if (ARMINSOPCODE(ins) == ARM_SWI) 
    {
      /* Actually, we could just consider system calls as normal
       * instructions, if we were sure that they are not long-jump-like
       * system calls. Since we do not know that for sure here, we must
       * assume that is will alter control flow => next instruction is a
       * leader */
      
      pc_def_detected_and_resolved = TRUE; 
      
      if (INS_NEXT(ins) != NULL)
	INS_ATTRIB((t_arm_ins *) INS_NEXT(ins)) |= IF_BBL_LEADER; 
    }
    /* }}} */
    /* 9. SWITCH TABLES {{{ */
    else if ((ARMINSOPCODE(ins) == ARM_LDR)                                          /* We have a load */
	&&   (ARMINSREGB(ins)==ARM_REG_R15)                                          /* from the PC  */
	&&   (!RegsetIsMutualExclusive(INS_REGS_USE(ins) , ARM_ALL_BUT_PC_AND_COND)) /* + a register (no immediate) */ 
	&&   (ARMINSSHIFTTYPE(ins)==ARM_SHIFT_TYPE_LSL_IMM)                          /* the register is shifted left */
	&&   (ARMINSSHIFTLENGTH(ins)==2)                                             /* over two positions */
	&&   (ARMINSCONDITION(ins)==ARM_CONDITION_LS)                                /* and it has a fallthrough (failing bound check) */ 
	)
    {
      t_reloc * rel;

      if (INS_NEXT(ins) == NULL) FATAL(("No next ins for switch statement")); 
      if (INS_NEXT(INS_NEXT(ins)) == NULL) FATAL(("No data for switch statement")); 
      if (ARMINSOPCODE(INS_NEXT(INS_NEXT(ins)))!=ARM_DATA) FATAL(("No data (2) for switch statement"));
      
      INS_ATTRIB(ins)|=IF_SWITCHJUMP;
      INS_ATTRIB(INS_NEXT(INS_NEXT(ins)))|=IF_SWITCHTABLE;
     
      /* Add a relocation to keep the jump table alive.  The relocation is only
       * meant to keep the data block alive as long as the switch instruction
       * is live, it never needs to be relocated, so the write part of the
       * program is empty. */
      
      rel=RelocTableAddRelocToRelocatable(obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(INS_NEXT(INS_NEXT(ins))),NULL_GENERIC,FALSE,NULL,NULL,NULL,"SA+\\*\\s0000$",TRUE);
      rel->label = StringDup("Switch table reloc");

      pc_use_detected_and_resolved=TRUE;
      pc_def_detected_and_resolved=TRUE;

      /* We should indicate the next instruction (fallthrough) as a bbl leader.
       * The targets of the switch table will be relocated as they are absolute
       * values and thus marked in 12. */
      
      INS_ATTRIB(T_ARM_INS(INS_NEXT(ins))) |= IF_BBL_LEADER;
    }
    /*}}}*/
    /* 10. LOADS USING THE PC {{{ */
    else if ((INS_TYPE(ins)==IT_LOAD)&&(ARMINSREGB(ins)==ARM_REG_R15)&&(ARMINSREGA(ins)!=ARM_REG_R15))
    {
      /* We can always ignore loads from the pc, since they will be removed
       * by the const/address producer reduction. If they produce an adress,
       * then it will have an associated relocation (or it is produced in one
       * of the above cases), so the correct basic blocks will be marked. */
      pc_use_detected_and_resolved=TRUE;
    }
    /* }}} */
    /* 11. FLOATING POINT LOADS USING THE PC {{{ */
    else if ((INS_TYPE(ins)==IT_FLT_LOAD) && (ARMINSREGB(ins)==ARM_REG_R15))
    {
      pc_use_detected_and_resolved=TRUE;
    }
    /* }}} */
    /* 12. THE RELOCATIONS (reason 3) {{{ */
    if (RELOCATABLE_REFED_BY(ins)) 
      INS_ATTRIB(ins) |= IF_BBL_LEADER;
    /* }}} */

    /* check for completeness */
    if ((RegsetIn(INS_REGS_DEF(ins) , ARM_REG_R15) && !pc_def_detected_and_resolved) || (RegsetIn(INS_REGS_USE(ins) , ARM_REG_R15) && !pc_use_detected_and_resolved))
    {
      /* so there are pc-defining instructions that we haven't trapped yet */
      FATAL(("Instruction @I defines or uses pc but is not recognized as such!\n",ins));
    }
  }


  /* Now we also mark the beginning of data-blocks as bbl-leaders. We do this by walking the subsections
   * and look for RO_DATA. The beginning of mapped RO_DATA subsections are marked as bbl-leaders.
   */
  {
    t_object * tmp, * tmp2;
    t_uint32 tel;
    t_section * sec;
    t_ins * tmp_ins;


    OBJECT_FOREACH_SUBOBJECT(obj,tmp,tmp2)
      OBJECT_FOREACH_SECTION(tmp,sec,tel)
      {
	  if (!SECTION_IS_MAPPED(sec)) continue;
	  tmp_ins=SecGetInsByAddress(code,SECTION_GET_CADDRESS(sec));
	  if (tmp_ins)
	  {
	    if (GENERIC_EQ(INS_CADDRESS(tmp_ins) , SECTION_GET_CADDRESS(sec)))
	    {
	      INS_ATTRIB(tmp_ins) |= IF_BBL_LEADER; 
	    }
	  }
      }
  }

  ins = code->data; 
  while(ins)
  {

    if (INS_ATTRIB(ins)&IF_BBL_LEADER) nleaders++;
    ins=INS_NEXT(ins);
  }
  return nleaders;
}
/* }}} */
/*!
 * \todo Document
 *
 * \param bbl_start
 * \param addr
 *
 * \return t_bbl *
*/
/* find_close_bbl_by_address {{{ */
static t_bbl * find_close_bbl_by_address(t_bbl * bbl_start, t_address addr)
{
  t_bbl * bbl=bbl_start;
  t_address base;

  while(bbl)
  {
    base = BBL_CADDRESS(bbl);
    if (GENERIC_LE(base,addr))
    {
      if (GENERIC_GT(GENERIC_ADD_UINT32(base,bbl->nins*4),addr))
      {
	return bbl;
      }
      else
      {
	bbl=BBL_NEXT(bbl);
      }
    }
    else
    {
      bbl=BBL_PREV(bbl);
    }
  }

  return NULL;

}
/* }}} */
/*!
 * \brief 
 *
 * \param start
 * \param reg
 * \param offset
 * \param cn
 *
 * \return t_ins *
*/
/* ArmFindConditionalInstructionThatDefinesRegisterInBbl {{{ */
static t_ins * ArmFindConditionalInstructionThatDefinesRegisterInBbl(t_ins * start, t_regset* reg, t_uint32 * offset,t_arm_condition_code cn)
{
  t_ins * i_ins=start;
  if (offset) *offset=-1;
  while((i_ins=INS_PREV(i_ins)))
    {
       if ((cn==ARMINSCONDITION(i_ins))&&(RegsetIsSubset(INS_REGS_DEF(i_ins), *reg)))
	return i_ins;
      if (offset) (*offset)--;
    }
 
  if (offset) (*offset)=0;
  return NULL;
}
/* }}} */
/*!
 * \brief 
 *
 * \param start
 * \param reg
 * \param offset
 *
 * \return t_ins *
*/
/* ArmFindInstructionThatDefinesRegisterInBbl {{{ */
static t_ins * ArmFindInstructionThatDefinesRegisterInBbl(t_ins * start, t_regset* reg, t_uint32 * offset)
{
  t_ins * i_ins=start;
  if (offset) *offset=-1;
  while((i_ins=INS_PREV(i_ins)))
  {
    if (RegsetIsSubset(INS_REGS_DEF(i_ins), *reg))
      return i_ins;
    if (offset) (*offset)--;
  }

  if (offset) (*offset)=0;
  return NULL;
}
/* }}} */
/*!
 * Turns position-dependent instructions (i.e. instructions that use the pc)
 * into pseudo-operations that no longer depend on the program conter. On the
 * arm the pc is mostly used to produce addresses, hence the name of this
 * function.
 *
 * \param code
 *
 * \return void 
*/
/* ArmMakeAddressProducers {{{ */
t_int32 nr_addr_prod = 0;
t_int32 nr_loaders = 0;
void ArmMakeAddressProducers(t_section * code) 
{
  t_object * obj=SECTION_OBJECT(code);
  t_bbl * bbl, * bbl2, * bbl3;
  t_ins * ins, * def;
  t_ins * ins2=NULL;
  t_reloc * rel=NULL;
  t_address address, def_address, load_address; 
  t_uint32 value;
  t_symbol * sym;


  /* 1. The addition of the PC with a register. We should be able to
   * find out what the value in the register is.  Normally it will be
   * a loaded self-relative value. We must convert this since it
   * introduces a dependency between the datablock in which the
   * self-relative value is stored and the instruction that does the
   * addition to the pc 
   *
   * {{{ */


  CFG_FOREACH_BBL(T_CFG(code->data),bbl)
  {
    if (!IS_DATABBL(bbl))
    {
      address = BBL_CADDRESS(bbl);
      BBL_FOREACH_INS(bbl,ins)
      {
	address=GENERIC_ADD_UINT32(address,4);

	if ((ARMINSOPCODE(ins)==ARM_ADD) &&  (! (ARMINSFLAGS(ins) & FL_IMMED)) && (INS_TYPE(ins) == IT_DATAPROC) && ((ARMINSREGB(ins) == ARM_REG_R15) || (ARMINSREGC(ins) == ARM_REG_R15) ))
	{
	  t_uint32 disp;
	  t_regset search=INS_REGS_USE(ins);
	  
	  RegsetSubReg(search,15);
	  RegsetSubReg(search,16);

	  /* The cases that we can handle: 
	   *
	   * a. Addition of a value to the pc: ADD pc, pc, reg
	   * if we find the definition of reg
	   * b. Switch statements: ADDLS pc,pc, reg << 8
	   * safe to ignore these, they are handled earlier on */

	  if ((ARMINSOPCODE(ins) == ARM_ADD)&&(ARMINSSHIFTTYPE(ins)==ARM_SHIFT_TYPE_LSL_IMM)) continue; /* Skip switches */
	  if (ARMINSOPCODE(ins) != ARM_ADD) FATAL(("Strange... expected and ADD pc, pc, reg or ADD pc, reg, pc"));

	  VERBOSE(1,("Handling         @G: @I",address,ins));

	  /* Find the definition of the register that is added to the pc */

	  if (!INS_IS_CONDITIONAL(ins)) {
	    def=ArmFindInstructionThatDefinesRegisterInBbl(ins,&search,&disp);
	  }
	  else 
	  {
	    RegsetSubReg(search,ARM_REG_Q_CONDITION);
	    RegsetSubReg(search,ARM_REG_V_CONDITION);
	    RegsetSubReg(search,ARM_REG_Z_CONDITION);
	    RegsetSubReg(search,ARM_REG_N_CONDITION);
	    RegsetSubReg(search,ARM_REG_C_CONDITION);
	    def=ArmFindConditionalInstructionThatDefinesRegisterInBbl(ins,&search,&disp,ARMINSCONDITION(ins));
	  }

	  /* Calculate the address of the definition */

	  def_address=GENERIC_ADD_UINT32(address,4*(disp-1));


	  /* We must find the definition or we cannot handle this */

	  if (!def) FATAL(("We did not find the definition of the used register. This is experimental, but essential code: mail this binary to bdebus at frodo.elis.rug.ac.be"));

	  VERBOSE(1,("Definition  =    @G: @I",def_address, def));

	  /* Check if it is a load, if not we cannot handle this */

	  if (ARMINSOPCODE(def)!=ARM_LDR) FATAL(("At @G: @I: PC-relative value is not loaded (no LDR, the instruction for the definition is @I). This is in experimental, but essential code: mail this binary along with the error message to bdebus at frodo.elis.rug.ac.be",address, ins, def));


	  /* Calculate the address from where we are loading */

	  if (!(ARMINSFLAGS(def) & FL_DIRUP))
	  {
	    load_address=GENERIC_SUB_UINT32(def_address,ARMINSIMMEDIATE(def)); /* Calculate the load address + 8 (pc=pc+8) */
	    load_address=GENERIC_ADD_UINT32(load_address,8); /* Calculate the load address + 8 (pc=pc+8) */
	  }
	  else
	  {
	    load_address=GENERIC_ADD_UINT32(def_address,ARMINSIMMEDIATE(def)+8); /* Calculate the load address + 8 (pc=pc+8) */
	  }

	  CountLoadAddress(load_address);
	  nr_addr_prod+=2;/* We add 2 because we have a load and a dataproc instruction to produce the address */
	  nr_loaders++;


	  /* Perform  the load, and get the relocation at that address */

	  bbl2=find_close_bbl_by_address(bbl,load_address);
	  if (!bbl2) FATAL(("BBL NOT FOUND!"));
	  else 
	  {
	    BBL_FOREACH_INS(bbl2,ins2)
	    {
	      if (GENERIC_EQ(INS_CADDRESS(ins2),load_address)) break;
	    }
	    if (!ins2) FATAL(("No ins found!"));

	  if (INS_TYPE(ins2)!=IT_DATA) FATAL(("Ins is not data!"));
	  if (RELOCATABLE_REFERS_TO(ins2)) rel=RELOCATABLE_REFERS_TO(ins2)->rel;
	  else rel=NULL;
	  value=ARMINSIMMEDIATE(ins2);
	  }

	  if (rel) VERBOSE(1,("Loaded reloc =    @R",rel));

	  /* If we do not have a relocation, than we have troubles: 
	   * this is probably weird assembler code
	   * we cannot handle */

	  if (!rel) FATAL(("The relocation (for a PC-relative load) was not found.  This is experimental, but essential code: mail this binary to bdebus at frodo.elis.rug.ac.be"));

	  /* If it is not a self-relative value, we cannot
	     handle it */

	  if (!PatternMatch("*P*",rel->code))
	    FATAL(("Relocation used for PC-relative displacement is not SREL32! This is experimental, but essential code: mail this binary to bdebus at frodo.elis.rug.ac.be"));

	  /* Replace the load with a no-op. We cannot delete
	     it since this would change addresses, and we
	     still need them! */

	  ArmInstructionMakeNoop(def);	
#ifdef KEEP_INFO
	  T_ARM_INS(def)->info=ArmAddrInfoNew();
	  T_ARM_INS(def)->info->original_type='L';
	  T_ARM_INS(def)->info->load_address=load_address-G_T_UINT32(INS_CADDRESS(def))+G_T_UINT32(INS_OLD_ADDRESS(def));
	  T_ARM_INS(def)->info->load_value=value;
#endif

	  /* No lets find out what the value is we want to
	   * produce: There are a number of possibilities:
	   * a. an ABS_SYMBOL. This value should never be
	   * relocated, so we can change it into a constant
	   * producer b. Otherwise its an address, so we
	   * should convert it to an address producer. There
	   * are still 2 posibilities: 1. The symbol is weak,
	   * than it has no real value. In that case we point
	   * to the value we're loading from (self-relativity)
	   * 2. Otherwise we produce the symbols address */

	  if   (!((REL_TO_RELOCATABLE(rel)==T_RELOCATABLE(abs_section)) || (REL_TO_RELOCATABLE(rel)==T_RELOCATABLE(undef_section))) )  /* b */
	  {
	    /* replace the ADD with produce pointer instruction*/
	    
	    if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(rel))==RT_INS)
	    {
	      FATAL(("Relocation still to instruction!"));
	    }
	    else if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(rel))==RT_SUBSECTION && SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(rel))==code)
	    {
	      FATAL(("Relocation still to code section"));
	    }
	    else if (REL_TO_RELOCATABLE(rel) == NULL) 
	    {
	      FATAL(("Implement 2"));
	      /*
	      CFG_FOREACH_BBL(T_CFG(code->data),bbl2)
	      {
		t_address base;
		base = BBL_CADDRESS(bbl2);
		if ((GENERIC_LE(base,addr))&&(GENERIC_GT(GENERIC_ADD_UINT32(base,bbl2->nins*4),addr)))
		{ 
		  break; 
		}
	      }

	      if (!bbl2) FATAL(("BBL not found!"));
	      sym = SymbolTableAddSymbolToBBL(obj->symbol_table, LOCAL_SYMBOL, T_RELOCATABLE(bbl2), TRUE);
	      rel=RelocTableAddReloc(obj->reloc_table, ins, NULL_GENERIC,sym,REPRESENT_VALUE, ADDRESS_PROC,RELOC_FROM_INSTRUCTION,rel->edge);
	      */
	    }
	    else if ((RELOCATABLE_TYPE(REL_TO_RELOCATABLE(rel))==RT_BBL) || (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(rel))==RT_SUBSECTION) || (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(rel))==RT_SECTION))
	    {
	      t_string code;

	      rel->addend=GENERIC_SUB(rel->addend,GENERIC_ADD_UINT32(GENERIC_SUB(load_address,INS_CADDRESS(ins)),-8));
	      if ((strncmp(rel->code,"SP-A+\\",6)==0) || (strncmp(rel->code,"SA+P-\\",6)==0))
	      {
		code="SA+\\l*w\\s0000$";
	      }
	      else if ((strncmp(rel->code,"SP-A+SM +\\",6)==0))
	      {
		code="SA+SM +\\l*w\\s0000$";
	      }
	      else 
	      {
		FATAL(("CODE: %s\n",rel->code));
		exit(0); /* Keeps the compiler happy */
	      }

	      rel=RelocTableAddRelocToRelocatable(obj->reloc_table, rel->addend, T_RELOCATABLE(ins), NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset, rel->hell,rel->edge,NULL,NULL,code,rel->first);
	    }
	    else
	    {
	      printf("%p %p\n",REL_TO_RELOCATABLE(rel),abs_section);
	      FATAL(("Implement @R %s",rel,rel->code));
	    }


	    ArmMakeAddressProducer(__FILE__,__LINE__,ins, value + G_T_UINT32(address) +4, rel);

#ifdef KEEP_INFO
	    T_ARM_INS(ins)->info=ArmAddrInfoNew();
	    T_ARM_INS(ins)->info->original_type='A';
#endif
	    /* we change the type of the relocation (it pointed to a code
	     * address, or at least it should: we catch this later on, so we
	     * change it to a relocation to a bbl (either data or code)) */

	    /* Find what we are pointing to, this should normaly be a bbl */
#if 0
	    if (rel->to->type != WEAK_UNDEFINED_SYMBOL) /* b2 */
	    {
	      CFG_FOREACH_BBL(T_CFG(code->data),bbl2)
	      {
		t_address base;
		base = BBL_CADDRESS(bbl2);
		if ((GENERIC_LE(base,rel->to->current_address))&&(GENERIC_GT(GENERIC_ADD_UINT32(base,bbl2->nins*4),rel->to->current_address)))
		{ 
		  rel->to->base = T_RELOCATABLE(bbl2); 
		  break; 
		}
	      }
	    } 
	    else /* b1 */
	    {
	      /* symbol is weak undef, and so we must make relocation point 
	       * to load address(load_address), TODO : check this */
	      CFG_FOREACH_BBL(T_CFG(code->data),bbl2)
	      {
		t_address base;
		base = BBL_CADDRESS(bbl2);
		if ((GENERIC_LE(base,load_address))&&(GENERIC_GT(GENERIC_ADD_UINT32(base,bbl2->nins*4),load_address)))
		{ 
		  rel->to->base = T_RELOCATABLE(bbl2); 
		  break; 
		}
	      }
	    }
	    if (!rel->to->base) FATAL(("Not to bbl, to = %d f%d!",rel->to->type,REL_TO_RELOCATABLE(rel)->type)); /* Catch the case where we cannot find a bbl */
#endif
	  } 
	  else  /* a */
	  { 
	    if ((REL_TO_RELOCATABLE(rel)) && (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(rel))==RT_SECTION) && (REL_TO_RELOCATABLE(rel)==T_RELOCATABLE(abs_section)))
            {
              /* case of ABS_SYMBOL */
              /* replace the ADD with an constant producer instruction */
              ArmMakeConstantProducer(ins,value + G_T_UINT32(address) + 4);
	      ARMINSFLAGS(ins) |= FL_ORIG_CONST_PROD;
#ifdef KEEP_INFO
	    T_ARM_INS(ins)->info=ArmAddrInfoNew();
	    T_ARM_INS(ins)->info->original_type='A';
#endif

              VERBOSE(1,("At @G  no reloc (ABS_SYMBOL) (immed=%d,loading from @G) (%x)!",address, ARMINSIMMEDIATE(ins), load_address, value)); 
            }
	    else if (strncmp(rel->code,"PP-A+\\",6)==0 || strncmp(rel->code,"PP-A+PM +\\",10)==0)
	    {
	      t_bbl * found=NULL;
	      DiabloPrint(stdout,"@G = @G = @G - @G\n",rel->addend,GENERIC_ADD_UINT32(GENERIC_SUB(load_address,INS_CADDRESS(ins)),-8),load_address,INS_CADDRESS(ins));

	      /* symbol is weak undef, and so we must make relocation point 
	       * to load address(load_address), TODO : check this */
	      CFG_FOREACH_BBL(T_CFG(code->data),bbl2)
	      {
		t_address base;
		base = BBL_CADDRESS(bbl2);
		if ((GENERIC_LE(base,load_address))&&(GENERIC_GT(GENERIC_ADD_UINT32(base,bbl2->nins*4),load_address)))
		{ 
		  found = T_BBL(bbl2); 
		  break; 
		}
	      }

	      if (!found) FATAL(("Implement"));
	     


	      rel=RelocTableAddRelocToRelocatable(obj->reloc_table, rel->addend,T_RELOCATABLE(ins), NULL_GENERIC, T_RELOCATABLE(found),GENERIC_SUB(load_address,BBL_CADDRESS(bbl)), rel->hell,rel->edge,NULL,NULL,"S\\l*w\\s0000$",rel->first);
	      rel->label = StringDup("Weak undefined symbol");
	      ArmMakeAddressProducer(__FILE__,__LINE__,ins, value + G_T_UINT32(address) +4, rel);
#ifdef KEEP_INFO
	    T_ARM_INS(ins)->info=ArmAddrInfoNew();
	    T_ARM_INS(ins)->info->original_type='A';
#endif

	    }
	    else
	      FATAL(("AY @R,code:%s\n",rel,rel->code));
	  }
	}
      }
    }
  }

  /* end of 1. }}} */

  /* 2. Load's relative to the pc.  This is very similar to the previous case.
   * No need to find any definition. We just need to remove it, since it
   * introduces dependencies between the pc and the address of the loaded value
   * 
   * {{{ */

  CFG_FOREACH_BBL(T_CFG(code->data),bbl)
  {
    if (IS_DATABBL(bbl)) continue; 
    address = BBL_CADDRESS(bbl);
    BBL_FOREACH_INS(bbl,ins) /* Check all instructions in the block */
    {

      if ((INS_TYPE(ins) == IT_LOAD) && (ARMINSREGB(ins) == ARM_REG_R15)  /* It has to be a load from the pc, that either  */


	  && ((RegsetIsMutualExclusive(INS_REGS_USE(ins) , ARM_ALL_BUT_PC_AND_COND)) || /* a. uses no other registers that the pc */ 
	    (!((ARMINSOPCODE(ins) == ARM_LDR)&&(ARMINSSHIFTTYPE(ins)==ARM_SHIFT_TYPE_LSL_IMM))) )) /*b. is not a switch */  
      {

	/* We should have a normal load or we cannot handle it */

	if (ARMINSOPCODE(ins)!=ARM_LDR && ARMINSOPCODE(ins)!=ARM_LDRH) FATAL(("PC-relative value is not loaded (no LDR, but @I). This is in experimental, but essential code: mail this binary along with the error message to bdebus at frodo.elis.rug.ac.be",ins));

	/* Calculate the address form which we are loading */

	if (!(ARMINSFLAGS(ins) & FL_DIRUP))
	{
	  load_address=GENERIC_SUB_UINT32(address,ARMINSIMMEDIATE(ins)); /* Calculate the load address + 8 (pc=pc+8) */
	  load_address=GENERIC_ADD_UINT32(load_address,8); /* Calculate the load address + 8 (pc=pc+8) */
	}
	else
	{
	  load_address=GENERIC_ADD_UINT32(address,ARMINSIMMEDIATE(ins)+8); /* Calculate the load address + 8 (pc=pc+8) */
	}

	CountLoadAddress(load_address);
	nr_addr_prod++;nr_loaders++;

	/* Perform the load */

	  bbl2=find_close_bbl_by_address(bbl,load_address);
	  if (!bbl2) FATAL(("BBL NOT FOUND!"));
	  else 
	  {
	    t_ins * ins2;
	    BBL_FOREACH_INS(bbl2,ins2)
	    {
	      if (GENERIC_EQ(INS_CADDRESS(ins2),load_address)) break;
	    }
	    if (!ins2) FATAL(("No ins found!"));


	    if (INS_TYPE(ins2)!=IT_DATA) FATAL(("Ins is not data!\nLoad: @I\nData: @I\n",ins,ins2));
	    if (RELOCATABLE_REFERS_TO(ins2)) rel=RELOCATABLE_REFERS_TO(ins2)->rel;
	    else rel=NULL;
	    value=ARMINSIMMEDIATE(ins2);
	    if (ARMINSOPCODE(ins)==ARM_LDRH) 
	      {
		value &=0xffff;
		if (rel) FATAL(("loading relocatable halfword !!!"));
	      }
	  }
	/* Two possibilies: 
	 *
	 * a. The load is used to produce a (large) constant value: no
	 * relocation or ABS relocation. It is also possible that we load a
	 * WEAK UNDEFINED value. In this case, 0 is produced sothe load
	 * becomes a constant producer, not an address producer.  
	 *
	 * b. An address is loaded (most common case) 
	 */

#ifdef KEEP_INFO
	  T_ARM_INS(ins)->info=ArmAddrInfoNew();
	  T_ARM_INS(ins)->info->original_type='L';
	  T_ARM_INS(ins)->info->load_address=load_address-G_T_UINT32(INS_CADDRESS(ins))+G_T_UINT32(INS_OLD_ADDRESS(ins));
	  T_ARM_INS(ins)->info->load_value=value;
#endif


	if ((!rel)||(REL_TO_RELOCATABLE(rel)==T_RELOCATABLE(abs_section))) /* a */
	{
	  ArmMakeConstantProducer(ins,value);
	  ARMINSFLAGS(ins) |= FL_ORIG_CONST_PROD;
	  VERBOSE(1,("At @G no (real) reloc (immed=%d,loading from @G) (%x)!",address,ARMINSIMMEDIATE(ins), load_address, value)); 
	}
	else  /* b */
	{
	  /* We only know ABS32 relocations = the absolute value is written at this address */

	  if ((strncmp(rel->code,"SA+\\",4)!=0) && (strncmp(rel->code,"SA+g-\\",6)!=0) && (strncmp(rel->code,"SZ+A+\\",4)!=0) && (strncmp(rel->code,"SR-\\",4)!=0)
#ifdef THUMB_SUPPORT
	      && (strncmp(rel->code,"SA+SM+\\",7)!=0)
#endif
	      ) 
	    FATAL(("Did not expect this relocation for @I %s at @G: @R",ins,rel->code,load_address,rel));


	  /* If it does not already point to a bbl make it point to a bbl when neccessary*/

	  if (REL_TO_RELOCATABLE(rel)->type!=RT_BBL)
	  {
	    if (REL_TO_RELOCATABLE(rel)->type==RT_INS) 
	    {
	      FATAL(("Should not happen!"));
	    }
	    else if ((REL_TO_RELOCATABLE(rel)->type!=RT_SUBSECTION)&&(REL_TO_RELOCATABLE(rel)->type!=RT_SECTION))
	    {
	      FATAL(("Implement @R\n",rel));
	    }
	    else if (T_SECTION(REL_TO_RELOCATABLE(rel))->type == CODE_SECTION)
	    {
	      FATAL(("This should be dead code!"));
	    }
	    else
	    {
	      rel=RelocTableAddRelocToRelocatable(obj->reloc_table, rel->addend, T_RELOCATABLE(ins), NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset, rel->hell, rel->edge, NULL /*rel->corresponding*/,rel->e.sec,rel->code,rel->first);
	      rel->label=StringDup("ADDR RELATIVE LOAD");
	      ArmMakeAddressProducer(__FILE__,__LINE__,ins,value , rel);
	    }
	  }
	  else
	  {
	    rel=RelocTableAddRelocToRelocatable(obj->reloc_table, rel->addend, T_RELOCATABLE(ins), NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset, rel->hell, rel->edge, NULL /*rel->corresponding*/,rel->e.sec,rel->code,rel->first);
	    rel->label=StringDup("ADDR RELATIVE LOAD (from bbl)");
	    ArmMakeAddressProducer(__FILE__,__LINE__,ins,value , rel);
	  }
	}
      }
      else if ((ARMINSOPCODE(ins)==ARM_LDF)&&(ARMINSREGB(ins)==ARM_REG_R15))
      {
	char * copy;
	if (!(ARMINSFLAGS(ins) & FL_DIRUP))
	{
	  load_address=GENERIC_SUB_UINT32(address,ARMINSIMMEDIATE(ins)); /* Calculate the load address + 8 (pc=pc+8) */
	  load_address=GENERIC_ADD_UINT32(load_address,8); /* Calculate the load address + 8 (pc=pc+8) */
	}
	else
	{
	  load_address=GENERIC_ADD_UINT32(address,ARMINSIMMEDIATE(ins)+8); /* Calculate the load address + 8 (pc=pc+8) */
	}


	bbl2=find_close_bbl_by_address(bbl,load_address);
	if (!bbl2) FATAL(("BBL NOT FOUND!"));
	else 
	{
	  BBL_FOREACH_INS(bbl2,ins2)
	  {
	    if (GENERIC_EQ(INS_CADDRESS(ins2),load_address)) break;
	  }
	  if (!ins2) FATAL(("No ins found!"));


	  if (INS_TYPE(ins2)!=IT_DATA) FATAL(("Ins is not data!\nLoad: @I\nData: @I\n",ins,ins2));
	  value=ARMINSIMMEDIATE(ins2);
	}

	if (ARMINSFLAGS(ins)&FL_FLT_PACKED) 
	  /* 96 bits*/
	{
	  FATAL(("Implement packed floats")); 
	  exit(0); /* Keep the compiler happy */
	}
	else if (ARMINSFLAGS(ins)&FL_FLT_DOUBLE_EXTENDED) 
	  /* 96 bits */
	{
	  FATAL(("Implement double extended floats")); 
	  exit(0); /* Keep the compiler happy */
	}
	else if (ARMINSFLAGS(ins)&FL_FLT_DOUBLE) 
	  /* 64 bits */
	{
	  if(INS_TYPE(INS_NEXT(ins2)) != IT_DATA) FATAL(("Ins->next is not data!\nData: @I\n",ins2));
	  copy=Malloc(8);/*64);*/
	  memcpy(copy,&value,4);/*64);*/
	  memcpy(copy+4,&ARMINSIMMEDIATE(INS_NEXT(ins2)),4);/*64);*/
	  /*if(memcmp(data,copy,8))*/
	}
	else 
	  /* 32 bits */
	{
	  copy=Malloc(4);
	  memcpy(copy,&value,4);
	}

	ArmMakeFloatProducer(ins,copy);
      }

      address=GENERIC_ADD_UINT32(address,4);
    }
  }

  /* end of 2. }}} */

  /* 3. Add or substract constants to/from the pc. This must be removed since
   * it introduces dependencies between the reffered value and this
   * instruction. The only hard thing is finding out what we reffer to. The
   * following rules are used:
   *
   * If the calculated address is a data address, we assume it points to
   * somewher in the corresponding datablock Else, we must find out if it will
   * be used for control-flow or for loads (which is possible, when we know
   * that there will always be a constant offset added to this block.
   *
   * {{{ */

  CFG_FOREACH_BBL(T_CFG(code->data),bbl)
  {
    if (!IS_DATABBL(bbl))
    {
      address = BBL_CADDRESS(bbl);
      BBL_FOREACH_INS(bbl,ins)
      {
	if (RegsetIn(INS_REGS_USE(ins) , 15))
	{
	  t_address addr;

	  if (((ARMINSOPCODE(ins)==ARM_ADD)||(ARMINSOPCODE(ins)==ARM_SUB)) && (ARMINSFLAGS(ins) & FL_IMMED))
	  {
	    if (ARMINSOPCODE(ins)==ARM_ADD)
	      addr=GENERIC_ADD_INT32(address,ARMINSIMMEDIATE(ins)+8);
	    else
	      addr=GENERIC_SUB_INT32(address,ARMINSIMMEDIATE(ins)-8);
	    
	    VERBOSE(1,("At @G:@I produced @G",address,ins,addr));

	    CFG_FOREACH_BBL(T_CFG(code->data),bbl2)
	    {
	      t_address base;
	      base = BBL_CADDRESS(bbl2);

	      /* Search for the basic block that contains this address */

	      if ((GENERIC_LE(base,addr))&&(GENERIC_GT(GENERIC_ADD_UINT32(base,bbl2->nins*4),addr)))
	      {

		/* Found one: is it data or code ? */
		if (ARMINSOPCODE(bbl2->ins_first)==ARM_DATA)
		{
		  /*TODO: check if this is entirely correct: I especially doubt the calculation program is correct (no addend???) */
		  rel=RelocTableAddRelocToRelocatable(obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(bbl2),GENERIC_SUB(addr,base), TRUE, NULL, NULL,NULL,"SA+\\l*w\\s0000$",TRUE);
		  rel->label=StringDup("ADDR ADD TO DATA");
		  nr_addr_prod++;
		  ArmMakeAddressProducer(__FILE__,__LINE__,ins,G_T_UINT32(addr),rel);
		}
		else /* It's code ! */
		{

		  if (!RegsetIn(INS_REGS_DEF(ins) , 14)) /* Are we producing the return register? */
		  {

		    t_bool found=FALSE;
		    t_ins * i_ins=ins;
		    /* Try to find the use */

		    while((i_ins=INS_NEXT(i_ins)))
		    {
		      if (RegsetIn(INS_REGS_USE(i_ins) , ARMINSREGA(ins)))
		      {
			found=TRUE;
			break;
		      }
		    }

		    if (!found)
		    {
		      /* assume the produced address can only be used for calling 
		       * the addressed bbl as a function. */
		      t_cfg_edge * edge;
		      printf("DANGEROUS: produced code address, use not found\n");
		      DiabloPrint(stdout,"-~-~- Relevant instruction: @I\n", ins);
		      VERBOSE(1,("Patching return! @I",ins));
		      /* sanity check: the produced address _should_ be the beginning
		       * of a basic block */
		      VERIFY_FASSERT(GENERIC_EQ(addr,base),("Produced code address not beginning of bbl"));
		      
		      edge=CfgEdgeNewCall(code->data,T_CFG(code->data)->hell_node,bbl2,NULL,NULL);
		      rel=RelocTableAddRelocToRelocatable(obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(bbl2),NULL_GENERIC,TRUE,edge,NULL,NULL,"SA+\\l*w\\s0000$",TRUE);
		      rel->label=StringDup("ADDR ADD TO UNKNOWN CODE");

		      nr_addr_prod++;
		      ArmMakeAddressProducer(__FILE__,__LINE__,ins,G_T_UINT32(addr),rel);
		    }
		    else
		    {

		      if (INS_TYPE(i_ins)==IT_LOAD)
		      {
			/* it appears we're making a data pointer, so we should look for the data block within 
			 * this section: the reloc should point to that block (with the proper addend of course) */
			t_address symadr=NULL_GENERIC; /* Just to keep the compiler happy. Never used! */
			t_bool found=FALSE;
			t_map_node * sub;
			t_section * sub_sec;
			sub=MapGetSection(obj->map,addr); 
			sub_sec=SectionGetFromObjectByName(sub->obj,sub->sec_name); 

			OBJECT_FOREACH_SYMBOL(obj,sym)
			{
			  if (GENERIC_LE(SECTION_GET_CADDRESS(sub_sec),sym->current_address) && (GENERIC_GT(GENERIC_ADD(SECTION_GET_CADDRESS(sub_sec),SECTION_CSIZE(sub_sec)),sym->current_address) ))
			  {
			    if (sym->name&&(strcmp("$d",sym->name)==0))
			    {
			      if (!found)
			      {
				found=TRUE;
				symadr=sym->current_address;
			      }
			      else if (strcmp(Basename(sub->obj->name),"f_a_p.l:fdiv.o")==0)
			      {
				/* special case from the arm ads floating point emul libs */
				symadr=SECTION_GET_CADDRESS(sub_sec);
			      }
			      else
			      {
				FATAL(("Two data blocks in section from file %s!",sub->obj->name));
				exit(0); /* Keep the compiler happy */
			      }
			    }
			  }
			}
			if (!found) FATAL(("Seems like we are making a datapointer, but we cannot find the data"));


			/* Find the real block = a datablock */
			CFG_FOREACH_BBL(T_CFG(code->data),bbl3)
			{
			  t_address base;
			  base = BBL_CADDRESS(bbl3);

			  if ((GENERIC_LE(base,symadr))&&(GENERIC_GT(GENERIC_ADD_UINT32(base,bbl3->nins*4),symadr)))
			  {
			    /* Ok found it */

			    rel=RelocTableAddRelocToRelocatable(obj->reloc_table,GENERIC_SUB(addr,symadr),T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(bbl3),NULL_GENERIC,TRUE, NULL,NULL,NULL,"SA+\\l*w\\s0000$",TRUE);
			    rel->label=StringDup("ADDR ADD TO DATA (guessed)");
			    nr_addr_prod++;
			    ArmMakeAddressProducer(__FILE__,__LINE__,ins,G_T_UINT32(addr),rel);
			    break;
			  }
			}
			VERBOSE(1,("Produced a pointer @G relative to @G",addr,symadr));
		      }
		      else FATAL(("Produced code address, used it in @I",i_ins));
		    }
		  }
		  else /* We are producing the return register! */
		  {
		    if (!GENERIC_EQ(GENERIC_SUB(addr,base),NULL_GENERIC)) FATAL(("Should not happen!"));
		    rel=RelocTableAddRelocToRelocatable(obj->reloc_table, NULL_GENERIC,T_RELOCATABLE(ins), NULL_GENERIC,T_RELOCATABLE(bbl2),GENERIC_SUB(addr,base), TRUE, NULL, NULL,NULL,"SA+\\l*w\\s0000$",rel->first);
		    rel->label=StringDup("ADDR RETURN");
		    nr_addr_prod++;
		    ArmMakeAddressProducer(__FILE__,__LINE__,ins,G_T_UINT32(addr),rel);
		  }

		  break; /* Stop searching for the block */
		}
	      }
	    }
	  }
	}
	address=GENERIC_ADD_UINT32(address,4);
      }
    }
  }

  /* End of 3. }}} */
  


#if 1
  CFG_FOREACH_BBL(T_CFG(code->data),bbl)
    {
      address = BBL_CADDRESS(bbl);
      BBL_FOREACH_INS(bbl,ins)
      {
	if (RegsetIn(INS_REGS_USE(ins) , 15))
	{
	  if ((ARMINSOPCODE(ins)==ARM_MOV) || (ARMINSOPCODE(ins)==ARM_STM))
	  {
	    /* Returns.... No need to handle this */
	  }
	  /* Just a test for the switches */
	  else if ((ARMINSOPCODE(ins)==ARM_ADD)&&(ARMINSSHIFTTYPE(ins)==ARM_SHIFT_TYPE_LSL_IMM)&&(ARMINSCONDITION(ins)==ARM_CONDITION_LS))
	  {
	  }
	  else if ((ARMINSOPCODE(ins)==ARM_LDR)&&(ARMINSSHIFTTYPE(ins)==ARM_SHIFT_TYPE_LSL_IMM)&&(ARMINSCONDITION(ins)==ARM_CONDITION_LS))
	  {
	  }
	  else if (INS_ATTRIB(ins) & IF_SWITCHJUMP)
	  {
	    /* Just a switch.... ignore it */ 
	  }
	  /* Drop out if we do not know the type */
	  else
	    FATAL(("Used pc, not handled: @I",ins));
	}
      }
  }
#endif

  /* temporary code to check for edges from datablocks */
  {
    t_cfg_edge * edge;
    printf("Checking for stray edges...\n");
    SECTION_FOREACH_EDGE(code,edge) 
    {
      t_bbl * from, * to;
      from = CFG_EDGE_HEAD(edge);
      to =CFG_EDGE_TAIL(edge); 

      if (from->ins_first)
	if (INS_TYPE(from->ins_first) == IT_DATA) 
	{
	  WARNING(("edge from data block! (@G)",BBL_CADDRESS(from)));
	}
      if (to->ins_first)
	if (INS_TYPE(to->ins_first) == IT_DATA)
	{
	  WARNING(("edge to data block! (@G)",BBL_CADDRESS(to)));
	}
    }

    printf("checking for odd basic blocks...\n");
    CFG_FOREACH_BBL(T_CFG(code->data),bbl)
    {
      int data = (bbl->ins_first ? (INS_TYPE(bbl->ins_first) == IT_DATA) : 0);
      BBL_FOREACH_INS(bbl,ins) 
      {
	if (data != (INS_TYPE(ins) == IT_DATA)) WARNING(("Odd block detected (@G)",BBL_CADDRESS(bbl)));
      }
    }
  }
  /*printf("Statistics address and constantproducers: %d\nStatistics loaders: %d",nr_addr_prod,nr_loaders);*/
}
/* }}} */
/*!
 * \brief 
 *
 * \param obj
 * \param code
 *
 * \return t_uint32 
*/
/* ArmPatchCallsToWeakSymbols {{{ */
t_uint32 ArmPatchCallsToWeakSymbols(t_object * obj, t_section * code)
{
  t_ins * ins;
  t_uint32 pc, ncalls=0;

  ins = code->data; 

  for (pc = 0; ins != NULL; ins = INS_NEXT(ins), pc += 4)
    {
      if (INS_TYPE(ins) == IT_BRANCH)
	if (ARMINSOPCODE(ins) == ARM_BL)
	  {
	    if ((ARMINSIMMEDIATE(ins) ==  -4))
	      {
		ncalls++;
		ArmInstructionMakeNoop(ins);	
	      }
	  }
    }
  return ncalls;
}
/* }}} */

static t_address addresses[50000] = {0};
t_int32 unique_addr = 0;
static void CountLoadAddress(t_address address)
{
  t_uint32 i;
  for(i=0; i < unique_addr; i++)
  {
    if(addresses[i] == address)
    {
      break;
    }
  }
  if(i == unique_addr)
  {
    unique_addr++;
    addresses[i] = address;
  }

}
/* vim: set shiftwidth=2 foldmethod=marker: */
