/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_iterators.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../diablo_options.h"
#include "arm_instructions.h"
#include "arm_registers.h"
#include "arm_utils.h"
/* }}} */
/* Static defines {{{ */
void ArmInlineProcedureWithOnlyOneIncomingNonCallEdge(t_cfg * cfg, t_function* callee);
t_uint32 ArmInlineSmallProcedure(t_object * obj,t_cfg * cfg, t_function * fun);
void ArmInlineFunAtCallsite(t_function * fun, t_bbl *call_site);
/* }}} */
/*!  
 * inline a procedure (that has multiple call sites) at one of it's call
 * sites. This is the most general form of inlining available 
 *
 * \param fun The function to inline
 * \param call_edge The call edge from the callsite at which we want to inline
 * to the function we want to inline
 *
 * \return void 
*/
/* ArmInlineProcedureAtOneCallsite  {{{ */
void ArmInlineProcedureAtOneCallsite(t_function * fun, t_cfg_edge * call_edge)
{

  /* iterators */
  t_bbl * i_bbl;
  t_ins * i_ins;
  t_cfg_edge * i_edge;
  
  /* local variables */
  t_cfg * cfg;
  t_cfg_edge * return_edge;
  t_cfg_edge * link_edge;

  t_bbl * call_site;
  t_bbl * return_site;

  t_bool ok;
  t_bool switch_condition;
  t_uint32 nr_return_points = 0;

  cfg = BBL_CFG(CFG_EDGE_HEAD(call_edge));
  call_site = CFG_EDGE_HEAD(call_edge);
  return_edge = call_edge->corr_edge;
  BBL_FOREACH_SUCC_EDGE(call_site,link_edge)
    if ((EDGE_CAT(link_edge) == ET_FUNLINK) || (EDGE_CAT(link_edge) == ET_IPFUNLINK))
      break;
  VERIFY_FASSERT(link_edge,("No link edge found for function call @E!\n", call_edge));
  return_site = CFG_EDGE_TAIL(link_edge);

  /* step 0: check if inlining this function is possible. We don't want to
   * inline functions that perform interprocedural jumps because it's
   * impossible to remodel control flow so that it stays correct in this case.
   * -- Optimization: we will allow interprocedural jumps that are tail calls,
   * 	because we can in effect remodel control flow for this situation by 
   * 	changing the jumps to function calls.
   */
  ok = TRUE;
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    /* the tmp2 field will be used as a boolean that indicates 
     * if a tail call starts from this block */
    ((t_bool) i_bbl->tmp2) = FALSE;


    /* skip return edges */
    if (i_bbl == FUN_BBL_LAST(fun))
      continue; 

    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      if (CfgEdgeIsInterproc(i_edge))
	if (EDGE_CAT(i_edge) != ET_CALL)
	{
	  if ((EDGE_CAT(i_edge) == ET_IPJUMP) && (CFG_EDGE_TAIL(i_edge) == FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_TAIL(i_edge)))))
	  {
	    /* only handle interprocedural edges of type IPJUMP, because the others are too complicated to handle well */
	    ((t_bool) i_bbl->tmp2) = TRUE;
	    DiabloPrint(stdout,"-~-~- tail call detected. still carrying on though\n");
	  }
	  else
	  {
	    ok = FALSE;
	  }
	}
    }
  }
  if (!ok)
  {
    VERBOSE(1,("No inline for fun @G because of interprocedural jumps!\n", BBL_CADDRESS(FUN_BBL_FIRST(fun))));
    return;
  }
  /* step 0b: check if all return instructions are of a format we understand.
   * At present that means they ought to be:
   * 		- mov pc,r14
   * 		- ldr pc,[r13,offset]
   * 		- ldmia r13!,{...,pc}     (ads-style return)
   * 		- ldmdb rXX, {...,r13,pc} (gcc-style return)
   */
  BBL_FOREACH_PRED_EDGE(FUN_BBL_LAST(fun),i_edge)
  {
    t_ins * lastins;

    /* skip compensating edges */
    if (EDGE_CAT(i_edge) == ET_COMPENSATING)
      continue;
    
    nr_return_points++;

    lastins = BBL_INS_LAST(CFG_EDGE_HEAD(i_edge));
    if ((ARMINSOPCODE(lastins) == ARM_MOV) && (ARMINSREGA(lastins) == ARM_REG_R15) && (ARMINSREGC(lastins) == ARM_REG_R14))
      continue;
    if ((ARMINSOPCODE(lastins) == ARM_LDR) && (ARMINSREGA(lastins) == ARM_REG_R15) && (ARMINSREGB(lastins) == ARM_REG_R13))
      continue;
    if ((ARMINSOPCODE(lastins) == ARM_LDM) && (ARMINSREGB(lastins) == ARM_REG_R13) && (ARMINSIMMEDIATE(lastins) & (1 << 15)) && ArmInsWriteBackHappens(lastins) && (ARMINSFLAGS(lastins) & FL_DIRUP))
      continue;
    if ((ARMINSOPCODE(lastins) == ARM_LDM) && (ARMINSREGB(lastins) != ARM_REG_R13) && (ARMINSIMMEDIATE(lastins) & ((1 << 13) | (1 << 15))) && !ArmInsWriteBackHappens(lastins) && !(ARMINSFLAGS(lastins) & FL_DIRUP))
      continue;

    ok = FALSE;
    break;
  }
  if (!ok)
  {
    VERBOSE(1,("No inline for fun @G because of abnormal return instruction\n", BBL_CADDRESS(FUN_BBL_FIRST(fun))));
    return;
  }

  DiabloPrint(stdout,"Inlining func %s from call edge @E\n",FUN_NAME(fun),call_edge);

  /* an optimization:
   * if the function that we are inlining only has one return point,
   * and the call to the function is conditional, we can gain an instruction
   * by inverting the condition and making it jump to the original fallthrough case.
   * we can then drop the branch to the return site at the return point and add a 
   * fallthrough edge from the return point to the return site.
   */
  if ((nr_return_points == 1) && (INS_IS_CONDITIONAL(BBL_INS_LAST(call_site))))
  {
    DiabloPrint(stdout,"Will swap condition on call\n");
    switch_condition = TRUE;
  }
  else
    switch_condition = FALSE;






  /* step 1: duplicate all basic blocks of the function */
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    t_bbl * copy;

    /* don't copy the return block */
    if (i_bbl == FUN_BBL_LAST(fun))
      continue;

    copy = BblNew(cfg);
    /* temporary references between the duplicate blocks, this is for easy
     * reference from original to copy and vice versa. these references will
     * of course be removed at the end of the inlining. */
    i_bbl->tmp = copy;
    copy->tmp = i_bbl;

    /* duplicate the instructions in the basic block */
    BBL_FOREACH_INS(i_bbl,i_ins)
    {
      t_ins * inscopy = InsDup(i_ins);
      InsAppendToBbl(inscopy, copy);
    }
  }

  /* step 1b: assign the copies of the basic blocks to the calling function */
  FUN_FOREACH_BBL_R(fun,i_bbl)
  {
    if (i_bbl!=FUN_BBL_LAST(fun))
    {	
      t_bbl * copy = T_BBL(i_bbl->tmp);
      copy->next_in_fun=call_site->next_in_fun;
      call_site->next_in_fun->prev_in_fun=copy;
      call_site->next_in_fun = copy;
      copy->prev_in_fun = call_site;
      copy->fun=call_site->fun;
    }
  }






  /* step 2: add edges to the newly created blocks */
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    /* skip the return edges */
    if (i_bbl == FUN_BBL_LAST(fun)) continue;

    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      /* intraprocedural edges should be duplicated between the copies of head
       * and tail. call edges should be duplicated between
       * the copy of the head and the original tail. interproc jump edges that
       * are used for tail calls are skipped for now. other interprocedural
       * edges should not occur. */
      t_bbl * newhead, * newtail;

      /* skip the edges to the return block */
      if (CFG_EDGE_TAIL(i_edge) == FUN_BBL_LAST(fun))
	continue;

      /* skip interprocedural edges (only tail calls should happen but we handle them later) */
      if (CfgEdgeIsInterproc(i_edge) && !(EDGE_CAT(i_edge) == ET_CALL))
	continue;

      newhead = T_BBL(i_bbl->tmp);
      if (EDGE_CAT(i_edge) == ET_CALL)
      {
	t_bbl * newreturn;

	newtail = CFG_EDGE_TAIL(i_edge);
	newreturn = T_BBL(CFG_EDGE_TAIL(i_edge->corr_edge)->tmp);
	CfgEdgeNewCall(cfg,newhead,newtail,newreturn,FUN_BBL_LAST(BBL_FUN(newtail)));
      }
      else
      {
	newtail = T_BBL(CFG_EDGE_TAIL(i_edge)->tmp);
	CfgEdgeNew(cfg,newhead,newtail,EDGE_CAT(i_edge));
      }
    }
  }






  /* step 3: change the function call into a jump */
  if (!switch_condition)
  {
    /* jump to the inlined part */
    ArmInstructionMakeCondBranch(BBL_INS_LAST(call_site),ARMINSCONDITION(BBL_INS_LAST(call_site)));
    CfgEdgeNew(cfg,call_site,T_BBL(FUN_BBL_FIRST(fun)->tmp),ET_JUMP);
  }
  else
  {
    /* invert the condition on the original call instruction and swap taken and fallthrough cases */
    t_cfg_edge * ft_edge;

    ArmInstructionMakeCondBranch(BBL_INS_LAST(call_site),ArmInvertCondition(ARMINSCONDITION(BBL_INS_LAST(call_site))));

    /* find the original fallthrough edge */
    BBL_FOREACH_SUCC_EDGE(call_site,ft_edge)
      if ((EDGE_CAT(ft_edge) == ET_FALLTHROUGH) || (EDGE_CAT(ft_edge) == ET_IPFALLTHRU))
	break;
    VERIFY_FASSERT(ft_edge,("Could not find fallthrough edge with conditional control transfer"));
    /* turn it into a jump edge */
    EDGE_CAT(ft_edge) = (EDGE_CAT(ft_edge) == ET_IPFALLTHRU) ? ET_IPJUMP : ET_JUMP;

    /* add a fallthrough edge from the call site to the entry of the inlined function */
    CfgEdgeNew(cfg,call_site,T_BBL(FUN_BBL_FIRST(fun)->tmp),ET_FALLTHROUGH);
  }






  /* step 4: change the return instructions into jumps */
  /* This is a little tricky: if the original link edge was interprocedural,
   * the  returning jumps should also be interprocedural */
  {
    t_bool return_is_interproc = (EDGE_CAT(link_edge) == ET_IPFUNLINK);

    BBL_FOREACH_PRED_EDGE(FUN_BBL_LAST(fun),i_edge)
    {
      t_bbl * block;
      t_ins * return_ins;

      /* skip compensating edges */
      if (EDGE_CAT(i_edge) == ET_COMPENSATING)
	continue;

      block = T_BBL(CFG_EDGE_HEAD(i_edge)->tmp);
      return_ins = BBL_INS_LAST(block);       
      
      /* as a reminder: in the beginning of the function we checked whether
       * the return instruction was of one of the following forms:
       * 	- mov pc,r14
       * 	- ldr pc,[r13,offset]
       * 	- ldmia r13!,{...,pc}
       * 	- ldmdb rXX, {...,r13,pc}
       * if we found any other return instructions, we refused to do the
       * inlining. */
      switch (ARMINSOPCODE(return_ins))
      {
	case ARM_MOV:
	  /* just turn this instruction into a (conditional) branch to the
	   * right destination, and add a jump edge to the return site 
	   * alternatively, if switch_condition is TRUE, add a fallthrough
	   * edge to the return site and remove the return instruction altogether */
	  if (!switch_condition)
	    ArmInstructionMakeCondBranch(return_ins,ARMINSCONDITION(return_ins));
	  else
	    InsKill(return_ins);
	  break;

	case ARM_LDR:
	  /* LDR's : Remove the LDR, add an indexing instruction if necessary and add the branch */

	  /* a small optimization:
	   * if r14 is dead at the beginning of the return site, we can just as well load the stack
	   * value in r14 (it was most probably stored from r14 as well). the idea is that we can
	   * reintroduce some symmetry in this way: in the beginning of the function we have STR r14,[r13]
	   * and we make sure that at the end of the function we have LDR r14,[r13]. This makes it easier
	   * for other analyses to remove this pair of stack instructions. Note that it isn't very useful to
	   * use another dead register for this purpose because that won't introduce the same symmetry, 
	   * and that will certainly make OptKillUseless change the load into a stack indexing operation, so we 
	   * might as well do that here in the first place. */
	  if (!RegsetIn(InsRegsLiveBefore(BBL_INS_FIRST(return_site)),ARM_REG_R14))
	  {
	    DiabloPrint(stdout,"-~-~- Optimized LDR\n");
	    ARMINSREGA(return_ins) = ARM_REG_R14;
	    INS_REGS_DEF(return_ins)=ArmDefinedRegisters(return_ins);

	    if (!switch_condition)
	    {
	      t_ins * j_ins = InsNewForCfg(cfg);
	      GenericSet_uint32(INS_CSIZE(j_ins),4);
	      GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
	      InsInsertAfter(j_ins,return_ins);
	      ArmInstructionMakeCondBranch(j_ins,ARMINSCONDITION(return_ins));
	    }
	  }
	  else
	  {
	    if (ArmInsWriteBackHappens(return_ins))
	    {
	      if (ARMINSFLAGS(return_ins) & FL_DIRUP)
	      {
		t_uint32 imm=ARMINSIMMEDIATE(return_ins);
		ArmInstructionMakeAdd(return_ins,ARMINSREGB(return_ins),ARMINSREGB(return_ins),ARM_REG_NONE,imm,ARMINSCONDITION(return_ins));

		if (!switch_condition)
		{
		  t_ins * j_ins = InsNewForCfg(cfg);
		  GenericSet_uint32(INS_CSIZE(j_ins),4);
		  GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
		  InsInsertAfter(j_ins,return_ins);
		  ArmInstructionMakeCondBranch(j_ins,ARMINSCONDITION(return_ins));
		}
	      }
	      else
		FATAL(("Implement stacks that grow in upward direction"));
	    }
	    else
	      FATAL(("Implement returns that don't restore the stack correctly"));
	  }
	  break;

	case ARM_LDM:
	  /* Load multiples: remove pc from multiple load, add a stack indexing instruction */ 
	  {
	    t_ins * j_ins;

	    if (!switch_condition)
	    {
	      /* Add the branch */
	      j_ins = InsNewForCfg(fun->cfg);
	      GenericSet_uint32(INS_CSIZE(j_ins),4);
	      GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
	      InsInsertAfter(j_ins,return_ins);
	      ArmInstructionMakeCondBranch(j_ins,ARMINSCONDITION(return_ins));
	    }

	    /* an optimization:
	     * if r14 is not live at the beginning of the return site, it means we can
	     * trash whatever is in it. This means we can use it to load the value that would 
	     * normally be loaded into the program counter. That way we don't have to add a stack
	     * indexing instruction to keep the stack pointer correct */
	    if (!RegsetIn(InsRegsLiveBefore(BBL_INS_FIRST(return_site)),ARM_REG_R14) && !(ARMINSIMMEDIATE(return_ins) & (1 << 14)))
	    {
	      DiabloPrint(stdout,"-~-~- Optimized LDM\n");
	      ARMINSIMMEDIATE(return_ins) &= ~(1 << 15);
	      ARMINSIMMEDIATE(return_ins) |= (1 << 14);
	      INS_REGS_DEF(return_ins)=ArmDefinedRegisters(return_ins);
	    }
	    else
	    {
	      /* check which kind of return instruction (gcc-style or ads-style)
	       * we have */
	      if (ARMINSREGB(return_ins) == ARM_REG_R13)
	      {
		/* ads-style */
		/* Add the indexing instruction before the branch, after the LDM */
		j_ins = InsNewForCfg(fun->cfg);
		GenericSet_uint32(INS_CSIZE(j_ins),4);
		GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
		InsInsertAfter(j_ins,return_ins);
		ArmInstructionMakeAdd(j_ins,ARMINSREGB(return_ins),ARMINSREGB(return_ins),ARM_REG_NONE,4,ARMINSCONDITION(return_ins));
		ARMINSIMMEDIATE(return_ins) &= ~(1<<15);
		INS_REGS_DEF(return_ins)=ArmDefinedRegisters(return_ins);
	      }
	      else
	      {
		/* gcc-style */
		/* Add the indexing instruction before the LDM */
		j_ins = InsNewForCfg(fun->cfg);
		GenericSet_uint32(INS_CSIZE(j_ins),4);
		GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
		InsInsertBefore(j_ins,return_ins);
		ArmInstructionMakeSub(j_ins,ARMINSREGB(return_ins),ARMINSREGB(return_ins),ARM_REG_NONE,4,ARMINSCONDITION(return_ins));
		ARMINSIMMEDIATE(return_ins) &= ~(1<<15);
		INS_REGS_DEF(return_ins)=ArmDefinedRegisters(return_ins);
	      }
	    }
	    break;
	  }

	default: FATAL(("Found a malformed return instruction. This should have been detected before!"));
      }

      /* add the jump edge */
      if (return_is_interproc)
      {
	t_cfg_edge * insert_edge = CfgEdgeNew(cfg,block,return_site,switch_condition ? ET_IPFALLTHRU : ET_IPJUMP);
	CfgEdgeNewCompensating(cfg,insert_edge);
      }
      else
      {
	CfgEdgeNew(cfg,block,return_site,switch_condition ? ET_FALLTHROUGH : ET_JUMP);
      }
    }
  }

  /* step 4b: handle the interprocedural tail calls.
   * This basically means turning the jump instruction into a function call and adding a
   * block to which the call can return */
  {
    t_bool return_is_interproc = (EDGE_CAT(link_edge) == ET_IPFUNLINK);

    FUN_FOREACH_BBL(fun,i_bbl)
    {
      t_bbl * copy;
      t_bbl * call_return;
      t_ins * jump_ins;
      
      if (!((t_bool) i_bbl->tmp2)) continue;

      copy = T_BBL(i_bbl->tmp);
      /* find the interprocedural jump edge */
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	if (EDGE_CAT(i_edge) == ET_IPJUMP)
	  break;
      VERIFY_FASSERT(i_edge,("interproc jump edge not found. this should not happen!"));

      /* change the last instruction of copy into a function call */
      jump_ins = BBL_INS_LAST(copy);
      ArmInstructionMakeCondBranchAndLink(jump_ins,ARMINSCONDITION(jump_ins));

      /* prepare a basic block for the call to return to */
      if (!switch_condition)
      {
	call_return = BblNew(cfg);

	call_return->next_in_fun = copy->next_in_fun;
	call_return->prev_in_fun = copy;
	copy->next_in_fun->prev_in_fun = call_return;
	copy->next_in_fun = call_return;

	BBL_FUN(call_return) = BBL_FUN(call_site);

	InsAppendToBbl(InsNewForCfg(cfg),call_return);
	GenericSet_uint32(INS_CSIZE(BBL_INS_FIRST(call_return)),4);
	GenericSet_uint32(INS_OLD_SIZE(BBL_INS_FIRST(call_return)),0);
	ArmInstructionMakeCondBranch(BBL_INS_FIRST(call_return),ARM_CONDITION_AL);
      }
      else
      {
	call_return = return_site;
      }

      /* Add the call and return edges */
      CfgEdgeNewCall(cfg,copy,CFG_EDGE_TAIL(i_edge),call_return,FUN_BBL_LAST(BBL_FUN(CFG_EDGE_TAIL(i_edge))));

      /* Add the funlink edge */
      if (!switch_condition)
      {
	/* always put a regular funlink edge between copy and call_return */
	CfgEdgeNew(cfg,copy,call_return,ET_FUNLINK);
      }
      else
      {
	/* this is slightly more complicated: the funlink edge points directly to
	 * return_site, which means that in the case of an interprocedural return,
	 * the funlink must be made interprocedural as well.
	 */
	if (return_is_interproc)
	{
	  t_cfg_edge * ipfunlink;
	  ipfunlink = CfgEdgeNew(cfg,copy,return_site,ET_IPFUNLINK);
	  CfgEdgeNewCompensating(cfg,ipfunlink);
	}
	else
	{
	  CfgEdgeNew(cfg,copy,return_site,ET_FUNLINK);
	}
      }

      /* if the conditions on the original call are not inverted,
       * we must now add a jump edge from the call_return block to the 
       * return site. */
      if (!switch_condition)
      {
	if (return_is_interproc)
	{
	  t_cfg_edge * ipjump = CfgEdgeNew(cfg,call_return,return_site,ET_IPJUMP);
	  CfgEdgeNewCompensating(cfg,ipjump);
	}
	else
	{
	  CfgEdgeNew(cfg,call_return,return_site,ET_JUMP);
	}

      }
    }
  }





  /* step 5: clean up the mess */
  /* this means: - remove the references from original to copy and vice versa
   *               that were stored in the tmp field of the basic blocks 
   *             - remove the original call, return and link edges
   */
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    /* skip the return block because it doesn't have a copy */
    if (i_bbl == FUN_BBL_LAST(fun))
      continue;

    T_BBL(i_bbl->tmp)->tmp = NULL;
    i_bbl->tmp = NULL;
  }
  CfgEdgeKill(call_edge);
  CfgEdgeKill(return_edge);
  CfgEdgeKill(link_edge);
}
/* }}} */
/*!
 * \todo Document
 *
 * \param cfg
 *
 * \return void 
*/
/* ArmInlineTrivial {{{ */
void ArmInlineTrivial(t_cfg * cfg)
{
  t_object * obj=CFG_OBJECT(cfg);
  t_ins * i_ins;
  t_cfg_edge * i_edge;
  t_function * i_fun;
  t_bbl * i_bbl;
  t_function* tmp;

  int nr_ins;
  int nr_edges;
  int nr_callees = 0;
  int nr_callers = 0;

  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    nr_ins = nr_edges = 0;

    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	nr_ins++;
      }

      BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
      {
	if (!CfgEdgeIsForwardInterproc(i_edge)) continue;
	if (EDGE_CAT(i_edge)==ET_CALL) nr_edges+=100;
	if (IsHellNode(EDGE_HEAD(i_edge))) nr_edges+=100;
	nr_edges++;
      }
    }

    if ((nr_ins==2 || nr_ins==1)&&(BBL_INS_LAST(i_fun->bbl_first)))
    {
      if (ARMINSOPCODE(BBL_INS_LAST(i_fun->bbl_first))==ARM_MOV && ARMINSREGA(BBL_INS_LAST(i_fun->bbl_first))==15 &&  ARMINSREGC(BBL_INS_LAST(i_fun->bbl_first))==14)
      {
	int nr = ArmInlineSmallProcedure(obj, cfg, i_fun);
	if (nr)
	  nr_callees++;
	nr_callers+=nr;
      }
    }
    else
      if (nr_edges<2)
      {
	if (nr_edges==1)
	{
	  nr_callees++;
	  nr_callers++;
	  /*	  ArmInlineProcedureWithOnlyOneIncomingNonCallEdge(cfg, i_fun);*/
	}
      }
  }

  printf("[inlining            ] %d callees inlined at %d callers\n",nr_callees,nr_callers);
}
/* }}} */
/*!
 * \todo Document
 *
 * \param cfg
 * \param callee
 *
 * \return void 
*/
/* ArmInlineProcedureWithOnlyOneIncomingNonCallEdge {{{ */
void ArmInlineProcedureWithOnlyOneIncomingNonCallEdge(t_cfg * cfg, t_function* callee)
{
  t_cfg_edge * tmp_edge;
  t_cfg_edge * i_edge=NULL;
  t_cfg_edge * call_edge;
  t_function * caller;
  t_bbl * i_bbl, *tmp, *last_bbl;

  /* Don't inline the hell function! */
  if (FUN_IS_HELL(callee)) return;

  FUN_FOREACH_BBL(callee,i_bbl) 
    BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
    {
      if (CfgEdgeIsForwardInterproc(i_edge)) goto found;
    }

found:
  call_edge = i_edge;

  /* Don't inline the entry point into the function containing the unique entry node */
  if (CFG_EDGE_HEAD(call_edge) == cfg->unique_entry_node) return;

  caller = BBL_FUN(EDGE_HEAD(call_edge));

  /* move the basic blocks */

  FUN_FOREACH_BBL_SAFE(callee,i_bbl,tmp)
  {
    /* insert the basic block */
    if (i_bbl==callee->bbl_last) continue;

    callee->bbl_first=i_bbl->next_in_fun;
    callee->bbl_first->prev_in_fun = NULL;

    caller->bbl_last->prev_in_fun->next_in_fun=i_bbl;
    i_bbl->prev_in_fun=caller->bbl_last->prev_in_fun;
    i_bbl->next_in_fun=caller->bbl_last;
    caller->bbl_last->prev_in_fun=i_bbl;
    i_bbl->fun=caller;      


    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
      if (EDGE_CAT(i_edge)!=ET_CALL && CfgEdgeIsForwardInterproc(i_edge) && BBL_FUN(CFG_EDGE_TAIL(i_edge))==caller)
      {
	switch (EDGE_CAT(i_edge))
	{
	  case ET_IPUNKNOWN: 
	    EDGE_CAT(i_edge)= ET_UNKNOWN; 
	    break;
	  case ET_IPFALLTHRU: 
	    EDGE_CAT(i_edge)=ET_FALLTHROUGH; 
	    break;
	  case ET_IPFUNLINK: 
	    EDGE_CAT(i_edge)=ET_FUNLINK; 
	    break;
	  case ET_IPSWILINK: 
	    EDGE_CAT(i_edge)=ET_SWILINK; 
	    break;
	  case ET_IPJUMP: 
	    EDGE_CAT(i_edge)=ET_JUMP; 
	    break;
	  default:
	    printf("we have a problem: unknown IP edge type\n");
	    exit(0);
	}
	if (i_edge->corr_edge)
	{
	  CfgEdgeKill(i_edge->corr_edge);
	  i_edge->corr_edge=NULL;
	}
      }
  }

  /* adapt the edge */

  if (call_edge->corr_edge)
  {
    CfgEdgeKill(call_edge->corr_edge);
    call_edge->corr_edge=NULL;
  }

  switch (EDGE_CAT(call_edge))
  {
    case ET_IPUNKNOWN: 
      EDGE_CAT(call_edge)= ET_UNKNOWN; 
      break;
    case ET_IPFALLTHRU: 
      EDGE_CAT(call_edge)=ET_FALLTHROUGH; 
      break;
    case ET_IPFUNLINK: 
      EDGE_CAT(call_edge)=ET_FUNLINK; 
      break;
    case ET_IPSWILINK: 
      EDGE_CAT(call_edge)=ET_SWILINK; 
      break;
    case ET_IPJUMP: 
      EDGE_CAT(call_edge)=ET_JUMP; 
      break;
    default:
      FATAL(("Unkown IP edge type: @E\n",call_edge));
      exit(0);
  }

  /* adapt all incoming edges in the callee's last block */

  BBL_FOREACH_PRED_EDGE_SAFE(callee->bbl_last,i_edge,tmp_edge)
  { 
    CFG_EDGE_TAIL(i_edge)=caller->bbl_last;

    EDGE_PRED_NEXT(i_edge)=NODE_PRED_FIRST(caller->bbl_last);

    if (NODE_PRED_FIRST(caller->bbl_last))
      EDGE_PRED_PREV(NODE_PRED_FIRST(caller->bbl_last))=T_EDGE(i_edge);

    EDGE_PRED_PREV(i_edge)=NULL;

    NODE_PRED_FIRST(caller->bbl_last)=T_EDGE(i_edge);

    if (NODE_PRED_LAST(caller->bbl_last)==NULL)
      NODE_PRED_LAST(caller->bbl_last)=T_EDGE(i_edge);
  }


  NODE_PRED_FIRST(callee->bbl_last)=NULL;
  NODE_PRED_LAST(callee->bbl_last)=NULL;

  NODE_SUCC_FIRST(callee->bbl_last)=NULL;
  NODE_SUCC_LAST(callee->bbl_last)=NULL;

  last_bbl = callee->bbl_last;

  BblKill(last_bbl);


  FunctionKill(callee);
}
/* }}} */
/*!
 * try to inline functions about which constant propagation indicates that 
 * inlining them loses enough instructions to make it worth while.
 * note that this must be executed _after_ constant propagation! 
 *
 * \param cfg
 *
 * \return void 
*/
/* ArmInlineSelectiveFromCPInfo {{{ */
void ArmInlineSelectiveFromCPInfo(t_cfg * cfg)
{
  t_function * fun;
  t_bbl * bbl;
  t_ins * ins;
  t_cfg_edge * edge;
  t_cfg_edge * tmp;

  t_uint32 fun_nr_ins;

  CFG_FOREACH_FUN(cfg,fun)
  {
    t_bool ok = TRUE;
    t_uint32 inlined_nr_ins = 0;
    t_uint32 nr_callsites = 0;

    if ((fun == cfg->hell_function) || (fun == cfg->call_hell_function))
      continue;

    /* count the number of instructions in the function */
    fun_nr_ins = 0;
    FUN_FOREACH_BBL(fun,bbl)
      BBL_FOREACH_INS(bbl,ins)
      fun_nr_ins++;

    /* first try: if we know all the call sites, and the number of
     * instructions left after inlining the function at every call site 
     * is less than the number of instructions in the function, do it.
     */

    FUN_FOREACH_BBL(fun,bbl)
    {

      BBL_FOREACH_PRED_EDGE(bbl,edge)
      {
	if (!CfgEdgeIsForwardInterproc(edge)) continue;

	if (IsHellNode(CFG_EDGE_HEAD(edge))) 
	{
	  ok = FALSE;
	  break;
	}

	/* to be safe, only consider functions that are only entered via the
	 * traditional way (ie via a call edge) */
	if (EDGE_CAT(edge) != ET_CALL)
	{
	  ok = FALSE;
	  break;
	}

	inlined_nr_ins += fun_nr_ins - edge->n_unevaluated_ins;
	nr_callsites++;
      }
    }

    if (ok && (nr_callsites > 1)) 	/* nr_callsites == 1 is handled in GeneralInlining */
    {
      /* we arbitrarily guess that every inlining automatically gains about
       * 3 instructions (through copy propagation and dismissal of
       * call/return sequences */
      if (inlined_nr_ins <= fun_nr_ins + nr_callsites*2)
      {
	DiabloPrint(stdout,"Trying to  inline fun %s at %d call sites\n",FUN_NAME(fun),nr_callsites);
	DiabloPrint(stdout,"fun has %d instructions, after inlining %d are necessary\n",fun_nr_ins,inlined_nr_ins);
	BBL_FOREACH_PRED_EDGE(FUN_BBL_FIRST(fun),edge)
	{
	  if (EDGE_CAT(edge) != ET_CALL) continue;
	  ArmInlineProcedureAtOneCallsite(fun,edge);
	}

	/* on to the next function! */
	continue;
      }	
    }

    /* second try: perhaps we can find an invocation of the function that
     * requires so little instructions we might just as well inline it.
     * the boundary for this is arbitrarily set to 2 instructions and 
     * we require the call to be a real call, not an interprocedural jump */
    FUN_FOREACH_BBL(fun,bbl)
    {
      BBL_FOREACH_PRED_EDGE_SAFE(bbl,edge,tmp)
      {
	if (EDGE_CAT(edge) != ET_CALL)
	  continue;
	if (IsHellNode(CFG_EDGE_HEAD(edge)))
	  continue;

	if (fun_nr_ins - edge->n_unevaluated_ins <= 2)
	{
	  /*DiabloPrint(stdout,"Trying to inline fun %s (@B) at @B, call edge @E\n",FUN_NAME(fun),FUN_BBL_FIRST(fun),CFG_EDGE_HEAD(edge),edge); */
	  ArmInlineProcedureAtOneCallsite(fun,edge);
	}
      }
    }
  }
}
/* }}} */
/*!
 * Inlines a funtion of 1 real block (maximum 2 instructions) (and one
 * exit block) at its callsites.
 *
 *
 * \todo Remove obj and cfg as parameters
 * 
 * \param obj 
 * \param cfg the control flowgraph to optimize
 * \param fun The function to inline
 *
 * \return t_uint32 The number of times the function got inlined
*/
/* ArmInlineSmallProcedure {{{ */
t_uint32 ArmInlineSmallProcedure(t_object * obj, t_cfg * cfg, t_function * fun)
{
  t_bbl* entry;
  t_bbl* i_bbl;
  t_cfg_edge * call_edge, * tmp;
  int nr_bbls=0;
  t_ins * new_ins;

  FUN_FOREACH_BBL(fun,i_bbl)
  {
    nr_bbls++;
  }

  if (nr_bbls!=2)
  {
    printf("not inlined: number of bbls is %d\n",nr_bbls++);
  }

  nr_bbls=0;

  entry = fun->bbl_first;


  BBL_FOREACH_PRED_EDGE_SAFE(entry,call_edge,tmp)
  {
    t_bbl* call_site = CFG_EDGE_HEAD(call_edge);
    t_ins* call_ins = BBL_INS_LAST(call_site);
    if (EDGE_CAT(call_edge)!=ET_CALL) continue;
    if (IsHellNode(call_site)) continue;
    if (ARMINSOPCODE(call_ins)!=ARM_BL) continue;
    if ((BBL_NINS(entry)==1 || BBL_NINS(entry)==2) && ((!ArmInstructionIsConditional(BBL_INS_FIRST(entry)) && !ArmInstructionIsConditional(BBL_INS_LAST(entry)) && !ArmInsUpdatesCond(BBL_INS_FIRST(entry)) && !ArmInsUpdatesCond(BBL_INS_LAST(entry))) || !ArmInstructionIsConditional(call_ins)))
      {
	t_cfg_edge * link_edge = NULL;
	t_cfg_edge * fall_through_edge = NULL;
	t_cfg_edge * edge;
	nr_bbls++;
	
	/*	DiabloPrint(stdout,"@eiB\n",call_site);*/
	
	/*	Export_Function(BBL_FUN(call_site),"inline.dot");*/
	
	BBL_EXEC_COUNT(entry)-=EDGE_EXEC_COUNT(call_edge);

	if (BBL_NINS(entry)==2)
	  {
	    new_ins = InsDup(BBL_INS_FIRST(entry));
	    RELOCATABLE_REFED_BY(new_ins)=NULL;
	    InsAppendToBbl(new_ins,call_site);
	    if (INS_IS_CONDITIONAL(call_ins))
	      {
		ARMINSCONDITION(new_ins)=ARMINSCONDITION(call_ins);
		INS_ATTRIB(new_ins) |= IF_CONDITIONAL;
		INS_REGS_USE(new_ins) = ArmUsedRegisters(new_ins);
		INS_REGS_DEF(new_ins) = ArmDefinedRegisters(new_ins);
	      }
	  }
	
	
	InsKill(call_ins);
	
	if (tmp==call_edge->corr_edge)
	  tmp = T_CFG_EDGE(EDGE_PRED_NEXT(tmp));
	
	CfgEdgeKill(call_edge->corr_edge);
	CfgEdgeKill(call_edge);
	
	BBL_FOREACH_SUCC_EDGE(call_site,edge)
	  {
	    if  (EDGE_CAT(edge)==ET_FUNLINK || EDGE_CAT(edge)==ET_IPFUNLINK)
	      link_edge = edge;
	    else if  (EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU)
	      fall_through_edge = edge;
	    else
	      FATAL(("OOPS: EDGE OTHER THAN LINK OR FALLTHROUGH EDGE DURING INLINING\n"));
	  }
	
	if (fall_through_edge)
	  {
	    CfgEdgeKill(link_edge);
	  }
	else
	  {
	    if (!link_edge) FATAL(("OOPS: NO LINK EDGE IN INLINING\n"));
	    if (EDGE_CAT(link_edge)==ET_FUNLINK)
	      EDGE_CAT(link_edge)=ET_FALLTHROUGH;
	    else
	      EDGE_CAT(link_edge)=ET_IPFALLTHRU;
	  }
      }
  }
  return nr_bbls;
  
}
/* }}} */
/*!
 * \todo Document
 *
 * \param cfg
 *
 * \return void 
*/
/* ArmGeneralInlining {{{ */
void ArmGeneralInlining(t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_cfg_edge * i_edge;
  t_bool flag = TRUE; /* in and outgoing forward interproc edges are only call edges? */
  t_function * i_fun;
  int nr_callers; /* nr of call sites */
  t_bbl * call_site=NULL;
  t_ins * last_ins;
  t_function * tmp;

  DiabloPrint(stdout,"inlining ");
  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    flag = TRUE;
    nr_callers = 0;

    if(IsHellNode(FUN_BBL_FIRST(i_fun))) continue;
    
    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	if (CfgEdgeIsForwardInterproc(i_edge) && EDGE_CAT(i_edge)!=ET_CALL)
	  flag = FALSE;

      BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
      {
	if (CfgEdgeIsForwardInterproc(i_edge))
	{
	  if (EDGE_CAT(i_edge)!=ET_CALL)
	    flag = FALSE;
	  else if (IsHellNode(EDGE_HEAD(i_edge)))
	    nr_callers+=2;
	  else
	  {
	    nr_callers++;
	    call_site = CFG_EDGE_HEAD(i_edge);
	  }
	}
      }
    }

    /*if (call_site == cfg->unique_entry_node) continue; */
    if (!flag || nr_callers!=1)
      continue;

    if (FunctionGetExitBlock(i_fun)==FUN_BBL_LAST(i_fun))
    {
      BBL_FOREACH_PRED_EDGE(FUN_BBL_LAST(i_fun),i_edge)
      {
	last_ins = BBL_INS_LAST(CFG_EDGE_HEAD(i_edge));
	if (!last_ins)
	{
	  FunctionPrint(i_fun,PRINT_INSTRUCTIONS);
	  FATAL(("Found a return block without instructions\n"));
	}
	if (ArmInsHasShiftedFlexible(last_ins))
	{
	  printf("POTENTIAL: NO INLINE 3\n");
	  /*InsPrint(last_ins,0); */
	  flag = FALSE;
	  break;
	}
	if (ARMINSOPCODE(last_ins)!=ARM_MOV && ARMINSOPCODE(last_ins)!=ARM_LDM && ARMINSOPCODE(last_ins)!=ARM_LDR) 
	{
	  printf("POTENTIAL: NO INLINE 2\n");
	  /*InsPrint(last_ins,0); */
	  flag = FALSE;
	  break;
	}
      }
    }

    if (!flag)
      printf("NO INLINE: PROBLEM WITH RETURN\n");
    else if (nr_callers==1 && ARMINSOPCODE(BBL_INS_LAST(call_site))==ARM_BL /*&& !INS_IS_CONDITIONAL(BBL_INS_LAST(call_site))*/)
    {
      char buffer[10000];
      DiabloPrint(stdout,", %s in %s", i_fun->name, BBL_FUN((call_site))->name); 
      sprintf(buffer,"inline-%s-%s.dot", i_fun->name, BBL_FUN((call_site))->name);
      ArmInlineFunAtCallsite(i_fun,call_site);

/*      Export_Function(BBL_FUN(call_site),buffer);*/
    }
  }	  
  DiabloPrint(stdout,"\n");
}
/* }}} */
/*!
 * \todo Document
 *
 * \param fun
 * \param call_site
 *
 * \return void 
*/
/* ArmInlineFunAtCallsite {{{ */
void ArmInlineFunAtCallsite(t_function * fun, t_bbl *call_site)
{
  t_bbl* entry;
  t_bbl* i_bbl;
  t_cfg_edge * call_edge, * tmp;
  t_cfg_edge * link_edge;  
  t_cfg_edge * return_edge;  
  t_cfg_edge * i_edge, *j_edge;

  t_bbl * tmp_bbl;
  t_bbl * return_point;

  t_bool last_is_ip=FALSE;
  int gain = 0;

  /* Change the call of the last ins into a branch instruction */
  ArmInstructionMakeCondBranch(BBL_INS_LAST(call_site),ARMINSCONDITION(BBL_INS_LAST(call_site)));

  /* Look at all the predecessors of the entry block of the function, to find
   * the call edge */

  entry = fun->bbl_first;
  BBL_FOREACH_PRED_EDGE_SAFE(entry,call_edge,tmp)
  {
    t_bbl* site = CFG_EDGE_HEAD(call_edge);
    if (site!=call_site) continue;
    break;
  }      
  if (!call_edge)
  {
    BBL_FOREACH_PRED_EDGE(entry,call_edge)
      DiabloPrint(stdout,"pred @E\n",call_edge);
    FATAL(("Could not find a call edge for call site @iB",call_site));
  }

  /* Look at all the successor edges of the callsite to find the link edge (and
   * thus the returnsite) */

  BBL_FOREACH_SUCC_EDGE_SAFE(call_site,link_edge,tmp)
  {
    if (EDGE_CAT(link_edge)==ET_FUNLINK || EDGE_CAT(link_edge)==ET_IPFUNLINK)
      break;
  }

  return_edge = EDGE_CORR(call_edge);
  return_point = link_edge?CFG_EDGE_TAIL(link_edge):NULL;



  /* If the link edge is interprocedural we need to remember this (the added
   * branches that replace the returns must be also interprocedural */
  
  if (link_edge && CfgEdgeIsInterproc(link_edge))
     last_is_ip=TRUE;

  /* Kill the edges */
  if (last_is_ip)
    CfgEdgeKill(link_edge->corr_edge);
  if (link_edge) CfgEdgeKill(link_edge);
  CfgEdgeKill(call_edge);
  if (return_edge) CfgEdgeKill(return_edge);

  /* Add the jump that replaces the call */
  
  CfgEdgeNew(fun->cfg,call_site,FUN_BBL_FIRST(fun),ET_JUMP);

  /* Add all blocks but the exit block to the calling function */
  FUN_FOREACH_BBL_SAFE_R(fun,i_bbl,tmp_bbl)
  {
    if (i_bbl!=FunctionGetExitBlock(fun))
    {	
      i_bbl->next_in_fun=call_site->next_in_fun;
      if (call_site->next_in_fun) call_site->next_in_fun->prev_in_fun=i_bbl;
      call_site->next_in_fun = i_bbl;
      i_bbl->prev_in_fun = call_site;
      i_bbl->fun=call_site->fun;
    }
  }

  if (FunctionGetExitBlock(fun)==FUN_BBL_LAST(fun))
  {
    fun->bbl_first = fun->bbl_last;
    fun->bbl_last->prev_in_fun = NULL;
    fun->bbl_last->next_in_fun = NULL;
  }
  else
  {
    fun->bbl_first = fun->bbl_last=NULL;
  }


  /* Look at the predecessors of the exit block, to identify all returnsites */

  if (FUN_BBL_LAST(fun))
  {
    BBL_FOREACH_PRED_EDGE_SAFE(FUN_BBL_LAST(fun),i_edge,tmp)
    {
      t_bbl * head = CFG_EDGE_HEAD(i_edge);
      t_ins * last_ins;
      last_ins = BBL_INS_LAST(head);
      if (!last_ins)
      {
	FunctionPrint(fun,PRINT_INSTRUCTIONS);
	FATAL(("This should be impossible: we encountered a returnsite without instructions. As a return needs an explicit return instruction this is impossible\n"));
      }

      /* Change last instruction of return site */

      /* Moves are easy: */

      if (ARMINSOPCODE(last_ins)==ARM_MOV)
      {
	ArmInstructionMakeCondBranch(last_ins,ARMINSCONDITION(last_ins));
      }

      /* Load multiples:
       * for starters, we assume the stack always grows downward.
       * there are several possibilities:
       * 	- ADS style: ldmia r13!, {...}
       * 	- GCC style: ldmdb rXX, {...,r13,...}
       * 	- possible but not yet encountered: ldmia rXX, {...,r13,...}
       */ 

      else if (ARMINSOPCODE(last_ins)==ARM_LDM)
      {
	t_ins * j_ins;

	/* Add a new instruction for the branch (as we cannot remove the last
	 * instruction) */

	j_ins = InsNewForCfg(fun->cfg);
	GenericSet_uint32(INS_CSIZE(j_ins),4);
	GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
	InsInsertAfter(j_ins,last_ins);

	/* Add the branch */
	ArmInstructionMakeCondBranch(j_ins,ARMINSCONDITION(last_ins));
	gain++;

	/* check for ADS-style instruction */
	if ((ARMINSREGB(last_ins) == ARM_REG_R13) && ArmInsWriteBackHappens(last_ins) && (ARMINSFLAGS(last_ins) & FL_DIRUP))
	{
	  /* we can avoid an indexing instruction by loading the value that
	   * comes from the pc into r14. of course r14 may not be used for
	   * anything else then. */
	  if (!(ARMINSIMMEDIATE(last_ins) & (1<<14)) && !RegsetIn(BBL_REGS_LIVE_OUT(head), ARM_REG_R14))
	  {
	    /* do the r14 trick */
	    ARMINSIMMEDIATE(last_ins) |= (1<<14);
	  }
	  else
	  {
	    /* Add the indexing instruction before the branch, after the LDM */
	    j_ins = InsNewForCfg(fun->cfg);
	    GenericSet_uint32(INS_CSIZE(j_ins),4);
	    GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
	    InsInsertAfter(j_ins,last_ins);
	    ArmInstructionMakeAdd(j_ins,ARMINSREGB(last_ins),ARMINSREGB(last_ins),ARM_REG_NONE,4,ARMINSCONDITION(last_ins));
	    gain++;
	  }
	  ARMINSIMMEDIATE(last_ins) &= ~(1<<15);
	  if(!ARMINSIMMEDIATE(last_ins))
	    InsMakeNoop(last_ins);
	  else
	    INS_REGS_DEF(last_ins)=ArmDefinedRegisters(last_ins);
	}
	/* check for GCC-style instruction */
	else if ((ARMINSREGB(last_ins) != ARM_REG_R13) && !ArmInsWriteBackHappens(last_ins) && !(ARMINSFLAGS(last_ins) & FL_DIRUP) && (ARMINSIMMEDIATE(last_ins) & (1<<13)))
	{
	  /* as we are loading downward in memory, r15 is the first register
	   * that gets loaded now. we can't just not load it because all other
	   * registers would be loaded four bytes too high in memory.
	   * There are two solutions: either add an indexing instruction
	   * _before_ the load or load the stack value in r14 instead of r15
	   */
	  if (!(ARMINSIMMEDIATE(last_ins) & (1<<14)) && !RegsetIn(BBL_REGS_LIVE_OUT(head),ARM_REG_R14))
	  {
	    /* use r14 instead of r15 */
	    ARMINSIMMEDIATE(last_ins) |= (1<<14);
	  }
	  else
	  {
	    /* the safe way: add an indexing instruction before the ldm */
	    j_ins = InsNewForCfg(fun->cfg);
	    GenericSet_uint32(INS_CSIZE(j_ins),4);
	    GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
	    InsInsertBefore(j_ins,last_ins);
	    ArmInstructionMakeSub(j_ins,ARMINSREGB(last_ins),ARMINSREGB(last_ins),ARM_REG_NONE,4,ARMINSCONDITION(last_ins));
	    gain++;
	  }
	  ARMINSIMMEDIATE(last_ins) &= ~(1<<15);
	  INS_REGS_DEF(last_ins)=ArmDefinedRegisters(last_ins);
	}
	else
	{
	  /* This is a kind of return instruction we haven't seen before. */
	  FATAL(("Implement @I as return instruction",last_ins));
	}
      }
      /* LDR's : Remove the LDR, add an indexing instruction if necessary and add the branch */
      else if (ARMINSOPCODE(last_ins)==ARM_LDR)
      {
	if ((ARMINSREGA(last_ins)!=ARM_REG_R15) || (ARMINSREGB(last_ins)!=ARM_REG_R13))
	{
	  BblPrint(INS_BBL(last_ins),0xf);
	  FATAL(("Strange last ins @I!",last_ins));
	}
	else
	{
	  if (ArmInsWriteBackHappens(last_ins))
	  {
	    if (ARMINSFLAGS(last_ins) & FL_DIRUP)
	    {
	      t_uint32 imm=ARMINSIMMEDIATE(last_ins);
	      t_ins * j_ins = InsNewForCfg(fun->cfg);
	      GenericSet_uint32(INS_CSIZE(j_ins),4);
	      GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
	      InsInsertAfter(j_ins,last_ins);
	      ArmInstructionMakeCondBranch(j_ins,ARMINSCONDITION(last_ins));
	      ArmInstructionMakeAdd(last_ins,ARMINSREGB(last_ins),ARMINSREGB(last_ins),ARM_REG_NONE,imm,ARMINSCONDITION(last_ins));
	    }
	    else
	      FATAL(("Implement 1"));
	  }
	  else
	    FATAL(("Implement 2"));
	}
      }
      else  FATAL(("NO GOOD LAST INSTRUCTION\n"));


      /* Kill the return edge, add a (IP)JUMP edge */
      CfgEdgeKill(i_edge);
      if (last_is_ip)
      {
	j_edge = CfgEdgeNew(fun->cfg,head,return_point,ET_IPJUMP);
	CfgEdgeNewCompensating(fun->cfg,j_edge);
      }
      else
      {
	j_edge=  CfgEdgeNew(fun->cfg,head,return_point,ET_JUMP);
      }
      j_edge->cfg = fun->cfg;
    }
    BblKill(fun->bbl_last);
  }

  FunctionKill(fun);

  printf("[gain= %d]",gain);
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
