/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/*#define DEBUG_LS_FWD*/
/*#define EVEN_MORE_UNSAFE*/
/* Includes {{{*/
#include "arm_ls_fwd.h"
#include "arm_utils.h"
#include "arm_instructions.h"
#include "arm_constantoptimizations.h"
#include "arm_description.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_procstate.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_liveness.h"
#include "../../kernel/diablo_addresses.h"
#include "../../kernel/diablo_dominator.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_extendedcopy.h"
#include "../../kernel/diablo_copy_analysis.h"
#include "../../utils/util_options.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../diablo_options.h"
#include <string.h>
#include <strings.h>
/*}}}*/

t_bool worst_case_untraced_load, worst_case_untraced_store;

void AddLiveRegisterBetween(t_bbl * from, t_bbl * to, t_reg reg, int lineno)
{
  t_cfg_edge * edge;
  t_uint32 npreds = 0;
  if (from==to) return;

  BBL_FOREACH_PRED_EDGE(from,edge)
  {
    if(EDGE_CAT(edge) != ET_FUNLINK && EDGE_CAT(edge) != ET_IPFUNLINK)
    {
      if(npreds) break;
      npreds++;
    }
  }
  
  if ((npreds > 1) && !BblDominates(to,from))
  {
    ExportFunctionGraph(BBL_FUN(from)->cfg->sec,"./dots-troubles");
    FATAL(("must update impossible path at line %d",lineno));
  }

  BBL_FOREACH_PRED_EDGE(from,edge)
    {
      if(EDGE_CAT(edge) != ET_FUNLINK && EDGE_CAT(edge) != ET_IPFUNLINK)
      {
	t_bbl * head = CFG_EDGE_HEAD(edge);
	RegsetAddReg(BBL_REGS_LIVE_OUT(head),reg);
	AddLiveRegisterBetween(head,to,reg, lineno);
      }
    }
  
  return;
}


t_uint32 load_count =0, store_count=0, load_tag_count=0, store_tag_count=0;
t_uint32 killed_store=0, changed_load=0, added_moves=0, added_adds=0, changed_loadmultiple=0, changed_storemultiple=0, killed_storemultiple=0, killed_loadmultiple=0, regs_to_load=0;

void ArmUseLSForwarding(t_arm_ins * j_ins, t_arm_ins * i_ins, t_partial_value * propregs, t_bool untraced_load, t_bool untraced_store, t_bool * instruction_is_killed);

/*! 
 *
 * Checks if i_ins accesses the same memory location as all instructions after
 * j_ins in the same bbl. If a pair is found, the function  calls
 * ArmUseLSForwarding, to make use of the results if possible. The base
 * register used in storing/loading is load_reg. If there is a intermediate
 * memory access, that could not be disambiguated, untraced_load is set to
 * TRUE. If intermediate calculations are done on the base address that can be
 * evaluated, it's results are kept in propregs. 
 *
 * \param j_ins Instruction we are currently forwarding TO
 * \param load_reg The register used a base register in the instruction we are forwarding FROM
 * \param propregs Datastructure to keep track of changes to other registers while forwarding
 * \param i_ins The instruction we are currently forwarding FROM
 * \param untraced_load A bool, indicates when an untraced load has happened
 * \param untraced_store A bool, indicates when an untraced store has appeared
 *
 * \return int 
*/

int ArmLoadStoreFwdInner2(t_arm_ins * j_ins, t_reg load_reg, t_partial_value * propregs, t_arm_ins * i_ins, t_bool * untraced_load, t_bool * untraced_store, t_bool * instruction_is_killed)/*{{{*/
{
  t_bool found = FALSE; /* A bool to indicate if we have found an instruction that load/stores to the same memory location */
  t_int32 addend = 0; /* Used in calculating the offset in some addressing modes */
  t_equation equations[max_nr_equations];
#ifdef EVEN_MORE_UNSAFE
  t_reloc * load_tag = NULL, * current_tag = NULL; /* tags used to see if intervening loads definitely overlap or not */
  t_lattice_level current_tag_level;
#endif
  t_lattice_level load_tag_level; 
  t_reg tmpreg;

#if 0 /* REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE*/
  load_tag_level = RegStateGetRegTag(T_INS(i_ins),&INS_REGS_USE(i_ins),ARMINSREGB(i_ins),&load_tag);
#else
  load_tag_level = BOT;
#endif
  while(j_ins != NULL) /* while there are no more instructions in this basic block */
  {
    /* When this instruction conditionally defines the base register, we just can't keep track of things anymore */

    if (INS_IS_CONDITIONAL(j_ins) &&
	RegsetIn(INS_REGS_DEF(j_ins),load_reg))
      return 0;
	
    /* If this instruction is no load or store, we can just skip to the next, except when the base register gets
     * defined in this instruction. We will then try to partially evaluate the instruction {{{*/
    if (INS_TYPE(j_ins) != IT_LOAD && INS_TYPE(j_ins) != IT_STORE && INS_TYPE(j_ins) != IT_STORE_MULTIPLE && INS_TYPE(j_ins) != IT_LOAD_MULTIPLE)
    {
      /* Is the base register defined? */
      if (RegsetIn(INS_REGS_DEF(j_ins),load_reg))
      {
	register_content tmp_offset;
	t_reg i_reg;

	tmp_offset.i = NULL_GENERIC;

	/* Try to track changes */
	if (ArmPartialEvaluation(j_ins,&tmp_offset,load_reg)) return 0;

	/* Add the offset to every register in propregs */
	if (!GENERICISNULL(tmp_offset.i))
	  REGSET_FOREACH_REG(propregs->registers,i_reg) propregs->offset[i_reg] += *(t_int32*)(&G_T_UINT32(tmp_offset.i));
      }
    }
    /*}}}*/

    /* This instruction is a load or a store, see if it uses the same base register or an alias of it */
    if(INS_TYPE(j_ins) == IT_LOAD || INS_TYPE(j_ins) == IT_STORE)
    {
      if(ARMINSREGB(j_ins) != load_reg)
      /*{{{*/
      {
	t_int32 distance;
	t_tristate alias = PERHAPS;

	if(global_options.copy_analysis)
	  BblCopyAnalysisUntilIns(T_INS(j_ins),equations);

	if(!ArmComputeLSOffset(j_ins,&distance))
	{
	  t_reg reg = ARMINSREGA(i_ins);
	  distance += propregs->offset[reg];
#ifdef DEBUG_LS_FWD
	  VERBOSE(1,("Found distance, it is: %d\n",distance));
#endif

	  if(global_options.copy_analysis)
	    {
	      /*	      DiabloPrint(stdout,"@I\n@I\n",i_ins,j_ins);*/
	      alias = EquationsRegsEqual(equations,load_reg,ARMINSREGB(j_ins),distance,NULL,NULL);
	    }
	  
	}

	if(alias == YES)
	{
	  found = TRUE;
	  /*DiabloPrint(stdout,"We can use copy info here!(Alias = TRUE)\n");*/
	  break;
	  /* We have certainly found a pair */
	}
	else if (alias == NO)
	{
	  /* We certainly did not find a pair */
	  /*DiabloPrint(stdout,"We can use copy info here!(Alias = FALSE)\n");*/
	  if (ARMINSREGA(j_ins) == load_reg && INS_TYPE(j_ins) == IT_LOAD) return 0;
	}
	else
	  /* It is unknown where this instruction accesses memory */
	{
	  /* If the load_reg gets overwritten, then we cannot continue anymore */
	  if (RegsetIn(INS_REGS_DEF(j_ins),load_reg)) return 0;

	  /* If the first ins (i_ins) was a store, we want to know if there is a load
	   * from memory which we couldn't trace, because if this is so, the store should
	   * never be removed */
#ifndef EVEN_MORE_UNSAFE
	  if(INS_TYPE(j_ins) == IT_LOAD) *untraced_load = TRUE;
	  else *untraced_store = TRUE;
#else
	  if (INS_TYPE(j_ins) == IT_LOAD && INS_TYPE(i_ins) == IT_STORE)
	  {
	    if(load_reg != ARM_REG_R13 && ARMINSREGB(j_ins) != ARM_REG_R13)
	    {
	      if (load_tag_level)
		*untraced_load = TRUE;
#if 0 /* REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE*/
	      else if ((current_tag_level = RegStateGetRegTag(T_INS(j_ins),&INS_REGS_USE(j_ins),ARMINSREGB(j_ins),&current_tag)))
#else 
		else if ((current_tag_level = BOT))
#endif
		*untraced_load = TRUE;
	      else if (REL_TO_RELOCATABLE(load_tag) == REL_TO_RELOCATABLE(current_tag))
		*untraced_load = TRUE;
	    }
	  }
	  /* If the first ins (i_ins) was a load, we want to know if there is a store
	   * to memory that cannot be traced. If this is so, we should never remove a load
	   * following the first load */
	  if (INS_TYPE(j_ins) == IT_STORE)
	  {
	    if(load_reg != ARM_REG_R13 && ARMINSREGB(j_ins) != ARM_REG_R13)
	    {
	      if (load_tag_level)
		*untraced_store = TRUE;
#if 0 /* REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE*/
	      else if ((current_tag_level = RegStateGetRegTag(T_INS(j_ins),&INS_REGS_USE(j_ins),ARMINSREGB(j_ins),&current_tag)))
#else 
		else if ((current_tag_level = BOT))
#endif
		*untraced_store = TRUE;
	      else if (REL_TO_RELOCATABLE(load_tag) == REL_TO_RELOCATABLE(current_tag))
		*untraced_store = TRUE;
	    }
	  }
#endif
	}
      }/*}}}*/
      else /* ARMINSREGB(j_ins) == load_reg */
      /*{{{*/
      {
	/* try to find out at which offset from the load_reg the memory is accessed,
	 * return if the offset is not computable */

	if (!ArmComputeLSOffset(j_ins,&addend)) 
	{
	  if (addend != 0 && (ARMINSFLAGS(j_ins) & FL_PREINDEX))
	  {
	    REGSET_FOREACH_REG(propregs->registers,tmpreg) propregs->offset[tmpreg] += addend;
	  }

	  REGSET_FOREACH_REG(propregs->registers,tmpreg)
	  {
	    if (propregs->offset[tmpreg] == 0) /* then we're accessing the same memory location */
	    {
	      if (found == FALSE) found = TRUE;
	      else
		/* FATAL because we're not dealing with multiple loads or stores */
		FATAL(("Strange load or store\n1)@I\n2)@I\n",i_ins,j_ins));
	      propregs->corresponding[tmpreg] = ARMINSREGA(j_ins);
	    }
	    else if (propregs->offset[tmpreg] <= 3 && propregs->offset[tmpreg] >= -3 && (ARMINSOPCODE(i_ins) == ARM_LDR || ARMINSOPCODE(i_ins) == ARM_STR || ARMINSOPCODE(j_ins) == ARM_LDR || ARMINSOPCODE(j_ins) == ARM_STR))
	    {
	      /* we are still accessing an overlapping memory location */
	      return 0;
	    }
	  }
	  /* see if we have to restore the offset, when there was no writeback{{{ */
	  if ( addend != 0 && (!(ARMINSFLAGS(j_ins) & FL_WRITEBACK) && (ARMINSFLAGS(j_ins) & FL_PREINDEX))) 
	  {
	    /*if(ARMINSFLAGS(j_ins) & FL_DIRUP)
	      {*/
	    REGSET_FOREACH_REG(propregs->registers,tmpreg) propregs->offset[tmpreg] -= addend;
	    /*}
	      else 
	      {
	      REGSET_FOREACH_REG(propregs->registers,tmpreg) propregs->offset[tmpreg] += addend;
	      }*/
	  }
	  /*}}}*/
	  if (found == TRUE) break;
	  if (RegsetIn(INS_REGS_DEF(j_ins),load_reg)) return 0; 
	}
	else /* Offset was not computable, we indicate that there was
		an untraced load or store */
	{
	  /* There is an offset in a register and there is writeback: we cannot go on */
	  if(RegsetIn(INS_REGS_DEF(j_ins),load_reg)) return 0;

	  /* There is no writeback */
	  if(INS_TYPE(j_ins) == IT_LOAD) *untraced_load = TRUE;
	  else *untraced_store = TRUE;
	}

      }/*}}}*/
    }

    if(INS_TYPE(j_ins) == IT_LOAD_MULTIPLE || INS_TYPE(j_ins) == IT_STORE_MULTIPLE)
    {
      t_regset tmp_set = NullRegs;

      RegsetAddMultipleRegs(tmp_set,ARMINSIMMEDIATE(j_ins));
      /* If load_reg is loaded, we cannot continue anymore */
      if(RegsetIn(tmp_set,load_reg) && INS_TYPE(j_ins) == IT_LOAD_MULTIPLE) return 0;
      if(ARMINSREGB(j_ins) != load_reg)
      {
	t_int32 difference;
	t_tristate known = PERHAPS;

	if(global_options.copy_analysis)
	{
	  BblCopyAnalysisUntilIns(T_INS(j_ins),equations);
/*          DiabloPrint(stdout,"CA until @I\n",j_ins);*/
/*          DiabloPrint(stdout,"Checking equations for @I and @I\n",i_ins,j_ins);*/
	  known = EquationsRegsDiffer(equations,load_reg,ARMINSREGB(j_ins),&difference);
	}
	if(known == YES)
	{
	  t_uint16 i = 0;
	  t_int16 i_reg;
	  t_reg j_reg;
#ifdef DEBUG_LS_FWD
	  VERBOSE(1,("We know a lot %d!\n",difference));
#endif
	  if(ARMINSFLAGS(j_ins) & FL_DIRUP)
	  {
	    if (ARMINSFLAGS(j_ins) & FL_PREINDEX) i++;
	    REGSET_FOREACH_REG(tmp_set,i_reg)
	    {
	      REGSET_FOREACH_REG(propregs->registers,j_reg)
	      {
		if (propregs->offset[j_reg] + (i*4) + difference == 0)
		{
		  propregs->corresponding[j_reg] = i_reg;
		  found = TRUE;
/*                  printf("Correspondig %d, %d",j_reg,i_reg);*/
		}
	      }
	      i++;
	    }
	    if (found == TRUE)
	    {
/*              DiabloPrint(stdout,"1:And we use it! (%d) @I, @I\n",difference,i_ins,j_ins);*/
	      break;
	    }
	  }
	  else
	  {
	    if (ARMINSFLAGS(j_ins) & FL_PREINDEX) i++;
	    REGSET_FOREACH_REG_INVERSE(tmp_set,i_reg)
	    {
	      REGSET_FOREACH_REG(propregs->registers,j_reg)
	      {
		if (propregs->offset[j_reg] - (i*4) + difference == 0)
		{
		  propregs->corresponding[j_reg] = i_reg;
		  found = TRUE;
/*                  printf("Corresponding2 %d, %d",j_reg,i_reg);*/
		}
	      }
	      i++;
	    }
	    if (found == TRUE) 
	    {
/*              DiabloPrint(stdout,"2:And we use it! (%d) @I, @I\n",difference,i_ins,j_ins);*/
	      break;
	    }
	  }
/*          DiabloPrint(stdout,"Skipping a multiple because of disambiguation (1)!\n");*/
	}
	else if (known == NO)
	  /* The tags are different so we will not have aliases. */
	{
/*          DiabloPrint(stdout,"Skipping a multiple because of disambiguation (2)!\n");*/
	}
	else /* known == PERHAPS */
	{
#ifndef EVEN_MORE_UNSAFE
	  if(INS_TYPE(j_ins) == IT_LOAD_MULTIPLE) *untraced_load = TRUE;
	  else *untraced_store = TRUE;
#endif
	}

      }
      else /* ARMINSREGB(j_ins) == load_reg */
      {
	t_reg j_reg;
	t_bool wb = ARMINSFLAGS(j_ins) & FL_WRITEBACK;
	t_uint16 i = 0;
	t_int16 i_reg;

	if (ARMINSFLAGS(j_ins) & FL_DIRUP)
	{
	  if (ARMINSFLAGS(j_ins) & FL_PREINDEX) i++;
	  REGSET_FOREACH_REG(tmp_set,i_reg)
	  {
	    REGSET_FOREACH_REG(propregs->registers,j_reg)
	    {
	      if (propregs->offset[j_reg] + (i*4) == 0)
	      {
		propregs->corresponding[j_reg] = i_reg;
		found = TRUE;
	      }
	    }
	    i++;
	  }
	  if (ARMINSFLAGS(j_ins) & FL_PREINDEX) i--;
	  if(wb) REGSET_FOREACH_REG(propregs->registers,j_reg) propregs->offset[j_reg] += i*4;
	  if (found == TRUE) break;
	}
	else /* no dirup */
	{
	  if (ARMINSFLAGS(j_ins) & FL_PREINDEX) i++;
	  REGSET_FOREACH_REG_INVERSE(tmp_set,i_reg)
	  {
	    REGSET_FOREACH_REG(propregs->registers,j_reg)
	    {
	      if (propregs->offset[j_reg] - (i*4) == 0)
	      {
		propregs->corresponding[j_reg] = i_reg;
		found = TRUE;
	      }
	    }
	    i++;
	  }
	  if (ARMINSFLAGS(j_ins) & FL_PREINDEX) i--;
	  if(wb) REGSET_FOREACH_REG(propregs->registers,j_reg) propregs->offset[j_reg] -= i*4;
	  if (found == TRUE) break;
	}
      }
    }
    
    /* See if the stored or loaded register(s) get changed in between the two memory operations,
     * if so try to keep track of these changes. With this info, we could avoid a load by undoing the change {{{*/
    /* TODO : rewrite this piece, it's not doing what it should, especially with multiple load/store! */
    if (!RegsetIsMutualExclusive(INS_REGS_DEF(j_ins),propregs->registers))
    {
      t_reg i_reg;
      t_regset tmpregset;
      register_content tmp_offset;
      tmp_offset.i = NULL_GENERIC;

      tmpregset = RegsetNewIntersect(INS_REGS_DEF(j_ins),propregs->registers);

      REGSET_FOREACH_REG(tmpregset,i_reg)
      {
	/* If there is still a value propagated along, we can continue partial evaluation */
	if(!propregs->level[i_reg])
	  propregs->level[i_reg] = ArmPartialEvaluation(j_ins,&tmp_offset,i_reg);
	if(!propregs->level[i_reg]) propregs->value[i_reg] += G_T_UINT32(tmp_offset.i);
	propregs->changed[i_reg] = TRUE;
      }

      /*if(RegsetIn(propregs->registers,ARMINSREGA(i_ins)))
      {
	if(!propregs->level[ARMINSREGA(i_ins)])
	  propregs->level[ARMINSREGA(i_ins)] = ArmPartialEvaluation(j_ins,&tmp_offset);
	if (!propregs->level[ARMINSREGA(i_ins)])
	{
	  propregs->value[ARMINSREGA(i_ins)] += G_T_UINT32(tmp_offset.i);
	}
      }
      REGSET_FOREACH_REG(INS_REGS_DEF(j_ins),i_reg) propregs->changed[i_reg] = TRUE;*/
    }/*}}}*/

    j_ins = T_ARM_INS(INS_NEXT(j_ins));
  }

  /* Didn't find a pair, go on with the next basic block */
  if (!found) return 1;
  
  /* We did find a pair, let's try to use it */
  ArmUseLSForwarding(j_ins,i_ins,propregs,*untraced_load,*untraced_store,instruction_is_killed);

  return 0;
}/*}}}*/

/*!
 * \todo Try to make use of the pairs of loads/stores that are found to be accessing the same memory location.
 * Whether this is possible depends on things such as intervening untraced load or stores, changes to the register
 * that was originally stored or loaded. This info is all contained in propregs, untraced_load and untraced_store.
 *
 * \param j_ins Instruction that is forwarded TO
 * \param i_ins Instruction that is forwarded FROM
 * \param propregs Info about registers that were originally stored or loaded (ie. in i_ins)
 * \param untraced_load Bool that indicates if an untraced load has happened
 * \param untraced_store Bool that indicates if an untraced store has happened
 *
 * \return void 
*/
/* ArmUseLSForwarding {{{*/
void ArmUseLSForwarding(t_arm_ins * j_ins, t_arm_ins * i_ins, t_partial_value * propregs, t_bool untraced_load, t_bool untraced_store, t_bool* instruction_is_killed)
{
  static int teller = 0;
  t_arm_ins * insert_ins;
  t_bool up = FALSE;

#ifdef DEBUG_LS_FWD
  if (global_options.debugcounter < teller) return;
  VERBOSE(0,("Case: %d",teller));
  VERBOSE(0,("Investigating pair @I and @I",i_ins,j_ins)); 
#endif /* DEBUG_LS_FWD */


  if (!untraced_load && ARMINSOPCODE(i_ins) == ARM_STR 
      && ARMINSOPCODE(j_ins) == ARM_STR && !INS_IS_CONDITIONAL(j_ins) 
       && BblPostdominates(INS_BBL(j_ins),INS_BBL(i_ins))
      && (!BBL_LOOPS(INS_BBL(i_ins)) || !BblIsLoopHeader(INS_BBL(i_ins))) )
  /* First case: two stores following each other, with no untraced load in between: kill the first
   * store, since it is useless. However check if the first store has a writeback! */
  /*{{{*/
  {
#ifdef DEBUG_LS_FWD
    DiabloPrint(stdout,"@I before @I\n",i_ins,j_ins);
    DiabloPrint(stdout,"Store killed or changed to dataproc!\n");
#endif /* DEBUG_LS_FWD */
    /* we can safely adjust the first store */
    if ((ARMINSFLAGS(i_ins) & FL_WRITEBACK) || !(ARMINSFLAGS(i_ins) & FL_PREINDEX))
    {
      if (ARMINSFLAGS(i_ins) & FL_DIRUP) up = TRUE;
      if (up) ArmInstructionMakeAdd(T_INS(i_ins),ARMINSREGB(i_ins),ARMINSREGB(i_ins),ARM_REG_NONE,4,ARMINSCONDITION(i_ins));
      else ArmInstructionMakeSub(T_INS(i_ins),ARMINSREGB(i_ins),ARMINSREGB(i_ins),ARM_REG_NONE,4,ARMINSCONDITION(i_ins));

      INS_REGS_DEF(i_ins) = ArmDefinedRegisters(T_INS(i_ins));
      INS_REGS_USE(i_ins) = ArmUsedRegisters(T_INS(i_ins));
      killed_store++;regs_to_load++;
    }
    else
    {
      InsKill(T_INS(i_ins));
      *instruction_is_killed = TRUE;
      killed_store++;regs_to_load++;
    }
    teller++;
    return;
  }/*}}}*/
  else if (!untraced_load && ARMINSOPCODE(i_ins) == ARM_STRB
      && ARMINSOPCODE(j_ins) == ARM_STRB && !INS_IS_CONDITIONAL(j_ins)
      && BblPostdominates(INS_BBL(j_ins),INS_BBL(i_ins))
      && (!BBL_LOOPS(INS_BBL(i_ins)) || !BblIsLoopHeader(INS_BBL(i_ins))) )
 /*{{{*/
  {
#ifdef DEBUG_LS_FWD
      VERBOSE(0,("@I before @I\n",i_ins,j_ins));
      VERBOSE(0,("Store killed or changed to dataproc!\n"));
#endif /* DEBUG_LS_FWD */
    /* we can safely adjust the first store */
    if ((ARMINSFLAGS(i_ins) & FL_WRITEBACK) || !(ARMINSFLAGS(i_ins) & FL_PREINDEX))
    {
      if (ARMINSFLAGS(i_ins) & FL_DIRUP) up = TRUE;
      if (up) ArmInstructionMakeAdd(T_INS(i_ins),ARMINSREGB(i_ins),ARMINSREGB(i_ins),ARM_REG_NONE,1,ARMINSCONDITION(i_ins));
      else ArmInstructionMakeSub(T_INS(i_ins),ARMINSREGB(i_ins),ARMINSREGB(i_ins),ARM_REG_NONE,1,ARMINSCONDITION(i_ins));

      INS_REGS_DEF(i_ins) = ArmDefinedRegisters(T_INS(i_ins));
      INS_REGS_USE(i_ins) = ArmUsedRegisters(T_INS(i_ins));
      killed_store++;regs_to_load++;
    }
    else
    {
      InsKill(T_INS(i_ins));
      *instruction_is_killed = TRUE;
      killed_store++;regs_to_load++;
    }
    teller++;
    return;
  }/*}}}*/
  else if ( (ARMINSOPCODE(i_ins) == ARM_STRB || ARMINSOPCODE(i_ins) == ARM_LDRB)
      && ARMINSOPCODE(j_ins) == ARM_LDRB && !untraced_store)
  /*{{{*/
  {
    if(!propregs->changed[ARMINSREGA(i_ins)])
    {
      if ((ARMINSFLAGS(j_ins) & FL_WRITEBACK) || !(ARMINSFLAGS(j_ins) & FL_PREINDEX))
      {
	insert_ins = InsNewForBbl(INS_BBL(j_ins));
	GenericSet_uint32(INS_CSIZE(insert_ins),4);
	GenericSet_uint32(INS_OLD_SIZE(insert_ins),0);

	InsInsertAfter(T_INS(insert_ins),T_INS(j_ins));
	
	if (ARMINSFLAGS(j_ins) & FL_DIRUP) up = TRUE;
	if (up) ArmInstructionMakeAdd(T_INS(insert_ins), ARMINSREGB(j_ins), ARMINSREGB(j_ins), ARM_REG_NONE, 1, ARMINSCONDITION(j_ins));
	else  ArmInstructionMakeSub(T_INS(insert_ins), ARMINSREGB(j_ins), ARMINSREGB(j_ins), ARM_REG_NONE, 1, ARMINSCONDITION(j_ins));
	INS_REGS_DEF(insert_ins) = ArmDefinedRegisters(T_INS(insert_ins));
	INS_REGS_USE(insert_ins) = ArmUsedRegisters(T_INS(insert_ins));
	added_adds++;
#ifdef DEBUG_LS_FWD
	VERBOSE(0,("Added instruction: @I and",insert_ins));
#endif
      }
#ifdef DEBUG_LS_FWD
      VERBOSE(0,("Changed @I to",j_ins));
#endif
      AddLiveRegisterBetween(INS_BBL(j_ins),INS_BBL(i_ins),ARMINSREGA(i_ins),__LINE__);
      ArmInstructionMakeMov(T_INS(j_ins),ARMINSREGA(j_ins),ARMINSREGA(i_ins),0,ARMINSCONDITION(j_ins));
#ifdef DEBUG_LS_FWD
      VERBOSE(0,("@I\n",j_ins));
#endif
      if (ARMINSREGA(j_ins) != ARMINSREGC(j_ins))
      {
	INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
      }
      else
      {
	INS_REGS_DEF(j_ins) = NullRegs;
	INS_REGS_USE(j_ins) = NullRegs;
      }
      changed_load++;regs_to_load++;
#ifdef DEBUG_LS_FWD
      teller++;
#endif
      return;
    }
      
  }/*}}}*/

  else if (
      (ARMINSOPCODE(i_ins) == ARM_STR 
      || ARMINSOPCODE(i_ins) == ARM_LDR 
      || ARMINSOPCODE(i_ins) == ARM_STM
      || ARMINSOPCODE(i_ins) == ARM_LDM
      )
      && ARMINSOPCODE(j_ins) == ARM_LDR 
      && !untraced_store)
  /* Second case: a load or store followed by a load, with no untraced store in between :
   * get rid of the second load and put a move instead. If necessary (writeback...) insert an extra
   * instruction to adjust the load_reg */
  /*{{{*/
  {
    t_reg tmpreg;

    if(ARMINSOPCODE(i_ins) == ARM_STR || ARMINSOPCODE(i_ins) == ARM_LDR)
      tmpreg = ARMINSREGA(i_ins);
    else
      /* i_ins is a load/store multiple */
      REGSET_FOREACH_REG(propregs->registers,tmpreg)
	if(propregs->corresponding[tmpreg] == ARMINSREGA(j_ins))
	  break;
    /*    VERBOSE(0,("@I, @I, changed? %d",i_ins,j_ins,propregs->changed[tmpreg]));*/
    if(tmpreg >= MAX_REG_ITERATOR)
    {
      VERBOSE(0,("Why do we get here then? @I , @I",i_ins,j_ins));
      return;
    }

    if(!propregs->changed[tmpreg])
      /* change the load in a copy instruction if possible */
    {
      if ((ARMINSFLAGS(j_ins) & FL_WRITEBACK) || !(ARMINSFLAGS(j_ins) & FL_PREINDEX))
      {

	insert_ins = InsNewForBbl(INS_BBL(j_ins));
	GenericSet_uint32(INS_CSIZE(insert_ins),4);
	GenericSet_uint32(INS_OLD_SIZE(insert_ins),0);

	InsInsertBefore(T_INS(insert_ins),T_INS(j_ins));

	if (ARMINSFLAGS(j_ins) & FL_DIRUP) up = TRUE;
	if (up) ArmInstructionMakeAdd(T_INS(insert_ins), ARMINSREGB(j_ins), ARMINSREGB(j_ins), ARM_REG_NONE, 4, ARMINSCONDITION(j_ins));
	else  ArmInstructionMakeSub(T_INS(insert_ins), ARMINSREGB(j_ins), ARMINSREGB(j_ins), ARM_REG_NONE, 4, ARMINSCONDITION(j_ins));

	INS_REGS_DEF(insert_ins) = ArmDefinedRegisters(T_INS(insert_ins));
	INS_REGS_USE(insert_ins) = ArmUsedRegisters(T_INS(insert_ins));
	added_adds++;
#ifdef DEBUG_LS_FWD
	VERBOSE(0,("Added instruction: @I and",insert_ins));
#endif
      }
#ifdef DEBUG_LS_FWD
      DiabloPrint(stdout,"Changed @I to",j_ins);
#endif
      /*      DiabloPrint(stdout,"Updating between @iB and @iB\n",INS_BBL(j_ins),INS_BBL(i_ins));*/
      AddLiveRegisterBetween(INS_BBL(j_ins),INS_BBL(i_ins),tmpreg,__LINE__);
      ArmInstructionMakeMov(T_INS(j_ins),ARMINSREGA(j_ins),tmpreg,0,ARMINSCONDITION(j_ins));
#ifdef DEBUG_LS_FWD
      DiabloPrint(stdout,"@I\n",j_ins);
#endif
      if (ARMINSREGA(j_ins) != ARMINSREGC(j_ins))
      {
	INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
      }
      else
      {
	INS_REGS_DEF(j_ins) = NullRegs;
	INS_REGS_USE(j_ins) = NullRegs;
      }
      changed_load++;regs_to_load++;
#ifdef DEBUG_LS_FWD
      teller++;
#endif
      return;
    }
    else
      /* look for a dead register, or try to regenerate the orig value by a calculation */
    {
      t_regset tmp_set = ARM_ALL_BUT_PC_AND_COND;
      t_arm_ins * tmp_ins = i_ins;
      t_int16 i_reg;
      t_cfg_edge * edge;
      t_uint32 realpreds = 0;

      if ((ARMINSFLAGS(j_ins) & FL_WRITEBACK) || !(ARMINSFLAGS(j_ins) & FL_PREINDEX))
      {

	/* since it is not likely to be profitable, we bail out here (BJORN)
	   remove the return to try safely otherwise 
	   */
	VERBOSE(0,("Unprofitable"));
	goto unused;/*return;  */

	insert_ins = InsNewForBbl(INS_BBL(j_ins));
	GenericSet_uint32(INS_CSIZE(insert_ins),4);
	GenericSet_uint32(INS_OLD_SIZE(insert_ins),0);

	InsInsertBefore(T_INS(insert_ins),T_INS(j_ins));

	if (ARMINSFLAGS(j_ins) & FL_DIRUP) up = TRUE;
	if (up) ArmInstructionMakeAdd(T_INS(insert_ins), ARMINSREGB(j_ins), ARMINSREGB(j_ins), ARM_REG_NONE, 4, ARMINSCONDITION(j_ins));
	else  ArmInstructionMakeSub(T_INS(insert_ins), ARMINSREGB(j_ins), ARMINSREGB(j_ins), ARM_REG_NONE, 4, ARMINSCONDITION(j_ins));

	INS_REGS_DEF(insert_ins) = ArmDefinedRegisters(T_INS(insert_ins));
	INS_REGS_USE(insert_ins) = ArmUsedRegisters(T_INS(insert_ins));
	added_adds++;
#ifdef DEBUG_LS_FWD
	VERBOSE(0,("Added instruction: @I and",insert_ins));
#endif
      }

      BBL_FOREACH_PRED_EDGE(INS_BBL(j_ins),edge)
      {
	if(EDGE_CAT(edge) == ET_FUNLINK || EDGE_CAT(edge) == ET_IPFUNLINK) continue;
	/*	if(EDGE_CAT(edge) == ET_SWITCH) continue;*/
	if(CFG_EDGE_HEAD(edge)==INS_BBL(i_ins))
	{
	  /* this test needs to be inserted as long as liveness
	     exploits calling conventions */
	  if (global_options.rely_on_calling_conventions)
	    if (EDGE_CAT(edge)==ET_CALL)
	    {
	      if (FUN_FLAGS(BBL_FUN(INS_BBL(j_ins))) & FF_IS_EXPORTED)
	      {
		RegsetSetIntersect(tmp_set,arm_description.callee_may_use);
	      }
	    }
	  continue;
	}
	if(realpreds) break;
	realpreds++;
      }

      RegsetSetDiff(tmp_set,InsRegsLiveAfter(T_INS(i_ins)));

      if(!RegsetIsEmpty(tmp_set))
      {
	if ( INS_BBL(j_ins) == INS_BBL(i_ins))
	{
	  tmp_ins = T_ARM_INS(INS_NEXT(i_ins));
	  while(tmp_ins != j_ins && !RegsetIsEmpty(tmp_set))
	  {
	    RegsetSetDiff(tmp_set,INS_REGS_DEF(tmp_ins));
	    tmp_ins = T_ARM_INS(INS_NEXT(tmp_ins));
	  }
	}
	else if (!realpreds)
	  /* There is only one incoming edge */
	{
	  tmp_ins = T_ARM_INS(INS_NEXT(i_ins));
	  while(tmp_ins && !RegsetIsEmpty(tmp_set))
	  {
	    RegsetSetDiff(tmp_set,INS_REGS_DEF(tmp_ins));
	    tmp_ins = T_ARM_INS(INS_NEXT(tmp_ins));
	  }

	  if(!tmp_ins) 
	  {
	    tmp_ins = T_ARM_INS(BBL_INS_FIRST(INS_BBL(j_ins)));
	  }

	  while(tmp_ins != j_ins && !RegsetIsEmpty(tmp_set))
	  {
	    RegsetSetDiff(tmp_set,INS_REGS_DEF(tmp_ins));
	    tmp_ins = T_ARM_INS(INS_NEXT(tmp_ins));
	  }
#ifdef DEBUG_LS_FWD
	  VERBOSE(0,("Crossed the basic block border!"));
#endif
	}
	else
	{
#ifdef DEBUG_LS_FWD
	  if(BblDominates(INS_BBL(i_ins),INS_BBL(j_ins)))
	    VERBOSE(0,("Use dominator info in LS! @I, @I",i_ins,j_ins));
#endif
	  goto unused;/*return;*/ /* \todo TODO: use dominator info here! */
	}
      }

      REGSET_FOREACH_REG(tmp_set,i_reg) break;

      if(!RegsetIsEmpty(tmp_set) ) /* We have found a dead register that we can use */
      {
	RegsetAddReg(BBL_REGS_LIVE_OUT(INS_BBL(i_ins)),i_reg);
	/*if (teller++ < global_options.debugcounter)*/
	{
	  /*DiabloPrint(stdout,"-----------------------------------------------------------------\n");*/
	  /*DiabloPrint(stdout,"going with r%d\n",i_reg);*/
	  /*DiabloPrint(stdout,"i_ins: @I\nj_ins: @I\n",i_ins,j_ins);*/

	  if (i_reg != tmpreg)
	  {
	    tmp_ins = InsNewForBbl(INS_BBL(i_ins));
	    GenericSet_uint32(INS_CSIZE(tmp_ins),4);
	    GenericSet_uint32(INS_OLD_SIZE(tmp_ins),4);

	    ArmInstructionMakeMov(T_INS(tmp_ins),i_reg,tmpreg,0,ARMINSCONDITION(i_ins));

	    InsInsertAfter(T_INS(tmp_ins),T_INS(i_ins));
	    INS_REGS_DEF(tmp_ins) = ArmDefinedRegisters(T_INS(tmp_ins));
	    INS_REGS_USE(tmp_ins) = ArmUsedRegisters(T_INS(tmp_ins));
	    added_moves++;
#ifdef DEBUG_LS_FWD
	    DiabloPrint(stdout,"Added @I\n", tmp_ins);
#endif
	  }

	  if (ARMINSREGA(j_ins) != i_reg)
	  {
#ifdef DEBUG_LS_FWD
	    VERBOSE(0,("@I, Changed @I",i_ins,j_ins));
#endif
	    AddLiveRegisterBetween(INS_BBL(j_ins),INS_BBL(i_ins),i_reg,__LINE__);
	    ArmInstructionMakeMov(T_INS(j_ins),ARMINSREGA(j_ins),i_reg,0,ARMINSCONDITION(j_ins));
	    INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	    INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
#ifdef DEBUG_LS_FWD
	    VERBOSE(0,(" to @I using a dead register",j_ins));
#endif
	    changed_load++;regs_to_load++;
	    /*DiabloPrint(stdout,"Changed second load to @I\n", j_ins);*/
	  }
	  else
	  {
#ifdef DEBUG_LS_FWD
	    DiabloPrint(stdout,"Removed @I as it became idempotent\n", j_ins);
#endif
	    InsKill(T_INS(j_ins));
	    *instruction_is_killed = TRUE;
	    changed_load++;regs_to_load++;
	  }
	}
#ifdef DEBUG_LS_FWD
	teller++;
#endif
	return;
      }
      else if (!propregs->level[tmpreg] && ArmIsEncodableConstantForOpcode(propregs->value[tmpreg],ARM_ADD) && BblDominates(INS_BBL(i_ins),INS_BBL(j_ins)))
	/* We do not have e dead register available, but if we know that we can calculate
	 * the value to be loaded, by simply adding a constant
	 * to a register, we change the load to an add */
      {
	DiabloPrint(stdout,"Before partial eval : @I, @I\n",i_ins,j_ins);
	if ((ARMINSFLAGS(j_ins) & FL_WRITEBACK) || !(ARMINSFLAGS(j_ins) & FL_PREINDEX))
	{
	  insert_ins = InsNewForBbl(INS_BBL(j_ins));
	  GenericSet_uint32(INS_CSIZE(insert_ins),4);
	  GenericSet_uint32(INS_OLD_SIZE(insert_ins),0);

	  InsInsertAfter(T_INS(insert_ins),T_INS(j_ins));

	  if (ARMINSFLAGS(j_ins) & FL_DIRUP) up = TRUE;
	  if (up) ArmInstructionMakeAdd(T_INS(insert_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4,ARMINSCONDITION(j_ins));
	  else  ArmInstructionMakeSub(T_INS(insert_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4,ARMINSCONDITION(j_ins));

	  INS_REGS_DEF(insert_ins) = ArmDefinedRegisters(T_INS(insert_ins));
	  INS_REGS_USE(insert_ins) = ArmUsedRegisters(T_INS(insert_ins));
	  added_adds++;
	}
	ArmInstructionMakeAdd(T_INS(j_ins),ARMINSREGA(j_ins),tmpreg,ARM_REG_NONE,propregs->value[i_reg],ARMINSCONDITION(j_ins));
	if (ARMINSIMMEDIATE(j_ins))
	{
	  INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	  INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
	}
	else
	{
	  INS_REGS_DEF(j_ins) = NullRegs;
	  INS_REGS_USE(j_ins) = NullRegs;
	}
	changed_load++;regs_to_load++;
	DiabloPrint(stdout,"After partial eval : @I\n",j_ins);
#ifdef DEBUG_LS_FWD
	teller++;
#endif
	return;
      }
#ifdef DEBUG_LS_FWD
      else
      {
	VERBOSE(0,("No dead register found to propagate value"));
	goto unused;
      }
#endif
    }
  }/*}}}*/

  else if (ARMINSOPCODE(i_ins) == ARM_STM && (ARMINSOPCODE(j_ins) == ARM_LDM || ARMINSOPCODE(j_ins)== ARM_LDR) )
  /* Third case: a store multiple followed by a load : depending on the fact that all registers
   * get restored(restored,nr_diff) and none is written in the mean time(nr_dirty,clean) we can kill one
   * or both instructions, or change the load into a move */
/*{{{*/
  {
    t_regset tmp_set1 = NullRegs;
    t_reg tmp_reg;
    t_bool restored = TRUE, clean = TRUE, count = TRUE;
    t_uint16 restore_value;
    t_int8 diff = 0; /* is the number of regs in the stm and ldm/ldr the same? */
    t_uint8 nr_diff=0; /* if not restored, how many regs are (inter)changed? */
    t_uint8 nr_dirty=0; /* how many aren't clean? */ 

    /*    DiabloPrint(stdout,"Now using LS-info :\n1)@I\n2)@I\n",i_ins,j_ins); */

    if (ARMINSOPCODE(j_ins) == ARM_LDM)
      RegsetAddMultipleRegs(tmp_set1,ARMINSIMMEDIATE(j_ins));
    else
      RegsetAddReg(tmp_set1,ARMINSREGA(j_ins));
    
    if ((diff = (RegsetCountRegs(propregs->registers) - RegsetCountRegs(tmp_set1))))
      count = FALSE;

    REGSET_FOREACH_REG(propregs->registers,tmp_reg)
    {
      if (propregs->corresponding[tmp_reg] != tmp_reg/* && propregs->corresponding[tmp_reg] != MAX_REG_ITERATOR*/)
      {
	if(propregs->corresponding[tmp_reg] == MAX_REG_ITERATOR && diff == 0) return;
	restored = FALSE;
	nr_diff++;
      }
      if (propregs->changed[tmp_reg])
      {
	clean = FALSE;
	nr_dirty++;
      }
    }
    if (ARMINSOPCODE(j_ins) == ARM_LDR)
    {
      REGSET_FOREACH_REG(propregs->registers,tmp_reg)
	if (propregs->corresponding[tmp_reg] == ARMINSREGA(j_ins)) break;
      if (!propregs->changed[tmp_reg]) clean = TRUE;
      else clean = FALSE;
      restored = FALSE;
      nr_diff = 1;
    }

    if (clean && restored ) 
    {
/*The load will surely disappear, the store only if we have a stack operation{{{*/
#ifdef DEBUG_LS_FWD
      VERBOSE(0,("Clean & restored :\n1)@I\n2)@I",i_ins,j_ins));
#endif
      /* TODO: if dominator AND postdominator info is available, we can do a little more */
      if (!diff && ARMINSREGB(j_ins) == ARM_REG_R13 && INS_BBL(j_ins) == INS_BBL(i_ins))
      {
	/* kill em both */
#ifdef DEBUG_LS_FWD
	VERBOSE(0,("Killed @I and @I",i_ins,j_ins));
#endif
	InsKill(T_INS(i_ins));
	*instruction_is_killed = TRUE;
	killed_storemultiple++;
	InsKill(T_INS(j_ins));
	killed_loadmultiple++;
	regs_to_load += 2* RegsetCountRegs(propregs->registers);
	teller++;
	return;
      }
      else
      {
	/* change the load into a sub or add */
	restore_value = RegsetCountRegs(tmp_set1);
	if (ARMINSFLAGS(j_ins) & FL_WRITEBACK)
	{
	  if (ARMINSFLAGS(j_ins) & FL_DIRUP) up = TRUE;
	  if (up) ArmInstructionMakeAdd(T_INS(j_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,restore_value * 4,ARMINSCONDITION(j_ins));
	  else ArmInstructionMakeSub(T_INS(j_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,restore_value * 4,ARMINSCONDITION(j_ins));
	  INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	  INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
#ifdef DEBUG_LS_FWD
	  VERBOSE(0,("Changed to : @I!!!!!!!!",j_ins));
#endif
	  changed_loadmultiple++;
	  regs_to_load += RegsetCountRegs(propregs->registers);

	}
	else
	{
#ifdef DEBUG_LS_FWD
	  VERBOSE(0,("1:Killed @I",j_ins));
#endif
	  InsKill(T_INS(j_ins));
	  *instruction_is_killed = TRUE;
	  killed_loadmultiple++;
	  regs_to_load += RegsetCountRegs(propregs->registers);
	}
#ifdef DEBUG_LS_FWD
	teller++;
#endif
	return;
      }/*}}}*/
    }
    else if (clean && !restored)
    {
/* In this case, we need to copy a reg into another, but the load can surely disappear{{{*/
#ifdef DEBUG_LS_FWD
      VERBOSE(0,("Clean but not restored: \n1)@I\n2)@I",i_ins,j_ins));
#endif
      if (nr_diff == 1)
      {
	t_reg tmpreg = MAX_REG_ITERATOR;
	
	/*REGSET_FOREACH_REG(propregs->registers,tmpreg)
	{
	  DiabloPrint(stdout,"%d corresponds to %d\n",tmpreg,propregs->corresponding[tmpreg]);
	}*/

	/* find out which register has not been restored */
	REGSET_FOREACH_REG(propregs->registers,tmpreg)
	{
	  if (propregs->corresponding[tmpreg] != MAX_REG_ITERATOR && ARMINSOPCODE(j_ins) == ARM_LDR) break;
	  if (propregs->corresponding[tmpreg] != tmpreg && propregs->corresponding[tmpreg] != MAX_REG_ITERATOR) break;
	}

	if (!diff && ARMINSREGB(j_ins) == ARM_REG_R13)
	{

	  if(tmpreg >= MAX_REG_ITERATOR) FATAL(("Have to break somewhere! @I, @I",i_ins, j_ins)); 
	  if(BblPostdominates(INS_BBL(j_ins),INS_BBL(i_ins)))
	  {
	    printf("Killing the store!\n");
	    InsKill(T_INS(i_ins));
	    *instruction_is_killed = TRUE;
	    killed_storemultiple++;
	    regs_to_load += RegsetCountRegs(propregs->registers);
	  }
	  else
	  {
	    return; /* lets return for now, until postdominators works fine again */
	    if(ARMINSOPCODE(j_ins) == ARM_LDM && (ARMINSFLAGS(j_ins) & FL_WRITEBACK))
	    {
	      insert_ins = InsNewForBbl(INS_BBL(j_ins));
	      GenericSet_uint32(INS_CSIZE(insert_ins),4);
	      GenericSet_uint32(INS_OLD_SIZE(insert_ins),0);

	      InsInsertBefore(T_INS(insert_ins),T_INS(j_ins));

	      if (ARMINSFLAGS(j_ins) & FL_DIRUP) up = TRUE;
	      if (up) ArmInstructionMakeAdd(T_INS(insert_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4 * RegsetCountRegs(tmp_set1),ARMINSCONDITION(j_ins));
	      else ArmInstructionMakeSub(T_INS(insert_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4 * RegsetCountRegs(tmp_set1),ARMINSCONDITION(j_ins));

	      INS_REGS_DEF(insert_ins) = ArmDefinedRegisters(T_INS(insert_ins));
	      INS_REGS_USE(insert_ins) = ArmUsedRegisters(T_INS(insert_ins));
	      added_adds++;
	    }
	  }
	  /* In this case we usually have a function return, like LDMIA 13!,{r4,pc} */
	  ArmInstructionMakeMov(T_INS(j_ins),propregs->corresponding[tmpreg],tmpreg,0,ARMINSCONDITION(j_ins));
	  INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	  INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
	  changed_loadmultiple++;
	  regs_to_load += RegsetCountRegs(propregs->registers);
#ifdef DEBUG_LS_FWD
	  teller++;
#endif
	  return;
	}
	else
	{
	  if (ARMINSOPCODE(j_ins) == ARM_LDM && (ARMINSFLAGS(j_ins) & FL_WRITEBACK))
	  {
	    insert_ins = InsNewForBbl(INS_BBL(j_ins));
	    GenericSet_uint32(INS_CSIZE(insert_ins),4);
	    GenericSet_uint32(INS_OLD_SIZE(insert_ins),0);

	    InsInsertBefore(T_INS(insert_ins),T_INS(j_ins));

	    if (ARMINSFLAGS(j_ins) & FL_DIRUP) up = TRUE;
	    if (up) ArmInstructionMakeAdd(T_INS(insert_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4 * RegsetCountRegs(tmp_set1),ARMINSCONDITION(j_ins));
	    else ArmInstructionMakeSub(T_INS(insert_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4 * RegsetCountRegs(tmp_set1),ARMINSCONDITION(j_ins));

	    INS_REGS_DEF(insert_ins) = ArmDefinedRegisters(T_INS(insert_ins));
	    INS_REGS_USE(insert_ins) = ArmUsedRegisters(T_INS(insert_ins));
	    added_adds++;
	  }
	  else if ((ARMINSFLAGS(j_ins) & FL_WRITEBACK) || !(ARMINSFLAGS(j_ins) & FL_PREINDEX))
	  {
	    insert_ins = InsNewForBbl(INS_BBL(j_ins));
	    GenericSet_uint32(INS_CSIZE(insert_ins),4);
	    GenericSet_uint32(INS_OLD_SIZE(insert_ins),0);

	    InsInsertAfter(T_INS(insert_ins),T_INS(j_ins));

	    if (ARMINSFLAGS(j_ins) & FL_DIRUP) up = TRUE;
	    if (up) ArmInstructionMakeAdd(T_INS(insert_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4 ,ARMINSCONDITION(j_ins));
	    else ArmInstructionMakeSub(T_INS(insert_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4 ,ARMINSCONDITION(j_ins));

	    INS_REGS_DEF(insert_ins) = ArmDefinedRegisters(T_INS(insert_ins));
	    INS_REGS_USE(insert_ins) = ArmUsedRegisters(T_INS(insert_ins));
	    added_adds++;
#ifdef DEBUG_LS_FWD
	    DiabloPrint(stdout,"Inserted @I\n",insert_ins);
#endif
	  }
	  if(tmpreg != MAX_REG_ITERATOR)
	  {


#ifdef DEBUG_LS_FWD
	  DiabloPrint(stdout,"Changed @I",j_ins);
#endif
	  AddLiveRegisterBetween(INS_BBL(j_ins),INS_BBL(i_ins),tmpreg,__LINE__);
	  ArmInstructionMakeMov(T_INS(j_ins),propregs->corresponding[tmpreg],tmpreg,0,ARMINSCONDITION(j_ins));
	  if (ARMINSREGA(j_ins) != ARMINSREGC(j_ins))
	  {
	    INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	    INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
	  }
	  else
	  {
	    INS_REGS_DEF(j_ins) = NullRegs;
	    INS_REGS_USE(j_ins) = NullRegs;
	  }
	  if (ARMINSOPCODE(j_ins) == ARM_LDM)
	  {
	    changed_loadmultiple++;
	    regs_to_load += RegsetCountRegs(propregs->registers);
	  }
	  else
	  {
	    changed_load++;
	    regs_to_load++;
	  }
#ifdef DEBUG_LS_FWD
	  DiabloPrint(stdout," to @I\n",j_ins);
#endif
	  }
	  else InsKill(T_INS(j_ins));
#ifdef DEBUG_LS_FWD
	  teller++;
#endif
	  return;
	}
      }
      /*{{{Contains bugs, don't use!*/
      /*else if (nr_diff == 2)
      {
	t_reg i_reg;
	t_bool do_it = TRUE;
	REGSET_FOREACH_REG(propregs->registers,i_reg)
	  if(RegsetIn(propregs->registers,propregs->corresponding[i_reg])) do_it = FALSE;
	if(do_it && !(ARMINSFLAGS(j_ins) & FL_WRITEBACK))
	{
	  REGSET_FOREACH_REG(propregs->registers,i_reg)
	    break;
	  insert_ins = InsNewForBbl(INS_BBL(j_ins));
	  InsInsertAfter(T_INS(insert_ins),T_INS(j_ins));
	  
	  ArmInstructionMakeMov(T_INS(insert_ins),propregs->corresponding[i_reg],i_reg,0,ARMINSCONDITION(j_ins));
	  added_moves++;
	  changed_loadmultiple++;
	  regs_to_load+=2;
	  REGSET_FOREACH_REG_INVERSE(propregs->registers,i_reg)
	    break;
	  ArmInstructionMakeMov(T_INS(j_ins),propregs->corresponding[i_reg],i_reg,0,ARMINSCONDITION(j_ins));
	  VERBOSE(0,("Added @I and @I",insert_ins,j_ins));
	  if(ARMINSREGB(i_ins) == ARM_REG_R13)
	  {
	    VERBOSE(0,("2:Killed @I",i_ins));
	    InsKill(T_INS(i_ins));
	    *instruction_is_killed = TRUE;
	    regs_to_load +=2;
	    killed_storemultiple++;
	  }
	}
      }*//*}}}*/
      else
      {
	DiabloPrint(stdout,"Missing opportunity! (%d) @I, @I\n",nr_diff, i_ins, j_ins);
	goto unused;
      }
      /* add some moves here, requires 1 dead registers in worst case */
    }/*}}}*/
    else if (restored) /* !clean and restored */
    /*In this case, we try to use dead registers if any or a copy instruction if possible{{{*/
    {
      t_regset tmp_set = ARM_ALL_BUT_PC_AND_COND,tmp_set2 = propregs->registers;
      t_arm_ins * tmp_ins = i_ins;
      t_int16 i_reg, j_reg;

#ifdef DEBUG_LS_FWD
      VERBOSE(0,("%d not clean, but restored: \n1)@I\n2)@I",nr_dirty,i_ins,j_ins));
#endif
      /* add some moves here, using dead registers if any */
      /*REGSET_FOREACH_REG(propregs->registers, tmp_reg) DiabloPrint(stdout,"%d) corresponding = %d, %s, offset = %d\n",tmp_reg,propregs->corresponding[tmp_reg],propregs->changed[tmp_reg]?"dirty":"clean",propregs->offset[tmp_reg]); */
      RegsetSetDiff(tmp_set,InsRegsLiveAfter(T_INS(i_ins)));
      REGSET_FOREACH_REG(propregs->registers,i_reg)
	if (propregs->changed[i_reg]) RegsetSubReg(tmp_set2,i_reg);
      RegsetSetDiff(tmp_set,tmp_set2);
      if(!RegsetIsEmpty(tmp_set))
      {
	if ( INS_BBL(j_ins) == INS_BBL(i_ins))
	{
	  tmp_ins = T_ARM_INS(INS_NEXT(i_ins));
	  while(tmp_ins != j_ins && !RegsetIsEmpty(tmp_set))
	  {
	    RegsetSetDiff(tmp_set,INS_REGS_DEF(tmp_ins));
	    tmp_ins = T_ARM_INS(INS_NEXT(tmp_ins));
	  }
	}
	else 
	  goto unused;/*return;*/
      }

      while(!RegsetIsEmpty(tmp_set) && nr_dirty)
      {
#ifdef DEBUG_LS_FWD
	VERBOSE(1,("Dead: "));
	REGSET_FOREACH_REG(tmp_set,i_reg) VERBOSE(1,("%d ",i_reg));
#endif
	REGSET_FOREACH_REG(tmp_set,i_reg) break;
	REGSET_FOREACH_REG(propregs->registers,j_reg)
	  if (propregs->changed[j_reg])
	  {
	    propregs->changed[j_reg] = FALSE;
	    break;
	  }
	/* change the store and load instructions to move's */
	if (nr_dirty == 1)
	{

#ifdef DEBUG_LS_FWD
	  VERBOSE(0,("Changed @I",i_ins));
#endif
	  ArmInstructionMakeMov(T_INS(i_ins),i_reg,j_reg,0,ARMINSCONDITION(i_ins));
	  if (ARMINSREGA(i_ins) != ARMINSREGC(i_ins))
	  {
	    INS_REGS_DEF(i_ins) = ArmDefinedRegisters(T_INS(i_ins));
	    INS_REGS_USE(i_ins) = ArmUsedRegisters(T_INS(i_ins));
	  }
	  else
	  {
	    INS_REGS_DEF(i_ins) = NullRegs;
	    INS_REGS_USE(i_ins) = NullRegs;
	  }
	  changed_storemultiple++;
#ifdef DEBUG_LS_FWD
	  VERBOSE(0,(" to @I",i_ins));
	  VERBOSE(0,("Changed @I",j_ins));
#endif
	  AddLiveRegisterBetween(INS_BBL(j_ins),INS_BBL(i_ins),i_reg,__LINE__);
	  ArmInstructionMakeMov(T_INS(j_ins),j_reg,i_reg,0,ARMINSCONDITION(j_ins));
	  if (ARMINSREGA(j_ins) != ARMINSREGC(j_ins))
	  {
	    INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	    INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
	  }
	  else
	  {
	    INS_REGS_DEF(j_ins) = NullRegs;
	    INS_REGS_USE(j_ins) = NullRegs;
	  }

#ifdef DEBUG_LS_FWD
	  VERBOSE(0,(" to @I",j_ins));
#endif
	  changed_loadmultiple++;
	  regs_to_load += RegsetCountRegs(propregs->registers) - 1;
#ifdef DEBUG_LS_FWD
	  teller++;
#endif
	  return;
	}
	/* add two move's to spill the register */
	else
	{
	  insert_ins = InsNewForBbl(INS_BBL(i_ins));
	  GenericSet_uint32(INS_CSIZE(insert_ins),4);
	  GenericSet_uint32(INS_OLD_SIZE(insert_ins),0);

	  InsInsertBefore(T_INS(insert_ins),T_INS(i_ins));

	  ArmInstructionMakeMov(T_INS(insert_ins),i_reg,j_reg,0,ARMINSCONDITION(i_ins));
	  if (ARMINSREGA(insert_ins) != ARMINSREGC(insert_ins))
	  {
	    INS_REGS_DEF(insert_ins) = ArmDefinedRegisters(T_INS(insert_ins));
	    INS_REGS_USE(insert_ins) = ArmUsedRegisters(T_INS(insert_ins));
	  }
	  else
	  {
	    INS_REGS_DEF(insert_ins) = NullRegs;
	    INS_REGS_USE(insert_ins) = NullRegs;
	  }
	  RegsetSubReg(ARMINSIMMEDIATE(i_ins),j_reg);
#ifdef DEBUG_LS_FWD
	  VERBOSE(0,("One reg less to load\n1)@I\n2)@I",insert_ins,i_ins));
#endif

	  insert_ins = InsNewForBbl(INS_BBL(j_ins));
	  GenericSet_uint32(INS_CSIZE(insert_ins),4);
	  GenericSet_uint32(INS_OLD_SIZE(insert_ins),0);
	  InsInsertBefore(T_INS(insert_ins),T_INS(j_ins));

	  ArmInstructionMakeMov(T_INS(insert_ins),j_reg,i_reg,0,ARMINSCONDITION(j_ins));
	  if (ARMINSREGA(insert_ins) != ARMINSREGC(insert_ins))
	  {
	    INS_REGS_DEF(insert_ins) = ArmDefinedRegisters(T_INS(insert_ins));
	    INS_REGS_USE(insert_ins) = ArmUsedRegisters(T_INS(insert_ins));
	  }
	  else
	  {
	    INS_REGS_DEF(insert_ins) = NullRegs;
	    INS_REGS_USE(insert_ins) = NullRegs;
	  }
	  RegsetSubReg(ARMINSIMMEDIATE(j_ins),j_reg);
	  if(ARMINSIMMEDIATE(j_ins) == 0) FATAL(("Make a noop!"));
	  regs_to_load++;
#ifdef DEBUG_LS_FWD
	  VERBOSE(0,("3)@I\n4)@I",insert_ins,j_ins));
#endif
	}
	nr_dirty--;
	RegsetSubReg(tmp_set,i_reg);
#ifdef DEBUG_LS_FWD
	teller++;
#endif
      }
      if (nr_dirty)
      {
	REGSET_FOREACH_REG(propregs->registers,j_reg)
	  if (!propregs->changed[j_reg])
	  {
	    RegsetSubReg(ARMINSIMMEDIATE(i_ins),j_reg);
	    RegsetSubReg(ARMINSIMMEDIATE(j_ins),j_reg);
	    regs_to_load++;
	  }
	INS_ATTRIB(i_ins) &= IF_CHANGED;
	INS_ATTRIB(j_ins) &= IF_CHANGED;
	INS_REGS_DEF(i_ins) = ArmDefinedRegisters(T_INS(i_ins));
	INS_REGS_USE(i_ins) = ArmUsedRegisters(T_INS(i_ins));
	INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
#ifdef DEBUG_LS_FWD
	VERBOSE(0,("Store and load not gone, now @I and @I",i_ins,j_ins));
#endif
      }
    }
    /*}}}*/
    else
    {
#ifdef DEBUG_LS_FWD
      VERBOSE(0,("%d not clean and %d not restored",nr_dirty,nr_diff));
#endif
      goto unused;
    }
    /*teller++;*/
  }/*}}}*/

  else if (ARMINSOPCODE(i_ins) == ARM_STM && ARMINSOPCODE(j_ins) == ARM_STM && BblPostdominates(INS_BBL(j_ins),INS_BBL(i_ins)))
  /* fourth case : two store multiples following each other, the second overwriting
   * what the first has written. Kill the first one, preserve side-effects */
  /*{{{*/
  {
    t_regset tmp_set1 = NullRegs;
    t_reg tmp_reg;
    t_bool overwrite = TRUE;
    t_int32 diff = 0;

    /*DiabloPrint(stdout,"Now using LS-info :\n1)@I\n2)@I\n",i_ins,j_ins); */
    RegsetAddMultipleRegs(tmp_set1,ARMINSIMMEDIATE(j_ins));
    diff = (RegsetCountRegs(propregs->registers) - RegsetCountRegs(tmp_set1));

    REGSET_FOREACH_REG(propregs->registers,tmp_reg)
    {
      if (propregs->corresponding[tmp_reg] == MAX_REG_ITERATOR) overwrite = FALSE;
    }
    if(!overwrite && diff <= 0) return;/*FATAL(("How is this possible?\n"));*/
    /* TODO: use this info!!!! */

    if (overwrite)
    {
      if (ArmInsWriteBackHappens(T_INS(i_ins)))
      {
#ifdef DEBUG_LS_FWD
        VERBOSE(0,("@I changed because it is immediately overwritten ",i_ins));
#endif
	if (ARMINSFLAGS(i_ins) & FL_DIRUP) up = TRUE;
	if (up) ArmInstructionMakeAdd(T_INS(i_ins),ARMINSREGB(i_ins),ARMINSREGB(i_ins),ARM_REG_NONE,4 * RegsetCountRegs(propregs->registers),ARMINSCONDITION(i_ins));
	else ArmInstructionMakeSub(T_INS(i_ins),ARMINSREGB(i_ins),ARMINSREGB(i_ins),ARM_REG_NONE,4 * RegsetCountRegs(propregs->registers),ARMINSCONDITION(i_ins));

	INS_REGS_DEF(i_ins) = ArmDefinedRegisters(T_INS(i_ins));
	INS_REGS_USE(i_ins) = ArmUsedRegisters(T_INS(i_ins));
        changed_storemultiple++;
	regs_to_load += RegsetCountRegs(propregs->registers);
#ifdef DEBUG_LS_FWD
        VERBOSE(0,("to @I",i_ins));
#endif
      }
      else
      {
#ifdef DEBUG_LS_FWD
	VERBOSE(0,("STM Killed because it is immediately overwritten!!! @I",i_ins));
#endif
	InsKill(T_INS(i_ins));
	*instruction_is_killed = TRUE;
	killed_storemultiple++;
	regs_to_load += RegsetCountRegs(propregs->registers);
      }
#ifdef DEBUG_LS_FWD
      teller++;
#endif
      return;
    }
    

  }/*}}}*/
  else if ((ARMINSOPCODE(i_ins) == ARM_LDR || ARMINSOPCODE(i_ins) == ARM_STR) && ARMINSOPCODE(j_ins) == ARM_LDM)
    /* {{{ */
  {
    t_reg loadreg = ARMINSREGA(i_ins);
    t_regset tmp_set = RegsetNew();
    t_bool first=FALSE,last=FALSE;
    t_reg i_reg,f_reg = ARM_REG_NONE,l_reg=ARM_REG_NONE;

    if(propregs->changed[loadreg]) return;
    if(!(ARMINSFLAGS(j_ins) & FL_WRITEBACK)) return;
    RegsetAddMultipleRegs(tmp_set,ARMINSIMMEDIATE(j_ins));
    REGSET_FOREACH_REG(tmp_set,i_reg)
    {
      if(f_reg == ARM_REG_NONE) f_reg = i_reg;
      l_reg = i_reg;
    }
    if (loadreg == f_reg) first = TRUE;
    if (loadreg == l_reg) last = TRUE;
    if (first || last) VERBOSE(0,("Ok we can do something with this! @I, @I",i_ins,j_ins));
    goto unused;
    
  }/*}}}*/
  else if (ARMINSOPCODE(i_ins) == ARM_LDR && ARMINSOPCODE(j_ins) == ARM_STR && ARMINSREGA(i_ins) == ARMINSREGA(j_ins) && !untraced_store
      && (!BBL_LOOPS(INS_BBL(j_ins)) || BblIsLoopHeader(INS_BBL(j_ins))))
  {
    t_reg loadreg = ARMINSREGA(i_ins);

    if(propregs->changed[loadreg]) goto unused;

    if(!ArmInsWriteBackHappens(T_INS(j_ins)))
    {
#ifdef DEBUG_LS_FWD
	VERBOSE(0,("STM Killed because it is unnecessary!!! @I",j_ins));
#endif
      InsKill(T_INS(j_ins));
      *instruction_is_killed = TRUE;
      regs_to_load++;
      killed_store++;
      teller++;
    }
    else
    {
#ifdef DEBUG_LS_FWD
        VERBOSE(0,("@I changed because it is unnecessary ",j_ins));
#endif
      if(ARMINSFLAGS(j_ins) & FL_DIRUP) ArmInstructionMakeAdd(T_INS(j_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4,ARMINSCONDITION(j_ins));
      else ArmInstructionMakeSub(T_INS(j_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4,ARMINSCONDITION(j_ins));
      regs_to_load++;
      changed_load++;
      teller++;
#ifdef DEBUG_LS_FWD
        VERBOSE(0,("to @I",j_ins));
#endif
    }
    return;
    
  }
  else if(ARMINSOPCODE(i_ins) == ARM_LDM && ARMINSOPCODE(j_ins) == ARM_STM && !untraced_store)
  {
    t_regset tmp_set1 = NullRegs;
    t_reg tmp_reg;
    t_bool do_it = TRUE;
    t_int32 diff = 0;

    RegsetAddMultipleRegs(tmp_set1,ARMINSIMMEDIATE(j_ins));
    diff = (RegsetCountRegs(propregs->registers) - RegsetCountRegs(tmp_set1));
    if(diff) goto unused;/* For the moment we return, but we can do better!*/

    REGSET_FOREACH_REG(propregs->registers,tmp_reg)
    {
      if(propregs->corresponding[tmp_reg] != tmp_reg) do_it = FALSE;
      if(propregs->changed[tmp_reg]) do_it = FALSE;
    }

    if(do_it)
    {
      if(ArmInsWriteBackHappens(T_INS(j_ins)))
      {
#ifdef DEBUG_LS_FWD
        VERBOSE(0,("@I changed because it is unnecessary ",j_ins));
#endif
	if(ARMINSFLAGS(j_ins) & FL_DIRUP) ArmInstructionMakeAdd(T_INS(j_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4 * RegsetCountRegs(propregs->registers),ARMINSCONDITION(j_ins));
	else ArmInstructionMakeSub(T_INS(j_ins),ARMINSREGB(j_ins),ARMINSREGB(j_ins),ARM_REG_NONE,4 * RegsetCountRegs(propregs->registers),ARMINSCONDITION(j_ins));

	INS_REGS_DEF(j_ins) = ArmDefinedRegisters(T_INS(j_ins));
	INS_REGS_USE(j_ins) = ArmUsedRegisters(T_INS(j_ins));
        changed_storemultiple++;
	regs_to_load += RegsetCountRegs(propregs->registers);

#ifdef DEBUG_LS_FWD
        VERBOSE(0,("to @I",j_ins));
#endif
      }
      else
      {
#ifdef DEBUG_LS_FWD
	VERBOSE(0,("STM Killed because it is unnecessary!!! @I",j_ins));
#endif
	InsKill(T_INS(j_ins));
	*instruction_is_killed = TRUE;
	killed_storemultiple++;
	regs_to_load += RegsetCountRegs(propregs->registers);
      }
      teller++;
      return;
    }
  }
#ifdef DEBUG_LS_FWD
  else
#endif
unused:
#ifdef DEBUG_LS_FWD
    VERBOSE(0,("Unused pair @I and @I",i_ins,j_ins));
#endif
  
  return;
}/*}}}*/
/*!
 * \todo Document
 *
 * \param i_bbl
 * \param load_reg
 * \param propregs
 * \param i_ins
 * \param untraced_load
 * \param untraced_store
 *
 * \return void 
*/
/* ArmLoadStoreFwdRecurse {{{ */
void ArmLoadStoreFwdRecurse(t_bbl * i_bbl, t_reg load_reg, t_partial_value * propregs, t_arm_ins * i_ins, t_bool * untraced_load, t_bool * untraced_store, t_bool * instruction_is_killed)
{
  t_cfg_edge * edge,* edge2;
  t_uint32 real_preds;
  t_bool untraced_load_traceback = *untraced_load, untraced_store_traceback = *untraced_store;
  t_partial_value propregs_traceback = *propregs;
  if (NodeIsMarked(i_bbl)) return;

  if(*untraced_store) worst_case_untraced_store = TRUE;
  if(*untraced_load) worst_case_untraced_load = TRUE;

  BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
  {
    t_bool worst_case = FALSE;
    *untraced_load = untraced_load_traceback;
    *untraced_store = untraced_store_traceback;
    *propregs = propregs_traceback;
    EdgeMark(edge);

    /* These edges are no real control flow, skip them, Hell node are useless to propagate to, skip them */
    if ((EDGE_CAT(edge)==ET_FUNLINK) || (EDGE_CAT(edge)==ET_IPFUNLINK) || (EDGE_CAT(edge)==ET_IPSWILINK) || (EDGE_CAT(edge)==ET_SWILINK) || IsHellNode(CFG_EDGE_TAIL(edge))) 
    {
      continue;
    }
    else
    {
      real_preds=0;
      BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge),edge2)
      {
	/* Fun link edges don't count as predecessor edges for load_store fwd */
	if ((EDGE_CAT(edge2)==ET_FUNLINK) || (EDGE_CAT(edge2)==ET_IPFUNLINK)) continue;
	if (edge == edge2) continue;
	if (EdgeIsMarked(edge2))
	{
	  /* We merge info here, used worst case assumptions (ie. when going through a diamondlike piece of the graph) */
	  worst_case = TRUE;
	  continue;
	}
	real_preds++;
	break;
      }
      if (real_preds==0)
      {
	if(worst_case)
	{
	  *untraced_load = worst_case_untraced_load;
	  *untraced_store = worst_case_untraced_store;
	}
	NodeMark(i_bbl);
	if (ArmLoadStoreFwdInner2(T_ARM_INS(BBL_INS_FIRST(CFG_EDGE_TAIL(edge))), load_reg, propregs, i_ins, untraced_load, untraced_store, instruction_is_killed))
	{
	  ArmLoadStoreFwdRecurse(CFG_EDGE_TAIL(edge),load_reg,propregs,i_ins, untraced_load, untraced_store, instruction_is_killed);
	}
      }
    }
  }
}/*}}}*/

/*!
 * \todo Document
 *
 * \param cfg
 *
 * \return void 
*/
/* ArmLoadStoreFwd {{{ */
void ArmLoadStoreFwd(t_cfg* cfg)
{
  t_bbl * i_bbl;
  t_arm_ins * i_ins, * j_ins, * tmp;
  /*t_function * fun;*/
  t_partial_value propregs;
  t_reg load_reg, tmpreg;
  t_bool untraced_load, untraced_store;
#if 0
  t_reloc * dummytag;
#endif
  t_uint16 nr_regs,i;
  
  load_count =0;
  store_count=0;
  load_tag_count=0;
  store_tag_count=0;


  /* we need liveness information and dominator information to do this correctly */
  ComDominators(cfg);
  ComLiveness(cfg,SENSITIVE); 

#if 0/*{{{*/
  CFG_FOREACH_FUN(cfg, fun)
    FUN_FOREACH_BBL(fun, i_bbl)
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	if(INS_TYPE(i_ins) == IT_LOAD)
	{
	  if(!RegsetIn(InsRegsLiveAfter(i_ins),ARMINSREGA(i_ins)))
	    DiabloPrint(stdout,"Why loading? @I\n",i_ins);
	}
	if(INS_TYPE(i_ins) == IT_LOAD_MULTIPLE)
	{
	  t_uint32 tmp = ARMINSIMMEDIATE(i_ins);
	  t_reg i = 0;
	  t_regset tmpset = InsRegsLiveAfter(i_ins);
	  while(tmp)
	  {
	    if (tmp & 0x1)
	    {
	      if(!RegsetIn(tmpset,i))
		DiabloPrint(stdout,"Shouldn't be loading reg %d in @I\n",i,i_ins);
	    }
	    tmp >>= 1;
	    i++;
	  }
	}
      }
#endif/*}}}*/
  ANALYSIS_START("ArmLoadStore");

  CFG_FOREACH_BBL(cfg, i_bbl)
  {
    t_bool instruction_is_killed = TRUE;

    worst_case_untraced_store = worst_case_untraced_load = FALSE;

    if (!BBL_FUN(i_bbl)) continue;

    /*    DiabloPrint(stdout,"In function @F\n@iB",BBL_FUN(i_bbl),i_bbl);*/

    while (instruction_is_killed)
      {
	instruction_is_killed = FALSE;

    /* Safe, because instructions can get killed */
	BBL_FOREACH_INS_SAFE(i_bbl,T_INS(i_ins),T_INS(tmp))
	  {
	    /* First some statistics: count loads and stores {{{*/
	    if (INS_TYPE(i_ins) == IT_LOAD || INS_TYPE(i_ins) == IT_LOAD_MULTIPLE)
	      {
		load_count++;
#if 0 /* REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE*/
		if(!RegStateGetRegTag(T_INS(i_ins),&INS_REGS_USE(i_ins),ARMINSREGB(i_ins),&dummytag)) load_tag_count++;
#else
		if (!BOT) load_tag_count++;
#endif
	      }
	    if (INS_TYPE(i_ins) == IT_STORE || INS_TYPE(i_ins) == IT_STORE_MULTIPLE)
	      {
		store_count++;
#if 0 /* REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE*/
		if(!RegStateGetRegTag(T_INS(i_ins),&INS_REGS_USE(i_ins),ARMINSREGB(i_ins),&dummytag)) store_tag_count++;
#else
		if (!BOT) store_tag_count++;
#endif
	      }
	    /* }}}*/
      
      

	    if (INS_IS_CONDITIONAL(i_ins)) continue;


	    /* TODO: The following could be less restricting, because loading/storing bytes and
	     * halfwords might be included as well, since they will overwrite memory, corrupting what
	     * was there (except when some does tricky things!) */

	    if (ARMINSOPCODE(i_ins) == ARM_STR || 
		ARMINSOPCODE(i_ins) == ARM_LDR ||
		ARMINSOPCODE(i_ins) == ARM_LDRB ||
		ARMINSOPCODE(i_ins) == ARM_STRB )
	      {
		if (ARMINSREGB(i_ins) == ARMINSREGA(i_ins) && INS_TYPE(i_ins) == IT_LOAD)
		  continue; /* Too hard to handle */

		memset(&propregs,0,sizeof(t_partial_value));

		FOREACH_REG(tmpreg) propregs.corresponding[tmpreg] = MAX_REG_ITERATOR;

		if (ArmComputeLSOffset(i_ins,&(propregs.offset[ARMINSREGA(i_ins)]))) continue;

		RegsetAddReg(propregs.registers,ARMINSREGA(i_ins));

		load_reg = ARMINSREGB(i_ins);

		/* Set load_offset so that 'Store_addres used' + load_offset = load_reg content */

		if (propregs.offset[ARMINSREGA(i_ins)] != 0
		    && (ARMINSFLAGS(i_ins) & FL_PREINDEX)
		    && !(ARMINSFLAGS(i_ins) & FL_WRITEBACK) )
		  propregs.offset[ARMINSREGA(i_ins)] = -(propregs.offset[ARMINSREGA(i_ins)]);
		else if (propregs.offset[ARMINSREGA(i_ins)] != 0 && !(ARMINSFLAGS(i_ins) & FL_PREINDEX))
		  { /* nothing to do */ }
		else if (propregs.offset[ARMINSREGA(i_ins)] != 0) propregs.offset[ARMINSREGA(i_ins)] = 0; 

		j_ins = T_ARM_INS(INS_NEXT(i_ins));
		untraced_load = FALSE;
		untraced_store = FALSE;
	
		/*	if(INS_NEXT(i_ins) && ARMINSOPCODE(INS_NEXT(i_ins)) == ARM_STR && INS_NEXT(INS_NEXT(i_ins)) && ARMINSOPCODE(INS_NEXT(INS_NEXT(i_ins))) == ARM_STR)*/
		/*	DiabloPrint(stdout,"Doing @I\n",i_ins);*/

		if (ArmLoadStoreFwdInner2(j_ins, load_reg, &propregs, i_ins, &untraced_load, &untraced_store,&instruction_is_killed))
		  {
		    InitNodeMarking();
		    InitEdgeMarking();
		    ArmLoadStoreFwdRecurse(i_bbl, load_reg, &propregs, i_ins, &untraced_load, &untraced_store, &instruction_is_killed);
		  }
	      }
	    else if ((ARMINSOPCODE(i_ins) == ARM_STM
		      || ARMINSOPCODE(i_ins) == ARM_LDM ))
	      {
		memset(&propregs,0,sizeof(t_partial_value));
		RegsetAddMultipleRegs(propregs.registers, ARMINSIMMEDIATE(i_ins));
		if (INS_TYPE(i_ins) == IT_LOAD_MULTIPLE && RegsetIn(propregs.registers,ARMINSREGB(i_ins)))
		  continue; /* Too hard to handle */

		FOREACH_REG(tmpreg) propregs.corresponding[tmpreg] = MAX_REG_ITERATOR;

		load_reg = ARMINSREGB(i_ins);

		/* Now for each register that gets loaded/stored : calculate the offset from the load_reg
		 * and put this in the propreg structure */
		nr_regs = RegsetCountRegs(propregs.registers);
		i = 0;
		if (ARMINSFLAGS(i_ins) & FL_DIRUP)
		  {
		    if (ARMINSFLAGS(i_ins) & FL_WRITEBACK)
		      {
			if (ARMINSFLAGS(i_ins) & FL_PREINDEX) i++;
			REGSET_FOREACH_REG(propregs.registers,tmpreg)
			  {
			    propregs.offset[tmpreg] = 4 * (nr_regs-i);
			    i++;
			  }
		      }
		    else /* no writeback, dirup */
		      {
			if (ARMINSFLAGS(i_ins) & FL_PREINDEX) i++;
			REGSET_FOREACH_REG(propregs.registers,tmpreg)
			  {
			    propregs.offset[tmpreg] = -4 * i;
			    i++;
			  }
		      }
		  }
		else /* no dirup */
		  {
		    if (ARMINSFLAGS(i_ins) & FL_WRITEBACK)
		      {
			if (!(ARMINSFLAGS(i_ins) & FL_PREINDEX)) i++;
			REGSET_FOREACH_REG(propregs.registers,tmpreg)
			  {
			    propregs.offset[tmpreg] = -4 * i;
			    i++;
			  }
		      }
		    else /* no writeback */
		      {
			if (!(ARMINSFLAGS(i_ins) & FL_PREINDEX)) i++;
			REGSET_FOREACH_REG(propregs.registers,tmpreg)
			  {
			    propregs.offset[tmpreg] = 4 * (nr_regs-i);
			    i++;
			  }
		      }
		  }

		j_ins = T_ARM_INS(INS_NEXT(i_ins));
		/*if(ARMINSOPCODE(i_ins) == ARM_LDM && ARMINSOPCODE(j_ins) == ARM_STM)*/
		/*DiabloPrint(stdout,"Doing @I and @I\n",i_ins,j_ins);*/
		untraced_load = FALSE;
		untraced_store = FALSE;
		if (ArmLoadStoreFwdInner2(j_ins, load_reg, &propregs, i_ins, &untraced_load, &untraced_store, &instruction_is_killed))
		  {
		    InitNodeMarking();
		    InitEdgeMarking();
		    ArmLoadStoreFwdRecurse(i_bbl, load_reg, &propregs, i_ins, &untraced_load, &untraced_store,&instruction_is_killed);
		  }
	      }
	    if (instruction_is_killed)
	      break;
	  }
      }
  }
  VERBOSE(0,("%d loads,  %d with tags",load_count,load_tag_count));
  VERBOSE(0,("%d stores, %d with tags",store_count,store_tag_count));
  VERBOSE(0,("%d = sum,  %d with tags",store_count+load_count,store_tag_count+load_tag_count));

  /*DiabloPrint(stdout,"[load-store forward'n] %d stores killed, %d loads changed, %d moves added, %d add's added, %d ldm's changed, %d stm's changed %d stm's killed, %d ldm's killed\n",killed_store,changed_load,added_moves,added_adds,changed_loadmultiple,changed_storemultiple,killed_storemultiple,killed_loadmultiple); */
  /*DiabloPrint(stdout,"[load-store forward'n] %d registers less to spill, %d load/stores less\n",regs_to_load,killed_store+changed_load+changed_loadmultiple+changed_storemultiple+killed_storemultiple+killed_loadmultiple); */
  ANALYSIS_END("ArmLoadStore",("%d registers less to spill, %d load/stores less",regs_to_load,killed_store+changed_load+changed_loadmultiple+changed_storemultiple+killed_storemultiple+killed_loadmultiple));

  /* clean up the dominator information */
  DominatorCleanup(cfg);

  return; /*killed_store+changed_load+added_moves+added_adds+changed_loadmultiple+killed_storemultiple+killed_loadmultiple;  */
}/*}}}*/
/*!
 * given an instruction, and an offset, we try to keep track
 * of the value of this offset. For the moment, we only try to trace add's and sub's
 * of a register with an immediate into that same register 
 *
 * \param ins
 * \param offset
 *
 * \return t_lattice_level 
*/
/* ArmPartialEvaluation{{{*/
t_lattice_level ArmPartialEvaluation(t_arm_ins* ins, register_content * offset, t_reg reg)
{
  /*register_content tmp_content; */
  if (ARMINSREGA(ins) == ARMINSREGB(ins) && ArmInsHasImmediate(ins) && ARMINSREGA(ins) == reg)
  {
    if(ARMINSOPCODE(ins) == ARM_ADD)
    {
#ifdef DEBUG_LS_FWD
      VERBOSE(0,("Add evaluated! @I\n",ins));
#endif
      offset->i = GENERIC_ADD_UINT32(offset->i,ARMINSIMMEDIATE(ins));
      return VALUE;
    }
    if(ARMINSOPCODE(ins) == ARM_SUB)
    {
#ifdef DEBUG_LS_FWD
      VERBOSE(0,("Sub evaluated! @I\n",ins));
#endif
      offset->i = GENERIC_SUB_UINT32(offset->i,ARMINSIMMEDIATE(ins));
      return VALUE;
    }
    else return BOT;
  }
  else if (ARMINSREGB(ins) == reg && (INS_TYPE(ins) == IT_LOAD || INS_TYPE(ins) == IT_STORE) && ARMINSFLAGS(ins) & FL_WRITEBACK && ARMINSFLAGS(ins) & FL_IMMED)
  {
    if(ARMINSFLAGS(ins) & FL_DIRUP)
    {
      offset->i = GENERIC_ADD_UINT32(offset->i,ARMINSIMMEDIATE(ins));
      return VALUE;
    }
    else
    {
      offset->i = GENERIC_SUB_UINT32(offset->i,ARMINSIMMEDIATE(ins));
      return VALUE;
    }
  }
  else if (ARMINSREGB(ins) == reg && (ARMINSOPCODE(ins) == ARM_STM || ARMINSOPCODE(ins) == ARM_LDM) && (ARMINSFLAGS(ins) & (FL_WRITEBACK) ))
  {
#if MAX_NR_REG_SUBSETS > 4
    FATAL(("Implement me with real regsets!"));
#else
    t_int32 nr_regs = RegsetCountRegs(ARMINSIMMEDIATE(ins));
    nr_regs *= 4;
    if(ARMINSFLAGS(ins) & FL_DIRUP)
    {
      offset->i = GENERIC_ADD_UINT32(offset->i,nr_regs);
      return VALUE;
    }
    else
    {
      offset->i = GENERIC_SUB_UINT32(offset->i,nr_regs);
      return VALUE;
    }
#endif
  }
#if 0
  else if (ARMINSREGA(ins) == ARMINSREGB(ins) && ARMINSREGC(ins) != ARM_REG_NONE && ARMINSREGA(ins) == reg)
  {
    if (RegsetIn(INS_REGS_USE(ins),ARMINSREGC(ins)))
    {
      register_content tmp_content;
#if 0 // REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE
      if(RegStateGetRegValue(ins,&INS_REGS_USE(ins),ARMINSREGC(ins),&tmp_content))
#else
	if (BOT)
#endif
	return BOT;
      if(ARMINSOPCODE(ins) == ARM_ADD)
      {
	DiabloPrint(stdout,"Add evaluated! @I\n",ins);
	FATAL((""));
	offset->i = GENERIC_ADD(offset->i,tmp_content.i);
	return VALUE;
      }
      if(ARMINSOPCODE(ins) == ARM_SUB)
      {
	DiabloPrint(stdout,"Sub evaluated! @I\n",ins);
	offset->i = GENERIC_SUB_UINT32(offset->i,tmp_content.i);
	return VALUE;
      }
      else return BOT;
    }
    else return BOT;
  }
  /* TODO: implement */
#endif
  return BOT;
}/*}}}*/

void ArmLoadStoreForwarding(t_cfg * cfg)
{
  t_function * ifun;
  t_bbl * ibbl;

  ComDominators(cfg);
  ComLiveness(cfg,SENSITIVE);
  ANALYSIS_START("ArmLoadStore");
  
  CFG_FOREACH_FUN(cfg,ifun)
    FUN_FOREACH_BBL(ifun,ibbl)
    {
    }

  ANALYSIS_SIMPLE_END("ArmLoadStore");
  DominatorCleanup(cfg);
  return;
}

/* vim: set shiftwidth=2 foldmethod=marker : */
