/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_map.h"
#include "../../utils/util_error.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_status.h"
#include "../../kernel/diablo_parsemap.h"
#include "../../diablo_options.h"
#include <string.h>
/* }}} */
/*!
 * The following function parses a line from an Arm ADS map file. 
 *
 * \param region The name of the parent section (for ADS binaries this is
 * ER_RO, ER_RW, or ER_ZI).
 * \param line The line to parse
 *
 * \return t_map_node * the parsed node
*/
/* ArmMapParseLine {{{ */
t_map_node * ArmMapParseLine(t_string region,t_string line)
{
  t_map_node * new_node = (t_map_node*) Malloc(sizeof(t_map_node));
  t_string_array * array=StringDivider(line," \t\n",FALSE,FALSE);
  t_bool is_eight_fielder;
  t_uint32 tmp=0;

  if (!region) FATAL(("Parse error!"));
  
  new_node->mapped_sec_name=StringDup(region);
  /* there should be 7 or 8 fields exept when there is padding in the map file.
   * Then there are 3 fields, and the last field is "PAD" */

  if ((StringArrayNElems(array)==3) && (strcmp(array->strings[2],"PAD")==0))
  {
    new_node->type=Dbug; /* This makes the reader throw away this element */
    new_node->object=StringDup("PAD");
    new_node->sec_name=StringDup("PAD");
	  
    StringArrayFree(array); /* Cleanup */
    return new_node;
  }

  /* the other should all have 7 or 8 fields */
  VERIFY_FASSERT(((StringArrayNElems(array)==8)||(StringArrayNElems(array)==7)),("Parse error in map file at line %s",line));

  /* Field 1: the start of the section in the final object */
  VERIFY_FASSERT(((array->strings[0][0]=='0')||(array->strings[0][1]=='x')), ("Parse error in map file at line %s: first field should be a hex-number",line));
  tmp=strtol(array->strings[0], NULL, 16);
  GenericSet_uint32(new_node->base_addr,tmp);
  /* Field 2: the size of the section in the final object */
  VERIFY_FASSERT(((array->strings[1][0]=='0')||(array->strings[1][1]=='x')), ("Parse error in map file at line %s: second field should be a hex-number",line));
  tmp=strtol(array->strings[1], NULL, 16);
  GenericSet_uint32(new_node->sec_size,tmp);
  
  is_eight_fielder=(StringArrayNElems(array)==8);

  /* Field 3: the type of the section in the final object */
  if (strcmp(array->strings[2],"Code")==0) new_node->type=Code;
  else if (strcmp(array->strings[2],"Data")==0) new_node->type=Data;
  else if (strcmp(array->strings[2],"Dbug")==0) new_node->type=Dbug;
  else if (strcmp(array->strings[2],"Zero")==0) new_node->type=Zero;
  else FATAL(("Unknown type for block!"));

  /* Field 4: the attributes of the section in the final object */
  if (strcmp(array->strings[3],"RO")==0) new_node->attr = RO;
  else if (strcmp(array->strings[3],"RW")==0) new_node->attr = RW;
  else FATAL(("Unknown attribute for block!"));

  /* Field 5: An index */ 
  new_node->idx=strtol(array->strings[4],NULL,10);
  
  /* Field 6(is_eight_fielder) : Marks an entry */ 
  if (is_eight_fielder) VERIFY_FASSERT((array->strings[5][0]=='*')&&(array->strings[5][1]=='\0'), ("Parse error, if fields==8, then the 5 element should be a star! in %s",line));

  /* Field 6(is_seven_fielder),7(is_eight_fielder): The sections original name */

  new_node->sec_name=StringDup(array->strings[6-!is_eight_fielder]);
  
  /* Field 7(is_seven_fielder),8(is_eight_fielder): The objects original name */

  if (PatternMatch("*(*)",array->strings[7-!is_eight_fielder]))
   {
      t_string dup=StringDup(array->strings[7-!is_eight_fielder]);
      t_string obj=strtok(dup,"(");
      t_string lib=strtok(NULL,")");
      t_string Name=Collate3(lib,":",obj);
      VERIFY_FASSERT(obj,("Illegal objectname in map: %s",array->strings[7-!is_eight_fielder]));
      VERIFY_FASSERT(lib,("Illegal objectname in map: %s",array->strings[7-!is_eight_fielder]));
      Free(dup);
      new_node->object=Name;
   }
   else
   {
      new_node->object=StringDup(array->strings[7-!is_eight_fielder]);
   }


  StringArrayFree(array); /* Cleanup */
  new_node->obj=NULL;
  new_node->flags = 0x0;	/* always initialise the flags field to zero! */
  
  return new_node;
}
/* }}}*/
/*!  
 * \param obj The object for which we want to read the map file 
 *
 * \param infile The name of the mapfile (without the .map extension)
 *
 * Parses the file `infile.map', and stores the parsed information in the `map'
 * field of the given object. If no map file exists map is NULL. 
 * This function actually only parses map files generated by the ADS. If the map
 * file was generated by gcc, this function will detect it and a generic GCC 
 * parser will be called.
 */
/* ArmReadMap {{{ */
void ArmReadMap(t_object * obj, t_string infile)
{
  t_string region=NULL;
  FILE * ifp;
  t_map * map=obj->map;
  t_string filename, parse_line = NULL;
  t_map_node * parsed_node;


  /* We initialize this to ADS, it can change when the map is a GNU ld map */
  obj->linker=ADS_LINKER;
  filename=Collate(infile,MAP_FILE_EXTENSION);

  ifp = fopen((const char*) filename, "r");

  if (!ifp) FATAL(("Map file not found. Compile the binary with (g)cc -Wl,-Map,<output_name>.map -o <output_name> <obj1.o> <obj...>"));

  while((parse_line = GetLine(ifp)) != NULL) /* EOF not reached */
  {
    if (*parse_line == '\n') /* Skip empty lines */
    {
      Free(parse_line);   /* avoids leaks */
      continue;
    }

    if (PatternMatch("*Execution Region*",parse_line))
    {
      t_string_array * array=StringDivider(parse_line," \t\n",FALSE,FALSE);
      if (region) Free(region);
      region=StringDup(array->strings[2]);
      Free(parse_line);   /* avoids leaks */
      StringArrayFree(array);
      continue;
    }
    /* If we see this line, this arm object is produced by the gcc compiler,
     * then we call the generic gcc parser */
    else if (PatternMatch("*Archive*member*included*because*of*file*", parse_line) || PatternMatch("*Memory*Configuration*", parse_line))
    {
      fclose(ifp);
      Free(filename);
      if (region) Free(region);
      Free(parse_line);   /* avoids leaks */
      /* Hand control over to the gcc map parser */
      GCCMapRead(obj, infile);
      return;
    }
    /* tricky test to see if this line is a line containing section-info,
     * relays on the fact that the address field start at the fifth character
     * of the line */
    else if ((strstr((const char*)parse_line, "0x") - (char *) parse_line) != 4)
    {
      /* It's not section info! */
      Free(parse_line);   /* avoids leaks */
      continue;
    }

    /* we only get here if the line is actually containing real map data */
    parsed_node = ArmMapParseLine(region,parse_line);
    
    if (parsed_node->type==Dbug) /* Debug nodes are not necessary, we free them */
      MapNodeFree(parsed_node);
    else /* All the others are inserted in the map */
      MapInsertNode(map, parsed_node);
    Free(parse_line);   /* avoids leaks */
  };

  fclose(ifp);
  Free(filename);
  if (region) Free(region);
  VERBOSE(1,("Parsing done"));
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
