/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_utils.h"
#include "arm_opcodes.h"
#include "arm_registers.h"
#include "arm_instructions.h"
#include "arm_emulation.h"
#include "../../utils/util_error.h"
#include "../../utils/util_bitmanipulation.h"
/* }}} */

/*!
 * checks whether writeback occurs with this instruction (this can only happen
 * with LDR,STR,LDM,STM and derivatives) 
 *
 * \param ins
 *
 * \return int 
*/
/* ArmInsWriteBackHappens {{{ */
int ArmInsWriteBackHappens(t_ins * ins) {

  if ( !( (INS_TYPE(ins) == IT_LOAD) || (INS_TYPE(ins) == IT_STORE) || (INS_TYPE(ins) == IT_LOAD_MULTIPLE) || (INS_TYPE(ins) == IT_STORE_MULTIPLE) || (INS_TYPE(ins) == IT_FLT_LOAD) || (INS_TYPE(ins) == IT_FLT_STORE)) ) {
    return 0;
  }

  /* LDM and STM treat the P and W flags differently compared to LDR and STR */
  if ((INS_TYPE(ins) == IT_LOAD_MULTIPLE) || (INS_TYPE(ins) == IT_STORE_MULTIPLE) )
  {
    if (ARMINSFLAGS(ins) & FL_WRITEBACK) 
      return 1; 
    else 
      return 0;
  }


  if (!(ARMINSFLAGS(ins) & FL_PREINDEX)) return 1;	/* postindex always includes writeback */

  /* if we reach this point, it's definitely preindex */
  if (ARMINSFLAGS(ins) & FL_WRITEBACK) return 1;
  else return 0;

  return 0;
}
/* }}}*/
/*!
 * \todo Document
 *
 * \param ins
 *
 * \return int 
*/
/* ArmInsHasShiftedFlexible {{{*/
int ArmInsHasShiftedFlexible(t_ins * ins) {
   if (ARMINSREGC(ins) == ARM_REG_NONE) return 0;
   if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_NONE) return 0;
   return 1;
}
/* }}} */
/*!
 * Returns a bitmask specifying which registers are use by this instruction the
 * bottommost sixteen bits represent the usual registers r0 - r15 bit 16 is
 * used to represent the CPSR (i.e. conditional instructions use this register)
 * bit 17 is used to represent the SPSR (possibly referenced in the MRS and MSR
 * instructions
 *
 * \param ins
 *
 * \return t_regset 
*/
/* ArmUsedRegisters {{{ */
t_regset ArmUsedRegisters(t_ins * ins) 
{

  t_regset mask = NullRegs;

  if (ins == NULL) return NullRegs;
  if (ArmInstructionIsNOOP(ins)) return NullRegs;

  /* conditional instructions use the CPSR */
  if (ArmInstructionIsConditional(ins))
    {
      switch(ARMINSCONDITION(ins))
	{
	case ARM_CONDITION_EQ :
	case ARM_CONDITION_NE :
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	  break;
	case ARM_CONDITION_CS :
	case ARM_CONDITION_CC :
	  RegsetAddReg(mask,ARM_REG_C_CONDITION);
	  break;
	case ARM_CONDITION_MI :
	case ARM_CONDITION_PL :
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  break;
	case ARM_CONDITION_VS : 
	case ARM_CONDITION_VC :
	  RegsetAddReg(mask,ARM_REG_V_CONDITION);
	  break;
	case ARM_CONDITION_HI :
	case ARM_CONDITION_LS :
	  RegsetAddReg(mask,ARM_REG_C_CONDITION);
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	  break;
	case ARM_CONDITION_GE :
	case ARM_CONDITION_LT : 
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  RegsetAddReg(mask,ARM_REG_V_CONDITION);
	  break;
	case ARM_CONDITION_GT : 
	case ARM_CONDITION_LE : 
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  RegsetAddReg(mask,ARM_REG_V_CONDITION);
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	  break;
	case ARM_CONDITION_AL:
	case ARM_CONDITION_NV : 
	  break; /* nothing ever happens */
	}
      /* Conditional registers can forward the defined registers */
    }
  
  switch (INS_TYPE(ins)) 
    {
    case IT_BRANCH:
      if (ARMINSOPCODE(ins) == ARM_BX) 
	RegsetAddReg(mask,ARMINSREGB(ins));
      break;
      
    case IT_SWI:
      /* SWI's use _all_ registers */
      {
	t_uint32 r;
	for (r=0; r<23; r++)
	  if (r != 15) /* exclude the PC, otherwise we run into trouble during leader detection */
	    RegsetAddReg(mask,r);
	
	switch (ARMINSIMMEDIATE(ins))
	  {
	  case 0x900005:
	  case 0x900004:
	  case 0x900003:
	    /*	      printf("AHA 2 %x %d\n",ARMINSIMMEDIATE(BBL_INS_LAST(head)),ARMINSIMMEDIATE(BBL_INS_LAST(head)));	      */
	    RegsetSubReg(mask,ARM_REG_R3);
	    break;
	  default:
	    break;
	    /*	      printf("VALUE 2 %x %d\n",ARMINSIMMEDIATE(BBL_INS_LAST(head)),ARMINSIMMEDIATE(BBL_INS_LAST(head)));	      */
	  }
      }
      break;

    case IT_DATAPROC:
      /* all dataprocessing instructions look the same in a certain way: the only registers */
      /* whose contents may be used in the instruction are ARM_REGB, ARM_REGC and ARM_REGS */
      if (ARMINSREGB(ins) != ARM_REG_NONE) RegsetAddReg(mask,ARMINSREGB(ins));
      if (ARMINSREGC(ins) != ARM_REG_NONE) RegsetAddReg(mask,ARMINSREGC(ins));
      if (ARMINSREGS(ins) != ARM_REG_NONE) RegsetAddReg(mask,ARMINSREGS(ins));
      if (ARMINSOPCODE(ins) == ARM_ADC) RegsetAddReg(mask,ARM_REG_C_CONDITION);
      if (ARMINSOPCODE(ins) == ARM_SBC) RegsetAddReg(mask,ARM_REG_C_CONDITION);
      if (ARMINSOPCODE(ins) == ARM_RSC) RegsetAddReg(mask,ARM_REG_C_CONDITION);
      if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_RRX) RegsetAddReg(mask,ARM_REG_C_CONDITION);
      break;

    case IT_MUL:
      /* the values in ARM_REGB and ARM_REGC are certainly read, ARM_REGA is never read but always written, */
      /* ARM_REGS is read with MLA and written with all but MUL and MLA  */
      RegsetAddReg(mask,ARMINSREGB(ins));
      RegsetAddReg(mask,ARMINSREGC(ins));
      if (ARMINSOPCODE(ins) == ARM_MLA) RegsetAddReg(mask,ARMINSREGS(ins));
      break;

    case IT_LOAD:
      /* ARM_REGB is always read, ARM_REGC if the offset is stored in a register,
       * ARM_REGS if the shift amount is stored in a register */
      RegsetAddReg(mask,ARMINSREGB(ins));
      if (ARMINSREGC(ins) != ARM_REG_NONE) RegsetAddReg(mask,ARMINSREGC(ins));
      if (ARMINSREGS(ins) != ARM_REG_NONE) RegsetAddReg(mask,ARMINSREGS(ins));
      if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_RRX) RegsetAddReg(mask,ARM_REG_C_CONDITION);
      break;

    case IT_STORE:
      /* ARM_REGA is always used (it's the register that is written to memory!) */
      /* ARM_REGB is always used, ARM_REGC only if the offset is stored in a register,
       * ARM_REGS if the shift amount is stored in a register */
      RegsetAddReg(mask,ARMINSREGA(ins));
      RegsetAddReg(mask,ARMINSREGB(ins));
      if (ARMINSREGC(ins) != ARM_REG_NONE) RegsetAddReg(mask,ARMINSREGC(ins));
      if (ARMINSREGS(ins) != ARM_REG_NONE) RegsetAddReg(mask,ARMINSREGS(ins));
      if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_RRX) RegsetAddReg(mask,ARM_REG_C_CONDITION);
      break;

    case IT_FLT_STORE:
      if (ARMINSOPCODE(ins)==ARM_STF)
      {
	RegsetAddReg(mask,ARMINSREGA(ins));
	RegsetAddReg(mask,ARMINSREGB(ins));
      }
      else /* Multiple store */
      {
	RegsetAddReg(mask,ARMINSREGB(ins));
	RegsetSetUnion(mask,ARMINSMULTIPLE(ins));
      }
      break;
    case IT_FLT_LOAD:
	RegsetAddReg(mask,ARMINSREGB(ins));
      break;
    case IT_FLT_ALU:
    case IT_FLT_INT:
      if (ARMINSREGC(ins)!=ARM_REG_NONE)
	RegsetAddReg(mask,ARMINSREGC(ins));
      RegsetAddReg(mask,ARMINSREGB(ins));
      break;
    case IT_LOAD_MULTIPLE:
    case IT_STORE_MULTIPLE:
      /* ARM_REGB is always used and the other used registers are stored in the immediate field, */
      /* provided that we're dealing with an STM instruction! */
      RegsetAddReg(mask,ARMINSREGB(ins));
      if (ARMINSOPCODE(ins) == ARM_STM) RegsetAddMultipleRegs(mask,ARMINSIMMEDIATE(ins));
      break;

    case IT_SWAP:
      RegsetAddReg(mask,ARMINSREGA(ins));
      RegsetAddReg(mask,ARMINSREGB(ins));
      RegsetAddReg(mask,ARMINSREGC(ins));
      break;

    case IT_STATUS:
      /* MRS uses CPSR or SPSR depending on whether FL_SPSR is set */
      /* MSR uses ARM_REGC if the value isn't given as an immediate */
      if (ARMINSOPCODE(ins) == ARM_MRS) {
	if (ARMINSFLAGS(ins) & FL_SPSR)
	  RegsetAddReg(mask,21);
	else {
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  RegsetAddReg(mask,ARM_REG_V_CONDITION);
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	  }
      } else 
	if (ARMINSREGC(ins) != ARM_REG_NONE) RegsetAddReg(mask,ARMINSREGC(ins));
      break;
    case IT_FLT_STATUS:
      RegsetAddReg(mask,ARMINSREGB(ins));
      break;

    case IT_DATA:
      mask = NullRegs;
      break;
    case IT_CONSTS:
      break;
    case IT_CALL:
      RegsetAddReg(mask,ARM_REG_R0);
      RegsetAddReg(mask,ARM_REG_R1);
      RegsetAddReg(mask,ARM_REG_R2);
      RegsetAddReg(mask,ARM_REG_R3);
      break;
    default:
      FATAL(("Implement type %d!",INS_TYPE(ins)));
    }

  return mask;
}
/* }}} */
/*!
 * Returns a bitmask specifying which registers are modified by this
 * instruction the bottommost sixteen bits represent the usual registers r0 -
 * r15 bit 16 is used to represent the CPSR (i.e. instructions having the S
 * flag change this register) bit 17 is used to represent the SPSR (possibly
 * referenced in the MRS and MSR instructions 
 * 
 * \param ins
 *
 * \return t_regset 
*/
/* ArmDefinedRegisters {{{ */
t_regset ArmDefinedRegisters(t_ins * ins) 
{

  t_regset mask = NullRegs;

  if (ins == NULL) return NullRegs;
  if (ArmInstructionIsNOOP(ins)) return NullRegs;

  /* instructions which have the S flag set define the CPSR */
  if (ARMINSFLAGS(ins) & FL_S)
    {
      switch ARMINSOPCODE(ins)
	{
	case ARM_CONSTANT_PRODUCER:
	case ARM_ADDRESS_PRODUCER:
	  break;
	case ARM_PSEUDO_CALL:
	  break;
	case ARM_MUL:
	case ARM_MLA:
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  RegsetAddReg(mask,ARM_REG_C_CONDITION);
	  /*mask |= (1<<ARM_REG_Z_CONDITION)|(1<<ARM_REG_N_CONDITION)|(1<<ARM_REG_C_CONDITION);*/
	  break;
	case ARM_AND:
	case ARM_TST:
	case ARM_BIC:
	case ARM_EOR:
	case ARM_ORR:
	case ARM_TEQ:
	case ARM_MOV:
	case ARM_MVN:
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  /*mask |= (1<<ARM_REG_Z_CONDITION)|(1<<ARM_REG_N_CONDITION);*/
	  break;
	case ARM_CMN:
	case ARM_CMP:
	case ARM_ADD:
	case ARM_ADC:
	case ARM_SUB:
	case ARM_SBC:
	case ARM_RSC:
	case ARM_RSB:
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  RegsetAddReg(mask,ARM_REG_C_CONDITION);
	  RegsetAddReg(mask,ARM_REG_V_CONDITION);
	  /*mask |= (1<<ARM_REG_Z_CONDITION)|(1<<ARM_REG_N_CONDITION)|(1<<ARM_REG_C_CONDITION)|(1<<ARM_REG_V_CONDITION);*/
	  break;
	case ARM_LDR:
	case ARM_LDRB:
	case ARM_STR:
	case ARM_STRB:
	case ARM_LDM:
	  break;
	  
	default:
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  RegsetAddReg(mask,ARM_REG_C_CONDITION);
	  RegsetAddReg(mask,ARM_REG_V_CONDITION);
	  /*mask |= (1<<ARM_REG_Z_CONDITION)|(1<<ARM_REG_N_CONDITION)|(1<<ARM_REG_C_CONDITION)|(1<<ARM_REG_V_CONDITION);*/
	}
      if (ArmInsHasShiftedFlexible(ins))
	RegsetAddReg(mask,ARM_REG_C_CONDITION);
	/*mask|=1<<ARM_REG_C_CONDITION;*/
    }
  if ((ARMINSOPCODE(ins)==ARM_CMP) && (!(ARMINSFLAGS(ins) & FL_S))) { FATAL(("CMP without set!\n")); } 
  if ((ARMINSOPCODE(ins)==ARM_CMFE) && (!(ARMINSFLAGS(ins) & FL_S))) { FATAL(("CMFE without set!\n")); } 

  switch (INS_TYPE(ins))
    {
    case IT_BRANCH:
      /* branches define the pc */
      RegsetAddReg(mask,15);/*mask |= (1 << 15);*/
      if (ARMINSOPCODE(ins)==ARM_BL)
	RegsetAddReg(mask,14);/*mask |= (1<<14);*/
      break;

    case IT_SWI:
      /* SWI's use _all_ registers */
      {
	t_uint32 r;
	for (r=0; r<5; r++)
	  RegsetAddReg(mask,r);
      }
      break;

    case IT_DATAPROC:
      /* all dataprocessing instructions look the same in a certain way: */
      /* only ARM_REGA can be defined */

      if (ARMINSREGA(ins) != ARM_REG_NONE)  RegsetAddReg(mask,ARMINSREGA(ins));/*mask |= (1 << ARMINSREGA(ins)); */
      break;

    case IT_MUL:
      /* the values in ARM_REGB and ARM_REGC are certainly read, ARM_REGA is never read but always written, */
      /* ARM_REGS is read with MLA and written with all but MUL and MLA  */
      RegsetAddReg(mask,ARMINSREGA(ins));/*mask |= (1 << ARMINSREGA(ins));*/
      if (!((ARMINSOPCODE(ins) == ARM_MLA) || (ARMINSOPCODE(ins) == ARM_MUL))) RegsetAddReg(mask,ARMINSREGS(ins));/*mask |= (1 << ARMINSREGS(ins));*/
      break;

    case IT_LOAD:
      /* ARM_REGA is always defined, ARM_REGB is defined if writeback is specified */
      RegsetAddReg(mask,ARMINSREGA(ins));/*mask |= (1 << ARMINSREGA(ins));*/
      if (ArmInsWriteBackHappens(ins)) RegsetAddReg(mask,ARMINSREGB(ins));/*mask |= (1 << ARMINSREGB(ins));*/
      break;

    case IT_STORE:
      /* ARM_REGB is defined if writeback is specified */
      if (ArmInsWriteBackHappens(ins)) RegsetAddReg(mask,ARMINSREGB(ins));/*mask |= (1 << ARMINSREGB(ins));*/
      break;
    
    case IT_LOAD_MULTIPLE:
    case IT_STORE_MULTIPLE:
      /* ARM_REGB is defined in case of writeback and the other defined registers are stored in the immediate field, */
      /* provided that we're dealing with an LDM instruction! */
      if (ArmInsWriteBackHappens(ins)) RegsetAddReg(mask,ARMINSREGB(ins));/*mask |= (1 << ARMINSREGB(ins));  */
      if (ARMINSOPCODE(ins) == ARM_LDM) RegsetAddMultipleRegs(mask,ARMINSIMMEDIATE(ins));/*mask |= ARMINSIMMEDIATE(ins);*/
      break;

    case IT_SWAP:
      RegsetAddReg(mask,ARMINSREGA(ins));/*mask |= (1 << ARMINSREGA(ins));*/
      break;

    case IT_STATUS:
      /* MRS defines ARM_REGA */
      /* MSR defines CPSR or SPSR depending on the value of the FL_SPSR flag */
      if (ARMINSOPCODE(ins) == ARM_MRS) {
	RegsetAddReg(mask,ARMINSREGA(ins));/*mask |= (1 << ARMINSREGA(ins));*/
      } else 
	if (ARMINSFLAGS(ins) & FL_SPSR)
	  RegsetAddReg(mask,17);/*mask |= (1 << 17);*/
	else {
	  RegsetAddReg(mask,ARM_REG_Z_CONDITION);
	  RegsetAddReg(mask,ARM_REG_N_CONDITION);
	  RegsetAddReg(mask,ARM_REG_C_CONDITION);
	  RegsetAddReg(mask,ARM_REG_V_CONDITION);
	  }
      break;
    case IT_FLT_STATUS:
      RegsetAddReg(mask,ARMINSREGA(ins));
      break;
    case IT_FLT_STORE:
      if (ArmInsWriteBackHappens(ins)) RegsetAddReg(mask,ARMINSREGB(ins));
      break;
    case IT_FLT_LOAD:
      if (ArmInsWriteBackHappens(ins)) RegsetAddReg(mask,ARMINSREGB(ins));
      if (ARMINSOPCODE(ins) == ARM_LFM) RegsetSetUnion(mask,ARMINSMULTIPLE(ins));
      else RegsetAddReg(mask,ARMINSREGA(ins));
      break;
    case IT_FLT_ALU:
    case IT_FLT_INT:
      if (ARMINSREGA(ins)!=ARM_REG_NONE)
	RegsetAddReg(mask,ARMINSREGA(ins));
      break;

    case IT_DATA:
      mask = NullRegs;
      break;
    case IT_CONSTS:
      RegsetAddReg(mask,ARMINSREGA(ins));
      break;
    case IT_CALL:
      RegsetAddReg(mask,ARM_REG_Z_CONDITION);
      RegsetAddReg(mask,ARM_REG_N_CONDITION);
      RegsetAddReg(mask,ARM_REG_C_CONDITION);
      RegsetAddReg(mask,ARM_REG_V_CONDITION);
      RegsetAddReg(mask,ARM_REG_R0);
      RegsetAddReg(mask,ARM_REG_R1);
      RegsetAddReg(mask,ARM_REG_R2);
      RegsetAddReg(mask,ARM_REG_R3);
      RegsetAddReg(mask,ARM_REG_R10);
      RegsetAddReg(mask,ARM_REG_R11);
      RegsetAddReg(mask,ARM_REG_R12);
      RegsetAddReg(mask,ARM_REG_R14);
      break;
    default:
      FATAL(("Implement type %d!",INS_TYPE(ins)));
    }
  return mask;
}
/* }}} */
/*!
 * \todo Document 
 *
 * \param ins
 *
 * \return t_bool 
*/
/* ArmInsIsCommutative {{{ */
t_bool ArmInsIsCommutative(t_ins * ins)
{
  switch(ARMINSOPCODE(ins))
  {
    case ARM_MUL:
    case ARM_MLA:
    case ARM_UMULL:
    case ARM_UMLAL:
    case ARM_SMULL:
    case ARM_SMLAL:
      return FALSE;
      /* this may strike you as odd, but you can't _always_ swap the registers here because mul 
       * cant have the first operand the same as the destination register */

    case ARM_AND:
    case ARM_EOR:
    case ARM_ADD:
    case ARM_ADC:
    case ARM_TST:
    case ARM_TEQ:
    case ARM_ORR:
      return TRUE;
    default:
      /* keep compiler happy */
      return FALSE;
  }
  /*return FALSE;*/
}
/* }}} */
/*!
 * Given a load or store instruction, try to find out at which offset from the
 * base register the memory is accessed 
 *
 * \param ins
 * \param offset
 *
 * \return t_lattice_level 
*/
/* ArmComputeLSOffset {{{ */
t_lattice_level ArmComputeLSOffset(t_arm_ins * ins, t_int32* offset)
{
  t_uint32 shiftimmediate;
  register_content tmp_value_to_shift;
  register_content tmp_shiftreg_value;
  t_bool carry_set=FALSE;
  t_bool tmp_flag;
  t_uint32 value_to_shift;
  t_uint32 shiftreg_value;
  
  *offset = 0;

  if (ArmInsHasShiftedFlexible(T_INS(ins)))
    /* Try to find out what is the offset to shift */
  {
#if 0 /* REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE*/
    if (RegStateGetRegValue(T_INS(ins),&INS_REGS_USE(ins),ARMINSREGC(ins),&tmp_value_to_shift))
    {
      return BOT;
    }
#else
    return BOT;
#endif
    if (ARMINSREGS(ins) != ARM_REG_NONE)
    {
#if 0 /* REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE*/
      if (RegStateGetRegValue(T_INS(ins),&INS_REGS_USE(ins),ARMINSREGS(ins),&tmp_shiftreg_value))
      {
	return BOT;
      }
#else
      return BOT;
#endif
      shiftreg_value = G_T_UINT32(tmp_shiftreg_value.i);
    }

    value_to_shift = G_T_UINT32(tmp_value_to_shift.i);
    shiftimmediate = ARMINSSHIFTLENGTH(ins);

    /* apply the shifting */
    switch (ARMINSSHIFTTYPE(ins)) 
    {
      case ARM_SHIFT_TYPE_LSL_IMM:
	if (shiftimmediate > 0) 
	  if (BIT_IS_SET(value_to_shift, (32-shiftimmediate))) 
	    carry_set=TRUE;
	*offset = value_to_shift << shiftimmediate;
	break;
      case ARM_SHIFT_TYPE_LSR_IMM:
	*offset = value_to_shift >> shiftimmediate;
	break;
      case ARM_SHIFT_TYPE_ASR_IMM:
	if ((t_int32) value_to_shift < 0)
	  *offset = (value_to_shift >> shiftimmediate) | ~(0xFFFFFFFF >> shiftimmediate);
	else 
	  *offset = value_to_shift >> shiftimmediate;
	break;
      case ARM_SHIFT_TYPE_ROR_IMM:
	if (shiftimmediate == 0)
	{
	  *offset = value_to_shift;
	  break;
	}
	*offset = (value_to_shift >> shiftimmediate) | (value_to_shift << (32 - shiftimmediate));
	break;
      case ARM_SHIFT_TYPE_LSL_REG:

	if (ARM_SHIFT_AMOUNT(shiftreg_value) == 0) /* no carry can be shifted out here */
	  *offset = value_to_shift;
	else if (ARM_SHIFT_AMOUNT(shiftreg_value) > 32) /* no carry can be shifted out here */
	  *offset = 0;
	else
	  *offset = value_to_shift << ARM_SHIFT_AMOUNT(shiftreg_value);
	break;
      case ARM_SHIFT_TYPE_LSR_REG: /* no carry can be shifted out here */
	*offset = (value_to_shift >> ARM_SHIFT_AMOUNT(shiftreg_value));
	break;
      case ARM_SHIFT_TYPE_ASR_REG: /* no carry can be shifted out here */
	if ((t_int32) value_to_shift < 0)
	  *offset = (value_to_shift >> ARM_SHIFT_AMOUNT(shiftreg_value)) | ~(0xFFFFFFFF >> ARM_SHIFT_AMOUNT(shiftreg_value));
	else *offset = value_to_shift >> ARM_SHIFT_AMOUNT(shiftreg_value);
	break;
      case ARM_SHIFT_TYPE_ROR_REG:
	if (ARM_ROTATE_AMOUNT(shiftreg_value) == 0)
	{
	  *offset = value_to_shift;
	  break;
	}
	*offset = (value_to_shift >> ARM_ROTATE_AMOUNT(shiftreg_value)) | (value_to_shift << (32 - ARM_ROTATE_AMOUNT(shiftreg_value)))  /* CHECK THIS */;
	break;
      case ARM_SHIFT_TYPE_RRX:
#if 0 /* REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE*/
	if (RegStateGetRegCond(T_INS(ins),&INS_REGS_USE(ins),ARM_REG_C_CONDITION,&tmp_flag))
	{
	  return BOT;
	}
#else
	return BOT;
#endif
	*offset = tmp_flag ? ((LL(0x1) << 31) & (value_to_shift >> 1)) : (value_to_shift >> 1);
	break;
      default: 
	FATAL(("Unknown shifttype while calculating offset!!"));
	break;
    }
  } 
  else if (ArmInsHasImmediate(ins) && INS_TYPE(ins) != IT_STORE_MULTIPLE && INS_TYPE(ins) != IT_LOAD_MULTIPLE)
  {
    *offset = ARMINSIMMEDIATE(ins);
  }
  else if (ARMINSREGC(ins) != ARM_REG_NONE)
  {
#if 0 /* REGSTATES ARE NEVER SET FOR THE MOMENT, AND THEREFORE NOT DECLARED ANYMORE*/
    if (RegStateGetRegValue(T_INS(ins),&INS_REGS_USE(ins),ARMINSREGC(ins),&tmp_value_to_shift))
    {
      return BOT;
    }
#else
    return BOT;
#endif
    *offset = G_T_UINT32(tmp_value_to_shift.i);
  }
  else if (INS_TYPE(ins) == IT_STORE_MULTIPLE || INS_TYPE(ins) == IT_LOAD_MULTIPLE)
  {
    t_regset tmp_set = RegsetNew();
    RegsetAddMultipleRegs(tmp_set,ARMINSIMMEDIATE(ins));
    *offset = RegsetCountRegs(tmp_set)*4;
  }
   
  if (!(ARMINSFLAGS(ins) & FL_DIRUP)) *offset = -(*offset);

  return VALUE;
  
}
/* }}} */
/*!
 * return the inverted condition code (eg AL becomes NV)
 *
 * \param condition
 *
 * \return  
 */
/* ArmInvertCondition {{{ */
t_arm_condition_code ArmInvertCondition(t_arm_condition_code condition)
{
  switch (condition)
  {  
    case ARM_CONDITION_EQ:
      return ARM_CONDITION_NE;
    case ARM_CONDITION_NE:
      return ARM_CONDITION_EQ;
    case ARM_CONDITION_CS:
      return ARM_CONDITION_CC;
    case ARM_CONDITION_CC:
      return ARM_CONDITION_CS;
    case ARM_CONDITION_MI:
      return ARM_CONDITION_PL;
    case ARM_CONDITION_PL:
      return ARM_CONDITION_MI;
    case ARM_CONDITION_VS:
      return ARM_CONDITION_VC;
    case ARM_CONDITION_VC:
      return ARM_CONDITION_VS;
    case ARM_CONDITION_HI:
      return ARM_CONDITION_LS;
    case ARM_CONDITION_LS:
      return ARM_CONDITION_HI;
    case ARM_CONDITION_GE:
      return ARM_CONDITION_LT;
    case ARM_CONDITION_LT:
      return ARM_CONDITION_GE;
    case ARM_CONDITION_GT:
      return ARM_CONDITION_LE;
    case ARM_CONDITION_LE:
      return ARM_CONDITION_GT;
    case ARM_CONDITION_AL:
      return ARM_CONDITION_NV;
    case ARM_CONDITION_NV:
      return ARM_CONDITION_AL;
  }
  /* just keep the compiler happy */
  return ARM_CONDITION_NV;
}
/* }}} */
/*!
 *
 * Checks if a constant can be encoded for a certain opcode
 *
 * \param cons the constant to encode
 *
 * \param opc the opcode
 *
 * \return t_bool TRUE if the value can be encoded else FALSE 
*/
/* ArmIsEncodableConstantForOpcode {{{ */
t_bool ArmIsEncodableConstantForOpcode(t_uint32 cons, t_uint16 opc) {
   int i;

   if ((opc == ARM_LDR) || (opc == ARM_LDRB) || (opc == ARM_STR) || (opc == ARM_STRB))
   {
     /* twelve-bit constants */
     return ((cons & 0xfff) == cons);
   }

   if ((opc == ARM_LDRH) || (opc == ARM_LDRSB) || (opc == ARM_LDRSH) || (opc == ARM_STRH))
   {
     /* eight-bit constants */
     return ((cons & 0xff) == cons);
   }

   /* standard case */
   for (i=0; i<32; i+=2) {
      if ((cons & rotate_right(0xff,i)) == cons)
	 return TRUE;
   }
   return FALSE;
}
/* }}} */
/*!
 * Checks if a constant can be produced in two instructions 
 *
 * \param val The value to produce
 *
 * \return t_bool TRUE if it can be produced, FALSE if not.
*/
/* ArmConstantIsProducableInTwoInstructions {{{ */
t_bool ArmConstantIsProducableInTwoInstructions(t_uint32 val)
{
  int i;
  t_uint32 a;
  t_uint32 val_invert = ~val;

  /* Try to generate the constant from scratch in two instructions */

  for (i = 0; i < 32; i += 2)
    if (((a = rotate_left (val, i)) & 0xff) != 0)
    {
      if (a & 0xff00)
      {
	if (a & ~ 0xffff)
	  continue;
      }
      else if (a & 0xff0000)
      {
	if (a & 0xff000000)
	  continue;
      }
      else
      {
      }

      return TRUE;
    }

  for (i = 0; i < 32; i += 2)
    if (((a = rotate_left (val_invert, i)) & 0xff) != 0)
    {
      if (a & 0xff00)
      {
	if (a & ~ 0xffff)
	  continue;
      }
      else if (a & 0xff0000)
      {
	if (a & 0xff000000)
	  continue;
      }
      else
      {
      }

      return TRUE;
    }

  return FALSE;
}

/* }}} */
/*!
 * Encodes a constant in 2 instructions. The constant is produced in the
 * destination register of the instruction that is passed to the function. A
 * second instruction is inserted after this instruction to encode the rest of
 * the value.   
 *
 * \param ins The first instruction. Register A will be used as the destination
 * register. The rest of the instruction will be overwritten. 
 * \param value The constant value to produce
 *
 * \return void 
*/
/* ArmEncodeConstantInTwoInstructions  {{{ */
void ArmEncodeConstantInTwoInstructions(t_ins * ins, t_uint32 value)
{
  t_ins * extra;
  t_uint32 val_invert = ~value;
  int i;
  t_uint32 a;
 
  extra = InsNewForBbl(INS_BBL(ins));
  GenericSet_uint32(INS_CSIZE(ins),4);
  GenericSet_uint32(INS_OLD_SIZE(ins),0);
  
  for (i = 0; i < 32; i += 2)
  {
    if (((a = rotate_left (value, i)) & 0xff) != 0)
    {
      t_uint32 firstpart,secondpart;

      if (a & 0xff00)
      {
	if (a & ~ 0xffff)
	{
	  continue;
	}
	else
	{
	  /* encode in two succeeding bytes */
	  firstpart = value & rotate_right(0xff,i);
	  secondpart = value & rotate_right(0xff00,i);
	}
      }
      else if (a & 0xff0000)
      {
	if (a & 0xff000000)
	{
	  continue;
	}
	else
	{
	  /* encode with one byte interleaved */
	  firstpart = value & rotate_right(0xff,i);
	  secondpart = value & rotate_right(0xff0000,i);
	}
      }
      else
      {
	/* encode with two bytes interleaved */
	firstpart = value & rotate_right(0xff,i);
	secondpart = value & rotate_right(0xff000000,i);
      }

      /* if we got here, the constant is encodable in two instructions */

      /* first instruction: the low byte */
      INS_TYPE(ins) = IT_DATAPROC;
      ARMINSOPCODE(ins) = ARM_MOV;
      ARMINSREGB(ins) = ARM_REG_NONE;
      ARMINSREGC(ins) = ARM_REG_NONE;
      ARMINSIMMEDIATE(ins) = firstpart;
      ARMINSREGS(ins) = ARM_REG_NONE;
      ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
      ARMINSSHIFTLENGTH(ins) = 0;
      ARMINSFLAGS(ins) = 0x0 | FL_IMMED; 
      /* flag the changes for the assembler */
      INS_ATTRIB(ins) |= IF_CHANGED;
      INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
      INS_REGS_USE(ins) = ArmUsedRegisters(ins);

      /* second instruction: the high byte */
      INS_TYPE(extra) = IT_DATAPROC;
      ARMINSCONDITION(extra) = ARMINSCONDITION(ins);
      ARMINSOPCODE(extra) = ARM_ORR;
      ARMINSREGA(extra) = ARMINSREGA(ins);
      ARMINSREGB(extra) = ARMINSREGA(ins);
      ARMINSREGC(extra) = ARM_REG_NONE;
      ARMINSIMMEDIATE(extra) = secondpart;
      ARMINSREGS(extra) = ARM_REG_NONE;
      ARMINSSHIFTTYPE(extra) = ARM_SHIFT_TYPE_NONE;
      ARMINSSHIFTLENGTH(extra) = 0;
      ARMINSFLAGS(extra) = 0x0 | FL_IMMED; 
      /* flag the changes for the assembler */
      INS_ATTRIB(extra) = INS_ATTRIB(ins) | IF_CHANGED;
      INS_REGS_DEF(extra) = ArmDefinedRegisters(extra);
      INS_REGS_USE(extra) = ArmUsedRegisters(extra);

      /* insert the extra instruction into the basic block */
      InsInsertAfter(extra,ins);

      return;
    }
  }

  for (i = 0; i < 32; i += 2)
  {
    if (((a = rotate_left (val_invert, i)) & 0xff) != 0)
    {
      t_uint32 firstpart,secondpart;

      if (a & 0xff00)
      {
	if (a & ~ 0xffff)
	{
	  continue;
	}
	else
	{
	  /* encode inverted in two succeeding bytes */
	  firstpart = val_invert & rotate_right(0xff,i);
	  secondpart = val_invert & rotate_right(0xff00,i);
	}
      }
      else if (a & 0xff0000)
      {
	if (a & 0xff000000)
	{
	  continue;
	}
	else
	{
	  /* encode inverted with one byte interleaved */
	  firstpart = val_invert & rotate_right(0xff,i);
	  secondpart = val_invert & rotate_right(0xff0000,i);
	}
      }
      else
      {
	/* encode inverted with two bytes interleaved */
	firstpart = val_invert & rotate_right(0xff,i);
	secondpart = val_invert & rotate_right(0xff000000,i);
      }

      /* if we got here, the constant can be encoded in two instructions */
 
      /* first instruction: the low byte */
      INS_TYPE(ins) = IT_DATAPROC;
      ARMINSOPCODE(ins) = ARM_MVN;
      ARMINSREGB(ins) = ARM_REG_NONE;
      ARMINSREGC(ins) = ARM_REG_NONE;
      ARMINSIMMEDIATE(ins) = firstpart;
      ARMINSREGS(ins) = ARM_REG_NONE;
      ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
      ARMINSSHIFTLENGTH(ins) = 0;
      ARMINSFLAGS(ins) = 0x0 | FL_IMMED; 
      /* flag the changes for the assembler */
      INS_ATTRIB(ins) |= IF_CHANGED;
      INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
      INS_REGS_USE(ins) = ArmUsedRegisters(ins);

      /* second instruction: the high byte */
      INS_TYPE(extra) = IT_DATAPROC;
      ARMINSCONDITION(extra) = ARMINSCONDITION(ins);
      ARMINSOPCODE(extra) = ARM_EOR;
      ARMINSREGA(extra) = ARMINSREGA(ins);
      ARMINSREGB(extra) = ARMINSREGA(ins);
      ARMINSREGC(extra) = ARM_REG_NONE;
      ARMINSIMMEDIATE(extra) = secondpart;
      ARMINSREGS(extra) = ARM_REG_NONE;
      ARMINSSHIFTTYPE(extra) = ARM_SHIFT_TYPE_NONE;
      ARMINSSHIFTLENGTH(extra) = 0;
      ARMINSFLAGS(extra) = 0x0 | FL_IMMED; 
      /* flag the changes for the assembler */
      INS_ATTRIB(extra) = INS_ATTRIB(ins) | IF_CHANGED;
      INS_REGS_DEF(extra) = ArmDefinedRegisters(extra);
      INS_REGS_USE(extra) = ArmUsedRegisters(extra);

      /* insert the extra instruction into the basic block */
      InsInsertAfter(extra,ins);

      return;
    }
  }

  /* if we got here, encoding in two constants was not possible */
  FATAL(("Could not encode constant %x in two instructions!\n",value));
}


/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
