/* Includes {{{ */
#include "thumb_assemble.h"
#include "thumb_assemble_one.h"
#include "arm_opcodes.h"
#include "arm_instructions.h"
#include "arm_registers.h"
#include "../../utils/util_error.h"
/* }}} */
/*!
 *
 * Assemble a Thumb instruction (t_thumb_ins to raw thumb instruction)
 *
 * \param i_ins The instruction to assemble
 *
 * \param data A buffer (must be at least 4 bytes long in which the instruction
 * is assembled
 *
 * \return void
*/
/* ThumbAssembleOne {{{ */
void ThumbAssembleOne(t_ins * i_ins, char * data)
{
    switch (INS_TYPE(i_ins)) {
    case IT_DATA:
	ThumbAssembleData(i_ins, (t_uint16*)data);
	break;
    case IT_BRANCH:
	if (ARMINSCONDITION(i_ins) != ARM_CONDITION_AL)
	    ThumbAssembleCondBranch(i_ins, (t_uint16*)data);
	else {
			if ((ARMINSOPCODE(i_ins) == ARM_BL) || (ARMINSOPCODE(i_ins) == ARM_BLX))
				ThumbAssembleBranchLink(i_ins, (t_uint32*)data);
			else
				ThumbAssembleBranch(i_ins, (t_uint16*)data);
	}
	break;
    case IT_DATAPROC:
	if (ARMINSSHIFTTYPE(i_ins) != ARM_SHIFT_TYPE_NONE) {
	    if (ARMINSREGS(i_ins) != ARM_REG_NONE)
		ThumbAssembleALU(i_ins, (t_uint16*)data);
	    else
		ThumbAssembleShifted(i_ins, (t_uint16*)data);
	}
	else {
	    switch (ARMINSOPCODE(i_ins)) {
	    case ARM_AND:
	    case ARM_EOR:
	    case ARM_ADC:
	    case ARM_SBC:
	    case ARM_ORR:
	    case ARM_BIC:
	    case ARM_TST:
	    case ARM_CMN:
	    case ARM_RSB:
	    case ARM_MVN:
		ThumbAssembleALU(i_ins, (t_uint16*)data);
		break;
	    case ARM_ADD:
	    case ARM_SUB:
		if (ARMINSFLAGS(i_ins) & FL_IMMED) {
		    if (ARMINSFLAGS(i_ins) & FL_S) {
			if(ARMINSREGA(i_ins) == ARMINSREGB(i_ins))
			    ThumbAssembleImm(i_ins, (t_uint16*)data);
			else
			    ThumbAssemble3Bit(i_ins, (t_uint16*)data);
		    }
		    else {
			if (ARMINSREGA(i_ins) == ARM_REG_R13)
			    ThumbAssembleStack(i_ins, (t_uint16*)data);
			else
			    ThumbAssembleLoadAddress(i_ins, (t_uint16*)data);
		    }
		}
		else {
		    if (ARMINSREGB(i_ins) > 0x7 || ARMINSREGC(i_ins) > 0x7)
			ThumbAssembleHiReg(i_ins, (t_uint16*)data);
		    else
			ThumbAssemble3Bit(i_ins, (t_uint16*)data);
		}
		break;
	    case ARM_MOV:
	    case ARM_CMP:
		if (ARMINSFLAGS(i_ins) & FL_IMMED)
		    ThumbAssembleImm(i_ins, (t_uint16*)data);
		else {
		    if (ARMINSREGB(i_ins) < 0x8 && ARMINSREGC(i_ins) < 0x8)
			ThumbAssembleALU(i_ins, (t_uint16*)data);
		    else
			ThumbAssembleHiReg(i_ins, (t_uint16*)data);
		}
		break;
	    default:
		FATAL(("No thumb instruction @I", i_ins));
	    }
	}
	break;
    case IT_LOAD:
    case IT_STORE:
	if (ARMINSREGB(i_ins) == ARM_REG_R13)
	    ThumbAssembleTransferSP(i_ins, (t_uint16 *) data);
	else {
	    if (ARMINSREGB(i_ins) == ARM_REG_R15)
		ThumbAssembleTransferPC(i_ins, (t_uint16 *) data);
	    else {
		if (ARMINSFLAGS(i_ins) & FL_IMMED) {
		    if (ARMINSOPCODE(i_ins) == ARM_STRH || ARMINSOPCODE(i_ins) == ARM_LDRH)
			ThumbAssembleTransferHalf(i_ins, (t_uint16 *) data);
		    else
			ThumbAssembleTransferImm(i_ins, (t_uint16 *) data);
		}
		else {
		    if (ARMINSOPCODE(i_ins) == ARM_STR || ARMINSOPCODE(i_ins) == ARM_LDR || ARMINSOPCODE(i_ins) == ARM_STRB || ARMINSOPCODE(i_ins) == ARM_LDRB)
			ThumbAssembleTransferRegOff(i_ins, (t_uint16 *) data);
		    else
			ThumbAssembleTransferSign(i_ins, (t_uint16 *) data);
		}
	    }
	}
	break;
    case IT_LOAD_MULTIPLE:
    case IT_STORE_MULTIPLE:
	if (ARMINSREGB(i_ins) == ARM_REG_R13)
	    ThumbAssemblePP(i_ins, (t_uint16 *) data);
	else
	    ThumbAssembleMultipleTransfer(i_ins, (t_uint16 *) data);
	break;
    case IT_SWI:
	ThumbAssembleSWI(i_ins, (t_uint16 *) data);
	break;
    case IT_MUL:
	ThumbAssembleALU(i_ins, (t_uint16 *) data);
	break;
    default:
    	ThumbAssembleUnsupported(i_ins, (t_uint16 *) data);
    }
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
