/* Includes {{{ */
#include "../../utils/util_error.h"
#include "../../utils/util_bitmanipulation.h"
#include "thumb_disassemble_one.h"
#include "thumb_opcodes.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_symbol.h"
#include "arm_utils.h"
#include "arm_registers.h"
#include "arm_opcodes.h"
#include "../../atom.h"
/*}}}*/
/*!
 * Disassemble Thumb Conditional Branches
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleCondBranch {{{ */
void ThumbDisassembleCondBranch(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	DiabloPrint(stdout, "Conditional Branch\n");
	INS_TYPE(ins) = IT_BRANCH;
	ARMINSCONDITION(ins) = (instr & 0x0f00) >> 8;
	ARMINSIMMEDIATE(ins) = 2 * (instr & 0x00ff);
	ARMINSFLAGS(ins) = FL_IMMED;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGA(ins) = ARM_REG_NONE;
	ARMINSREGB(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSOPCODE(ins) = ARM_B;
}
/*}}}*/
/*!
 * Disassemble Thumb Long Branch with Link (BL)
 *
 */
/* ThumbDisassembleBranchLink {{{ */
void ThumbDisassembleBranchLink(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_BRANCH;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGA(ins) = ARMINSREGB(ins) = ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = ((instr & 0x07ff0000) >> 15) | ((instr & 0x000007ff) << 12);
	ARMINSFLAGS(ins) = FL_IMMED;
	ARMINSOPCODE(ins) = ((instr & 0x10000000) == 0x10000000) ? ARM_BL : ARM_BLX;
}
/*}}}*/
/*!
 * Disassemble Thumb Branches (B, BX)
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleBranch {{{ */
void ThumbDisassembleBranch(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_BRANCH;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGA(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSFLAGS(ins) = 0x0;
	switch(opc) {
	case TH_B:
		ARMINSREGB(ins) = ARM_REG_NONE;
		ARMINSIMMEDIATE(ins) = (instr & 0x07ff) << 1;
		ARMINSFLAGS(ins) |= FL_IMMED;
		ARMINSOPCODE(ins) = ARM_B;
		break;
	case TH_BX:
		ARMINSREGB(ins) = (instr & 0x0078) >> 3;
		ARMINSIMMEDIATE(ins) = 0;
		ARMINSOPCODE(ins) = ARM_BX;
	};
}
/*}}}*/
/*!
 * Disassemble Thumb Software Interrupts (SWI)
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 * ThumbDisassembleSWI {{{ */
void ThumbDisassembleSWI(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_SWI;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSFLAGS(ins) = FL_IMMED;
	ARMINSREGA(ins) = ARMINSREGB(ins) = ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSOPCODE(ins) = ARM_SWI;
	ARMINSIMMEDIATE(ins) = instr & 0x00ff;
}
/*}}}*/
/*!
 * Disassemble Format 2: add/subtract
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassemble3Bit {{{ */
void ThumbDisassemble3Bit(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_DATAPROC;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSREGA(ins) = instr & 0x0007;
	ARMINSREGB(ins) = (instr & 0x0038) >> 3;
	ARMINSFLAGS(ins) = FL_S;
	switch (opc) {
	case TH_ADD_3R:
	case TH_SUB_3R:
		ARMINSREGC(ins) = (instr & 0x01c0) >> 6;
		ARMINSIMMEDIATE(ins) = 0;
		ARMINSOPCODE(ins) = (opc == TH_ADD_3R) ? ARM_ADD : ARM_SUB;
		break;
	case TH_ADD_3I:
	case TH_SUB_3I:
		ARMINSREGC(ins) = ARM_REG_NONE;
		ARMINSIMMEDIATE(ins) = (instr & 0x01c0) >> 6;
		ARMINSFLAGS(ins) |= FL_IMMED;
		ARMINSOPCODE(ins) = (opc == TH_ADD_3I) ? ARM_ADD : ARM_SUB;
	};
}
/*}}}*/
/*!
 * Disassemble Format 3: move/compare/add/subtract immediate
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleImm {{{ */
void ThumbDisassembleImm(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_DATAPROC;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = instr & 0x00ff;
	ARMINSFLAGS(ins) = FL_S | FL_IMMED;
	switch(opc) {
	case TH_MOV_I:
		ARMINSOPCODE(ins) = ARM_MOV;
		ARMINSREGA(ins) = (instr & 0x0700) >> 8;
		ARMINSREGB(ins) = ARM_REG_NONE;
		break;
	case TH_CMP_I:
		ARMINSOPCODE(ins) = ARM_CMP;
		ARMINSREGA(ins) = ARM_REG_NONE;
		ARMINSREGB(ins) = (instr & 0x0700) >> 8;
		break;
	case TH_ADD_2I:
	case TH_SUB_2I:
		ARMINSOPCODE(ins) = (opc == TH_ADD_2I) ? ARM_ADD : ARM_SUB;
		ARMINSREGA(ins) = (instr & 0x0700) >> 8;
		ARMINSREGB(ins) = (instr & 0x0700) >> 8;
	};
}
/*}}}*/
/*!
 * Disassemble Format 1: move shifted register
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleShifted {{{ */
void ThumbDisassembleShifted(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_DATAPROC;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSOPCODE(ins) = ARM_MOV;
	ARMINSFLAGS(ins) = FL_S;
	ARMINSIMMEDIATE(ins) = 0;
	ARMINSREGA(ins) = instr & 0x0007;
	ARMINSREGB(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = (instr & 0x0038) >> 3;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTLENGTH(ins) = (instr & 0x07c0) >> 6;
	if (opc == TH_LSL_I)
		ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_LSL_IMM;
	else if (opc == TH_LSR_I)
		ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_LSR_IMM;
	else
		ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_ASR_IMM;
}
/*}}}*/
/*!
 * Disassemble Thumb ALU operations
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleALU {{{ */
void ThumbDisassembleALU(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSIMMEDIATE(ins) = 0;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSFLAGS(ins) = FL_S;
	switch(opc) {
	case TH_AND:
	case TH_EOR:
	case TH_ADC:
	case TH_SBC:
	case TH_ORR:
	case TH_BIC:
		INS_TYPE(ins) = IT_DATAPROC;
		ARMINSREGA(ins) = instr & 0x0007;
		ARMINSREGB(ins) = instr & 0x0007;
		ARMINSREGC(ins) = (instr & 0x0038) >> 3;
		ARMINSREGS(ins) = ARM_REG_NONE;
		ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
		if (opc == TH_AND)
			ARMINSOPCODE(ins) = ARM_AND;
		else if (opc == TH_EOR)
			ARMINSOPCODE(ins) = ARM_EOR;
		else if (opc == TH_ADC)
			ARMINSOPCODE(ins) = ARM_ADC;
		else if (opc == TH_SBC)
			ARMINSOPCODE(ins) = ARM_SBC;
		else if (opc == TH_ORR)
			ARMINSOPCODE(ins) = ARM_ORR;
		else
			ARMINSOPCODE(ins) = ARM_BIC;
		break;
	case TH_LSL_R:
	case TH_LSR_R:
	case TH_ASR_R:
	case TH_ROR:
		INS_TYPE(ins) = IT_DATAPROC;
		ARMINSOPCODE(ins) = ARM_MOV;
		ARMINSREGA(ins) = instr & 0x0007;
		ARMINSREGB(ins) = ARM_REG_NONE;
		ARMINSREGC(ins) = instr & 0x0007;
		ARMINSREGS(ins) = (instr & 0x0038) >> 3;
		if (opc == TH_LSL_R)
			ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_LSL_REG;
		else if (opc == TH_LSR_R)
			ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_LSR_REG;
		else if (opc == TH_ASR_R)
			ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_ASR_REG;
		else
			ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_ROR_REG;
		break;
	case TH_TST:
	case TH_CMN:
	case TH_CMP_R:
		INS_TYPE(ins) = IT_DATAPROC;
		ARMINSREGA(ins) = ARM_REG_NONE;
		ARMINSREGB(ins) = instr & 0x0007;
		ARMINSREGC(ins) = (instr & 0x0038) >> 3;
		ARMINSREGS(ins) = ARM_REG_NONE;
		ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
		if (opc == TH_TST)
			ARMINSOPCODE(ins) = ARM_TST;
		else if (opc == TH_CMN)
			ARMINSOPCODE(ins) = ARM_CMN;
		else
			ARMINSOPCODE(ins) = ARM_CMP;
		break;
	case TH_NEG:
		INS_TYPE(ins) = IT_DATAPROC;
		ARMINSOPCODE(ins) = ARM_RSB;
		ARMINSREGA(ins) = instr & 0x0007;
		ARMINSREGB(ins) = (instr & 0x0038) >> 3;
		ARMINSREGC(ins) = ARM_REG_NONE;
		ARMINSREGS(ins) = ARM_REG_NONE;
		ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
		ARMINSFLAGS(ins) |= FL_IMMED;
		break;
	case TH_MUL:
		INS_TYPE(ins) = IT_MUL;
		ARMINSOPCODE(ins) = ARM_MUL;
		ARMINSREGA(ins) = instr & 0x0007;
		ARMINSREGB(ins) = (instr & 0x0038) >> 3;
		ARMINSREGC(ins) = instr & 0x0007;
		ARMINSREGS(ins) = ARM_REG_NONE;
		ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
		break;
	case TH_MVN:
		INS_TYPE(ins) = IT_DATAPROC;
		ARMINSOPCODE(ins) = ARM_MVN;
		ARMINSREGA(ins) = instr & 0x0007;
		ARMINSREGB(ins) = ARM_REG_NONE;
		ARMINSREGC(ins) = (instr & 0x0038) >> 3;
		ARMINSREGS(ins) = ARM_REG_NONE;
		ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	};
}
/*}}}*/
/*!
 * Disassemble Format 5: Hi register operations
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleHiReg {{{ */
void ThumbDisassembleHiReg(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_DATAPROC;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGC(ins) = (instr & 0x0078) >> 3;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = 0;
	ARMINSFLAGS(ins) = 0x0;
	switch(opc) {
	case TH_ADD_2RH:
		ARMINSOPCODE(ins) = ARM_ADD;
		ARMINSREGA(ins) = ((instr & 0x0080) >> 4) | (instr & 0x0007);
		ARMINSREGB(ins) = ((instr & 0x0080) >> 4) | (instr & 0x0007);
		break;
	case TH_CMP_RH:
		ARMINSOPCODE(ins) = ARM_CMP;
		ARMINSREGA(ins) = ARM_REG_NONE;
		ARMINSREGB(ins) = ((instr & 0x0080) >> 4) | (instr & 0x0007);
		ARMINSFLAGS(ins) |= FL_S;
		break;
	case TH_MOV_RH:
		ARMINSOPCODE(ins) = ARM_MOV;
		ARMINSREGA(ins) = ((instr & 0x0080) >> 4) | (instr & 0x0007);
		ARMINSREGB(ins) = ARM_REG_NONE;
	};
}
/*}}}*/
/*!
 * Disassemble Load address
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleLoadAddress {{{ */
void ThumbDisassembleLoadAddress(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_DATAPROC;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSOPCODE(ins) = ARM_ADD;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGA(ins) = (instr & 0x0700) >> 8;
	ARMINSREGB(ins) = (opc == TH_ADD_PC) ? ARM_REG_R15 : ARM_REG_R13;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = (instr & 0x00ff) << 2;
	ARMINSFLAGS(ins) = FL_IMMED;
}
/*}}}*/
/*!
 * Disassemble Format 13: add offset to SP
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleStack {{{ */
void ThumbDisassembleStack(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_DATAPROC;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGA(ins) = ARM_REG_R13;
	ARMINSREGB(ins) = ARM_REG_R13;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = (instr & 0x007f) << 2;
	ARMINSFLAGS(ins) = FL_IMMED;
	ARMINSOPCODE(ins) = (opc == TH_ADD_SPI) ? ARM_ADD : ARM_SUB;
}
/*}}}*/
/*!
 * Disassemble Format 9: load/store with immediate offset
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleTransferImm {{{ */
void ThumbDisassembleTransferImm(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSFLAGS(ins) = FL_PREINDEX | FL_DIRUP | FL_IMMED;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGA(ins) = instr & 0x0007;
	ARMINSREGB(ins) = (instr & 0x0038) >> 3;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	switch (opc) {
	case TH_STR_I:
		INS_TYPE(ins) = IT_STORE;
		ARMINSOPCODE(ins) = ARM_STR;
		ARMINSIMMEDIATE(ins) = 4 * ((instr & 0x07c0) >> 6);
		break;
	case TH_LDR_I:
		INS_TYPE(ins) = IT_LOAD;
		ARMINSOPCODE(ins) = ARM_LDR;
		ARMINSIMMEDIATE(ins) = 4 * ((instr & 0x07c0) >> 6);
		break;
	case TH_STRB_I:
		INS_TYPE(ins) = IT_STORE;
		ARMINSOPCODE(ins) = ARM_STRB;
		ARMINSIMMEDIATE(ins) = 1 * ((instr & 0x07c0) >> 6);
		break;
	case TH_LDRB_I:
		INS_TYPE(ins) = IT_LOAD;
		ARMINSOPCODE(ins) = ARM_LDRB;
		ARMINSIMMEDIATE(ins) = 1 * ((instr & 0x07c0) >> 6);
	};
}
/*}}}*/
/*!
 * Disassemble Format 10: load/store halfword
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleTransferHalf {{{ */
void ThumbDisassembleTransferHalf(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSREGA(ins) = instr & 0x0007;
	ARMINSREGB(ins) = (instr & 0x0038) >> 3;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSFLAGS(ins) |= FL_PREINDEX | FL_DIRUP | FL_IMMED;
	ARMINSIMMEDIATE(ins) = 2 * ((instr & 0x07c0) >> 6);
	switch (opc) {
	case TH_STRH_I:
		INS_TYPE(ins) = IT_STORE;
		ARMINSOPCODE(ins) = ARM_STRH;
		break;
	case TH_LDRH_I:
		INS_TYPE(ins) = IT_LOAD;
		ARMINSOPCODE(ins) = ARM_LDRH;
	};
}
/*}}}*/
/*!
 * Disassemble Format 8: load/store sign-extended byte/halfword
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleTransferSign {{{ */
void ThumbDisassembleTransferSign(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSREGA(ins) = instr & 0x0007;
	ARMINSREGB(ins) = (instr & 0x0038) >> 3;
	ARMINSREGC(ins) = (instr & 0x01c0) >> 6;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSFLAGS(ins) = FL_PREINDEX | FL_DIRUP;
	ARMINSIMMEDIATE(ins) = 0;
	switch (opc) {
	case TH_STRH_R:
		INS_TYPE(ins) = IT_STORE;
		ARMINSOPCODE(ins) = ARM_STRH;
		break;
	case TH_LDRH_R:
		INS_TYPE(ins) = IT_LOAD;
		ARMINSOPCODE(ins) = ARM_LDRH;
		break;
	case TH_LDRSB:
		INS_TYPE(ins) = IT_LOAD;
		ARMINSOPCODE(ins) = ARM_LDRSB;
		break;
	case TH_LDRSH:
		INS_TYPE(ins) = IT_LOAD;
		ARMINSOPCODE(ins) = ARM_LDRSH;
	};
}
/*}}}*/
/*!
 * Disassemble Format 7: load/store with register offset
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleTransferRegOff {{{ */
void ThumbDisassembleTransferRegOff(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSREGA(ins) = instr & 0x0007;
	ARMINSREGB(ins) = (instr & 0x0038) >> 3;
	ARMINSREGC(ins) = (instr & 0x01c0) >> 6;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSFLAGS(ins) = FL_PREINDEX | FL_DIRUP;
	ARMINSIMMEDIATE(ins) = 0;
	switch (opc) {
	case TH_STR_R:
		INS_TYPE(ins) = IT_STORE;
		ARMINSOPCODE(ins) = ARM_STR;
		break;
	case TH_LDR_R:
		INS_TYPE(ins) = IT_LOAD;
		ARMINSOPCODE(ins) = ARM_LDR;
		break;
	case TH_STRB_R:
		INS_TYPE(ins) = IT_STORE;
		ARMINSOPCODE(ins) = ARM_STRB;
		break;
	case TH_LDRB_R:
		INS_TYPE(ins) = IT_LOAD;
		ARMINSOPCODE(ins) = ARM_LDRB;
	};
}
/*}}}*/
/*!
 * Disassemble Format 6: PC-relative load
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleTransferPC {{{ */
void ThumbDisassembleTransferPC(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	INS_TYPE(ins) = IT_LOAD;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSREGA(ins) = (instr & 0x0700) >> 8;
	ARMINSREGB(ins) = ARM_REG_R15;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSFLAGS(ins) = FL_PREINDEX | FL_DIRUP | FL_IMMED;
	ARMINSIMMEDIATE(ins) = 4 * (instr & 0x00ff);
	ARMINSOPCODE(ins) = ARM_LDR;
}
/*}}}*/
/*!
 * Disassemble Format 11: SP-relative load
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleTransferSP {{{ */
void ThumbDisassembleTransferSP(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSREGA(ins) = (instr & 0x0700) >> 8;
	ARMINSREGB(ins) = ARM_REG_R13;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSFLAGS(ins) = FL_PREINDEX | FL_DIRUP | FL_IMMED;
	ARMINSIMMEDIATE(ins) = 4 * (instr & 0x00ff);
	switch (opc) {
	case TH_STR_SP:
		INS_TYPE(ins) = IT_STORE;
		ARMINSOPCODE(ins) = ARM_STR;
		break;
	case TH_LDR_SP:
		INS_TYPE(ins) = IT_LOAD;
		ARMINSOPCODE(ins) = ARM_LDR;
	};
}
/*}}}*/
/*!
 * Disassemble Format 15: multiple load/store
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassembleMultipleTransfer {{{ */
void ThumbDisassembleMultipleTransfer(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSFLAGS(ins) = 0x0;

	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGS(ins) = ARM_REG_NONE;

	ARMINSREGA(ins) = ARMINSREGC(ins) = ARM_REG_NONE;

	switch(opc) {
	case TH_STMIA:
	case TH_LDMIA:
		INS_TYPE(ins) = (opc == TH_STMIA) ? IT_STORE_MULTIPLE : IT_LOAD_MULTIPLE;
		ARMINSOPCODE(ins) = (opc == TH_STMIA) ? ARM_STM : ARM_LDM;
		ARMINSREGB(ins) = (instr & 0x0700) >> 8;
		ARMINSIMMEDIATE(ins) = instr & 0x00ff;
		ARMINSFLAGS(ins) |= FL_DIRUP | FL_WRITEBACK;
	};
}
/*}}}*/
/*!
 * Disassemble Thumb Push & Pop
 *
 * \param ins The thumb instruction that gets filled in
 * \param instr The encoded thumb instruction
 * \param opc The opcode
 *
 * \return void
 */
/* ThumbDisassemblePP {{{ */
 void ThumbDisassemblePP(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	/*ARMINSTEXT(ins) = instr; */
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSFLAGS(ins) = 0x0;

	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSREGA(ins) = ARMINSREGC(ins) = ARM_REG_NONE;
	INS_TYPE(ins) = (opc == TH_PUSH) ? IT_STORE_MULTIPLE : IT_LOAD_MULTIPLE;
	ARMINSOPCODE(ins) = (opc == TH_PUSH) ? ARM_STM : ARM_LDM;
	ARMINSREGB(ins) = ARM_REG_R13;
	ARMINSIMMEDIATE(ins) = 0 | (instr & 0x00ff) | ((instr & 0x0100) << ((opc == TH_PUSH) ? 6 : 7));
	ARMINSFLAGS(ins) |= FL_WRITEBACK | ( (opc == TH_PUSH) ? FL_PREINDEX : FL_DIRUP );
}
/* }}} */
/* ThumbDisassembleUnsupported {{{ */
void ThumbDisassembleUnsupported(t_ins * ins, t_uint32 instr, t_uint16 opc)
{
	FATAL(("Unsupported thumb instruction. Instr = 0x%x\n", instr));
}
/* }}} */
/*!
 * \todo Document
 *
 *
 * \param ins
 * \param instr
 * \param opc
 *
 * \return void
*/
/* Data {{{ */
void ThumbDisassembleData(t_ins * ins, t_uint32 instr, t_uint16 opc ) {
	  /* fills in a thumb_ins with the appropriate values if we're dealing with a data word */
	INS_TYPE(ins) = IT_DATA;
	ARMINSOPCODE(ins) = ARM_DATA;
	ARMINSCONDITION(ins) = ARM_CONDITION_AL;
	ARMINSREGA(ins) = ARMINSREGB(ins) = ARMINSREGC(ins) = ARMINSREGS(ins) = ARM_REG_NONE;
	ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
	ARMINSSHIFTLENGTH(ins) = 0;
	ARMINSIMMEDIATE(ins) = instr;
	ARMINSFLAGS(ins) = FL_THUMB | FL_IMMED;
	INS_ATTRIB(ins) = IF_DATA;
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
