/***********************************************************
 * Copyright 2001,2002,2003: {{{
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * }}}
 **********************************************************/

#ifdef INSTRUMENTATION

#define USE_PSEUDO
#define DIRECT_ARGS


#include "i386_instructions.h"
#include "i386_opcode.h"
#include "i386_registers.h"
#include "../../atom_internal.h"
#include "../../diablo_options.h"
#include "../../kernel/diablo_liveness.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_stack_machine.h"
#include "../../kernel/diablo_instrument.h"
#include "../../kernel/diablo_print.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include <string.h>

extern t_bool use_old_addresses;
static t_section * stack_save_area;
static t_section * fpu_save_area;
static t_section * int_save_area;

#ifdef USE_PSEUDO
t_regset always_stack;
#endif

#define FPU_SAVE_AREA_SIZE	108
#define STACK_SAVE_AREA_SIZE	4
#define INTREG_SAVE_AREA_SIZE	32


/**************************************************************************/
/**                       GENERAL INSTRUMENTATION                        **/
/**************************************************************************/


/* patch functions */
/* {{{ I386PatchRepPrefixes */
void I386PatchRepPrefixes(t_object * obj)
{
  t_cfg * cfg = SECTION_CFG(obj->code[0]);
  t_bbl * bbl; 
  t_ins * ins;
  t_regset affected = RegsetNew();

  ComLiveness(cfg,TRIVIAL);
  ComputeSavedChangedRegisters(CFG_SECTION(cfg));
  ComLiveness(cfg,SENSITIVE);
  ComputeSavedChangedRegisters(CFG_SECTION(cfg));
  ComLiveness(cfg,SENSITIVE);

  RegsetAddReg(affected,I386_CONDREG_AF);
  RegsetAddReg(affected,I386_CONDREG_ZF);
  RegsetAddReg(affected,I386_CONDREG_PF);
  RegsetAddReg(affected,I386_CONDREG_SF);
  RegsetAddReg(affected,I386_CONDREG_OF);
  RegsetAddReg(affected,I386_CONDREG_CF);

  /* change instructions of the form 
   *   rep stosb ...
   * into
   *   <pushf>
   *   cmp 0, %ecx
   *   je skip
   *  next:     _____ InstBefore
   *   stosb ..._____ InstAfter
   *   loop next
   *  skip:
   *   <popf>
   *   ...      _____ BlockAfter
   */

  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS(bbl,ins)
    {
      t_bbl * before, * pblock, * after;
      t_ins * loop, * cmp, * je, * nop, * pushf, * popf;
      t_regset tmp;
      t_bool save_condregs;

      if (!I386_INS_HAS_PREFIX(ins, I386_PREFIX_REP | I386_PREFIX_REPZ | I386_PREFIX_REPNZ))
	continue;
      if (!INS_TEMP(ins))  /* don't patch if there aren't any markers attached to the instruction */
	continue;

      /* check if the condition registers are live before ins */
      RegsetDup(tmp,affected);
      RegsetSetIntersect(tmp,InsRegsLiveBefore(ins));
      save_condregs = !RegsetIsEmpty(tmp);


      if (!INS_NEXT(ins))
      {
	nop = InsNewForBbl(bbl);
	I386InstructionMakeNoop(nop);
	InsAppendToBbl(nop,bbl);
	INS_OLD_ADDRESS(nop) = GENERIC_ADD(INS_OLD_ADDRESS(ins),INS_OLD_SIZE(ins));
	INS_OLD_SIZE(nop) = GENERIC_SUB(INS_CSIZE(ins),INS_CSIZE(ins));
      }

      /* split the block */
      before = bbl;
      pblock = AtomSplitBlock(bbl,ins,TRUE);
      after = AtomSplitBlock(pblock,ins,FALSE);

      /* add the edges */
      CfgEdgeNew(cfg,before,pblock,ET_FALLTHROUGH);
      CfgEdgeNew(cfg,before,after,ET_JUMP);
      CfgEdgeNew(cfg,pblock,pblock,ET_JUMP);
      CfgEdgeNew(cfg,pblock,after,ET_FALLTHROUGH);

      /* add the pushf and popf instructions if necessary */
      if (save_condregs)
      {
	I386MakeInsForBbl(Simple,Append,pushf,before,I386_PUSHF);
	I386MakeInsForBbl(Simple,Prepend,popf,after,I386_POPF);
      }

      /* add the jmp and cmp instructions */
      cmp = InsNewForBbl(before);
      I386InstructionMakeCmp(cmp,I386_REG_ECX,I386_REG_NONE,0);
      InsAppendToBbl(cmp,before);
      je = InsNewForBbl(before);
      I386InstructionMakeCondJump(je,I386_CONDITION_Z);
      InsAppendToBbl(je,before);

      /* add the loop instruction */
      loop = InsNewForBbl(pblock);
      InsAppendToBbl(loop,pblock);
      if (I386_INS_OPCODE(ins) != I386_CMPSB && I386_INS_OPCODE(ins) != I386_CMPSD &&
	  I386_INS_OPCODE(ins) != I386_SCASB && I386_INS_OPCODE(ins) != I386_SCASD)
      {
	/* REP prefix */
	I386InstructionMakeCondJump(loop,I386_CONDITION_LOOP);
      }
      else
      {
	/* REPE/REPNE prefix */
	if (I386_INS_HAS_PREFIX(ins,I386_PREFIX_REPZ))
	  I386InstructionMakeCondJump(loop,I386_CONDITION_LOOPZ);
	else
	  I386InstructionMakeCondJump(loop,I386_CONDITION_LOOPNZ);
      }
      I386_INS_PREFIXES(ins) &= ~(I386_PREFIX_REP | I386_PREFIX_REPE | I386_PREFIX_REPNE);
      INS_CSIZE(ins) = GENERIC_SUB_UINT32(INS_CSIZE(ins),1);
    }
}
/* }}} */
/* {{{ I386PatchCmps */
void I386PatchCmps(t_object * obj)
{
  /* for cmps instructions, there are two source operands in memory.
   * if EffLoadAddrValue is queried for these, we add a second query for 
   * EffLoadAddrValue2, which returns the address of the second load from memory */
  t_cfg * cfg = SECTION_CFG(obj->code[0]);
  t_bbl * bbl;
  t_ins * ins;

  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS(bbl,ins)
    {
      if (I386_INS_OPCODE(ins) == I386_CMPSB || I386_INS_OPCODE(ins) == I386_CMPSD)
      {
	t_atom_instrument_marker * m;
	for (m = INS_TEMP(ins); m; m = m->next)
	{
	  int i;
	  t_atom_instrument_marker * new;
	  t_atom_proto * p = m->proto;
	  t_bool dup = FALSE;

	  for (i = 0; i < p->nargs; i++)
	    if (p->args[i] == ATOM_SPECIAL_VALUE && (m->values[i].query == EffAddrValue || m->values[i].query == EffLoadAddrValue))
	    {
	      dup = TRUE;
	      break;
	    }
	  if (!dup) continue;

	  new = AtomInstrumentMarkerNew(m->when,p);
	  new->values = Malloc(sizeof(t_atom_value)*p->nargs);
	  memcpy(new->values, m->values, sizeof(t_atom_value)*p->nargs);
	  for (i = 0; i < p->nargs; i++)
	    if (p->args[i] == ATOM_SPECIAL_VALUE && (m->values[i].query == EffAddrValue || m->values[i].query == EffLoadAddrValue))
	      new->values[i].query = EffLoadAddrValue2;
	  new->next = m->next;
	  m->next = new;
	}
      }
    }
}
/* }}} */

/* {{{ I386AtomPatchTranslateFunctions */
void I386AtomPatchTranslateFunctions(t_section *code)
{
  t_ins * ins;
  t_address f;
  t_uint32 * data;
  GenericSet_uint32(f,0xffffffff);
  SECTION_FOREACH_INS(code,ins)
  {
    if (!GENERICISNULL(SECTION_CSIZE(added_section)))
    {
      data = (t_uint32 *)added_section->data;

      if (I386_INS_OPCODE(ins) == I386_CMP &&
	  I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_imm &&
	  I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins)) == 0xfefefefe &&
	  I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg &&
	  I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_EAX)
      {
	/* fefefefe: upper limit of transtable */
	I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins)) = data[4*data[0]+1];
      }
      if (I386_INS_OPCODE(ins) == I386_CMP &&
	  I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_imm &&
	  I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins)) == 0xfdfdfdfd &&
	  I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg &&
	  I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_EAX)
      {
	/* fdfdfdfd: lower limit of transtable */
	I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins)) = data[4];
      }
      if (I386_INS_OPCODE(ins) == I386_MOV &&
	  I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm &&
	  I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) == 0xfcfcfcfc &&
	  I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg &&
	  I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_EDX)
      {
	/* fcfcfcfc: (table_size+1)*16 */
	I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) = (data[0]+1)*16;
      }
    }

    if (!GENERICISNULL(SECTION_CSIZE(added_section2)))
    {
      data = (t_uint32 *)added_section2->data;

      if (I386_INS_OPCODE(ins) == I386_CMP &&
	  I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_imm &&
	  I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins)) == 0xf9f9f9f9 &&
	  I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg &&
	  I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_EAX)
      {
	/* f9f9f9f9: upper limit of reverse_transtable */
	I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins)) = data[4*data[0]+1];
      }
      if (I386_INS_OPCODE(ins) == I386_CMP &&
	  I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_imm &&
	  I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins)) == 0xf8f8f8f8 &&
	  I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg &&
	  I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_EAX)
      {
	/* f8f8f8f8: lower limit of reverse_transtable */
	I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins)) = data[4];
      }
      if (I386_INS_OPCODE(ins) == I386_MOV &&
	  I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm &&
	  I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) == 0xf7f7f7f7 &&
	  I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg &&
	  I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_EDX)
      {
	/* f7f7f7f7: (table_size+1)*16 */
	I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) = (data[0]+1)*16;
      }
    }
  }
} /* }}} */

/** The real instrumentation code **/
/* {{{ Estimate changed registers */
t_regset I386AtomEstimateChangedRegisters(t_atom_instrument_marker * m)
{
  t_atom_proto * p = m->proto;
  t_function * fun = BBL_FUN(p->bbl);
  t_uint32 i;
  t_regset ret = NullRegs;
  t_regset tmp;

  /* add the registers that are changed by the called function */
  tmp = FUN_REGS_CHANGED(fun);
  RegsetSetDiff(tmp,FUN_REGS_SAVED(fun));
  RegsetSetUnion(ret,tmp);

  /* if there is more than one argument, the condition registers will certainly
   * be changed by the instruction that allocates space on the stack */
  if (p->nargs > 1)
  {
#define addf(x) RegsetAddReg(ret,I386_CONDREG_##x)
    addf(OF); addf(SF); addf(ZF); addf(AF); addf(PF); addf(CF);
#undef addf
  }

  for (i=0; i<p->nargs; i++)
  {
    switch (p->args[i])
    {
      case ATOM_REG_VALUE:
      case ATOM_CHAR_VALUE:
      case ATOM_INT_VALUE:
      case ATOM_LONG_VALUE:
      case ATOM_LONG_LONG_VALUE:
	/* these do not change any registers */
	break;
      case ATOM_POINTER_VALUE:
	FATAL(("implement"));
	break;
      case ATOM_SPECIAL_VALUE:
	switch (m->values[i].query)
	{
	  case FunctionName:
	    /* no registers changed */
	    break;
	  case BrCondValue:
	    /* %eax is changed:
	     *   - %ah is used to save the condition flags while allocating stack space for
	     *     the function arguments
	     *   - %eax is used as a temporary register while evaluating loop(n)z conditions */
	    RegsetAddReg(ret,I386_REG_EAX);
	    break;
	  case EffAddrValue:
	  case EffLoadAddrValue:
	  case EffLoadAddrValue2:
	  case EffStoreAddrValue:
	    /* these change either %eax, %ecx or %edx */
	    RegsetAddReg(ret,I386_REG_EAX);
	    RegsetAddReg(ret,I386_REG_ECX);
	    RegsetAddReg(ret,I386_REG_EDX);
	    /* these also call Translate(), so we have to add the changed registers of this function as well */
	    tmp = FUN_REGS_CHANGED(translate);
	    RegsetSetDiff(tmp,FUN_REGS_SAVED(translate));
	    RegsetSetUnion(ret,tmp);
	    break;
	  default:
	    FATAL(("IMPLEMENT"));
	}
	break;
      default:
	FATAL(("IMPLEMENT"));
    }
  }
  return ret;
} /* }}} */

/* {{{ Add register saves */
t_uint32 I386AtomAddSave(t_bbl * in, t_regset liveregs, t_regset changedregs)
{
  t_ins * save_regs, * save_flags, * save_fpu;
  t_regset condflags = NullRegs;
  t_regset fpuregs = NullRegs;
  t_regset tmp;
  t_regset free_regs;
  t_reg r;
  t_uint32 stack_change = 0;

  //if (RegsetIsEmpty(condflags)) 
  {
#define addf(flag)  RegsetAddReg(condflags,I386_CONDREG_##flag)
    addf(AF); addf(CF); addf(DF); addf(OF); addf(PF); addf(SF); addf(ZF);
#undef addf
#define addf(flag)  RegsetAddReg(fpuregs,I386_CONDREG_##flag)
#define addr(reg)   RegsetAddReg(fpuregs,I386_REG_##reg)
    addf(C0); addf(C1); addf(C2); addf(C3);
    addr(ST0); addr(ST1); addr(ST2); addr(ST3); addr(ST4); addr(ST5); addr(ST6); addr(ST7);
#undef addf
#undef addr
  }
  
  RegsetDup(tmp,changedregs);
  RegsetSetIntersect(tmp,liveregs);
  RegsetSubReg(tmp,I386_REG_ESP); /* no need to change esp */

  RegsetSetIntersect(tmp,in->cfg->sec->obj->description->int_registers);

  RegsetDup(free_regs,changedregs);
  RegsetSetUnion(free_regs,liveregs);
  RegsetSetInvers(free_regs);
  RegsetSetIntersect(free_regs,in->cfg->sec->obj->description->int_registers);
  RegsetSubReg(free_regs,I386_REG_ESP);

  stack_change = 0;
  REGSET_FOREACH_REG(tmp,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForBbl(Push,Append,save_regs,in,r,0);
    stack_change += 4;
#else
    if (RegsetIn(always_stack,r))
    {
      if (!RegsetIsEmpty(free_regs))
      {
	t_reg r2;
	REGSET_FOREACH_REG(free_regs,r2)
	  break;
	RegsetSubReg(free_regs,r2);
	I386MakeInsForBbl(MovToReg,Append,save_regs,in,r2,r,0);
      }
      else
      {
	I386MakeInsForBbl(Push,Append,save_regs,in,r,0);
	stack_change += 4;
      }
    }
    else
      I386MakeInsForBbl(PseudoSave,Append,save_regs,in,r);
#endif
  }
  
  RegsetDup(tmp,liveregs);
  RegsetSetIntersect(tmp,condflags);
  if (!RegsetIsEmpty(tmp))
  {
    I386MakeInsForBbl(Simple,Append,save_flags,in,I386_PUSHF);
    stack_change += 4;
  }

#if 0
  RegsetDup(tmp,changedregs);
  RegsetSetIntersect(tmp,liveregs);
  RegsetSetIntersect(tmp,fpuregs);
  if (!RegsetIsEmpty(tmp))
  {
    t_reloc * rel;
    /* save the fpu state */
    I386MakeInsForBbl(FSaveRstor,Append,save_fpu,in,I386_FSAVE,0,I386_REG_NONE,I386_REG_NONE,0);
    I386_OP_FLAGS(I386_INS_DEST(save_fpu)) |= I386_OPFLAG_ISRELOCATED;
    I386_OP_MEMOPSIZE(I386_INS_DEST(save_fpu)) = 0;
    rel = RelocTableAddRelocToRelocatable(BBL_CFG(in)->sec->obj->reloc_table,
	NULL_GENERIC,(t_relocatable *)save_fpu,GENERIC_ADD_UINT32(NULL_GENERIC,2),
	(t_relocatable *)fpu_save_area,NULL_GENERIC,FALSE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);
    rel->label = StringDup("fpu_save_area");
  }
#endif

  return stack_change;
} /* }}} */

/* {{{ Add register restores */
void I386AtomAddRestore(t_bbl * out,t_regset liveregs,t_regset changedregs,t_uint32 stack_change) 
{
  t_ins * restore_regs, * restore_flags, * restore_fpu;
  t_regset condflags = NullRegs;
  t_regset fpuregs = NullRegs;
  t_regset tmp;
  t_reg r;

  {
#define addf(flag)  RegsetAddReg(condflags,I386_CONDREG_##flag)
    addf(AF); addf(CF); addf(DF); addf(OF); addf(PF); addf(SF); addf(ZF);
#undef addf
#define addf(flag)  RegsetAddReg(fpuregs,I386_CONDREG_##flag)
#define addr(reg)   RegsetAddReg(fpuregs,I386_REG_##reg)
    addf(C0); addf(C1); addf(C2); addf(C3);
    addr(ST0); addr(ST1); addr(ST2); addr(ST3); addr(ST4); addr(ST5); addr(ST6); addr(ST7);
#undef addf
#undef addr
  }

  RegsetDup(tmp,changedregs);
  RegsetSetIntersect(tmp,liveregs);
  RegsetSubReg(tmp,I386_REG_ESP);
  RegsetSetIntersect(tmp,out->cfg->sec->obj->description->int_registers);

#ifdef USE_PSEUDO
  t_regset free_regs;
  RegsetDup(free_regs,changedregs);
  RegsetSetUnion(free_regs,liveregs);
  RegsetSetInvers(free_regs);
  RegsetSetIntersect(free_regs,out->cfg->sec->obj->description->int_registers);
  RegsetSubReg(free_regs,I386_REG_ESP);
#endif

  REGSET_FOREACH_REG(tmp,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForBbl(Pop,Prepend,restore_regs,out,r);
#else
    if (RegsetIn(always_stack,r))
    {
      if (!RegsetIsEmpty(free_regs))
      {
	t_reg r2;
	REGSET_FOREACH_REG(free_regs,r2)
	  break;
	RegsetSubReg(free_regs,r2);
	I386MakeInsForBbl(MovToReg,Prepend,restore_regs,out,r,r2,0);
      }
      else
	I386MakeInsForBbl(Pop,Prepend,restore_regs,out,r);
    }
    else
      I386MakeInsForBbl(PseudoLoad,Prepend,restore_regs,out,r);
#endif
  }

  RegsetDup(tmp,liveregs);
  RegsetSetIntersect(tmp,condflags);
  if (!RegsetIsEmpty(tmp))
    I386MakeInsForBbl(Simple,Prepend,restore_flags,out,I386_POPF);

#if 0
  RegsetDup(tmp,changedregs);
  RegsetSetIntersect(tmp,liveregs);
  RegsetSetIntersect(tmp,fpuregs);
  if (!RegsetIsEmpty(tmp))
  {
    t_reloc * rel;
    /* restore the fpu state */
    I386MakeInsForBbl(FSaveRstor,Prepend,restore_fpu,out,I386_FRSTOR,0,I386_REG_NONE,I386_REG_NONE,0);
    I386_OP_FLAGS(I386_INS_SOURCE1(restore_fpu)) |= I386_OPFLAG_ISRELOCATED;
    I386_OP_MEMOPSIZE(I386_INS_SOURCE1(restore_fpu)) = 0;
    rel = RelocTableAddRelocToRelocatable(BBL_CFG(out)->sec->obj->reloc_table,
	NULL_GENERIC,(t_relocatable *)restore_fpu,GENERIC_ADD_UINT32(NULL_GENERIC,2),
	(t_relocatable *)fpu_save_area,NULL_GENERIC,FALSE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);
    rel->label = StringDup("fpu_save_area");
  }
#endif

} /* }}} */

/* {{{ I386PutParamOnStack -- helper function
 * moves the parameter on the stack, but if there is only one parameter it uses a push instead of a mov
 * (this saves an instruction) */
t_ins * I386PutParamOnStack(t_bbl * bbl, t_reg reg, t_uint32 imm, t_uint32 paramno)
{
  t_ins * param;
  t_uint32 paramoffset = paramno*4;
  I386MakeInsForBbl(MovToMem,Append,param,bbl,paramoffset,I386_REG_ESP,I386_REG_NONE,0,reg,imm);
  return param;
}
/* }}} */
/* {{{ I386AddCalcEffAddrValue -- helper function */
static t_ins * I386AddCalcEffAddrValue(t_bbl * in, t_i386_operand * op, t_reloc * rel, t_uint32 paramno, t_uint32 stack_change)
{
  t_reg tmpreg;
  t_ins * lea, * mov;
  t_ins * push, * pseudo_call, * add;
  t_uint32 offset;
  t_bool must_translate = TRUE;

  /* determine whether address translation is necessary */
  if (I386_OP_BASE(op) == I386_REG_ESP)
    must_translate = FALSE;
  if (I386_OP_BASE(op) == I386_REG_EBP && (FUN_FLAGS(BBL_FUN(in)) & FF_HAS_FRAMEPOINTER))
    must_translate = FALSE;
  if ((I386_OP_FLAGS(op) & I386_OPFLAG_ISRELOCATED)	/* OPTIMIZATION: SEE BELOW */
      && (!PatternMatch("*P-*",rel->code) && !PatternMatch("*R-*",rel->code)))
    must_translate = FALSE;
  if (use_old_addresses)
    must_translate = FALSE;

  /* choose in which register to evaluate the address */
  if (must_translate)
    tmpreg = I386_REG_EAX;
  else
  {
    if (I386_OP_BASE(op) == I386_REG_EAX || I386_OP_INDEX(op) == I386_REG_EAX)
    {
      if (I386_OP_BASE(op) == I386_REG_EDX || I386_OP_INDEX(op) == I386_REG_EDX)
	tmpreg = I386_REG_ECX;
      else
	tmpreg = I386_REG_EDX;
    }
    else
      tmpreg = I386_REG_EAX;
  }

  /* compensate for altered stack pointer */
  offset = I386_OP_IMMEDIATE(op);
  if (I386_OP_BASE(op) == I386_REG_ESP)
    offset += stack_change;

  /* compute the effective address */
  I386MakeInsForBbl(Lea,Append,lea,in,tmpreg,offset,I386_OP_BASE(op),I386_OP_INDEX(op),I386_OP_SCALE(op));
  if (I386_OP_FLAGS(op) & I386_OPFLAG_ISRELOCATED)
  {
    if ((!PatternMatch("*P-*",rel->code) && !PatternMatch("*R-*",rel->code)) || use_old_addresses)
    {
      /* optimization: if the relocation is absolute (or we're working with old addresses),
       * the old address can be calculated right now, so we can skip address translation */
      t_address old = StackExec(rel->code,rel,NULL,FALSE,2);
      I386_OP_IMMEDIATE(I386_INS_SOURCE1(lea)) = GENERIC_EXTRACT_UINT32(old);
    }
    else
    {
      t_reloc * tmp;
      I386_OP_FLAGS(I386_INS_SOURCE1(lea)) |= I386_OPFLAG_ISRELOCATED;
      tmp = RelocTableAddRelocToRelocatable(in->cfg->sec->obj->reloc_table,rel->addend,(t_relocatable *)lea,
	  NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset,FALSE,NULL,NULL,NULL,rel->code,TRUE);
      tmp->label = StringDup(rel->label);
      if (rel->e.sec != NULL)
      {
	if (REL_E_IS_TO_RELOCATABLE(rel))
	  RelocESetToRelocatable(tmp,rel->e.sec);
	else
	  FATAL(("extra field of @R points to symbol",rel));
      }
    }
  }
  
  /* translate address */
  if (must_translate)
  {
#ifndef DIRECT_ARGS
    I386MakeInsForBbl(Push,Append,push,in,tmpreg,0);
    I386MakeInsForBbl(PseudoCall,Append,pseudo_call,in,translate);
    I386MakeInsForBbl(Arithmetic,Append,add,in,I386_ADD,I386_REG_ESP,I386_REG_NONE,4);
#else
    I386MakeInsForBbl(PseudoCall,Append,pseudo_call,in,translate);
#endif
  }

  mov = I386PutParamOnStack(in,tmpreg,0,paramno);

  return lea;
} /* }}} */

/* {{{ I386AtomAddParams */
typedef struct _t_ins_patch {
  t_ins * ins;
  t_i386_operand * operand;
  int offset_diff;
  enum {make_push,adjust_offset} action;
} t_ins_patch;

void I386AtomAddParams(t_bbl * in, t_ins * ins, t_bbl * orig, t_atom_instrument_marker * marker,t_uint32 stack_change, t_address pc)
{
  t_ins ** first_inses_for_params=NULL;
  t_ins * parstart=BBL_INS_LAST(in);
  t_atom_proto * proto=marker->proto;
  int tel;
  if (proto->nargs) first_inses_for_params=Calloc(proto->nargs+1,sizeof(t_ins *));

  /* {{{ allocate space on the stack for the function arguments */
  if (proto->nargs > 0)
  {
    t_uint32 nbytes = proto->nargs * 4;
    t_ins * stack_alloc;
    I386MakeInsForBbl(Arithmetic,Append,stack_alloc,in,I386_SUB,I386_REG_ESP,I386_REG_NONE,nbytes);
    stack_change += nbytes;
    parstart = stack_alloc;

  } /* }}} */

  /* {{{ add parameter construction instructions */
  for (tel=0; tel<proto->nargs; tel++)
  {

    switch(proto->args[tel])
    {
      case ATOM_REG_VALUE:
	if (marker->values[tel].reg==REG_PC)
	{
	  t_ins * param;
	  /* I386 implementation {{{*/
	  param = I386PutParamOnStack(in,I386_REG_NONE,G_T_UINT32(pc),tel);
	  first_inses_for_params[tel]=param;
	  /* }}} */
	}
	else if ((marker->values[tel].reg==REG_SP) || (marker->values[tel].reg == REG_I386_ESP))
	{
          t_i386_operand dummy;

	  /* I386AddCalcEffAddrValue can do this for us if we construct the
	   * right kind of operand */
	  I386_OP_TYPE(&dummy) = i386_optype_mem;
	  I386_OP_REGMODE(&dummy) = i386_regmode_full32;
	  I386_OP_MEMOPSIZE(&dummy) = 4;
	  I386_OP_BASE(&dummy) = I386_REG_ESP;
	  I386_OP_INDEX(&dummy) = I386_REG_NONE;
	  I386_OP_IMMEDIATE(&dummy) = 0;
	  I386_OP_SCALE(&dummy) = 0;
	  I386_OP_FLAGS(&dummy) = 0x0;
	  first_inses_for_params[tel] = I386AddCalcEffAddrValue(in,&dummy,NULL,tel,stack_change);
	}
	else if (marker->values[tel].reg >= REG_I386_EAX && marker->values[tel].reg <= REG_I386_EBP)
	{
	  first_inses_for_params[tel] = I386PutParamOnStack(in,marker->values[tel].reg - REG_I386_EAX + I386_REG_EAX,0,tel);
	}
	else
	{
	  FATAL(("Implement other registers (reg=%d)!",marker->values[tel].reg));
	}
	break;
      case ATOM_SPECIAL_VALUE:
	{
	  switch(marker->values[tel].query)
	  {
	    case CondValue:
	      if (!ins) FATAL(("Can only query CondValue for instructions!"));
	      /* {{{ i386 implementation */
	      if (!I386InsIsConditional(ins))
		FATAL(("CondValue is not defined for @I", ins));

	      if (I386_INS_OPCODE(ins) != I386_LOOPZ && I386_INS_OPCODE(ins) != I386_LOOPNZ)
	      {
		t_ins * setcc, * mov0;
		t_i386_condition_code cond = I386_INS_CONDITION(ins);

		mov0 = I386PutParamOnStack(in,I386_REG_NONE,0,tel);
		if (cond == I386_CONDITION_ECXZ || cond == I386_CONDITION_LOOP)
		{
		  t_ins * cmp = InsNewForBbl(in);
		  I386InstructionMakeCmp(cmp,I386_REG_ECX,I386_REG_NONE,0);
		  InsAppendToBbl(cmp,in);
		  cond = (cond == I386_CONDITION_ECXZ) ? I386_CONDITION_Z : I386_CONDITION_NZ;
		}
		setcc = InsNewForBbl(in);
		I386InstructionMakeSetcc(setcc,cond,tel*4,I386_REG_ESP,I386_REG_NONE,I386_SCALE_1,TRUE);
		InsAppendToBbl(setcc,in);

		first_inses_for_params[tel] = mov0;
	      }
	      else
	      {
		/* loopz and loopnz: these are rather difficult.
		 *   mov 0. %eax
		 *   setz/nz %al
		 *   cmp 0, %ecx
		 *   setnz %ah
		 *   and %ah,%al
		 *   xor %ah,%ah
		 *   mov %eax, paramoffset(%esp)
		 */
		t_ins * setcc1, * setcc2, * cmp, * and, * xor, * mov0, * mov1;
		t_i386_condition_code cond = I386_INS_OPCODE(ins) == I386_LOOPZ ? I386_CONDITION_Z : I386_CONDITION_NZ;

		FATAL(("this code should work but has never before been tested. just remove this line, but make sure to check the results!"));
		I386MakeInsForBbl(MovToReg,Append,mov0,in,I386_REG_EAX,I386_REG_NONE,0);

		I386MakeInsForBbl(Setcc,Append,setcc1,in,cond,0,I386_REG_EAX,I386_REG_NONE,0,FALSE);
		I386_OP_REGMODE(I386_INS_DEST(setcc1)) = i386_regmode_lo8;
		RegsetAddReg(INS_REGS_USE(setcc1),I386_REG_EAX);

		I386MakeInsForBbl(Cmp,Append,cmp,in,I386_REG_ECX,I386_REG_NONE,0);

		I386MakeInsForBbl(Setcc,Append,setcc2,in,cond,0,I386_REG_EAX,I386_REG_NONE,0,FALSE);
		I386_OP_REGMODE(I386_INS_DEST(setcc2)) = i386_regmode_hi8;
		RegsetAddReg(INS_REGS_USE(setcc2),I386_REG_EAX);
		
		I386MakeInsForBbl(Arithmetic,Append,and,in,I386_AND,I386_REG_EAX,I386_REG_EAX,0);
		I386_OP_REGMODE(I386_INS_DEST(and)) = i386_regmode_lo8;
		I386_OP_REGMODE(I386_INS_SOURCE1(and)) = i386_regmode_hi8;
		RegsetAddReg(INS_REGS_USE(and),I386_REG_EAX);

		I386MakeInsForBbl(Arithmetic,Append,xor,in,I386_XOR,I386_REG_EAX,I386_REG_EAX,0);
		I386_OP_REGMODE(I386_INS_DEST(xor)) = i386_regmode_hi8;
		I386_OP_REGMODE(I386_INS_SOURCE1(xor)) = i386_regmode_hi8;
		RegsetAddReg(INS_REGS_USE(xor),I386_REG_EAX);

		mov1 = I386PutParamOnStack(in,I386_REG_EAX,0,tel);

		first_inses_for_params[tel] = mov0;
	      }
	      /* }}} */
	      break;
	    case EffStoreAddrValue:
do_EffStoreAddrValue:
	      if (!ins) FATAL(("Can only query EffStoreAddrValue for instructions!"));

	      /* {{{ i386 implementation */
	      {
		t_i386_ins * real = T_I386_INS(ins)->ap_original ? T_I386_INS(ins)->ap_original : T_I386_INS(ins);
		t_i386_operand * op = I386InsGetMemStoreOp((t_ins *)real);
		t_i386_operand dummy;
		t_reloc * rel = NULL;

		if (!op)
		{
		  /* handle instructions in which the memory access is implicit */
		  op = &dummy;

		  I386_OP_FLAGS(op) = 0x0;
		  switch (I386_INS_OPCODE(real))
		  {
		    case I386_PUSH:
		    case I386_PUSHF:
		    case I386_CALL:
		      /* stack pointer - 4 */
		      I386OpSetMem(op,-4,I386_REG_ESP,I386_REG_NONE,I386_SCALE_1,4);
		      break;
		    case I386_PUSHA:
		      /* stack pointer - 32 */
		      I386OpSetMem(op,-32,I386_REG_ESP,I386_REG_NONE,I386_SCALE_1,4);
		      break;
		    default:
		      FATAL(("implement @I\n", ins));
		  }
		}

		if (I386_OP_FLAGS(op) & I386_OPFLAG_ISRELOCATED)
		{
		  t_reloc_ref * rr = RELOCATABLE_REFERS_TO(real);
		  /* get the relevant relocation */
		  VERIFY_FASSERT(rr,("relocations corrupt for @I",real));
		  
		  if (rr->next)
		  {
		    /* there are two relocations: which one is it? */
		    if (op == I386InsGetFirstRelocatedOp((t_ins *)real))
		      rel = rr->rel;
		    else
		      rel = rr->next->rel;
		  }
		  else
		    rel = rr->rel;
		}

		first_inses_for_params[tel] = I386AddCalcEffAddrValue(in,op,rel,tel,stack_change);
	      }
	      /* }}} */
	      break;

	    case EffLoadAddrValue:
do_EffLoadAddrValue:
	      if (!ins) FATAL(("Can only query EffLoadAddrValue for instructions!"));

	      /* {{{ i386 implementation */
	      {
		t_i386_ins * real = T_I386_INS(ins)->ap_original ? T_I386_INS(ins)->ap_original : T_I386_INS(ins);
		t_i386_operand * op = I386InsGetMemLoadOp((t_ins *)real);
		t_i386_operand dummy;
		t_reloc * rel;

		if (!op)
		{
		  /* handle instructions in which the memory access is implicit */
		  op = &dummy;

		  I386_OP_FLAGS(op) = 0x0;
		  switch (I386_INS_OPCODE(real))
		  {
		    case I386_POP:
		      if (T_INS(real) != ins)
		      {
			/* this is another ugly hack, but it is necessary to conserve correctness */
			/* if we get here, we're in the middle of a removed address producer. the previous
			 * instruction would have been a call, which would have pushed the return address on
			 * the stack. so the "virtual" stack pointer points four bytes lower than the actual stack
			 * pointer. we have to take this into account */
			I386OpSetMem(op,-4,I386_REG_ESP,I386_REG_NONE,I386_SCALE_1,4);
			break;
		      }
		      else
		      {
			/* fall through to the other pop-cases */
		      }
		    case I386_POPA:
		    case I386_POPF:
		    case I386_RET:
		      /* stack pointer */
		      I386OpSetMem(op,0,I386_REG_ESP,I386_REG_NONE,I386_SCALE_1,4);
		      break;
		    case I386_LEAVE:
		      /* ebp */
		      I386OpSetMem(op,0,I386_REG_EBP,I386_REG_NONE,I386_SCALE_1,4);
		      break;
		    default:
		      FATAL(("implement @I\n", ins));
		  }
		}

		if (I386_OP_FLAGS(op) & I386_OPFLAG_ISRELOCATED)
		{
		  t_reloc_ref * rr = RELOCATABLE_REFERS_TO(real);
		  /* get the relevant relocation */
		  VERIFY_FASSERT(rr,("relocations corrupt"));
		  
		  if (rr->next)
		  {
		    /* there are two relocations: which one is it? */
		    if (op == I386InsGetFirstRelocatedOp((t_ins *)real))
		      rel = rr->rel;
		    else
		      rel = rr->next->rel;
		  }
		  else
		    rel = rr->rel;
		}

		first_inses_for_params[tel] = I386AddCalcEffAddrValue(in,op,rel,tel,stack_change);
	      }
	      /* }}} */
	      break;

	    case EffLoadAddrValue2:
	      if (!ins) FATAL(("EffLoadAddrValue only works for instructions"));

	      if (I386_INS_OPCODE(ins) != I386_CMPSB && I386_INS_OPCODE(ins) != I386_CMPSD)
		FATAL(("don't know of any second load value for @I\n", ins));
	      first_inses_for_params[tel] = I386AddCalcEffAddrValue(in,I386_INS_SOURCE2(ins),NULL,tel,stack_change);
	      break;

	    case EffAddrValue:
	      if (!ins) FATAL(("Can only query EffAddrValue for instructions"));
	      if (IsInstType(ins, InstTypeStore) && IsInstType(ins, InstTypeLoad))
		FATAL(("@I is both store and load. Disambiguate by using EffLoadAddrValue and EffStoreAddrValue",ins));
	      if (IsInstType(ins, InstTypeStore))
		goto do_EffStoreAddrValue;
	      if (IsInstType(ins, InstTypeLoad))
		goto do_EffLoadAddrValue;
	      FATAL(("@I is neither load nor store. There is no effective address value to be calculated", ins));
	      break;

	    case FunctionName:
	      /* FunctionName {{{ */
	      {
		t_section * added_section;
		t_object * obj;
		t_section * sec;
		t_string secname;
		t_ins * param;
		t_address null;
		t_reloc * rel;
		if (!(BBL_FUN(orig))) FATAL(("Oeps!"));
		obj=SECTION_OBJECT(CFG_SECTION(BBL_CFG(orig)));
		sec=SectionGetFromObjectByName(obj,".rodata");
		if (!sec) FATAL(("Implement no rodata!"));

		if (BBL_FUN(orig)->name)
		{
		  secname=Collate("funname_",BBL_FUN(orig)->name);
		  added_section=InstrumentAddExtraSection(obj,sec,RO_DATA_SECTION,secname,sec->name,strlen(BBL_FUN(orig)->name)+1,0x0);
		  strcpy(added_section->data,BBL_FUN(orig)->name);
		  Free(secname);
		}
		else 
		{
		  secname=Collate("funname_","noname");
		  added_section=InstrumentAddExtraSection(obj,sec,RO_DATA_SECTION,secname,sec->name,strlen("noname"),0x0);
		  strcpy(added_section->data,"noname");
		  Free(secname);
		}

		GenericSet_uint32(null,0);

		param = I386PutParamOnStack(in,I386_REG_NONE,0,tel);
		I386_OP_FLAGS(I386_INS_SOURCE1(param)) |= I386_OPFLAG_ISRELOCATED;
		rel=RelocTableAddRelocToRelocatable(obj->reloc_table,null,T_RELOCATABLE(param),null,T_RELOCATABLE(added_section),null,FALSE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);
		first_inses_for_params[tel] = param;

	      }
	      /* }}} */
	      break;
	    default:
	      FATAL(("Implement special values!"));
	  }
	  break;
	}

      case ATOM_INT_VALUE:
	{
	  t_ins * param = I386PutParamOnStack(in,I386_REG_NONE,marker->values[tel].int32,tel);
	  first_inses_for_params[tel] = param;
	}
	break;

      default:
	FATAL(("Implement argument type %d\n",proto->args[tel]));
	/* TODO: Add the right parameter */
    }
  } /* }}} */

  AtomReorderParameters(proto,first_inses_for_params,parstart,in);

  /* {{{ add some fixups to keep the code correct / optimize the parameter construction */
  if (proto->nargs > 0)
  {
    t_ins * ins;
    int bytes_alloced, below_alloced, current_in_prealloc;
    t_bool must_prealloc = FALSE;
    t_ins_patch patches[BBL_NINS(INS_BBL(parstart))];
    int npatches = 0;
    t_bool call_to_translate = FALSE;

    /* {{{ is it really necessary to preallocate stack space for the function arguments? */
    if (I386_INS_OPCODE(parstart) != I386_SUB)
      FATAL(("@I is wrong parstart",parstart));
    
    bytes_alloced = I386_OP_IMMEDIATE(I386_INS_SOURCE1(parstart));
    current_in_prealloc = bytes_alloced;
    below_alloced = 0;

    for (ins = INS_NEXT(parstart); ins; ins = INS_NEXT(ins))
    {
      t_i386_operand * store_op, * load_op;

      if (I386_INS_OPCODE(ins) == I386_PSEUDOCALL && (((t_function *) I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins))) == translate))
	call_to_translate = TRUE;

      if (!RegsetIn(INS_REGS_USE(ins),I386_REG_ESP) && !RegsetIn(INS_REGS_DEF(ins),I386_REG_ESP))
	continue;
      
      store_op = I386InsGetMemStoreOp(ins);
      load_op = I386InsGetMemLoadOp(ins);

      if (store_op)
      {
	if (I386_OP_BASE(store_op) == I386_REG_ESP)
	{
	  int offset = I386_OP_IMMEDIATE(store_op);
	  if (I386_OP_INDEX(store_op) != I386_REG_NONE)
	  {
	    must_prealloc = TRUE;
	    break;
	  }
	  
	  if (offset < below_alloced) /* OK */
	    continue;
	  else if (offset >= below_alloced + current_in_prealloc) /* OK */
	  {
	    patches[npatches].operand = store_op;
	    patches[npatches].offset_diff = current_in_prealloc;
	    patches[npatches].action = adjust_offset;
	    npatches++;
	    continue;
	  }
	  else if ((offset == current_in_prealloc - 4) && (below_alloced == 0))
	  {
	    if (I386_INS_OPCODE(ins) == I386_MOV)
	    {
	      /* move current_in_prealloc one slot down */
	      current_in_prealloc -= 4;
	      patches[npatches].ins = ins;
	      patches[npatches].action = make_push;
	      npatches++;
	      continue;
	    }
	    else
	    {
	      must_prealloc = TRUE;
	      continue;
	    }
	  }
	  else
	  {
	    /* either the allocated slots are not filled in order or there are still 
	     * some bytes alloced below the prealloced space when the next slot is being filled */
	    must_prealloc = TRUE;
	    break;
	  }
	}
      }
      if (load_op)
      {
	if (I386_OP_BASE(load_op) == I386_REG_ESP)
	{
	  int offset = I386_OP_IMMEDIATE(load_op);
	  if (I386_OP_INDEX(load_op) != I386_REG_NONE)
	  {
	    must_prealloc = TRUE;
	    break;
	  }

	  if (offset < below_alloced) /* OK */
	    continue;
	  else if (offset >= below_alloced + current_in_prealloc) /* OK */
	  {
	    patches[npatches].operand = load_op;
	    patches[npatches].offset_diff = current_in_prealloc;
	    patches[npatches].action = adjust_offset;
	    npatches++;
	    continue;
	  }
	  else
	  {
	    /* memory access to the area of preallocated space that
	     * has not yet been initialised. as this is something really
	     * weird, just stay away from it */
	    must_prealloc = TRUE;
	    break;
	  }
	}
      }

      /* if we reach this point, the instruction is either push, pop,
       * add $...,%esp or sub $...,%esp or lea ...,...(%esp) or something that is too 
       * complicated to understand, in which case we just keep our hands 
       * off of it */
      switch (I386_INS_OPCODE(ins))
      {
	case I386_PUSH:
	  below_alloced += 4;
	  continue;
	case I386_POP:
	  below_alloced -= 4;
	  if (below_alloced <= 0)
	    must_prealloc = TRUE;	/* this is something funny, just be conservative about it */
	  break;
	case I386_SUB:
	  if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg &&
	      I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_ESP &&
	      I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	    below_alloced += I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins));
	  else
	    must_prealloc = TRUE;
	  break;
	case I386_ADD:
	  if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg &&
	      I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_ESP &&
	      I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	    below_alloced -= I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins));
	  else
	    must_prealloc = TRUE;
	  if (below_alloced < 0) must_prealloc = TRUE;
	  break;
	case I386_LEA:
	  if (I386_OP_INDEX(I386_INS_SOURCE1(ins)) != I386_REG_NONE)
	    must_prealloc = TRUE;
	  else
	  {
	    int offset = I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins));
	    if (offset < below_alloced)
	    { /* OK */ }
	    else if (offset >= below_alloced + current_in_prealloc)
	    { /* OK */ 
	      patches[npatches].operand = I386_INS_SOURCE1(ins);
	      patches[npatches].offset_diff = current_in_prealloc;
	      patches[npatches].action = adjust_offset;
	      npatches++;
	    }
	    else
	    {
	      /* not OK */
	      must_prealloc = TRUE;
	    }
	  }
	  break;
	default:
	  /* something we don't understand: be conservative about it */
	  DiabloPrint(stdout,"conservative because of @I\n",ins);
	  must_prealloc = TRUE;
      }
      if (must_prealloc)
	break;
    } /* }}} */

    if (!must_prealloc)
    {
      /* {{{ patch instructions and kill stack preallocation */
      int i;
      for (i = 0; i < npatches; i++)
      {
	if (patches[i].action == make_push)
	{
	  t_reg reg = I386_REG_NONE;
	  t_uint32 imm = 0;
	  if (I386_OP_TYPE(I386_INS_SOURCE1(patches[i].ins)) == i386_optype_reg)
	    reg = I386_OP_BASE(I386_INS_SOURCE1(patches[i].ins));
	  else if (I386_OP_TYPE(I386_INS_SOURCE1(patches[i].ins)) == i386_optype_imm)
	    imm = I386_OP_IMMEDIATE(I386_INS_SOURCE1(patches[i].ins));
	  else
	    FATAL(("unexpected operand type"));

	  I386InstructionMakePush(patches[i].ins,reg,imm);
	}
	else
	  I386_OP_IMMEDIATE(patches[i].operand) -= patches[i].offset_diff;
      }
      InsKill(parstart);
      /* small optimization:
       * if translate() is called during argument construction, the following is generated:
       *   push <address>
       *   pseudocall translate
       *   add $4, %esp
       *   push %eax
       * we can collapse the last two instructions into mov %eax, (%esp)
       */
      if (call_to_translate)
      {
	BBL_FOREACH_INS(in,ins)
	{
	  if (I386_INS_OPCODE(ins) == I386_ADD 
	      && I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg && I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_ESP 
	      && I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm && I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) == 4)
	  {
	    t_ins * next = INS_NEXT(ins);
	    if (next && I386_INS_OPCODE(next) == I386_PUSH
		&& I386_OP_TYPE(I386_INS_SOURCE1(next)) == i386_optype_reg && I386_OP_BASE(I386_INS_SOURCE1(next)) == I386_REG_EAX)
	    {
	      InsKill(next);
	      I386InstructionMakeMovToMem(ins,0,I386_REG_ESP,I386_REG_NONE,0,I386_REG_EAX,0);
	    }
	  }
	}
      }
      /* }}} */
    }
    else
    {
      /* {{{ add instructions to preserve the condition flags if necessary */
      t_ins * ins;
      t_regset changed, used, tmp;
      t_regset condflags = NullRegs;
      t_bool preserve_flags, preserve_eax;

      //if (RegsetIsEmpty(condflags)) 
      {
#define addf(flag)  RegsetAddReg(condflags,I386_CONDREG_##flag)
	addf(AF); addf(CF); addf(OF); addf(PF); addf(SF); addf(ZF);
#undef addf
      }

      /* check if we need to preserve the condition flags after stack space allocation
       * (the sub <space>,%esp instruction changes the condition flags). also check if 
       * we need to preserve %eax (if not, we can use it for saving/restoring flags. the
       * alternative is doing it over the stack, which is pretty slow) */
      changed = NullRegs;
      used = NullRegs;
      for (ins = INS_NEXT(parstart); ins; ins = INS_NEXT(ins))
      {
	/* used = used U (Use(ins) \ changed); changed = changed U Def(ins) */
	RegsetDup(tmp,INS_REGS_USE(ins));
	RegsetSetDiff(tmp,changed);
	RegsetSetUnion(used,tmp);
	RegsetSetUnion(changed,INS_REGS_DEF(ins));
      }
      preserve_eax = RegsetIn(used, I386_REG_EAX);
      RegsetDup(tmp,used);
      RegsetSetIntersect(tmp,condflags);
      preserve_flags = !RegsetIsEmpty(tmp);

      if (preserve_flags)
      {
	if (preserve_eax || RegsetIn(used,I386_CONDREG_OF))
	{
	  /* save over stack */
	  t_ins *push, *popf;
	  int flags_offset;

	  /* we assume the condition flags will only be relevant in the parameter passing code
	   * if they are also relevant in the original program. therefore, we assume they 
	   * will already be saved by AtomAddSave() (directly above the stack space we've allocated
	   * for the arguments). we will restore them from there. */
	  VERIFY_FASSERT(I386_INS_OPCODE(parstart) == I386_SUB,("parstart == @I, expected sub instruction!",parstart));
	  VERIFY_FASSERT(I386_OP_TYPE(I386_INS_SOURCE1(parstart)) == i386_optype_imm,("expected immediate argument to @I",parstart));
	  flags_offset = I386_OP_IMMEDIATE(I386_INS_SOURCE1(parstart));
	  push = InsNewForBbl(in);
	  I386InstructionMakePushMem(push,flags_offset,I386_REG_ESP,I386_REG_NONE,I386_SCALE_1);
	  InsInsertAfter(push,parstart);
	  popf = InsNewForBbl(in);
	  I386InstructionMakeSimple(popf,I386_POPF);
	  InsInsertAfter(popf,push);
	}
	else
	{
	  /* save over %ah */
	  t_ins *lahf, *sahf;
	  lahf = InsNewForBbl(in);
	  I386InstructionMakeLSahf(lahf,I386_LAHF);
	  InsInsertBefore(lahf,parstart);
	  sahf = InsNewForBbl(in);
	  I386InstructionMakeLSahf(sahf,I386_SAHF);
	  InsInsertAfter(sahf,parstart);
	}
      }
      /* }}} */
    }
  }
  /* }}} */

} /* }}} */

/* {{{ add instructions that remove parameters from stack */
void I386AtomAddKillParams(t_bbl * out, t_atom_instrument_marker * marker)
{
  t_atom_proto * proto = marker->proto;
  t_uint32 tel;
  t_uint32 paramsize = 0;
  for (tel = 0; tel < proto->nargs; tel++)
  {
    switch (proto->args[tel])
    {
      case ATOM_INT_VALUE:
      case ATOM_REG_VALUE:
	paramsize += 4;
	break;
      case ATOM_SPECIAL_VALUE:
	switch (marker->values[tel].query)
	{
	  case BrCondValue:
	  case EffAddrValue:
	  case EffLoadAddrValue:
	  case EffLoadAddrValue2:
	  case EffStoreAddrValue:
	  case FunctionName:
	    paramsize += 4;
	    break;
	  default:
	    FATAL(("implement special value"));
	}
	break;
      default:
	FATAL(("implement argument type %d\n",proto->args[tel]));
    }
  }

  if (paramsize > 0)
  {
    t_ins * cleanup = InsNewForBbl(out);
    InsPrependToBbl(cleanup,out);
    I386InstructionMakeArithmetic(cleanup, I386_ADD, I386_REG_ESP, I386_REG_NONE, paramsize);
  }
} /* }}} */

/* {{{ I386AtomAddCheckForExitCall */
void I386AtomAddCheckForExitCall(t_bbl * check, t_bbl * call, t_bbl * restore)
{
  t_ins * cmp, * jne;
  cmp = InsNewForBbl(check);
  I386InstructionMakeCmp(cmp,I386_REG_EAX,I386_REG_NONE,1);
  InsAppendToBbl(cmp,check);
  jne = InsNewForBbl(check);
  I386InstructionMakeCondJump(jne,I386_CONDITION_NZ);
  InsAppendToBbl(jne,check);

  CfgEdgeNew(BBL_CFG(check),check,call,ET_FALLTHROUGH);
  CfgEdgeNew(BBL_CFG(check),check,restore,ET_JUMP);
} /* }}} */

/* {{{ I386AtomAddCall */
void I386AtomAddCall(t_bbl * in, t_bbl * out, t_atom_proto * proto)
{
  t_cfg * cfg = FUN_CFG(BBL_FUN(in));
  t_bbl * instr_entry = proto->bbl;
  t_ins * call;

  I386MakeInsForBbl(Call,Append,call,in);
  CfgEdgeNewCall(cfg,in,instr_entry,out,FUN_BBL_LAST(BBL_FUN(instr_entry)));
  CfgEdgeNew(cfg,in,out,ET_FUNLINK);
} /* }}} */

/* {{{ pre-instrumentation initialization */
void I386AtomInit(t_object * obj)
{
  int i;
  t_section * parsec = NULL;
  t_object * linker_subobject;
  t_address size;

  /* rewrite some complicated program structures to ensure correctness of the
   * instrumentation results */
  I386PatchRepPrefixes(obj); 
  I386PatchCmps(obj);
  /* frame pointer analysis tries to determine whether a function sets the frame pointer at entry.
   * this knowledge allows us to add less address translation code to the instrumented program */
  I386FramePointerAnalysis(T_CFG(obj->code[0]->data));

  linker_subobject = ObjectGetFromCache("Linker",obj);
  for (i=0; i<obj->ndatas; i++)
    if (!strncmp(obj->data[i]->name,INSTRUMENT_PREFIX,strlen(INSTRUMENT_PREFIX)))
      parsec = obj->data[i];
  if (!parsec)
    FATAL(("Implement no instrumentation data sections"));

  /* fpu state save area */
  GenericSet_uint32(size,FPU_SAVE_AREA_SIZE);
  fpu_save_area = SectionCreateForObject(linker_subobject,DATA_SECTION,parsec,size,"FPUSaveArea");
  GenericSet_uint32(SECTION_ALIGN(fpu_save_area),32);
  SECTION_FLAGS(fpu_save_area) |= SECTION_FLAG_KEEP;
  /* stack save area, for some special cases of instrumentation */
  GenericSet_uint32(size,STACK_SAVE_AREA_SIZE);
  stack_save_area = SectionCreateForObject(linker_subobject,DATA_SECTION,parsec,size,"StackSaveArea");
  GenericSet_uint32(SECTION_ALIGN(stack_save_area),4);
  SECTION_FLAGS(stack_save_area) |= SECTION_FLAG_KEEP;
  /* general purpose register save area */
  GenericSet_uint32(size,INTREG_SAVE_AREA_SIZE);
  int_save_area = SectionCreateForObject(linker_subobject,DATA_SECTION,parsec,size,"IntRegSaveArea");
  GenericSet_uint32(SECTION_ALIGN(int_save_area),32);
  SECTION_FLAGS(int_save_area) |= SECTION_FLAG_KEEP;

#ifdef USE_PSEUDO
  /* always save %eax and %edx to the stack for the best results.
   * this results in somewhat more pushes and pops, but reduces code size,
   * which is beneficial for the I-cache behaviour. */
  always_stack = NullRegs;
  RegsetAddReg(always_stack,I386_REG_EAX);
  RegsetAddReg(always_stack,I386_REG_EDX);
  RegsetAddReg(always_stack,I386_REG_ECX);
#endif
} /* }}} */

/* {{{ I386IsInstType */
t_bool I386IsInstType(Inst * ins, t_atom_inst_type type)
{
  switch(type)
  {
    case InstTypeLoad:
      return I386InsIsLoad((t_ins *)ins) || (T_I386_INS(ins)->ap_original && I386InsIsLoad((t_ins *)T_I386_INS(ins)->ap_original));
    case InstTypeStore:
      return I386InsIsStore(ins) || ((t_ins *)T_I386_INS(ins)->ap_original && I386InsIsStore((t_ins *)T_I386_INS(ins)->ap_original));
    case InstTypeCondBr:
      return (I386_INS_OPCODE(ins) == I386_Jcc   || I386_INS_OPCODE(ins) == I386_JECXZ ||
              I386_INS_OPCODE(ins) == I386_LOOPZ || I386_INS_OPCODE(ins) == I386_LOOPNZ);
    case InstTypeConditional:
      return I386InsIsConditional((t_ins *)ins);
    default:
      FATAL(("Implement IsInstType"));
  }
  return FALSE;
} /* }}} */

/* {{{ turn pseudo saves and loads into real instructions */
void I386AtomExpandPseudoSavesAndLoads(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  t_reg reg;
  t_address null, two, offset;

  GenericSet_uint32(two,2);
  null = GENERIC_SUB(two,two);
  
  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS(bbl,ins)
    {
      if (INS_TYPE(ins) == IT_PSEUDO_LOAD)
      {
	reg = I386_OP_BASE(I386_INS_DEST(ins));
	GenericSet_uint32(offset,reg*4);

	I386InstructionMakeMovFromMem(ins,reg,0,I386_REG_NONE,I386_REG_NONE,0);
	I386_OP_FLAGS(I386_INS_SOURCE1(ins)) |= I386_OPFLAG_ISRELOCATED;
	RelocTableAddRelocToRelocatable(cfg->sec->obj->reloc_table,
	    null, (t_relocatable *)ins, two,
	    (t_relocatable *)int_save_area, offset,
	    FALSE, NULL, NULL, NULL, CALC_ABS "\\" WRITE_32, FALSE );
      }
      else if (INS_TYPE(ins) == IT_PSEUDO_SAVE)
      {
	reg = I386_OP_BASE(I386_INS_SOURCE1(ins));
	GenericSet_uint32(offset,reg*4);

	I386InstructionMakeMovToMem(ins,0,I386_REG_NONE,I386_REG_NONE,0,reg,0);
	I386_OP_FLAGS(I386_INS_DEST(ins)) |= I386_OPFLAG_ISRELOCATED;
	RelocTableAddRelocToRelocatable(cfg->sec->obj->reloc_table,
	    null, (t_relocatable *)ins, two,
	    (t_relocatable *)int_save_area, offset,
	    FALSE, NULL, NULL, NULL, CALC_ABS "\\" WRITE_32, FALSE );
      }
    }
}
/* }}} */

/** Some peephole optimizations that limit the instrumentation overhead **/
/* {{{ redundant fsave removal (turned off for now) */
void I386PeepHoleRedundantFSaves(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  t_regset fpuregs = NullRegs;
  static int counter = 0;

  /* TODO: this function does nothing as long as we fpu state saving/restoring is disabled */
  return;

#define addf(flag)  RegsetAddReg(fpuregs,I386_CONDREG_##flag)
#define addr(reg)   RegsetAddReg(fpuregs,I386_REG_##reg)
  addf(C0); addf(C1); addf(C2); addf(C3);
  addr(ST0); addr(ST1); addr(ST2); addr(ST3); addr(ST4); addr(ST5); addr(ST6); addr(ST7);
#undef addf
#undef addr
  
  CFG_FOREACH_BBL(cfg,bbl)
  {
    BBL_FOREACH_INS(bbl,ins)
    {
      if (I386_INS_OPCODE(ins) == I386_FRSTOR)
	if ( 1 &&
	  RELOCATABLE_REFERS_TO(ins) &&
	  REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ins)->rel) == (t_relocatable *)fpu_save_area)
      {
	t_ins * save;
	t_regset changed = NullRegs;

	/* find next fsave */
	for (save = INS_NEXT(ins); save; save = INS_NEXT(save))
	{
	  if (I386_INS_OPCODE(save) == I386_FSAVE &&
	      RELOCATABLE_REFERS_TO(save) &&
	      REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(save)->rel) == (t_relocatable *)fpu_save_area)
	    break;
	  RegsetSetUnion(changed,INS_REGS_DEF(save));
	}
	if (!save) continue;
	if (!RegsetIsMutualExclusive(changed,fpuregs)) continue;

	/* fsave is redundant. remove it. */
	InsKill(save);
	++counter;
      }
    }
  }
  VERBOSE(0,("[redundant fsave removal] %d times",counter));
}
/* }}} */

/* {{{ peephole specifically for myprof, does some specialized constant propagation
 * so we don't need the complete constprop */
void I386PeepHoleMyprof(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS(bbl,ins)
    {
      if (I386_INS_OPCODE(ins) == I386_INC
	  && I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_mem
	  && I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_NONE
	  && I386_OP_INDEX(I386_INS_DEST(ins)) != I386_REG_NONE
	  && RELOCATABLE_REFERS_TO(ins))
      {
	t_reg index = I386_OP_INDEX(I386_INS_DEST(ins));
	t_ins * iter;
	for (iter = INS_PREV(ins); iter; iter = INS_PREV(iter))
	  if (RegsetIn(INS_REGS_DEF(iter),index))
	    break;
	if (!iter) continue;
	if (I386_INS_OPCODE(iter) == I386_MOV)
	{
	  t_i386_operand * dest = I386_INS_DEST(iter);
	  t_i386_operand * src = I386_INS_SOURCE1(iter);
	  if (I386_OP_TYPE(src) == i386_optype_imm
	      && I386_OP_TYPE(dest) == i386_optype_reg
	      && I386_OP_REGMODE(dest) == i386_regmode_full32
	      && I386_OP_BASE(dest) == index
	      && !RELOCATABLE_REFERS_TO(iter))
	  {
	    /* encode this immediate value directly in the relocation of ins */
	    t_reloc * rel = RELOCATABLE_REFERS_TO(ins)->rel;
	    rel->to_offset = GENERIC_ADD_UINT32(rel->to_offset,I386_OP_IMMEDIATE(src) * (1 << I386_OP_SCALE(I386_INS_DEST(ins))));
	    I386_OP_INDEX(I386_INS_DEST(ins)) = I386_REG_NONE;
	    I386_OP_SCALE(I386_INS_DEST(ins)) = 0;
	    INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
	    INS_REGS_DEF(ins) = I386InsDefinedRegisters(ins);
	    /*DiabloPrint(stdout,"@iB\n",bbl);*/
	  }
	}
      }
    }
} /* }}} */

void I386PeepHoleSimpleCopyPropagation(t_cfg * cfg)
{
  /* simple, stupid copy propagation */

  t_bbl * bbl;
  t_ins * ins, * tmp;

  static int kills = 0;
  static int substs = 0;

  int fixedop;

  CFG_FOREACH_BBL(cfg,bbl)
  {
    int copies[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
    BBL_FOREACH_INS_SAFE(bbl,ins,tmp)
    {
      t_regset subst;
      t_reg r;
      t_i386_operand * op;

      if (I386_INS_OPCODE(ins) == I386_PSEUDOSAVE)
	continue;	/* don't perform substitutions here */

      RegsetDup(subst,INS_REGS_USE(ins));
      RegsetSetDiff(subst,INS_REGS_DEF(ins));
      RegsetSetIntersect(subst,cfg->sec->obj->description->int_registers);

      for (fixedop = 2, op = I386_INS_SOURCE1(ins); op; 
	  op = (op == I386_INS_SOURCE1(ins) ? I386_INS_SOURCE2(ins) : NULL), fixedop <<= 1)
      {
	if (I386_OP_TYPE(op) == i386_optype_reg && !(i386_opcode_table[I386_INS_OPCODE(ins)].fixedops & fixedop))
	{
	  r = I386_OP_BASE(op);
	  if (RegsetIn(subst,r) && copies[r] != -1 && I386_OP_REGMODE(op) == i386_regmode_full32)
	  {
	    I386_OP_BASE(op) = copies[r];
	    INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
	    ++substs;
	  }
	}
	else if (I386_OP_TYPE(op) == i386_optype_mem)
	{
	  r = I386_OP_BASE(op);
	  if (r != I386_REG_NONE && RegsetIn(subst,r) && copies[r] != -1)
	  {
	    I386_OP_BASE(op) = copies[r];
	    INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
	    ++substs;
	  }
	  r = I386_OP_INDEX(op);
	  if (r != I386_REG_NONE && RegsetIn(subst,r) && copies[r] != -1)
	  {
	    I386_OP_INDEX(op) = copies[r];
	    INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
	    ++substs;
	  }
	}
      }

      if (I386_INS_OPCODE(ins) == I386_MOV
	  && I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg
	  && I386_OP_REGMODE(I386_INS_SOURCE1(ins)) == i386_regmode_full32
	  && I386IsGeneralPurposeReg(I386_OP_BASE(I386_INS_SOURCE1(ins)))
	  && I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg
	  && I386_OP_REGMODE(I386_INS_DEST(ins)) == i386_regmode_full32
	  && I386IsGeneralPurposeReg(I386_OP_BASE(I386_INS_DEST(ins))))
      {
	t_reg from = I386_OP_BASE(I386_INS_SOURCE1(ins));
	t_reg to = I386_OP_BASE(I386_INS_DEST(ins));
	if (to == from)
	{
	  InsKill(ins);
	  ++kills;
	}
	else
	{
	  copies[to] = from;
	  int i;
	  for (i=0; i<8; i++)
	    if (copies[i] == to)
	      copies[i] = -1;
	}
      }
      else
      {
	REGSET_FOREACH_REG(INS_REGS_DEF(ins),r)
	{
	  if (I386IsGeneralPurposeReg(r))
	  {
	    int i;
	    copies[r] = -1;
	    for (i=0; i<8; i++)
	      if (copies[i] == r)
		copies[i] = -1;
	  }
	}
      }
    }
  }
  DiabloPrint(stdout,"[SimpleCopy] %d substitutions, %d kills\n",substs,kills);
}

void I386PeepHolePushESP(t_cfg * cfg)
{
  /* mov %esp, %reg
   * push %reg  |  mov %reg, (%esp) 
   *   =>
   * push %esp  |  mov %esp, (%esp)
   */
  t_bbl * bbl;
  t_ins * ins;

  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS(bbl,ins)
    {
      t_reg reg;
      t_ins *next = INS_NEXT(ins);

      if (!next) continue;
      if (I386_INS_OPCODE(ins) != I386_MOV) continue;
      if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) != i386_optype_reg) continue;
      if (I386_OP_BASE(I386_INS_SOURCE1(ins)) != I386_REG_ESP) continue;
      if (I386_OP_TYPE(I386_INS_DEST(ins)) != i386_optype_reg) continue;
      reg = I386_OP_BASE(I386_INS_DEST(ins));
      
      if (!(I386_INS_OPCODE(next) == I386_PUSH
	    && I386_OP_TYPE(I386_INS_SOURCE1(next)) == i386_optype_reg
	    && I386_OP_BASE(I386_INS_SOURCE1(next)) == reg
	   )
	  &&
	  !(I386_INS_OPCODE(next) == I386_MOV
	    && I386_OP_TYPE(I386_INS_SOURCE1(next)) == i386_optype_reg
	    && I386_OP_BASE(I386_INS_SOURCE1(next)) == reg
	    && I386_OP_TYPE(I386_INS_DEST(next)) == i386_optype_mem
	    && I386_OP_BASE(I386_INS_DEST(next)) == I386_REG_ESP
	    && I386_OP_INDEX(I386_INS_DEST(next)) == I386_REG_NONE
	    && I386_OP_IMMEDIATE(I386_INS_DEST(next)) == 0
	   )
	 ) continue;

      if (I386_INS_OPCODE(next) == I386_PUSH)
	I386InstructionMakePush(ins,I386_REG_ESP,0);
      else
	I386InstructionMakeMovToMem(ins,0,I386_REG_ESP,I386_REG_NONE,0,I386_REG_ESP,0);
      InsKill(next);
    }
}

/**************************************************************************/
/**                       PRECISE INSTRUMENTATION                        **/
/**************************************************************************/

/* {{{ determine which relocs should be calculated with old addresses */
void I386RelocWithOldAddresses(t_cfg * cfg)
{
  t_object * obj = cfg->sec->obj;
  t_ins * ins;
  t_reloc * rel, * safe;
  t_uint32 address;

  OBJECT_FOREACH_RELOC_SAFE(obj,rel,safe)
  {
    switch (RELOCATABLE_TYPE(REL_FROM(rel)))
    {
      case RT_INS:
	ins = (t_ins *)REL_FROM(rel);
	address = G_T_UINT32(BBL_OLD_ADDRESS(INS_BBL(ins)));
	if (address == 0 || address == 0xffffffff) 
	  continue;

	if (INS_TYPE(ins) == IT_DATA)
	  FATAL(("implement"));
	else
	{
	  /* skip relocs from jump and call instructions (these are handled
	   * in AdjustJumpOffsets) */
	  if (I386_INS_OPCODE(ins) == I386_JMP ||
	      I386_INS_OPCODE(ins) == I386_Jcc ||
	      I386_INS_OPCODE(ins) == I386_JECXZ ||
	      I386_INS_OPCODE(ins) == I386_CALL ||
	      I386_INS_OPCODE(ins) == I386_LOOP ||
	      I386_INS_OPCODE(ins) == I386_LOOPZ ||
	      I386_INS_OPCODE(ins) == I386_LOOP)
	  {
	    if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	      continue;
	  }

	  /* find out which operand has been relocated */
	  t_i386_operand * op;
	  t_reloc_ref * rr = RELOCATABLE_REFERS_TO(ins);
	  t_reloc_ref * rr2;
	  VERIFY_FASSERT(rr,("relocations corrupt"));
	  VERIFY_FASSERT(rr->rel == rel || (rr->next && rr->next->rel == rel),("relocations corrupt")); 

	  if (rr->rel == rel) 
	    rr2 = NULL;
	  else
	  {
	    rr2 = rr;
	    rr = rr->next;
	  }

	  if (rr2 && GENERIC_GT(rel->from_offset,rr2->rel->from_offset))
	    op = I386InsGetSecondRelocatedOp((t_ins *)ins);
	  else
	    op = I386InsGetFirstRelocatedOp((t_ins *)ins);

	  if (I386_OP_TYPE(op) == i386_optype_imm)
	  {
	    /* OK, use old address */
	  }
	  else if (I386_OP_TYPE(op) == i386_optype_mem)
	  {
	    /* for lea instructions, use old addresses
	     * for other instructions, use old addresses only if the
	     * relocation appears to be relative (i.e. either "R-" or
	     * "P-" appears in the stack machine code). */
	    if (I386_INS_OPCODE(ins) != I386_LEA && 
		!(PatternMatch("*P-*",rel->code) || PatternMatch("*R-*",rel->code)))
	    {
	      /* skip, can be relocated with new address immediately */
	      continue;
	    }
	  }
	  else
	  {
            FATAL(("funny op type with relocation"));
	  }

	  REL_FLAGS(rel) |= REL_FLAG_USE_OLD_ADDRESSES;
	}
	break;
      case RT_SECTION:
      case RT_SUBSECTION:
	/* perform all relocations in _original_ code sections with old addresses */
	address = G_T_UINT32(SECTION_GET_OLD_ADDRESS(T_SECTION(REL_FROM(rel))));
	if (address == 0 || address == 0xffffffff) 
	  continue;
	REL_FLAGS(rel) |= REL_FLAG_USE_OLD_ADDRESSES;
	/*StackExec(rel->code,rel,T_SECTION(REL_FROM(rel))->data,TRUE,2);*/
	/*RelocTableRemoveReloc(obj->reloc_table,rel);*/
	break;
      default:
	FATAL(("Relocations corrupt: @R"));
    }
  }
} /* }}} */

/* {{{ is the operand a simple address expression?
 * returns true if the operand is a memory operand like:
 * (%reg) or offset or (,%reg,1) */
static t_bool IsSimpleAddressExpression(t_i386_operand * op)
{
  if (I386_OP_TYPE(op) != i386_optype_mem) return FALSE;
  if (I386_OP_BASE(op) == I386_REG_NONE && I386_OP_INDEX(op) == I386_REG_NONE) return TRUE; /* offset only */
  if (I386_OP_IMMEDIATE(op) == 0 && I386_OP_INDEX(op) == I386_REG_NONE) return TRUE;
  if (I386_OP_IMMEDIATE(op) == 0 && I386_OP_BASE(op) == I386_REG_NONE && I386_OP_SCALE(op) == I386_SCALE_1) return TRUE;
  return FALSE;
} /* }}} */

/* {{{ patch string instructions that use only %esi or %edi, not both 
 * in order to preserve the condition flags, this results in rather 
 * contrived code:
 *
 * INS                       | SCAS
 * ---                       | ----
 *                           | 
 * pushf                     | push %edi
 * push %esi                 | push <...>
 * push <...>                | push %edi
 * push %esi                 | call ReverseTranslate
 * call ReverseTranslate     | add 0x4,%esp
 * add 0x4,%esp              | mov %eax,%edi
 * mov %eax,%esi             | pop <...>
 * pop <...>                 | sub %edi,(%esp)
 * sub %esi,(%esp)           | scas
 * ins                       | push %eax
 * add (%esp),%esi           | pushf
 * add 0x4,%esp              | pop %eax
 * popf                      | add 0x4(%esp),%edi
 *                           | mov %eax,0x4(%esp)
 *                           | pop %eax
 *                           | popf
 *
 * The difference between scas and the other instructions is caused by
 * the fact that scas updates the eflags register and the others don't
 */
static void DoPatchForSingleStringIns(t_ins * ins)
{
  t_reg adreg,r;
  t_regset save,change;
  t_ins * added;
  t_bool ins_sets_flags = (I386_INS_OPCODE(ins) == I386_SCASB || I386_INS_OPCODE(ins) == I386_SCASD);

  /*DiabloPrint(stdout,"patching @I:\n",ins);*/

  /* which register do we have to translate */
  switch (I386_INS_OPCODE(ins))
  {
    case I386_INSB:  case I386_INSD:
    case I386_STOSB: case I386_STOSD:
    case I386_SCASB: case I386_SCASD:
      adreg = I386_REG_EDI;
      break;
    case I386_OUTSB: case I386_OUTSD:
    case I386_LODSB: case I386_LODSD:
      adreg = I386_REG_ESI;
      break;
    default:
      FATAL(("not the right kind of instruction"));
  }

  /* which registers need to be saved? */
  RegsetDup(change,FUN_REGS_CHANGED(reversetranslate));
  RegsetSetDiff(change,FUN_REGS_SAVED(reversetranslate));
  RegsetAddReg(change,I386_REG_EAX);
  RegsetDup(save,InsRegsLiveBefore(ins));
  RegsetSetIntersect(save,BBL_CFG(INS_BBL(ins))->sec->obj->description->int_registers);
  RegsetSetIntersect(save,change);
  RegsetSubReg(save,I386_REG_ESP);
  RegsetSubReg(save,adreg);

  /* the real patching */

  if (!ins_sets_flags)
    I386MakeInsForIns(Simple,Before,added,ins,I386_PUSHF);

  I386MakeInsForIns(Push,Before,added,ins,adreg,0);
  REGSET_FOREACH_REG(save,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForIns(Push,Before,added,ins,r,0);
#else
    if (RegsetIn(always_stack,r))
      I386MakeInsForIns(Push,Before,added,ins,r,0);
    else
      I386MakeInsForIns(PseudoSave,Before,added,ins,r);
#endif
  }

#ifndef DIRECT_ARGS
  I386MakeInsForIns(Push,Before,added,ins,adreg,0);
  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
  I386MakeInsForIns(Arithmetic,Before,added,ins,I386_ADD,I386_REG_ESP,I386_REG_NONE,4);
#else
  I386MakeInsForIns(MovToReg,Before,added,ins,I386_REG_EAX,adreg,0);
  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
#endif
  I386MakeInsForIns(MovToReg,Before,added,ins,adreg,I386_REG_EAX,0);

  REGSET_FOREACH_REG_INVERSE(save,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForIns(Pop,Before,added,ins,r);
#else
    if (RegsetIn(always_stack,r))
      I386MakeInsForIns(Pop,Before,added,ins,r);
    else
      I386MakeInsForIns(PseudoLoad,Before,added,ins,r);
#endif
  }
  I386MakeInsForIns(ArithmeticToMem,Before,added,ins,I386_SUB,0,I386_REG_ESP,I386_REG_NONE,0,adreg,0);

  /* instructions that come after ins should be inserted in inverted order! */
  if (!ins_sets_flags)
  {
    I386MakeInsForIns(Simple,After,added,ins,I386_POPF);
    I386MakeInsForIns(Arithmetic,After,added,ins,I386_ADD,I386_REG_ESP,I386_REG_NONE,4);
    I386MakeInsForIns(ArithmeticFromMem,After,added,ins,I386_ADD,adreg,0,I386_REG_ESP,I386_REG_NONE,0);
  }
  else
  {
    I386MakeInsForIns(Simple,After,added,ins,I386_POPF);
    I386MakeInsForIns(Pop,After,added,ins,I386_REG_EAX);
    I386MakeInsForIns(MovToMem,After,added,ins,4,I386_REG_ESP,I386_REG_NONE,0,I386_REG_EAX,0);
    I386MakeInsForIns(ArithmeticFromMem,After,added,ins,I386_ADD,adreg,4,I386_REG_ESP,I386_REG_NONE,0);
    I386MakeInsForIns(Pop,After,added,ins,I386_REG_EAX);
    I386MakeInsForIns(Simple,After,added,ins,I386_PUSHF);
    I386MakeInsForIns(Push,After,added,ins,I386_REG_EAX,0);
  }
  
  /*DiabloPrint(stdout,"end result: @iB\n",INS_BBL(ins));*/
} /* }}} */

/* {{{ patch string instructions that use both %edi and %esi
 * again, this is pretty messy, but it works:
 * 
 * MOVS                                 | CMPS
 * ----                                 | ----
 *                                      | 
 *  pushf                               |
 *  push %esi                           | push %esi
 *  push %edi                           | push %edi
 *  push <...>                          | push <...>
 *  push %edi                           | push %edi
 *  call ReverseTranslate               | call ReverseTranslate
 *  mov %eax,(%esp)                     | mov %eax,(%esp)
 *  push <stack slot of esi>(%esp)      | push <stack slot of esi>(%esp)
 *  call ReverseTranslate               | call ReverseTranslate
 *  add 0x4,%esp                        | add 0x4,%esp
 *  mov %eax,%esi                       | mov %eax,%esi
 *  pop %edi                            | pop %edi
 *  pop <...>                           | pop <...>
 *  sub %edi,(%esp)                     | sub %edi,(%esp)
 *  sub %esi,0x4(%esp)                  | sub %esi,0x4(%esp)
 *  movs                                | movs
 *  add (%esp),%edi                     | push %eax
 *  add 0x4(%esp),%esi                  | pushf
 *  add 0x8,%esp                        | pop %eax
 *  popf                                | add 0x8(%esp),%esi
 *                                      | add 0x4(%esp),%edi
 *                                      | mov %eax,0x8(%esp)
 *                                      | pop %eax
 *                                      | add 0x4,%esp
 *                                      | popf
 *
 */
static void DoPatchForDoubleStringIns(t_ins * ins)
{
  t_reg r;
  t_regset save,change;
  t_ins * added;
  t_bool ins_sets_flags = (I386_INS_OPCODE(ins) == I386_CMPSB || I386_INS_OPCODE(ins) == I386_CMPSD);
  int nsaved = 0;

  if (!(I386_INS_OPCODE(ins) == I386_MOVSB || I386_INS_OPCODE(ins) == I386_MOVSD ||
	I386_INS_OPCODE(ins) == I386_CMPSB || I386_INS_OPCODE(ins) == I386_CMPSD))
    FATAL(("wrong kind of instruction"));

  /*DiabloPrint(stdout,"patching @I:\n",ins);*/

  /* which registers need to be saved? */
  RegsetDup(change,FUN_REGS_CHANGED(reversetranslate));
  RegsetSetDiff(change,FUN_REGS_SAVED(reversetranslate));
  RegsetAddReg(change,I386_REG_EAX);
  RegsetDup(save,InsRegsLiveBefore(ins));
  RegsetSetIntersect(save,BBL_CFG(INS_BBL(ins))->sec->obj->description->int_registers);
  RegsetSetIntersect(save,change);
  RegsetSubReg(save,I386_REG_ESP);
  RegsetSubReg(save,I386_REG_ESI);
  RegsetSubReg(save,I386_REG_EDI);

  /* do the actual patching */
  if (!ins_sets_flags)
    I386MakeInsForIns(Simple,Before,added,ins,I386_PUSHF);
  I386MakeInsForIns(Push,Before,added,ins,I386_REG_ESI,0);
  I386MakeInsForIns(Push,Before,added,ins,I386_REG_EDI,0);
  REGSET_FOREACH_REG(save,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForIns(Push,Before,added,ins,r,0);
    ++nsaved;
#else
    if (RegsetIn(always_stack,r))
    {
      I386MakeInsForIns(Push,Before,added,ins,r,0);
      ++nsaved;
    }
    else
      I386MakeInsForIns(PseudoSave,Before,added,ins,r);
#endif
  }

#ifndef DIRECT_ARGS
  I386MakeInsForIns(Push,Before,added,ins,I386_REG_EDI,0);
  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
  I386MakeInsForIns(MovToMem,Before,added,ins,0,I386_REG_ESP,I386_REG_NONE,0,I386_REG_EAX,0);
  I386MakeInsForIns(PushMem,Before,added,ins,4*(nsaved+2),I386_REG_ESP,I386_REG_NONE,0);
  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
  I386MakeInsForIns(Arithmetic,Before,added,ins,I386_ADD,I386_REG_ESP,I386_REG_NONE,4);
#else
  I386MakeInsForIns(MovToReg,Before,added,ins,I386_REG_EAX,I386_REG_EDI,0);
  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
  I386MakeInsForIns(Push,Before,added,ins,I386_REG_EAX,0);
  I386MakeInsForIns(MovFromMem,Before,added,ins,I386_REG_EAX,4*(nsaved+2),I386_REG_ESP,I386_REG_NONE,0);
  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
#endif
  I386MakeInsForIns(MovToReg,Before,added,ins,I386_REG_ESI,I386_REG_EAX,0);
  I386MakeInsForIns(Pop,Before,added,ins,I386_REG_EDI);
  REGSET_FOREACH_REG_INVERSE(save,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForIns(Pop,Before,added,ins,r);
#else
    if (RegsetIn(always_stack,r))
      I386MakeInsForIns(Pop,Before,added,ins,r);
    else
      I386MakeInsForIns(PseudoLoad,Before,added,ins,r);
#endif
  }
  I386MakeInsForIns(ArithmeticToMem,Before,added,ins,I386_SUB,0,I386_REG_ESP,I386_REG_NONE,0,I386_REG_EDI,0);
  I386MakeInsForIns(ArithmeticToMem,Before,added,ins,I386_SUB,4,I386_REG_ESP,I386_REG_NONE,0,I386_REG_ESI,0);
  /* instructions added after the original instruction should be added in reverse order */
  if (!ins_sets_flags)
  {
    /* MOVS */
    I386MakeInsForIns(Simple,After,added,ins,I386_POPF);
    I386MakeInsForIns(Arithmetic,After,added,ins,I386_ADD,I386_REG_ESP,I386_REG_NONE,8);
    I386MakeInsForIns(ArithmeticFromMem,After,added,ins,I386_ADD,I386_REG_ESI,4,I386_REG_ESP,I386_REG_NONE,0);
    I386MakeInsForIns(ArithmeticFromMem,After,added,ins,I386_ADD,I386_REG_EDI,0,I386_REG_ESP,I386_REG_NONE,0);
  }
  else
  {
    /* CMPS */
    I386MakeInsForIns(Simple,After,added,ins,I386_POPF);
    I386MakeInsForIns(Arithmetic,After,added,ins,I386_ADD,I386_REG_ESP,I386_REG_NONE,4);
    I386MakeInsForIns(Pop,After,added,ins,I386_REG_EAX);
    I386MakeInsForIns(MovToMem,After,added,ins,8,I386_REG_ESP,I386_REG_NONE,0,I386_REG_EAX,0);
    I386MakeInsForIns(ArithmeticFromMem,After,added,ins,I386_ADD,I386_REG_EDI,4,I386_REG_ESP,I386_REG_NONE,0);
    I386MakeInsForIns(ArithmeticFromMem,After,added,ins,I386_ADD,I386_REG_ESI,8,I386_REG_ESP,I386_REG_NONE,0);
    I386MakeInsForIns(Pop,After,added,ins,I386_REG_EAX);
    I386MakeInsForIns(Simple,After,added,ins,I386_PUSHF);
    I386MakeInsForIns(Push,After,added,ins,I386_REG_EAX,0);
  }

  /*DiabloPrint(stdout,"end result: @iB\n",INS_BBL(ins));*/
} /* }}} */

/* {{{ DoPatchForRegularIns */
static void DoPatchForRegularIns(t_ins * ins, t_i386_operand * memop)
{
  t_regset int_registers = BBL_CFG(INS_BBL(ins))->sec->obj->description->int_registers;
  t_regset liveafter = InsRegsLiveAfter(ins);
  t_regset used, killed, livebeforecopy;
  t_regset condflags = NullRegs;
  t_bool save_condflags = FALSE;
  t_bool esp_dependent = FALSE;

  t_reg adreg;
  t_bool adreg_is_live;

  t_ins * copy = InsDup(ins);
  t_i386_operand * memop2 = (t_i386_operand *) (((void *)copy) + (((void *)memop) - ((void *)ins)));

  t_ins * added;

#define addf(flag)  RegsetAddReg(condflags,I386_CONDREG_##flag)
    addf(AF); addf(CF); addf(DF); addf(OF); addf(PF); addf(SF); addf(ZF);
#undef addf

  I386OpSetMem(memop2,0,I386_REG_NONE,I386_REG_NONE,0,4);
  I386_OP_FLAGS(memop2) &= ~I386_OPFLAG_ISRELOCATED;
  INS_REGS_USE(copy) = I386InsUsedRegisters(copy);
  INS_REGS_DEF(copy) = I386InsDefinedRegisters(copy);

  killed = INS_REGS_DEF(copy);
  used = INS_REGS_USE(copy);
  RegsetDup(livebeforecopy,liveafter);
  RegsetSetDiff(livebeforecopy,killed);
  RegsetSetUnion(livebeforecopy,used);

  if (RegsetIn(killed,I386_REG_ESP) || RegsetIn(used,I386_REG_ESP))
    esp_dependent = TRUE;

  /* {{{ determine which register to use while computing the real address */
  adreg = I386_REG_EAX;
  adreg_is_live = FALSE;
  RegsetSubReg(int_registers,I386_REG_ESP);
  if (RegsetIn(used,I386_REG_EAX) || (esp_dependent && RegsetIn(livebeforecopy,I386_REG_EAX)))
  {
    /* find another register to store the address */
    /* try to find a suitable dead register */
    REGSET_FOREACH_REG(int_registers,adreg)
      if (!RegsetIn(livebeforecopy,adreg)) 
	break;
    /* if no dead register can be found, take a live register */
    if (!I386IsGeneralPurposeReg(adreg))
      REGSET_FOREACH_REG(int_registers,adreg)
	if (!RegsetIn(used,adreg))
	  break;
    VERIFY_FASSERT(I386IsGeneralPurposeReg(adreg),("could not find suitable register"));
  }

  if (RegsetIn(livebeforecopy,adreg))
    adreg_is_live = TRUE;
  /* }}} */

  /* {{{ add saves */
  t_regset save,change,tmp;
  t_reg r;

  /* find out which registers should be saved */
  RegsetDup(change,FUN_REGS_CHANGED(reversetranslate));
  RegsetSetDiff(change,FUN_REGS_SAVED(reversetranslate));
  RegsetAddReg(change,I386_REG_EAX);
  RegsetAddReg(change,adreg);
  RegsetDup(save,livebeforecopy);
  RegsetSetIntersect(save,int_registers);
  RegsetSetIntersect(save,change);
  RegsetSubReg(save,I386_REG_ESP);
  RegsetDup(tmp,livebeforecopy);
  RegsetSetIntersect(tmp,condflags);
  save_condflags = !RegsetIsEmpty(tmp);

#ifndef USE_PSEUDO
  if (adreg_is_live)
  {
    RegsetSubReg(save,adreg);
    if (!esp_dependent)
    {
      /* push adreg */
      I386MakeInsForIns(Push,Before,added,ins,adreg,0);
    }
    else
    {
      t_reloc * rel;
      /* save adreg in specially designated area */
      I386MakeInsForIns(MovToMem,Before,added,ins,0,I386_REG_NONE,I386_REG_NONE,0,adreg,0);
      I386_OP_FLAGS(I386_INS_DEST(added)) |= I386_OPFLAG_ISRELOCATED;
      rel = RelocTableAddRelocToRelocatable(BBL_CFG(INS_BBL(ins))->sec->obj->reloc_table,
	  NULL_GENERIC,(t_relocatable *)added,GENERIC_ADD_UINT32(NULL_GENERIC,1),
	  (t_relocatable *)stack_save_area,NULL_GENERIC,FALSE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);
      rel->label = StringDup("stack_save_area");
    }
  }
  /* push other registers */
  REGSET_FOREACH_REG(save,r)
    I386MakeInsForIns(Push,Before,added,ins,r,0);
#else
  if (adreg_is_live)
  {
    RegsetSubReg(save,adreg);
    if (RegsetIn(always_stack,adreg) && !esp_dependent)
      I386MakeInsForIns(Push,Before,added,ins,adreg,0);
    else
      I386MakeInsForIns(PseudoSave,Before,added,ins,adreg);
  }

  REGSET_FOREACH_REG(save,r)
  {
    if (RegsetIn(always_stack,r) && (r != adreg))
      I386MakeInsForIns(Push,Before,added,ins,r,0);
    else
      I386MakeInsForIns(PseudoSave,Before,added,ins,r);
  }
#endif

  /* pushf */
  if (save_condflags)
    I386MakeInsForIns(Simple,Before,added,ins,I386_PUSHF);
  /* }}} */

  /* {{{ compute address and do reverse translate */
  /* lea <address expression>, %eax */
#ifndef DIRECT_ARGS
  if (IsSimpleAddressExpression(memop))
  {
    /* shortcut: immediately push the address instead of calculating it in %eax first */
    if (I386_OP_BASE(memop) != I386_REG_NONE)
      I386MakeInsForIns(Push,Before,added,ins,I386_OP_BASE(memop),0);
    else if (I386_OP_INDEX(memop) != I386_REG_NONE && I386_OP_SCALE(memop) == I386_SCALE_1)
      I386MakeInsForIns(Push,Before,added,ins,I386_OP_INDEX(memop),0);
    else
      I386MakeInsForIns(Push,Before,added,ins,I386_REG_NONE,I386_OP_IMMEDIATE(memop));
    if (I386_OP_FLAGS(memop) & I386_OPFLAG_ISRELOCATED)
    {
      t_reloc * rel = I386GetRelocForOp(ins,memop);
      RelocSetFrom(rel,(t_relocatable *)added);
      I386_OP_FLAGS(I386_INS_SOURCE1(added)) |= I386_OPFLAG_ISRELOCATED;
      I386_OP_FLAGS(memop) &= ~I386_OPFLAG_ISRELOCATED;
    }
  }
  else
#endif
  {
    I386MakeInsForIns(Lea,Before,added,ins,I386_REG_EAX,I386_OP_IMMEDIATE(memop),I386_OP_BASE(memop),I386_OP_INDEX(memop),I386_OP_SCALE(memop));
    if (I386_OP_FLAGS(memop) & I386_OPFLAG_ISRELOCATED)
    {
      t_reloc * rel = I386GetRelocForOp(ins,memop);
      RelocSetFrom(rel,(t_relocatable *)added);
      I386_OP_FLAGS(I386_INS_SOURCE1(added)) |= I386_OPFLAG_ISRELOCATED;
      I386_OP_FLAGS(memop) &= ~I386_OPFLAG_ISRELOCATED;
    }
#ifndef DIRECT_ARGS
    /* push %eax */
    I386MakeInsForIns(Push,Before,added,ins,I386_REG_EAX,0);
#endif
  }
  /* call ReverseTranslate */
  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
#ifndef DIRECT_ARGS
  /* add $4, %esp */
  I386MakeInsForIns(Arithmetic,Before,added,ins,I386_ADD,I386_REG_ESP,I386_REG_NONE,4);
#endif
  if (adreg != I386_REG_EAX)
  {
    /* mov %eax,adreg */
    I386MakeInsForIns(MovToReg,Before,added,ins,adreg,I386_REG_EAX,0);
  }
  /* }}} */

  /* {{{ restore all saved registers apart from adreg */
  /* popf */
  if (save_condflags)
    I386MakeInsForIns(Simple,Before,added,ins,I386_POPF);
#ifndef USE_PSEUDO
  /* pop registers */
  REGSET_FOREACH_REG_INVERSE(save,r)
    I386MakeInsForIns(Pop,Before,added,ins,r);
#else
  /* restore all but adreg */
  REGSET_FOREACH_REG_INVERSE(save,r)
  {
    if (RegsetIn(always_stack,r))
      I386MakeInsForIns(Pop,Before,added,ins,r);
    else
      I386MakeInsForIns(PseudoLoad,Before,added,ins,r);
  }
#endif
  /* }}} */

  /* patch the original instruction */
  I386OpSetMem(memop,0,adreg,I386_REG_NONE,0,I386_OP_MEMOPSIZE(memop));
  INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
  INS_REGS_DEF(ins) = I386InsDefinedRegisters(ins);

  /* {{{ restore adreg */
  if (adreg_is_live)
  {
#ifndef USE_PSEUDO
    if (!esp_dependent)
      I386MakeInsForIns(Pop,After,added,ins,adreg);
    else
    {
      t_reloc * rel;
      I386MakeInsForIns(MovFromMem,After,added,ins,adreg,0,I386_REG_NONE,I386_REG_NONE,0);
      I386_OP_FLAGS(I386_INS_SOURCE1(added)) |= I386_OPFLAG_ISRELOCATED;
      rel = RelocTableAddRelocToRelocatable(BBL_CFG(INS_BBL(ins))->sec->obj->reloc_table,
	  NULL_GENERIC,(t_relocatable *)added,GENERIC_ADD_UINT32(NULL_GENERIC,1),
	  (t_relocatable *)stack_save_area,NULL_GENERIC,FALSE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);
      rel->label = StringDup("stack_save_area");
    }
#else
    if (RegsetIn(always_stack,adreg) && !esp_dependent)
      I386MakeInsForIns(Pop,After,added,ins,adreg);
    else
      I386MakeInsForIns(PseudoLoad,After,added,ins,adreg);
#endif
  } /* }}} */

  while (RELOCATABLE_REFERS_TO(copy))
    RelocTableRemoveReloc(BBL_CFG(INS_BBL(ins))->sec->obj->reloc_table,RELOCATABLE_REFERS_TO(copy)->rel);
  Free(copy);
} /* }}} */

/* {{{ DoPatchForRet 
 * translate the return address on the stack.
 * this inserts the following code:
 * push <...>
 * pushf
 * push <offset of return address>(%esp) 
 * call ReverseTranslate 
 * add $0x4,%esp
 * mov %eax, <offset of return address>(%esp)
 * popf 
 * pop <...>
 * ret
 */
static void DoPatchForRet(t_ins * ins)
{
  t_ins * added;
  t_regset save, change,tmp;
  t_regset condflags = NullRegs;
  t_reg r;
  t_bool save_condflags;
  int nsaved = 0;

#define addf(flag)  RegsetAddReg(condflags,I386_CONDREG_##flag)
    addf(AF); addf(CF); addf(DF); addf(OF); addf(PF); addf(SF); addf(ZF);
#undef addf

  /*DiabloPrint(stdout,"patching @I:\n",ins);*/

  /* determine which registers to save */
  RegsetDup(change,FUN_REGS_CHANGED(reversetranslate));
  RegsetSetDiff(change,FUN_REGS_SAVED(reversetranslate));
  RegsetAddReg(change,I386_REG_EAX);
  RegsetDup(save,InsRegsLiveBefore(ins));
  RegsetSetIntersect(save,BBL_CFG(INS_BBL(ins))->sec->obj->description->int_registers);
  RegsetSetIntersect(save,change);
  RegsetSubReg(save,I386_REG_ESP);
  RegsetDup(tmp,InsRegsLiveBefore(ins));
  RegsetSetIntersect(tmp,condflags);
  save_condflags = !RegsetIsEmpty(tmp);

  /* apply the actual patch */
  REGSET_FOREACH_REG(save,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForIns(Push,Before,added,ins,r,0);
    ++nsaved;
#else
    if (RegsetIn(always_stack,r))
    {
      I386MakeInsForIns(Push,Before,added,ins,r,0);
      ++nsaved;
    }
    else
      I386MakeInsForIns(PseudoSave,Before,added,ins,r);
#endif
  }
  if (save_condflags)
  {
    I386MakeInsForIns(Simple,Before,added,ins,I386_PUSHF);
    ++nsaved;
  }
#ifndef DIRECT_ARGS
  I386MakeInsForIns(PushMem,Before,added,ins,4*nsaved,I386_REG_ESP,I386_REG_NONE,0);
  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
  I386MakeInsForIns(Arithmetic,Before,added,ins,I386_ADD,I386_REG_ESP,I386_REG_NONE,4);
#else
  I386MakeInsForIns(MovFromMem,Before,added,ins,I386_REG_EAX,4*nsaved,I386_REG_ESP,I386_REG_NONE,0);
  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
#endif
  I386MakeInsForIns(MovToMem,Before,added,ins,4*nsaved,I386_REG_ESP,I386_REG_NONE,0,I386_REG_EAX,0);
  if (save_condflags)
    I386MakeInsForIns(Simple,Before,added,ins,I386_POPF);
  REGSET_FOREACH_REG_INVERSE(save,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForIns(Pop,Before,added,ins,r);
#else
    if (RegsetIn(always_stack,r))
      I386MakeInsForIns(Pop,Before,added,ins,r);
    else
      I386MakeInsForIns(PseudoLoad,Before,added,ins,r);
#endif
  }

  /*DiabloPrint(stdout,"end result: @iB\n",INS_BBL(ins));*/

} /* }}} */

/* {{{ patch indirect jump instructions */
static void DoPatchForJmp(t_ins * ins, t_i386_operand * memop)
{
  t_i386_operand * op = I386_INS_SOURCE1(ins);
  t_ins * added;
  t_regset save, change, liveafter, tmp;
  t_regset condflags = NullRegs;
  t_reg adreg,r;
  int nsaved = 0, stack_delta = 0;
  t_bool has_dead_reg,has_memop,translate_memop,was_originally_call;
  t_bool save_condflags;

#define addf(flag)  RegsetAddReg(condflags,I386_CONDREG_##flag)
  addf(AF); addf(CF); addf(DF); addf(OF); addf(PF); addf(SF); addf(ZF);
#undef addf

  /*DiabloPrint(stdout,"patching @I:\n",ins);*/

  liveafter = InsRegsLiveAfter(ins);
  /* determine which registers to save */
  RegsetDup(change,FUN_REGS_CHANGED(reversetranslate));
  RegsetSetDiff(change,FUN_REGS_SAVED(reversetranslate));
  RegsetAddReg(change,I386_REG_EAX);
  RegsetDup(save,liveafter);
  RegsetSetIntersect(save,BBL_CFG(INS_BBL(ins))->sec->obj->description->int_registers);
  RegsetSetIntersect(save,change);
  RegsetSubReg(save,I386_REG_ESP);
  RegsetDup(tmp,InsRegsLiveBefore(ins));
  RegsetSetIntersect(tmp,condflags);
  save_condflags = !RegsetIsEmpty(condflags);

  /* find out how hard the translation is going to be */
  REGSET_FOREACH_REG(BBL_CFG(INS_BBL(ins))->sec->obj->description->int_registers,adreg)
    if (!RegsetIn(liveafter,adreg))
      break;

  {
    t_cfg_edge * edge;
    BBL_FOREACH_SUCC_EDGE(INS_BBL(ins),edge)
      if (EdgeTestCategoryOr(edge,ET_CALL|ET_FUNLINK))
	break;
    was_originally_call = (edge ? TRUE : FALSE);
  }
  has_dead_reg = I386IsGeneralPurposeReg(adreg);
  has_memop = (I386_OP_TYPE(op) == i386_optype_mem);
  translate_memop = (memop != NULL);

  /* now do the actual patching */

  /* if the instruction originally was a call, the old return address is already pushed on the stack -> we need to compensate this */
  if (was_originally_call)
    stack_delta += 4;

  /* if there is no suitable dead register, reserve a stack slot for the translated address */
  if (!has_dead_reg)
  {
    I386MakeInsForIns(Arithmetic,Before,added,ins,I386_SUB,I386_REG_ESP,I386_REG_NONE,4);
    stack_delta += 4;
  }
  REGSET_FOREACH_REG(save,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForIns(Push,Before,added,ins,r,0);
    ++nsaved;
#else
    if (RegsetIn(always_stack,r))
    {
      I386MakeInsForIns(Push,Before,added,ins,r,0);
      ++nsaved;
    }
    else
      I386MakeInsForIns(PseudoSave,Before,added,ins,r);
#endif
  }
  stack_delta += 4*nsaved;
  if (save_condflags)
  {
    I386MakeInsForIns(Simple,Before,added,ins,I386_PUSHF);
    stack_delta += 4;
  }

  /* if necessary, translate the address expression first */
  if (translate_memop)
  {
#ifndef DIRECT_ARGS
    if (IsSimpleAddressExpression(memop))
    {
      /* {{{ shortcut: immediately push the address instead of calculating it in %eax first */
      if (I386_OP_BASE(memop) != I386_REG_NONE)
	I386MakeInsForIns(Push,Before,added,ins,I386_OP_BASE(memop),0);
      else if (I386_OP_INDEX(memop) != I386_REG_NONE && I386_OP_SCALE(memop) == I386_SCALE_1)
	I386MakeInsForIns(Push,Before,added,ins,I386_OP_INDEX(memop),0);
      else
	I386MakeInsForIns(Push,Before,added,ins,I386_REG_NONE,I386_OP_IMMEDIATE(memop));
      if (I386_OP_FLAGS(memop) & I386_OPFLAG_ISRELOCATED)
      {
	t_reloc * rel = I386GetRelocForOp(ins,memop);
	RelocSetFrom(rel,(t_relocatable *)added);
	I386_OP_FLAGS(I386_INS_SOURCE1(added)) |= I386_OPFLAG_ISRELOCATED;
	I386_OP_FLAGS(memop) &= ~I386_OPFLAG_ISRELOCATED;
      } /* }}} */
    }
    else
#endif
    {
      /* {{{ lea ...,%eax ; push %eax */
      /* we needn't worry that the base reg is %esp, because then we just wouldn't translate the memop in the first place */
      I386MakeInsForIns(Lea,Before,added,ins,I386_REG_EAX,I386_OP_IMMEDIATE(memop),I386_OP_BASE(memop),I386_OP_INDEX(memop),I386_OP_SCALE(memop));
      if (I386_OP_FLAGS(memop) & I386_OPFLAG_ISRELOCATED)
      {
	t_reloc * rel = I386GetRelocForOp(ins,memop);
	RelocSetFrom(rel,(t_relocatable *)added);
	I386_OP_FLAGS(I386_INS_SOURCE1(added)) |= I386_OPFLAG_ISRELOCATED;
	I386_OP_FLAGS(memop) &= ~I386_OPFLAG_ISRELOCATED;
      }
#ifndef DIRECT_ARGS
      I386MakeInsForIns(Push,Before,added,ins,I386_REG_EAX,0);
#endif
      /* }}} */
    }
    I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
#ifndef DIRECT_ARGS
    I386MakeInsForIns(Arithmetic,Before,added,ins,I386_ADD,I386_REG_ESP,I386_REG_NONE,4);
#endif
  }

  /* push the jump destination on the stack */
  if (!has_memop)
  {
    /* jmp *%reg */
#ifndef DIRECT_ARGS
    I386MakeInsForIns(Push,Before,added,ins,I386_OP_BASE(op),0);
#else
    I386MakeInsForIns(MovToReg,Before,added,ins,I386_REG_EAX,I386_OP_BASE(op),0);
#endif
  }
  else if (!translate_memop)
  {
    /* jmp *<memop that needn't be translated> */
#ifndef DIRECT_ARGS
    I386MakeInsForIns(PushMem,Before,added,ins,I386_OP_IMMEDIATE(op),I386_OP_BASE(op),I386_OP_INDEX(op),I386_OP_SCALE(op));
#else
    I386MakeInsForIns(MovFromMem,Before,added,ins,I386_REG_EAX,I386_OP_IMMEDIATE(op),I386_OP_BASE(op),I386_OP_INDEX(op),I386_OP_SCALE(op));
#endif
    if (I386_OP_FLAGS(op) & I386_OPFLAG_ISRELOCATED)
    {
      t_reloc * rel = I386GetRelocForOp(ins,op);
      RelocSetFrom(rel,(t_relocatable *)added);
      I386_OP_FLAGS(I386_INS_SOURCE1(added)) |= I386_OPFLAG_ISRELOCATED;
      I386_OP_FLAGS(op) &= ~I386_OPFLAG_ISRELOCATED;
    }
    /* if the base register is the stack pointer, we have to compensate for all the stuff we have pushed on the stack */
    if (I386_OP_BASE(op) == I386_REG_ESP)
      I386_OP_IMMEDIATE(I386_INS_SOURCE1(added)) += stack_delta;
  }
  else
  {
    /* jmp *<memop that has just been translated> */
#ifndef DIRECT_ARGS
    I386MakeInsForIns(PushMem,Before,added,ins,0,I386_REG_EAX,I386_REG_NONE,0);
#else
    I386MakeInsForIns(MovFromMem,Before,added,ins,I386_REG_EAX,0,I386_REG_EAX,I386_REG_NONE,0);
#endif
  }

  I386MakeInsForIns(PseudoCall,Before,added,ins,reversetranslate);
#ifndef DIRECT_ARGS
  I386MakeInsForIns(Arithmetic,Before,added,ins,I386_ADD,I386_REG_ESP,I386_REG_NONE,4);
#endif
  if (has_dead_reg)
    I386MakeInsForIns(MovToReg,Before,added,ins,adreg,I386_REG_EAX,0);
  else
    I386MakeInsForIns(MovToMem,Before,added,ins,stack_delta-4,I386_REG_ESP,I386_REG_NONE,0,I386_REG_EAX,0);
  if (save_condflags)
    I386MakeInsForIns(Simple,Before,added,ins,I386_POPF);
  REGSET_FOREACH_REG_INVERSE(save,r)
  {
#ifndef USE_PSEUDO
    I386MakeInsForIns(Pop,Before,added,ins,r);
#else
    if (RegsetIn(always_stack,r))
      I386MakeInsForIns(Pop,Before,added,ins,r);
    else
      I386MakeInsForIns(PseudoLoad,Before,added,ins,r);
#endif
  }

  if (has_dead_reg)
  {
    I386OpSetReg(I386_INS_SOURCE1(ins),adreg,i386_regmode_full32);
    INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
    INS_REGS_DEF(ins) = I386InsDefinedRegisters(ins);
  }
  else
    I386InstructionMakeSimple(ins,I386_RET);

  /*DiabloPrint(stdout,"end result: @iB\n",INS_BBL(ins));*/
  
} /* }}} */

/* {{{ Patch address consumers */
void I386PatchAddressConsumers(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;

  CFG_FOREACH_BBL(cfg,bbl)
  {
    if (!BBL_FUN(bbl)) continue;	/* skip unconnected blocks */

    BBL_FOREACH_INS(bbl,ins)
    {
      t_uint32 oldaddr = G_T_UINT32(INS_OLD_ADDRESS(ins));
      int nmemops = 0,i;
      t_i386_operand * ops[3];
      t_i386_operand * memop = NULL;

      if (oldaddr == 0 || oldaddr == 0xffffffff) continue;

      /* first patch all calls so that they push the _old_ return address onto the stack */
      if (I386_INS_OPCODE(ins) == I386_CALL)
      {
	t_ins * push;
	t_uint32 oldreturn = G_T_UINT32(GENERIC_ADD(INS_OLD_ADDRESS(ins),INS_OLD_SIZE(ins)));
	push = InsNewForBbl(bbl);
	I386InstructionMakePush(push,I386_REG_NONE,oldreturn);
	InsInsertBefore(push,ins);
	I386_INS_OPCODE(ins) = I386_JMP;
      }

      if (I386_INS_OPCODE(ins) == I386_LEA) continue;

      /* get the memory operand */
      ops[0] = I386_INS_DEST(ins);
      ops[1] = I386_INS_SOURCE1(ins);
      ops[2] = I386_INS_SOURCE2(ins);
      for (i=0; i<3; i++)
      {
	if (I386_OP_TYPE(ops[i]) == i386_optype_mem)
	{
	  /* skip operands that are still relocated */
	  if ((I386_OP_FLAGS(ops[i]) & I386_OPFLAG_ISRELOCATED) && !(REL_FLAGS(I386GetRelocForOp(ins,ops[i])) & REL_FLAG_USE_OLD_ADDRESSES))
	    continue;
	  if (I386_OP_BASE(ops[i]) == I386_REG_ESP)
	    continue;
	  if (I386_OP_BASE(ops[i]) == I386_REG_EBP && (FUN_FLAGS(BBL_FUN(bbl)) & FF_HAS_FRAMEPOINTER))
	    continue;
	  nmemops++;
	  memop = ops[i];
	}
      }
      
      /* apply the appropriate patch */
      switch (I386_INS_OPCODE(ins))
      {
	case I386_JMP:
	  if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) != i386_optype_imm)
	    DoPatchForJmp(ins,memop);
	  break;
	case I386_RET:
	  DoPatchForRet(ins);
	  break;
	case I386_INT:
	  /* translation of system call arguments will happen in-kernel via a special patch */
	  break;
	case I386_MOVSB: case I386_MOVSD:
	case I386_CMPSB: case I386_CMPSD:
	  /* esi and edi */
	  DoPatchForDoubleStringIns(ins);
	  break;
	case I386_INSB:  case I386_INSD:
	case I386_STOSB: case I386_STOSD:
	case I386_SCASB: case I386_SCASD:
	  /* edi */
	case I386_OUTSB: case I386_OUTSD:
	case I386_LODSB: case I386_LODSD:
	  /* esi */
	  DoPatchForSingleStringIns(ins);
	  break;
	default:
	  if (nmemops > 1) FATAL(("should only have more than one memop with string instructions"));
	  if (memop)
	    DoPatchForRegularIns(ins,memop);
      }
    }
  }
} /* }}} */

/* {{{ perform relocation with old addresses */
void I386RealRelocWithOldAddresses(t_object * obj)
{
  t_reloc * rel, * tmp;
  OBJECT_FOREACH_RELOC_SAFE(obj,rel,tmp)
  {
    if (!(REL_FLAGS(rel) & REL_FLAG_USE_OLD_ADDRESSES)) continue;

    if (RELOCATABLE_TYPE(REL_FROM(rel)) == RT_INS)
    {
      t_i386_ins * ins = T_I386_INS(REL_FROM(rel));

      if (INS_TYPE(ins) == IT_DATA)
      {
	t_uint32 data;
	t_uint8 *buf = (t_uint8 *)&data;
	t_i386_ins * iter;

	//DiabloPrint(stdout, "relocing data in the code section at @G\n",INS_CADDRESS(ins));

	if (!GENERICISNULL(rel->from_offset)) FATAL(("Implement data with non-zero from_offset"));
	StackExec(rel->code,rel,buf,TRUE,2);

	/* TODO: this only works on little-endian machines */
	ins->data = buf[0];
	iter = (t_i386_ins *) INS_NEXT(ins);  iter->data = buf[1];
	iter = (t_i386_ins *) INS_NEXT(iter); iter->data = buf[2];
	iter = (t_i386_ins *) INS_NEXT(iter); iter->data = buf[3];
      }
      else
      {
	/* skip relocs from jump and call instructions (these are handled
	 * in AdjustJumpOffsets) */
	if (I386_INS_OPCODE(ins) == I386_JMP ||
	    I386_INS_OPCODE(ins) == I386_Jcc ||
	    I386_INS_OPCODE(ins) == I386_JECXZ ||
	    I386_INS_OPCODE(ins) == I386_CALL ||
	    I386_INS_OPCODE(ins) == I386_LOOP ||
	    I386_INS_OPCODE(ins) == I386_LOOPZ ||
	    I386_INS_OPCODE(ins) == I386_LOOP)
	{
	  if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	    continue;
	}

	/* find out which operand has been relocated */
	t_address ret;
	t_i386_operand * op;
	t_reloc_ref * rr = RELOCATABLE_REFERS_TO(ins);
	t_reloc_ref * rr2;
	VERIFY_FASSERT(rr,("relocations corrupt"));
	VERIFY_FASSERT(rr->rel == rel || (rr->next && rr->next->rel == rel),("relocations corrupt")); 

	if (rr->rel == rel) 
	  rr2 = NULL;
	else
	{
	  rr2 = rr;
	  rr = rr->next;
	}

	if (rr2 && GENERIC_GT(rel->from_offset,rr2->rel->from_offset))
	  op = I386InsGetSecondRelocatedOp((t_ins *)ins);
	else
	  op = I386InsGetFirstRelocatedOp((t_ins *)ins);

	ret = StackExec(rel->code,rel,NULL,FALSE,2);
	//DiabloPrint(stdout,"ins @I relocing 0x%x new @G\n",ins,I386_OP_IMMEDIATE(op),ret);
	I386_OP_IMMEDIATE(op) = GENERIC_EXTRACT_UINT32(ret);
	I386_OP_FLAGS(op) &= ~I386_OPFLAG_ISRELOCATED;
      }
    }
    else
    {
      //DiabloPrint(stdout,"relocing in section %s at @G old 0x%x new ",T_SECTION(REL_FROM(rel))->name,GENERIC_ADD(REL_FROM(rel)->caddress,rel->from_offset), *(t_uint32 *)(T_SECTION(REL_FROM(rel))->data + rel->from_offset));
      t_address ret = StackExec(rel->code,rel,T_SECTION(REL_FROM(rel))->data,TRUE,2);
      //DiabloPrint(stdout,"%x\n",*(t_uint32 *)(T_SECTION(REL_FROM(rel))->data + rel->from_offset));
      VERIFY_FASSERT(GENERICISNULL(ret),("Could not properly execute @R",rel));
    }
    RelocTableRemoveReloc(obj->reloc_table,rel);
  }
} /* }}} */

/* {{{ patch system calls in instrumentation code so they tell the kernel not
 * to translate their arguments (this is only necessary when we instrument with
 * old addresses) */
void I386PatchInstrumentationSyscalls(t_cfg * cfg) { t_bbl * bbl; t_ins * ins;
  t_address ffffffff;

  GenericSet_uint32(ffffffff,-1);

  CFG_FOREACH_BBL(cfg,bbl)
  {
    t_ins * added;
    t_cfg_edge * edge;

    if (!GENERIC_EQ(BBL_OLD_ADDRESS(bbl),ffffffff)) continue;

    ins = BBL_INS_LAST(bbl);
    if (!ins || I386_INS_OPCODE(ins) != I386_INT || I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) != 0x80) continue;

    BBL_FOREACH_SUCC_EDGE(bbl,edge)
      if (EDGE_CAT(edge) == ET_SWILINK || EDGE_CAT(edge) == ET_IPSWILINK)
	break;

    I386MakeInsForIns(MovToMem,Before,added,ins,0,I386_REG_NONE,I386_REG_NONE,0,I386_REG_NONE,0);
    RelocTableAddRelocToRelocatable(cfg->sec->obj->reloc_table,NULL_GENERIC,
	(t_relocatable *) added         , GENERIC_ADD_UINT32(NULL_GENERIC,2),
	(t_relocatable *) added_section2, GENERIC_ADD_UINT32(NULL_GENERIC,4),
	FALSE, NULL, NULL, NULL, CALC_ABS "\\" WRITE_32, FALSE);
    I386_OP_FLAGS(I386_INS_DEST(added)) |= I386_OPFLAG_ISRELOCATED;
    
    I386MakeInsForBbl(MovToMem,Prepend,added,CFG_EDGE_TAIL(edge),0,I386_REG_NONE,I386_REG_NONE,0,I386_REG_NONE,1);
    RelocTableAddRelocToRelocatable(cfg->sec->obj->reloc_table,NULL_GENERIC,
	(t_relocatable *) added         , GENERIC_ADD_UINT32(NULL_GENERIC,2),
	(t_relocatable *) added_section2, GENERIC_ADD_UINT32(NULL_GENERIC,4),
	FALSE, NULL, NULL, NULL, CALC_ABS "\\" WRITE_32, FALSE);
    I386_OP_FLAGS(I386_INS_DEST(added)) |= I386_OPFLAG_ISRELOCATED;
  }
} /* }}} */

#endif

/* vim: set shiftwidth=2 foldmethod=marker:*/
