/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* new i386 layout and deflowgraph algorithm, by Dominique
 * this is intended to be a lot more general than the old 
 * algorithm, with support for splitting the merged code
 * section as needed for linux kernel optimization */

#include <stdlib.h>
#include <string.h>
#include "i386_instructions.h"
#include "i386_assemble.h"
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_code_layout.h"
#include "../../kernel/diablo_section.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_stack_machine.h"
#include "../../kernel/diablo_layout.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../diablo_options.h"

#define SHORT_DISPL_HI_BOUND CHAR_MAX
#define SHORT_DISPL_LO_BOUND CHAR_MIN
#define FULL_DISPL_HI_BOUND  INT_MAX
#define FULL_DISPL_LO_BOUND  INT_MIN                                

#ifdef OBFUSCATION_SUPPORT
#include "../../obfuscation/obf_branch.h"
void I386FillOpcodesObf(t_object * obj);
#endif
#ifdef INSTRUMENTATION
#include "../../atom_internal.h"
#endif

/*typedef struct _t_chain_holder {
  t_bbl ** chains;
  t_uint32 nchains;
} t_chain_holder;
*/

/*static t_address AssignAddressesInChain(t_bbl * bbl, t_address start)
{
  for (; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
  {
    BblSetAddressSuper(bbl,start);
    start = GENERIC_ADD(start,BBL_CSIZE(bbl));
  }
  return start;
}*/

/* {{{ create chains of basic blocks */
#define ET_MUSTCHAIN  	ET_FALLTHROUGH | ET_IPFALLTHRU | ET_FUNLINK | ET_IPFUNLINK | ET_SWILINK | ET_IPSWILINK
/*TODO: SWITCH TABLES!!! */
static void I386CreateChains(t_cfg * cfg, t_chain_holder * ch)
{
  t_bbl * i_bbl;
  t_uint32 nbbls = 0, nchains = 0;
  int i;
 
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    BBL_FIRST_IN_CHAIN(i_bbl) = BBL_LAST_IN_CHAIN(i_bbl) = NULL;
    BBL_NEXT_IN_CHAIN(i_bbl) = BBL_PREV_IN_CHAIN(i_bbl) = NULL;
    /* count all bbls except return blocks and hell nodes (those will not be chained) */
    if (!IsHellNode(i_bbl) && (!BBL_FUN(i_bbl) || (FunctionGetExitBlock(BBL_FUN(i_bbl)) != i_bbl)))
      nbbls++;
  }
  nchains = nbbls;
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_cfg_edge * i_edge;

    /* skip return blocks and hell nodes */
    if (IsHellNode(i_bbl) || (BBL_FUN(i_bbl) && (FunctionGetExitBlock(BBL_FUN(i_bbl)) == i_bbl)))
      continue;

    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      if (EdgeTestCategoryOr(i_edge,ET_MUSTCHAIN))
	break;
    }
    if (i_edge)
    {
      t_bbl * tail = CFG_EDGE_TAIL(i_edge);
      BBL_PREV_IN_CHAIN(tail) = i_bbl;
      BBL_NEXT_IN_CHAIN(i_bbl) = tail;
      nchains--; /* each time two blocks get chained, the number of chains decreases by one */
    }
  }

  ch->chains = Malloc(nchains*sizeof(t_bbl *));
  ch->nchains = nchains;

  i = 0;
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_bbl * j_bbl, * prev;

    /* skip return blocks and hell nodes */
    if (IsHellNode(i_bbl) || (BBL_FUN(i_bbl) && (FunctionGetExitBlock(BBL_FUN(i_bbl)) == i_bbl)))
      continue;

    if (BBL_PREV_IN_CHAIN(i_bbl)) continue; /* we're looking for the heads of the chains here */

    ch->chains[i++] = i_bbl;
    for (j_bbl = i_bbl; j_bbl; prev = j_bbl, j_bbl = BBL_NEXT_IN_CHAIN(j_bbl))
    {
      BBL_FIRST_IN_CHAIN(j_bbl) = i_bbl;
    }
    for (j_bbl = prev; j_bbl; j_bbl = BBL_PREV_IN_CHAIN(j_bbl))
    {
      BBL_LAST_IN_CHAIN(j_bbl) = prev;
    }
  }
}
/* }}} */

/* {{{ determine the layout order for a set of layout chains
 * after the final order has been determined, all chains
 * are concatenated to ch->chains[0] */

/* {{{ helper function that merges two chains */
void I386MergeTwoChains(t_bbl * ca, t_bbl * cb)
{
  t_bbl * bbl;
  t_address length_of_ca = GENERIC_ADD(BBL_CADDRESS(BBL_LAST_IN_CHAIN(ca)),BBL_CSIZE(BBL_LAST_IN_CHAIN(ca)));

  BBL_NEXT_IN_CHAIN(BBL_LAST_IN_CHAIN(ca)) = cb;
  BBL_PREV_IN_CHAIN(cb) = BBL_LAST_IN_CHAIN(ca);
  for (bbl = ca; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
  {
    BBL_FIRST_IN_CHAIN(bbl) = ca;
    BBL_LAST_IN_CHAIN(bbl) = BBL_LAST_IN_CHAIN(cb);
  }
  AssignAddressesInChain(cb,length_of_ca);
}
/* }}} */

#ifndef KERNEL_OPT
/* {{{ find candidate for appending */
int I386FindAppendable(t_chain_holder * ch, t_address chainsizes[], int index)
{
  t_bbl * bbl;
  t_ins * ins;
  t_cfg_edge * edge;
  int ret = -1;

  bbl = BBL_LAST_IN_CHAIN(ch->chains[index]);
  while (bbl && !BBL_INS_LAST(bbl))
    bbl = BBL_PREV_IN_CHAIN(bbl);

  while (bbl && (G_T_UINT32(GENERIC_SUB(chainsizes[index],GENERIC_ADD(BBL_CADDRESS(bbl),BBL_CSIZE(bbl)))) < 128))
  {
    t_uint32 dist_to_end_of_chain = G_T_UINT32(GENERIC_SUB(chainsizes[index],GENERIC_ADD(BBL_CADDRESS(bbl),BBL_CSIZE(bbl))));

    ins = BBL_INS_LAST(bbl);
    if (INS_TYPE(ins) == IT_BRANCH && I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm && I386_INS_OPCODE(ins)!=I386_RET)
    {
      t_bbl * tail;
      BBL_FOREACH_SUCC_EDGE(bbl,edge)
	if (EdgeTestCategoryOr(edge,ET_JUMP|ET_IPJUMP|ET_CALL))
	  break;
      VERIFY_FASSERT(edge,("need edge @ieB",bbl));
      tail = CFG_EDGE_TAIL(edge);
      if (BBL_FIRST_IN_CHAIN(tail) && BBL_FIRST_IN_CHAIN(tail) != BBL_FIRST_IN_CHAIN(bbl))
      {
	if (G_T_UINT32(BBL_CADDRESS(tail)) + dist_to_end_of_chain < 128)
	{
	  /* the chain of block tail is eligible for appending to the current chain.
	   * we just need to find its index in the array of chains */
	  for (ret = 0; ret < ch->nchains; ret++)
	    if (ch->chains[ret] == BBL_FIRST_IN_CHAIN(tail))
	      return ret;
#ifndef KERNEL_OPT
	  VERIFY_FASSERT(ret < ch->nchains,("could not find chain of bbl @ieB",tail));
#endif
	}
      }
    }
    do { bbl = BBL_PREV_IN_CHAIN(bbl); } while (bbl && !BBL_INS_LAST(bbl));
  }
  return -1;
}
/* }}} */
/* {{{ find candidate for prepending */
int I386FindPrependable(t_chain_holder * ch, t_address chainsizes[], int index)
{
  t_bbl * bbl;
  t_ins * ins;
  t_cfg_edge * edge;

  bbl = ch->chains[index];
  while (bbl && !BBL_INS_LAST(bbl))
    bbl = BBL_NEXT_IN_CHAIN(bbl);

  while (bbl && (G_T_UINT32(GENERIC_ADD(BBL_CADDRESS(bbl),BBL_CSIZE(bbl))) < 128))
  {
    ins = BBL_INS_LAST(bbl);
    if (INS_TYPE(ins) == IT_BRANCH && I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm && I386_INS_OPCODE(ins)!=I386_RET)
    {
      t_bbl * tail;
      BBL_FOREACH_SUCC_EDGE(bbl,edge)
	if (EdgeTestCategoryOr(edge,ET_JUMP|ET_IPJUMP|ET_CALL))
	  break;
      VERIFY_FASSERT(edge,("need edge"));
      tail = CFG_EDGE_TAIL(edge);
      if (BBL_FIRST_IN_CHAIN(tail) && BBL_FIRST_IN_CHAIN(tail) != BBL_FIRST_IN_CHAIN(bbl))
      {
	int chainno;
	t_uint32 dist_to_end_of_chain;
	/* find the chain index of the chain of tail */
	for (chainno=0; chainno<ch->nchains; chainno++)
	  if (ch->chains[chainno] == BBL_FIRST_IN_CHAIN(tail))
	    break;
	if (chainno < ch->nchains)
	{
	  dist_to_end_of_chain = G_T_UINT32(GENERIC_SUB(chainsizes[chainno],BBL_CADDRESS(tail)));
	  if (G_T_UINT32(GENERIC_ADD(BBL_CADDRESS(bbl),BBL_CSIZE(bbl))) + dist_to_end_of_chain < 128)
	    return chainno;
	}
#ifndef KERNEL_OPT
	else
	  FATAL(("could not find chain of bbl @ieB",tail));
#endif
      }
    }
    do { bbl = BBL_NEXT_IN_CHAIN(bbl); } while (bbl && !BBL_INS_LAST(bbl));
  }
  return -1;
}
/* }}} */
/* {{{ I386OrderChains */
static void I386OrderChains(t_chain_holder * ch)
{
  t_bbl * bbl, * prev;
  int i;
  t_address * chainsizes = Malloc(ch->nchains * sizeof(t_address *));
  t_address null = GENERIC_SUB(BBL_CADDRESS(ch->chains[0]),BBL_CADDRESS(ch->chains[0]));

  /* store in the c_address field of each ins and bbl their offset to the start of their chain */
  for (i=0; i<ch->nchains; i++)
    chainsizes[i] = AssignAddressesInChain(ch->chains[i],null);

  printf("starting with %d chains...\n", ch->nchains);
  /* try to find chains that can be appended so that jump offsets stay small */
  i = 0;
  while (i < ch->nchains)
  {
    int j = I386FindAppendable(ch,chainsizes,i);
    while (j != -1)
    {
      I386MergeTwoChains(ch->chains[i],ch->chains[j]);
      chainsizes[i] = GENERIC_ADD(chainsizes[i],chainsizes[j]);
      memmove(&(ch->chains[j]),&(ch->chains[j+1]),sizeof(t_bbl *)*(ch->nchains-j-1));
      memmove(&(chainsizes[j]),&(chainsizes[j+1]),sizeof(t_address)*(ch->nchains-j-1));
      ch->nchains--;
      if (j < i) i--;
      j = I386FindAppendable(ch,chainsizes,i);
    }
    ++i;
  }
  
  /* try to find chains that can be prepended so that jump offsets stay small */
  i = 0;
  while (i < ch->nchains)
  {
    int j = I386FindPrependable(ch,chainsizes,i);
    while (j != -1)
    {
      I386MergeTwoChains(ch->chains[j],ch->chains[i]);
      chainsizes[j] = GENERIC_ADD(chainsizes[i],chainsizes[j]);
      if (j > i)
      {
	memmove(&(ch->chains[i]),&(ch->chains[i+1]),sizeof(t_bbl *)*(ch->nchains-i-1));
	memmove(&(chainsizes[i]),&(chainsizes[i+1]),sizeof(t_address)*(ch->nchains-i-1));
      }
      else if (j < i)
      {
	ch->chains[i] = ch->chains[j];
	chainsizes[i] = chainsizes[j];
	memmove(&(ch->chains[j]),&(ch->chains[j+1]),sizeof(t_bbl *)*(ch->nchains-j-1));
	memmove(&(chainsizes[j]),&(chainsizes[j+1]),sizeof(t_address)*(ch->nchains-j-1));
	--i;
      }
      ch->nchains--;
  
      j = I386FindPrependable(ch,chainsizes,i);
    }
    ++i;
  }

#ifdef KERNEL_OPT
  /* make sure the entry point is the first block of the .text section */
  {
    t_cfg * cfg = BBL_CFG(ch->chains[0]);
    t_bbl * entry = cfg->entries->entry_bbl;
    VERIFY_FASSERT(BBL_PREV_IN_CHAIN(entry) == NULL,("entry block should be first in chain"));
    /* move the entry chain to the front of the section */
    for (i=0; i<ch->nchains; i++)
      if (ch->chains[i] == entry)
	break;
    if (i < ch->nchains)
    {
      DiabloPrint(stdout,"Moving kernel entry from position %d to the front of the chain...\n",i);
      ch->chains[i] = ch->chains[0];
      ch->chains[0] = entry;
    }
  }
#endif

  printf("merging remaining %d chains...\n",ch->nchains);
  /* merge all remaining chains */
  prev = NULL;
  for (i=0; i<ch->nchains; i++)
  {
    for (bbl = ch->chains[i]; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    {
      BBL_FIRST_IN_CHAIN(bbl) = ch->chains[0];
      BBL_LAST_IN_CHAIN(bbl) = BBL_LAST_IN_CHAIN(ch->chains[ch->nchains-1]);
      if (!BBL_PREV_IN_CHAIN(bbl) && prev)
      {
	BBL_PREV_IN_CHAIN(bbl) = prev;
	BBL_NEXT_IN_CHAIN(prev) = bbl;
      }
      prev = bbl;
    }
  }
  ch->nchains = 1;

  Free(chainsizes);
} /* }}} */ 
#else
/* {{{ helper function for qsort */
static int __helper_order_chains(const void *a, const void *b)
{
  t_bbl * blocka = *((t_bbl **)a);
  t_bbl * blockb = *((t_bbl **)b);
  if (GENERIC_LT(BBL_OLD_ADDRESS(blocka),BBL_OLD_ADDRESS(blockb)))
    return -1;
  else if (GENERIC_GT(BBL_OLD_ADDRESS(blocka),BBL_OLD_ADDRESS(blockb)))
    return 1;
  
  return 0;
}
/* }}} */
/* {{{ I386OrderChains */
static void I386OrderChains(t_chain_holder * ch)
{
  int i;
  /* simple layout: just put all chains in their original order */
  qsort(ch->chains,ch->nchains,sizeof(t_bbl *),__helper_order_chains);
  for (i=1;i<ch->nchains;i++)
    I386MergeTwoChains(ch->chains[0],ch->chains[i]);
  ch->nchains = 1;
} /* }}} */
#endif
/* }}} */

#ifdef KERNEL_OPT
/* {{{ create new section of type DEFLOWGRAPHED_SECTION */
static t_section * CreateDeflowgraphedSection(t_object * obj, t_chain_holder * ch, t_string name)
{
  t_section * sec;
  t_bbl * bbl;
  t_ins * ins;
  t_address size = NULL_GENERIC;

  sec = SectionNew(obj,DEFLOWGRAPHED_CODE_SECTION);
  SectionSetCAddressSuper(sec,BBL_CADDRESS(ch->chains[0]));
  SectionSetOldAddressSuper(sec,BBL_CADDRESS(ch->chains[0]));
  sec->name = StringDup(name);
  sec->data = ch->chains[0];
  /* insert in the section hash table. the node will probably exist already,
   * so we just have to add the section to the node */
  {
    t_section_hash_node * shn = HashLookup(obj->section_table, name);
    if (!shn)
      HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(sec->name,sec));
    else if (!shn->section)
      shn->section = sec;
    else
      FATAL(("section %s already existed",name));
  }

  for (bbl = ch->chains[0]; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    BBL_FOREACH_INS(bbl,ins)
      INS_SECTION(ins) = sec;

  for (bbl = ch->chains[0]; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    size = GENERIC_ADD(size,BBL_CSIZE(bbl));
  SECTION_OLD_SIZE(sec) = size;
  SECTION_CSIZE(sec) = size;

  return sec;
}
/* }}} */
#endif

/* {{{ copy all instructions from the chain of the DEFLOWGRAPHED_SECTION into a linear buffer to create a DISASSEMBLED_SECTION
 * while we're at it, migrate all relocations to the copied instructions */
static void I386SectionFromDeflowgraphedToDisassembled(t_object * obj)
{
  t_bbl * bbl;
  t_ins * ins;
  int i;

  for (i = 0; i < obj->ncodes; i++)
  {
    t_section * sec = obj->code[i];
    t_uint32 nins = 0, count;

    if (SECTION_TYPE(sec) != DEFLOWGRAPHED_CODE_SECTION)
      FATAL(("Section %s has type %c",sec->name,sec->type));

    for (bbl = sec->data; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
      BBL_FOREACH_INS(bbl,ins)
	nins++;

    SECTION_TMP_BUF(sec)=Malloc(sizeof(t_i386_ins)*nins);

    count = 0;
    for (bbl = sec->data; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    {
      BBL_FOREACH_INS(bbl,ins)
      {
	t_ins * copy = (t_ins *) &(T_I386_INS(SECTION_TMP_BUF(sec))[count++]);
	/* copy ins */
	INS_COPY(ins) = copy;
	memmove(copy,ins,sizeof(t_i386_ins));
	/* migrate relocations from instruction */
	RELOCATABLE_REFERS_TO(copy) = NULL;
	while (RELOCATABLE_REFERS_TO(ins))
	{
	  t_reloc * rel = RELOCATABLE_REFERS_TO(ins)->rel;
	  RelocSetFrom(rel,(t_relocatable *)copy);
	  rel->edge = NULL;
	}
      }
      /* migrate relocations to bbl */
      if (RELOCATABLE_REFED_BY(bbl))
      {
	if (BBL_INS_FIRST(bbl))
	{
	  t_ins * to = INS_COPY(BBL_INS_FIRST(bbl));
	  while (RELOCATABLE_REFED_BY(bbl))
	  {
	    t_reloc * rel = RELOCATABLE_REFED_BY(bbl)->rel;
	    RelocSetToRelocatable(rel,(t_relocatable *)to);
	    rel->addend = GENERIC_SUB(rel->addend,GENERIC_SUB(INS_CADDRESS(to),BBL_CADDRESS(bbl)));
	    rel->edge = NULL;
	  }
	}
	else
	{
	  /* move relocation to next block with instructions in the chain */
	  t_bbl * next = bbl;
	  t_bbl * prev = bbl;
	  while (!BBL_INS_FIRST(next))
	  {
	    next = BBL_NEXT_IN_CHAIN(next);
	    VERIFY_FASSERT(next,("should have next in chain"));
	    VERIFY_FASSERT(BBL_SUCC_FIRST(prev),("bbl should have successor"));
	    VERIFY_FASSERT(BBL_SUCC_FIRST(prev) == BBL_SUCC_LAST(prev),("bbl should have exactly one successor"));
	    VERIFY_FASSERT(EDGE_CAT(BBL_SUCC_FIRST(prev)) & (ET_FALLTHROUGH | ET_IPFALLTHRU), ("should be fallthrough edge"));
	    prev = next;
	  }
	  while (RELOCATABLE_REFED_BY(bbl))
	    RelocSetToRelocatable(RELOCATABLE_REFED_BY(bbl)->rel,(t_relocatable *)next);
	}
      }
    }
    printf("Done, wrote %d instructions \n",count);

    for (count = 0; count < nins; count++) 
    {
      INS_NEXT(&(T_I386_INS(SECTION_TMP_BUF(sec))[count]))=T_INS(&(T_I386_INS(SECTION_TMP_BUF(sec))[count+1]));
      INS_BBL(&(T_I386_INS(SECTION_TMP_BUF(sec))[count])) = NULL;
    }
    INS_NEXT(&(T_I386_INS(SECTION_TMP_BUF(sec))[nins-1]))=NULL;

    sec->data = SECTION_TMP_BUF(sec);
    SECTION_TMP_BUF(sec) = NULL;
    sec->type = DISASSEMBLED_CODE_SECTION;
    SECTION_NINS(sec) = nins;
  }
}
/* }}} */

/* {{{ KillUselessJumps */
static void KillUselessJumps(t_object * obj)
{
  int i;
  t_uint32 nkills = 0;

  for (i=0; i<obj->ncodes; i++)
  {
    t_bbl * bbl;
    t_section * sec = obj->code[i];
    VERIFY_FASSERT(SECTION_TYPE(sec) == DEFLOWGRAPHED_CODE_SECTION,("Section %s has unexpected type %c",sec->name,sec->type));

    for (bbl = sec->data; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    {
      t_ins * ins = BBL_INS_LAST(bbl);
      t_cfg_edge * edge;

      if (!ins) continue;
      if (I386_INS_OPCODE(ins) != I386_JMP && I386_INS_OPCODE(ins) != I386_Jcc && I386_INS_OPCODE(ins) != I386_JECXZ) 
	continue;

      BBL_FOREACH_SUCC_EDGE(bbl,edge)
	if (EdgeTestCategoryOr(edge,ET_JUMP | ET_IPJUMP | ET_SWITCH | ET_IPSWITCH))
	  break;
#ifndef INSTRUMENTATION
      VERIFY_FASSERT(edge,("Could not find jump edge for @I in block @ieB",ins,bbl));
#else
      /* when using old addresses, calls are changed into jumps, but the call
       * edges remain. */
      if (!edge) continue;
#endif

      if (EDGE_CAT(edge) == ET_SWITCH) continue;
      
      if (CFG_EDGE_TAIL(edge) == BBL_NEXT_IN_CHAIN(bbl))
      {
#ifdef OBFUSCATION_SUPPORT
	if(global_options.obf_dis)
	{
	  if(I386_INS_CORR(ins)!=NULL)
	    InsKill(I386_INS_CORR(ins));
	}
#endif

#ifdef KERNEL_OPT
	/* if the control registers have been updated in the same block, 
	 * the jump has a side effect: the prefetch queue is flushed.
	 * consequently, we can't delete the jump instruction */
	{
	  t_bbl * bbl = INS_BBL(ins);
	  t_ins * iter;
	  t_regset usedef = NullRegs;
	  BBL_FOREACH_INS(bbl,iter)
	  {
	    RegsetSetUnion(usedef,INS_REGS_USE(iter));
	    RegsetSetUnion(usedef,INS_REGS_DEF(iter));
	  }
	  if (RegsetIn(usedef,I386_REG_CR0) ||
	      RegsetIn(usedef,I386_REG_CR1) ||
	      RegsetIn(usedef,I386_REG_CR2) ||
	      RegsetIn(usedef,I386_REG_CR3) ||
	      RegsetIn(usedef,I386_REG_CR4))
	    continue;
	}
#endif
	InsKill(ins);
	nkills++;
	EDGE_CAT(edge) = (EDGE_CAT(edge) == ET_JUMP) ? ET_FALLTHROUGH : ET_IPFALLTHRU;
      }
    }
  }
  DiabloPrint(stdout,"[KillUselessJumps] Killed %d useless jumps\n", nkills);
}
/* }}} */

/* {{{ adjust jump offsets (encode them in 1 byte if possible, in 4 if necessary) */

/* look for short jumps that should become long jumps and change the instructions accordingly */
static t_bool AdjustJumpOffsetsUp(t_object * obj)
{
  int i;
  t_bbl * i_bbl;
  t_bool change = FALSE;

  for (i = 0; i < obj->ncodes; i++)
  {
    for (i_bbl = T_BBL(obj->code[i]->data); i_bbl; i_bbl = BBL_NEXT_IN_CHAIN(i_bbl))
    {
      t_address offset;
      t_int32 sval;
      t_bbl * dest;
      t_cfg_edge * i_edge;
      t_ins * last = BBL_INS_LAST(i_bbl);
      t_i386_operand * op;

      if (!last || (I386_INS_OPCODE(last) != I386_JMP && I386_INS_OPCODE(last) != I386_Jcc 
	    && I386_INS_OPCODE(last) != I386_JECXZ && I386_INS_OPCODE(last) != I386_LOOP
	    && I386_INS_OPCODE(last) != I386_LOOPZ && I386_INS_OPCODE(last) != I386_LOOPNZ))
	continue;

      /* we're only interested in jumps with an immediate short offset */
      op = I386_INS_SOURCE1(last);
      if (I386_OP_TYPE(op) != i386_optype_imm || I386_OP_IMMEDSIZE(op) != 1)
	continue;

      /* find the destination of the control transfer */
      BBL_FOREACH_SUCC_EDGE(i_bbl, i_edge)
	if (EdgeTestCategoryOr(i_edge,ET_CALL | ET_JUMP | ET_IPJUMP))
	  break;
      VERIFY_FASSERT(i_edge,("Could not find edge for @I",last));
      dest = CFG_EDGE_TAIL(i_edge);

      offset = GENERIC_SUB(BBL_CADDRESS(dest),GENERIC_ADD(INS_CADDRESS(last),INS_CSIZE(last)));
      sval = GENERIC_EXTRACT_INT32(offset);
      
      if (sval < -128 || sval > 127)
      {
	if (I386_INS_OPCODE(last) != I386_JECXZ && I386_INS_OPCODE(last) != I386_LOOP
	    && I386_INS_OPCODE(last) != I386_LOOPZ && I386_INS_OPCODE(last) != I386_LOOPNZ)
	{
	  /* make long jump */
	  I386_OP_IMMEDSIZE(op) = 4;
	  I386_OP_IMMEDIATE(op) = G_T_UINT32(offset);
	  INS_CSIZE(last) = I386InsGetSize(last);
	}
	else
	{
	  /* no long jump version of jecxz/loop(nz) available: we'll have to do something rather complicated instead
	   * jecxz <dest> becomes:
	   *
	   * jecxz <trampoline>
	   * jmp <fallthrough>
	   * trampoline: jmp <dest>
	   * fallthrough: ...
	   */
	  t_bbl * blk_ft, * blk_dest;
	  t_ins * jmp_ft, * jmp_dest;
	  t_cfg * cfg = BBL_CFG(i_bbl);
	  t_cfg_edge * edge;

	  /* add blocks */
	  blk_ft = BblNew(cfg);
	  blk_dest = BblNew(cfg);
	  BBL_NEXT_IN_CHAIN(blk_ft) = blk_dest;
	  BBL_PREV_IN_CHAIN(blk_dest) = blk_ft;
	  BBL_NEXT_IN_CHAIN(blk_dest) = BBL_NEXT_IN_CHAIN(i_bbl);
	  BBL_PREV_IN_CHAIN(blk_ft) = i_bbl;
	  BBL_NEXT_IN_CHAIN(i_bbl) = blk_ft;
	  BBL_PREV_IN_CHAIN(BBL_NEXT_IN_CHAIN(blk_dest)) = blk_dest;

	  /* add instructions */
	  jmp_ft = InsNewForBbl(blk_ft);
	  InsAppendToBbl(jmp_ft,blk_ft);
	  I386InstructionMakeJump(jmp_ft);
	  jmp_dest = InsNewForBbl(blk_dest);
	  InsAppendToBbl(jmp_dest,blk_dest);
	  I386InstructionMakeJump(jmp_dest);

	  /* replace the edges */
	  BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
	    if (EDGE_CAT(edge) == ET_FALLTHROUGH)
	      break;
	  VERIFY_FASSERT(edge, ("Could not find fallthrough edge after @I\n", last));
	  CfgEdgeKill(edge);
	  BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
	    if (EDGE_CAT(edge) == ET_JUMP)
	      break;
	  VERIFY_FASSERT(edge, ("Could not find jump edge after @I\n", last));
	  CfgEdgeNew(cfg,blk_dest,CFG_EDGE_TAIL(edge),ET_JUMP);
	  CfgEdgeKill(edge);
	  CfgEdgeNew(cfg,i_bbl,blk_ft,ET_FALLTHROUGH);
	  CfgEdgeNew(cfg,blk_ft,BBL_NEXT_IN_CHAIN(blk_dest),ET_JUMP);
	  CfgEdgeNew(cfg,i_bbl,blk_dest,ET_JUMP);

	  VERBOSE(0,("@I out of reach: extending by trampoline", last));
	}
	change = TRUE;
      }
    }
  }
  return change;
}

static t_bool AdjustJumpOffsetsDown(t_object * obj)
{
  int i;
  t_bbl * i_bbl;
  t_bool change = FALSE;

  for (i = 0; i < obj->ncodes; i++)
  {
    for (i_bbl = T_BBL(obj->code[i]->data); i_bbl; i_bbl = BBL_NEXT_IN_CHAIN(i_bbl))
    {
      t_address offset;
      t_int32 sval;
      t_bbl * dest;
      t_cfg_edge * i_edge;
      t_ins * last = BBL_INS_LAST(i_bbl);
      t_i386_operand * op;

      if (!last || (I386_INS_OPCODE(last) != I386_JMP && I386_INS_OPCODE(last) != I386_Jcc))
	continue;

      /* we're only interested in jumps with an immediate long offset */
      op = I386_INS_SOURCE1(last);
      if (I386_OP_TYPE(op) != i386_optype_imm || I386_OP_IMMEDSIZE(op) != 4)
	continue;

      /* find the destination of the control transfer */
      BBL_FOREACH_SUCC_EDGE(i_bbl, i_edge)
	if (EdgeTestCategoryOr(i_edge,ET_CALL | ET_JUMP | ET_IPJUMP))
	  break;
      VERIFY_FASSERT(i_edge,("Could not find edge for @I",last));
      dest = CFG_EDGE_TAIL(i_edge);

      offset = GENERIC_SUB(BBL_CADDRESS(dest),GENERIC_ADD(INS_CADDRESS(last),INS_CSIZE(last)));
      sval = GENERIC_EXTRACT_INT32(offset);
      
      if (sval >= -128 && sval < 128)
      {
	/* make short jump */
	I386_OP_IMMEDSIZE(op) = 1;
	I386_OP_IMMEDIATE(op) = G_T_UINT32(offset);
	INS_CSIZE(last) = I386InsGetSize(last);
	change = TRUE;
      }
      
    }
  }
  return change;
}
/* }}} */

#ifdef OBFUSCATION_SUPPORT
/* I386FillOpcodesObf {{{ */
void I386FillOpcodesObf(t_object * obj)
{
  int i;
  t_bbl * i_bbl;
  for (i=0; i<obj->ncodes; i++)
  {
    for (i_bbl = T_BBL(obj->code[i]->data); i_bbl; i_bbl = BBL_NEXT_IN_CHAIN(i_bbl))
    {
      t_ins * i_ins;
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	if(I386_INS_CORR(i_ins)!=NULL)
	{
	  t_ins * h_ins=I386_INS_CORR(i_ins);
	  t_uint8 buf[15];

	  I386AssembleIns(i_ins,buf);
	  I386OpSetImm(I386_INS_SOURCE1(h_ins),buf[0],1);
	}

      }
    }
  }
}
/* }}} */
#endif

/* {{{ UpdateJumpDisplacements */
static void UpdateJumpDisplacements(t_object * obj)
{
  int i;
  t_bbl * i_bbl;
  for (i=0; i<obj->ncodes; i++)
  {
    for (i_bbl = T_BBL(obj->code[i]->data); i_bbl; i_bbl = BBL_NEXT_IN_CHAIN(i_bbl))
    {
      t_i386_operand * op;
      t_ins * last = BBL_INS_LAST(i_bbl);
      t_address offset;
      t_cfg_edge * i_edge;
      t_bbl * dest;

      if (!last) continue;
      switch (I386_INS_OPCODE(last))
      {
	case I386_CALL:
	case I386_JMP: case I386_Jcc: case I386_JECXZ:
	case I386_LOOP: case I386_LOOPZ: case I386_LOOPNZ:
	  break;
	default:
	  continue;
      }
      op = I386_INS_SOURCE1(last);
      if (I386_OP_TYPE(op) != i386_optype_imm)
	continue;

      /* find the jump destination */
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	if (EdgeTestCategoryOr(i_edge,ET_CALL | ET_JUMP | ET_IPJUMP))
	  break;
      VERIFY_FASSERT(i_edge,("Could not find edge for @I",last));
      dest = CFG_EDGE_TAIL(i_edge);

      offset = GENERIC_SUB(BBL_CADDRESS(dest),GENERIC_ADD(INS_CADDRESS(last),INS_CSIZE(last)));
      I386_OP_IMMEDIATE(op) = G_T_UINT32(offset);
    }
  }
}
/* }}} */

#ifdef OBFUSCATION_SUPPORT
/* {{{ Recalculate the relocations */
static void CalcRelocs(t_object * obj)
{
  t_reloc * rel;
  t_uint32 code=0x12345678;

  OBJECT_FOREACH_RELOC(obj,rel)
  {
    if (RELOCATABLE_TYPE(REL_FROM(rel)) == RT_INS)
    {
      t_i386_ins * ins = T_I386_INS(REL_FROM(rel));

      /* skip relocs from jump and call instructions (these are handled
       * in AdjustJumpOffsets) */
      if (I386_INS_OPCODE(ins) == I386_JMP ||
	  I386_INS_OPCODE(ins) == I386_Jcc ||
	  I386_INS_OPCODE(ins) == I386_JECXZ ||
	  I386_INS_OPCODE(ins) == I386_CALL ||
	  I386_INS_OPCODE(ins) == I386_LOOP ||
	  I386_INS_OPCODE(ins) == I386_LOOPZ ||
	  I386_INS_OPCODE(ins) == I386_LOOP)
      {
	if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	  continue;
      }
      
      if (INS_TYPE(ins) == IT_DATA)
      {
	t_uint32 data;
	t_uint8 *buf = (t_uint8 *)&data;
	t_i386_ins * iter;

	//DiabloPrint(stdout, "relocing data in the code section at @G\n",INS_CADDRESS(ins));

	if (!GENERICISNULL(rel->from_offset)) FATAL(("Implement data with non-zero from_offset"));
	StackExec(rel->code,rel,buf,TRUE,0);

	/* TODO: this only works on little-endian machines */
	ins->data = buf[0];
	iter = (t_i386_ins *) INS_NEXT(ins);  iter->data = buf[1];
	iter = (t_i386_ins *) INS_NEXT(iter); iter->data = buf[2];
	iter = (t_i386_ins *) INS_NEXT(iter); iter->data = buf[3];
      }
      else
      {
	/* find out which operand has been relocated */
	t_address ret;
	t_i386_operand * op;
	t_reloc_ref * rr = RELOCATABLE_REFERS_TO(ins);
	t_reloc_ref * rr2;
	VERIFY_FASSERT(rr,("relocations corrupt"));
	VERIFY_FASSERT(rr->rel == rel || (rr->next && rr->next->rel == rel),("relocations corrupt")); 

	if (!rr->next)
	  rr2 = NULL;
	else if (rr->next && rr->rel == rel)
	  rr2 = rr->next;
	else if (rr->next && rr->next->rel == rel)
	{
	  rr2 = rr;
	  rr = rr->next;
	}
	else
	  FATAL(("Relocations corrupt"));

	if (rr2 && GENERIC_GT(rel->from_offset,rr2->rel->from_offset))
	  op = I386InsGetSecondRelocatedOp((t_ins *)ins);
	else
	  op = I386InsGetFirstRelocatedOp((t_ins *)ins);

	if((!strcmp(global_options.obfuscation,"branch_obf")) && ((INS_ATTRIB(ins)&IF_BRANCH_OBF)!=0))
	{
	  rel->addend=code;
	  ret = StackExec(rel->code,rel,NULL,FALSE,0);
	  // DiabloPrint(stdout,"XXins @I relocing 0x%x new @G\n",ins,I386_OP_IMMEDIATE(op),ret);
	  I386_OP_IMMEDIATE(op) = GENERIC_EXTRACT_UINT32(ret);
	}
	else
	{
	  ret = StackExec(rel->code,rel,NULL,FALSE,0);
	  // DiabloPrint(stdout,"ins @I relocing 0x%x new @G\n",ins,I386_OP_IMMEDIATE(op),ret);
	  I386_OP_IMMEDIATE(op) = GENERIC_EXTRACT_UINT32(ret);
	}
      }
    }
    else
    {
      //DiabloPrint(stdout,"relocing in section %s at @G old 0x%x new ",T_SECTION(REL_FROM(rel))->name,GENERIC_ADD(REL_FROM(rel)->caddress,rel->from_offset), *(t_uint32 *)(T_SECTION(REL_FROM(rel))->data + rel->from_offset));
      t_address ret = StackExec(rel->code,rel,T_SECTION(REL_FROM(rel))->data,TRUE,0);
      //DiabloPrint(stdout,"%x\n",*(t_uint32 *)(T_SECTION(REL_FROM(rel))->data + rel->from_offset));
      VERIFY_FASSERT(GENERICISNULL(ret),("Could not properly execute @R",rel));
    }
  }

  if(!strcmp(global_options.obfuscation,"branch_obf"))
  {
    DiabloPrint(stdout,"CODE:");
    PrintCode(code);
    DiabloPrint(stdout,"\n");
  }

}
/* }}} */
#else
/* {{{ Recalculate the relocations */
static void CalcRelocs(t_object * obj)
{
  t_reloc * rel;

  OBJECT_FOREACH_RELOC(obj,rel)
  {
    if (RELOCATABLE_TYPE(REL_FROM(rel)) == RT_INS)
    {
      t_i386_ins * ins = T_I386_INS(REL_FROM(rel));

      if (INS_TYPE(ins) == IT_DATA)
      {
	t_uint32 data;
	t_uint8 *buf = (t_uint8 *)&data;
	t_i386_ins * iter;

	/*DiabloPrint(stdout, "relocing data in the code section at @G\n",INS_CADDRESS(ins)); */

	if (!GENERICISNULL(rel->from_offset)) FATAL(("Implement data with non-zero from_offset"));
	StackExec(rel->code,rel,buf,TRUE,0);

	/* TODO: this only works on little-endian machines */
	ins->data = buf[0];
	iter = (t_i386_ins *) INS_NEXT(ins);  iter->data = buf[1];
	iter = (t_i386_ins *) INS_NEXT(iter); iter->data = buf[2];
	iter = (t_i386_ins *) INS_NEXT(iter); iter->data = buf[3];
      }
      else
      {
	t_address ret;
	t_i386_operand * op;
	t_reloc_ref * rr;
	t_reloc_ref * rr2;

	/* skip relocs from jump and call instructions (these are handled
	 * in AdjustJumpOffsets) */
	if (I386_INS_OPCODE(ins) == I386_JMP ||
	    I386_INS_OPCODE(ins) == I386_Jcc ||
	    I386_INS_OPCODE(ins) == I386_JECXZ ||
	    I386_INS_OPCODE(ins) == I386_CALL ||
	    I386_INS_OPCODE(ins) == I386_LOOP ||
	    I386_INS_OPCODE(ins) == I386_LOOPZ ||
	    I386_INS_OPCODE(ins) == I386_LOOP)
	{
	  if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	    continue;
	}

	/* find out which operand has been relocated */
	rr = RELOCATABLE_REFERS_TO(ins);
	VERIFY_FASSERT(rr,("relocations corrupt"));
	VERIFY_FASSERT(rr->rel == rel || (rr->next && rr->next->rel == rel),("relocations corrupt")); 

	if (rr->rel == rel) 
	  rr2 = NULL;
	else
	{
	  rr2 = rr;
	  rr = rr->next;
	}

	if (rr2 && GENERIC_GT(rel->from_offset,rr2->rel->from_offset))
	  op = I386InsGetSecondRelocatedOp((t_ins *)ins);
	else
	  op = I386InsGetFirstRelocatedOp((t_ins *)ins);

	ret = StackExec(rel->code,rel,NULL,FALSE,0);
	/*DiabloPrint(stdout,"ins @I relocing 0x%x new @G\n",ins,I386_OP_IMMEDIATE(op),ret); */
	I386_OP_IMMEDIATE(op) = GENERIC_EXTRACT_UINT32(ret);
      }
    }
    else
    {
      /*DiabloPrint(stdout,"relocing in section %s at @G old 0x%x new ",T_SECTION(REL_FROM(rel))->name,GENERIC_ADD(REL_FROM(rel)->caddress,rel->from_offset), *(t_uint32 *)(T_SECTION(REL_FROM(rel))->data + rel->from_offset)); */
      t_address ret = StackExec(rel->code,rel,T_SECTION(REL_FROM(rel))->data,TRUE,0);
      /*DiabloPrint(stdout,"%x\n",*(t_uint32 *)(T_SECTION(REL_FROM(rel))->data + rel->from_offset)); */
      VERIFY_FASSERT(GENERICISNULL(ret),("Could not properly execute @R",rel));
    }
  }
}
/* }}} */
#endif

/* {{{ Set the right instruction sizes */
static t_bool SetInsSizes(t_object * obj)
{
  int i;
  t_uint8 buf[15];
  t_bbl * bbl;
  t_ins * ins;
  t_bool change = FALSE;

  for (i = 0; i < obj->ncodes; i++)
  {
    for (bbl = T_BBL(obj->code[i]->data); bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    {
      BBL_FOREACH_INS(bbl,ins)
      {
	t_uint32 len = I386AssembleIns(ins,buf);
	if (len != G_T_UINT32(INS_CSIZE(ins)))
	{
	  change = TRUE;
	  GenericSet_uint32(INS_CSIZE(ins),len);
	}
      }
    }
  }
  return change;
}
/* }}} */

/* {{{ set the right immediate size for operands */
static void DoSetSize(t_ins * ins, t_i386_operand * op)
{
  static t_uint32 next_proposed[] = {1,4,4,4,1000};

  t_uint32 imm = I386_OP_IMMEDIATE(op);
  t_int32 simm = (t_int32) I386_OP_IMMEDIATE(op);
  t_uint32 orig_immedsize = I386_OP_IMMEDSIZE(op);
  t_uint32 proposed_immedsize = 0;
  t_i386_opcode_entry * forms[10];
  int nfitting;

  if (I386_OP_TYPE(op) != i386_optype_imm && I386_OP_TYPE(op) != i386_optype_mem)
    return;
  if (I386_OP_FLAGS(op) & I386_OPFLAG_ISRELOCATED)
    return;
  if (orig_immedsize == 2) 
    /* stay away from 16-bit immediates. they are complicated and not worth the hassle */
    return;

  if (imm == 0)
    proposed_immedsize = 0;
  else if (imm > 0 && imm < 256)
    proposed_immedsize = 1;
  else if (simm >= -128 && simm < 128)
    proposed_immedsize = 1;
  else
    proposed_immedsize = 4;

  while (proposed_immedsize <= 4)
  {
    I386_OP_IMMEDSIZE(op) = proposed_immedsize;
    nfitting = I386GetPossibleEncodings(ins,forms);
    if (nfitting) 
    {
      /*if (proposed_immedsize != orig_immedsize) DiabloPrint(stdout,"changed: @I from %d to %d\n",ins,orig_immedsize,proposed_immedsize);*/
      return; /* found a valid encoding */
    }
    proposed_immedsize = next_proposed[proposed_immedsize];
  }
  /* if we get here, we couldn't find a good encoding. emit a warning and restore the original immedsize */
  FATAL(("@I: found no good encoding",ins));
  I386_OP_IMMEDSIZE(op) = proposed_immedsize;
}

static void SetImmedSizes(t_object * obj)
{
  int i;
  t_bbl * bbl;
  t_ins * ins;
  /*static int teller = 0; */
  
#ifdef STEGO
  return;
#endif

  for  (i=0; i<obj->ncodes; i++)
    for (bbl = T_BBL(obj->code[i]->data); bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    {
      /*if (teller++ < global_options.debugcounter) */
      BBL_FOREACH_INS(bbl,ins)
      {
	/* don't do jumps and calls and stuff, they are done during the 
	 * deflowgraph loop */
	if (INS_TYPE(ins) == IT_BRANCH) continue;

	DoSetSize(ins,I386_INS_DEST(ins));
	DoSetSize(ins,I386_INS_SOURCE1(ins));
	DoSetSize(ins,I386_INS_SOURCE2(ins));
      }
    }
}
/* }}} */

#ifdef KERNEL_OPT
static void AlignPageDirectory(t_object * obj)
{
  t_symbol * sym = SymbolTableGetSymbolByName(obj->sub_symbol_table,"swapper_pg_dir");
  t_bbl * bbl, * fill;
  int diff,i;
  t_ins * ins;
  VERIFY_FASSERT(sym,("could not find sym swapper_pg_dir"));

  bbl = (t_bbl *)sym->base;
  diff = 0xc0101000 - G_T_UINT32(BBL_CADDRESS(bbl));
  fill = BBL_PREV_IN_CHAIN(bbl);
  if (BBL_INS_LAST(fill) &&
      I386_INS_OPCODE(BBL_INS_LAST(fill)) != I386_NOP)
  {
    /* no fill block yet. add one */
    fill = BblNew(BBL_CFG(bbl));
    BBL_PREV_IN_CHAIN(fill) = BBL_PREV_IN_CHAIN(bbl);
    BBL_NEXT_IN_CHAIN(fill) = bbl;
    BBL_NEXT_IN_CHAIN(BBL_PREV_IN_CHAIN(fill)) = fill;
    BBL_PREV_IN_CHAIN(bbl) = fill;
  }

  if (diff < 0)
  {
    VERIFY_FASSERT(BBL_NINS(fill) >= -diff,("cannot place swapper_page_dir"));
    for (i = diff; i < 0; i++)
      InsKill(BBL_INS_LAST(fill));
  }
  else
  {
    for (i = 0; i < diff; i++)
      I386MakeInsForBbl(Noop,Append,ins,fill);
  }
  
}
#endif

/* {{{ fixpoint that places sections, updates relocations and adjusts 
 * the operand sizes of jumps */
#define UPDATE_ADDRESSES	for (i=0; i<obj->ncodes; i++) AssignAddressesInChain(obj->code[i]->data,SECTION_GET_CADDRESS(obj->code[i]))
static void I386DeflowFixpoint(t_object * obj)
{
  t_bool loop = TRUE;
  int i;

  SetImmedSizes(obj);
  SetInsSizes(obj);
  KillUselessJumps(obj);
  UPDATE_ADDRESSES;

restart_loop:
  while (loop)
  {
    loop = FALSE;
    PlaceSections(obj,FALSE);
    UPDATE_ADDRESSES;
#ifdef KERNEL_OPT
    AlignPageDirectory(obj);
    UPDATE_ADDRESSES;
#endif
    if (AdjustJumpOffsetsUp(obj)) 
    {
      loop = TRUE;
      UPDATE_ADDRESSES;
    }
    if (AdjustJumpOffsetsDown(obj)) 
    {
      loop = TRUE;
      UPDATE_ADDRESSES;
    }
  }
  UpdateJumpDisplacements(obj);
  CalcRelocs(obj);
  /* there is a *slight* possibility that the recalculation of the relocations
   * causes the length of some instructions to change. if so, we have to start
   * it all over... */
  if (SetInsSizes(obj))
  {
    loop = TRUE;
    goto restart_loop;
  }
}
#undef UPDATE_ADDRESSES
/* }}} */

#ifdef KERNEL_OPT
/* {{{ deflowgraph algorithm for kernel optimization */
void I386DeflowgraphSection(t_object * obj, t_section * code)
{
  t_cfg * cfg = SECTION_CFG(code);
  t_bbl * i_bbl;
  t_chain_holder allchains, textchains, initchains;
  t_address split;
  t_section * text, * init;
  t_reloc * rel;
  int i;

  I386CreateChains(cfg,&allchains);

  /* {{{ split the chains between the .text and .text.init code sections */
  {
    t_uint32 ntext = 0, ninit = 0;
    t_uint32 texti, initi;

    /* decide which chain goes where */
    split = GENERIC_ADD(obj->orig_codes[0].base,obj->orig_codes[0].size);
    for (i = 0; i < allchains.nchains; i++)
    {
      t_bool in_init = TRUE;
      for (i_bbl = allchains.chains[i]; i_bbl; i_bbl = BBL_NEXT_IN_CHAIN(i_bbl))
	if (GENERIC_LT(BBL_OLD_ADDRESS(i_bbl),split))
	{
	  in_init = FALSE;
	  break;
	}
      if (in_init)
      {
	ninit++;
	/* hijack the prev_in_chain field of the chain head to indicate this chain belongs in the .text.init section */
	BBL_PREV_IN_CHAIN(allchains.chains[i]) = (t_bbl *)0x1;
      }
      else
      {
	ntext++;
      }
    }

    /* do the copying */
    textchains.chains = Malloc(sizeof(t_bbl *)*ntext);
    initchains.chains = Malloc(sizeof(t_bbl *)*ninit);
    textchains.nchains = ntext;
    initchains.nchains = ninit;
    texti = initi = 0;
    for (i = 0; i < allchains.nchains; i++)
    {
      if (BBL_PREV_IN_CHAIN(allchains.chains[i]))
      {
	/* .text.init section */
	initchains.chains[initi++] = allchains.chains[i];
	BBL_PREV_IN_CHAIN(allchains.chains[i]) = NULL;
      }
      else
      {
	/* .text section */
	textchains.chains[texti++] = allchains.chains[i];
      }
    }

    /* free the allchains list */
    Free(allchains.chains);
    allchains.chains = NULL;
  }
  /* }}} */

  /* {{{ report how many blocks that originally were in .text.init are now in .text */
  {
    t_uint32 nblocks;
    t_address lostsize = NULL_GENERIC;
    for (i=0; i<textchains.nchains; i++)
    {
      for (i_bbl=textchains.chains[i]; i_bbl; i_bbl=BBL_NEXT_IN_CHAIN(i_bbl))
      {
	if (GENERIC_GE(BBL_OLD_ADDRESS(i_bbl),split))
	{
	  nblocks++;
	  lostsize = GENERIC_ADD(lostsize,BBL_CSIZE(i_bbl));
	}
      }
    }
    DiabloPrint(stdout,"[Layout   ] %d blocks (@G bytes) moved from .text.init to .text\n",nblocks,lostsize);
  } /* }}} */

  I386OrderChains(&textchains);
  I386OrderChains(&initchains);

  text = CreateDeflowgraphedSection(obj,&textchains,".text");
  init = CreateDeflowgraphedSection(obj,&initchains,".text.init");

  /* {{{ move the TO_SECTION relocations pointing to the merged code section to the split sections */
  OBJECT_FOREACH_RELOC(obj,rel)
  {
    if (REL_TO_RELOCATABLE(rel) != (t_relocatable *)code) continue;
    DiabloPrint(stdout,"Migrating rel label %s",rel->label);
    if (GENERIC_GE(rel->to_offset,GENERIC_SUB(obj->orig_codes[1].base,obj->orig_codes[0].base)))
    {
      /* migrate to .text.init section */
      printf(" to .text.init\n");
      RelocSetToRelocatable(rel,(t_relocatable *)init);
      rel->to_offset = GENERIC_SUB(rel->to_offset,GENERIC_SUB(obj->orig_codes[1].base,obj->orig_codes[0].base));
    }
    else
    {
      /* migrate to .text section */
      printf(" to .text\n");
      RelocSetToRelocatable(rel,(t_relocatable *)text);
    }
  }
  /* }}} */

  /* kill the old merged code section */
  obj->code[0]->data = NULL;	/* prevent SectionFree from freeing the cfg. we still need it! */
  CFG_SECTION(cfg) = text;
  SectionFree(obj->code[0]);
  obj->ncodes = 2;
  Free(obj->code);
  obj->code = Malloc(sizeof(t_section *)*2);
  obj->code[0] = text;
  obj->code[1] = init;

  /* move the __init_begin relocations to the beginning of the .text.init section */
  OBJECT_FOREACH_RELOC(obj,rel)
  {
    if (!strcmp(rel->label,"__init_begin") || !strcmp(rel->label,"Copied__init_begin"))
    {
      DiabloPrint(stdout,"moving rel label %s code %s base @G offset @G addend @G\n",rel->label,rel->code,REL_TO_RELOCATABLE(rel)->caddress,rel->to_offset,rel->addend);
      REL_TO_RELOCATABLE(rel) = init;
      rel->to_offset = NULL_GENERIC;
    }
  }

  I386DeflowFixpoint(obj);
  I386SectionFromDeflowgraphedToDisassembled(obj);

  /* {{{ calculate gain */
  {
    t_uint32 ns, os;
    ns = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(text));
    os = GENERIC_EXTRACT_UINT32(obj->orig_codes[0].size);
    printf("ns = 0x%x, os = 0x%x, ns/os = %lf\n",ns,os,((double)ns)/((double)os));
    printf("GAIN in .text: %lf%%\n",100.0*(1.0-((double)ns)/((double)os)));
    ns = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(init));
    os = GENERIC_EXTRACT_UINT32(obj->orig_codes[1].size);
    printf("GAIN in .text.init: %lf%%\n",100.0*(1.0-((double)ns)/((double)os)));
  }
  /* }}}*/

  /* {{{ debug code: list the final program */
  {
    char * filename = Collate(global_options.output_name,".list");
    FILE * f = fopen(filename,"w");
    if (f)
    {
      t_ins * ins;
      for (ins=obj->code[0]->data; ins; ins=INS_NEXT(ins))
	DiabloPrint(f,"@I\n",ins);
      DiabloPrint(f,"==============================[.text.init]==============================\n");
      for (ins=obj->code[1]->data; ins; ins=INS_NEXT(ins))
	DiabloPrint(f,"@I\n",ins);
      fclose(f);
    }
    else
      VERBOSE(0,("Could not open %s for writing!",filename));
    Free(filename);
  } /* }}} */

  obj->entry=INS_CADDRESS(BBL_INS_FIRST(cfg->entries->entry_bbl));

  CFG_SECTION(cfg) = NULL; 	/* prevent CfgFree from freeing text->data, which we still need */
  CfgFree(cfg);

  Free(textchains.chains);
  Free(initchains.chains);
} /* }}} */
#else
/* {{{ regular deflowgraph algorithm */
void I386DeflowgraphSection(t_object * obj, t_section * code)
{
  t_cfg * cfg = SECTION_CFG(code);
  t_chain_holder chains;

  I386CreateChains(cfg,&chains);
#ifndef STEGO
  I386OrderChains(&chains);
#else
  StegoOrderChains(&chains,NULL,obj);
#endif


  code->type = DEFLOWGRAPHED_CODE_SECTION;
  code->data = chains.chains[0];

  I386DeflowFixpoint(obj);

#ifdef OBFUSCATION_SUPPORT
  if(global_options.obf_dis)
    I386FillOpcodesObf(obj);
#endif

  DiabloPrint(stdout,"old object entry was @G\n",obj->entry);
  obj->entry=INS_CADDRESS(BBL_INS_FIRST(cfg->entries->entry_bbl));
  DiabloPrint(stdout,"new object entry is @G\n", obj->entry);

#ifdef INSTRUMENTATION
  if (use_old_addresses)
  {
    int count = 0;
    t_bbl * bbl;
    CFG_FOREACH_BBL(cfg,bbl)
      if (BBL_ATTRIB(bbl) & BBL_IS_POSSIBLE_TARGET_OF_INDIRECT_CONTROL_FLOW)
	count++;
    bbl_translate_list = Malloc((2*count+1)*sizeof(t_address));
    GenericSet_uint32(bbl_translate_list[0],count);
    bbl_translate_list++;
    count = 0;
    CFG_FOREACH_BBL(cfg,bbl)
      if (BBL_ATTRIB(bbl) & BBL_IS_POSSIBLE_TARGET_OF_INDIRECT_CONTROL_FLOW)
      {
	bbl_translate_list[2*count] = BBL_OLD_ADDRESS(bbl);
	bbl_translate_list[2*count+1] = BBL_CADDRESS(bbl);
	count++;
      }
    bbl_translate_list--;
  }
#endif

  I386SectionFromDeflowgraphedToDisassembled(obj);

#ifndef INSTRUMENTATION
  /* {{{ debug code: list the final program */
  {
    char * filename = Collate(global_options.output_name,".list");
    FILE * f = fopen(filename,"w");
    if (f)
    {
      t_ins * ins;
      for (ins=obj->code[0]->data; ins; ins=INS_NEXT(ins))
	DiabloPrint(f,"@I\n",ins);
      fclose(f);
    }
    else
      VERBOSE(0,("Could not open %s for writing!",filename));
    Free(filename);
  } /* }}} */
#endif

  /* {{{ compute code size gain */
  {
    double gain, os, ns;
    int i;
    t_address osize, nsize;

    osize = nsize = NULL_GENERIC;
    for (i=0; i < obj->ncodes; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->code[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->code[i]));
    }
    DiabloPrint(stdout,"Code section old size @G new size @G\n",osize,nsize);
    os = (double) G_T_UINT32(osize);
    ns = (double) G_T_UINT32(nsize);
    gain = (os - ns)/os;
    printf("GAIN %lf%% in code section\n",100.0 * gain);
    for (i=0; i < obj->nrodatas; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->rodata[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->rodata[i]));
    }
    for (i=0; i < obj->ndatas; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->data[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->data[i]));
    }
    for (i=0; i < obj->nbsss; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->bss[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->bss[i]));
    }
    os = (double) G_T_UINT32(osize);
    ns = (double) G_T_UINT32(nsize);
    gain = (os - ns)/os;
    printf("GAIN %lf%% total\n",100.0 * gain);
  } /* }}} */

  CFG_SECTION(cfg) = NULL; 	/* prevent CfgFree from freeing code->data, which we still need */
  CfgFree(cfg);

  Free(chains.chains);
}
/* }}} */
#endif


/* vim: set shiftwidth=2 foldmethod=marker: */
