/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "i386_description.h"
#include "i386_instructions.h"
#include "i386_inline.h"
#include "i386_peephole.h"
#include "i386_stack.h"
#include "../../kernel/diablo_opt_misc.h"

#ifdef INSTRUMENTATION
#include "../../atom_internal.h"
#include "i386_atom.h"
#endif

static char * names[] =
{
  "EAX", "EBX", "ECX", "EDX", "ESI", "EDI", "EBP", "ESP",                   /* GP registers */
  "ST0", "ST1", "ST2", "ST3", "ST4", "ST5", "ST6", "ST7",                   /* FP registers */
  "CS" , "DS" , "ES" , "FS" , "GS" , "SS",                                  /* segment registers */
  "CR0", "CR1", "CR2", "CR3", "CR4",					    /* control registers */
  "DR0", "DR1", "DR2", "DR3", "DR4", "DR5", "DR6", "DR7", 		    /* debug registers */
  "AF" , "CF" , "DF" , "IF" , "NT" , "OF" , "PF" , "RF" , "SF", "TF", "ZF", /* eflags       */
  "C0" , "C1" , "C2" , "C3"                                                 /* FP condition */
};

static t_arch_specific_optimization i386_specific_opts[] = {
  {"MergeBbls",MergeBbls},
  {"BranchElimination",BranchElimination},
  {"I386InlineSimple",I386InlineSmallFunctions},
  {"I386InlineFunctionsWithOneCallSite",I386InlineFunctionsWithOneCallSite},
  {"I386PeepHoles",I386PeepHoles},
  {"I386StackHeightAnalysis",I386StackHeightAnalysis},
  {"I386PeepHolesInstrumentation",I386PeepHolesInstrumentation},
  { NULL,NULL}
};

#ifdef INSTRUMENTATION
static t_atom_actions i386_atom_actions = {
  I386AtomInit,
  I386AtomEstimateChangedRegisters,
  I386AtomAddSave,
  I386AtomAddRestore,
  I386AtomAddParams,
  I386AtomAddKillParams,
  I386AtomAddCheckForExitCall,
  I386AtomAddCall,
  I386IsInstType
};
#endif

const t_architecture_description i386_description =
{
   8, /* encoded minimal size [1 byte] */
   120, /* encoded maximum size [15 byte] */
   8, /* mod size */
   1, /* bundel size */
   0, /* bundel template size */
   sizeof(t_i386_ins), /* size of an instruction */
   ADDRSIZE32,
   8,  /* int regs */
   8,  /* float regs */
   0,  /* predicate regs */
   0,  /* branch regs */
   34, /* special regs */
   names,
#if MAX_REG_ITERATOR > 64

#ifndef BIT64ADDRSUPPORT
/* all registers            */   {MAX_REG_ITERATOR, {0xffff,0xffff,0xffff,0x0003}},
/* int registers            */   {MAX_REG_ITERATOR, {0x00ff}},
/* float registers          */   {MAX_REG_ITERATOR, {0xff00}},
/* predicate registers      */   {MAX_REG_ITERATOR, {0x0000,0x0000,0xfff8,0x0003}},
/* callee saved             */   {MAX_REG_ITERATOR, {0x0072}},
/* callee may use           */   {MAX_REG_ITERATOR, {0x0080,0xffff,0x0000,0x0}},
/* callee may change        */   {MAX_REG_ITERATOR, {0xff8d,0x0000,0xfff8,0x0003}},
/* callee may return        */   {MAX_REG_ITERATOR, {0x0109}},
/* always live              */   {MAX_REG_ITERATOR, {0x0080}},
/* registers prop over hell */   {MAX_REG_ITERATOR, {0x0000}},
/* const registers          */   {MAX_REG_ITERATOR, {0x0000}},
/* dead over call           */   {MAX_REG_ITERATOR, {0x0000,0x0000,0xfff8,0x0003}},
#else
#error The regsets on the calling conventions are incorrect, see below
/* all registers            */   {MAX_REG_ITERATOR, {0xffffULL,0xffffULL,0xffffULL,0x0003ULL}},
/* int registers            */   {MAX_REG_ITERATOR, {0x00ffULL}},
/* float registers          */   {MAX_REG_ITERATOR, {0xff00ULL}},
/* predicate registers      */   {MAX_REG_ITERATOR, {0x0000ULL,0x0000ULL,0xfff0ULL,0x0003ULL}},
/* callee saved             */   {MAX_REG_ITERATOR, {0x0072ULL}},
/* callee may use           */   {MAX_REG_ITERATOR, {0xffffULL,0xffffULL,0xffffULL,0x0003ULL}},
/* callee may change        */   {MAX_REG_ITERATOR, {0xffffULL,0xffffULL,0xffffULL,0x0003ULL}},
/* callee may return        */   {MAX_REG_ITERATOR, {0x0000ULL}},
/* always live              */   {MAX_REG_ITERATOR, {0x0000ULL}},
/* registers prop over hell */   {MAX_REG_ITERATOR, {0x0000ULL}},
/* const registers          */   {MAX_REG_ITERATOR, {0x0000ULL}},
/* dead over call           */#endif

#elif MAX_REG_ITERATOR > 32
/* all registers            */   0x3ffffffffffffLL,
/* int registers            */   0x00000000000ffLL,
/* float registers          */   0x000000000ff00LL,
/* predicate registers      */   0x3fff800000000LL,
/* callee saved             */   0x0000000000072LL,
/* callee may use           */   0x00000ffff0080LL,
/* callee may change        */   0x3fff80000ff8dLL, /* callee may change the stack pointer: see ret $0x... instruction */
/* callee may return        */   0x0000000000109LL,
/* always live              */   0x0000000000080LL, /* stack pointer is always live */
/* registers prop over hell */   0x0LL,
/* const registers          */   0x0LL,
/* dead over call           */   0x3fff800000000LL,
#else
#error   i386 has more than 32 registers!!!
#endif
   REG_NONE,
   i386_specific_opts,
#ifdef INSTRUMENTATION
   &i386_atom_actions,
#endif
   NULL,   /* An array with names of unbehaved functions */
   NULL,
   I386InsIsProcedureCall,
   NullRegs,
   NULL
};
/* vim: set shiftwidth=2: */





