/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "../../utils/util_error.h"
#include "../../kernel/diablo_iterators.h"
#include "i386_instructions.h"
#include "i386_description.h"
#include "i386_registers.h"
#include "i386_opcode.h"
#include "i386_table.h"
/*#ifdef WATERMARK
#include "../../diablo_options.h"
#include "../../watermark/water_hydan.h"
#include <string.h>
#endif
*/

/** {{{ Parse instruction prefixes */
static t_uint32 HandlePrefixes(t_i386_ins * ins, char * codep)
{
  char * run = codep;
  I386_INS_PREFIXES(ins) = 0;
  while (1)
  {
    switch (*run)
    {
      case '\xf0':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_LOCK;
	break;

      case '\xf2':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_REPNZ;
	break;
      case '\xf3':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_REP;
	break;

      case '\x2e':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_CS_OVERRIDE;
	break;
      case '\x36':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_SS_OVERRIDE;
	break;
      case '\x3e':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_DS_OVERRIDE;
	break;
      case '\x26':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_ES_OVERRIDE;
	break;
      case '\x64':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_FS_OVERRIDE;
	break;
      case '\x65':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_GS_OVERRIDE;
	break;

      case '\x66':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_OPSIZE_OVERRIDE;
	break;
      case '\x67':
	I386_INS_PREFIXES(ins) |= I386_PREFIX_ADDRSIZE_OVERRIDE;
	break;

      default:
	return (run - codep);
    }
    run++;
  }

  /* keep the compiler happy */
  return 0;
} /* }}} */

/** {{{ get opcode and ModR/M byte for floating point instructions */
static t_uint32 DoFloat(t_i386_ins * ins, char * codep, t_uint32 * modrm_ret, t_uint32 * sib_ret, t_i386_opcode_entry ** disentry_ret)
{
  t_uint8 * run = (t_uint8 *) codep;
  t_uint32 modrm, sib;
  t_uint32 ext;
  t_i386_opcode opc;

  modrm = run[1];
  ext = (modrm >> 3) & 7;

  if ((modrm >> 6) == 3)
  {
    /* register operations */
    opc = dis386_fpu_reg[*run - 0xd8][ext][modrm & 7].opcode;
    *disentry_ret = &(dis386_fpu_reg[*run - 0xd8][ext][modrm & 7]);
    sib = -1;
  }
  else
  {
    /* memory operations */
    opc = dis386_fpu_mem[*run - 0xd8][ext].opcode;
   *disentry_ret = &(dis386_fpu_mem[*run - 0xd8][ext]);
    if ((modrm & 7) == 4)
      sib = *(run + 2);
    else
      sib = -1;
  }

  I386_INS_OPCODE(ins) = opc;
  *modrm_ret = modrm;
  *sib_ret = sib;
  
  if (sib == -1)
    return 2;
  else
    return 3;
} /* }}} */

/** {{{ Get opcode, ModR/M and SIB byte for a non-fpu instruction */
static t_uint32 GetOpcodeModRMAndSIB(t_i386_ins * ins, char * codep, t_uint32 * modrm_ret, t_uint32 * sib_ret, t_i386_opcode_entry ** disentry_ret)
{
  t_uint8 * run = (t_uint8 *)codep;
  t_i386_opcode opc = dis386[*run].opcode;
  t_bool has_modrm = dis386[*run].has_modrm;
  t_uint32 modrm = -1, sib = -1;

#ifdef STEGO
  ins->opcode_byte = *codep;
#endif
  
  *disentry_ret = &(dis386[*run]);

  run++;

  /* floating point instructions are somewhat more complex. they are handled
   * in a separate function */
  if (opc == FPU_ESC)
    return DoFloat(ins,codep,modrm_ret,sib_ret,disentry_ret);

  if (opc == TWOBYTE_ESC)
  {
    opc = dis386_twobyte[*run].opcode;
    has_modrm = dis386_twobyte[*run].has_modrm;
    *disentry_ret = &(dis386_twobyte[*run]);
    run++;
  }

  if (has_modrm)
  {
    modrm = (t_uint32) *run;
    run++;
  }

  if (opc < 0)
  {
    /* we haven't determined a precise opcode yet.
     * we'll have to look at the opcode extension bits in the modrm byte */
    t_uint32 ext = (modrm & 0x38) >> 3;

    /* sanity check: if there is no modrm byte something is definitely wrong */
    VERIFY_FASSERT(has_modrm,("Disassembly tables corrupt!"));

    *disentry_ret = &(dis386_grps[opc - GRP1a][ext]);
    opc = dis386_grps[opc - GRP1a][ext].opcode;
  }

  /* if mod != 11b && rm == 100b && addressing mode is 32-bit, we have an sib
   * byte */
  if (has_modrm && !I386_ADSZPREF(ins) && ((modrm >> 6) != 3) && ((modrm & 7) == 4))
  {
    sib = (t_uint32) *run;
    run++;
  }

  I386_INS_OPCODE(ins) = opc;
  *modrm_ret = modrm;
  *sib_ret = sib;

  return (run - ((t_uint8 *)codep));
} /* }}} */

/** {{{ set condition field for conditional instructions */
static void SetConditionField(t_i386_ins * ins,t_uint8 * orig_opcode)
{
  switch (I386_INS_OPCODE(ins))
  {
    case I386_Jcc:
    case I386_SETcc:
    case I386_CMOVcc:
      if (orig_opcode[0] == 0x0f)
      {
	/* two-byte opcode: four last bits of second opcode byte */
	I386_INS_CONDITION(ins) = ((t_uint32) orig_opcode[1]) & 0xf;
      }
      else
      {
         /* one-byte opcode: four last bits of the first opcode byte */
	  I386_INS_CONDITION(ins) = ((t_uint32) orig_opcode[0]) & 0xf;
      }
      break;

    case I386_FCMOVcc:
      {
	t_uint32 selector = ((orig_opcode[0] & 0x1) << 2) + ((orig_opcode[1] & 0x18) >> 3);
	switch (selector)
	{
	  case 0:
	    I386_INS_CONDITION(ins) = I386_CONDITION_B;
	    break;
	  case 1:
	    I386_INS_CONDITION(ins) = I386_CONDITION_Z;
	    break;
	  case 2:
	    I386_INS_CONDITION(ins) = I386_CONDITION_BE;
	    break;
	  case 3:
	    I386_INS_CONDITION(ins) = I386_CONDITION_P;
	    break;
	  case 4:
	    I386_INS_CONDITION(ins) = I386_CONDITION_AE;
	    break;
	  case 5:
	    I386_INS_CONDITION(ins) = I386_CONDITION_NZ;
	    break;
	  case 6:
	    I386_INS_CONDITION(ins) = I386_CONDITION_A;
	      ;
	    break;
	  case 7:
	    I386_INS_CONDITION(ins) = I386_CONDITION_NP;
	    break;
	}
      }
      break;
    case I386_LOOPNZ:
      I386_INS_CONDITION(ins) = I386_CONDITION_LOOPNZ;
      break;
    case I386_LOOPZ:
      I386_INS_CONDITION(ins) = I386_CONDITION_LOOPZ;
      break;
    case I386_LOOP:
      I386_INS_CONDITION(ins) = I386_CONDITION_LOOP;
      break;
    case I386_JECXZ:
      I386_INS_CONDITION(ins) = I386_CONDITION_ECXZ;
      break;
    default:
      I386_INS_CONDITION(ins) = I386_CONDITION_NONE;
  }

  if (I386_INS_CONDITION(ins) != I386_CONDITION_NONE)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;

} /* }}} */

/* {{{ set the instruction type field */
static void SetInsType(t_i386_ins * ins)
{
  INS_TYPE(ins) = i386_opcode_table[I386_INS_OPCODE(ins)].type;
} /* }}} */

/** {{{ helper function: sort the refers_to of a relocatable on increasing offset */
static int __rel_cmp(const void * a, const void * b)
{
  t_reloc * rela = *(t_reloc **)a;
  t_reloc * relb = *(t_reloc **)b;
  t_address addra = GENERIC_ADD(REL_FROM(rela)->caddress, rela->from_offset);
  t_address addrb = GENERIC_ADD(REL_FROM(relb)->caddress, relb->from_offset);
  return GENERIC_EXTRACT_INT32(GENERIC_SUB(addra,addrb));
}

static t_reloc ** SortRefersTo(t_section * code, int * nrels_ret)
{
  int nrels,i;
  t_reloc ** rels;
  t_reloc * rel;

  nrels = 0;
  OBJECT_FOREACH_RELOC(code->obj,rel)
    if ((RELOCATABLE_TYPE(REL_FROM(rel)) == RT_SUBSECTION) && (SECTION_PARENT_SECTION(REL_FROM(rel)) == code))
      nrels++;
  
  rels = Malloc(nrels*sizeof(t_reloc *));
  i = 0;
  OBJECT_FOREACH_RELOC(code->obj,rel)
    if ((RELOCATABLE_TYPE(REL_FROM(rel)) == RT_SUBSECTION) && (SECTION_PARENT_SECTION(REL_FROM(rel)) == code))
      rels[i++] = rel;

  qsort(rels,nrels,sizeof(t_reloc *),__rel_cmp);

  *nrels_ret = nrels;
  return rels;
}
/* }}} */

/** {{{ disassemble one  instruction */
#define UpdateCurrentAndRun()   run += len; current = GENERIC_ADD_UINT32(current,len)
#define CheckOpRelocated(op)    if ((curr_rel < max_rel) && GENERIC_EQ(current,GENERIC_ADD(rels[curr_rel]->from_offset,REL_FROM(rels[curr_rel])->caddress)) && (len > 0)) \
                                { \
				  I386_OP_FLAGS(op) |= I386_OPFLAG_ISRELOCATED; \
				  curr_rel++; \
				}
#define CheckNoRelocsBefore()   if ((curr_rel < max_rel) && GENERIC_GT(current,GENERIC_ADD(rels[curr_rel]->from_offset,REL_FROM(rels[curr_rel])->caddress))) \
                                  FATAL(("Relocs in the non-operand parts of instructions: @R from offset @G base @G current @G",rels[curr_rel], \
					 rels[curr_rel]->from_offset, REL_FROM(rels[curr_rel])->caddress,current));

/*#ifdef WATERMARK
static t_uint32 bitcounter=0;
static char charakter;
static t_bool markfinished;
#endif
*/
t_uint32 I386DisassembleOne(t_i386_ins * ins, char * codep, t_reloc ** rels, int max_rel, int curr_rel)
{
  t_uint32 len,prefix_len;
  char * run;
  t_uint32 modrm = -1, sib = -1;
  t_i386_opcode_entry * disentry;
  t_address current;
/*#ifdef WATERMARK
  t_uint8 orig_opcode = *codep;
#endif
*/
  run = codep;
  current = INS_CADDRESS(ins);

  /* handle instruction prefixes */
  len = prefix_len = HandlePrefixes(ins,run);
  UpdateCurrentAndRun();
  
  /* determine opcode bytes + modr/m and sib bytes */
  len = GetOpcodeModRMAndSIB(ins,run,&modrm,&sib,&disentry);
  UpdateCurrentAndRun();

  /* disassemble the instruction operands */
  CheckNoRelocsBefore();
  /* 1. the destination operand */
  len = disentry->op1dis(ins,run,modrm,sib,I386_INS_DEST(ins),disentry->op1bm);
  CheckOpRelocated(I386_INS_DEST(ins));
  UpdateCurrentAndRun();
  if (disentry->usedefpattern & DU)
    I386_INS_FLAGS(ins) |= I386_IF_DEST_IS_SOURCE;

  /* 2. the first source operand */
  len = disentry->op2dis(ins,run,modrm,sib,I386_INS_SOURCE1(ins),disentry->op2bm);
  CheckOpRelocated(I386_INS_SOURCE1(ins));
  UpdateCurrentAndRun();
  if (disentry->usedefpattern & S1D)
    I386_INS_FLAGS(ins) |= I386_IF_SOURCE1_DEF;

  /* 3. the second source operand */
  len = disentry->op3dis(ins,run,modrm,sib,I386_INS_SOURCE2(ins),disentry->op3bm);
  CheckOpRelocated(I386_INS_SOURCE2(ins));
  UpdateCurrentAndRun();
  if (disentry->usedefpattern & S2D)
    I386_INS_FLAGS(ins) |= I386_IF_SOURCE2_DEF;
  
  SetConditionField(ins,(t_uint8 *)codep+prefix_len);
  SetInsType(ins);
  INS_REGS_USE(ins) = I386InsUsedRegisters((t_ins*)ins);
  INS_REGS_DEF(ins) = I386InsDefinedRegisters((t_ins*)ins);

  if( I386_INS_OPCODE(ins) != I386_XOR )
    INS_ATTRIB(ins) |= IF_FAST_CP_EVAL;
  
/*#ifdef WATERMARK
  if(!strcmp(global_options.watermark,"yellow"))
  {
    int i;
    char read;
    int nr_possible;

    if((i=WaterHydanInsIsReplacable(T_INS(ins)))!=-1 && !markfinished)
    {
      read = WaterHydanGetIndex(T_INS(ins),i,orig_opcode,&nr_possible);
      charakter += (read<<bitcounter);
      bitcounter+=nr_possible;
      if(bitcounter >= 8){
	
	printf("%c",charakter);
	if(charakter==0){
	  printf("\n");
	  markfinished=TRUE;
	}
	bitcounter = bitcounter - 8;
	charakter = (read >> (nr_possible - bitcounter));
      }
    }
  }
#endif
*/  
  return (run - codep);
}
#undef UpdateCurrentAndRun
#undef CheckOpRelocated 
#undef CheckNoRelocsBefore
/* }}} */

/** {{{ helper function: determine old address from current address */
static t_address ComputeOldAddress(t_object * obj, t_address current)
{
  int i;
  for (i = 0; i < obj->n_orig_codes; i++)
  {
    if (GENERIC_LE(obj->orig_codes[i].base, current) && GENERIC_GT(GENERIC_ADD(obj->orig_codes[i].base, obj->orig_codes[i].size),current))
    {
#ifdef INSTRUMENTATION
      if (G_T_UINT32(obj->orig_codes[i].obase) == 0x0U)
      {
	t_address ret;
	GenericSet_uint32(ret,0xffffffffU);
	return ret;
      }
#endif
      return GENERIC_ADD(obj->orig_codes[i].obase, GENERIC_SUB(current,obj->orig_codes[i].base));
    }
  }
  /* if we get here, current belongs to none of the original code sections. just return NULL_GENERIC */
  return NULL_GENERIC;
}
/* }}} */

/** {{{ disassemble a code section */
void I386DisassembleSection(t_object * obj, t_section * code)
{
  t_uint8 * buf = code->data;
  t_uint8 * run = buf;
  t_uint32 inslen;
  t_address old, current, offset;
  t_i386_ins *ins, *prev;
  t_reloc ** sorted_relocs;
  int nrels, relindex;

  I386InitOpcodeTable();
  I386CreateOpcodeHashTable();
  sorted_relocs = SortRefersTo(code, &nrels);

  old = SECTION_GET_OLD_ADDRESS(code);
  current = SECTION_GET_CADDRESS(code);

  prev = NULL;
  GenericSet_uint32(offset,0);
  relindex = 0;
  while (GENERIC_LT(offset, SECTION_CSIZE(code)))
  {
    /* let relindex point to the first relocation at or beyond the current address */
    while ((relindex < nrels) && GENERIC_LT(GENERIC_ADD(sorted_relocs[relindex]->from_offset,REL_FROM(sorted_relocs[relindex])->caddress),current))
      relindex++;

    ins = InsNewForSec(code);
    INS_CADDRESS(ins) = current;
    INS_OLD_ADDRESS(ins) = ComputeOldAddress(obj,current);

    if (SymbolTableGetDataType(obj, current) == ADDRESS_IS_DATA)
    {
      INS_TYPE(ins) = IT_DATA;
      INS_ATTRIB(ins) = IF_DATA;
      I386_INS_OPCODE(ins) = I386_DATA;
      I386_INS_CONDITION(ins) = I386_CONDITION_NONE;
              
      GenericSet_uint32(INS_CSIZE(ins),1);
      GenericSet_uint32(INS_OLD_SIZE(ins),1);

      ins->data = (t_uint8) (*run);
    }
    else
    {
      t_i386_ins * dummy, * pdummy;
      int i;

      inslen = I386DisassembleOne(ins,run,sorted_relocs,nrels,relindex);
      VERIFY_FASSERT(inslen,("Disassembly of instruction at @G (old address) failed!",current));

      GenericSet_uint32(INS_CSIZE(ins),inslen);
      GenericSet_uint32(INS_OLD_SIZE(ins),inslen);

      /* add dummy instructions to the list, otherwise SecGetInsByAddress
       * won't work */
      pdummy = ins;
      for (i = 1; i < inslen; i++)
      {
	dummy = InsNewForSec(code);
	INS_ATTRIB(dummy) = IF_DUMMY;
	INS_CADDRESS(dummy) = GENERIC_ADD_UINT32(INS_CADDRESS(ins),i);
	INS_OLD_ADDRESS(dummy) = GENERIC_ADD_UINT32(INS_OLD_ADDRESS(ins),i);
	INS_NEXT(pdummy) = (t_ins *) dummy;
	INS_PREV(dummy) = (t_ins *) pdummy;
	pdummy = dummy;
      }
    }

    run=GENERIC_ADD_DISPL(run,INS_CSIZE(ins));
    offset=GENERIC_ADD(offset,INS_CSIZE(ins));
    current=GENERIC_ADD(current,INS_CSIZE(ins));
    old=GENERIC_ADD(old,INS_OLD_SIZE(ins));

    if (prev) INS_NEXT(prev) = (t_ins *)ins;
    INS_PREV(ins) = (t_ins *)prev;
    prev = ins;
    while (INS_NEXT(prev))((t_ins *)prev) = INS_NEXT(prev); /* make prev point to the last dummy */
  }
  Free(sorted_relocs);
  
}
/* }}} */

/* vim: set shiftwidth=2 foldmethod=marker: */
