/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_iterators.h"
#include "i386_flowgraph.h"
#include "i386_instructions.h"
#include "i386_registers.h"
#include <string.h>

void I386FlowgraphSection(t_object * obj, t_section * section) 
{    
  t_uint32 ret;

#ifdef KERNEL_OPT
  /* for the linux kernel, the entry point is the start of the 
   * .text section, not the entry point advertised in the elf header
   */
  obj->entry = SECTION_GET_CADDRESS(obj->code[0]);
#endif

  /* Find the leaders in the instruction list */
  ANALYSIS_START("Leader detection");
  ret=I386FindBBLLeaders(obj,section);
  ANALYSIS_END("Leader detection",("%d leaders",ret));

  /* Create the basic blocks (platform independent) */
  ANALYSIS_START("Creating Basic Blocks");
  ret=CfgCreateBasicBlocks(section);
  ANALYSIS_END("Creating Basic Blocks",("%d blocks",ret));

  /* Create the edges between basic blocks */
  ANALYSIS_START("Creating Basic Block graph");
  ret=I386AddBasicBlockEdges(obj,section);
  ANALYSIS_END("Creating Basic Block graph",("%d edges added",ret));

  /* {{{ kill all relocations from the immediate operands of jump and call instructions: 
   * they have no use any more because their information is represented by flow graph edges */
  {
    t_cfg * cfg = SECTION_CFG(section);
    t_bbl * bbl;
    t_ins * ins;
    CFG_FOREACH_BBL(cfg,bbl)
    {
      BBL_FOREACH_INS(bbl,ins)
      {
	if (I386_INS_OPCODE(ins) == I386_CALL || I386_INS_OPCODE(ins) == I386_JMP || I386_INS_OPCODE(ins) == I386_Jcc)
	{
	  if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	  {
	    while (RELOCATABLE_REFERS_TO(ins))
	      RelocTableRemoveReloc(obj->reloc_table,RELOCATABLE_REFERS_TO(ins)->rel);
	    I386_OP_FLAGS(I386_INS_SOURCE1(ins)) &= ~I386_OPFLAG_ISRELOCATED;
	  }
	}
      }
    }
  } /* }}} */
}

/* {{{ I386AddBasicBlockEdges */
t_uint32 I386AddBasicBlockEdges(t_object * obj, t_section * section)
{
  /* Basic Block Edges
   * -----------------
   *
   * Add edges between the newly created basic blocks. Both the lineair (array)
   * representation of the instructions and the basic block representation are
   * available: you can access the (copy) basic block by using INS_BBL of a
   * leader in the lineair representation.
   *
   * Although connecting the blocks isn't that hard, it could be difficult to
   * find out if a given jump instructions is a call or just a jump (without
   * return). Getting this completly safe is probably not what you want: if you
   * implement this completly safe, you'll probably end up with one big
   * function, which will degrade the performance of the rest of diablo or hell
   * nodes after each jump which will degrade the performance even further.
   *
   * On the other hand, if you get this wrong (e.g add a jump edge where there
   * should have been a call edge), then chances are that your flowgraph (and
   * thus the produced binary) is wrong (the successor block of the call block
   * will have no incoming (return) edge, and thus this block is dead). 
   *
   * The solution is to look at code that is produced by your compiler(s).
   * It'll probably have only a few possibilties to produce calls. If you
   * handle them as safe as possible, than you're off the hook.  */

  t_ins *ins, *iter,  *block_end, *block_start, *prev, *prev_controltransfer;
  t_bbl * block, * ft_block;
  t_cfg * flowgraph=SECTION_CFG(section);
  t_reloc_ref * ref;
  t_uint32 nedges=0; 

#ifdef STEGO
  if(!global_options.stegoDecode){
#endif
  /* {{{ first detect switch tables and add the necessary switch edges.
   * we don't have to do this, but otherwise a lot of hell edges will
   * be added that substantially degrade the accuracy and performance of a 
   * number of analyses */
  for (prev = prev_controltransfer = NULL, ins = section->data; ins; ins = INS_NEXT(ins))
  {
    /* switch constructs have the following pattern:
     *   cmp $ubound, %reg
     *   ...
     *   ja <default_case>
     *   ...
     *   jmp *<jumptable_base>(,%reg,4)
     */

    t_i386_operand * op = I386_INS_SOURCE1(ins);
    t_reg jumpreg;
    t_reloc * jumptablerel;
    t_uint32 bound;
    t_regset changed;
    t_ins * cmp;
    t_reloc_ref * rr;
    //Xiangyu
    t_ins * jmp_def=NULL;
    t_i386_operand * jmp_def_op;
    t_reg tmp_reg;
    t_ins * tmp_ins;
    int flaged=0;

    //printf("%x\n",BBL_CADDRESS(ins));
    if (BBL_CADDRESS(ins)==0x805e963) flaged=1;
    //if (BBL_CADDRESS(ins)==0x8048221) flaged=1;
    if (flaged) printf ("INNNNNN\n"); 
    if (INS_PREV(ins) && !(INS_ATTRIB(INS_PREV(ins)) & IF_DUMMY))
      prev = INS_PREV(ins);
    if (prev && I386InsIsControlTransfer(prev))
      prev_controltransfer = prev;

    if (INS_ATTRIB(ins) & IF_DUMMY) continue;
    if (!prev_controltransfer || (I386_INS_OPCODE(prev_controltransfer) != I386_Jcc)) continue;
         if (flaged) printf("good 0.1\n");
    if (I386_INS_CONDITION(prev_controltransfer) != I386_CONDITION_A) continue;
         if (flaged) printf("good 0.2\n");
    if (I386_INS_OPCODE(ins) != I386_JMP) continue;
         if (flaged) printf("good 0.3\n");
 
    jumpreg = I386_OP_INDEX(op);
   //Xiangyu
    if (I386_OP_TYPE(op) != i386_optype_mem ) {
         if (flaged) printf("good 0.4\n");
         if (I386_OP_TYPE(op)==i386_optype_imm) continue;
         if (flaged) printf("good 0.5\n");
         tmp_reg=I386_INS_SOURCE1(ins);
	 //if (tmp_reg==I386_REG_NONE) continue;
	 for (iter=INS_PREV(ins);iter; iter=INS_PREV(iter)) { 
            if (RegsetIn(INS_REGS_DEF(iter),jumpreg)) {//the jump %eax is defined
                tmp_ins=iter;
	        break;
	    }
         }
         if (flaged) printf("good 1\n");
         if (!tmp_ins) continue;
	 op=I386_INS_SOURCE1(tmp_ins);
         if (flaged) printf("good 2 %d@%x\n",I386_OP_TYPE(op),BBL_CADDRESS(ins));
         if (I386_OP_TYPE(op)!=i386_optype_mem) continue;
    } else {
        if (flaged) printf("good 2.5 %d NONE=%d\n",I386_OP_BASE(op), I386_REG_NONE);
        if (I386_OP_BASE(op) != I386_REG_NONE) continue;

        if (I386_OP_SCALE(op) != I386_SCALE_4) FATAL(("unexpected scale in switch jump @I",ins));

         /* check to see if the jumpreg isn't changed between the ja and the jmp */
         changed = NullRegs;
         for (iter = INS_PREV(ins); iter != prev_controltransfer; iter = INS_PREV(iter))
             RegsetSetUnion(changed, INS_REGS_DEF(iter));
   
         if (flaged) printf("good 3\n");
         //8/16/05
         //if (RegsetIn(changed,jumpreg))
         //    continue;
    }

    if (flaged) printf("good 4\n");
    //if (!RELOCATABLE_REFERS_TO(BBL_INS_LAST(INS_BBL(ins)))) continue; 
    //if (!RELOCATABLE_REFERS_TO(BBL_INS_LAST(INS_BBL(ins)))) FATAL(("could not find jumptable relocation for @I",ins));
    //jumptablerel = RELOCATABLE_REFERS_TO(BBL_INS_LAST(INS_BBL(ins)))->rel;
    
    jumptablerel=NULL;
    for (iter=BBL_INS_LAST(INS_BBL(ins));iter && iter!=prev_controltransfer; iter=INS_PREV(iter)) {
          if (RELOCATABLE_REFERS_TO(iter)) {
              jumptablerel=RELOCATABLE_REFERS_TO(iter)->rel;
              break;
          }
    }
    if (!jumptablerel) {
        //FATAL(("NO jumptable"));
        continue;
    }
    

    if (flaged) printf("good 5\n");
    if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(jumptablerel)) != RT_SUBSECTION) continue;
    if (SECTION_TYPE(REL_TO_RELOCATABLE(jumptablerel)) != RO_DATA_SECTION) continue;

    /* find compare instruction that indicates the switch bound */
  
    changed = NullRegs;
    for (cmp = INS_PREV(prev_controltransfer); cmp; cmp = INS_PREV(cmp))
    {
    /*
      if (INS_ATTRIB(cmp) & IF_DUMMY) continue;
      if (I386_INS_OPCODE(cmp) == I386_CMP && 
	  I386_OP_TYPE(I386_INS_SOURCE1(cmp)) == i386_optype_reg &&
	  I386_OP_BASE(I386_INS_SOURCE1(cmp)) == jumpreg &&
	  I386_OP_TYPE(I386_INS_SOURCE2(cmp)) == i386_optype_imm)
	break;

      RegsetSetUnion(changed, INS_REGS_DEF(cmp));

      if (I386InsIsControlTransfer(cmp) || RegsetIn(changed,jumpreg) ||
	  RegsetIn(changed,I386_CONDREG_CF) || RegsetIn(changed,I386_CONDREG_ZF))
      {
	cmp = NULL;
	break;
      }
      */
      if (INS_ATTRIB(cmp) & IF_DUMMY) continue;
      if (I386_INS_OPCODE(cmp) == I386_CMP) {
	  if (I386_OP_TYPE(I386_INS_SOURCE1(cmp))==i386_optype_imm) {
             bound = I386_OP_IMMEDIATE(I386_INS_SOURCE1(cmp));
          } else if  (I386_OP_TYPE(I386_INS_SOURCE2(cmp))==i386_optype_imm) {
             bound = I386_OP_IMMEDIATE(I386_INS_SOURCE2(cmp));
          } 
	  break;
      }
    }
/*
    if (!cmp) continue;

    bound = I386_OP_IMMEDIATE(I386_INS_SOURCE2(cmp));
    VERIFY_FASSERT(G_T_UINT32(SECTION_CSIZE(T_SECTION(REL_TO_RELOCATABLE(jumptablerel)))) >= 4*(bound+1), ("jump table too small for switch bound"));
*/
    for (rr = RELOCATABLE_REFERS_TO(REL_TO_RELOCATABLE(jumptablerel)); rr; rr = rr->next)
    {
      t_reloc * rel = rr->rel;
      if (GENERIC_GE(rel->from_offset,jumptablerel->to_offset) &&
	  GENERIC_LE(rel->from_offset,GENERIC_ADD_UINT32(jumptablerel->to_offset,4*bound)))
      {
	t_cfg_edge * edge;
	rel->hell = FALSE;
	if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(rel)) != RT_BBL) break;
	//printf("\tCandidate SWITCH EDGE %x\n", BBL_CADDRESS(INS_BBL(ins)));
	//VERIFY_FASSERT(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(rel)) == RT_BBL,("@R should point to bbl",rel));
	printf("SWITCH EDGE %x-%x\n", BBL_CADDRESS(INS_BBL(ins)), BBL_CADDRESS(T_BBL(REL_TO_RELOCATABLE(rel))));
        edge = CfgEdgeNew(flowgraph,INS_BBL(ins),T_BBL(REL_TO_RELOCATABLE(rel)),ET_SWITCH);
	edge->rel = rel;
	edge->switchvalue = G_T_UINT32(GENERIC_SUB(rel->from_offset,jumptablerel->to_offset))/4;
      }
    }
  } /* }}} */
#ifdef STEGO
  }
#endif
  /* walk through the original(!) array (we need offsets) and add
   * all possible successors to every node */

  for(ins=section->data; ins!=NULL; ins=INS_NEXT(ins))
  {
    /* we're only interested in the first and last instructions of a basic block */
    if (!(INS_ATTRIB(ins) & IF_BBL_LEADER)) 
      continue;

    block_start = ins;
    block=INS_BBL(block_start);

    block_end = block_start;
    iter = INS_NEXT(block_end);
    while (iter)
    {
      if (INS_ATTRIB(iter) & IF_DUMMY)
      {
	iter = INS_NEXT(iter);
	continue;
      }
      if (INS_ATTRIB(iter) & IF_BBL_LEADER)
	break;
      else
	block_end = iter;
      iter = INS_NEXT(iter);
    }
    /* block_end now points to the last "real" instruction of the block in the linear list,
     * iter points to the first instruction of the next block (or NULL if there is no next block) */
    ft_block = iter ? INS_BBL(iter) : NULL;

#ifdef STEGO
    if(!global_options.stegoDecode){
#endif
    /* {{{ 1. ADD EDGES FOR RELOCATIONS:
     *
     * each relocation gets it's own edge. These edges will be automatically
     * deleted when the relocation is removed. There's no need to add an edge
     * for the obvious relocations (such as the relocations that are present
     * for the calculation of the offsets in intermodular jumps), since they
     * are completly described by the edge itself (the reason for the edge is
     * that there is a jump, not that there is a relocation)
     */

    for (ref=RELOCATABLE_REFED_BY(block); ref!=NULL; ref=ref->next)
    {
      /* Skip the obvious relocations and reloctions to data */
      if (ref->rel->hell)
      {
	t_cfg_edge * i_edge;

	/* do not add hell edges to data blocks */
	if (IS_DATABBL(block))
	{
	  /* the hell flag shouldn't be on for the relocation either: turn it off */
	  ref->rel->hell = FALSE;
	  continue;
	}

	/* skip jump relocations: they are already represented by jump and call edges in the cfg */
	if (RELOCATABLE_TYPE(REL_FROM(ref->rel))==RT_INS)
	{
	  if (I386InsIsControlTransfer((t_ins *)REL_FROM(ref->rel)))
	  {
	    ref->rel->hell = FALSE;
	    continue;
	  }
	}

	/* check if a hell edge already exists */
	BBL_FOREACH_PRED_EDGE(block,i_edge)
	  if ((CFG_EDGE_HEAD(i_edge)==flowgraph->hell_node)&&(EDGE_CAT(i_edge)==ET_CALL))
	    break;

	if (i_edge)
	{
	  /* found a hell edge: just increment the refcount */
	  ref->rel->edge=i_edge;
	  i_edge->refcount++;
	  /* also increment the refcount of the corresponding return edge!!! */
	  VERIFY_FASSERT(i_edge->corr_edge,("Call edge @E does not have a corresponding edge!",i_edge));
	  i_edge->corr_edge->refcount++;
	}
	else
	{
	  /* not found: create a new edge */
	  ref->rel->edge=CfgEdgeNewCall(flowgraph, flowgraph->hell_node, block,NULL,NULL);
	  nedges++;
	}
      }
    } /* }}} */
#ifdef STEGO
    }
#endif
    /* 2. ADD EDGES FOR DATA
     *
     * No edges enter the data block, no edges leave the data block.
     */
    if (IS_DATABBL(block)) continue;

    /* 3. ADD NORMAL EDGES
     *
     * now determine the successors of the block and their types
     */
    if (I386InsIsControlTransfer((t_ins *)block_end))
    {
      if (I386InsIsConditional(block_end))
      {
	/* add fallthrough edge */
	VERIFY_FASSERT(ft_block && !IS_DATABBL(ft_block),("conditional instruction but non-suitable fallthrough block\nblock @iBft_block @iB", block, ft_block));
	CfgEdgeNew(flowgraph, block, ft_block, ET_FALLTHROUGH);
	nedges++;
      }

      switch (I386_INS_OPCODE(block_end))
      {
	case I386_JMP:
	case I386_Jcc:
	case I386_JECXZ:
	  {
	    if (I386_OP_TYPE(I386_INS_SOURCE1(block_end)) == i386_optype_imm)
	    {
	      /* we can determine the jump destination */
	      t_bbl * target;
	      t_ins * target_ins;
	      t_address dest = GENERIC_ADD(INS_CSIZE(block_end),INS_CADDRESS(block_end));
	      dest = GENERIC_ADD_UINT32(dest,I386_OP_IMMEDIATE(I386_INS_SOURCE1(block_end)));
	      target_ins = SecGetInsByAddress(section,dest);
	      if (target_ins)
	      {
		target = INS_BBL(target_ins);
		CfgEdgeNew(flowgraph,block,target,ET_JUMP);
		nedges++;
	      }
	      else
	      {
		VERIFY_FASSERT(GENERICISNULL(dest),("could not find destination of @I",block_end));

		/* jmp 0 was encountered.
		 *         movl %eax <- $0
		 *         testl %eax,%eax
		 *         je label
		 *         ...
		 *         jmp 0
		 *         ...
		 *   label: ...
		 *
		 * The BBL with the call 0 as block_end is either a dead block or data
		 * (otherwise the program would crash because of the call 0)
		 * We will add a jump edge that goes back to the block itself. this is an incorrect
		 * modelling, but we don't care as the block will never be executed anyway
		 */
		CfgEdgeNew(flowgraph, block, block, ET_JUMP);
		nedges++;
	      }
	    }
	    else
	    {
	      /* destination unknown: add a hell edge, except if we're dealing with a 
	       * switch here: in that case the switch edges that are already in place 
	       * suffice. */
	      t_bool is_switch = FALSE;
	      t_cfg_edge * edge;
	      BBL_FOREACH_SUCC_EDGE(block,edge)
		if (EDGE_CAT(edge) == ET_SWITCH)
		{
		  is_switch = TRUE;
		  break;
		}
	      if (!is_switch)
	      {
		CfgEdgeNew(flowgraph,block,flowgraph->hell_node,ET_JUMP);
		nedges++;
	      }
	    }
	  }
	  break;
	case I386_CALL:
	  {
	    t_bool nolink = !ft_block || IS_DATABBL(ft_block);

	    if (I386_OP_TYPE(I386_INS_SOURCE1(block_end)) == i386_optype_imm)
	    {
	      /* we can determine the call destination */
	      t_bbl * target;
	      t_ins * target_ins;
	      t_address dest = GENERIC_ADD(INS_CSIZE(block_end),INS_CADDRESS(block_end));
	      dest = GENERIC_ADD_UINT32(dest,I386_OP_IMMEDIATE(I386_INS_SOURCE1(block_end)));
	      target_ins = SecGetInsByAddress(section,dest);
	      if (target_ins)
	      {
		target = INS_BBL(target_ins);
		if (!nolink)
		{
		  CfgEdgeNewCall(flowgraph, block, target, ft_block, NULL);
		  nedges++;
		  CfgEdgeNew(flowgraph,block,ft_block,ET_FUNLINK);
		  nedges++;
		}
		else
		{
		  /* if the function return would end up in data or out of the code section,
		   * we can't add a link edge. Therefore we assume the call is to a non-returning
		   * function and so we can just as well model it with a jump edge */
		  CfgEdgeNew(flowgraph,block,target,ET_JUMP);
		  nedges++;
		}
	      }
	      else
	      {
		VERIFY_FASSERT(GENERICISNULL(dest),("could not find destination of @I",block_end));

		/* call 0 was encountered.
		 *         movl %eax <- $0
		 *         testl %eax,%eax
		 *         je label
		 *         ...
		 *         call 0
		 *         ...
		 *   label: ...
		 *
		 * The BBL with the call 0 as block_end is either a dead block or data
		 * (otherwise the program would crash because of the call 0)
		 * We will add a jump edge that goes back to the block itself. this is an incorrect
		 * modelling, but we don't care as the block will never be executed anyway
		 */
		CfgEdgeNew(flowgraph, block, block, ET_JUMP);
		nedges++;
	      }
	    }
	    else
	    {
	      /* destination unknown: add a hell edge */
	      if (!nolink)
	      {
		CfgEdgeNewCall(flowgraph,block,flowgraph->hell_node,ft_block,NULL);
		nedges++;
		CfgEdgeNew(flowgraph,block,ft_block,ET_FUNLINK);
		nedges++;
	      }
	      else
	      {
		CfgEdgeNew(flowgraph,block,flowgraph->hell_node,ET_JUMP);
		nedges++;
	      }
	    }
	  }
	  break;
	case I386_LOOP:
	case I386_LOOPZ:
	case I386_LOOPNZ:
	  {
	    t_ins * target_ins;
	    t_bbl * target;
	    t_address dest = GENERIC_ADD(INS_CADDRESS(block_end),INS_CSIZE(block_end));
	    dest = GENERIC_ADD_UINT32(dest,I386_OP_IMMEDIATE(I386_INS_SOURCE1(block_end)));
	    target_ins = SecGetInsByAddress(section,dest);
	    VERIFY_FASSERT(target_ins,("Could not find destination for @I",block_end));
	    target = INS_BBL(target_ins);
	    CfgEdgeNew(flowgraph,block,target,ET_JUMP);
	    nedges++;
	  }
	  break;
	case I386_RET:
	  /* add a return edge that at the moment points to hell */
	  CfgEdgeNew(flowgraph,block,flowgraph->exit_hell_node,ET_RETURN);
	  nedges++;
	  break;
	case I386_INT:
	case I386_INTO:
	case I386_INT3:
	  CfgEdgeNew(flowgraph,block,flowgraph->swi_hell_node,ET_SWI);
	  nedges++;
	  if (ft_block && !IS_DATABBL(ft_block))
	  {
	    CfgEdgeNew(flowgraph,block,ft_block,ET_SWILINK);
	    nedges++;
	  }
	  break;
	case I386_UD2:
	case I386_BOUND:
	  /* for all these strange instructions, we just add a jump edge to hell
	   * (as the control flow may be altered in inpredictible ways), and a 
	   * fallthrough edge to be on the safe side */
	  CfgEdgeNew(flowgraph,block,flowgraph->hell_node,ET_JUMP);
	  nedges++;
	  if (ft_block && !IS_DATABBL(ft_block))
	  {
	    CfgEdgeNew(flowgraph,block,ft_block,ET_FALLTHROUGH);
	    nedges++;
	  }
	  break;
	case I386_HLT:
	  /* for a hlt instruction, we assume the program will never be resumed. 
	   * therefore, it suffices to add a jump edge back to the beginning of 
	   * the block (the hlt instruction will almost certainly be alone in a block */
	  /* TODO for kernels, we will have to forego this assumption, as the hlt instruction
	   * is used within the idle task */
	  CfgEdgeNew(flowgraph,block,block,ET_JUMP);
	  nedges++;
	  break;
	case I386_IRET:
	  /* add a hell edge... a correct modeling will be decided upon later */
	  CfgEdgeNew(flowgraph,block,flowgraph->exit_hell_node,ET_RETURN);
	  break;

	case I386_JMPF:
	  if (I386_OP_TYPE(I386_INS_SOURCE1(block_end)) == i386_optype_mem)
	  {
	    CfgEdgeNew(flowgraph,block,flowgraph->hell_node,ET_JUMP);
	    nedges++;
	  }
	  else
	  {
	    /* destination given as an immediate operand. assume the destination
	     * code segment is the same as the current code segment, because I
	     * wouldn't know what to do otherwise :-( */
	    t_address dest;
	    t_ins * target;
	    t_bbl * target_block;
	    GenericSet_uint32(dest,I386_OP_IMMEDIATE(I386_INS_SOURCE1(block_end)));
	    target = SecGetInsByAddress(section,dest);
	    if (target)
	    {
	      target_block = INS_BBL(target);
	      CfgEdgeNew(flowgraph,block,target_block,ET_JUMP);
	      nedges++;
	    }
	    else
	    {
	      /* if we can't find the destination, it's supposedly a jump to 
	       * a different segment, so we just add a hell edge */
	      CfgEdgeNew(flowgraph,block,flowgraph->hell_node,ET_JUMP);
	      nedges++;
	    }
	  }
	  break;
	case I386_CALLF:
	  {
	    t_bool nolink = !ft_block || IS_DATABBL(ft_block);

	    if (I386_OP_TYPE(I386_INS_SOURCE1(block_end)) == i386_optype_mem)
	    {
	      /* add hell edge */
	      if (!nolink)
	      {
		CfgEdgeNewCall(flowgraph,block,flowgraph->hell_node,ft_block,NULL);
		nedges++;
		CfgEdgeNew(flowgraph,block,ft_block,ET_FUNLINK);
		nedges++;
	      }
	      else
	      {
		CfgEdgeNew(flowgraph,block,flowgraph->hell_node,ET_JUMP);
		nedges++;
	      }
	    }
	    else
	    {
	      /* destination given as an immediate operand. assume the destination
	       * code segment is the same as the current code segment, because I
	       * wouldn't know what to do otherwise :-( */
	      t_address dest;
	      t_ins * target;
	      t_bbl * target_block;
	      GenericSet_uint32(dest,I386_OP_IMMEDIATE(I386_INS_SOURCE1(block_end)));
	      target = SecGetInsByAddress(section,dest);
	      if (target)
	      {
		target_block = INS_BBL(target);

		if (!nolink)
		{
		  CfgEdgeNewCall(flowgraph,block,target_block,ft_block,NULL);
		  nedges++;
		  CfgEdgeNew(flowgraph,block,ft_block,ET_FUNLINK);
		  nedges++;
		}
		else
		{
		  CfgEdgeNew(flowgraph,block,target_block,ET_JUMP);
		  nedges++;
		}
	      }
	      else
	      {
		/* apparently it's a call to a different segment, which is pretty odd */
		FATAL(("implement call to another code segment"));
	      }
	    }
	  }
	  break;
	case I386_RETF:
	  /* add a return edge that at the moment points to hell */
	  CfgEdgeNew(flowgraph,block,flowgraph->exit_hell_node,ET_RETURN);
	  nedges++;
	  break;
	default:
	  FATAL(("unknown control transfer instruction @I",block_end));
      }
    }
    else
    {
      /* just add a fallthrough edge to the next block */
      if (ft_block && !IS_DATABBL(ft_block))
	CfgEdgeNew(flowgraph, block, ft_block, ET_FALLTHROUGH);
    }
  }
  return nedges;
}
/* }}} */

/* {{{ I386FindBBLLeaders */
t_uint32 I386FindBBLLeaders(t_object * obj, t_section * code)
{
  /* Basic leader-detection algorithm.
   * ---------------------------------
   *
   * See e.g. the Red Dragon Book ("Compilers: Principles, Techniques and
   * Tools", by Alfred V. Aho, Ravi Sethi, and Jeffrey D. Ullman) for more
   * details.
   *
   * The aim of this function is to identify all start-addresses of basic
   * blocks (and mark them). There are four reasons to mark an instructions as
   * a basic block leader:
   *
   * 1. The instruction is the target of a direct or indirect jump. This is not
   * as easy as it seems, especially when the program counter is explicit:
   * every instruction that changes the pc can be considered as a jump
   * instruction.
   *
   * 2. It's the successor of one of the above (if a jump is conditional, than
   * the next instruction will be the first instruction of the fall-through
   * block, else it is either target of an other jump or a dead block. In the
   * latter case, it will be removed by the unreachable block removal routines,
   * so we just mark it as a basic block leader).
   *
   * 3. There is an address produced to this basic block, this is either
   *
   *    a. done directly, using the program counter (we assume no real
   *    code-address calculations are present (so not address of function + x) 
   *
   *    b. using a relocation (and not necessary detected when scanning the
   *    instructions)
   *
   * For both cases we need to assume there will be an indirect jump to this
   * address.
   *
   * 4. The start of data blocks in code or instructions following datablocks
   * are considered basic block leaders.
   *
   */

  /* We can still assume that all instructions are stored sequentially in the
   * array Ains the last instruction stored can be recognised with INS_NEXT(ins) ==
   * NULL
   */

  t_i386_ins * ins, * next;
  t_uint32 pc;
  t_uint32 nleaders=0;
  
#ifdef STEGO
  if(global_options.stegoDecode && !global_options.restoreState){
    CreateMissedBblLeadersAndRelocatedOperands(code);
  }
#endif

#ifdef KERNEL_OPT
  /* divide the data blocks in the code as they were meant to be
   * divided (the automatic leader detection fails here) */
  {
    t_symbol * sym;
    int i;

    sym = SymbolTableGetSymbolByName(obj->sub_symbol_table,"stack_start");
    if (!sym) FATAL(("could not find stack_start"));
    INS_ATTRIB(sym->base) |= IF_BBL_LEADER;
    sym = SymbolTableGetSymbolByName(obj->sub_symbol_table,"idt_descr");
    if (!sym) FATAL(("could not find idt_descr"));
    INS_ATTRIB(sym->base) |= IF_BBL_LEADER;
    for (ins = T_I386_INS(sym->base), i=0; i<14; i++)
      ins = INS_NEXT(ins);
    INS_ATTRIB(ins) |= IF_BBL_LEADER;
    sym = SymbolTableGetSymbolByName(obj->sub_symbol_table,"swapper_pg_dir");
    if (!sym) FATAL(("could not find swapper_pg_dir"));
    INS_ATTRIB(sym->base) |= IF_BBL_LEADER;
  }
#endif
  
  /* Initialisation: The first instruction of the program is always a leader.
   * The reasoning is equal to that of reason 2: Either there will be a jump to
   * this address or it's the entry point or it is dead. */

  /* Instructions are still in a sequential list, stored in t_section->data */
  ins = T_I386_INS(code->data); 
  INS_ATTRIB(ins) |= IF_BBL_LEADER;
  
  for (pc = 0; ins != NULL; pc += G_T_UINT32(INS_CSIZE(ins)), ins = (t_i386_ins *) INS_NEXT(ins))
  {
    /* skip dummy instructions */
    while (ins && (INS_ATTRIB(ins) & IF_DUMMY)) 
      ins = (t_i386_ins *) INS_NEXT(ins);
    if (!ins) break;

    /* the entry point is always a leader */
    if (GENERIC_EQ(INS_CADDRESS(ins),obj->entry))
      INS_ATTRIB(ins) |= IF_BBL_LEADER;

    /* reason 4: ins is the start of a data block */
    if (INS_TYPE(ins) == IT_DATA)
    {
      if (INS_PREV(ins) && (INS_TYPE(INS_PREV(ins)) != IT_DATA))
	INS_ATTRIB(ins) |= IF_BBL_LEADER;
      continue;
    }
    /* reason 4b: ins is the first after a data block */
    if (INS_PREV(ins) && (INS_TYPE(INS_PREV(ins)) == IT_DATA))
      INS_ATTRIB(ins) |= IF_BBL_LEADER;

    /* reason 3b: ins is pointed to by a relocation */
    if (RELOCATABLE_REFED_BY(ins))
      INS_ATTRIB(ins) |= IF_BBL_LEADER;

    /* reason 2: ins following a control transfer */
    if (I386InsIsControlTransfer((t_ins *)ins))
    {
      /* find next non-dummy instruction */
      next = (t_i386_ins *) INS_NEXT(ins);
      while (next && (INS_ATTRIB(next) & IF_DUMMY))
	next = (t_i386_ins *) INS_NEXT(next);
      if (next) INS_ATTRIB(next) |= IF_BBL_LEADER;
    }

    /* reason 1: target of a control transfer */
    switch (I386_INS_OPCODE(ins))
    {
      case I386_Jcc:
      case I386_JMP:
      case I386_JECXZ:
      case I386_CALL:
      case I386_LOOP:
      case I386_LOOPZ:
      case I386_LOOPNZ:
	if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	{
	  t_address dest = GENERIC_ADD(INS_CADDRESS(ins),INS_CSIZE(ins));
	  t_ins *  target;
	  dest = GENERIC_ADD_UINT32(dest,I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)));
	  target = SecGetInsByAddress(code,dest);
	  if (target) 
	    INS_ATTRIB(target) |= IF_BBL_LEADER;
	  else
	    VERIFY_FASSERT(GENERICISNULL(dest),("no destination for @I!"));
	}
	break;
      case I386_JMPF:
      case I386_CALLF:
	if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_farptr)
	{
	  /* assume the destination segment is equal to the current code segment (what else could it be?) */
	  t_address dest;
	  t_ins * target;
	  GenericSet_uint32(dest,I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)));
	  target = SecGetInsByAddress(code,dest);
	  if (target) 
	    INS_ATTRIB(target) |= IF_BBL_LEADER;
	  else
	  {
	    /* apparently it's a jump to another segment, we'll just add a hell edge */
	  }
	}
	break;
      default:
	/* do nothing but it saves us a ton of compiler warnings to include this default case here */
	break;
    }
  }

  /* count the number of basic block leaders */
  nleaders = 0;
  for (ins = (t_i386_ins *) code->data; ins; ins = (t_i386_ins *) INS_NEXT(ins))
    if (INS_ATTRIB(ins) & IF_BBL_LEADER)
      nleaders++;

  return nleaders;
} /* }}} */

/* {{{ I386MakeAddressProducers */
void I386MakeAddressProducers(t_section * sec)
{
  t_cfg * cfg = SECTION_CFG(sec);
  t_bbl * bbl, * next;
  t_ins * ins;

  /* GLIBC:
   * we're looking for the following pattern:
   *    call <next ins>
   *    pop %reg
   *    add $offset, %reg
   *
   * this code only appears in the initialization and finalization code
   * of glibc, but it is dangerous code: if the pop and add instructions 
   * get moved, the address calculation will fail.
   * we can easily remove these three instructions and insert a 
   *    mov $dest, %reg 
   * instruction instead. */

  CFG_FOREACH_BBL(cfg,bbl)
  {
    t_i386_operand * op;
    t_reg destreg;
    t_ins * call, * pop, * add;
    t_ins * mov;
    t_int32 offset;
    t_reloc * rel;
#ifdef INSTRUMENTATION
    t_ins * nop1, * nop2;
#endif

 
    ins = BBL_INS_LAST(bbl);
    if (!ins || (I386_INS_OPCODE(ins) != I386_CALL)) continue;
    op = I386_INS_SOURCE1(ins);
    if (I386_OP_TYPE(op) != i386_optype_imm) continue;
    if (I386_OP_IMMEDIATE(op) != 0) continue;

    /* we've found a call to the next instruction. check for the rest of the pattern */
    call = ins;
    next = CFG_EDGE_TAIL(BBL_SUCC_FIRST(bbl));
    pop = BBL_INS_FIRST(next);
    if (!pop || (I386_INS_OPCODE(pop) != I386_POP))
    {
#ifndef INSTRUMENTATION
      if(global_options.restoreState)
      {
	printf("here\n");
	WARNING(("found call to next ins but unknown pattern (@I), possibly long jump", call));
	return;
      }
#endif
      FATAL(("found call to next ins but unknown pattern (@I)", call));
    }
    if (I386_OP_TYPE(I386_INS_DEST(pop)) != i386_optype_reg)
      FATAL(("found call to next ins but unknown pattern (@I)", call));
    destreg = I386_OP_BASE(I386_INS_DEST(pop));

    add = INS_NEXT(pop);
    if (!add || (I386_INS_OPCODE(add) != I386_ADD))
      FATAL(("found call to next ins but unknown pattern (@I)", call));
    if ((I386_OP_TYPE(I386_INS_DEST(add)) != i386_optype_reg) || (I386_OP_BASE(I386_INS_DEST(add)) != destreg))
      FATAL(("found call to next ins but unknown pattern (@I)", call));
    if (I386_OP_TYPE(I386_INS_SOURCE1(add)) != i386_optype_imm)
      FATAL(("found call to next ins but unknown pattern (@I)", call));
    if (!RELOCATABLE_REFERS_TO(add))
      FATAL(("found call to next ins but unknown pattern (@I)", call));
    rel = RELOCATABLE_REFERS_TO(add)->rel;
    offset = (t_int32) I386_OP_IMMEDIATE(I386_INS_SOURCE1(add));

    /* make mov instruction */
    I386MakeInsForBbl(MovToReg,Append,mov,bbl,destreg,I386_REG_NONE,G_T_UINT32(INS_CADDRESS(pop)) + offset);
    I386_OP_FLAGS(I386_INS_SOURCE1(mov)) |= I386_OPFLAG_ISRELOCATED;
    RelocTableAddRelocToRelocatable(sec->obj->reloc_table,
	GENERIC_SUB(rel->addend, GENERIC_ADD(rel->from_offset, INS_CSIZE(pop))),
	(t_relocatable *) mov,
	GENERIC_ADD_UINT32(NULL_GENERIC,1),
	REL_TO_RELOCATABLE(rel),
	rel->to_offset,
	rel->hell,
	rel->edge,
	NULL,
	NULL,
	CALC_ABS "\\" WRITE_32,
	TRUE);
    INS_OLD_SIZE(mov) = INS_OLD_SIZE(call);
    INS_OLD_ADDRESS(mov) = INS_OLD_ADDRESS(call);
#ifdef INSTRUMENTATION
    /* save information about the original instruction */
    T_I386_INS(mov)->ap_original = Malloc(sizeof(t_i386_ins));
    memcpy(T_I386_INS(mov)->ap_original,call,sizeof(t_i386_ins));
    RELOCATABLE_REFERS_TO(T_I386_INS(mov)->ap_original) = NULL;
    RELOCATABLE_REFED_BY(T_I386_INS(mov)->ap_original) = NULL;

    /* make two noops to replace the pop and add instructions */
    I386MakeInsForBbl(Noop,Prepend,nop2,next);
    INS_OLD_SIZE(nop2) = INS_OLD_SIZE(add);
    INS_OLD_ADDRESS(nop2) = INS_OLD_ADDRESS(add);
    /* save information about the original instruction */
    T_I386_INS(nop2)->ap_original = Malloc(sizeof(t_i386_ins));
    memcpy(T_I386_INS(nop2)->ap_original,add,sizeof(t_i386_ins));
    RELOCATABLE_REFERS_TO(T_I386_INS(nop2)->ap_original) = NULL;
    RELOCATABLE_REFED_BY(T_I386_INS(nop2)->ap_original) = NULL;

    I386MakeInsForBbl(Noop,Prepend,nop1,next);
    INS_OLD_SIZE(nop1) = INS_OLD_SIZE(pop);
    INS_OLD_ADDRESS(nop1) = INS_OLD_ADDRESS(pop);
    /* save information about the original instruction */
    T_I386_INS(nop1)->ap_original = Malloc(sizeof(t_i386_ins));
    memcpy(T_I386_INS(nop1)->ap_original,pop,sizeof(t_i386_ins));
    RELOCATABLE_REFERS_TO(T_I386_INS(nop1)->ap_original) = NULL;
    RELOCATABLE_REFED_BY(T_I386_INS(nop1)->ap_original) = NULL;
#endif

    /* kill the pattern */
    InsKill(call);
    InsKill(pop);
    InsKill(add);
    while (BBL_SUCC_FIRST(bbl))
    {
      if (BBL_SUCC_FIRST(bbl)->corr_edge)
	CfgEdgeKill(BBL_SUCC_FIRST(bbl)->corr_edge);
      CfgEdgeKill(BBL_SUCC_FIRST(bbl));
    }

    /* add a fallthrough edge to connect the blocks involved in the pattern */
    CfgEdgeNew(cfg,bbl,next,ET_FALLTHROUGH);
  }

  /* UCLIBC
   * a slightly different pattern:
   * call __i686.get_pc_thunk.bx
   * add $offset, %ebx
   *
   * where __i686.get_pc_thunk.bx looks like
   * mov (%esp), %ebx
   * ret
   */
  CFG_FOREACH_BBL(cfg,bbl)
  {
    t_cfg_edge * edge, * tmp;
    t_ins * ins;
    t_i386_operand * src, * dest;
    t_reg destreg;

    /* look for __i686.get_pc_thunk.bx first, work back from there */
    ins = BBL_INS_FIRST(bbl);
    src = I386_INS_SOURCE1(ins);
    dest = I386_INS_DEST(ins);
    if (BBL_NINS(bbl) != 2) continue;
    if (I386_INS_OPCODE(BBL_INS_LAST(bbl)) != I386_RET) continue;
    if (I386_INS_OPCODE(ins) != I386_MOV) continue;
    if (I386_OP_TYPE(src) != i386_optype_mem ||
	I386_OP_BASE(src) != I386_REG_ESP ||
	I386_OP_INDEX(src) != I386_REG_NONE ||
	I386_OP_IMMEDIATE(src) != 0)
      continue;
    if (I386_OP_TYPE(dest) != i386_optype_reg) continue;
    destreg = I386_OP_BASE(dest);

    /* now go back over all incoming call edges of this basic block */
    BBL_FOREACH_PRED_EDGE_SAFE(bbl,edge,tmp)
    {
      t_bbl * head = CFG_EDGE_HEAD(edge);
      t_bbl * ret;
      t_cfg_edge * link;
      t_reloc * rel;
      t_ins * mov, * call;

      if (EDGE_CAT(edge) != ET_CALL)
	FATAL(("@B should only have incoming call edges!",bbl));

      BBL_FOREACH_SUCC_EDGE(head,link)
	if (EDGE_CAT(link) == ET_FUNLINK)
	  break;
      VERIFY_FASSERT(link,("Could not find link edge"));
      ret = CFG_EDGE_TAIL(link);
      call = BBL_INS_LAST(head);
      ins = BBL_INS_FIRST(ret);
      src = I386_INS_SOURCE1(ins);
      dest = I386_INS_DEST(ins);
      if (I386_INS_OPCODE(ins) != I386_ADD ||
	  I386_OP_TYPE(dest) != i386_optype_reg ||
	  I386_OP_BASE(dest) != destreg ||
	  I386_OP_TYPE(src) != i386_optype_imm)
	FATAL(("Unexpected @I",ins));
      if (!RELOCATABLE_REFERS_TO(ins))
	FATAL(("Expected some PC-relative relocation"));
      rel = RELOCATABLE_REFERS_TO(ins)->rel;

      /* create the address producer */
      I386MakeInsForBbl(MovToReg,Append,mov,head,destreg,I386_REG_NONE,G_T_UINT32(BBL_CADDRESS(ret)) + I386_OP_IMMEDIATE(src));
      I386_OP_FLAGS(I386_INS_SOURCE1(mov)) |= I386_OPFLAG_ISRELOCATED;
      RelocTableAddRelocToRelocatable(sec->obj->reloc_table,
	  GENERIC_SUB(rel->addend, rel->from_offset),
	  (t_relocatable *) mov,
	  GENERIC_ADD_UINT32(NULL_GENERIC,1),
	  REL_TO_RELOCATABLE(rel),
	  rel->to_offset,
	  rel->hell,
	  rel->edge,
	  NULL,
	  NULL,
	  CALC_ABS "\\" WRITE_32,
	  TRUE);
      INS_OLD_SIZE(mov) = INS_OLD_SIZE(call);
      INS_OLD_ADDRESS(mov) = INS_OLD_ADDRESS(call);

      /* kill the call instruction */
      InsKill(call);
      /* turn the add instruction into a noop */
      I386InstructionMakeNoop(ins);
      /* kill the call and return edges */
      if (edge->corr_edge)
	CfgEdgeKill(edge->corr_edge);
      CfgEdgeKill(edge);
      /* turn the funlink edge into a fallthrough edge */
      EDGE_CAT(link) = ET_FALLTHROUGH;
    }

  }
} /* }}}*/

/* vim: set shiftwidth=2 foldmethod=marker: */
