/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/

#include "i386_instructions.h"
#include "i386_registers.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_function.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_print.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../diablo_options.h"

/* {{{ check whether it is possible to inline the function */
static t_bool CheckInlinable(t_function * fun)
{
  t_cfg_edge * edge;
  t_bbl * retblock = FunctionGetExitBlock(fun);

  if (!retblock) return FALSE;
  /* incoming compensating edges => outgoing interprocedural edges => don't
   * inline the function */
  BBL_FOREACH_PRED_EDGE(retblock,edge)
    if (EDGE_CAT(edge) == ET_COMPENSATING) return FALSE;
  /* outgoing compensating edges => incoming interprocedural edges => don't
   * inline the function */
  BBL_FOREACH_SUCC_EDGE(retblock,edge)
    if (EDGE_CAT(edge) == ET_COMPENSATING) return FALSE;

  return TRUE;
} /* }}} */

/*TODO: this is in fact architecture-independent */
/* {{{ copy the structure of the function we want to inline */
static void CopyFunctionStructureToCallSite(t_function * fun, t_bbl * call_site)
{
  t_bbl * i_bbl;
  t_ins * i_ins;
  t_cfg_edge * i_edge;
  t_cfg * cfg = FUN_CFG(fun);

  /* step 1: duplicate all basic blocks of the function */
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    t_bbl * copy;

    /* don't copy the return block */
    if (i_bbl == FUN_BBL_LAST(fun))
      continue;

    copy = BblNew(cfg);
    /* temporary references between the duplicate blocks, this is for easy
     * reference from original to copy and vice versa. these references will
     * of course be removed at the end of the inlining. */
    i_bbl->tmp = copy;
    copy->tmp = i_bbl;

    /* duplicate the instructions in the basic block */
    BBL_FOREACH_INS(i_bbl,i_ins)
    {
      t_ins * inscopy = InsDup(i_ins);
      INS_OLD_ADDRESS(inscopy) = INS_OLD_ADDRESS(i_ins);
      InsAppendToBbl(inscopy, copy);
    }
  }

  /* step 1b: assign the copies of the basic blocks to the calling function */
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    if (i_bbl!=FUN_BBL_LAST(fun))
    {	
      t_bbl * copy = T_BBL(i_bbl->tmp);
      BblInsertInFunction(copy,BBL_FUN(call_site));
      /*copy->next_in_fun=call_site->next_in_fun;
	call_site->next_in_fun->prev_in_fun=copy;
	call_site->next_in_fun = copy;
	copy->prev_in_fun = call_site;
	copy->fun=call_site->fun; */
    }
  }

  /* step 2: add edges to the newly created blocks */
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    /* skip the return edges */
    if (i_bbl == FUN_BBL_LAST(fun)) continue;

    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      /* intraprocedural edges should be duplicated between the copies of head
       * and tail. call edges should be duplicated between
       * the copy of the head and the original tail. interproc jump edges that
       * are used for tail calls are skipped for now. other interprocedural
       * edges should not occur. */
      t_bbl * newhead, * newtail;

      /* skip the edges to the return block */
      if (CFG_EDGE_TAIL(i_edge) == FUN_BBL_LAST(fun))
	continue;

      /* skip interprocedural edges (only tail calls should happen but we handle them later) */
      if (CfgEdgeIsInterproc(i_edge) && !(EDGE_CAT(i_edge) == ET_CALL))
	continue;

      newhead = T_BBL(i_bbl->tmp);
      if (EDGE_CAT(i_edge) == ET_CALL)
      {
	t_bbl * newreturn;

	newtail = CFG_EDGE_TAIL(i_edge);
	newreturn = T_BBL(CFG_EDGE_TAIL(i_edge->corr_edge)->tmp);
	CfgEdgeNewCall(cfg,newhead,newtail,newreturn,FUN_BBL_LAST(BBL_FUN(newtail)));
      }
      else
      {
	newtail = T_BBL(CFG_EDGE_TAIL(i_edge)->tmp);
	CfgEdgeNew(cfg,newhead,newtail,EDGE_CAT(i_edge));
      }
    }
  }
}
/* }}} */

/* {{{ generic function inlining */
t_bool I386InlineFunAtCallSite(t_function * fun, t_bbl * callsite)
{
  t_cfg_edge * ecall, * eret, * elink, * i_edge;
  t_bbl * bentry = FUN_BBL_FIRST(fun);
  t_bbl * bexit = FUN_BBL_LAST(fun);
  t_bbl * returnsite;
  t_ins * cjmp, * rjmp, * sub;
  t_cfg * cfg = BBL_CFG(callsite);

  /*if (!CheckInlinable(fun)) return FALSE;*/

  /* find call, link and return edges */
  BBL_FOREACH_SUCC_EDGE(callsite, ecall)
    if (EDGE_CAT(ecall) == ET_CALL && CFG_EDGE_TAIL(ecall) == bentry)
      break;
  VERIFY_FASSERT(ecall,("inlining: could not find call edge"));
  BBL_FOREACH_SUCC_EDGE(callsite, elink)
    if (EDGE_CAT(elink) == ET_FUNLINK || EDGE_CAT(elink) == ET_IPFUNLINK)
      break;
  VERIFY_FASSERT(elink,("inlining: could not find link edge"));
  returnsite = CFG_EDGE_TAIL(elink);
  BBL_FOREACH_PRED_EDGE(returnsite,eret)
    if (EDGE_CAT(eret) == ET_RETURN && CFG_EDGE_HEAD(eret) == bexit)
      break;
  VERIFY_FASSERT(eret,("inlining: could not find return edge"));

  /* refuse to inline if the funlink edge is interprocedural */
  if (EDGE_CAT(elink) == ET_IPFUNLINK) return FALSE;

  /*if (!BBL_PRED_FIRST(FUN_BBL_LAST(fun))) return FALSE;
  static int teller = 0;
  if (teller++ >= global_options.debugcounter) return FALSE;
  DiabloPrint(stdout,"-inlining fun %s...\n",FUN_NAME(fun));
  DiabloPrint(stdout,"--call site @iB\n",callsite);
  DiabloPrint(stdout,"--return site @iB\n",returnsite);
  DiabloPrint(stdout,"--function: @iB\n",FUN_BBL_FIRST(fun)); */

  /* copy the structure of the function into the caller. */
  CopyFunctionStructureToCallSite(fun,callsite);

  /* change the call instruction into a jump + move the stack pointer 4 bytes
   * down. (this is necessary to keep all stack references in the inlined
   * function correct) */
  InsKill(BBL_INS_LAST(callsite));
  I386MakeInsForBbl(Arithmetic,Append,sub,callsite,I386_SUB,I386_REG_ESP,I386_REG_NONE,4);
  I386MakeInsForBbl(Jump,Append,cjmp,callsite);
  CfgEdgeNew(cfg,callsite,T_BBL(bentry->tmp),ET_JUMP);
  
  /* change the return instructions into jumps + move the stack pointer 4
   * bytes up */
  BBL_FOREACH_PRED_EDGE(bexit,i_edge)
  {
    t_bbl * bret = CFG_EDGE_HEAD(i_edge)->tmp;
    t_ins * ret = BBL_INS_LAST(bret);
    int popbytes = 4;
    VERIFY_FASSERT(EDGE_CAT(i_edge) == ET_JUMP,("flow graph corrupt"));
    VERIFY_FASSERT(I386_INS_OPCODE(ret) == I386_RET,("weird return instruction @I",ret));

    if (I386_OP_TYPE(I386_INS_SOURCE1(ret)) == i386_optype_imm)
      popbytes += I386_OP_IMMEDIATE(I386_INS_SOURCE1(ret));
    I386InstructionMakeArithmetic(ret,I386_ADD,I386_REG_ESP,I386_REG_NONE,popbytes);
    I386MakeInsForBbl(Jump,Append,rjmp,bret);
    CfgEdgeNew(cfg,bret,returnsite,ET_JUMP);
  }

  /* clean up the old edges */
  CfgEdgeKill(ecall);
  CfgEdgeKill(eret);
  CfgEdgeKill(elink);
  
  return TRUE;
} /* }}} */

/* {{{ helper function: determines whether it is necessary to keep the 
 * function's stack behaviour exactly the same before and after inlining */
/* store the stack offset in the tmp field of every instruction */
static t_bool CheckFunctionSimpleness(t_function * fun, t_bool * fp_ret)
{
  t_bbl * bbl = FUN_BBL_FIRST(fun);
  t_ins * ins;
  t_bool simple = TRUE;
  t_int32 offset = 0;
  t_bool fp = FALSE;
  t_uint32 fp_offset = 0;

  ins = BBL_INS_FIRST(bbl);
  /* if the first block of tile instruction is empty, something funny is going on - just be conservative */
  if (!ins) return FALSE;

  /* does the function use a frame pointer? */
  if (I386_INS_OPCODE(ins) == I386_PUSH && I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_EBP)
  {
    (t_int32) INS_TEMP(ins) = offset;
    ins = INS_NEXT(ins);
    offset += 4;

    if (ins && I386_INS_OPCODE(ins) == I386_MOV && 
	I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg && I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_ESP &&
	I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg && I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_EBP)
    {
      (t_int32)INS_TEMP(ins) = offset;
      fp = TRUE;
      fp_offset = 4;
      ins = INS_NEXT(ins);
    }
  }

  for (; ins; ins = INS_NEXT(ins))
  {
    (t_int32) INS_TEMP(ins) = offset;

    /* skip instructions that don't influence the stack */
    if (!RegsetIn(INS_REGS_USE(ins),I386_REG_ESP) && !RegsetIn(INS_REGS_DEF(ins),I386_REG_ESP) &&
	(!fp || (!RegsetIn(INS_REGS_USE(ins),I386_REG_EBP) && !RegsetIn(INS_REGS_DEF(ins),I386_REG_EBP))))
      continue;

    switch (I386_INS_OPCODE(ins))
    {
      case I386_PUSH:
      case I386_PUSHF:
	offset += 4;
	break;
      case I386_PUSHA:
	offset += 32;
	break;
      case I386_POP:
      case I386_POPF:
	offset -= 4;
	break;
      case I386_POPA:
	offset -= 32;
	break;

      case I386_MOV:
	{
	  t_i386_operand * src = I386_INS_SOURCE1(ins), * dest = I386_INS_DEST(ins);

	  if (I386_OP_TYPE(src) == i386_optype_reg)
	    if (I386_OP_BASE(src) == I386_REG_ESP || (fp && I386_OP_BASE(src) == I386_REG_EBP))
	    {
	      /* danger of aliasing */
	      simple = FALSE;
	      break;
	    }
	  if (I386_OP_TYPE(dest) == i386_optype_reg)
	    if (I386_OP_BASE(dest) == I386_REG_ESP || (fp && I386_OP_BASE(dest) == I386_REG_EBP))
	    {
	      /* we don't know the offsets in the stack frame any more */
	      simple = FALSE;
	      break;
	    }

	  /* load and store relative to esp or ebp without index register are well understood and
	   * ok. all other cases fail the test. */
	  if (I386_OP_TYPE(dest) == i386_optype_mem && I386_OP_INDEX(dest) == I386_REG_NONE && 
	      (I386_OP_BASE(dest) == I386_REG_ESP || (fp && I386_OP_BASE(dest) == I386_REG_EBP)))
	  { /* ok */ }
	  else if (I386_OP_TYPE(src) == i386_optype_mem && I386_OP_INDEX(src) == I386_REG_NONE &&
	      (I386_OP_BASE(src) == I386_REG_ESP || (fp && I386_OP_BASE(src) == I386_REG_EBP)))
	  { /* ok */ }
	  else
	  {
	    /* not ok */
	    simple = FALSE;
	    break;
	  }
	}
	break;

      case I386_ADD:
	{
	  t_i386_operand * src = I386_INS_SOURCE1(ins), * dest = I386_INS_DEST(ins);
	  if (I386_OP_TYPE(dest) == i386_optype_reg)
	  {
	    if (I386_OP_BASE(dest) == I386_REG_ESP)
	    {
	      if (I386_OP_TYPE(src) == i386_optype_imm)
		offset -= I386_OP_IMMEDIATE(src);
	      else
	      {
		simple = FALSE;
		break;
	      }
	    }
	    if (fp && I386_OP_BASE(dest) == I386_REG_EBP)
	    {
	      if (I386_OP_TYPE(src) == i386_optype_imm)
		fp_offset -= I386_OP_IMMEDIATE(src);
	      else
	      {
		simple = FALSE;
		break;
	      }
	    }
	  }
	  /* other cases are all ok */
	}
	break;

      case I386_SUB:
	{
	  t_i386_operand * src = I386_INS_SOURCE1(ins), * dest = I386_INS_DEST(ins);
	  if (I386_OP_TYPE(dest) == i386_optype_reg)
	  {
	    if (I386_OP_BASE(dest) == I386_REG_ESP)
	    {
	      if (I386_OP_TYPE(src) == i386_optype_imm)
		offset += I386_OP_IMMEDIATE(src);
	      else
	      {
		simple = FALSE;
		break;
	      }
	    }
	    if (fp && I386_OP_BASE(dest) == I386_REG_EBP)
	    {
	      if (I386_OP_TYPE(src) == i386_optype_imm)
		fp_offset += I386_OP_IMMEDIATE(src);
	      else
	      {
		simple = FALSE;
		break;
	      }
	    }
	  }
	  /* other cases are all ok */
	}
	break;

      case I386_RET:
	/* ok */
	break;

      default:
	/* ok as long as stack pointer (and frame pointer) are only used
	 * as base register in memory operands */
	{
	  t_i386_operand * src1 = I386_INS_SOURCE1(ins), * src2 = I386_INS_SOURCE2(ins),* dest = I386_INS_DEST(ins);

	  if (!RegsetIn(INS_REGS_DEF(ins),I386_REG_ESP) && (!fp || (!RegsetIn(INS_REGS_DEF(ins),I386_REG_EBP))))
	  {
	    if (I386_OP_BASE(dest) == I386_REG_ESP || (fp && I386_OP_BASE(dest) == I386_REG_EBP))
	    {
	      if (I386_OP_TYPE(dest) == i386_optype_mem && I386_OP_INDEX(dest) == I386_REG_NONE)
	      { /* ok */ }
	      else
	      {
		simple = FALSE;
		break;
	      }
	    }
	    if (I386_OP_BASE(src1) == I386_REG_ESP || (fp && I386_OP_BASE(src1) == I386_REG_EBP))
	    {
	      if (I386_OP_TYPE(src1) == i386_optype_mem && I386_OP_INDEX(src1) == I386_REG_NONE)
	      { /* ok */ }
	      else
	      {
		simple = FALSE;
		break;
	      }
	    }
	    if (I386_OP_BASE(src2) == I386_REG_ESP || (fp && I386_OP_BASE(src2) == I386_REG_EBP))
	    {
	      if (I386_OP_TYPE(src2) == i386_optype_mem && I386_OP_INDEX(src2) == I386_REG_NONE)
	      { /* ok */ }
	      else
	      {
		simple = FALSE;
		break;
	      }
	    }
	  }
	  else
	  {
	    simple = FALSE;
	    break;
	  }
	}
	break;
    }

    if (fp && fp_offset != 4) simple = FALSE;
    if (offset < 0) simple = FALSE;
    if (!simple)
    {
      /*DiabloPrint(stdout, "simpleness test failed on @I\n", ins);*/
      break;
    }

  }
  /*if (simple) DiabloPrint(stdout, "eligible function %s passes simpleness test\n", FUN_NAME(fun)); */

  *fp_ret = fp;
  return simple;
} /* }}} */

/* {{{ helper function: remove superfluous stack operations after inlining */
static void RemoveUnnecessaryStackOps(t_bbl * callsite, t_bool fp)
{
  t_bbl * inlined;
  t_ins * ins;

  VERIFY_FASSERT(BBL_SUCC_FIRST(callsite),("need outgoing edge"));
  VERIFY_FASSERT(BBL_SUCC_FIRST(callsite) == BBL_SUCC_LAST(callsite),("should only be one outgoing edge"));
  inlined = CFG_EDGE_TAIL(BBL_SUCC_FIRST(callsite));
  /* kill the stack instructions */
  ins = INS_PREV(BBL_INS_LAST(callsite));
  VERIFY_FASSERT(I386_INS_OPCODE(ins) == I386_SUB,("did not expect @I",ins));
  VERIFY_FASSERT(I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_ESP,("did not expect @I",ins));
  VERIFY_FASSERT(I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) == 4,("did not expect @I",ins));
  InsKill(ins);
  ins = INS_PREV(BBL_INS_LAST(inlined));
  VERIFY_FASSERT(I386_INS_OPCODE(ins) == I386_ADD,("did not expect @I",ins));
  VERIFY_FASSERT(I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_ESP,("did not expect @I",ins));
  if (I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) > 4)
    I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) -= 4;
  else
    InsKill(ins);

  /* iterate through the inlined instructions and adapt stack offsets */
  BBL_FOREACH_INS(inlined,ins)
  {
    t_int32 offset = (t_int32) INS_TEMP(ins);
    t_i386_operand *src1 = I386_INS_SOURCE1(ins), *src2 = I386_INS_SOURCE2(ins), *dest = I386_INS_DEST(ins);
    if (I386_OP_TYPE(dest) == i386_optype_mem)
    {
      if (I386_OP_BASE(dest) == I386_REG_ESP)
      {
	if (I386_OP_IMMEDIATE(dest) > offset) I386_OP_IMMEDIATE(dest) -= 4;
	else if (I386_OP_IMMEDIATE(dest) == offset) FATAL(("function references return address on the stack"));
      }
      if (I386_OP_BASE(dest) == I386_REG_EBP)
      {
	if (I386_OP_IMMEDIATE(dest) > 4) I386_OP_IMMEDIATE(dest) -= 4;
	else if (I386_OP_IMMEDIATE(dest) == 4) FATAL(("function references return address on the stack"));
      }
    }
    if (I386_OP_TYPE(src1) == i386_optype_mem)
    {
      if (I386_OP_BASE(src1) == I386_REG_ESP)
      {
	if (I386_OP_IMMEDIATE(src1) > offset) I386_OP_IMMEDIATE(src1) -= 4;
	else if (I386_OP_IMMEDIATE(src1) == offset) FATAL(("function references return address on the stack"));
      }
      if (I386_OP_BASE(src1) == I386_REG_EBP)
      {
	if (I386_OP_IMMEDIATE(src1) > 4) I386_OP_IMMEDIATE(src1) -= 4;
	else if (I386_OP_IMMEDIATE(src1) == 4) FATAL(("function references return address on the stack"));
      }
    }
    if (I386_OP_TYPE(src1) == i386_optype_mem)
    {
      if (I386_OP_BASE(src2) == I386_REG_ESP)
      {
	if (I386_OP_IMMEDIATE(src2) > offset) I386_OP_IMMEDIATE(src2) -= 4;
	else if (I386_OP_IMMEDIATE(src2) == offset) FATAL(("function references return address on the stack"));
      }
      if (I386_OP_BASE(src2) == I386_REG_EBP)
      {
	if (I386_OP_IMMEDIATE(src2) > 4) I386_OP_IMMEDIATE(src2) -= 4;
	else if (I386_OP_IMMEDIATE(src2) == 4) FATAL(("function references return address on the stack"));
      }
    }
  }
}
/* }}} */

/* {{{ inline simple functions of only one basic block */
#define MAX_NR_INS	2
void I386InlineSmallFunctions(t_cfg * cfg)
{
  t_function * fun;
  CFG_FOREACH_FUN(cfg,fun)
  {
    t_cfg_edge * edge, * tmp;
    t_bool simple, use_fp;
    
    if (!FUN_BBL_FIRST(fun)) continue;
    if (BBL_NEXT_IN_FUN(FUN_BBL_FIRST(fun)) != FUN_BBL_LAST(fun)) continue;
    if (BBL_NINS(FUN_BBL_LAST(fun)) != 0) continue;
    if (FUN_IS_HELL(fun)) continue;
    
    /* avoid unreasonable expansion of the program: cap the number of
     * instructions in the function */
    if (BBL_NINS(FUN_BBL_FIRST(fun)) > MAX_NR_INS) continue;

    if (!CheckInlinable(fun)) continue;

    simple = CheckFunctionSimpleness(fun,&use_fp);

    /*DiabloPrint(stdout,"IS: will try to inline @iB\n",FUN_BBL_FIRST(fun));*/

    BBL_FOREACH_PRED_EDGE_SAFE(FUN_BBL_FIRST(fun),edge,tmp)
    {
      t_bbl * callsite = CFG_EDGE_HEAD(edge);

      if (EDGE_CAT(edge) != ET_CALL) continue;
      if (IsHellNode(callsite)) continue;

      DiabloPrint(stdout,"Inlining %s at call site @B\t", fun->name, callsite);
      if (I386InlineFunAtCallSite(fun,callsite))
      {
	/*printf("[ OK ]\n");*/
	if (simple) RemoveUnnecessaryStackOps(callsite,use_fp);
      }
      /*else*/
	/*printf("[FAIL]\n");*/
    }
    
  }
/*  DiabloPrint(stdout,"[small inlining]: %d times\n",counter);*/
} /* }}} */

/* {{{ inline functions with only one call site
 * until we have a good stack analysis, this will probably always increase code size :-( */
void I386InlineFunctionsWithOneCallSite(t_cfg * cfg)
{
  t_function * fun;
  static int count = 0;

  ANALYSIS_START("Inlining functions with one call site");
  CFG_FOREACH_FUN(cfg,fun)
  {
    t_cfg_edge * edge, * call = NULL;
    int ncallers = 0;
    t_bool ok;

    if (FUN_IS_HELL(fun)) continue;
    if (!FUN_BBL_FIRST(fun)) continue;
    if (!CheckInlinable(fun)) continue;
    BBL_FOREACH_PRED_EDGE(FUN_BBL_FIRST(fun),edge)
      if (EDGE_CAT(edge) == ET_CALL)
      {
	call = edge;
	if (IsHellNode(CFG_EDGE_HEAD(edge)))
	  ncallers += 10;
	else
	  ncallers++;
      }
    if (ncallers != 1) continue;

    /*if (count >= global_options.debugcounter) continue;*/
    /*DiabloPrint(stdout, "fun %s has only one call site at @B, inlining...\n",fun->name,CFG_EDGE_HEAD(call));*/

    ok = I386InlineFunAtCallSite(fun,CFG_EDGE_HEAD(call));
    if (!ok) printf("failed!\n");
    else count++;

  }
  VERBOSE(0,("inlined %d functions at one call site\n",count));
  ANALYSIS_SIMPLE_END("Inlining functions with one call site");
}
/* }}} */

/* vim: set shiftwidth=2 foldmethod=marker : */
