/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include "i386_assemble.h"
#include "i386_instructions.h"
#include "i386_opcode.h"
#include "i386_registers.h"
#include "../../utils/util_error.h"
#include <string.h>

/** {{{ ins is conditional? */
t_bool I386InsIsConditional(t_ins * ins)
{
  return I386_INS_CONDITION(ins) != I386_CONDITION_NONE;
}
/* }}} */

/** {{{ ins has side effect? */
t_bool I386InsHasSideEffect(t_ins * ins)
{

  if (I386InsIsStore(ins))
    return TRUE;

  if (I386InsIsControlTransfer(ins))
    return TRUE;

  /* all instructions that define the stack pointer have a side effect */
  if (RegsetIn(INS_REGS_DEF(ins),I386_REG_ESP))
    return TRUE;

  /* some special cases */
  switch (I386_INS_OPCODE(ins))
  {
    case I386_IN:
    case I386_INSB:
    case I386_INSD:
    case I386_OUT:
    case I386_OUTSB:
    case I386_OUTSD:
    case I386_LEAVE:
    case I386_LDS:
    case I386_LES:
    case I386_LFS:
    case I386_LGS:
    case I386_LSS:
    case I386_FLDCW:
    case I386_FSTCW:
    case I386_FLDENV:
    case I386_FSTENV:
    case I386_FSAVE:
    case I386_FRSTOR:
      return TRUE;
    default:
      break;
  }

  /* all instructions that push/pop the fpu stack have a side effect
   * (they change the top-of-stack pointer, which is an invisible and
   * unmodeled register) */
  if (i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pops ||
      i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pushes)
    return TRUE;

  return FALSE;
} /* }}} */

/** {{{ ins writes to memory? */
t_bool I386InsIsStore(t_ins * ins)
{
  /* first the obvious cases */
  switch (I386_INS_OPCODE(ins))
  {
    case I386_PUSH:
    case I386_PUSHA:
    case I386_PUSHF:
    case I386_ENTER:
    case I386_MOVSB:
    case I386_MOVSD:
    case I386_STOSB:
    case I386_STOSD:
    case I386_OUTSB:
    case I386_OUTSD:
    /* calls store their return address on the stack */
    case I386_CALL:
    case I386_CALLF:
    /* pseudo save, used in instrumentation */
    case I386_PSEUDOSAVE:
      return TRUE;
    default:
      break;
  }

  if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_mem)
    return TRUE;
  if (I386_INS_HAS_FLAG(ins, I386_IF_SOURCE1_DEF))
    if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_mem)
      return TRUE;
  if (I386_INS_HAS_FLAG(ins, I386_IF_SOURCE2_DEF))
    if (I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_mem)
      return TRUE;

  return FALSE;

} /* }}} */

/** {{{ ins reads from memory? */
t_bool I386InsIsLoad(t_ins * ins)
{
  /* instructions in which the memory operation is implicit */
  switch (I386_INS_OPCODE(ins))
  {
    case I386_POP:
    case I386_POPA:
    case I386_POPF:
    case I386_LEAVE:
    case I386_MOVSB:
    case I386_MOVSD:
    case I386_CMPSB:
    case I386_CMPSD:
    case I386_LODSB:
    case I386_LODSD:
      /* return instructions pop the return address off the stack */
    case I386_RET:
    case I386_RETF:
    case I386_IRET:
      /* pseudo load, used in instrumentation */
    case I386_PSEUDOLOAD:
      return TRUE;
      /* lea has a memory operand but doesn't actually perform a load from it */
    case I386_LEA:
      return FALSE;
    default:
      break;
  }

  if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_mem)
    return TRUE;
  if (I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_mem)
    return TRUE;
  if (I386_INS_HAS_FLAG(ins, I386_IF_DEST_IS_SOURCE))
    if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_mem)
      return TRUE;

  return FALSE;

} /* }}} */

/** {{{ ins is control transfer? */
t_bool I386InsIsSystemControlTransfer(t_ins * ins)
{
  switch (I386_INS_OPCODE(ins))
  {
    case I386_INT:
    case I386_INTO:
    case I386_BOUND:
    case I386_UD2:
    case I386_HLT:
    case I386_SYSENTER:
    case I386_SYSEXIT:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}

t_bool I386InsIsRegularControlTransfer(t_ins * ins)
{
  switch (I386_INS_OPCODE(ins))
  {
    case I386_JMP:
    case I386_JMPF:
    case I386_Jcc:
    case I386_JECXZ:
    case I386_LOOP:
    case I386_LOOPZ:
    case I386_LOOPNZ:
    case I386_CALL:
    case I386_CALLF:
    case I386_RET:
    case I386_RETF:
    case I386_IRET:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}

t_bool I386InsIsControlTransfer(t_ins * ins)
{
  if (I386InsIsSystemControlTransfer(ins) || I386InsIsRegularControlTransfer(ins))
    return TRUE;

  /* in fact there's no need to see this as control transfer instructions
     if (I386_INS_HAS_PREFIX(ins,I386_PREFIX_REP) || I386_INS_HAS_PREFIX(ins,I386_PREFIX_REPNZ))
    return TRUE; */

  return FALSE;
} /* }}} */

/** {{{ clear the current contents of the instruction */
static void ClearIns(t_ins * ins)
{
  memset(I386_INS_DEST(ins), 0, sizeof(t_i386_operand));
  memset(I386_INS_SOURCE1(ins), 0, sizeof(t_i386_operand));
  memset(I386_INS_SOURCE2(ins), 0, sizeof(t_i386_operand));
  I386_INS_FLAGS(ins) = 0x0;
  I386_INS_PREFIXES(ins) = 0x0;
  I386_INS_CONDITION(ins) = I386_CONDITION_NONE;
#ifdef OBFUSCATION_SUPPORT
  I386_INS_CORR(ins)=NULL;
#endif
  while (RELOCATABLE_REFERS_TO(ins))
    RelocTableRemoveReloc(INS_BBL(ins)->cfg->sec->obj->reloc_table,RELOCATABLE_REFERS_TO(ins)->rel);
} /* }}}*/

/** {{{ set some generic characteristics of the instruction */
static void SetGenericInsInfo(t_ins * ins)
{
  INS_TYPE(ins) = i386_opcode_table[I386_INS_OPCODE(ins)].type;
  INS_CSIZE(ins) = I386InsGetSize(ins);
  INS_REGS_USE(ins) =  I386InsUsedRegisters(ins);
  INS_REGS_DEF(ins) =  I386InsDefinedRegisters(ins);
}
/* }}} */

/** {{{ make noop */
void I386InstructionMakeNoop(t_ins * ins)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = I386_NOP;
  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make call */
void I386InstructionMakeCall(t_ins * ins)
{
  t_i386_operand * op;

  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_CALL;
  op = I386_INS_SOURCE1(ins);
  I386_OP_TYPE(op) = i386_optype_imm;
  I386_OP_IMMEDSIZE(op) = 4;
  I386_OP_IMMEDIATE(op) = 0;

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make jump */
void I386InstructionMakeJump(t_ins * ins)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_JMP;
  I386OpSetImm(I386_INS_SOURCE1(ins),0,4);

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make jump reg */
void I386InstructionMakeJumpReg(t_ins * ins, t_reg src)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_JMP;

  if (src != I386_REG_NONE)
    I386OpSetReg(I386_INS_SOURCE1(ins),src,i386_regmode_full32);

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make conditional jump (also jecxz and loop variants) */
void I386InstructionMakeCondJump(t_ins * ins, t_i386_condition_code condition)
{
  ClearIns(ins);

  switch (condition)
  {
    case I386_CONDITION_ECXZ:
      I386_INS_OPCODE(ins) = I386_JECXZ;
      break;
    case I386_CONDITION_LOOP:
      I386_INS_OPCODE(ins) = I386_LOOP;
      break;
    case I386_CONDITION_LOOPZ:
      I386_INS_OPCODE(ins) = I386_LOOPZ;
      break;
    case I386_CONDITION_LOOPNZ:
      I386_INS_OPCODE(ins) = I386_LOOPNZ;
      break;
    case I386_CONDITION_NONE:
      FATAL(("cannot make unconditional conditional jump!"));
    default:
      I386_INS_OPCODE(ins) = I386_Jcc;
  }
  I386_INS_CONDITION(ins) = condition;
  I386OpSetImm(I386_INS_SOURCE1(ins),0,1);

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make pop */
void I386InstructionMakePop(t_ins * ins, t_reg reg)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = I386_POP;
  I386OpSetReg(I386_INS_DEST(ins),reg,i386_regmode_full32);
  SetGenericInsInfo(ins);
}
/* }}} */

/** {{{ make push */
void I386InstructionMakePush(t_ins * ins, t_reg reg, t_uint32 immval)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_PUSH;
  if (reg == I386_REG_NONE)
    I386OpSetImm(I386_INS_SOURCE1(ins),immval,4);
  else
    I386OpSetReg(I386_INS_SOURCE1(ins),reg,i386_regmode_full32);

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make push memory operand */
void I386InstructionMakePushMem(t_ins * ins, t_uint32 offset, t_reg base, t_reg index, int scale)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_PUSH;
  I386OpSetMem(I386_INS_SOURCE1(ins),offset,base,index,scale,4);

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make an arithmetic instruction of the form "opc src, dest" or "opc imm, dest" with dest a register */
void I386InstructionMakeArithmetic(t_ins * ins, t_i386_opcode opc, t_reg dest, t_reg src, t_uint32 imm)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = opc;
  I386OpSetReg(I386_INS_DEST(ins),dest,i386_regmode_full32);
  if (src != I386_REG_NONE)
    I386OpSetReg(I386_INS_SOURCE1(ins),src,i386_regmode_full32);
  else
    I386OpSetImm(I386_INS_SOURCE1(ins),imm,4);

  /* these instructions have in common that they all use their destination
   * operand */
  I386_INS_FLAGS(ins) |= I386_IF_DEST_IS_SOURCE;

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make an arithmetic instruction of the form "opc src, dest" or "opc imm, dest" with dest a memory location */
void I386InstructionMakeArithmeticToMem(t_ins * ins, t_i386_opcode opc, t_uint32 offset, t_reg base, t_reg index, int scale, t_reg src, t_uint32 imm)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = opc;
  I386OpSetMem(I386_INS_DEST(ins),offset,base,index,scale,4);
  if (src != I386_REG_NONE)
    I386OpSetReg(I386_INS_SOURCE1(ins),src,i386_regmode_full32);
  else
    I386OpSetImm(I386_INS_SOURCE1(ins),imm,4);

  /* these instructions have in common that they all use their destination
   * operand */
  I386_INS_FLAGS(ins) |= I386_IF_DEST_IS_SOURCE;

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make an arithmetic instruction of the form "opc src, dest" where src is a memory location */
void I386InstructionMakeArithmeticFromMem(t_ins * ins, t_i386_opcode opc, t_reg dest, t_uint32 offset, t_reg base, t_reg index, int scale)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = opc;
  I386OpSetReg(I386_INS_DEST(ins),dest,i386_regmode_full32);
  I386OpSetMem(I386_INS_SOURCE1(ins),offset,base,index,scale,4);

  /* these instructions have in common that they all use their destination
   * operand */
  I386_INS_FLAGS(ins) |= I386_IF_DEST_IS_SOURCE;

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make an instruction from reg/imm to reg */
void I386InstructionMakeRM32IMM8(t_ins * ins, t_reg dest, t_reg src, t_uint32 imm, t_i386_opcode opcode)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = opcode;
  I386OpSetReg(I386_INS_DEST(ins),dest,i386_regmode_full32);
  if (src != I386_REG_NONE)
    I386OpSetReg(I386_INS_SOURCE1(ins),src,i386_regmode_full32);
  else
    I386OpSetImm(I386_INS_SOURCE1(ins),imm,1);

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make an instruction from reg/imm to reg */
void I386InstructionMakeRM32IMM32(t_ins * ins, t_reg dest, t_reg src, t_uint32 imm, t_i386_opcode opcode)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = opcode;
  I386OpSetReg(I386_INS_DEST(ins),dest,i386_regmode_full32);
  if (src != I386_REG_NONE)
    I386OpSetReg(I386_INS_SOURCE1(ins),src,i386_regmode_full32);
  else
    I386OpSetImm(I386_INS_SOURCE1(ins),imm,4);

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make a mov instruction from reg/imm to reg */
void I386InstructionMakeMovToReg(t_ins * ins, t_reg dest, t_reg src, t_uint32 imm)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_MOV;
  I386OpSetReg(I386_INS_DEST(ins),dest,i386_regmode_full32);
  if (src != I386_REG_NONE)
    I386OpSetReg(I386_INS_SOURCE1(ins),src,i386_regmode_full32);
  else
    I386OpSetImm(I386_INS_SOURCE1(ins),imm,4);

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make a mov instruction to memory + len*/
void I386InstructionMakeMovToMemLen(t_ins * ins, t_uint32 offset, t_reg base, t_reg index, int scale, t_reg src, t_uint32 imm, t_uint32 imm_len)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_MOV;
  I386OpSetMem(I386_INS_DEST(ins),offset,base,index,scale,imm_len);
  if (src != I386_REG_NONE)
    I386OpSetReg(I386_INS_SOURCE1(ins),src,i386_regmode_full32);
  else
    I386OpSetImm(I386_INS_SOURCE1(ins),imm,imm_len);

  SetGenericInsInfo(ins);
}
/* }}} */

/** {{{ make a mov instruction to memory */
void I386InstructionMakeMovToMem(t_ins * ins, t_uint32 offset, t_reg base, t_reg index, int scale, t_reg src, t_uint32 imm)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_MOV;
  I386OpSetMem(I386_INS_DEST(ins),offset,base,index,scale,4);
  if (src != I386_REG_NONE)
    I386OpSetReg(I386_INS_SOURCE1(ins),src,i386_regmode_full32);
  else
    I386OpSetImm(I386_INS_SOURCE1(ins),imm,4);

  SetGenericInsInfo(ins);
}
/* }}} */

/** {{{ make a mov instruction from memory */
void I386InstructionMakeMovFromMem(t_ins * ins, t_reg dest, t_uint32 offset, t_reg base, t_reg index, int scale)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = I386_MOV;
  I386OpSetReg(I386_INS_DEST(ins),dest,i386_regmode_full32);
  I386OpSetMem(I386_INS_SOURCE1(ins),offset,base,index,scale,4);
  SetGenericInsInfo(ins);
}
/* }}} */

/** {{{ make a lea instruction */
void I386InstructionMakeLea(t_ins * ins, t_reg dest, t_uint32 offset, t_reg base, t_reg index, int scale)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_LEA;
  I386OpSetReg(I386_INS_DEST(ins),dest,i386_regmode_full32);
  I386OpSetMem(I386_INS_SOURCE1(ins),offset,base,index,scale,0);

  SetGenericInsInfo(ins);
}
/* }}} */

/** {{{ compare register to register or immediate */
void I386InstructionMakeCmp(t_ins * ins, t_reg reg, t_reg cmpreg, t_uint32 cmpimm)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_CMP;
  I386OpSetReg(I386_INS_SOURCE1(ins),reg,i386_regmode_full32);
  if (cmpreg == I386_REG_NONE)
    I386OpSetImm(I386_INS_SOURCE2(ins),cmpimm,4);
  else
    I386OpSetReg(I386_INS_SOURCE2(ins),cmpreg,i386_regmode_full32);

  SetGenericInsInfo(ins);
} /* }}} */

/** {{{ make a setcc instruction */
void I386InstructionMakeSetcc(t_ins * ins, t_i386_condition_code cond, t_uint32 offset, t_reg base, t_reg index, int scale, t_bool memop)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = I386_SETcc;
  I386_INS_CONDITION(ins) = cond;
  if (memop)
    I386OpSetMem(I386_INS_DEST(ins),offset,base,index,scale,1);
  else
    I386OpSetReg(I386_INS_DEST(ins),base,i386_regmode_full32);

  SetGenericInsInfo(ins);
}
/* }}} */

/** {{{ make a pseudo call (used in instrumentation) */
void I386InstructionMakePseudoCall(t_ins * ins, t_function * to)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = I386_PSEUDOCALL;
  /* this is really ugly but we don't care. pseudo calls are hackish as hell anyway */
  ((t_function *) I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins))) = to;

  INS_CSIZE(ins) = NULL_GENERIC;
  INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
  INS_REGS_DEF(ins) = I386InsDefinedRegisters(ins);
  INS_TYPE(ins) = IT_CALL;
}
/* }}} */

/** {{{ make a pseudo save (used in instrumentation) */
void I386InstructionMakePseudoSave(t_ins * ins, t_reg reg)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = I386_PSEUDOSAVE;
  INS_TYPE(ins) = IT_PSEUDO_SAVE;

  I386OpSetReg(I386_INS_SOURCE1(ins),reg,i386_regmode_full32);
  GenericSet_uint32(INS_CSIZE(ins),6);
  INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
  INS_REGS_DEF(ins) = I386InsDefinedRegisters(ins);
}
/* }}} */

/** {{{ make a pseudo load (used in instrumentation) */
void I386InstructionMakePseudoLoad(t_ins * ins, t_reg reg)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = I386_PSEUDOLOAD;
  INS_TYPE(ins) = IT_PSEUDO_LOAD;

  I386OpSetReg(I386_INS_DEST(ins),reg,i386_regmode_full32);
  GenericSet_uint32(INS_CSIZE(ins),6);
  INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
  INS_REGS_DEF(ins) = I386InsDefinedRegisters(ins);
}
/* }}} */

/** {{{ make a simple no-argument instruction with a given opcode */
void I386InstructionMakeSimple(t_ins * ins, t_i386_opcode opc)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = opc;
  SetGenericInsInfo(ins);
}
/* }}} */

/** {{{ make LAHF/SAHF instruction */
void I386InstructionMakeLSahf(t_ins * ins, t_i386_opcode opc)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = opc;
  if (opc == I386_LAHF)
    I386OpSetReg(I386_INS_DEST(ins),I386_REG_EAX,i386_regmode_hi8);
  else if (opc == I386_SAHF)
    I386OpSetReg(I386_INS_SOURCE1(ins),I386_REG_EAX,i386_regmode_hi8);
  else 
    FATAL(("wrong opcode"));
  SetGenericInsInfo(ins);
}
/* }}}*/

/** {{{ make FSAVE/FRSTOR instruction */
void I386InstructionMakeFSaveRstor(t_ins * ins, t_i386_opcode opc, t_uint32 offset, t_reg base, t_reg index, int scale)
{
  ClearIns(ins);

  I386_INS_OPCODE(ins) = opc;
  if (opc == I386_FSAVE)
    I386OpSetMem(I386_INS_DEST(ins),offset,base,index,scale,0);
  else if (opc == I386_FRSTOR)
    I386OpSetMem(I386_INS_SOURCE1(ins),offset,base,index,scale,0);
  else
    FATAL(("wrong opcode"));

  SetGenericInsInfo(ins);
}
/* }}} */

/** {{{ make inc/dec instruction */
void I386InstructionMakeIncDec(t_ins * ins, t_i386_opcode opc, t_reg dest)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = opc;
  I386OpSetReg(I386_INS_DEST(ins),dest,i386_regmode_full32);
  I386_INS_FLAGS(ins) |= I386_IF_DEST_IS_SOURCE;
  SetGenericInsInfo(ins);
}
/* }}}*/

/** {{{ make test instruction */
void I386InstructionMakeTest(t_ins * ins, t_reg src1, t_reg src2, t_uint32 imm)
{
  ClearIns(ins);
  I386_INS_OPCODE(ins) = I386_TEST;
  I386OpSetReg(I386_INS_SOURCE1(ins),src1,i386_regmode_full32);
  if (src2 != I386_REG_NONE)
    I386OpSetReg(I386_INS_SOURCE2(ins),src2,i386_regmode_full32);
  else
    I386OpSetImm(I386_INS_SOURCE2(ins),imm,4);
  SetGenericInsInfo(ins);
}
/* }}} */




t_bool I386IsUseException(t_ins * ins){
  switch(I386_INS_OPCODE(ins)){
    case I386_IMULexp2:
      if(I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins))!=0)
	return FALSE;
    case I386_SUB:
    case I386_XOR:
      return 
	(
	 (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg && I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg)
	 &&
	 (I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_OP_BASE(I386_INS_DEST(ins)))
	);
    case I386_AND:
      return
	(
	 (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg && I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	 &&
	 (I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins))==0)
	);
    default:
      return FALSE;
  }
}
/* {{{ I386InsUsedRegisters */
t_regset I386InsUsedRegisters(t_ins * ins)
{
  t_regset use = NullRegs;
  t_i386_operand * op;

  if (INS_TYPE(ins) == IT_DATA)
    return use;

  /* {{{ used condition flags */
  if (I386_INS_CONDITION(ins) == I386_CONDITION_NONE)
    RegsetSetUnion(use, i386_opcode_table[I386_INS_OPCODE(ins)].cf_used);
  else
  {
    switch (I386_INS_CONDITION(ins))
    {
      case I386_CONDITION_O:
      case I386_CONDITION_NO:
	RegsetAddReg(use, I386_CONDREG_OF);
	break;
      case I386_CONDITION_B:
      case I386_CONDITION_AE:
	RegsetAddReg(use, I386_CONDREG_CF);
	break;
      case I386_CONDITION_Z:
      case I386_CONDITION_NZ:
	RegsetAddReg(use, I386_CONDREG_ZF);
	break;
      case I386_CONDITION_BE:
      case I386_CONDITION_A:
	RegsetAddReg(use, I386_CONDREG_CF);
	RegsetAddReg(use, I386_CONDREG_ZF);
	break;
      case I386_CONDITION_S:
      case I386_CONDITION_NS:
	RegsetAddReg(use, I386_CONDREG_SF);
	break;
      case I386_CONDITION_P:
      case I386_CONDITION_NP:
	RegsetAddReg(use, I386_CONDREG_PF);
	break;
      case I386_CONDITION_L:
      case I386_CONDITION_GE:
	RegsetAddReg(use, I386_CONDREG_SF);
	RegsetAddReg(use, I386_CONDREG_OF);
	break;
      case I386_CONDITION_LE:
      case I386_CONDITION_G:
	RegsetAddReg(use, I386_CONDREG_SF);
	RegsetAddReg(use, I386_CONDREG_OF);
	RegsetAddReg(use, I386_CONDREG_ZF);
	break;
      case I386_CONDITION_LOOPZ:
      case I386_CONDITION_LOOPNZ:
	RegsetAddReg(use, I386_CONDREG_ZF);
	break;
      case I386_CONDITION_ECXZ:
      case I386_CONDITION_LOOP:
	/* no condition flags used */
	break;
      default:
	FATAL(("unexpected condition value for @I",ins));
    }
  } /* }}} */

  /* {{{ source operands */
  op = I386_INS_SOURCE1(ins);
  switch (I386_OP_TYPE(op))
  {
    case i386_optype_reg:
      if(!I386IsUseException(ins))
      RegsetAddReg(use, I386_OP_BASE(op));
      break;
    case i386_optype_mem:
      if (I386_OP_BASE(op) != I386_REG_NONE)
	RegsetAddReg(use, I386_OP_BASE(op));
      if (I386_OP_INDEX(op) != I386_REG_NONE)
	RegsetAddReg(use, I386_OP_INDEX(op));
      break;
    default:
      /* keep the compiler happy */
      break;
  }
  op = I386_INS_SOURCE2(ins);
  switch (I386_OP_TYPE(op))
  {
    case i386_optype_reg:
      RegsetAddReg(use, I386_OP_BASE(op));
      break;
    case i386_optype_mem:
      if (I386_OP_BASE(op) != I386_REG_NONE)
	RegsetAddReg(use, I386_OP_BASE(op));
      if (I386_OP_INDEX(op) != I386_REG_NONE)
	RegsetAddReg(use, I386_OP_INDEX(op));
      break;
    default:
      /* keep the compiler happy */
      break;
  }
  /* instructions where the destination operand is also used as a source operand */
  if (I386_INS_HAS_FLAG(ins,I386_IF_DEST_IS_SOURCE))
  {
    op = I386_INS_DEST(ins);
    switch (I386_OP_TYPE(op))
    {
      case i386_optype_reg:
	if(!I386IsUseException(ins))
	RegsetAddReg(use, I386_OP_BASE(op));
	break;
      case i386_optype_mem:
	if (I386_OP_BASE(op) != I386_REG_NONE)
	  RegsetAddReg(use, I386_OP_BASE(op));
	if (I386_OP_INDEX(op) != I386_REG_NONE)
	  RegsetAddReg(use, I386_OP_INDEX(op));
	break;
      default:
	/* keep the compiler happy */
	break;
    }
  }
  /* }}} */

  /* {{{ the destination operand */
  op = I386_INS_DEST(ins);
  if (I386_OP_TYPE(op) == i386_optype_mem)
  {
    if (I386_OP_BASE(op) != I386_REG_NONE)
      RegsetAddReg(use, I386_OP_BASE(op));
    if (I386_OP_INDEX(op) != I386_REG_NONE)
      RegsetAddReg(use, I386_OP_INDEX(op));
  } /* }}} */

  /* {{{ prefixes */ 
  /* the REP/REPNZ prefixes */
  if (I386_INS_HAS_PREFIX(ins, I386_PREFIX_REP) ||
      I386_INS_HAS_PREFIX(ins, I386_PREFIX_REPNZ))
    RegsetAddReg(use, I386_REG_ECX);
  
  /* used segment registers: instructions always use the CS register, because this register
   * is used by the processor to determine from where it has to load the instruction. Instructions
   * that have memory operands usually get them from the data segment, so these instructions also 
   * the DS register. However, this can be overridden using the segment override prefix. The string
   * instructions also implicitly use the ES register (all of this is primarily of importance for 
   * kernel level code, not for user level code) */
#if 0
  if (I386_INS_HAS_PREFIX(ins,I386_PREFIX_CS_OVERRIDE))
    RegsetAddReg(use, I386_REG_CS);
  else if (I386_INS_HAS_PREFIX(ins,I386_PREFIX_DS_OVERRIDE))
    RegsetAddReg(use, I386_REG_DS);
  else if (I386_INS_HAS_PREFIX(ins,I386_PREFIX_ES_OVERRIDE))
    RegsetAddReg(use, I386_REG_ES);
  else if (I386_INS_HAS_PREFIX(ins,I386_PREFIX_FS_OVERRIDE))
    RegsetAddReg(use, I386_REG_FS);
  else if (I386_INS_HAS_PREFIX(ins,I386_PREFIX_GS_OVERRIDE))
    RegsetAddReg(use, I386_REG_GS);
  else if (I386_INS_HAS_PREFIX(ins,I386_PREFIX_SS_OVERRIDE))
    RegsetAddReg(use, I386_REG_SS);
  else
  {
    t_i386_operand * op;
    /* instructions with memory operands use the DS register, except when the 
     * reference is relative to the %e{b,s}p register */
    op = I386_INS_DEST(ins);
    if (I386_OP_TYPE(op) == i386_optype_mem)
    {
      if (I386_OP_BASE(op) == I386_REG_EBP || I386_OP_BASE(op) == I386_REG_ESP)
	RegsetAddReg(use, I386_REG_SS);
      else
	RegsetAddReg(use, I386_REG_DS);
    }
    op = I386_INS_SOURCE1(ins);
    if (I386_OP_TYPE(op) == i386_optype_mem)
    {
      if (I386_OP_BASE(op) == I386_REG_EBP || I386_OP_BASE(op) == I386_REG_ESP)
	RegsetAddReg(use, I386_REG_SS);
      else
	RegsetAddReg(use, I386_REG_DS);
    }
    op = I386_INS_SOURCE2(ins);
    if (I386_OP_TYPE(op) == i386_optype_mem)
    {
      if (I386_OP_BASE(op) == I386_REG_EBP || I386_OP_BASE(op) == I386_REG_ESP)
	RegsetAddReg(use, I386_REG_SS);
      else
	RegsetAddReg(use, I386_REG_DS);
    }
  }
  
  if (I386_INS_OPCODE(ins) == I386_MOVSB || I386_INS_OPCODE(ins) == I386_MOVSD ||
      I386_INS_OPCODE(ins) == I386_SCASB || I386_INS_OPCODE(ins) == I386_SCASD ||
      I386_INS_OPCODE(ins) == I386_STOSB || I386_INS_OPCODE(ins) == I386_STOSD ||
      I386_INS_OPCODE(ins) == I386_CMPSB || I386_INS_OPCODE(ins) == I386_CMPSD ||
      I386_INS_OPCODE(ins) == I386_INSB  || I386_INS_OPCODE(ins) == I386_INSD  )
    RegsetAddReg(use, I386_REG_ES);

  if (I386_INS_OPCODE(ins) == I386_PUSH || I386_INS_OPCODE(ins) == I386_POP ||
      I386_INS_OPCODE(ins) == I386_PUSHA || I386_INS_OPCODE(ins) == I386_POPA ||
      I386_INS_OPCODE(ins) == I386_PUSHF || I386_INS_OPCODE(ins) == I386_POPF )
    RegsetAddReg(use, I386_REG_SS);
#endif
  /* }}}*/
  
  /* {{{ partially defined registers */
  /* if the destination of the instruction is a partial register,
   * the destination register is also used (to preserve the 
   * unaffected bits of the register) */
  op = I386_INS_DEST(ins);
  if (I386_OP_TYPE(op) == i386_optype_reg && 
      I386IsGeneralPurposeReg(I386_OP_BASE(op)) && 
      I386_OP_REGMODE(op) != i386_regmode_full32)
  {
    RegsetAddReg(use,I386_OP_BASE(op));
  }
  if (I386_INS_HAS_FLAG(ins, I386_IF_SOURCE1_DEF))
  {
    op = I386_INS_SOURCE1(ins);
    if (I386_OP_TYPE(op) == i386_optype_reg && 
	I386IsGeneralPurposeReg(I386_OP_BASE(op)) && 
	I386_OP_REGMODE(op) != i386_regmode_full32)
    {
      RegsetAddReg(use,I386_OP_BASE(op));
    }
  }
  if (I386_INS_HAS_FLAG(ins, I386_IF_SOURCE2_DEF))
  {
    op = I386_INS_SOURCE2(ins);
    if (I386_OP_TYPE(op) == i386_optype_reg && 
	I386IsGeneralPurposeReg(I386_OP_BASE(op)) && 
	I386_OP_REGMODE(op) != i386_regmode_full32)
    {
      RegsetAddReg(use,I386_OP_BASE(op));
    }
  }
  /* }}} */

  /* {{{ fp instructions that push or pop the fp stack */
  if (i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pops != 0)
  {
    int i;
    for (i = 0; i < 8-i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pops; i++)
      RegsetAddReg(use, I386_REG_ST0 + i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pops + i);
  }
  if (i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pushes != 0)
  {
    int i;
    for (i = 0; i < 8-i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pushes; i++)
      RegsetAddReg(use, I386_REG_ST0 + i);
  }
  /* }}} */

  /* special cases */
  switch (I386_INS_OPCODE(ins))
  {
    /* {{{ push and pop implicitly use the stack pointer */
    case I386_PUSH:
    case I386_PUSHA:
    case I386_PUSHF:
    case I386_POP:
    case I386_POPA:
    case I386_POPF:
      RegsetAddReg(use, I386_REG_ESP);
      /* pusha uses all general purpose registers */
      if (I386_INS_OPCODE(ins) == I386_PUSHA)
      {
	RegsetAddReg(use, I386_REG_EAX);
	RegsetAddReg(use, I386_REG_EBX);
	RegsetAddReg(use, I386_REG_ECX);
	RegsetAddReg(use, I386_REG_EDX);
	RegsetAddReg(use, I386_REG_ESI);
	RegsetAddReg(use, I386_REG_EDI);
	RegsetAddReg(use, I386_REG_ESP);
	RegsetAddReg(use, I386_REG_EBP);
      }
      break;
      /* }}} */

    /* {{{ call and return instructions use the stack pointer */
    case I386_CALL:
    case I386_CALLF:
    case I386_RET:
    case I386_RETF:
    case I386_IRET:
      RegsetAddReg(use, I386_REG_ESP);
      break;
      /* }}} */

    /* {{{ enter and leave instructions use %esp and %ebp */
    case I386_ENTER:
      RegsetAddReg(use, I386_REG_ESP);
      RegsetAddReg(use, I386_REG_EBP);
      break;
    case I386_LEAVE:
      RegsetAddReg(use, I386_REG_EBP);
      break;
      /* }}}*/
    
    /* {{{ div and idiv instructions use edx */
    case I386_DIV:
    case I386_IDIV:
      if (I386_OP_REGMODE(I386_INS_DEST(ins)) == i386_regmode_full32 || 
	  I386_OP_REGMODE(I386_INS_DEST(ins)) == i386_regmode_lo16 )
	RegsetAddReg(use, I386_REG_EDX);
      break;
      /* }}}*/

    case I386_INT:
    case I386_INTO:
    case I386_INT3:
      RegsetAddReg(use, I386_REG_EAX);
      RegsetAddReg(use, I386_REG_EBX);
      RegsetAddReg(use, I386_REG_ECX);
      RegsetAddReg(use, I386_REG_EDX);
      RegsetAddReg(use, I386_REG_ESI);
      RegsetAddReg(use, I386_REG_EDI);
      RegsetAddReg(use, I386_REG_ESP);
      RegsetAddReg(use, I386_REG_EBP);
      break;

    /* jecxz and loop(z/nz) use the %ecx register */
    case I386_JECXZ:
    case I386_LOOP:
    case I386_LOOPZ:
    case I386_LOOPNZ:
      RegsetAddReg(use, I386_REG_ECX);
      break;
    /*cmpxchg uses eax*/
    case I386_CMPXCHG:
      RegsetAddReg(use, I386_REG_EAX);
      break;

    default:
      break;
  }

  return use;
} /*}}} */

/* {{{ I386InsDefinedRegisters */
t_regset I386InsDefinedRegisters(t_ins * ins)
{
  t_regset def = NullRegs;

  if (INS_TYPE(ins) == IT_DATA)
    return def;

  /* defined condition flags */
  RegsetSetUnion(def, i386_opcode_table[I386_INS_OPCODE(ins)].cf_defined);

  /* if the destination operand is a register, this register is, of course, defined */
  if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg)
    RegsetAddReg(def, I386_OP_BASE(I386_INS_DEST(ins)));

  /* source operands that are defined as well */
  if (I386_INS_HAS_FLAG(ins, I386_IF_SOURCE1_DEF))
    if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg)
      RegsetAddReg(def, I386_OP_BASE(I386_INS_SOURCE1(ins)));
  if (I386_INS_HAS_FLAG(ins, I386_IF_SOURCE2_DEF))
    if (I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_reg)
      RegsetAddReg(def, I386_OP_BASE(I386_INS_SOURCE2(ins)));

  /* the REP/REPNZ prefixes */
  if (I386_INS_HAS_PREFIX(ins, I386_PREFIX_REP) ||
      I386_INS_HAS_PREFIX(ins, I386_PREFIX_REPNZ))
  {
    RegsetAddReg(def, I386_REG_ECX);
  }

  /* floating point instructions that push or pop the stack
   * redefine the fpu registers */
  if (i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pops != 0)
  {
    int i;
    for (i = 0; i < 8-i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pops; i++)
      RegsetAddReg(def, I386_REG_ST0 + i);
  }
  if (i386_opcode_table[I386_INS_OPCODE(ins)].fpstack_pushes != 0)
  {
    int i;
    for (i = 0; i < 8; i++)
      RegsetAddReg(def, I386_REG_ST0 + i);
  }

  /* special cases */
  switch (I386_INS_OPCODE(ins))
  {
    /* {{{ 1. string instructions: these define %edi and %esi */
    case I386_MOVSB: case I386_MOVSD:
    case I386_CMPSB: case I386_CMPSD:
      RegsetAddReg(def, I386_REG_EDI);
      RegsetAddReg(def, I386_REG_ESI);
      break;
    case I386_LODSB: case I386_LODSD:
    case I386_OUTSB: case I386_OUTSD:
      RegsetAddReg(def, I386_REG_ESI);
      break;
    case I386_STOSB: case I386_STOSD:
    case I386_SCASB: case I386_SCASD:
    case I386_INSB:  case I386_INSD:
      RegsetAddReg(def, I386_REG_EDI);
      break;
      /* }}}*/

    /* {{{ 2. call and return instructions change the stack pointer */
    case I386_CALL:
    case I386_CALLF:
    case I386_RET:
    case I386_RETF:
    case I386_IRET:
      RegsetAddReg(def, I386_REG_ESP);
      break;
      /* }}} */
      
    /* {{{ 3. enter and leave instructions change %esp and %ebp */
    case I386_ENTER:
    case I386_LEAVE:
      RegsetAddReg(def, I386_REG_ESP);
      RegsetAddReg(def, I386_REG_EBP);
      break;
      /* }}}*/
      
    /* 4. {{{ mul, imul, div and idiv implicitly define %edx
     * if executed with 16-bit or 32-bit operands */
    case I386_MUL:
    case I386_IMUL:
    case I386_DIV:
    case I386_IDIV:
      if (I386_OP_REGMODE(I386_INS_DEST(ins)) == i386_regmode_full32 || 
	  I386_OP_REGMODE(I386_INS_DEST(ins)) == i386_regmode_lo16 )
	RegsetAddReg(def, I386_REG_EDX);
      break;
      /* }}} */

    /* 5. fstsw modifies the %eax register */
    case I386_FSTSW:
      RegsetAddReg(def, I386_REG_EAX);
      break;

    /* 6. {{{ push and pop implicitly define the stack pointer */
    case I386_PUSH:
    case I386_PUSHA:
    case I386_PUSHF:
    case I386_POP:
    case I386_POPA:
    case I386_POPF:
      RegsetAddReg(def, I386_REG_ESP);

      /* POPA defines all general purpose registers */
      if (I386_INS_OPCODE(ins) == I386_POPA)
      {
	RegsetAddReg(def, I386_REG_EAX);
	RegsetAddReg(def, I386_REG_EBX);
	RegsetAddReg(def, I386_REG_ECX);
	RegsetAddReg(def, I386_REG_EDX);
	RegsetAddReg(def, I386_REG_ESI);
	RegsetAddReg(def, I386_REG_EDI);
	RegsetAddReg(def, I386_REG_ESP);
	RegsetAddReg(def, I386_REG_EBP);
      }
      break;
      /* }}} */

    case I386_INT:
    case I386_INTO:
    case I386_INT3:
      /* syscalls define %eax */
      RegsetAddReg(def, I386_REG_EAX);
      break;

    case I386_LOOP:
    case I386_LOOPZ:
    case I386_LOOPNZ:
      /* loop instructions define ecx */
      RegsetAddReg(def, I386_REG_ECX);
      break;

    /* calling conventions state that any function call (and consequently all pseudo calls)
     * may change %eax, %ecx and %edx */
    case I386_PSEUDOCALL:
      RegsetAddReg(def, I386_REG_EAX);
      RegsetAddReg(def, I386_REG_ECX);
      RegsetAddReg(def, I386_REG_EDX);
      break;
    case I386_CMPXCHG:
      RegsetAddReg(def, I386_REG_EAX);
      break;

    default:
      break;
  }
  return def;
} /* }}} */

t_address I386InsGetSize(t_ins * ins)
{
  t_uint8 buf[15];
  t_address ret;
  t_uint32 len;
  len = I386AssembleIns(ins,buf);
  GenericSet_uint32(ret,len);
  return ret;
}

/** {{{ Get first relocated operand */
t_i386_operand * I386InsGetFirstRelocatedOp(t_ins * ins)
{
  if (I386_OP_FLAGS(I386_INS_DEST(ins)) & I386_OPFLAG_ISRELOCATED)
    return I386_INS_DEST(ins);
  else if (I386_OP_FLAGS(I386_INS_SOURCE1(ins)) & I386_OPFLAG_ISRELOCATED)
    return I386_INS_SOURCE1(ins);
  else if (I386_OP_FLAGS(I386_INS_SOURCE2(ins)) & I386_OPFLAG_ISRELOCATED)
    return I386_INS_SOURCE2(ins);
  else FATAL(("Could not find a relocated operand"));
  return NULL;
} /* }}} */

/** {{{ Get second relocated operand */
t_i386_operand * I386InsGetSecondRelocatedOp(t_ins * ins)
{
  t_bool foundfirst = FALSE;

  if (I386_OP_FLAGS(I386_INS_DEST(ins)) & I386_OPFLAG_ISRELOCATED)
    foundfirst = TRUE;

  if (I386_OP_FLAGS(I386_INS_SOURCE1(ins)) & I386_OPFLAG_ISRELOCATED)
  {
    if (foundfirst)
      return I386_INS_SOURCE1(ins);
    else
      foundfirst = TRUE;
  }

  if (I386_OP_FLAGS(I386_INS_SOURCE2(ins)) & I386_OPFLAG_ISRELOCATED)
  {
    if (foundfirst)
      return I386_INS_SOURCE2(ins);
    else 
      FATAL(("found only one relocated operand"));
  }

  FATAL(("not enough relocated operands"));
  return NULL;
} /* }}} */

/** {{{ Get reloc for operand */
t_reloc * I386GetRelocForOp(t_ins * ins, t_i386_operand * op)
{
  t_i386_operand * ops[3];
  int i;
  t_bool second_relocated = FALSE;
  t_reloc_ref * rr, * rr2, * tmp;

  if (!(I386_OP_FLAGS(op) & I386_OPFLAG_ISRELOCATED)) return NULL;

  ops[0] = I386_INS_DEST(ins);
  ops[1] = I386_INS_SOURCE1(ins);
  ops[2] = I386_INS_SOURCE2(ins);
  
  for (i=0; i<3; i++)
  {
    if (ops[i] == op) break;
    if (I386_OP_FLAGS(ops[i]) & I386_OPFLAG_ISRELOCATED)
      second_relocated = TRUE;
  }

  rr = RELOCATABLE_REFERS_TO(ins);
  VERIFY_FASSERT(rr, ("need at least one reloc in this instruction: @I",ins));
  rr2 = rr->next;
  VERIFY_FASSERT(!second_relocated || rr2,("need two relocs in this instruction: @I",ins));

  if (rr2 && GENERIC_GT(rr->rel->from_offset,rr2->rel->from_offset))
  {
    /* swap */
    tmp = rr2;
    rr2 = rr;
    rr = tmp;
  }
  
  if (second_relocated) return rr2->rel;
  return rr->rel;
} /* }}} */

t_tristate I386IsSyscallExit(t_ins * ins)
{
  t_uint32 syscallno;
  
  if (I386_INS_OPCODE(ins) != I386_INT) return NO;
  if (I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) != 0x80) return NO;

  syscallno = I386GetSyscallNo(ins);
  /* store this syscall number in the instruction, this might be handy */
  I386_OP_IMMEDIATE(I386_INS_SOURCE2(ins)) = syscallno;
  
  if (syscallno == -1)
    return PERHAPS;
  else if (syscallno == 0x1 || syscallno == 0xfc)
    return YES;

  return NO;
}

t_uint32 I386GetSyscallNo(t_ins * ins)
{
  t_ins * iter;

  if (I386_INS_OPCODE(ins) != I386_INT) return -1;
  if (I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) != 0x80) return -1;

  /* find the instruction that defines %eax */
  for (iter = INS_PREV(ins); iter; iter = INS_PREV(iter))
  {
    if (RegsetIn(INS_REGS_DEF(iter), I386_REG_EAX))
      break;
  }
  if (!iter)
    return -1;

  if (I386_INS_OPCODE(iter) == I386_MOV && I386_OP_TYPE(I386_INS_SOURCE1(iter)) == i386_optype_imm)
    return I386_OP_IMMEDIATE(I386_INS_SOURCE1(iter));
  else
    FATAL(("Syscall @iB @I\n eax defined by @I\nIMPLEMENT!",INS_BBL(ins),ins,iter));

  return -1;
}

void I386OpSetReg(t_i386_operand * op, t_reg reg, t_i386_regmode mode)
{
  I386_OP_TYPE(op) = i386_optype_reg;
  I386_OP_BASE(op) = reg;
  I386_OP_REGMODE(op) = mode;
  I386_OP_FLAGS(op) = 0x0;
}

void I386OpSetImm(t_i386_operand * op, t_uint32 imm, t_uint32 immedsize)
{
  I386_OP_TYPE(op) = i386_optype_imm;
  I386_OP_IMMEDIATE(op) = imm;
  I386_OP_IMMEDSIZE(op) = immedsize;
  I386_OP_FLAGS(op) = 0x0;
}

void I386OpSetMem(t_i386_operand * op, t_uint32 offset, t_reg base, t_reg index, int scale, t_uint32 memopsize)
{
  I386_OP_TYPE(op) = i386_optype_mem;
  I386_OP_IMMEDIATE(op) = offset;
  I386_OP_IMMEDSIZE(op) = 4;
  I386_OP_MEMOPSIZE(op) = memopsize;
  I386_OP_BASE(op) = base;
  I386_OP_INDEX(op) = index;
  I386_OP_SCALE(op) = scale;
  I386_OP_FLAGS(op) = 0x0;
}

/* returns the memory operand of an instruction. if there is none, returns NULL */
t_i386_operand * I386InsGetMemLoadOp(t_ins * ins)
{

  if (I386_INS_FLAGS(ins) & I386_IF_DEST_IS_SOURCE)
    if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_mem)
      return I386_INS_DEST(ins);

  if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_mem)
    return I386_INS_SOURCE1(ins);

  if (I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_mem)
    return I386_INS_SOURCE2(ins);

  return NULL;
}

t_i386_operand * I386InsGetMemStoreOp(t_ins * ins)
{
  if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_mem)
    return I386_INS_DEST(ins);

  if (I386_INS_FLAGS(ins) & I386_IF_SOURCE1_DEF)
    if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_mem)
      return I386_INS_SOURCE1(ins);
  
  if (I386_INS_FLAGS(ins) & I386_IF_SOURCE2_DEF)
    if (I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_mem)
      return I386_INS_SOURCE2(ins);

  return NULL;
}

t_bool I386InsIsIndirectCall(t_ins * ins)
{
  return (I386_INS_OPCODE(ins) == I386_CALL) && (I386_OP_TYPE(I386_INS_SOURCE1(ins)) != i386_optype_imm);
}

t_bool I386InsIsProcedureCall(t_ins * ins)
{
  return I386_INS_OPCODE(ins) == I386_CALL;
}

void I386InsCleanup(t_ins * ins)
{
  return;
}

t_bool I386OpsAreIdentical(t_i386_operand * op1, t_i386_operand * op2)
{
  return 
    I386_OP_TYPE(op1)         == I386_OP_TYPE(op2) &&
    I386_OP_BASE(op1)         == I386_OP_BASE(op2)  &&
    I386_OP_INDEX(op1)        == I386_OP_INDEX(op2)  &&
    I386_OP_IMMEDIATE(op1)    == I386_OP_IMMEDIATE(op2) &&
    I386_OP_SEGSELECTOR(op1)  == I386_OP_SEGSELECTOR(op2) &&
    I386_OP_SCALE(op1)        == I386_OP_SCALE(op2) &&
    I386_OP_REGMODE(op1)      == I386_OP_REGMODE(op2) &&
    I386_OP_IMMEDSIZE(op1)    == I386_OP_IMMEDSIZE(op2)  &&
    I386_OP_MEMOPSIZE(op1)    == I386_OP_MEMOPSIZE(op2) &&
    I386_OP_FLAGS(op1)        == I386_OP_FLAGS(op2); 
}

t_bool I386InsAreIdentical(t_ins * ins1, t_ins * ins2)
{
  return  
    I386_INS_OPCODE(ins1) == I386_INS_OPCODE(ins2) &&
    I386_INS_DATA(ins1)	==I386_INS_DATA(ins2) &&
    I386_INS_PREFIXES(ins1) ==	I386_INS_PREFIXES(ins2)	&&
    I386_INS_FLAGS(ins1) == I386_INS_FLAGS(ins2) &&
    I386OpsAreIdentical(I386_INS_DEST(ins1),I386_INS_DEST(ins2)) &&
    I386OpsAreIdentical(I386_INS_SOURCE1(ins1),I386_INS_SOURCE1(ins2)) &&
    I386OpsAreIdentical(I386_INS_SOURCE2(ins1),I386_INS_SOURCE2(ins2)) &&
    I386_INS_CONDITION(ins1) ==	I386_INS_CONDITION(ins2);
}

t_uint32 I386BblHashKey(t_bbl * bbl)
{
  t_ins * ins;
  t_uint32 return_value = 0;
  if (BBL_NINS(bbl)<3) return 0;
  ins = BBL_INS_FIRST(bbl);
  return_value|=I386_INS_OPCODE(ins);
  return_value<<=3;
  return_value|=I386_OP_TYPE(I386_INS_DEST(ins));
  return_value<<=3;
  return_value|=I386_OP_TYPE(I386_INS_SOURCE1(ins));
  return_value<<=3;
  ins = INS_NEXT(ins);
  return_value|=I386_INS_OPCODE(ins);
  return_value<<=3;
  return_value|=I386_OP_TYPE(I386_INS_DEST(ins));
  return_value<<=3;
  return_value|=I386_OP_TYPE(I386_INS_SOURCE1(ins));
  return_value<<=3;
  ins = INS_NEXT(ins);
  return_value|=I386_INS_OPCODE(ins);
  return_value<<=3;
  return_value|=I386_OP_TYPE(I386_INS_DEST(ins));
  return_value<<=3;
  return_value|=I386_OP_TYPE(I386_INS_SOURCE1(ins));

  return return_value;
}
/* vim: set shiftwidth=2 foldmethod=marker: */
