/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef I386_INS_TYPEDEFS
#define I386_INS_TYPEDEFS
typedef struct _t_i386_ins t_i386_ins;
typedef struct _t_i386_operand t_i386_operand;
#endif

#ifndef I386_INSTRUCTION_H
#define I386_INSTRUCTION_H

#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_regset.h"
#include "i386_opcode.h"
#include "i386_table.h"
	
typedef enum {
  i386_optype_none = 0,    /* no operand */
  i386_optype_reg,         /* register only */
  i386_optype_imm,         /* immediate only */
  i386_optype_mem,         /* scale-index-base-displacement */
  i386_optype_farptr,	   /* 48-bit memory pointer (16-bit segment selector + 32-bit offset) */
  i386_optype_invalid	 
} t_i386_optype;

typedef enum {
  i386_regmode_full32 = 0, /* the full register, e.g. %eax */
  i386_regmode_lo16,       /* low 16 bits, e.g. %ax */
  i386_regmode_lo8,        /* low 8 bits, e.g. %al */
  i386_regmode_hi8         /* bits 8-15, e.g. %ah */
} t_i386_regmode;

struct _t_i386_operand {
  /** defines the type of operand (not all types are possible for all instructions) */
  t_i386_optype type:8;
  /** operand flags */
  t_uint8 flags;
  /** scale for sib type */
  t_uint8 scale;
  /** register mode (8, 16 or 32 bits) */
  t_i386_regmode regmode:8;
  /** base register for sib or plain_reg types */
  t_reg base;
  /** index register for sib type */
  t_reg index;
  /** immediate for plain_immed, displacement for sib and 32-bit offset for far_ptr */
  t_uint32 immediate;
  /** segment selector for far_ptr */
  t_uint16 segment_selector;
  /** size in bytes of the immediate field */
  t_uint8 immed_size;
  /** memory operand size (1, 2 or 4 bytes) */
  t_uint8 memop_size;
  /*superoptimizer*/
  t_uint8 super;
};

#define I386_OP_TYPE(op)         (op)->type
#define I386_OP_BASE(op)         (op)->base
#define I386_OP_INDEX(op)        (op)->index
#define I386_OP_IMMEDIATE(op)    (op)->immediate
#define I386_OP_SEGSELECTOR(op)  (op)->segment_selector
#define I386_OP_SCALE(op)        (op)->scale
#define I386_OP_REGMODE(op)      (op)->regmode
#define I386_OP_IMMEDSIZE(op)    (op)->immed_size
#define I386_OP_MEMOPSIZE(op)    (op)->memop_size
#define I386_OP_FLAGS(op)        (op)->flags

/** instruction operand flags */
#define I386_OPFLAG_ISRELOCATED       (1 << 0)

/** register scale possibilities for sib operands */
#define I386_SCALE_1         0
#define I386_SCALE_2         1
#define I386_SCALE_4         2
#define I386_SCALE_8         3
#define I386_SCALE_INVALID   255

/* operand iterator */
#define I386_INS_FOREACH_OP(ins,op)	\
	for (op = I386_INS_DEST(ins); op; op = (op == I386_INS_SOURCE2(ins) ? NULL : op+1))


/* operand iterator */
#define I386_INS_FOREACH_OP(ins,op)	\
	for (op = I386_INS_DEST(ins); op; op = (op == I386_INS_SOURCE2(ins) ? NULL : op+1))

typedef enum _t_i386_condition_code
{
  I386_CONDITION_O  = 0,
  I386_CONDITION_NO = 1,
  I386_CONDITION_B  = 2,
  I386_CONDITION_AE = 3,
  I386_CONDITION_Z  = 4,
  I386_CONDITION_NZ = 5,
  I386_CONDITION_BE = 6,
  I386_CONDITION_A  = 7,
  I386_CONDITION_S  = 8,
  I386_CONDITION_NS = 9,
  I386_CONDITION_P  = 10,
  I386_CONDITION_NP = 11,
  I386_CONDITION_L  = 12,
  I386_CONDITION_GE = 13,
  I386_CONDITION_LE = 14,
  I386_CONDITION_G  = 15,
  I386_CONDITION_LOOP = 251,    /* ecx != 0 */
  I386_CONDITION_LOOPZ = 252,   /* ecx != 0 &&  ZF */
  I386_CONDITION_LOOPNZ = 253,  /* ecx != 0 && !ZF */
  I386_CONDITION_ECXZ = 254,    /* ecx == 0 */
  I386_CONDITION_NONE = 255
} t_i386_condition_code;

/** i386 instruction representation */
struct _t_i386_ins {
  /** base class */
  t_ins generic;

  /** instruction opcode (internal diablo representation) */
  t_i386_opcode opcode;
  /** source and destination operands */
  t_i386_operand ops[3];
  /** address producer info - useful for instrumentation */
  t_i386_ins * ap_original;
  /** bit field indicating all instruction prefixes */
  t_uint16 prefixes;
  /** i386-specific instruction flags */
  t_uint16 flags;
  /** condition code for conditional instructions */
  t_i386_condition_code condition:8;
  /** if this structure represents data in the code section,
   * this field contains the data */
  t_uint8 data;
#ifdef OBFUSCATION_SUPPORT
  /* obfuscation !! */
  t_i386_ins * corr;
#endif
#ifdef STEGO
  t_uint8 encoding;
  t_uint8 opcode_byte;
#endif
};

#define T_I386_INS(ins)		((t_i386_ins *)(ins))

/** field selectors */
#define I386_INS_OPCODE(ins)		(T_I386_INS(ins)->opcode)
#define I386_INS_DATA(ins)		(T_I386_INS(ins)->data)
#define I386_INS_PREFIXES(ins)		(T_I386_INS(ins)->prefixes)
#define I386_INS_FLAGS(ins)		(T_I386_INS(ins)->flags)
#define I386_INS_DEST(ins)		(&(T_I386_INS(ins)->ops[0]))
#define I386_INS_SOURCE1(ins)		(&(T_I386_INS(ins)->ops[1]))
#define I386_INS_SOURCE2(ins)		(&(T_I386_INS(ins)->ops[2]))
#define I386_INS_CONDITION(ins)		(T_I386_INS(ins)->condition)
#ifdef OBFUSCATION_SUPPORT
#define I386_INS_CORR(ins)		(T_I386_INS(ins)->corr)
#endif


/* shortcuts */
#define I386_INS_HAS_FLAG(ins,flag)	(I386_INS_FLAGS(ins) & (flag))
#define I386_INS_HAS_PREFIX(ins,pref)   (I386_INS_PREFIXES(ins) & (pref))
#define I386_ADSZPREF(ins)	 	I386_INS_HAS_PREFIX(ins,I386_PREFIX_ADDRSIZE_OVERRIDE)
#define I386_OPSZPREF(ins)	 	I386_INS_HAS_PREFIX(ins,I386_PREFIX_OPSIZE_OVERRIDE)
#define I386_INS_SET_FLAG(ins, flag)	(I386_INS_FLAGS(ins) |= (flag))
#define I386_INS_FLIP_FLAG(ins, flag)	(I386_INS_FLAGS(ins) ^= (flag))

/** possible instruction prefixes */
#define I386_PREFIX_LOCK	(1 << 0)
#define I386_PREFIX_REPNZ	(1 << 1)
#define I386_PREFIX_REP		(1 << 2)
#define I386_PREFIX_REPZ	I386_PREFIX_REP
#define I386_PREFIX_REPE	I386_PREFIX_REP
#define I386_PREFIX_REPNE	I386_PREFIX_REPNZ

#define I386_PREFIX_CS_OVERRIDE	(1 << 3)
#define I386_PREFIX_DS_OVERRIDE	(1 << 4)
#define I386_PREFIX_ES_OVERRIDE	(1 << 5)
#define I386_PREFIX_FS_OVERRIDE	(1 << 6)
#define I386_PREFIX_GS_OVERRIDE	(1 << 7)
#define I386_PREFIX_SS_OVERRIDE	(1 << 8)

#define I386_PREFIX_OPSIZE_OVERRIDE	(1 << 9)
#define I386_PREFIX_ADDRSIZE_OVERRIDE	(1 << 10)

/** instruction flags */
#define I386_IF_DEST_IS_SOURCE		(1 << 0)
#define I386_IF_SOURCE1_DEF		(1 << 1)
#define I386_IF_SOURCE2_DEF		(1 << 2)

/** function prototypes */
t_bool I386InsIsConditional(t_ins * ins);
t_bool I386InsHasSideEffect(t_ins * ins);
t_bool I386InsIsStore(t_ins * ins);
t_bool I386InsIsLoad(t_ins * ins);
t_bool I386InsIsControlTransfer(t_ins * ins);
t_bool I386InsIsSystemControlTransfer(t_ins * ins);
t_bool I386InsIsRegularControlTransfer(t_ins * ins);
t_bool I386InsIsIndirectCall(t_ins * ins);
t_tristate I386IsSyscallExit(t_ins * ins);
t_uint32 I386GetSyscallNo(t_ins * ins);


void I386InstructionMakeNoop(t_ins * ins);
void I386InstructionMakeCall(t_ins * ins);
void I386InstructionMakeJump(t_ins * ins);
void I386InstructionMakeJumpReg(t_ins * ins, t_reg src);
void I386InstructionMakeCondJump(t_ins * ins, t_i386_condition_code condition);
void I386InstructionMakePush(t_ins * ins, t_reg reg, t_uint32 immval);
void I386InstructionMakePushMem(t_ins * ins, t_uint32 offset, t_reg base, t_reg index, int scale);
void I386InstructionMakePop(t_ins * ins, t_reg reg);
void I386InstructionMakeArithmetic(t_ins * ins, t_i386_opcode opc, t_reg dest, t_reg src, t_uint32 imm);
void I386InstructionMakeArithmeticToMem(t_ins * ins, t_i386_opcode opc, t_uint32 offset, t_reg base, t_reg index, int scale, t_reg src, t_uint32 imm);
void I386InstructionMakeArithmeticFromMem(t_ins * ins, t_i386_opcode opc, t_reg dest, t_uint32 offset, t_reg base, t_reg index, int scale);
void I386InstructionMakeMovToMem(t_ins * ins, t_uint32 offset, t_reg base, t_reg index, int scale, t_reg src, t_uint32 imm);
void I386InstructionMakeMovToMemLen(t_ins * ins, t_uint32 offset, t_reg base, t_reg index, int scale, t_reg src, t_uint32 imm, t_uint32 imm_len);
void I386InstructionMakeMovFromMem(t_ins * ins, t_reg dest, t_uint32 offset, t_reg base, t_reg index, int scale);
void I386InstructionMakeMovToReg(t_ins * ins, t_reg dest, t_reg src, t_uint32 imm);
void I386InstructionMakeRM32IMM32(t_ins * ins, t_reg dest, t_reg src, t_uint32 imm, t_i386_opcode opcode);
void I386InstructionMakeRM32IMM8(t_ins * ins, t_reg dest, t_reg src, t_uint32 imm, t_i386_opcode opcode);
void I386InstructionMakeLea(t_ins * ins, t_reg dest, t_uint32 offset, t_reg base, t_reg index, int scale);
void I386InstructionMakeCmp(t_ins * ins, t_reg reg, t_reg cmpreg, t_uint32 cmpimm);
void I386InstructionMakeSetcc(t_ins * ins, t_i386_condition_code cond, t_uint32 offset, t_reg base, t_reg index, int scale, t_bool memop);
void I386InstructionMakePseudoCall(t_ins * ins, t_function * to);
void I386InstructionMakePseudoLoad(t_ins * ins, t_reg reg);
void I386InstructionMakePseudoSave(t_ins * ins, t_reg reg);
void I386InstructionMakeSimple(t_ins * ins, t_i386_opcode opc);
void I386InstructionMakeLSahf(t_ins * ins, t_i386_opcode opc);
void I386InstructionMakeFSaveRstor(t_ins * ins, t_i386_opcode opc, t_uint32 offset, t_reg base, t_reg index, int scale);
void I386InstructionMakeIncDec(t_ins * ins, t_i386_opcode opc, t_reg dest);
void I386InstructionMakeTest(t_ins * ins, t_reg src1, t_reg src2, t_uint32 imm);

t_regset I386InsUsedRegisters(t_ins * ins);
t_regset I386InsDefinedRegisters(t_ins * ins);
t_address I386InsGetSize(t_ins * ins);
t_i386_operand * I386InsGetFirstRelocatedOp(t_ins * ins);
t_i386_operand * I386InsGetSecondRelocatedOp(t_ins * ins);
t_i386_operand * I386InsGetMemLoadOp(t_ins * ins);
t_i386_operand * I386InsGetMemStoreOp(t_ins * ins);
t_reloc * I386GetRelocForOp(t_ins * ins, t_i386_operand * op);

void I386OpSetReg(t_i386_operand * op, t_reg reg, t_i386_regmode mode);
void I386OpSetImm(t_i386_operand * op, t_uint32 imm, t_uint32 immedsize);
void I386OpSetMem(t_i386_operand * op, t_uint32 offset, t_reg base, t_reg index, int scale, t_uint32 memopsize);

/* convenient shortcut when dynamically generating code.
 * e.g. I386MakeInsForBbl(Noop,Append,ins,bbl) will append a noop to bbl */
#define I386MakeInsForBbl(type,where,ins,bbl, ...)    \
do {                                                  \
    ins = InsNewForBbl(bbl);                          \
    I386InstructionMake ## type(ins, ## __VA_ARGS__); \
    Ins ## where ## ToBbl(ins,bbl);                   \
} while (0)


#define I386MakeInsForIns(type,where,ins,existing, ...) \
do {                                                    \
    ins = InsNewForBbl(INS_BBL(existing));              \
    I386InstructionMake ## type(ins, ## __VA_ARGS__);   \
    InsInsert ## where (ins,existing);                  \
} while (0)

void I386InsCleanup(t_ins * ins);
t_bool I386InsIsProcedureCall(t_ins * ins);
#endif
