%{
/*
 * A2I
 * Copyright (c) 1998 Lacroix Pascal <placr@hotmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:

* 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)

* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/* 
 * Please contact me if you find bugs or if you want to improve A2I.
 * Lacroix Pascal <placr@hotmail.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../../../utils/util_error.h"
#include "i386_a2i.h"

#define VERSION "0.4"

/* forward declarations */

void warning(char *);
void fatal();

void yyerror(char *);

void mov_sd(const char*, const char*, const char*);

/* command line switches */
int no_info = 0/*, opt = 0*/;
long baseaddr = 0;

char * string_pointer;
char output_string[250];

/* buffer used for strings and IDs */
char strbuf[512];
long strbuflen;

/* symbol structure for externals */
struct symbol {
  char *name;
  char isextern;
  char islocal;  
  struct symbol *next;
};

#define HASH_BITS 32
#define HASH_SIZE 509

static struct symbol *hashtab[HASH_SIZE];

static unsigned int hash(char *name) {
	unsigned int hash = 0;

while (*name) {
		hash = hash << 3 | hash >>( HASH_BITS - 3);
		hash += *name++;
	}

return hash % HASH_SIZE;
}

struct symbol *addexternal(char *);
void killexternal(char *s);
struct symbol *sym_lookup(char *);

void *xmalloc(int size)
{
  void *p = (void *) malloc(size);
  if (!p) {
    printf("error: not enough memory\n");
    exit(0);
  }
  return p;
}

%}

%union {
  char *str;
  long type;
  long num;
};

%token <str> ID STRING SEGMENT REG ONEBYTE STR JMP SET ARITH LEA INC CLTD
%token <str> PUSH IMUL IDIV TRANSLAT NEG DIV LSEG LGDT MOV_SD

%token _FILE ASCII GLOBAL _LONG _WORD _BYTE IDENT
%token ALIGN P2ALIGN COMM LCOMM SPACE CALL INT _VERSION
%token SECTION UNSUPPORTED UNSUP_2 LOCAL

%token <str> ST MATH

%token <num> NUMBER

%type <str> memref1 memref scale offset memptr varptr arg math_reg

%{

#define tDWORD 1
#define tWORD 2
#define tBYTE 3

/* return the correct instruction without 'l', 'w' or 'b' at the end */
static char *op(const char *s) {
  static char tmp[50];
  strcpy(tmp, s);
  tmp[strlen(tmp)-1] = 0;
  return strdup(tmp);
}

/* get instruction size: dword, word or byte */
static char *op_size(const char *s) {
	int c;
	if (strlen(s)>1) c = s[strlen(s)-1];
	else c = 0;

if (c == 'l')  return "dword";
	else
	if (c == 'w') return "word";
	else
	if (c == 'b') return "byte";
	else
	{ fprintf(stderr, "Internal error: op_size=%c (in %s line %d)\n", c, __FILE__, __LINE__);
	  yyerror("unknown instruction");
	  return "[error]";
	}
}

/* convert a number to a string but always with a sign (+ or -) */
static char *signed_num(const long x) {
	static char s[16];
	sprintf(s, "%s%ld", (x>=0 ? "+" : ""), x);
	return strdup(s);
}

/* useful for jumps */
static char *is_near(const char *s) {
	if (strcmp(s, "jmp") == 0) return ""; else return " near";
}


static char *translate_instr(char *instr) {
    char s[20];
    strcpy(s, instr);
    if (s[strlen(s)-1] == 'l') {
	s[strlen(s)-1] = 'd';
    }
    else {
	char buf[100];
	sprintf(buf, "translating %s to %s", instr, s);
	warning(buf);
    }

return strdup(s);
}

static char *translate_math(char *instr) {
    char s[30], c;
    strcpy(s, instr);
    c = s[strlen(s)-1];
    if (c=='s') {
	s[strlen(s)-1] = 0;
	strcat(s, " dword");
	return strdup(s);
    }
    else
    if (c=='l') {
	s[strlen(s)-1] = 0;
	strcat(s, " qword");
	return strdup(s);
    }
    else
    if (c=='p') {
	return strdup(s);
    }
    else warning("unknown math instruction");

return instr;
}

struct tok {
	char *name;
	long token;
};

struct tok tokens[] = {
    {".file", _FILE},
    {".text", SEGMENT},
    {".data", SEGMENT},
    {".bss", SEGMENT},
    {".section", SECTION},
    {".ascii", ASCII},
    {".string", ASCII},
    {".globl", GLOBAL},
    {".version", _VERSION},
    {".local", LOCAL},

{".long", _LONG},
    {".word", _WORD},
    {".byte", _BYTE},

{".p2align", P2ALIGN},
    {".align", ALIGN},
    {".space", SPACE},
    {".zero", SPACE},

{"pushl", PUSH},
    {"pushw", PUSH},

{"popl", PUSH},
    {"popw", PUSH},

{"movl", ARITH},
    {"movw", ARITH},
    {"movb", ARITH},

{"xorl", ARITH},
    {"xorw", ARITH},
    {"xorb", ARITH},

{"call", CALL},

{"addl", ARITH},
    {"addw", ARITH},
    {"addb", ARITH},

{"subl", ARITH},
    {"subw", ARITH},
    {"subb", ARITH},

{"cmpl", ARITH},
    {"cmpw", ARITH},
    {"cmpb", ARITH},

{"leal", LEA},
    {"leaw", LEA},

{"incl", INC},
    {"incw", INC},
    {"incb", INC},

{"decl", INC},
    {"decw", INC},
    {"decb", INC},

{"ja", JMP},
    {"jae", JMP},
    {"jb", JMP},
    {"jbe", JMP},
    {"jc", JMP},
    {"jcxz", JMP},
    {"je", JMP},
    {"jg", JMP},
    {"jge", JMP},
    {"jl", JMP},
    {"jle", JMP},
    {"jmp", JMP},
    {"jna", JMP},
    {"jnae", JMP},
    {"jnb", JMP},
    {"jnbe", JMP},
    {"jnc", JMP},
    {"jne", JMP},
    {"jng", JMP},
    {"jnge", JMP},
    {"jnl", JMP},
    {"jnle", JMP},
    {"jno", JMP},
    {"jnp", JMP},
    {"jns", JMP},
    {"jnz", JMP},
    {"jo", JMP},
    {"jp", JMP},
    {"jpe", JMP},
    {"jpo", JMP},
    {"js", JMP},
    {"jz", JMP},
    {"jcxz", JMP},
    {"jecxz", JMP},
    {"loop", JMP},
    {"loopz", JMP},
    {"loope", JMP},
    {"loopnz", JMP},
    {"loopne", JMP},


{"sarl", ARITH},
    {"sarw", ARITH},

{"sall", ARITH},
    {"salw", ARITH},

{"movswl", MOV_SD},
    {"movslw", MOV_SD},
    {"movsbl", MOV_SD},
    {"movsbw", MOV_SD},

{"movzwl", MOV_SD},
    {"movzlw", MOV_SD},
    {"movzbl", MOV_SD},
    {"movzbw", MOV_SD},

{"testl", ARITH},
    {"testw", ARITH},
    {"testb", ARITH},

{"negl", NEG},
    {"negw", NEG},
    {"negb", NEG},

{"notl", NEG},
    {"notw", NEG},
    {"notb", NEG},

{"setne", SET},
    {"sete", SET},
    {"setg", SET},

{"andl", ARITH},
    {"andw", ARITH},
    {"andb", ARITH},

{".comm", COMM},
    {".lcomm", LCOMM},

{"shrl", ARITH},
    {"shrw", ARITH},
    {"shrb", ARITH},

{"shll", ARITH},
    {"shlw", ARITH},
    {"shlb", ARITH},

{"imull", IMUL},
    {"imulw", IMUL},

{"idivl", IDIV},
    {"idivw", IDIV},
    {"idivb", IDIV},

{"divl", DIV},
    {"divw", DIV},

{"orl", ARITH},
    {"orw", ARITH},
    {"orb", ARITH},

{"sbbl", ARITH},
    {"sbbw", ARITH},
    {"sbbb", ARITH},

{"eax", REG},
    {"ebx", REG},
    {"ecx", REG},
    {"edx", REG},
    {"esi", REG},
    {"edi", REG},
    {"ebp", REG},
    {"esp", REG},

{"cr0", REG},
    {"cr2", REG},
    {"cr3", REG},
    {"cr4", REG},

{"ax", REG},
    {"bx", REG},
    {"cx", REG},
    {"dx", REG},
    {"si", REG},
    {"di", REG},
    {"bp", REG},
    {"sp", REG},

{"ah", REG},
    {"al", REG},
    {"bh", REG},
    {"bl", REG},
    {"ch", REG},
    {"cl", REG},
    {"dh", REG},
    {"dl", REG},

{"cs", REG},
    {"ds", REG},
    {"es", REG},
    {"fs", REG},
    {"gs", REG},
    {"ss", REG},

{"cld", ONEBYTE},
    {"cmc", ONEBYTE},
    {"std", ONEBYTE},
    {"stc", ONEBYTE},
    {"ret", ONEBYTE},
    {"nop", ONEBYTE},

{"rep", ONEBYTE},
    {"repe", ONEBYTE},
    {"repne", ONEBYTE},
    {"repnz", ONEBYTE},
    {"repz", ONEBYTE},

{"cmpsb", ONEBYTE},
    {"leave", ONEBYTE},

{"movs", ONEBYTE},
    {"movsw", ONEBYTE},
    {"movsb", ONEBYTE},
    {"aaa", ONEBYTE},
    {"aad", ONEBYTE},

{"lods", ONEBYTE},
    {"lodsb", ONEBYTE},
    {"lodsw", ONEBYTE},
    {"lodsl", TRANSLAT},

{"movsl", TRANSLAT},
    {"pushal", TRANSLAT},
    {"popal", TRANSLAT},

{"outl", ARITH},
    {"outw", ARITH},
    {"outb", ARITH},
    {"inl", ARITH},
    {"inb", ARITH},
    {"inw", ARITH},

{"btrl", ARITH},
    {"btrw", ARITH},
    {"btrb", ARITH},

{"cmps", ONEBYTE},
    {"cmpsb", ONEBYTE},
    {"cmpsw", ONEBYTE},
    {"cmpsl", TRANSLAT},

{"scasb", ONEBYTE},
    {"scasw", ONEBYTE},
    {"scasl", TRANSLAT},

{"adcl", ARITH},
    {"adcw", ARITH},
    {"adcb", ARITH},

{"btl", ARITH},
    {"btw", ARITH},
    {"btb", ARITH},

{"cli", ONEBYTE},
    {"sti", ONEBYTE},
    {"daa", ONEBYTE},
    {"clc", ONEBYTE},
    {"das", ONEBYTE},

{"int", INT},

{"lds", LSEG},
    {"les", LSEG},
    {"lfs", LSEG},
    {"lgs", LSEG},
    {"lss", LSEG},

{"lgdt", LGDT},
    {"sgdt", LGDT},
    {"lidt", LGDT},
    {"sidt", LGDT},
    {".type", UNSUPPORTED},
    {".size", UNSUPPORTED},
    {".weak", UNSUPPORTED},
    {".set", UNSUPPORTED},
    {".4byte", UNSUPPORTED},
    {".ident", IDENT},

{"roll", ARITH},
    {"rolw", ARITH},
    {"rolb", ARITH},

{"mull", DIV},
    {"mulw", DIV},
    {"mulb", DIV},

{"st", ST},
    {"flds", MATH},
    {"fldl", MATH},
    {"fdivs", MATH},
    {"fdivl", MATH},
    {"fstps", MATH},
    {"fstpl", MATH},
    {"fadds", MATH},
    {"faddl", MATH},
    {"faddp", MATH},
    {"faddrp", MATH},
    {"fsubs", MATH},
    {"fsubl", MATH},
    {"fsubp", MATH},
    {"fsubrp", MATH},
    {"fmuls", MATH},
    {"fmull", MATH},
    {"fdivrp", MATH},
    {"fstp", MATH},
    {"filds", MATH},
    {"fildl", MATH},
{"bsfl", ARITH},
{"cltd", CLTD}
};

/* slower than in math.h but without math instructions */
long pow2(long to)
{
  long i, j;
  j = 1;
  for (i = 0; i < to; i++) j *= 2;
  return j;
}

/* used by the parser */
char tmpstr1[100], tmpstr2[100], tmpstr3[100], tmpstr4[100], tmpstr5[100],
	tmpstr6[100];

/* execute something if there is no `--no-info' switch */
#define D(x) { if (no_info == 0) (x); }

%}

%start line

%%

line	:
	| PUSH REG		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }
	| PUSH varptr		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }
	| PUSH memref		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }
	| PUSH memptr		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }

| ARITH REG ',' REG	{ sprintf(output_string, "%s %s %s, %s\n", op($1), op_size($1), $4, $2); }
	| ARITH REG ',' memref
				{ sprintf(output_string, "%s %s %s, %s\n", op($1), op_size($1), $4, $2); }
	| ARITH memref ',' REG
				{ sprintf(output_string, "%s %s %s, %s\n", op($1), op_size($1), $4, $2); }

| ARITH REG ',' memptr	{ sprintf(output_string, "%s %s %s, %s\n", op($1), op_size($1), $4, $2); }
	| ARITH memptr ',' REG	{ sprintf(output_string, "%s %s %s, %s\n", op($1), op_size($1), $4, $2); }

| ARITH varptr ',' REG	{ sprintf(output_string, "%s %s %s, %s\n", op($1), op_size($1), $4, $2); }
	| ARITH varptr ',' memref
				{ sprintf(output_string, "%s %s %s, %s\n", op($1), op_size($1), $4, $2); }
	| ARITH varptr ',' memptr
				{ sprintf(output_string, "%s %s %s, %s\n", op($1), op_size($1), $4, $2); }

| INC REG		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }
	| INC memref		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }
	| INC memptr		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }

| LEA memref ',' REG
				{ sprintf(output_string, "lea %s %s, %s\n", op_size($1), $4, $2); }
	| LEA memptr ',' REG	{ sprintf(output_string, "lea %s %s, %s\n", op_size($1), $4, $2); }

| CALL ID		{ sprintf(output_string, "call %s\n", $2); }
	| CALL '*' REG		{ sprintf(output_string, "call %s\n", $3); }

| JMP ID		{ sprintf(output_string, "%s%s %s\n", $1, is_near($1), $2); }
	| JMP memref		{ sprintf(output_string, "%s%s %s\n", $1, is_near($1), $2); }
	| JMP '*' REG		{ sprintf(output_string, "%s%s %s\n", $1, is_near($1), $3); }
	| JMP '*' ID		{ sprintf(output_string, "%s%s %s\n", $1, is_near($1), $3); }

| IMUL REG		{ sprintf(output_string, "imul %s %s\n", op_size($1), $2); }
	| IMUL REG ',' REG	{ sprintf(output_string, "imul %s %s, %s\n", op_size($1), $4, $2); }
	| IMUL memref ',' REG	{ sprintf(output_string, "imul %s %s, %s\n", op_size($1), $4, $2); }
	| IMUL memptr ',' REG	{ sprintf(output_string, "imul %s %s, %s\n", op_size($1), $4, $2); }

| IDIV memref		{ sprintf(output_string, "idiv %s %s\n", op_size($1), $2); }
	| IDIV REG		{ sprintf(output_string, "idiv %s %s\n", op_size($1), $2); }
	| IDIV memptr		{ sprintf(output_string, "idiv %s %s\n", op_size($1), $2); }

| DIV memref		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }
	| DIV REG		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }

| MOV_SD arg ',' arg	{ mov_sd($1, $4, $2); }

| NEG REG		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }
	| NEG memref		{ sprintf(output_string, "%s %s %s\n", op($1), op_size($1), $2); }

| SET REG		{ sprintf(output_string, "%s %s\n", $1, $2); }

| INT NUMBER		{ sprintf(output_string, "int 0x%lx\n", $2); }

| LSEG ID ',' REG	{ sprintf(output_string, "%s %s, %s\n", $1, $4, $2); }

| LGDT ID		{ sprintf(output_string, "%s [%s]\n", $1, $2); }

| ONEBYTE		{ sprintf(output_string, "%s\n", $1); }

| CLTD			{ sprintf(output_string, "cdq\n"); }

| TRANSLAT		{ sprintf(output_string, "%s\n", translate_instr($1)); }

| MATH arg              { sprintf(output_string, "%s %s\n", translate_math($1), $2); }
	| MATH math_reg		{ sprintf(output_string, "%s %s\n", translate_math($1), $2); }
	| MATH math_reg ',' math_reg
				{ sprintf(output_string, "%s %s, %s\n", translate_math($1), $4, $2); }

| UNSUPPORTED		{ ; }

| error
	;

arg	: REG			{ $$ = $1; }
    | memptr		{ $$ = $1; }
	| memref		{ $$ = $1; }
	| varptr		{ $$ = $1; }
	;

memptr	: ID			{ sprintf(tmpstr1, "[%s+0x%lx]", $1, baseaddr); $$ = tmpstr1; }
       | ID '+' NUMBER		{ sprintf(tmpstr1, "[%s+%ld+0x%lx]", $1, $3, baseaddr); $$ = tmpstr1; }
	| ID NUMBER		{ sprintf(tmpstr1, "[%s%ld+0x%lx]", $1, $2, baseaddr); $$ = tmpstr1;
					if ($2>0) yyerror("expression error"); }
	;

memref	: memref1               { sprintf(tmpstr6, "[%s]", $1); $$ = tmpstr6; }
       | REG ':' memref1	{ sprintf(tmpstr6, "[%s:%s]", $1, $3); $$ = tmpstr6; }
	;

memref1	: '(' REG ')'		{ sprintf(tmpstr2, "%s", $2); $$ = tmpstr2;}
	| '(' REG ',' REG ')'	{ sprintf(tmpstr2, "%s + %s", $2, $4); $$ = tmpstr2; }
	| '(' REG ',' REG ',' scale ')'
				{ sprintf(tmpstr2, "%s + %s*%s", $2, $4, $6); $$ = tmpstr2; }
	| offset '(' REG ')'	{ sprintf(tmpstr2, "%s%s", $3, $1); $$ = tmpstr2; }
	| offset '(' REG ',' REG ')'
				{ sprintf(tmpstr2, "%s + %s %s", $3, $5, $1); $$ = tmpstr2; }
	| offset '(' REG ',' REG ',' scale ')'
				{ sprintf(tmpstr2, "%s + %s*%s %s", $3, $5, $7, $1); $$ = tmpstr2; }
	| offset '(' ',' REG ')'
				{ sprintf(tmpstr2, "%s %s", $4, $1); $$ = tmpstr2; }
	| offset '(' ',' REG ',' scale ')'
				{ sprintf(tmpstr2, "%s*%s %s", $4, $6, $1); $$ = tmpstr2; }
	;

varptr	: '$' NUMBER		{ $$ = signed_num($2); }
       | '$' ID		{ sprintf(tmpstr3, "%s+0x%lx", $2, baseaddr); $$ = tmpstr3; }
	| '$' ID NUMBER		{ sprintf(tmpstr3, "%s%ld+0x%lx", $2, $3, baseaddr); $$ = tmpstr3;
					if ($3>0) yyerror("expression error"); }
	| '$' ID '+' NUMBER	{ sprintf(tmpstr3, "%s+%ld+0x%lx", $2, $4, baseaddr); $$ = tmpstr3; }
	;

scale	: REG			{ $$ = $1; }
      | NUMBER		{ $$ = signed_num($1); }
	;

offset	: NUMBER		{ $$ = signed_num($1); }
       | ID			{ strcpy(tmpstr4, "+ "); strcat(tmpstr4, $1); $$ = tmpstr4;}
	| ID NUMBER		{ strcpy(tmpstr4, "+ "); strcat(tmpstr4, $1); strcat(tmpstr4, signed_num($2)); $$ = tmpstr4; }
	| '*' ID		{ strcpy(tmpstr4, "+ "); strcat(tmpstr4, $2); $$ = tmpstr4; }
	;

math_reg: ST                    { $$ = "st"; }
	| ST '(' NUMBER ')'     { sprintf(tmpstr4, "st%d", (int)$3); $$ = tmpstr4; }
	;
%%

/* output an error, exit if more than 15 errors */
void yyerror(char *s)
{
  FATAL(("error in translation from AT&T to intel syntax.\n Possible causes: don't use 0x... for constants and use imull eax,ebx instead of imul eax,ebx etc."));
}

/* output a warning */
void warning(char *s)
{
printf("in warning\n");
return;
}

/* function used for token sorting */
static int tok_cmp (const void *e1, const void *e2)
{
  return strcmp (((struct tok *)e1)->name, ((struct tok *)e2)->name);
}

/* THE lexer */
int yylex(void)
{
  static int was_percent = 0;
  int c, oldc, i;
  struct tok *token, t;

do {
	c = *string_pointer;
	string_pointer++;
  } while (c == ' ' || c == '\t');

switch (c) {
	case 'a' ... 'z':
	case 'A' ... 'Z':
	case '.':
	case '@':
	case '_':
	case '%':
		if (c != '%') { string_pointer--; was_percent = 0; } else was_percent = 1;

		sscanf(string_pointer, "%[a-zA-Z0-9_$.@]", strbuf);
		string_pointer+=strlen(strbuf);


t.name = strbuf;

/* bsearch is quicker */
		token = bsearch(&t, tokens, sizeof(tokens) / sizeof(tokens[0]),
			sizeof(tokens[0]), tok_cmp);

if (token) {
			yylval.str = strdup(token->name);
			if (token->token == UNSUPPORTED || token->token == UNSUP_2) {
				char tmp[100];
				if (token->token == UNSUPPORTED) {
					sprintf(tmp, "unsupported keyword `%s'", token->name);
					warning(tmp);
					warning("(only appears one time)");
					token->token = UNSUP_2;
				}
				/* eat line */
				while ((c = *string_pointer ) != EOF && c != '\n') string_pointer++;
				string_pointer--;
				return UNSUPPORTED;
			}
					/* not a real REG */
			if ((token->token == REG || token->token == ST) && (was_percent == 0) ) {
				yylval.str = strdup (strbuf);
				addexternal(strbuf);
				printf("%s\n", strbuf);
				return ID;
			}

return token->token;
		}
		if (strbuf[0] == '.') strbuf[0] = '$';
		yylval.str = strdup(strbuf);
		addexternal(strbuf);	/* add it to the symbol table */
		return ID;

case '0' ... '9':
		string_pointer--;
		sscanf(string_pointer, "%li", &yylval.num);
		sprintf(strbuf,"%li",yylval.num);
		string_pointer+=strlen(strbuf);
		return NUMBER;

case '-':
		string_pointer--;
		sscanf(string_pointer, "%ld", &yylval.num);
		sprintf(strbuf,"%ld",yylval.num);
		string_pointer+=strlen(strbuf);
		return NUMBER;

case '"':	/* a string */
	case '\'':
		oldc = c;
		i = 0;
		while (1) {
			c = *string_pointer;
			string_pointer++;
			if (c == oldc) {
				strbuflen = i;
				strbuf[i] = 0;
				yylval.str = strdup(strbuf);
				return STRING;
			}
			switch (c) {
				case '\\':
					c = * string_pointer;
					string_pointer++;
					switch (c) {
						case '0': strbuf[i++] = 0; break;
						case 'n': /*strbuf[i++] = '\n';*/ break;
						case 't': strbuf[i++] = '\t'; break;
						default:
							if (c >= '0' && c<= '9') {
								string_pointer--;

sscanf(string_pointer, "%i", &c);
								sprintf(strbuf,"%i",c);
								string_pointer+=strlen(strbuf);
								if (c == 12) c = 10;
								else
								if (c == 11) c = 9;

strbuf[i++] = c;
								break;
							}
							else strbuf[i++] = c;
					}
					break;
				default: strbuf[i++] = c; break;
			}
		}
		abort();
	case '/':
                printf("UNCOMMENT\n");
/*		c = ';';
		while (c != '\n') {
			if (no_info == 0) putc(c, outfile);
			c = *string_pointer;
			string_pointer++;
			//c = getc(infile);
		}
		c = '\n';
		if (no_info == 0) putc(c, outfile);
		return '\n';*/
	case ';':
	case '#':
		while ((c = *string_pointer )!= '\n') string_pointer++;
		c = '\n';

default:
       return c;
  }
}


void mov_sd(const char *mov, const char *arg1, const char *arg2)
{
printf("UNCOMME\n");
/*	int src, dst;
	char *s, *d;
	src = mov[4]; dst = mov[5];
	if (src == 'l') s = "dword";
	else
	if (src == 'w') s = "word";
	else
	if (src == 'b') s = "byte";
	else
	{ yyerror("mov?? unknown"); s = "[error]"; }

if (dst == 'l') d = "dword";
	else
	if (dst == 'w') d = "word";
	else
	if (dst == 'b') d = "byte";
	else
	{ yyerror("mov?? unknown"); d = "[error]"; }

if (strncmp(mov, "movs", 4) == 0)
		fprintf(outfile, "\tmovsx %s %s, %s %s\n",
			d, arg1, s, arg2);
	else
	if (strncmp(mov, "movz", 4) == 0)
		fprintf(outfile, "\tmovzx %s %s, %s %s\n",
			d, arg1, s, arg2);
	else yyerror("mov?? unknown");*/
}

/* look for a symbol */
struct symbol *sym_lookup(char *name) {
	int i;
	struct symbol *s;

i = hash(name);
	for (s = hashtab[i]; s; s = s->next) {
		if (!strcmp(s->name, name))
			return s;
	}
	return (struct symbol *)NULL;
}

/* add a symbol to the table */

struct symbol* addexternal(char *name)
{
	int i;
	struct symbol* s = sym_lookup(name);
	if (s == NULL) {	/* not found, add it */
		i = hash(name);
		s = (struct symbol *) xmalloc(sizeof (struct symbol));
		s->name = (char *) xmalloc(strlen(name) + 1);
		strcpy(s->name, name);
		s->isextern = 1;
		s->islocal = 0; 
		s->next = hashtab[i];
		hashtab[i] = s;
		return s;
	}
	return s;
}

int initialized = 0;
char * i386_a2i(char * input){
  if(!initialized){
    qsort(tokens, sizeof(tokens) / sizeof(tokens[0]), sizeof(tokens[0]), tok_cmp);
    initialized=1;
  }

  string_pointer = input;
  yyparse();
  return output_string;
}

/*
int main(int argc, char **argv)
{
	qsort(tokens, sizeof(tokens) / sizeof(tokens[0]), sizeof(tokens[0]), tok_cmp);

        printf("tada: %s",a2i("imull   %ebx, %eax"));
	string_pointer = "imull   %ebx, %eax";
	yyparse();
	printf("resultaat: %s",output_string);
	string_pointer = "leal    0(,%eax,8), %edx";
	yyparse();
	printf("resultaat: %s",output_string);
	string_pointer = "subl    %eax, %edx";
	yyparse();
	printf("resultaat: %s",output_string);
	string_pointer = "movl    %ecx, %eax";
	yyparse();
	printf("resultaat: %s",output_string);
	string_pointer = "decl    %edx";
	yyparse();
	printf("resultaat: %s",output_string);
	string_pointer = "imull   %ecx, %eax";
	yyparse();
	printf("resultaat: %s",output_string);
	string_pointer = "cmpl    %eax, %edx";
	yyparse();
	printf("resultaat: %s",output_string);
	string_pointer = "je      .L5";
	yyparse();
	printf("resultaat: %s",output_string);
	return 0;
}*/
