/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
#include "i386_assemble.h"
#include "i386_instructions.h"
#include "i386_opcode.h"
#include "i386_registers.h"
#include "i386_stack.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_reloc.h"
#include "../../kernel/diablo_liveness.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include <string.h>

/* {{{ stack optimization */
#define STACK_SIZE	20000

#define PushReg(_reg,ins)                  \
  do {                                     \
    top++;                                 \
    if (top == STACK_SIZE)                 \
      goto function_exit;                  \
    stack[top].kind = reg;                 \
    stack[top].u.reg = _reg;               \
    stack[top].refcount = 0;               \
    stack[top].def = ins;                  \
    stack[top].changed = NullRegs;         \
  } while (0)

#define PushImm(_imm,ins)                  \
  do {                                     \
    top++;                                 \
    if (top == STACK_SIZE)                 \
      goto function_exit;                  \
    stack[top].kind = imm;                 \
    stack[top].u.imm = _imm;               \
    stack[top].refcount = 0;               \
    stack[top].def = ins;                  \
    stack[top].changed = NullRegs;         \
  } while (0)

#define PushRel(_rel,ins)                  \
  do {                                     \
    top++;                                 \
    if (top == STACK_SIZE)                 \
      goto function_exit;                  \
    stack[top].kind = rel;                 \
    stack[top].u.rel = _rel;               \
    stack[top].refcount = 0;               \
    stack[top].def = ins;                  \
    stack[top].changed = NullRegs;         \
  } while (0)


#define PushUnk(ins)                       \
  do {                                     \
    top++;                                 \
    if (top == STACK_SIZE)                 \
      goto function_exit;                  \
    stack[top].kind = unk;                 \
    stack[top].refcount = 0;               \
    stack[top].def = ins;                  \
    stack[top].changed = NullRegs;         \
  } while (0)

#define PopCheck()                         \
  do {                                     \
    if (top < 0)                           \
      goto stack_reset;                    \
  } while (0)

typedef struct _t_stack_cont {
  enum {reg,imm,rel,unk} kind;
  union {
    t_reg reg;
    t_uint32 imm;
    t_reloc * rel;
  } u;
  int refcount;
  t_ins * def;
  t_regset changed;
} t_stack_cont;

static t_stack_cont stack[STACK_SIZE];
static int top = -1;

static void RemoveStackSlot(int slot, t_ins * alloc, t_ins * release)
{
  t_ins * ins;

  /* {{{ adjust all stack references between allocation and deletion of the slot */
  for (ins = INS_NEXT(alloc); ins != release; ins = INS_NEXT(ins))
  {
    t_i386_operand *dest = I386_INS_DEST(ins), *src1 = I386_INS_SOURCE1(ins), *src2 = I386_INS_SOURCE2(ins);
    int local_top = (int) INS_TEMP(ins);
    int thres = (local_top - slot) * 4;

    /* local stack top is one slot lower after the slot is removed */
    ((int) INS_TEMP(ins))--;

    if (!RegsetIn(INS_REGS_USE(ins),I386_REG_ESP)) continue;
    
    if (I386_OP_TYPE(dest) == i386_optype_mem && I386_OP_BASE(dest) == I386_REG_ESP)
    {
      if (I386_OP_IMMEDIATE(dest) > thres) 
	I386_OP_IMMEDIATE(dest) -= 4;
      else if (I386_OP_IMMEDIATE(dest) == thres)
	FATAL(("stack propagation algorithm is corrupt"));
    }
    if (I386_OP_TYPE(src1) == i386_optype_mem && I386_OP_BASE(src1) == I386_REG_ESP)
    {
      if (I386_OP_IMMEDIATE(src1) > thres) 
	I386_OP_IMMEDIATE(src1) -= 4;
      else if (I386_OP_IMMEDIATE(src1) == thres)
	FATAL(("stack propagation algorithm is corrupt"));
    }
    if (I386_OP_TYPE(src2) == i386_optype_mem && I386_OP_BASE(src2) == I386_REG_ESP)
    {
      if (I386_OP_IMMEDIATE(src2) > thres) 
	I386_OP_IMMEDIATE(src2) -= 4;
      else if (I386_OP_IMMEDIATE(src2) == thres)
	FATAL(("stack propagation algorithm is corrupt"));
    }
  } /* }}} */

  /* {{{ remove allocation of the slot */
  switch (I386_INS_OPCODE(alloc))
  {
    case I386_PUSH:
      InsKill(alloc);
      break;
    case I386_SUB:
      I386_OP_IMMEDIATE(I386_INS_SOURCE1(alloc)) -= 4;
      if (I386_OP_IMMEDIATE(I386_INS_SOURCE1(alloc)) == 0)
	InsKill(alloc);
      break;
    default:
      FATAL(("unknown stack slot allocation @I",alloc));
  }
  /* }}} */

  /* {{{ remove release of the slot */
  switch (I386_INS_OPCODE(release))
  {
    case I386_ADD:
      I386_OP_IMMEDIATE(I386_INS_SOURCE1(release)) -= 4;
      if (I386_OP_IMMEDIATE(I386_INS_SOURCE1(release)) == 0)
	I386InstructionMakeNoop(release);	/* do not kill this instruction immediately, it is still referenced in the main loop of PropBlock */
      break;
    case I386_POP:
      I386InstructionMakeNoop(release);		/* do not kill this instruction immediately, it is still referenced in the main loop of PropBlock */
      break;
    default:
      FATAL(("unknown stack slot release @I",alloc));
  }
  /* }}} */
}

static t_bool PropBlock(t_bbl * bbl)
{
  t_bool change = FALSE;
  t_ins * ins, * tmp;
  t_regset changed = NullRegs;
  int i;

  top = -1;

  BBL_FOREACH_INS_SAFE(bbl,ins,tmp)
  {
    t_i386_operand * src1 = I386_INS_SOURCE1(ins);
    /*t_i386_operand * src2 = I386_INS_SOURCE2(ins); */
    t_i386_operand * dest = I386_INS_DEST(ins);

    (int) INS_TEMP(ins) = top;

    /* skip instructions that do not read from or write to memory, and that do
     * not use or define the stack pointer */
    if (!RegsetIn(INS_REGS_USE(ins),I386_REG_ESP) && !RegsetIn(INS_REGS_DEF(ins),I386_REG_ESP) && !I386InsIsLoad(ins) && !I386InsIsStore(ins))
    {
      for (i=0; i<=top; i++) RegsetSetUnion(stack[i].changed,INS_REGS_DEF(ins));
      continue;
    }

    /* 1. stack manipulation */
    switch (I386_INS_OPCODE(ins))
    {
      case I386_PUSH:
	switch (I386_OP_TYPE(src1))
	{
	  case i386_optype_reg:
	    if (I386_OP_REGMODE(src1) == i386_regmode_full32)
	      PushReg(I386_OP_BASE(src1),ins);
	    else
	      PushUnk(ins);
	    break;
	  case i386_optype_imm:
	    if (!I386_OP_FLAGS(src1) & I386_OPFLAG_ISRELOCATED)
	      PushImm(I386_OP_IMMEDIATE(src1),ins);
	    else if (!PatternMatch("*P*",RELOCATABLE_REFERS_TO(ins)->rel->code))
	      PushRel(RELOCATABLE_REFERS_TO(ins)->rel,ins);
	    else		/* pc-relative relocations are too hard to propagate. we just skip them */
	      PushUnk(ins);
	    break;
	  default:
	    PushUnk(ins);
	}
	break;
      case I386_PUSHF:
	PushUnk(ins);
	break;
      case I386_PUSHA:
	PushReg(I386_REG_EAX,ins);
	PushReg(I386_REG_ECX,ins);
	PushReg(I386_REG_EDX,ins);
	PushReg(I386_REG_EBX,ins);
	PushReg(I386_REG_ESP,ins);
	PushReg(I386_REG_EBP,ins);
	PushReg(I386_REG_ESI,ins);
	PushReg(I386_REG_EDI,ins);
	break;
      case I386_POP:
	PopCheck();
	if (stack[top].kind == reg && !RegsetIn(stack[top].changed,stack[top].u.reg) && stack[top].refcount == 0)
	{
	  if (stack[top].u.reg == I386_OP_BASE(dest))
	  {
/*            DiabloPrint(stdout, "@I pops value that should never have been saved\n",ins);*/
	    RemoveStackSlot(top,stack[top].def,ins);
	    change = TRUE;
	  }
	  else
	  {
	    t_reg d = I386_OP_BASE(dest), s = stack[top].u.reg;
	    RemoveStackSlot(top,stack[top].def,ins);
	    I386InstructionMakeMovToReg(ins,d,s,0);
	    change = TRUE;
	  }
	}
	else if (stack[top].kind == imm && stack[top].refcount == 0)
	{
	  t_reg d = I386_OP_BASE(dest);
	  t_uint32 imm = stack[top].u.imm;

	  RemoveStackSlot(top,stack[top].def,ins);
	  I386InstructionMakeMovToReg(ins,d,I386_REG_NONE,imm);
	  change = TRUE;
	}
	else if (stack[top].kind == rel && !PatternMatch("*P*",stack[top].u.rel->code) && stack[top].refcount == 0)
	{
	  t_reg d = I386_OP_BASE(dest);
	  t_reloc * rel = stack[top].u.rel;
	  t_address addend = rel->addend;
	  t_relocatable * to = REL_TO_RELOCATABLE(rel);
	  t_address to_offset = rel->to_offset;
	  t_bool hell = rel->hell;
	  t_cfg_edge * edge = rel->edge;
	  t_reloc * corresponding = rel->corresponding;
	  t_relocatable * sec = rel->e.sec;
	  t_string code = StringDup(rel->code);
	  
	  RemoveStackSlot(top,stack[top].def,ins);
	  I386InstructionMakeMovToReg(ins,d,I386_REG_NONE,0);
	  I386_OP_FLAGS(I386_INS_SOURCE1(ins)) |= I386_OPFLAG_ISRELOCATED;
	  RelocTableAddRelocToRelocatable(INS_OBJECT(ins)->reloc_table,addend,T_RELOCATABLE(ins),NULL_GENERIC,to,to_offset,hell,edge,corresponding,sec,code,FALSE);
	  change = TRUE;
	}
	--top;
	break;
      case I386_POPF:
	PopCheck();
	--top;
	break;
      case I386_POPA:
	top -= 8;
	PopCheck();
	break;

      case I386_ADD:
	if (I386_OP_TYPE(dest) == i386_optype_reg && I386_OP_BASE(dest) == I386_REG_ESP)
	{
	  int slots,i;

	  if (I386_OP_TYPE(src1) != i386_optype_imm)
	    goto stack_reset; 	/* we can't know the new stack pointer value, so we reset the  propagation */
	  slots = I386_OP_IMMEDIATE(src1) / 4;
	  if (slots > top + 1)
	    goto stack_reset;	/* stack popped below the level at the beginning of the block; reset the propagation */
	  /* find out if any of the freed stack slots was never referenced.
	   * if so, it should not have been allocated in the first place */
	  {
	    t_bool killable[slots];
	    int nkillable = 0;

	    for (i=0; i<slots; i++)
	    {
	      if (stack[top-i].refcount == 0) 
	      {
		killable[i] = TRUE;
		nkillable++;
	      }
	      else 
		killable[i] = FALSE;
	    }
	    while (nkillable > 0)
	    {
	      nkillable--;

	      for (i=0; i<slots; i++)
	      {
		if (killable[i])
		{
		  change = TRUE;
		  RemoveStackSlot(top-i,stack[top-i].def,ins);
		  /* update datastructures to reflect the new stack layout */
		  memmove(&(stack[top-i]), &(stack[top-i+1]), i*sizeof(t_stack_cont));
		  memmove(&(killable[i]),&(killable[i+1]),(slots-i-1)*sizeof(t_bool));
		  slots--;
		  top--;
		}
	      }
	    }
	  }
	  top -= slots;
	}
	break;

      case I386_SUB:
	if (I386_OP_TYPE(dest) == i386_optype_reg && I386_OP_BASE(dest) == I386_REG_ESP)
	{
	  int slots,i;

	  if (I386_OP_TYPE(src1) != i386_optype_imm)
	    goto stack_reset; 	/* we can't know the new stack pointer value, so we reset the propagation */
	  slots = I386_OP_IMMEDIATE(src1) / 4;
	  for (i=0; i<slots; i++)
	    PushUnk(ins);
	}
	break;

      case I386_CALL:
      case I386_RET:
	/* ignore these, they always come at the end of a bbl so propagation
	 * stops anyway */
	break;

      default:
	if (RegsetIn(INS_REGS_DEF(ins),I386_REG_ESP))
	{
	  /* uncaught stack manipulation: terminate propagation */
/*          DiabloPrint(stdout, "do not understand this stack manipulation: @I\n", ins);*/
	  goto function_exit;
	}
    }

    /* 2. stores */
    if (I386InsIsStore(ins))
    {
      t_i386_operand * op = I386InsGetMemStoreOp(ins);
      if (op)
      {
	/* possible aliasing? */
	if (I386_OP_BASE(op) != I386_REG_ESP && I386_OP_BASE(op) != I386_REG_NONE)
	  goto stack_reset;
	/* stack-relative? */
	if (I386_OP_BASE(op) == I386_REG_ESP)
	{
	  if (I386_OP_INDEX(op) != I386_REG_NONE)
	  {
	    /* there is an index register, so we can't know for sure which
	     * stack slot we change: make them all unknown */
	    int i;
	    for (i=0; i<=top; i++) 
	    {
	      stack[i].kind = unk;
	      stack[i].refcount++;
	    }
	  }
	  else
	  {
	    /* TODO: stores van registers en immediate waarden */
	    int slot = top - (I386_OP_IMMEDIATE(op) / 4);
	    int nslots = (I386_OP_MEMOPSIZE(op) + 3) / 4; /* round up, not down */
	    int i;
	    for (i = slot; i >= 0 && i > slot - nslots; i--)
	    {
	      stack[i].kind = unk;
	      stack[i].refcount++;
	    }
	  }
	}
      }
    }


    /* 3. loads */
    if (I386InsIsLoad(ins))
    {
      t_i386_operand * op = I386InsGetMemLoadOp(ins);
      if (op)
      {
	/* possible aliasing? */
	if (I386_OP_BASE(op) != I386_REG_ESP && I386_OP_BASE(op) != I386_REG_NONE)
	{
	  int i;
	  for (i=0; i<=top; i++) stack[i].refcount++;
	}
	/* stack-relative? */
	if (I386_OP_BASE(op) == I386_REG_ESP)
	{
	  if (I386_OP_INDEX(op) != I386_REG_NONE)
	  {
	    /* there is an index register, so we can't know for sure which
	     * stack slot we change: increase their refcounts */
	    int i;
	    for (i=0; i<=top; i++) stack[i].refcount++;
	  }
	  else
	  {
	    int slot = top - (I386_OP_IMMEDIATE(op) / 4);
	    int nslots = (I386_OP_MEMOPSIZE(op) + 3) / 4; /* round up, not down */
	    int i;
	    for (i = slot; i >= 0 && i > slot - nslots; i--)
	    {
	      stack[i].refcount++;
	      if (nslots == 1)
	      {
		t_i386_operand saved;
		t_i386_opcode_entry * forms[10];
		memcpy(&saved,op,sizeof(t_i386_operand));

		/* optimization only possible if slots are read one at a time */
		if (stack[i].kind == imm)
		{
/*                  DiabloPrint(stdout,"@I loads immediate %x\n",ins,stack[i].u.imm);*/
		  I386OpSetImm(op,stack[i].u.imm,4);
		  INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
		  INS_REGS_DEF(ins) = I386InsDefinedRegisters(ins);
		  if (I386GetPossibleEncodings(ins,forms) == 0)
		  {
		    /* undo */
		    memcpy(op,&saved,sizeof(t_i386_operand));
		  }
		  else
		  {
		    change = TRUE;
		  }
		}
		else if (stack[i].kind == rel)
		{
		  t_reloc * rel = stack[i].u.rel;
/*                  DiabloPrint(stdout,"@I loads relocated immediate @R\n",ins,stack[i].u.rel);*/
		  VERIFY_FASSERT(!RELOCATABLE_REFERS_TO(ins),("ins already refers to something!"));
		  I386OpSetImm(op,0,4);
		  if (I386GetPossibleEncodings(ins,forms) == 0)
		  {
		    /* undo */
		    memcpy(op,&saved,sizeof(t_i386_operand));
		  }
		  else
		  {
		    I386_OP_FLAGS(op) |= I386_OPFLAG_ISRELOCATED;
		    RelocTableAddRelocToRelocatable(INS_OBJECT(ins)->reloc_table,NULL_GENERIC,
			T_RELOCATABLE(ins),NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset,
			rel->hell,rel->edge,rel->corresponding,rel->e.sec,StringDup(rel->code),
			rel->first);
		    RELOCATABLE_REFERS_TO(ins)->rel->label = StringDup(rel->label);
		    INS_REGS_USE(ins) = I386InsUsedRegisters(ins);
		    INS_REGS_DEF(ins) = I386InsDefinedRegisters(ins);
		    change = TRUE;
		  }
		}
		else if (stack[i].kind == reg && !RegsetIn(changed,stack[i].u.reg))
		{
/*                  DiabloPrint(stdout, "@I loads unchanged reg %d\n",ins,changed,stack[top-slot].u.reg);*/
		}
	      }
	    }
	  }
	}
      }
    }

    /* 4. lea instructions. if they reference a stack slot, the refcount of this slot should be incremented */
    if (I386_INS_OPCODE(ins) == I386_LEA)
    {
      if (I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_ESP)
      {
	if (I386_OP_INDEX(I386_INS_SOURCE1(ins)) != I386_REG_NONE)
	{
	  /* we don't know which stack slot is referenced, mark them all */
	  int i;
	  for (i=0; i<=top; i++) stack[i].refcount++;
	}
	else
	{
	  int slot = top - I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins))/4;
	  if (slot >= 0 && slot <= top) stack[slot].refcount++;
	}
      }
      else
      {
	/* possible aliasing: increment the refcount of all stack slots */
	int i;
	for (i=0; i<=top; i++) stack[i].refcount++;
      }
    }
 
    for (i=0; i<=top; i++) RegsetSetUnion(stack[i].changed,INS_REGS_DEF(ins));

    continue;

stack_reset:
    top = -1;
  }

function_exit:
  return change;
}

static void I386PeepHoleStack(t_cfg * cfg)
{
  t_bbl * bbl;
  CFG_FOREACH_BBL(cfg,bbl)
  {
    while (PropBlock(bbl)) {}
  }
}
/* }}} */

/* {{{ idempotent instruction removal */
static void I386PeepHoleIdempotent(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins, * tmp;
  t_uint32 kills = 0;
  int nop,lea,mov,add_sub;
  nop = lea = mov = add_sub = 0;
  
  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS_SAFE(bbl,ins,tmp)
    {
      t_i386_operand * dest = I386_INS_DEST(ins), * src1 = I386_INS_SOURCE1(ins) /* , *src2 = I386_INS_SOURCE2(ins) */ ;

      switch (I386_INS_OPCODE(ins))
      {
	case I386_NOP:
	  kills++;
	  nop++;
	  InsKill(ins);
	  break;
	case I386_MOV:
	  if (I386_OP_TYPE(dest) == i386_optype_reg && I386_OP_TYPE(src1) == i386_optype_reg)
	    if (I386_OP_BASE(dest) == I386_OP_BASE(src1) && I386_OP_REGMODE(dest) == I386_OP_REGMODE(src1))
	    {
	      kills++;
	      mov++;
	      InsKill(ins);
	    }
	  break;
	case I386_LEA:
	  if (I386_OP_BASE(dest) == I386_OP_BASE(src1) &&
	      I386_OP_IMMEDIATE(src1) == 0 && I386_OP_INDEX(src1) == I386_REG_NONE)
	  {
	    kills++;
	    lea++;
	    InsKill(ins);
	  }
	  break;
	case I386_ADD:
	case I386_SUB:
	  if (I386_OP_TYPE(src1) == i386_optype_imm && I386_OP_IMMEDIATE(src1) == 0)
	  {
	    kills++;
	    add_sub++;
	    InsKill(ins);
	  }
	  break;
	default:
	  break;
      }
    }
  VERBOSE(0,("[idempotent instruction removal] killed %d instructions",kills));
  /*printf("nop %d mov %d lea %d add/sub %d\n", nop,mov,lea,add_sub);*/
} /* }}} */

/* {{{ redundant pop/push removal */
/* look for patterns like
 *  pop reg
 *  ... some stuff that doesn't change %esp or reg
 *  push reg
 *
 * we can then change the pop into a load and remove the push,
 * if we can adjust all intermediate uses of the stack pointer
 */

/* helper function */
static t_bool StackUseIsWellbehaved(t_ins * ins)
{
  t_i386_operand * op1, * op2;
  t_bool found_esp_use = FALSE;
  t_bool well_behaved = TRUE;

  op1 = I386InsGetMemLoadOp(ins);
  op2 = I386InsGetMemStoreOp(ins);
  
  if (op1)
  {
    if (I386_OP_TYPE(op1) == i386_optype_mem)
    {
      if (I386_OP_BASE(op1) == I386_REG_ESP)
	found_esp_use = TRUE;
    }
    else
      well_behaved = FALSE;
  }
  if (op2)
  {
    if (I386_OP_TYPE(op2) == i386_optype_mem)
    {
      if (I386_OP_BASE(op2) == I386_REG_ESP)
	found_esp_use = TRUE;
    }
    else
      well_behaved = FALSE;
  }
  return found_esp_use && well_behaved;
}

void I386PeepHoleRedundantPushes(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  /*static int teller = 0; */

  CFG_FOREACH_BBL(cfg,bbl)
  {
    t_bool rerun = TRUE;
    while (rerun)
    {
      t_ins * pop = NULL;
      t_regset use,def;
      t_reg reg;
      
      rerun = FALSE;
      use = def = NullRegs;
      BBL_FOREACH_INS(bbl,ins)
      {
	if (I386_INS_OPCODE(ins) == I386_POP)
	{
	  pop = ins;
	  reg = I386_OP_BASE(I386_INS_DEST(ins));
	  use = def = NullRegs;
	  continue;
	}

	/* for speed */
	if (!pop) continue;

	if (I386_INS_OPCODE(ins) == I386_PUSH)
	{
	  t_i386_operand * op = I386_INS_SOURCE1(ins);
	  if (I386_OP_TYPE(op) == i386_optype_reg && I386_OP_BASE(op) == reg)
	  {
	    t_ins * iter;
	    /* change the pop instruction */
	    I386InstructionMakeMovFromMem(pop,reg,0,I386_REG_ESP,I386_REG_NONE,I386_SCALE_1);
	    /* adjust all intermediate %esp uses */
	    for (iter = INS_NEXT(pop); iter != ins; iter = INS_NEXT(iter))
	    {
	      t_i386_operand *op1, *op2;
	      if (!RegsetIn(INS_REGS_USE(iter),I386_REG_ESP)) continue;
	      
	      op1 = I386InsGetMemLoadOp(iter);
	      op2 = I386InsGetMemStoreOp(iter);

	      if (op1 && I386_OP_BASE(op1) == I386_REG_ESP)
		I386OpSetMem(op1,I386_OP_IMMEDIATE(op1)+4,I386_OP_BASE(op1),I386_OP_INDEX(op1),I386_OP_SCALE(op1),I386_OP_MEMOPSIZE(op1));
	      if (op2 && op2 != op1 && I386_OP_BASE(op2) == I386_REG_ESP)
		I386OpSetMem(op2,I386_OP_IMMEDIATE(op2)+4,I386_OP_BASE(op2),I386_OP_INDEX(op2),I386_OP_SCALE(op2),I386_OP_MEMOPSIZE(op2));
	    }
	    /* kill the push instruction */
	    InsKill(ins);
	    /* restart examination of the block to find further opportunities */
	    rerun = TRUE;
	    break;
	  }
	}

	if (RegsetIn(INS_REGS_DEF(ins),reg) || RegsetIn(INS_REGS_DEF(ins),I386_REG_ESP))
	{
	  pop = NULL;
	  continue;
	}
	
	if (RegsetIn(INS_REGS_USE(ins),I386_REG_ESP))
	{
	  /* we can tolerate this, if it is possible to adjust the use to the changed stack pointer */
	  if (!StackUseIsWellbehaved(ins))
	    pop = NULL;
	}
      }
    }
  }

#if 0
  /* perform liveness analysis in order to kill as much of the mov (%esp),%reg 
   * instructions as possible, and turn the rest into pop/push combinations
   * (this reduces code section bloat and speeds up instrumentation) */
  ComLiveness(cfg,INSENSITIVE);
  ComputeSavedChangedRegisters(cfg->sec);
  ComLiveness(cfg,SENSITIVE);
  ComputeSavedChangedRegisters(cfg->sec);
  ComLiveness(cfg,SENSITIVE);
  OptKillUseless(cfg);

  CFG_FOREACH_BBL(cfg,bbl)
  {
    BBL_FOREACH_INS(bbl,ins)
    {
      if (I386_INS_OPCODE(ins) == I386_MOV
	  && I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_mem
	  && I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_ESP
	  && I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins)) == 0
	  && I386_OP_INDEX(I386_INS_SOURCE1(ins)) == I386_REG_NONE
	  && I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg
	  && I386_OP_BASE(I386_INS_DEST(ins)) != I386_REG_ESP
	  && I386IsGeneralPurposeReg(I386_OP_BASE(I386_INS_DEST(ins))))
      {
	t_ins * added;
	t_reg reg = I386_OP_BASE(I386_INS_DEST(ins));
	I386InstructionMakePop(ins,reg);
	I386MakeInsForIns(Push,After,added,ins,reg,0);
	DiabloPrint(stdout,"UNdoing\n");
      }
    }
  }
#endif
}
/* }}} */

/* {{{ substitute instructions with smaller, equivalent forms */
static void I386PeepHoleShorterEquivalent(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  t_regset condflags = NullRegs;

  RegsetAddReg(condflags,I386_CONDREG_OF);
  RegsetAddReg(condflags,I386_CONDREG_SF);
  RegsetAddReg(condflags,I386_CONDREG_ZF);
  RegsetAddReg(condflags,I386_CONDREG_AF);
  RegsetAddReg(condflags,I386_CONDREG_PF);
  RegsetAddReg(condflags,I386_CONDREG_CF);

  /* for most substitutions, we need accurate liveness information */
  ComLiveness(cfg,INSENSITIVE);
  ComputeSavedChangedRegisters(cfg->sec);
  ComLiveness(cfg,SENSITIVE);
  ComputeSavedChangedRegisters(cfg->sec);
  ComLiveness(cfg,SENSITIVE);

  CFG_FOREACH_BBL(cfg,bbl)
  {
    BBL_FOREACH_INS(bbl,ins)
    {
      t_i386_operand * dest = I386_INS_DEST(ins);
      t_i386_operand * src  = I386_INS_SOURCE1(ins);
      t_i386_operand * src2 = I386_INS_SOURCE2(ins);

      switch (I386_INS_OPCODE(ins))
      {
	case I386_MOV:
	  if (I386_OP_TYPE(dest) == i386_optype_reg && I386_OP_TYPE(src) == i386_optype_imm)
	  {
	    if (I386IsGeneralPurposeReg(I386_OP_BASE(dest)) && I386_OP_REGMODE(dest) == i386_regmode_full32)
	    {
	      if (I386_OP_IMMEDIATE(src) == 0 && (!(I386_OP_FLAGS(src) & I386_OPFLAG_ISRELOCATED) || REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ins)->rel) == (t_relocatable *) abs_section))
	      {
		/* mov $0,%reg -> xor %reg,%reg */
		if (RegsetIsMutualExclusive(InsRegsLiveAfter(ins),condflags))
		{
		  t_reg d = I386_OP_BASE(dest);
		  /*DiabloPrint(stdout,"SR: @I -> ",ins);*/
		  I386InstructionMakeArithmetic(ins,I386_XOR,d,d,0);
		  /*DiabloPrint(stdout,"@I\n",ins);*/
		}
	      }
	      else if (I386_OP_IMMEDIATE(src) == -1 && (!(I386_OP_FLAGS(src) & I386_OPFLAG_ISRELOCATED) || REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ins)->rel) == (t_relocatable *) abs_section))
	      {
		
		/* mov $-1,%reg -> or $-1,%reg */
		if (RegsetIsMutualExclusive(InsRegsLiveAfter(ins),condflags))
		{
		  t_reg d = I386_OP_BASE(dest);
		  /*DiabloPrint(stdout,"SR: @I -> ",ins);*/
		  I386InstructionMakeArithmetic(ins,I386_OR,d,I386_REG_NONE,-1);
		  /*DiabloPrint(stdout,"@I\n",ins);*/
		}
	      }
	    }
	  }
	  break;
	case I386_ADD:
	  if (I386_OP_TYPE(dest) == i386_optype_reg && I386_OP_TYPE(src) == i386_optype_imm)
	  {
	    if (I386IsGeneralPurposeReg(I386_OP_BASE(dest)) && I386_OP_REGMODE(dest) == i386_regmode_full32)
	    {
	      if (!RegsetIn(InsRegsLiveAfter(ins),I386_CONDREG_CF))
	      {
		if (I386_OP_IMMEDIATE(src) == 1 && (!(I386_OP_FLAGS(src) & I386_OPFLAG_ISRELOCATED) || REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ins)->rel) == (t_relocatable *) abs_section))
		{
		  /* add $1,%reg -> inc %reg */
		  /*DiabloPrint(stdout,"SR: @I -> ",ins);*/
		  I386InstructionMakeIncDec(ins,I386_INC,I386_OP_BASE(dest));
		  /*DiabloPrint(stdout,"@I\n",ins);*/
		}
		else if (I386_OP_IMMEDIATE(src) == -1 && (!(I386_OP_FLAGS(src) & I386_OPFLAG_ISRELOCATED) || REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ins)->rel) == (t_relocatable *) abs_section))
		{
		  /* add $-1,%reg -> dec %reg */
		  /*DiabloPrint(stdout,"SR: @I -> ",ins);*/
		  I386InstructionMakeIncDec(ins,I386_DEC,I386_OP_BASE(dest));
		  /*DiabloPrint(stdout,"@I\n",ins);*/
		}
	      }
	    }
	  }
	  break;
	case I386_SUB:
	  if (I386_OP_TYPE(dest) == i386_optype_reg && I386_OP_TYPE(src) == i386_optype_imm)
	  {
	    if (I386IsGeneralPurposeReg(I386_OP_BASE(dest)) && I386_OP_REGMODE(dest) == i386_regmode_full32)
	    {
	      if (!RegsetIn(InsRegsLiveAfter(ins),I386_CONDREG_CF))
	      {
		if (I386_OP_IMMEDIATE(src) == 1 && (!(I386_OP_FLAGS(src) & I386_OPFLAG_ISRELOCATED) || REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ins)->rel) == (t_relocatable *) abs_section))
		{
		  /* sub $1,%reg -> dec %reg */
		  /*DiabloPrint(stdout,"SR: @I -> ",ins);*/
		  I386InstructionMakeIncDec(ins,I386_DEC,I386_OP_BASE(dest));
		  /*DiabloPrint(stdout,"@I\n",ins);*/
		}
		else if (I386_OP_IMMEDIATE(src) == -1 && (!(I386_OP_FLAGS(src) & I386_OPFLAG_ISRELOCATED) || REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ins)->rel) == (t_relocatable *) abs_section))
		{
		  /* sub $-1,%reg -> inc %reg */
		  /*DiabloPrint(stdout,"SR: @I -> ",ins);*/
		  I386InstructionMakeIncDec(ins,I386_INC,I386_OP_BASE(dest));
		  /*DiabloPrint(stdout,"@I\n",ins);*/
		}
	      }
	    }
	  }
	  break;
	case I386_CMP:
	  if (I386_OP_TYPE(src) == i386_optype_reg && I386_OP_TYPE(src2) == i386_optype_imm)
	  {
	    if (I386IsGeneralPurposeReg(I386_OP_BASE(src)) && I386_OP_REGMODE(src) == i386_regmode_full32)
	    {
	      if (I386_OP_IMMEDIATE(src2) == 0 && (!(I386_OP_FLAGS(src2) & I386_OPFLAG_ISRELOCATED) || REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ins)->rel) == (t_relocatable *) abs_section))
	      {
		if (RegsetIsMutualExclusive(InsRegsLiveAfter(ins),condflags))
		{
		  /* cmp $0,%reg -> test %reg,%reg */
		  /*DiabloPrint(stdout,"SR: @I -> ",ins);*/
		  I386InstructionMakeTest(ins,I386_OP_BASE(src),I386_OP_BASE(src),0);
		  /*DiabloPrint(stdout,"@I\n",ins);*/
		}
	      }
	    }
	  }
	  break;
	case I386_LEA:
	  /* lea (%rega),%regb -> mov %rega,%regb */
	  if (I386_OP_IMMEDIATE(src) == 0 && I386_OP_INDEX(src) == I386_REG_NONE)
	  {
	    I386InstructionMakeMovToReg(ins,I386_OP_BASE(dest),I386_OP_BASE(src),0);
	  }
	  /* lea <abs address>,%reg -> mov $<abs address>,%reg */
	  else if (I386_OP_BASE(src) == I386_REG_NONE && I386_OP_INDEX(src) == I386_REG_NONE)
	  {
	    t_ins * extra;
	    I386MakeInsForIns(MovToReg,Before,extra,ins,I386_OP_BASE(dest),I386_REG_NONE,I386_OP_IMMEDIATE(src));
	    if (I386_OP_FLAGS(src) & I386_OPFLAG_ISRELOCATED)
	    {
	      RelocSetToRelocatable(RELOCATABLE_REFERS_TO(ins)->rel,(t_relocatable *)extra);
	      I386_OP_FLAGS(I386_INS_SOURCE1(extra)) |= I386_OPFLAG_ISRELOCATED;
	    }
	    InsKill(ins);
	    ins = extra;
	  }
	  break;
	default:
	  /* keep the compiler happy */
	  break;
      }
    }
  }
}
/* }}} */

/* {{{ subsume adjacent add/sub combinations in one instruction */
static void I386PeepHoleAddSub(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins, * tmp;
  t_regset condflags = NullRegs;

  RegsetAddReg(condflags,I386_CONDREG_OF);
  RegsetAddReg(condflags,I386_CONDREG_SF);
  RegsetAddReg(condflags,I386_CONDREG_ZF);
  RegsetAddReg(condflags,I386_CONDREG_AF);
  RegsetAddReg(condflags,I386_CONDREG_PF);
  RegsetAddReg(condflags,I386_CONDREG_CF);

  /* we need accurate liveness information */
  ComLiveness(cfg,INSENSITIVE);
  ComputeSavedChangedRegisters(cfg->sec);
  ComLiveness(cfg,SENSITIVE);
  ComputeSavedChangedRegisters(cfg->sec);
  ComLiveness(cfg,SENSITIVE);

  CFG_FOREACH_BBL(cfg,bbl)
  {
    BBL_FOREACH_INS_SAFE(bbl,ins,tmp)
    {
      if (I386_INS_OPCODE(ins) == I386_ADD || I386_INS_OPCODE(ins) == I386_SUB)
      {
	t_i386_operand * dest = I386_INS_DEST(ins), * src = I386_INS_SOURCE1(ins);
	t_i386_operand * dest2, * src2;

	if (I386_OP_TYPE(dest) == i386_optype_reg && I386_OP_REGMODE(dest) == i386_regmode_full32 && I386IsGeneralPurposeReg(I386_OP_BASE(dest))
	    && I386_OP_TYPE(src) == i386_optype_imm && !(I386_OP_FLAGS(src) & I386_OPFLAG_ISRELOCATED))
	{
	  t_regset def = INS_REGS_DEF(ins);
	  t_reg d = I386_OP_BASE(dest);
          t_bool found = FALSE;
	  t_ins * iter;
	  for (iter = INS_NEXT(ins); iter; iter = INS_NEXT(iter))
	  {
	    dest2 = I386_INS_DEST(iter);
	    src2 = I386_INS_SOURCE1(iter);

	    if (RegsetIsMutualExclusive(INS_REGS_USE(iter),def)) continue;
	    /* should be add or sub */
	    if (I386_INS_OPCODE(iter) == I386_ADD || I386_INS_OPCODE(iter) == I386_SUB)
	    {
	      if (I386_OP_TYPE(dest2) == i386_optype_reg && I386_OP_REGMODE(dest2) == i386_regmode_full32 && I386_OP_BASE(dest2) == d
		  && I386_OP_TYPE(src2) == i386_optype_imm && !(I386_OP_FLAGS(src2) & I386_OPFLAG_ISRELOCATED)
		  && RegsetIsMutualExclusive(InsRegsLiveAfter(iter),condflags))
		found = TRUE;
	    }
	    break;
	  }

	  /* TODO: merge the two instructions */
	  if (found)
	  {
	    t_i386_opcode opcode = I386_INS_OPCODE(iter);
	    t_uint32 imm;
	    /*DiabloPrint(stdout,"AS: will merge @I and @I\n",ins,iter);*/
	    imm = I386_OP_IMMEDIATE(src2) + 
	         (I386_INS_OPCODE(iter) == I386_INS_OPCODE(ins) ? 1 : -1)*I386_OP_IMMEDIATE(src);
	    if (imm == 128)
	    {
	      /* 128 is encoded in 4 bytes, -128 in 1 byte */
	      opcode = (opcode == I386_SUB) ? I386_ADD : I386_SUB;
	      imm = -imm;
	    }
	    I386_INS_OPCODE(iter) = opcode;
	    I386_OP_IMMEDIATE(src2) = imm;
	    InsKill(ins);
	    /*DiabloPrint(stdout,"--> @I\n",iter);*/
	  }
	}
      }
    }
  }
}
/* }}} */

/* {{{ if possible, remove frame pointer set/reset from one-block functions 
 * this helps for simple function inlining */
void I386PeepHoleFramePointerRemoval(t_cfg * cfg)
{
  t_function * fun;

  I386FramePointerAnalysis(cfg);
  CFG_FOREACH_FUN(cfg,fun)
  {
    t_bbl * first = FUN_BBL_FIRST(fun);
    t_ins * ins, * prologue, * epilogue;
    int height = 0;
    t_bool can_remove = TRUE;

    if (!(FUN_FLAGS(fun) & FF_HAS_FRAMEPOINTER)) continue;
    if (!first) continue;
    if (BBL_NEXT_IN_FUN(first) && BBL_NEXT_IN_FUN(first) != FunctionGetExitBlock(fun)) continue;

    /* find frame pointer set/reset prologue and epilogue */
    ins = BBL_INS_FIRST(first);
    if (I386_INS_OPCODE(ins) != I386_PUSH || 
	I386_OP_TYPE(I386_INS_SOURCE1(ins)) != i386_optype_reg ||
	I386_OP_BASE(I386_INS_SOURCE1(ins)) != I386_REG_EBP ||
	I386_OP_REGMODE(I386_INS_SOURCE1(ins)) != i386_regmode_full32)
      continue;
    ins = INS_NEXT(ins);
    if (I386_INS_OPCODE(ins) != I386_MOV ||
	I386_OP_TYPE(I386_INS_SOURCE1(ins)) != i386_optype_reg ||
	I386_OP_BASE(I386_INS_SOURCE1(ins)) != I386_REG_ESP ||
	I386_OP_REGMODE(I386_INS_SOURCE1(ins)) != i386_regmode_full32 ||
	I386_OP_TYPE(I386_INS_DEST(ins)) != i386_optype_reg ||
	I386_OP_BASE(I386_INS_DEST(ins)) != I386_REG_EBP ||
	I386_OP_REGMODE(I386_INS_DEST(ins)) != i386_regmode_full32)
      continue;
    prologue = INS_NEXT(ins);

    ins = BBL_INS_LAST(first);
    if (I386_INS_OPCODE(ins) != I386_RET)
      continue;
    ins = INS_PREV(ins);
    if (!(
	  (I386_INS_OPCODE(ins) == I386_LEAVE) || 
	  (I386_INS_OPCODE(ins) == I386_POP &&
	   I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg &&
	   I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_EBP &&
	   I386_OP_REGMODE(I386_INS_DEST(ins)) == i386_regmode_full32)
	  ))
      continue;
    ins = INS_PREV(ins);
    if (I386_INS_OPCODE(ins) == I386_MOV &&
	I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg &&
	I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_EBP &&
	I386_OP_REGMODE(I386_INS_SOURCE1(ins)) == i386_regmode_full32 &&
	I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg &&
	I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_ESP &&
	I386_OP_REGMODE(I386_INS_DEST(ins)) == i386_regmode_full32)
      ins = INS_PREV(ins);
    epilogue = INS_NEXT(ins);

    /* check if we can really remove the frame pointer */
    for (ins = prologue; ins != epilogue; ins = INS_NEXT(ins))
    {
      if (RegsetIn(INS_REGS_USE(ins),I386_REG_EBP) ||
	  RegsetIn(INS_REGS_DEF(ins),I386_REG_EBP))
      {
	can_remove = FALSE;
	break;
      }
      if (RegsetIn(INS_REGS_USE(ins),I386_REG_EBP) ||
	  RegsetIn(INS_REGS_DEF(ins),I386_REG_EBP))
      {
	if (I386_INS_OPCODE(ins) == I386_PUSH)
	  height += 4;
	else if (I386_INS_OPCODE(ins) == I386_POP)
	  height -= 4;
	else if (I386_INS_OPCODE(ins) == I386_SUB &&
	    I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg &&
	    I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_ESP &&
	    I386_OP_REGMODE(I386_INS_DEST(ins)) == i386_regmode_full32 &&
	    I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	  height += I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins));
	else if (I386_INS_OPCODE(ins) == I386_ADD &&
	    I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg &&
	    I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_ESP &&
	    I386_OP_REGMODE(I386_INS_DEST(ins)) == i386_regmode_full32 &&
	    I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_imm)
	  height -= I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins));
	else
	{
	  can_remove = FALSE;
	  break;
	}

	if (height < 0)
	{
	  can_remove = FALSE;
	  break;
	}
      }
    }

    /* remove the prologue and epilogue */
    if (can_remove)
    {
      /*DiabloPrint(stdout,"FPR:\nbefore @iB\n",first);*/
      while (INS_PREV(prologue))
	InsKill(INS_PREV(prologue));
      ins = INS_PREV(epilogue);
      while (INS_PREV(BBL_INS_LAST(first)) != ins)
	InsKill(INS_PREV(BBL_INS_LAST(first)));
      /*DiabloPrint(stdout,"after: @iB\n",first);*/
    }

  }
  
  I386FramePointerAnalysis(cfg);
}
/* }}} */

#ifdef INSTRUMENTATION
/* {{{ mov $val,%reg ; push %reg --> mov $val,%reg ; push $val 
 * in fact, this will only do something for instrumentation code, and only for
 * as long as copy propagation and constant propagation are not properly implemented
 * for i386 */
static void I386PeepHoleMovPush(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  static int count = 0;

  CFG_FOREACH_BBL(cfg,bbl)
  {
    BBL_FOREACH_INS(bbl,ins)
    {
      t_i386_operand * op1, * op2, * op3;
      t_ins * next;

      if (I386_INS_OPCODE(ins) != I386_MOV) 
	continue;
      op1 = I386_INS_DEST(ins);
      if (I386_OP_TYPE(op1) != i386_optype_reg || !I386IsGeneralPurposeReg(I386_OP_BASE(op1)))
	continue;
      next = INS_NEXT(ins);
      if (!next || I386_INS_OPCODE(next) != I386_PUSH)
	continue;
      op2 = I386_INS_SOURCE1(next);
      if (I386_OP_TYPE(op2) != i386_optype_reg || I386_OP_BASE(op2) != I386_OP_BASE(op1) || I386_OP_REGMODE(op2) != I386_OP_REGMODE(op1))
	continue;
      op3 = I386_INS_SOURCE1(ins);
      if (I386_OP_TYPE(op3) == i386_optype_mem && (I386_OP_BASE(op3) == I386_OP_BASE(op1) || I386_OP_INDEX(op3) == I386_OP_BASE(op1)))
	continue;
	
      /*if (count >= global_options.debugcounter) continue;*/

      count++;
      memcpy(op2,op3,sizeof(t_i386_operand));
      if (I386_OP_FLAGS(op3) & I386_OPFLAG_ISRELOCATED)
      {
	t_reloc * rel = I386GetRelocForOp(ins,op3);
	t_reloc * rel2;
	VERIFY_FASSERT(rel,("relocations corrupt"));
	rel2 = RelocTableAddRelocToRelocatable(cfg->sec->obj->reloc_table, rel->addend, (t_relocatable *)next, rel->from_offset,
	    REL_TO_RELOCATABLE(rel),rel->to_offset,rel->hell,rel->edge,rel->corresponding,rel->e.sec,rel->code,TRUE);
	REL_FLAGS(rel2) = REL_FLAGS(rel);
      }
      INS_REGS_USE(next) = I386InsUsedRegisters(next);
      INS_REGS_DEF(next) = I386InsDefinedRegisters(next);
    }
  }
  VERBOSE(0,("[mov-push elimination]: %d times",count));
} /* }}} */
#endif

void I386PeepHoles(t_cfg * cfg)
{
  I386PeepHoleFramePointerRemoval(cfg);
  I386PeepHoleStack(cfg);
  I386PeepHoleRedundantPushes(cfg);
  I386PeepHoleShorterEquivalent(cfg);
  I386PeepHoleAddSub(cfg);
  I386PeepHoleIdempotent(cfg);
}

void I386PeepHolesInstrumentation(t_cfg * cfg)
{
#ifdef INSTRUMENTATION
  I386PeepHoleFramePointerRemoval(cfg);
  if (!use_old_addresses)
    I386PeepHoleStack(cfg);
  I386PeepHoleRedundantPushes(cfg);
  I386PeepHoleRedundantFSaves(cfg);
  I386PeepHoleShorterEquivalent(cfg);
  I386PeepHoleAddSub(cfg);
  I386PeepHoleIdempotent(cfg);
  I386PeepHoleMovPush(cfg);
  I386PeepHolePushESP(cfg);
  I386PeepHoleMyprof(cfg);
  I386PeepHoleSimpleCopyPropagation(cfg);
  if (!use_old_addresses)
    I386PeepHoleStack(cfg);
#endif
}
/* vim: set shiftwidth=2 foldmethod=marker : */
