/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "i386_instructions.h"
#include "i386_opcode.h"
#include "i386_registers.h"
#include "../../datastructures/data_string.h"
#include "../../utils/util_error.h"
#include <ctype.h>
#include <string.h>

static t_string conditioncodes[] = {
  "o", "no", "b", "ae", "e", "ne", "be", "a",
  "s", "ns", "p", "np", "l", "ge", "le", "g"
};

static t_string regs32[] = {
  "%eax", "%ebx", "%ecx", "%edx",
  "%esi", "%edi", "%ebp", "%esp"
};

static t_string regs16[] = {
  "%ax", "%bx", "%cx", "%dx",
  "%si", "%di", "%bp", "%sp"
};

static t_string regs8hi[] = {
  "%ah", "%bh", "%ch", "%dh"
};

static t_string regs8lo[] = {
  "%al", "%bl", "%cl", "%dl"
};

static t_string regsseg[] = {
  "%cs", "%ds", "%es", "%fs", "%gs", "%ss"
};

static t_string regsfp[] = {
  "%st", "%st1", "%st2", "%st3",
  "%st4", "%st5", "%st6", "%st7"
};

static t_string regscontrol[] = {
  "%cr0", "%cr1", "%cr2", "%cr3",
  "%cr4", "%cr5", "%cr6", "%cr7"
};

static t_string regsdebug[] = {
  "%dr0", "%dr1", "%dr2", "%dr3",
  "%dr4", "%dr5", "%dr6", "%dr7"
};

static void PrintPrefixes(t_i386_ins * ins, t_string buf)
{
  /* for now, only print repz and repnz prefixes */
  if (I386_INS_HAS_PREFIX(ins, I386_PREFIX_REPNZ))
  {
    sprintf(buf,"repnz ");
  }
  else if (I386_INS_HAS_PREFIX(ins, I386_PREFIX_REP))
  {
    sprintf(buf,"repz ");
  }
  else
    buf[0] = '\0';
}

static void PrintReg(t_i386_operand * op, t_string buf)
{
  t_reg reg = I386_OP_BASE(op);
  int index;

  if (reg >= I386_REG_EAX && reg <= I386_REG_ESP)
  {
    index = reg - I386_REG_EAX;
    switch (I386_OP_REGMODE(op))
    {
      case i386_regmode_full32:
	sprintf(buf,"%s",regs32[index]);
	break;
      case i386_regmode_lo16:
	sprintf(buf,"%s",regs16[index]);
	break;
      case i386_regmode_lo8:
	sprintf(buf,"%s",regs8lo[index]);
	break;
      case i386_regmode_hi8:
	sprintf(buf,"%s",regs8hi[index]);
	break;
    }
  }
  else if (reg >= I386_REG_ST0 && reg <= I386_REG_ST7)
  {
    index = reg - I386_REG_ST0;
    sprintf(buf,"%s",regsfp[index]);
  }
  else if (reg >= I386_REG_CS && reg <= I386_REG_SS)
  {
    index = reg - I386_REG_CS;
    sprintf(buf,"%s",regsseg[index]);
  }
  else if (reg >= I386_REG_CR0 && reg <= I386_REG_CR4)
  {
    index = reg - I386_REG_CR0;
    sprintf(buf,"%s",regscontrol[index]);
  }
  else if (reg >= I386_REG_DR0 && reg <= I386_REG_DR7)
  {
    index = reg - I386_REG_DR0;
    sprintf(buf,"%s",regsdebug[index]);
  }
  else
    FATAL(("unknown register %d",reg));
}

static void PrintMem(t_i386_operand * op, t_string buf)
{
  t_bool base = (I386_OP_BASE(op) != I386_REG_NONE);
  t_bool index = (I386_OP_INDEX(op) != I386_REG_NONE);
  t_bool longform = base || index;
  t_uint32 pos = 0;

  if (I386_OP_IMMEDIATE(op) != 0 || !longform)
    pos = sprintf(buf, "0x%x", I386_OP_IMMEDIATE(op));

  if (longform)
  {
    pos += sprintf(buf+pos,"(%s%c",base ? regs32[I386_OP_BASE(op)] : "", index ? ',' : ')');
    if (index)
      sprintf(buf+pos,"%s,%d)", regs32[I386_OP_INDEX(op)], 1 << I386_OP_SCALE(op));
  }
}

static t_bool PrintOp(t_i386_ins * ins, t_i386_operand * op, t_string buf)
{
  switch (I386_OP_TYPE(op))
  {
    case i386_optype_none:
      return FALSE;

    case i386_optype_reg:
      if (INS_TYPE(ins) == IT_BRANCH)
      {
	/* add a '*' before the operand */
	buf[0] = '*';
	buf++;
      }
      PrintReg(op,buf);
      break;

    case i386_optype_imm:
      if (INS_TYPE(ins) != IT_BRANCH || I386_INS_OPCODE(ins) == I386_RET)
      {
	sprintf(buf, "$0x%x", I386_OP_IMMEDIATE(op));
      }
      else
      {
	sprintf(buf, "%x", G_T_UINT32(GENERIC_ADD(INS_CADDRESS(ins),INS_CSIZE(ins))) + I386_OP_IMMEDIATE(op));
      }
      break;

    case i386_optype_mem:
      if (INS_TYPE(ins) == IT_BRANCH)
      {
	/* add a '*' before the operand */
	buf[0] = '*';
	buf++;
      }
      PrintMem(op,buf);
      break;

    case i386_optype_farptr:
      sprintf(buf, "_FAR_PTR_: TODO");
      break;

    default:
      FATAL(("invalid i386 instruction"));
  }
  return TRUE;
}

void I386InstructionPrint(t_ins * ins, t_string outputstring) 
{
  char pref[20], opc[20], dest[20], src1[20], src2[20];
  t_uint32 pos;
  t_bool has_dest, has_src1, has_src2;

  /* prefixes */
  PrintPrefixes((t_i386_ins *) ins,pref);

  /* the opcode */
  sprintf(opc, "%s", i386_opcode_table[I386_INS_OPCODE(ins)].textual);
  if (PatternMatch("*CC*",opc))
  {
    /* print condition code */
    for (pos = 0; pos < strlen(opc) - 1; pos++) 
      if (opc[pos] == 'C' && opc[pos+1] == 'C')
	break;
    sprintf(opc+pos, "%s", conditioncodes[I386_INS_CONDITION(ins)]);
  }

  /* the operand size suffix */
  /* this is only necessary if the right operand size can not be
   * determined without the suffix, i.e. if there are no register operands */
  if (INS_TYPE(ins) != IT_BRANCH)
  {
    t_bool suffix_needed = FALSE, suffix_not_needed = FALSE;
    if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg ||
	I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg ||
	I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_reg)
      suffix_not_needed = TRUE;
    
    if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_mem ||
	I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_mem ||
	I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_mem)
      suffix_needed = TRUE;

    if (suffix_needed && !suffix_not_needed)
    {
      char suffix;
      int opsize = 0;
      if (I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_mem)
	opsize = I386_OP_MEMOPSIZE(I386_INS_DEST(ins));
      if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_mem)
	opsize = I386_OP_MEMOPSIZE(I386_INS_SOURCE1(ins));
      if (I386_OP_TYPE(I386_INS_SOURCE2(ins)) == i386_optype_mem)
	opsize = I386_OP_MEMOPSIZE(I386_INS_SOURCE2(ins));
      
      if (opsize == 1)
	suffix = 'b';
      else if (opsize == 2)
	suffix = 'w';
      else if (opsize == 4)
	suffix = 'l';
      else
	suffix = ' ';

      opc[strlen(opc)+1] = '\0';
      opc[strlen(opc)] = suffix;
    }
  }

  /* the destination operand */
  has_dest = PrintOp((t_i386_ins *) ins, I386_INS_DEST(ins), dest);

  /* the first source operand */
  has_src1 = PrintOp((t_i386_ins *) ins, I386_INS_SOURCE1(ins), src1);

  /* the second source operand */
  has_src2 = PrintOp((t_i386_ins *) ins, I386_INS_SOURCE2(ins), src2);
  

  /* put it all together */
  sprintf(outputstring, "%s%-7s%s%s%s%s%s",
      pref, opc,
      has_src2 ? src2 : "", (has_src2 && (has_src1 || has_dest)) ? "," : "",
      has_src1 ? src1 : "", (has_src1 && has_dest) ? "," : "",
      has_dest ? dest : "" );

  /* special case: data */
  if (I386_INS_OPCODE(ins) == I386_DATA)
    sprintf(outputstring, "%-7s%c %02x", "data", isprint(I386_INS_DATA(ins)) ? I386_INS_DATA(ins) : '.', (t_uint8) I386_INS_DATA(ins));
}
/* vim: set shiftwidth=2 foldmethod=marker: */
