/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file routines 
 * 
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/

#include "../../utils/util_error.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_function.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_constprop.h"
#include "../../kernel/diablo_print.h"
#include "i386_opcode.h"
#include "i386_registers.h"
#include <string.h>

/* {{{ I386FunComputeStackSavedRegisters */
#define MAX_NR_REG  16
void I386FunComputeStackSavedRegisters(t_function * fun)
{
  int i;
  t_bool use_fp;
  t_bbl *entryblock, *exitblock, *iter;
  t_ins * ins;
  t_cfg_edge * edge;
  t_bool conservative = FALSE;
  t_bool ret_pops_only_return_address = TRUE;

  /* be conservative with respect to anything that looks even remotely weird */
  entryblock = FUN_BBL_FIRST(fun);
  exitblock = FunctionGetExitBlock(fun);

  if (!exitblock) /* this can actually happen if the exitblock is killed because it was unreachable */
    conservative = TRUE;
  if (BBL_NINS(entryblock) == 0)
    conservative = TRUE;
  if (FUN_IS_HELL(fun))
    conservative = TRUE;
  if (entryblock == exitblock)
    conservative = TRUE;
  FUN_FOREACH_BBL(fun,iter)
    BBL_FOREACH_PRED_EDGE(iter,edge)
    {
      if (CfgEdgeIsForwardInterproc(edge) && EDGE_CAT(edge)!=ET_CALL)
	conservative = TRUE;
    }
  FUN_FOREACH_BBL(fun,iter)
    BBL_FOREACH_SUCC_EDGE(iter,edge)
    {
      if (CfgEdgeIsForwardInterproc(edge) && EDGE_CAT(edge)!=ET_CALL)
	conservative = TRUE;
    }

  if (!conservative)
  {
    t_reg savedregs[MAX_NR_REG];
    t_reg loadedregs[MAX_NR_REG];
    int nsaved = 0;
    int nloaded = 0;
    t_regset changedregs;
    
    ins = BBL_INS_FIRST(entryblock);

    /* {{{ the entry block */
    if (!(I386_INS_OPCODE(ins) == I386_PUSH && 		/* push %ebp */
	  I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg && 
	  I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_EBP))
      use_fp = FALSE;
    else
      use_fp = TRUE;

    if (use_fp)
    {
      ins = INS_NEXT(ins);
      if (!(ins && I386_INS_OPCODE(ins) == I386_MOV &&		/* mov %esp, %ebp */
	    I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg &&
	    I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg &&
	    I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_EBP &&
	    I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_ESP ))
      {
	conservative = TRUE;
      }
      if (ins) ins = INS_NEXT(ins);
    }

    changedregs = NullRegs;
    for (; ins; ins = INS_NEXT(ins))
    {
      t_i386_operand * op = I386_INS_SOURCE1(ins);

      RegsetSetUnion(changedregs,INS_REGS_DEF(ins));

      if (!I386InsIsStore(ins) && !RegsetIn(INS_REGS_DEF(ins),I386_REG_ESP))
	continue;

      if (I386_INS_OPCODE(ins) != I386_PUSH) break;
      if (I386_OP_TYPE(op) != i386_optype_reg) break;
      if (RegsetIn(changedregs,I386_OP_BASE(op))) break;

      savedregs[nsaved++] = I386_OP_BASE(op);
    }
    for (i = nsaved; i < MAX_NR_REG; i++)
      savedregs[i] = I386_REG_NONE;
    /* }}} */

    /* {{{ the exit blocks */
    nloaded = -1;
    BBL_FOREACH_PRED_EDGE(exitblock,edge)
    {
      int tmploaded = 0;

      iter = CFG_EDGE_HEAD(edge);
      
      if (BBL_NINS(iter) == 0)
      {
	conservative = TRUE;
	break;
      }

      ins = BBL_INS_LAST(iter);
      if (I386_INS_OPCODE(ins) != I386_RET)
	FATAL(("procedure exit but no ret instruction: @iB\nexitblock @iB\nedge @E",iter,exitblock,edge));
      if (I386_OP_TYPE(I386_INS_SOURCE1(ins)) != i386_optype_none)
	ret_pops_only_return_address = FALSE;

      ins = INS_PREV(ins);

      if (use_fp)
      {
	if (!(ins && 
	      (I386_INS_OPCODE(ins) == I386_LEAVE ||
	       (I386_INS_OPCODE(ins) == I386_POP && 
		I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg && 
		I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_EBP))))
	{
	  conservative = TRUE;
	  nloaded = 0;
	  break;
	}
	if (ins) ins = INS_PREV(ins);
      }
      
      for (; ins && tmploaded != nloaded; ins = INS_PREV(ins))
      {
	t_i386_operand * op = I386_INS_DEST(ins);

	if (I386_INS_OPCODE(ins) != I386_POP) break;
	if (I386_OP_TYPE(op) != i386_optype_reg) break;
	
	if (nloaded == -1)
	{
	  loadedregs[tmploaded++] = I386_OP_BASE(op);
	}
	else
	{
	  if (loadedregs[tmploaded] == I386_OP_BASE(op))
	    tmploaded++;
	  else
	    break;
	}
      }
      if (nloaded == -1) 
	nloaded = tmploaded;
      else
	nloaded = (nloaded < tmploaded) ? nloaded : tmploaded;
    }
    for (i = nloaded; i < MAX_NR_REG; i++)
      loadedregs[i] = I386_REG_NONE;
    /* }}} */

    /* {{{ check if the saved & loaded registers match */
    RegsetDup(FUN_REGS_SAVED(fun),NullRegs);
    for (i = 0; (i < nsaved) && (i < nloaded); i++)
    {
      if (savedregs[i] == loadedregs[i])
	RegsetAddReg(FUN_REGS_SAVED(fun),savedregs[i]);
      else
	break;
    }
    if (use_fp)
      RegsetAddReg(FUN_REGS_SAVED(fun),I386_REG_EBP);
    if (ret_pops_only_return_address)
      RegsetAddReg(FUN_REGS_SAVED(fun),I386_REG_ESP);
    /* }}} */
  }

  if (conservative)
  {
    RegsetDup(FUN_REGS_SAVED(fun),NullRegs);
  }
} /* }}} */

/* {{{ I386FramePointerAnalysis */
void I386FramePointerAnalysis(t_cfg * cfg)
{
  /* for each function in the cfg, check if it sets the frame pointer (%ebp),
   * and set the FF_HAS_FRAMEPOINTER flag accordingly. If we can't find out
   * whether the frame pointer is set, we assume it isn't set */
  t_function * fun;
  CFG_FOREACH_FUN(cfg,fun)
  {
    t_bbl * first = FUN_BBL_FIRST(fun);
    t_bbl * last = FunctionGetExitBlock(fun);
    t_ins * ins;
    t_bool fp_set_up = FALSE;
    t_bool fp_reset = TRUE;

    if (!first || !last)
    {
      FUN_FLAGS(fun) &= ~FF_HAS_FRAMEPOINTER;
      continue;
    }

    /* is the frame pointer set up at the entry point? */
    {
      t_regset changed = NullRegs;
      t_bool push_ebp = FALSE, set_ebp = FALSE;
      BBL_FOREACH_INS(first,ins)
      {
	if (!push_ebp && RegsetIn(changed,I386_REG_EBP))
	  break;
	if (RegsetIn(changed,I386_REG_ESP))	/* the stack changed induced by the push %ebp is not recorded in changed */
	  break;
	if (!push_ebp 
	    && I386_INS_OPCODE(ins) == I386_PUSH
	    && I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg
	    && I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_EBP)
	{
	  push_ebp = TRUE;
	  continue;
	}
	if (push_ebp
	    && I386_INS_OPCODE(ins) == I386_MOV
	    && I386_OP_TYPE(I386_INS_SOURCE1(ins)) == i386_optype_reg
	    && I386_OP_BASE(I386_INS_SOURCE1(ins)) == I386_REG_ESP
	    && I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg
	    && I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_EBP)
	{
	  set_ebp = TRUE;
	  break;
	}
	RegsetSetUnion(changed,INS_REGS_DEF(ins));
      }
      fp_set_up = set_ebp;
    }

    if (fp_set_up)
    {
      t_cfg_edge * edge;
      BBL_FOREACH_PRED_EDGE(last,edge)
      {
	t_bbl * head = CFG_EDGE_HEAD(edge);
	if (EDGE_CAT(edge) == ET_COMPENSATING) continue;
	ins = BBL_INS_LAST(head);
	if (!ins || I386_INS_OPCODE(ins) != I386_RET)
	{
	  fp_reset = FALSE;
	  break;
	}
	ins = INS_PREV(ins);
	if (ins && (I386_INS_OPCODE(ins) == I386_LEAVE || 
	      (I386_INS_OPCODE(ins) == I386_POP && I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_EBP)))
	  continue;

	/* if we get here, we didn't see the instructions we expected */
	fp_reset = FALSE;
	break;
      }
    }

    if (fp_set_up && fp_reset)
      FUN_FLAGS(fun) |= FF_HAS_FRAMEPOINTER;
    else
      FUN_FLAGS(fun) &= ~FF_HAS_FRAMEPOINTER;
  }
}

t_bool I386FunIsGlobal(t_function* fun)
{
  if (fun->function_flags & FF_IS_EXPORTED)
  {
    return TRUE;
  }
  return FALSE;
}

/* }}} */


/* stack height analysis
 * ---------------------
 *
 * This analysis tries to determine for each basic block 
 * in a function the height of the stack at the beginning
 * of that block.
 *
 * The analysis is context-sensitive: in a first step, 
 * it tries to determine the stack delta caused by following
 * any interprocedural edge (this is the same for all edges
 * that have the same tail). This information is stored
 * in the edge->stack_delta field.
 *
 * The second step calculates the stack height at the beginning
 * each basic block by setting the stack height to 0 at every 
 * procedure entry point (i.e. all blocks where forward
 * interprocedural edges enter a function), and doing an 
 * intraprocedural fixpoint calculation using the information
 * on the outgoing interprocedural edges. The stack height
 * at the beginning of each basic block is stored in 
 * bbl->stack_in.
 *
 * The analysis tries to account for the use of a frame 
 * pointer, by detecting whether one is set at the function
 * entry and whether the stack pointer is reset to it
 * at the end of the function */

#define DEBUG_STACKHEIGHT

#define STACKBOT ((t_int32) 0x80000000)
#define STACKTOP ((t_int32) 0x7fffffff)

/* for each bbl, fill in following fields:
 * bbl->reset_to_framepointer: TRUE if the block resets the stack pointer
 * to the value stored in the frame pointer
 * bbl->set_framepointer: TRUE if the block sets the frame pointer
 * bbl->framepointer_offset: stack height at which the frame pointer points
 * relative to the beginning stack height at the beginning of the bbl
 * bbl->stack_delta: if the stack pointer is not reset to the frame pointer,
 * this field contains the number of bytes the stack grows or shrinks in this
 * bbl. if the stack pointer _is_ reset to the frame pointer, this field 
 * indicates what the stack delta was _after_ this reset.
 */
/* {{{ ComputeStackDelta */
static void ComputeStackDelta(t_bbl * bbl)
{
  t_int32 delta = 0;
  t_ins * ins;

  bbl->reset_to_framepointer = FALSE;
  bbl->set_framepointer = FALSE;
  
  if (IsHellNode(bbl) || IS_DATABBL(bbl))
  {
    bbl->stack_delta = 0;
    return;
  }

  BBL_FOREACH_INS(bbl,ins)
  {
    t_i386_operand * src = I386_INS_SOURCE1(ins);

    if (I386_INS_OPCODE(ins) == I386_MOV &&
	I386_OP_TYPE(I386_INS_DEST(ins)) == i386_optype_reg && I386_OP_BASE(I386_INS_DEST(ins)) == I386_REG_EBP && 
	I386_OP_TYPE(src) == i386_optype_reg && I386_OP_BASE(src) == I386_REG_ESP)
    {
      /* mov %esp, %ebp */
      bbl->set_framepointer = TRUE;
      bbl->framepointer_offset = delta;
    }
    
    if (!RegsetIn(INS_REGS_DEF(ins),I386_REG_ESP)) continue;
    if (delta == STACKBOT)
    {
      if (I386_INS_OPCODE(ins) != I386_LEAVE 
	  && I386_INS_OPCODE(ins) != I386_LEA 
	  && I386_INS_OPCODE(ins) != I386_MOV) 
	continue;
    }

    switch (I386_INS_OPCODE(ins))
    {
      case I386_POP:
      case I386_POPF:
	delta -= 4;
	break;
      case I386_POPA:
	delta -= 32;
	break;
      case I386_PUSH:
      case I386_PUSHF:
	delta += 4;
	break;
      case I386_PUSHA:
	delta += 32;
	break;
      case I386_ADD:
	if (I386_OP_TYPE(src) == i386_optype_imm)
	  delta -= I386_OP_IMMEDIATE(src);
	else
	  delta = STACKBOT;
	break;
      case I386_SUB:
	if (I386_OP_TYPE(src) == i386_optype_imm)
	  delta += I386_OP_IMMEDIATE(src);
	else
	  delta = STACKBOT;
	break;
      case I386_MOV:
	if (I386_OP_TYPE(src) == i386_optype_reg && I386_OP_BASE(src) == I386_REG_EBP)
	  bbl->reset_to_framepointer = TRUE;
	delta = 0;
	break;
      case I386_LEAVE:
	bbl->reset_to_framepointer = TRUE;
	delta = -4;
	break;
      case I386_LEA:
	if (I386_OP_BASE(src) == I386_REG_EBP && I386_OP_INDEX(src) == I386_REG_NONE)
	{
	  bbl->reset_to_framepointer = TRUE;
	  delta = -(t_int32)I386_OP_IMMEDIATE(src);
	}
	else
	  delta = STACKBOT;
	break;
      case I386_CALL:
      case I386_INT:
	/* for this analysis we act as if they don't change the stack pointer */
	break;
      case I386_RET:
	if (I386_OP_TYPE(src) == i386_optype_imm)
	  delta -= I386_OP_IMMEDIATE(src);
	break;
      default:
	/*DiabloPrint(stdout,"STACKBOTTING out for @I\n",ins);*/
	delta = STACKBOT;
    }
    /*if (delta == STACKBOT) DiabloPrint(stdout,"STACKBOT for @I\n",ins);*/
  }
  bbl->stack_delta = delta;
} /* }}} */

/* {{{ join functions */
static t_int32 Join(t_int32 a, t_int32 b)
{
  if (a == STACKBOT || b == STACKBOT) return STACKBOT;
  if (a == STACKTOP) return b;
  if (b == STACKTOP) return a;
  if (a == b) return a;
  return STACKBOT;
}

static t_bool EdgeDeltaJoin(t_cfg_edge * edge, t_int32 delta)
{
  if (delta != edge->stack_delta)
  {
    edge->stack_delta = delta;
    return TRUE;
  }
  return FALSE;
}

static t_bool BlockDeltaJoin(t_bbl * bbl, t_int32 delta)
{
  delta = Join(bbl->stack_in,delta);
  if (delta != bbl->stack_in)
  {
    bbl->stack_in = delta;
    return TRUE;
  }
  return FALSE;
}
/* }}} */

/* {{{ function stack height propagation for a particular entry block */
/* after the fixpoint is reached, call this function with entry == NULL
 * to perform a final intraprocedural propagation that takes into account
 * all entry blocks at once */
static void DoRealFunctionProp(t_function * fun, t_bbl * entry)
{
  t_bbl * bbl;
  t_bbl * exit_block = FunctionGetExitBlock(fun);
  t_cfg_edge * edge;
  t_bool entry_sets_framepointer;
  t_int32 framepointer_offset;
  t_int32 fun_stack_delta;

  if (FUN_IS_HELL(fun)) return;

  if (!entry)
  {
    /* not during fixpoint calculations - we don't care about frame pointers */
    entry_sets_framepointer = FALSE;
    framepointer_offset = 0;
  }
  else
  {
    entry_sets_framepointer = entry->set_framepointer;
    framepointer_offset = entry->framepointer_offset;
    /* {{{ is the framepointer set at function entry? */
    /* we only consider the frame pointer to be set when it is set in the entry block and nowhere else */
    FUN_FOREACH_BBL(fun,bbl)
    {
      t_bool allowed_def = TRUE, found_allowed_def = FALSE;
      if (bbl == entry) continue;
      if (!RegsetIn(BBL_REGS_DEFINED(bbl),I386_REG_EBP)) continue;
      /* the only definitions of %ebp that are allowed are leave and pop %ebp in functions that reset the stack pointer to the frame pointer */
      if (bbl->reset_to_framepointer)
      {
	t_ins * ins;
	BBL_FOREACH_INS(bbl,ins)
	{
	  if (RegsetIn(INS_REGS_DEF(ins),I386_REG_EBP))
	  {
	    if (!found_allowed_def && (I386_INS_OPCODE(ins) == I386_LEAVE ||
		  (I386_INS_OPCODE(ins) == I386_POP && INS_PREV(ins) && I386_INS_OPCODE(INS_PREV(ins)) == I386_MOV &&
		   I386_OP_TYPE(I386_INS_DEST(INS_PREV(ins))) == i386_optype_reg && I386_OP_BASE(I386_INS_DEST(INS_PREV(ins))) == I386_REG_ESP &&
		   I386_OP_TYPE(I386_INS_SOURCE1(INS_PREV(ins))) == i386_optype_reg && I386_OP_BASE(I386_INS_SOURCE1(INS_PREV(ins))) == I386_REG_EBP)))
	      found_allowed_def = TRUE;
	    else
	      allowed_def = FALSE;
	  }
	}
      }
      if (!allowed_def)
	entry_sets_framepointer = FALSE;
    }
    /* }}} */
  }

  /* {{{ initialization */
  FUN_FOREACH_BBL(fun,bbl)
    bbl->stack_in = STACKTOP;
  if (entry)
  {
    entry->stack_in = 0;
    FunctionMarkBbl(fun,entry);
  }
  else
  {
    /* mark all blocks with incoming forward interprocedural edges */
    FUN_FOREACH_BBL(fun,bbl)
    {
      BBL_FOREACH_PRED_EDGE(bbl,edge)
	if (CfgEdgeIsForwardInterproc(edge))
	  break;
      if (edge)
      {
	bbl->stack_in = 0;
	FunctionMarkBbl(fun,bbl);
      }
    }
  } /* }}} */

  while (FunctionUnmarkBbl(fun,&bbl))
  {
    /* real propagation */
    t_int32 stack_out;
    t_cfg_edge * edge;

    /* {{{ determine stack_out */
    if (bbl->reset_to_framepointer)
    {
      if (entry_sets_framepointer)
	stack_out = framepointer_offset + bbl->stack_delta;
      else
	stack_out = STACKBOT;
    }
    else
    {
      stack_out = (bbl->stack_in == STACKBOT || bbl->stack_delta == STACKBOT) ? STACKBOT : (bbl->stack_in + bbl->stack_delta);
    } /* }}} */

    /* propagation over non-interprocedural edges */
    BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      if (CfgEdgeIsInterproc(edge) || EDGE_CAT(edge) == ET_FUNLINK) continue;
      if (BlockDeltaJoin(CFG_EDGE_TAIL(edge),stack_out))
	FunctionMarkBbl(fun,CFG_EDGE_TAIL(edge));
    }

    /* propagation over interprocedural edges (but ignore swi-edges, syscalls don't change the stack) */
    BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      t_cfg_edge * corr_edge = edge->corr_edge;
      t_bbl * next;
      t_int32 tmp_stack_out;

      if (!CfgEdgeIsForwardInterproc(edge) || EDGE_CAT(edge) == ET_SWI) continue;

      if (!corr_edge) FATAL(("expected corresponding edge for @E",edge));
      next = CFG_EDGE_TAIL(corr_edge);
      if (BBL_FUN(next) != fun)
      {
	DiabloPrint(stdout,"!xx!!xx! vind een oplossing voor interprocedurale funlinks\n");
	continue;
      }

      tmp_stack_out = (stack_out == STACKBOT || edge->stack_delta == STACKBOT) ? STACKBOT : stack_out + edge->stack_delta;
      if (BlockDeltaJoin(next,tmp_stack_out))
	FunctionMarkBbl(fun,next);
    }
  }

  /* determine fun_stack_delta */
  if (entry)
  {
    if (!exit_block || (BBL_ATTRIB(exit_block) & BBL_UNREACHABLE))
    {
      /* assume that functions that do not return could just as well be considered to be well-behaved */
      fun_stack_delta = 0;
    }
    else if (exit_block->stack_in == STACKTOP)
    {
      /* this means the exit block of the function is not reachable for this given entry point:
       * entering the function via this point means it will never return... */
      fun_stack_delta = 0;
    }
    else
    {
      fun_stack_delta = exit_block->stack_in;
    }

    BBL_FOREACH_PRED_EDGE(entry,edge)
    {
      if (!CfgEdgeIsForwardInterproc(edge)) continue;
      if (EdgeDeltaJoin(edge,fun_stack_delta))
	CfgMarkFun(fun->cfg,BBL_FUN(CFG_EDGE_HEAD(edge)));
    }
  }
}
/* }}} */

static void FunctionStackHeightAnalysis(t_function * fun)
{
  t_bbl * entry;
  t_cfg_edge * edge;

  FUN_FOREACH_BBL(fun,entry)
  {
    BBL_FOREACH_PRED_EDGE(entry,edge)
      if (CfgEdgeIsForwardInterproc(edge))
	break;
    if (!edge) continue;

    /* propagate function for this entry block */
    DoRealFunctionProp(fun,entry);
  }
}

#ifdef DEBUG_STACKHEIGHT
static void labeled_Export_Function(t_function * fun, const char * filename);
#endif

void I386StackHeightAnalysis(t_cfg * cfg)
{
  t_function * fun;
  t_bbl * bbl;
  t_cfg_edge * edge;
  int nfunc = 0, nzero = 0, nbot = 0, nother = 0;
  t_bool bot,zero,other;
  t_bool multi_entry;


  /* initialize */
  InitEdgeMarking();
  CfgUnmarkAllFun(cfg);
  CFG_FOREACH_FUN(cfg,fun)
  {
    FunctionUnmarkAllEdges(fun);
    FunctionUnmarkAllBbls(fun);
  }
  CFG_FOREACH_BBL(cfg,bbl)
    ComputeStackDelta(bbl);
  CFG_FOREACH_EDGE(cfg,edge)
    edge->stack_delta = 0;

  BBL_FOREACH_PRED_EDGE(cfg->hell_node,edge)
    edge->stack_delta = STACKBOT;
  BBL_FOREACH_PRED_EDGE(cfg->call_hell_node,edge)
    edge->stack_delta = STACKBOT;
  
  /* start propagation */
  CFG_FOREACH_FUN(cfg,fun)
  {
    if (fun != cfg->call_hell_function && fun != cfg->hell_function)
      CfgMarkFun(cfg,fun);
  }

  while (CfgUnmarkFun(cfg,&fun))
    FunctionStackHeightAnalysis(fun);

  /* propagate final information into all basic blocks */
  CFG_FOREACH_FUN(cfg,fun)
    DoRealFunctionProp(fun,NULL);

#ifdef DEBUG_STACKHEIGHT
  /* {{{ report statistics and dump annotated dot files */
    DirectoryAssure("./dots-stack",FALSE);
  /* print out gained info */
  CFG_FOREACH_FUN(cfg,fun)
  {
    t_cfg_edge * edge;
    t_bbl * bbl;
    char fname[100];
    nfunc++;

    bot=zero=other=FALSE;
    multi_entry=FALSE;
    FUN_FOREACH_BBL(fun,bbl)
      BBL_FOREACH_PRED_EDGE(bbl,edge)
      {
	if (CfgEdgeIsForwardInterproc(edge))
	{
	  bot = bot || (edge->stack_delta == STACKBOT);
	  zero = zero || (edge->stack_delta == 0);
	  other = other || (edge->stack_delta != STACKBOT && edge->stack_delta != 0);
	  if (bbl != FUN_BBL_FIRST(fun))
	    multi_entry=TRUE;
	}
      }
    if (other)
    {
      nother++;
      sprintf(fname,"./dots-stack/%s-other.dot",fun->name);
    }
    else if (bot)
    {
      nbot++;
      sprintf(fname,"./dots-stack/%s-bot.dot",fun->name);
    }
    else
    {
      nzero++;
      sprintf(fname,"./dots-stack/%s-zero.dot",fun->name);
    }
    if (multi_entry)
    {
      sprintf(fname+strlen(fname)-4,"-me.dot");
    }
    labeled_Export_Function(fun,fname);
  }
  DiabloPrint(stdout,"functions: %d zero, %d bot, %d other = %d total\n",nzero,nbot,nother,nfunc);
  /* }}} */
#endif
}

#ifdef DEBUG_STACKHEIGHT
/* {{{ export function graph with stack height info */
static void labeled_Export_Function(t_function * fun, const char * filename)
{
  t_ins * ins;
  t_bbl * bbl;
  t_cfg_edge * edge;
  t_bool show_callers = TRUE;
  int callers = 0;

  FILE * out = fopen(filename,"w");
  VERIFY_FASSERT(out,("Could not open %s for writing!",filename));

  /* only show callers if there are less than 10 */
  BBL_FOREACH_PRED_EDGE(FUN_BBL_FIRST(fun),edge)
    if (EDGE_CAT(edge) == ET_CALL)
      callers++;
  show_callers = (callers < 10);

  /* graph header */
  if (FUN_NAME(fun) && strlen(FUN_NAME(fun)))
    DiabloPrint(out, "digraph \"%s\" {\n\tnode [shape=box];\n",FUN_NAME(fun));
  else
    DiabloPrint(out, "digraph \"noname_@G\" {\n\tnode [shape=box]\n",BBL_CADDRESS(FUN_BBL_FIRST(fun)));

  /* graph body */
  /* {{{  regular nodes */
  FUN_FOREACH_BBL(fun,bbl)
  {
    if (bbl == bbl->cfg->hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->call_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"CALL HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->unique_entry_node)
    {
      fprintf(out,"\t\"%p\" [label=\"Entry\"];\n",bbl);
    }
    else if (bbl == FunctionGetExitBlock(fun))
    {
      fprintf(out,"\t\"%p\" [label=\"RETURN\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->exit_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"EXIT HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->exit_call_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"EXIT CALL HELL\"];\n",bbl);
    }
    else
    {
      fprintf(out,"\t\"%p\" ",bbl);
      DiabloPrint(out, "[label=\"@B\\n", bbl);

      if (bbl->stack_in == STACKBOT)
	DiabloPrint(out,"h: STACKBOT\\n");
      else if (bbl->stack_in == STACKTOP)
	DiabloPrint(out,"h: STACKTOP\\n");
      else
	DiabloPrint(out,"h: %d\\n",bbl->stack_in);

      if (bbl->stack_delta == STACKBOT)
	DiabloPrint(out,"delta: STACKBOT\\n");
      else if (bbl->stack_delta == STACKTOP)
	DiabloPrint(out,"delta: STACKTOP\\n");
      else
	DiabloPrint(out,"delta: %d\\n",bbl->stack_delta);

      BBL_FOREACH_INS(bbl,ins)
	DiabloPrint(out,"@I\\l",ins);
      fprintf(out,"\"];\n");
    }
  } /* }}} */

  /* {{{ successor edges */
  FUN_FOREACH_BBL(fun,bbl)
  {
    BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      t_bbl * tail = CFG_EDGE_TAIL(edge);

      if (!EdgeTestCategoryOr(edge,ET_INTERPROC))
      {
	t_string style, color;

	switch (EDGE_CAT(edge))
	{
	  case ET_FUNLINK:
	  case ET_SWILINK:
	    style = "solid";
	    color = "yellow";
	    break;
	  case ET_FALLTHROUGH:
	    style = "solid";
	    color = "green";
	    break;
	  case ET_UNKNOWN:
	    style = "dashed";
	    color = "purple";
	    break;
	  case ET_JUMP:
	  case ET_SWITCH:
	    style = "solid";
	    color = "black";
	    break;
	  default:
	    style = "dashed";
	    color = "blue";
	}

	/* intraprocedural edges */
	if (EDGE_CAT(edge)==ET_SWITCH) fprintf(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s,label=\"%d\"];\n",bbl,tail,style,color,edge->switchvalue);
	else fprintf(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",bbl,tail,style,color);
      }
      else
      {
	/* interprocedural edges */
	t_string estyle, ecolor;
	t_string ncolor;
	t_string nodetext;
	char nodename[20];
	char label[40];
	t_bool filled = FALSE;
	t_string destfun = BBL_FUN(tail)?(FUN_NAME(BBL_FUN(tail)) ? FUN_NAME(BBL_FUN(tail)) : "noname"):"nofun";

	/* don't print successors of the hell or call hell nodes (there's no way you can get them all to fit nicely in a drawing) */
	if (bbl == bbl->cfg->hell_node || bbl == bbl->cfg->call_hell_node)
	  continue;

	switch (EDGE_CAT(edge))
	{
	  case ET_CALL:
	    estyle = "solid";
	    ecolor = "red";
	    ncolor = "yellow";
	    sprintf(nodename, "f%p", tail);
	    nodetext = sDiabloPrint("%s (@B)",destfun,tail);
	    filled = TRUE;
	    break;
	  case ET_RETURN:
	    estyle = "solid";
	    ecolor = "black";
	    ncolor = "green";
	    sprintf(nodename, "r%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    break;
	  case ET_COMPENSATING:
	    estyle = "dashed";
	    ecolor = "black";
	    ncolor = "green";
	    sprintf(nodename, "c%p", tail);
	    nodetext = sDiabloPrint("return of %s",destfun);
	    break;
	  case ET_SWI:
	    estyle = "dashed";
	    ecolor = "red";
	    ncolor = "red";
	    sprintf(nodename,"i%p",tail); /* tail will be the hell node */
	    nodetext = sDiabloPrint("HELL");
	    break;
	  case ET_IPUNKNOWN:
	    estyle = "bold";
	    ecolor = "purple";
	    ncolor = "magenta";
	    filled = TRUE;
	    sprintf(nodename, "i%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    break;
	  case ET_IPFALLTHRU:
	  case ET_IPFUNLINK:
	  case ET_IPJUMP:
	  case ET_IPSWILINK:
	  case ET_IPSWITCH:
	    estyle = "solid";
	    ecolor = "red";
	    ncolor = "black";
	    sprintf(nodename, "i%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    break;
	  default:
	    /* make it stand out as something we haven't seen yet */
	    estyle = "bold";
	    ecolor = "magenta";
	    ncolor = "blue";
	    sprintf(nodename, "i%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    filled = TRUE;
	    break;
	}

	/* catch hell node and call hell node */
	if (tail == bbl->cfg->hell_node || tail == bbl->cfg->call_hell_node)
	{
	  sprintf(nodename,"h%p",tail);
	  Free(nodetext);
	  nodetext = sDiabloPrint("%sHELL", tail == bbl->cfg->hell_node ? "" : "CALL ");
	  ncolor = "red";
	  filled = TRUE;
	}

	if (CfgEdgeIsForwardInterproc(edge) && EDGE_CAT(edge) != ET_SWI)
	  sprintf(label,"delta %d",edge->stack_delta);
	else
	  label[0] = '\0';

	if (show_callers || EDGE_CAT(edge) != ET_RETURN)
	{
	  fprintf(out,"\t\"%p\" -> \"%s\" [style=%s,color=%s,label=\"%s\"];\n",bbl,nodename,estyle,ecolor,label);
	  fprintf(out,"\t\"%s\" [label=\"%s\",color=\"%s\"%s];\n",nodename,nodetext,ncolor,filled?",style=filled":"");
	}
	Free(nodetext);
      }
    }
  } /* }}} */

  /* {{{ predecessor edges */
  FUN_FOREACH_BBL(fun,bbl)
  {
    BBL_FOREACH_PRED_EDGE(bbl,edge)
    {
      t_bbl * head = CFG_EDGE_HEAD(edge);
      t_string destfun = BBL_FUN(head)?(FUN_NAME(BBL_FUN(head)) ? FUN_NAME(BBL_FUN(head)) : "noname"):"nofun";
      t_string estyle, ecolor;
      t_string ncolor;
      t_string nodetext;
      char label[40];
      char nodename[20];
      t_bool filled = FALSE;

      if (!EdgeTestCategoryOr(edge,ET_INTERPROC)) continue; /* all intraprocedural edges have already been handled as predecessors */
      if (bbl == bbl->cfg->hell_node || bbl == bbl->cfg->call_hell_node) continue; /* don't do this for hell nodes */

      switch (EDGE_CAT(edge))
      {
	case ET_CALL:
	  estyle = "solid";
	  ecolor = "red";
	  ncolor = "yellow";
	  sprintf(nodename,"F%p",head);
	  nodetext = sDiabloPrint("@B in %s",head,destfun);
	  break;
	case ET_IPUNKNOWN:
	  estyle = "bold";
	  ecolor = "purple";
	  ncolor = "blue";
	  sprintf(nodename,"I%p",head);
	  nodetext = sDiabloPrint("@B in %s",head,destfun);
	  break;
	case ET_IPFALLTHRU:
	case ET_IPFUNLINK:
	case ET_IPSWILINK:
	case ET_IPJUMP:
	case ET_IPSWITCH:
	  estyle = "solid";
	  ecolor = "red";
	  ncolor = "cyan";
	  sprintf(nodename,"I%p",head);
	  nodetext = sDiabloPrint("@B in %s",head,destfun);
	  break;
	case ET_COMPENSATING:
	  estyle = "solid";
	  ecolor = "blue";
	  ncolor = "black";
	  sprintf(nodename,"C%p",head);
	  nodetext = sDiabloPrint("return from %s",destfun);
	  break;
	case ET_RETURN:
	  estyle = "solid";
	  ecolor = "blue";
	  ncolor = NULL;
	  nodetext = NULL;
	  if (FUN_BBL_FIRST(BBL_FUN(head)) == bbl->cfg->hell_node || FUN_BBL_FIRST(BBL_FUN(head)) == bbl->cfg->call_hell_node)
	    sprintf(nodename,"h%p", FUN_BBL_FIRST(BBL_FUN(head)));
	  else
	    sprintf(nodename,"f%p", FUN_BBL_FIRST(BBL_FUN(head)));
	  break;
	default:
	  FATAL(("unexpected edge type"));
      }

      /* catch hell node and call hell node */
      if (head == bbl->cfg->hell_node || head == bbl->cfg->call_hell_node)
      {
	sprintf(nodename,"h%p",head);
	if (nodetext) Free(nodetext);
	nodetext = sDiabloPrint("%sHELL", head == bbl->cfg->hell_node ? "" : "CALL ");
	ncolor = "red";
	filled = TRUE;
      }
      /* catch the entry node */
      if (head == bbl->cfg->unique_entry_node)
      {
	sprintf(nodename,"e%p",head);
	Free(nodetext);
	nodetext = sDiabloPrint("Entry");
	ncolor = "blue";
	filled = TRUE;
      }

      if (CfgEdgeIsForwardInterproc(edge))
	sprintf(label,"delta %d",edge->stack_delta);
      else
	label[0] = '\0';

      if (show_callers || EDGE_CAT(edge) != ET_CALL)
      {
	fprintf(out,"\t\"%s\" -> \"%p\" [style=%s,color=%s,label=\"%s\"];\n",nodename,bbl,estyle,ecolor,label);
	if (nodetext)
	  fprintf(out,"\t\"%s\" [label=\"%s\",color=%s%s];\n",nodename,nodetext,ncolor,filled?",style=filled":"");
      }
      if (nodetext) 
	Free(nodetext);
    }
  } /* }}} */
  
  /* graph footer */
  DiabloPrint(out, "}\n");
  fclose(out);
} /* }}} */
#endif

/* vim: set shiftwidth=2 foldmethod=marker: */
