/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_constant_propagation.h"
#include "ia64_registers.h"
#include "ia64_instructions.h"
#include "ia64_bundle.h"
#include "ia64_utils.h"
#include "ia64_opcodes.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_liveness.h"

#define ABS(a) ((a)>=0?(a):-(a))


static t_uint16 Ia64InsDefinedRegCount(t_ins * ins) 
{
  t_uint16 i, count;
  t_regset mask;

  mask = Ia64DefinedRegisters(ins);
  count = 0;

  for (i=0; i</*IA64_TOTALREGS*/IA64_NONE; i++) 
    if (RegsetIn(mask, i)) count++;
  
  return count;
}







  

static void Ia64TestAndSetSingleCondition(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false, t_reg flag)
{
  register_content reg;

  if (BOT==ProcStateGetReg(state,flag,&reg))
    {
      
      register_content waarde;
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);

      waarde.i = TRUE;
      ProcStateSetReg(*state_true,flag,waarde);
      waarde.i = FALSE;
      ProcStateSetReg(*state_false,flag,waarde);
      
    }
  else if (G_T_UINT64(reg.i))
    {
      *state_true = state;
      *state_false = NULL;
    }
  else
    {
      *state_false = state;
      *state_true = NULL;
    }
    
}


/* This is a function that is attached to system call edges */

void Ia64TestAndSetConditionSetAllToBot(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = ProcStateNew();
  *state_false = state;
  ProcStateSetAllBot(*state_true,AllRegs());
}

/* This is a function for return and compensating edges: information
   must be propagated over all outgoing edges of an exit block, at
   least as far as we are concerned. The constant propagator might add
   different constraints, such as in a context-sensitive constant
   propagation. This function is also used for hell edges 

   In general it is used where more than two successor edges are found
   in the graph, which model anomalous control flow. They are not
   taken or not-taken edges, but reflect different execution contexts.

*/

void Ia64TestAndSetConditionAllPaths(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = ProcStateNew();
  *state_false = state;
  ProcStateDup(*state_true,state);
}



void Ia64TestAndSetCondition_P0(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true = state;
  *state_false = NULL;
}  

void Ia64TestAndSetCondition_P1(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P1);
}  

void Ia64TestAndSetCondition_P2(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P2);
}  

void Ia64TestAndSetCondition_P3(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P3);
}  

void Ia64TestAndSetCondition_P4(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P4);
}  

void Ia64TestAndSetCondition_P5(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P5);
}  

void Ia64TestAndSetCondition_P6(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P6);
}  

void Ia64TestAndSetCondition_P7(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P7);
}  

void Ia64TestAndSetCondition_P8(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P8);
}  


void Ia64TestAndSetCondition_P9(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P9);
}  

void Ia64TestAndSetCondition_P10(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P10);
}  

void Ia64TestAndSetCondition_P11(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P11);
}  

void Ia64TestAndSetCondition_P12(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P12);
}  

void Ia64TestAndSetCondition_P13(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P13);
}  

void Ia64TestAndSetCondition_P14(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P14);
}  

void Ia64TestAndSetCondition_P15(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P15);
}  

void Ia64TestAndSetCondition_P16(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P16);
}  

void Ia64TestAndSetCondition_P17(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P17);
}  

void Ia64TestAndSetCondition_P18(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P18);
}  

void Ia64TestAndSetCondition_P19(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P19);
}  

void Ia64TestAndSetCondition_P20(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P20);
}  

void Ia64TestAndSetCondition_P21(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P21);
}  

void Ia64TestAndSetCondition_P22(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P22);
}  

void Ia64TestAndSetCondition_P23(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P23);
}  

void Ia64TestAndSetCondition_P24(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P24);
}  

void Ia64TestAndSetCondition_P25(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P25);
}  

void Ia64TestAndSetCondition_P26(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P26);
}  

void Ia64TestAndSetCondition_P27(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P27);
}  

void Ia64TestAndSetCondition_P28(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P28);
}  

void Ia64TestAndSetCondition_P29(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P29);
}  

void Ia64TestAndSetCondition_P30(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P30);
}  

void Ia64TestAndSetCondition_P31(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P31);
}  

void Ia64TestAndSetCondition_P32(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P32);
}  

void Ia64TestAndSetCondition_P33(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P33);
}  

void Ia64TestAndSetCondition_P34(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P34);
}  

void Ia64TestAndSetCondition_P35(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P35);
}  

void Ia64TestAndSetCondition_P36(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P36);
}  

void Ia64TestAndSetCondition_P37(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P37);
}  

void Ia64TestAndSetCondition_P38(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P38);
}  

void Ia64TestAndSetCondition_P39(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P39);
}  

void Ia64TestAndSetCondition_P40(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P40);
}  

void Ia64TestAndSetCondition_P41(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P41);
}  

void Ia64TestAndSetCondition_P42(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P42);
}  

void Ia64TestAndSetCondition_P43(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P43);
}  

void Ia64TestAndSetCondition_P44(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P44);
}  

void Ia64TestAndSetCondition_P45(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P45);
}  

void Ia64TestAndSetCondition_P46(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P46);
}  

void Ia64TestAndSetCondition_P47(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P47);
}  

void Ia64TestAndSetCondition_P48(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P48);
}  

void Ia64TestAndSetCondition_P49(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P49);
}  

void Ia64TestAndSetCondition_P50(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P50);
}  

void Ia64TestAndSetCondition_P51(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P51);
}  

void Ia64TestAndSetCondition_P52(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P52);
}  

void Ia64TestAndSetCondition_P53(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P53);
}  

void Ia64TestAndSetCondition_P54(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P54);
}  

void Ia64TestAndSetCondition_P55(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P55);
}  

void Ia64TestAndSetCondition_P56(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P56);
}  

void Ia64TestAndSetCondition_P57(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P57);
}  

void Ia64TestAndSetCondition_P58(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P58);
}  

void Ia64TestAndSetCondition_P59(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P59);
}  

void Ia64TestAndSetCondition_P60(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P60);
}  

void Ia64TestAndSetCondition_P61(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P61);
}  

void Ia64TestAndSetCondition_P62(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P62);
}  

void Ia64TestAndSetCondition_P63(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  Ia64TestAndSetSingleCondition(edge,state, state_true, state_false, IA64_P63);
}  




void Ia64TestAndSetCondition_Always(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true = state;
  *state_false = NULL;
}

void Ia64TestAndSetCondition_Never(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_false = state;
  *state_true = NULL;
}


TestAndSetConditionFunction Ia64TestAndSetConditionFunctionsArray [] = 
{ 
  &Ia64TestAndSetCondition_P0, &Ia64TestAndSetCondition_P1, &Ia64TestAndSetCondition_P2, &Ia64TestAndSetCondition_P3,
  &Ia64TestAndSetCondition_P4, &Ia64TestAndSetCondition_P5, &Ia64TestAndSetCondition_P6, &Ia64TestAndSetCondition_P7,
  &Ia64TestAndSetCondition_P8, &Ia64TestAndSetCondition_P9, &Ia64TestAndSetCondition_P10, &Ia64TestAndSetCondition_P11,
  &Ia64TestAndSetCondition_P12, &Ia64TestAndSetCondition_P13, &Ia64TestAndSetCondition_P14, &Ia64TestAndSetCondition_P15,
  &Ia64TestAndSetCondition_P16, &Ia64TestAndSetCondition_P17, &Ia64TestAndSetCondition_P18, &Ia64TestAndSetCondition_P19,
  &Ia64TestAndSetCondition_P20, &Ia64TestAndSetCondition_P21, &Ia64TestAndSetCondition_P22, &Ia64TestAndSetCondition_P23,
  &Ia64TestAndSetCondition_P24, &Ia64TestAndSetCondition_P25, &Ia64TestAndSetCondition_P26, &Ia64TestAndSetCondition_P27,
  &Ia64TestAndSetCondition_P28, &Ia64TestAndSetCondition_P29, &Ia64TestAndSetCondition_P30, &Ia64TestAndSetCondition_P31,
  &Ia64TestAndSetCondition_P32, &Ia64TestAndSetCondition_P33, &Ia64TestAndSetCondition_P34, &Ia64TestAndSetCondition_P35,
  &Ia64TestAndSetCondition_P36, &Ia64TestAndSetCondition_P37, &Ia64TestAndSetCondition_P38, &Ia64TestAndSetCondition_P39,
  &Ia64TestAndSetCondition_P40, &Ia64TestAndSetCondition_P41, &Ia64TestAndSetCondition_P42, &Ia64TestAndSetCondition_P43,
  &Ia64TestAndSetCondition_P44, &Ia64TestAndSetCondition_P45, &Ia64TestAndSetCondition_P46, &Ia64TestAndSetCondition_P47,
  &Ia64TestAndSetCondition_P48, &Ia64TestAndSetCondition_P49, &Ia64TestAndSetCondition_P50, &Ia64TestAndSetCondition_P51,
  &Ia64TestAndSetCondition_P52, &Ia64TestAndSetCondition_P53, &Ia64TestAndSetCondition_P54, &Ia64TestAndSetCondition_P55,
  &Ia64TestAndSetCondition_P56, &Ia64TestAndSetCondition_P57, &Ia64TestAndSetCondition_P58, &Ia64TestAndSetCondition_P59,
  &Ia64TestAndSetCondition_P60, &Ia64TestAndSetCondition_P61, &Ia64TestAndSetCondition_P62, &Ia64TestAndSetCondition_P63

};



void Ia64EdgePropagator(t_cfg_edge * edge, t_ins * ins)
{
  t_ia64_reg condition = ins?IA64INSPRED(ins):IA64_P0;
  t_uint16 edgetype = EDGE_CAT(edge);
/*
  printf("Ia64EdgePropagator called... ");
  if(ins != NULL)
    InsPrint(ins, 0);
  else
    printf("Got NULL instruction\n");
*/    
  
  if (IsHellNode(EDGE_HEAD(edge)))
    {
      edge->testcondition = &Ia64TestAndSetConditionAllPaths;
      return;
    }

  switch (edgetype)
    {
    case ET_JUMP:
/*
      if (CFG_EDGE_FLAGS(edge) & EF_FROM_SWITCH_TABLE)
      	{
      	  edge->testcondition = &Ia64TestAndSetConditionAllPaths;
      	  break;
      	}
*/
    case ET_CALL:
    case ET_IPJUMP:
      if(ins && (Ia64IsModOrLoopBranch(ins) || Ia64IsCheck(ins)))
      {
	edge->testcondition = &Ia64TestAndSetConditionAllPaths;
	return;
      }
	
      if(!IA64_IS_PR(condition))
	FATAL(("Ongeldige conditie meegegeven: p%d !!!", condition-IA64_P0));
      edge->testcondition = Ia64TestAndSetConditionFunctionsArray[condition-IA64_P0];
      break;
    case ET_FALLTHROUGH:
    case ET_IPFALLTHRU:
    case ET_UNKNOWN:
    case ET_IPUNKNOWN:
/*
      if(edgetype == ET_FALLTHROUGH && IA64INSTYPE(ins) == IA64_TYPE_DUMMY && IA64INSOPCODE(ins) == DUMMY_LIVE_OPCODE)
      {
	edge->testcondition = &Ia64TestAndSetConditionAllPaths;
	return;
      }
*/      
      edge->testcondition = &Ia64TestAndSetCondition_Always;
      break;
    case ET_SWITCH:
    case ET_RETURN:
    case ET_COMPENSATING:
      edge->testcondition = &Ia64TestAndSetConditionAllPaths;
      break;
    case ET_FUNLINK:
    case ET_IPFUNLINK:
    case ET_SWI:
      edge->testcondition = &Ia64TestAndSetCondition_Never;
      break;
    case ET_SWILINK:
    case ET_IPSWILINK:
      edge->testcondition = &Ia64TestAndSetConditionSetAllToBot;
      break;
    default:
      FATAL(("Ia64SetConditionToProp invalid type %lx\n",edgetype));
      break;
    }

}

t_bool Ia64InstructionUnconditionalizer(t_ins * ins)
{
  if (IA64INSPRED(ins)!=IA64_P0)
  {
    IA64INSPRED(ins)=IA64_P0;
    INS_REGS_USE(ins)=Ia64UsedRegisters(ins); 
    INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins); 
    INS_ATTRIB(ins) &= ~IF_CONDITIONAL;
    
    return TRUE;
  }
  return FALSE;
}






t_bool RegIsEqualAndKnown(t_reg reg, t_bool isFloat, t_procstate * state_a, t_procstate * state_b)
{
  register_content value1;
  register_content value2;

  if (ProcStateGetReg(state_a,reg, &value1))
    return FALSE;

  if (ProcStateGetReg(state_b,reg, &value2))
    return FALSE;

  if(isFloat)
  {
    FATAL(("implement"));
   // return FP_EQ(value1.f, value2.f);
  }
  else
    if (GENERIC_EQ(value1.i,value2.i))
      return TRUE;

  return FALSE;
}

t_bool ConditionsAreEqualAndKnown(t_ia64_reg p1, t_ia64_reg p2, t_procstate * state_a, t_procstate * state_b)
{

  register_content value_a1;

  register_content value_b1;

  if (ProcStateGetReg(state_a, p1 ,&value_a1))
    return FALSE;
  if (ProcStateGetReg(state_b, p1 ,&value_b1))
    return FALSE;
  if (!GENERIC_EQ(value_a1.i,value_b1.i))
    return FALSE;

  if (ProcStateGetReg(state_a, p2 ,&value_a1))
    return FALSE;
  if (ProcStateGetReg(state_b, p2 ,&value_b1))
    return FALSE;
  if (!GENERIC_EQ(value_a1.i,value_b1.i))
    return FALSE;
  
  return TRUE;
}











t_bool Ia64InstructionsConstantOptimizer(t_ins * ins, t_procstate * prev_state, t_procstate * next_state)
{
  t_bool doesnothing;
  t_uint32 i;
  t_ia64_instr_opcode opc = IA64OPCODEINFO(ins).instr_opcode;

  //printf("In functie Ia64InstructionsConstantOptimizer...\n");

  //  FATAL(("Implement"));




  /* ZET TIJDELIJK AF !!!!!!!!!!!!!*/
  
  //return TRUE;





 /*databbls overslaan !?*/



  

  
  if(opc == IA64_INSTR_DUMMY)
    return TRUE; 


  /* TODO: geldt ook voor ia64-compare-instructies -> wel verfijnen: enkel kijken naar de predicaten die effectief zouden kunnen wijzigen...*/
  if (opc == IA64_INSTR_CMP || opc == IA64_INSTR_CMP4 || opc == IA64_INSTR_TBIT || opc == IA64_INSTR_TNAT)
  {
    if (ConditionsAreEqualAndKnown(IA64INSOUT1(ins), IA64INSOUT2(ins), prev_state, next_state))
    {
      InsKill(ins);
      return FALSE; 
    }
  }
  doesnothing = TRUE;
  for (i=0; i<IA64_TOTALREGS; i++)
  {
    if (RegsetIn(INS_REGS_DEF(ins),i)) 
      if (!RegIsEqualAndKnown(i, IA64_IS_FR(i), prev_state, next_state))
      {
	doesnothing = FALSE;
	break;
      }
  }

  if (doesnothing && !Ia64InsHasSideEffect(ins))
  {
    //      printf("Useless instruction found. Killing\n");
    //      ProcStatePrint(stdout,1,prev_state);
    //      InsPrint(ins,0);
    //      ProcStatePrint(stdout,1,next_state);
    InsKill(ins);
    return FALSE;
  }


  if (Ia64InsDefinedRegCount(ins) == 1 && IA64INSOUT1(ins)!=IA64_NONE && !IA64INS_IS_GR_MOV(ins) && !Ia64InsHasSideEffect(ins))
  {
    register_content prod_value;
    t_reloc * prod_reloc;
    t_lattice_level prod_value_level;
    t_lattice_level prod_reloc_level;

    prod_value_level = ProcStateGetReg(next_state,IA64INSOUT1(ins),&prod_value);
    prod_reloc_level = ProcStateGetTag(next_state,IA64INSOUT1(ins),&prod_reloc);


    if (!prod_value_level && (!prod_reloc_level || TOP==prod_reloc_level))
    { /*\TODO: we bekijken enkel de globale registers...*/
      for (i=IA64_GR; i<=IA64_MAX_GR; i++)
      {
	if (i!=IA64INSOUT1(ins))
	{
	  register_content other_value;
	  t_reloc * other_reloc;
	  t_lattice_level other_value_level;
	  t_lattice_level other_reloc_level;

	  t_regset livebefore = InsRegsLiveBefore(ins);

	  if (!(RegsetIn(livebefore, i))) 
	    continue;

	  //  printf("i: %d\n",i);fflush(stdout);

	  other_value_level = ProcStateGetReg(next_state,i,&other_value);
	  other_reloc_level = ProcStateGetTag(next_state,i,&other_reloc);

	  if (!other_value_level && (!other_reloc_level || TOP==other_reloc_level))
	  {
	    if (GENERIC_EQ(other_value.i,prod_value.i))
	    {
	      // printf("values %x %x\n",G_T_UINT64(other_value.i),G_T_UINT64(prod_value.i)); 

	      if ((TOP==other_reloc_level && TOP==prod_reloc_level) || 
		  (!other_reloc_level && !prod_reloc_level && 
		   (prod_reloc==other_reloc || (REL_TO_RELOCATABLE(prod_reloc)==REL_TO_RELOCATABLE(other_reloc) && (strcmp(prod_reloc->code,other_reloc->code)==0)))))
	      {
		IA64INSOPCODE(ins) = GetAddsOpcode();
		IA64INSTYPE(ins) = IA64_TYPE_A;
		/* 
		* opgepast: instructiegroep moet nieuw zijn: misschien wordt i maar in dezelfde instructiegroep geschreven 
		* zal dan bij het layouten evt wel weer gecorrigeerd worden...moest deze onderstelling verkeerd blijken te zijn
		* */
		//het defining predicaat mag blijven zoal het is... 
		IA64INSINSTRGROUP(ins) = GetCurrentInstrGroupAndInc();
		IA64INSIN1(ins) = i;
		IA64INSIN2(ins) = IA64_NONE;
		IA64INSIN3(ins) = IA64_NONE;
		IA64INSIMM1(ins) = 0;
		IA64INSIMM2(ins) = 0;
		IA64INSIMM3(ins) = 0;
		INS_ATTRIB(ins) |= IF_CHANGED;
		InsFreeReferedRelocsDelayed(ins);

		INS_REGS_DEF(ins) = Ia64DefinedRegisters(ins);
		INS_REGS_USE(ins) = Ia64UsedRegisters(ins);

		// RegsetAddReg(INS_REGS_USE(ins), i);

		// InsPrint(ins,0);
	      }
	    }
	  }
	}
      }
    }

    if (!prod_value_level && !prod_reloc_level)
    {
      if ((G_T_UINT64(prod_value.i) & 0xfffffffffffffff0ULL) == G_T_UINT64(prod_value.i))
      {
	for (i=IA64_GR; i<=IA64_MAX_GR; i++)
	{
	  register_content other_value;
	  t_reloc * other_reloc;
	  t_lattice_level other_value_level;
	  t_lattice_level other_reloc_level;

	  t_regset livebefore = InsRegsLiveBefore(ins);

	  if (!(RegsetIn(livebefore, i))) 
	    continue;

	  // printf("i: %d\n",i);fflush(stdout);

	  other_value_level = ProcStateGetReg(prev_state,i,&other_value);
	  other_reloc_level = ProcStateGetTag(prev_state,i,&other_reloc);

	  if (!other_value_level && !other_reloc_level)
	    if ((G_T_UINT64(other_value.i) & 0xfffffffffffffff0ULL) == G_T_UINT64(other_value.i))
	      if ((RELOCATABLE_TYPE(REL_TO_RELOCATABLE(prod_reloc))==RT_BBL && RELOCATABLE_TYPE(REL_TO_RELOCATABLE(other_reloc))==RT_BBL) || 
		  (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(prod_reloc))==RT_SECTION && RELOCATABLE_TYPE(REL_TO_RELOCATABLE(other_reloc))==RT_SECTION))
		if (REL_TO_RELOCATABLE(prod_reloc) == REL_TO_RELOCATABLE(other_reloc))
		{
		  t_int64 diff;

		  if(G_T_UINT64(prod_value.i)>=G_T_UINT64(other_value.i))
		    diff = G_T_UINT64(prod_value.i) - G_T_UINT64(other_value.i);
		  else
		    diff = -(t_int64)(G_T_UINT64(other_value.i) - G_T_UINT64(prod_value.i));

		  /*kijken voor addl imm22 */
		  if(i>=IA64_R0 && i <= IA64_R3 && ABS(diff) < (1<<21)-1)
		  {
		    IA64INSOPCODE(ins) = GetAddlOpcode();
		    IA64INSTYPE(ins) = IA64_TYPE_A;
		    /* 
		    * opgepast: instructiegroep moet nieuw zijn: misschien wordt i maar in dezelfde instructiegroep geschreven 
		    * zal dan bij het layouten evt wel weer gecorrigeerd worden...moest deze onderstelling verkeerd blijken te zijn
		    * */
		    //het defining predicaat mag blijven zoal het is... 
		    IA64INSINSTRGROUP(ins) = GetCurrentInstrGroupAndInc();
		    IA64INSIN1(ins) = i;
		    IA64INSIN2(ins) = IA64_NONE;
		    IA64INSIN3(ins) = IA64_NONE;
		    IA64INSIMM1(ins) = diff;
		    IA64INSIMM2(ins) = 0;
		    IA64INSIMM3(ins) = 0;
		    INS_ATTRIB(ins) |= IF_CHANGED;
		    InsFreeReferedRelocsDelayed(ins);

		    INS_REGS_DEF(ins) = Ia64DefinedRegisters(ins);
		    INS_REGS_USE(ins) = Ia64UsedRegisters(ins);
		    // RegsetAddReg(INS_REGS_USE(ins), i);

		    // InsPrint(ins,0);
		  }
		  else /*kijken voor adds imm14*/ 
		    if (ABS(diff) < (1<<13)-1)
		    {
		      IA64INSOPCODE(ins) = GetAddsOpcode();
		      IA64INSTYPE(ins) = IA64_TYPE_A;
		      /* 
		      * opgepast: instructiegroep moet nieuw zijn: misschien wordt i maar in dezelfde instructiegroep geschreven 
		      * zal dan bij het layouten evt wel weer gecorrigeerd worden...moest deze onderstelling verkeerd blijken te zijn
		      * */
		      //het defining predicaat mag blijven zoal het is... 
		      IA64INSINSTRGROUP(ins) = GetCurrentInstrGroupAndInc();
		      IA64INSIN1(ins) = i;
		      IA64INSIN2(ins) = IA64_NONE;
		      IA64INSIN3(ins) = IA64_NONE;
		      IA64INSIMM1(ins) = diff;
		      IA64INSIMM2(ins) = 0;
		      IA64INSIMM3(ins) = 0;
		      INS_ATTRIB(ins) |= IF_CHANGED;
		      InsFreeReferedRelocsDelayed(ins);

		      INS_REGS_DEF(ins) = Ia64DefinedRegisters(ins);
		      INS_REGS_USE(ins) = Ia64UsedRegisters(ins);

		      // RegsetAddReg(INS_REGS_USE(ins), i);

		      // InsPrint(ins,0);
		    }
		}
	}
      }
    }  

  }


  /*\TODO: nog het algemene (voor gewone registers-zonder relocaties...)*/
  /* sideeffect is nodig om bv geen functiecalls e.d. zomaar te wijzigen in een mov of zo...*/
  if(Ia64InsDefinedRegCount(ins) == 1 && IA64INSOUT1(ins) != IA64_NONE && !Ia64InsHasSideEffect(ins))
  {
    t_ia64_reg regOut= IA64INSOUT1(ins);
    register_content reg;
    t_int64 ivalue;
    t_float fvalue;
    t_lattice_level level;
    t_lattice_level level_reloc;
    t_reloc * reloc;

    level = ProcStateGetReg(next_state,regOut,&reg);
    if (!level)
    {
      if(IA64_IS_FR(regOut))
      {
	FATAL(("implement"));
	/*fvalue = reg.f;*/
      }
      else
	ivalue = (t_int64)G_T_UINT64(reg.i);
    }
    level_reloc = ProcStateGetTag(next_state,regOut,&reloc);

    if (level_reloc==TOP)
    {
      /* eerst onderscheid maken per type register...*/
      /*\TODO evt ook uitbreiden: ook gaan kijken tussen verschillende*/


      if(IA64_IS_GR(regOut))
      {
	/*is het geen mov r=imm22*/
	if(!(IA64INSIN1(ins) == IA64_R0 && IA64INSOPCODE(ins) == GetAddlOpcode()) && (ABS(ivalue) < (1<<21)-1))
	{
	  IA64INSOPCODE(ins) = GetAddlOpcode();
	  IA64INSTYPE(ins) = IA64_TYPE_A;
	  /* 
	  * Hier staat de instructiegroep OK; het is toch een zuivere constante die erin wordt geschreven en
	  * R0 kan toch nooit wijzigen...
	  * */
	  //het defining predicaat mag blijven zoal het is... 
	  IA64INSIN1(ins) = IA64_R0;
	  IA64INSIN2(ins) = IA64_NONE;
	  IA64INSIN3(ins) = IA64_NONE;
	  IA64INSIMM1(ins) = ivalue;
	  IA64INSIMM2(ins) = 0;
	  IA64INSIMM3(ins) = 0;
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  InsFreeReferedRelocsDelayed(ins);

	  INS_REGS_DEF(ins) = Ia64DefinedRegisters(ins);
	  INS_REGS_USE(ins) = Ia64UsedRegisters(ins);

	  // RegsetAddReg(INS_REGS_USE(ins), i);

	  // InsPrint(ins,0);

	}
      }
#if 0      
      else if (IA64_IS_FR(regOut))
      {
	/* je kan niet rechtstreeks een constante in een Floating-pointregister steken...*/
      }


      else if (IA64_IS_BR(regOut))
      {
      }

#endif
    }
  }

  return TRUE;

}
/* vim: set shiftwidth=2: */
