/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_disassemble_one_a.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"



#ifdef _DEBUG_DISASSEMBLE_
#include "../ia64_print/ia64_print_a.h"
static char buffer[256];
#endif


/* A1: Integer Alu-Register-Register */
void Ia64DisassembleA1(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA1(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}

/* A2: Integer Alu-shift left and add */
void Ia64DisassembleA2(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM3_A2_A10(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA2(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* A3: Integer Alu-immediate8 register */
void Ia64DisassembleA3(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM8_A3_A8_I27_M30(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA3(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* A4: Integer Alu-add immediate14*/
void Ia64DisassembleA4(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);
  
  IA64INSIMM1(ins) = IMM14_A4(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA4(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}

/* A5: Integer Alu-add immediate22*/
void Ia64DisassembleA5(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GETBITS64(instr,20,2);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM22_A5(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA5(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}

/* A6: Integer compare-register-register*/
void Ia64DisassembleA6(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_PR_P1(instr);
  IA64INSOUT2(ins) = GET_PR_P2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA6(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* A7: Integer comparetozero-register*/
void Ia64DisassembleA7(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);
    
  IA64INSOUT1(ins) = GET_PR_P1(instr);
  IA64INSOUT2(ins) = GET_PR_P2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA7(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* A8: Integer compare-immediate-register*/
void Ia64DisassembleA8(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_PR_P1(instr);
  IA64INSOUT2(ins) = GET_PR_P2(instr);
  
  IA64INSIMM1(ins) = IMM8_A3_A8_I27_M30(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA8(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* A9: Multimedia ALU*/
void Ia64DisassembleA9(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA9(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* A10: Multimedia shift and add*/
void Ia64DisassembleA10(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM3_A2_A10(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintA10(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* vim: set shiftwidth=2: */

