/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_disassemble_one_f.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"


#ifdef _DEBUG_DISASSEMBLE_
#include "../ia64_print/ia64_print_f.h"
static char buffer[256];
#endif


/* F1: Floating-point multiply add */
void Ia64DisassembleF1(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F3(instr);
  IA64INSIN2(ins) = GET_FR_F4(instr);
  IA64INSIN3(ins) = GET_FR_F2(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF1(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* F2: Fixed-point multiply add */
void Ia64DisassembleF2(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F3(instr);
  IA64INSIN2(ins) = GET_FR_F4(instr);
  IA64INSIN3(ins) = GET_FR_F2(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF2(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* F3: Parallel Floating-point select */
void Ia64DisassembleF3(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F3(instr);
  IA64INSIN2(ins) = GET_FR_F4(instr);
  IA64INSIN3(ins) = GET_FR_F2(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF3(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* F4: Floating-point compare */
void Ia64DisassembleF4(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);
  IA64INSIN2(ins) = GET_FR_F3(instr);

  IA64INSOUT1(ins) = GET_PR_P1(instr);
  IA64INSOUT2(ins) = GET_PR_P2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF4(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* F5: Floating-point class */
void Ia64DisassembleF5(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);

  IA64INSOUT1(ins) = GET_PR_P1(instr);
  IA64INSOUT2(ins) = GET_PR_P2(instr);

  IA64INSIMM1(ins) = IMM9_F5(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF5(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* F6: Floating-point reciprocal approxiimation */
void Ia64DisassembleF6(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);
  IA64INSIN2(ins) = GET_FR_F3(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);
  IA64INSOUT2(ins) = GET_PR_P2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF6(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* F7: Floating-point reciprocal square root approximation */
void Ia64DisassembleF7(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F3(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);
  IA64INSOUT2(ins) = GET_PR_P2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF7(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* F8: minimum/maximum and parallel compare */
void Ia64DisassembleF8(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);
  IA64INSIN2(ins) = GET_FR_F3(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF8(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* F9: merge and logical */
void Ia64DisassembleF9(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);
  IA64INSIN2(ins) = GET_FR_F3(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF9(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* F10: convert floating-point to fixed-point */
void Ia64DisassembleF10(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF10(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* F11: convert fixed-point to floating-point */
void Ia64DisassembleF11(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF11(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* F12: floating-point set controls*/
void Ia64DisassembleF12(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM2(ins) = IMM_IMM7c(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF12(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* F13: floating-point clear flags*/
void Ia64DisassembleF13(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF13(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* F14: floating-point check flags*/
void Ia64DisassembleF14(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = TARGET25_F14(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF14(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* F15: break/nop*/
void Ia64DisassembleF15(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = IMM_IMM21_F15(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintF15(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}



/* vim: set shiftwidth=2: */
