/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_disassemble_one_i.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"


#ifdef _DEBUG_DISASSEMBLE_
#include "../ia64_print/ia64_print_i.h"
static char buffer[256];
#endif



/* I1: Multimedia multiply and shift */
void Ia64DisassembleI1(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = (IMM_IMM2d(instr) == 0)?0:(IMM_IMM2d(instr)==1)?7:(IMM_IMM2d(instr)==2)?15:16;

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI1(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* I2: Multimedia multiply/mix/pack/unpack */
void Ia64DisassembleI2(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI2(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* I3: Multimedia mux1 */
void Ia64DisassembleI3(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM_IMM4c(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI3(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* I4: Multimedia mux2 */
void Ia64DisassembleI4(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM_IMM8c(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI4(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}

/* I5: Multimedia mux2 */
void Ia64DisassembleI5(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);
  IA64INSIN2(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI5(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* I6: Multimedia shift right fixec */
void Ia64DisassembleI6(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM_IMM5b(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI6(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* I7: Multimedia shift left variable */
void Ia64DisassembleI7(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI7(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* I8: Multimedia shift left fixed */
void Ia64DisassembleI8(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = 31 - IMM_IMM5c(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI8(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* I9: Population count */
void Ia64DisassembleI9(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI9(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}

/* I10: shift right pair */
void Ia64DisassembleI10(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM_IMM6d(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI10(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I11: extract */
void Ia64DisassembleI11(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM_IMM6b(instr);
  IA64INSIMM2(ins) = IMM_IMM6d(instr)+1;

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI11(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}

/* I12: zero and deposit */
void Ia64DisassembleI12(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  
  IA64INSOUT1(ins) = GET_GR_R1(instr);


  IA64INSIMM1(ins) = 63 - IMM_IMM6c(instr);
  IA64INSIMM2(ins) = IMM_IMM6d(instr) + 1;

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI12(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I13: zero and deposit immediate*/
void Ia64DisassembleI13(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSOUT1(ins) = GET_GR_R1(instr);
  IA64INSIMM1(ins) = SIGN_EXT((IMM_S_I_BIT(instr)<<7)|IMM_IMM7b(instr),8);
  IA64INSIMM2(ins) = 63 - IMM_IMM6c(instr);
  IA64INSIMM3(ins) = IMM_IMM6d(instr)+1;

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI13(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I14: deposit immediate*/
void Ia64DisassembleI14(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);
  IA64INSOUT1(ins) = GET_GR_R1(instr);
  IA64INSIMM1(ins) = SIGN_EXT(IMM_S_I_BIT(instr),1);
  IA64INSIMM2(ins) = 63 - IMM_IMM6b(instr);
  IA64INSIMM3(ins) = IMM_IMM6d(instr)+1;

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI14(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I15: deposit*/
void Ia64DisassembleI15(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = 63 - IMM_IMM6e(instr);
  IA64INSIMM2(ins) = IMM_IMM4d(instr)+1;

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI15(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I16: test bit*/
void Ia64DisassembleI16(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_PR_P1(instr);
  IA64INSOUT2(ins) = GET_PR_P2(instr);

  IA64INSIMM1(ins) = IMM_IMM6b(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI16(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}

/* I17: test naT*/
void Ia64DisassembleI17(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_PR_P1(instr);
  IA64INSOUT2(ins) = GET_PR_P2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI17(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I19: break\nop*/
void Ia64DisassembleI19(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = IMM21_I19_M37(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI19(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}

/* I20: integer speculation unit*/
void Ia64DisassembleI20(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSIMM1(ins) = TARGET25_I20_M20_M21(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI20(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I21: move to BR*/
void Ia64DisassembleI21(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_BR_B1(instr);

  IA64INSIMM1(ins) = TAG13_I21(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI21(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I22: move from BR*/
void Ia64DisassembleI22(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_BR_B2(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI22(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I23: move to predicates -register*/
void Ia64DisassembleI23(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSIMM1(ins) = MASK17_I23(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI23(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}

/* I24: move to predicates - immediate44*/
void Ia64DisassembleI24(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = IMM44_I24(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI24(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I25: move from predicates ip*/
void Ia64DisassembleI25(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI25(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}

/* I26: move to AR - register*/
void Ia64DisassembleI26(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_AR_R3(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI26(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}

/* I27: move to AR - immediate8*/
void Ia64DisassembleI27(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSOUT1(ins) = GET_AR_R3(instr);

  IA64INSIMM1(ins) = IMM8_A3_A8_I27_M30(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI27(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}


/* I28: move from AR - immediate8*/
void Ia64DisassembleI28(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_AR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI28(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}

/* I29: sign/zero extend/compute zero index */
void Ia64DisassembleI29(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintI29(ins, buffer);
	printf(buffer);
	printf("\n");
#endif
}









/* vim: set shiftwidth=2: */
