/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_disassemble_one_m.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"


#ifdef _DEBUG_DISASSEMBLE_
#include "../ia64_print/ia64_print_m.h"
static char buffer[256];
#endif


/* M1: Integer Load */
void Ia64DisassembleM1(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM1(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* M2: Integer Load - Increment by Register */
void Ia64DisassembleM2(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);
  IA64INSIN2(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM2(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M3: Integer Load - Increment by immediate */
void Ia64DisassembleM3(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM9_M3(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM3(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}



/* M4: Integer Store */
void Ia64DisassembleM4(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSIN2(ins) = GET_GR_R3(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM4(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M5: Integer Store - Increment by immediate */
void Ia64DisassembleM5(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);
  
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSIMM1(ins) = IMM9_M5(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM5(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}

/* M6: Floating-point Load */
void Ia64DisassembleM6(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM6(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M7: Floating-point Load - Increment by Register */
void Ia64DisassembleM7(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);
  IA64INSIN2(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM7(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M8: Floating-point Load - Increment by immediate */
void Ia64DisassembleM8(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

  IA64INSIMM1(ins) = IMM9_M8(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM8(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}



/* M9: Floating-point Store */
void Ia64DisassembleM9(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);

  IA64INSIN2(ins) = GET_GR_R3(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM9(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* M10: Floating-point Store - Increment by immediate */
void Ia64DisassembleM10(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);

  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSIMM1(ins) = IMM9_M10(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM10(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}




/* M11: Floating-point Load Pair */
void Ia64DisassembleM11(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);
  IA64INSOUT2(ins) = GET_FR_F2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM11(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M12: Floating-point Load Pair - Increment by Immediate*/
void Ia64DisassembleM12(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);
  IA64INSOUT2(ins) = GET_FR_F2(instr);

  if((GETBITS64(instr, 30, 6) & 1) == 0)
  	IA64INSIMM1(ins) = 8; //als x6 even is
  else
  	IA64INSIMM1(ins) = 16;//als x6 oneven is
  
#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM12(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}



/* M13: Line Prefetch */
void Ia64DisassembleM13(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM13(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M14: Line Prefetch - Increment by Register */
void Ia64DisassembleM14(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);
  IA64INSIN2(ins) = GET_GR_R2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM14(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M15: Line Prefetch - Increment by Immediate */
void Ia64DisassembleM15(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSIMM1(ins) = IMM9_M15(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM15(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M16: Exchange/Compare and Exchange */
void Ia64DisassembleM16(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);
  IA64INSIN2(ins) = GET_GR_R2(instr);
  if(GETBITS64(instr, 30, 6)<0x08)
    IA64INSIN3(ins) = IA64_AR_CCV;
  else
    IA64INSIN3(ins) = IA64_NONE;

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM16(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M17: Fetch and Add - Immediate */
void Ia64DisassembleM17(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = INC3_M17(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM17(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M18: Set FR */
void Ia64DisassembleM18(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM18(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M19: Get FR */
void Ia64DisassembleM19(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM19(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M20: Integer Speculation Check (M-Unit) */
void Ia64DisassembleM20(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSIMM1(ins) = TARGET25_M20(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM20(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M21: Floating-point Speculation Check */
void Ia64DisassembleM21(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F2(instr);

  IA64INSIMM1(ins) = TARGET25_M21(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM21(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M22: Integer Advanced Load Check */
void Ia64DisassembleM22(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = TARGET25_M22(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM22(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M23: Floating-point Advanced Load Check */
void Ia64DisassembleM23(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F1(instr);

  IA64INSIMM1(ins) = TARGET25_M23(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM23(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M24: Sync/Fence/Serialize/ALAT Control */
void Ia64DisassembleM24(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM24(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* M25: RSE Control */
void Ia64DisassembleM25(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM25(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M26: Integer ALAT Entry Invalidate */
void Ia64DisassembleM26(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM26(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M27: Floating-point ALAT Entry Invalidate */
void Ia64DisassembleM27(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_FR_F1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM27(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M28: Flush Cache/Purge Translation Cache Entry */
void Ia64DisassembleM28(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM28(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M29: Move to AR - Register (M-Unit) */
void Ia64DisassembleM29(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_AR_R3(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM29(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M30: Move to AR - Immediate8 (M-Unit) */
void Ia64DisassembleM30(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSOUT1(ins) = GET_AR_R3(instr);

  IA64INSIMM1(ins) = IMM8_M30(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM30(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M31: Move from AR (M-Unit) */
void Ia64DisassembleM31(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_AR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM31(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M32: Move to CR */
void Ia64DisassembleM32(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_CR_R3(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM32(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M33: Move from CR */
void Ia64DisassembleM33(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_CR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM33(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M34: Allocate Register Stack Frame */
void Ia64DisassembleM34(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  t_uint8 sof, sol, sor;
	
  /* OPGELET ! hier worden SOL, SOF en SOR<<3 bijgehouden !!! */

  sof = GETBITS64(instr, 13, 7);
  sol = GETBITS64(instr, 20, 7);
  sor = GETBITS64(instr, 27, 4) << 3;
  
  IA64INSOUT1(ins) = GET_GR_R1(instr);
  
  IA64INSIMM1(ins) = sol; 		
  IA64INSIMM2(ins) = sof; 		
  IA64INSIMM3(ins) = sor; 		

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM34(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* M35: Move to PSR */
void Ia64DisassembleM35(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

  if(GETBITS64(instr, 27, 6) == 0x2D)
    IA64INSOUT1(ins) = IA64_PSR_L;
  else
    IA64INSOUT1(ins) = IA64_PSR_UM;

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM35(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M36: Move from PSR */
void Ia64DisassembleM36(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  if(GETBITS64(instr, 27, 6) == 0x25)
    IA64INSIN1(ins) = IA64_PSR;
  else
    IA64INSIN1(ins) = IA64_PSR_UM;

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM36(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M37: Break/Nop (M-Unit) */
void Ia64DisassembleM37(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = IMM21_M37(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM37(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* M38: Probe - Register */
void Ia64DisassembleM38(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);
  IA64INSIN2(ins) = GET_GR_R2(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM38(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M39: Probe - Immediate2 */
void Ia64DisassembleM39(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = GETBITS64(instr, 13, 2);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM39(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M40: Probe Fault - Immediate2 */
void Ia64DisassembleM40(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSIMM1(ins) = GETBITS64(instr, 13, 2);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM40(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M41: Translation Cache Insert */
void Ia64DisassembleM41(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM41(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M42: Move to Indirect Register/Translation Register Insert */
void Ia64DisassembleM42(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  /* Opgelet/Merk op: de indexregister r3 wordt ook als inputregister beschouwd !!! */
  IA64INSIN1(ins) = GET_GR_R3(instr);
  IA64INSIN2(ins) = GET_GR_R2(instr);

  switch(GETBITS64(instr, 27, 6))
  {
    case 0x00:
      IA64INSOUT1(ins) = IA64_RR;
      break;
    case 0x01:
      IA64INSOUT1(ins) = IA64_DBR;
      break;
    case 0x02:
      IA64INSOUT1(ins) = IA64_IBR;
      break;
    case 0x03:
      IA64INSOUT1(ins) = IA64_PKR;
      break;
    case 0x04:
      IA64INSOUT1(ins) = IA64_PMC;
      break;
    case 0x05:
      IA64INSOUT1(ins) = IA64_PMD;
      break;
    case 0x0E:
      IA64INSOUT1(ins) = IA64_DTR;
      break;
    case 0x0F:
      IA64INSOUT1(ins) = IA64_ITR;
      break;
  } 

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM42(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M43: Move from Indirect Register */
void Ia64DisassembleM43(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  switch(GETBITS64(instr, 27, 6))
  {
    case 0x10:
      IA64INSIN1(ins) = IA64_RR;
      break;
    case 0x11:
      IA64INSIN1(ins) = IA64_DBR;
      break;
    case 0x12:
      IA64INSIN1(ins) = IA64_IBR;
      break;
    case 0x13:
      IA64INSIN1(ins) = IA64_PKR;
      break;
    case 0x14:
      IA64INSIN1(ins) = IA64_PMC;
      break;
    case 0x15:
      IA64INSIN1(ins) = IA64_PMD;
      break;
    case 0x17:
      IA64INSIN1(ins) = IA64_CPUID;
      break;
  }
  IA64INSIN2(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM43(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M44: Set/Rest User/System Mask */
void Ia64DisassembleM44(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = IMM24_M44(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM44(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M45: Translation Purge */
void Ia64DisassembleM45(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);
  IA64INSIN2(ins) = GET_GR_R2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM45(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* M46: Translation Access */
void Ia64DisassembleM46(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_GR_R3(instr);

  IA64INSOUT1(ins) = GET_GR_R1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintM46(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}







/* vim: set shiftwidth=2: */
