/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "ia64_flowgraph.h"
#include "ia64_instructions.h"
#include "ia64_opcodes.h"
#include "ia64_registers.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "ia64_utils.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_iterators.h"

#include "../../diablo_options.h"

//#ifdef _DEBUG_FLOWGRAPH_
#include "ia64_print.h"
//#endif

t_uint32 Ia64RemoveImm21BRelocs(t_object * obj)
{
  t_reloc * rel, *tmp;
  t_uint32 ret=0;
  OBJECT_FOREACH_RELOC_SAFE(obj,rel,tmp)
    if (REL_FROM(rel)->type == RT_INS && Ia64IsIPRelativeBranchOrCall(T_INS(REL_FROM(rel))))
      {
	RelocTableRemoveReloc(obj->reloc_table,rel);
	ret++;
      }
  return ret;
}

t_uint32 Ia64AddBasicBlockEdges(t_object * obj, t_section * section);
t_uint32 Ia64FindBBLLeaders(t_object * obj, t_section * code);

void Ia64FlowgraphSection(t_object * obj, t_section * section) 
{
  /*t_cfg_edge * i_edge, *i_safe;*/
  /*t_function * function;*/
  
  t_uint32 ret;

  /* Find the leaders in the instruction list */
  ANALYSIS_START("Leader detection");
  ret=Ia64FindBBLLeaders(obj,section);
  ANALYSIS_END("Leader detection",("%d leaders",ret));

  /* Create the basic blocks (platform independent) */
  ANALYSIS_START("Creating Basic Blocks");
  ret=CfgCreateBasicBlocks(section);
  ANALYSIS_END("Creating Basic Blocks",("%d blocks",ret));
/*
  {
    t_cfg * cfg=SECTION_CFG(section);
    t_bbl * i_bbl;

    CFG_FOREACH_BBL(cfg,i_bbl)
    {
      t_ins *tmp_ins = i_bbl->ins_first;

      while(tmp_ins)
      {
	InsPrint(tmp_ins, 0);
	tmp_ins = INS_NEXT(tmp_ins);
      }
      printf("\n");
    }

  }
*/
  
  /* Create the edges between basic blocks */
  ANALYSIS_START("Creating Basic Block graph");
  ret=Ia64AddBasicBlockEdges(obj,section);
  ANALYSIS_END("Creating Basic Block graph",("%d edges added",ret));

}

t_uint32 Ia64AddBasicBlockEdges(t_object * obj, t_section * section) 
{


  /* Basic Block Edges 
  * -----------------
  *
  * Add edges between the newly created basic blocks. Both the lineair (array)
  * representation of the instructions and the basic block representation are
  * available: you can access the (copy) basic block by using INS_BBL of a
  * leader in the lineair representation.
  *
  * Although connecting the blocks isn't that hard, it could be difficult to
  * find out if a given jump instructions is a call or just a jump (without
  * return). Getting this completly safe is probably not what you want: if you
  * implement this completly safe, you'll probably end up with one big
  * function, which will degrade the performance of the rest of diablo or hell
  * nodes after each jump which will degrade the performance even further.
  *
  * On the other hand, if you get this wrong (e.g add a jump edge where there
  * should have been a call edge), then chances are that your flowgraph (and
  * thus the produced binary) is wrong (the successor block of the call block
  * will have no incoming (return) edge, and thus this block is dead). 
  *
  * The solution is to look at code that is produced by your compiler(s).
  * It'll probably have only a few possibilties to produce calls. If you
  * handle them as safe as possible, than you're off the hook.  */

  t_ins *ins/*, * iter, *stop*/, * block_end, * block_start, * jump_target;
  t_bbl * block, * ft_block, *link_block;
  /*t_bbl * bbbl;*/
  /*t_cfg_edge * edge, * j_edge;*/
  t_cfg * flowgraph=SECTION_CFG(section);
  /*t_bool flag;*/
  t_reloc_ref * ref;
  t_uint32 nedges=0;
  t_uint8 slot = 0, currslot;
  t_symbol *symptr;
  t_object * tmp2, * tmp;

  Ia64RemoveImm21BRelocs(obj);
  
//#if 0
  /*seeking for special $diablo-symbols*/
  for (tmp2=obj->mapped_first;tmp2!=NULL; tmp2=tmp2->next)
  { 
    for (tmp=tmp2; tmp!=NULL; tmp=tmp->equal)
    {     
      t_address generic;
      t_uint32 tel;
      
      for (symptr=tmp->symbol_table->first; symptr!=NULL; symptr=symptr->next)
      {
	if(!symptr->name)
	  continue;
	
	if (strncmp(symptr->name,"$diablo:",8)==0)
	{
//#if 0
	  int begin = strstr(symptr->name,"@-@")-symptr->name;
	  char save = symptr->name[begin];
	  t_symbol *s1, *s2;
	  t_address a1, a2;
#ifdef IA64_DEBUG_RELOCS
	  printf("symbol: %s @ 0x%llx\n", symptr->name, symptr->base->address);
#endif	  
	  symptr->name[begin] = '\0';
	  

	  /*TODO ook checken als het symbool van zelfde .o-file komt*/
	  s1=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+8);
	  s2=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+begin+3);

	  a1 = GENERIC_ADD(SECTION_GET_CADDRESS(symptr->base), s1->current_address);
	  a2 = GENERIC_ADD(SECTION_GET_CADDRESS(symptr->base), s2->current_address);

/*
	  printf("%s @ 0x%llx;  %s @ 0x%llx\n", s1->name, a1, s2->name, a2);
          fflush(stdout);
*/	  
	  /*char herstellen*/
	  symptr->name[begin] = save;
	  
	  /*controleren of wel bestaat !*/
	  if(!s1|| !s2)
	    FATAL(("Invalid special diablo symbol !!!"));


	  if(!(INS_ATTRIB(SecGetInsByAddress(section, a1)) & IF_BBL_LEADER))
	    FATAL(("First instruction of patched function has NOT been marked as BBL_LEADER !!!"));
	  CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, a1)),NULL,NULL);        


	  if(!(INS_ATTRIB(SecGetInsByAddress(section, a2)) & IF_BBL_LEADER))
	    FATAL(("First instruction of patched function has NOT been marked as BBL_LEADER !!!"));
	  CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, a2)),NULL,NULL);        
//#endif
	}
      }

    }
  }

//#endif
  
#if 0
  //PATCH
  //waarsch moeten we ook nog leaders maken => moet eig vooraf of zo gebeuren...
  FORALL_SYMBOLS(obj,symptr)
  {
    if(!symptr->name)
      continue;
    if (strncmp(symptr->name,"$diablo:",8)==0)
    {
      int begin = strstr(symptr->name,"@-@")-symptr->name;
      char save = symptr->name[begin];
      t_symbol *s1, *s2;

      symptr->name[begin] = '\0';

      /*TODO ook checken als het symbool van zelfde .o-file komt*/
      s1=SymbolTableGetSymbolByName(obj->symbol_table, symptr->name+8);
      s2=SymbolTableGetSymbolByName(obj->symbol_table, symptr->name+begin+3);

      printf("%d %s %s\n", begin, symptr->name+8, symptr->name+begin+3);

      /*controleren of wel bestaat !*/
      if(!s1|| !s2)
	FATAL(("Invalid special diablo symbol !!!"));

      /*char herstellen*/
      symptr->name[begin] = save;


      if(!(INS_ATTRIB(SecGetInsByAddress(section, s1->current_address)) & IF_BBL_LEADER))
	FATAL(("First instruction of patched function has NOT been marked as BBL_LEADER !!!"));
      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, s1->current_address)),NULL,NULL);	

      if(!(INS_ATTRIB(SecGetInsByAddress(section, s2->current_address)) & IF_BBL_LEADER))
	FATAL(("First instruction of patched function has NOT been marked as BBL_LEADER !!!"));
      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, s2->current_address)),NULL,NULL);	
    }
  }
#endif  
#if 0
  /* First a quick-hack patch in order to get some arrows form the hellnode to the blocks...*/
  {
    t_section * opd=SectionGetFromObjectByName(obj,".opd");
    if (opd)
    {
      t_uint64 *data = (t_uint64*)opd->data;
      t_uint32 cnt=0;
      t_uint32 opd_size = GENERIC_EXTRACT_UINT32(SECTION_SIZE(opd));
      t_uint32 currsize =0;
      t_uint64 max_address = G_T_UINT64(GENERIC_ADD(SECTION_GET_CADDRESS(section), SECTION_SIZE(section)));

      printf("max_text_address: %llx\n", max_address);

      while(currsize < opd_size)
      {
	if(*data != 0 && *data < max_address)
	{
	  t_address addr;
	  GenericSet_uint64(addr,*data);
	  if(!(INS_ATTRIB(SecGetInsByAddress(section, addr)) & IF_BBL_LEADER))
	    FATAL(("First instruction of patched function has NOT been marked as BBL_LEADER !!!"));
	  //          printf("INS_BBL: %llx\n",INS_BBL(SecGetInsByAddress(section, addr)));
	  CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, addr)),NULL,NULL);
	  //	  InsPrint(SecGetInsByAddress(section, addr), 0);
	  printf("Patched edge to function pointer address: %llx\n", *data);
	}

	currsize +=16;
	data +=2;
	cnt++;
      }

      printf("Number of patched edges: %d\n", cnt);
      //    exit(0);
    }
    else
      FATAL(("Section .opd NOT found !!!"));   
  }
#endif

  /* now a patch to add arrows*/
#if 0  
  {
    t_bbl *bbl;

    CFG_FOREACH_BBL(flowgraph, bbl)
    {
      for(ref=RELOCATABLE_REFED_BY(bbl); ref != NULL; ref=ref->next)
      {
	t_reloc *reloc = ref->rel;

	FATAL(("Implement"));
	/*check for PCREL64I-type*/
	if(reloc->write_type == INSTRUCTION_IMMEDIATE64 && reloc->type == REPRESENT_SELFREL_VALUE)
	{
	  t_address target_addr = GENERIC_ADD_UINT32(BBL_BASE_ADDRESS(T_BBL(reloc->to->base)), 16);
	  if(reloc->offset != 2) /* als deze fout zou optreden, betekent dit volgens mij dat de mov_from_ip niet in de bundel direct na de movl zit => dan nog opvangen !!!*/
	    FATAL(("relocation type PCREL64I expected value 2 as offset !"));
#ifdef IA64_DEBUG_RELOCS
	  printf("Target address found: 0x%llx\n", target_addr);
#endif	  
	  if(!(INS_ATTRIB(SecGetInsByAddress(section, target_addr)) & IF_BBL_LEADER))
	    FATAL(("First instruction of PCREL64I-target-entry has NOT been marked as BBL_LEADER !!!"));
	  CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, target_addr)),NULL,NULL);
	}

#if 0

	/*check for PCREL64LSB-type*/
	if(reloc->write_type == ABS64 && reloc->type == REPRESENT_SELFREL_VALUE)
	{
	  t_address switch_addr = GENERIC_ADD(BBL_BASE_ADDRESS(T_BBL(reloc->to->base)), reloc->addend);
	  printf("Switch address found: 0x%llx\n", switch_addr);
	  if(!(INS_ATTRIB(SecGetInsByAddress(section, switch_addr)) & IF_BBL_LEADER))
	    FATAL(("First instruction of switch-entry has NOT been marked as BBL_LEADER !!!"));
	  CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, switch_addr)),NULL,NULL);
	}

	/*check for DIR64LSB-type*/
	if(reloc->write_type == ABS64 && reloc->type == REPRESENT_VALUE)
	{
	  t_address dir_addr = GENERIC_ADD(BBL_BASE_ADDRESS(T_BBL(reloc->to->base)), reloc->addend);
	  printf("Direct address found: 0x%llx\n", dir_addr);
	  if(!(INS_ATTRIB(SecGetInsByAddress(section, dir_addr)) & IF_BBL_LEADER))
	    FATAL(("First instruction of direct-entry has NOT been marked as BBL_LEADER !!!"));
	  CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, dir_addr)),NULL,NULL);
	}
#endif
      }
    }

  }
#endif


  /* walk through the original(!) array (we need offsets) and add 
  * all possible successors to every node */

  for(ins=section->data; ins!=NULL; ins=INS_NEXT(ins), slot=(slot+1)%3)
  {
    /* we're only interested in the first and last instructions of a basic block */

#if 0
    if(G_T_UINT64(INS_CADDRESS(ins))==0x4000000000000340ULL) 
      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, INS_CADDRESS(ins))),NULL,NULL);

    if(G_T_UINT64(INS_CADDRESS(ins))==0x40000000000a7ec0ULL) 
      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, INS_CADDRESS(ins))),NULL,NULL);

    if(G_T_UINT64(INS_CADDRESS(ins))==0x40000000000a7be0ULL) 
      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, INS_CADDRESS(ins))),NULL,NULL);
    if(G_T_UINT64(INS_CADDRESS(ins))==0x40000000000aad40ULL) 
      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, INS_CADDRESS(ins))),NULL,NULL);
    if(G_T_UINT64(INS_CADDRESS(ins))==0x40000000000a9d60ULL) 
      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, INS_CADDRESS(ins))),NULL,NULL);
#endif    

    if (!(INS_ATTRIB(ins) & IF_BBL_LEADER)) 
      continue;
#if 0
    if(nedges >= global_options.debugcounter)
    {
      printf("gestopt na %d edges\n", nedges);

      return nedges;
    }
#endif    

    block_start = ins;
    block=INS_BBL(block_start);

    currslot = (slot + block->nins - 1)%3;

    block_end = T_INS(&(T_IA64_INS(ins)[block->nins - 1])); /* Warning: This is not the same as the end of the bbl, since the end is a copy of the original end */
    ft_block = INS_NEXT(block_end)?INS_BBL(INS_NEXT(block_end)):NULL;  /* the FallThrough basic block */


    /*    
	  if ((INS_ATTRIB(ins) & IF_ADDRESS_TAKEN))
	  {

    //	  DiabloPrint(stdout,"TAKEN @B\n",block);
    CfgEdgeNewCall(flowgraph, flowgraph->hell_node, block,NULL,NULL);
    VERBOSE(1,("Adding INS arrow from hell! (Address taken)"));
    }
     */


    /* 1. ADD EDGES FOR RELOCATIONS: 
    * 
    * each relocation gets it's own edge. These edges will be automatically
    * deleted when the relocation is removed. There's no need to add an edge
    * for the obvious relocations (such as the relocations that are present
    * for the calculation of the offsets in intermodular jumps), since they
    * are completly described by the edge itself (the reason for the edge is
    * that there is a jump, not that there is a relocation) */
    //#if 0
    for (ref=RELOCATABLE_REFED_BY(block); ref!=NULL; ref=ref->next)
    {
      //       if ((ref->rel->write_type!=LAST_BITS_24) && (INS_TYPE(block_start)!=IT_DATA))
      t_bool hellNodeToBeAdded = TRUE;
//#if 0

/*
      if(REL_FROM(ref->rel)->type == RT_INS && Ia64IsIPRelativeBranchOrCall(T_INS(REL_FROM(ref->rel))))
      {
	//printf("skipping INSTRUCTION_IMMEDIATE21B\n");
	hellNodeToBeAdded = FALSE;
      }
*/
      
      if(hellNodeToBeAdded)
      {
	/* Some extra checks if the relocation is from an instruction */
//	if (REL_FROM(ref->rel)->type == RT_INS)
	{
	  t_cfg_edge * found=NULL;
	  t_cfg_edge * i_edge;
	  //if ((INS_REFERS_TO(ref->rel->from)->rel)!=ref->rel) FATAL(("Relocation wrong @I @R",ref->rel->from,ref->rel));


	  BBL_FOREACH_PRED_EDGE(block,i_edge)
	  {
	    if ((CFG_EDGE_HEAD(i_edge)==flowgraph->hell_node) && (EDGE_CAT(i_edge) == ET_CALL))
	    {
	      found=i_edge;
	      break;
	    }
	  }

	  if (found)
	  {
	    ref->rel->edge=found;
	    found->refcount++;
	    if (found->corr_edge) found->corr_edge->refcount++;
	  }
	  else
	  {
	    ref->rel->edge=CfgEdgeNewCall(flowgraph, flowgraph->hell_node, block,NULL,NULL);
	    nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	    printf("edge(reloc) %d\n", nedges);
#endif	    
	  }
	  /* Some comments on this */
	  VERBOSE(1,("Adding [perhaps INS] arrow from hell! (relocation @R, @B)",ref->rel,block));
	}
      }
//#endif
    }
    //#endif    


    if(IA64INSTYPE(ins) == IA64_TYPE_DATA)
    {
      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, block,NULL,NULL);
      nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
      printf("edge(data) %d\n", nedges);
#endif      
      continue;
    }



    /* 2. ADD NORMAL EDGES */
    /* now determine the successors of the block and their types */
    if(Ia64IsBranch(block_end))
    {
      /* If the branch is conditional, add a fall-through edge */
      if (ft_block != NULL && 
	  IA64INSPRED(block_end) != IA64_P0)
      {
	CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
	//	InsPrint(ft_block->ins_first, 0);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(br-ft) %d\n", nedges);
#endif      
      }
      
      if (ft_block != NULL && Ia64IsModOrLoopBranch(block_end))
      {
	//InsPrint(block_end, 0);
	CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(br-mod-ft) %d\n", nedges);
#endif      
	//	InsPrint(ft_block->ins_first, 0);
      }

      if(Ia64IsIPRelativeBranch(block_end))
      {
	/* get the instruction we're jumping to */
	jump_target = (t_ins *) (T_IA64_INS(block_end) - currslot + (IA64INSIMM1(block_end)>>4)*3);
	CfgEdgeNew(flowgraph, block, INS_BBL(jump_target),ET_JUMP);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(br-ip) %d\n", nedges);
#endif      
#ifdef _DEBUG_FLOWGRAPH_
	{
	  static char buffer[256]; 
	  Ia64PrintFunction_table[IA64OPCODEINFO(jump_target).instr_format](jump_target, buffer);
	  printf("currslot: %d %s\n", currslot, buffer);
	  //      exit(0);
	}
#endif
      }
      else
      {
	CfgEdgeNew(flowgraph, block, flowgraph->hell_node,ET_JUMP);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(br-jmp) %d\n", nedges);
#endif      
      }

      continue;
    }

    if(Ia64IsCheck(block_end))
    {
      CfgEdgeNew(flowgraph, block, ft_block, ET_FALLTHROUGH);
      nedges ++;
#ifdef IA64_DEBUG_FLOWGRAPH
      printf("edge(chk-ft) %d\n", nedges);
#endif      

      jump_target = (t_ins *) (T_IA64_INS(block_end) - currslot + (IA64INSIMM1(block_end)>>4)*3);
      CfgEdgeNew(flowgraph, block, INS_BBL(jump_target),ET_JUMP);
      nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
      printf("edge(chk-jmp) %d\n", nedges);
#endif      

      continue;
    }
//#if 0
    if(Ia64IsCall(block_end))
    {
      t_ia64_ins * tmp;

      /* insert new bbl, for dummy: liveness and constprop */
      t_bbl * dummy_bbl = BblNew(flowgraph);

      BBL_CADDRESS(dummy_bbl) = BBL_CADDRESS(ft_block);
      
      /* 
      * verwijzen naar laatste instructie van deze bundle (-1)
      * hebben we nodig om te bepalen of er wel nog een volgende bundle bestaat... 
       */      
      tmp = T_IA64_INS(block_end) - currslot + 3 - 1;
      /* linkblok op hellnode zetten als er blijkbaar niets meer op volgt... */
      link_block = INS_NEXT(tmp)?INS_BBL(INS_NEXT(tmp)):flowgraph->hell_node;

      if(link_block != ft_block)
      {
       if(IA64OPCODEINFO(BBL_INS_FIRST(ft_block)).instr_opcode != IA64_INSTR_NOP)
       {
	 DiabloPrint(stdout, "curr:@I\n link:@I\n ft:@I\n", block_end, link_block, ft_block);
	 FATAL(("Linkblock different from fallthrough in CALL !!!"));
       }
       else
	 ft_block = link_block; /*skip this nop-instruction in slot 2 or so...*/
      }

      /* If the branch is conditional, add a fall-through edge */
      if (ft_block != NULL && IA64INSPRED(block_end) != IA64_P0)
      {
	CfgEdgeNew(flowgraph, block,dummy_bbl, ET_FALLTHROUGH);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(call-ft-dummy) %d\n", nedges);
#endif      
	//CfgEdgeNew(flowgraph, dummy_bbl, ft_block, ET_FALLTHROUGH);
	//nedges++;
      }

      
      /* make some place for the dummy-bbl ! */
      CfgEdgeNew(flowgraph, dummy_bbl, link_block, ET_FALLTHROUGH);
      nedges ++;
#ifdef IA64_DEBUG_FLOWGRAPH
      printf("edge(call-ft-link) %d\n", nedges);
#endif      
//      CfgEdgeNew(flowgraph, dummy_bbl, link_block, ET_FUNLINK);
//      nedges ++;
      
      
#ifdef _DEBUG_FLOWGRAPH_
      {
	static char buffer[256]; 
	Ia64PrintFunction_table[IA64OPCODEINFO(INS_NEXT(tmp)).instr_format](INS_NEXT(tmp), buffer);
	printf("%s\n", buffer);
      }
#endif

      if(Ia64IsIPRelativeCall(block_end))
      {
	/* get the instruction we're jumping to */
	jump_target = (t_ins *) (T_IA64_INS(block_end) - currslot + (IA64INSIMM1(block_end)>>4)*3);
	CfgEdgeNewCall(flowgraph, block, INS_BBL(jump_target),dummy_bbl,NULL);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(call-jmp) %d\n", nedges);
#endif      
      }
      else
      {
	CfgEdgeNewCall(flowgraph, block, flowgraph->hell_node,dummy_bbl,NULL);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(call-ind) %d\n", nedges);
#endif      
      }

      /* also add an edge of type ET_FUNLINK to link call site to return site */
      CfgEdgeNew(flowgraph, block, dummy_bbl, ET_FUNLINK);
      nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
      printf("edge(call-fun) %d\n", nedges);
#endif      

      continue;
    }
//#endif
    if(Ia64IsReturn(block_end))
    {
      /* If the return is conditional, add a fall-through edge */
      if (ft_block != NULL && IA64INSPRED(block_end) != IA64_P0)
      { 
	CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(ret-ft) %d\n", nedges);
#endif      
      } 

      CfgEdgeNew(flowgraph, block, flowgraph->exit_hell_node,ET_RETURN);
      nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
      printf("edge(ret) %d\n", nedges);
#endif      

      continue;
    }

    /* SWI's go to SWI-hell */
    if(Ia64IsBreak(block_end))
    {
      
      /* If the SWI is conditional, we add a fall-throug edge */
      if (ft_block != NULL && IA64INSPRED(block_end) != IA64_P0)
      {
	CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(break-ft) %d\n", nedges);
#endif      
      }

      if (ft_block != NULL)
      {
	/* link call and return site of the SWI */
	CfgEdgeNew(flowgraph, block,ft_block, ET_SWILINK);
	nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
	printf("edge(break-link) %d\n", nedges);
#endif      

      }
      else
	FATAL(("Could not place SWI link arrow because the end of the section was reached!"));


      /* since we don't know where a SWI jumps to, we'll add hell as a successor */
      CfgEdgeNew(flowgraph, block,flowgraph->swi_hell_node,ET_SWI);
      nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
      printf("edge(break) %d\n", nedges);
#endif      

      continue;
    }

    /* instructions that end up here do not alter the control flow, they just happen to be */
    /* the end of a basic block */

    if (INS_NEXT(block_end) != NULL)
    {
      CfgEdgeNew(flowgraph,  block, ft_block, ET_FALLTHROUGH);
      nedges++;
#ifdef IA64_DEBUG_FLOWGRAPH
      printf("edge(ft) %d\n", nedges);
#endif      
    }

  }


  return nedges;
}





t_uint32 Ia64FindBBLLeaders(t_object * obj, t_section * code) 
{

  /* Basic leader-detection algorithm. 
  * ---------------------------------
  *
  * See e.g. the Red Dragon Book ("Compilers: Principles, Techniques and
  * Tools", by Alfred V. Aho, Ravi Sethi, and Jeffrey D. Ullman) for more
  * details. 
  *
  * The aim of this function is to identify all start-addresses of basic
  * blocks (and mark them). There are four reasons to mark an instructions as
  * a basic block leader:
  * 
  * 1. The instruction is the target of a direct or indirect jump. This is not
  * as easy as it seems, especially when the program counter is explicit:
  * every instruction that changes the pc can be considered as a jump
  * instruction.
  *
  * 2. It's the successor of one of the above (if a jump is conditional, than
  * the next instruction will be the first instruction of the fall-through
  * block, else it is either target of an other jump or a dead block. In the
  * latter case, it will be removed by the unreachable block removal routines,
  * so we just mark it as a basic block leader).
  *
  * 3. There is an address produced to this basic block, this is either 
  *
  *    a. done directly, using the program counter (we assume no real
  *    code-address calculations are present (so not address of function + x) 
  *
  *    b. using a relocation (and not necessary detected when scanning the
  *    instructions) 
  *
  * For both cases we need to assume there will be an indirect jump to this
  * address.
  *
  * 4. The start of data blocks in code or instructions following datablocks
  * are considered basic block leaders.
  *
   */

  /* We can still assume that all instructions are stored sequentially in the
  * array Ains the last instruction stored can be recognised with INSNEXT ==
  * NULL */

  t_ins * ins;
  t_uint64 pc;
  t_uint32 nleaders=0;
  t_uint8 slot=0;

  /*  RelocTablePrint(obj->reloc_table);*/

  /* Initialisation: The first instruction of the program is always a leader.
  * The reasoning is equal to that of reason 2: Either there will be a jump to
  * this address or it's the entry point or it is dead. */

  /* Instructions are still in a sequential list, stored in t_section->data */
  ins = code->data; 
  INS_ATTRIB(ins) |= IF_BBL_LEADER;

  /* pc maar met 16 verhogen na het doorlopen van een bundle */
  for (pc = 0; ins != NULL; ins = INS_NEXT(ins), pc += ((slot<2)?0:16), slot=(slot+1)%3)
  {
#if 0
    /* patchke, zolang niet alle relocs goed staan... */
    if(G_T_UINT64(INS_CADDRESS(ins))==0x4000000000000340ULL)
    {
      printf("Patched 0x4000000000000340 is Bbl-leader !\n");
      INS_ATTRIB(ins) |= IF_BBL_LEADER;
    }
#endif    
    /* 1. NORMAL BRANCHES (reason 1 and 2) */
    if (Ia64IsBranchOrCall(ins) ||Ia64IsReturn(ins)||Ia64IsCheck(ins) )
    {
      t_uint64 addr64;
      t_ins * target;
      /*t_address addr;*/

      /*special patch for those silly undefined calls and jumps*/
      if(Ia64IsIPRelativeBranchOrCall(ins) && RELOCATABLE_REFERS_TO(ins))
      {
	t_reloc *rel = RELOCATABLE_REFERS_TO(ins)->rel;
	if(GENERIC_EQ(REL_TO_RELOCATABLE(rel)->caddress, NULL_GENERIC))
	{
	  IA64INSIMM1(ins) = 0;
	  RelocTableRemoveReloc(obj->reloc_table, rel);
	 
//#if 0 /*special patchke die de ambetante calls omzet in jumps, waardoor extra dingen verwijderd kunnen worden!!!*/
	  if(Ia64IsCall(ins))
	  {
	    /*convert these into simple jumps*/
	    IA64INSOPCODE(ins) = GetBrCondOpcode("br.cond.dpnt.few");
	  }
//#endif
	}
      }

      /* The instruction following the branch instruction is a basic block leader */
      if (INS_NEXT(ins) != NULL) INS_ATTRIB((t_ia64_ins *) INS_NEXT(ins)) |= IF_BBL_LEADER;

      /* If it is a call, then the return will be to the next bundle */
      if(Ia64IsCall(ins))
      {
	t_ins * return_from_call = ins;
	t_uint8 cnt = 3 - slot;
	while(cnt--)
	  return_from_call = INS_NEXT(return_from_call);
	if (return_from_call != NULL) 
	  INS_ATTRIB((t_ia64_ins *) return_from_call) |= IF_BBL_LEADER;
      }

      /* The target of the branch will also be a basic block leader */

      if(Ia64IsIPRelativeBranchOrCall(ins)||Ia64IsCheck(ins) )
      {
	t_object * obj=code->obj;
	t_uint64 c_offset;

	addr64 = pc + IA64INSIMM1(ins);
	/*        GenericSet_uint64(addr,addr64);*/

	c_offset = obj->description->decoded_instruction_size*((addr64>>4)*3); 

	target = ((t_ins *) (((char *)  code->data) +c_offset ));

	/** \TODO: dit moet er nog bij, maar hij kent het veld address nog niet...
	  addr=GENERIC_ADD(addr,code->address);

	  target=SecGetInsByAddress(code,addr);
	 */        
	if (addr64<(SECTION_NINS(code)<<4)/3) 
	  INS_ATTRIB(target) |= IF_BBL_LEADER; 
#ifdef _DEBUG_FLOWGRAPH_
	else
	  printf("Target outside code section !!!\n");
#endif 
      }  

    }
    else
      /* 2. SYSTEM CALLS */
      if (Ia64IsBreak(ins)) 
      {
	/* Actually, we could just consider system calls as normal
	* instructions, if we were sure that they are not long-jump-like
	* system calls. Since we do not now that for sure here, we must
	* assume that is will alter control flow => next instruction is a
	* leader */

	if (INS_NEXT(ins) != NULL)
	  INS_ATTRIB((t_ia64_ins *) INS_NEXT(ins)) |= IF_BBL_LEADER; 
      }
      else
	if(IA64INSTYPE(ins) == IA64_TYPE_DATA && (!INS_PREV(ins) || IA64INSTYPE(INS_PREV(ins)) != IA64_TYPE_DATA))
	  INS_ATTRIB(ins) |= IF_BBL_LEADER;   

    /* moet ook altijd uitgevoerd worden, voor PCREL64I-type: deze markeert immers andere instructies als leader */
    /* 3. THE RELOCATIONS (reason 3) */
    if (RELOCATABLE_REFED_BY(ins))
    { 
      t_reloc_ref * ref;
      if(slot == 0) 
	INS_ATTRIB(ins) |= IF_BBL_LEADER;
      else
	FATAL(("Relocation inside a bundle"));
      //#if 0
      /* now a patch to add switch and other-arrows - should work for all (LSB) IA64-binaries */
#if 0
      for(ref=RELOCATABLE_REFED_BY(ins); ref != NULL; ref=ref->next)
      {
	t_reloc *reloc = ref->rel;

	FATAL(("Implement"));
	/*check for PCREL64I-type*/
	if(reloc->write_type == INSTRUCTION_IMMEDIATE64 && reloc->type == REPRESENT_SELFREL_VALUE)
	{
	  t_address target_addr = GENERIC_ADD_UINT32(INS_CADDRESS(reloc->to->base), 16);
	  if(reloc->offset != 2) /* als deze fout zou optreden, betekent dit volgens mij dat de mov_from_ip niet in de bundel direct na de movl zit => dan nog opvangen !!!*/
	    FATAL(("relocation type PCREL64I expected value 2 as offset !"));
#ifdef IA64_DEBUG_RELOCS
	  printf("Target address found: 0x%llx\n", target_addr);
#endif	  
	  INS_ATTRIB(SecGetInsByAddress(code, target_addr)) |=IF_BBL_LEADER;
	}




#if 0
	/*check for PCREL64LSB-type*/
	if(reloc->write_type == ABS64 && reloc->type == REPRESENT_SELFREL_VALUE)
	{
	  t_address switch_addr = GENERIC_ADD(INS_CADDRESS(reloc->to->base), reloc->addend);
	  printf("Switch address found: 0x%llx\n", switch_addr);
	  INS_ATTRIB(SecGetInsByAddress(code, switch_addr)) |= IF_BBL_LEADER;
	}

	/*check for DIR64LSB-type*/
	if(reloc->write_type == ABS64 && reloc->type == REPRESENT_VALUE)
	{
	  t_address dir_addr = GENERIC_ADD(INS_CADDRESS(reloc->to->base), reloc->addend);
	  printf("Direct address found: 0x%llx\n", dir_addr);
	  INS_ATTRIB(SecGetInsByAddress(code, dir_addr)) |= IF_BBL_LEADER;
	}
#endif 
      }

#endif
    }
  }
//#if 0
  {
    t_object *tmp, *tmp2;
    t_symbol *symptr;
    
    /*seeking for special $diablo-symbols*/
    for (tmp2=obj->mapped_first;tmp2!=NULL; tmp2=tmp2->next)
    { 
      for (tmp=tmp2; tmp!=NULL; tmp=tmp->equal)
      {     
	for (symptr=tmp->symbol_table->first; symptr!=NULL; symptr=symptr->next)
	{
	  if(!symptr->name)
	    continue;

	  if (strncmp(symptr->name,"$diablo:",8)==0)
	  {
	    int begin = strstr(symptr->name,"@-@")-symptr->name;
	    char save = symptr->name[begin];
	    t_symbol *s1, *s2;
	    t_address a1, a2;

//#if 0

#ifdef IA64_DEBUG_RELOCS
	    printf("symbol: %s @ 0x%llx\n", symptr->name, symptr->base->address);
#endif	    
	    symptr->name[begin] = '\0';


	    /*TODO ook checken als het symbool van zelfde .o-file komt*/
	    s1=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+8);
	    s2=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+begin+3);

	    a1 = GENERIC_ADD(SECTION_GET_CADDRESS(symptr->base), s1->current_address);
	    a2 = GENERIC_ADD(SECTION_GET_CADDRESS(symptr->base), s2->current_address);

/*
	    printf("%s @ 0x%llx;  %s @ 0x%llx\n", s1->name, a1, s2->name, a2);
	    fflush(stdout);
*/
	    /*char herstellen*/
	    symptr->name[begin] = save;
	    
	    /*controleren of wel bestaat !*/
	    if(!s1|| !s2)
	      FATAL(("Invalid special diablo symbol !!!"));


	    INS_ATTRIB(SecGetInsByAddress(code, a1))|= IF_BBL_LEADER;
#ifdef IA64_DEBUG_RELOCS
	    printf("First instruction of special diablo-symbol at %s had been marked as BBL_LEADER.\n", s1->name);
#endif
	    INS_ATTRIB(SecGetInsByAddress(code, a2))|= IF_BBL_LEADER;
#ifdef IA64_DEBUG_RELOCS
	    printf("First instruction of special diablo-symbol at %s had been marked as BBL_LEADER.\n", s2->name);
#endif	   
//#endif
	  }
	}

      }
    }

  }
//#endif
#if 0
  /* First a quick-hack patch in order to get some arrows form the hellnode to the blocks...*/
  /* This piece of code ensures that every such function will start in a new BBL_address*/
  {
    t_section * opd=SectionGetFromObjectByName(obj,".opd");
    if (opd)
    {
      t_uint64 *data = (t_uint64*)opd->data;
      t_uint32 cnt=0;
      t_uint32 opd_size = GENERIC_EXTRACT_UINT32(SECTION_SIZE(opd));
      t_uint32 currsize =0;
      t_uint64 max_address = G_T_UINT64(GENERIC_ADD(SECTION_GET_CADDRESS(code), SECTION_SIZE(code)));

      printf("max_text_address: %llx\n", max_address);

      while(currsize < opd_size)
      {
	if(*data != 0 && *data < max_address)
	{
	  t_address addr;
	  GenericSet_uint64(addr,*data);
	  if(!(INS_ATTRIB(SecGetInsByAddress(code, addr)) & IF_BBL_LEADER))
	  {
	    printf("Patched a BBL_LEADER to function pointer address: %llx\n", *data);
	    INS_ATTRIB(SecGetInsByAddress(code, addr)) |= IF_BBL_LEADER;
	    cnt++;
	  }
	}

	currsize +=16;
	data +=2;
      }

      printf("Number of patched functions as BBL_LEADER: %d\n", cnt);
      //    exit(0);
    }
    else
      FATAL(("Section .opd NOT found !!!"));   
  }
#endif


  /**/  
  ins = code->data; 
  while(ins)
  {
    if (INS_ATTRIB(ins)&IF_BBL_LEADER)
    {
#ifdef _DEBUG_FLOWGRAPH_
      printf("\n");
#endif   
      nleaders++;
    }
#ifdef _DEBUG_FLOWGRAPH_
    {
      InsPrint(ins, 0);
      //      static char buffer[256]; 
      //      Ia64PrintFunction_table[IA64OPCODEINFO(ins).instr_format](ins, buffer);
      //      printf("%s\n", buffer);
    }
#endif
    ins=INS_NEXT(ins);
  }
  return nleaders;
}

void Ia64MakeAddressProducers(/*t_object * obj, */t_section * sec)
{
  /* just a stub */
  return;
}

t_bool Ia64FunIsGlobal(t_function * fun)
{
  if (fun->function_flags & FF_IS_EXPORTED)
  {
    return TRUE;
  }
  return FALSE;

}
/* vim: set shiftwidth=2: */
