/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdlib.h>
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_function.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_liveness.h"
#include "../../kernel/diablo_constprop.h"
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_iterators.h"

#include "ia64_opcodes.h"
#include "ia64_utils.h"
#include "ia64_dummy.h"
#include "ia64_instructions.h"

#include "../../utils/util_options.h"

#include "ia64_inlining.h"




t_bool Ia64InlineFunAtCallsite(t_function * fun, t_bbl *call_site);

void Ia64Inlining(t_cfg *cfg)
{
  /* I suppose this still needs some work, no ? */

  t_bbl * i_bbl;
  t_cfg_edge * i_edge;
  t_bool flag = TRUE;
  t_function * i_fun;
  t_int32 teller, inlineCnt = 0, totalCnt = 0;
  t_bbl * call_site=NULL;
  t_ins * last_ins;
  t_function * tmp;

  static t_uint32 __inlineCnt = 0;

  printf("Trying to inline...\n");
  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    flag = TRUE;
    teller = 0;
    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	if (CfgEdgeIsForwardInterproc(i_edge) && EDGE_CAT(i_edge)!=ET_CALL)
	  flag = FALSE;

      BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
      {
	if (CfgEdgeIsForwardInterproc(i_edge))
	{
	  if (EDGE_CAT(i_edge)!=ET_CALL)
	    flag = FALSE;
	  else if (IsHellNode(EDGE_HEAD(i_edge)))
	    teller+=2;
	  else
	  {
	    teller++;
	    call_site = CFG_EDGE_HEAD(i_edge);
	  }
	}
      }
    }

    if (!flag || teller!=1)
      continue;

    BBL_FOREACH_PRED_EDGE(FUN_BBL_LAST(i_fun),i_edge)
    {
      last_ins = BBL_INS_LAST(CFG_EDGE_HEAD(i_edge));
      if (!last_ins)
      {
	FunctionPrint(i_fun,PRINT_INSTRUCTIONS);
	FATAL(("Found a return block without instructions\n"));
	//flag = FALSE;
	//break;
      }

      if(!Ia64IsReturn((t_ins*)last_ins))
      {
	printf("POTENTIAL: NO INLINE 2\n");
	flag = FALSE;
	break;
      }
      
    }

    if (!flag)
      printf("NO INLINE: PROBLEM WITH RETURN\n");
    else if (teller==1 && Ia64IsCall(BBL_INS_LAST(call_site)))
    {
//#define IA64_INLINING_DEBUGGING
#ifdef IA64_INLINING_DEBUGGING      
      if(++__inlineCnt>=global_options.debugcounter)
      {
	printf("******************** stop\n");
	BblPrint(call_site, 0);
	FunctionPrint(i_fun,PRINT_INSTRUCTIONS);
	return;
      }
#endif	
      inlineCnt += Ia64InlineFunAtCallsite(i_fun,call_site);
      totalCnt++;
#ifdef IA64_INLINING_DEBUGGING
      if(__inlineCnt+1>=global_options.debugcounter)
      {
	printf("******************** voorlaatste stop\n");
	BblPrint(call_site, 0);
	FunctionPrint(BBL_FUN(call_site),PRINT_INSTRUCTIONS);
	//return;
      }
#endif
    }
  }	  

  printf("\nInlining: %d/%d functions could be inlined !\n", inlineCnt, totalCnt);
  
}

//#if 0

/*
* Inlining is a bit more difficult on the IA-64, as there is a register-stack. This means the
* alloc of the calling function has to be enlarged [+ adjust all dummies in that function !]
* and all registers of the called (and inlined) function have to be adjusted !
* There cannot be rotating registers in the called function either.
* */



#define IA64_INLINING_WITH_CALLERS



t_bool Ia64InlineFunAtCallsite(t_function * fun, t_bbl *call_site)
{
  t_bbl* entry;
  t_bbl* i_bbl;
  t_cfg_edge * call_edge, * tmp;
  t_cfg_edge * link_edge;  
  t_cfg_edge * return_edge;  
  t_cfg_edge * i_edge, *j_edge;

  t_cfg_edge *funlink_edge;
  t_bbl *funlink_bbl = NULL;

  t_bbl * tmp_bbl;
  t_bbl * return_point;

  t_bool last_is_ip=FALSE;

  t_ins *alloc1 = NULL, *alloc2 = NULL;

//  t_uint32 skew = 0, expand = 0;

  t_uint32 cntShiftRegsInlinedFunction = 0;
  t_uint32 expandLocals = 0;
  t_uint32 expandOuts = 0;
  
  t_ins *ins;


  t_ins *dummy1 = INS_PREV(BBL_INS_LAST(call_site));
  t_ins *dummy2;

  t_cfg_edge * new_edge; 
  t_uint64 taken, evaluated;

  /* check for dummy1 */
  if(!IA64ISLIVEDUMMY( dummy1 ))
    FATAL(("No special dummy before call !"));


  /* make sure we know the alloc -> we still gonna have to search it...*/
  if(IA64INSIMM2(dummy1) == IA64_ALLOC_MISSING)
  {
    printf("WARNING: missing opportunity: cannot inline %s in %s because alloc is missing !\n",
	fun->name, BBL_FUN((call_site))->name);
    return FALSE;
  }

  /*search the first alloc  */
  alloc1 = Ia64SearchForAlloc(BBL_FUN(call_site));

  if(alloc1 == NULL || !FUN_ISALLOCCLEAN(BBL_FUN(call_site)))
  {
    printf("WARNING: missing opportunity: cannot inline %s in %s because alloc1 is missing !\n",
	fun->name, BBL_FUN((call_site))->name);

    return FALSE;

  }


  /* now search for the alloc in the called function, and see if there are no rotating regs !!! */
  /* als er geen alloc is, dan sol = 0, sof = aantal outputregs van vorige functie */

  alloc2 = Ia64SearchForAlloc(fun);


  if(CountAllocsInFun(fun) > 1 || !FUN_ISALLOCCLEAN(fun))
  {
    printf("WARNING: missing opportunity: cannot inline %s in %s because of multiple allocs in i_fun !\n",
	fun->name, BBL_FUN((call_site))->name);

    return FALSE;
    
  }

  if(CountAllocsInFun(BBL_FUN((call_site))) > 1)
  {
    printf("WARNING: missing opportunity: cannot inline %s in %s because of multiple allocs at call_site !\n",
	fun->name, BBL_FUN((call_site))->name);

    return FALSE;
    
  }
  
  /*check for rotating regs, indicated in alloc */

  if(IA64INSIMM3(alloc1) != 0)
  {
    printf("WARNING: missing opportunity: cannot inline %s in %s because sor != 0 in alloc1 !\n",
	fun->name, BBL_FUN((call_site))->name);

    return FALSE;      
  }

  if(alloc2 != NULL)
  {
    if(IA64INSIMM3(alloc2) != 0)
    {
      printf("WARNING: missing opportunity: cannot inline %s in %s because sor != 0 in alloc2 !\n",
	  fun->name, BBL_FUN((call_site))->name);

      return FALSE;
    }
  }
  else
   {
    printf("WARNING: no alloc2 found for %s !\n", fun->name);
  }


  /*CHANGED !!!!!!*/
#if 0
    FUN_FOREACH_BBL(fun, i_bbl)
    {
      BBL_FOREACH_SUCC_EDGE(i_bbl, i_edge)
      {
	if (IsHellNode(EDGE_TAIL(i_edge)))
	  return FALSE;

	if (CfgEdgeIsForwardInterproc(i_edge) && EDGE_CAT(i_edge)!=ET_CALL)
	{
          printf("WARNING: missing opportunity: ...\n");
	  return FALSE;
	}
      }
    }
#endif


  
  
#if 0  
  /* check if i_fun has any outgoing calls => there could be problems if so !!*/

  if(alloc2 != NULL)
  {
    FUN_FOREACH_BBL(fun, i_bbl)
    {
      BBL_FOREACH_SUCC_EDGE(i_bbl, i_edge)
      {
	if(EDGE_CAT(i_edge) == ET_CALL)
	{
	  /*ZEKER nog verfijnen !!! wel dan de dummies daarrond ook aanpassen !!! viz imm2, imm3*/


	  printf("WARNING: missing opportunity: cannot inline %s in %s because function is not leaf !\n",
	      fun->name, BBL_FUN((call_site))->name);	  

	  return FALSE;
	}
      }
    }
  }

#endif




  

  /* if there is an alloc2, expand alloc2 properly ! */

//  skew = IA64INSIMM1(alloc1);

  cntShiftRegsInlinedFunction = IA64INSIMM1(alloc1);
  
  if(alloc2 != NULL)
  {
//    expand = MAX(0, IA64INSIMM2(alloc2) - (IA64INSIMM2(alloc1) - IA64INSIMM1(alloc1)));
    expandLocals = IA64INSIMM1(alloc2);
    expandOuts = MAX(0, IA64INSIMM2(alloc2) - IA64INSIMM1(alloc2) - (IA64INSIMM2(alloc1) - IA64INSIMM1(alloc1)));
  }
  
#ifdef IA64_INLINING_WITH_CALLERS 
  /* don't do inlining with indirect calls...we cannot fully adjust everything...is suppose*/
  if(expandLocals + expandOuts > 0)
  {
    FUN_FOREACH_BBL(BBL_FUN((call_site)), i_bbl)
    {
      if(i_bbl->ins_last && Ia64IsBranchOrCall(i_bbl->ins_last))
      {
	if(!IA64_IS_BR(IA64INSIN1(i_bbl->ins_last)))
	  continue;

	printf("WARNING: missing opportunity: cannot inline %s in %s because of indirect branch or call !\n",
	    fun->name, BBL_FUN((call_site))->name);	  

	return FALSE;
      }

    }

  }

#if 0  
  FUN_FOREACH_BBL(BBL_FUN((call_site)), i_bbl)
  {

    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(Ia64IsModOrLoopBranch(ins))
      {
	printf("WARNING: missing opportunity: cannot inline %s in %s because of mod-or-loop-branch !\n",
	    fun->name, BBL_FUN((call_site))->name);	  

	return FALSE;

      }
    }
  }

  FUN_FOREACH_BBL(fun, i_bbl)
  {

    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(Ia64IsModOrLoopBranch(ins))
      {
	printf("WARNING: missing opportunity: cannot inline %s in %s because of mod-or-loop-branch !\n",
	    fun->name, BBL_FUN((call_site))->name);	  

	return FALSE;

      }
    }
  }
#endif  
  
#endif  

  if(expandLocals + expandOuts != 0)
  {
    if(IA64INSIMM2(alloc1) + expandLocals + expandOuts >= 96)
    {
      printf("WARNING: missing opportunity: cannot inline %s in %s because expanding alloc >= 96 !\n",
	  fun->name, BBL_FUN((call_site))->name);

      return FALSE; 
    }

    //IA64INSIMM2(alloc1)+=expand;
    
    /*pas alloc aan, en de cache*/
    IA64INSIMM2(alloc1)+=expandLocals + expandOuts;
    IA64INSIMM1(alloc1)+=expandLocals;

    FUN_SOL(BBL_FUN(call_site)) = IA64INSIMM1(alloc1);
    FUN_SOF(BBL_FUN(call_site)) = IA64INSIMM2(alloc1);
    FUN_SOO(BBL_FUN(call_site)) += expandOuts;
    

  }



  /* TODO: eventueel problemen bij functie 1, doordat de grootte van soo(utput) evt wijzigt
  * => kan evt voor problemen zorgen wanneer deze zelfde functie andere functies gaat oproepen... */

  printf("---> %s in %s\n", fun->name, BBL_FUN((call_site))->name); 


  /*remove the two dummies !*/
  InsKill(dummy1);

  BBL_FOREACH_SUCC_EDGE(call_site, funlink_edge)
  {
    if(EDGE_CAT(funlink_edge) & ET_FUNLINK)
      funlink_bbl=CFG_EDGE_TAIL(funlink_edge);
  }

  if(funlink_bbl == NULL)
    FATAL(("Could not find link-block"));

  dummy2 = funlink_bbl->ins_first;

  if(!IA64ISLIVEDUMMY(dummy2))
    FATAL(("No special dummy after call, at link-block !"));

  InsKill(dummy2);

  /* don't forget to delete alloc2 */
  if(alloc2 != NULL)
  {
    t_ia64_reg pfs_save = IA64INSOUT1(alloc2);
    
    InsKill(alloc2);
    
    /*bovendien, als er een mov.i ar.pfs, rx is, met alloc rx=ar.pfs, ... => deze moet ook weg !*/

    FUN_FOREACH_BBL(fun, i_bbl)
    {
      BBL_FOREACH_INS(i_bbl, ins)
      {
	if(IA64INSOUT1(ins) == IA64_AR_PFS && 
	    IA64INSIN1(ins) == pfs_save && /*dit evt achterwege laten...*/
	    IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_MOV_AR)
	{
	  InsKill(ins);
	  break;
	}
      }
    }

        
  }



  /*en nu nog alle GR's boven r31 in i_fun verschuiven !!!*/

#define IA64_IS_STACKED_REG(reg) ((reg) >= IA64_R32 && (reg) <= IA64_R127)

  FUN_FOREACH_BBL(fun,i_bbl)
  {
    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(IA64_IS_STACKED_REG(IA64INSIN1(ins)))
	IA64INSIN1(ins)+= cntShiftRegsInlinedFunction;

      if(IA64_IS_STACKED_REG(IA64INSIN2(ins)))
	IA64INSIN2(ins)+= cntShiftRegsInlinedFunction;

      if(IA64_IS_STACKED_REG(IA64INSIN3(ins)))
	IA64INSIN3(ins)+= cntShiftRegsInlinedFunction;

      if(IA64_IS_STACKED_REG(IA64INSOUT1(ins)))
	IA64INSOUT1(ins)+= cntShiftRegsInlinedFunction;

      if(IA64_IS_STACKED_REG(IA64INSOUT2(ins)))
	IA64INSOUT2(ins)+= cntShiftRegsInlinedFunction;

      INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
      INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);
      

    }
  }
#if 0
#define IA64_IS_OUT_REG(reg, shift) ((reg) >= IA64_R32 + shift && (reg) <= IA64_R127)  
  /*test: ook de output vand e huidige functie opschuiven...*/
  FUN_FOREACH_BBL(BBL_FUN((call_site)),i_bbl)
  {
    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(IA64_IS_OUT_REG(IA64INSIN1(ins), skew))
	IA64INSIN1(ins)+= expand;

      if(IA64_IS_OUT_REG(IA64INSIN2(ins), skew))
	IA64INSIN2(ins)+= expand;

      if(IA64_IS_OUT_REG(IA64INSIN3(ins), skew))
	IA64INSIN3(ins)+= expand;

      if(IA64_IS_OUT_REG(IA64INSOUT1(ins), skew))
	IA64INSOUT1(ins)+= expand;

      if(IA64_IS_OUT_REG(IA64INSOUT2(ins), skew))
	IA64INSOUT2(ins)+= expand;

    }
  }
#endif


  
#ifdef IA64_INLINING_WITH_CALLERS  

#define IA64_IS_OUT_REG(reg, shift) ((reg) >= IA64_R32 + shift && (reg) <= IA64_R127)
  {
    t_bool activated = FALSE;
    
    FUN_FOREACH_BBL_R(BBL_FUN((call_site)),i_bbl) /* in reverse werken */
    {
      if(i_bbl == call_site)
	activated = TRUE; /* vanaf hier moet je niets verschuiven... */
      else
	if(activated && i_bbl->ins_last && Ia64IsCall(i_bbl->ins_last))
	  activated = FALSE;

      if(!activated)
	BBL_FOREACH_INS(i_bbl, ins)
	{
	  if(IA64_IS_OUT_REG(IA64INSIN1(ins), cntShiftRegsInlinedFunction))
	    IA64INSIN1(ins)+= expandLocals;

	  if(IA64_IS_OUT_REG(IA64INSIN2(ins), cntShiftRegsInlinedFunction))
	    IA64INSIN2(ins)+= expandLocals;

	  if(IA64_IS_OUT_REG(IA64INSIN3(ins), cntShiftRegsInlinedFunction))
	    IA64INSIN3(ins)+= expandLocals;

	  if(IA64_IS_OUT_REG(IA64INSOUT1(ins), cntShiftRegsInlinedFunction))
	    IA64INSOUT1(ins)+= expandLocals;

	  if(IA64_IS_OUT_REG(IA64INSOUT2(ins), cntShiftRegsInlinedFunction))
	    IA64INSOUT2(ins)+= expandLocals;
	
      	  INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
          INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);

	}
    }
  }  
#endif
  

  
  


  /* ook alle dummies van deze functie at callsite aanpassen !*/
  FUN_FOREACH_BBL(BBL_FUN((call_site)),i_bbl)
  {
    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(IA64ISLIVEDUMMY(ins))
      {
	IA64INSIMM2(ins) = IA64INSIMM1(alloc1);
	IA64INSIMM3(ins) = IA64INSIMM2(alloc1);
      }
    }
  }
  
#ifdef IA64_INLINING_WITH_CALLERS  
  /* en zelfs alle dummies at i_fun*/
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(IA64ISLIVEDUMMY(ins))
      {
	IA64INSIMM2(ins) = IA64INSIMM1(alloc1);
	IA64INSIMM3(ins) = IA64INSIMM2(alloc1);
      }
    }
  }

  
#endif  



  /* Look at all the predecessors of the entry block of the function, to find
  * the call edge */

  entry = fun->bbl_first;
  BBL_FOREACH_PRED_EDGE_SAFE(entry,call_edge,tmp)
  {
    t_bbl* site = CFG_EDGE_HEAD(call_edge);
    if (site!=call_site) continue;
    break;
  }      

  /* Look at all the successor edges of the callsite to find the link edge (and
  * thus the returnsite) */

  BBL_FOREACH_SUCC_EDGE_SAFE(call_site,link_edge,tmp)
  {
    if (EDGE_CAT(link_edge)==ET_FUNLINK || EDGE_CAT(link_edge)==ET_IPFUNLINK)
      break;
  }

  return_edge = EDGE_CORR(call_edge);
  return_point = CFG_EDGE_TAIL(link_edge);



  /* If the link edge is interprocedural we need to remember this (the added
  * branches that replace the returns must be also interprocedural */

  if (CfgEdgeIsInterproc(link_edge))
    last_is_ip=TRUE;

  /* Kill the edges */
  if (last_is_ip)
    CfgEdgeKill(link_edge->corr_edge);
  taken = EDGE_TAKEN(call_edge);
  evaluated = EDGE_EVALUATED(call_edge);
  CfgEdgeKill(link_edge);
  CfgEdgeKill(call_edge);
  CfgEdgeKill(return_edge);


  /* Change the call of the last ins into a branch instruction */
  if(IA64INSPRED(BBL_INS_LAST(call_site)) == IA64_P0)
  {
    InsKill(BBL_INS_LAST(call_site));
    new_edge = CfgEdgeNew(fun->cfg,call_site,fun->bbl_first,ET_FALLTHROUGH);
  }
  else  
  {
    Ia64InstructionMakeCondBranch(BBL_INS_LAST(call_site), IA64INSPRED(BBL_INS_LAST(call_site)));
    /* Add the jump that replaces the call */
    new_edge = CfgEdgeNew(fun->cfg,call_site,FUN_BBL_FIRST(fun),ET_JUMP);
    if(BBL_PROFILE_INFO(FUN_BBL_FIRST(fun))>0)
      EDGE_TAKEN(new_edge)=taken;
  }

  /* Add all blocks but the exit block to the calling function */
  FUN_FOREACH_BBL_SAFE_R(fun,i_bbl,tmp_bbl)
  {
    if (i_bbl!=FUN_BBL_LAST(fun))
    {	
      i_bbl->next_in_fun=call_site->next_in_fun;
      call_site->next_in_fun->prev_in_fun=i_bbl;
      call_site->next_in_fun = i_bbl;
      i_bbl->prev_in_fun = call_site;
      i_bbl->fun=call_site->fun;
    }
  }

  fun->bbl_first = fun->bbl_last;
  fun->bbl_last->prev_in_fun = NULL;
  fun->bbl_last->next_in_fun = NULL;


  /* Look at the predecessors of the exit block, to identify all returnsites */

  BBL_FOREACH_PRED_EDGE_SAFE(FUN_BBL_LAST(fun),i_edge,tmp)
  {
    t_bbl * head = CFG_EDGE_HEAD(i_edge);
    t_ins * last_ins;
    t_cfg_edge * new_edge;
    last_ins = BBL_INS_LAST(head);
    if (!last_ins)
    {
      FunctionPrint(fun,PRINT_INSTRUCTIONS);
      FATAL(("This should be impossible: we encountered a returnsite without instructions. As a return needs an explicit return instruction this is impossible\n"));
    }

    /* Change last instruction of return site */

    if (Ia64IsReturn((last_ins)))
    {
      Ia64InstructionMakeCondBranch(last_ins,IA64INSPRED(last_ins));
    }
    else  
      FATAL(("NO GOOD LAST INSTRUCTION\n"));




    /* Kill the return edge, add a (IP)JUMP edge */
    CfgEdgeKill(i_edge);

    if (last_is_ip)
    {
      j_edge = CfgEdgeNew(fun->cfg,head,return_point,ET_IPJUMP);
      CfgEdgeNewCompensating(fun->cfg,j_edge);
    }
    else
    {
      j_edge = CfgEdgeNew(fun->cfg,head,return_point,ET_JUMP);
    }

    j_edge->cfg = fun->cfg;
  }

  BblKill(fun->bbl_last);
  FunctionKill(fun);

  return TRUE;

}
//#endif

/* vim: set shiftwidth=2: */
