/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_instructions.h"
#include "ia64_opcodes.h"
#include "../../utils/util_error.h"
#include "../../utils/util_malloc.h"
#include <string.h>

#include "ia64_utils.h"

const t_ia64_unit templateField [32] [3] =
        {
        {IA64_UNIT_M,IA64_UNIT_I,IA64_UNIT_I},{IA64_UNIT_M,IA64_UNIT_I,IA64_UNIT_I},
        {IA64_UNIT_M,IA64_UNIT_I,IA64_UNIT_I},{IA64_UNIT_M,IA64_UNIT_I,IA64_UNIT_I},
        {IA64_UNIT_M,IA64_UNIT_L,IA64_UNIT_X},{IA64_UNIT_M,IA64_UNIT_L,IA64_UNIT_X},
        {IA64_UNIT_NIL,IA64_UNIT_NIL,IA64_UNIT_NIL},{IA64_UNIT_NIL,IA64_UNIT_NIL,IA64_UNIT_NIL},
        {IA64_UNIT_M,IA64_UNIT_M,IA64_UNIT_I},{IA64_UNIT_M,IA64_UNIT_M,IA64_UNIT_I},
        {IA64_UNIT_M,IA64_UNIT_M,IA64_UNIT_I},{IA64_UNIT_M,IA64_UNIT_M,IA64_UNIT_I},
        {IA64_UNIT_M,IA64_UNIT_F,IA64_UNIT_I},{IA64_UNIT_M,IA64_UNIT_F,IA64_UNIT_I},
        {IA64_UNIT_M,IA64_UNIT_M,IA64_UNIT_F},{IA64_UNIT_M,IA64_UNIT_M,IA64_UNIT_F},
        {IA64_UNIT_M,IA64_UNIT_I,IA64_UNIT_B},{IA64_UNIT_M,IA64_UNIT_I,IA64_UNIT_B},
        {IA64_UNIT_M,IA64_UNIT_B,IA64_UNIT_B},{IA64_UNIT_M,IA64_UNIT_B,IA64_UNIT_B},
        {IA64_UNIT_NIL,IA64_UNIT_NIL,IA64_UNIT_NIL},{IA64_UNIT_NIL,IA64_UNIT_NIL,IA64_UNIT_NIL},
        {IA64_UNIT_B,IA64_UNIT_B,IA64_UNIT_B},{IA64_UNIT_B,IA64_UNIT_B,IA64_UNIT_B},
        {IA64_UNIT_M,IA64_UNIT_M,IA64_UNIT_B},{IA64_UNIT_M,IA64_UNIT_M,IA64_UNIT_B},
        {IA64_UNIT_NIL,IA64_UNIT_NIL,IA64_UNIT_NIL},{IA64_UNIT_NIL,IA64_UNIT_NIL,IA64_UNIT_NIL},
        {IA64_UNIT_M,IA64_UNIT_F,IA64_UNIT_B},{IA64_UNIT_M,IA64_UNIT_F,IA64_UNIT_B},
        {IA64_UNIT_NIL,IA64_UNIT_NIL,IA64_UNIT_NIL},{IA64_UNIT_NIL,IA64_UNIT_NIL,IA64_UNIT_NIL},
        };


void Ia64InstructionMakeNoop(t_ins * ins)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);

  t_reloc_table * reloc_table;  
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  IA64INSTYPE(ins) = IA64_TYPE_M;
  IA64INSOPCODE(ins) = GetNopOpcode(IA64_UNIT_M);

  IA64INSPRED(ins) = IA64_P0;

  IA64INSIN1(ins) = IA64_NONE;
  IA64INSIN2(ins) = IA64_NONE;
  IA64INSIN3(ins) = IA64_NONE;

  IA64INSOUT1(ins) = IA64_NONE;
  IA64INSOUT2(ins) = IA64_NONE;

  IA64INSIMM1(ins) = 0;
  IA64INSIMM2(ins) = 0;
  IA64INSIMM3(ins) = 0;

#ifndef IA64_NEW_BBL_ADDRESS
  IA64INSORIGINSTR(ins)=IA64INSOPCODE(ins);
#endif 
  
  INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
  INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);


  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */ 
     
  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RELOCATABLE_REFERS_TO(ins) = RELOCATABLE_REFERS_TO(ins)->next;

    RelocTableRemoveReloc(reloc_table, relref->rel);
    Free(relref);
    relref = RELOCATABLE_REFERS_TO(ins);
  }


}



void Ia64InstructionMakeAdd(t_ins * ins, t_bool useAddl, t_ia64_reg out, t_int64 imm, t_ia64_reg in)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);

  t_reloc_table * reloc_table;  
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  IA64INSTYPE(ins) = IA64_TYPE_A;
  if(useAddl)
    IA64INSOPCODE(ins) = GetAddlOpcode();
  else
    IA64INSOPCODE(ins) = GetAddlOpcode();

  if(useAddl)
  {
    if(in > IA64_R3)
      FATAL(("Only R0...R3 are allowed for ADDL"));
    
    if(imm < 0)
    {
      if((-imm) > (1<<21))
	FATAL(("Constant too small for ADDL"));
    }
    else
      if(imm > (1<<21))
	FATAL(("Constant too large for ADDL"));
    
  }
  else
  {
    if(imm < 0)
    {
      if((-imm) > (1<<13))
	FATAL(("Constant too small for ADDS"));
    }
    else
      if(imm >= (1<<13))
	FATAL(("Constant too large for ADDS"));
  }
  

  IA64INSIN1(ins) = in;
  IA64INSIN2(ins) = IA64_NONE;
  IA64INSIN3(ins) = IA64_NONE;

  IA64INSOUT1(ins) = out;
  IA64INSOUT2(ins) = IA64_NONE;

  IA64INSIMM1(ins) = imm;
  IA64INSIMM2(ins) = 0;
  IA64INSIMM3(ins) = 0;

#ifndef IA64_NEW_BBL_ADDRESS
  IA64INSORIGINSTR(ins)=IA64INSOPCODE(ins);
#endif 
  
  INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
  INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);


  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */ 
     
  while ((relref=RELOCATABLE_REFERS_TO(ins)))
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
  }


  
}



/** \TODO is gebruik van InsNewForSec wel ok ??? */
t_ins * Ia64InstructionCreateNoop(t_section *code, t_ia64_unit unit)
{
  t_ins * ins_s;
  ins_s = InsNewForSec(code);
  if (!ins_s) FATAL(("Could not create Nop!"));

  INS_SECTION(ins_s) = code;

  RELOCATABLE_REFERS_TO(ins_s)=NULL;  /* TODO: Check if this is necessary */
  RELOCATABLE_REFED_BY(ins_s)=NULL; /* TODO: Check if this is necessary */

  IA64INSTYPE(ins_s) = unit == IA64_UNIT_L ? IA64_TYPE_DUMMY : unit2instr_type(0x0ULL, unit);        
  IA64INSOPCODE(ins_s) = GetNopOpcode(unit);

  IA64INSPRED(ins_s) = IA64_P0;

  IA64INSIN1(ins_s) = IA64_NONE;
  IA64INSIN2(ins_s) = IA64_NONE;
  IA64INSIN3(ins_s) = IA64_NONE;

  IA64INSOUT1(ins_s) = IA64_NONE;
  IA64INSOUT2(ins_s) = IA64_NONE;

  IA64INSIMM1(ins_s) = 0;
  IA64INSIMM2(ins_s) = 0;
  IA64INSIMM3(ins_s) = 0;

#ifndef IA64_NEW_BBL_ADDRESS
  IA64INSORIGINSTR(ins_s)=IA64INSOPCODE(ins_s);
#endif   
  
  INS_PREV(ins_s) = NULL;
  INS_NEXT(ins_s) = NULL;


  INS_REGS_USE(ins_s)=Ia64UsedRegisters(ins_s);
  INS_REGS_DEF(ins_s)=Ia64DefinedRegisters(ins_s);


  return ins_s;
}


void Ia64InstructionMakeCondBranch(t_ins * ins, t_ia64_reg pred)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  /* Make a conditional branch */


  IA64INSTYPE(ins) = IA64_TYPE_B;
  
  /* afhankelijk van het feit of het al dan niet conditioneel is ... */
  if(pred == IA64_P0)
    IA64INSOPCODE(ins) = GetBrCondOpcode("br.cond.sptk.many");
  else
    IA64INSOPCODE(ins) = GetBrCondOpcode("br.cond.dpnt.few");

  IA64INSPRED(ins) = pred;

  if (pred != IA64_P0)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_BRANCH;
  

  IA64INSIN1(ins) = IA64_NONE;
  IA64INSIN2(ins) = IA64_NONE;
  IA64INSIN3(ins) = IA64_NONE;

  IA64INSOUT1(ins) = IA64_NONE;
  IA64INSOUT2(ins) = IA64_NONE;

  IA64INSIMM1(ins) = 0; /* het zou kunnen dat dit nog ingevuld zal moeten worden !!! */
  IA64INSIMM2(ins) = 0;
  IA64INSIMM3(ins) = 0;

#ifndef IA64_NEW_BBL_ADDRESS
  IA64INSORIGINSTR(ins)=IA64INSOPCODE(ins);
#endif 
  
  INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
  INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);

  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_ATTRIB(ins) |= IF_PCDEF;


  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    FATAL(("Reimplement"));
#if 0
    t_reloc_ref * reffedby = NULL,*prev = NULL;
    if (relref->rel->to->to_type == TO_BBL) 
    {  
      /* TODO: Bbl's should also have a reffed by field */
      reffedby=NULL;
    }
    else if(relref->rel->to->to_type == TO_INSTRUCTION) 
    { 
      reffedby = RELOCATABLE_REFED_BY(relref->rel->to->base); 
    }
    else if(relref->rel->to->to_type == TO_SUB_SECTION)
    {
      reffedby=NULL;
    }
    else if (relref->rel->to->to_type == TO_NONE)
    {
      reffedby=NULL;
    }
    else
    {
      //	printf("%x\n",relref->rel->to->to_type);
      FATAL(("Relocation ttype not implemented!"));
    }

    RELOCATABLE_REFERS_TO(ins) = RELOCATABLE_REFERS_TO(ins)->next;


    if (reffedby)
    {
      while (reffedby != NULL && reffedby->rel != relref->rel)
      {
	prev = reffedby;
	reffedby = reffedby->next;
      }

      if (reffedby == NULL) FATAL(("Relocations corrupt!\n"));
      if (prev) prev->next = reffedby->next;
      if (reffedby) Free(reffedby);
    }
    RelocTableRemoveReloc(reloc_table, relref->rel);
    Free(relref);
    relref = RELOCATABLE_REFERS_TO(ins);
#endif
  }
  
}









#define IA64_OP(instr) (((instr) >> 37) & 0x0F)


t_ia64_unit * DecodeTemplate(const t_bundle *bundle)
{
	if (templateField[bundle->template][0]==IA64_UNIT_NIL) FATAL(("NOT A VALID TEMPLATE, POSSIBLY DATA"));
	else return (t_ia64_unit *)templateField[bundle->template];
	return NULL; /* keep compiler happy */
}

t_ia64_instr_type unit2instr_type(t_uint64 instruction, t_ia64_unit unit)
{
  t_ia64_instr_type type;
  t_uint8 op;

  op = IA64_OP(instruction);

  if (op >= 8 && (unit == IA64_UNIT_I || unit == IA64_UNIT_M))
      type = IA64_TYPE_A;
  else
    {
      switch (unit)
    	{
    	case IA64_UNIT_I:
    	  type = IA64_TYPE_I; break;
    	case IA64_UNIT_M:
    	  type = IA64_TYPE_M; break;
    	case IA64_UNIT_B:
    	  type = IA64_TYPE_B; break;
    	case IA64_UNIT_F:
    	  type = IA64_TYPE_F; break;
      case IA64_UNIT_L:
    	case IA64_UNIT_X:
    	  type = IA64_TYPE_X; break;
    	default:
    	  type = IA64_TYPE_NIL;
    	}
    }
  return type;
}


t_bool IsFollowedByStop(const t_bundle *bundle, t_uint8 slot)
{
	if(slot == 0 && (bundle->template == 0x0A || bundle->template == 0x0B))
		return 1;

	if(slot == 1 && (bundle->template == 0x02 || bundle->template == 0x03))
		return 1;

	if(slot == 2 && (bundle->template & 0x1) == 1)
		return 1;
	
		
	return 0;
}


/* zou nog veel duidelijker en sneller moeten gemaakt worden !!! */
void MakeTemplate(t_bundle * bundle, const t_ia64_instr_type *instr_types, const t_bool *followedByStop)
{
  if(followedByStop[0])
    bundle->template=0x0A;
  else
    if(followedByStop[1])
      bundle->template=0x02;
    else
      if(instr_types[2] == IA64_TYPE_X)	
	bundle->template=0x04;
      else
	if(instr_types[2] == IA64_TYPE_F)	
	  bundle->template=0x0E;
	else
	  if(instr_types[0] == IA64_TYPE_B)	
	    bundle->template=0x16;
	  else
	    if(instr_types[1] == IA64_TYPE_F && instr_types[2] == IA64_TYPE_B)	
	      bundle->template=0x1C;
	    else
	      if(instr_types[1] == IA64_TYPE_F)	
		bundle->template=0x0C;
	      else
		if(instr_types[1] == IA64_TYPE_B)	
		  bundle->template=0x12;
		else
		  if(instr_types[2] == IA64_TYPE_B)	
		  { /* 10 en 18 zijn nog mogelijk */
		    /* merk op dat we op slot1 voor type A eig de keuze hebben: 
		    * evt nog optimaliseren naargelang de reeds gebruikte units;
		    * nu steken we hem standaard op I 
		     */
		    if(instr_types[1] == IA64_TYPE_M)
		      bundle->template=0x18;
		    else
		      bundle->template=0x10;
		  }
		  else
		  { /* 00 en 08 zijn nog mogelijk */
		    /* merk op dat we op slot1 voor type A eig de keuze hebben: 
		    * evt nog optimaliseren naargelang de reeds gebruikte units;
		    * nu steken we hem standaard op I 
		     */
		    if(instr_types[1] == IA64_TYPE_M)
		      bundle->template=0x08;
		    else
		      bundle->template=0x00;

		  }

  if(followedByStop[2])
    bundle->template++;
}





t_bool Ia64IsBranch(t_ins *ins)
{
  t_ia64_instr_format format = 
    IA64OPCODEINFO(ins).instr_format;
    
  switch(format)
  {
    case IA64_INSTR_FORMAT_B1:
    case IA64_INSTR_FORMAT_B2:
/*    case IA64_INSTR_FORMAT_B6:*/
/*    case IA64_INSTR_FORMAT_B7:*/
    case IA64_INSTR_FORMAT_X3:
      return 1;
    case IA64_INSTR_FORMAT_B4:
      if(strncmp("br.ret.", IA64OPCODEINFO(ins).desc, 7) == 0)
        return 0;
      else
        return 1;
    default:
      return 0;
  }
}


t_bool Ia64IsCheck(t_ins *ins)
{
  return IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_CHK;
}

t_bool Ia64IsCall(t_ins *ins)
{
  t_ia64_instr_format format = 
    IA64OPCODEINFO(ins).instr_format;
    
  switch(format)
  {
    case IA64_INSTR_FORMAT_B3:
    case IA64_INSTR_FORMAT_B5:
    case IA64_INSTR_FORMAT_X4:
      return 1;
    default:
      return 0;
  }
}


t_bool Ia64IsBranchOrCall(t_ins *ins)
{
  return Ia64IsBranch(ins) || Ia64IsCall(ins);
}


t_bool Ia64IsIPRelativeBranch(t_ins *ins)
{
  t_ia64_instr_format format = 
    IA64OPCODEINFO(ins).instr_format;
    
  switch(format)
  {
    case IA64_INSTR_FORMAT_B1:
    case IA64_INSTR_FORMAT_B2:
/*    case IA64_INSTR_FORMAT_B6:*/
    case IA64_INSTR_FORMAT_X3:
      return 1;
    default:
      return 0;
  }
}

t_bool Ia64IsIPRelativeCall(t_ins *ins)
{
  t_ia64_instr_format format = 
    IA64OPCODEINFO(ins).instr_format;
    
  switch(format)
  {
    case IA64_INSTR_FORMAT_B3:
    case IA64_INSTR_FORMAT_X4:
      return 1;
    default:
      return 0;
  }
}

t_bool Ia64IsIPRelativeBranchOrCall(t_ins *ins)
{
  return Ia64IsIPRelativeBranch(ins) || Ia64IsIPRelativeCall(ins);
}

t_bool Ia64IsReturn(t_ins *ins)
{
  return IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_B4 
    && strncmp("br.ret.", IA64OPCODEINFO(ins).desc, 7) == 0;
}

t_bool Ia64IsModOrLoopBranch(t_ins *ins)
{
  return IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_B2 ||
    (IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_B1 && 
     strncmp("br.cond.", IA64OPCODEINFO(ins).desc, 8) != 0);
}

t_tristate Ia64IsSyscallExit(t_ins * ins)
{
  return PERHAPS;
}

t_bool Ia64InsIsIndirectCall(t_ins * ins)
{
  return IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_B5;
}

/* vim: set shiftwidth=2: */
