/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_opcodes_a.h"
#include "../../../utils/util_error.h"

#include "../ia64_instructions.h"
#include "../ia64_disassemble/ia64_disassemble_one.h"
#include "../../../simple_types.h"

/* instruction bit fields: */
#define bC(x)		(((t_uint64) ((x) & 0x1)) << 12)
#define bImm14(x)	((((t_uint64) (((x) >>  0) & 0x7f)) << 13) | \
			 (((t_uint64) (((x) >>  7) & 0x3f)) << 27) | \
			 (((t_uint64) (((x) >> 13) & 0x01)) << 36))
#define bR3a(x)		(((t_uint64) ((x) & 0x7f)) << 20)
#define bR3b(x)		(((t_uint64) ((x) & 0x3)) << 20)
#define bTa(x)		(((t_uint64) ((x) & 0x1)) << 33)
#define bTb(x)		(((t_uint64) ((x) & 0x1)) << 36)
#define bVe(x)		(((t_uint64) ((x) & 0x1)) << 33)
#define bX(x)		(((t_uint64) ((x) & 0x1)) << 33)
#define bX2(x)		(((t_uint64) ((x) & 0x3)) << 34)
#define bX2a(x)		(((t_uint64) ((x) & 0x3)) << 34)
#define bX2b(x)		(((t_uint64) ((x) & 0x3)) << 27)
#define bX4(x)		(((t_uint64) ((x) & 0xf)) << 29)
#define bZa(x)		(((t_uint64) ((x) & 0x1)) << 36)
#define bZb(x)		(((t_uint64) ((x) & 0x1)) << 33)


/* instruction bit masks: */
#define mC	bC (-1)
#define mImm14	bImm14 (-1)
#define mR3a	bR3a (-1)
#define mR3b	bR3b (-1)
#define mTa	bTa (-1)
#define mTb	bTb (-1)
#define mVe	bVe (-1)
#define mX	bX (-1)
#define mX2	bX2 (-1)
#define mX2a	bX2a (-1)
#define mX2b	bX2b (-1)
#define mX4	bX4 (-1)
#define mZa	bZa (-1)
#define mZb	bZb (-1)

#define OpR3b(a,b)		(bOp (a) | bR3b (b)), (mOp | mR3b)
#define OpX2aVe(a,b,c)		(bOp (a) | bX2a (b) | bVe (c)), \
				(mOp | mX2a | mVe)
#define OpX2aVeR3a(a,b,c,d)	(bOp (a) | bX2a (b) | bVe (c) | bR3a (d)), \
				(mOp | mX2a | mVe | mR3a)
#define OpX2aVeImm14(a,b,c,d)	(bOp (a) | bX2a (b) | bVe (c) | bImm14 (d)), \
				(mOp | mX2a | mVe | mImm14)
#define OpX2aVeX4(a,b,c,d)	(bOp (a) | bX2a (b) | bVe (c) | bX4 (d)), \
				(mOp | mX2a | mVe | mX4)
#define OpX2aVeX4X2b(a,b,c,d,e)	\
	(bOp (a) | bX2a (b) | bVe (c) | bX4 (d) | bX2b (e)), \
	(mOp | mX2a | mVe | mX4 | mX2b)
#define OpX2TbTaC(a,b,c,d,e) \
	(bOp (a) | bX2 (b) | bTb (c) | bTa (d) | bC (e)), \
	(mOp | mX2 | mTb | mTa | mC)
#define OpX2TaC(a,b,c,d)	(bOp (a) | bX2 (b) | bTa (c) | bC (d)), \
				(mOp | mX2 | mTa | mC)
#define OpX2aZaZbX4(a,b,c,d,e) \
	(bOp (a) | bX2a (b) | bZa (c) | bZb (d) | bX4 (e)), \
	(mOp | mX2a | mZa | mZb | mX4)
#define OpX2aZaZbX4X2b(a,b,c,d,e,f) \
	(bOp (a) | bX2a (b) | bZa (c) | bZb (d) | bX4 (e) | bX2b (f)), \
	(mOp | mX2a | mZa | mZb | mX4 | mX2b)






const t_ia64_opcode ia64_opcode_table_a[] =
  {
    /* A-type instruction encodings (sorted according to major opcode) */

    {"add",      OpX2aVeX4X2b (8, 0, 0, 0, 0), IA64_INSTR_FORMAT_A1, IA64_INSTR_ADD},
    {"add",      OpX2aVeX4X2b (8, 0, 0, 0, 1), IA64_INSTR_FORMAT_A1, IA64_INSTR_ADD},
    {"sub",      OpX2aVeX4X2b (8, 0, 0, 1, 1), IA64_INSTR_FORMAT_A1, IA64_INSTR_SUB},
    {"sub",      OpX2aVeX4X2b (8, 0, 0, 1, 0), IA64_INSTR_FORMAT_A1, IA64_INSTR_SUB},
    {"addp4",    OpX2aVeX4X2b (8, 0, 0, 2, 0), IA64_INSTR_FORMAT_A1, IA64_INSTR_ADDP4},
    {"and",      OpX2aVeX4X2b (8, 0, 0, 3, 0), IA64_INSTR_FORMAT_A1, IA64_INSTR_AND},
    {"andcm",    OpX2aVeX4X2b (8, 0, 0, 3, 1), IA64_INSTR_FORMAT_A1, IA64_INSTR_ANDCM},
    {"or",       OpX2aVeX4X2b (8, 0, 0, 3, 2), IA64_INSTR_FORMAT_A1, IA64_INSTR_OR},
    {"xor",      OpX2aVeX4X2b (8, 0, 0, 3, 3), IA64_INSTR_FORMAT_A1, IA64_INSTR_XOR},

    {"shladd",   OpX2aVeX4 (8, 0, 0, 4), IA64_INSTR_FORMAT_A2, IA64_INSTR_SHLADD},
    {"shladdp4", OpX2aVeX4 (8, 0, 0, 6), IA64_INSTR_FORMAT_A2, IA64_INSTR_SHLADDP4},

    {"sub",      OpX2aVeX4X2b (8, 0, 0, 9, 1), IA64_INSTR_FORMAT_A3, IA64_INSTR_SUB},
    {"and",      OpX2aVeX4X2b (8, 0, 0, 0xb, 0), IA64_INSTR_FORMAT_A3, IA64_INSTR_AND},
    {"andcm",    OpX2aVeX4X2b (8, 0, 0, 0xb, 1), IA64_INSTR_FORMAT_A3, IA64_INSTR_ANDCM},
    {"or",       OpX2aVeX4X2b (8, 0, 0, 0xb, 2), IA64_INSTR_FORMAT_A3, IA64_INSTR_OR},
    {"xor",      OpX2aVeX4X2b (8, 0, 0, 0xb, 3), IA64_INSTR_FORMAT_A3, IA64_INSTR_XOR},

    {"adds",     OpX2aVe (8, 2, 0), IA64_INSTR_FORMAT_A4, IA64_INSTR_ADD},
    {"addp4",    OpX2aVe (8, 3, 0), IA64_INSTR_FORMAT_A4, IA64_INSTR_ADDP4},

    {"padd1",            OpX2aZaZbX4X2b (8, 1, 0, 0, 0, 0), IA64_INSTR_FORMAT_A9, IA64_INSTR_PADD},
    {"padd2",            OpX2aZaZbX4X2b (8, 1, 0, 1, 0, 0), IA64_INSTR_FORMAT_A9, IA64_INSTR_PADD},
    {"padd4",            OpX2aZaZbX4X2b (8, 1, 1, 0, 0, 0), IA64_INSTR_FORMAT_A9, IA64_INSTR_PADD},
    {"padd1.sss",        OpX2aZaZbX4X2b (8, 1, 0, 0, 0, 1), IA64_INSTR_FORMAT_A9, IA64_INSTR_PADD},
    {"padd2.sss",        OpX2aZaZbX4X2b (8, 1, 0, 1, 0, 1), IA64_INSTR_FORMAT_A9, IA64_INSTR_PADD},
    {"padd1.uuu",        OpX2aZaZbX4X2b (8, 1, 0, 0, 0, 2), IA64_INSTR_FORMAT_A9, IA64_INSTR_PADD},
    {"padd2.uuu",        OpX2aZaZbX4X2b (8, 1, 0, 1, 0, 2), IA64_INSTR_FORMAT_A9, IA64_INSTR_PADD},
    {"padd1.uus",        OpX2aZaZbX4X2b (8, 1, 0, 0, 0, 3), IA64_INSTR_FORMAT_A9, IA64_INSTR_PADD},
    {"padd2.uus",        OpX2aZaZbX4X2b (8, 1, 0, 1, 0, 3), IA64_INSTR_FORMAT_A9, IA64_INSTR_PADD},
    {"psub1",            OpX2aZaZbX4X2b (8, 1, 0, 0, 1, 0), IA64_INSTR_FORMAT_A9, IA64_INSTR_PSUB},
    {"psub2",            OpX2aZaZbX4X2b (8, 1, 0, 1, 1, 0), IA64_INSTR_FORMAT_A9, IA64_INSTR_PSUB},
    {"psub4",            OpX2aZaZbX4X2b (8, 1, 1, 0, 1, 0), IA64_INSTR_FORMAT_A9, IA64_INSTR_PSUB},
    {"psub1.sss",        OpX2aZaZbX4X2b (8, 1, 0, 0, 1, 1), IA64_INSTR_FORMAT_A9, IA64_INSTR_PSUB},
    {"psub2.sss",        OpX2aZaZbX4X2b (8, 1, 0, 1, 1, 1), IA64_INSTR_FORMAT_A9, IA64_INSTR_PSUB},
    {"psub1.uuu",        OpX2aZaZbX4X2b (8, 1, 0, 0, 1, 2), IA64_INSTR_FORMAT_A9, IA64_INSTR_PSUB},
    {"psub2.uuu",        OpX2aZaZbX4X2b (8, 1, 0, 1, 1, 2), IA64_INSTR_FORMAT_A9, IA64_INSTR_PSUB},
    {"psub1.uus",        OpX2aZaZbX4X2b (8, 1, 0, 0, 1, 3), IA64_INSTR_FORMAT_A9, IA64_INSTR_PSUB},
    {"psub2.uus",        OpX2aZaZbX4X2b (8, 1, 0, 1, 1, 3), IA64_INSTR_FORMAT_A9, IA64_INSTR_PSUB},
    {"pavg1",            OpX2aZaZbX4X2b (8, 1, 0, 0, 2, 2), IA64_INSTR_FORMAT_A9, IA64_INSTR_PAVG},
    {"pavg2",            OpX2aZaZbX4X2b (8, 1, 0, 1, 2, 2), IA64_INSTR_FORMAT_A9, IA64_INSTR_PAVG},
    {"pavgsub1",         OpX2aZaZbX4X2b (8, 1, 0, 0, 3, 2), IA64_INSTR_FORMAT_A9, IA64_INSTR_PAVGSUB},
    {"pavgsub2",         OpX2aZaZbX4X2b (8, 1, 0, 1, 3, 2), IA64_INSTR_FORMAT_A9, IA64_INSTR_PAVGSUB},
    {"pcmp1.eq",         OpX2aZaZbX4X2b (8, 1, 0, 0, 9, 0), IA64_INSTR_FORMAT_A9, IA64_INSTR_PCMP},
    {"pcmp2.eq",         OpX2aZaZbX4X2b (8, 1, 0, 1, 9, 0), IA64_INSTR_FORMAT_A9, IA64_INSTR_PCMP},
    {"pcmp4.eq",         OpX2aZaZbX4X2b (8, 1, 1, 0, 9, 0), IA64_INSTR_FORMAT_A9, IA64_INSTR_PCMP},
    {"pcmp1.gt",         OpX2aZaZbX4X2b (8, 1, 0, 0, 9, 1), IA64_INSTR_FORMAT_A9, IA64_INSTR_PCMP},
    {"pcmp2.gt",         OpX2aZaZbX4X2b (8, 1, 0, 1, 9, 1), IA64_INSTR_FORMAT_A9, IA64_INSTR_PCMP},
    {"pcmp4.gt",         OpX2aZaZbX4X2b (8, 1, 1, 0, 9, 1), IA64_INSTR_FORMAT_A9, IA64_INSTR_PCMP},

    {"pshladd2",	 OpX2aZaZbX4 (8, 1, 0, 1, 4), IA64_INSTR_FORMAT_A10, IA64_INSTR_PSHLADD2},
    {"pshradd2",	 OpX2aZaZbX4 (8, 1, 0, 1, 6), IA64_INSTR_FORMAT_A10, IA64_INSTR_PSHRADD2},

    {"addl",		 Op    (9), IA64_INSTR_FORMAT_A5, IA64_INSTR_ADD},

    {"cmp.lt",            OpX2TbTaC (0xc, 0, 0, 0, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.lt.unc",        OpX2TbTaC (0xc, 0, 0, 0, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.eq.and",        OpX2TbTaC (0xc, 0, 0, 1, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.ne.and",        OpX2TbTaC (0xc, 0, 0, 1, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp4.lt",           OpX2TbTaC (0xc, 1, 0, 0, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.lt.unc",       OpX2TbTaC (0xc, 1, 0, 0, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.eq.and",       OpX2TbTaC (0xc, 1, 0, 1, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.ne.and",       OpX2TbTaC (0xc, 1, 0, 1, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp.gt.and",        OpX2TbTaC (0xc, 0, 1, 0, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp.le.and",        OpX2TbTaC (0xc, 0, 1, 0, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp.ge.and",        OpX2TbTaC (0xc, 0, 1, 1, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp.lt.and",        OpX2TbTaC (0xc, 0, 1, 1, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp4.gt.and",       OpX2TbTaC (0xc, 1, 1, 0, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp4.le.and",       OpX2TbTaC (0xc, 1, 1, 0, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp4.ge.and",       OpX2TbTaC (0xc, 1, 1, 1, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp4.lt.and",       OpX2TbTaC (0xc, 1, 1, 1, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp.lt",            OpX2TaC   (0xc, 2, 0, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp.lt.unc",        OpX2TaC   (0xc, 2, 0, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp.eq.and",        OpX2TaC   (0xc, 2, 1, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp.ne.and",        OpX2TaC   (0xc, 2, 1, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp4.lt",        OpX2TaC   (0xc, 3, 0, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp4.lt.unc",        OpX2TaC   (0xc, 3, 0, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp4.eq.and",       OpX2TaC   (0xc, 3, 1, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp4.ne.and",       OpX2TaC   (0xc, 3, 1, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp.ltu",           OpX2TbTaC (0xd, 0, 0, 0, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.ltu.unc",       OpX2TbTaC (0xd, 0, 0, 0, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.eq.or",         OpX2TbTaC (0xd, 0, 0, 1, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.ne.or",         OpX2TbTaC (0xd, 0, 0, 1, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.gt.or",        OpX2TbTaC (0xd, 0, 1, 0, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp.le.or",        OpX2TbTaC (0xd, 0, 1, 0, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp.ge.or",        OpX2TbTaC (0xd, 0, 1, 1, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp.lt.or",        OpX2TbTaC (0xd, 0, 1, 1, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp4.ltu",          OpX2TbTaC (0xd, 1, 0, 0, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.ltu.unc",      OpX2TbTaC (0xd, 1, 0, 0, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.eq.or",        OpX2TbTaC (0xd, 1, 0, 1, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.ne.or",        OpX2TbTaC (0xd, 1, 0, 1, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.gt.or",        OpX2TbTaC (0xd, 1, 1, 0, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp4.le.or",        OpX2TbTaC (0xd, 1, 1, 0, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp4.ge.or",        OpX2TbTaC (0xd, 1, 1, 1, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp4.lt.or",        OpX2TbTaC (0xd, 1, 1, 1, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp.ltu",           OpX2TaC   (0xd, 2, 0, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp.ltu.unc",       OpX2TaC   (0xd, 2, 0, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp.eq.or",         OpX2TaC   (0xd, 2, 1, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp.ne.or",         OpX2TaC   (0xd, 2, 1, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp4.ltu",          OpX2TaC   (0xd, 3, 0, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp4.ltu.unc",      OpX2TaC   (0xd, 3, 0, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp4.eq.or",        OpX2TaC   (0xd, 3, 1, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp4.ne.or",        OpX2TaC   (0xd, 3, 1, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp.eq",            OpX2TbTaC (0xe, 0, 0, 0, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.eq.unc",        OpX2TbTaC (0xe, 0, 0, 0, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.eq.or.andcm",   OpX2TbTaC (0xe, 0, 0, 1, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp.ne.or.andcm",   OpX2TbTaC (0xe, 0, 0, 1, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP},
    {"cmp4.eq",           OpX2TbTaC (0xe, 1, 0, 0, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.eq.unc",       OpX2TbTaC (0xe, 1, 0, 0, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.eq.or.andcm",  OpX2TbTaC (0xe, 1, 0, 1, 0), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp4.ne.or.andcm",  OpX2TbTaC (0xe, 1, 0, 1, 1), IA64_INSTR_FORMAT_A6, IA64_INSTR_CMP4},
    {"cmp.gt.or.andcm",   OpX2TbTaC (0xe, 0, 1, 0, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp.le.or.andcm",   OpX2TbTaC (0xe, 0, 1, 0, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp.ge.or.andcm",   OpX2TbTaC (0xe, 0, 1, 1, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp.lt.or.andcm",   OpX2TbTaC (0xe, 0, 1, 1, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP},
    {"cmp4.gt.or.andcm",  OpX2TbTaC (0xe, 1, 1, 0, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp4.le.or.andcm",  OpX2TbTaC (0xe, 1, 1, 0, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp4.ge.or.andcm",  OpX2TbTaC (0xe, 1, 1, 1, 0), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp4.lt.or.andcm",   OpX2TbTaC (0xe, 1, 1, 1, 1), IA64_INSTR_FORMAT_A7, IA64_INSTR_CMP4},
    {"cmp.eq",            OpX2TaC   (0xe, 2, 0, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp.eq.unc",        OpX2TaC   (0xe, 2, 0, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp.eq.or.andcm",   OpX2TaC   (0xe, 2, 1, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp.ne.or.andcm",   OpX2TaC   (0xe, 2, 1, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP},
    {"cmp4.eq",           OpX2TaC   (0xe, 3, 0, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp4.eq.unc",       OpX2TaC   (0xe, 3, 0, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp4.eq.or.andcm",  OpX2TaC   (0xe, 3, 1, 0), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},
    {"cmp4.ne.or.andcm",  OpX2TaC   (0xe, 3, 1, 1), IA64_INSTR_FORMAT_A8, IA64_INSTR_CMP4},

    {"Unsupported", 0x0ULL, 0x0ULL, IA64_INSTR_FORMAT_UNSUPPORTED, IA64_INSTR_UNSUPPORTED}
  };





#undef bC
#undef bImm14
#undef bR3a
#undef bR3b
#undef bTa
#undef bTb
#undef bVe
#undef bX
#undef bX2
#undef bX2a
#undef bX2b
#undef bX4
#undef bZa
#undef bZb
#undef mC
#undef mImm14
#undef mR3a
#undef mR3b
#undef mTa
#undef mTb
#undef mVe
#undef mX
#undef mX2
#undef mX2a
#undef mX2b
#undef mX4
#undef mZa
#undef mZb
#undef OpR3a
#undef OpR3b
#undef OpX2aVe
#undef OpX2aVeImm14
#undef OpX2aVeX4
#undef OpX2aVeX4X2b
#undef OpX2TbTaC
#undef OpX2TaC
#undef OpX2aZaZbX4
#undef OpX2aZaZbX4X2b

/* vim: set shiftwidth=2: */
