/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_opcodes_f.h"
#include "../../../utils/util_error.h"

#include "../ia64_instructions.h"
#include "../ia64_disassemble/ia64_disassemble_one.h"
#include "../../../simple_types.h"

/* instruction bit fields: */

#define bF2(x)	(((t_uint64) ((x) & 0x7f)) << 13)
#define bF4(x)	(((t_uint64) ((x) & 0x7f)) << 27)
#define bQ(x)	(((t_uint64) ((x) & 0x1)) << 36)
#define bRa(x)	(((t_uint64) ((x) & 0x1)) << 33)
#define bRb(x)	(((t_uint64) ((x) & 0x1)) << 36)
#define bSf(x)	(((t_uint64) ((x) & 0x3)) << 34)
#define bTa(x)	(((t_uint64) ((x) & 0x1)) << 12)
#define bXa(x)	(((t_uint64) ((x) & 0x1)) << 36)
#define bXb(x)	(((t_uint64) ((x) & 0x1)) << 33)
#define bX2(x)	(((t_uint64) ((x) & 0x3)) << 34)
#define bX6_(x)	(((t_uint64) ((x) & 0x3f)) << 27)

/* instruction bit masks: */

#define mF2	bF2 (-1)
#define mF4	bF4 (-1)
#define mQ	bQ (-1)
#define mRa	bRa (-1)
#define mRb	bRb (-1)
#define mSf	bSf (-1)
#define mTa	bTa (-1)
#define mXa	bXa (-1)
#define mXb	bXb (-1)
#define mX2	bX2 (-1)
#define mX6	bX6_ (-1)

#define OpXa(a,b)	(bOp (a) | bXa (b)), (mOp | mXa)
#define OpXaSf(a,b,c)	(bOp (a) | bXa (b) | bSf (c)), (mOp | mXa | mSf)
#define OpXaSfF2(a,b,c,d) \
	(bOp (a) | bXa (b) | bSf (c) | bF2 (d)), (mOp | mXa | mSf | mF2)
#define OpXaSfF4(a,b,c,d) \
	(bOp (a) | bXa (b) | bSf (c) | bF4 (d)), (mOp | mXa | mSf | mF4)
#define OpXaSfF2F4(a,b,c,d,e) \
	(bOp (a) | bXa (b) | bSf (c) | bF2 (d) | bF4 (e)), \
	(mOp | mXa | mSf | mF2 | mF4)
#define OpXaX2(a,b,c)	(bOp (a) | bXa (b) | bX2 (c)), (mOp | mXa | mX2)
#define OpXaX2F2(a,b,c,d) \
	(bOp (a) | bXa (b) | bX2 (c) | bF2 (d)), (mOp | mXa | mX2 | mF2)
#define OpRaRbTaSf(a,b,c,d,e) \
	(bOp (a) | bRa (b) | bRb (c) | bTa (d) | bSf (e)), \
	(mOp | mRa | mRb | mTa | mSf)
#define OpTa(a,b)	(bOp (a) | bTa (b)), (mOp | mTa)
#define OpXbQSf(a,b,c,d) \
	(bOp (a) | bXb (b) | bQ (c) | bSf (d)), (mOp | mXb | mQ | mSf)
#define OpXbX6(a,b,c) \
	(bOp (a) | bXb (b) | bX6_ (c)), (mOp | mXb | mX6)
#define OpXbX6F2(a,b,c,d) \
struct ia64 bXb (b) | bX6_ (c) | bF2 (d)), (mOp | mXb | mX6 | mF2)
#define OpXbX6Sf(a,b,c,d) \
	(bOp (a) | bXb (b) | bX6_ (c) | bSf (d)), (mOp | mXb | mX6 | mSf)

const t_ia64_opcode ia64_opcode_table_f[] =
  {
    /* F-type instruction encodings (sorted according to major
opcode) */

    {"frcpa.s0",        OpXbQSf (0, 1, 0, 0), IA64_INSTR_FORMAT_F6, IA64_INSTR_FRCPA},
    {"frcpa.s1",        OpXbQSf (0, 1, 0, 1), IA64_INSTR_FORMAT_F6, IA64_INSTR_FRCPA},
    {"frcpa.s2",        OpXbQSf (0, 1, 0, 2), IA64_INSTR_FORMAT_F6, IA64_INSTR_FRCPA},
    {"frcpa.s3",        OpXbQSf (0, 1, 0, 3), IA64_INSTR_FORMAT_F6, IA64_INSTR_FRCPA},

    {"frsqrta.s0",      OpXbQSf (0, 1, 1, 0), IA64_INSTR_FORMAT_F7, IA64_INSTR_FRSQRTA},
    {"frsqrta.s1",      OpXbQSf (0, 1, 1, 1), IA64_INSTR_FORMAT_F7, IA64_INSTR_FRSQRTA},
    {"frsqrta.s2",      OpXbQSf (0, 1, 1, 2), IA64_INSTR_FORMAT_F7, IA64_INSTR_FRSQRTA},
    {"frsqrta.s3",      OpXbQSf (0, 1, 1, 3), IA64_INSTR_FORMAT_F7, IA64_INSTR_FRSQRTA},

    {"fmin.s0",         OpXbX6Sf (0, 0, 0x14, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FMIN},
    {"fmin.s1",         OpXbX6Sf (0, 0, 0x14, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FMIN},
    {"fmin.s2",         OpXbX6Sf (0, 0, 0x14, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FMIN},
    {"fmin.s3",         OpXbX6Sf (0, 0, 0x14, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FMIN},
    {"fmax.s0",         OpXbX6Sf (0, 0, 0x15, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FMAX},
    {"fmax.s1",         OpXbX6Sf (0, 0, 0x15, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FMAX},
    {"fmax.s2",         OpXbX6Sf (0, 0, 0x15, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FMAX},
    {"fmax.s3",         OpXbX6Sf (0, 0, 0x15, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FMAX},
    {"famin.s0",        OpXbX6Sf (0, 0, 0x16, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FAMIN},
    {"famin.s1",        OpXbX6Sf (0, 0, 0x16, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FAMIN},
    {"famin.s2",        OpXbX6Sf (0, 0, 0x16, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FAMIN},
    {"famin.s3",        OpXbX6Sf (0, 0, 0x16, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FAMIN},
    {"famax.s0",        OpXbX6Sf (0, 0, 0x17, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FAMAX},
    {"famax.s1",        OpXbX6Sf (0, 0, 0x17, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FAMAX},
    {"famax.s2",        OpXbX6Sf (0, 0, 0x17, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FAMAX},
    {"famax.s3",        OpXbX6Sf (0, 0, 0x17, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FAMAX},

    {"fmerge.s",        OpXbX6   (0, 0, 0x10), IA64_INSTR_FORMAT_F9, IA64_INSTR_FMERGE},
    {"fmerge.ns",       OpXbX6   (0, 0, 0x11), IA64_INSTR_FORMAT_F9, IA64_INSTR_FMERGE},

    {"fmerge.se",       OpXbX6 (0, 0, 0x12), IA64_INSTR_FORMAT_F9, IA64_INSTR_FMERGE},
    {"fmix.lr",         OpXbX6 (0, 0, 0x39), IA64_INSTR_FORMAT_F9, IA64_INSTR_FMIX},
    {"fmix.r",          OpXbX6 (0, 0, 0x3a), IA64_INSTR_FORMAT_F9, IA64_INSTR_FMIX},
    {"fmix.l",          OpXbX6 (0, 0, 0x3b), IA64_INSTR_FORMAT_F9, IA64_INSTR_FMIX},
    {"fsxt.r",          OpXbX6 (0, 0, 0x3c), IA64_INSTR_FORMAT_F9, IA64_INSTR_FSXT},
    {"fsxt.l",          OpXbX6 (0, 0, 0x3d), IA64_INSTR_FORMAT_F9, IA64_INSTR_FSXT},
    {"fpack",           OpXbX6 (0, 0, 0x28), IA64_INSTR_FORMAT_F9, IA64_INSTR_FPACK},
    {"fswap",           OpXbX6 (0, 0, 0x34), IA64_INSTR_FORMAT_F9, IA64_INSTR_FSWAP},
    {"fswap.nl",        OpXbX6 (0, 0, 0x35), IA64_INSTR_FORMAT_F9, IA64_INSTR_FSWAP},
    {"fswap.nr",        OpXbX6 (0, 0, 0x36), IA64_INSTR_FORMAT_F9, IA64_INSTR_FSWAP},
    {"fand",            OpXbX6 (0, 0, 0x2c), IA64_INSTR_FORMAT_F9, IA64_INSTR_FAND},
    {"fandcm",          OpXbX6 (0, 0, 0x2d), IA64_INSTR_FORMAT_F9, IA64_INSTR_FANDCM},
    {"for",             OpXbX6 (0, 0, 0x2e), IA64_INSTR_FORMAT_F9, IA64_INSTR_FOR},
    {"fxor",            OpXbX6 (0, 0, 0x2f), IA64_INSTR_FORMAT_F9, IA64_INSTR_FXOR},

    {"fcvt.fx.s0",              OpXbX6Sf (0, 0, 0x18, 0), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fx.s1",              OpXbX6Sf (0, 0, 0x18, 1), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fx.s2",              OpXbX6Sf (0, 0, 0x18, 2), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fx.s3",              OpXbX6Sf (0, 0, 0x18, 3), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fxu.s0",             OpXbX6Sf (0, 0, 0x19, 0), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fxu.s1",             OpXbX6Sf (0, 0, 0x19, 1), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fxu.s2",             OpXbX6Sf (0, 0, 0x19, 2), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fxu.s3",             OpXbX6Sf (0, 0, 0x19, 3), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fx.trunc.s0",        OpXbX6Sf (0, 0, 0x1a, 0), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fx.trunc.s1",        OpXbX6Sf (0, 0, 0x1a, 1), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fx.trunc.s2",        OpXbX6Sf (0, 0, 0x1a, 2), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fx.trunc.s3",        OpXbX6Sf (0, 0, 0x1a, 3), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fxu.trunc.s0",       OpXbX6Sf (0, 0, 0x1b, 0), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fxu.trunc.s1",       OpXbX6Sf (0, 0, 0x1b, 1), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fxu.trunc.s2",       OpXbX6Sf (0, 0, 0x1b, 2), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},
    {"fcvt.fxu.trunc.s3",       OpXbX6Sf (0, 0, 0x1b, 3), IA64_INSTR_FORMAT_F10, IA64_INSTR_FCVT_FX},

    {"fcvt.xf",         OpXbX6 (0, 0, 0x1c), IA64_INSTR_FORMAT_F11, IA64_INSTR_FCVT_XF},

    {"fsetc.s0",        OpXbX6Sf (0, 0, 0x04, 0), IA64_INSTR_FORMAT_F12, IA64_INSTR_FSETC},
    {"fsetc.s1",        OpXbX6Sf (0, 0, 0x04, 1), IA64_INSTR_FORMAT_F12, IA64_INSTR_FSETC},
    {"fsetc.s2",        OpXbX6Sf (0, 0, 0x04, 2), IA64_INSTR_FORMAT_F12, IA64_INSTR_FSETC},
    {"fsetc.s3",        OpXbX6Sf (0, 0, 0x04, 3), IA64_INSTR_FORMAT_F12, IA64_INSTR_FSETC},
    {"fclrf.s0",        OpXbX6Sf (0, 0, 0x05, 0), IA64_INSTR_FORMAT_F13, IA64_INSTR_FCLRF},
    {"fclrf.s1",        OpXbX6Sf (0, 0, 0x05, 1), IA64_INSTR_FORMAT_F13, IA64_INSTR_FCLRF},
    {"fclrf.s2",        OpXbX6Sf (0, 0, 0x05, 2), IA64_INSTR_FORMAT_F13, IA64_INSTR_FCLRF},
    {"fclrf.s3",        OpXbX6Sf (0, 0, 0x05, 3), IA64_INSTR_FORMAT_F13, IA64_INSTR_FCLRF},
    {"fchkf.s0",        OpXbX6Sf (0, 0, 0x08, 0), IA64_INSTR_FORMAT_F14, IA64_INSTR_FCHKF},
    {"fchkf.s1",        OpXbX6Sf (0, 0, 0x08, 1), IA64_INSTR_FORMAT_F14, IA64_INSTR_FCHKF},
    {"fchkf.s2",        OpXbX6Sf (0, 0, 0x08, 2), IA64_INSTR_FORMAT_F14, IA64_INSTR_FCHKF},
    {"fchkf.s3",        OpXbX6Sf (0, 0, 0x08, 3), IA64_INSTR_FORMAT_F14, IA64_INSTR_FCHKF},

    {"break.f",         OpXbX6 (0, 0, 0x00), IA64_INSTR_FORMAT_F15, IA64_INSTR_BREAK},
    {"nop.f",           OpXbX6 (0, 0, 0x01), IA64_INSTR_FORMAT_F15, IA64_INSTR_NOP},

    {"fprcpa.s0",       OpXbQSf (1, 1, 0, 0), IA64_INSTR_FORMAT_F6, IA64_INSTR_FPRCPA},
    {"fprcpa",          OpXbQSf (1, 1, 0, 0), IA64_INSTR_FORMAT_F6, IA64_INSTR_FPRCPA},
    {"fprcpa.s1",       OpXbQSf (1, 1, 0, 1), IA64_INSTR_FORMAT_F6, IA64_INSTR_FPRCPA},
    {"fprcpa.s2",       OpXbQSf (1, 1, 0, 2), IA64_INSTR_FORMAT_F6, IA64_INSTR_FPRCPA},
    {"fprcpa.s3",       OpXbQSf (1, 1, 0, 3), IA64_INSTR_FORMAT_F6, IA64_INSTR_FPRCPA},

    {"fprsqrta.s0",     OpXbQSf (1, 1, 1, 0), IA64_INSTR_FORMAT_F7, IA64_INSTR_FPRSQRTA},
    {"fprsqrta",        OpXbQSf (1, 1, 1, 0), IA64_INSTR_FORMAT_F7, IA64_INSTR_FPRSQRTA},
    {"fprsqrta.s1",     OpXbQSf (1, 1, 1, 1), IA64_INSTR_FORMAT_F7, IA64_INSTR_FPRSQRTA},
    {"fprsqrta.s2",     OpXbQSf (1, 1, 1, 2), IA64_INSTR_FORMAT_F7, IA64_INSTR_FPRSQRTA},
    {"fprsqrta.s3",     OpXbQSf (1, 1, 1, 3), IA64_INSTR_FORMAT_F7, IA64_INSTR_FPRSQRTA},

    {"fpmin.s0",        OpXbX6Sf (1, 0, 0x14, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPMIN},
    {"fpmin.s1",        OpXbX6Sf (1, 0, 0x14, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPMIN},
    {"fpmin.s2",        OpXbX6Sf (1, 0, 0x14, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPMIN},
    {"fpmin.s3",        OpXbX6Sf (1, 0, 0x14, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPMIN},
    {"fpmax.s0",        OpXbX6Sf (1, 0, 0x15, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPMAX},
    {"fpmax.s1",        OpXbX6Sf (1, 0, 0x15, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPMAX},
    {"fpmax.s2",        OpXbX6Sf (1, 0, 0x15, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPMAX},
    {"fpmax.s3",        OpXbX6Sf (1, 0, 0x15, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPMAX},
    {"fpamin.s0",       OpXbX6Sf (1, 0, 0x16, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPAMIN},
    {"fpamin.s1",       OpXbX6Sf (1, 0, 0x16, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPAMIN},
    {"fpamin.s2",       OpXbX6Sf (1, 0, 0x16, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPAMIN},
    {"fpamin.s3",       OpXbX6Sf (1, 0, 0x16, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPAMIN},
    {"fpamax.s0",       OpXbX6Sf (1, 0, 0x17, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPAMAX},
    {"fpamax.s1",       OpXbX6Sf (1, 0, 0x17, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPAMAX},
    {"fpamax.s2",       OpXbX6Sf (1, 0, 0x17, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPAMAX},
    {"fpamax.s3",       OpXbX6Sf (1, 0, 0x17, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPAMAX},

    {"fpcmp.eq.s0",     OpXbX6Sf (1, 0, 0x30, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.eq.s1",     OpXbX6Sf (1, 0, 0x30, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.eq.s2",     OpXbX6Sf (1, 0, 0x30, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.eq.s3",     OpXbX6Sf (1, 0, 0x30, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.lt.s0",     OpXbX6Sf (1, 0, 0x31, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.lt.s1",     OpXbX6Sf (1, 0, 0x31, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.lt.s2",     OpXbX6Sf (1, 0, 0x31, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.lt.s3",     OpXbX6Sf (1, 0, 0x31, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.le.s0",     OpXbX6Sf (1, 0, 0x32, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.le.s1",     OpXbX6Sf (1, 0, 0x32, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.le.s2",     OpXbX6Sf (1, 0, 0x32, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.le.s3",     OpXbX6Sf (1, 0, 0x32, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.unord.s0",  OpXbX6Sf (1, 0, 0x33, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.unord.s1",  OpXbX6Sf (1, 0, 0x33, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.unord.s2",  OpXbX6Sf (1, 0, 0x33, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.unord.s3",  OpXbX6Sf (1, 0, 0x33, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.neq.s0",    OpXbX6Sf (1, 0, 0x34, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.neq.s1",    OpXbX6Sf (1, 0, 0x34, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.neq.s2",    OpXbX6Sf (1, 0, 0x34, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.neq.s3",    OpXbX6Sf (1, 0, 0x34, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.nlt.s0",    OpXbX6Sf (1, 0, 0x35, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.nlt.s1",    OpXbX6Sf (1, 0, 0x35, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.nlt.s2",    OpXbX6Sf (1, 0, 0x35, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.nlt.s3",    OpXbX6Sf (1, 0, 0x35, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.nle.s0",    OpXbX6Sf (1, 0, 0x36, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.nle.s1",    OpXbX6Sf (1, 0, 0x36, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.nle.s2",    OpXbX6Sf (1, 0, 0x36, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.nle.s3",    OpXbX6Sf (1, 0, 0x36, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.ord.s0",    OpXbX6Sf (1, 0, 0x37, 0), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.ord.s1",    OpXbX6Sf (1, 0, 0x37, 1), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.ord.s2",    OpXbX6Sf (1, 0, 0x37, 2), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},
    {"fpcmp.ord.s3",    OpXbX6Sf (1, 0, 0x37, 3), IA64_INSTR_FORMAT_F8, IA64_INSTR_FPCMP},

    {"fpmerge.s",       OpXbX6   (1, 0, 0x10), IA64_INSTR_FORMAT_F9, IA64_INSTR_FPMERGE},
    {"fpmerge.ns",      OpXbX6   (1, 0, 0x11), IA64_INSTR_FORMAT_F9, IA64_INSTR_FPMERGE},
    {"fpmerge.se",      OpXbX6 (1, 0, 0x12), IA64_INSTR_FORMAT_F9, IA64_INSTR_FPMERGE},

    {"fpcvt.fx.s0",             OpXbX6Sf (1, 0, 0x18, 0), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fx.s1",             OpXbX6Sf (1, 0, 0x18, 1), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fx.s2",             OpXbX6Sf (1, 0, 0x18, 2), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fx.s3",             OpXbX6Sf (1, 0, 0x18, 3), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fxu.s0",            OpXbX6Sf (1, 0, 0x19, 0), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fxu.s1",            OpXbX6Sf (1, 0, 0x19, 1), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fxu.s2",            OpXbX6Sf (1, 0, 0x19, 2), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fxu.s3",            OpXbX6Sf (1, 0, 0x19, 3), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fx.trunc.s0",       OpXbX6Sf (1, 0, 0x1a, 0), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fx.trunc.s1",       OpXbX6Sf (1, 0, 0x1a, 1), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fx.trunc.s2",       OpXbX6Sf (1, 0, 0x1a, 2), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fx.trunc.s3",       OpXbX6Sf (1, 0, 0x1a, 3), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fxu.trunc.s0",      OpXbX6Sf (1, 0, 0x1b, 0), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fxu.trunc.s1",      OpXbX6Sf (1, 0, 0x1b, 1), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fxu.trunc.s2",      OpXbX6Sf (1, 0, 0x1b, 2), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},
    {"fpcvt.fxu.trunc.s3",      OpXbX6Sf (1, 0, 0x1b, 3), IA64_INSTR_FORMAT_F10, IA64_INSTR_FPCVT_FX},

    {"fcmp.eq.s0",        OpRaRbTaSf (4, 0, 0, 0, 0), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.eq.s1",        OpRaRbTaSf (4, 0, 0, 0, 1), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.eq.s2",        OpRaRbTaSf (4, 0, 0, 0, 2), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.eq.s3",        OpRaRbTaSf (4, 0, 0, 0, 3), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.lt.s0",        OpRaRbTaSf (4, 0, 1, 0, 0), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.lt.s1",        OpRaRbTaSf (4, 0, 1, 0, 1), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.lt.s2",        OpRaRbTaSf (4, 0, 1, 0, 2), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.lt.s3",        OpRaRbTaSf (4, 0, 1, 0, 3), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.le.s0",        OpRaRbTaSf (4, 1, 0, 0, 0), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.le.s1",        OpRaRbTaSf (4, 1, 0, 0, 1), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.le.s2",        OpRaRbTaSf (4, 1, 0, 0, 2), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.le.s3",        OpRaRbTaSf (4, 1, 0, 0, 3), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.unord.s0",     OpRaRbTaSf (4, 1, 1, 0, 0), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.unord.s1",     OpRaRbTaSf (4, 1, 1, 0, 1), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.unord.s2",     OpRaRbTaSf (4, 1, 1, 0, 2), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.unord.s3",     OpRaRbTaSf (4, 1, 1, 0, 3), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.eq.unc.s0",    OpRaRbTaSf (4, 0, 0, 1, 0), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.eq.unc.s1",    OpRaRbTaSf (4, 0, 0, 1, 1), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.eq.unc.s2",    OpRaRbTaSf (4, 0, 0, 1, 2), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.eq.unc.s3",    OpRaRbTaSf (4, 0, 0, 1, 3), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.lt.unc.s0",    OpRaRbTaSf (4, 0, 1, 1, 0), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.lt.unc.s1",    OpRaRbTaSf (4, 0, 1, 1, 1), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.lt.unc.s2",    OpRaRbTaSf (4, 0, 1, 1, 2), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.lt.unc.s3",    OpRaRbTaSf (4, 0, 1, 1, 3), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.le.unc.s0",    OpRaRbTaSf (4, 1, 0, 1, 0), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.le.unc.s1",    OpRaRbTaSf (4, 1, 0, 1, 1), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.le.unc.s2",    OpRaRbTaSf (4, 1, 0, 1, 2), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.le.unc.s3",    OpRaRbTaSf (4, 1, 0, 1, 3), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.unord.unc.s0", OpRaRbTaSf (4, 1, 1, 1, 0), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.unord.unc.s1", OpRaRbTaSf (4, 1, 1, 1, 1), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.unord.unc.s2", OpRaRbTaSf (4, 1, 1, 1, 2), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},
    {"fcmp.unord.unc.s3", OpRaRbTaSf (4, 1, 1, 1, 3), IA64_INSTR_FORMAT_F4, IA64_INSTR_FCMP},

    {"fclass.m",        OpTa (5, 0), IA64_INSTR_FORMAT_F5, IA64_INSTR_FCLASS},
    {"fclass.m.unc",    OpTa (5, 1), IA64_INSTR_FORMAT_F5, IA64_INSTR_FCLASS},

    /* note: fnorm and fcvt.xuf have identical encodings! */
    {"fma.s0",          OpXaSf (0x8, 0, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.s1",          OpXaSf (0x8, 0, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.s2",          OpXaSf (0x8, 0, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.s3",          OpXaSf (0x8, 0, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.s.s0",        OpXaSf (0x8, 1, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.s.s1",        OpXaSf (0x8, 1, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.s.s2",        OpXaSf (0x8, 1, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.s.s3",        OpXaSf (0x8, 1, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},

    {"fma.d.s0",        OpXaSf (0x9, 0, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.d.s1",        OpXaSf (0x9, 0, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.d.s2",        OpXaSf (0x9, 0, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},
    {"fma.d.s3",        OpXaSf (0x9, 0, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMA},

    {"fpma.s0",         OpXaSf   (0x9, 1, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPMA},
    {"fpma.s1",         OpXaSf   (0x9, 1, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPMA},
    {"fpma.s2",         OpXaSf   (0x9, 1, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPMA},
    {"fpma.s3",         OpXaSf   (0x9, 1, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPMA},

   {"fms.s0",          OpXaSf   (0xa, 0, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.s1",          OpXaSf   (0xa, 0, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.s2",          OpXaSf   (0xa, 0, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.s3",          OpXaSf   (0xa, 0, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.s.s0",        OpXaSf   (0xa, 1, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.s.s1",        OpXaSf   (0xa, 1, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.s.s2",        OpXaSf   (0xa, 1, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.s.s3",        OpXaSf   (0xa, 1, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.d.s0",        OpXaSf   (0xb, 0, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.d.s1",        OpXaSf   (0xb, 0, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.d.s2",        OpXaSf   (0xb, 0, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},
    {"fms.d.s3",        OpXaSf   (0xb, 0, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FMS},

    {"fpms.s0",         OpXaSf (0xb, 1, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPMS},
    {"fpms.s1",         OpXaSf (0xb, 1, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPMS},
    {"fpms.s2",         OpXaSf (0xb, 1, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPMS},
    {"fpms.s3",         OpXaSf (0xb, 1, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPMS},

    {"fnma.s0",         OpXaSf (0xc, 0, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.s1",         OpXaSf (0xc, 0, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.s2",         OpXaSf (0xc, 0, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.s3",         OpXaSf (0xc, 0, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.s.s0",       OpXaSf (0xc, 1, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.s.s1",       OpXaSf (0xc, 1, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.s.s2",       OpXaSf (0xc, 1, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.s.s3",       OpXaSf (0xc, 1, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.d.s0",       OpXaSf (0xd, 0, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.d.s1",       OpXaSf (0xd, 0, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.d.s2",       OpXaSf (0xd, 0, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},
    {"fnma.d.s3",       OpXaSf (0xd, 0, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FNMA},

    {"fpnma.s0",        OpXaSf   (0xd, 1, 0), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPNMA},
    {"fpnma.s1",        OpXaSf   (0xd, 1, 1), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPNMA},
    {"fpnma.s2",        OpXaSf   (0xd, 1, 2), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPNMA},
    {"fpnma.s3",        OpXaSf   (0xd, 1, 3), IA64_INSTR_FORMAT_F1, IA64_INSTR_FPNMA},


    {"xma.l",           OpXaX2 (0xe, 1, 0), IA64_INSTR_FORMAT_F2, IA64_INSTR_XMA},
    {"xma.h",           OpXaX2 (0xe, 1, 3), IA64_INSTR_FORMAT_F2, IA64_INSTR_XMA},
    {"xma.hu",          OpXaX2 (0xe, 1, 2), IA64_INSTR_FORMAT_F2, IA64_INSTR_XMA},

    {"fselect",         OpXa (0xe, 0), IA64_INSTR_FORMAT_F3, IA64_INSTR_FSELECT},

    {"Unsupported", 0x0ULL, 0x0ULL, IA64_INSTR_FORMAT_UNSUPPORTED, IA64_INSTR_UNSUPPORTED}
  };

#undef bF2
#undef bF4
#undef bQ
#undef bRa
#undef bRb
#undef bSf
#undef bTa
#undef bXa
#undef bXb
#undef bX2
#undef bX6_
#undef mF2
#undef mF4
#undef mQ
#undef mRa
#undef mRb
#undef mSf
#undef mTa
#undef mXa
#undef mXb
#undef mX2
#undef mX6
#undef OpXa
#undef OpXaSf
#undef OpXaSfF2
#undef OpXaSfF4
#undef OpXaSfF2F4
#undef OpXaX2
#undef OpRaRbTaSf
#undef OpTa
#undef OpXbQSf
#undef OpXbX6
#undef OpXbX6F2
#undef OpXbX6Sf

/* vim: set shiftwidth=2: */
