/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_opcodes_m.h"
#include "../../../utils/util_error.h"

#include "../ia64_instructions.h"
#include "../ia64_disassemble/ia64_disassemble_one.h"
#include "../../../simple_types.h"

/* instruction bit fields: */
#define bM(x)		(((t_uint64) ((x) & 0x1)) << 36)
#define bX(x)		(((t_uint64) ((x) & 0x1)) << 27)
#define bX2(x)		(((t_uint64) ((x) & 0x3)) << 31)
#define bX3(x)		(((t_uint64) ((x) & 0x7)) << 33)
#define bX4_(x)		(((t_uint64) ((x) & 0xf)) << 27)
#define bX6a(x)		(((t_uint64) ((x) & 0x3f)) << 30)
#define bX6b(x)		(((t_uint64) ((x) & 0x3f)) << 27)
#define bHint(x)	(((t_uint64) ((x) & 0x3)) << 28)

#define mM	bM (-1)
#define mX	bX (-1)
#define mX2	bX2 (-1)
#define mX3	bX3 (-1)
#define mX4	bX4_ (-1)
#define mX6a	bX6a (-1)
#define mX6b	bX6b (-1)
#define mHint	bHint (-1)

#define OpX3(a,b) 		(bOp (a) | bX3 (b)), (mOp | mX3)
#define OpX3X6b(a,b,c) 		(bOp (a) | bX3 (b) | bX6b (c)), \
				(mOp | mX3 | mX6b)
#define OpX3X4(a,b,c)	 	(bOp (a) | bX3 (b) | bX4_ (c)), \
				(mOp | mX3 | mX4)
#define OpX3X4X2(a,b,c,d) 	(bOp (a) | bX3 (b) | bX4_ (c) | bX2 (d)), \
				(mOp | mX3 | mX4 | mX2)
#define OpX6aHint(a,b,c) 	(bOp (a) | bX6a (b) | bHint (c)), \
				(mOp | mX6a | mHint)
#define OpXX6aHint(a,b,c,d) 	(bOp (a) | bX (b) | bX6a (c) | bHint (d)), \
				(mOp | mX | mX6a | mHint)
#define OpMXX6a(a,b,c,d) \
	(bOp (a) | bM (b) | bX (c) | bX6a (d)), (mOp | mM | mX | mX6a)
#define OpMXX6aHint(a,b,c,d,e) \
	(bOp (a) | bM (b) | bX (c) | bX6a (d) | bHint (e)), \
	(mOp | mM | mX | mX6a | mHint)

const t_ia64_opcode ia64_opcode_table_m[] =
  {
    /* M-type instruction encodings (sorted according to major 
opcode) */

    {"chk.a.nc",	OpX3 (0, 4), IA64_INSTR_FORMAT_M22, IA64_INSTR_CHK},
    {"chk.a.clr",	OpX3 (0, 5), IA64_INSTR_FORMAT_M22, IA64_INSTR_CHK},
    
    {"chk.a.nc",	OpX3 (0, 6), IA64_INSTR_FORMAT_M23, IA64_INSTR_CHK},
    {"chk.a.clr",	OpX3 (0, 7), IA64_INSTR_FORMAT_M23, IA64_INSTR_CHK},

    {"invala",		OpX3X4X2 (0, 0, 0, 1), IA64_INSTR_FORMAT_M24, IA64_INSTR_INVALA},
    {"fwb",		OpX3X4X2 (0, 0, 0, 2), IA64_INSTR_FORMAT_M24, IA64_INSTR_FWB},
    {"mf",		OpX3X4X2 (0, 0, 2, 2), IA64_INSTR_FORMAT_M24, IA64_INSTR_MF},
    {"mf.a",		OpX3X4X2 (0, 0, 3, 2), IA64_INSTR_FORMAT_M24, IA64_INSTR_MF},
    {"srlz.d",		OpX3X4X2 (0, 0, 0, 3), IA64_INSTR_FORMAT_M24, IA64_INSTR_SRLZ},
    {"srlz.i",		OpX3X4X2 (0, 0, 1, 3), IA64_INSTR_FORMAT_M24, IA64_INSTR_SRLZ},
    {"sync.i",		OpX3X4X2 (0, 0, 3, 3), IA64_INSTR_FORMAT_M24, IA64_INSTR_SYNC},
    
    {"flushrs",		OpX3X4X2 (0, 0, 0xc, 0), IA64_INSTR_FORMAT_M25, IA64_INSTR_FLUSHRS,
      IA64_OPCODE_FIRST | IA64_OPCODE_NO_PRED},
    {"loadrs",		OpX3X4X2 (0, 0, 0xa, 0), IA64_INSTR_FORMAT_M25, IA64_INSTR_LOADRS,
      IA64_OPCODE_FIRST | IA64_OPCODE_NO_PRED},
    
    {"invala.e",	OpX3X4X2 (0, 0, 2, 1), IA64_INSTR_FORMAT_M26, IA64_INSTR_INVALA},
    
    {"invala.e",	OpX3X4X2 (0, 0, 3, 1), IA64_INSTR_FORMAT_M27, IA64_INSTR_INVALA},
    
    {"mov.m",		OpX3X4X2 (0, 0, 8, 2), IA64_INSTR_FORMAT_M30, IA64_INSTR_MOV_AR},

    {"break.m",		OpX3X4X2 (0, 0, 0, 0), IA64_INSTR_FORMAT_M37, IA64_INSTR_BREAK},
    {"nop.m",		OpX3X4X2 (0, 0, 1, 0), IA64_INSTR_FORMAT_M37, IA64_INSTR_NOP},

    {"sum",		OpX3X4 (0, 0, 4), IA64_INSTR_FORMAT_M44, IA64_INSTR_SUM},
    {"rum",		OpX3X4 (0, 0, 5), IA64_INSTR_FORMAT_M44, IA64_INSTR_RUM},
    {"ssm",		OpX3X4 (0, 0, 6), IA64_INSTR_FORMAT_M44, IA64_INSTR_SSM, 
      IA64_OPCODE_PRIV},
    {"rsm",		OpX3X4 (0, 0, 7), IA64_INSTR_FORMAT_M44, IA64_INSTR_RSM, 
      IA64_OPCODE_PRIV},

    {"mov.m",	OpX3X6b (1, 0, 0x2a), IA64_INSTR_FORMAT_M29, IA64_INSTR_MOV_AR},
    
    {"mov.m",	OpX3X6b (1, 0, 0x22), IA64_INSTR_FORMAT_M31, IA64_INSTR_MOV_AR},
    
    {"mov",	OpX3X6b (1, 0, 0x2c), IA64_INSTR_FORMAT_M32, IA64_INSTR_MOV_CR, 
      IA64_OPCODE_PRIV},
    
    {"mov",	OpX3X6b (1, 0, 0x24), IA64_INSTR_FORMAT_M33, IA64_INSTR_MOV_CR, 
      IA64_OPCODE_PRIV},

    {"alloc",	OpX3 (1, 6), IA64_INSTR_FORMAT_M34, IA64_INSTR_ALLOC, 
      IA64_OPCODE_FIRST | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},

    {"mov",	OpX3X6b (1, 0, 0x2d), IA64_INSTR_FORMAT_M35, IA64_INSTR_MOV_PSR, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x29), IA64_INSTR_FORMAT_M35, IA64_INSTR_MOV_PSR},
    
    {"mov",	OpX3X6b (1, 0, 0x25), IA64_INSTR_FORMAT_M36, IA64_INSTR_MOV_PSR, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x21), IA64_INSTR_FORMAT_M36, IA64_INSTR_MOV_PSR},
    
    {"probe.r",	OpX3X6b (1, 0, 0x38), IA64_INSTR_FORMAT_M38, IA64_INSTR_PROBE},
    {"probe.w",	OpX3X6b (1, 0, 0x39), IA64_INSTR_FORMAT_M38, IA64_INSTR_PROBE},
    
    {"probe.r", OpX3X6b (1, 0, 0x18), IA64_INSTR_FORMAT_M39, IA64_INSTR_PROBE},
    {"probe.w", OpX3X6b (1, 0, 0x19), IA64_INSTR_FORMAT_M39, IA64_INSTR_PROBE},
    
    {"probe.rw.fault", OpX3X6b (1, 0, 0x31), IA64_INSTR_FORMAT_M40, IA64_INSTR_PROBE},
    {"probe.r.fault",  OpX3X6b (1, 0, 0x32), IA64_INSTR_FORMAT_M40, IA64_INSTR_PROBE},
    {"probe.w.fault",  OpX3X6b (1, 0, 0x33), IA64_INSTR_FORMAT_M40, IA64_INSTR_PROBE},
    
    {"itc.d",	OpX3X6b (1, 0, 0x2e), IA64_INSTR_FORMAT_M41, IA64_INSTR_ITC, 
      IA64_OPCODE_LAST | IA64_OPCODE_PRIV},
    {"itc.i",	OpX3X6b (1, 0, 0x2f), IA64_INSTR_FORMAT_M41, IA64_INSTR_ITC, 
      IA64_OPCODE_LAST | IA64_OPCODE_PRIV},

    {"mov",	OpX3X6b (1, 0, 0x00), IA64_INSTR_FORMAT_M42, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x01), IA64_INSTR_FORMAT_M42, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x02), IA64_INSTR_FORMAT_M42, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x03), IA64_INSTR_FORMAT_M42, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x04), IA64_INSTR_FORMAT_M42, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x05), IA64_INSTR_FORMAT_M42, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x06), IA64_INSTR_FORMAT_M42, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"itr.d",	OpX3X6b (1, 0, 0x0e), IA64_INSTR_FORMAT_M42, IA64_INSTR_ITR, 
      IA64_OPCODE_PRIV},
    {"itr.i",	OpX3X6b (1, 0, 0x0f), IA64_INSTR_FORMAT_M42, IA64_INSTR_ITR, 
      IA64_OPCODE_PRIV},

    {"mov",	OpX3X6b (1, 0, 0x10), IA64_INSTR_FORMAT_M43, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x11), IA64_INSTR_FORMAT_M43, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x12), IA64_INSTR_FORMAT_M43, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x13), IA64_INSTR_FORMAT_M43, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x14), IA64_INSTR_FORMAT_M43, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x15), IA64_INSTR_FORMAT_M43, IA64_INSTR_MOV_IREG},
    {"mov",	OpX3X6b (1, 0, 0x16), IA64_INSTR_FORMAT_M43, IA64_INSTR_MOV_IREG, 
      IA64_OPCODE_PRIV},
    {"mov",	OpX3X6b (1, 0, 0x17), IA64_INSTR_FORMAT_M43, IA64_INSTR_MOV_IREG},

    {"ptc.l",	OpX3X6b (1, 0, 0x09), IA64_INSTR_FORMAT_M45, IA64_INSTR_PTC_L, 
      IA64_OPCODE_PRIV},
    {"ptc.g",	OpX3X6b (1, 0, 0x0a), IA64_INSTR_FORMAT_M45, IA64_INSTR_PTC_G, 
      IA64_OPCODE_LAST | IA64_OPCODE_PRIV},
    {"ptc.ga",	OpX3X6b (1, 0, 0x0b), IA64_INSTR_FORMAT_M45, IA64_INSTR_PTC_G, 
      IA64_OPCODE_LAST | IA64_OPCODE_PRIV},
    {"ptr.d",	OpX3X6b (1, 0, 0x0c), IA64_INSTR_FORMAT_M45, IA64_INSTR_PTR, 
      IA64_OPCODE_PRIV},
    {"ptr.i",	OpX3X6b (1, 0, 0x0d), IA64_INSTR_FORMAT_M45, IA64_INSTR_PTR, 
      IA64_OPCODE_PRIV},

    {"thash",	OpX3X6b (1, 0, 0x1a), IA64_INSTR_FORMAT_M46, IA64_INSTR_THASH},
    {"ttag",	OpX3X6b (1, 0, 0x1b), IA64_INSTR_FORMAT_M46, IA64_INSTR_TTAG},
    {"tpa",	OpX3X6b (1, 0, 0x1e), IA64_INSTR_FORMAT_M46, IA64_INSTR_TPA, 
      IA64_OPCODE_PRIV},
    {"tak",	OpX3X6b (1, 0, 0x1f), IA64_INSTR_FORMAT_M46, IA64_INSTR_TAK, 
      IA64_OPCODE_PRIV},

    {"chk.s.m",	OpX3 (1, 1), IA64_INSTR_FORMAT_M20, IA64_INSTR_CHK},
    
    {"chk.s",	OpX3 (1, 3), IA64_INSTR_FORMAT_M21, IA64_INSTR_CHK},

    {"fc",	OpX3X6b (1, 0, 0x30), IA64_INSTR_FORMAT_M28, IA64_INSTR_FC},
    {"ptc.e",	OpX3X6b (1, 0, 0x34), IA64_INSTR_FORMAT_M28, IA64_INSTR_PTC_E, 
      IA64_OPCODE_PRIV},

    /* integer load */
    {"ld1",		OpMXX6aHint (4, 0, 0, 0x00, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.nt1",		OpMXX6aHint (4, 0, 0, 0x00, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.nta",		OpMXX6aHint (4, 0, 0, 0x00, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2",		OpMXX6aHint (4, 0, 0, 0x01, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.nt1",		OpMXX6aHint (4, 0, 0, 0x01, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.nta",		OpMXX6aHint (4, 0, 0, 0x01, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4",		OpMXX6aHint (4, 0, 0, 0x02, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.nt1",		OpMXX6aHint (4, 0, 0, 0x02, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.nta",		OpMXX6aHint (4, 0, 0, 0x02, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8",		OpMXX6aHint (4, 0, 0, 0x03, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.nt1",		OpMXX6aHint (4, 0, 0, 0x03, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.nta",		OpMXX6aHint (4, 0, 0, 0x03, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.s",		OpMXX6aHint (4, 0, 0, 0x04, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.s.nt1",	OpMXX6aHint (4, 0, 0, 0x04, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.s.nta",	OpMXX6aHint (4, 0, 0, 0x04, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.s",		OpMXX6aHint (4, 0, 0, 0x05, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.s.nt1",	OpMXX6aHint (4, 0, 0, 0x05, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.s.nta",	OpMXX6aHint (4, 0, 0, 0x05, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.s",		OpMXX6aHint (4, 0, 0, 0x06, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.s.nt1",	OpMXX6aHint (4, 0, 0, 0x06, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.s.nta",	OpMXX6aHint (4, 0, 0, 0x06, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.s",		OpMXX6aHint (4, 0, 0, 0x07, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.s.nt1",	OpMXX6aHint (4, 0, 0, 0x07, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.s.nta",	OpMXX6aHint (4, 0, 0, 0x07, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.a",		OpMXX6aHint (4, 0, 0, 0x08, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.a.nt1",	OpMXX6aHint (4, 0, 0, 0x08, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.a.nta",	OpMXX6aHint (4, 0, 0, 0x08, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.a",		OpMXX6aHint (4, 0, 0, 0x09, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.a.nt1",	OpMXX6aHint (4, 0, 0, 0x09, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.a.nta",	OpMXX6aHint (4, 0, 0, 0x09, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.a",		OpMXX6aHint (4, 0, 0, 0x0a, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.a.nt1",	OpMXX6aHint (4, 0, 0, 0x0a, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.a.nta",	OpMXX6aHint (4, 0, 0, 0x0a, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.a",		OpMXX6aHint (4, 0, 0, 0x0b, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.a.nt1",	OpMXX6aHint (4, 0, 0, 0x0b, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.a.nta",	OpMXX6aHint (4, 0, 0, 0x0b, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.sa",		OpMXX6aHint (4, 0, 0, 0x0c, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.sa.nt1",	OpMXX6aHint (4, 0, 0, 0x0c, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.sa.nta",	OpMXX6aHint (4, 0, 0, 0x0c, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.sa",		OpMXX6aHint (4, 0, 0, 0x0d, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.sa.nt1",	OpMXX6aHint (4, 0, 0, 0x0d, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.sa.nta",	OpMXX6aHint (4, 0, 0, 0x0d, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.sa",		OpMXX6aHint (4, 0, 0, 0x0e, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.sa.nt1",	OpMXX6aHint (4, 0, 0, 0x0e, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.sa.nta",	OpMXX6aHint (4, 0, 0, 0x0e, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.sa",		OpMXX6aHint (4, 0, 0, 0x0f, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.sa.nt1",	OpMXX6aHint (4, 0, 0, 0x0f, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.sa.nta",	OpMXX6aHint (4, 0, 0, 0x0f, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.bias",	OpMXX6aHint (4, 0, 0, 0x10, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.bias.nt1",	OpMXX6aHint (4, 0, 0, 0x10, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.bias.nta",	OpMXX6aHint (4, 0, 0, 0x10, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.bias",	OpMXX6aHint (4, 0, 0, 0x11, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.bias.nt1",	OpMXX6aHint (4, 0, 0, 0x11, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.bias.nta",	OpMXX6aHint (4, 0, 0, 0x11, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.bias",	OpMXX6aHint (4, 0, 0, 0x12, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.bias.nt1",	OpMXX6aHint (4, 0, 0, 0x12, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.bias.nta",	OpMXX6aHint (4, 0, 0, 0x12, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.bias",	OpMXX6aHint (4, 0, 0, 0x13, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.bias.nt1",	OpMXX6aHint (4, 0, 0, 0x13, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.bias.nta",	OpMXX6aHint (4, 0, 0, 0x13, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.acq",		OpMXX6aHint (4, 0, 0, 0x14, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.acq.nt1",	OpMXX6aHint (4, 0, 0, 0x14, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.acq.nta",	OpMXX6aHint (4, 0, 0, 0x14, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.acq",		OpMXX6aHint (4, 0, 0, 0x15, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.acq.nt1",	OpMXX6aHint (4, 0, 0, 0x15, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.acq.nta",	OpMXX6aHint (4, 0, 0, 0x15, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.acq",		OpMXX6aHint (4, 0, 0, 0x16, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.acq.nt1",	OpMXX6aHint (4, 0, 0, 0x16, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.acq.nta",	OpMXX6aHint (4, 0, 0, 0x16, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.acq",		OpMXX6aHint (4, 0, 0, 0x17, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.acq.nt1",	OpMXX6aHint (4, 0, 0, 0x17, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.acq.nta",	OpMXX6aHint (4, 0, 0, 0x17, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.fill",	OpMXX6aHint (4, 0, 0, 0x1b, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.fill.nt1",	OpMXX6aHint (4, 0, 0, 0x1b, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.fill.nta",	OpMXX6aHint (4, 0, 0, 0x1b, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.c.clr",	OpMXX6aHint (4, 0, 0, 0x20, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.c.clr.nt1",	OpMXX6aHint (4, 0, 0, 0x20, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.c.clr.nta",	OpMXX6aHint (4, 0, 0, 0x20, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.c.clr",	OpMXX6aHint (4, 0, 0, 0x21, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.c.clr.nt1",	OpMXX6aHint (4, 0, 0, 0x21, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.c.clr.nta",	OpMXX6aHint (4, 0, 0, 0x21, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.c.clr",	OpMXX6aHint (4, 0, 0, 0x22, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.c.clr.nt1",	OpMXX6aHint (4, 0, 0, 0x22, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.c.clr.nta",	OpMXX6aHint (4, 0, 0, 0x22, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.c.clr",	OpMXX6aHint (4, 0, 0, 0x23, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.c.clr.nt1",	OpMXX6aHint (4, 0, 0, 0x23, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.c.clr.nta",	OpMXX6aHint (4, 0, 0, 0x23, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.c.nc",	OpMXX6aHint (4, 0, 0, 0x24, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.c.nc.nt1",	OpMXX6aHint (4, 0, 0, 0x24, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.c.nc.nta",	OpMXX6aHint (4, 0, 0, 0x24, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.c.nc",	OpMXX6aHint (4, 0, 0, 0x25, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.c.nc.nt1",	OpMXX6aHint (4, 0, 0, 0x25, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.c.nc.nta",	OpMXX6aHint (4, 0, 0, 0x25, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.c.nc",	OpMXX6aHint (4, 0, 0, 0x26, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.c.nc.nt1",	OpMXX6aHint (4, 0, 0, 0x26, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.c.nc.nta",	OpMXX6aHint (4, 0, 0, 0x26, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.c.nc",	OpMXX6aHint (4, 0, 0, 0x27, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.c.nc.nt1",	OpMXX6aHint (4, 0, 0, 0x27, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.c.nc.nta",	OpMXX6aHint (4, 0, 0, 0x27, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.c.clr.acq",	  OpMXX6aHint (4, 0, 0, 0x28, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.c.clr.acq.nt1", OpMXX6aHint (4, 0, 0, 0x28, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld1.c.clr.acq.nta", OpMXX6aHint (4, 0, 0, 0x28, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.c.clr.acq",	  OpMXX6aHint (4, 0, 0, 0x29, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.c.clr.acq.nt1", OpMXX6aHint (4, 0, 0, 0x29, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld2.c.clr.acq.nta", OpMXX6aHint (4, 0, 0, 0x29, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.c.clr.acq",	  OpMXX6aHint (4, 0, 0, 0x2a, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.c.clr.acq.nt1", OpMXX6aHint (4, 0, 0, 0x2a, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld4.c.clr.acq.nta", OpMXX6aHint (4, 0, 0, 0x2a, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.c.clr.acq",	  OpMXX6aHint (4, 0, 0, 0x2b, 0), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.c.clr.acq.nt1", OpMXX6aHint (4, 0, 0, 0x2b, 1), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},
    {"ld8.c.clr.acq.nta", OpMXX6aHint (4, 0, 0, 0x2b, 3), IA64_INSTR_FORMAT_M1, IA64_INSTR_LD},

    /* integer load w/increment by register */
#define LDINCREG(c,h) OpMXX6aHint (4, 1, 0, c, h)
    {"ld1",		LDINCREG (0x00, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.nt1",		LDINCREG (0x00, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.nta",		LDINCREG (0x00, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2",		LDINCREG (0x01, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.nt1",		LDINCREG (0x01, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.nta",		LDINCREG (0x01, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4",		LDINCREG (0x02, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.nt1",		LDINCREG (0x02, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.nta",		LDINCREG (0x02, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8",		LDINCREG (0x03, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.nt1",		LDINCREG (0x03, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.nta",		LDINCREG (0x03, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.s",		LDINCREG (0x04, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.s.nt1",	LDINCREG (0x04, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.s.nta",	LDINCREG (0x04, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.s",		LDINCREG (0x05, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.s.nt1",	LDINCREG (0x05, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.s.nta",	LDINCREG (0x05, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.s",		LDINCREG (0x06, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.s.nt1",	LDINCREG (0x06, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.s.nta",	LDINCREG (0x06, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.s",		LDINCREG (0x07, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.s.nt1",	LDINCREG (0x07, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.s.nta",	LDINCREG (0x07, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.a",		LDINCREG (0x08, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.a.nt1",	LDINCREG (0x08, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.a.nta",	LDINCREG (0x08, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.a",		LDINCREG (0x09, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.a.nt1",	LDINCREG (0x09, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.a.nta",	LDINCREG (0x09, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.a",		LDINCREG (0x0a, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.a.nt1",	LDINCREG (0x0a, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.a.nta",	LDINCREG (0x0a, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.a",		LDINCREG (0x0b, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.a.nt1",	LDINCREG (0x0b, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.a.nta",	LDINCREG (0x0b, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.sa",		LDINCREG (0x0c, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.sa.nt1",	LDINCREG (0x0c, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.sa.nta",	LDINCREG (0x0c, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.sa",		LDINCREG (0x0d, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.sa.nt1",	LDINCREG (0x0d, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.sa.nta",	LDINCREG (0x0d, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.sa",		LDINCREG (0x0e, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.sa.nt1",	LDINCREG (0x0e, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.sa.nta",	LDINCREG (0x0e, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.sa",		LDINCREG (0x0f, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.sa.nt1",	LDINCREG (0x0f, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.sa.nta",	LDINCREG (0x0f, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.bias",	LDINCREG (0x10, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.bias.nt1",	LDINCREG (0x10, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.bias.nta",	LDINCREG (0x10, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.bias",	LDINCREG (0x11, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.bias.nt1",	LDINCREG (0x11, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.bias.nta",	LDINCREG (0x11, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.bias",	LDINCREG (0x12, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.bias.nt1",	LDINCREG (0x12, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.bias.nta",	LDINCREG (0x12, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.bias",	LDINCREG (0x13, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.bias.nt1",	LDINCREG (0x13, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.bias.nta",	LDINCREG (0x13, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.acq",		LDINCREG (0x14, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.acq.nt1",	LDINCREG (0x14, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.acq.nta",	LDINCREG (0x14, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.acq",		LDINCREG (0x15, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.acq.nt1",	LDINCREG (0x15, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.acq.nta",	LDINCREG (0x15, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.acq",		LDINCREG (0x16, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.acq.nt1",	LDINCREG (0x16, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.acq.nta",	LDINCREG (0x16, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.acq",		LDINCREG (0x17, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.acq.nt1",	LDINCREG (0x17, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.acq.nta",	LDINCREG (0x17, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.fill",	LDINCREG (0x1b, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.fill.nt1",	LDINCREG (0x1b, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.fill.nta",	LDINCREG (0x1b, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr",	LDINCREG (0x20, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.nt1",	LDINCREG (0x20, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.nta",	LDINCREG (0x20, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr",	LDINCREG (0x21, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.nt1",	LDINCREG (0x21, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.nta",	LDINCREG (0x21, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr",	LDINCREG (0x22, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.nt1",	LDINCREG (0x22, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.nta",	LDINCREG (0x22, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr",	LDINCREG (0x23, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.nt1",	LDINCREG (0x23, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.nta",	LDINCREG (0x23, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.nc",	LDINCREG (0x24, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.nc.nt1",	LDINCREG (0x24, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.nc.nta",	LDINCREG (0x24, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.nc",	LDINCREG (0x25, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.nc.nt1",	LDINCREG (0x25, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.nc.nta",	LDINCREG (0x25, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.nc",	LDINCREG (0x26, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.nc.nt1",	LDINCREG (0x26, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.nc.nta",	LDINCREG (0x26, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.nc",	LDINCREG (0x27, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.nc.nt1",	LDINCREG (0x27, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.nc.nta",	LDINCREG (0x27, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.acq",	  LDINCREG (0x28, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.acq.nt1", LDINCREG (0x28, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.acq.nta", LDINCREG (0x28, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.acq",	  LDINCREG (0x29, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.acq.nt1", LDINCREG (0x29, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.acq.nta", LDINCREG (0x29, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.acq",	  LDINCREG (0x2a, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.acq.nt1", LDINCREG (0x2a, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.acq.nta", LDINCREG (0x2a, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.acq",	  LDINCREG (0x2b, 0), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.acq.nt1", LDINCREG (0x2b, 1), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.acq.nta", LDINCREG (0x2b, 3), IA64_INSTR_FORMAT_M2, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
#undef LDINCREG

    {"st1",		OpMXX6aHint (4, 0, 0, 0x30, 0), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st1.nta",		OpMXX6aHint (4, 0, 0, 0x30, 3), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st2",		OpMXX6aHint (4, 0, 0, 0x31, 0), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st2.nta",		OpMXX6aHint (4, 0, 0, 0x31, 3), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st4",		OpMXX6aHint (4, 0, 0, 0x32, 0), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st4.nta",		OpMXX6aHint (4, 0, 0, 0x32, 3), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st8",		OpMXX6aHint (4, 0, 0, 0x33, 0), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st8.nta",		OpMXX6aHint (4, 0, 0, 0x33, 3), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st1.rel",		OpMXX6aHint (4, 0, 0, 0x34, 0), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st1.rel.nta",	OpMXX6aHint (4, 0, 0, 0x34, 3), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st2.rel",		OpMXX6aHint (4, 0, 0, 0x35, 0), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st2.rel.nta",	OpMXX6aHint (4, 0, 0, 0x35, 3), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st4.rel",		OpMXX6aHint (4, 0, 0, 0x36, 0), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st4.rel.nta",	OpMXX6aHint (4, 0, 0, 0x36, 3), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st8.rel",		OpMXX6aHint (4, 0, 0, 0x37, 0), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st8.rel.nta",	OpMXX6aHint (4, 0, 0, 0x37, 3), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st8.spill",	OpMXX6aHint (4, 0, 0, 0x3b, 0), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},
    {"st8.spill.nta",	OpMXX6aHint (4, 0, 0, 0x3b, 3), IA64_INSTR_FORMAT_M4, IA64_INSTR_ST},

#define CMPXCHG(c,h)	OpMXX6aHint (4, 0, 1, c, h)
    {"cmpxchg1.acq",		CMPXCHG (0x00, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg1.acq.nt1",	CMPXCHG (0x00, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg1.acq.nta",	CMPXCHG (0x00, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg2.acq",		CMPXCHG (0x01, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg2.acq.nt1",	CMPXCHG (0x01, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg2.acq.nta",	CMPXCHG (0x01, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg4.acq",		CMPXCHG (0x02, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg4.acq.nt1",	CMPXCHG (0x02, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg4.acq.nta",	CMPXCHG (0x02, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg8.acq",		CMPXCHG (0x03, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg8.acq.nt1",	CMPXCHG (0x03, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg8.acq.nta",	CMPXCHG (0x03, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg1.rel",		CMPXCHG (0x04, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg1.rel.nt1",	CMPXCHG (0x04, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg1.rel.nta",	CMPXCHG (0x04, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg2.rel",		CMPXCHG (0x05, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg2.rel.nt1",	CMPXCHG (0x05, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg2.rel.nta",	CMPXCHG (0x05, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg4.rel",		CMPXCHG (0x06, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg4.rel.nt1",	CMPXCHG (0x06, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg4.rel.nta",	CMPXCHG (0x06, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg8.rel",		CMPXCHG (0x07, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg8.rel.nt1",	CMPXCHG (0x07, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
    {"cmpxchg8.rel.nta",	CMPXCHG (0x07, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_CMPXCHG},
#undef CMPXCHG
    {"xchg1",		  OpMXX6aHint (4, 0, 1, 0x08, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg1.nt1",	  OpMXX6aHint (4, 0, 1, 0x08, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg1.nta",	  OpMXX6aHint (4, 0, 1, 0x08, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg2",		  OpMXX6aHint (4, 0, 1, 0x09, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg2.nt1",	  OpMXX6aHint (4, 0, 1, 0x09, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg2.nta",	  OpMXX6aHint (4, 0, 1, 0x09, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg4",		  OpMXX6aHint (4, 0, 1, 0x0a, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg4.nt1",	  OpMXX6aHint (4, 0, 1, 0x0a, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg4.nta",	  OpMXX6aHint (4, 0, 1, 0x0a, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg8",		  OpMXX6aHint (4, 0, 1, 0x0b, 0), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg8.nt1",	  OpMXX6aHint (4, 0, 1, 0x0b, 1), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},
    {"xchg8.nta",	  OpMXX6aHint (4, 0, 1, 0x0b, 3), IA64_INSTR_FORMAT_M16, IA64_INSTR_XCHG},

    {"fetchadd4.acq",	  OpMXX6aHint (4, 0, 1, 0x12, 0), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd4.acq.nt1", OpMXX6aHint (4, 0, 1, 0x12, 1), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd4.acq.nta", OpMXX6aHint (4, 0, 1, 0x12, 3), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd8.acq",	  OpMXX6aHint (4, 0, 1, 0x13, 0), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd8.acq.nt1", OpMXX6aHint (4, 0, 1, 0x13, 1), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd8.acq.nta", OpMXX6aHint (4, 0, 1, 0x13, 3), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd4.rel",	  OpMXX6aHint (4, 0, 1, 0x16, 0), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd4.rel.nt1", OpMXX6aHint (4, 0, 1, 0x16, 1), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd4.rel.nta", OpMXX6aHint (4, 0, 1, 0x16, 3), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd8.rel",	  OpMXX6aHint (4, 0, 1, 0x17, 0), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd8.rel.nt1", OpMXX6aHint (4, 0, 1, 0x17, 1), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},
    {"fetchadd8.rel.nta", OpMXX6aHint (4, 0, 1, 0x17, 3), IA64_INSTR_FORMAT_M17, IA64_INSTR_FETCHADD},

    {"getf.sig",	  OpMXX6a (4, 0, 1, 0x1c), IA64_INSTR_FORMAT_M19, IA64_INSTR_GETF},
    {"getf.exp",	  OpMXX6a (4, 0, 1, 0x1d), IA64_INSTR_FORMAT_M19, IA64_INSTR_GETF},
    {"getf.s",		  OpMXX6a (4, 0, 1, 0x1e), IA64_INSTR_FORMAT_M19, IA64_INSTR_GETF},
    {"getf.d",		  OpMXX6a (4, 0, 1, 0x1f), IA64_INSTR_FORMAT_M19, IA64_INSTR_GETF},

    /* integer load w/increment by immediate */
#define LDINCIMMED(c,h) OpX6aHint (5, c, h)
    {"ld1",		LDINCIMMED (0x00, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.nt1",		LDINCIMMED (0x00, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.nta",		LDINCIMMED (0x00, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2",		LDINCIMMED (0x01, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.nt1",		LDINCIMMED (0x01, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.nta",		LDINCIMMED (0x01, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4",		LDINCIMMED (0x02, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.nt1",		LDINCIMMED (0x02, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.nta",		LDINCIMMED (0x02, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8",		LDINCIMMED (0x03, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.nt1",		LDINCIMMED (0x03, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.nta",		LDINCIMMED (0x03, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.s",		LDINCIMMED (0x04, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.s.nt1",	LDINCIMMED (0x04, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.s.nta",	LDINCIMMED (0x04, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.s",		LDINCIMMED (0x05, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.s.nt1",	LDINCIMMED (0x05, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.s.nta",	LDINCIMMED (0x05, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.s",		LDINCIMMED (0x06, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.s.nt1",	LDINCIMMED (0x06, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.s.nta",	LDINCIMMED (0x06, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.s",		LDINCIMMED (0x07, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.s.nt1",	LDINCIMMED (0x07, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.s.nta",	LDINCIMMED (0x07, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.a",		LDINCIMMED (0x08, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.a.nt1",	LDINCIMMED (0x08, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.a.nta",	LDINCIMMED (0x08, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.a",		LDINCIMMED (0x09, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.a.nt1",	LDINCIMMED (0x09, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.a.nta",	LDINCIMMED (0x09, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.a",		LDINCIMMED (0x0a, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.a.nt1",	LDINCIMMED (0x0a, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.a.nta",	LDINCIMMED (0x0a, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.a",		LDINCIMMED (0x0b, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.a.nt1",	LDINCIMMED (0x0b, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.a.nta",	LDINCIMMED (0x0b, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.sa",		LDINCIMMED (0x0c, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.sa.nt1",	LDINCIMMED (0x0c, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.sa.nta",	LDINCIMMED (0x0c, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.sa",		LDINCIMMED (0x0d, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.sa.nt1",	LDINCIMMED (0x0d, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.sa.nta",	LDINCIMMED (0x0d, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.sa",		LDINCIMMED (0x0e, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.sa.nt1",	LDINCIMMED (0x0e, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.sa.nta",	LDINCIMMED (0x0e, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.sa",		LDINCIMMED (0x0f, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.sa.nt1",	LDINCIMMED (0x0f, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.sa.nta",	LDINCIMMED (0x0f, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.bias",	LDINCIMMED (0x10, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.bias.nt1",	LDINCIMMED (0x10, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.bias.nta",	LDINCIMMED (0x10, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.bias",	LDINCIMMED (0x11, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.bias.nt1",	LDINCIMMED (0x11, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.bias.nta",	LDINCIMMED (0x11, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.bias",	LDINCIMMED (0x12, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.bias.nt1",	LDINCIMMED (0x12, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.bias.nta",	LDINCIMMED (0x12, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.bias",	LDINCIMMED (0x13, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.bias.nt1",	LDINCIMMED (0x13, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.bias.nta",	LDINCIMMED (0x13, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.acq",		LDINCIMMED (0x14, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.acq.nt1",	LDINCIMMED (0x14, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.acq.nta",	LDINCIMMED (0x14, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.acq",		LDINCIMMED (0x15, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.acq.nt1",	LDINCIMMED (0x15, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.acq.nta",	LDINCIMMED (0x15, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.acq",		LDINCIMMED (0x16, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.acq.nt1",	LDINCIMMED (0x16, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.acq.nta",	LDINCIMMED (0x16, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.acq",		LDINCIMMED (0x17, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.acq.nt1",	LDINCIMMED (0x17, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.acq.nta",	LDINCIMMED (0x17, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.fill",	LDINCIMMED (0x1b, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.fill.nt1",	LDINCIMMED (0x1b, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.fill.nta",	LDINCIMMED (0x1b, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr",	LDINCIMMED (0x20, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.nt1",	LDINCIMMED (0x20, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.nta",	LDINCIMMED (0x20, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr",	LDINCIMMED (0x21, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.nt1",	LDINCIMMED (0x21, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.nta",	LDINCIMMED (0x21, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr",	LDINCIMMED (0x22, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.nt1",	LDINCIMMED (0x22, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.nta",	LDINCIMMED (0x22, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr",	LDINCIMMED (0x23, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.nt1",	LDINCIMMED (0x23, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.nta",	LDINCIMMED (0x23, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.nc",	LDINCIMMED (0x24, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.nc.nt1",	LDINCIMMED (0x24, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.nc.nta",	LDINCIMMED (0x24, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.nc",	LDINCIMMED (0x25, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.nc.nt1",	LDINCIMMED (0x25, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.nc.nta",	LDINCIMMED (0x25, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.nc",	LDINCIMMED (0x26, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.nc.nt1",	LDINCIMMED (0x26, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.nc.nta",	LDINCIMMED (0x26, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.nc",	LDINCIMMED (0x27, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.nc.nt1",	LDINCIMMED (0x27, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.nc.nta",	LDINCIMMED (0x27, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.acq",	  LDINCIMMED (0x28, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.acq.nt1", LDINCIMMED (0x28, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld1.c.clr.acq.nta", LDINCIMMED (0x28, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.acq",	  LDINCIMMED (0x29, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.acq.nt1", LDINCIMMED (0x29, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld2.c.clr.acq.nta", LDINCIMMED (0x29, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.acq",	  LDINCIMMED (0x2a, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.acq.nt1", LDINCIMMED (0x2a, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld4.c.clr.acq.nta", LDINCIMMED (0x2a, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.acq",	  LDINCIMMED (0x2b, 0), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.acq.nt1", LDINCIMMED (0x2b, 1), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
    {"ld8.c.clr.acq.nta", LDINCIMMED (0x2b, 3), IA64_INSTR_FORMAT_M3, IA64_INSTR_LD, 
      IA64_OPCODE_POSTINC},
#undef LDINCIMMED

    /* store w/increment by immediate */
#define STINCIMMED(c,h) OpX6aHint (5, c, h)
    {"st1",		STINCIMMED (0x30, 0), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st1.nta",		STINCIMMED (0x30, 3), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st2",		STINCIMMED (0x31, 0), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st2.nta",		STINCIMMED (0x31, 3), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st4",		STINCIMMED (0x32, 0), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st4.nta",		STINCIMMED (0x32, 3), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st8",		STINCIMMED (0x33, 0), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st8.nta",		STINCIMMED (0x33, 3), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st1.rel",		STINCIMMED (0x34, 0), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st1.rel.nta",	STINCIMMED (0x34, 3), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st2.rel",		STINCIMMED (0x35, 0), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st2.rel.nta",	STINCIMMED (0x35, 3), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st4.rel",		STINCIMMED (0x36, 0), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st4.rel.nta",	STINCIMMED (0x36, 3), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st8.rel",		STINCIMMED (0x37, 0), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st8.rel.nta",	STINCIMMED (0x37, 3), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st8.spill",	STINCIMMED (0x3b, 0), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
    {"st8.spill.nta",	STINCIMMED (0x3b, 3), IA64_INSTR_FORMAT_M5, IA64_INSTR_ST, 
      IA64_OPCODE_POSTINC},
#undef STINCIMMED

    /* floating-point load */
    {"ldfs",		OpMXX6aHint (6, 0, 0, 0x02, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.nt1",	OpMXX6aHint (6, 0, 0, 0x02, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.nta",	OpMXX6aHint (6, 0, 0, 0x02, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd",		OpMXX6aHint (6, 0, 0, 0x03, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.nt1",	OpMXX6aHint (6, 0, 0, 0x03, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.nta",	OpMXX6aHint (6, 0, 0, 0x03, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8",		OpMXX6aHint (6, 0, 0, 0x01, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.nt1",	OpMXX6aHint (6, 0, 0, 0x01, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.nta",	OpMXX6aHint (6, 0, 0, 0x01, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe",		OpMXX6aHint (6, 0, 0, 0x00, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.nt1",	OpMXX6aHint (6, 0, 0, 0x00, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.nta",	OpMXX6aHint (6, 0, 0, 0x00, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.s",		OpMXX6aHint (6, 0, 0, 0x06, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.s.nt1",	OpMXX6aHint (6, 0, 0, 0x06, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.s.nta",	OpMXX6aHint (6, 0, 0, 0x06, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.s",		OpMXX6aHint (6, 0, 0, 0x07, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.s.nt1",	OpMXX6aHint (6, 0, 0, 0x07, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.s.nta",	OpMXX6aHint (6, 0, 0, 0x07, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.s",		OpMXX6aHint (6, 0, 0, 0x05, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.s.nt1",	OpMXX6aHint (6, 0, 0, 0x05, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.s.nta",	OpMXX6aHint (6, 0, 0, 0x05, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.s",		OpMXX6aHint (6, 0, 0, 0x04, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.s.nt1",	OpMXX6aHint (6, 0, 0, 0x04, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.s.nta",	OpMXX6aHint (6, 0, 0, 0x04, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.a",		OpMXX6aHint (6, 0, 0, 0x0a, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.a.nt1",	OpMXX6aHint (6, 0, 0, 0x0a, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.a.nta",	OpMXX6aHint (6, 0, 0, 0x0a, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.a",		OpMXX6aHint (6, 0, 0, 0x0b, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.a.nt1",	OpMXX6aHint (6, 0, 0, 0x0b, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.a.nta",	OpMXX6aHint (6, 0, 0, 0x0b, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.a",		OpMXX6aHint (6, 0, 0, 0x09, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.a.nt1",	OpMXX6aHint (6, 0, 0, 0x09, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.a.nta",	OpMXX6aHint (6, 0, 0, 0x09, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.a",		OpMXX6aHint (6, 0, 0, 0x08, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.a.nt1",	OpMXX6aHint (6, 0, 0, 0x08, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.a.nta",	OpMXX6aHint (6, 0, 0, 0x08, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.sa",		OpMXX6aHint (6, 0, 0, 0x0e, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.sa.nt1",	OpMXX6aHint (6, 0, 0, 0x0e, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.sa.nta",	OpMXX6aHint (6, 0, 0, 0x0e, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.sa",		OpMXX6aHint (6, 0, 0, 0x0f, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.sa.nt1",	OpMXX6aHint (6, 0, 0, 0x0f, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.sa.nta",	OpMXX6aHint (6, 0, 0, 0x0f, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.sa",		OpMXX6aHint (6, 0, 0, 0x0d, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.sa.nt1",	OpMXX6aHint (6, 0, 0, 0x0d, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.sa.nta",	OpMXX6aHint (6, 0, 0, 0x0d, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.sa",		OpMXX6aHint (6, 0, 0, 0x0c, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.sa.nt1",	OpMXX6aHint (6, 0, 0, 0x0c, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.sa.nta",	OpMXX6aHint (6, 0, 0, 0x0c, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf.fill",	OpMXX6aHint (6, 0, 0, 0x1b, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf.fill.nt1",	OpMXX6aHint (6, 0, 0, 0x1b, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf.fill.nta",	OpMXX6aHint (6, 0, 0, 0x1b, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.c.clr",	OpMXX6aHint (6, 0, 0, 0x22, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.c.clr.nt1",	OpMXX6aHint (6, 0, 0, 0x22, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.c.clr.nta",	OpMXX6aHint (6, 0, 0, 0x22, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.c.clr",	OpMXX6aHint (6, 0, 0, 0x23, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.c.clr.nt1",	OpMXX6aHint (6, 0, 0, 0x23, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.c.clr.nta",	OpMXX6aHint (6, 0, 0, 0x23, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.c.clr",	OpMXX6aHint (6, 0, 0, 0x21, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.c.clr.nt1",	OpMXX6aHint (6, 0, 0, 0x21, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.c.clr.nta",	OpMXX6aHint (6, 0, 0, 0x21, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.c.clr",	OpMXX6aHint (6, 0, 0, 0x20, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.c.clr.nt1",	OpMXX6aHint (6, 0, 0, 0x20, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.c.clr.nta",	OpMXX6aHint (6, 0, 0, 0x20, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.c.nc",	OpMXX6aHint (6, 0, 0, 0x26, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.c.nc.nt1",	OpMXX6aHint (6, 0, 0, 0x26, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfs.c.nc.nta",	OpMXX6aHint (6, 0, 0, 0x26, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.c.nc",	OpMXX6aHint (6, 0, 0, 0x27, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.c.nc.nt1",	OpMXX6aHint (6, 0, 0, 0x27, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfd.c.nc.nta",	OpMXX6aHint (6, 0, 0, 0x27, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.c.nc",	OpMXX6aHint (6, 0, 0, 0x25, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.c.nc.nt1",	OpMXX6aHint (6, 0, 0, 0x25, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldf8.c.nc.nta",	OpMXX6aHint (6, 0, 0, 0x25, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.c.nc",	OpMXX6aHint (6, 0, 0, 0x24, 0), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.c.nc.nt1",	OpMXX6aHint (6, 0, 0, 0x24, 1), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},
    {"ldfe.c.nc.nta",	OpMXX6aHint (6, 0, 0, 0x24, 3), IA64_INSTR_FORMAT_M6, IA64_INSTR_LDF},

    /* floating-point load w/increment by register */
#define FLDINCREG(c,h) OpMXX6aHint (6, 1, 0, c, h)
    {"ldfs",		FLDINCREG (0x02, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.nt1",	FLDINCREG (0x02, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.nta",	FLDINCREG (0x02, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd",		FLDINCREG (0x03, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.nt1",	FLDINCREG (0x03, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.nta",	FLDINCREG (0x03, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8",		FLDINCREG (0x01, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.nt1",	FLDINCREG (0x01, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.nta",	FLDINCREG (0x01, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe",		FLDINCREG (0x00, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.nt1",	FLDINCREG (0x00, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.nta",	FLDINCREG (0x00, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.s",		FLDINCREG (0x06, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.s.nt1",	FLDINCREG (0x06, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.s.nta",	FLDINCREG (0x06, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.s",		FLDINCREG (0x07, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.s.nt1",	FLDINCREG (0x07, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.s.nta",	FLDINCREG (0x07, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.s",		FLDINCREG (0x05, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.s.nt1",	FLDINCREG (0x05, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.s.nta",	FLDINCREG (0x05, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.s",		FLDINCREG (0x04, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.s.nt1",	FLDINCREG (0x04, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.s.nta",	FLDINCREG (0x04, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.a",		FLDINCREG (0x0a, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.a.nt1",	FLDINCREG (0x0a, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.a.nta",	FLDINCREG (0x0a, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.a",		FLDINCREG (0x0b, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.a.nt1",	FLDINCREG (0x0b, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.a.nta",	FLDINCREG (0x0b, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.a",		FLDINCREG (0x09, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.a.nt1",	FLDINCREG (0x09, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.a.nta",	FLDINCREG (0x09, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.a",		FLDINCREG (0x08, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.a.nt1",	FLDINCREG (0x08, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.a.nta",	FLDINCREG (0x08, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.sa",		FLDINCREG (0x0e, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.sa.nt1",	FLDINCREG (0x0e, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.sa.nta",	FLDINCREG (0x0e, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.sa",		FLDINCREG (0x0f, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.sa.nt1",	FLDINCREG (0x0f, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.sa.nta",	FLDINCREG (0x0f, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.sa",		FLDINCREG (0x0d, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.sa.nt1",	FLDINCREG (0x0d, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.sa.nta",	FLDINCREG (0x0d, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.sa",		FLDINCREG (0x0c, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.sa.nt1",	FLDINCREG (0x0c, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.sa.nta",	FLDINCREG (0x0c, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf.fill",	FLDINCREG (0x1b, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf.fill.nt1",	FLDINCREG (0x1b, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf.fill.nta",	FLDINCREG (0x1b, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.clr",	FLDINCREG (0x22, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.clr.nt1",	FLDINCREG (0x22, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.clr.nta",	FLDINCREG (0x22, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.clr",	FLDINCREG (0x23, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.clr.nt1",	FLDINCREG (0x23, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.clr.nta",	FLDINCREG (0x23, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.clr",	FLDINCREG (0x21, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.clr.nt1",	FLDINCREG (0x21, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.clr.nta",	FLDINCREG (0x21, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.clr",	FLDINCREG (0x20, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.clr.nt1",	FLDINCREG (0x20, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.clr.nta",	FLDINCREG (0x20, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.nc",	FLDINCREG (0x26, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.nc.nt1",	FLDINCREG (0x26, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.nc.nta",	FLDINCREG (0x26, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.nc",	FLDINCREG (0x27, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.nc.nt1",	FLDINCREG (0x27, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.nc.nta",	FLDINCREG (0x27, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.nc",	FLDINCREG (0x25, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.nc.nt1",	FLDINCREG (0x25, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.nc.nta",	FLDINCREG (0x25, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.nc",	FLDINCREG (0x24, 0), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.nc.nt1",	FLDINCREG (0x24, 1), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.nc.nta",	FLDINCREG (0x24, 3), IA64_INSTR_FORMAT_M7, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
#undef FLDINCREG

    /* floating-point store */
    {"stfs",		OpMXX6aHint (6, 0, 0, 0x32, 0), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},
    {"stfs.nta",	OpMXX6aHint (6, 0, 0, 0x32, 3), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},
    {"stfd",		OpMXX6aHint (6, 0, 0, 0x33, 0), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},
    {"stfd.nta",	OpMXX6aHint (6, 0, 0, 0x33, 3), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},
    {"stf8",		OpMXX6aHint (6, 0, 0, 0x31, 0), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},
    {"stf8.nta",	OpMXX6aHint (6, 0, 0, 0x31, 3), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},
    {"stfe",		OpMXX6aHint (6, 0, 0, 0x30, 0), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},
    {"stfe.nta",	OpMXX6aHint (6, 0, 0, 0x30, 3), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},
    {"stf.spill",	OpMXX6aHint (6, 0, 0, 0x3b, 0), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},
    {"stf.spill.nta",	OpMXX6aHint (6, 0, 0, 0x3b, 3), IA64_INSTR_FORMAT_M9, IA64_INSTR_STF},

    /* floating-point load pair */
    {"ldfps",		OpMXX6aHint (6, 0, 1, 0x02, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.nt1",	OpMXX6aHint (6, 0, 1, 0x02, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.nta",	OpMXX6aHint (6, 0, 1, 0x02, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd",		OpMXX6aHint (6, 0, 1, 0x03, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.nt1",	OpMXX6aHint (6, 0, 1, 0x03, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.nta",	OpMXX6aHint (6, 0, 1, 0x03, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8",		OpMXX6aHint (6, 0, 1, 0x01, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.nt1",	OpMXX6aHint (6, 0, 1, 0x01, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.nta",	OpMXX6aHint (6, 0, 1, 0x01, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.s",		OpMXX6aHint (6, 0, 1, 0x06, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.s.nt1",	OpMXX6aHint (6, 0, 1, 0x06, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.s.nta",	OpMXX6aHint (6, 0, 1, 0x06, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.s",		OpMXX6aHint (6, 0, 1, 0x07, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.s.nt1",	OpMXX6aHint (6, 0, 1, 0x07, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.s.nta",	OpMXX6aHint (6, 0, 1, 0x07, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.s",		OpMXX6aHint (6, 0, 1, 0x05, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.s.nt1",	OpMXX6aHint (6, 0, 1, 0x05, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.s.nta",	OpMXX6aHint (6, 0, 1, 0x05, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.a",		OpMXX6aHint (6, 0, 1, 0x0a, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.a.nt1",	OpMXX6aHint (6, 0, 1, 0x0a, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.a.nta",	OpMXX6aHint (6, 0, 1, 0x0a, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.a",		OpMXX6aHint (6, 0, 1, 0x0b, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.a.nt1",	OpMXX6aHint (6, 0, 1, 0x0b, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.a.nta",	OpMXX6aHint (6, 0, 1, 0x0b, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.a",		OpMXX6aHint (6, 0, 1, 0x09, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.a.nt1",	OpMXX6aHint (6, 0, 1, 0x09, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.a.nta",	OpMXX6aHint (6, 0, 1, 0x09, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.sa",	OpMXX6aHint (6, 0, 1, 0x0e, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.sa.nt1",	OpMXX6aHint (6, 0, 1, 0x0e, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.sa.nta",	OpMXX6aHint (6, 0, 1, 0x0e, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.sa",	OpMXX6aHint (6, 0, 1, 0x0f, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.sa.nt1",	OpMXX6aHint (6, 0, 1, 0x0f, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.sa.nta",	OpMXX6aHint (6, 0, 1, 0x0f, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.sa",	OpMXX6aHint (6, 0, 1, 0x0d, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.sa.nt1",	OpMXX6aHint (6, 0, 1, 0x0d, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.sa.nta",	OpMXX6aHint (6, 0, 1, 0x0d, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.c.clr",	OpMXX6aHint (6, 0, 1, 0x22, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.c.clr.nt1",	OpMXX6aHint (6, 0, 1, 0x22, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.c.clr.nta",	OpMXX6aHint (6, 0, 1, 0x22, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.c.clr",	OpMXX6aHint (6, 0, 1, 0x23, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.c.clr.nt1",	OpMXX6aHint (6, 0, 1, 0x23, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.c.clr.nta",	OpMXX6aHint (6, 0, 1, 0x23, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.c.clr",	OpMXX6aHint (6, 0, 1, 0x21, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.c.clr.nt1",	OpMXX6aHint (6, 0, 1, 0x21, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.c.clr.nta",	OpMXX6aHint (6, 0, 1, 0x21, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.c.nc",	OpMXX6aHint (6, 0, 1, 0x26, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.c.nc.nt1",	OpMXX6aHint (6, 0, 1, 0x26, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfps.c.nc.nta",	OpMXX6aHint (6, 0, 1, 0x26, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.c.nc",	OpMXX6aHint (6, 0, 1, 0x27, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.c.nc.nt1",	OpMXX6aHint (6, 0, 1, 0x27, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfpd.c.nc.nta",	OpMXX6aHint (6, 0, 1, 0x27, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.c.nc",	OpMXX6aHint (6, 0, 1, 0x25, 0), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.c.nc.nt1",	OpMXX6aHint (6, 0, 1, 0x25, 1), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},
    {"ldfp8.c.nc.nta",	OpMXX6aHint (6, 0, 1, 0x25, 3), IA64_INSTR_FORMAT_M11, IA64_INSTR_LDFP},

    /* floating-point load pair w/increment by immediate */
#define LD(a,b,c) OpMXX6aHint (6, 1, 1, a, b)
    {"ldfps",		LD (0x02, 0, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.nt1",	LD (0x02, 1, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.nta",	LD (0x02, 3, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd",		LD (0x03, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.nt1",	LD (0x03, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.nta",	LD (0x03, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8",		LD (0x01, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.nt1",	LD (0x01, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.nta",	LD (0x01, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.s",		LD (0x06, 0, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.s.nt1",	LD (0x06, 1, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.s.nta",	LD (0x06, 3, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.s",		LD (0x07, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.s.nt1",	LD (0x07, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.s.nta",	LD (0x07, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.s",		LD (0x05, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.s.nt1",	LD (0x05, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.s.nta",	LD (0x05, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.a",		LD (0x0a, 0, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.a.nt1",	LD (0x0a, 1, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.a.nta",	LD (0x0a, 3, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.a",		LD (0x0b, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.a.nt1",	LD (0x0b, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.a.nta",	LD (0x0b, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.a",		LD (0x09, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.a.nt1",	LD (0x09, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.a.nta",	LD (0x09, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.sa",	LD (0x0e, 0, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.sa.nt1",	LD (0x0e, 1, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.sa.nta",	LD (0x0e, 3, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.sa",	LD (0x0f, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.sa.nt1",	LD (0x0f, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.sa.nta",	LD (0x0f, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.sa",	LD (0x0d, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.sa.nt1",	LD (0x0d, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.sa.nta",	LD (0x0d, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.c.clr",	LD (0x22, 0, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.c.clr.nt1",	LD (0x22, 1, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.c.clr.nta",	LD (0x22, 3, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.c.clr",	LD (0x23, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.c.clr.nt1",	LD (0x23, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.c.clr.nta",	LD (0x23, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.c.clr",	LD (0x21, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.c.clr.nt1",	LD (0x21, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.c.clr.nta",	LD (0x21, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.c.nc",	LD (0x26, 0, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.c.nc.nt1",	LD (0x26, 1, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfps.c.nc.nta",	LD (0x26, 3, C8), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.c.nc",	LD (0x27, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.c.nc.nt1",	LD (0x27, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfpd.c.nc.nta",	LD (0x27, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.c.nc",	LD (0x25, 0, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.c.nc.nt1",	LD (0x25, 1, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
    {"ldfp8.c.nc.nta",	LD (0x25, 3, C16), IA64_INSTR_FORMAT_M12, IA64_INSTR_LDFP, 
      IA64_OPCODE_POSTINC},
#undef LD

/* line prefetch */
    {"lfetch",                  OpMXX6aHint (6, 0, 0, 0x2c, 0), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.nt1",              OpMXX6aHint (6, 0, 0, 0x2c, 1), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.nt2",              OpMXX6aHint (6, 0, 0, 0x2c, 2), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.nta",              OpMXX6aHint (6, 0, 0, 0x2c, 3), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.excl",             OpMXX6aHint (6, 0, 0, 0x2d, 0), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.excl.nt1",         OpMXX6aHint (6, 0, 0, 0x2d, 1), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.excl.nt2",         OpMXX6aHint (6, 0, 0, 0x2d, 2), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.excl.nta",         OpMXX6aHint (6, 0, 0, 0x2d, 3), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.fault",            OpMXX6aHint (6, 0, 0, 0x2e, 0), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.fault.nt1",        OpMXX6aHint (6, 0, 0, 0x2e, 1), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.fault.nt2",        OpMXX6aHint (6, 0, 0, 0x2e, 2), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.fault.nta",        OpMXX6aHint (6, 0, 0, 0x2e, 3), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.fault.excl",       OpMXX6aHint (6, 0, 0, 0x2f, 0), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.fault.excl.nt1",   OpMXX6aHint (6, 0, 0, 0x2f, 1), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.fault.excl.nt2",   OpMXX6aHint (6, 0, 0, 0x2f, 2), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},
    {"lfetch.fault.excl.nta",   OpMXX6aHint (6, 0, 0, 0x2f, 3), IA64_INSTR_FORMAT_M13, IA64_INSTR_LFETCH},

    /* line prefetch w/increment by register */
#define LFETCHINCREG(c,h) OpMXX6aHint (6, 1, 0, c, h)
    {"lfetch",			LFETCHINCREG (0x2c, 0), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.nt1",		LFETCHINCREG (0x2c, 1), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.nt2",		LFETCHINCREG (0x2c, 2), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.nta",		LFETCHINCREG (0x2c, 3), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.excl",		LFETCHINCREG (0x2d, 0), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.excl.nt1",		LFETCHINCREG (0x2d, 1), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.excl.nt2",		LFETCHINCREG (0x2d, 2), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.excl.nta",		LFETCHINCREG (0x2d, 3), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault",		LFETCHINCREG (0x2e, 0), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.nt1",	LFETCHINCREG (0x2e, 1), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.nt2",	LFETCHINCREG (0x2e, 2), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.nta",	LFETCHINCREG (0x2e, 3), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.excl",	LFETCHINCREG (0x2f, 0), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.excl.nt1",	LFETCHINCREG (0x2f, 1), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.excl.nt2",	LFETCHINCREG (0x2f, 2), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.excl.nta",	LFETCHINCREG (0x2f, 3), IA64_INSTR_FORMAT_M14, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
#undef LFETCHINCREG

    /* semaphore operations */
    {"setf.sig",	OpMXX6a (6, 0, 1, 0x1c), IA64_INSTR_FORMAT_M18, IA64_INSTR_SETF},
    {"setf.exp",	OpMXX6a (6, 0, 1, 0x1d), IA64_INSTR_FORMAT_M18, IA64_INSTR_SETF},
    {"setf.s",		OpMXX6a (6, 0, 1, 0x1e), IA64_INSTR_FORMAT_M18, IA64_INSTR_SETF},
    {"setf.d",		OpMXX6a (6, 0, 1, 0x1f), IA64_INSTR_FORMAT_M18, IA64_INSTR_SETF},

    /* floating-point load w/increment by immediate */
#define FLDINCIMMED(c,h) OpX6aHint (7, c, h)
    {"ldfs",		FLDINCIMMED (0x02, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.nt1",	FLDINCIMMED (0x02, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.nta",	FLDINCIMMED (0x02, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd",		FLDINCIMMED (0x03, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.nt1",	FLDINCIMMED (0x03, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.nta",	FLDINCIMMED (0x03, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8",		FLDINCIMMED (0x01, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.nt1",	FLDINCIMMED (0x01, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.nta",	FLDINCIMMED (0x01, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe",		FLDINCIMMED (0x00, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.nt1",	FLDINCIMMED (0x00, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.nta",	FLDINCIMMED (0x00, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.s",		FLDINCIMMED (0x06, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.s.nt1",	FLDINCIMMED (0x06, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.s.nta",	FLDINCIMMED (0x06, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.s",		FLDINCIMMED (0x07, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.s.nt1",	FLDINCIMMED (0x07, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.s.nta",	FLDINCIMMED (0x07, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.s",		FLDINCIMMED (0x05, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.s.nt1",	FLDINCIMMED (0x05, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.s.nta",	FLDINCIMMED (0x05, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.s",		FLDINCIMMED (0x04, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.s.nt1",	FLDINCIMMED (0x04, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.s.nta",	FLDINCIMMED (0x04, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.a",		FLDINCIMMED (0x0a, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.a.nt1",	FLDINCIMMED (0x0a, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.a.nta",	FLDINCIMMED (0x0a, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.a",		FLDINCIMMED (0x0b, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.a.nt1",	FLDINCIMMED (0x0b, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.a.nta",	FLDINCIMMED (0x0b, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.a",		FLDINCIMMED (0x09, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.a.nt1",	FLDINCIMMED (0x09, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.a.nta",	FLDINCIMMED (0x09, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.a",		FLDINCIMMED (0x08, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.a.nt1",	FLDINCIMMED (0x08, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.a.nta",	FLDINCIMMED (0x08, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.sa",		FLDINCIMMED (0x0e, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.sa.nt1",	FLDINCIMMED (0x0e, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.sa.nta",	FLDINCIMMED (0x0e, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.sa",		FLDINCIMMED (0x0f, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.sa.nt1",	FLDINCIMMED (0x0f, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.sa.nta",	FLDINCIMMED (0x0f, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.sa",		FLDINCIMMED (0x0d, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.sa.nt1",	FLDINCIMMED (0x0d, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.sa.nta",	FLDINCIMMED (0x0d, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.sa",		FLDINCIMMED (0x0c, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.sa.nt1",	FLDINCIMMED (0x0c, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.sa.nta",	FLDINCIMMED (0x0c, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf.fill",	FLDINCIMMED (0x1b, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf.fill.nt1",	FLDINCIMMED (0x1b, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf.fill.nta",	FLDINCIMMED (0x1b, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.clr",	FLDINCIMMED (0x22, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.clr.nt1",	FLDINCIMMED (0x22, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.clr.nta",	FLDINCIMMED (0x22, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.clr",	FLDINCIMMED (0x23, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.clr.nt1",	FLDINCIMMED (0x23, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.clr.nta",	FLDINCIMMED (0x23, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.clr",	FLDINCIMMED (0x21, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.clr.nt1",	FLDINCIMMED (0x21, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.clr.nta",	FLDINCIMMED (0x21, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.clr",	FLDINCIMMED (0x20, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.clr.nt1",	FLDINCIMMED (0x20, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.clr.nta",	FLDINCIMMED (0x20, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.nc",	FLDINCIMMED (0x26, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.nc.nt1",	FLDINCIMMED (0x26, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfs.c.nc.nta",	FLDINCIMMED (0x26, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.nc",	FLDINCIMMED (0x27, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.nc.nt1",	FLDINCIMMED (0x27, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfd.c.nc.nta",	FLDINCIMMED (0x27, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.nc",	FLDINCIMMED (0x25, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.nc.nt1",	FLDINCIMMED (0x25, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldf8.c.nc.nta",	FLDINCIMMED (0x25, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.nc",	FLDINCIMMED (0x24, 0), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.nc.nt1",	FLDINCIMMED (0x24, 1), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
    {"ldfe.c.nc.nta",	FLDINCIMMED (0x24, 3), IA64_INSTR_FORMAT_M8, IA64_INSTR_LDF, 
      IA64_OPCODE_POSTINC},
#undef FLDINCIMMED

    /* floating-point store w/increment by immediate */
#define FSTINCIMMED(c,h) OpX6aHint (7, c, h)
    {"stfs",		FSTINCIMMED (0x32, 0), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
    {"stfs.nta",	FSTINCIMMED (0x32, 3), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
    {"stfd",		FSTINCIMMED (0x33, 0), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
    {"stfd.nta",	FSTINCIMMED (0x33, 3), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
    {"stf8",		FSTINCIMMED (0x31, 0), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
    {"stf8.nta",	FSTINCIMMED (0x31, 3), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
    {"stfe",		FSTINCIMMED (0x30, 0), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
    {"stfe.nta",	FSTINCIMMED (0x30, 3), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
    {"stf.spill",	FSTINCIMMED (0x3b, 0), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
    {"stf.spill.nta",	FSTINCIMMED (0x3b, 3), IA64_INSTR_FORMAT_M10, IA64_INSTR_STF, 
      IA64_OPCODE_POSTINC},
#undef FSTINCIMMED

    /* line prefetch w/increment by immediate */
#define LFETCHINCIMMED(c,h) OpX6aHint (7, c, h)
    {"lfetch",			LFETCHINCIMMED (0x2c, 0), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.nt1",		LFETCHINCIMMED (0x2c, 1), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.nt2",		LFETCHINCIMMED (0x2c, 2), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.nta",		LFETCHINCIMMED (0x2c, 3), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.excl",		LFETCHINCIMMED (0x2d, 0), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.excl.nt1",		LFETCHINCIMMED (0x2d, 1), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.excl.nt2",		LFETCHINCIMMED (0x2d, 2), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.excl.nta",		LFETCHINCIMMED (0x2d, 3), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault",		LFETCHINCIMMED (0x2e, 0), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.nt1",	LFETCHINCIMMED (0x2e, 1), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.nt2",	LFETCHINCIMMED (0x2e, 2), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.nta",	LFETCHINCIMMED (0x2e, 3), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.excl",	LFETCHINCIMMED (0x2f, 0), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.excl.nt1",	LFETCHINCIMMED (0x2f, 1), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.excl.nt2",	LFETCHINCIMMED (0x2f, 2), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
    {"lfetch.fault.excl.nta",	LFETCHINCIMMED (0x2f, 3), IA64_INSTR_FORMAT_M15, IA64_INSTR_LFETCH, 
      IA64_OPCODE_POSTINC},
#undef LFETCHINCIMMED

    {"Unsupported", 0x0ULL, 0x0ULL, IA64_INSTR_FORMAT_UNSUPPORTED, IA64_INSTR_UNSUPPORTED}
  };

#undef bM
#undef bX
#undef bX2
#undef bX3
#undef bX4_
#undef bX6a
#undef bX6b
#undef bHint
#undef mM
#undef mX
#undef mX2
#undef mX3
#undef mX4
#undef mX6a
#undef mX6b
#undef mHint
#undef OpX3
#undef OpX3X6b
#undef OpX3X4
#undef OpX3X4X2
#undef OpX6aHint
#undef OpXX6aHint
#undef OpMXX6a
#undef OpMXX6aHint


/* vim: set shiftwidth=2: */
