/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_opt_special.h"
#include "../../utils/util_error.h"


#include <stdlib.h>
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_function.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_liveness.h"
#include "../../kernel/diablo_constprop.h"
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_iterators.h"

#include "ia64_opcodes.h"
#include "ia64_utils.h"
#include "ia64_dummy.h"
#include "ia64_instructions.h"
#include "ia64_opcode_label.h"



#include "../../utils/util_options.h"


/*OPGELET: ik gebruik een vlagje om de "stoute" functies mee te markeren. 
* stoute functies zijn deze die de branchregisters ook voor andere dingen gebruiken.
* */

/*ik gebruik meer bepaald het veld function_flags uit t_function */
/*het gevaar zit hem in overlap...*/

#define FF_IA64_BAD_FUNCTION_FOR_BRANCHES 0x80

/* ******** */


static t_bool Ia64IsBadFunctionForBranches(t_function *fun)
{
  return (FUN_FLAGS(fun) & FF_IA64_BAD_FUNCTION_FOR_BRANCHES) != 0;
}

static void Ia64SetAsBadFunctionForBranches(t_function *fun)
{
  FUN_FLAGS(fun) |= FF_IA64_BAD_FUNCTION_FOR_BRANCHES;
}


static t_bool Ia64GebruiktAndereBranchRegs(t_function * i_fun, t_bool *_heeftBranchRegs)
{
  t_bool heeftBranchRegs = FALSE;
  t_bool gebruiktAndereBranchRegs = FALSE;
  t_bbl *i_bbl;

  FUN_FOREACH_BBL(i_fun,i_bbl)
  {
    t_ins *ins;

    /*skip dangerous functions...*/
    if(i_bbl->ins_last && Ia64IsBranchOrCall(i_bbl->ins_last))
    {
      if(IA64_IS_BR(IA64INSIN1(i_bbl->ins_last)))
      {
	/*CHANGED !!!!!*/
	Ia64SetAsBadFunctionForBranches(i_fun);
        gebruiktAndereBranchRegs = TRUE;
	
        continue;
      }
    }


    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(IA64_IS_BR(IA64INSIN1(ins)))
      {
	heeftBranchRegs = TRUE;

	if(IA64INSIN1(ins) != IA64_B0 && IA64INSIN1(ins) != IA64_B6)
	{
	  Ia64SetAsBadFunctionForBranches(i_fun);
	  gebruiktAndereBranchRegs = TRUE;
	  break;
	}
      }
      if(IA64_IS_BR(IA64INSOUT1(ins)))
      {
	heeftBranchRegs = TRUE;

	if(IA64INSOUT1(ins) != IA64_B0 && IA64INSOUT1(ins) != IA64_B6)
	{
	  Ia64SetAsBadFunctionForBranches(i_fun);
	  gebruiktAndereBranchRegs = TRUE;
	  break;
	}
      }

    }
  }

  if(_heeftBranchRegs != NULL)
    *_heeftBranchRegs = heeftBranchRegs;
  
  return gebruiktAndereBranchRegs;
}


/*indirecte branches en calls worden nu a.h.v. b6 gedaan, althans in gcc...*/
/*b6 is eerste scratchregister*/
static void Ia64UseB7AsIndirectRegister(t_cfg *cfg)
{
  t_function * i_fun, *tmp;
  
  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    t_bbl * i_bbl;
    t_bool heeftBranchRegs = FALSE;
    t_bool gebruiktAndereBranchRegs = FALSE;

    gebruiktAndereBranchRegs = Ia64GebruiktAndereBranchRegs(i_fun, &heeftBranchRegs);

    if(heeftBranchRegs && !gebruiktAndereBranchRegs)
    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      t_ins *ins;

      BBL_FOREACH_INS(i_bbl, ins)
      {
	if(IA64INSOUT1(ins) == IA64_B6)
	{
	  IA64INSOUT1(ins) = IA64_B7;

	  INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
	  INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);
	}

	if(IA64INSIN1(ins) == IA64_B6)
	{
	  IA64INSIN1(ins) = IA64_B7;

	  INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
	  INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);
	}

      }
    }
  }  
}


t_uint32 Ia64TryToEliminateMovBr(t_function *fun, t_ia64_reg newLinkReg, t_uint32 *cntFun)
{
  t_bbl * bbl;

  t_uint32 gain = 0; /*this will be 0 or 2..or even more, if there are multiple restores.*/
  t_ins *save_br = NULL, *restore_br = NULL;
  t_bool multipleRestores = FALSE;
  t_uint32 eliminationOK = TRUE;

  t_uint8 incfuncnt = 0;

  /* dit is waarschijnlijk niet nodig, maar kan geen kwaad...komt toch niet veel voor*/
  if(Ia64IsBadFunctionForBranches(fun))
    return 0;

  FUN_FOREACH_BBL(fun,bbl)
  {
    t_ins *ins_last = bbl->ins_last;

    if(ins_last && Ia64IsReturn(ins_last))
    {
      if(IA64INSIN1(ins_last) != newLinkReg)
	incfuncnt = 1;
    }
  }

  (*cntFun)+= incfuncnt;

  FUN_FOREACH_BBL(fun, bbl)
  {
    t_ins *ins;
    t_ins *ins_last = bbl->ins_last;

    if(ins_last && Ia64IsCall(ins_last))
    {
      /*if there is still a call with bx as linkregister => cannot execute elimination of mov_br's yet...*/
      if(IA64INSOUT1(ins_last) == newLinkReg)
      {
	eliminationOK = FALSE;
	break;
      }
    }


    BBL_FOREACH_INS(bbl, ins)
    {
      if(IA64INSIN1(ins) == newLinkReg && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_MOV_BR)
      {
	if(save_br != NULL)
	{
	  /* dit gebeurt dus effectief meerdere keren ! bv in __libc_realloc: daar wordt b0 ook in een outputregister gestoken !!!*/
	  /* is dit om sneller te kunnen terugkeren? */
//	  printf("WARNING: b%d is being saved multiple times in function %s\n", newLinkReg - IA64_BR, fun->name);
	  eliminationOK = FALSE;
	}
	else
	  save_br = ins;
      }

      if(IA64INSOUT1(ins) == newLinkReg && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_MOV_BR)
      {
	/* ook dit komt voor => hier kijken we wel of de restore niet van hetzelfde register zou komen */
	if(restore_br != NULL)
	{
	  if(IA64INSOUT1(ins) == IA64INSOUT1(restore_br))
	  {
	    multipleRestores = TRUE;
//	    printf("WARNING: b%d is being restored multiple times in function %s, BUT from same register\n", newLinkReg - IA64_BR, fun->name);
	  }
	  else
	  {
//	    printf("WARNING: b%d is being restored multiple times in function %s, BUT NOT from the same register\n", newLinkReg - IA64_BR, fun->name);
	    eliminationOK = FALSE;
	  }
	}
	else
	  restore_br = ins;
      }

    }

    if(!eliminationOK)
      break;
  }

  if(eliminationOK)
  {
    if(save_br != NULL && restore_br != NULL)
    {
      InsKill(save_br);
      gain++;

      if(multipleRestores)
      {
	/* do it manually*/
	FUN_FOREACH_BBL(fun, bbl)
	{
	  t_ins *ins, *tmp;

	  BBL_FOREACH_INS_SAFE(bbl, ins, tmp)
	  {
	    if(IA64INSOUT1(ins) == newLinkReg && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_MOV_BR)
	    {
	      InsKill(ins);
	      gain++;
	    }
	  }
	}
      }
      else
      {
	InsKill(restore_br);
	gain++;
      }
    }
/*    
    else
      if(save_br != NULL && restore_br != NULL)
	printf("WARNING: found a save_br or restore_br but not both, in function %s", fun->name);
*/	
  }

  return gain;
}

static void Ia64OptimizeBranchRegistersCore(t_cfg *cfg)
{
  t_bool flag = TRUE;
  t_function * i_fun;
  t_function * tmp;
  t_bbl *i_bbl;
  t_cfg_edge *i_edge;
  
  t_uint32 cnt = 0;
  t_uint32 gain = 0;
  t_uint32 cntFun = 0;

  t_uint32 iterations = 0;

  t_bool changed = FALSE;


  do
  {
    changed = FALSE;

    CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
    {
      /*in het beste geval kunnen we b6 gebruiken (als het een leaf is bv...)*/
      t_ia64_reg maxBr = IA64_B6;

      flag = TRUE;

      /*logisch*/
      if(FUN_IS_HELL(i_fun))
	continue;

      /*skip the bad functions immediately !!!!*/
      if(Ia64IsBadFunctionForBranches(i_fun))
      {
//	printf("INFO in Ia64OptimizeBranchRegistersForLeaves: skipping bad function %s\n", i_fun->name);
	continue;
      }

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
	BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	{
	  if (CfgEdgeIsForwardInterproc(i_edge) && EDGE_CAT(i_edge)!=ET_CALL)
	    flag = FALSE;
	  else
	    if (EDGE_CAT(i_edge) == ET_CALL)
	    {
	      t_bbl *call_site = CFG_EDGE_HEAD(i_edge);
	      t_ins *ins_last = call_site->ins_last;


	      if (IsHellNode(EDGE_TAIL(i_edge)))
	      {
//		printf("INFO: cannot optimize function %s because of outgoing call to hell\n", i_fun->name);
		flag = FALSE;
		break;
	      }

	      if(ins_last && Ia64IsCall(ins_last))
	      {
		  /*kijken of het al dan niet indirecte call is...*/
		  if(IA64_IS_BR(IA64INSIN1(ins_last)))
		  {
		    flag = FALSE;
		    break;
		  }

		  maxBr = MAX(IA64_B0, MIN(maxBr, IA64INSOUT1(ins_last) - 1 ));
	      }
	      else
		FATAL(("expected br.call 1 !!!"));
	    }

	  /* zal niet voor deze functie zijn...*/
	  if(maxBr == IA64_B0)
	    flag = FALSE;
	}

	BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
	{
	  if (CfgEdgeIsForwardInterproc(i_edge))
	  {
	    if (EDGE_CAT(i_edge)!=ET_CALL)
	      flag = FALSE;
	    else if (IsHellNode(EDGE_HEAD(i_edge)))
	      flag = FALSE;
	  }
	}
      }

      if (!flag)
	continue;

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
	t_ins *ins_last = i_bbl->ins_last;

	if(ins_last && Ia64IsReturn(ins_last))
	{
	  if(IA64INSIN1(ins_last) != maxBr)
	  {
	    changed = TRUE;

	    IA64INSIN1(ins_last) = maxBr;

	    INS_REGS_USE(ins_last)=Ia64UsedRegisters(ins_last);
	    INS_REGS_DEF(ins_last)=Ia64DefinedRegisters(ins_last);
	  }
	}
      }

      if(!changed)
	continue;

//      printf("--> Optimizing branch-registers for function %s\n", i_fun->name);
      cnt++;

      /* we can now change the return-register very nicely into maxBr. We also have to change all
      * the registers at all call_sites for this function...*/

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
	BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
	{
	  if (CfgEdgeIsForwardInterproc(i_edge))
	  {
	    if (EDGE_CAT(i_edge) == ET_CALL)
	    {
	      t_bbl *call_site = CFG_EDGE_HEAD(i_edge);
	      t_ins *ins_last = call_site->ins_last;

	      if(ins_last && Ia64IsCall(ins_last))
	      {
		IA64INSOUT1(ins_last) = maxBr;

		INS_REGS_USE(ins_last)=Ia64UsedRegisters(ins_last);
		INS_REGS_DEF(ins_last)=Ia64DefinedRegisters(ins_last);



	      }
	      else
		FATAL(("expected br.call 2 !!!"));
	    }
	    else
	      FATAL(("This is not possible, as we filtered this case already out"));
	  }
	}
      }

    } 

    iterations++;

  } while (changed);

  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    /*and now the interesting part: check if it is possible to remove redundant save, restores of b0 !!!*/
    gain += Ia64TryToEliminateMovBr(i_fun, IA64_B0, &cntFun);
  }
  
  printf("Adjusted %d clean functions in %d iterations and %d redundant MOV_BR instructions could be eliminated.\n", cntFun, iterations, gain);
}




/*eerst alle functies die een inkomende call-pijl vanuit hell, of vanuit een bad-function hebben, markeren als bad function. */
static void Ia64OptimizeBranchRegistersMarkFromHell(t_cfg *cfg)
{  
  t_function * i_fun, *tmp;
  t_cfg_edge *i_edge;

  t_bool changed = FALSE, skip = FALSE;

  do 
  {
    changed = FALSE;

    CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
    {
      t_bbl * i_bbl;

      /* deze functie is al gemarkeerd */
      if(Ia64IsBadFunctionForBranches(i_fun))
	continue;

/*      
      if(i_fun->name && strcmp(i_fun->name, "main") == 0)
	continue;
*/
      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
	BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
	{
	  if (CfgEdgeIsForwardInterproc(i_edge))
	  {
	    if (EDGE_CAT(i_edge)!=ET_CALL)
	      skip = TRUE;
	    else 
	      if (IsHellNode(EDGE_HEAD(i_edge)) || Ia64IsBadFunctionForBranches(BBL_FUN(EDGE_HEAD(i_edge))))
	      {
/*		
		if(IsHellNode(EDGE_HEAD(i_edge)))
		  printf("Function %s has incoming hell-node\n", i_fun->name);
*/		
		changed = TRUE;
		Ia64SetAsBadFunctionForBranches(i_fun);
	      }
	  }

	}

      }  
    }
    
  } while(changed);

}



/* this function will try to exploit the full potential of the 8 branchregisters of the IA-64*/
void Ia64OptimizeBranchRegisters(t_cfg *cfg)
{
//  t_uint32 iteratie = 0;

  static t_bool alreadyDone = FALSE;

  /* kan evt verfijnd worden; wel rekening houden met inlining en zo !*/
  if(alreadyDone)
    return;

  alreadyDone = TRUE;
  
  printf("Starting Optimizing BranchRegisters\n");

  printf("Number of SAVE/RESTORE instructions: %d\n", Ia64CalcNinsWithOpcodeInCfg(cfg, IA64_INSTR_MOV_BR));

  /*eerst alle indirecte sprongen a.h.v. b7, ipv b6, kwestie van het een beetje op te ruimen...*/
  Ia64UseB7AsIndirectRegister(cfg);

  /*nu zijn b0 t.e.m. b6 ter beschikking; b0 is de default, b1->b6 zijn vrij te gebruiken, tenzij
  * in de "vervelende" functies...*/


  //werkt toch niet, door supermassale aantal hellpijlen...
  //Ia64OptimizeBranchRegistersMarkFromHell(cfg);
  
  
  printf("Starting Optimizing clean functions\n");
  Ia64OptimizeBranchRegistersCore(cfg);
  printf("Finished Optimizing clean functions\n");

  

  printf("Finished OK\n");
  
}





/*-----------------------------------------------------------------------------------------*/


t_uint32 Ia64UpdateMovBr(t_function *fun, t_ia64_reg newLinkReg, t_uint32 *winst, t_uint32 *verlies, t_uint32 *cntFun)
{
  t_bbl * bbl;
  
  t_uint32 localWinst = 0, localVerlies = 0;
  t_ins *save_br = NULL, *restore_br = NULL;
  t_bool multipleRestores = FALSE;
  t_uint32 eliminationOK = TRUE;

  /* dit is waarschijnlijk niet nodig, maar kan geen kwaad...komt toch niet veel voor*/

  /*TODO: evt hier ook extra saven !!!!*/
  
  if(Ia64IsBadFunctionForBranches(fun))
    return 0;

  (*cntFun)++;
  
  FUN_FOREACH_BBL(fun, bbl)
  {
    t_ins *ins;
    t_ins *ins_last = bbl->ins_last;

    if(ins_last && Ia64IsCall(ins_last))
    {
      /*if there is still a call with bx as linkregister => cannot execute elimination of mov_br's yet...*/
      if(IA64INSOUT1(ins_last) == newLinkReg)
      {
	(*cntFun)--;
	eliminationOK = FALSE;
	break;
      }
    }
    

    BBL_FOREACH_INS(bbl, ins)
    {
      if(IA64INSIN1(ins) == newLinkReg && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_MOV_BR)
      {
	if(save_br != NULL)
	{
	  /* dit gebeurt dus effectief meerdere keren ! bv in __libc_realloc: daar wordt b0 ook in een outputregister gestoken !!!*/
	  /* is dit om sneller te kunnen terugkeren? */
	  printf("WARNING: b%d is being saved multiple times in function %s\n", newLinkReg - IA64_BR, fun->name);
	  eliminationOK = FALSE;
	}
	else
  	  save_br = ins;
      }

      if(IA64INSOUT1(ins) == newLinkReg && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_MOV_BR)
      {
	/* ook dit komt voor => hier kijken we wel of de restore niet van hetzelfde register zou komen */
	if(restore_br != NULL)
	{
	  if(IA64INSOUT1(ins) == IA64INSOUT1(restore_br))
	  {
	    multipleRestores = TRUE;
	    printf("WARNING: b%d is being restored multiple times in function %s, BUT from same register\n", newLinkReg - IA64_BR, fun->name);
	  }
	  else
	  {
	    printf("WARNING: b%d is being restored multiple times in function %s, BBUT NOT from the same register\n", newLinkReg - IA64_BR, fun->name);
	    eliminationOK = FALSE;
	  }
	}
	else
	  restore_br = ins;
      }
      
    }

    if(!eliminationOK)
      break;
  }

  if(eliminationOK)
  {
    if(save_br != NULL && restore_br != NULL)
    {
      InsKill(save_br);
      localWinst++;
      
      if(multipleRestores)
      {
	/* do it manually*/
	 FUN_FOREACH_BBL(fun, bbl)
         {
	   t_ins *ins, *tmp;
	   
	   BBL_FOREACH_INS_SAFE(bbl, ins, tmp)
	   {
	     if(IA64INSOUT1(ins) == newLinkReg && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_MOV_BR)
	     {
	       InsKill(ins);
	       localWinst++;
	     }
	   }
	 }
      }
      else
      {
	InsKill(restore_br);
	localWinst++;
      }
    }
    else
      if(save_br != NULL && restore_br != NULL)
	printf("WARNING: found a save_br or restore_br but not both, in function %s", fun->name);
  }
  else
  {
    t_cfg_edge *i_edge;

    t_bool hasIncomingFromHell = FALSE;

    FUN_FOREACH_BBL(fun,bbl)
    {
      BBL_FOREACH_PRED_EDGE(bbl,i_edge)
	if (IsHellNode(EDGE_HEAD(i_edge)))
	{
	  hasIncomingFromHell = TRUE;
	  break;
	}

      if (hasIncomingFromHell)
	break;
    }

    if(hasIncomingFromHell)
    {
      t_uint8 i = 0;
      t_uint8 newTmps = 0;
      
      
      /*kijken hoeveel bewaren...*/
      t_bool saveBr[8];

      for(i = 0; i < 8; i++)
//	if (i < (newLinkReg - IA64_BR)) /*merk op dat we newLinkReg er niet bij nemen aangezien deze standaard al bewaard wordt */
//	  saveBr[i] = TRUE;
//        else
	  saveBr[i] = FALSE;

      FUN_FOREACH_BBL(fun,bbl)
      {
	if(bbl->ins_last && Ia64IsCall(bbl->ins_last) && IA64INSOUT1(bbl->ins_last) < newLinkReg && !saveBr[IA64INSOUT1(bbl->ins_last)-IA64_BR])
	{
	  newTmps++;
	  localVerlies+=2;
	  saveBr[IA64INSOUT1(bbl->ins_last)-IA64_BR] = TRUE;
	}
      }
	    
      printf("Todo: bewaren van ");
      for(i = 0; i < 8; i++)
	if (saveBr[i])
	  printf("b%d", i);

      printf("\n");
    }
  }

  *winst += localWinst;
  *verlies += localVerlies;
  
  return localWinst - localVerlies;
}




static void Ia64OptimizeImprovedBranchRegistersCore(t_cfg *cfg, t_ia64_reg newLinkReg)
{
  t_bool flag = TRUE;
  t_function * i_fun;
  t_function * tmp;
  t_bbl *i_bbl;
  t_cfg_edge *i_edge;

  t_uint32 cnt = 0;
  t_uint32 gain = 0;
  t_uint32 brutowinst = 0, verlies = 0;
  
  t_uint32 iterations = 0;
  t_uint32 cntFun = 0;

  t_bool changed = FALSE;


  do
  {
    changed = FALSE;

    CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
    {
      /*in het beste geval kunnen we b6 gebruiken (als het een leaf is bv...)*/
      t_ia64_reg maxBr = IA64_B6;

      flag = TRUE;

      /*logisch*/
      if(FUN_IS_HELL(i_fun))
	continue;

      /*skip the bad functions immediately !!!!*/
      if(Ia64IsBadFunctionForBranches(i_fun))
      {
	printf("INFO in Ia64OptimizeBranchRegistersForLeaves: skipping bad function %s\n", i_fun->name);
	continue;
      }

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
	BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	{
	  if (CfgEdgeIsForwardInterproc(i_edge) && EDGE_CAT(i_edge)!=ET_CALL)
	    flag = FALSE;
	  else
	    if (EDGE_CAT(i_edge) == ET_CALL)
	    {
	      t_bbl *call_site = CFG_EDGE_HEAD(i_edge);
	      t_ins *ins_last = call_site->ins_last;

/*
	      if (IsHellNode(EDGE_TAIL(i_edge)))
	      {
		printf("INFO: cannot optimize function %s because of outgoing call to hell\n", i_fun->name);
		flag = FALSE;
		break;
	      }
*/
	      if(ins_last && Ia64IsCall(ins_last))
	      {
		/*kijken of het al dan niet indirecte call is...*/
/*		
		if(IA64_IS_BR(IA64INSIN1(ins_last)))
		{
		  flag = FALSE;
		  break;
		}
*/
		maxBr = MAX(IA64_B0, MIN(maxBr, IA64INSOUT1(ins_last) - 1 ));
	      }
	      else
		FATAL(("expected br.call 1 !!!"));
	    }

	  /* zal niet voor deze functie zijn...*/
	  if(maxBr == newLinkReg)
	    flag = FALSE;
	}

	BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
	{
	  if (CfgEdgeIsForwardInterproc(i_edge))
	  {
	    if (EDGE_CAT(i_edge)!=ET_CALL)
	      flag = FALSE;
	    else if (IsHellNode(EDGE_HEAD(i_edge)))
	      flag = FALSE;
	  }
	}
      }

      if (!flag)
	continue;

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
	t_ins *ins_last = i_bbl->ins_last;

	if(ins_last && Ia64IsReturn(ins_last))
	{
	  if(IA64INSIN1(ins_last) != maxBr)
	  {
	    changed = TRUE;

	    IA64INSIN1(ins_last) = maxBr;

	    INS_REGS_USE(ins_last)=Ia64UsedRegisters(ins_last);
	    INS_REGS_DEF(ins_last)=Ia64DefinedRegisters(ins_last);
	  }
	}
      }

      if(!changed)
	continue;

      printf("--> Optimizing branch-registers for function %s\n", i_fun->name);
      cnt++;

      /* we can now change the return-register very nicely into maxBr. We also have to change all
      *       * the registers at all call_sites for this function...*/

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
	BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
	{
	  if (CfgEdgeIsForwardInterproc(i_edge))
	  {
	    if (EDGE_CAT(i_edge) == ET_CALL)
	    {
	      t_bbl *call_site = CFG_EDGE_HEAD(i_edge);
	      t_ins *ins_last = call_site->ins_last;

	      if(ins_last && Ia64IsCall(ins_last))
	      {
		IA64INSOUT1(ins_last) = maxBr;

		INS_REGS_USE(ins_last)=Ia64UsedRegisters(ins_last);
		INS_REGS_DEF(ins_last)=Ia64DefinedRegisters(ins_last);



	      }
	      else
		FATAL(("expected br.call 2 !!!"));
	    }
	    else
	      FATAL(("This is not possible, as we filtered this case already out"));
	  }
	}
      }

    } 

    iterations++;

  } while (changed);

  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    /*and now the interesting part: check if it is possible to remove redundant save, restores of bx !!!*/
//    gain += Ia64TryToEliminateMovBr(i_fun, newLinkReg);
    gain += Ia64UpdateMovBr(i_fun, newLinkReg, &brutowinst, &verlies, &cntFun);
  }

  printf("Adjusted %d clean functions in %d iterations and %d redundant MOV_BR instructions could be eliminated [+%d, -%d].\n", cntFun, iterations, gain, brutowinst, verlies);
}


//vertaalt b0 naar newLinkReg
void Ia64UseBxAsStandardLinkregisters(t_cfg *cfg, t_ia64_reg newLinkReg)
{
  t_bbl *bbl;
  
  if(!IA64_IS_BR(newLinkReg))
    FATAL(("Het moet wel een branchregister zijn"));

  CFG_FOREACH_BBL(cfg, bbl)
  {
    t_ins *ins;

    BBL_FOREACH_INS(bbl, ins)
    {
      if(IA64INSOUT1(ins) == IA64_B0)
      {
	IA64INSOUT1(ins) = newLinkReg;

	INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
	INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);
      }

      if(IA64INSIN1(ins) == IA64_B0)
      {
	IA64INSIN1(ins) = newLinkReg;

	INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
	INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);
      }

    }

  }
}


/* this function will try to exploit the full potential of the 8 branchregisters of the IA-64*/
void Ia64OptimizeBranchRegisters2(t_cfg *cfg, t_ia64_reg newLinkReg)
{
  t_uint32 iteratie = 0;

  static t_bool alreadyDone = FALSE;

  /* kan evt verfijnd worden; wel rekening houden met inlining en zo !*/
  if(alreadyDone)
    return;

  alreadyDone = TRUE;

  FATAL(("Gelieve deze optimalisatie toch niet te gebruiken!"));
  
  printf("Starting Optimizing BranchRegisters with the Improved Algorithm\n");

  printf("Number of SAVE/RESTORE instructions: %d\n", Ia64CalcNinsWithOpcodeInCfg(cfg, IA64_INSTR_MOV_BR));

  /*eerst alle indirecte sprongen a.h.v. b7, ipv b6, kwestie van het een beetje op te ruimen...*/
  Ia64UseB7AsIndirectRegister(cfg);

  /*nu zijn b0 t.e.m. b6 ter beschikking; b0 is de default, b1->b6 zijn vrij te gebruiken, tenzij
  * in de "vervelende" functies...*/


  //werkt toch niet, door supermassale aantal hellpijlen...
  //Ia64OptimizeBranchRegistersMarkFromHell(cfg);
  
  newLinkReg = IA64_B3;
  
  Ia64UseBxAsStandardLinkregisters(cfg, newLinkReg);
  
  printf("Starting Optimizing clean functions\n");
  Ia64OptimizeImprovedBranchRegistersCore(cfg, newLinkReg);
  printf("Finished Optimizing clean functions\n");

  

  printf("Finished OK\n");
  
}



















/*-----------------------------------------------------------------------------------------*/


void Ia64OptimizeGpUsageExamineOPD(t_cfg *cfg)
{
  t_object *obj = CFG_OBJECT(cfg);
  
    /* first apply a patch for the opd entry that may exist for this bbl */
    t_section * opd=SectionGetFromObjectByName(obj,".opd");
    t_section * got=SectionGetFromObjectByName(obj,".got");

    /*see Conventions; p.13-1 13.1.2 */
    t_uint64 gp = G_T_UINT64(SECTION_GET_CADDRESS(got));


    if (opd)
    {
      t_uint64 *data = (t_uint64*)opd->data;
      t_uint32 opd_size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(opd));
      t_uint32 currsize =0;

      while(currsize < opd_size)
      {
	if(*data != 0)
	{
	  if(*data >> 60 == 0x4)
	  {
	    //printf("0x%llx\n", *data);
	    if(*(data+1) != gp)
	      FATAL(("OPD-gp-entry does not equal default gp !"));
	  }
	}
	
	currsize +=16;
	data +=2;
      }

    }
    else
      FATAL(("Section .opd NOT found !!!"));  
}



void Ia64OptimizeGpDeleteGPLoadsBeforeIndirectCalls(t_cfg *cfg)
{
  t_function *i_fun, *tmp;

  t_uint32 gain = 0;
  
  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    t_bbl * i_bbl;

    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      t_ins *ins_last = i_bbl->ins_last;
      
      /* search for indirect calls*/
      if(ins_last && IA64_IS_BR(IA64INSIN1(ins_last)) && Ia64IsCall(ins_last))
      {
	/* merk op dat deze code er wel van uitgaat dat r1 altijd de geldige waarde bevat => evt checken !!!*/

	t_ins *ins, *tmp;

	/*in eerste instantie enkel naar dezelfde bbl kijken...*/
	BBL_FOREACH_INS_R_SAFE(i_bbl, ins, tmp)
	{
	  if(IA64INSOUT1(ins) == IA64_R1 && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_LD)
	  {
	    InsKill(ins);
	    gain++;
	  }
	}
      }

    }  
  }

  
  printf("Removed %d redundant ld.8 r1=[rx] instructions before indirect calls.\n", gain);

}


void Ia64OptimizeGpDeleteGeneralGPLoads(t_cfg *cfg)
{
  t_function *i_fun, *tmp;

  t_uint32 gain = 0;
  
  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    t_bbl * i_bbl;

    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      t_ins *ins, *tmp2;

      BBL_FOREACH_INS_SAFE(i_bbl, ins, tmp2)
      {
	if(IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_LD && IA64INSIMM1(ins) == 0)
	{
	  if(IA64INSOUT1(ins) == IA64_R1)
	  {
	    InsKill(ins);
	    gain++;
	  }
	}
      }
    }      

  }

  
  printf("Removed %d redundant ld.8 r1=[rx] instructions.\n", gain);

}



t_bool Ia64PropagateAllUsagesOfGpInFunction(t_function *i_fun, t_bbl *startBbl, t_ins *startIns, t_ia64_reg reg)
{
  t_bbl *i_bbl;
  t_bool foundbbl = FALSE;
  t_bool foundins = FALSE;
  t_bool stop = FALSE;
  
  FUN_FOREACH_BBL(i_fun,i_bbl)
  {
    t_ins *ins;
    if(!foundbbl && i_bbl != startBbl)
      continue;
    else
      foundbbl = TRUE;

    if(i_bbl != startBbl && NODE_PRED_FIRST(i_bbl) != NODE_PRED_LAST(i_bbl))
      return FALSE;
      
    
//#if 0
    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(!foundins && ins != startIns)
	continue;
      else
	foundins = TRUE;

      /*wel deze instructie nog overslaan*/
      if(ins == startIns)
	continue;
      
      
      /*heeft nieuwe definitie => stop propageren*/
      /*wat als dit geprediceerd is???*/
//      if((IA64INSOUT1(ins) == reg || IA64INSOUT2(ins) == reg) && IA64INSPRED(ins) != IA64_P0)
      if(RegsetIn(INS_REGS_DEF(ins), reg)/* || IA64INSPRED(ins) != IA64_P0*/)
      {
	stop = TRUE;
	break;
      }
    }

    if (stop)
      break;
//#endif
  }

  foundbbl = FALSE;
  foundins = FALSE;

  FUN_FOREACH_BBL(i_fun,i_bbl)
  {
    t_ins *ins;
    if(!foundbbl && i_bbl != startBbl)
      continue;
    else
      foundbbl = TRUE;
    

    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(!foundins && ins != startIns)
	continue;
      else
	foundins = TRUE;

      /*wel deze instructie nog overslaan*/
      if(ins == startIns)
	continue;

      if(RegsetIn(INS_REGS_DEF(ins), reg))
	return TRUE;

      if(IA64INSIN1(ins) == reg)      
	IA64INSIN1(ins) = IA64_R1;

      if(IA64INSIN2(ins) == reg)      
	IA64INSIN2(ins) = IA64_R1;
      
      if(IA64INSIN3(ins) == reg)      
	IA64INSIN3(ins) = IA64_R1;
      

      INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
      INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);

      /*heeft nieuwe definitie => stop propageren*/
      /*wat als dit geprediceerd is???*/
/*      
      if((IA64INSOUT1(ins) == reg || IA64INSOUT2(ins) == reg))
	return TRUE;
*/
    }

  }
  
  
  return TRUE;
  
}




t_bool Ia64CheckInsIsSaveOfGP(t_function *i_fun, t_bbl *i_bbl, t_ins *ins)
{
  t_bbl *nextBbl = i_bbl;
  t_ins *restoreIns;
  
  /*simple analysis*/

  /*PATCH !!!!!!!*/
  if(i_fun->name && (strcmp(i_fun->name, "__do_global_dtors_aux") == 0 || strcmp(i_fun->name, "PolarPoint_InitClass") == 0))
    return FALSE;

//  return FALSE;
  return TRUE;

#if 0  
  while(nextBbl && nextBbl->ins_last && !Ia64IsCall(nextBbl->ins_last))
    nextBbl = BBL_NEXT(nextBbl);
  
  if(!nextBbl)
    return FALSE;

  nextBbl = BBL_NEXT(nextBbl);
  
  if(!nextBbl)
    return FALSE;
  
  restoreIns = nextBbl->ins_first;
  
  if(!restoreIns)
    return FALSE;
  
  if(IA64INSOUT1(restoreIns) == IA64_R1 && IA64INSIN1(restoreIns) == IA64INSOUT1(ins) &&
      IA64OPCODEINFO(restoreIns).instr_opcode == IA64_INSTR_ADD && IA64INSIMM1(restoreIns) == 0 && 
      IA64INSIN2(restoreIns) == IA64_NONE)
    return TRUE;
  
  return FALSE;
#endif  
}




void Ia64OptimizeGpDeleteGPMovs(t_cfg *cfg)
{
  t_function *i_fun, *tmp;

  t_uint32 gain1 = 0, gain2 = 0, opportunity = 0;

  
  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    t_bbl * i_bbl;
    
    t_bool isNiceFunction = !Ia64DoesThisFunctionAskForTrouble(i_fun);

    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      t_ins *ins, *tmp2;

      BBL_FOREACH_INS_SAFE(i_bbl, ins, tmp2)
      {
	/*zoek addl, adds met constante 0 => mov*/
	if(IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_ADD && IA64INSIMM1(ins) == 0 && 
	    IA64INSIN2(ins) == IA64_NONE)
	{
	  if(IA64INSOUT1(ins) == IA64_R1)
	  {
	    InsKill(ins);
	    gain1++;
	  }
	  else
	    if(IA64INSIN1(ins) == IA64_R1 && Ia64CheckInsIsSaveOfGP(i_fun, i_bbl, ins))
	    {
	      InsKill(ins);
	      gain2++;
	    }
	    else
#if 0
	  else
	    /*CHANGED*/
	    if(/*isNiceFunction && */ IA64INSIN1(ins) == IA64_R1 && IA64INSOUT1(ins) >= IA64_R2 && IA64INSOUT1(ins) <= IA64_R31)
	    {
		if(Ia64PropagateAllUsagesOfGpInFunction(i_fun, i_bbl, ins, IA64INSOUT1(ins)))
		{
//		  InsKill(ins);
		  gain2++;
		}
	    }
//#if 0	  /*CHANGED !!!!*/
	    else
	      /* werkt niet ... liveness werkt niet meer voor bzip2
	      * => misschien ietske intelligenter zijn...: ganse functie afscannen en alle gebruiken kopieren
	      * => zou evt al door copypropagatie kunnen gedaan zijn...
	      * */
	      if(FUN_ISALLOCVALID(i_fun) && 
		  IA64INSIN1(ins) == IA64_R1 && IA64INSOUT1(ins) >= IA64_R0 && 
		  IA64INSOUT1(ins) <= FUN_SOL(i_fun) && isNiceFunction)
	      {
		if(Ia64PropagateAllUsagesOfGpInFunction(i_fun, i_bbl, ins, IA64INSOUT1(ins)))
		{
		  //InsKill(ins);
		  gain2++;
		}
	      }
	      else
		/*CHANGED !!!!*/
//#if 0		
		if (IA64INSIN1(ins) == IA64_R1 && CountAllocsInFun(i_fun) == 0 && 
		    IA64INSOUT1(ins) >= IA64_R32 && IA64INSOUT1(ins) <= IA64_R127 && isNiceFunction)
		{
		  if(Ia64PropagateAllUsagesOfGpInFunction(i_fun, i_bbl, ins, IA64INSOUT1(ins)))
		  {
		    //InsKill(ins);
		    gain2++;
		  }

		}
		else
//#endif		  
#endif		  
		  if(IA64INSIN1(ins) == IA64_R1)
		  {
		    //InsPrint(ins, 0);
		    opportunity++;
		  }
	  //#endif	  
	  //#endif	  

	}
	
      }
      
    }  
  }

  
  printf("Removed %d redundant mov r1=rx instructions.\n", gain1);
  printf("Removed %d redundant mov rx=r1 instructions.\n", gain2);
  printf("There are still %d mov out=r1 instructions.\n", opportunity);

}

t_bool Ia64IsAddressFromSection(t_cfg *cfg, t_string sectionname, t_uint64 address)
{
  t_section * sec=SectionGetFromObjectByName(CFG_OBJECT(cfg),sectionname);
  if(sec == NULL)
    return FALSE;

  if(address < SECTION_GET_CADDRESS(sec))
    return FALSE;
      
  if(address >= SECTION_GET_CADDRESS(sec) + SECTION_CSIZE(sec))
    return FALSE;

  return TRUE;
}


/*offset: in het aantal bytes*/
/*dez functie heette vroeger FromGOT*/
t_bool Ia64ReadFromSection(t_section *section, t_uint64 offset, t_uint64 *value)
{
  /*hij leest blijkbaar ook vanaf de sectie volgend op got, namelijk sdata en zelfs verder !...*/
  if(offset >= SECTION_CSIZE(section))
  {
//    printf("Trying to read outside GOT section: offset %ld <-> size %ld !\n", offset, SECTION_SIZE(got));
    return FALSE;
  }
  
  *value = *(t_uint64 *)(((char*)section->data) + offset);

  return TRUE;
}


t_bool Ia64HasNearbyLoad(t_function *i_fun, t_bbl *startBbl, t_ins *startIns, t_ins **loadIns)
{
  t_ia64_reg loadReg = IA64INSOUT1(startIns);

  t_bbl *i_bbl;
  t_bool foundbbl = FALSE;
  t_bool foundins = FALSE;
//  t_ins *ins;
  
  FUN_FOREACH_BBL(i_fun,i_bbl)
  {
    t_ins *ins;
    
    if(!foundbbl && i_bbl != startBbl)
      continue;
    else
      foundbbl = TRUE;

#if 0
  /*warning: als over alle functions => dit weer uitschakelen !!!!!!!!*/
  i_bbl = startBbl;
#endif
  
    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(!foundins && ins != startIns)
	continue;
      else
	foundins = TRUE;

      if(ins == startIns)
	continue;

      /*volgende bbls mogen maar 1 binnenkomende pijl hebben*/
      if(i_bbl != startBbl)
      {
	t_cfg_edge *edge;
	t_uint32 cnt = 0;
	BBL_FOREACH_PRED_EDGE(i_bbl, edge)
	  cnt++;

	if(cnt > 1)
	  return FALSE;
      }

      if(IA64INSIN1(ins) == loadReg && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_LD)
      {
	if(IA64INSIMM1(ins) != 0)
	{
	  printf("WARNING: postinc !");
	  //InsPrint(ins, 0);
	  return FALSE;
	}
	else
	{
	  /* TODO: ook kijken of dit ook niet verder voor iets anders gebruikt wordt !!!!*/

	  *loadIns = ins;
	  
	  return TRUE;
	}
      }
      else
	if(IA64INSOUT1(ins) == loadReg) /* new definition => we are too late !!! */
	  return FALSE;
    }

  }

  return FALSE;
}



void Ia64PropagateReloc(t_cfg *cfg, t_ins *ins, t_ins *newIns)
{
  t_reloc *origReloc, *newReloc;
  
  t_reloc_ref *ref;

  t_bbl *bbl;

  t_section *got = SectionGetFromObjectByName(CFG_OBJECT(cfg),".got");


    for(ref=RELOCATABLE_REFERS_TO(ins); ref != NULL; ref=ref->next)
    {
      t_reloc *reloc = ref->rel;

      if(T_INS(REL_FROM(reloc)) == ins && REL_E_TO_RELOCATABLE(reloc) == T_RELOCATABLE(got) /*&& strcmp(REL_TO_RELOCATABLE(rel)->name, "GOT_ENTRY") == 0*/)
      {
	t_reloc_ref *ptr, *newRef;

/*	
	printf("Found reloc type %d addr 0x%llx for ins ", reloc->type, reloc->from->address);
	InsPrint(ins, 0);

	fflush(stdout);

	return;
*/	
/*
	InsPrint(T_INS(reloc->from), 0);
*/	
	origReloc = reloc;

	newReloc=RelocTableAddRelocToRelocatable(SECTION_OBJECT(INS_SECTION(ins))->reloc_table, origReloc->addend, T_RELOCATABLE(newIns), origReloc->from_offset, REL_TO_RELOCATABLE(origReloc), origReloc->to_offset, origReloc->hell, origReloc->edge, origReloc->corresponding, origReloc->e.sec, "s0000$", origReloc->first);
/*	
	newReloc=RelocTableAddRelocWithAddend(INS_SECTION(ins)->obj->reloc_table, origReloc->from, origReloc->offset,
	    origReloc->addend, origReloc->to, origReloc->type, origReloc->write_type, origReloc->from_type, origReloc->edge);
*/
	/*newReloc = RelocTableDupReloc(INS_SECTION(ins)->obj->reloc_table, origReloc);*/

//	T_INS(newReloc->from) = newIns;
//	reloc->type = newType;

	/*nu nog toevoegen aan juiste bbl */
/*	
	newRef = RelocRefNew(newReloc);
	
	
	ptr = RELOCATABLE_REFERS_TO(newIns); 
	
	if(ptr == NULL)
	  RELOCATABLE_REFERS_TO(newIns) = newRef; 
	else
	{
	  while(ptr->next != NULL)
	    ptr = ptr->next;

	  ptr->next = newRef;
	  newRef->prev = ptr->next;
	}
*/

	
	return;
      }
    }  
  
    DiabloPrint(stdout, "WARNING: Could not find reloc for ins @I\n", ins);
}



#if 0
void Ia64PropagateReloc(t_cfg *cfg, t_ins *ins, t_ins *newIns, t_reloc_types newType)
{
  t_reloc *origReloc, *newReloc;
  
  t_reloc_ref *ref;

  t_bbl *bbl;


    for(ref=INS_REFERS_TO(ins)/*BBL_REFED_BY(INS_BBL(ins))*/; ref != NULL; ref=ref->next)
    {
      t_reloc *reloc = ref->rel;

      if(reloc->write_type == INSTRUCTION_IMMEDIATE22 && /*reloc->type != REPRESENT_LT_OPD_POINTER  &&*/ T_INS(reloc->from) == ins)
      {
	t_reloc_ref *ptr, *newRef;

/*	
	printf("Found reloc type %d addr 0x%llx for ins ", reloc->type, reloc->from->address);
	InsPrint(ins, 0);

	fflush(stdout);

	return;
*/	
/*
	InsPrint(T_INS(reloc->from), 0);
*/	
	origReloc = reloc;
	
	
	newReloc=RelocTableAddRelocWithAddend(INS_SECTION(ins)->obj->reloc_table, origReloc->from, origReloc->offset,
	    origReloc->addend, origReloc->to, origReloc->type, origReloc->write_type, origReloc->from_type, origReloc->edge);
	/*newReloc = RelocTableDupReloc(INS_SECTION(ins)->obj->reloc_table, origReloc);*/

	T_INS(newReloc->from) = newIns;
	reloc->type = newType;

	/*nu nog toevoegen aan juiste bbl */
	newRef = RelocRefNew(newReloc);
	
	
	ptr = INS_REFERS_TO(newIns); //BBL_REFED_BY(INS_BBL(newIns));
	
	if(ptr == NULL)
	  INS_REFERS_TO(newIns) = newRef; //BBL_REFED_BY(INS_BBL(newIns)) = newRef;
	else
	{
	  while(ptr->next != NULL)
	    ptr = ptr->next;

	  ptr->next = newRef;
	}


	
	return;
      }
    }  
  
    printf("WARNING: Could not find reloc <new: %s> for ins ", 
	newType == REPRESENT_GOTREL_VALUE? "GOT" : "RODATA"
	); 
    InsPrint(ins, 0);
}
#endif

#define Ia64ValueCanBeCodeAddress(addr) ((addr)>>60 == 0x4)
#define Ia64ValueCanBeDataAddress(addr) ((addr)>>60 == 0x6)

void Ia64OptimizeGotLoadsViaGp(t_cfg *cfg, t_bool do_code_addresses)
{
  t_section * got=SectionGetFromObjectByName(CFG_OBJECT(cfg),".got");
  t_section * rodata=SectionGetFromObjectByName(CFG_OBJECT(cfg),".rodata");
  t_function *i_fun, *tmp;

  t_uint64 gp, r3;

  t_uint32 totalCount = 0, gotCount = 0, dataCount = 0, 
  codeCount = 0, valueCount = 0, nearbyLoadCount = 0, dangerCount = 0, loadCount = 0,
  rodataCount = 0, rodataLoadCount = 0, todoCount = 0;

  t_uint32 datagotcount= 0, datadatacount = 0, datasdatacount = 0, databsscount = 0, 
    datasbsscount = 0, dataothercount = 0;

  if(got == NULL)
    FATAL(("Couldn't find GOT-section"));

  gp = G_T_UINT64(SECTION_GET_CADDRESS(got));

  r3 = G_T_UINT64(SECTION_GET_CADDRESS(rodata));



  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    t_bbl * i_bbl;

/*    
    t_bool isNiceFunction = !Ia64DoesThisFunctionAskForTrouble(i_fun);

    if(!isNiceFunction)
      continue;
*/

    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      t_ins *ins;

      BBL_FOREACH_INS(i_bbl, ins)
      {
	t_ins *loadIns;

	/*zoek addl, adds, met r1 als input*/
	if(IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_ADD && IA64INSIMM1(ins) != 0 &&
	    IA64INSIN2(ins) == IA64_NONE && IA64INSIN1(ins) == IA64_R1 && IA64INSPRED(ins) == IA64_P0 )
	{
	  t_ia64_reg gotAddrReg = IA64INSOUT1(ins);
	  t_uint64 gotOffset = IA64INSIMM1(ins);

	  t_uint64 value;

	  totalCount++;

	  if(Ia64ReadFromSection(got, gotOffset, &value))
	  {


	    gotCount++;

	    /*
	       printf("Value read: 0x%llx for instruction ", value);
	       InsPrint(ins, 0);
	     */

	    if(gotAddrReg >= IA64_R32)
	    {
	      /*problemen, zoals elders hierboven*/
	      if(!(FUN_ISALLOCVALID(i_fun) && gotAddrReg <= FUN_SOL(i_fun)))
	      {
		/*zou evt kunenn gebruikt worden als param; hoewel: ik zou niet weten waarvoor dit nuttig zou zijn 
		* => waarschijnlijk mogen deze toch ook gedaan worden... */
#if 0		
		printf("Possible danger at ");
		InsPrint(ins, 0);
#endif		
		dangerCount++;

		/*deze voorlopig overslaan*/
		/*CHANGED*/
		//continue;
	      }
	    }

	    /*todo: zie Ia64HasNearbyLoad: kijken of dat register verder niet gebruikt wordt*/
	    /*todo: kijken of de predicaten wel overeenkomen*/
	    if(Ia64HasNearbyLoad(i_fun, i_bbl, ins, &loadIns))
	    {
	      nearbyLoadCount++;

	      if(Ia64ValueCanBeCodeAddress(value))
	      {
/*		
		if(value < G_T_UINT64(SECTION_GET_CADDRESS(rodata)))
		  printf("Init-section code address 0x%llx @ ", value);
		else
		  printf("Other-section code address 0x%llx @ ", value);
*/		  
		if(value >= G_T_UINT64(SECTION_GET_CADDRESS(rodata)) && do_code_addresses)
		{
		  t_uint64 rovalue = 0x0123456789ABCDEFULL;

		  if(Ia64ReadFromSection(rodata, value - G_T_UINT64(SECTION_GET_CADDRESS(rodata)), &rovalue))
		  {
		    t_ins *rodataLoadIns = NULL;
		    

		    rodataCount++;

		    Ia64InstructionMakeAdd(loadIns, TRUE, IA64INSOUT1(loadIns), (t_int64)value - (t_int64)r3, IA64_R3);
		    /*evt zelfs nog stap verder zetten: ook data van rodata al inlezen !!!*/

		    /*now also propagate the relocation so that blocks wont be killed */
		    //FATAL(("Reimplement"));
		    
		    Ia64PropagateReloc(cfg, ins, loadIns);
		    //Ia64PropagateReloc(cfg, ins, loadIns, REPRESENT_RODATAREL_VALUE);
		    

		    
		    if(Ia64HasNearbyLoad(i_fun, INS_BBL(loadIns), loadIns, &rodataLoadIns))
		    {
		      //InsPrint(loadIns, 0);
		      //InsPrint(rodataLoadIns, 0);
		      rodataLoadCount++;
		    }

		    loadCount++;

		    
		    //printf(".rodata address 0x%llx @ ", value);
		    //InsPrint(ins, 0);
		    //printf("%lld\n", value - G_T_UINT64(SECTION_GET_ADDRESS(rodata)));

		    //printf("value in .rodata: 0x%llx\n", rovalue);
		    
		    
		  }
		  
		}
		else
		  todoCount++;
		//InsPrint(ins, 0);
		
		codeCount++;
	      }
	      else if(Ia64ValueCanBeDataAddress(value) /*&& value >= gp*/)
	      {
		if(Ia64IsAddressFromSection(cfg, ".got", value))
		  datagotcount++;
		else if (Ia64IsAddressFromSection(cfg, ".data", value))
		  datadatacount++;
		else if (Ia64IsAddressFromSection(cfg, ".sdata", value))
		  datasdatacount++;
		else if (Ia64IsAddressFromSection(cfg, ".sbss", value))
		  datasbsscount++;
		else if (Ia64IsAddressFromSection(cfg, ".bss", value))
		  databsscount++;
		else
		  dataothercount++;
		
		loadCount++;
		
		Ia64InstructionMakeAdd(loadIns, TRUE, IA64INSOUT1(loadIns), (t_int64)value - (t_int64)gp, IA64_R1);

		/*now also propagate the relocation so that blocks wont be killed */
//		FATAL(("Reimplement"));
		
		Ia64PropagateReloc(cfg, ins, loadIns);
		//Ia64PropagateReloc(cfg, ins, loadIns, REPRESENT_GOTREL_VALUE);

		dataCount++;
	      }
	      else
	      {
#if 0		
		printf("Value encountered: 0x%llx at ", value);
		InsPrint(ins, 0);
#endif		
		valueCount++;

		if(value == 0)
		{
		  /*liveness zal deze evt wel wissen !!! => is correctere modellering*/
		  //InsKill(ins);

		  Ia64InstructionMakeAdd(loadIns, TRUE, IA64INSOUT1(loadIns), 0, IA64_R0);

		/*now also propagate the relocation so that blocks wont be killed */
//		FATAL(("Reimplement"));
		
		Ia64PropagateReloc(cfg, ins, loadIns);
		//Ia64PropagateReloc(cfg, ins, loadIns, REPRESENT_GOTREL_VALUE);

		  
		  loadCount++;
		}
		else
		  todoCount++;
	      }

	    }
	  }
	}
      }
    }
  }    
  
  printf("\n");
  printf("--  GOT statistics  --\n");
  printf("Total:          %6d\n", totalCount);
  printf("GOT-hits:       %6d\n", gotCount);
  printf("Code-addresses: %6d\n", codeCount);
  printf("Data-addresses: %6d\n", dataCount);
  printf("-> .got:        %6d\n", datagotcount);
  printf("-> .data:       %6d\n", datadatacount);
  printf("-> .sdata:      %6d\n", datasdatacount);
  printf("-> .bss:        %6d\n", databsscount);
  printf("-> .sbss:       %6d\n", datasbsscount);
  printf("-> other:       %6d\n", dataothercount);
  printf("Values:         %6d\n", valueCount);
  printf("Rodatas:        %6d\n", rodataCount);
  printf("Effective loads:%6d\n", nearbyLoadCount);
  printf("Rodata loads:   %6d\n", rodataLoadCount);
  printf("Possible danger:%6d\n", dangerCount);
  printf("Reduced loads  :%6d\n", loadCount);
  printf("Todo loads     :%6d\n", todoCount);
  printf("--  END statistics  --\n");
  printf("\n");
}



void Ia64CheckGpRels(t_cfg *cfg)
{
  t_function *i_fun;
  t_bbl *i_bbl;

  t_uint32 cnt = 0;

return;

  CFG_FOREACH_FUN(cfg, i_fun)
  {
    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      t_ins *ins, *tmp;
      t_bool changed = FALSE;

      BBL_FOREACH_INS_SAFE(i_bbl, ins, tmp)
      {
       if(IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_ADD && IA64INSIMM1(ins) != 0 &&
            IA64INSIN2(ins) == IA64_NONE && IA64INSIN1(ins) != IA64_R1)
       {
	//if(INS_REFERS_TO(ins) != NULL)
	{
		InsPrint(ins, 0);
		cnt++;
	}
       }
      }
    }
  }

  printf("found: %d\n", cnt);

}



/*copyright D. Copperfield */
void Ia64MakeR3Disappear(t_cfg *cfg)
{
  t_function *i_fun;
  t_bbl *i_bbl;

  t_ia64_reg newReg = IA64_R13;

  CFG_FOREACH_FUN(cfg, i_fun)
  {

    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      t_ins *ins, *tmp;
      t_bool changed = FALSE;

      BBL_FOREACH_INS_SAFE(i_bbl, ins, tmp)
      {
	if(IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_ADD && IA64INSIMM1(ins) != 0 &&
	    IA64INSIN2(ins) == IA64_NONE && IA64INSIN1(ins) == IA64_R3)
	{
	  t_int64 imm = IA64INSIMM1(ins);
	  t_bool shrink = FALSE;
	  
	  if(imm >= 0)
	  {
	    if(imm >= (1<<13))
	      shrink = TRUE;
	  }
	  else
	  {
	    if(-imm > (1<<13))
	      shrink = TRUE;
	  }

	  if(shrink)
	  {
	    t_ins *newIns;

	    IA64INSIN1(ins) = IA64_R0;


	    INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
	    INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);

	    

	    /* nu nog nieuwe instructie bijmaken !*/

	    newIns = Ia64InstructionCreateNoop(FUN_SECTION(i_fun), IA64_UNIT_M);
	    IA64INSINSTRGROUP(newIns) = GetCurrentInstrGroupAndInc();
	    IA64INSPRED(newIns) = IA64INSPRED(ins); /*erf predicaat*/
	    IA64INSTYPE(newIns) = IA64_TYPE_A;
	    IA64INSOPCODE(newIns) = IA64_ADD_OPCODE;
	    IA64INSOUT1(newIns) = IA64INSOUT1(ins);
	    IA64INSIN1(newIns) = IA64INSOUT1(ins);
	    IA64INSIN2(newIns) = IA64_R3;
	    IA64INSIMM1(newIns) = imm;

	    if(INS_NEXT(ins))
	    	INS_PREV(INS_NEXT(ins)) = newIns;
	    else
	    {
	      BBL_INS_LAST(INS_BBL(ins)) = newIns;
	    }
	    
	    INS_NEXT(newIns) = INS_NEXT(ins);
	    INS_PREV(newIns) = ins;
	    INS_NEXT(ins) = newIns;

	    INS_BBL(newIns) = INS_BBL(ins);
	    INS_SECTION(newIns) = INS_SECTION(ins);


	    INS_REGS_USE(newIns)=Ia64UsedRegisters(newIns);
	    INS_REGS_DEF(newIns)=Ia64DefinedRegisters(newIns);

	    BBL_NINS(i_bbl)++;

	    printf("************** ");
	    //InsPrint(newIns,0);

	    tmp = newIns;


	  }
	  else
	  {
	    IA64INSOPCODE(ins) = GetAddsOpcode();
	    changed = TRUE;
	  }
	  
	}

	/*Small patch: initial definition of r4 to 0 is not necessary (used in function _start)*/
/*	
	if(IA64INSOUT1(ins) == IA64_R4 && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_ADD && IA64INSIN1(ins) == IA64_R0)
	{

	  InsKill(ins);
	  continue;
	}
*/	
#if 0
	if(IA64INSIN1(ins) == IA64_R4 || IA64INSIN2(ins) == IA64_R4 || IA64INSIN3(ins) == IA64_R4 ||
	   IA64INSOUT1(ins) == IA64_R4 || IA64INSOUT2(ins) == IA64_R4)
	{
	  InsPrint(ins, 0);
		if(!FUN_ISALLOCVALID(i_fun) || FUN_SOF(i_fun)==96)
			printf("--> Dat wordt lastig !\n");
	  InsKill(ins);
	  continue;
	}
#endif

	if(IA64INSTYPE(ins) == IA64_TYPE_DUMMY)
	  continue;

	if(IA64INSIN1(ins) == IA64_R3)      
	{
	  changed = TRUE;
	  IA64INSIN1(ins) = newReg;
	}

	if(IA64INSIN2(ins) == IA64_R3)      
	{
	  changed = TRUE;
	  IA64INSIN2(ins) = newReg;
	}

	if(IA64INSIN3(ins) == IA64_R3)      
	{
	  changed = TRUE;
	  IA64INSIN3(ins) = newReg;
	}

	if(IA64INSOUT1(ins) == IA64_R3)      
	{
	  changed = TRUE;
	  IA64INSOUT1(ins) = newReg;
	}

	if(IA64INSOUT2(ins) == IA64_R3)      
	{
	  changed = TRUE;
	  IA64INSOUT2(ins) = newReg;
	}

	if(changed)
	{
	  INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
	  INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);
	}


      }

    }
  }  
}

void Ia64InsertDefinitionOfR3AtEntryPoint(t_cfg *cfg)
{
  t_bbl *entry = cfg->entries->entry_bbl;
  
  t_ins *newIns;
  t_ins *ins = entry->ins_first;

  t_section *rodata = SectionGetFromObjectByName(CFG_OBJECT(cfg),".rodata");

  /* nu nog nieuwe instructie bijmaken !*/

  newIns = Ia64InstructionCreateNoop(CFG_SECTION(cfg), IA64_UNIT_X);
  IA64INSINSTRGROUP(newIns) = IA64INSINSTRGROUP(ins);
  IA64INSPRED(newIns) = IA64_P0; 
  IA64INSTYPE(newIns) = IA64_TYPE_X;
  IA64INSOPCODE(newIns) = IA64_MOVL_OPCODE;
  IA64INSOUT1(newIns) = IA64_R3;
  IA64INSIMM1(newIns) = G_T_UINT64(SECTION_GET_CADDRESS(rodata)); /*start rodata opgeven*/

  INS_PREV(INS_NEXT(ins)) = newIns;
  INS_NEXT(newIns) = INS_NEXT(ins);
  INS_PREV(newIns) = ins;
  INS_NEXT(ins) = newIns;

  INS_BBL(newIns) = INS_BBL(ins);
  INS_SECTION(newIns) = INS_SECTION(ins);


  INS_REGS_USE(newIns)=Ia64UsedRegisters(newIns);
  INS_REGS_DEF(newIns)=Ia64DefinedRegisters(newIns);

  BBL_NINS(entry)++;

}




/*tweede optimalisatie: deze rond gebruik van gp */

/* this function will try to exploit the fact that there normally is only 1 value as gp */
void Ia64OptimizeGpUsage(t_cfg *cfg)
{
  t_uint32 iteratie = 0;

  static t_bool alreadyDone = FALSE;

  t_bool do_code_addresses = TRUE;

  /* kan evt verfijnd worden; wel rekening houden met inlining en zo !*/
  if(alreadyDone)
    return;

  alreadyDone = TRUE;
  
  printf("Starting Optimizing gp usage\n");

  printf("Number of loads before gp-optimization: %d\n", Ia64CalcNinsWithOpcodeInCfg(cfg, IA64_INSTR_LD));
  
  /*
  * Mogelijk optimalisaties:
  * 1) inladen van constante uit got
  *    
  *    addl rx = offset, r1
  *    ld   ry = [rx]
  *
  * 2) bij indirecte calls: inladen van r1 uit opd (eerst checken of alle r1's uit opd wel gelijk zijn...)
  *  
  *    ld8 r1 = [r14]
  *
  */
  
  Ia64OptimizeGpUsageExamineOPD(cfg);

  printf("  Starting Optimizing deleting gp-loads before indirect calls\n");
  Ia64OptimizeGpDeleteGPLoadsBeforeIndirectCalls(cfg);
  printf("  Finished Optimizing deleting gp-loads before indirect calls\n");

  printf("  Starting Optimizing deleting general gp-loads\n");
  Ia64OptimizeGpDeleteGeneralGPLoads(cfg);
  printf("  Finished Optimizing deleting general gp-loads\n");

  Ia64CheckGpRels(cfg);

  printf("  Starting Optimizing deleting gp-movs\n");
  Ia64OptimizeGpDeleteGPMovs(cfg);
  printf("  Finished Optimizing deleting gp-movs\n");

  Ia64CheckGpRels(cfg);

  printf("  Starting Optimizing eliminating GOT-loads via gp\n");
  if(do_code_addresses)
  {
    printf("  Making R3 disappear\n");
    Ia64MakeR3Disappear(cfg);
    Ia64InsertDefinitionOfR3AtEntryPoint(cfg);
  }
  Ia64OptimizeGotLoadsViaGp(cfg, do_code_addresses);
  printf("  Finished Optimizing eliminating GOT-loads via gp\n");
  
  printf("Finished OK\n");
  
}






/* vim: set shiftwidth=2: */
