/*
* Copyright 2001,2002: Bruno De Bus
*                      Bjorn De Sutter
*                      Dominique Chanet
*                      Ludo Van Put
*
*
* This file is part of Diablo (Diablo is a better link-time optimizer)
*
* Diablo is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* Diablo is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Diablo; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* Headers for profilegraph manipulation 
*
* THIS FILE IS PART OF DIABLO (KERNEL)
*
***********************************************************/
#include "ia64_prograph.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_iterators.h"

/*! Allocate a new control flow graph for section sec.*/ 

t_pfg * PfgNew()
{
  t_pfg * ret=Malloc(sizeof(t_pfg));
  GraphInit(T_GRAPH(ret),sizeof(t_pfg_node), sizeof(t_pfg_edge));
  return ret;
}

void PfgFree(t_pfg * graph)
{
  t_pfg_edge * i_edge, * i_tmp1;
  t_pfg_node * i_node, * i_tmp2;

  /* Remove all edges (these may be freed) */
  GRAPH_FOREACH_EDGE_SAFE(T_GRAPH(graph), T_EDGE(i_edge), T_EDGE(i_tmp1))
  {
    GraphRemoveEdge(T_GRAPH(graph), T_EDGE(i_edge));
  }

  /* Remove all nodes from the graph (nodes should not be freed) */
  GRAPH_FOREACH_NODE_SAFE(T_GRAPH(graph), T_EDGE(i_node), T_EDGE(i_tmp2))
  {
    GraphRemoveNodeNoFree(T_GRAPH(graph), T_NODE(i_node));
  }

  Free(T_GRAPH(graph));
}

t_pfg_node * PfgNewNode(t_pfg * graph, t_bbl * bbl)
{
  t_pfg_node * temp;
  VERIFY_FASSERT((T_GRAPH(graph))->node_size,("Can't use GraphNodeNew when node_size == 0")); 
  temp = Malloc((T_GRAPH(graph))->node_size);

  temp->bbl = bbl;
  temp->layout_last=temp;
  temp->layout_first=temp;
  temp->layout_next=NULL;
  temp->layout_prev=NULL;
  temp->breakable_next = TRUE;
  temp->breakable_prev = TRUE;
  temp->header = TRUE;
  GraphInitNode(T_GRAPH(graph),T_NODE(temp));
  return temp;
}

t_pfg_edge * PfgNewEdge(t_pfg * graph, t_pfg_node * head, t_pfg_node * tail, t_uint64 counter)
{
  t_pfg_edge * temp;

  if(head == NULL || tail == NULL)
    FATAL(("this shouldn't happen."));
  temp = Calloc(1,(T_GRAPH(graph))->edge_size);

  temp->counter=counter;

  EDGE_HEAD(T_EDGE(temp)) = T_NODE(head);
  EDGE_TAIL(T_EDGE(temp)) = T_NODE(tail);
  //	EdgeSetCategory(temp,edge_category);
  EDGE_NEXT(T_EDGE(temp)) = NULL;

  if (NODE_SUCC_LAST(T_NODE(head)))
  {
    EDGE_SUCC_NEXT(T_EDGE(temp))=NULL;
    EDGE_SUCC_PREV(T_EDGE(temp))=NODE_SUCC_LAST(T_NODE(head));
    EDGE_SUCC_NEXT(NODE_SUCC_LAST(T_NODE(head)))=T_EDGE(temp);
    NODE_SUCC_LAST(T_NODE(head))=T_EDGE(temp);
  }
  else
  {
    EDGE_SUCC_NEXT(T_EDGE(temp))=EDGE_SUCC_PREV(T_EDGE(temp))=NULL;
    NODE_SUCC_LAST(T_NODE(head))=NODE_SUCC_FIRST(T_NODE(head))=T_EDGE(temp);
  }


  if (NODE_PRED_LAST(T_NODE(tail)))
  {
    EDGE_PRED_NEXT(T_EDGE(temp))=NULL;
    EDGE_PRED_PREV(T_EDGE(temp))=NODE_PRED_LAST(T_NODE(tail));
    EDGE_PRED_NEXT(NODE_PRED_LAST(T_NODE(tail)))=T_EDGE(temp);
    NODE_PRED_LAST(T_NODE(tail))=T_EDGE(temp);
  }
  else
  {
    EDGE_PRED_NEXT(T_EDGE(temp))=EDGE_PRED_PREV(T_EDGE(temp))=NULL;
    NODE_PRED_LAST(T_NODE(tail))=NODE_PRED_FIRST(T_NODE(tail))=T_EDGE(temp);
  }


  // Automatically insert into the global edge list 
  if (T_GRAPH(graph)->edge_last) 
  {
    EDGE_NEXT(T_GRAPH(graph)->edge_last) = T_EDGE(temp);
    EDGE_PREV(T_EDGE(temp)) = T_GRAPH(graph)->edge_last;
    T_GRAPH(graph)->edge_last = T_EDGE(temp);
  } 
  else 
  {
    T_GRAPH(graph)->edge_first = T_GRAPH(graph)->edge_last = T_EDGE(temp);
    EDGE_PREV(T_EDGE(temp)) = NULL;
  }
  T_GRAPH(graph)->nedges++;
  return temp;
}
void PfgMergeByBbl(t_pfg * graaf, t_bbl * head_bbl, t_bbl * tail_bbl)
{
  t_pfg_node * head_node=NULL;
  t_pfg_node * tail_node=NULL;
  t_pfg_node * iter = (t_pfg_node *) GRAPH_NODE_FIRST(graaf);
  t_bool trouble = FALSE;
  do 
  {
    //if(iter==NULL) FATAL(("trouble in paradise"));
    if(BBL_OLD_ADDRESS(PFG_BBL(iter))==BBL_OLD_ADDRESS(head_bbl))
      head_node = iter;
    if(BBL_OLD_ADDRESS(PFG_BBL(iter))==BBL_OLD_ADDRESS(tail_bbl))
    {
      tail_node = iter;
    }
    iter=(t_pfg_node *)NODE_NEXT(iter);
  }while ((head_node == NULL || tail_node == NULL)&&iter!=NULL);
  if(head_node ==NULL || tail_node == NULL)
  {
    return;
  }
/*	  head_node = PfgNewNode(graaf, head_bbl);
	  BblPrint(head_bbl,0);
	  printf("PATCHING!!!!!!!!!!!!!!!!!!!!!!!!!\n");
  }
  if(tail_node ==NULL)
  {
	  tail_node = PfgNewNode(graaf, tail_bbl);
	  BblPrint(tail_bbl,0);
	  printf("PATCHING!!!!!!!!!!!!!!!!!!!!!!!!!\n");
  }*/
  PfgMergeByNodes(graaf, head_node, tail_node);
}
void PfgMergeByNodes(t_pfg * graaf, t_pfg_node * head, t_pfg_node * tail)
{
  t_pfg_edge * conn_edge=NULL, * i_edge;
  NODE_FOREACH_SUCC_EDGE(T_NODE(head),T_EDGE(i_edge))
  {
    if(PFG_TAIL(i_edge)==tail)
      conn_edge=i_edge;
//      FATAL(("PfgMergeByNodes bedoeld voor niet verbonden knopen (kettingen vereist door FT e.d.)"));
  }
  NODE_FOREACH_PRED_EDGE(T_NODE(head),T_EDGE(i_edge))
  {
    if(PFG_HEAD(i_edge)==tail)
      conn_edge = i_edge;
//      FATAL(("PfgMergeByNodes bedoeld voor niet verbonden knopen (kettingen vereist door FT e.d.)"));
  }
  /*edge = PfgNewEdge(graaf, head, tail, -1);*/
  head->layout_last->breakable_next = FALSE;
  tail->breakable_prev = FALSE;
  while (head->layout_prev)
    head = head->layout_prev;
  if (conn_edge == NULL)
  {
    //FATAL(("het gebeurt"));
    conn_edge = PfgNewEdge(graaf, head, tail, -1);
  }
  PfgMergeByEdge(graaf,conn_edge);
}


void PfgMergeByEdge(t_pfg * graaf, t_pfg_edge * the_edge)
{
  t_pfg_node * tail = PFG_TAIL(the_edge);
  t_pfg_node * head = PFG_HEAD(the_edge);
  t_pfg_node * iter;
  t_pfg_edge * edge;
  t_pfg_edge * corr_edge = NULL;
  t_pfg_node * node;
  int i=0;
  t_bbl * bbl;
  if(head ==NULL || tail == NULL)
    FATAL(("dat zal moeilijk gaan."));
  if(BBL_OLD_ADDRESS(PFG_BBL(tail))==BBL_OLD_ADDRESS(PFG_BBL(head)))
  {
    GraphRemoveEdge(T_GRAPH(graaf), T_EDGE(the_edge));
    return;
  }
//    printf("%llx - %llx\n", BBL_OLD_ADDRESS(PFG_BBL(head)),BBL_OLD_ADDRESS(PFG_BBL(tail)));
  tail->header = FALSE;
  
  GraphRemoveEdge(T_GRAPH(graaf), T_EDGE(the_edge));
  if(head->layout_first == tail->layout_first)
    return;//FATAL(("trouble\n"));
  
  tail->layout_first->layout_prev = head->layout_last;
  head->layout_last->layout_next = tail->layout_first;
  iter=head->layout_first;
  while(iter)
  {
    iter->layout_last = tail->layout_last;
    iter->layout_first = head->layout_first;
    iter = PFG_NODE_NEXT(iter);  
  }

  
  NODE_FOREACH_SUCC_EDGE(T_NODE(tail),T_EDGE(edge))
  {
    NODE_FOREACH_SUCC_EDGE(T_NODE(head), T_EDGE(corr_edge))
    {
      if(PFG_BBL(PFG_TAIL(corr_edge))==PFG_BBL(PFG_TAIL(edge)))
	break;
    }
    if(corr_edge == NULL)
    {
      NODE_FOREACH_PRED_EDGE(T_NODE(head), T_EDGE(corr_edge))
      {
	if(PFG_BBL(PFG_HEAD(corr_edge))==PFG_BBL(PFG_TAIL(edge)))
	  break;
      }
    }
    if(corr_edge == NULL){
      if(PFG_TAIL(edge)!=head)
	PfgNewEdge(graaf, head, PFG_TAIL(edge),PFG_COUNTER(edge));
      GraphRemoveEdge(T_GRAPH(graaf), T_EDGE(edge));
    }
    else 
    {
      PFG_COUNTER(corr_edge)+=PFG_COUNTER(edge);
      GraphRemoveEdge(T_GRAPH(graaf), T_EDGE(edge));
    }
  }
  NODE_FOREACH_PRED_EDGE(T_NODE(tail),T_EDGE(edge))
  {
    
    NODE_FOREACH_SUCC_EDGE(T_NODE(head), T_EDGE(corr_edge))
    {
      if(PFG_BBL(PFG_TAIL(corr_edge))==PFG_BBL(PFG_HEAD(edge)))
	break;
    }
    if(corr_edge == NULL)
    {
      NODE_FOREACH_PRED_EDGE(T_NODE(head), T_EDGE(corr_edge))
      {
	if(PFG_BBL(PFG_HEAD(corr_edge))==PFG_BBL(PFG_HEAD(edge)))
	  break;
      }
    }
    if(corr_edge == NULL)
    {
      if(PFG_HEAD(edge)!=head)
	PfgNewEdge(graaf, head, PFG_HEAD(edge),PFG_COUNTER(edge));
      GraphRemoveEdge(T_GRAPH(graaf), T_EDGE(edge));
    }
    else 
    {
      PFG_COUNTER(corr_edge)+=PFG_COUNTER(edge);
      GraphRemoveEdge(T_GRAPH(graaf), T_EDGE(edge));
    }
  }
}
void MeetAantalBblsPerCluster(t_cfg * cfg)
{
  int aantal_clusters=0, aantal_bbls=0;
  t_bbl * i_bbl, * j_bbl;
  t_cfg_edge * i_edge;
  t_bool voorgaand_doorvalpad;

  InitNodeMarking();
  
  CFG_FOREACH_BBL(cfg, i_bbl)
  {
    if(NodeIsMarked(i_bbl))
      continue;
    voorgaand_doorvalpad=FALSE;
    BBL_FOREACH_PRED_EDGE(i_bbl, i_edge)
    {
      if (EDGE_CAT(i_edge)==ET_FUNLINK ||
          EDGE_CAT(i_edge)==ET_SWILINK ||
          EDGE_CAT(i_edge)==ET_IPFUNLINK ||
          EDGE_CAT(i_edge)==ET_IPSWILINK ||
          EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
          EDGE_CAT(i_edge)==ET_IPFALLTHRU )
	voorgaand_doorvalpad = TRUE;
    }
    if(voorgaand_doorvalpad == TRUE)
      continue;
    
    aantal_clusters++;
  }
  CFG_FOREACH_BBL(cfg, i_bbl)
  {
    if(i_bbl->ins_first && IA64INSTYPE(i_bbl->ins_first) != IA64_TYPE_DATA)
      aantal_bbls ++;
  }
  printf("%f\n",aantal_bbls*1.0/aantal_clusters);    
  FATAL(("stop"));
}
void PfgClosestIsBest(t_pfg * graaf)
{
  
  t_pfg_edge * largest_edge;
  t_node * node;
  int i=0; 
  printf("start van PFGCLOSESTISBEST\n");
  while(T_GRAPH(graaf)->nedges!=0)
  {
//    printf("e: %d\n", T_GRAPH(graaf)->nedges);
//    printf("%d\n",i++);
/*    GRAPH_FOREACH_NODE(T_GRAPH(graaf),node)
    {
      if(((t_pfg_node *)node)->header == FALSE)
      {
	if(NODE_SUCC_FIRST(node) != NULL)
	  FATAL(("probleempje."));
	if(NODE_PRED_FIRST(node) != NULL)
	  FATAL(("probleempje2."));
      }
    }
*/
    largest_edge = PfgFindLargestEdge(graaf);
    
   // if(PFG_HEAD(largest_edge)==PFG_TAIL(largest_edge))
   //   FATAL(("dit is geen goed teken"));
    
    PfgMergeByEdge(graaf, largest_edge);
  }
}

t_pfg_edge * PfgFindLargestEdge(t_pfg * graaf)
{
  t_edge * ret = GRAPH_EDGE_FIRST(T_GRAPH(graaf));
  t_edge * edge;
  GRAPH_FOREACH_EDGE(T_GRAPH(graaf),edge)
  {
    if(PFG_COUNTER(edge)>PFG_COUNTER(ret))
      ret = edge;
  }
  return (t_pfg_edge *)ret;
}

void PfgAddEdge(t_pfg * graaf, t_bbl * bbl, t_uint64 address, t_uint64 counter)
{
  
  t_pfg_node * head=NULL;
  t_pfg_node * tail=NULL;
  t_pfg_node * iter =(t_pfg_node *) GRAPH_NODE_FIRST(graaf);
  do 
  {
    if(iter==NULL) FATAL(("trouble in paradise"));
    if(PFG_BBL(iter)==bbl)
      head = iter;
    if(BBL_OLD_ADDRESS(PFG_BBL(iter))==address)
    {
      tail = iter;
    }
    iter=(t_pfg_node *)NODE_NEXT(iter);
  }while (head == NULL || tail == NULL);
//  BblPrint(head->bbl, 0);
//  BblPrint(tail->bbl, 0);
  PfgNewEdge(graaf, head, tail, counter);
}

/* Maak gebruik van de profileringsinformatie van edges */
t_pfg * Ia64InitializeGraph(t_cfg * cfg)
{
  t_bbl * bbl;
  t_cfg_edge * edge;
  t_pfg * graaf;
  int teller=0;
  graaf = PfgNew();

  /*voeg knopen toe aan profileringsgraaf*/
  /*??????????????,waar dienen lege bbl's weer voor,????????????*/

  /*  CFG_FOREACH_BBL(cfg,bbl)
  {
    if(!IsHellNode(bbl) && (!bbl->ins_first ||
	IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA))
    {
      PfgNewNode(graaf,bbl);
    }
  }
  */
  
  CFG_FOREACH_BBL(cfg, bbl)
  {
    if(bbl->ins_first && IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)
      if(BBL_PROFILE_INFO(bbl)>0)
      {
	PfgNewNode(graaf,bbl);
      }
  }

  /*voeg bogen toe aan profileringsgraaf*/
  /*CFG_FOREACH_BBL(cfg,bbl)
  {
    if(!IsHellNode(bbl) && 
	bbl->ins_first && 
	IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)

      BBL_FOREACH_SUCC_EDGE(bbl,edge)
      {
	if ((EDGE_CAT(edge)==ET_JUMP || 
	      EDGE_CAT(edge)==ET_IPJUMP || 
	      EDGE_CAT(edge)==ET_CALL)&& !Ia64IsReturn(bbl->ins_last)&&!Ia64IsCheck(bbl->ins_last))
	{
	    PfgAddEdge(graaf, bbl, EDGE_TARGET(edge), EDGE_TAKEN(edge));
	}
      }
  }*/
  CFG_FOREACH_BBL(cfg,bbl)
  {
//    printf("i: %d\n",teller++);
    if(!IsHellNode(bbl) && 
	bbl->ins_first && 
	IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)

      BBL_FOREACH_SUCC_EDGE(bbl,edge)
      {
/*	if ((EDGE_CAT(edge)==ET_JUMP || 
	      EDGE_CAT(edge)==ET_IPJUMP || 
	      EDGE_CAT(edge)==ET_CALL)&& !Ia64IsReturn(bbl->ins_last)&&!Ia64IsCheck(bbl->ins_last))
	{*/
	  if(BBL_PROFILE_INFO(bbl)>0 && EDGE_TAKEN(edge)>0)
	    PfgAddEdge(graaf, bbl, EDGE_TARGET(edge), EDGE_TAKEN(edge));
/*	}
	if ((EDGE_CAT(edge)==ET_FUNLINK ||
	    EDGE_CAT(edge)==ET_SWILINK ||
	    EDGE_CAT(edge)==ET_IPFUNLINK ||
	    EDGE_CAT(edge)==ET_IPSWILINK ||
	    EDGE_CAT(edge)==ET_FALLTHROUGH ||
	    EDGE_CAT(edge)==ET_IPFALLTHRU))
	{*/
	  //printf("%llx, %d, %d, %d \n",BBL_OLD_ADDRESS(bbl),BBL_PROFILE_INFO(bbl),EDGE_TAKEN(edge),BBL_PROFILE_INFO(CFG_EDGE_TAIL(edge)));
	//  if(BBL_PROFILE_INFO(bbl)>0 && EDGE_TAKEN(edge)>0 /*&&BBL_OLD_ADDRESS(bbl)!=0x4000000000000240&&BBL_OLD_ADDRESS(bbl)!=0x4000000000001090*//*&& BBL_PROFILE_INFO(CFG_EDGE_TAIL(edge))>0*/)
/*	  {
	    //printf("%llx %llx %d\n",BBL_OLD_ADDRESS(bbl),EDGE_TARGET(edge),EDGE_TAKEN(edge));
	    PfgAddEdge(graaf, bbl, EDGE_TARGET(edge), EDGE_TAKEN(edge));
	  }
	}*/

      }
  }
  //FATAL(("stop"));
  return graaf;
}

void Ia64PrintPrograph(t_string bestand, t_pfg * pfg)
{
  FILE * out;
  static char buffer1[512];
  static char buffer2[512];
  t_node * node;
  t_edge * edge;
  t_bbl * bbl, * i_bbl, * j_bbl;
  out = fopen(bestand,"w+");
  VERIFY_FASSERT(out,("Could not open for writing!"));
  fprintf(out, "digraph \"%llx\" {\n\tnode [shape=box]\n",0x0000000000000000);
  GRAPH_FOREACH_NODE(T_GRAPH(pfg),node)
  {
    bbl = PFG_BBL(node);
    //if(bbl->ins_first && IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)
    //  if(BBL_PROFILE_INFO(bbl)>0)
    //  {

//	Ia64InstructionPrint(bbl->ins_first, buffer1);
//	Ia64InstructionPrint(bbl->ins_last, buffer2);
  	if(((t_pfg_node *)node)->header)
	  fprintf(out, "\t\"%llx\" [label=\"H:%d - (0x%llx)\"]\n",
	      BBL_OLD_ADDRESS(bbl), BBL_PROFILE_INFO(bbl),BBL_OLD_ADDRESS(bbl));
	else 
	  fprintf(out, "\t\"%llx\" [label=\"F:%d - (0x%llx)\"]\n",
	    BBL_OLD_ADDRESS(bbl), BBL_PROFILE_INFO(bbl),BBL_OLD_ADDRESS(bbl));
	if(PFG_NODE_NEXT(node))
	{
	  fprintf(out, "\t\"%llx\" -> \"%llx\" [style=dotted,color=\"black\"];\n",BBL_OLD_ADDRESS(PFG_BBL(node)), BBL_OLD_ADDRESS(PFG_BBL(PFG_NODE_NEXT(node))));
	}

  //    }
  }
/*  GRAPH_FOREACH_NODE(T_GRAPH(pfg),node)
  {
    bbl = PFG_BBL(node);
    if(PFG_NODE_NEXT(node))
    {
      fprintf(out, "\t\"%llx\" -> \"%llx\" [style=solid,color=\"black\"];\n",BBL_OLD_ADDRESS(bbl), BBL_OLD_ADDRESS(PFG_BBL(PFG_NODE_NEXT(node))));
    }
  }*/
  GRAPH_FOREACH_EDGE(T_GRAPH(pfg),edge)
  {
    i_bbl = PFG_BBL(PFG_HEAD(edge));
    j_bbl = PFG_BBL(PFG_TAIL(edge));
    fprintf(out, "\t\"%llx\" -> \"%llx\" [label=\"%d\",style=solid,color=\"black\"];\n",BBL_OLD_ADDRESS(i_bbl), BBL_OLD_ADDRESS(j_bbl),PFG_COUNTER(edge));
  }

/*  GRAPH_FOREACH_NODE(T_GRAPH(pfg),node)
  {
    bbl = PFG_BBL(node);
    //if(bbl->ins_first && IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)
    //  if(BBL_PROFILE_INFO(bbl)>0)
    //  {

//	Ia64InstructionPrint(bbl->ins_first, buffer1);
//	Ia64InstructionPrint(bbl->ins_last, buffer2);
	fprintf(out, "\t\"%llx\" [label=\"%d - (0x%llx)\"]\n",
	    BBL_OLD_ADDRESS(bbl), BBL_PROFILE_INFO(bbl),BBL_OLD_ADDRESS(bbl));
  //    }
  }
*/

  fprintf(out, "}\n");
  fclose(out);
}
/* vim: set shiftwidth=2: */

