/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "mips_assemble.h"
#include "mips_opcodes.h"
#include "mips_registers.h"
#include "mips_print.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../utils/util_bitmanipulation.h"

#define ASM_RS(_reg)	(_reg<<21)
#define ASM_BASE(_reg)	ASM_RS(_reg)
#define ASM_RT(_reg)	(_reg<<16)
#define ASM_RD(_reg)	(_reg<<11)
#define ASM_SA(_reg)	(_reg<<6)
#define ASM_FT(_reg)	ASM_RT(_reg)
#define ASM_FS(_reg)	ASM_RD(_reg)
#define ASM_FD(_reg)	ASM_SA(_reg)
#define ASM_FMT(_reg)	ASM_RS(_reg)
#define ASM_SUB(_reg)	ASM_RS(_reg)
#define ASM_COND(_reg)	ASM_SA(_reg)
#define ASM_CC(_reg)	(_reg<<8)
   
int RegisterControle(int reg)
{
  if(reg==MIPS_REG_NONE)
    return 0;
  else 
    return reg;
}

void MipsAssembleCPU(t_ins * ins, t_uint32 * instr)
{
  if(MIPSINSOPCODE(ins)!=MIPS_BREAK)
  {
    *instr= mips_opcode_table[MIPSINSOPCODE(ins)].opcode;  /* Set the opcode */
    *instr |= ASM_RS(RegisterControle(MIPSINSRS(ins)));
    *instr |= ASM_RT(RegisterControle(MIPSINSRT(ins)));
    *instr |= ASM_RD(RegisterControle(MIPSINSRD(ins)));
    *instr |= ASM_SA(RegisterControle(MIPSINSSA(ins)));
    *instr |= MIPSINSINSTR_INDEX(ins);
    *instr |= MIPSINSIMMEDIATE(ins);
  }
  else
  {
    *instr= mips_opcode_table[MIPSINSOPCODE(ins)].opcode;  /* Set the opcode */
    *instr |= (MIPSINSIMMEDIATE(ins)<<6);
  }
}

void MipsAssembleCOP0(t_ins * ins, t_uint32 * instr)
{
  *instr= mips_opcode_table[MIPSINSOPCODE(ins)].opcode;  /* Set the opcode */
  *instr |= ASM_RT(RegisterControle(MIPSINSRS(ins)));
  *instr |= ASM_RD(RegisterControle(MIPSINSRS(ins)));
  *instr |= MIPSINSSEL(ins);
}

void MipsAssembleCOP1Branch(t_ins * ins, t_uint32 * instr)
{
  *instr= mips_opcode_table[MIPSINSOPCODE(ins)].opcode;  /* Set the opcode */
  *instr |= ASM_CC(MIPSINSCC(ins));
  *instr |= MIPSINSOFFSET(ins);
}

void MipsAssembleCOP1LoadStore(t_ins * ins, t_uint32 * instr)
{
  *instr= mips_opcode_table[MIPSINSOPCODE(ins)].opcode;  /* Set the opcode */
  *instr |= ASM_BASE(MIPSINSBASE(ins));
  *instr |= ASM_FT((t_uint8)(MIPSINSFT(ins)-MIPS_REG_F0));
  *instr |= MIPSINSOFFSET(ins);
}

void MipsAssembleCOP1Dataproc(t_ins * ins, t_uint32 * instr)
{
  *instr= mips_opcode_table[MIPSINSOPCODE(ins)].opcode;  /* Set the opcode */
  switch(MIPSINSOPCODE(ins))
  {
    case MIPS_MFC1:
    case MIPS_CFC1:
    case MIPS_MTC1:
    case MIPS_CTC1:
      *instr |= ASM_RT(MIPSINSRT(ins));
      *instr |= ASM_FS((t_uint8)(MIPSINSFS(ins)-MIPS_REG_F0));
      break;
    case MIPS_ADD_:
    case MIPS_SUB_:
    case MIPS_MUL_:
    case MIPS_DIV_:
      *instr |= ASM_FT((t_uint8)(MIPSINSFT(ins)-MIPS_REG_F0));
    case MIPS_SQRT:
    case MIPS_ABS:
    case MIPS_NEG:
      *instr |= ASM_FMT(MIPSINSFMT(ins));
      *instr |= ASM_FS((t_uint8)(MIPSINSFS(ins)-MIPS_REG_F0));
      *instr |= ASM_FD((t_uint8)(MIPSINSFD(ins)-MIPS_REG_F0));
      break;
    case MIPS_C:
      *instr |= ASM_FT((t_uint8)(MIPSINSFT(ins)-MIPS_REG_F0));
      *instr |= ASM_FMT(MIPSINSFMT(ins));
      *instr |= ASM_FS((t_uint8)(MIPSINSFS(ins)-MIPS_REG_F0));
      *instr |= MIPSINSCOND(ins);
      *instr |= ASM_CC(MIPSINSCC(ins));
      break;
    case MIPS_CVT_D: 
    case MIPS_CVT_S:
    case MIPS_CVT_W:
    case MIPS_ROUND_W:
    case MIPS_TRUNC_W:
    case MIPS_CEIL_W:
    case MIPS_FLOOR_W:
    case MIPS_MOV:
      *instr |= ASM_FMT(MIPSINSFMT(ins));
      *instr |= ASM_FS((t_uint8)(MIPSINSFS(ins)-MIPS_REG_F0));
      *instr |= ASM_FD((t_uint8)(MIPSINSFD(ins)-MIPS_REG_F0));
      break;
    case MIPS_MOVF_:
    case MIPS_MOVT_:
      *instr |= ASM_FMT(MIPSINSFMT(ins));
      *instr |= ASM_CC(MIPSINSCC(ins));
      *instr |= ASM_FS((t_uint8)(MIPSINSFS(ins)-MIPS_REG_F0));
      *instr |= ASM_FD((t_uint8)(MIPSINSFD(ins)-MIPS_REG_F0));
      break;
    case MIPS_MOVZ_:
    case MIPS_MOVN_:
      *instr |= ASM_FMT(MIPSINSFMT(ins));
      *instr |= ASM_RT(MIPSINSRT(ins));
      *instr |= ASM_FS((t_uint8)(MIPSINSFS(ins)-MIPS_REG_F0));
      *instr |= ASM_FD((t_uint8)(MIPSINSFD(ins)-MIPS_REG_F0));
      break;
    case MIPS_MOVF:
    case MIPS_MOVT:
      *instr |= ASM_RS(MIPSINSRS(ins));
      *instr |= ASM_RD(MIPSINSRD(ins));
      *instr |= ASM_CC(MIPSINSCC(ins));
      break;
  }
}

/* vim: set shiftwidth=2: */
