/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "mips_flowgraph.h"
#include <string.h>
#include "mips_instructions.h"
#include "mips_registers.h"
#include "mips_opcodes.h"
#include "mips_utils.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../kernel/diablo_symbol.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_iterators.h"

t_uint32 MipsAddBasicBlockEdges(t_object * obj, t_section * section);
t_uint32 MipsFindBBLLeaders(t_object * obj, t_section * code);
t_uint32 MipsChangeGot16ToHi16(t_object * obj,t_section * section);
t_uint32 RedefineEdges(t_object * obj, t_section * section); 
t_uint32 MipsRemoveGPComputations(t_section * code);

void MipsFlowgraphSection(t_object * obj, t_section * section) 
{
  t_uint32 ret;

  /* Find the leaders in the instruction list */
  ANALYSIS_START("Counting different instructions");
  MipsDifferentInstructions(obj,section);
  ANALYSIS_SIMPLE_END("Counting different instructions");

  /* Find the leaders in the instruction list */
  ANALYSIS_START("Leader detection");
  ret=MipsFindBBLLeaders(obj,section);
  ANALYSIS_END("Leader detection",("%d leaders",ret));

  /* Create the basic blocks (platform independent) */
  ANALYSIS_START("Creating Basic Blocks");
  ret=CfgCreateBasicBlocks(section);
  ANALYSIS_END("Creating Basic Blocks",("%d blocks",ret));

  /* GOT16-LO16 change to HI16-LO16 */
  ANALYSIS_START("Changing GOT16-LO16 to HI16-LO16");
  ret=MipsChangeGot16ToHi16(obj,section);
  ANALYSIS_END("Changing GOT16-LO16 to HI16-LO16",("%d changed",ret));

  /* Create the edges between basic blocks */
  ANALYSIS_START("Creating Basic Block graph");
  ret=MipsAddBasicBlockEdges(obj,section);
  ANALYSIS_END("Creating Basic Block graph",("%d edges added",ret));

  /* Rebuild Got-table */
  ANALYSIS_START("Rebuild Got-table");
  ret=MipsRebuildGotTableRelocs(obj,section);
  ANALYSIS_END("Rebuild Got-table",("%d changed",ret));
  
  /* Rebuild Got-table */
  ANALYSIS_START("Kill useless gp-calculations");
  ret=MipsRemoveGPComputations(section);
  ANALYSIS_END("Kill useless gp-calculations",("%d killed",ret));


}

t_uint32 MipsRemoveGPComputations(t_section * code) 
{
  t_ins *ins;
  t_bbl * bbl, *entry_bbl2,*entry_bbl3;
  t_cfg_edge * edge;
  t_cfg * cfg=SECTION_CFG(code);
  t_ins *tmp;
  t_uint32 deleted=0;

  t_bbl* entry_bbl1 = cfg->unique_entry_node;

  BBL_FOREACH_SUCC_EDGE(entry_bbl1,edge)
  {
    entry_bbl2 = CFG_EDGE_TAIL(edge);
    break;
  }

  BBL_FOREACH_SUCC_EDGE(entry_bbl2,edge)
  {
    entry_bbl3 = CFG_EDGE_TAIL(edge);
    break;
  }
  
  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS_SAFE(bbl,ins,tmp)
    if(MIPSINSOPCODE(ins)==MIPS_NOP)
    {
      InsKill(ins);
      deleted++;
    }

  CFG_FOREACH_BBL(cfg,bbl)
  {
    {
      if (bbl!=entry_bbl1 && bbl!=entry_bbl2 && bbl!=entry_bbl3)
      {
	BBL_FOREACH_INS_SAFE(bbl,ins,tmp)
	{
	  if (MIPSINSOPCODE(ins)==MIPS_LUI && MIPSINSRT(ins)==MIPS_REG_GP)
	  {
	    //MipsInstructionMakeNoop(ins);
	    InsKill(ins);
	    deleted++;
	  }
	  else if (MIPSINSOPCODE(ins)==MIPS_ADDIU && MIPSINSRT(ins)==MIPS_REG_GP)
	  {
	    //MipsInstructionMakeNoop(ins);
	    InsKill(ins);
	    deleted++;
	  }
	  else if (MIPSINSOPCODE(ins)==MIPS_ADDU && MIPSINSRD(ins)==MIPS_REG_GP)
	  {
	    //MipsInstructionMakeNoop(ins);
	    InsKill(ins);
	    deleted++;
	  }
	  else if (MIPSINSOPCODE(ins)==MIPS_SW && MIPSINSRT(ins)==MIPS_REG_GP)
	  {
	    //MipsInstructionMakeNoop(ins);
	    InsKill(ins);
	    deleted++;
	  }
	  else if(MIPSINSOPCODE(ins)==MIPS_LW && MIPSINSRT(ins)==MIPS_REG_GP)
	  {
	    //MipsInstructionMakeNoop(ins);
	    InsKill(ins);
	    deleted++;
	  }

	}
      }
    }
  }
  return deleted;
}

t_uint32 MipsRebuildGotTableRelocs(t_object * obj,t_section * section)
{
  t_object * sub, * tmp;
  t_section * sec;
  t_uint32 deleted=0;
  t_uint32 i,j;
  OBJECT_FOREACH_SECTION(obj,sec,i)
  {
    if (!strcmp(sec->name,".got"))
    {
      i=0x1000000;
      break;
    }
  }
  VERIFY_FASSERT(sec,("no got section"));
  OBJECT_FOREACH_SUBOBJECT(obj,sub,tmp)
  {
    t_section * ss;
    OBJECT_FOREACH_SECTION(sub,ss,i)
    {
      if (SECTION_PARENT_SECTION(ss) == sec&&
	  RELOCATABLE_REFED_BY(ss)!=NULL&&
	  RELOCATABLE_REFERS_TO(ss)!=NULL&&
	  (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ss)->rel))==RT_BBL||
	   RELOCATABLE_TYPE(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ss)->rel))==RT_SECTION))
      {

	t_section * ss2;
	OBJECT_FOREACH_SECTION(sub,ss2,j)
	{
	  if (SECTION_PARENT_SECTION(ss2) == sec&&
	      RELOCATABLE_REFERS_TO(ss2)&&
	      RELOCATABLE_REFERS_TO(ss2)->rel && RELOCATABLE_REFERS_TO(ss)->rel &&
	      REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ss2)->rel)==REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ss)->rel) )
	  {
	    if(ss2==ss)
	    {
	      j=0x1000000;
	      break;
	    }
	    else
	    {
	      RelocSetToRelocatable(RELOCATABLE_REFED_BY(ss)->rel,T_RELOCATABLE(ss2));
	      deleted++;
	      j=0x1000000;
	      break;
	    }
	  }
	}
      }
    }
  }
  return deleted;
}


t_uint32 MipsChangeGot16ToHi16(t_object * obj,t_section * section)
{
  t_uint32 teller;
  t_ins * i_ins, * ins;
  t_bbl * block;
  t_reloc * reloc=NULL;
  t_reloc * hulp=NULL;

  SECTION_FOREACH_INS(section,ins)
  {

    if (!(INS_ATTRIB(ins) & IF_BBL_LEADER)) 
      continue;
    block=INS_BBL(ins);

    BBL_FOREACH_INS(block,i_ins)
    {
      if(RELOCATABLE_REFERS_TO(i_ins))
	reloc=RELOCATABLE_REFERS_TO(i_ins)->rel;
      if(MIPSINSOPCODE(i_ins) == MIPS_LW && reloc && reloc->corresponding && MIPSINSBASE(i_ins)==MIPS_REG_GP)
      {
	hulp=RelocTableAddRelocToRelocatable(obj->reloc_table,reloc->addend,REL_FROM(reloc),reloc->from_offset,REL_TO_RELOCATABLE(reloc->corresponding),reloc->corresponding->to_offset,FALSE,reloc->corresponding->edge,NULL,NULL,"A #0f S+ = sffff& #0f  - s0010> \\ sffff& = l iffff0000&|w\\l sffff &-$", TRUE);
	reloc->corresponding->corresponding=NULL;
	reloc->corresponding=NULL;
	RelocTableRemoveReloc(section->obj->reloc_table,reloc);
	INS_TYPE(i_ins)=IT_DATAPROC;
	MIPSINSOPCODE(i_ins)=MIPS_LUI;
	MIPSINSBASE(i_ins)=MIPS_REG_NONE;
	MIPSINSOFFSET(i_ins)=0;
	INS_REGS_USE(i_ins) = MipsUsedRegisters(i_ins);
	INS_REGS_DEF(i_ins) = MipsDefinedRegisters(i_ins);
	INS_ATTRIB(i_ins) |= IF_CHANGED;
	teller++;
      }
    }
  }
  return teller;
}


t_uint32 MipsAddBasicBlockEdges(t_object * obj, t_section * section) 
{

  /* Basic Block Edges 
   * -----------------
   *
   * Add edges between the newly created basic blocks. Both the lineair (array)
   * representation of the instructions and the basic block representation are
   * available: you can access the (copy) basic block by using INS_BBL of a
   * leader in the lineair representation.
   *
   * Although connecting the blocks isn't that hard, it could be difficult to
   * find out if a given jump instructions is a call or just a jump (without
   * return). Getting this completly safe is probably not what you want: if you
   * implement this completly safe, you'll probably end up with one big
   * function, which will degrade the performance of the rest of diablo or hell
   * nodes after each jump which will degrade the performance even further.
   *
   * On the other hand, if you get this wrong (e.g add a jump edge where there
   * should have been a call edge), then chances are that your flowgraph (and
   * thus the produced binary) is wrong (the successor block of the call block
   * will have no incoming (return) edge, and thus this block is dead). 
   *
   * The solution is to look at code that is produced by your compiler(s).
   * It'll probably have only a few possibilties to produce calls. If you
   * handle them as safe as possible, than you're off the hook.  */

  t_ins *ins, * block_end, * block_start, * block_branch, * jump_target ;
  t_bbl * block, * ft_block;
  t_cfg * flowgraph=SECTION_CFG(section);
  t_reloc_ref * ref;
  t_uint32 nedges=0;

  /* walk through the original(!) array (we need offsets) and add 
   * all possible successors to every node */

  SECTION_FOREACH_INS(section,ins)
  {
    /* we're only interested in the first and last instructions of a basic block */

    if (!(INS_ATTRIB(ins) & IF_BBL_LEADER)) 
      continue;

    block_start = ins;
    block=INS_BBL(block_start);

    block_branch = T_INS(&(T_MIPS_INS(ins)[block->nins - 2])); 

    /* Warning: This is not the same as the end of the bbl, since the end is a copy of the original end */
    block_end = T_INS(&(T_MIPS_INS(ins)[block->nins - 1])); 
    ft_block = INS_NEXT(block_end)?INS_BBL(INS_NEXT(block_end)):NULL;  /* the FallThrough basic block */


    /* 1. ADD EDGES FOR RELOCATIONS: 
     * 
     * each relocation gets it's own edge. These edges will be automatically
     * deleted when the relocation is removed. There's no need to add an edge
     * for the obvious relocations (such as the relocations that are present
     * for the calculation of the offsets in intermodular jumps), since they
     * are completly described by the edge itself (the reason for the edge is
     * that there is a jump, not that there is a relocation) */


    for (ref=RELOCATABLE_REFED_BY(block); ref!=NULL; ref=ref->next)
    {

      //      DiabloPrint(stdout,"IEKS: @R %x %x\n",ref->rel,ref->rel->addend,ref->rel->hell);

      /* Skip the obvious relocations and reloctions to data */

      if (ref->rel->hell && (INS_TYPE(block_start)!=IT_DATA))
      {

	/* Some extra checks if the relocation is from an instruction */
	if (REL_FROM(ref->rel)->type==RT_INS)
	{
	  t_cfg_edge * found=NULL;
	  t_cfg_edge * i_edge;
	  if (RELOCATABLE_REFERS_TO(REL_FROM(ref->rel))->next)
	  {
	    DiabloPrint(stdout,"1. @R\n",RELOCATABLE_REFERS_TO(REL_FROM(ref->rel))->rel);
	    DiabloPrint(stdout,"2. @R\n",RELOCATABLE_REFERS_TO(REL_FROM(ref->rel))->next->rel);
	    FATAL(("Multiple relocs from instruction?"));
	  }
	  if ((RELOCATABLE_REFERS_TO(REL_FROM(ref->rel))->rel)!=ref->rel) FATAL(("Relocation wrong"));


	  BBL_FOREACH_PRED_EDGE(block,i_edge)
	  {
	    if ((CFG_EDGE_HEAD(i_edge)==flowgraph->hell_node) && (EDGE_CAT(i_edge) == ET_CALL))
	    {
	      found=i_edge;
	      break;
	    }
	  }

	  if (found)
	  {
	    ref->rel->edge=found;
	    found->refcount++;
	    /* also increment the refcount of the corresponding return edge!!! */
	    VERIFY_FASSERT(found->corr_edge,("Call edge @E does not have a corresponding edge!",found));
	    found->corr_edge->refcount++;
	  }
	  else
	  {
	    //t_ins * h_ins=SecGetInsByAddress(section,REL_FROM(ref->rel)->caddress);
	    {
	      //DiabloPrint(stdout,"FROM HELL TO 1: @R\n",ref->rel);
	      //DiabloPrint(stdout,"@iB\n", block);
	      ref->rel->edge=CfgEdgeNewCall(flowgraph, flowgraph->hell_node, block,NULL,NULL);
	      nedges++;
	    }
	  }
	  /* Some comments on this */

	  VERBOSE(1,("Adding INS arrow from hell! (relocation @R, @B)",ref->rel,block));
	}
	else
	{
	  t_cfg_edge * found=NULL;
	  t_cfg_edge * i_edge;
	  /* Add the edge */
	  BBL_FOREACH_PRED_EDGE(block,i_edge)
	  {
	    if ((CFG_EDGE_HEAD(i_edge)==flowgraph->hell_node) && (EDGE_CAT(i_edge) == ET_CALL))
	    {
	      found=i_edge;
	      break;
	    }
	  }

	  if (found)
	  {
	    ref->rel->edge=found;
	    found->refcount++;
	    /* also increment the refcount of the corresponding return edge!!! */
	    VERIFY_FASSERT(found->corr_edge,("Call edge @E does not have a corresponding edge!",found));
	    found->corr_edge->refcount++;
	  }
	  else
	  {

	    //DiabloPrint(stdout,"FROM HELL TO 2 : @B\n",block);
	    //DiabloPrint(stdout,"FROM HELL TO 2: @R\n",ref->rel);
	    //DiabloPrint(stdout,"@iB\n", block);
	    ref->rel->edge=CfgEdgeNewCall(flowgraph, flowgraph->hell_node, block,NULL,NULL);
	    nedges++;
	  }
	  VERBOSE(1,("Adding arrow from hell! (relocation @R, @B)",ref->rel,block));
	}
      }
    }
    /* }}} */

/* 2. ADD NORMAL EDGES */
/* now determine the successors of the block and their types */

if (INS_TYPE(block_branch) == IT_BRANCH)
{
  t_uint32 addr32;//address;
  t_address hulp_addr,hulp_addr2;
  GenericSet_uint32(hulp_addr,0xf0000000);
  /* The target of the branch will also be a basic block leader */
  switch (MIPSINSOPCODE(block_branch))
  {
    case MIPS_J:
      GenericSet_uint32(hulp_addr2,MIPSINSINSTR_INDEX(block_branch)<<2);
      jump_target = SecGetInsByAddress(section,GENERIC_OR(GENERIC_AND(INS_OLD_ADDRESS(block_end),hulp_addr),hulp_addr2));
      CfgEdgeNew(flowgraph, block, INS_BBL(jump_target),ET_JUMP);
      nedges++;
      while (RELOCATABLE_REFERS_TO(block_branch))
	RelocTableRemoveReloc(section->obj->reloc_table,RELOCATABLE_REFERS_TO(block_branch)->rel);
      break;
    case MIPS_JAL:

      //  DiabloPrint(stdout,"@I\n",block->ins_last);
      //  DiabloPrint(stdout,"JAL:@I\n",INS_PREV(block->ins_last));
      if(
	  RELOCATABLE_REFERS_TO(INS_PREV(block->ins_last)) &&
	  RELOCATABLE_REFERS_TO(INS_PREV(block->ins_last))->rel
	)
      {
	t_ins * h_ins;
	h_ins=INS_PREV(block->ins_last);
	t_address h_address;
	//  DiabloPrint(stdout,"@I\n",h_ins);

	h_address=GENERIC_ADD(INS_REFERS_TO(h_ins)->rel->to_offset,REL_TO_RELOCATABLE(INS_REFERS_TO(h_ins)->rel)->caddress); 
	CfgEdgeNewCall(flowgraph, block, INS_BBL(SecGetInsByAddress(section,h_address)),ft_block,NULL);
	nedges++;
	CfgEdgeNew(flowgraph,block,ft_block,ET_FUNLINK);
	nedges++;
	if (INS_REFERS_TO(h_ins)->rel)
	{
	  //  DiabloPrint(stdout,"@R\n",INS_REFERS_TO(h_ins)->rel);
	  RelocTableRemoveReloc(section->obj->reloc_table,INS_REFERS_TO(h_ins)->rel); 
	}
	break;
      }

      /* skip jal with target 0, this is an inefficiency of gcc */
      if(MIPSINSINSTR_INDEX(block_branch)==0)
	break;
      GenericSet_uint32(hulp_addr2,MIPSINSINSTR_INDEX(block_branch)<<2);
      jump_target = SecGetInsByAddress(section,GENERIC_OR(GENERIC_AND(INS_OLD_ADDRESS(block_end),hulp_addr),hulp_addr2));
      CfgEdgeNewCall(flowgraph, block, INS_BBL(jump_target),ft_block,NULL);
      nedges++;
      CfgEdgeNew(flowgraph,block,ft_block,ET_FUNLINK);
      nedges++;
      while (RELOCATABLE_REFERS_TO(block_branch))
	RelocTableRemoveReloc(section->obj->reloc_table,RELOCATABLE_REFERS_TO(block_branch)->rel);
      break;
    case MIPS_JR:
      if(MIPSINSRS(block_branch)==MIPS_REG_R31)
      {
	CfgEdgeNew(flowgraph,  block,  flowgraph->exit_hell_node, ET_RETURN);
	nedges++;
      }
      else 
      {
	CfgEdgeNew(flowgraph, block, flowgraph->hell_node, ET_JUMP);
	nedges++;
      }
      break;
    case MIPS_JALR:
      if(INS_PREV(block_branch) &&
	  INS_PREV(INS_PREV(block_branch)) && 
	  MIPSINSOPCODE(INS_PREV(INS_PREV(block_branch)))==MIPS_ADDIU &&
	  MIPSINSRT(INS_PREV(INS_PREV(block_branch)))==MIPSINSRS(block_branch) &&
	  RELOCATABLE_REFERS_TO(INS_PREV(INS_PREV(INS_PREV(block->ins_last)))) &&
	  RELOCATABLE_REFERS_TO(INS_PREV(INS_PREV(INS_PREV(block->ins_last))))->rel)
      {
	t_ins * h_ins;
	t_ins * jalr_ins;
	h_ins=INS_PREV(INS_PREV(INS_PREV(block->ins_last)));
	jalr_ins=INS_PREV(block->ins_last);
	t_address h_address;

	h_address=GENERIC_ADD(INS_REFERS_TO(h_ins)->rel->to_offset,REL_TO_RELOCATABLE(INS_REFERS_TO(h_ins)->rel)->caddress); 
	CfgEdgeNewCall(flowgraph, block, INS_BBL(SecGetInsByAddress(section,h_address)),ft_block,NULL);
	nedges++;

	//DiabloPrint(stdout,"WAS:block:@iB\n",block);
	//DiabloPrint(stdout,"address:%x\n",h_address);

	INS_TYPE(jalr_ins) =IT_BRANCH;
	MIPSINSOPCODE(jalr_ins)=MIPS_JAL;
	MIPSINSRS(jalr_ins)=MIPS_REG_NONE;
	MIPSINSRD(jalr_ins)=MIPS_REG_NONE;
	MIPSINSOFFSET(jalr_ins)=0;
	INS_REGS_USE(jalr_ins) = MipsUsedRegisters(jalr_ins);
	INS_REGS_DEF(jalr_ins) = MipsDefinedRegisters(jalr_ins);
	INS_ATTRIB(jalr_ins) |= IF_CHANGED;
	//DiabloPrint(stdout,"NU:block:@iB\n",block);
      }

      else if(INS_PREV(block_branch) &&
	  MIPSINSOPCODE(INS_PREV(block_branch))==MIPS_ADDIU &&
	  MIPSINSRT(INS_PREV(block_branch))==MIPSINSRS(block_branch) &&
	  RELOCATABLE_REFERS_TO(INS_PREV(INS_PREV(block->ins_last))) &&
	  RELOCATABLE_REFERS_TO(INS_PREV(INS_PREV(block->ins_last)))->rel)
      {
	t_ins * h_ins;
	t_ins * jalr_ins;
	jalr_ins=INS_PREV(block->ins_last);
	h_ins=INS_PREV(INS_PREV(block->ins_last));
	t_address h_address;

	h_address=GENERIC_ADD(INS_REFERS_TO(h_ins)->rel->to_offset,REL_TO_RELOCATABLE(INS_REFERS_TO(h_ins)->rel)->caddress); 
	CfgEdgeNewCall(flowgraph, block, INS_BBL(SecGetInsByAddress(section,h_address)),ft_block,NULL);
	nedges++;

	//DiabloPrint(stdout,"WAS:block:@iB\n",block);
	//DiabloPrint(stdout,"address:%x\n",h_address);

	INS_TYPE(jalr_ins) =IT_BRANCH;
	MIPSINSOPCODE(jalr_ins)=MIPS_JAL;
	MIPSINSRS(jalr_ins)=MIPS_REG_NONE;
	MIPSINSRD(jalr_ins)=MIPS_REG_NONE;
	MIPSINSOFFSET(jalr_ins)=0;
	INS_REGS_USE(jalr_ins) = MipsUsedRegisters(jalr_ins);
	INS_REGS_DEF(jalr_ins) = MipsDefinedRegisters(jalr_ins);
	INS_ATTRIB(jalr_ins) |= IF_CHANGED;
	//DiabloPrint(stdout,"NU:block:@iB\n",block);

      }

      else if(INS_PREV(block_branch) &&
	  INS_PREV(INS_PREV(block_branch)) && 
	  MIPSINSOPCODE(INS_PREV(INS_PREV(block_branch)))==MIPS_LW &&
	  MIPSINSRT(INS_PREV(INS_PREV(block_branch)))==MIPSINSRS(block_branch) &&
	  RELOCATABLE_REFERS_TO(INS_PREV(INS_PREV(INS_PREV(block->ins_last)))) &&
	  RELOCATABLE_REFERS_TO(INS_PREV(INS_PREV(INS_PREV(block->ins_last))))->rel)
      {
	t_ins * h_ins;
	t_ins * jalr_ins;
	jalr_ins=INS_PREV(block->ins_last);
	h_ins=INS_PREV(INS_PREV(INS_PREV(block->ins_last)));
	t_address h_address;

	if(RELOCATABLE_REFERS_TO(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(h_ins)->rel))->rel)
	{
	  t_reloc * h_reloc=RELOCATABLE_REFERS_TO(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(h_ins)->rel))->rel;
	  h_address=BBL_CADDRESS(REL_TO_RELOCATABLE(h_reloc));
	//DiabloPrint(stdout,"WAS:block:@iB\n",block);
	//DiabloPrint(stdout,"address:%x\n",h_address);


	  if(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(h_reloc))==RT_BBL)
	  {	
	    CfgEdgeNewCall(flowgraph, block, INS_BBL(SecGetInsByAddress(section,h_address)),ft_block,NULL);
	    nedges++;
	    INS_TYPE(jalr_ins) =IT_BRANCH;
	    MIPSINSOPCODE(jalr_ins)=MIPS_JAL;
	    MIPSINSRS(jalr_ins)=MIPS_REG_NONE;
	    MIPSINSRD(jalr_ins)=MIPS_REG_NONE;
	    MIPSINSOFFSET(jalr_ins)=0;
	    INS_REGS_USE(jalr_ins) = MipsUsedRegisters(jalr_ins);
	    INS_REGS_DEF(jalr_ins) = MipsDefinedRegisters(jalr_ins);
	    INS_ATTRIB(jalr_ins) |= IF_CHANGED;
	  }
	  else
	  {
	    CfgEdgeNewCall(flowgraph, block, flowgraph->hell_node,ft_block,NULL);
	    nedges++;
	  }
	//DiabloPrint(stdout,"NU:block:@iB\n",block);
	}
	else
	  FATAL(("NO REL TO RELOCATABLE"));

      }

      else
      {
	CfgEdgeNewCall(flowgraph, block, flowgraph->hell_node,ft_block,NULL);
	nedges++;
      }
      CfgEdgeNew(flowgraph, block,ft_block, ET_FUNLINK);
      nedges++;
      break;
    case MIPS_BEQ:
    case MIPS_BGEZ:
    case MIPS_BGTZ:
    case MIPS_BLEZ:
    case MIPS_BLTZ:
    case MIPS_BNE:
    case MIPS_BC1F:
    case MIPS_BC1T:
      /* If the branch is conditional, add a fall-through edge */
      if ((ft_block != NULL)&&(!IS_DATABBL(ft_block)))
      {
	CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
	nedges++;
      }
      /* the address is the address of the delay slot (pc+4) and the immediate signed and
       * shifted 2 to left */
      addr32=((t_int16)MIPSINSIMMEDIATE(block_branch)) ;
      jump_target = (t_ins *) (T_MIPS_INS(block_end) + addr32); 
//      DiabloPrint(stdout,"@I\n%x \n",block_branch,jump_target);
      CfgEdgeNew(flowgraph, block, INS_BBL(jump_target),ET_JUMP);
      nedges++;
      break;
    case MIPS_BLTZAL:
    case MIPS_BGEZAL:
      addr32=((t_int16)MIPSINSIMMEDIATE(block_branch)) ;
      jump_target = (t_ins *) (T_MIPS_INS(block_end) + addr32); 
      CfgEdgeNewCall(flowgraph, block, INS_BBL(jump_target),ft_block,NULL);
      nedges++;
      CfgEdgeNew(flowgraph, block,ft_block, ET_FUNLINK);
      CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
      nedges++;
      break;
    default:
      FATAL(("Not implemented in flowgraph: AddBasicBlockEdges %d",(MIPSINSOPCODE(block_branch))));
      break;
  }
}
else if (INS_TYPE(block_end) == IT_SWI)
{
  CfgEdgeNew(flowgraph, block,flowgraph->swi_hell_node,ET_SWI);
  nedges++;
  if (ft_block != NULL)
  {
    CfgEdgeNew(flowgraph, block,ft_block, ET_SWILINK);
    nedges++;
  }
/*  if (ft_block != NULL)
  {
    CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
    nedges++;
  }*/
}
else
{
  if ((ft_block != NULL)&&(!IS_DATABBL(ft_block)))
  {
    CfgEdgeNew(flowgraph, block,ft_block, ET_FALLTHROUGH);
    nedges++;
  }
}

}
return nedges;
}


t_uint32 MipsFindBBLLeaders(t_object * obj, t_section * code) 
{

  /* Basic leader-detection algorithm. 
   * ---------------------------------
   *
   * See e.g. the Red Dragon Book ("Compilers: Principles, Techniques and
   * Tools", by Alfred V. Aho, Ravi Sethi, and Jeffrey D. Ullman) for more
   * details. 
   *
   * The aim of this function is to identify all start-addresses of basic
   * blocks (and mark them). There are four reasons to mark an instructions as
   * a basic block leader:
   * 
   * 1. The instruction is the target of a direct or indirect jump. This is not
   * as easy as it seems, especially when the program counter is explicit:
   * every instruction that changes the pc can be considered as a jump
   * instruction.
   *
   * 2. It's the successor of one of the above (if a jump is conditional, than
   * the next instruction will be the first instruction of the fall-through
   * block, else it is either target of an other jump or a dead block. In the
   * latter case, it will be removed by the unreachable block removal routines,
   * so we just mark it as a basic block leader).
   *
   * 3. There is an address produced to this basic block, this is either 
   *
   *    a. done directly, using the program counter (we assume no real
   *    code-address calculations are present (so not address of function + x) 
   *
   *    b. using a relocation (and not necessary detected when scanning the
   *    instructions) 
   *
   * For both cases we need to assume there will be an indirect jump to this
   * address.
   *
   * 4. The start of data blocks in code or instructions following datablocks
   * are considered basic block leaders.
   *
   */

  /* We can still assume that all instructions are stored sequentially in the
   * array Ains the last instruction stored can be recognised with INSNEXT ==
   * NULL */

  t_ins * ins;
  t_uint32 pc, nleaders=0;
  /* Initialisation: The first instruction of the program is always a leader.
   * The reasoning is equal to that of reason 2: Either there will be a jump to
   * this address or it's the entry point or it is dead. */

  /* Instructions are still in a sequential list, stored in t_section->data */
  ins = code->data; 
  INS_ATTRIB(ins) |= IF_BBL_LEADER;

  for (pc = 0; ins != NULL; ins = INS_NEXT(ins), pc += 4)
  {

    /* 1. DATA (reason 4) */ 

    if (INS_TYPE(ins) == IT_DATA)
    {
      /*t_section * sub_sec;*/
      /* The start of a bunch of data is considered a basic block leader */
      INS_ATTRIB(ins) |= IF_BBL_LEADER; 


      /* Process all the other data */
      while (ins && INS_TYPE(ins) == IT_DATA) {  
	ins = INS_NEXT(ins); 
	pc += 4; 
      } 

      /* The instruction following the data is a basic block leader */
      if (ins) INS_ATTRIB(ins) |= IF_BBL_LEADER; 
      else
	/* If there are no instructions left, we can exit the outer loop
	 * (all blocks are marked) */
	break; 
    }

    /* 2. NORMAL BRANCHES BY THE CPU (reason 1 and 2) */
    if (INS_TYPE(ins) == IT_BRANCH && MIPSINSPROCESSOR(ins)==CPU)
    {
      t_uint32 addr32=pc+4;
      t_address offset;
      t_ins * target;
      t_address addr,hulp_addr;
      GenericSet_uint32(hulp_addr,0xf0000000);
      /* The instruction following the delay slot of the branch instruction is a basic block leader */
      if (INS_NEXT(ins) != NULL && INS_NEXT(INS_NEXT(ins)) != NULL) 
	INS_ATTRIB((t_mips_ins *) INS_NEXT(INS_NEXT(ins))) |= IF_BBL_LEADER;

      /* The target of the branch will also be a basic block leader */
      switch (MIPSINSOPCODE(ins))
      {
	case MIPS_J:
	case MIPS_JAL:
	  /* we suppose here that this instruction will never be executed, due to
	   * a previous test (we noticed this behaviour in gcc) */
	  if(MIPSINSINSTR_INDEX(ins)==0)
	    break;
	  GenericSet_uint32(addr,addr32);
	  addr=GENERIC_AND(hulp_addr,SECTION_GET_OLD_ADDRESS(code));
	  GenericSet_uint32(offset,MIPSINSINSTR_INDEX(ins)<<2);
	  addr=GENERIC_OR(addr,offset);
	  target=SecGetInsByAddress(code,addr);
	  if (target) INS_ATTRIB(target) |= IF_BBL_LEADER; 
	  else FATAL(("BRANCH outside of section: @I",ins));
	  break;
	case MIPS_JALR:
	case MIPS_JR:
	  /* destination unknown */
	  break;
	default:
	  /* the address is the address of the delay slot (pc+4) and the immediate signed and
	   * shifted 2 to left */
	  addr32+=(((t_int16)MIPSINSIMMEDIATE(ins))*4);
	  GenericSet_uint32(addr,addr32);
	  addr=GENERIC_ADD(addr,SECTION_GET_OLD_ADDRESS(code));
	  target=SecGetInsByAddress(code,addr);
	  if (addr32<SECTION_NINS(code)*4) INS_ATTRIB(target) |= IF_BBL_LEADER; 
	  else FATAL(("BRANCH outside of section: @I",target));
	  break;
      }
    }

    /* 3. SYSTEM CALLS {{{ */
    else if (INS_TYPE(ins) == IT_SWI) 
    {
      /* Actually, we could just consider system calls as normal
       * instructions, if we were sure that they are not long-jump-like
       * system calls. Since we do not know that for sure here, we must
       * assume that is will alter control flow => next instruction is a
       * leader */

      if (INS_NEXT(ins) != NULL )
	INS_ATTRIB((t_mips_ins *) INS_NEXT(ins)) |= IF_BBL_LEADER; 
    }
    /* }}} */

    /* 4. NORMAL BRANCHES BY THE COP1 (reason 1 and 2) */
    else if (INS_TYPE(ins) == IT_BRANCH && MIPSINSPROCESSOR(ins)==COP1)
    {
      t_uint32 addr32=pc;
      t_ins * target;
      t_address addr;
      /* The instruction following the delay slot of the branch instruction is a basic block leader */
      if (INS_NEXT(ins) != NULL && INS_NEXT(INS_NEXT(ins)) != NULL) 
	INS_ATTRIB((t_mips_ins *) INS_NEXT(INS_NEXT(ins))) |= IF_BBL_LEADER;

      /* The target of the branch will also be a basic block leader */
      addr32=pc+4+ ((t_int16)MIPSINSIMMEDIATE(ins))*4;
      GenericSet_uint32(addr,addr32);
      addr=GENERIC_ADD(addr,SECTION_GET_OLD_ADDRESS(code));
      target=SecGetInsByAddress(code,addr);
      if (addr32<SECTION_NINS(code)*4) INS_ATTRIB(target) |= IF_BBL_LEADER; 
      else FATAL(("BRANCH outside of section: @I",target));
    }
    /* 4. THE RELOCATIONS (reason 3) */
    if (RELOCATABLE_REFED_BY(ins)) 
    {
      INS_ATTRIB(ins) |= IF_BBL_LEADER;
    }
  }

  /* Now we also mark the beginning of data-blocks as bbl-leaders. We do this by walking the subsections
   * and look for RO_DATA. The beginning of mapped RO_DATA subsections are marked as bbl-leaders.
   */
  {
    t_object * tmp, * tmp2;
    t_uint32 tel;
    t_section * sec;
    t_ins * tmp_ins;


    OBJECT_FOREACH_SUBOBJECT(obj,tmp,tmp2)
      OBJECT_FOREACH_SECTION(tmp,sec,tel)
      {
	if (!SECTION_IS_MAPPED(sec)) continue;
	tmp_ins=SecGetInsByAddress(code,SECTION_GET_CADDRESS(sec));
	if (tmp_ins)
	{
	  if (GENERIC_EQ(INS_CADDRESS(tmp_ins) , SECTION_GET_CADDRESS(sec)))
	  {
	    INS_ATTRIB(tmp_ins) |= IF_BBL_LEADER; 
	  }
	}
      }
  }

  ins=code->data;
  while(ins)
  {
    if(INS_ATTRIB(ins)&IF_BBL_LEADER)
      nleaders++;
    ins=INS_NEXT(ins);
  }

  return nleaders;
}

void MipsMakeAddressProducers(t_section * sec)
{
  /* just a stub */
  return;
}

t_bool MipsFunIsGlobal(t_function * fun)
{
  if (fun->function_flags & FF_IS_EXPORTED)
  {
    return TRUE;
  }
  return FALSE;
}

/* vim: set shiftwidth=2: */
