/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "mips_instructions.h"
#include "mips_opcodes.h"
#include "mips_registers.h"
#include "mips_utils.h"
#include "../../utils/util_error.h"
#include "../../utils/util_malloc.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_iterators.h"

t_bool MipsIsControlflow(t_ins * ins)
{
  if(INS_TYPE(ins) == IT_BRANCH) return TRUE;
  if(INS_TYPE(ins) == IT_SWI) return TRUE;
  return FALSE;
}

t_bool MipsInsHasSideEffect(t_ins * ins)
{
  if ((INS_TYPE(ins) == IT_STORE) || (INS_ATTRIB(ins) & IF_PCDEF))
    return TRUE;
  if (MIPSINSOPCODE(ins) == MIPS_MTC1 || MIPSINSOPCODE(ins)==MIPS_LWC1 || MIPSINSOPCODE(ins)==MIPS_CTC1 || MIPSINSOPCODE(ins)==MIPS_CFC1)
    return TRUE;
  if (INS_TYPE(ins) == IT_STORE_MULTIPLE) 
  {
    return TRUE;
  }
  if (INS_TYPE(ins) == IT_SWI) 
  {
    return TRUE;
  }

  return FALSE;
}

void MipsInstructionMakeNoop(t_ins * ins)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  /* Make a noop */

   INS_TYPE(ins)=IT_NOP;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=MIPS_NOP;
   MIPSINSRS(ins)=MIPS_REG_NONE;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins) =0;
   MIPSINSINSTR_INDEX(ins)=0;

  INS_REGS_USE(ins) = MipsUsedRegisters(ins);
  INS_REGS_DEF(ins) = MipsDefinedRegisters(ins);
  INS_ATTRIB(ins) |= IF_CHANGED;

  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
 // relref = RELOCATABLE_REFERS_TO(ins);
  /*while (relref)
  {
    RELOCATABLE_REFERS_TO(ins) = RELOCATABLE_REFERS_TO(ins)->next;
    
    RelocTableRemoveReloc(reloc_table, relref->rel);
    Free(relref);
    relref = RELOCATABLE_REFERS_TO(ins);
  }*/
}

void MipsMakeAddressProducer(t_ins * ins)
{
  INS_TYPE(ins) = IT_CONSTS;
  MIPSINSOPCODE(ins) = MIPS_ADDRESS_PRODUCER;
  INS_ATTRIB(ins) |= IF_CHANGED;

}


void MipsDifferentInstructions(t_object * obj, t_section * section) 
{
  t_ins * ins_s;
  t_int32 cpu,cop0,cop1;
  t_int32 pc;

  cpu=0;cop0=0;cop1=0;

  ins_s = section->data; 
  for (pc = 0; ins_s != NULL; ins_s = INS_NEXT(ins_s), pc += 4)
  {
      switch(MIPSINSPROCESSOR(ins_s))
      {
	case CPU:
	  cpu++;
	  break;
	case COP0:
	  cop0++;
	  break;
	case COP1:
	  cop1++;
	  break;
	default:
	  FATAL(("ins:@I\nProcessor niet gedef:%d\n",ins_s,MIPSINSPROCESSOR(ins_s)));
	  break;
      }
  }
  DiabloPrint(stdout,"INSTR: CPU-instructions:%d\nINSTR: COP0-instructions:%d\nINSTR: COP1-instructions:%d\n",cpu,cop0,cop1);
}

void MipsDifferentInstructions2(t_section * code) 
{
  t_bbl * bbl;
  t_int32 cpu,cop0,cop1;
  cpu=0;cop0=0;cop1=0;

  CFG_FOREACH_BBL(T_CFG(code->data),bbl)
  {
    //t_uint32 addr;
    t_mips_ins * current=(t_mips_ins *) bbl->ins_first;

    while(current)
    {
      switch(MIPSINSPROCESSOR(current))
      {
	case CPU:
	  cpu++;
	  break;
	case COP0:
	  cop0++;
	  break;
	case COP1:
	  cop1++;
	  break;
	default:
	  FATAL(("ins:@I\nProcessor niet gedef:%d\n",current,MIPSINSPROCESSOR(current)));
	  break;
      }

      current=(t_mips_ins *) INS_NEXT(current);
    }

  }
  DiabloPrint(stdout,"INSTR: CPU-instructions:%d\nINSTR: COP0-instructions:%d\nINSTR: COP1-instructions:%d\n",cpu,cop0,cop1);
}

t_tristate MipsIsSyscallExit(t_ins * ins)
{
  return PERHAPS;
}

t_bool MipsInsIsIndirectCall(t_ins * ins)
{
  switch(MIPSINSOPCODE(ins))
  {
    case MIPS_JAL:
    case MIPS_BLTZAL:
    case MIPS_BGEZAL:
      return FALSE;
      break;
    default:
      switch(MIPSINSOPCODE(INS_PREV(ins)))
      {
	case MIPS_JAL:
	case MIPS_BLTZAL:
	case MIPS_BGEZAL:
	  return FALSE;
	  break;
	default:
	  DiabloPrint(stdout,"@I\n",INS_PREV(ins));
	  FATAL(("IMPLEMENT ME"));
	  break;
      }
  }
  return TRUE;
}

t_bool MipsInsIsIndirectJump(t_ins * ins)
{
  switch(MIPSINSOPCODE(ins))
  {
    case MIPS_JR:
    case MIPS_BC1F:
    case MIPS_BC1T:
      return TRUE;
    case MIPS_J:
    case MIPS_BNE:
    case MIPS_BEQ:
    case MIPS_BLEZ:
    case MIPS_BGEZ:
    case MIPS_BGTZ:
    case MIPS_BLTZ:
      return FALSE;
      break;
    default:
      if(INS_PREV(ins))
      switch(MIPSINSOPCODE(INS_PREV(ins)))
      {
	case MIPS_JR:
	case MIPS_BC1F:
	case MIPS_BC1T:
	  return TRUE;
	case MIPS_J:
	case MIPS_BNE:
	case MIPS_BEQ:
	case MIPS_BLEZ:
	case MIPS_BGEZ:
	case MIPS_BGTZ:
	case MIPS_BLTZ:
	  return FALSE;
	  break;
	default:
	  DiabloPrint(stdout,"@I\n",INS_PREV(ins));
	  FATAL(("IMPLEMENT ME"));
	  break;
      }
  }
  FATAL(("IMPLEMENT ME"));
  return TRUE;
}

t_bool MIPSInsIsProcedureCall(t_ins * ins)
{
  switch(MIPSINSOPCODE(ins))
  {
    case MIPS_JR:
    case MIPS_BC1F:
    case MIPS_BC1T:
    case MIPS_J:
    case MIPS_BNE:
    case MIPS_BEQ:
    case MIPS_BLEZ:
    case MIPS_BGEZ:
    case MIPS_BGTZ:
    case MIPS_BLTZ:
      return FALSE;
      break;
    case MIPS_JALR:
    case MIPS_JAL:
    case MIPS_BLTZAL:
    case MIPS_BGEZAL:
      return TRUE;
      break;
    default:
      if(INS_PREV(ins))
      switch(MIPSINSOPCODE(INS_PREV(ins)))
      {
	case MIPS_JR:
	case MIPS_BC1F:
	case MIPS_BC1T:
	case MIPS_J:
	case MIPS_BNE:
	case MIPS_BEQ:
	case MIPS_BLEZ:
	case MIPS_BGEZ:
	case MIPS_BGTZ:
	case MIPS_BLTZ:
	  return FALSE;
	  break;
	case MIPS_JALR:
	case MIPS_JAL:
	case MIPS_BLTZAL:
	case MIPS_BGEZAL:
	  return TRUE;
	  break;
	default:
	  DiabloPrint(stdout,"@I\n",INS_PREV(ins));
	  FATAL(("IMPLEMENT ME"));
	  break;
      }
  }
  DiabloPrint(stdout,"@I\n",(ins));
  FATAL(("IMPLEMENT ME"));
  return TRUE;
}

/* vim: set shiftwidth=2: */
