/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "mips_utils.h"
#include "mips_opcodes.h"
#include "mips_registers.h"
#include "mips_instructions.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_regset.h"

t_regset MipsUsedRegisters(t_ins * ins) 
{
  /* returns a bitmask specifying which registers are use by this instruction */
  /* the bottommost 32 bits represent the usual registers r0 - r31 */
  /* Then the registers SR, LO, HI, BAD, CAUSE, PC */

  t_regset mask = NullRegs;

  if(INS_TYPE(ins)==IT_UNKNOWN||
      INS_TYPE(ins)==IT_DATA||
      INS_TYPE(ins)==IT_NOP||
      ins==NULL)
    return NullRegs;

  if(MIPSINSPROCESSOR(ins)==CPU)
  {
  switch(INS_TYPE(ins))
  {
    case IT_DATAPROC:
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_SLL:
	case MIPS_SRL:
	case MIPS_SRA:
          RegsetAddReg(mask,MIPSINSRT(ins));
	  break;
	case MIPS_ADD:
	case MIPS_ADDU:
	case MIPS_SUB:
	case MIPS_SUBU:
	case MIPS_AND:
	case MIPS_OR:
	case MIPS_XOR:
	case MIPS_NOR:
	case MIPS_SLT:
	case MIPS_SLTU:
	case MIPS_SLLV:
	case MIPS_SRLV:
	case MIPS_SRAV:
	case MIPS_MOVZ:
	case MIPS_MOVN:
          RegsetAddReg(mask,MIPSINSRT(ins));
          RegsetAddReg(mask,MIPSINSRS(ins));
	  break;
	case MIPS_ADDI:
	case MIPS_ADDIU:
	case MIPS_SLTI:
	case MIPS_SLTIU:
	case MIPS_ANDI:
	case MIPS_ORI:
	case MIPS_XORI:
	case MIPS_CLZ:
	case MIPS_CLO:
          RegsetAddReg(mask,MIPSINSRS(ins));
	  break;
	case MIPS_LUI:
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_MUL:
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_MADDU:
	case MIPS_MADD:
	case MIPS_MSUB:
	case MIPS_MSUBU:
          RegsetAddReg(mask,MIPS_REG_LO);
	  RegsetAddReg(mask,MIPS_REG_HI);
          RegsetAddReg(mask,MIPSINSRT(ins));
          RegsetAddReg(mask,MIPSINSRS(ins));
	  break;
	case MIPS_MULT:
	case MIPS_MULTU:
	case MIPS_DIV:
	case MIPS_DIVU:
	case MIPS_MUL:
          RegsetAddReg(mask,MIPSINSRT(ins));
          RegsetAddReg(mask,MIPSINSRS(ins));
	  break;
	case MIPS_MTHI:
	case MIPS_MTLO:
          RegsetAddReg(mask,MIPSINSRS(ins));
	  break;
	case MIPS_MFLO:
          RegsetAddReg(mask,MIPS_REG_LO);
	  break;
	case MIPS_MFHI:
	  RegsetAddReg(mask,MIPS_REG_HI);
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_BRANCH:
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_JR:
          RegsetAddReg(mask,MIPSINSRS(ins));
	  break;
	case MIPS_JALR:
	case MIPS_BLTZ:
	case MIPS_BGEZ:
	case MIPS_BLTZAL:
	case MIPS_BGEZAL:
	case MIPS_BLEZ:
	case MIPS_BGTZ:
          RegsetAddReg(mask,MIPSINSRS(ins));
	  RegsetAddReg(mask,MIPS_REG_PC);
	  break;
	case MIPS_BEQ:
	case MIPS_BNE:
          RegsetAddReg(mask,MIPSINSRT(ins));
          RegsetAddReg(mask,MIPSINSRS(ins));
	  RegsetAddReg(mask,MIPS_REG_PC);
	  break;
	case MIPS_J:
	case MIPS_JAL:
	  RegsetAddReg(mask,MIPS_REG_PC);
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_STORE:
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_SB:
	case MIPS_SH:
	case MIPS_SW:
	case MIPS_SWL:
	case MIPS_SWR:
	case MIPS_SC:
          RegsetAddReg(mask,MIPSINSRT(ins));
          RegsetAddReg(mask,MIPSINSBASE(ins));
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_LOAD:
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_LB:
	case MIPS_LH:
	case MIPS_LW:
	case MIPS_LBU:
	case MIPS_LHU:
	case MIPS_LL:
          RegsetAddReg(mask,MIPSINSBASE(ins));
	  break;
	case MIPS_LWL:
	case MIPS_LWR:
	  RegsetAddReg(mask,MIPSINSBASE(ins));
	  RegsetAddReg(mask,MIPSINSRT(ins));
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_SWI:
      break;
    default:
      FATAL(("Implement type %d!",INS_TYPE(ins)));
      break;
  }
  }
  else if(MIPSINSPROCESSOR(ins)==COP0)
  {
    if(MIPSINSOPCODE(ins)==MIPS_MTC0)
      RegsetAddReg(mask,MIPSINSRT(ins));
  }
  else if(MIPSINSPROCESSOR(ins)==COP1)
  {
    switch(INS_TYPE(ins))
    {
      case IT_STORE:
      /*  RegsetAddReg(mask,MIPSINSFT(ins));
        RegsetAddReg(mask,MIPSINSFT(ins)+1);
        RegsetAddReg(mask,MIPSINSBASE(ins));*/
	//BEGIN
	switch (MIPSINSOPCODE(ins))
	{
	  case MIPS_SWC1:
	    RegsetAddReg(mask,MIPSINSFT(ins));
	    RegsetAddReg(mask,MIPSINSBASE(ins));
	    break;
	  case MIPS_SDC1:
	    RegsetAddReg(mask,MIPSINSFT(ins));
	    RegsetAddReg(mask,MIPSINSBASE(ins));
	    RegsetAddReg(mask,MIPSINSFT(ins)+1);
	    break;
	  default:
	    FATAL(("ERROR!\n"));
	    break;
	}
	//EINDE
	break;
      case IT_LOAD:
        RegsetAddReg(mask,MIPSINSBASE(ins));
	break;
      case IT_BRANCH:
	RegsetAddReg(mask,MIPS_REG_FCCR);
	break;
      case IT_DATAPROC:
	switch (MIPSINSOPCODE(ins))
	{
	  case MIPS_MTC1:
	  case MIPS_CTC1:
            RegsetAddReg(mask,MIPSINSRT(ins));
	    RegsetAddReg(mask,MIPS_REG_FCSR);
	    break;
	  case MIPS_C:
	    /*RegsetAddReg(mask,MIPSINSFS(ins));
	    RegsetAddReg(mask,MIPSINSFS(ins)+1);
	    RegsetAddReg(mask,MIPSINSFT(ins));
	    RegsetAddReg(mask,MIPSINSFT(ins)+1);
	    *///BEGIN
	    switch(MIPSINSFMT(ins))
	    {
	      case 16:
	      case 20:
		RegsetAddReg(mask,MIPSINSFS(ins));
		RegsetAddReg(mask,MIPSINSFT(ins));
		break;
	      case 17:
	      case 21:
		RegsetAddReg(mask,MIPSINSFS(ins));
		RegsetAddReg(mask,MIPSINSFS(ins)+1);
		RegsetAddReg(mask,MIPSINSFT(ins));
		RegsetAddReg(mask,MIPSINSFT(ins)+1);
		break;
	      default:
		FATAL(("ERROR!"));
		break;
	    }
	    //EINDE
	    break;
	  case MIPS_MOVZ_:
	  case MIPS_MOVN_:
	    /*
            RegsetAddReg(mask,MIPSINSRT(ins));
	    RegsetAddReg(mask,MIPSINSFS(ins));
	    RegsetAddReg(mask,MIPSINSFS(ins)+1);
	    */
	    //BEGIN
	    switch(MIPSINSFMT(ins))
	    {
	      case 16:
		RegsetAddReg(mask,MIPSINSRT(ins));
		RegsetAddReg(mask,MIPSINSFS(ins));
		break;
	      case 17:
		RegsetAddReg(mask,MIPSINSRT(ins));
		RegsetAddReg(mask,MIPSINSFS(ins));
		RegsetAddReg(mask,MIPSINSFS(ins)+1);
		break;
	      default:
		FATAL(("ERROR!"));
		break;
	    }
	    //EINDE
	    break;
	  case MIPS_MOVF:
	  case MIPS_MOVT:
            RegsetAddReg(mask,MIPSINSRS(ins));
	    break;
	  case MIPS_ADD_:
	  case MIPS_SUB_:
	  case MIPS_MUL_:
	  case MIPS_DIV_:
	    /*
	    RegsetAddReg(mask,MIPSINSFT(ins));
	    RegsetAddReg(mask,MIPSINSFT(ins)+1);
	    RegsetAddReg(mask,MIPS_REG_FCSR);
	    RegsetAddReg(mask,MIPSINSFS(ins));
	    RegsetAddReg(mask,MIPSINSFS(ins)+1);
	    */
	    //BEGIN
	    switch(MIPSINSFMT(ins))
	    {
	      case 16:
		RegsetAddReg(mask,MIPSINSFT(ins));
		RegsetAddReg(mask,MIPS_REG_FCSR);
		RegsetAddReg(mask,MIPSINSFS(ins));
		break;
	      case 17:
		RegsetAddReg(mask,MIPSINSFT(ins));
		RegsetAddReg(mask,MIPSINSFT(ins)+1);
		RegsetAddReg(mask,MIPS_REG_FCSR);
		RegsetAddReg(mask,MIPSINSFS(ins));
		RegsetAddReg(mask,MIPSINSFS(ins)+1);
		break;
	      default:
		FATAL(("ERROR!"));
		break;
	    }
	    //EINDE
	    break;
	  case MIPS_MFC1:
	    RegsetAddReg(mask,MIPS_REG_FCSR);
	    RegsetAddReg(mask,MIPSINSFS(ins));
	    break;
	    /*
	  case MIPS_CFC1:
	  case MIPS_SQRT:
	  case MIPS_CVT_D:
	  case MIPS_CVT_S:
	  case MIPS_CVT_W:
	    RegsetAddReg(mask,MIPS_REG_FCSR);
	    RegsetAddReg(mask,MIPSINSFS(ins));
	    RegsetAddReg(mask,MIPSINSFS(ins)+1);
	    break;
	    */
	    //BEGIN
	  case MIPS_CFC1:
	    RegsetAddReg(mask,MIPS_REG_FCSR);
	    RegsetAddReg(mask,MIPSINSFS(ins));
	    break;
	  case MIPS_SQRT:
	    switch(MIPSINSFMT(ins))
	    {
	      case 16:
		RegsetAddReg(mask,MIPS_REG_FCSR);
		RegsetAddReg(mask,MIPSINSFS(ins));
		break;
	      case 17:
		RegsetAddReg(mask,MIPS_REG_FCSR);
		RegsetAddReg(mask,MIPSINSFS(ins));
		RegsetAddReg(mask,MIPSINSFS(ins)+1);
		break;
	      default:
		FATAL(("ERROR!"));
		break;
	    }
	    break;
	  case MIPS_CVT_D:
	  case MIPS_CVT_S:
	  case MIPS_CVT_W:
	    switch(MIPSINSFMT(ins))
	    {
	      case 16:
	      case 20:
		RegsetAddReg(mask,MIPS_REG_FCSR);
		RegsetAddReg(mask,MIPSINSFS(ins));
		break;
	      case 17:
	      case 21:
		RegsetAddReg(mask,MIPS_REG_FCSR);
		RegsetAddReg(mask,MIPSINSFS(ins));
		RegsetAddReg(mask,MIPSINSFS(ins)+1);
		break;
	      default:
		FATAL(("ERROR!"));
		break;
	    }
	    break;
	    //EINDE
	  case MIPS_ABS:
	  case MIPS_NEG:
	  case MIPS_ROUND_W:
	  case MIPS_TRUNC_W:
	  case MIPS_CEIL_W:
	  case MIPS_FLOOR_W:
	  case MIPS_MOV:
	  case MIPS_MOVF_:
	  case MIPS_MOVT_:
	    /*
	    RegsetAddReg(mask,MIPSINSFS(ins));
	    RegsetAddReg(mask,MIPSINSFS(ins)+1);
	    break
	    */
	    //BEGIN
	    switch(MIPSINSFMT(ins))
	    {
	      case 16:
		RegsetAddReg(mask,MIPSINSFS(ins));
		break;
	      case 17:
		RegsetAddReg(mask,MIPSINSFS(ins));
		RegsetAddReg(mask,MIPSINSFS(ins)+1);
		break;
	      default:
		FATAL(("ERROR!"));
		break;
	    }
	    break;
	    //EINDE
	  default:
            FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	    break;
	}
	break;
      default:
        FATAL(("Implement ins_type %d!",INS_TYPE(ins)));
	break;
    }
	
  }
  return mask;
}

t_regset MipsDefinedRegisters(t_ins * ins) 
{
  /* returns a bitmask specifying which registers are use by this instruction */
  /* the bottommost 32 bits represent the usual registers r0 - r31 */
  /* Then the registers SR, LO, HI, BAD, CAUSE, PC */

  t_regset mask = NullRegs;

  if(INS_TYPE(ins)==IT_UNKNOWN||
      INS_TYPE(ins)==IT_DATA||
      INS_TYPE(ins)==IT_NOP||
      ins==NULL)
    return NullRegs;

  if(MIPSINSPROCESSOR(ins)==CPU)
  {
  switch(INS_TYPE(ins))
  {
    case IT_DATAPROC:
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_ADDI:
	case MIPS_ADDIU:
	case MIPS_SLTI:
	case MIPS_SLTIU:
	case MIPS_ANDI:
	case MIPS_ORI:
	case MIPS_XORI:
	case MIPS_LUI:
          RegsetAddReg(mask,MIPSINSRT(ins));
	  break;
	case MIPS_ADDU:
	case MIPS_ADD:
	case MIPS_SUB:
	case MIPS_SUBU:
	case MIPS_AND:
	case MIPS_OR:
	case MIPS_XOR:
	case MIPS_NOR:
	case MIPS_SLT:
	case MIPS_SLTU:
	case MIPS_CLZ:
	case MIPS_CLO:
	case MIPS_SLL:
	case MIPS_SRL:
	case MIPS_SRA:
	case MIPS_SLLV:
	case MIPS_SRLV:
	case MIPS_SRAV:
	  RegsetAddReg(mask,MIPSINSRD(ins));
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_MUL:
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_MTHI:
          RegsetAddReg(mask,MIPS_REG_HI);
	  break;
	case MIPS_MTLO:
          RegsetAddReg(mask,MIPS_REG_LO);
	  break;
	case MIPS_MFHI:
	case MIPS_MFLO:
	case MIPS_MUL:
          RegsetAddReg(mask,MIPSINSRD(ins));
	  break;
	case MIPS_MULT:
	case MIPS_MULTU:
	case MIPS_DIV:
	case MIPS_DIVU:
          RegsetAddReg(mask,MIPS_REG_LO);
          RegsetAddReg(mask,MIPS_REG_HI);
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_BRANCH:
      RegsetAddReg(mask,MIPS_REG_PC);
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_JR:
	case MIPS_BLTZ:
	case MIPS_BGEZ:
	case MIPS_BLEZ:
	case MIPS_BGTZ:
	case MIPS_BEQ:
	case MIPS_BNE:
	case MIPS_J:
	  break;
	case MIPS_JALR:
          RegsetAddReg(mask,MIPSINSRD(ins));
	  break;
	case MIPS_JAL:
	case MIPS_BLTZAL:
	case MIPS_BGEZAL:
          RegsetAddReg(mask,MIPS_REG_R31);
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_STORE:
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_SB:
	case MIPS_SH:
	case MIPS_SW:
	case MIPS_SWL:
	case MIPS_SWR:
	  break;
	case MIPS_SC:
          RegsetAddReg(mask,MIPSINSRT(ins));
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_LOAD:
      switch(MIPSINSOPCODE(ins))
      {
	case MIPS_LB:
	case MIPS_LH:
	case MIPS_LW:
	case MIPS_LBU:
	case MIPS_LHU:
	case MIPS_LWL:
	case MIPS_LWR:
	case MIPS_LL:
          RegsetAddReg(mask,MIPSINSRT(ins));
	  break;
	default:
          FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	  break;
      }
      break;
    case IT_SWI:
      break;
    default:
      FATAL(("Implement type %d!",INS_TYPE(ins)));
      break;
  }
  }
  else if(MIPSINSPROCESSOR(ins)==COP0)
  {
    switch (MIPSINSOPCODE(ins))
    {
      case MIPS_MFC0:
        RegsetAddReg(mask,MIPSINSRT(ins));
	break;
      case MIPS_MTC0:
	break;
      default:
        FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	break;
    }
  }
  else if(MIPSINSPROCESSOR(ins)==COP1)
  {
    switch(INS_TYPE(ins))
    {
      case IT_LOAD:
	/*
        RegsetAddReg(mask,MIPSINSFT(ins));
        RegsetAddReg(mask,MIPSINSFT(ins)+1);
	*/
	//BEGIN
	switch (MIPSINSOPCODE(ins))
	{
	  case MIPS_LWC1:
	    RegsetAddReg(mask,MIPSINSFT(ins));
	    break;
	  case MIPS_LDC1:
	    RegsetAddReg(mask,MIPSINSFT(ins));
	    RegsetAddReg(mask,MIPSINSFT(ins)+1);
	    break;
	  default:
	    FATAL(("ERROR!"));
	    break;
	}
	//EINDE
	break;
      case IT_STORE:
	break;
      case IT_BRANCH:
        RegsetAddReg(mask,MIPS_REG_PC);
	break;
      case IT_DATAPROC:
	switch (MIPSINSOPCODE(ins))
	{
	  case MIPS_MFC1:
	  case MIPS_CFC1:
            RegsetAddReg(mask,MIPSINSRT(ins));
	    break;
	  case MIPS_MOVF:
	  case MIPS_MOVT:
            RegsetAddReg(mask,MIPSINSRD(ins));
	    break;
	  case MIPS_MTC1:
	  case MIPS_CTC1:
	    RegsetAddReg(mask,MIPSINSFS(ins));
	    RegsetAddReg(mask,MIPS_REG_FCSR);
	    break;
	  case MIPS_C:
	    RegsetAddReg(mask,MIPS_REG_FCSR);
	    RegsetAddReg(mask,MIPS_REG_FCCR);
	    break;
	  case MIPS_ROUND_W:
	  case MIPS_TRUNC_W:
	  case MIPS_CEIL_W:
	  case MIPS_FLOOR_W:
	    RegsetAddReg(mask,MIPS_REG_FCSR);
	    RegsetAddReg(mask,MIPSINSFD(ins));
	    //RegsetAddReg(mask,MIPSINSFD(ins)+1);
	    break;
	    /*
	  case MIPS_SQRT:
	  case MIPS_ADD_:
	  case MIPS_SUB_:
	  case MIPS_MUL_:
	  case MIPS_DIV_:
	  case MIPS_ABS:
	  case MIPS_NEG:
	  case MIPS_CVT_D:
	  case MIPS_CVT_S:
	  case MIPS_CVT_W:
	  case MIPS_MOV:
	  case MIPS_MOVZ_:
	  case MIPS_MOVN_:
	  case MIPS_MOVF_:
	  case MIPS_MOVT_:
	    RegsetAddReg(mask,MIPSINSFD(ins));
	    RegsetAddReg(mask,MIPSINSFD(ins)+1);
	    break;
	    */
	    //BEGIN
	  case MIPS_SQRT:
	  case MIPS_ADD_:
	  case MIPS_SUB_:
	  case MIPS_MUL_:
	  case MIPS_DIV_:
	  case MIPS_ABS:
	  case MIPS_NEG:
	  case MIPS_MOV:
	  case MIPS_MOVZ_:
	  case MIPS_MOVN_:
	  case MIPS_MOVF_:
	  case MIPS_MOVT_:
	    switch(MIPSINSFMT(ins))
	    {
	      case 16:
		RegsetAddReg(mask,MIPSINSFD(ins));
		break;
	      case 17:
		RegsetAddReg(mask,MIPSINSFD(ins));
		RegsetAddReg(mask,MIPSINSFD(ins)+1);
		break;
	      default:
		FATAL(("ERROR!"));
		break;
	    }
	    break;
	  case MIPS_CVT_D:
	    RegsetAddReg(mask,MIPSINSFD(ins));
	    RegsetAddReg(mask,MIPSINSFD(ins)+1);
	    break;
	  case MIPS_CVT_S:
	  case MIPS_CVT_W:
	    RegsetAddReg(mask,MIPSINSFD(ins));
	    break;
	    //EINDE
	  default:
	    FATAL(("Implement type %d!",MIPSINSOPCODE(ins)));
	    break;
	}
	break;
      default:
        FATAL(("Implement type %d!",INS_TYPE(ins)));
	break;
    }
  }
  return mask;
}

/* vim: set shiftwidth=2: */
