/* Copyright {{{ 
 * 2001,2002,2003: Bruno De Bus
 *                 Bjorn De Sutter
 *                 Dominique Chanet
 *                 Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * }}} */

/* This code is based upon the threaded avl trees as implemented in libavl. It
 * is however strongly modified. The original copyright is included:
 * Original Copyright {{{ 
 * libavl: manipulates AVL trees.  Copyright (C) 1998, 1999 Free Software
 * Foundation, Inc.  
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.  
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.  
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * The author may be contacted at <pfaffben@pilot.msu.edu> on the Internet, or
 * as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA through more mundane
 * means. 
 * This is file avlt.c in libavl. 
 * }}} */ 
/* Includes {{{ */
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "data_avlt.h"
/* }}} */

/* Tag types. */
#define NON_THREADED +1
#define THREADED -1

#define NODE_TO_BASE(tree,x) ((void *) (((char *) x)-tree->node_offset))
#define BASE_TO_NODE(tree,x) ((t_threaded_avl_tree_node *) (((char *) x)+tree->node_offset))

/* ThreadedAvlTreeNew {{{ */
/*! Creates an AVL tree.  
 *
 * \param cmp is a order function for the data to be stored in the tree.  
 * 
 * \param param is arbitrary data that becomes an argument to the comparison
 * function. 
 */

t_threaded_avl_tree * ThreadedAvlTreeNew (t_uint32 node_offset, t_avl_comparison_func cmp)
{
  t_threaded_avl_tree *tree;

  VERIFY_FASSERT(cmp != NULL,("Threaded avl trees need a compare function. No compare function is supplied"));
  tree = Malloc(sizeof (t_threaded_avl_tree));

  tree->root.link[0] = tree->root.link[1] = &tree->root;
  tree->root.tag[0] = THREADED;
  tree->root.tag[1] = NON_THREADED;
  tree->node_offset=node_offset;
  tree->cmp = cmp;
  tree->count = 0;

  return tree;
}
/*}}}*/
/*ThreadedAvlTreeDestroy {{{ */
/*! Destroy tree.  
 *
 * \param tree is the tree that will be freed
 * \param free_func is called for every node in the tree as it is destroyed.
 */

void ThreadedAvlTreeDestroy (t_threaded_avl_tree *tree, t_avl_node_func free_func)
{
  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  
  if (tree->root.link[0] != &tree->root)
    {
      /* Uses Knuth's Algorithm 2.3.1T as modified in exercise 13
	 (postorder traversal). */
      
      /* T1. */
      t_threaded_avl_tree_node *an[AVL_MAX_HEIGHT];	/* Stack A: nodes. */
      char ab[AVL_MAX_HEIGHT];		/* Stack A: bits. */
      int ap = 0;			/* Stack A: height. */
      t_threaded_avl_tree_node *p = tree->root.link[0];

      for (;;)
	{
	  /* T2. */
	  for (;;)
	    {
	      /* T3. */
	      ab[ap] = 0;
	      an[ap++] = p;
	      if (p->tag[0] == THREADED)
		break;
	      p = p->link[0];
	    }

	  /* T4. */
	  for (;;)
	    {
	      if (ap == 0)
		goto done;

	      p = an[--ap];
	      if (ab[ap] == 0)
		{
		  ab[ap++] = 1;
		  if (p->tag[1] == THREADED)
		    continue;
		  p = p->link[1];
		  break;
		}
      
	      if (free_func)
		free_func (NODE_TO_BASE(tree,p));
	    }
	}
    }

 done:
  Free (tree);
}
/*}}}*/
/* ThreadedAvlTreeFree {{{ */
/*! avlt_destroy() with FREE_FUNC hardcoded as free(). */
void ThreadedAvlTreeFree (t_threaded_avl_tree *tree)
{
  ThreadedAvlTreeDestroy (tree, (t_avl_node_func) free);
}
/*}}}*/
/* ThreadedAvlTreeCount {{{ */
/*! Return the number of nodes in tree. */
int ThreadedAvlTreeCount (const t_threaded_avl_tree *tree)
{
  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  return tree->count;
}
/*}}}*/
/* ThreadedAvlTreeCopy {{{ */
/* Copy the contents of TREE to a new tree.  
 *
 * \param copy If copy is non-NULL, then each data item is passed to function
 * COPY, and the return values are inserted into the new tree; otherwise, the
 * items are copied verbatim from the old tree to the new tree.  
 *
 * \return the new tree. */

t_threaded_avl_tree *
ThreadedAvlTreeCopy (const t_threaded_avl_tree *tree, t_avl_copy_func copy)
{
  /* Knuth's Algorithm 2.3.1C (copying a binary tree).  Additionally
     uses Algorithm 2.3.1I (insertion into a threaded binary tree) and
     Algorithm 2.3.1 exercise 17 (preorder successor in threaded
     binary tree). */

  FATAL(("Implement"));
#if 0
  t_threaded_avl_tree *new_tree;

  const t_threaded_avl_tree_node *p;
  t_threaded_avl_tree_node *q;
  
  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  new_tree = ThreadedAvlTreeNew (tree->cmp, tree->param);
  new_tree->count = tree->count;
  p = &tree->root;
  if (p->link[0] == p)
    return new_tree;
  q = &new_tree->root;

  for (;;)
    {
      /* C4.  This is Algorithm 2.3.1I modified for insertion to the
         left.  Step I2 is not necessary. */
      if (p->tag[0] == NON_THREADED)
	{
	  t_threaded_avl_tree_node *r = Malloc(sizeof (t_threaded_avl_tree_node));

	  r->link[0] = q->link[0];
	  r->tag[0] = q->tag[0];
	  q->link[0] = r;
	  q->tag[0] = NON_THREADED;
	  r->link[1] = q;
	  r->tag[1] = THREADED;
	}

      /* C5: Find preorder successors of P and Q.  This is Algorithm
         2.3.1 exercise 17 but applies its actions to Q as well as
         P. */
      if (p->tag[0] == NON_THREADED)
	{
	  p = p->link[0];
	  q = q->link[0];
	}
      else
	{
	  while (p->tag[1] == THREADED)
	    {
	      p = p->link[1];
	      q = q->link[1];
	    }
	  p = p->link[1];
	  q = q->link[1];
	}

      /* C6. */
      if (p == &tree->root)
	{
	   VERIFY_FASSERT(tree != NULL,("NULL-tree"));
	  return new_tree;
	}
      
      /* C2.  This is Algorithm 2.3.1I.  Step I2 is not necessary. */
      if (p->tag[1] == NON_THREADED)
	{
	  t_threaded_avl_tree_node *r = Malloc(sizeof (t_threaded_avl_tree_node));

	  r->link[1] = q->link[1];
	  r->tag[1] = q->tag[1];
	  q->link[1] = r;
	  q->tag[1] = NON_THREADED;
	  r->link[0] = q;
	  r->tag[0] = THREADED;
	}

      /* C3. */
      q->bal = p->bal;
      if (copy == NULL)
	q->data = p->data;
      else
	q->data = copy (p->data, tree->param);
    }
#endif
  return NULL;
}
/*}}}*/
/* ThreadedAvlTreeWalk {{{ */
/* Walk tree TREE in inorder, calling walk_func at each node.  
 * 
 * \param walk_func is called at each node
 * \param param is passed to walk_func */
void ThreadedAvlTreeWalk (const t_threaded_avl_tree *tree, t_avl_node_func walk_func)
{
  const t_threaded_avl_tree_node *p;

  /* Uses Knuth's algorithm 2.3.1D (threaded inorder successor). */
  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  VERIFY_FASSERT(walk_func != NULL,("No walk function"));

  p = &tree->root;
  for (;;)
    {
      if (p->tag[1] == THREADED)
	p = p->link[1];
      else
	{
	  p = p->link[1];
	  while (p->tag[0] == NON_THREADED)
	    p = p->link[0];
	}

      if (p == &tree->root)
	return;

      walk_func (NODE_TO_BASE(tree,p));
    }
}


static void RecursePrint(FILE * fp, const t_threaded_avl_tree *tree, const t_threaded_avl_tree_node * node, t_avl_node_print_func print_func)
{
  if (node->tag[0] !=THREADED)
  {
    print_func(fp,NODE_TO_BASE(tree,node));
    fprintf(fp," -> ");
    print_func(fp,NODE_TO_BASE(tree,node->link[0]));
    fprintf(fp," \n ");
    fflush(fp);
    if (node->link [0] != &tree->root)
      RecursePrint(fp,tree,node->link[0],print_func);
  }
  if (node->tag[1] !=THREADED)
  {
    print_func(fp,NODE_TO_BASE(tree,node));
    fprintf(fp," -> ");
    print_func(fp,NODE_TO_BASE(tree,node->link[1]));
    fprintf(fp," \n ");
    fflush(fp);
    if (node->link [1] != &tree->root)
      RecursePrint(fp,tree,node->link[1],print_func);
  }


}
void ThreadedAvlTreePrint(FILE * fp, const t_threaded_avl_tree *tree, t_avl_node_print_func print_func)
{
   fprintf(fp,"digraph test {\n");
   RecursePrint(fp,tree,&tree->root,print_func);
   fprintf(fp,"}\n");
}

/*}}}*/
/* ThreadedAvlTreeTraverse {{{ */
/*! Each call to this function for a given TREE and TRAV return the next item
 * in the tree in inorder.  Initialize the first element of TRAV (init) to 0
 * before calling the first time.  Returns NULL when out of elements.  */
void * ThreadedAvlTreeTraverse (const t_threaded_avl_tree *tree, t_threaded_avl_tree_traverser *trav)
{
  const t_threaded_avl_tree_node *p;
  
  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  VERIFY_FASSERT(trav != NULL,("No traversal function"));

  if (trav->init == 0)
    {
      p = &tree->root;
      trav->init = 1;
    }
  else
    p = trav->p;

  /* Knuth's Algorithm 2.3.1S (threaded inorder successor). */
  if (p->tag[1] == THREADED)
    p = p->link[1];
  else
    {
      p = p->link[1];
      while (p->tag[0] == NON_THREADED)
	p = p->link[0];
    }

  if (p == &tree->root)
    {
      trav->init = 0;
      return NULL;
    }
  else
    {
      trav->p = p;
      return NODE_TO_BASE(tree,p);
    }
}
/*}}}*/
/* ThreadedAvlTreeNext {{{ */
/* Given ITEM, a pointer to a data item in TREE (or NULL), returns a pointer to
 * the next item in the tree in comparison order, or NULL if ITEM is the last
 * item. */
void * ThreadedAvlTreeNext (const t_threaded_avl_tree *tree, void *item)
{
  const t_threaded_avl_tree_node *p;

  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  if (item == NULL)
    p = &tree->root;
  else
    p = BASE_TO_NODE(tree,item); 

  /* Knuth's Algorithm 2.3.1S (threaded inorder successor). */
  if (p->tag[1] == THREADED)
    p = p->link[1];
  else
    {
      p = p->link[1];
      while (p->tag[0] == NON_THREADED)
	p = p->link[0];
    }

  if (p == &tree->root)
    return NULL;

  return NODE_TO_BASE(tree,p);
}
/*}}}*/
/* ThreadedAvlTreePrev {{{ */
/* Given ITEM, a pointer to a data item in TREE (or NULL), returns a pointer to
 * the previous item in the tree in comparison order, or NULL if ITEM is the
 * first item. */
void * ThreadedAvlTreePrev (const t_threaded_avl_tree *tree, void *item)
{
  const t_threaded_avl_tree_node *p;

  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  if (item == NULL)
    {
      /* Find node with greatest value. */
      p = tree->root.link[0];
      if (p == &tree->root)
	return NULL;
      while (p->tag[1] == NON_THREADED)
	p = p->link[1];
    }
  else
    {
      p = BASE_TO_NODE(tree,item); 

      /* Knuth's Algorithm 2.3.1S (threaded inorder successor)
	 modified to find the predecessor node. */
      if (p->tag[0] == THREADED)
	p = p->link[0];
      else
	{
	  VERIFY_FASSERT(p->tag[0] == NON_THREADED,("Malicious ordered tree"));
	  p = p->link[0];
	  while (p->tag[1] == NON_THREADED)
	    p = p->link[1];
	}
    }

  if (p == &tree->root)
    return NULL;

  return NODE_TO_BASE(tree,p);
}
/*}}}*/
/* ThreadedAvlTreeProbe {{{ */
/* Search TREE for an item matching ITEM.  If found, returns a pointer to the
 * address of the item.  If none is found, ITEM is inserted into the tree, and
 * a pointer to the address of ITEM is returned.  In either case, the pointer
 * returned can be changed by the caller, or the returned data item can be
 * directly edited, but the key data in the item must not be changed. */

void * ThreadedAvlTreeProbe (t_threaded_avl_tree *tree, void * item)
{
  /* Uses Knuth's Algorithm 6.2.3A (balanced tree search and
     insertion), modified for a threaded binary tree.  Caches results
     of comparisons.  In empirical tests this eliminates about 25% of
     the comparisons seen under random insertions.  */

  /* A1. */
  t_threaded_avl_tree_node * node_item=BASE_TO_NODE(tree,item);
  t_threaded_avl_tree_node *t;
  t_threaded_avl_tree_node *s, *p, *q, *r;

  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  t = &tree->root;
  s = p = t->link[0];

  if (t->tag[0] == THREADED)
    {
      tree->count++;
      VERIFY_FASSERT(tree->count  == 1,("Malicious tree"));
      t->tag[0] = NON_THREADED;
      q = t->link[0] = node_item; 
      q->link[0] = q->link[1] = t;
      q->tag[0] = THREADED; 
	 q->tag[1] = THREADED;
      q->bal = 0;
      return item; 
    }

  for (;;)
    {
      /* A2. */
      int diff = tree->cmp (item, NODE_TO_BASE(tree,p));

      /* A3. */
      if (diff < 0)
	{
	  p->cache = 0;
	  q = p->link[0];
	  if (p->tag[0] == THREADED)
	    {
	      q = node_item; 
	      q->link[0] = p->link[0];
	      q->tag[0] = p->tag[0];
	      p->link[0] = q;
	      p->tag[0] = NON_THREADED;
	      q->link[1] = p;
	      q->tag[1] = THREADED;
	      break;
	    }
	}
      /* A4. */
      else if (diff > 0)
	{
	  p->cache = 1;
	  q = p->link[1];
	  if (p->tag[1] == THREADED)
	    {
	      q = node_item; 
	      q->link[1] = p->link[1];
	      q->tag[1] = p->tag[1];
	      p->link[1] = q;
	      p->tag[1] = NON_THREADED;
	      q->link[0] = p;
	      q->tag[0] = THREADED;
	      break;
	    }
	}
      else
      {
	/* A2. */
	return NODE_TO_BASE(tree,p);
      }

      /* A3, A4. */
      if (q->bal != 0)
	t = p, s = q;
      p = q;
    }
  
  /* A5. */
  tree->count++;
  q->bal = 0;
  
  /* A6. */
  r = p = s->link[(int) s->cache];
  while (p != q)
    {
      p->bal = p->cache * 2 - 1;
      p = p->link[(int) p->cache];
    }

  /* A7. */
  if (s->cache == 0)
    {
      /* a = -1. */
      if (s->bal == 0)
	{
	  s->bal = -1;
	  return item; 
	}
      else if (s->bal == +1)
	{
	  s->bal = 0;
	  return item; 
	}
      
      VERIFY_FASSERT(s->bal==-1,("Malicious tree"));
      if (r->bal == -1)
	{
	  /* A8. */
	  p = r;
	  s->link[0] = r->link[1];
	  s->tag[0] = r->tag[1];
	  r->link[1] = s;
	  r->tag[1] = NON_THREADED;
	  if (s->link[0] == s)
	    {
	      s->link[0] = r;
	      s->tag[0] = THREADED;
	    }
	  r->tag[0] = r->tag[1] = NON_THREADED;
	  s->bal = r->bal = 0;
	}
      else
	{
	  /* A9. */
	  VERIFY_FASSERT(r->bal==1,("Malicious tree"));
	  p = r->link[1];
	  r->link[1] = p->link[0];
	  p->link[0] = r;
	  s->link[0] = p->link[1];
	  p->link[1] = s;
	  if (p->bal == -1)
	    s->bal = 1, r->bal = 0;
	  else if (p->bal == 0)
	    s->bal = r->bal = 0;
	  else 
	    {
	      VERIFY_FASSERT(p->bal==1,("Malicious tree"));
	      s->bal = 0, r->bal = -1;
	    }
	  p->bal = 0;
	  p->tag[0] = p->tag[1] = NON_THREADED;
	  if (s->link[0] == s)
	    {
	      s->link[0] = p;
	      s->tag[0] = THREADED;
	    }
	  if (r->link[1] == r)
	    {
	      r->link[1] = p;
	      r->tag[1] = THREADED;
	    }
	}
    }
  else
    {
      /* a == +1. */
      if (s->bal == 0)
	{
	  s->bal = 1;
	  return item;
	}
      else if (s->bal == -1)
	{
	  s->bal = 0;
	  return item;
	}

      VERIFY_FASSERT(s->bal==1,("Malicious tree"));
      if (r->bal == +1)
	{
	  /* A8. */
	  p = r;
	  s->link[1] = r->link[0];
	  s->tag[1] = r->tag[0];
	  r->link[0] = s;
	  r->tag[0] = NON_THREADED;
	  if (s->link[1] == s)
	    {
	      s->link[1] = r;
	      s->tag[1] = THREADED;
	    }
	  s->bal = r->bal = 0;
	}
      else
	{
	  /* A9. */
	  VERIFY_FASSERT(r->bal==-1,("Malicious tree"));
	  p = r->link[0];
	  r->link[0] = p->link[1];
	  p->link[1] = r;
	  s->link[1] = p->link[0];
	  p->link[0] = s;
	  if (p->bal == +1)
	    s->bal = -1, r->bal = 0;
	  else if (p->bal == 0)
	    s->bal = r->bal = 0;
	  else 
	    {
              VERIFY_FASSERT(p->bal==-1,("Malicious tree"));
	      s->bal = 0, r->bal = 1;
	    }
	  p->tag[0] = p->tag[1] = NON_THREADED;
	  if (s->link[1] == s)
	    {
	      s->link[1] = p;
	      s->tag[1] = THREADED;
	    }
	  if (r->link[0] == r)
	    {
	      r->link[0] = p;
	      r->tag[0] = THREADED;
	    }
	  p->bal = 0;
	}
    }
		
  /* A10. */
  if (t != &tree->root && s == t->link[1])
    t->link[1] = p;
  else
    t->link[0] = p;

  return item;
}
/*}}}*/
/* ThreadedAvlTreeFind {{{*/ 
/* Search TREE for an item matching ITEM, and return a pointer to it if found.
 * */
void ** ThreadedAvlTreeFind (const t_threaded_avl_tree *tree, const void *item)
{
  const t_threaded_avl_tree_node *p;

  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  if (tree->root.tag[0] == THREADED)
  {
    /* Tree is empty. */
    return NULL;
  }

  p = tree->root.link[0];
  for (;;)
    {
      int diff = tree->cmp (item, NODE_TO_BASE(tree,p));
      int t;

      /* A3. */
      if (diff < 0)
	t = 0;
      else if (diff > 0)
	t = 1;
      else
	return NODE_TO_BASE(tree,p);

      if (p->tag[t] == NON_THREADED)
      {
	p = p->link[t];
      }
      else
      {
	return NULL;
      }
    }
}
/*}}}*/
/* ThreadedAvlTreeFindClose {{{ */
/* Search TREE for an item close to the value of ITEM, and return it.  This
 * function will return a null pointer only if TREE is empty. */
void ** ThreadedAvlTreeFindClose (const t_threaded_avl_tree *tree, const void *item)
{
  const t_threaded_avl_tree_node *p;

  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  if (tree->root.tag[0] == THREADED)
    /* Tree is empty. */
    return NULL;

  p = tree->root.link[0];
  for (;;)
    {
      int diff = tree->cmp (item, NODE_TO_BASE(tree,p));
      int t;

      /* A3. */
      if (diff < 0)
	t = 0;
      else if (diff > 0)
	t = 1;
      else
	return NODE_TO_BASE(tree,p);

      if (p->tag[t] == NON_THREADED)
	p = p->link[t];
      else
	return NODE_TO_BASE(tree,p);
    }
}
/*}}}*/
/* ThreadedAvlTreeDelete {{{ */
/* Searches AVL tree TREE for an item matching ITEM.  If found, the item is
 * removed from the tree and the actual item found is returned to the caller.
 * If no item matching ITEM exists in the tree, returns NULL. */
void printfunc(void *, void *);
void * ThreadedAvlTreeDelete (t_threaded_avl_tree *tree, const void *item)
{
  /* Uses my Algorithm DT, which can be found at
     http://www.msu.edu/user/pfaffben/avl.  Algorithm DT is based on
     Knuth's Algorithms 6.2.2D (Tree deletion), 6.2.3A (Balanced tree
     search and insertion), 2.3.1I (Insertion into a threaded binary
     trees), and the notes on pages 465-466 of Vol. 3. */

  /* D1. */
  t_threaded_avl_tree_node *pa[AVL_MAX_HEIGHT];	/* Stack P: Nodes. */
  unsigned char a[AVL_MAX_HEIGHT];	/* Stack P: Bits. */
  int k = 1;				/* Stack P: Pointer. */
  
  t_threaded_avl_tree_node *p;
  void * ret;

  VERIFY_FASSERT(tree != NULL,("NULL-tree"));

  if (tree->root.tag[0] == THREADED)
    /* Empty tree. */
    return NULL;

  a[0] = 0;
  pa[0] = &tree->root;
  p = tree->root.link[0];


  /* Search for the element to delete, build the path to this node in pa, p */

  for (;;)
    {
      int diff = tree->cmp (item, NODE_TO_BASE(tree,p));

      if (diff == 0)
	break;

      pa[k] = p;
      if (diff < 0)
	{
	  if (p->tag[0] == NON_THREADED)
	    {
	      p = p->link[0];
	      a[k] = 0;
	    }
	  else
	    return NULL;
	}
      else if (diff > 0)
	{
	  if (p->tag[1] == NON_THREADED)
	    {
	      p = p->link[1];
	      a[k] = 1;
	    }
	  else
	    return NULL;
	}

      k++;
    }
  tree->count--;

  /* p holds the node that needs to be deleted */

  {
    t_threaded_avl_tree_node *to_delete = p;
    t_threaded_avl_tree_node **q = &pa[k - 1]->link[(int) a[k - 1]];

    /* D5. */
    if (to_delete->tag[1] == THREADED)
      {
	if (to_delete->tag[0] == NON_THREADED)
	  {
	    t_threaded_avl_tree_node *const x = to_delete->link[0];

	    *q = x;
	    (*q)->bal = 0;
	    if (x->tag[1] == THREADED)
	      {
		if (a[k - 1] == 1)
		  x->link[1] = to_delete->link[1];
		else
		  x->link[1] = pa[k - 1];
	      }
	  }
	else
	  {
	    pa[k- 1]->link[a[k-1]] = to_delete->link[a[k - 1]];
	    pa[k - 1]->tag[a[k - 1]] = THREADED;
	  }
      }
    else
      {
	/* D6. */
	t_threaded_avl_tree_node *r = to_delete->link[1];
	if (r->tag[0] == THREADED)
	  {
	    r->link[0] = to_delete->link[0];
	    r->tag[0] = to_delete->tag[0];
	    r->bal = to_delete->bal;
	    if (r->tag[0] == NON_THREADED)
	      {
		t_threaded_avl_tree_node *s = r->link[0];
		while (s->tag[1] == NON_THREADED)
		  s = s->link[1];
		VERIFY_FASSERT(s->tag[1] == THREADED,("Malformed tree"));
		s->link[1] = r;
	      }
	    *q = r;
	    a[k] = 1;
	    pa[k++] = r;
	  }
	else
	  {
	    t_uint32 delete_place=k;
	    /* D7.: Zoek het kleinste van de die groter dan het te verwijderen. 
	     *
	     *
	     *         TO DELETE
	     *           /    \ 
	     *                R     
	     *               /
	     *              S
	     *             T
	     *
	     * */
	    t_threaded_avl_tree_node *s = r->link[0];
	    t_threaded_avl_tree_node *largest_predecessor;

	    a[k] = 1;
	    pa[k++] = NULL;

	    a[k] = 0;
	    pa[k++] = r;
	    
	    /* D8. Naar links tot we het kleinste hebben */
	    while (s->tag[0] != THREADED)
	      {
		r = s;
		s = r->link[0];
		a[k] = 0;
		pa[k++] = r;
	      }

	    /* D9. */

	    /* The predecessor of the deleted node could be threaded, change
	     * the pointer */

	    if (to_delete->tag[0]==THREADED)
	    {
	      FATAL(("Impossible... tree must be corrupt"));
	    }
	    else
	    {
	      largest_predecessor=to_delete->link[0];
	      while (largest_predecessor->tag[1] != THREADED)
	      {
		largest_predecessor = largest_predecessor->link[1];
	      }

	      if (largest_predecessor->link[1]!=to_delete) FATAL(("Impossible... tree is corrupt"));
	      largest_predecessor->link[1]=s;
	    }

	    
	    if (s->tag[1] == THREADED)
	      {
		r->tag[0] = THREADED;
		r->link[0] = s;
	      }
	    else
	      {
		r->link[0] = s->link[1];
		if (s->link[1]->tag[0] == THREADED)
		  s->link[1]->link[0] = s;
	      }

	    s->link[0]=to_delete->link[0];
	    s->link[1]=to_delete->link[1];
	    s->tag[0]=to_delete->tag[0];
	    s->tag[1]=to_delete->tag[1];
	    s->bal=to_delete->bal;
	    pa[delete_place-1]->link[a[delete_place-1]]=s;
	    pa[delete_place]=s;
	  }
      }
  }

  ret=NODE_TO_BASE(tree,p);
  VERIFY_FASSERT(k>0,("Malformed tree"));
  /* D10. */
  while (--k)
    {
      t_threaded_avl_tree_node *const s = pa[k];

      if (a[k] == 0)
	{
	  t_threaded_avl_tree_node *const r = s->link[1];
	 
	  /* D10. */
	  if (s->bal == -1)
	    {
	      s->bal = 0;
	      continue;
	    }
	  else if (s->bal == 0)
	    {
	      s->bal = +1;
	      break;
	    }

	  VERIFY_FASSERT(s->bal==1,("Malformed tree"));
	  if (s->tag[1] == THREADED || r->bal == 0)
	    {
	      /* D11. */

	      s->link[1] = r->link[0];
	      r->link[0] = s;
	      r->bal = -1;
	      pa[k - 1]->link[(int) a[k - 1]] = r;
	      break;
	    }
	  else if (r->bal == +1)
	    {
	      /* D12. */
	      if (NON_THREADED == (s->tag[1] = r->tag[0]))
		s->link[1] = r->link[0];
	      r->link[0] = s;
	      r->tag[0] = NON_THREADED;
	      s->bal = r->bal = 0;
	      pa[k - 1]->link[a[k - 1]] = r;
	    }
	  else 
	    {
	      /* D13. */
	      VERIFY_FASSERT(r->bal==-1,("Malformed tree"));
	      p = r->link[0];
	      if (NON_THREADED == (r->tag[0] = p->tag[1]))
		r->link[0] = p->link[1];
	      p->link[1] = r;
	      p->tag[1] = NON_THREADED;
	      if (THREADED == (s->tag[1] = p->tag[0]))
		s->link[1] = p;
	      else
		s->link[1] = p->link[0];
	      p->link[0] = s;
	      p->tag[0] = NON_THREADED;
	      if (p->bal == +1)
		s->bal = -1, r->bal = 0;
	      else if (p->bal == 0)
		s->bal = r->bal = 0;
	      else
		{
		  VERIFY_FASSERT(p->bal==-1,("Malformed tree"));
		  s->bal = 0, r->bal = +1;
		}
	      p->bal = 0;
	      pa[k - 1]->link[(int) a[k - 1]] = p;
	      pa[k - 1]->tag[(int) a[k - 1]] = NON_THREADED;
	    }
	}
      else
	{
	  t_threaded_avl_tree_node *const r = s->link[0];
	 
	  /* D10. */
	  if (s->bal == +1)
	    {
	      s->bal = 0;
	      continue;
	    }
	  else if (s->bal == 0)
	    {
	      s->bal = -1;
	      break;
	    }

	  VERIFY_FASSERT(s->bal==-1,("Malformed tree"));
	  if (s->tag[0] == THREADED || r->bal == 0)
	    {
	      /* D11. */
	      s->link[0] = r->link[1];
	      r->link[1] = s;
	      r->bal = +1;
	      pa[k - 1]->link[(int) a[k - 1]] = r;
	      break;
	    }
	  else if (r->bal == -1)
	    {
	      /* D12. */
	      if (NON_THREADED == (s->tag[0] = r->tag[1]))
		s->link[0] = r->link[1];
	      r->link[1] = s;
	      r->tag[1] = NON_THREADED;
	      s->bal = r->bal = 0;
	      pa[k - 1]->link[a[k - 1]] = r;
	    }
	  else 
	    {
	      /* D13. */
	      VERIFY_FASSERT(r->bal==+1,("Malformed tree"));
	      p = r->link[1];
	      if (NON_THREADED == (r->tag[1] = p->tag[0]))
		r->link[1] = p->link[0];
	      p->link[0] = r;
	      p->tag[0] = NON_THREADED;
	      if (THREADED == (s->tag[0] = p->tag[1]))
		s->link[0] = p;
	      else
		s->link[0] = p->link[1];
	      p->link[1] = s;
	      p->tag[1] = NON_THREADED;
	      if (p->bal == -1)
		s->bal = +1, r->bal = 0;
	      else if (p->bal == 0)
		s->bal = r->bal = 0;
	      else
		{
		  VERIFY_FASSERT(p->bal==+1,("Malformed tree"));
		  s->bal = 0, r->bal = -1;
		}
	      p->bal = 0;
	      pa[k - 1]->link[(int) a[k - 1]] = p;
	      pa[k - 1]->tag[(int) a[k - 1]] = NON_THREADED;
	    }
	}
    }
      
  return ret; 
}
/*}}}*/
/* ThreadedAvlTreeInsert {{{ */
/* Inserts ITEM into TREE.  Returns NULL if the item was inserted, otherwise a
 * pointer to the duplicate item. */
void * ThreadedAvlTreeInsert (t_threaded_avl_tree *tree, void *item)
{
  void **p;
  
  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  
  p = ThreadedAvlTreeProbe (tree, item);
  return (*p == item) ? NULL : *p;
}
/*}}}*/
/* ThreadedAvlTreeReplace {{{ */
/* If ITEM does not exist in TREE, inserts it and returns NULL.  If a matching
 * item does exist, it is replaced by ITEM and the item replaced is returned.
 * The caller is responsible for freeing the item returned. */
void * ThreadedAvlTreeReplace (t_threaded_avl_tree *tree, void *item)
{
  void **p;

  VERIFY_FASSERT(tree != NULL,("NULL-tree"));
  
  p = ThreadedAvlTreeProbe (tree, item);
  if (*p == item)
    return NULL;
  else
    {
      void *r = *p;
      *p = item;
      return r;
    }
}
/*}}}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
