/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../config.h" 
#include <stdio.h>
#include <string.h> /* For memmove */
#include "../utils/util_malloc.h" 
#include "data_graph.h"
#include "../utils/util_error.h"
#include "../kernel/diablo_flowgraph.h"

t_uint32 global_node_marking_number=1;
t_uint32 global_node_marking_number2=1;
t_uint32 global_edge_marking_number=1;

void InitNodeMarking(void)
{
  if (++global_node_marking_number==0x7fffffff)
    {
      FATAL(("Marking number overflow\n"));
    }
}
void InitNodeMarking2(void)
{
  if (++global_node_marking_number2==0x7fffffff)
    {
      FATAL(("Marking number overflow\n"));
    }
}

void InitEdgeMarking(void)
{
  if (++global_edge_marking_number==0x7fffffff)
    {
      FATAL(("Edge marking number overflow\n"));
    }
}

void GraphInit(t_graph * temp, t_uint32 node_size, t_uint32 edge_size)
{
  temp->node_size=node_size;
  temp->edge_size=edge_size;
  temp->node_first=NULL;
  temp->node_last=NULL;
  temp->edge_first=NULL;
  temp->edge_last=NULL;
  temp->nnodes=0;
  temp->nedges=0;
}

t_graph * GraphNew(t_uint32 node_size, t_uint32 edge_size)
{
  t_graph * temp = Malloc(sizeof(t_graph));
  GraphInit(temp,node_size,edge_size);
  return temp;
}

void GraphInitNode(t_graph * graph, t_node * temp)
{
  
  NODE_NEXT(temp) = NULL;
  if (graph->node_last)
    {
      NODE_PREV(temp)=graph->node_last;
      NODE_NEXT(graph->node_last)=temp;
      graph->node_last=temp;
    }
  else
    {
      NODE_PREV(temp)=NULL;
      graph->node_first=graph->node_last=temp;
    }

  graph->nnodes++;
  temp->marked_number=0;
  NODE_PRED_FIRST(temp)=NODE_PRED_LAST(temp)=NODE_SUCC_FIRST(temp)=NODE_SUCC_LAST(temp)=NULL;
}

t_node * GraphNewNode(t_graph * graph, t_uint32 node_category)
{
  t_node * temp;
  VERIFY_FASSERT(graph->node_size,("Can't use GraphNodeNew when node_size == 0")); 
  temp = Malloc(graph->node_size);


  /*  NodeSetCategory(temp,node_category);*/
  GraphInitNode(graph,temp);
  return temp;
}

t_edge * GraphNewEdge(t_graph * graph, t_node * head, t_node * tail, t_uint32 edge_category)
{
  t_edge * temp;

  temp = Calloc(1,graph->edge_size);

  EDGE_HEAD(temp) = head;
  EDGE_TAIL(temp) = tail;
  EdgeSetCategory(temp,edge_category);
  EDGE_NEXT(temp) = NULL;

  if (NODE_SUCC_LAST(head))
    {
      EDGE_SUCC_NEXT(temp)=NULL;
      EDGE_SUCC_PREV(temp)=NODE_SUCC_LAST(head);
      EDGE_SUCC_NEXT(NODE_SUCC_LAST(head))=temp;
      NODE_SUCC_LAST(head)=temp;
    }
  else
    {
      EDGE_SUCC_NEXT(temp)=EDGE_SUCC_PREV(temp)=NULL;
      NODE_SUCC_LAST(head)=NODE_SUCC_FIRST(head)=temp;
    }

 
  if (NODE_PRED_LAST(tail))
    {
      EDGE_PRED_NEXT(temp)=NULL;
      EDGE_PRED_PREV(temp)=NODE_PRED_LAST(tail);
      EDGE_PRED_NEXT(NODE_PRED_LAST(tail))=temp;
      NODE_PRED_LAST(tail)=temp;
    }
  else
    {
      EDGE_PRED_NEXT(temp)=EDGE_PRED_PREV(temp)=NULL;
      NODE_PRED_LAST(tail)=NODE_PRED_FIRST(tail)=temp;
    }

  
  /* Automatically insert into the global edge list */
  if (graph->edge_last) 
    {
      EDGE_NEXT(graph->edge_last) = temp;
      EDGE_PREV(temp) = graph->edge_last;
      graph->edge_last = temp;
    } 
  else 
    {
      graph->edge_first = graph->edge_last = temp;
      EDGE_PREV(temp) = NULL;
    }
  return temp;
}

/* removes an edge (that isn't connected to any nodes) from the linked list of edges */
static void GraphKillEdge(t_graph * graph, t_edge * edge)
{

  /* remove from list of edges */
  if (edge == graph->edge_first) 
    {
      graph->edge_first = EDGE_NEXT(edge);
      if (graph->edge_first) 
	EDGE_PREV(graph->edge_first) = NULL;
      else 
	graph->edge_first = graph->edge_last = NULL;
    } 
  else 
    {
      if (edge == graph->edge_last) 
	{
	  graph->edge_last = EDGE_PREV(edge);
	  if (graph->edge_last) 
	    EDGE_NEXT(graph->edge_last) = NULL;
	} 
      else 
	{
	  EDGE_PREV(EDGE_NEXT(edge)) = EDGE_PREV(edge);
	  EDGE_NEXT(EDGE_PREV(edge)) = EDGE_NEXT(edge);
	}
    }
  
  graph->nedges--;

  /* free the memory */
  Free(edge);
}

static void GraphKillNodeNoFree(t_graph * graph, t_node * node)
{

  /* remove from list of nodes */
  if (node == graph->node_first) 
    {
      graph->node_first = NODE_NEXT(node);
      if (graph->node_first) 
	NODE_PREV(graph->node_first) = NULL;
      else 
	graph->node_first = graph->node_last = NULL;
    } 
  else 
    {
      if (node == graph->node_last) 
	{
	  graph->node_last = NODE_PREV(node);
	  if (graph->node_last) 
	    NODE_NEXT(graph->node_last) = NULL;
	} 
      else 
	{
	  NODE_PREV(NODE_NEXT(node)) = NODE_PREV(node);
	  NODE_NEXT(NODE_PREV(node)) = NODE_NEXT(node);
	}
    }

  /* zero out all node-related fields (necessary in case we're not working 
   * with a bare-bones node, but a structure derived from t_node that is 
   * allocated and deallocated separately. e.g. t_ins in diablo */
  NODE_NEXT(node) = NODE_PREV(node) = NULL;
  NODE_SUCC_FIRST(node) = NODE_SUCC_LAST(node) = NULL;
  NODE_PRED_FIRST(node) = NODE_PRED_LAST(node) = NULL;
  node->marked_number = 0;
  NODE_SUCC_FIRST(node) = NODE_SUCC_LAST(node) = NULL;

  graph->nnodes--;
}

static void GraphKillNode(t_graph * graph, t_node * node)
{
  GraphKillNodeNoFree(graph, node);
  Free(node);
}

/* removes an edge from the graph, this means: remove it from the nodes AND from the linked list of edges */
void GraphRemoveEdge(t_graph * graph, t_edge * edge) 
{
  t_node * tmp_node;

  if (!edge) FATAL(("Null edge!"));
  /* Delete from predecessor list */

  tmp_node = EDGE_TAIL(edge);
  tmp_node = EDGE_HEAD(edge);

  if (EDGE_PRED_PREV(edge))
    EDGE_PRED_NEXT(EDGE_PRED_PREV(edge))=EDGE_PRED_NEXT(edge);
  else
    {
      /* removing the first predecessor */
      NODE_PRED_FIRST(EDGE_TAIL(edge))=EDGE_PRED_NEXT(edge);
    }
  
  if (EDGE_PRED_NEXT(edge))
    EDGE_PRED_PREV(EDGE_PRED_NEXT(edge))=EDGE_PRED_PREV(edge);
  else
    {
      /* removing the last predecessor */
      NODE_PRED_LAST(EDGE_TAIL(edge))=EDGE_PRED_PREV(edge);
    }

  /* Delete from successor list */
  if (EDGE_SUCC_PREV(edge))
    EDGE_SUCC_NEXT(EDGE_SUCC_PREV(edge))=EDGE_SUCC_NEXT(edge);
  else
    {
      /* removing the first predecessor */
      NODE_SUCC_FIRST(EDGE_HEAD(edge))=EDGE_SUCC_NEXT(edge);
    }

  if (EDGE_SUCC_NEXT(edge))
    EDGE_SUCC_PREV(EDGE_SUCC_NEXT(edge))=EDGE_SUCC_PREV(edge);
  else
    {
      /* removing the first predecessor */
      NODE_SUCC_LAST(EDGE_HEAD(edge))=EDGE_SUCC_PREV(edge);
    }

  /* now kill the edge itself */
  GraphKillEdge(graph, edge);
}

void GraphRemoveNode(t_graph * graph, t_node * node)
{
  if (NODE_PREV(node))
    NODE_NEXT(NODE_PREV(node))=NODE_NEXT(node);
  
  if (NODE_NEXT(node))
    NODE_PREV(NODE_NEXT(node))=NODE_PREV(node);

  /* now kill the node itself */
  GraphKillNode(graph, node);
  
}

/* completely remove node from graph, but don't free the node itself */
void GraphRemoveNodeNoFree(t_graph * graph, t_node * node)
{
  if (NODE_PREV(node))
    NODE_NEXT(NODE_PREV(node))=NODE_NEXT(node);
  
  if (NODE_NEXT(node))
    NODE_PREV(NODE_NEXT(node))=NODE_PREV(node);

  /* now kill the node itself */
  GraphKillNodeNoFree(graph, node);
}
    
void GraphRemoveAllEdgesFromTo(t_graph * graph, t_node * head, t_node * tail)
{

  t_edge * edge;
  t_edge * next;
   
  for(edge=NODE_SUCC_FIRST(head); edge!=NULL; edge=next)
    {
      next=EDGE_SUCC_NEXT(edge);
      if (EDGE_TAIL(edge)==tail) 
	{
	  GraphRemoveEdge(graph,edge);
	}
    }
}

t_bool NodeIsPred(t_node * succ, t_node * pred, t_uint32 edge_category) /* Arg2 is predecessor of Arg1 */
{
  t_edge * edge;
  if (NODE_SUCC_FIRST(pred)==NULL) return FALSE;
  if (NODE_PRED_FIRST(succ)==NULL) return FALSE;

  NODE_FOREACH_SUCC_EDGE(pred,edge)
    if (EDGE_TAIL(edge)==succ && edge->category & edge_category) 
      return TRUE;
  
  return FALSE;
}

t_bool NodeIsSucc(t_node * pred, t_node * succ, t_uint32 edge_category) /* Arg2 is successor of Arg1*/
{
  return (NodeIsPred(succ,pred,edge_category));
}

t_edge * GraphPrependNode(t_graph * graph, t_node * before, t_node * insert, t_uint32 edge_category) /*                                                                                                             Prepend arg2 to arg1 */
{
  return  GraphNewEdge(graph, insert, before, edge_category);
}

t_edge * GraphAppendNode(t_graph * graph, t_node * after, t_node * insert, t_uint32 edge_category) /*  Append arg2 to arg1 */
{
  return GraphNewEdge(graph, after, insert, edge_category);
}

void GraphWhateverTraversalWithCheckAndData(t_graph * graph, t_node * start, 
					    t_bool (*CallBack)(t_node * curr, t_node * par,void * data), 
					    t_uint32 edge_category_or_mask,
					    void * data)
{
  t_edge * edge;
  t_node * node=start;
  t_bool change;
  
  InitNodeMarking();

  if (CallBack(node,NULL,data))
    return;
  
  NodeMark(node);
  
  do 
    {
      change = FALSE;
      GRAPH_FOREACH_NODE(graph,node)
	{
	  if (!NodeIsMarked(node)) continue;
	  NodeUnmark(node);
	  NODE_FOREACH_SUCC_EDGE(node,edge)
	    if (EdgeTestCategoryOr(edge,edge_category_or_mask) && !CallBack(EDGE_TAIL(edge),node,data))
	      {
		/*		DiabloPrint(stdout,"@B\n",T_BBL(EDGE_TAIL(edge)));*/
		NodeMark(EDGE_TAIL(edge));
		change = TRUE;
	      }
	}
    } while (change);
}


void GraphDFTraversalWithCheckAndData(t_node * start, 
				      t_bool (*CallBack)(t_node * curr, t_node * par,void * data), 
				      t_uint32 edge_category_or_mask,
				      void * data)
{
  t_node ** pstack=Malloc(sizeof(t_node *) * CONFIG_GT_STACK_ALLOC_SIZE);
  t_edge ** nstack=Malloc(sizeof(t_edge *) * CONFIG_GT_STACK_ALLOC_SIZE);
  int stack_size=CONFIG_GT_STACK_ALLOC_SIZE;
  int stack_pointer=0;
  t_edge * current_succ_edge=NODE_SUCC_FIRST(start);
  t_node * current_node=start;
  t_node * par=NULL; 
  t_bool ret=FALSE;

  InitNodeMarking();

  ret=CallBack(current_node,par,data);

  do
  {
    NodeMark(current_node);
    if ((!current_succ_edge) || (ret))
    {
      ret=FALSE;
      if (stack_pointer)
      {
	/* Pop old successors (and nodes) form the stack */
	current_node=pstack[stack_pointer-1];
	current_succ_edge=nstack[stack_pointer-1];
	stack_pointer--;
	current_succ_edge=EDGE_SUCC_NEXT(current_succ_edge);
      }
      else
      {
	break;
      }
    }
    else
    {
      if (!NodeIsMarked(EDGE_TAIL(current_succ_edge)) && EdgeTestCategoryOr(current_succ_edge,edge_category_or_mask))
      {
	if (stack_pointer==stack_size)
	{
	  nstack=Realloc(nstack,(stack_size+CONFIG_GT_STACK_ALLOC_SIZE)*sizeof(t_edge *));
	  pstack=Realloc(pstack,(stack_size+CONFIG_GT_STACK_ALLOC_SIZE)*sizeof(t_node *));
	  stack_size+=CONFIG_GT_STACK_ALLOC_SIZE;
	}
	par=current_node;
	nstack[stack_pointer]=current_succ_edge;
	pstack[stack_pointer]=current_node;
	stack_pointer++;

	/*DiabloPrint(stdout, "@iB\n", current_node);*/

	current_node=EDGE_TAIL(current_succ_edge);
	current_succ_edge=NODE_SUCC_FIRST(current_node);
	ret=CallBack(current_node,par,data);
      }
      else
      {
	current_succ_edge=EDGE_SUCC_NEXT(current_succ_edge);
      }
    }
  } while (1);


  Free(nstack);
  Free(pstack);
}

typedef struct _bf_queue_node {
  t_node * node;
  t_node * parent;
  struct _bf_queue_node * next;
} bf_queue_node;

void GraphBFTraversal(t_node * start, void (*CallBack)(t_node * curr, t_node * par), t_uint32 edge_category_or_mask)
{

  bf_queue_node * Qhead = NULL;
  bf_queue_node * Qtail = NULL;
  bf_queue_node * temp;
  t_edge * edge;

  t_node * curr;
  t_node * par; 

  InitNodeMarking();

  if (start != NULL) 
    {
      Qhead = Qtail = Malloc(sizeof(bf_queue_node));
      Qhead->node = start;
      NodeMark(Qhead->node);
      Qhead->parent = NULL;
      Qhead->next = NULL;
    }
  
  while (Qhead != NULL)
    {
      
      curr = Qhead->node;
      par = Qhead->parent;
      
      CallBack(curr,par);
      
      /* add successors to the queue */
      NODE_FOREACH_SUCC_EDGE(curr,edge) 
	{
	  if (!NodeIsMarked(edge->tail) &&(edge->category & edge_category_or_mask)) 
	    { /* only add to queue if not already processed or in queue */
	      Qtail->next = Malloc(sizeof(bf_queue_node)); 
	      Qtail = Qtail->next; 
	      Qtail->next = NULL; 
	      Qtail->node = edge->tail;
	      NodeMark(edge->tail);
	      Qtail->parent = curr;
	    }
	}
      
      /* remove the element we just processed from the queue */
      temp = Qhead;
      Qhead = Qhead->next;
      Free(temp);
      
    }
  
}
#if 0

void PrintIt(t_node * n, t_node * par)
{
  static int i=0;
  printf("%d: %x (parent=%x)\n",++i,(int) n, (int) par);
}
#endif
/* void testcb(t_node * node, t_node * parent) { */
/*  if (parent != NULL) printf("ouder: %skind: %s",(char *) (parent->data), (char *) (node->data));  */
/* } */

/* int main(int argc, char ** argv) { */
/*   int i,j; */

/*   t_node * graph; */
/*   char * tekst = "dit is de hoofdnode\n"; */
/*   char * ctekst[3] = {"000\n", "001\n", "002\n"}; */
/*   char * jtekst[9] = {"k00\n", "k01\n","k02\n","k10\n", "k11\n","k12\n","k20\n", "k21\n","k22\n"}; */

/*   graph = GraphNew(tekst); */

/*   for (i = 0; i < 3; i++) { */
/*     t_node * child; */
/*     GraphRealAppendNode(graph,child = GraphNew(ctekst[i])); */

/*     for (j = 0; j < 3; j++) */
/*       GraphRealAppendNode(child, GraphNew(jtekst[i*3+j])); */
/*   }; */

/*   GraphBFSTraverse(graph, testcb); */
/*   // GraphDFSTraverse(graph, testcb); */
/* }; */



#if 0
typedef struct _dfs_bt_data
{
  int child;
  t_node * parent;
} dfs_bt_data;

static int GraphDomSetup(t_node * start, t_node *** dfs_array, t_node *** dfs_parents) 
{
  dfs_bt_data * stack=NULL;
  int stack_size=0;
  int stack_pointer=0;
  int current_succ=0;
  t_node * curr=start;
  t_node * par=NULL; 
  int curflag=globflag++;
  int ret=1;

  *dfs_array=Malloc(ret*sizeof(t_node));
  *dfs_parents=Malloc(ret*sizeof(t_node));
  (*dfs_array)[ret-1]=curr;
  (*dfs_parents)[ret-1]=par;
  
  curr->data=(void *) (ret-1);

  do
    {
      curr->flag=curflag;
      if (curr->nsuccs==current_succ)
	{
	  if (stack_pointer)
	    {
	      curr=stack[stack_pointer-1].parent;
	      current_succ=stack[stack_pointer-1].child;
	      stack_pointer--;
	      current_succ++;
	    }
	  else break;
	}
      else
	{
	  if (curr->succs[current_succ]->flag!=curflag)
	    {
	      if (stack_pointer==stack_size)
		{
		  stack=Realloc(stack,(stack_size+CONFIG_GT_STACK_ALLOC_SIZE)*sizeof(dfs_bt_data));
		  stack_size+=CONFIG_GT_STACK_ALLOC_SIZE;
		}
	      par=curr;
	      stack[stack_pointer].child=current_succ;
	      stack[stack_pointer].parent=curr;
	      stack_pointer++;
	      curr=curr->succs[current_succ];
	      current_succ=0;
	      ret++;
	      *dfs_array=Realloc(*dfs_array,ret*sizeof(t_node));
	      *dfs_parents=Realloc(*dfs_parents,ret*sizeof(t_node));
	      (*dfs_array)[ret-1]=curr;
	      (*dfs_parents)[ret-1]=par;
	      curr->data=(void *) (ret-1);
	    }
	  else current_succ++;
	}
    } while (1);


  Free(stack);
  return ret;
}


static void DomCompress(int tc, int * ancestor, int *label, int *semi)
{
  if (ancestor[ancestor[tc]]!=-1)
    {
      DomCompress(ancestor[tc],ancestor,label,semi);
      if (semi[label[ancestor[tc]]] < semi[label[tc]])
	{
	  label[tc]=label[ancestor[tc]];
	}
      ancestor[tc]=ancestor[ancestor[tc]];
    }
}

typedef struct _bucket
{
  struct _bucket * next;
  int data;
} bucket;


void GraphComputeDominatorTree(t_node * start)
{
  t_node ** dfs_array=NULL;
  t_node ** dfs_parents=NULL;
  int nnodes=GraphDomSetup(start,&dfs_array,&dfs_parents);
  int * semi=Malloc(nnodes*sizeof(int));
  int * ancestor=Malloc(nnodes*sizeof(int)); 
  int * label=Malloc(nnodes*sizeof(int)); 
  int tel,tel2,tmp;
  bucket ** buckets=Malloc(nnodes*sizeof(bucket *));
  bucket * bucket_tmp,* bucket_next;
  printf("%d nodes in graph\n",nnodes);
  for (tel=nnodes-1; tel>=0; tel--) semi[tel]=tel;
  for (tel=nnodes-1; tel>=0; tel--) label[tel]=tel;
  for (tel=nnodes-1; tel>=0; tel--) ancestor[tel]=-1;
  for (tel=nnodes-1; tel>=0; tel--) buckets[tel]=NULL;

  for (tel=nnodes-1; tel>0; tel--)
    {
      /*printf("%3d: %x (%d) (parent %x (%d))\n",tel,dfs_array[tel],dfs_array[tel]->data,dfs_parents[tel],dfs_parents[tel]?dfs_parents[tel]->data:-1);  */
      for (tel2=0; tel2<dfs_array[tel]->npreds; tel2++)
	{
	  /* EVAL */
	  tmp=(int) dfs_array[tel]->preds[tel2]->data; 
	  if (ancestor[tmp]!=-1)
	    {
	      /* COMPRESS tmp */
	      DomCompress(tmp,ancestor,label,semi);
	      /* END COMPRESS */ 
	      tmp=label[tmp];
	    }
	  /* END EVAL */
	  if (semi[tmp]<semi[tel]) semi[tel]=semi[tmp];
	  /* add to bucket */
	  bucket_tmp=Malloc(sizeof(bucket));
	  bucket_tmp->data=tel;
	  bucket_tmp->next=buckets[semi[tel]];
	  buckets[semi[tel]]=bucket_tmp;
	  /* end add */
	    
	  /* LINK */
	  ancestor[tel]=(int) dfs_parents[tel]->data;
	  /* END LINK */

	  for (bucket_tmp=buckets[(int) dfs_parents[tel]->data]; bucket_tmp!=NULL; bucket_tmp=bucket_next)
	    {
	      /* EVAL */
	      /* END EVAL */
	      bucket_next=bucket_tmp->next;
	      Free(bucket_tmp);
	    }

	  buckets[(int) dfs_parents[tel]->data]=NULL;

	  /*printf("Pred %d: %x (%d) (semi==%d)\n",tel2,dfs_array[tel]->preds[tel2],dfs_array[tel]->preds[tel2]->data,semi[tel]); */
	}
      /*	printf("SEMI=%d\n",semi[tel]); */
    }

  Free(dfs_array);
  Free(dfs_parents);
  Free(semi);
  Free(label);
  Free(ancestor);
}

#endif

void GraphDFReverseTraversalWithCheckAndData(t_node * start, 
				      t_bool (*CallBack)(t_node * curr, t_node * par,void * data), 
				      t_uint32 edge_category_or_mask,
				      void * data)
{
  t_node ** pstack=Malloc(sizeof(t_node *) * CONFIG_GT_STACK_ALLOC_SIZE);
  t_edge ** nstack=Malloc(sizeof(t_edge *) * CONFIG_GT_STACK_ALLOC_SIZE);
  int stack_size=CONFIG_GT_STACK_ALLOC_SIZE;
  int stack_pointer=0;
  t_edge * current_pred_edge=NODE_PRED_FIRST(start);
  t_node * current_node=start;
  t_node * par=NULL; 
  t_bool ret=FALSE;

  InitNodeMarking();
    
  ret=CallBack(current_node,par,data);

  do
    {
      NodeMark(current_node);
      if ((!current_pred_edge) || (ret))
	{
	  ret=FALSE;
	  if (stack_pointer)
	    {
	      /* Pop old successors (and nodes) form the stack */
	      current_node=pstack[stack_pointer-1];
	      current_pred_edge=nstack[stack_pointer-1];
	      stack_pointer--;
	      current_pred_edge=EDGE_PRED_NEXT(current_pred_edge);
	    }
	  else
	    {
	      break;
	    }
	}
      else
	{
	  if (!NodeIsMarked(EDGE_HEAD(current_pred_edge)) && EdgeTestCategoryOr(current_pred_edge,edge_category_or_mask))
	    {
	      if (stack_pointer==stack_size)
		{
		  nstack=Realloc(nstack,(stack_size+CONFIG_GT_STACK_ALLOC_SIZE)*sizeof(t_edge *));
		  pstack=Realloc(pstack,(stack_size+CONFIG_GT_STACK_ALLOC_SIZE)*sizeof(t_node *));
		  stack_size+=CONFIG_GT_STACK_ALLOC_SIZE;
		}
	      par=current_node;
	      nstack[stack_pointer]=current_pred_edge;
	      pstack[stack_pointer]=current_node;
	      stack_pointer++;
	      current_node=EDGE_HEAD(current_pred_edge);
	      current_pred_edge=NODE_PRED_FIRST(current_node);
	      ret=CallBack(current_node,par,data);
	    }
	  else
	    {
	      current_pred_edge=EDGE_PRED_NEXT(current_pred_edge);
	    }
	}
    } while (1);


  Free(nstack);
  Free(pstack);
}

void GraphOrder(t_graph * graph, NodeCmpFun fun){
  /* Orders the lists of nodes, edges, successors and predecessors {{{*/
  t_node * current_node;
  t_node * next_node;
  t_edge * current_edge;
  t_edge * next_edge;
  int nr_cessors;
  int i;
  
  /*order nodes {{{*/
  for(i=0;i<GRAPH_NNODES(graph);i++){
    
    current_node = GRAPH_NODE_FIRST(graph);
    if(current_node==NULL)
      return;
    next_node = NODE_NEXT(current_node);

    while(next_node != NULL){
      if(fun(current_node, next_node)==1)
      {

	NODE_PREV(next_node)=NODE_PREV(current_node);
	NODE_NEXT(current_node)=NODE_NEXT(next_node);

	if(NODE_PREV(next_node)==NULL)
	  GRAPH_NODE_FIRST(graph)=next_node;
	else NODE_NEXT(NODE_PREV(next_node)) = next_node;

	if(NODE_NEXT(current_node)==NULL)
	  GRAPH_NODE_LAST(graph)=current_node;
	else NODE_PREV(NODE_NEXT(current_node)) = current_node;

	NODE_NEXT(next_node)=current_node;
	NODE_PREV(current_node)=next_node;
      }
      else{
       	current_node = NODE_NEXT(current_node);      
      }
      next_node = NODE_NEXT(current_node);
    }
  }
  /*}}}*/

  /*Order list of edges {{{*/
  for(i=0;i<GRAPH_NEDGES(graph);i++){
    
    current_edge = GRAPH_EDGE_FIRST(graph);
    if(current_edge==NULL)
      return;
    next_edge = EDGE_NEXT(current_edge);

    while(next_edge != NULL){
      t_uint8 tmp = fun(EDGE_TAIL(current_edge), EDGE_TAIL(next_edge));
      t_uint8 tmp2 = fun(EDGE_HEAD(current_edge), EDGE_HEAD(next_edge));
      if(tmp==1 || (tmp==0&&tmp2==1))
      {
	EDGE_PREV(next_edge)=EDGE_PREV(current_edge);
	EDGE_NEXT(current_edge)=EDGE_NEXT(next_edge);

	if(EDGE_PREV(next_edge)==NULL)
	  GRAPH_EDGE_FIRST(graph)=next_edge;
	else EDGE_NEXT(EDGE_PREV(next_edge)) = next_edge;

	if(EDGE_NEXT(current_edge)==NULL)
	  GRAPH_EDGE_LAST(graph)=current_edge;
	else EDGE_PREV(EDGE_NEXT(current_edge)) = current_edge;

	EDGE_NEXT(next_edge)=current_edge;
	EDGE_PREV(current_edge)=next_edge;
      }
      else{
	/*
	  if(tmp==0 && tmp2==0)
	  FATAL(("darn"));
	*/
       	current_edge = EDGE_NEXT(current_edge);      
      }
      next_edge = EDGE_NEXT(current_edge);
    }
  }
  /*}}}*/

  /*Order list of successors {{{*/
  GRAPH_FOREACH_NODE(graph, current_node){
    nr_cessors = 0;
    NODE_FOREACH_SUCC_EDGE(current_node, current_edge){
      nr_cessors++;
    }

    for(i=0;i<nr_cessors;i++){
      
      current_edge = NODE_SUCC_FIRST(current_node);
      if(current_edge==NULL){
	return;
      }
      next_edge = EDGE_SUCC_NEXT(current_edge);

      while(next_edge != NULL){
	if(fun(EDGE_TAIL(current_edge), EDGE_TAIL(next_edge))==1)
	{
	  EDGE_SUCC_PREV(next_edge)=EDGE_SUCC_PREV(current_edge);
	  EDGE_SUCC_NEXT(current_edge)=EDGE_SUCC_NEXT(next_edge);

	  if(EDGE_SUCC_PREV(next_edge)==NULL)
	    NODE_SUCC_FIRST(current_node)=next_edge;
	  else EDGE_SUCC_NEXT(EDGE_SUCC_PREV(next_edge)) = next_edge;

	  if(EDGE_SUCC_NEXT(current_edge)==NULL)
	    NODE_SUCC_LAST(current_node)=current_edge;
	  else EDGE_SUCC_PREV(EDGE_SUCC_NEXT(current_edge)) = current_edge;

	  EDGE_SUCC_NEXT(next_edge)=current_edge;
	  EDGE_SUCC_PREV(current_edge)=next_edge;
	}
	else{
	  current_edge = EDGE_SUCC_NEXT(current_edge);      
	}
	next_edge = EDGE_SUCC_NEXT(current_edge);
      }
    }
  }
  /*}}}*/
  
  /*Order list of predecessors {{{*/
  GRAPH_FOREACH_NODE(graph, current_node){
    nr_cessors = 0;
    NODE_FOREACH_PRED_EDGE(current_node, current_edge){
      nr_cessors++;
    }

    for(i=0;i<nr_cessors;i++){
      
      current_edge = NODE_PRED_FIRST(current_node);
      if(current_edge==NULL){
	return;
      }
      next_edge = EDGE_PRED_NEXT(current_edge);

      while(next_edge != NULL){
	if(fun(EDGE_TAIL(current_edge), EDGE_TAIL(next_edge))==1)
	{
	  EDGE_PRED_PREV(next_edge)=EDGE_PRED_PREV(current_edge);
	  EDGE_PRED_NEXT(current_edge)=EDGE_PRED_NEXT(next_edge);

	  if(EDGE_PRED_PREV(next_edge)==NULL)
	    NODE_PRED_FIRST(current_node)=next_edge;
	  else EDGE_PRED_NEXT(EDGE_PRED_PREV(next_edge)) = next_edge;

	  if(EDGE_PRED_NEXT(current_edge)==NULL)
	    NODE_PRED_LAST(current_node)=current_edge;
	  else EDGE_PRED_PREV(EDGE_PRED_NEXT(current_edge)) = current_edge;

	  EDGE_PRED_NEXT(next_edge)=current_edge;
	  EDGE_PRED_PREV(current_edge)=next_edge;
	}
	else{
	  current_edge = EDGE_PRED_NEXT(current_edge);      
	}
	next_edge = EDGE_PRED_NEXT(current_edge);
      }
    }
  }
  /*}}}*/
}
/*}}}*/
