/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*!\addtogroup GENERIC_DATATYPES  Generic Datatypes*/
/*@{ */
/*! \defgroup GRAPH Graphs 
 * 
 * We have one generic data structure for storing directed graphs. A
 * graph consists of nodes and edges. Specific graphs in a program
 * will be stored in graphs that are derived from this graph. 
 *
 * For control flow graphs, e.g., we use the t_cfg data type, which is
 * derived from graph. It consists of basic blocks (type t_bbl) and
 * control flow edges (type t_cfg_edge), which are derived from the
 * t_node and t_edge datatypes. This inheritance works as follows: the
 * first member in a derived struct is the base struct. This allows
 * casting between the two types. Depending on the information needed
 * about a derived type, it should be casted to its base type or
 * not. Basic algorithms will work with the base types. For graphs,
 * e.g., this might be a dominator analysis. Algorithms on control
 * flow graphs, e.g., will use the derived types, as they contain more
 * information. 
 *
 * Most (if not all) data members can be accessed using a
 * set of macro's. We strongly recomment the use of those macro's, as
 * they contain the necessary castings.

*/

/*@{ */
#ifndef _GRAPH_TYPEDEFS
#define _GRAPH_TYPEDEFS
typedef struct _t_edge t_edge;
typedef struct _t_node t_node;
typedef struct _t_graph t_graph;
#endif

#ifndef _GRAPH_H
#define _GRAPH_H
#include "../simple_types.h"

extern t_uint32 global_node_marking_number;
extern t_uint32 global_node_marking_number2;
extern t_uint32 global_edge_marking_number;

/*! \defgroup GRAPH_EDGES Graph Edges*/
/*@{*/

/*! This structure holds all the data about (generic) edges in generic graphs */
struct _t_edge
{
  /*! Each edge has a head */
  struct _t_node * head;
  /*! And a tail */
  struct _t_node * tail;

  /*! Edges are stored in double linked lists: one list per
    graph. Hence a next member. */
  struct _t_edge * next; /* global list */
  /*! And a prev member. */
  struct _t_edge * prev; /* global list */
  /*! Each edges is part of two other double linked lists: one of the
    successors of its head and one of the predecessors of its
    tail. This points to the next successor of this edge's head.*/
  struct _t_edge * succ_next; /* succ list */
  /*! Each edges is part of two other double linked lists: one of the
    successors of its head and one of the predecessors of its
    tail. This points to the previous successor of this edge's head.*/
  struct _t_edge * succ_prev; /* succ list */
  /*! Each edges is part of two other double linked lists: one of the
    successors of its head and one of the predecessors of its
    tail. This points to the next predecessor of this edge's tail.*/
  struct _t_edge * pred_next; /* pred list */
  /*! Each edges is part of two other double linked lists: one of the
    successors of its head and one of the predecessors of its
    tail. This points to the previous predecessor of this edge's tail.*/
  struct _t_edge * pred_prev; /* pred list */
  /*! edges have types, they are stored in this member */
  t_uint32 category;
  /*! We have quite a strange scheme to mark and unmark blocks and
  edges: they get a number: having the "current marking number" means
  being marked, otherwise they are unmarked. All edges are easily
  unmarked by just incrementing the "current marking number". */
  t_uint32 marked_number; 
}; 

#define T_EDGE(x)  ((t_edge *) (x))

#define EdgeSetCategory(edge,_category)   (T_EDGE(edge)->category |= _category)
#define EdgeClearCategory(edge,_category) (T_EDGE(edge)->category &= ~_category)
#define EdgeTestCategoryOr(edge,_category) (T_EDGE(edge)->category & (_category))
#define EdgeTestCategoryAnd(edge,_category) (T_EDGE(edge)->category & (_category) == (_category))

#define EDGE_HEAD(edge)          (T_EDGE(edge)->head)
#define EDGE_TAIL(edge)          (T_EDGE(edge)->tail)
#define EDGE_NEXT(edge)          (T_EDGE(edge)->next)
#define EDGE_PREV(edge)          (T_EDGE(edge)->prev)
#define EDGE_SUCC_NEXT(edge)     (T_EDGE(edge)->succ_next)
#define EDGE_SUCC_PREV(edge)     (T_EDGE(edge)->succ_prev)
#define EDGE_PRED_NEXT(edge)     (T_EDGE(edge)->pred_next)
#define EDGE_PRED_PREV(edge)     (T_EDGE(edge)->pred_prev)
#define EDGE_CAT(edge)          (T_EDGE(edge)->category)

#define EdgeIsMarked(edge) (T_EDGE(edge)->marked_number==global_edge_marking_number)
#define EdgeMark(edge)     (T_EDGE(edge)->marked_number=global_edge_marking_number)
#define EdgeUnmark(edge)   (T_EDGE(edge)->marked_number=0)

void InitEdgeMarking(void);
t_edge * GraphNewEdge(t_graph * graph, t_node * from, t_node * to, t_uint32 edge_category);
void GraphRemoveEdge(t_graph * graph, t_edge * edge);
void GraphRemoveAllEdgesFromTo(t_graph * graph, t_node * from, t_node * to);
/*@}*/

/*! \defgroup GRAPH_NODES Graph Nodes*/
/*@{*/

/*! This structure holds all the data about (generic) nodes in generic graphs */
struct _t_node
{
  /*! For each graph, we have a double linked list of blocks. Hence the next member */
  struct _t_node * next;
  /*! For each graph, we have a double linked list of blocks. Hence the prev member */
  struct _t_node * prev;

  /*! For each node, we have a double linked list of successors. We
    have a pointer to the first successor (edge!!!) */
  t_edge * succ_first;
  /*! For each node, we have a double linked list of successors. We
    have a pointer to the last successor (edge!!!) */
  t_edge * succ_last;

  /*! For each node, we have a double linked list of predecessors. We
    have a pointer to the first predecessor (edge!!!) */
  t_edge * pred_first; 
  /*! For each node, we have a double linked list of predecessors. We
    have a pointer to the last predecessor (edge!!!) */
  t_edge * pred_last; 
  /*! nodes have types, they are stored in this member */
  /*  t_uint32 category; */
  /*! We have quite a strange scheme to mark and unmark blocks and
    edges: they get a number: having the "current marking number" means
    being marked, otherwise they are unmarked. All edges are easily
    unmarked by just incrementing the "current marking number". */
  t_uint32 marked_number;
  t_uint32 marked_number2; 
};

/* node categories */

#define NODE_IS_DATA 0x1
#define NODE_IS_CODE 0x2

#define T_NODE(x)  ((t_node *) (x))

#define NodeSetCategory(node,_category)   (T_NODE(node))->category |= _category
#define NodeClearCategory(node,_category) (T_NODE(node))->category &= ~_category
#define NodeTestCategoryOr(node,_category) ((T_NODE(node))->category & (_category))
#define NodeTestCategoryAnd(node,_category) ((T_NODE(node))->category & (_category) == (_category))

#define NODE_NEXT(node)          ((T_NODE(node))->next) 
#define NODE_PREV(node)          ((T_NODE(node))->prev) 
#define NODE_SUCC_FIRST(node)    ((T_NODE(node))->succ_first)
#define NODE_SUCC_LAST(node)     ((T_NODE(node))->succ_last) 
#define NODE_PRED_FIRST(node)    ((T_NODE(node))->pred_first) 
#define NODE_PRED_LAST(node)     ((T_NODE(node))->pred_last) 

#define NodeIsMarked(node) (T_NODE(node)->marked_number==global_node_marking_number)
#define NodeMark(node)     (T_NODE(node)->marked_number=global_node_marking_number)
#define NodeUnmark(node)   (T_NODE(node)->marked_number=0)

#define NodeIsMarked2(node) (T_NODE(node)->marked_number2==global_node_marking_number2)
#define NodeMark2(node)     (T_NODE(node)->marked_number2=global_node_marking_number2)
#define NodeUnmark2(node)   (T_NODE(node)->marked_number2=0)


void InitNodeMarking(void);
void InitNodeMarking2(void);
t_node * GraphNewNode(t_graph * graph, t_uint32 node_category);
void GraphInitNode(t_graph * graph, t_node * temp);
void GraphInit(t_graph * temp, t_uint32 node_size, t_uint32 edge_size);
t_bool NodeIsPred(t_node * succ, t_node * pred, t_uint32 edge_category);
t_bool NodeIsSucc(t_node * pred, t_node * succ, t_uint32 edge_category);
t_edge * GraphPrependNode(t_graph * graph, t_node * before, t_node * insert, t_uint32 edge_category);
t_edge * GraphAppendNode(t_graph * graph, t_node * after, t_node * insert, t_uint32 edge_category);
void GraphRemoveNode(t_graph * graph, t_node * node);
void GraphRemoveNodeNoFree(t_graph * graph, t_node * node);
/*@}*/

/*! \defgroup GRAPH_GRAPH Graphs themselves*/
/*@{*/

/*! This structure holds all the data about (generic) graphs */
struct _t_graph
{
  /*! For each graph, we store the number of nodes in it */
  t_uint32 nnodes;
  /*! For each graph, we store the number of edges in it */
  t_uint32 nedges;
  /*! For each graph, we need to know the size of the edge datatypes
    (base or derived) to provide space on allocation */
  t_uint32 edge_size;
  /*! For each graph, we need to know the size of the node datatypes
    (base or derived) to provide space on allocation */
  t_uint32 node_size;
  /*! For each graph, we hold a double linked list of nodes. Hence
    this member to the first node in that list */
  t_node * node_first;
  /*! For each graph, we hold a double linked list of nodes. Hence
    this member to the last node in that list */
  t_node * node_last;
  /*! For each graph, we hold a double linked list of edges. Hence
    this member to the first edge in that list */
  t_edge * edge_first;
  /*! For each graph, we hold a double linked list of edges. Hence
    this member to the last edge in that list */
  t_edge * edge_last;
  
};

#define T_GRAPH(graph)   ((t_graph*) (graph))

#define GRAPH_NODE_SIZE(graph)     (T_GRAPH(graph)->node_size)
#define GRAPH_EDGE_SIZE(graph)     (T_GRAPH(graph)->edge_size)
#define GRAPH_NODE_FIRST(graph)    (T_GRAPH(graph)->node_first)
#define GRAPH_NODE_LAST(graph)     (T_GRAPH(graph)->node_last)
#define GRAPH_EDGE_FIRST(graph)    (T_GRAPH(graph)->edge_first)
#define GRAPH_EDGE_LAST(graph)     (T_GRAPH(graph)->edge_last)
#define GRAPH_NNODES(graph)        (T_GRAPH(graph)->nnodes)
#define GRAPH_NEDGES(graph)        (T_GRAPH(graph)->nedges)

#define GRAPH_EDGE_SIZE(graph)     (T_GRAPH(graph)->edge_size)
#define GRAPH_EDGE_SIZE(graph)     (T_GRAPH(graph)->edge_size)

t_graph * GraphNew(t_uint32 node_size, t_uint32 edge_size);

void GraphDFTraversalWithCheckAndData(t_node * start, t_bool (*CallBack)(t_node * curr, t_node * par,void * data), t_uint32 edge_category_mask,void * data);
void GraphWhateverTraversalWithCheckAndData(t_graph * graph, t_node * start, t_bool (*CallBack)(t_node * curr, t_node * par,void * data), t_uint32 edge_category_mask,void * data);
void GraphBFTraversal(t_node * start, void (*CallBack)(t_node * curr, t_node * par), t_uint32 edge_category_or_mask);
void GraphDFReverseTraversalWithCheckAndData(t_node * start,t_bool (*CallBack)(t_node * curr, t_node * par,void * data), t_uint32 edge_category_mask,void * data); 

#define GRAPH_FOREACH_NODE(graph,node) for (node=GRAPH_NODE_FIRST(graph); node!=NULL; node=NODE_NEXT(node))
#define GRAPH_FOREACH_NODE_SAFE(graph,node,tmp) for (node=GRAPH_NODE_FIRST(graph), tmp=node?NODE_NEXT(node):NULL; node!=NULL; node=tmp, tmp=node?NODE_NEXT(node):NULL)

#define GRAPH_FOREACH_EDGE(graph,edge) for (edge=GRAPH_EDGE_FIRST(graph); edge!=NULL; edge=EDGE_NEXT(edge))
#define GRAPH_FOREACH_EDGE_SAFE(graph,edge,tmp) for (edge=GRAPH_EDGE_FIRST(graph), tmp=edge?EDGE_NEXT(edge):NULL; edge!=NULL; edge=tmp, tmp=edge?EDGE_NEXT(edge):NULL)

#define NODE_FOREACH_SUCC_EDGE(node,edge) for (edge=NODE_SUCC_FIRST(node); edge!=NULL; edge=EDGE_SUCC_NEXT(edge))

#define NODE_FOREACH_SUCC_EDGE_SAFE(node,edge,tmp) for (edge=NODE_SUCC_FIRST(node), tmp=edge?EDGE_SUCC_NEXT(edge):NULL; edge!=NULL; edge= tmp, tmp=edge?EDGE_SUCC_NEXT(edge):NULL)

#define NODE_FOREACH_PRED_EDGE(node,edge) for (edge=NODE_PRED_FIRST(node); edge!=NULL; edge=EDGE_PRED_NEXT(edge))

#define NODE_FOREACH_PRED_EDGE_SAFE(node,edge,tmp) for (edge=NODE_PRED_FIRST(node), tmp=edge?EDGE_PRED_NEXT(edge):NULL; edge!=NULL; edge= tmp, tmp=edge?EDGE_PRED_NEXT(edge):NULL)

typedef t_int8 (*NodeCmpFun) (t_node * node1, t_node * node2);
void GraphOrder(t_graph * graph, NodeCmpFun fun);

#endif
/*@}*/
/*@}*/
