/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//* $Header: /home/cvs/cvsroot/diablo/datastructures/data_hash.c,v 1.1.1.1 2003/10/10 12:12:21 bdebus Exp $ 
 * 
 * This files contains the functions for hash tables 
 * 
 * 
 */

#include "../config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef TEST_HASH
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#else
#define Malloc malloc
#define Realloc realloc
#define Free free 
#endif
#include "data_hash.h"
#define DATA_HASH_NODE(x) ((data_hash_table *)(((char*)(x))+table->offset))

 
/* Two different string hash functions */ 

#if 0
/* This is much slower than the following */
t_uint32 string_hash (data_hash_table * table, void * vkey)
{
   char * key=(char *) vkey;
   t_uint32 n = 0;
   while (*key != 0) n = ((n + n + *(key++)) % table->tsize);
   return n;
}
#else
t_uint32 string_hash (data_hash_table * table, void * vkey)
{
   char * key=(char *) vkey;
   t_uint32 n = 0;
   while (*key != 0) n = ((n*131 + *(key++)));
   
   n %= table->tsize;
   return n;
}
#endif

/* Create an empty hash table */

data_hash_table * HashTableNew(t_uint32 size, t_uint32 offset, t_uint32 (*hash_func)(data_hash_table *, void *) , int (*cmp)(void *, void *), void (*node_free)(void *,void *)) 
{
   t_uint32 tel;
   data_hash_table * ret=(data_hash_table *) Malloc(sizeof(data_hash_table));
   ret->nodes = (data_hash_node **) Malloc (size*sizeof(data_hash_node *));
   ret->tsize = size;
   ret->offset = offset;
   ret->hash_func = hash_func;
   ret->hash_cmp = cmp; 
   ret->node_free = node_free; 

   for (tel = 0; tel <ret->tsize; tel++)
      ret->nodes[tel] = NULL;

   return ret;
}

/* Free the hash table and all elements in it */

void HashFree(data_hash_table * table)
{
   HashWalk(table, (void (*)(void *,void *) )table->node_free,NULL); 
   Free(table->nodes);
   Free(table);
}

/* Insert an allocated node in the hash table. The key of the node has to be initialized ! */

void HashInsert (data_hash_table * table, void * vnode) 
{
   data_hash_node * node=(data_hash_node *) (((char *) vnode)+table->offset);
   void * vit;
   data_hash_node * it;
   t_uint32 n = table->hash_func (table, node->key);
   node->equal=NULL;
   node->next=NULL;
   for (vit=table->nodes[n], it = vit?(data_hash_node *) (((char *) vit)+table->offset):NULL ; it != NULL; vit = it->next, it=vit?(data_hash_node *) (((char *) vit)+table->offset):NULL)
   {
      if (table->hash_cmp(it->key,node->key)==0)
      {
	 node->equal=it->equal;
	 it->equal=vnode;
	 return;
      }
   
   }
   if (table->nodes[n]) 
      node->next=table->nodes[n];
   else
      node->next=NULL;

   table->nodes[n]=vnode;
   
}


void HashPrepend (data_hash_table * table, void * vnode) 
{
   data_hash_node * node=(data_hash_node *) (((char *) vnode)+table->offset);
   void * vit, * vprev;
   data_hash_node * it, * prev=NULL;
   t_uint32 n = table->hash_func (table, node->key);
   node->equal=NULL;
   node->next=NULL;
   for (vit=table->nodes[n], it = vit?(data_hash_node *) (((char *) vit)+table->offset):0 ; it != NULL; prev=it, vprev=vit, vit = it->next, it=vit?(data_hash_node *) (((char *) vit)+table->offset):NULL)
   {
      if (table->hash_cmp(it->key,node->key)==0)
      {
	 node->equal=vit;
	 node->next=it->next;
	 if (prev)
	    prev->next=vnode;
	 else
	    table->nodes[n]=vnode;
	 return;
      }
   }

   if (table->nodes[n]) 
      node->next=table->nodes[n];
   else
      node->next=NULL;
   table->nodes[n]=vnode;
}







/* Delete a node in the hashtable (second argument=node!!!)
 * Not the most elegant code, but it works (pruts) */

void HashDelete(data_hash_table * table, void * vnode)
{
   data_hash_node * node=(data_hash_node *) (((char *) vnode)+table->offset);
   void * vit, * vprev;
   data_hash_node * it, * prev=NULL;
   t_uint32 n = table->hash_func (table, node->key);

   if(table->nodes[n]==NULL) FATAL(("Trying to delete a node that is no longer in the table!!!!"));
   
   for (vit=table->nodes[n], it = vit?(data_hash_node *) (((char *) vit)+table->offset):0 ; it != NULL; prev=it, vprev=vit, vit = it->next, it=vit?(data_hash_node *) (((char *) vit)+table->offset):NULL)
   {
      if (table->hash_cmp(it->key,node->key)==0) 			/*iterate to first node with same key*/
      {
	 if(vit==vnode)							/*if the first node with the same key is the node to be deleted*/
	 {
	  
	  if(it->equal)							/*if the equal field is not NULL, then the equal node comes in place*/
	  {
           ((data_hash_node *) (((char*) it->equal)+table->offset))->next=it->next;
           if(prev) prev->next=node->equal;
	   else table->nodes[n]=node->equal;
	  }

	  
	  else								/*otherwise the next node comes in place*/
	  {
	   if(prev) prev->next=node->next;
	   else table->nodes[n]=node->next;
	  }
	  table->node_free(vit,NULL);					/*free the deleted node*/
	  return;
	 }
      
   
	 for(vprev=vit,prev=it,vit=it->equal,it=vit?(data_hash_node *) (((char *) vit)+table->offset):NULL;it!=NULL;vprev=vit,prev=it,vit=it->equal,it=vit?(data_hash_node *) (((char *) vit)+table->offset):NULL)			/*search the correct node via the equal fields*/
	 {
	  if(vit==vnode)							/*the node is found*/
	  {
           prev->equal=it->equal;
	   table->node_free(vit,NULL);
	   return;
	  }
	 }
      }
   }    
   FATAL(("Trying to delete a hash node which is not in the table!!, no matching key found"));
}

/* Find a node in the hashtable (the node that was inserted first in case of equal keys) */

void * HashLookup (data_hash_table * table,  void * key)
{
   data_hash_node * it;
   void * vit;
   t_uint32 n = table->hash_func (table, key);

   for (vit=table->nodes[n], it = vit?(data_hash_node *) (((char *) vit)+table->offset):NULL ; it != NULL; vit = it->next, it=(data_hash_node *) (((char *) vit)+table->offset))
   {
      if (table->hash_cmp(it->key,key)==0)
      {
	 return vit;
      }
   }

   return NULL;
}

/* Visit all elements in the hash_table and perform function on them */ 

void HashWalk(data_hash_table * table, void (*func)(void *,void *),void * data)
{
   t_uint32 n;
   data_hash_node * it, * itn, * it2, * it2n;

   void * vit,  * vit2, *vit2n;
   
   for (n=0; n<table->tsize; n++)
   {
      for (vit=table->nodes[n], it = vit?(data_hash_node *) (((char *) table->nodes[n])+table->offset):NULL ; it != NULL; vit = itn, it=vit?(data_hash_node *) (((char *) vit)+table->offset):NULL)
      {
	 itn=it->next;
	 vit2n=it->equal;
	 it2n=vit2n?(data_hash_node *) (((char *) vit2n)+table->offset):NULL;
	 func(vit,data);
	 for (vit2=vit2n, it2 = it2n; it2!=NULL; vit2=vit2n, it2= it2n )
	 {
	    vit2n=it2->equal;
	    it2n=vit2n?(data_hash_node *) (((char *) vit2n)+table->offset):0;
	    func(vit2,data);
	 }
      }
   }
}

void HashStats(data_hash_table * table)
{
   t_uint32 n;
   data_hash_node * it;
   t_uint32 this;
   t_uint32 max=0;
   t_uint32 maxfill=0;
   t_uint32 minfill=((t_uint32) 1)<<31;
   t_uint32 totalfill=0;
   
   for (n=0; n<table->tsize; n++)
   {
      this=0;
      for (it = table->nodes[n]; it != NULL; it = it->next)
      {
	 this++;
      }
      if (this>maxfill) { maxfill=this; max=n; }
      if (this<minfill) minfill=this;
      totalfill+=this;
   }

   printf("Average fill= %f\n",((float) totalfill)/n);
   printf("Max fill= %d\n",maxfill);
   printf("Min fill= %d\n",minfill);
   printf("Nelems= %d\n",totalfill);

   /* This will fail if the hash_key is not a string! */
   printf("MAX\n");
   for (it = table->nodes[max]; it != NULL; it = it->next)
   {
      printf("%s\n",(char *) it->key);
   }
   printf("END MAX\n");

}

int string_cmp(void * a , void * b)
{
	return strcmp((char *) a, (char *) b);
}


#ifdef TEST_HASH

char * keys[]=
{
   NULL,
   "Search",
   "seek",
   "and",
   "know",
   "how",
   "this",
   "foul",
   "murder",
   "comes",
   "-----",
   "Who",
   "seeks",
   "and", 
   "will", 
   "not", 
   "take", 
   "when", 
   "once", 
   "'tis", 
   "offer'd",
   "Shall", 
   "never", 
   "find", 
   "it", 
   "more",
   "-----",
   "Search",
   "seek",
   "and",
   "know",
   "how",
   "this",
   "foul",
   "murder",
   "comes",
   NULL
};

typedef struct _test1
{
   data_hash_node node;
   char * string2;
} hntest1;

typedef struct _test2
{
   char * string2;
   data_hash_node node;
} hntest2;

void SP(void * in,void * ea)
{
   hntest1 * p=in;
   printf("p->string=%s => %p\n",p->string2,in);
}

void SP2(void * in,void * ea)
{
   hntest2 * p=in;
   printf("p->string=%s => %p\n",p->string2,in);
}
#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

int main()
{
   char ** it=keys;
   hntest1 * insert;
   hntest2 * insert2;
   data_hash_table * test1;

   printf("Creating simple hash table\n"); 
   /* We use small hash table size to allow thourough testing */
   test1=HashTableNew(3,0,string_hash,string_cmp,NULL);
   printf("Done\n");

   it++;
   printf("Inserting elements\n");
   do
   {
      insert=Malloc(sizeof(hntest1));
      printf("   Insert: %s => %p\n",*it,insert);
      insert->node.key=*it;
      insert->string2=strdup(*it);
      HashInsert(test1,insert);
   }
   while(*(++it));

   it--;
   do
   {
      printf("   Lookup: %s\n",*it);
      insert=HashLookup(test1,*it);
      printf("   Verify: %s\n",insert->string2);
      if (strcmp(insert->string2,*it)) { printf("Failed!\n"); exit(-1); }
   }
   while(*(--it));

   HashWalk(test1,SP,NULL);


   printf("Prepending elements\n");
   printf("Creating simple hash table\n"); 
   /* We use small hash table size to allow thourough testing */
   test1=HashTableNew(3,0,string_hash,string_cmp,NULL);
   printf("Done\n");

   it++;
   printf("Inserting elements\n");
   do
   {
      insert=Malloc(sizeof(hntest1));
      printf("   Prepend: %s => %p\n",*it,insert);
      insert->node.key=*it;
      insert->string2=strdup(*it);
      HashPrepend(test1,insert);
   }
   while(*(++it));

   it--;
   do
   {
      printf("   Lookup: %s\n",*it);
      insert=HashLookup(test1,*it);
      printf("   Verify: %s\n",insert->string2);
      if (strcmp(insert->string2,*it)) { printf("Failed!\n"); exit(-1); }
   }
   while(*(--it));

   HashWalk(test1,SP,NULL);

   
   
   
   printf("Creating displaced (%d) hash table\n",offsetof(hntest2,node)); 
   /* We use small hash table size to allow thourough testing */
   test1=HashTableNew(3,offsetof(hntest2,node),string_hash,string_cmp,NULL);
   printf("Done\n");

   it++;
   printf("Inserting elements\n");
   do
   {
      insert2=Malloc(sizeof(hntest2));
      printf("   Insert: %s => %p\n",*it,insert2);
      insert2->node.key=*it;
      insert2->string2=strdup(*it);
      HashInsert(test1,insert2);
   }
   while(*(++it));

   it--;
   do
   {
      printf("   Lookup: %s\n",*it);
      insert2=HashLookup(test1,*it);
      printf("   Verify: %s\n",insert2->string2);
      if (strcmp(insert2->string2,*it)) { printf("Failed!\n"); exit(-1); }
   }
   while(*(--it));

   HashWalk(test1,SP2,NULL);

   printf("Creating displaced (%d) hash table\n",offsetof(hntest2,node)); 
   /* We use small hash table size to allow thourough testing */
   test1=HashTableNew(3,offsetof(hntest2,node),string_hash,string_cmp,NULL);
   printf("Done\n");

   it++;
   printf("Prepend elements\n");
   do
   {
      insert2=Malloc(sizeof(hntest2));
      printf("   Prepend: %s => %p\n",*it,insert2);
      insert2->node.key=*it;
      insert2->string2=strdup(*it);
      HashPrepend(test1,insert2);
   }
   while(*(++it));

   it--;
   do
   {
      printf("   Lookup: %s\n",*it);
      insert2=HashLookup(test1,*it);
      printf("   Verify: %s\n",insert2->string2);
      if (strcmp(insert2->string2,*it)) { printf("Failed!\n"); exit(-1); }
   }
   while(*(--it));

   HashWalk(test1,SP2,NULL);


   return 0;

}
#endif
