/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifdef ARM_SUPPORT
#include "../../config.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_ins.h"
#include "../../arch/arm/arm_opcodes.h"
#include "../../arch/arm/arm_disassemble.h"
#include "../../arch/arm/arm_assemble.h"
#include "../../arch/arm/arm_instructions.h"
#include "elf_common32.h"
#include "elf_arm.h"
#include "elf_endian.h"
#include "elf_main.h"
#include "elf_armsupport.h"
#include "../../utils/util_machine.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_object.h"

#define sign_extend(x,p)        (((x) ^ (1 << (p))) - (1 << (p)))
#define SELECT_BITS(i,b,e) ((i&((1<<(b+1))-1))>>e)

/* Common operations on objectfiles */

/* Write an elf object file. The elfhdr is supplied by the arch-specific code */

t_bool IsElfArmSameEndianOnMsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   Elf32_Ehdr hdr;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=MSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2LSB:
	 return FALSE;
      case ELFDATA2MSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_ARM)
      return FALSE; 

   return TRUE;
}

t_bool IsElfArmSwitchedEndianOnMsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   Elf32_Ehdr hdr;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=MSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2MSB:
	 return FALSE;
      case ELFDATA2LSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_SWITCHED_ARM)
      return FALSE; 

   return TRUE;
}

t_bool IsElfArmSwitchedEndianOnLsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   Elf32_Ehdr hdr;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=LSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2LSB:
	 return FALSE;
      case ELFDATA2MSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_SWITCHED_ARM)
      return FALSE; 

   return TRUE;
}



t_bool IsElfArmSameEndianOnLsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   Elf32_Ehdr hdr;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=LSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2MSB:
	 return FALSE;
      case ELFDATA2LSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_ARM)
      return FALSE; 

   return TRUE;
}

void ElfWriteArmSameEndian(t_object * obj, t_string filename)
{
   FILE * fp=fopen(filename,"w");
   int tel=0;
   Elf32_Ehdr hdr;
   ANALYSIS_START("Writing arm binary");
   hdr.e_ident[EI_MAG0]=ELFMAG0;
   hdr.e_ident[EI_MAG1]=ELFMAG1;
   hdr.e_ident[EI_MAG2]=ELFMAG2;
   hdr.e_ident[EI_MAG3]=ELFMAG3;
   hdr.e_ident[EI_VERSION]=EV_CURRENT;
   if (global_machine_info.endian==LSB)
      hdr.e_ident[EI_DATA]=ELFDATA2LSB;
   else
      hdr.e_ident[EI_DATA]=ELFDATA2MSB;

   hdr.e_ident[EI_CLASS]=ELFCLASS32;
   for (tel=EI_PAD; tel<16; tel++)
   {
      hdr.e_ident[tel]=0;
   }
   hdr.e_type=ET_EXEC;
   hdr.e_machine=EM_ARM;
   hdr.e_version=EV_CURRENT;
   hdr.e_entry=G_T_UINT32(obj->entry);
   hdr.e_phoff=0;
   hdr.e_flags=EF_ARM_HASENTRY | EF_ARM_SYMSARESORTED;
   hdr.e_ehsize=0;
   hdr.e_phentsize=0;
   hdr.e_phnum=0;
   hdr.e_shentsize=0; /* Set in common */
   hdr.e_shnum=0; /* Set in common */
   hdr.e_shstrndx=0; /* Set in common */

   ElfWriteCommonSameEndian(fp,&hdr,obj,2<<1); 
   ANALYSIS_SIMPLE_END("Writing arm binary");
}

void ElfReadArmSameEndian(FILE * fp, t_object * obj)
{
   Elf32_Shdr * shdrs=NULL;
   Elf32_Sym * symbol_table=NULL;
   t_section ** sec_table=NULL;
   char * strtab=NULL;
   int numsyms=0;
   int tel;
   char * sechdrstrtab=NULL;
   Elf32_Ehdr hdr32;
   t_symbol ** table;
   if (fread((&hdr32),sizeof(Elf32_Ehdr),1,fp)!=1)
   {
      FATAL(("Could not read elf header!"));
   }

   /* Reading elf headers does not depend on the type of objectfile */
   ElfReadCommon32(fp, &hdr32, obj, &shdrs, &symbol_table, &numsyms, &strtab, &sechdrstrtab, &sec_table, &table,FALSE); 

   obj->reloc_table=RelocTableNew();
   if (hdr32.e_shnum)
   {
      tel=0;
      while (tel<hdr32.e_shnum)
      {
	 /* STEP 4b: architecture specific flags specific  */


	 if ((shdrs[tel].sh_flags & SHF_ARM_COMDEF) && (shdrs[tel].sh_flags & SHF_ALLOC))
	 {
	    t_section * sec=SectionGetFromObjectByName(obj, sechdrstrtab + shdrs[tel].sh_name);
	    if (!sec) FATAL(("Common defined section %s not found!",sechdrstrtab + shdrs[tel].sh_name));
	    else sec->is_common_defined=TRUE;
	 }

	 
	 /* STEP 5: Find the relocations */ 
	 if (PatternMatch(".rel.debug*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    /* if we ignore debug info, we can also ignore debug relocations */
	 }
	 else if (PatternMatch(".rela*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    FATAL(("RELA relocations not implemented!"));
	 }
	 else if (PatternMatch(".rel.stab*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	   /* Lets ignore these */
	 }
	 else if (PatternMatch(".rel*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    Elf32_Rel * rel=Malloc(shdrs[tel].sh_size);
	    t_uint32 tel2;
#ifdef NAME_BASED
	    t_section_hash_node * sh;
#endif
	    t_section * corrsec;
	    fseek(fp,obj->streampos+shdrs[tel].sh_offset,SEEK_SET);
	    fread(rel,shdrs[tel].sh_size,1,fp);
#ifdef NAME_BASED 
	    sh=(t_section_hash_node *) HashLookup(obj->section_table,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);
	    VERIFY_FASSERT( sh!=NULL ,("Base section for relocations not found!"));
	    corrsec=sh->section;
#else
	    corrsec=sec_table[shdrs[tel].sh_info]; 
#endif



	    for (tel2=0 ; tel2<(shdrs[tel].sh_size/sizeof(Elf32_Rel)); tel2++)
	    {
	       VERIFY_FASSERT(symbol_table && strtab,("Symtab or stringtab not found. Implement me"));
	       /*if (ELF32_R_TYPE(rel[tel2].r_info)==27) continue;*/ /* PLT enries */
	       switch(ELF32_R_TYPE(rel[tel2].r_info))
	       {
		  case R_ARM_NONE:
		     break;
		  case R_ARM_PLT32:
		  case R_ARM_PC24:
		     {
			t_uint32 data;
			t_symbol * sym;
			t_address generic;
			t_address addend;

			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			GenericSet_uint32(generic,rel[tel2].r_offset);

			data=SectionGetData32(corrsec,generic);

			data=SELECT_BITS(data,23,0);
			data=sign_extend(data,23);
			data<<=2;
			GenericSet_uint32(addend,data);

			RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(corrsec),generic,sym,FALSE,NULL,NULL,NULL,CALC_REL "\\" WRITE_PC24,FALSE);
		     }
		     break; 
		  case R_ARM_ABS32:
		     {
			t_uint32 data;
			t_address generic;
			t_address addend;
			t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			GenericSet_uint32(generic,rel[tel2].r_offset);
			data=SectionGetData32(corrsec,generic);
			GenericSet_uint32(addend,data);
#ifdef THUMB_SUPPORT
			RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(corrsec),generic,sym,TRUE,NULL,NULL,NULL,CALC_ABS "SM+\\" WRITE_32,FALSE);
#else
			RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(corrsec),generic,sym,TRUE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);
#endif
		     }
		     break;
		  case R_ARM_REL32:
		     {
			t_uint32 data;
			t_address generic;
			t_address addend;
			t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			
			GenericSet_uint32(generic,rel[tel2].r_offset);
			if (corrsec->type!=CODE_SECTION)
			{
			   FATAL(("REL 32 not in code section!"));
			}
			data=SectionGetData32(corrsec,generic);
			GenericSet_uint32(addend,data);

			RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(corrsec),generic,sym,TRUE,NULL,NULL,NULL,CALC_REL "SM+\\" WRITE_32,FALSE);
		     }
		     break;
		  case R_ARM_GOTPC:
		     {
			t_uint32 data;
			t_address generic;
			t_address addend;
			t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];

			GenericSet_uint32(generic,rel[tel2].r_offset);
			data=SectionGetData32(corrsec,generic);
			GenericSet_uint32(addend,data);

			RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(corrsec),generic,sym,TRUE,NULL,NULL,NULL,"SA+P-" "\\" WRITE_32,FALSE);
		     }
		     break;
		  case R_ARM_GOT32:
		     {
			t_uint32 data;
			t_address generic;
			t_address addend;
			t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];

			GenericSet_uint32(generic,rel[tel2].r_offset);
			data=SectionGetData32(corrsec,generic);
			GenericSet_uint32(addend,data);

			RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(corrsec),generic,sym,TRUE,NULL,NULL,NULL,"GA+" "\\" WRITE_32,FALSE);

		     }
		     break;
		  case R_ARM_THM_PC22:
		     {
			t_uint32 data, tmp;
			t_symbol * sym;
			t_address generic;
			t_address addend;

			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			GenericSet_uint32(generic,rel[tel2].r_offset);

			data=SectionGetData32(corrsec,generic);

			tmp=SELECT_BITS(data,10,0);
			tmp<<=11;
			tmp=sign_extend(tmp,21);
			data=SELECT_BITS(data,26,16);
			data |=tmp;
			data <<=1;
			GenericSet_uint32(addend,data);

			RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(corrsec),generic,sym,FALSE,NULL,NULL,NULL,"SSM?P:Pifffffffc&!-A+\\s0001 > = i003ff800 & s000b > % i000007ff & s0010 < | l if800f800 & | SM?:iefffffff&! w\\s0000$" ,FALSE);
		     }
		     break;
		  case R_ARM_GOTOFF:
		     {
			t_uint32 data;
			t_symbol * sym;
			t_address generic;
			t_address addend;

			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			GenericSet_uint32(generic,rel[tel2].r_offset);

			data=SectionGetData32(corrsec,generic);
			GenericSet_uint32(addend,data);

			RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(corrsec),generic,sym,TRUE,NULL,NULL,NULL,"SA+g-\\" WRITE_32,FALSE);
 		

		     }
		     break;
		  case R_ARM_PC13:
		     FATAL(("Implement ARMPC13!"));
		  case R_ARM_ABS16:
		     FATAL(("Implement ARMABS16!"));
		  case R_ARM_ABS12:
		     FATAL(("Implement ARMABS12!"));
		  case R_ARM_THM_ABS5:
		     FATAL(("Implement ARMTHM_ABS5!"));
		  case R_ARM_ABS8:
		     FATAL(("Implement ARMABS8!"));
		  case R_ARM_SBREL32:
		     FATAL(("Implement ARMSBREL32!"));
		  case R_ARM_THM_PC8:
		     FATAL(("Implement ARMTHBPC8!"));
		  case R_ARM_AMP_VCALL9:
		     FATAL(("Implement ARMAMPVCALL9!"));
		  case R_ARM_SWI24:
		     FATAL(("Implement ARMSWI24!"));
		  case R_ARM_THM_SWI8:
		     FATAL(("Implement ARMTHBSWI8!"));
		  case R_ARM_XPC25:
		     FATAL(("Implement ARMXPC25!"));
		  case R_ARM_THM_XPC22:
		     FATAL(("Implement ARMXPC22!"));
		  default:
		     FATAL(("Implement %d!",ELF32_R_TYPE(rel[tel2].r_info)));
	       }
	    }
	    Free(rel);
	 }
	 tel++;
      }
   }


   if (table) Free(table);
   if (shdrs) Free(shdrs);
   if (sechdrstrtab) Free(sechdrstrtab);
   if (symbol_table) Free(symbol_table);
   if (strtab) Free(strtab);
}

void ElfReadArmSwitchedEndian(FILE * fp, t_object * obj)
{
   Elf32_Shdr * shdrs=NULL;
   Elf32_Sym * symbol_table=NULL;
   char * strtab=NULL;
   int numsyms=0;
   int tel;
   char * sechdrstrtab=NULL;
   t_section ** sec_table=NULL;
   Elf32_Ehdr hdr32;
   t_symbol ** table;
   if (fread((&hdr32),sizeof(Elf32_Ehdr),1,fp)!=1)
   {
      FATAL(("Could not read elf header!"));
   }


   Elf32HdrSwitchEndian(&hdr32);
   /* Reading elf headers does not depend on the type of objectfile */
   ElfReadCommon32(fp, &hdr32, obj, &shdrs, &symbol_table, &numsyms, &strtab, &sechdrstrtab, &sec_table, &table, TRUE); 
   obj->reloc_table=RelocTableNew();
   if (hdr32.e_shnum)
   {
      tel=0;
      /* STEP 5: Find the relocations */ 
      while (tel<hdr32.e_shnum)
      {
	 if (PatternMatch(".rel.debug*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    /* if we ignore debug info, we can also ignore debug relocations */
	 }
	 else if (PatternMatch(".rela*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    FATAL(("RELA relocations not implemented!"));
	 }
	 else if (PatternMatch(".rel.stab*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	   /* Lets ignore these */
	 }
	 else if (PatternMatch(".rel*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    Elf32_Rel * rel=Malloc(shdrs[tel].sh_size);
	    t_uint32 tel2;
#ifdef NAME_BASED
	    t_section_hash_node * sh;
#endif
	    t_section * corrsec;
	    fseek(fp,obj->streampos+shdrs[tel].sh_offset,SEEK_SET);
	    fread(rel,shdrs[tel].sh_size,1,fp);

	    for (tel2=0; tel2<(shdrs[tel].sh_size/sizeof(Elf32_Rel)); tel2++)
	    {
	       Elf32RelSwitchEndian(&rel[tel2]);
	    }
#ifdef NAME_BASED 
	    sh=(t_section_hash_node *) HashLookup(obj->section_table,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);
	    VERIFY_FASSERT( sh!=NULL ,("Base section for relocations not found!"));
	    corrsec=sh->section;
#else
	    corrsec=sec_table[shdrs[tel].sh_info]; 
#endif
	    for (tel2=0 ; tel2<(shdrs[tel].sh_size/sizeof(Elf32_Rel)); tel2++)
	    {
	       VERIFY_FASSERT(symbol_table && strtab,("Symtab or stringtab not found. Implement me"));
	       VERIFY_FASSERT(ELF32_R_TYPE(rel[tel2].r_info)<=16, ("Implement relocation %d",ELF32_R_TYPE(rel[tel2].r_info)));
	       switch(ELF32_R_TYPE(rel[tel2].r_info))
	       {
		  case R_ARM_NONE:
		     FATAL(("None relocation!"));
		  case R_ARM_PC24:
		     {
			void * ret;
			t_symbol * sym;
			t_address generic;
			t_uint32 data=0;
			ret=ArmDisassembleOneInstructionForSection(corrsec, rel[tel2].r_offset, NULL);
			VERIFY_FASSERT(ret!=NULL,("No instruction for PC24 relocation"));
			VERIFY_FASSERT(INS_TYPE(ret) == IT_BRANCH, ("Illegal instruction type for PC24 relocation")); 
			/* If this is the case we can safely skip this relocation, 
			 * the targets can be easily deduced during flowgraph construction. */
			if (ARMINSOPCODE(ret) != ARM_BL)
			   if (ARMINSOPCODE(ret) != ARM_B)
			   {
			      FATAL(("@I is Illegal instruction for PC24 relocation",ret));
			   }
			Free(ret);
			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
#define sign_extend(x,p)      (((x) ^ (1 << (p))) - (1 << (p)))
			data=4 * sign_extend(SectionGetData32(corrsec,generic) & 0x00ffffff,23);

			VERIFY_FASSERT((data==(-8)),("PC24 with addend!!!!!!"));
		
			FATAL(("RELOCTYPES VERTALEN"));
		     }
		     break; 
		  case R_ARM_ABS32:
		     {
			t_uint32 data=0;
			t_address generic;
			t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			data=SectionGetData32(corrsec,generic);



			FATAL(("RELOCTYPES VERTALEN"));
		     }
		     break;
		  case R_ARM_REL32:
		     {
			t_uint32 data;
			t_address generic;
			t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			if (corrsec->type!=CODE_SECTION)
			{
			   FATAL(("REL 32 not in code section!"));
			}
			data=SectionGetData32(corrsec,generic);


			FATAL(("RELOCTYPES VERTALEN"));
		     }
		     break;
		  case R_ARM_PC13:
		  case R_ARM_ABS16:
		  case R_ARM_ABS12:
		  case R_ARM_THM_ABS5:
		  case R_ARM_ABS8:
		  case R_ARM_SBREL32:
		  case R_ARM_THM_PC22:
		  case R_ARM_THM_PC8:
		  case R_ARM_AMP_VCALL9:
		  case R_ARM_SWI24:
		  case R_ARM_THM_SWI8:
		  case R_ARM_XPC25:
		  case R_ARM_THM_XPC22:
		     FATAL(("Implement %d!",ELF32_R_TYPE(rel[tel2].r_info)));
	       }
	    }
	    Free(rel);
	 }
	 tel++;
      }
   }

   if (table) Free(table);
   if (shdrs) Free(shdrs);
   if (sechdrstrtab) Free(sechdrstrtab);
   if (symbol_table) Free(symbol_table);
   if (strtab) Free(strtab);
}

#endif
