/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifdef BIT64ADDRSUPPORT
#include "../../config.h"
#include "../../simple_types.h"
#include <string.h>
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "elf_main.h"
#include "elf_types.h"
#include "elf_endian.h"

typedef struct 
{
  t_section * sec;
  Elf64_Shdr hdr;
  t_int32 prepadding;
  t_int32 described_by_pdr;
} sec_store_info;



/* Common operations on objectfiles */
/* ElfWriteCommonSameEndian64
 *
 * Current layout:
 * 
 * +--------------------------+
 * |        ELF HEADER        |
 * +--------------------------+
 * |    PROGRAM HEADERS(s)    |
 * +--------------------------+
 * |   NULL SECTION HEADER    |
 * |  NORMAL SECTION HEADERS  |
 * | SECSTRTAB SECTION HEADER |
 * +--------------------------+
 * |   NORMAL SECTION DATA    |
 * |   SECTION STRING TAB     |
 * +--------------------------+
 *
 */
#define SYMTABNAME ".symtab"
#define SYMSTRTABNAME ".strtab"
/*#define SYMTAB  */

#define VERBOSE_WRITE

#define MAX_PDR_DIFF 100
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_section.h"

int __sort_secs(const void * a, const void * b)
{
  sec_store_info * seca = (sec_store_info *)a, * secb = (sec_store_info *)b;

  if (seca->hdr.sh_addr > secb->hdr.sh_addr)
    return 1;
  else if (seca->hdr.sh_addr < secb->hdr.sh_addr)
    return -1;

  return 0;
}

void ElfWriteCommonSameEndian64(FILE * fp, Elf64_Ehdr * hdr, t_object * obj, int pagesize)
{
  Elf64_Phdr * phdrs=NULL;
  Elf64_Shdr nullshdr;
  Elf64_Shdr sectabshdr;
  t_uint32 tel;
  t_uint32 nphdrs=0;
  t_int64 run=0;
  t_bool sec_first=TRUE;
  t_uint64 can_be_placed_at; 
  sec_store_info * secs=NULL;

  t_string secstrtab;
  t_uint64 secstrtabsz=1;
  t_uint64 runstrtabsz=1;
  t_uint32 nsecs=0;

  /* Print section layout {{{*/
#ifdef VERBOSE_WRITE 
  for (tel=0; tel<obj->ncodes; tel++)
  {
    DiabloPrint(stdout,"Code section %d: %s at @G size @G\n",tel,obj->code[tel]->name,SECTION_GET_CADDRESS(obj->code[tel]),SECTION_CSIZE(obj->code[tel]));
  }

  for (tel=0; tel<obj->nrodatas; tel++)
  {
    DiabloPrint(stdout,"Rodata section %d: %s at @G size @G\n",tel,obj->rodata[tel]->name,SECTION_GET_CADDRESS(obj->rodata[tel]),SECTION_CSIZE(obj->rodata[tel]));
  }

  for (tel=0; tel<obj->ndatas; tel++)
  {
    DiabloPrint(stdout,"Data section %d: %s at @G size @G\n",tel,obj->data[tel]->name,SECTION_GET_CADDRESS(obj->data[tel]),SECTION_CSIZE(obj->data[tel]));
  }
  for (tel=0; tel<obj->nnotes; tel++)
  {
    DiabloPrint(stdout,"Note section %d: %s at @G size @G\n",tel,obj->note[tel]->name,SECTION_GET_CADDRESS(obj->note[tel]),SECTION_CSIZE(obj->note[tel]));
  }
  for (tel=0; tel<obj->nbsss; tel++)
  {
    DiabloPrint(stdout,"Bss section %d: %s at @G size @G\n",tel,obj->bss[tel]->name,SECTION_GET_CADDRESS(obj->bss[tel]),SECTION_CSIZE(obj->bss[tel]));
  }
#endif

  /* End of testing }}} */

  /* Calculate the section - string table size "\0.section\0... and the number of sections" {{{ */

  for (tel=0; tel<obj->ncodes; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->code[tel];
    secstrtabsz+=strlen(obj->code[tel]->name)+1;
    nsecs++;
  }
  for (tel=0; tel<obj->nrodatas; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->rodata[tel];
    secstrtabsz+=strlen(obj->rodata[tel]->name)+1;
    nsecs++;
  }
  for (tel=0; tel<obj->ndatas; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->data[tel];
    secstrtabsz+=strlen(obj->data[tel]->name)+1;
    nsecs++;
  }
for (tel=0; tel<obj->nnotes; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->note[tel];
    secstrtabsz+=strlen(obj->note[tel]->name)+1;
    nsecs++;
  }
  for (tel=0; tel<obj->nbsss; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->bss[tel];
    secstrtabsz+=strlen(obj->bss[tel]->name)+1;
    nsecs++;
  }
  secstrtabsz+=strlen(".shstrtab")+1;
#ifdef VERBOSE_WRITE
  DiabloPrint(stdout,"Total string table size - .shstrtab :  %d\n",secstrtabsz);
  DiabloPrint(stdout,"Total number of sections :  %d\n",nsecs);
#endif

  /* End of Calculate the section - string table size }}} */
  /* Allocate space for string table and initialize it {{{ */
  secstrtab=Malloc(secstrtabsz);
  secstrtab[0]='\0';
  /* }}} */

  /* Calculate section headers {{{ */
  nsecs=0;

  /* the zero sectionheader */
  nullshdr.sh_name=0;
  nullshdr.sh_type=SHT_NULL;
  nullshdr.sh_flags=0;
  nullshdr.sh_addr=0;
  nullshdr.sh_offset=0;
  nullshdr.sh_size=0;
  nullshdr.sh_link=SHN_UNDEF;
  nullshdr.sh_info=0;
  nullshdr.sh_addralign=0;
  nullshdr.sh_entsize=0;

  /* The normal sectionheaders */
  for (tel=0; tel<obj->ncodes; tel++)
  {
    secs[nsecs].hdr.sh_name=runstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_EXECINSTR;
    secs[nsecs].hdr.sh_addr=G_T_UINT64(SECTION_GET_CADDRESS(obj->code[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT64(SECTION_CSIZE(obj->code[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF; /*ter duidelijkheid MACRO gebruiken en niet 0*/
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT64(SECTION_ALIGN(obj->code[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    sprintf(secstrtab+runstrtabsz,"%s",obj->code[tel]->name);
    runstrtabsz+=strlen(obj->code[tel]->name)+1;
    nsecs++; 
  }

  for (tel=0; tel<obj->nrodatas; tel++)
  {
    secs[nsecs].hdr.sh_name=runstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC;
    secs[nsecs].hdr.sh_addr=G_T_UINT64(SECTION_GET_CADDRESS(obj->rodata[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT64(SECTION_CSIZE(obj->rodata[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT64(SECTION_ALIGN(obj->rodata[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    sprintf(secstrtab+runstrtabsz,"%s",obj->rodata[tel]->name);
    runstrtabsz+=strlen(obj->rodata[tel]->name)+1;
    nsecs++; 
  }

  for (tel=0; tel<obj->ndatas; tel++)
  {
    secs[nsecs].hdr.sh_name=runstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_WRITE;
    secs[nsecs].hdr.sh_addr=G_T_UINT64(SECTION_GET_CADDRESS(obj->data[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT64(SECTION_CSIZE(obj->data[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT64(SECTION_ALIGN(obj->data[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    sprintf(secstrtab+runstrtabsz,"%s",obj->data[tel]->name);
    runstrtabsz+=strlen(obj->data[tel]->name)+1;
    nsecs++; 
  }
for (tel=0; tel<obj->nnotes; tel++)
  {
    secs[nsecs].hdr.sh_name=runstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_WRITE;
    secs[nsecs].hdr.sh_addr=G_T_UINT64(SECTION_GET_CADDRESS(obj->note[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT64(SECTION_CSIZE(obj->note[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT64(SECTION_ALIGN(obj->note[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    sprintf(secstrtab+runstrtabsz,"%s",obj->note[tel]->name);
    runstrtabsz+=strlen(obj->note[tel]->name)+1;
    nsecs++; 
  }


  for (tel=0; tel<obj->nbsss; tel++)
  {
    secs[nsecs].hdr.sh_name=runstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_NOBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_WRITE;
    secs[nsecs].hdr.sh_addr=G_T_UINT64(SECTION_GET_CADDRESS(obj->bss[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT64(SECTION_CSIZE(obj->bss[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT64(SECTION_ALIGN(obj->bss[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    sprintf(secstrtab+runstrtabsz,"%s",obj->bss[tel]->name);
    runstrtabsz+=strlen(obj->bss[tel]->name)+1;
    nsecs++; 
  }

  /* sort the sections */
  qsort(secs,nsecs,sizeof(sec_store_info),__sort_secs);

  /* data_string table */
  sectabshdr.sh_name=runstrtabsz;
  sectabshdr.sh_type=SHT_STRTAB;
  sectabshdr.sh_addr=0;
  sectabshdr.sh_size=runstrtabsz+strlen(".shstrtab")+1;
  sectabshdr.sh_flags=0;
  sectabshdr.sh_offset=0;
  sectabshdr.sh_link=0;
  sectabshdr.sh_info=0;
  sectabshdr.sh_addralign=0;
  sectabshdr.sh_entsize=0;
  sprintf(secstrtab+runstrtabsz,"%s",".shstrtab");
  runstrtabsz+=strlen(".shstrtab")+1;
  /* }}} */

  VERIFY_FASSERT(runstrtabsz==secstrtabsz,("Something went wrong while building the string table!"));


  /* Calculate the number of phdrs needed. The phdrs describes which segments
   * the OS/debugger/whatever needs to make to run the program. So it describes
   * a collection of sections */

  for (tel=0; tel<nsecs; tel++)
  {
    if (secs[tel].hdr.sh_flags&SHF_ALLOC)
    {
      /* Alloc new segment when necessary */
      if ((secs[tel].hdr.sh_addr>run+MAX_PDR_DIFF) || (sec_first))
      {
	printf("Allocing new program header (%d) for %s, diff between previous programheader = %lld bytes\n",nphdrs,secstrtab+secs[tel].hdr.sh_name,secs[tel].hdr.sh_addr-run);
	phdrs=Realloc(phdrs,sizeof(Elf64_Phdr)*(nphdrs+1));
	phdrs[nphdrs].p_type=PT_LOAD;
	phdrs[nphdrs].p_offset=0; /* Depends on nphdrs */

	phdrs[nphdrs].p_paddr=phdrs[nphdrs].p_vaddr=secs[tel].hdr.sh_addr;
	phdrs[nphdrs].p_memsz=0;
	phdrs[nphdrs].p_align=pagesize;
	phdrs[nphdrs].p_flags=PF_R;
	phdrs[nphdrs].p_filesz=0;
	nphdrs++;
	secs[tel].prepadding=-1;
	sec_first=FALSE;
      }
      else if (secs[tel].hdr.sh_addr<run)
      {
	FATAL(("Sections overlap (or are not sorted as expected) for section %s at @G!(run = 0x%x)",secs[tel].sec->name,SECTION_GET_CADDRESS(secs[tel].sec),run));
      }
      else
      {
	secs[tel].prepadding=secs[tel].hdr.sh_addr-run;
	phdrs[nphdrs-1].p_memsz+=secs[tel].prepadding;
	phdrs[nphdrs-1].p_filesz+=(secs[tel].hdr.sh_type&SHT_PROGBITS)?secs[tel].prepadding:0;
	printf("Hoisting %s in previous program header, diff with previous section = %lld bytes\n",secstrtab+secs[tel].hdr.sh_name,secs[tel].hdr.sh_addr-run);
      }
      phdrs[nphdrs-1].p_filesz+=(secs[tel].hdr.sh_type&SHT_PROGBITS)?secs[tel].hdr.sh_size:0;
      phdrs[nphdrs-1].p_memsz+=secs[tel].hdr.sh_size;

      secs[tel].described_by_pdr=nphdrs-1;
      if (secs[tel].hdr.sh_flags & SHF_WRITE) phdrs[nphdrs-1].p_flags|=PF_W;
      if (secs[tel].hdr.sh_flags & SHF_EXECINSTR) phdrs[nphdrs-1].p_flags|=PF_X;
      run=secs[tel].hdr.sh_addr+secs[tel].hdr.sh_size;
    }
    else
    {
      secs[tel].described_by_pdr=-1;
      printf("Skipping section: %s (not in program header)\n",secstrtab+secs[tel].hdr.sh_name);
    }
  }

  /* Now place all segments in the file: segments file offsets need to be
   * congruent with the virtual address modulo the pagesize */  

  can_be_placed_at=sizeof(Elf64_Ehdr)+nphdrs*sizeof(Elf64_Phdr)+(nsecs+2)*sizeof(Elf64_Shdr);

  for (tel=0; tel<nsecs; tel++)
  {
    if (secs[tel].hdr.sh_flags&SHF_ALLOC)
    {
      if (secs[tel].prepadding==-1)
      {
	if ((can_be_placed_at&(pagesize-1))<(secs[tel].hdr.sh_addr&(pagesize-1)))
	{
	  secs[tel].prepadding=(secs[tel].hdr.sh_addr&(pagesize-1))-(can_be_placed_at&(pagesize-1));

	  printf("Segment prepadding =%d\n",secs[tel].prepadding);
	}
	else  if ((can_be_placed_at&(pagesize-1))>(secs[tel].hdr.sh_addr&(pagesize-1)))
	{
	  secs[tel].prepadding=pagesize+(secs[tel].hdr.sh_addr&(pagesize-1))-(can_be_placed_at&(pagesize-1));
	  printf("Segment prepadding =%d\n",secs[tel].prepadding);
	}
	else
	{
	  secs[tel].prepadding=0;
	}
	can_be_placed_at+=secs[tel].prepadding;
	phdrs[secs[tel].described_by_pdr].p_offset=can_be_placed_at;
	secs[tel].hdr.sh_offset=can_be_placed_at;
	if (secs[tel].hdr.sh_type&SHT_PROGBITS) can_be_placed_at+=secs[tel].hdr.sh_size;
      }
      else
      {
	can_be_placed_at+=secs[tel].prepadding;
	secs[tel].hdr.sh_offset=can_be_placed_at;
	if (secs[tel].hdr.sh_type&SHT_PROGBITS) can_be_placed_at+=secs[tel].hdr.sh_size;
      }
    }
  }



  sectabshdr.sh_offset=can_be_placed_at;



  /* Set ELFheaders */
  /* Ident is already set */
  /* Type is already set */
  /* Machine is already set */
  /* Version is already set */
  /* Entry is already set */
  hdr->e_phoff=sizeof(Elf64_Ehdr);
  hdr->e_shoff=sizeof(Elf64_Ehdr)+nphdrs*sizeof(Elf64_Phdr);
  /* Flags : TODO */
  /* Ehsize is already set */ 
  hdr->e_phentsize=sizeof(Elf64_Phdr);
  hdr->e_phnum=nphdrs;
  hdr->e_shentsize=sizeof(Elf64_Shdr);
  /* #entries in section header table =
   * #normal sections + 1 zero-section + 1 section name string table */
  hdr->e_shnum=nsecs+2;
  hdr->e_shstrndx=nsecs+1;

  /* Write ELF header + Program headers {{{ */
  fwrite(hdr,1,sizeof(Elf64_Ehdr),fp);
  fwrite(phdrs,sizeof(Elf64_Phdr),nphdrs,fp);
  /* }}} */

  /* Write section headers {{{ */

  /* Verify correctness: section table should be at hdr->e_shoff */
  VERIFY_FASSERT(ftell(fp)==hdr->e_shoff,("Section header table is not at %x as said in the program header, but at %x",hdr->e_shoff,ftell(fp)));
  /* 1 - zero header */
  fwrite(&nullshdr,sizeof(Elf64_Shdr),1,fp);
  /* 2 - normal section headers */
  for (tel=0; tel<nsecs; tel++)
    fwrite(&secs[tel].hdr,sizeof(Elf64_Shdr),1,fp);
  /* 3 - section name string table header */ 
  fwrite(&sectabshdr,sizeof(Elf64_Shdr),1,fp);
  /* }}} */

  /* Write section data {{{ */
  for (tel=0; tel<nsecs; tel++)
  {
    if (secs[tel].hdr.sh_flags&SHF_ALLOC)
    {
      static char padd[1<<14]; /*todo..aanpassen aan pagesize=16kB [OK]*/
      fwrite(padd, sizeof(char), secs[tel].prepadding, fp);
      printf("Writing section %d(%s) at 0x%x\n", tel, secs[tel].sec->name, ftell(fp));
      VERIFY_FASSERT(secs[tel].hdr.sh_offset==ftell(fp),("Section %d not at the right place (should be %x, is %x)",tel,secs[tel].hdr.sh_offset,ftell(fp)));
      if (secs[tel].hdr.sh_type&SHT_PROGBITS)fwrite(secs[tel].sec->data,sizeof(char),G_T_UINT64(SECTION_CSIZE(secs[tel].sec)),fp);
    }
  }
  /* }}} */

  /* Write section string table {{{ */

  VERIFY_FASSERT(sectabshdr.sh_offset==ftell(fp),("String table not at the right place: should be %x but is %x",sectabshdr.sh_offset,ftell(fp)));
  printf("Writing secstrtab(length %d) at Ox%x\n", secstrtabsz, (t_int32) ftell(fp));
  fwrite(secstrtab,sizeof(char),secstrtabsz,fp);

  fflush(fp);

  /* }}} */

  if (nphdrs) Free(phdrs);
  Free(secs);
  Free(secstrtab);
}

/* FUN : ElfReadCommon64
 * PAR : A file pointer pointing to the data of an objectfile after the header, the header, an obsolete/unused parameter, 
 *       the t_object structure to fill, A pointer to return the section headers in, 
 *       A pointer to return the symbol table in, A pointer to return the number of syms in, 
 *       A pointer to return the stringtable in, A pointer to return the section header stringtable in
 *       t_address (= the value of the symbol), type_uint8 (TODO= the type of the symbol)
 * RET : nothing 
 * DESC: This function reads the common information in a 64 bit ELF file and adapts the corresponding 
 *       t_object structure. Structures that need further parsing are returned. 
 *       TODO: The symboltable change probably made most returns obsolete ... */    


void ElfReadCommon64(FILE * fp, void * hdr,  
    t_object * obj, Elf64_Shdr ** shdr_ret, 
    Elf64_Sym ** symbol_table_ret, t_uint32 * numsyms_ret, 
    char ** strtab_ret, char ** sechdrstrtab_ret, 
    t_symbol *** lookup, t_bool switch_endian)
{
  char * dynstrtab=NULL; /* Will hold the dynamic string table */

  Elf64_Dyn * dynamic_table=NULL; /* Will hold the dynamic table */

  t_address tmp_addr, tmp_sz, tmp_align; /* Temporaries used in conversion from uint32 -> generic_address */

  Elf64_Ehdr * hdr64=(Elf64_Ehdr *) hdr; /* The Objectfileheader (casted correctly, it's passed as a union) */
  t_uint32 section_tel=0; /* Counter for the sections */
  t_uint32 dynamic_tel=0; /* Counter for the dynamic table entries */
  t_uint32 symbol_tel=0;  /* Counter for the symbols */

  if (!obj) FATAL(("Object = NULL"));
  /* Let the object know that it has to be careful if the endianness is switched */
  obj->switched_endian=switch_endian;
  /* Check if the use of this function is correct: TODO: remove in the final version */ 
  VERIFY_FASSERT(shdr_ret!=NULL, ("Must supply a shdr_return"));
  VERIFY_FASSERT(symbol_table_ret!=NULL, ("Must supply a symbol_table_return"));
  VERIFY_FASSERT(numsyms_ret!=NULL, ("Must supply a numsyms_return"));
  VERIFY_FASSERT(strtab_ret!=NULL, ("Must supply a strtab_return"));
  VERIFY_FASSERT(sechdrstrtab_ret!=NULL, ("Must supply a strtab_return"));
  VERIFY_FASSERT(lookup!=NULL, ("Must supply a symbol lookup table return"));

  /* We need to know the endianness to check this */
  VERIFY_FASSERT (hdr64->e_version==EV_CURRENT, ("Unknown elf version!"));

  /* Store the entry point in the object structure  */
  GenericSet_uint64(obj->entry,hdr64->e_entry);

  if (hdr64->e_shnum) /* If there are sections */
  {
    /* Create a hash table to store the section headers in for named lookups */
    obj->section_table=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,SectionHashNodeFree);

    /* STEP 1: Read the headers */ 
    /* Allocated place to store the headers */
    (*shdr_ret)=(Elf64_Shdr *) Malloc(sizeof(Elf64_Shdr)*hdr64->e_shnum);
    /* Look for the headers in the file 
     * streampos is used for when we are directly reading from an archive */
    fseek(fp,obj->streampos+hdr64->e_shoff,SEEK_SET);
    /* Read the headers */
    fread((*shdr_ret),sizeof(Elf64_Shdr),hdr64->e_shnum,fp);
    /* If we are doing a switched endian read, endian-swap all headers */

    if (switch_endian)
    {
      t_uint32 tel=0;

      for (tel=0; tel<hdr64->e_shnum; tel++)
	Elf64ShdrSwitchEndian(&(*shdr_ret)[tel]);
    }

    /* STEP 2: Find the sectionheader string table */ 
    (*sechdrstrtab_ret)=(char *) Malloc((*shdr_ret)[hdr64->e_shstrndx].sh_size);
    fseek(fp,obj->streampos+(*shdr_ret)[hdr64->e_shstrndx].sh_offset,SEEK_SET);
    fread((*sechdrstrtab_ret),(*shdr_ret)[hdr64->e_shstrndx].sh_size,1,fp);

    /* STEP 3: find the symboltable and the (symbol) stringtable and the dynamic stringtable*/
    while (section_tel<hdr64->e_shnum)
    {
      if ((*shdr_ret)[section_tel].sh_type==SHT_SYMTAB)
      {
	/* A normal symbol table.... current ABI says we should only have one symbol table */
	VERIFY_FASSERT(!(*symbol_table_ret),("More than one symbol_tablele in %s",obj->name)); 
	VERIFY_FASSERT(!((*shdr_ret)[section_tel].sh_size%sizeof(Elf64_Sym)),("Symtab size is not a multiple of %d in %s",sizeof(Elf64_Sym),obj->name));
	(*numsyms_ret)=(*shdr_ret)[section_tel].sh_size/sizeof(Elf64_Sym);
	(*symbol_table_ret)=Malloc((*shdr_ret)[section_tel].sh_size);
	fseek(fp,obj->streampos+(*shdr_ret)[section_tel].sh_offset,SEEK_SET);
	fread((*symbol_table_ret),(*shdr_ret)[section_tel].sh_size,1,fp);

	if (switch_endian)
	{
	  FATAL(("Switched endian not implemented!"));
	}
      }
      else if (strcmp((*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name,".strtab")==0)
      {
	/*  A normal string table... current ABI says we should only have one (normal) string table */
	VERIFY_FASSERT(!(*strtab_ret),("More than one stringtable in %s",obj->name)); 
	(*strtab_ret)=Malloc((*shdr_ret)[section_tel].sh_size); 
	fseek(fp,obj->streampos+(*shdr_ret)[section_tel].sh_offset,SEEK_SET);
	fread((*strtab_ret),(*shdr_ret)[section_tel].sh_size,1,fp);
      }
      else if (strcmp((*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name,".dynstr")==0)
      {
	/*  A dynamic string table... current ABI says we should only have one (dynamic) string table */
	VERIFY_FASSERT(!(dynstrtab),("More than one dynamic stringtable in %s",obj->name)); 
	(dynstrtab)=Malloc((*shdr_ret)[section_tel].sh_size); 
	fseek(fp,obj->streampos+(*shdr_ret)[section_tel].sh_offset,SEEK_SET);
	fread((dynstrtab),(*shdr_ret)[section_tel].sh_size,1,fp);
      }
      section_tel++;
    }

    /* STEP 4: find the loaded sections */
    section_tel=0; 
    while (section_tel<hdr64->e_shnum)
    {
      switch((*shdr_ret)[section_tel].sh_type)
      {
	case SHT_NULL:
	  /* The null section: an inactive section, probably only the first section i
	   * in the binary. It has no associated data */
	  if (section_tel!=0) FATAL(("NULL SECTION as %d-th section while reading %s!",section_tel,obj->name));
	  break;
	case SHT_PROGBITS: 

	  /* This section holds information defined by the program, whose format
	   * and meaning are determined solely by the program. */
	  VERBOSE(1,("Found program specific information. Looking at the flags to determine it's data")); 
	  if (((*shdr_ret)[section_tel].sh_flags & SHF_ALLOC))  /* This section will be allocated at run-time */
	  {
	    if (((*shdr_ret)[section_tel].sh_flags & SHF_EXECINSTR)) /* This section contains code */
	    {
	      if (((*shdr_ret)[section_tel].sh_flags & SHF_WRITE))  /* It's a writable section */
	      {
		WARNING(("Writable code section %s",(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name));
		VERBOSE(1,("   It's an executable (code) section named %s (size=%d)",(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name,(*shdr_ret)[section_tel].sh_size));
		GenericSet_uint64(tmp_addr,(*shdr_ret)[section_tel].sh_addr);
		GenericSet_uint64(tmp_sz,(*shdr_ret)[section_tel].sh_size);
		GenericSet_uint64(tmp_align,(*shdr_ret)[section_tel].sh_addralign);
		ObjectAddSectionFromFile(obj,CODE_SECTION,TRUE,fp, obj->streampos+(*shdr_ret)[section_tel].sh_offset, tmp_addr, tmp_sz, tmp_align, (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name );

	      }
	      else /* Normal code section */
	      {
		VERBOSE(1,("   It's an executable (code) section named %s (size=%d)",(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name,(*shdr_ret)[section_tel].sh_size));

		GenericSet_uint64(tmp_addr,(*shdr_ret)[section_tel].sh_addr);
		GenericSet_uint64(tmp_sz,(*shdr_ret)[section_tel].sh_size);
		GenericSet_uint64(tmp_align,(*shdr_ret)[section_tel].sh_addralign);
		ObjectAddSectionFromFile(obj,CODE_SECTION,TRUE,fp, obj->streampos+(*shdr_ret)[section_tel].sh_offset, tmp_addr, tmp_sz, tmp_align, (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name );
	      }
	    }
	    else /* Non-code sections */
	    {
	      if (((*shdr_ret)[section_tel].sh_flags & SHF_WRITE)) /* Writable section */
	      {
		VERBOSE(1,("   It's a writable data section named %s",(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name));
		GenericSet_uint64(tmp_addr,(*shdr_ret)[section_tel].sh_addr);
		GenericSet_uint64(tmp_sz,(*shdr_ret)[section_tel].sh_size);
		GenericSet_uint64(tmp_align,(*shdr_ret)[section_tel].sh_addralign);
		ObjectAddSectionFromFile(obj,DATA_SECTION,FALSE,fp, obj->streampos+(*shdr_ret)[section_tel].sh_offset, tmp_addr, tmp_sz, tmp_align, (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name );
	      }
	      else /* Constant section */
	      {
		VERBOSE(1,("   It's a constant data section named %s",(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name));
		GenericSet_uint64(tmp_addr,(*shdr_ret)[section_tel].sh_addr);
		GenericSet_uint64(tmp_sz,(*shdr_ret)[section_tel].sh_size);
		GenericSet_uint64(tmp_align,(*shdr_ret)[section_tel].sh_addralign);
		ObjectAddSectionFromFile(obj,DATA_SECTION,TRUE,fp, obj->streampos+(*shdr_ret)[section_tel].sh_offset, tmp_addr, tmp_sz,tmp_align,  (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name );
	      }
	    }
	  }
	  else /* Not allocated section */
	  {
	    /* Non alloced progbits are normally data describing the program, unused during dynamic 
	     * linking or normal execution*/ 
	    /* stab, comment, ??? */
	    VERBOSE(1,("   It's a debug section named %s",(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name));
	  }
	  /* The only other real sections are bss sections (SHT_NOBITS) */
	  break;
	case SHT_SYMTAB: 
	  /* This section hold a symbol table. There can be only one! (currently). We ignore it 
	   * here: symbol_tableles should be fetched in STEP 3 */
	  VERBOSE(1,("Found a symbol table name %s",(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name)); 
	  break;
	case SHT_STRTAB: 
	  /* This section holds a string table. We can have multiple string tables (in theory, the current ABI does not allow them) ! We ignore it
	   * here: all necessary stringtables should be fetched in STEP 3 */ 
	  VERBOSE(1,("Found a string table")); 
	  break;
	case SHT_RELA: 
	  /* This section contains relocations with explicit addends. Currently unsupported... */
	  VERBOSE(1,("Found relocations with explicit addends for section %s", (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name+5));
	  break;
	case SHT_HASH: 
	  /* This section contains a hash table (for dynamic linking). There can be only one! (currently).
	   * No need to load this, although we could verify it's correctness, but we can regenerate it when
	   * the binary is written*/
	  VERBOSE(1,("Found a hash table")); 
	  break;
	case SHT_DYNAMIC: 
	  /* TODO */
	  /* The section holds information for dynamic linking. There can be only one! (currently) */
	  VERBOSE(1,("Found a dynamic section....")); 
	  VERIFY_FASSERT(!dynamic_table,("Found 2 dynamic tables. This does not conform to current ELF (SYSV - ARM - INTEL) ABI specifications"));
	  dynamic_table=Malloc((*shdr_ret)[section_tel].sh_size);
	  fseek(fp,obj->streampos+(*shdr_ret)[section_tel].sh_offset,SEEK_SET);
	  fread(dynamic_table,1,(*shdr_ret)[section_tel].sh_size,fp);

	  while(dynamic_table[dynamic_tel].d_tag!=DT_NULL) /* An entry with a DT_NULL tag marks the end of the _DYNAMIC array. */
	  {
	    switch(dynamic_table[dynamic_tel].d_tag)
	    {
	      /* Holds a string table offset to a NULL-determined string, given the name of a needed libraryi. 
	       * The offset points in the dynamic table to a DT_STRTAB entry*/
	      case DT_NEEDED:
		printf("NEEDED: %s\n",dynstrtab + dynamic_table[dynamic_tel].d_un.d_val);
		VERBOSE(1,("Need something dynamic: %s",dynstrtab + dynamic_table[dynamic_tel].d_un.d_val));
		break;
		/* Total size in bytes of the plt. Must be present if we have a DT_JMPREL entry */
	      case DT_PLTRELSZ:
		printf("PLTSIZE = %lld\n",dynamic_table[dynamic_tel].d_un.d_val);
		break;
		/* An address associated with the PLT or GOT */
	      case DT_PLTGOT:
		printf("PLTGOT = %llx\n",dynamic_table[dynamic_tel].d_un.d_ptr);
		break;
		/* The address of the (dynamic) Symbol hash table */
	      case DT_HASH:
		printf("HASH = %llx\n",dynamic_table[dynamic_tel].d_un.d_ptr);
		break;
		/* The address of the (dynamic) string table */
	      case DT_STRTAB:
		printf("STRTAB = %llx\n",dynamic_table[dynamic_tel].d_un.d_ptr);
		break;
		/* The address of the (dynamic) symbol table */
	      case DT_SYMTAB:
		printf("SYMTAB = %llx\n",dynamic_table[dynamic_tel].d_un.d_ptr);
		break;
		/* The size in bytes of the (dynamic) string table */
	      case DT_STRSZ:
		printf("STRSIZE = %lld\n",dynamic_table[dynamic_tel].d_un.d_val);
		break;
		/* The size in bytes of a (dynamic) symbol table entry*/
	      case DT_SYMENT:
		printf("SYMENTSIZE = %lld\n",dynamic_table[dynamic_tel].d_un.d_val);
		break;
		/* The address of a relocation table  with explicit addends. The dynamic linker sees only one table! */
	      case DT_RELA:
		printf("RELA TABLE\n");
		break;
		/* The size of the the relocation table  with explicit addends. */
	      case DT_RELASZ:
		printf("RELASZ\n");
		break;
		/* The size of an element in the relocation table  with explicit addends.*/
	      case DT_RELAENT:
		printf("RELAENT\n");
		break;
	      case DT_INIT:
		printf("INIT\n");
		break;
	      case DT_FINI:
		printf("FINI\n");
		break;
		/* The address of a relocation table  with implicit addends. The dynamic linker sees only one table! */
	      case DT_REL:
		printf("REL TABLE at %llx\n",dynamic_table[dynamic_tel].d_un.d_ptr);
		break;
		/* The size of the reloc table */
	      case DT_RELSZ:
		printf("RELSIZE = %lld\n",dynamic_table[dynamic_tel].d_un.d_val);
		break;
	      case DT_RELENT:
		printf("RELENT\n");
		break;
	      case DT_PLTREL:
		printf("PLTREL\n");
		break;
		/* Used for debugging. Not specified by the ABI. Can be ignored?. */
	      case DT_DEBUG:
		printf("DEBUG\n");
		break;
	      case DT_JMPREL:
		printf("JMPREL\n");
		break;
	      default:
		if ((dynamic_table[dynamic_tel].d_tag>=DT_LOOS)&&(dynamic_table[dynamic_tel].d_tag<=DT_HIOS))
		{
		  WARNING(("Ignoring OS specific dynamic!"));
		}
		else if ((dynamic_table[dynamic_tel].d_tag>=DT_LOPROC)&&(dynamic_table[dynamic_tel].d_tag<=DT_HIPROC))
		{
		  WARNING(("Ignoring Processor specific dynamic!"));
		}
		else
		{
		  FATAL(("Unknown d_tag: %d %x",dynamic_table[dynamic_tel].d_tag,dynamic_table[dynamic_tel].d_tag));
		}
	    }
	    dynamic_tel++;
	  }
	  WARNING(("Object has cross - object references ! "));
	  exit(0);

	  break;
	case SHT_NOTE:
	  /* TODO */
	  /* The section holds information that marks the file in some way */
	  VERBOSE(1,("Found a section marking the binary")); 
	  GenericSet_uint64(tmp_addr,(*shdr_ret)[section_tel].sh_addr);
	  GenericSet_uint64(tmp_sz,(*shdr_ret)[section_tel].sh_size);
	  GenericSet_uint64(tmp_align,(*shdr_ret)[section_tel].sh_addralign);
	  ObjectAddSectionFromFile(obj,NOTE_SECTION,TRUE,fp, obj->streampos+(*shdr_ret)[section_tel].sh_offset, tmp_addr, tmp_sz,tmp_align,  (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name );


	  break;
	case SHT_NOBITS:
	  /* A section of this type occupies no space in the file but otherwise resembles SHT_PROGBITS. 
	   * Although this section contains no bytes, the sh_offset member contains the conceptual 
	   * file offset. */
	  VERBOSE(1,("Found program specific information but no bits. Looking at the flags to determine it's data")); 
	  if (((*shdr_ret)[section_tel].sh_flags & SHF_ALLOC)) /* Must have alloc flags */
	  {
	    if (((*shdr_ret)[section_tel].sh_flags & SHF_EXECINSTR)) FATAL(("Executable NOBITS!"));
	    if (((*shdr_ret)[section_tel].sh_flags & SHF_WRITE)) /* Must be writable */
	    {
	      VERBOSE(1,("   It's a writable data section named %s",(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name));
	      GenericSet_uint64(tmp_addr,(*shdr_ret)[section_tel].sh_addr);
	      GenericSet_uint64(tmp_sz,(*shdr_ret)[section_tel].sh_size);
	      GenericSet_uint64(tmp_align,(*shdr_ret)[section_tel].sh_addralign);
	      ObjectAddSectionFromFile(obj,BSS_SECTION,FALSE,fp, obj->streampos+(*shdr_ret)[section_tel].sh_offset, tmp_addr, tmp_sz,tmp_align,  (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name );
	    }
	    else FATAL(("No bits section without write(useless)"));
	  }
	  else FATAL(("No bits section without alloc (useless)"));

	  break;
	case SHT_REL: 
	  /* This section contains relocations without addends. We ignore those here, because they get
	   * processed in STEP 5 */
	  VERBOSE(1,("Found relocations without explicit addends for section %s", (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name+4));
	  break;
	case SHT_SHLIB:
	  /* SHLIB is named in the System V ABI specs, but it's meaning is unspecified... Programs 
	   * conformant to the ABI should not contain it */
	  FATAL(("Found a SHT_SHLIB section. This is named in the System V ABI specs, but it's meaning is unspecified... Programs conformant to the ABI should not contain it. Send more info on this file to bdebus@frodo.elis.rug.ac.be"));
	  break;
	case SHT_DYNSYM: 
	  /* TODO */
	  VERBOSE(1,("SHT_DYNSYM")); 
	  break;
	default:
	  if (((*shdr_ret)[section_tel].sh_type>=SHT_LOOS)&&((*shdr_ret)[section_tel].sh_type<=SHT_HIOS))
	  {
	    /* TODO */
	    WARNING(("Ignoring operating specific section type %x, name = %s",((*shdr_ret)[section_tel].sh_type) , (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name   ));
	  }
	  else if (((*shdr_ret)[section_tel].sh_type>=SHT_LOPROC)&&((*shdr_ret)[section_tel].sh_type<=SHT_HIPROC))
	  {
	    /* TODO */
	    if (obj->subtype==OBJECT_FILE_SUBTYPE_IA64ELF && ((*shdr_ret)[section_tel].sh_type)==SHT_IA_64_UNWIND)
	    {
	      if ((*shdr_ret)[section_tel].sh_size!=0)
	      {
		if (!(((*shdr_ret)[section_tel].sh_flags & SHF_ALLOC))) FATAL(("Ia 64 unwind is not alloc!"));
		if (((*shdr_ret)[section_tel].sh_flags & SHF_EXECINSTR)) FATAL(("Ia 64 unwind is exec!"));
		if (((*shdr_ret)[section_tel].sh_flags & SHF_WRITE)) FATAL(("Ia 64 unwind is non constant!"));
		VERBOSE(1,("   It's a constant unwind section named %s",(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name));
		GenericSet_uint64(tmp_addr,(*shdr_ret)[section_tel].sh_addr);
		GenericSet_uint64(tmp_sz,(*shdr_ret)[section_tel].sh_size);
		GenericSet_uint64(tmp_align,(*shdr_ret)[section_tel].sh_addralign);
		ObjectAddSectionFromFile(obj,DATA_SECTION,TRUE,fp, obj->streampos+(*shdr_ret)[section_tel].sh_offset, tmp_addr, tmp_sz,tmp_align,  (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name );
	      }

	    }
	    else
	      WARNING(("Ignoring processor specific section type %x, name = %s",((*shdr_ret)[section_tel].sh_type),  (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name ));
	  }
	  else if (((*shdr_ret)[section_tel].sh_type>=SHT_LOUSER)&&((*shdr_ret)[section_tel].sh_type<=SHT_HIUSER))
	  {
	    /* TODO */
	    WARNING(("Ignoring user specific section type %x, name = %s",((*shdr_ret)[section_tel].sh_type) ,  (*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name ));
	  }
	  else
	  {
	    FATAL(("Unknown section type! %x, name %s",((*shdr_ret)[section_tel].sh_type),(*sechdrstrtab_ret) + (*shdr_ret)[section_tel].sh_name));
	  }
      }
      section_tel++;
    }
  }
  /* Read the symbol table and put it in a generic form... */
  if (hdr64->e_shnum)
  {
    /* Create a new symbol table */
    obj->symbol_table=SymbolTableNew(obj);

    *lookup=(t_symbol **) Calloc((*numsyms_ret) , sizeof(t_symbol *));

    /* Add all symbols */
    for (symbol_tel=0; symbol_tel<(*numsyms_ret); symbol_tel++)
    {
      if (symbol_tel==0) continue; /* ignore the reserved symbol */

      /* regular symbols: all but undefined, absolute and to_common */
      if (((*symbol_table_ret)[symbol_tel].st_shndx!=SHN_UNDEF)&&((*symbol_table_ret)[symbol_tel].st_shndx!=SHN_ABS)&&((*symbol_table_ret)[symbol_tel].st_shndx!=SHN_COMMON))
      {
	t_uint32 type;
	t_section * sec=SectionGetFromObjectByName(obj,(*sechdrstrtab_ret)+(*shdr_ret)[(*symbol_table_ret)[symbol_tel].st_shndx].sh_name);
	t_address generic;

	if (!sec) continue; 
	VERIFY_FASSERT(sec,("Could not find section %s",(*sechdrstrtab_ret)+(*shdr_ret)[(*symbol_table_ret)[symbol_tel].st_shndx].sh_name));
	
	GenericSet_uint64(generic,(*symbol_table_ret)[symbol_tel].st_value);

	if (ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)==STT_NOTYPE)
	{
	  if (strcmp("$d",(*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name)==0)
	    (*lookup)[symbol_tel]=SymbolTableAddSymbolToSection(obj->symbol_table,LOCAL_SYMBOL,T_RELOCATABLE(sec),generic,((*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name), TRUE);
	  else if (strcmp("$a",(*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name)==0)
	    (*lookup)[symbol_tel]=SymbolTableAddSymbolToSection(obj->symbol_table,LOCAL_SYMBOL,T_RELOCATABLE(sec),generic,((*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name), TRUE);
	  else
	  {
	    if (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_GLOBAL)
	    {
	      type=GLOBAL_SYMBOL;
	    }
	    else if (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_LOCAL)
	    {
	      type=LOCAL_SYMBOL;
	    }
	    else
	    {
	      type=WEAK_SYMBOL;
	    }

	    (*lookup)[symbol_tel]=SymbolTableAddSymbolToSection(obj->symbol_table,type,T_RELOCATABLE(sec),generic,((*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name), (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_LOCAL));
	    WARNING(("symbol %d: %s has an unknown type %d\n",symbol_tel,(*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name,ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)));
	  }
	  continue;
	}
	else if (ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)==STT_OBJECT)
	{
	  if (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_GLOBAL)
	  {
	    type=GLOBAL_SYMBOL;
	  }
	  else if (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_LOCAL)
	  {
	    type=LOCAL_SYMBOL;
	  }
	  else
	  {
	    type=WEAK_SYMBOL;
	  }
	}
	else if (ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)==STT_FUNC)
	{
	  type=FUNCTION_SYMBOL;
	}
	else if (ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)==STT_SECTION)
	{
	  type=ORIGINAL_SECTION_SYMBOL;
	}
	else 
	{
	  FATAL(("unknown type %d\n",ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)));
	}
	(*lookup)[symbol_tel]=SymbolTableAddSymbolToSection(obj->symbol_table,type,(t_relocatable *) sec,generic,((*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name), (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_LOCAL));
      }
      /* Common symbols: not yet allocated bss */
      else if ((*symbol_table_ret)[symbol_tel].st_shndx==SHN_COMMON)
      {
	t_uint32 type;
	t_uint32 size,alignment;
	t_address generic, generic_align, generic_sz;
	if (ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)==STT_NOTYPE)
	{
	  WARNING(("symbol %s has an unknown type %d\n",(*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name,ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)));
	  continue;
	}
	else if (ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)==STT_OBJECT)
	{
	  if (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_GLOBAL)
	  {
	    type=GLOBAL_SYMBOL;
	  }
	  else if (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_LOCAL)
	  {
	    type=LOCAL_SYMBOL;
	  }
	  else
	  {
	    type=WEAK_SYMBOL;
	  }
	}
	else if (ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)==STT_FUNC)
	{
	  type=FUNCTION_SYMBOL;
	}
	else if (ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)==STT_SECTION)
	{
	  type=ORIGINAL_SECTION_SYMBOL;
	}
	else 
	{
	  FATAL(("unknown type %d\n",ELF64_ST_TYPE((*symbol_table_ret)[symbol_tel].st_info)));
	}

//	size = (*symbol_table_ret)[symbol_tel].st_size;
//	alignment = (*symbol_table_ret)[symbol_tel].st_value;
	/* now this is a bit tricky: we want to encode both the requested size
	 * of the common symbol and it's alignment. The problem is, both
	 * values are 32-bits and we only have one 32-bit field to encode the
	 * both of them: current_address. So we just hope that neither is 
	 * bigger than 2**16, and try to encode them in halfwords
	 */
//	VERIFY_FASSERT(size < (1 << 16), ("COMMON symbol encoding trick won't work: size is too big"));
//	VERIFY_FASSERT(alignment < (1 << 16), ("COMMON symbol encoding trick won't work: alignment is too big"));
//	GenericSet_uint64(generic, (alignment << 16) + size);

	/*CHANGED*/
	/* for COMMON symbols, the alignment is stored in sym->current_address, the size in sym>size */
	size = (*symbol_table_ret)[symbol_tel].st_size;
	alignment = (*symbol_table_ret)[symbol_tel].st_value;
	GenericSet_uint64(generic_align, alignment);
	GenericSet_uint64(generic_sz, size);
	

	(*lookup)[symbol_tel]=SymbolTableAddSymbolToCommon(obj->symbol_table,type,generic_align,generic_sz,((*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name),(ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_LOCAL));
      }
      /* Constants */
      else if ((*symbol_table_ret)[symbol_tel].st_shndx==SHN_ABS)
      {
	t_address generic;
	GenericSet_uint64(generic,(*symbol_table_ret)[symbol_tel].st_value);
	(*lookup)[symbol_tel]=SymbolTableAddAbsoluteSymbol(obj->symbol_table,generic,((*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name), (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_LOCAL));
      }
      else if ((*symbol_table_ret)[symbol_tel].st_shndx==SHN_UNDEF)
      {
	if (ELF64_ST_BIND((*symbol_table_ret)[symbol_tel].st_info)==STB_WEAK)
	  (*lookup)[symbol_tel]=SymbolTableAddWeakUndefinedSymbol(obj->symbol_table,((*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name));
	else
	  (*lookup)[symbol_tel]=SymbolTableAddUndefinedSymbol(obj->symbol_table,((*strtab_ret) + (*symbol_table_ret)[symbol_tel].st_name));
      }
    }
  }
}
#endif
/* vim: set shiftwidth=2: */
