/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/* Symbol Version Table
 *
 * The Symbol Version Table is contained in the special section .gnu.version which has a section type 
 * of SHT_GNU_versym. This section has the same number of entries as the Dynamic
 * Symbol Table.
 *
 * This section contains an array of elements of type Elfxx_Half. Each entry specifies the 
 * version defined for or required by the corresponding symbol in the Dynamic Symbol Table.
 *
 * The values in the Symbol Version Table are unique to the object in which they are located. 
 * These values are identifiers that are provided by the the vna_other member of the
 * Elfxx_Vernaux structure or the vd_ndx member of the Elfxx_Verdef structure.
 *
 * The values 0 and 1 are reserved.
 *
 * 0: The symbol is local, not available outside the object.
 * 1: The symbol is defined in this object and is globally available.
 *
 * All other values are used to identify version strings located in one of the other Symbol 
 * Version sections. The value itself is not the version associated with the symbol. The string 
 * identified by the value defines the version of the symbol. */


/* Version Definitions
 *
 * Symbol definitions are contained in the special section .gnu.version_d which has a section 
 * type of SHT_GNU_verdef. The number of entries in this section is contained in the
 * DT_VERDEFNUM entry of the Dynamic Section. The sh_link member of the section header points 
 * to the section that contains the strings referenced by this section.
 */

#include "elf_types.h"
#include <stdio.h>

typedef struct {
   Elf32_Half    vd_version; /* Version revision. This value is currently set to 1, 
				and will be reset if the versioning implementation 
				is incompatibly altered. */
   Elf32_Half    vd_flags;   /* Version information flag bitmask */
   Elf32_Half    vd_ndx;     /* Version index numeric value referencing the SHT_GNU_versym section. */ 
   Elf32_Half    vd_cnt;     /* Number of associated verdaux array entries. */
   Elf32_Word    vd_hash;    /* Version name hash value (ELF hash function). */
   Elf32_Word    vd_aux;     /* Offset to a corresponding entry in the verdaux array, in bytes. */
   Elf32_Word    vd_next;    /* Offset to the next verdef entry, in bytes. */
} Elf32_Verdef;

typedef struct {
   Elf32_Word    vda_name; /* Offset to the version or dependency name string in the section header, in bytes. */
   Elf32_Word    vda_next; /* Offset to the next verdaux entry, in bytes. */
} Elf32_Verdaux;


/* Version Requirements
 *
 * Symbol definitions are contained in the special section .gnu.version_r which has a 
 * section type of SHT_GNU_verneed. The number of entries in this section is contained in the
 * DT_VERNEEDNUM entry of the Dynamic Section. The sh_link member of the section header points 
 * to the section that contains the strings referenced by this section. */

typedef struct {
   Elf32_Half    vn_version; /* Version of structure. This value is currently set to 1, 
				and will be reset if the versioning implementation is 
				incompatibly altered. */
   Elf32_Half    vn_cnt;     /* Number of associated verneed array entries. */
   Elf32_Word    vn_file;    /* Offset to the file name string in the section header, in bytes. */
   Elf32_Word    vn_aux;     /* Offset to a corresponding entry in the vernaux array, in bytes. */
   Elf32_Word    vn_next;    /* Offset to the next verneed entry, in bytes. */
} Elf32_Verneed;


typedef struct {
   Elf32_Word    vna_hash;   /* Dependency name hash value (ELF hash function). */
   Elf32_Half    vna_flags;  /* Dependency information flag bitmask. */
   Elf32_Half    vna_other;  /* Object file version identifier used in the .gnu.version symbol version array. 
				Bit number 15 controls whether or not the object is hidden; 
				if this bit is set, the object cannot be used and
				the static linker will ignore the symbol's presence in the object. */
   Elf32_Word    vna_name;   /* Offset to the dependency name string in the section header, in bytes. */
   Elf32_Word    vna_next;   /* Offset to the next vernaux entry, in bytes. */
} Elf32_Vernaux;


/* Startup Sequence
 *
 * When loading a sharable object, version definition data from the loaded object is 
 * analyzed to assure that it meets the version requirements of the calling object. 
 * The dynamic loader retrieves the entries in the caller's Elfxx_Verneed array and 
 * attempts to find matching definition information in the loaded Elfxx_Verdef table.
 * Each object and dependency is tested in turn. If a symbol definition is missing, 
 * the loader returns an error. A warning is issued instead of a hard error when the 
 * vna_flags bit for VER_FLG_WEAK is set in the Elfxx_Vernaux entry.
 * When the versions referenced by undefined symbols in the loaded object are found, 
 * version availability is certified. The test completes without error and the object 
 * is made available. */

/* Symbol Resolution
 *
 * When symbol versioning is used in an object, relocations extend the performance of 
 * definition testing beyond the simple match of symbol name strings: the version of 
 * the reference must also equal the name of the definition. The same index that is used 
 * in the symbol table can be referenced in the SHT_GNU_versym section, and the value of 
 * this index is then used to acquire name data. The corresponding requirement string is 
 * retrieved from the Elfxx_Verneed array, and likewise, the corresponding definition 
 * string from the Elfxx_Verdef table.
 *
 * Bit number 15 of the version symbol controls whether or not the object is hidden; if this 
 * bit is set, the object cannot be used and the static linker will ignore the symbol's 
 * presence in the object.
 *
 * Results differ in the interaction of objects that variously use symbol versioning. 
 *
 * The object with the reference and the object with the definitions may both use 
 * versioning. All described matching is processed in this case. A fatal error is triggered 
 * when no matching definition can be found in the object whose name is the one referenced 
 * by the vn_name element in the Elfxx_Verneed entry.
 *
 * The object with the reference may not use versioning, while the object with the definitions 
 * does. In this instance, only the definition with index numbers 1 and 2 will be used in the
 * reference match, the same identified by the static linker as the base definition. In infrequent 
 * cases where the static linker was not used, as in calls to dlopen(), a version that does not
 * have the base definition index is acceptable as long as it is the only version for which the 
 * symbol is defined.
 *
 * The object with the reference may use versioning, but the object with the definitions specifies 
 * none. A matching symbol is accepted in this case. A fatal error is triggered in the
 * unlikely event that a corruption in the required symbols list obscured an outdated object file 
 * and caused a match on the object filename in the Elfxx_Verneed entry.
 *
 * Finally, both the object with the reference and the object with the definitions may not use versioning. 
 * The behavior in this instance defaults to pre-existing symbol rules.*/

void PrintLinuxVersionInfo(Elf32_Verdef * verdef)
{
   printf("VERSION %d\n",verdef->vd_version);
   printf("FLAGS: "); /*PrintBits(verdef->vd_flags);*/ printf("\n");
   
}
