/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <string.h>
#include "../../config.h"
#include "../../kernel/diablo_symbol.h"
#include "../../diablo_options.h"
#include "../../kernel/diablo_iterators.h"
#ifdef ARM_SUPPORT
#include "elf_arm.h"
#endif

#ifdef I386_SUPPORT
#include "elf_i386.h"
#endif

#ifdef SPARC_SUPPORT
#include "elf_sparc.h"
#endif

#ifdef ALPHA_SUPPORT
#include "elf_alpha.h"
#endif

#ifdef IA64_SUPPORT
#include "elf_ia64.h"
#endif

#ifdef MIPS_SUPPORT
#include "elf_mips.h"
#endif

#ifdef POWERPC_SUPPORT
#include "elf_powerpc.h"
#endif


#include "elf_main.h"
#include "elf_types.h"
#include "../../simple_types.h"
#include "../../utils/util_error.h"
#include "../../utils/util_machine.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_status.h"
#include "../../datastructures/data_string.h"
#include "../../utils/util_term.h"
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_section.h"
#ifdef COLORS
#include "../../utils/util_color.h"
#endif

t_uint64 ElfGetSizeofHeaders(t_object * obj, t_layout_rule * layoutscript);
t_fileformat_action_struct elf_actions = {
  ElfGetSizeofHeaders
};

/* UNIX ELF hash function (needed to build/use the dynamic hash tables) */
#if 0
static unsigned long ElfHash(char *name)
{
  unsigned long h = 0, g;
  while ( *name ) {
    h = ( h << 4 ) + *name++;
    if ( (g = (h & 0xF0000000)) != 0)
      h ^= g >> 24;
    h &= ~g;
  }
  return h;
}
#endif
/* This structure describes how we handle a particular Elf format (based on the information in the Elf(32/64) header 
 * TODO: We should have a write in this (since different ABI's write in different ways) 
 * */
#if 0
typedef struct
{
  t_objectsubtype type;  /* the type of ELF-file */
  Elf32_Byte class;       /* TODO */
  my_byteorder objectbo;  /* Little or Big endian objectfile */ 
  my_byteorder machinebo; /* Little or Big endian machine */
  Elf32_Word machine;     /* TODO */
} ElfHandler;

/* Common */ 

ElfHandler elf_type_ids[]=
{
#ifdef ARM_SUPPORT
   {OBJECT_FILE_SUBTYPE_ARMELFLSB ,ELFCLASS32,  LSB,LSB,EM_ARM},
   {OBJECT_FILE_SUBTYPE_ARMELFMSB ,ELFCLASS32,  MSB,MSB,EM_ARM},
   {OBJECT_FILE_SUBTYPE_ARMELFLSB ,ELFCLASS32,  LSB,MSB,EM_SWITCHED_ARM},
   {OBJECT_FILE_SUBTYPE_ARMELFMSB ,ELFCLASS32,  MSB,LSB,EM_SWITCHED_ARM},
#endif
#ifdef SPARC_SUPPORT
   {OBJECT_FILE_SUBTYPE_SPARCELF  ,ELFCLASS32,  MSB,LSB,EM_SWITCHED_SPARC},
   {OBJECT_FILE_SUBTYPE_SPARCELF  ,ELFCLASS32,  MSB,MSB,EM_SPARC},
#endif
#ifdef I386_SUPPORT
   {OBJECT_FILE_SUBTYPE_I386ELF    ,ELFCLASS32,  LSB,LSB,EM_386},
   {OBJECT_FILE_SUBTYPE_I386ELF    ,ELFCLASS32,  LSB,MSB,EM_SWITCHED_386},
#endif
#ifdef ALPHA_SUPPORT
   {OBJECT_FILE_SUBTYPE_ALPHAELF  ,ELFCLASS64,  LSB,LSB,EM_ALPHA},
   {OBJECT_FILE_SUBTYPE_ALPHAELF  ,ELFCLASS64,  LSB,MSB,EM_SWITCHED_ALPHA},
#endif
#ifdef IA64_SUPPORT
   {OBJECT_FILE_SUBTYPE_IA64ELF,ELFCLASS64,  LSB,LSB,EM_IA_64},
   {OBJECT_FILE_SUBTYPE_IA64ELF,ELFCLASS64,  LSB,MSB,EM_SWITCHED_IA_64},
#endif
#ifdef POWERPC_SUPPORT
   {OBJECT_FILE_SUBTYPE_POWERPCELFLSB ,ELFCLASS32,  LSB,LSB,EM_POWERPC},
   {OBJECT_FILE_SUBTYPE_POWERPCELFMSB ,ELFCLASS32,  MSB,MSB,EM_POWERPC},
   {OBJECT_FILE_SUBTYPE_POWERPCELFLSB ,ELFCLASS32,  LSB,MSB,EM_SWITCHED_POWERPC},
   {OBJECT_FILE_SUBTYPE_POWERPCELFMSB ,ELFCLASS32,  MSB,LSB,EM_SWITCHED_POWERPC},
#endif
   {ST_UNKNOWN   ,ELFCLASSNONE,LSB,LSB,EM_NONE}
};
#endif

t_bool IsElf(FILE * fp)
{
   Elf32_Byte buffer[4];
   long fpos=ftell(fp);
   int ret=fread(buffer,sizeof(Elf32_Byte),4,fp); 
   fseek(fp,fpos,SEEK_SET);
   
   if (ret!=4) return FALSE;
   if ((buffer[0]!=ELFMAG0)||(buffer[1]!=ELFMAG1)||(buffer[2]!=ELFMAG2)||(buffer[3]!=ELFMAG3))
   {
      return FALSE;
   }
   VERBOSE(1,("File is an ELF objectfile")); 
   return TRUE;
}

/* This struct is used to pass the right type of elf header to the elf-reader */
typedef union 
{
   Elf32_Ehdr h32;
   Elf64_Ehdr h64;
} mixed_elf_hdr;

/*! This function will detect if a file is a supported ELF type, and if so,
 *  read in the elf objectfile and translate it to an object.
 *
 *  \par fp an open filepointer to the start of the object (so it can also read from archives
 *  \par obj a pointer to an allocated object that will be filled in.*/

void ReadElf(FILE * fp, t_object * obj)
{
#if 0
   Elf32_Byte buffer[EI_NIDENT];
   my_byteorder byte_order;
   mixed_elf_hdr hdr;
   int tel=0;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
#endif

   FATAL(("This is just a stub.... Should never be called!"));

#if 0
   
   Filepos save=ftell(fp);

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
   {
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
   }

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
	 FATAL(("Invalid ELF data type in %s!",obj->name));
      case ELFDATA2LSB:
	 byte_order=LSB;
	 break;
      case ELFDATA2MSB:
	 byte_order=MSB;
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr.h32),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 machine=hdr.h32.e_machine;
	 break;
      case ELFCLASS64: 
	 if (fread(&(hdr.h64),sizeof(Elf64_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 machine=hdr.h64.e_machine;
	 break;
      case ELFCLASSNONE:
      default:
	 FATAL(("Invalid ELF class in %s!",obj->name));
   }
/*
   while(elfhandler[tel].type!=ST_UNKNOWN)
   {
      if ((elfhandler[tel].class==buffer[EI_CLASS])&&(elfhandler[tel].machinebo==global_machine_info.endian)
	 &&(elfhandler[tel].objectbo==byte_order)&&(elfhandler[tel].machine==machine))
      {
	 obj->actions=elfhandler[tel].acts;
	 obj->subtype=elfhandler[tel].type;
	 if (buffer[EI_CLASS]==ELFCLASS32)
	    elfhandler[tel].read(fp, &hdr.h32, elfhandler[tel].data, obj);
	 else
	    elfhandler[tel].read(fp, &hdr.h64, elfhandler[tel].data, obj);
	 return;
      }
      tel++;
   }
*/
   FATAL_USAGE(("Unknown machine (0x%x=%d) - class (%d) - byteorder combination (M: %d, O: %d)",machine,machine,buffer[EI_CLASS],global_machine_info.endian,byte_order));
#endif
}

t_uint64 ElfGetSizeofHeaders(t_object * obj, t_layout_rule * layoutscript)
{
  t_uint64 elfhdrsz, phdrsz, shdrsz;
  int gaps;
  t_layout_rule * r;
  int prev = -1;
  /* this is a bit of a problem: we cannot accurately predict the number 
     of needed phdrs as long as we don't know the complete section layout.
     on the other hand, we need the size of the phdrs to determine the
     section layout :-(
     we'll just have to make an educated guess:
     # of phdrs = # of possible gaps between sections + 1
   */
  if (obj->description->address_size == ADDRSIZE32)
  {
    elfhdrsz = sizeof(Elf32_Ehdr);
    phdrsz = sizeof(Elf32_Phdr);
    shdrsz = sizeof(Elf32_Shdr);
  }
  else if (obj->description->address_size == ADDRSIZE64)
  {
    elfhdrsz = sizeof(Elf64_Ehdr);
    phdrsz = sizeof(Elf64_Phdr);
    shdrsz = sizeof(Elf64_Shdr);
  }
  else
  {
    FATAL(("Unknown address size for this object"));
    exit(0); /* Keeps the compiler happy */
  }

#define PREV_WAS_NOT_ASSIGN	1
#define PREV_WAS_ASSIGN		2


  {
     t_section * sec;
     t_uint32 tel;
     OBJECT_FOREACH_SECTION(obj,sec,tel)
     {
	sec->tmp=0;
     }
  }
  
  /* now count the number of gaps */
  gaps = 0;
  prev = 3;
  for (r = layoutscript; r; r=r->next)
  {
    if (r->kind == ASSIGN)
    {
      if (strcmp(r->u.assign->lhs,".")) continue;
      prev = PREV_WAS_ASSIGN;
    }
    else if (r->kind == SECTION_SPEC) 
    {
       t_section * sec=SectionGetFromObjectByName(obj,r->u.secspec->name);
       if (!sec) continue;

       sec->tmp=1;
       if (prev == PREV_WAS_ASSIGN || (prev == 3)) { printf("Section %s\n",r->u.secspec->name); gaps++;  }
       prev = PREV_WAS_NOT_ASSIGN;
    }
    else if (r->kind == PUT_SECTIONS)
    {
       t_uint32 tel;
       t_uint32 nplaced=1;
       if (r->u.sectype=='N')
       {
	  for (tel=0; tel<obj->nnotes; tel++)
	     nplaced&=obj->note[tel]->tmp;
       }
       else if (r->u.sectype=='C')
       {
	  for (tel=0; tel<obj->ncodes; tel++)
	     nplaced&=obj->code[tel]->tmp;
       }
       else if (r->u.sectype=='R')
       {
	  for (tel=0; tel<obj->nrodatas; tel++)
	  {
	     nplaced&=obj->rodata[tel]->tmp;
	  }
       }
       else if (r->u.sectype=='D')
       {
	  for (tel=0; tel<obj->ndatas; tel++)
	  {
	     nplaced&=obj->data[tel]->tmp;
	  }
       }
       else if (r->u.sectype=='B')
       {
	  for (tel=0; tel<obj->nbsss; tel++)
	     nplaced&=obj->bss[tel]->tmp;
       }

       if (nplaced) continue;
	  
       if ((prev == PREV_WAS_ASSIGN) || (prev == 3)) { printf("Putsection\n"); gaps++; }
       prev = PREV_WAS_NOT_ASSIGN;
    }
    else if (r->kind == SEG_END)
    {
       prev=3;
    }
    else if (r->kind == SEG_START)
    {
    }
    else
      FATAL(("unexpected kind of rule"));
  }
  /* adjust if the last rule was an assignment (this creates no gap) */
  if (prev == PREV_WAS_ASSIGN) gaps--;
#undef PREV_WAS_NOT_ASSIGN
#undef PREV_WAS_ASSIGN

  printf("Guessed %d segments\n",gaps);
  printf("Returning %d\n",(t_uint32) (elfhdrsz+(gaps)*phdrsz+(2+obj->ncodes+obj->nrodatas+obj->ndatas+obj->nbsss+obj->nnotes)*shdrsz));
  return (t_uint64) (elfhdrsz+(gaps)*phdrsz+(2+obj->ncodes+obj->nrodatas+obj->ndatas+obj->nbsss+obj->nnotes)*shdrsz);
}
