/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*!\addtogroup FILE_FORMATS Handling different file formats */
/*@{ */
/*! \defgroup ELF Elf */ 
/*@{ */
#ifndef MY_ELF_H
#define MY_ELF_H
#include "elf_types.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_action.h"

#define EI_MAG0    0
#define EI_MAG1    1
#define EI_MAG2    2
#define EI_MAG3    3
#define EI_CLASS   4
#define EI_DATA    5
#define EI_VERSION 6
#define EI_PAD     7
#define EI_NIDENT 16

#define ET_NONE   0
#define ET_REL    1
#define ET_EXEC   2
#define ET_DYN    3
#define ET_CORE   4
#define ET_LOPROC 0xff00
#define ET_HIPROC 0xffff

#define EM_NONE            0x0	
#define EM_M32             0x0001 /* AT&T WE 32100 */
#define EM_SWITCHED_M32    0x0100
#define EM_SPARC           0x0002
#define EM_SWITCHED_SPARC  0x0200
#define EM_386             0x0003
#define EM_SWITCHED_386    0x0300
#define EM_MIPS		   0x0008
#define EM_SWITCHED_MIPS   0x0800
#define EM_ARM             0x0028 /* ARM-ELF if the endianness is the same */ 
#define EM_SWITCHED_ARM    0x2800 /* ARM-ELF if the endianness is switched */ 
#define EM_SH              0x002a /* SH-ELF if the endianness is the same */ 
#define EM_SWITCHED_SH     0x2a00 /* SH-ELF if the endianness is switched */ 
#define EM_IA_64	   0x0032 /* Itanium if the endianness is the same */
#define EM_SWITCHED_IA_64  0x3200 /* Itanium if the endianness is switched */
#define EM_ALPHA           0x9026 /* Alpha if the endianness is the same */
#define EM_SWITCHED_ALPHA  0x2690 /* Alpha if the endianness is switched */
#define EM_PPC		   0x0014	/* PowerPC */
#define EM_SWITCHED_PPC	   0x1400	/* PowerPC */
#define EV_NONE    0
#define EV_CURRENT 1

/*! \name Elf-Magic */ 
/*@{*/
#define ELFMAG0      0x7f
#define ELFMAG1      'E' 
#define ELFMAG2      'L'
#define ELFMAG3      'F'
/*@}*/

/*! \name Elf-Classes */ 
/*@{*/
#define ELFCLASSNONE 0
#define ELFCLASS32   1
#define ELFCLASS64   2
/*@}*/

/*! \name Elf-Datatypes */ 
/*@{*/
#define ELFDATANONE  0
#define ELFDATA2LSB  1
#define ELFDATA2MSB  2
/*@}*/

/* PROGRAM HEADER CONSTS */

#define PT_NULL         0
#define PT_LOAD         1
#define PT_DYNAMIC      2
#define PT_INTERP       3
#define PT_NOTE         4
#define PT_SHLIB        5
#define PT_PHDR         6

#define PT_LOPROC       0x70000000
#define PT_HIPROC       0x7fffffff

#define PF_X           0x1
#define PF_W           0x2
#define PF_R           0x4
#define PF_ARM_ENTRY   0x80000000 

/* SECTION HEADER CONSTS */
/** \name Section types */
/*@{*/
/* 1.1. These are specified by  System V Application Binary Interface */

#define SHT_NULL          0
#define SHT_PROGBITS      1
#define SHT_SYMTAB        2
#define SHT_STRTAB        3
#define SHT_RELA          4
#define SHT_HASH          5
#define SHT_DYNAMIC       6
#define SHT_NOTE          7
#define SHT_NOBITS        8
#define SHT_REL           9
#define SHT_SHLIB         10
#define SHT_DYNSYM        11
#define SHT_LOPROC        0x70000000
#define SHT_HIPROC        0x7fffffff
#define SHT_LOUSER        0x80000000
#define SHT_HIUSER        0x8fffffff

/* 1.2. Linux And Sun add OS specific relocations */

#define SHT_LOOS         0x60000000     /* Lower bound of OS-specific section-types */
#define SHT_HIOS         0x6fffffff     /* Upper bound of OS-specific section-types */

/* 1.3. Special Linux Section types (Linux Standard Base Specification) */

#define SHT_GNU_verdef    0x6ffffffd    /* This section contains the symbol versions that are provided. */
#define SHT_GNU_verneed   0x6ffffffe    /* This section contains the symbol versions that are required. */
#define SHT_GNU_versym    0x6fffffff    /* This section contains the Symbol Version Table. */


/* 1.4. Special IA64 types */ 
#define SHT_IA_64_UNWIND	(SHT_LOPROC + 1) /* unwind bits */

/* 1.5. Special MIPS types */
#define SHT_MIPS_REGINFO  0x70000006	/* Needed to read in the .reginfo section, to recover the original gp-value */
/*@}*/

/** \name Section flags */
/*@{*/
#define SHF_WRITE         0x1 
#define SHF_ALLOC         0x2 
#define SHF_EXECINSTR     0x4 
#define SHF_ARM_ENTRYSECT 0x10000000 
#define SHF_ARM_COMDEF    0x80000000
#define SHF_MASKPROC      0xf0000000
/*@}*/

/** \name Special section */
/*@{*/
#define SHN_UNDEF  0
#define SHN_LORESERVE    0xff00
#define SHN_ABS          0xfff1
#define SHN_COMMON       0xfff2
#define SHN_HIRESERVE    0xffff
/*@}*/

/** \name SYMBOL CONSTANTS */
/*@{*/
/* utils */
#define ELF32_ST_BIND(i) ((i)>>4)
#define ELF32_ST_TYPE(i) ((i)&0xf)
#define ELF32_ST_INFO(b,t) (((b)<<4)+((t)&0xf))

#define ELF64_ST_BIND(val)              ELF32_ST_BIND (val)
#define ELF64_ST_TYPE(val)              ELF32_ST_TYPE (val)
#define ELF64_ST_INFO(bind, type)       ELF32_ST_INFO ((bind), (type))

/* Symbol binding */
#define STB_LOCAL  0
#define STB_GLOBAL 1
#define STB_WEAK   2
#define STB_LOPROC 13
#define STB_HIPROC 15

/* Symbol types */
#define STT_NOTYPE  0
#define STT_OBJECT  1
#define STT_FUNC    2
#define STT_SECTION 3
#define STT_FILE    4
#define STT_LOPROC  13
#define STT_HIPROC  15
/*@}*/

/** \name i386 relocs.  */
/*@{*/
#define R_386_32        1
#define R_386_PC32      2
#define R_386_GOT32     3
#define R_386_PLT32     4
#define R_386_COPY      5
#define R_386_GLOB_DAT  6
#define R_386_JMP_SLOT  7
#define R_386_RELATIVE  8
#define R_386_GOTOFF    9
#define R_386_GOTPC     10
/*@}*/

/** Elf 32 Header */
typedef struct 
{
   Elf32_Byte e_ident[EI_NIDENT];
   Elf32_Half e_type;
   Elf32_Half e_machine;
   Elf32_Word e_version;
   Elf32_Addr e_entry; 
   Elf32_Off  e_phoff;
   Elf32_Off  e_shoff;
   Elf32_Word e_flags;
   Elf32_Half e_ehsize;
   Elf32_Half e_phentsize;
   Elf32_Half e_phnum;  
   Elf32_Half e_shentsize;
   Elf32_Half e_shnum;   
   Elf32_Half e_shstrndx; 
} Elf32_Ehdr; 

/** Elf 64 Header */
typedef struct
{
   unsigned char e_ident[EI_NIDENT];
   Elf64_Half    e_type;
   Elf64_Half    e_machine;
   Elf64_Word    e_version;
   Elf64_Addr    e_entry;
   Elf64_Off     e_phoff;
   Elf64_Off     e_shoff;
   Elf64_Word    e_flags;
   Elf64_Half    e_ehsize;
   Elf64_Half    e_phentsize;
   Elf64_Half    e_phnum;
   Elf64_Half    e_shentsize;
   Elf64_Half    e_shnum;
   Elf64_Half    e_shstrndx;
} Elf64_Ehdr;

/** Program header table entry */
typedef struct
{
   Elf32_Word p_type;
   Elf32_Off  p_offset;
   Elf32_Addr p_vaddr;
   Elf32_Addr p_paddr;
   Elf32_Word p_filesz;
   Elf32_Word p_memsz;
   Elf32_Word p_flags;
   Elf32_Word p_align;
} Elf32_Phdr;


/** Program header table entry */
typedef struct
{
   Elf64_Word p_type;
   Elf64_Word p_flags;
   Elf64_Off  p_offset;
   Elf64_Addr p_vaddr;
   Elf64_Addr p_paddr;
   Elf64_Xword p_filesz;
   Elf64_Xword p_memsz;
   Elf64_Xword p_align;
} Elf64_Phdr;




/** Section header table entry */
typedef struct
{
   Elf32_Word sh_name;
   Elf32_Word sh_type;
   Elf32_Word sh_flags;
   Elf32_Addr sh_addr;
   Elf32_Off  sh_offset;
   Elf32_Word sh_size;
   Elf32_Word sh_link;
   Elf32_Word sh_info;
   Elf32_Word sh_addralign;
   Elf32_Word sh_entsize;
} Elf32_Shdr;


/** Section header table entry (64 bit) */
typedef struct
{
   Elf64_Word    sh_name;
   Elf64_Word    sh_type;
   Elf64_Xword   sh_flags;
   Elf64_Addr    sh_addr;
   Elf64_Off     sh_offset;
   Elf64_Xword   sh_size;
   Elf64_Word    sh_link;
   Elf64_Word    sh_info;
   Elf64_Xword   sh_addralign;
   Elf64_Xword   sh_entsize;
} Elf64_Shdr;



/** Symbol table entry (32 bit)*/
typedef struct
{
   Elf32_Word st_name;
   Elf32_Addr st_value;
   Elf32_Word st_size;
   Elf32_Byte st_info;
   Elf32_Byte st_other;
   Elf32_Half st_shndx; 
} Elf32_Sym;


/** Symbol table entry (64 bit)*/
typedef struct
{
   Elf64_Word    st_name;
   Elf64_Byte    st_info;
   Elf64_Byte    st_other;
   Elf64_Section st_shndx;
   Elf64_Addr    st_value;
   Elf64_Xword   st_size;
} Elf64_Sym;

/** Relocation table entry */
typedef struct
{
   Elf32_Addr r_offset;
   Elf32_Word r_info;
} Elf32_Rel;

  
typedef struct  
{ 
   Elf64_Addr    r_offset;
   Elf64_Xword   r_info;
} Elf64_Rel;


typedef struct
{
  Elf32_Addr	r_offset;
  Elf32_Word	r_info;	
  Elf32_Sword	r_addend;
} Elf32_Rela;

typedef struct
{
  Elf64_Addr	r_offset;
  Elf64_Xword	r_info;	
  Elf64_Sxword	r_addend;
} Elf64_Rela;


/** \name Relocation entry utils*/
/*@{*/
#define ELF32_R_SYM(i)     ((i) >> 8)
#define ELF32_R_TYPE(i)    ((unsigned char)  i)
#define ELF32_R_INFO(s, t) (((s) << 8) + ((unsigned char) t))

#define ELF64_R_SYM(i)			((i) >> 32)
#define ELF64_R_TYPE(i)			((i) & 0xffffffff)
#define ELF64_R_INFO(sym,type)		((((Elf64_Xword) (sym)) << 32) + (type))
/*@}*/


typedef struct
{
  Elf32_Sword d_tag;
  union {
    Elf32_Word d_val;
    Elf32_Addr d_ptr;
  } d_un;
} Elf32_Dyn; 


typedef struct
{
   Elf64_Sxword  d_tag;
   union
   {
      Elf64_Xword d_val;
      Elf64_Addr d_ptr;
   } d_un;
} Elf64_Dyn;

typedef struct
{
  Elf32_Word	ri_gprmask;
  Elf32_Word	ri_cprmask[4];
  Elf32_Sword ri_gp_value;
} Elf_RegInfo;


#define DT_NULL     0
#define DT_NEEDED   1
#define DT_PLTRELSZ 2
#define DT_PLTGOT   3
#define DT_HASH     4
#define DT_STRTAB   5
#define DT_SYMTAB   6
#define DT_RELA     7
#define DT_RELASZ   8
#define DT_RELAENT  9 
#define DT_STRSZ    10
#define DT_SYMENT   11
#define DT_INIT     12
#define DT_FINI     13
#define DT_REL      17
#define DT_RELSZ    18
#define DT_RELENT   19
#define DT_PLTREL   20 
#define DT_DEBUG    21 
#define DT_JMPREL   23 
#define DT_LOOS     0x60000000
#define DT_HIOS     0x6fffffff 
#define DT_LOPROC   0x70000000
#define DT_HIPROC   0x7fffffff
t_bool IsElf(FILE * fp);
void ReadElf(FILE * fp,t_object * obj);
void ElfAction(t_action action,...);
#ifdef ARM_SUPPORT
#include "elf_arm.h"
#endif
#ifdef SH_SUPPORT
#include "elf_sh3.h"
#endif
#ifdef IA64_SUPPORT
#include "elf_ia64.h"
#endif
#ifdef I386_SUPPORT
#include "elf_i386.h"
#endif
#ifdef ALPHA_SUPPORT
#include "elf_alpha.h"
#endif
#ifdef MIPS_SUPPORT
#include "elf_mips.h"
#endif
#ifdef POWERPC_SUPPORT
#include "elf_powerpc.h"
#endif

extern t_fileformat_action_struct elf_actions;

#endif
/*@} */
/*@} */
