/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/

/* Includes {{{*/
#include "../config.h"
#include "diablo_action.h" 
#include "diablo_print.h" 
#include "diablo_object.h"
#include "diablo_flowgraph.h"
#include "diablo_function.h"
#include "diablo_iterators.h"
#include "../diablo_options.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include <stdarg.h>
#include "../utils/util_term.h"
#ifdef INSTRUMENTATION
#include "../atom.h"
#endif
/*}}}*/

/* ObjectAction {{{ */
/*!
 * \param action  the type of action to take. The other parameters depends on this 
 */ 

void ObjectAction(t_action action,...)
{
  va_list val;
  switch(action)
  {
    case GROUP_INFO:
      FATAL(("Virtual GROUP_INFO called"));
    case READ_MAP:
      {
        t_object * obj;
        t_string name;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        name=va_arg(val,t_string);
        VERIFY_FASSERT(obj->actions,("No actions supported for this object"));
        VERIFY_FASSERT(obj->actions->ReadMap,("Read Linkmap not supported for this object!"));
        obj->actions->ReadMap(obj,name);
      }
      break;
    case DISASSEMBLE:
      {
        t_object * obj;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        ObjectDisassemble(obj);
        break;
      }	
    case DISASSEMBLE_SECTION:
      {
        t_object * obj;
        t_section* sec;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        sec=va_arg(val,t_section *);
        VERIFY_FASSERT(obj,("Illegal number of arguments for action DISASSEMBLE SECTION"));
        VERIFY_FASSERT(sec,("Illegal number of arguments for action DISASSEMBLE SECTION"));
        VERIFY_FASSERT(obj->actions,("No actions for this object!"));
        VERIFY_FASSERT(obj->actions->DisassembleSec,("Section Disassembly not supported for this object!"));
        ANALYSIS_START("Disassemble section");
        SectionInitDisassembly(sec); 
        obj->actions->DisassembleSec(obj,sec);
        SectionFiniDisassembly(obj,sec);
        ANALYSIS_END("Disassemble",("nins"));
        break;
      }
    case FLOWGRAPH:
      {
        t_object * obj;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        ObjectFlowgraph(obj);
        break;
      }
    case FLOWGRAPH_SECTION:
      {
        t_uint32 ret;
        t_object * obj;
        t_section* sec;
        t_function * function;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        sec=va_arg(val,t_section *);
        VERIFY_FASSERT(obj,("Illegal number of arguments for action FLOWGRAPH SECTION"));
        VERIFY_FASSERT(sec,("Illegal number of arguments for action FLOWGRAPH SECTION"));
        VERIFY_FASSERT(obj->actions,("No actions for this object!"));
        VERIFY_FASSERT(obj->actions->FlowgraphSec,("Section Flowgraph not supported for this object!"));
        ANALYSIS_START("Flowgraph construction");
        SectionInitFlowgraph(sec);
        obj->actions->FlowgraphSec(obj,sec);
        SectionFiniFlowgraph(sec); 
        ANALYSIS_SIMPLE_END("Flowgraph construction");
	
        CfgCountBlocks(sec->data);

        /* Identify the address producers */
#ifdef STEGO
	if(!global_options.stegoDecode || global_options.restoreState)
#endif
	  ObjectAction(MAKE_ADDRESS_PRODUCERS,obj,obj->code[0]);

#ifdef INSTRUMENTATION
	/* {{{ add hell edges for instrumentation functions */
	AtomProtosResolve(obj);
	{
	  t_section * code = obj->code[0];
	  t_cfg * cfg = SECTION_CFG(code);
	  t_bbl * block;
	  t_symbol * sym = SymbolTableGetSymbolByName(obj->sub_symbol_table,"__diablo_inst_Translate");
	  if (!sym) FATAL(("Could not find instrumentation Translate"));

	  block = T_BBL(sym->base);
	  if (!block || RELOCATABLE_TYPE(block) != RT_BBL)
	    FATAL(("symbol Translate does not point to code"));

	  CfgEdgeNewCall(cfg,cfg->hell_node,block,NULL,NULL);
	}
	{
	  t_section * code = obj->code[0];
	  t_cfg * cfg = SECTION_CFG(code);
	  t_bbl * block;
	  t_symbol * sym = SymbolTableGetSymbolByName(obj->sub_symbol_table,"__diablo_inst_ReverseTranslate");
	  if (!sym) FATAL(("Could not find instrumentation function ReverseTranslate"));

	  block = T_BBL(sym->base);
	  if (!block || RELOCATABLE_TYPE(block) != RT_BBL)
	    FATAL(("symbol ReverseTranslate does not point to code"));

	  CfgEdgeNewCall(cfg,cfg->hell_node,block,NULL,NULL);
	}
	/* }}} */
#endif

        /* Find possible functions */
        ANALYSIS_START("Function list construction");
        ret=CfgCreateFunctions(sec->data);
        ANALYSIS_END("Function list construction",("%d functions",ret));
        CfgCountBlocks(sec->data);

        /* Associate basic blocks to a function */
        ANALYSIS_START("Association of basic blocks to their functions");
        FOREACH_FUNCTION(sec,function) FunctionAssociateBbls(function);

	/* Pick up blocks that are only target of a jump from hell */
	{
	  t_bbl* bbl, * prev_bbl;
	  t_bool changed;
	  prev_bbl = NULL;
	  CFG_FOREACH_BBL(T_CFG(sec->data),bbl)
	  {
	    t_cfg_edge * edge; 
	    if (!BBL_FUN(bbl))
	      BBL_FOREACH_PRED_EDGE(bbl,edge)
		if (IsHellNode(EDGE_HEAD(edge)) && EDGE_CAT(edge)==ET_JUMP)
		  {
		    if (BBL_FUN(prev_bbl))
		      {
			t_cfg_edge * tmp, * edge2;
			/*			DiabloPrint(stdout,"inserting @ieB\n",bbl);fflush(stdout);*/
			BblInsertInFunction(bbl,BBL_FUN(prev_bbl));
			BBL_FOREACH_SUCC_EDGE_SAFE(bbl,edge2,tmp)
			  {
			    if ((CFG_EDGE_TAIL(edge2) == BBL_FUN(prev_bbl)->cfg->hell_node) && (EDGE_CAT(edge2) == ET_RETURN) && (edge2->corr_edge == NULL))
			      {
				ret++;
				CfgEdgeKill( edge2);
				edge2=(t_cfg_edge *) GraphAppendNode(T_GRAPH(BBL_FUN(prev_bbl)->cfg), T_NODE(bbl), (t_node *) FUN_BBL_LAST(BBL_FUN(prev_bbl)), ET_JUMP);	
				edge2->cfg=(t_cfg *) BBL_FUN(prev_bbl)->cfg;
				edge2->refcount=1;
				/* this shouldn't be ET_FALLTHROUGH, because in the case
				   of a conditional return there already is a
				   fallthrough edge! */
			      }
			    else if (EDGE_CAT(edge2) == ET_RETURN) 
			      {
				FATAL(("How did we get here? @ieB @E",bbl,edge2));
			      }
			  }
		      }
		    else
		      FATAL(("OOPS\n"));
		    break;
		  }
	    if (BBL_FUN(bbl))
	      prev_bbl = bbl;
	  }

	  printf("part 2\n");

	  changed = TRUE;
	  
	  while (changed)
	    {
	      changed = FALSE;
	      
	      CFG_FOREACH_BBL(T_CFG(sec->data),bbl)
		{
		  if (BBL_FUN(bbl) && !IsHellNode(bbl))
		    {
		      t_cfg_edge * edge; 
		      BBL_FOREACH_SUCC_EDGE(bbl,edge)
			{
			  t_cfg_edge * tmp, * edge2;
			  if (EDGE_CAT(edge)==ET_CALL || EDGE_CAT(edge)==ET_RETURN) continue;
			  if (BBL_FUN(EDGE_TAIL(edge))) continue;
			  BblInsertInFunction(CFG_EDGE_TAIL(edge),BBL_FUN(bbl));
			  BBL_FOREACH_SUCC_EDGE_SAFE(bbl,edge2,tmp)
			    {
			      if ((CFG_EDGE_TAIL(edge2) == BBL_FUN(prev_bbl)->cfg->hell_node) && (EDGE_CAT(edge2) == ET_RETURN) && (edge2->corr_edge == NULL))
				{
				  ret++;
				  CfgEdgeKill( edge2);
				  edge2=(t_cfg_edge *) GraphAppendNode(T_GRAPH(BBL_FUN(prev_bbl)->cfg), T_NODE(bbl), T_NODE(FUN_BBL_LAST(BBL_FUN(prev_bbl))), ET_JUMP);	
				  edge2->cfg=(t_cfg *) BBL_FUN(prev_bbl)->cfg;
				  edge2->refcount=1;
				  /* this shouldn't be ET_FALLTHROUGH, because in the case
				     of a conditional return there already is a
				     fallthrough edge! */
				}
			      else if (EDGE_CAT(edge2) == ET_RETURN) 
				{
				  FATAL(("How did we get here? @ieB @E",bbl,edge2));
				}
			    }
			  changed = TRUE;
			}
		    }
		}
	    }
	}
	
        ANALYSIS_SIMPLE_END("Association of basic blocks to their functions");

        CfgCountBlocks(sec->data);

        /* Convert edges to interprocedural version where necessary */ 
        ANALYSIS_START("Patching edges to function representation");
        ret=CfgPatchInterProcedural(sec->data);
        ANALYSIS_END("Patching edges to function representation",("Patched %d edges",ret));

        if (global_options.generate_dots)
        {
          ANALYSIS_START("Flowgraph export");
#ifdef HAVE_DIRS
	  DirectoryAssure("./dots",FALSE);
	  Export_Flowgraph(T_BBL(SECTION_CFG(sec)->entries->entry_bbl), 0xffff, "./dots/flowgraph.dot");
#else
	  Export_Flowgraph(T_BBL(SECTION_CFG(sec)->entries->entry_bbl), 0xffff, "./flowgraph.dot");
#endif
	  ANALYSIS_SIMPLE_END("Flowgraph export");
        }
        break;
      }
    case MAKE_ADDRESS_PRODUCERS:
      {
        t_object * obj;
        t_section* sec;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        sec=va_arg(val,t_section *);
        VERIFY_FASSERT(obj,("Illegal number of arguments for action MAKE ADDRESS PRODUCERS"));
        VERIFY_FASSERT(sec,("Illegal number of arguments for action MAKE ADDRESS PRODUCERS"));
        VERIFY_FASSERT(obj->actions,("No actions for this object!"));
        VERIFY_FASSERT(obj->actions->MakeAddressProducers,("Detection and handling of address producers not supported for this object!"));

        ANALYSIS_START("Detecting address and constant producers");
        obj->actions->MakeAddressProducers(sec);
        ANALYSIS_SIMPLE_END("Detecting address and constant producers");
        break;
      }
    case ASSEMBLE_SECTION:
      {
        t_object * obj;
        t_section* sec;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        sec=va_arg(val,t_section *);
        VERIFY_FASSERT(obj,("Illegal number of arguments for action ASSEMBLE SECTION"));
        VERIFY_FASSERT(sec,("Illegal number of arguments for action ASSEMBLE SECTION"));
        VERIFY_FASSERT(obj->actions,("No actions for this object!"));
        VERIFY_FASSERT(obj->actions->AssembleSec,("Section Assembly not supported for this object!"));

        ANALYSIS_START("Assemble section");
        SectionInitAssembly(obj,sec,0);
        obj->actions->AssembleSec(obj,sec);
        SectionFiniAssembly(obj,sec);
        ANALYSIS_END("Assemble section",("nins"));
        break;
      }
    case ASSEMBLE:
      {
        t_object * obj;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        VERIFY_FASSERT(obj,("Illegal number of arguments for action ASSEMBLE"));
        ObjectAssemble(obj);
        break;
      }
    case WRITE:
      {
        t_object * obj;
        t_string filename;

        va_start(val,action);
        obj=va_arg(val,t_object *);
        filename=va_arg(val,t_string);

        VERIFY_FASSERT(obj,("Illegal number of arguments for action WRITE"));
        VERIFY_FASSERT(filename,("Illegal number of arguments for action WRITE"));
        VERIFY_FASSERT(obj->actions->Write,("Write not supported for this object!"));
        obj->actions->Write(obj,filename);
      }
      break;
    case DEFLOWGRAPH:
      {
        t_object * obj;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        ObjectDeflowgraph(obj);
        break;
      }
    case DEFLOWGRAPH_SECTION:
      {
        t_object * obj;
        t_section* sec;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        sec=va_arg(val,t_section *);
        VERIFY_FASSERT(obj,("Illegal number of arguments for action DEFLOWGRAPHING SECTION"));
        VERIFY_FASSERT(sec,("Illegal number of arguments for action DEFLOWGRAPHING SECTION"));
        VERIFY_FASSERT(obj->actions,("No actions for this object!"));
        VERIFY_FASSERT(obj->actions->DeFlowgraphSec,("Section Deflowgraph not supported for this object!"));
        ANALYSIS_START("Deflowgraphing");
        SectionInitDeflowgraph(obj,sec);
        obj->actions->DeFlowgraphSec(obj,sec);
        ANALYSIS_SIMPLE_END("Deflowgraphing");
        break;
      }
    case CONVERT_DATA_TO_BLOCKS:
      {
        t_object * obj;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        ObjectConvertToDataBlocks(obj);
        break;
      }
    case CONVERT_SECTION_TO_BLOCKS:
      {
        t_object * obj;
        t_section* sec;
        va_start(val,action);
        obj=va_arg(val,t_object *);
        sec=va_arg(val,t_section *);
        VERIFY_FASSERT(obj,("Illegal number of arguments for action CONVERTING SECTION TO BLOCKS"));
        VERIFY_FASSERT(sec,("Illegal number of arguments for action CONVERTING SECTION TO BLOCKS"));
        VERIFY_FASSERT(obj->actions,("No actions for this object!"));
        VERIFY_FASSERT(obj->actions->ConvertToBlocksSec,("Section convertion to blocks not supported for this object!"));

        ANALYSIS_START("Convert to blocks");
        obj->actions->ConvertToBlocksSec(obj,sec);
        ANALYSIS_END("Convertion",("nins"));
        break;
      }
    default:
      FATAL(("Unknown action %d",action));
  }
}
/* }}} */
/* ActionPrint {{{ */

/*! 
 * \param action an t_action_struct * to print
 * \param start_char The column where printing will start
 * This functions prints the supported actions for a given action_struct */ 

void ActionPrint(int start_char, const t_action_struct * action)
{
  if (!action)
  {
    TermPrintClean(stdout,"Not supported!",start_char,40,FINAL);
    return;
  }

  start_char=TermPrintClean(stdout,"Read",start_char,40,0);
  if (action->DisassembleSec)
    start_char=TermPrintClean(stdout," Disassembly",start_char,40,0);
  if (action->FlowgraphSec) 
    start_char=TermPrintClean(stdout," Flowgraph",start_char,40,0);
  if (action->DeFlowgraphSec) 
    start_char=TermPrintClean(stdout," Deflowgraph",start_char,40,0);
  if (action->AssembleSec) 
    start_char=TermPrintClean(stdout," Assemble",start_char,40,0);
  if (action->Write) 
    start_char=TermPrintClean(stdout," Write",start_char,40,0);
  if (action->Print) 
    start_char=TermPrintClean(stdout," Print",start_char,40,0);
  printf("\n");
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker: */
