/***********************************************************
 * Copyright 2001,2002,2003: {{{
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * }}}
 **********************************************************/

/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup BBLS Basic blocks */
/*@{ */
#ifndef DIABLO_BBL_TYPEDEFS
#define DIABLO_BBL_TYPEDEFS
typedef struct _t_bbl t_bbl;
#endif

#ifndef UD_CHAIN_TYPEDEFS
#define UD_CHAIN_TYPEDEFS
typedef struct _t_ud_chain t_ud_chain;
typedef struct _t_ud_chain_list t_ud_chain_list;
#endif

#ifndef DIABLO_BBL_H
#define DIABLO_BBL_H

#include "../datastructures/data_graph.h"
#include "../datastructures/data_ll.h"
#include "diablo_function.h"
#include "diablo_flowgraph.h"
#include "diablo_ins.h"
#include "diablo_procstate.h"
#include "diablo_udchain.h"
#include "diablo_relocatable.h"
#include "diablo_loop.h"
#include "diablo_extendedcopy.h"
#include "diablo_dominator.h"

/*! \defgroup BBL_ATTR  Basic block attributes */
/*@{ */
/*! If set, this block must be placed at the base_address specified */
#define BBL_FIXED_ADDR	0x1
/*! If set, this block has been given an definitive address during placing */
#define BBL_EMITTED	0x2
/*! If set, this block contains an address producer */
#define BBL_ADDR_PROD	0x4	
#define BBL_ENLARGED    0x8
/*! If set, this block is a data block */ 
#define BBL_IS_DATA     0x10
#define BBL_IS_INSTRUMENT_CODE 0x20
#define BBL_IS_FUN_ENTRY 0x40
/*! if set, the block is an unreachable return block (e.g. from exit()) */
#define BBL_UNREACHABLE  0x80
#define BBL_DATA_POOL            0x100
#define BBL_DATA_POOL_MIGRATED   0x200
/*! block is function return block */
#define BBL_IS_EXITBLOCK	0x400
/*! useful for instrumentation (code address translation) */
#define BBL_IS_POSSIBLE_TARGET_OF_INDIRECT_CONTROL_FLOW	0x800
/*! block contains hand-written assembler code */
#define BBL_IS_HANDWRITTEN_ASSEMBLY	0x1000
/*@} */

/*! This structure represents basic blocks in a function. The first member is a
 * relocatable (as we can assign addresses to it). It can also be casted to a
 * node  */

struct _t_bbl 
{
  /*! The first member is a t_relocatable, so each basic block can be casted to a
    generic (graph) node. */
  t_relocatable relocatable;
  /*! Function to which the bbl belongs, is NULL for data blocks and
    (unreachable) blocks that have not been assigned to procedures
    (yet) */
  t_function * fun;
  /*! CFG to which the bbl belongs */
  t_cfg * cfg;
  /*! Next bbl in bbl's fun. (For each function, there is a double
    linked list of all its basic blocks */
  struct _t_bbl * next_in_fun;
  /*! Prev bbl in bbl's fun */
  struct _t_bbl * prev_in_fun;
  /*! Next marked bbl in bbl's fun: for each function, there is a list
    of marked procedures. It is used a stack during constant
    propagation */
  /*  struct _t_bbl * next_marked_in_fun; */
  /*! First instruction in a bbl */
  t_ins * ins_first;
  /*! Last instruction in a bbl */
  t_ins * ins_last;
  /*! Number of instruction in a bbl */
  t_uint32 nins;
  /*! Bitvector to store some \ref BBL_ATTR or properties from basic
    blocks. */
  t_uint32 bbl_attributes;
  /*! Incoming ud-chain during ud_chain analysis */
  t_ud_chain * ud_chain_in;
  /*! Incoming procstate during constant propagation */
  t_procstate * procstate_in; 
  /*! Outgoing procstate during constant propagation */
  /*  t_procstate * procstate_out; */
  /*! Registers live at the end of the block */
  t_regset regs_live_out;
  /*! Registers who's value coming in to the block is comsumed in the block */
  t_regset regs_used;
  /*! Registers defined in the block */
  t_regset regs_defined;

  t_regset regs_defined_in;

  /*! for use with use-def analysis: registers that certainly are defined in the block */
  t_regset regs_def_certain;
  /*! for use with use-def analysis: registers that perhaps are defined in the block */
  t_regset regs_def_perhaps;

  /*  t_regset stack_copies_in; */
  /*  t_regset stack_copies_out; */
  /*! Offset of the stack pointer at entry to this block, compared to
      its value on entry to the procedure */
  /*  t_int32 stack_offset;  */
  /*! During layout, chains of basic blocks are used. A chain is a
    sequence of basic blocks with fixed layout order. They consist of
    double linked lists. So there is a prev_in_chain */
  struct _t_bbl * prev_in_chain; 
  /*! And a next_in_chain*/
  struct _t_bbl * next_in_chain; 
  /*! To easily merge chains, we also hold for each of its blocks a
     reference to the first block in the chain */
  struct _t_bbl * first_in_chain; 
  /*! And to the last block in it. */
  struct _t_bbl * last_in_chain; 
  /*! Position of the block in the layout of the text
      section during code layout */
  int layout_number; 
  /*! A linked list of the bll's that dominate this bbl */
  data_ll_elem * tmp_dominated;
  /*! The bbl that immediately dominates this bbl */
  data_ll_elem * dominated;
  /*! A linked list of the bbl's that post-dominate this bbl */
  data_ll_elem * tmp_p_dominated;
  /*! The bbl that immediately post-dominates this bbl */
  data_ll_elem * p_dominated;
  /*! The bbl that is the loop header of the most inner loop if this bbl is in a loop */
  t_loopref * loops;
  /*  t_reg copies[16];*/
  /*  t_reg copies_rev[16];*/
  /*  t_bool offset_set[16];*/
  /*  t_uint32 offset[16];*/

  t_int32  execution_count;

  void * tmp;
  void * tmp2;
  
#ifdef IA64_SUPPORT  
  t_uint64 profile_info;
#endif  

#ifdef SH_SUPPORT
  t_uint32 switchbound;
#endif
  t_equations eqs_in;
  /*  t_equations eqs_cte;*/

#ifdef I386_SUPPORT
  t_int32 stack_delta;
  t_bool reset_to_framepointer;
  t_bool set_framepointer;
  t_int32 framepointer_offset;
  t_int32 stack_in;
#endif

#ifdef STEGO
  t_bbl * next_equal_chain;
  t_bool leadership_can_be_guessed;
  t_bool incoming_hell_edge_can_be_guessed;
  t_bbl * canonical;
#endif
};


#ifdef IA64_SUPPORT

#define BBL_PROFILE_INFO(bbl) (T_BBL(bbl)->profile_info)

#endif


/* (down)Cast to basic block */

#define T_BBL(x)     ((t_bbl*)(x))

/*! \defgroup BBL_SELECT Macro's to access fields of the basic block structure  */
/*@{*/
#define BBL_NEXT_IN_FUN(bbl)    (T_BBL(bbl)->next_in_fun)
#define BBL_PREV_IN_FUN(bbl)    (T_BBL(bbl)->prev_in_fun)

#define BBL_NEXT(bbl)           (T_BBL(NODE_NEXT(bbl)))
#define BBL_PREV(bbl)           (T_BBL(NODE_PREV(bbl)))

#define BBL_NEXT_IN_CHAIN(bbl)           (T_BBL(bbl)->next_in_chain)
#define BBL_PREV_IN_CHAIN(bbl)           (T_BBL(bbl)->prev_in_chain)
/*#define BBL_NEXT_IN_CHAIN(bbl)           (T_BBL(T_BBL(bbl)->dominated))*/
/*#define BBL_PREV_IN_CHAIN(bbl)           (T_BBL(T_BBL(bbl)->p_dominated))*/

#define BBL_FIRST_IN_CHAIN(bbl)           (T_BBL(T_BBL(bbl)->first_in_chain))
#define BBL_LAST_IN_CHAIN(bbl)           (T_BBL(T_BBL(bbl)->last_in_chain))

#define BBL_INS_FIRST(ibbl)    (T_BBL(ibbl)->ins_first)
#define BBL_INS_LAST(ibbl)     (T_BBL(ibbl)->ins_last)

#define BBL_OLD_ADDRESS(ibbl) (T_BBL(ibbl)->relocatable.old_address)
#define BBL_MIN_ADDRESS(ibbl) (T_BBL(ibbl)->relocatable.min_address)
#define BBL_CADDRESS(ibbl) (T_BBL(ibbl)->relocatable.caddress)
#define BBL_CSIZE(ibbl) (T_BBL(ibbl)->relocatable.csize)
#define BBL_OLD_SIZE(ibbl) (T_BBL(ibbl)->relocatable.old_size)
#define BBL_MIN_SIZE(ibbl) (T_BBL(ibbl)->relocatable.min_size)

#define BBL_FUN(ibbl)          (T_BBL(ibbl)->fun)
#define BBL_CFG(ibbl)          (T_BBL(ibbl)->cfg)
#define BBL_SECTION(ibbl)      (CFG_SECTION(BBL_CFG(ibbl)))
#define BBL_OBJECT(ibbl)       (SECTION_OBJECT(BBL_SECTION(ibbl)))

#define BBL_PROCSTATE_IN(ibbl) (T_PROCSTATE((T_BBL(ibbl)->procstate_in)))

#define BBL_REGS_LIVE_OUT(ibbl)       (((t_bbl *) ibbl)->regs_live_out)
#define BBL_REGS_USED(ibbl)           (((t_bbl *) ibbl)->regs_used)
#define BBL_REGS_DEFINED(ibbl)        (((t_bbl *) ibbl)->regs_defined)
#define BBL_REGS_DEFINED_IN(ibbl)     (((t_bbl *) ibbl)->regs_defined_in)

#define BBL_EXEC_COUNT(ibbl)          (((t_bbl *) ibbl)->execution_count)

#define BBL_REGS_DEF_CERTAIN(ibbl)    (((t_bbl *) ibbl)->regs_def_certain)
#define BBL_REGS_DEF_PERHAPS(ibbl)    (((t_bbl *) ibbl)->regs_def_perhaps)

#define BBL_NEXT_MARKED_IN_FUN(ibbl)  (T_BBL(T_BBL(ibbl)->tmp))
#define BBL_NINS(ibbl)                (T_BBL(ibbl)->nins)

#define BBL_STACK_COPIES_IN(ibbl)     (T_BBL(ibbl)->stack_copies_in)
#define BBL_STACK_COPIES_OUT(ibbl)    (T_BBL(ibbl)->stack_copies_out)

#define BBL_PRED_FIRST(ibbl)  	      (T_CFG_EDGE(NODE_PRED_FIRST(T_NODE(ibbl))))
#define BBL_PRED_LAST(ibbl)  	      (T_CFG_EDGE(NODE_PRED_LAST(T_NODE(ibbl))))

#define BBL_SUCC_FIRST(ibbl)  	      (T_CFG_EDGE(NODE_SUCC_FIRST(T_NODE(ibbl))))
#define BBL_SUCC_LAST(ibbl)  	      (T_CFG_EDGE(NODE_SUCC_LAST(T_NODE(ibbl))))

#define BBL_DOMINATED_BY(ibbl)	(T_BBL(ibbl)->dominated)
#define BBL_TMP_DOMINATED(ibbl)	(T_BBL(ibbl)->tmp_dominated)
#define BBL_PDOMINATED_BY(ibbl)	(T_BBL(ibbl)->p_dominated)
#define BBL_TMP_PDOMINATED(ibbl)	(T_BBL(ibbl)->tmp_p_dominated)

#define BBL_LOOPS(ibbl)	(T_BBL(ibbl)->loops)
#define BBL_EQS_IN(ibbl) 	(T_BBL(ibbl)->eqs_in)
/*#define BBL_EQS_CTE(ibbl)	(T_BBL(ibbl)->tmp2)*/
/*@}*/

#define BblIsMarked(bbl)        NodeIsMarked(bbl)      
#define BblMark(bbl)            NodeMark(bbl)          
#define BblUnmark(bbl)          NodeUnmark(bbl)        

#define BBL_ATTRIB(x)         (T_BBL(x)->bbl_attributes)
#define BblSetAttributes(x,attrib)     (T_BBL(x)->bbl_attributes |= (attrib))
#define BblClearAttributes(x,attrib)   (T_BBL(x)->bbl_attributes &= ~(attrib))
#define BBLTestAttributesOr(x,attrib)  (T_BBL(x)->bbl_attributes & (attrib))
#define BBLTestAttributesAnd(x,attrib) ((T_BBL(x)->bbl_attributes & (attrib)) == (attrib))

#define HAS_FIXED_ADDR(bbl)	(((t_bbl*)(bbl))->bbl_attributes & BBL_FIXED_ADDR)
#define IS_EMITTED(bbl)		(((t_bbl*)(bbl))->bbl_attributes & BBL_EMITTED)
#define HAS_ADDR_PROD(bbl)	(((t_bbl*)(bbl))->bbl_attributes & BBL_ADDR_PROD)
#define IS_ENLARGED(bbl)        (((t_bbl*)(bbl))->flags & BBL_ENLARGED)
#define BBL_PREV_DATA_POOL(bbl)	(((t_bbl*)(bbl))->bbl_attributes & BBL_DATA_POOL)

#define BBL_DESCRIPTOR(bbl)     (BBL_CFG(bbl)->sec->obj->description)

#define IsHellNode(x)	((T_BBL(x) == T_BBL(x)->cfg->hell_node) || (T_BBL(x) == T_BBL(x)->cfg->call_hell_node) || (T_BBL(x) == T_BBL(x)->cfg->exit_hell_node) || (T_BBL(x) == T_BBL(x)->cfg->exit_call_hell_node) || (T_BBL(x) == T_BBL(x)->cfg->unique_entry_node) || (T_BBL(x) == T_BBL(x)->cfg->unique_exit_node) || (T_BBL(x) == T_BBL(x)->cfg->swi_hell_node) || (T_BBL(x) == T_BBL(x)->cfg->exit_swi_hell_node))
/*#define IS_DATABBL(bbl)		((((t_bbl*)(bbl))->ins_first!=NULL)?((t_bbl*)(bbl))->ins_first->Attributes & IF_DATA : 0)*/
#define IS_DATABBL(bbl)		((((t_bbl*)(bbl))->ins_first!=NULL)?INS_TYPE(((t_bbl*)(bbl))->ins_first) == IT_DATA : 0)
#define IS_SWITCH_TABLE(bbl)		((((t_bbl*)(bbl))->ins_first!=NULL)?((t_bbl*)(bbl))->ins_first->Attributes & IF_SWITCHTABLE : 0)

#define BBL_IS_LAST(ibbl)      (FUN_BBL_LAST(BBL_FUN(ibbl))==ibbl)
#define BBL_IS_FIRST(ibbl)     (FUN_BBL_FIRST(BBL_FUN(ibbl))==ibbl)
#define BBL_DATA_POOL_IS_MIGRATED(i_bbl) 	(((t_bbl*)(i_bbl))->bbl_attributes & BBL_DATA_POOL_MIGRATED)

extern t_bbl * hell_node;
extern t_bbl * call_hell_node;
extern t_bbl * swi_hell_node;
extern t_bbl * exit_hell_node;
extern t_bbl * exit_call_hell_node;
extern t_bbl * exit_swi_hell_node;
extern t_bbl * unique_entry_node;
extern t_bbl * unique_exit_node;


/*! \defgroup BBL_ALLOC Constructors and destructors */
/*@{*/
t_bbl * BblNew(t_cfg * graph);
void BblKill(t_bbl * bbl);
/*@} */

void BblSetDefinedRegs(t_bbl * bbl);
t_regset BblRegsDef(t_bbl * bbl);
t_regset BblRegsUse(t_bbl * bbl);
t_regset BblRegsLiveAfter(t_bbl * bbl);
t_regset BblRegsLiveBefore(t_bbl * bbl);
t_regset BblRegsMaybeDef(t_bbl * bbl);

void realBblInsertInFunction(char * file, int lnno, t_bbl * bbl, t_function * func);
#define BblInsertInFunction(bbl,func) realBblInsertInFunction(__FILE__,__LINE__,bbl,func)
void InsAppendToBbl(t_ins * ins, t_bbl * bbl);
void InsPrependToBbl(t_ins * ins, t_bbl * bbl);
void InsInsertAfter(t_ins * ins, t_ins *after);
void InsInsertBefore(t_ins * ins, t_ins *before);
void InsSwap(t_ins * ins_first, t_ins *ins_next);
void BblMoveInstructionAfter(t_ins * move_ins, t_ins * dest_ins);
void BblMoveInstructionBefore(t_ins * move_ins, t_ins * dest_ins);
#endif
/*@} */
/*@} */
void BblSetAddressSuper(t_bbl * bbl,t_address address);
void BblSetAllAddressSuper(t_bbl * bbl,t_address old, t_address new);
t_cfg_edge * BblGetSuccEdgeOfType(t_bbl * bbl, t_uint32 types);
t_cfg_edge * BblGetPredEdgeOfType(t_bbl * bbl, t_uint32 types);
t_bbl * BblSplitBlock(t_bbl * orig_bbl, t_ins * where, t_bool before);
/* vim: set shiftwidth=2 foldmethod=marker: */
