/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/
#include <string.h>
#include "diablo_extendedcopy.h"
#include "diablo_copy_analysis.h"
#ifdef ARM_SUPPORT
#include "../arch/arm/arm_registers.h"
#endif
#include "diablo_print.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include "../diablo_options.h"
#ifdef I386_SUPPORT
#include "../arch/i386/i386_registers.h"
#endif

/*#define DEBUG_COPY_ANALYSIS  |+ Uncomment for debugging purposes +|*/
/*#define VERBOSE_COPY_ANALYSIS |+ Uncomment for a whole bunch of write-out's +|*/
const t_equation BotEquation = {REG_BOT,REG_BOT,0,NULL,NULL};
const t_equation TopEquation = {REG_TOP,REG_TOP,0,NULL,NULL};
t_equation EquationSum(t_equation eq1, t_equation eq2);
t_equation EquationSub(t_equation eq1, t_equation eq2);
t_bool EquationsDiffer(t_equations eqs1, t_equations eqs2);

/* Allocate equations and set all but the last to TOP. The last equation is unused,
 * but must always be set to BOT!
 */
t_equations EquationsNew(void)
{
  int i;
  t_equations ret = Malloc((max_nr_equations)*sizeof(t_equation));
  for(i=0;i<max_nr_equations-1;i++)
    ret[i] = TopEquation;
  ret[max_nr_equations-1] = BotEquation;
  return ret;
}

/* Set all equations to BOT */
void EquationsSetAllBot(t_equations eqs)
{
  int i;
  for(i=0;i<=max_nr_equations-1;i++)
    eqs[i] = BotEquation;
}

/* Set all equations to TOP */
void EquationsSetAllTop(t_equations eqs)
{
  int i;
  for(i=0;i<max_nr_equations-1;i++)
    eqs[i] = TopEquation;
}

/* Join two sets of equations, eqs1 and eqs2, into eqs1. Returns TRUE if eqs1 differs
 * before and after the join. If eqs1 is still TOP, then we copy eqs2 into eqs1. */
t_bool EquationsJoin(t_equations eqs1, t_equations eqs2)
{
  t_equation join[max_nr_equations];
  t_equation tmp;
  t_reg i; 
  
  /* {{{ */
#ifdef DEBUG_COPY_ANALYSIS  
  if(eqs1[max_nr_equations-1].rega != REG_BOT) FATAL(("Not bot anymore!"));
  if(eqs2[max_nr_equations-1].rega != REG_BOT) FATAL(("Not bot anymore!"));
  for(i=0;i<max_nr_equations-2;i++)
  {
    if(eqs1[i].rega != REG_TOP && eqs1[i+1].rega == REG_TOP)
    {
      printf("eqs1:\n"); EquationsPrint(eqs1);
      printf("eqs2:\n"); EquationsPrint(eqs2);
      FATAL(("Troubles here!"));
    }
    if(eqs2[i].rega != REG_TOP && eqs2[i+1].rega == REG_TOP)
    {
      printf("eqs1:\n"); EquationsPrint(eqs1);
      printf("eqs2:\n"); EquationsPrint(eqs2);
      FATAL(("Troubles here!"));
    }
  }
#endif
  
#ifdef VERBOSE_COPY_ANALYSIS  
  DiabloPrint(stdout,"Joining to:\n");
  //EquationsPrint(eqs1);
  //DiabloPrint(stdout,"and:\n");
  //EquationsPrint(eqs2);
  //DiabloPrint(stdout,"Into:\n");
#endif
  /* }}} */

  if(eqs1[0].rega == REG_TOP)
  /* We haven't visited these equations yet, just copy eqs2 into eqs1 */
  {
    if(eqs2[0].rega == REG_TOP) return FALSE;

    EquationsCopy(eqs2,eqs1);
#ifdef VERBOSE_COPY_ANALYSIS  
    EquationsPrint(eqs1);
#endif
    return TRUE;
  }
  
  /* Real join!
   * First take a copy of the original eqs1 so we can detect changes at the end of the join-operation
   */
  EquationsCopy(eqs1,join);
  EquationsSetAllBot(eqs1);
  
  for(i=0;i<max_nr_equations-1;i++)
  {
    if(eqs2[i].rega == REG_BOT) continue; /* No information available, skip! */
    if(eqs2[i].rega == REG_TOP)
    {
      WARNING(("Still to top!"));
      return FALSE;
    }
      
    tmp = eqs2[i];
    while(1)
    {
      if(EquationsRegsEqual(join,tmp.rega,tmp.regb,0-tmp.constant,tmp.taga,tmp.tagb) == TRUE)
      /* if this equation also holds in the original eqs1, it is common and will appear in the join */
	EquationInsert(eqs1,tmp.rega,tmp.regb,tmp.constant,tmp.taga,tmp.tagb);
      
      if(eqs2[tmp.regb].rega != REG_BOT)
#ifdef DEBUG_COPY_ANALYSIS
        if(eqs2[tmp.regb].rega == REG_TOP)
	  FATAL(("Still to top!"));
	else
#else
        if(eqs2[tmp.regb].rega != REG_TOP)
#endif
      /* we can produce another equation by summing two equations, lets investigate these */
	{
	  tmp = EquationSum(tmp,eqs2[tmp.regb]);
	}
	else break; 
#ifndef DEBUG_COPY_ANALYSIS
      else break; 
#endif
    }
  }
#ifdef VERBOSE_COPY_ANALYSIS  
  EquationsPrint(eqs1);
#endif
  for(i=0;i<max_nr_equations-2;i++)
  {
    if(eqs1[i].rega != REG_TOP && eqs1[i+1].rega == REG_TOP)
      FATAL(("Troubles here!"));
    if(eqs2[i].rega != REG_TOP && eqs2[i+1].rega == REG_TOP)
      FATAL(("Troubles here!"));
  }
  if(EquationsDiffer(join,eqs1)) return TRUE;
  return FALSE;
}

/* Answers the question 'reg1 + taga ?= reg2 + tagb + constant'. We return PERHAPS when eqs == NULL,
 * because in this case there is just no information available which is conservative
 */
t_tristate EquationsRegsEqual(t_equations eqs, t_reg reg1, t_reg reg2, t_int32 constant, t_reloc * taga, t_reloc * tagb)
{
  int i,j,cte = 0;
  t_reloc * tagi, * tagj;
  t_equation tmp;

  if(!eqs) return PERHAPS;

#ifdef DEBUG_COPY_ANALYSIS
  if(eqs[max_nr_equations-1].rega != REG_BOT)
    FATAL(("Not bot anymore!"));
#endif

  if(reg1<reg2)
  { i = reg1; j = reg2; cte -= constant; tagi = taga; tagj = tagb;}
  else
  { i = reg2; j = reg1; cte += constant; tagi = tagb; tagj = taga;}
  tmp = eqs[i];
  if(tmp.rega == REG_BOT) return PERHAPS;
  if(tmp.rega == REG_TOP) return PERHAPS;
  
  while(1)
  {
    if(tmp.regb == j)
    {
      if(tmp.constant == cte && tmp.taga == tagi && tmp.tagb == tagj) return YES;
      else return NO;
    }
    else if(eqs[tmp.regb].rega != REG_BOT
#ifdef DEBUG_COPY_ANALYSIS
	&& eqs[tmp.regb].rega != REG_TOP
#endif
	)
    {
      tmp = EquationSum(tmp,eqs[tmp.regb]);
    }
    else return PERHAPS;
  }
}

/* Returns YES when there is an integer offset between the two registers and this offset is
 * filled in in the t_int32* parameter, returns NO
 * when there is a known difference between the registers but it is a difference in tags. Returns
 * PERHAPS when nothing is known about the registers and we have to be conservative */
t_tristate EquationsRegsDiffer(t_equations eqs, t_reg reg1, t_reg reg2, t_int32 * constant)
{
  t_int32 i,j;
  t_equation tmp;
  
  if(reg1<reg2)
  { i = reg1; j = reg2; }
  else
  { i = reg2; j = reg1; }
  tmp = eqs[i];
  if(tmp.rega == REG_BOT) return PERHAPS;
  if(tmp.rega == REG_TOP) return PERHAPS;
  
  while(1)
  {
    if(tmp.regb == j)
    {
      if(!tmp.taga && !tmp.tagb)
      {
	*constant = tmp.constant;
	return YES;
      }
      return NO;
    }
    else if(eqs[tmp.regb].rega != REG_BOT
#ifdef DEBUG_COPY_ANALYSIS
	&& eqs[tmp.regb].rega != REG_TOP
#endif
	)
    {
      tmp = EquationSum(tmp,eqs[tmp.regb]);
    }
    else return PERHAPS;
  }
  return PERHAPS;
}

/* Returns YES when there is a tag offset between the two registers, the tag is returned in reloc.
 * Returns NO when there is a known difference between the registers but it is not a difference in tags. Returns
 * PERHAPS when nothing is known about the registers and we have to be conservative */
t_tristate EquationsRegsDifferByTag(t_equations eqs, t_reg reg1, t_reg reg2, t_reloc** reloc)
{
  t_int32 i,j;
  t_equation tmp;
  
  if(reg1<reg2)
  { i = reg1; j = reg2; }
  else
  { i = reg2; j = reg1; }
  tmp = eqs[i];
  if(tmp.rega == REG_BOT) return PERHAPS;
  if(tmp.rega == REG_TOP) return PERHAPS;
  
  while(1)
  {
    if(tmp.regb == j)
    {
      if(tmp.taga)
      {
	*reloc = tmp.taga;
	return YES;
      }
      if(tmp.tagb)
      {
	*reloc = tmp.tagb;
	return YES;
      }
      return NO;
    }
    else if(eqs[tmp.regb].rega != REG_BOT
#ifdef DEBUG_COPY_ANALYSIS
	&& eqs[tmp.regb].rega != REG_TOP
#endif
	)
    {
      tmp = EquationSum(tmp,eqs[tmp.regb]);
    }
    else return PERHAPS;
  }
  return PERHAPS;
}


/* Add the equation "defined - used + constant - taga + tagb = 0" to the set of equations, hereby
 * invalidating some of the existing equations. */
void EquationsAdd(t_equations eqs, t_reg defined, t_reg used, t_int32 constant, t_reloc* taga, t_reloc* tagb)
{
  int i;
  if(defined != used)
  {
    EquationsInvalidate(eqs,defined);
    EquationInsert(eqs,defined,used,constant,taga,tagb);
  }
  else
  {
#ifdef DEBUG_COPY_ANALYSIS
    if(taga != NULL || tagb != NULL) FATAL(("Tags should be NULL!"));
#endif
    for(i=0; i<defined; i++)
    {
      if(eqs[i].regb == defined) eqs[i].constant -= constant;
    }
    if(eqs[defined].rega != REG_BOT) eqs[defined].constant += constant;
  }
  return;
}

void EquationInsert(t_equations eqs, int reg1, int reg2, int constant, t_reloc* taga, t_reloc* tagb)
{
  t_equation tmp, old;
  t_reloc * reloc;
  int i;
  /* Sort the registerpair */
  if(reg1 < reg2)
  {tmp.rega = reg1; tmp.regb = reg2;tmp.constant = constant; tmp.taga = taga; tmp.tagb = tagb;}
  else
  {tmp.rega = reg2; tmp.regb = reg1;tmp.constant = 0-constant; tmp.taga = tagb; tmp.tagb = taga;}

#ifdef VERBOSE_COPY_ANALYSIS
  printf("Inserting ");  EquationPrint(&tmp);  printf("\n");
#endif

  if((taga && EquationsRegsDifferByTag(eqs,reg2,null_register,&reloc)==YES) || (tagb && EquationsRegsDifferByTag(eqs,reg1,null_register,&reloc)==YES) )
    /* Don't insert, we cannot handle the cases where relocs get added or substracted */
    return;


  /* See if the highest numbered register pair isn't in an equation already */
  for(i=0;i < tmp.regb; i++)
    if(eqs[i].regb == tmp.regb)
    {
      /* Substract the equations, we don't lose any information, but we will insert them
       * in an ordered way */
      tmp = EquationSub(tmp,eqs[i]);

      if(tmp.rega != tmp.regb)
	/* Info is not redundant, so insert */
        EquationInsert(eqs,tmp.rega,tmp.regb,tmp.constant,tmp.taga,tmp.tagb);
      else if (tmp.constant != 0) FATAL(("Conflict when inserting an equation"));
      return;
    }

  /* If there is already info available over the smallest numbered reg, see if the highest numbered
   * reg is lower than regb in the existing info */
  if(eqs[tmp.rega].rega != REG_BOT && eqs[tmp.rega].rega != REG_TOP)
  {
    if(eqs[tmp.rega].regb > tmp.regb)
    {
    /* Insert the new info, and insert the substraction of the new and old info */
      old = eqs[tmp.rega];
      eqs[tmp.rega] = tmp;
      tmp = EquationSub(old,tmp);
      EquationInsert(eqs,tmp.rega,tmp.regb,tmp.constant,tmp.taga,tmp.tagb);
      return;
    }
    else if (eqs[tmp.rega].regb < tmp.regb)
    {
    /* Leave the old info and insert the substraction of the new and old info */
      tmp = EquationSub(tmp,eqs[tmp.rega]);
      EquationInsert(eqs,tmp.rega,tmp.regb,tmp.constant,tmp.taga,tmp.tagb);
      return;
    }
  }
  /* There was no info available yet, just insert it */
  eqs[tmp.rega] = tmp;

  return;
}

void EquationsInvalidate(t_equations equations, t_reg reg1)
{
  t_uint32 i,j = REG_BOT,k = REG_BOT;
  t_equation tmp;

  /* skip non-int registers */
  if (reg1 >= max_nr_equations-1) return;

  for(i = 0; i <= reg1/*max_nr_equations-1*/; i++)
  {
    if(equations[i].rega == reg1)
      j = i;
    if(equations[i].regb == reg1)
      k = i;
  }
  if(j != REG_BOT && k != REG_BOT)
  /* Invalidate both equation, but first insert the sum of both, so we don't lose that information */
  {
    tmp = EquationSum(equations[j], equations[k]);
    equations[j] = BotEquation;
    equations[k] = BotEquation;
    EquationInsert(equations,tmp.rega,tmp.regb,tmp.constant,tmp.taga,tmp.tagb);
    return;
  }
  if(j != REG_BOT)
    equations[j] = BotEquation;
  if(k != REG_BOT)
    equations[k] = BotEquation;
  return;
}

t_equation EquationSum(t_equation eq1, t_equation eq2)
{
  t_equation ret;
  if(eq1.rega == eq2.regb)
  {
    ret.rega = eq2.rega;
    ret.regb = eq1.regb;
    ret.constant = eq1.constant + eq2.constant;
  }
  else if (eq1.regb == eq2.rega)
  {
    ret.rega = eq1.rega;
    ret.regb = eq2.regb;
    ret.constant = eq1.constant + eq2.constant;
  }
  else
  {
    EquationPrint(&eq1);printf("\n");
    EquationPrint(&eq2);printf("\n");
    FATAL(("Must have equal register pair for addition\n"));
  }

  if(eq1.taga == eq2.tagb)
  {
    if(eq1.tagb == eq2.taga)
    {
      ret.taga = NULL;
      ret.tagb = NULL;
    }
    else
    {
      ret.taga = eq2.taga;
      ret.tagb = eq1.tagb;
    }
  }
  else if(eq1.tagb == eq2.taga)
  {
    ret.taga = eq1.taga;
    ret.tagb = eq2.tagb;
  }
  else if(!eq2.taga && !eq2.tagb)
  {
    ret.taga = eq1.taga;
    ret.tagb = eq1.tagb;
  }
  else if(!eq1.taga && !eq1.tagb)
  {
    ret.taga = eq2.taga;
    ret.tagb = eq2.tagb;
  }
  else
  {
    FATAL(("Tag troubles in EquationSum"));
  }

  if(ret.taga && ret.taga == ret.tagb)
  {
    ret.taga = NULL;
    ret.tagb = NULL;
  }
  return ret;
}

t_equation EquationSub(t_equation eq1, t_equation eq2)
{
  t_equation ret;
  if(eq1.rega == eq2.rega)
  {
    ret.rega = eq2.regb;
    ret.regb = eq1.regb;
    ret.constant = eq1.constant - eq2.constant;
  }
  else if (eq1.regb == eq2.regb)
  {
    ret.rega = eq1.rega;
    ret.regb = eq2.rega;
    ret.constant = eq1.constant - eq2.constant;
  }
  else
  {
    EquationPrint(&eq1);EquationPrint(&eq2);
    FATAL(("Can only sub when regA or regB is the same in both equations\n"));
  }

  if(eq1.taga == eq2.taga)
  {
    ret.taga = eq2.tagb;
    ret.tagb = eq1.tagb;
  }
  else if(eq1.tagb == eq2.tagb)
  {
    ret.taga = eq1.taga;
    ret.tagb = eq2.taga;
  }
  else if(!eq2.taga && !eq2.tagb)
  {
    ret.taga = eq1.taga;
    ret.tagb = eq1.tagb;
  }
  else if(!eq1.taga && !eq1.tagb)
  {
    ret.taga = eq2.tagb;
    ret.tagb = eq2.taga;
  }
  else
  {
    FATAL(("Tag troubles in EquationSub"));
  }
  
  if(ret.taga && ret.taga == ret.tagb)
  {
    ret.taga = NULL;
    ret.tagb = NULL;
  }
  return ret;
}


void EquationsPrint(t_equations eqs)
{
  t_uint32 i;
#ifdef DEBUG_COPY_ANALYSIS
  t_equation tmp;
#endif
  /*if(eqs[0].rega == REG_TOP)
  {
    DiabloPrint(stdout,"REG_TOP\n");
    return;
  }*/
  for(i = 0; i < max_nr_equations-2; i++)
  {
    if (eqs[i].rega != REG_BOT && eqs[i].rega != REG_TOP)
    {
      EquationPrint(eqs+i);
#ifdef DEBUG_COPY_ANALYSIS
      tmp = eqs[i];
      while(eqs[tmp.regb].rega != REG_BOT && eqs[tmp.regb].rega != REG_TOP)
      {
        fprintf(stdout," ; ");
        fflush(stdout);
        tmp = EquationSum(tmp,eqs[tmp.regb]);
	EquationPrint(&tmp);
        fflush(stdout);
      }
#endif
      fprintf(stdout,"\n");
    }
    else if (eqs[i].rega == REG_TOP)
      VERBOSE(0,("%d : REG_TOP",i));
  }
  if(eqs[max_nr_equations-2].rega != REG_BOT && eqs[max_nr_equations-2].rega != REG_TOP)
  {
    EquationPrint(&eqs[max_nr_equations-2]);
    DiabloPrint(stdout,"\n");
  }
  else if (eqs[max_nr_equations-2].rega == REG_TOP)
    VERBOSE(0,("%d : REG_TOP",max_nr_equations-2));
}

void EquationPrint(t_equation* eq)
{
  if(eq->rega == REG_TOP)
  {
    fprintf(stdout,"top\n");
    return;
  }
  if(eq->rega == REG_BOT)
  {
    fprintf(stdout,"bot\n");
    return;
  }
#ifdef ARM_SUPPORT
  if(eq->constant <= 0)
  {
    fprintf(stdout,"%s = %s + %d + (%p - %p)",eq->rega==null_register?"Nullreg":ArmRegisterName(eq->rega),eq->regb==null_register?"Nullreg":ArmRegisterName(eq->regb),0-eq->constant, eq->taga, eq->tagb);
  }
  else
  {
    fprintf(stdout,"%s = %s + %d + (%p - %p)",eq->regb==null_register?"Nullreg":ArmRegisterName(eq->regb),eq->rega==null_register?"Nullreg":ArmRegisterName(eq->rega),eq->constant, eq->tagb, eq->taga);
  }
#elif defined I386_SUPPORT
  if(eq->constant <= 0)
  {
    fprintf(stdout,"%s = %s + %d + (%p - %p)",eq->rega==null_register?"Nullreg":I386RegisterName(eq->rega),eq->regb==null_register?"Nullreg":I386RegisterName(eq->regb),0-eq->constant, eq->taga, eq->tagb);
  }
  else
  {
    fprintf(stdout,"%s = %s + %d + (%p - %p)",eq->regb==null_register?"Nullreg":I386RegisterName(eq->regb),eq->rega==null_register?"Nullreg":I386RegisterName(eq->rega),eq->constant, eq->tagb, eq->taga);
  }
#else
  fprintf(stdout,"!!! This is not implemented for anything other than the ARM port...\n");
#endif
}

void EquationsCopy(t_equations src, t_equations dest)
{
  if(!src || !dest) FATAL(("Null equations!"));
  memcpy(dest,src,(max_nr_equations)*sizeof(t_equation));
  /*for(i=0;i<max_nr_equations;i++)
    dest[i] = src[i];
  return;*/
}

t_bool EquationsDiffer(t_equations eqs1, t_equations eqs2)
{
  if(!eqs1 || !eqs2)
  {
    FATAL(("Null equations!"));
    return TRUE;
  }
#ifdef DEBUG_COPY_ANALYSIS  
  if(eqs1[max_nr_equations-1].rega != REG_BOT) FATAL(("Not bot anymore!"));
  if(eqs2[max_nr_equations-1].rega != REG_BOT) FATAL(("Not bot anymore!"));
#endif
  if(memcmp(eqs1,eqs2,(max_nr_equations)*sizeof(t_equation))) return TRUE;
  return FALSE;
}

#if 0 /*for testing*/
int main(void)
{
  int i;
  t_equation eqs[max_nr_equations-1];
  t_equation eq1[max_nr_equations-1];
  t_equation eq2[max_nr_equations-1];
  EquationsResetAll(eqs);
  EquationsResetAll(eq1);
  EquationsResetAll(eq2);

  EquationsAdd(eq1, 1, 2, 0);
  EquationsAdd(eq1, 2, 3, 0);
  EquationsAdd(eq1, 3, 4, 0);
  EquationsAdd(eq2, 1, 3, 0);
  EquationsAdd(eq2, 3, 4, 0);
  EquationsAdd(eqs, 4, 5, -3);
  /*EquationsAdd(eqs, 1, 6, 0);
  EquationsAdd(eqs, 7, 6, 0);
  EquationsAdd(eqs, 4, 2, 0);
  EquationsAdd(eqs, 1, 2, 0);*/
  EquationsPrint(eq1);
  /*EquationsAdd(eqs, 1, 4, 0);*/
  EquationsPrint(eq2);
  /*EquationsAdd(eqs, 7, 2, 0);
  EquationsAdd(eqs, 1, 2, 2);
  EquationsPrint(eqs);*/
  EquationsJoin(eq1,eq2);
  EquationsPrint(eq1);
  if (EquationsRegsEqual(eq1, 2, 1, 0)) printf("2 and 1 equal\n");
  if (EquationsRegsEqual(eq1, 4, 1, 0)) printf("1 and 4 equal\n");
}
#endif
/* vim: set shiftwidth=2 foldmethod=marker: */
