/******************************************************************************
 * Copyright 2001,2002: {{{
 *
 * Bruno De Bus 
 * Bjorn De Sutter 
 * Dominique Chanet 
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This file contains all instruction-related functions for libdiablo. It
 * should not contain architecture specific operations, but it needs to now the
 * size of an instruction struct for this arch! 
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/

/* Includes {{{ */
#include <string.h> /* For memcopy */
#include "../config.h"
#include "diablo_ins.h"
#include "diablo_object.h"
#include "diablo_reloc.h"
#include "../utils/util_error.h"
#include "../utils/util_malloc.h"
#include "diablo_print.h"
#include "diablo_iterators.h"
#ifdef MIPS_SUPPORT
#include "../arch/mips/mips_instructions.h"
#include "../arch/mips/mips_opcodes.h"
#else
#ifdef ARM_SUPPORT
#include "../arch/arm/arm_opcodes.h"
#include "../arch/arm/arm_registers.h"
#endif
#endif
/*}}}*/
/* RealGetInsByNumber {{{ */
/**  
 * \param filename The calling-file-name
 * \param lnno     The calling-line-number
 * \param obj      The object of the instruction
 * \param section  The section of the instruction
 * \param ins_num  the number of the instruction 
 * \return a void * (= a void-casted struct ins)
 * 
 * This function returns instruction ins_num in
 * a section in an object. The behaviour is different
 * for different types of sections: When we are
 * disassembling we can are indexing the
 * to_disassemble_buffer, else we are indexing a
 * buffer with all the instructions in it.*/

void * RealGetInsByNumber(char * filename, int lnno, t_section * section, t_uint32 ins_num)    
{
  t_object * obj=section->obj;
  if ((section->type!=DISASSEMBLING_CODE_SECTION)  /* Instruction is in the tmp buffer */
      &&(section->type!=DISASSEMBLED_CODE_SECTION) /* instruction is in the data buffer */
      &&(section->type!=FLOWGRAPHING_CODE_SECTION) /* instruction is in the data buffer */
      &&(section->type!=ASSEMBLING_CODE_SECTION))  /* instruction is in the data buffer */
    FATAL(("Instruction number not supported in fase %c",section->type));

  if (ins_num>=SECTION_NINS(section)) { 
    printf("Num %d nins %d\n",ins_num,SECTION_NINS(section));
    FATAL(("Instruction not yet allocated at line %d in file %s!",lnno,filename)); 
    return NULL;
  }
  return ((section->type==DISASSEMBLING_CODE_SECTION)?(((char *) SECTION_TMP_BUF(section)) + (ins_num * obj->description->decoded_instruction_size)):(((char *) section->data) + (ins_num * obj->description->decoded_instruction_size)));
}
/*}}}*/
/* InsNewForSec {{{ */
/** 
 * \param sec The section for which we create an instruction
 * \return A new (void -casted) instruction
 * 
 * InsNewForSec allocates a new instruction (for this section) and returns it. 
 * \todo Add buffer overrun check 
 * \todo If we can garantee that no pointers are created we could reallocate the buffers used for storing instructions (much more efficient)
 */

void * InsNewForSec(t_section * sec)
{
  t_object * obj=sec->obj;

  if ((sec->type!=DISASSEMBLING_CODE_SECTION)&&(sec->type!=FLOWGRAPHING_CODE_SECTION)&&(sec->type!=DEFLOWGRAPHING_CODE_SECTION)&&(sec->type!=FLOWGRAPHED_CODE_SECTION))
  {
    FATAL(("Cannot allocate instructions in fase %c",sec->type));
  }

  if ((sec->type==DISASSEMBLING_CODE_SECTION)||(sec->type==DISASSEMBLED_CODE_SECTION))
  {
    t_uint32 nins=++(SECTION_NINS(sec));
    if (nins>1)
    {
      INS_PREV(AINS(obj,sec,nins-1))=AINS(obj,sec,nins-2);
      INS_NEXT(AINS(obj,sec,nins-2))=AINS(obj,sec,nins-1);
      INS_NEXT(AINS(obj,sec,nins-1))=NULL;
    }
    else
    {
      INS_PREV(AINS(obj,sec,nins-1))=NULL;
      INS_NEXT(AINS(obj,sec,nins-1))=NULL;
    }

    INS_SECTION(T_INS(AINS(obj,sec,nins-1)))=sec;

    RELOCATABLE_REFERS_TO(AINS(obj,sec,nins-1))=NULL;
    RELOCATABLE_REFED_BY(AINS(obj,sec,nins-1))=NULL;
    RELOCATABLE_TYPE(AINS(obj,sec,nins-1))=RT_INS;
    INS_COPY(AINS(obj,sec,nins-1))=NULL;
    return AINS(obj,sec,nins-1);
  }
  else if (sec->type==FLOWGRAPHING_CODE_SECTION || sec->type==DEFLOWGRAPHING_CODE_SECTION || sec->type==FLOWGRAPHED_CODE_SECTION )
  {
    t_ins * ret=Calloc(1,obj->description->decoded_instruction_size);
#ifdef BIT32ADDRSUPPORT
    GenericSet_uint32(INS_CSIZE(ret), obj->description->minimal_encoded_instruction_size/8);
#else
    GenericSet_uint64(INS_CSIZE(ret), obj->description->minimal_encoded_instruction_size/8);
#endif    
    RELOCATABLE_TYPE(ret) = RT_INS;
    INS_SECTION(T_INS(ret))=sec;
    return ret;
  }
  return NULL; /* keep the compiler happy */
}
/*}}}*/
/* InsNewForCfg {{{ */
/** 
 * \param cfg The cfg for which we create an instruction
 * \return A new (void -casted) instruction
 * 
 * InsNewForCfg allocates a new instruction (in the cfg) and returns it. 
 */

void * InsNewForCfg(t_cfg * cfg)
{
  t_section * sec=CFG_SECTION(cfg);
  t_object * obj=sec->obj;
  t_ins * ret=Calloc(1,obj->description->decoded_instruction_size);
  INS_SECTION(ret)=sec;
#ifdef BIT32ADDRSUPPORT  
  GenericSet_uint32(INS_CSIZE(ret), obj->description->minimal_encoded_instruction_size/8);
  GenericSet_uint32(INS_MIN_SIZE(ret), obj->description->minimal_encoded_instruction_size/8);
#else
  GenericSet_uint64(INS_CSIZE(ret), obj->description->minimal_encoded_instruction_size/8);
  GenericSet_uint64(INS_MIN_SIZE(ret), obj->description->minimal_encoded_instruction_size/8);
#endif  
  RELOCATABLE_TYPE(ret) = RT_INS;
  return ret;
}
/*}}}*/
/* InsNewForBbl {{{ */
/** 
 * \param bbl The bbl for which we create an instruction
 * \return A new (void -casted) instruction
 * 
 * InsNewForBbl allocates a new instruction (in the cfg of the basic block) and returns it. 
 */

void * InsNewForBbl(t_bbl * bbl)
{
  static int nr_allocated_ins =0;
  t_section * sec=BBL_SECTION(bbl);
  t_object * obj=sec->obj;
  t_ins * ret=Calloc(1,obj->description->decoded_instruction_size);

  if ((nr_allocated_ins++%100000)==0)
    printf("ins allocated %d\n",nr_allocated_ins);

#if 0
  if (nr_allocated_ins>=400000) 
    {
      FATAL(("TOO MUCH INS\n"));
    }
#endif

  INS_SECTION(ret)=sec;
  INS_BBL(ret)=bbl;
#ifdef BIT32ADDRSUPPORT
  GenericSet_uint32(INS_CSIZE(ret), obj->description->minimal_encoded_instruction_size/8);
#else
  GenericSet_uint64(INS_CSIZE(ret), obj->description->minimal_encoded_instruction_size/8);
#endif  
  RELOCATABLE_TYPE(ret) = RT_INS;
  return ret;
}
/*}}}*/

void InsFreeAllReferedRelocsBase(t_ins * ins)
{
  while (RELOCATABLE_REFERS_TO(ins)) 
  {
    t_reloc_ref * tmp=RELOCATABLE_REFERS_TO(ins)->next;
    Free(RELOCATABLE_REFERS_TO(ins));
    RELOCATABLE_REFERS_TO(ins)=tmp;
  }
  while (RELOCATABLE_REFED_BY(ins)) 
  {
    t_reloc_ref * tmp=RELOCATABLE_REFED_BY(ins)->next;
    Free(RELOCATABLE_REFED_BY(ins));
    RELOCATABLE_REFED_BY(ins)=tmp;
  }
}

void InsFreeReferedRelocs(t_ins * ins)
{
  t_reloc_ref * iter=RELOCATABLE_REFERS_TO(ins);
  t_reloc_ref * itern;

  while (iter) 
  {
    itern=iter->next;
    if (RELOCATABLE_TYPE(REL_FROM(RELOCATABLE_REFERS_TO(ins)->rel))==RT_INS)
      RelocTableRemoveReloc(INS_OBJECT(ins)->reloc_table,RELOCATABLE_REFERS_TO(ins)->rel);
    else
      FATAL(("Illegal relocation for instruction!"));
    iter=itern;
  }
}

/* InsFree {{{ */
/** 
 * \param ins The instruction to free 
 *
 * \todo Needs a callback to free arch-specific allocated data
 *
 * InsFree frees the memory of an instruction.
 * It has some side effects: If the instruction refers something else in the 
 * program (with a relocation) then by killing this instruction we remove the relocation  
 */
void InsFree(t_ins * ins)
{
    if(INS_OBJECT(ins)->actions->InsCleanup) INS_OBJECT(ins)->actions->InsCleanup(ins);
    InsFreeReferedRelocs(ins);

    if (RELOCATABLE_REFED_BY(ins)) 
    {
      FATAL(("Removing referenced instruction!"));
    }
    Free(ins);
  }
  /*}}}*/
  /* InsKill {{{ */
  /** 
   * \param ins The instruction to kill 
   * 
   * InsKill unlinks the instruction and then calls InsFree
   */

  void InsKill(t_ins * ins)
  {

    if (INS_NEXT(ins))
    {
      INS_PREV(INS_NEXT(ins))=INS_PREV(ins);
    }
    else
    {
      BBL_INS_LAST(INS_BBL(ins))=INS_PREV(ins);
    }
    if (INS_PREV(ins))
    {
      INS_NEXT(INS_PREV(ins))=INS_NEXT(ins);
    }
    else
    {
      BBL_INS_FIRST(INS_BBL(ins))=INS_NEXT(ins);
    }

    INS_BBL(ins)->nins--;


    /* if the instruction is part of a ud-graph, it should be removed from this graph.
     * notice that the ud graph won't be correct afterwards, but this is impossible
     * because we haven't got enough information in the ud graph to correct it. */

    if (InsInUdGraph(ins))
    {
      if (BBL_CFG(INS_BBL(ins)))
      {
	t_ud_graph * udg;
	t_ud_edge * i_edge, *tmp;

	if ((udg = CFG_UDGRAPH(BBL_CFG(INS_BBL(ins)))))
	{
	  /* remove all predecessor edges */
	  NODE_FOREACH_PRED_EDGE_SAFE(ins,i_edge,tmp)
	    GraphRemoveEdge(udg,i_edge);
	  /* remove all successor edges */
	  NODE_FOREACH_SUCC_EDGE_SAFE(ins,i_edge,tmp)
	    GraphRemoveEdge(udg,i_edge);
	  /* Remove the instruction from the graph */
	  GraphRemoveNodeNoFree(udg,T_NODE(ins));
	}
      }
    }
    InsFree(ins);
  }
  /*}}}*/
  /* InsDup {{{ */
  /*!
   * \brief Duplicate an instruction
   *
   * \param ins The instruction to duplicate
   *
   * \return t_ins * A duplicated instruction
   */

  t_ins * InsDup(t_ins * ins)
  {
    t_ins * ret;
    t_section * sec;
    t_object * obj;
    t_reloc_ref * refs=RELOCATABLE_REFERS_TO(ins);

    sec=InsGetSection(ins);
    obj=sec->obj;

    if (sec->type!=FLOWGRAPHING_CODE_SECTION && sec->type!=FLOWGRAPHED_CODE_SECTION)
    {
      FATAL(("Cannot duplicate instructions in fase %c",sec->type));
    }

    ret=InsNewForSec(sec);
    memcpy(ret,ins,obj->description->decoded_instruction_size);
    INS_OLD_ADDRESS(ret)= NULL_GENERIC;
    INS_CADDRESS(ret)= NULL_GENERIC;
    INS_OLD_SIZE(ret)= NULL_GENERIC;
    INS_CSIZE(ret)=INS_CSIZE(ins);
    T_RELOCATABLE(ret)->refedby = NULL;
    INS_PREV(ret)=INS_NEXT(ret)=NULL;

    /* all node structures (cf ud-graph) must be set to NULL.
     * the copied instruction is _never_ part of the ud-graph
     * until it is inserted in the graph */
    NODE_NEXT(ret) = NODE_PREV(ret) = NULL;
    NODE_SUCC_FIRST(ret) = NODE_SUCC_LAST(ret) = NULL;
    NODE_PRED_FIRST(ret) = NODE_PRED_LAST(ret) = NULL;

    RELOCATABLE_REFERS_TO(ret)=NULL;
    RELOCATABLE_REFED_BY(ret)=NULL;

    INS_COPY(ins)=ret;

    INS_EXEC_COUNT(ret)=INS_EXEC_COUNT(ins);

    while (refs)
  {
    /* Duplicate the reference */
    t_reloc * copy=RelocTableDupReloc(obj->reloc_table,refs->rel);

    VERIFY_FASSERT(RELOCATABLE_TYPE(REL_FROM(copy))==RT_INS,("Illegal reference with instruction @I: @R!",ins,copy));
    VERIFY_FASSERT(REL_TO_RELOCATABLE(copy),("@R has no to (code= %s)",copy,copy->code));

    RelocSetFrom(copy,T_RELOCATABLE(ret));
    refs=refs->next;
  }

  if(INS_OBJECT(ins)->actions->InsDupDynamic)
    INS_OBJECT(ins)->actions->InsDupDynamic(ret,ins);
  return ret;
}
/*}}}*/

/*!
 * \brief Check if the instruction has a side effect
 *
 * Architecture-independent wrapper around the architecture specific callbacks
 * for InsHasSideEffect
 *
 * \param ins The instruction to check
 *
 * \return t_bool TRUE if the instruction has a side effect, FALSE otherwise 
 */

t_bool InsHasSideEffect(t_ins * ins)
{
  t_section * sec = InsGetSection(ins);
  return sec->obj->actions->InsHasSideEffect(ins);
}

void InsMakeNoop(t_ins * ins)
{
  t_section * sec = InsGetSection(ins);
  sec->obj->actions->MakeNoop(ins);
}

t_ins * SecGetInsByAddress(t_section * sec, t_address addr)
{
#ifdef GENERICADDRSUPPORT
  if (addr.type==BIT32)
  {
#endif
#ifdef BIT32ADDRSUPPORT
    t_object * obj=sec->obj;
    t_uint32 addr32=G_T_UINT32(addr);
    t_int32 offset32=addr32-G_T_UINT32(SECTION_GET_CADDRESS(sec));
    if ((offset32 < 0) || (offset32 >= G_T_UINT32(SECTION_CSIZE(sec))))
      return NULL;
    else
    {
      t_int32 c_offset=((offset32*8)/obj->description->encoded_instruction_mod_size)*obj->description->decoded_instruction_size;
      
      t_ins * ins=((t_ins *) (((char *)  sec->data) +c_offset ));

      while (INS_ATTRIB(ins) & IF_DUMMY)
      {
        ins = INS_PREV(ins);
        if (!ins) FATAL(("Troubles!"));
      }
      return ins;
    }
#endif
#ifdef GENERICADDRSUPPORT
  }
  else if (addr.type==BIT64)
  {
#endif
#ifdef BIT64ADDRSUPPORT
#define BIT64_TOTAL_BUNDLESIZE (obj->description->encoded_instruction_mod_size*obj->description->bundle_size +                     obj->description->template_size) /*this expressed in bits*/
    t_object * obj=sec->obj;
    t_uint64 addr64=G_T_UINT64(addr);
    t_uint64 offset64=addr64-G_T_UINT64(SECTION_GET_CADDRESS(sec));
    t_uint64 c_offset=((obj->description->bundle_size*offset64*8)/BIT64_TOTAL_BUNDLESIZE
        + (addr64 & ((BIT64_TOTAL_BUNDLESIZE/8)-1)) /*this last thing is to deal with instructions within bundles; it assumes though that those are referenced with 0, 1, 2, ... as least x bits*/
        ) * obj->description->decoded_instruction_size;
#undef BIT64_TOTAL_BUNDLESIZE
    return ((t_ins *) (((char *)  sec->data) +c_offset ));
#endif
    FATAL(("Implement!"));
#ifdef GENERICADDRSUPPORT
  }
  else FATAL(("Implement!"));
#endif
}

t_ins * InsFindMovableDownEqualInsInBbl(t_ins * search, t_bbl * bbl)
{
  t_bool ins_can_move;
  t_ins * j_ins;
  t_bool load_seen=FALSE;
  t_bool store_seen=FALSE;
  t_regset use2=RegsetNew(), def2=RegsetNew(), tmp=RegsetNew();
  if (!BBL_OBJECT(bbl)->actions->CompareInstructions) return NULL;
  BBL_FOREACH_INS_R(bbl,j_ins)
  {
    ins_can_move=TRUE;
    if ((INS_TYPE(j_ins)==IT_LOAD) || (INS_TYPE(j_ins)==IT_LOAD_MULTIPLE))
      if (store_seen) ins_can_move=FALSE;
    if ((INS_TYPE(j_ins)==IT_STORE) || (INS_TYPE(j_ins)==IT_STORE_MULTIPLE))
      if (store_seen||load_seen) ins_can_move=FALSE;
    if ((INS_TYPE(j_ins)==IT_LOAD) || (INS_TYPE(j_ins)==IT_LOAD_MULTIPLE))
      load_seen=TRUE;
    if ((INS_TYPE(j_ins)==IT_STORE) || (INS_TYPE(j_ins)==IT_STORE_MULTIPLE))
      store_seen=TRUE; 
    RegsetDup(tmp,INS_REGS_DEF(j_ins));
    RegsetSetIntersect(tmp,use2);
    if (!RegsetIsEmpty(tmp)) ins_can_move=FALSE; 
    RegsetDup(tmp,INS_REGS_USE(j_ins));
    RegsetSetIntersect(tmp,def2);
    if (!RegsetIsEmpty(tmp)) ins_can_move=FALSE; 
    RegsetSetUnion(use2,INS_REGS_USE(j_ins));
    RegsetSetUnion(def2,INS_REGS_DEF(j_ins));
    RegsetSetUnion(use2,INS_REGS_DEF(j_ins));
    if (!ins_can_move) continue;
    if (BBL_OBJECT(bbl)->actions->CompareInstructions(search,j_ins)) { return j_ins; } 
  }
  return NULL;
}

/* iterator function useful during the layout phase */
t_ins * InsNextInChain(t_ins * ins)
{
  t_bbl * bbl = INS_BBL(ins);
  if (INS_NEXT(ins)) return INS_NEXT(ins);
  while (1)
  {
    if (!BBL_NEXT_IN_CHAIN(bbl)) return NULL;
    bbl = BBL_NEXT_IN_CHAIN(bbl);
    if (BBL_INS_FIRST(bbl)) return BBL_INS_FIRST(bbl);
  }
  /* keep the compiler happy */
  return NULL;
}

/** Calculate the registers live after ins, starting from the out
    registers in the block */
t_regset InsRegsLiveAfter(t_ins * ins)
{
  t_ins * i_ins;
  
  t_bbl * bbl = INS_BBL(ins);
  t_regset live = RegsetNew();

  RegsetDup(live,BBL_REGS_LIVE_OUT(bbl));

#ifdef DEBUG_LIVENESS
  printf("--------------- %d\n",BBL_NINS(bbl));
#endif

  for (i_ins = BBL_INS_LAST(bbl); ((i_ins!=NULL) && (i_ins != ins)); i_ins = INS_PREV(i_ins)) 
    {
#ifdef DEBUG_LIVENESS
      DiabloPrint(stdout,"@I\n",i_ins);
#endif
      if (!INS_IS_CONDITIONAL(i_ins))
	RegsetSetDiff(live,INS_REGS_DEF(i_ins));
      RegsetSetUnion(live,INS_REGS_USE(i_ins));
#ifdef DEBUG_LIVENESS
      DiabloPrintArch(stdout,BBL_FUN(INS_BBL(i_ins))->cfg->sec->obj->description,"live: @A\n",live);
#endif
    }

  if (!i_ins)
    {
#ifdef DEBUG_LIVENESS
      printf("*************\n");
      InsPrint(ins,0);
      BblPrint(bbl,PRINT_INSTRUCTIONS);
#endif 
    }
  if (!i_ins) FATAL(("ins is not in his own basic block!"));

  return live;
}

/** Calculate the registers live before ins, starting from the out
    registers in the block */
t_regset InsRegsLiveBefore(t_ins * ins)
{
  t_ins * i_ins;

  t_bbl * bbl = INS_BBL(ins);
  t_regset live = RegsetNew();

  RegsetDup(live,BBL_REGS_LIVE_OUT(bbl));

  for (i_ins = BBL_INS_LAST(bbl) ; ((i_ins!=NULL) && (i_ins != INS_PREV(ins))); i_ins = INS_PREV(i_ins)) 
    {
      if (!INS_IS_CONDITIONAL(i_ins))
	RegsetSetDiff(live,INS_REGS_DEF(i_ins));
      RegsetSetUnion(live,INS_REGS_USE(i_ins));
    }
  
#ifdef DEBUG_LIVENESS
  if (!i_ins) FATAL(("ins is not in his own basic block!"));
#endif

  return live;
}

t_regset InsRegsLiveOrUsedOrDefinedBefore(t_ins * ins)
{
  t_ins * i_ins;

  t_bbl * bbl = INS_BBL(ins);
  t_regset live = RegsetNew();

  RegsetDup(live,BBL_REGS_LIVE_OUT(bbl));

  for (i_ins = BBL_INS_LAST(bbl) ; ((i_ins!=NULL) && (i_ins != INS_PREV(ins))); i_ins = INS_PREV(i_ins)) 
    {
      if (!INS_IS_CONDITIONAL(i_ins))
	RegsetSetDiff(live,INS_REGS_DEF(i_ins));
      RegsetSetUnion(live,INS_REGS_USE(i_ins));
    }
  
  i_ins = BBL_INS_FIRST(INS_BBL(ins));
  while (i_ins && i_ins!=ins)
    {
      RegsetSetUnion(live,INS_REGS_USE(i_ins));
      RegsetSetUnion(live,INS_REGS_DEF(i_ins));
      i_ins = INS_NEXT(i_ins);
    }

#ifdef DEBUG_LIVENESS
  if (!i_ins) FATAL(("ins is not in his own basic block!"));
#endif

  return live;
}

/** Check if register r is not modified in the bbl after instruction
    ins */
t_bool InsRegUnmodifiedAfter(t_reg reg, t_ins * ins)
{

  t_ins * i_ins = ins;

  for (i_ins = INS_NEXT(i_ins) ; i_ins!=NULL ; i_ins = INS_NEXT(i_ins))
    if (RegsetIn(INS_REGS_DEF(i_ins),reg))
      return FALSE;
  
  return TRUE;
}

t_regset InsRegsDefBefore(t_ins * ins)
{
  t_regset def;
  t_function * fun = BBL_FUN(INS_BBL(ins));

  if( FUN_IS_HELL(fun) )
    {
      FATAL(("implement"));
    }

  def = RegsetNew();
        
  for (ins = INS_PREV(ins) ; ins!=NULL; ins = INS_PREV(ins))
    {
      RegsetSetUnion(def,INS_REGS_DEF(ins));
    }

  return def;
}

void InsChangePlace(t_ins * first_ins, t_ins * sec_ins)
{
  t_ins * prev_ins=INS_PREV(first_ins);
  t_ins * next_ins=INS_NEXT(sec_ins);
  
  if(INS_BBL(first_ins)!=INS_BBL(sec_ins))
    FATAL(("Not allowed to change place of 2 instructions from a differt BBL"));

  if(first_ins == sec_ins)
    FATAL(("Not allowed"));
  
  if(INS_NEXT(first_ins)!=sec_ins)
    FATAL(("Second ins (@I) must be just after first ins (@I, just before @I)",sec_ins,first_ins,INS_NEXT(first_ins)));

  if(INS_PREV(sec_ins)!=first_ins)
    FATAL(("First ins (@I) must be just before sec ins (@I)",first_ins,sec_ins));
  
  if(prev_ins)
    INS_NEXT(prev_ins)=sec_ins;
  else
    BBL_INS_FIRST(INS_BBL(sec_ins))=sec_ins;
  if(next_ins)
    INS_PREV(next_ins)=first_ins;
  else
    BBL_INS_LAST(INS_BBL(first_ins))=first_ins;

  INS_PREV(sec_ins)=prev_ins;
  INS_NEXT(sec_ins)=first_ins;
  INS_PREV(first_ins)=sec_ins;
  INS_NEXT(first_ins)=next_ins;
}

static t_inslist2 * delayed_ins_kill_list;
static t_inslist2 * delayed_ins_free_list;

void InsKillDelayed(t_ins * ins)
{
  t_inslist2 * new_entry = Malloc(sizeof(t_inslist2));
  new_entry->ins = ins;
  new_entry->next = delayed_ins_kill_list;
  delayed_ins_kill_list = new_entry;
}

void InitDelayedInsKilling()
{
  delayed_ins_kill_list = NULL;
}

void ApplyDelayedInsKilling()
{
  t_inslist2 * iterator = delayed_ins_free_list;

  while (iterator)
    {
      InsFreeReferedRelocs(iterator->ins);
      iterator = iterator->next;
      Free(delayed_ins_free_list);
      delayed_ins_free_list = iterator;
    }

  iterator = delayed_ins_kill_list;

  while (iterator)
    {
      InsKill(iterator->ins);
      iterator = iterator->next;
      Free(delayed_ins_kill_list);
      delayed_ins_kill_list = iterator;
    }
}

void InsFreeReferedRelocsDelayed(t_ins * ins)
{
  t_inslist2 * new_entry = Malloc(sizeof(t_inslist2));
  new_entry->ins = ins;
  new_entry->next = delayed_ins_free_list;
  delayed_ins_free_list = new_entry;
}

#ifdef ARM_SUPPORT
#include "../arch/arm/arm_utils.h"
#include "../arch/arm/arm_instructions.h"
#endif

t_bool InsIsFirstInConditionalBundle(t_ins * ins)
{
#ifndef ARM_SUPPORT
  return FALSE;
#else
  t_ins * prev_ins = INS_PREV(ins);
  t_uint32 cond;
  
  if (!(INS_ATTRIB(ins) & IF_CONDITIONAL)) return FALSE;

  cond = ARMINSCONDITION(ins);
  
  while (prev_ins)
    {
      if (ArmInsUpdatesCond(prev_ins))
	return TRUE;
      if (ARMINSCONDITION(prev_ins)==cond)
	return FALSE;
      if (ARMINSCONDITION(prev_ins)==ArmInvertCondition(cond))
	return FALSE;
      prev_ins = INS_PREV(prev_ins);
    }
  
  return TRUE;
#endif
}

t_uint32 ConditionalBundleSize(t_ins * first_ins)
{
#ifndef ARM_SUPPORT
  return 0;
#else
  t_ins * next_ins = first_ins;
  t_uint32 cond;
  int count = 0;

  if (!InsIsFirstInConditionalBundle(first_ins)) return 0;

  cond = ARMINSCONDITION(first_ins);
  
  while (next_ins)
    {
      if (ARMINSCONDITION(next_ins)==cond)
	count++;
      if (ArmInsUpdatesCond(next_ins))
	return count;
      next_ins = INS_NEXT(next_ins);
    }
  return count;
#endif
}



/* vim: set shiftwidth=2 foldmethod=marker: */
