/******************************************************************************
 * Copyright 2001,2002, 2003: {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file section 
 * handling routines  
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/
/* Includes {{{ */
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stddef.h>
#include "../config.h"
#include "../simple_types.h"
#include "diablo_layout.h"
#include "diablo_section.h"
#include "diablo_object.h"
#include "diablo_relocate.h"
#include "diablo_flowgraph.h"
#include "diablo_ins.h"
#include "diablo_relocate.h"
#include "diablo_print.h"
#include "diablo_description.h"
#include "diablo_iterators.h"
#include "diablo_addresses.h"
#include "../utils/util_malloc.h"
#include "../utils/util_status.h"
#include "../datastructures/data_string.h"
#include "../utils/util_error.h"
#include "../utils/util_endian.h"
#include "../diablo_options.h"
/*}}}*/

/*#define VERBOSE_PLACEMENT*/
/*#define VERBOSE_PLACEMENT2*/

t_segment * gseg=NULL;

/* {{{ Helper functions */
static int _sec_compare_caddresses(t_section ** s1, t_section ** s2)
{
  if (GENERIC_LT(SECTION_GET_CADDRESS(*s1),SECTION_GET_CADDRESS(*s2)))
    return -1;
  if (GENERIC_GT(SECTION_GET_CADDRESS(*s1),SECTION_GET_CADDRESS(*s2)))
    return 1;
  return 0;
}

/* sort sections by old address, but put all code sections up front */
static int _sec_reorder(t_section ** a, t_section ** b)
{
  t_section * seca = *a;
  t_section * secb = *b;


  if (SECTION_TYPE(seca) == SECTION_TYPE(secb))
    return GENERIC_LT(SECTION_GET_OLD_ADDRESS(seca), SECTION_GET_OLD_ADDRESS(secb)) ? -1 : 1;
  else if (SECTION_TYPE(seca) == CODE_SECTION)
    return -1;
  else if (SECTION_TYPE(secb) == CODE_SECTION)
    return 1;
  else
    return GENERIC_LT(SECTION_GET_OLD_ADDRESS(seca), SECTION_GET_OLD_ADDRESS(secb)) ? -1 : 1;

  /* keep the compiler happy */
  return 0;
}

static t_address SetAddress(long long addr, t_object * obj)
{
  t_address ret=NULL_GENERIC; /* Just to keep the compiler happy */
  if (obj->description->address_size == ADDRSIZE32)
  {
#ifdef BIT32ADDRSUPPORT
    GenericSet_uint32(ret,(t_uint32)addr);
#else
    FATAL(("Needed 32-bit addresses but 32bit address support was not enabled!"));
    exit(0); /* Keeps the compiler happy */
#endif
  }
  else
  if (obj->description->address_size == ADDRSIZE64)
  {
#ifdef BIT64ADDRSUPPORT
    GenericSet_uint64(ret,(t_uint64)addr);
#else
    FATAL(("Needed 64-bit addresses but 64bit address support was not enabled!"));
    exit(0); /* Keeps the compiler happy */
#endif
  }
  return ret;
}
/* }}} */
/* {{{ PlaceSectionsContiguously: place a given set of parent sections contiguously in memory,
 * adjusting address and size for all of them (and their subsecs)
 * PARAMS:
 * 	- secs: the array of parent sections
 * 	- nsecs: the number of sections in the array
 * 	- start: address in memory where placing starts
 * 	- end_align: enlarge the last section so that 
 * 	  the first free address is aligned to end_align
 * 	- pad_sections: indicates whether the parent sections
 *        should be resized to fill the gaps created by 
 *        alignment. We want to do this on the final section
 *        layout, but not while reordering sections before code 
 *        section merging (screws up $switch symbols for padding
 *        in code sections...)
 * RETURNS:
 *   the address of the first free byte of memory
 */
static t_address PlaceSectionsContiguously(t_section ** secs, t_uint32 nsecs, t_address start, t_address end_align, t_bool pad_sections)
{
  t_uint32 i,j;

  t_section * subsec, * sec, * prev_sec;
  t_address first_free, tmp_addr;

  t_uint32 nsubsecs = 0;
  t_section ** subsecarr = NULL;

  /* the start address should be at the right alignment */
  VERIFY_FASSERT(start == AlignAddress(SECTION_ALIGN(secs[0]),start),("Start address @G not properly aligned",start));

  first_free = start;
  for (i = 0; i < nsecs; i++)
  {
    sec = secs[i];
    prev_sec = i ? secs[i-1] : NULL;

    subsecarr = SectionGetSubsections(sec, &nsubsecs);

    /* Align the section. if padding must be added, enlarge the previous
     * section accordingly */
    tmp_addr = AlignAddress(SECTION_ALIGN(sec),first_free);
    if (pad_sections && prev_sec) SECTION_CSIZE(prev_sec) = GENERIC_ADD(SECTION_CSIZE(prev_sec),GENERIC_SUB(tmp_addr,first_free));
    first_free = tmp_addr;
    /* set the new address of the parent section */
    SectionSetCAddressSuper(sec, first_free);
    /*DiabloPrint(stdout,"Section %s old @G new @G align @G\n", sec->name, SECTION_OLD_ADDRESS(sec), SECTION_CADDRESS(sec), SECTION_ALIGN(sec)); */

    /* set the new address of the subsections */
    for (j=0; j<nsubsecs; j++)
    {
      subsec = subsecarr[j];
      first_free = AlignAddress(SECTION_ALIGN(subsec),first_free);
      SectionSetCAddressSuper(subsec, first_free);
      first_free = GENERIC_ADD(first_free,SECTION_CSIZE(subsec));
      /*DiabloPrint(stdout,"    Subsection %s old @G new @G align @G\n", subsec->name, SECTION_OLD_ADDRESS(subsec), SECTION_CADDRESS(subsec), SECTION_ALIGN(subsec)); */
    }

    /* set the size of the parent section */
    SECTION_CSIZE(sec) = GENERIC_SUB(first_free,SECTION_GET_CADDRESS(sec));

    if (subsecarr) Free(subsecarr);
  }

  /* change the size of the last section so that the first free address is
   * properly aligned */
  tmp_addr = AlignAddress(end_align,first_free);
  SECTION_CSIZE(secs[nsecs-1]) = GENERIC_ADD(SECTION_CSIZE(secs[nsecs-1]),GENERIC_SUB(tmp_addr,first_free));
  first_free = tmp_addr;

  return first_free;
} /* }}} */
/* PutSectionAtAddress {{{ */

static int _sec_compare_align(t_section ** s1, t_section ** s2)
{
  if (GENERIC_LT(SECTION_ALIGN(*s1),SECTION_ALIGN(*s2)))
    return 1;
  if (GENERIC_GT(SECTION_ALIGN(*s1),SECTION_ALIGN(*s2)))
    return -1;
  return 0;
}

t_uint64 PutSectionAtAddress(t_section * sec, t_uint64 start, t_layout_rule * r,t_bool min)
{
  t_object * obj = sec->obj;
  t_uint64 secaddr, endaddr, tmpaddr, padding;
  int i,j;
  t_uint32 nsubs, align;
  t_section ** subs;

  /* check if the script specifies a start address */
  if (r && r->u.secspec->address)
    secaddr = LayoutScriptCalcExp(r->u.secspec->address,start,obj);
  else
    secaddr = start;
  VERIFY_FASSERT(secaddr >= start, ("Section %s: requested address < first free address", sec->name));

  /* make sure the section is aligned properly */
  align = GENERIC_EXTRACT_UINT32(SECTION_ALIGN(sec));
  if (align) secaddr = (secaddr + (align-1)) & ~((t_uint64)(align-1));
  start = secaddr;

#ifdef VERBOSE_PLACEMENT
  VERBOSE(0,("section %s address 0x%llx",sec->name,secaddr)); 
#endif

  if (min)
    SectionSetMinAddressSuper(sec,SetAddress(secaddr,obj));
  else
    SectionSetCAddressSuper(sec,SetAddress(secaddr,obj));
  subs = SectionGetSubsections(sec,&nsubs);
#if 1
  if (SECTION_TYPE(sec) == DATA_SECTION || 
      SECTION_TYPE(sec) == RO_DATA_SECTION || 
      SECTION_TYPE(sec) == BSS_SECTION)
  {
    qsort(subs,nsubs,sizeof(t_section *),(int (*) (const void *,const void *)) _sec_compare_align);
    tmpaddr=secaddr;
    for (i=0; i<nsubs; i++)
    {
      t_section * sub = subs[i];
      align = GENERIC_EXTRACT_UINT32(SECTION_ALIGN(sub));

      padding=align?((tmpaddr + (align-1)) & ~((t_uint64)(align-1)))-tmpaddr:0;
      if (padding)
      {
	int bestj=0;
	int minpad=padding;
	for (j=nsubs-1; j>i; j--)
	{
	  t_section * sub2 = subs[j];
	  if (GENERIC_EXTRACT_UINT32(SECTION_ALIGN(sub2))>1) break;
	  if (GENERIC_EXTRACT_UINT64(SECTION_CSIZE(sub2))==0) continue;
	  if ((GENERIC_EXTRACT_UINT64(SECTION_CSIZE(sub2))% align)==0) continue;
	  if (((tmpaddr+GENERIC_EXTRACT_UINT64(SECTION_CSIZE(sub2)) + (align-1)) & ~((t_uint64)(align-1)))-tmpaddr-GENERIC_EXTRACT_UINT64(SECTION_CSIZE(sub2))>=minpad) continue;
	  minpad=(((tmpaddr+GENERIC_EXTRACT_UINT64(SECTION_CSIZE(sub2)) + (align-1)) & ~((t_uint64)(align-1)))-tmpaddr-GENERIC_EXTRACT_UINT64(SECTION_CSIZE(sub2)));
	  bestj=j;
	  if (minpad==0) break;
	}

	if (minpad!=padding)
	{
	  t_section * newi=subs[bestj];

	  for (j=bestj; j>i; j--)
	  {
	    subs[j]=subs[j-1];
	  }
	  subs[i]=newi;
	  i--;
	  continue;

	}
      }
      if (align) tmpaddr = (tmpaddr + (align-1)) & ~((t_uint64)(align-1));
      tmpaddr += GENERIC_EXTRACT_UINT64(SECTION_CSIZE(sub));
    }
  }
#endif
  
  for (i=0; i<nsubs; i++)
  {
    t_section * sub = subs[i];
    align = GENERIC_EXTRACT_UINT32(SECTION_ALIGN(sub));
    if (align) secaddr = (secaddr + (align-1)) & ~((t_uint64)(align-1));
    if (min)
      SectionSetMinAddressSuper(sub,SetAddress(secaddr,obj));
    else
      SectionSetCAddressSuper(sub,SetAddress(secaddr,obj));
#ifdef VERBOSE_PLACEMENT2
    VERBOSE(0,("subsection %s address 0x%llx", sub->name, secaddr));
#endif
    secaddr += GENERIC_EXTRACT_UINT64(SECTION_CSIZE(sub));
  }

  /* determine the end address of the section */
  if (r && r->u.secspec->internal_rule && !strcmp(".",r->u.secspec->internal_rule->lhs))
    endaddr = LayoutScriptCalcExp(r->u.secspec->internal_rule->rhs,secaddr,obj);
  else
    endaddr = secaddr;
  VERIFY_FASSERT(secaddr <= endaddr, ("section %s: specified end address < first free address",sec->name));

  if (endaddr == start && (SECTION_TYPE(sec) == DEFLOWGRAPHING_CODE_SECTION || SECTION_TYPE(sec) ==FLOWGRAPHED_CODE_SECTION))
  {
    /* the deflowgraphing code section is different: no subsections here,
     * the size must be calculated from the basic block sizes */
    t_cfg * cfg = SECTION_CFG(sec);
    t_bbl * bbl;
    t_uint32 nins = 0;
    CFG_FOREACH_BBL(cfg,bbl)
    {
      if (min)
	endaddr += GENERIC_EXTRACT_UINT64(BBL_MIN_SIZE(bbl));
      else
	endaddr += GENERIC_EXTRACT_UINT64(BBL_CSIZE(bbl));
      nins += BBL_NINS(bbl);
    }
    if (SECTION_TYPE(sec) !=FLOWGRAPHED_CODE_SECTION) SECTION_NINS(sec) = nins;
  }
  else if (endaddr == start && SECTION_TYPE(sec) == DEFLOWGRAPHED_CODE_SECTION)
  {
    /* almost the same as deflowgraphing code section, only 
     * the basic blocks are stored in a linked list here
     * (through BBL_NEXT_IN_CHAIN) */
    t_bbl * bbl;
    t_uint32 nins = 0;

    for (bbl = sec->data; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    {
      if (min)
	endaddr += GENERIC_EXTRACT_UINT64(BBL_MIN_SIZE(bbl));
      else
	endaddr += GENERIC_EXTRACT_UINT64(BBL_CSIZE(bbl));
      nins += BBL_NINS(bbl);
    }
    SECTION_NINS(sec) = nins;
  }
  else if (endaddr == start && SECTION_TYPE(sec) == ASSEMBLING_CODE_SECTION)
  {
    t_ins * ins;
    t_uint32 nins = 0;
    if (min) FATAL(("implement minimal address calculation in this case"));
    
    for (ins = (t_ins *) sec->data; ins; ins = INS_NEXT(ins), nins++)
      endaddr += GENERIC_EXTRACT_UINT64(INS_CSIZE(ins));
    SECTION_NINS(sec) = nins;
  }

  if (min)
    SECTION_MIN_SIZE(sec) = SetAddress(endaddr - start, obj);
  else
    SECTION_CSIZE(sec) = SetAddress(endaddr - start, obj);

#ifdef INSTRUMENTATION
  /* this is a very ugly hack, but it seems to be the only thing that works for now...
   * the problem is that with instrumentation, the PlaceSections code is called 
   * _before_ the subsections of the .got section are created. as a result the 
   * size of the .got section is set to 0, which we don't want... */
  if (!strcmp(SECTION_NAME(sec),".got"))
  {
    SECTION_CSIZE(sec) = SECTION_OLD_SIZE(sec);
    endaddr = GENERIC_EXTRACT_UINT64(GENERIC_ADD(SECTION_GET_CADDRESS(sec),SECTION_CSIZE(sec)));
  }
#endif

#ifdef VERBOSE_PLACEMENT
  VERBOSE(0,("-> 0x%llx size 0x%x",endaddr,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec))));
#endif

  if (subs) Free(subs);

  return endaddr;
}
/* }}} */
/* SectionGetGeneralType {{{ */
/* return general section type (ie the normal section
 * type for non-code sections and CODE_SECTION
 * for all variants of the code section (disassembled,
 * flowgraphed, etc.) */
static char SectionGetGeneralType(t_section * sec)
{
  switch (SECTION_TYPE(sec))
  {
    case RO_DATA_SECTION:
    case DATA_SECTION:
    case BSS_SECTION:
    case NOTE_SECTION:
      return SECTION_TYPE(sec);
  }
  return CODE_SECTION;
}
/* }}} */
/* ExecLayoutRule {{{ */
t_uint64 ExecLayoutRule(t_layout_rule * r, t_uint64 currpos, t_object * obj,t_bool min)
{
  /* skip symbol assignments */
  if (r->kind == ASSIGN && strcmp(r->u.assign->lhs, "."))
    return currpos;

  if (r->kind == ASSIGN)
  {
    currpos = LayoutScriptCalcExp(r->u.assign->rhs,currpos,obj);
#ifdef VERBOSE_PLACEMENT
    VERBOSE(0,(". = %llx", currpos));
#endif
  }
  else if (r->kind == SECTION_SPEC)
  {
    /* place section rule */
    t_string secname = r->u.secspec->name;
    t_section_hash_node * shn = HashLookup(obj->section_table, secname);

    if (shn && shn->section)
      currpos = PutSectionAtAddress(shn->section,currpos,r,min);
  }
  else if (r->kind == SEG_START)
  {
    t_segment * seg = ObjectGetSegmentByName(obj,r->u.segment_ident);
    if (seg) {
      SEGMENT_CADDRESS(seg) = SetAddress(currpos,obj);
      seg->hoist_headers=FALSE;
    }
    else {
      seg=SegmentNew(r->u.segment_ident);
      seg->next=NULL;
      seg->hoist_headers=FALSE;
      if (obj->segment_first)
      {
	obj->segment_last->next=seg;
	seg->prev=obj->segment_last;
	obj->segment_last=seg;
      }
      else
      {
	seg->prev=NULL;
	obj->segment_first=obj->segment_last=seg;
      }
      SEGMENT_CADDRESS(seg) = SetAddress(currpos,obj);
    }
    gseg=seg;
  }
  else if (r->kind == SEG_END)
  {
    t_segment * seg = ObjectGetSegmentByName(obj,r->u.segment_ident);
    t_address tmp;
    if (seg)
    {
      tmp = SetAddress(currpos,obj);
      SEGMENT_CSIZE(seg) = GENERIC_SUB(tmp,SEGMENT_CADDRESS(seg));
    }
    gseg=NULL;
  }
  else if (r->kind == PUT_SECTIONS)
  {
    t_section * sec;
    t_uint32 i;
    OBJECT_FOREACH_SECTION(obj,sec,i)
    {
      if (SectionGetGeneralType(sec) == r->u.sectype)
	if (!(SECTION_FLAGS(sec) & SF_MARKED))
	  currpos = PutSectionAtAddress(sec,currpos,NULL,min);
    }
  }
  else
    FATAL(("unknown type of layout rule"));

  return currpos;
}
/* }}} */
/* PlaceSections {{{ */
void PlaceSections(t_object * obj,t_bool min)
{
  t_layout_rule * script = obj->layout_script;
  t_layout_rule * r;
  t_uint64 currpos;
  int i;
  t_section * sec;
#ifdef VERBOSE_PLACEMENT
  VERBOSE(0,("Layout started!"));
#endif
  /* turn off the marked and placed flags for all sections */
  OBJECT_FOREACH_SECTION(obj,sec,i)
    SECTION_FLAGS(sec) &= ~(SF_PLACED | SF_MARKED);

  if (!script)
  {
    /* read and parse the linker script */
    script = obj->layout_script = ObjectGetAndParseLayoutScript(obj);
  }

  /* {{{ first step: mark all sections that are mentioned in the layout script
   * for those that are unmarked after this step, we will have to decide 
   * ourselves where to place them */
  for (r = script; r; r = r->next)
  {
    t_section * sec;
    t_section_hash_node * shn;

    if (r->kind != SECTION_SPEC) continue;
    shn = HashLookup(obj->section_table, r->u.secspec->name);
    if (!shn) continue;		/* section not in binary */
    sec = shn->section;
    if (!sec) continue;		/* section removed by diablo */

    SECTION_FLAGS(sec) |= SF_MARKED;
  } /* }}} */

  /* second step: execute layout script */
  currpos = 0LL;
  for (r = script; r; r = r->next)
  {
    currpos = ExecLayoutRule(r, currpos, obj, min);
  }

  /* {{{ third step: reorder the sections in the obj->code[], obj->data[], ... arrays
   * this is necessary as the section order may have changed during the placing of the 
   * sections, and there is code in diablo that expects the order to be correct */
  qsort(obj->code,obj->ncodes,sizeof(t_section *),(int (*) (const void *,const void *)) _sec_compare_caddresses);
  qsort(obj->rodata,obj->nrodatas,sizeof(t_section *),(int (*) (const void *,const void *)) _sec_compare_caddresses);
  qsort(obj->data,obj->ndatas,sizeof(t_section *),(int (*) (const void *,const void *)) _sec_compare_caddresses);
  qsort(obj->bss,obj->nbsss,sizeof(t_section *),(int (*) (const void *,const void *)) _sec_compare_caddresses);
  qsort(obj->note,obj->nnotes,sizeof(t_section *),(int (*) (const void *,const void *)) _sec_compare_caddresses);
  /* }}} */

#ifdef VERBOSE_PLACEMENT
  VERBOSE(0,("Layout finished!"));
#endif
}
/* }}} */
/* OrderCodeSectionsContiguously {{{ */
void OrderCodeSectionsContiguously(t_object * obj)
{
  int i;
  t_section ** secarr;
  t_uint32 nsecs,nfixed;
  t_address tmp_align;

  /* we only do this for parent objects */
  VERIFY_FASSERT(obj->subobject_cache, ("Can't do this for sub objects"));

  nsecs = obj->ncodes + obj->nrodatas + obj->ndatas + obj->nbsss;
  secarr = (t_section **) Calloc(nsecs,sizeof(t_section *));

  memcpy(secarr,obj->code,sizeof(t_section *)*obj->ncodes);
  memcpy(secarr+obj->ncodes,obj->rodata,sizeof(t_section *)*obj->nrodatas);
  memcpy(secarr+obj->ncodes+obj->nrodatas,obj->data,sizeof(t_section *)*obj->ndatas);
  memcpy(secarr+obj->ncodes+obj->nrodatas+obj->ndatas,obj->bss,sizeof(t_section *)*obj->nbsss);
  memcpy(secarr+obj->ncodes+obj->nrodatas+obj->ndatas+obj->nbsss,obj->note,sizeof(t_section *)*obj->nnotes);

  /* remove sections with a fixed address from the array */
  nfixed = 0;
  for (i=0; i<nsecs; i++)
  {
    t_section * tmp;
    tmp = secarr[i];
    secarr[i] = NULL;
    if (tmp->is_fixed)
      nfixed++;
    else
      secarr[i-nfixed] = tmp;
  }

  /* reorder the sections: code sections will come first */
  qsort(secarr,nsecs-nfixed,sizeof(t_section *),(int (*) (const void *,const void *)) _sec_reorder);

  /* this is a guess: we don't know anything about the end alignment so we take 8 byte,
   * as this is a safe value for the heap begin on 32-bit and 64-bit architectures */
  tmp_align = GENERIC_ADD_UINT32(NULL_GENERIC,8);

  PlaceSectionsContiguously(secarr, nsecs - nfixed, SECTION_GET_CADDRESS(secarr[0]), tmp_align, FALSE);
}
/* }}} */

#include "diablo_iterators.h"
void LayoutPrint(t_object * obj)
{
  t_object * sub, * tmp;

  t_section * sec;
  t_uint32 i;
  OBJECT_FOREACH_SUBOBJECT(obj,sub,tmp)
  {
    OBJECT_FOREACH_SECTION(sub,sec,i)
    {
      if (SECTION_TYPE(sec)==CODE_SECTION) continue;
      DiabloPrint(stdout,"%s %s old: @G new: @G\n",sub->name,sec->name,SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_CADDRESS(sec));
    }
  }
}

typedef struct _t_layout_chain
{
  t_bbl * chain_head;
  t_int32 max_length;
  t_int32 alignment;
  t_address address;
  int prev_index;
} t_layout_chain;

t_layout_chain * layouted_chains;

static t_address local_split_address;

int cmp_chain_alignment(const void * a , const void *b)
{
  t_layout_chain * lca = (t_layout_chain*)a; 
  t_layout_chain * lcb = (t_layout_chain*)b;
  
  if (GENERIC_LT(lca->address,local_split_address) && !GENERIC_LT(lcb->address,local_split_address)) return -1;
  if (GENERIC_LT(lcb->address,local_split_address) && !GENERIC_LT(lca->address,local_split_address)) return -1;

  if (!GENERICISNULL(lca->address)) return -1;
  if (!GENERICISNULL(lcb->address)) return 1;

  if (lca->alignment>0 && lcb->alignment<=0) return -1;
  if (lcb->alignment>0 && lca->alignment<=0) return 1;

  return 0;
}

t_int32 BblMaxLength(t_bbl * bbl)
{
  t_int32 return_value = 0;
  t_ins * ins;
  BBL_FOREACH_INS(bbl,ins)
    return_value+=INS_CSIZE(ins);
  return return_value+4;
}

#define BblMaxLength(x) (4*BBL_NINS(x))
#define BblAlignment(x) 0

void MergeChains(t_bbl * ca, t_bbl * cb)
{
  t_bbl * bbl;

  BBL_NEXT_IN_CHAIN(BBL_LAST_IN_CHAIN(ca)) = cb;
  BBL_PREV_IN_CHAIN(cb) = BBL_LAST_IN_CHAIN(ca);
  for (bbl = ca; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    {
      BBL_FIRST_IN_CHAIN(bbl) = ca;
      BBL_LAST_IN_CHAIN(bbl) = BBL_LAST_IN_CHAIN(cb);
    }
  
}
#ifndef IA64_SUPPORT
void LayoutChains(t_cfg * cfg, t_address begin_address_txt, t_address old_address_split)
{
  t_layout_chain * tmp;
  t_bbl * bbl;
  int nr_chains = 1;
  int nr_aligned_bbls = 0;
  int filling_after = 0;
  t_bbl * first_bbl = NULL;
  int i = 0,j;
  t_uint32 space_between;

  local_split_address = old_address_split;

  layouted_chains = (t_layout_chain*) malloc (sizeof(t_layout_chain)*1000000);

  tmp = layouted_chains;

  /* start block */

  tmp->chain_head = NULL;
  tmp->max_length = 0;
  tmp->alignment  = 0;
  tmp->address    = begin_address_txt;
  tmp->prev_index = -1;

  /* create chains for other chains */
  
  CFG_FOREACH_BBL(cfg,bbl)
    if (BBL_FIRST_IN_CHAIN(bbl)==bbl)
      {
	tmp[nr_chains].chain_head = bbl;
	tmp[nr_chains].max_length = BblMaxLength(bbl);
	tmp[nr_chains].alignment = BblAlignment(bbl);
	if (BblAlignment(bbl)>0)
	  nr_aligned_bbls++;
	tmp[nr_chains].address = 0;
	tmp[nr_chains].prev_index=-1;
	nr_chains++;
      }
 
  /* layout the chains that need alignment */
 
  if (nr_aligned_bbls>0)
    {
      t_address next_free_address = begin_address_txt;
      t_uint32 start_address;

      qsort(layouted_chains,nr_chains,sizeof(t_layout_chain),cmp_chain_alignment);
      
      for (i=1;i<nr_aligned_bbls+1;i++)
	{
	  start_address = GENERIC_EXTRACT_UINT32(next_free_address)+tmp[i].alignment;
	  start_address &= tmp[i].alignment-1;
#if BIT32ADDRSUPPORT
	  GenericSet_uint32(tmp[i].address,start_address);
#else
	  GenericSet_uint64(tmp[i].address,start_address);
#endif

	  BBL_CADDRESS(tmp[i].chain_head)=start_address;
	  next_free_address = tmp[i].address;
	  next_free_address=GENERIC_ADD_UINT32(next_free_address,tmp[i].max_length);
	  
	}
    }

  /* layout the other chains */
  
  space_between = GENERIC_EXTRACT_UINT32(tmp[1].address)-GENERIC_EXTRACT_UINT32(tmp[0].address);

  for (;i<nr_chains;i++)
    {
      while (!space_between>=tmp[i].max_length)
	{
	  filling_after++;
	  if (filling_after==nr_aligned_bbls)
	    space_between = 300*1024*1024;
	  else
	    space_between = GENERIC_EXTRACT_UINT32(tmp[filling_after+1].address)-GENERIC_EXTRACT_UINT32(tmp[filling_after].address)-tmp[filling_after].max_length;
	}
      
      tmp[i].prev_index = filling_after;
      space_between -= tmp[i].max_length;
    }

  /* chain the first blocks */
  
  for (i=nr_aligned_bbls+1;i<nr_chains;i++)
    {
      if (tmp[i].prev_index==0)
	{
	  if (first_bbl==NULL)
	    first_bbl = tmp[i].chain_head;
	  else
	    MergeChains(first_bbl,layouted_chains[i].chain_head);
	}
    }

  /* chain the rest */

  for (i=1;i<nr_aligned_bbls+1;i++)
    {
      for (j=0;j<nr_chains;j++)
	{
	  if (layouted_chains[j].prev_index==-1) continue;
	  if (layouted_chains[j].prev_index==i)
	    {
	      MergeChains(layouted_chains[i].chain_head,layouted_chains[j].chain_head);
	    }
	}
      MergeChains(first_bbl,layouted_chains[i].chain_head);
    }
}
#endif
/* vim: set shiftwidth=2 foldmethod=marker: */
