%{
/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file routines 
 * 
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/
#include <string.h>
#include "diablo_layout_script.h"
#include "diablo_object.h"
#include "../utils/util_error.h"
#include "../utils/util_malloc.h"
#include "../kernel/diablo_print.h"
#include "../fileformats/elf/elf_main.h"
#include "../simple_types.h"

extern t_segment * gseg;
int lparselex(void);
static int lparseerror(char * s);

static t_layout_exp * new_exp(int operator, void * arg1, void * arg2);

static t_layout_rule * last_rule = NULL;
static t_layout_rule * complete_script = NULL;
%}


%union {
  t_string string;
  long long longlong;
  t_layout_exp * layout_exp;
  t_layout_rule * layout_rule;
  t_layout_assign * layout_assign;
  t_layout_secspec * layout_secspec;
}

%type <layout_rule> rule rules command
%type <layout_assign> assign opt_endalign
%type <layout_secspec> section_spec;
%type <layout_exp> exp opt_filler opt_address

%token SEGMENT_START
%token SEGMENT_END
%token PUT_REMAINING_SECTIONS
%token SECTIONS
%token ALIGN
%token PROVIDE
%token <string> IDENTIFIER
%token <longlong> NUMBER

%left '|'
%left '&'
%left '+' '-'
%left '*' '/'
%left '~'
%left UMINUS

%start layout_script

%%

layout_script: SECTIONS '{' rules '}' { complete_script = $3; }
             ;

rules: rules rule	{ 
	$$ = $1; 
	VERIFY_FASSERT(last_rule, ("last_rule NULL but not first rule"));
	last_rule->next = $2; 
	last_rule = last_rule->next;
}
     | rule		{ $$ = last_rule = $1;}
     ;

rule: section_spec	{
	t_layout_rule * tmp = Malloc(sizeof(t_layout_rule));
	tmp->kind = SECTION_SPEC;
	tmp->u.secspec = $1;
	tmp->next = NULL;
	$$ = tmp;
}
    | assign ';'	{
 	t_layout_rule * tmp = Malloc(sizeof(t_layout_rule));
 	tmp->kind = ASSIGN;
 	tmp->u.assign = $1;
 	tmp->next = NULL;
 	$$ = tmp;
}
    | command ';'	{
	$$ = $1;
}
    ;

command: SEGMENT_START '(' '"' IDENTIFIER '"' ')'	{
	t_layout_rule * tmp = Malloc(sizeof(t_layout_rule));
	tmp->kind = SEG_START;
	tmp->u.segment_ident = $4;
	tmp->next = NULL;
	$$ = tmp;
}
    | SEGMENT_END '(' '"' IDENTIFIER '"' ')'		{
	t_layout_rule * tmp = Malloc(sizeof(t_layout_rule));
	tmp->kind = SEG_END;
	tmp->u.segment_ident = $4;
	tmp->next = NULL;
	$$ = tmp;
}
    | PUT_REMAINING_SECTIONS '(' IDENTIFIER ')'	{
	t_layout_rule * tmp = Malloc(sizeof(t_layout_rule));
	tmp->kind = PUT_SECTIONS;

	/* check the validity of the identifier */
	if (!strcmp($3,"CODE_SECTION"))
	{
	  tmp->u.sectype = CODE_SECTION;
	}
	else if (!strcmp($3,"RO_DATA_SECTION"))
	{
	  tmp->u.sectype = RO_DATA_SECTION;
	}
	else if (!strcmp($3,"DATA_SECTION"))
	{
	  tmp->u.sectype = DATA_SECTION;
	}
	else if (!strcmp($3,"BSS_SECTION"))
	{
	  tmp->u.sectype = BSS_SECTION;
	}
	else if (!strcmp($3,"NOTE_SECTION"))
	{
	  tmp->u.sectype = NOTE_SECTION;
	}
	else
	{
	  FATAL(("not a valid section type identifier: %s",$3));
	}
	Free($3);
	tmp->next = NULL;
	$$ = tmp;
}
    ;

section_spec: IDENTIFIER opt_address ':' '{' opt_endalign '}' opt_filler	{
	t_layout_secspec * tmp = Malloc(sizeof(t_layout_secspec));
	tmp->name = $1;
	tmp->address = $2;
	tmp->internal_rule = $5;
	tmp->filler_exp = $7;
	$$ = tmp;
}
            ;

opt_address: /* empty */	{ $$ = NULL; }
           | exp		{ $$ = $1; }
	   ;

opt_endalign: /* empty */	{ $$ = NULL; }
            | assign ';'	{ $$ = $1; }
	    ;

opt_filler: /* empty */ 	{ $$ = NULL; }
          | '=' exp		{ $$ = $2; }
	  ;

assign: IDENTIFIER '=' exp	{ 
	t_layout_assign * tmp = Malloc(sizeof(t_layout_assign));
	tmp->lhs = $1;
	tmp->rhs = $3;
	tmp->provide = FALSE;
	$$ = tmp;
}
      | PROVIDE '(' IDENTIFIER '=' exp ')'	{
	t_layout_assign * tmp = Malloc(sizeof(t_layout_assign));
	tmp->lhs = $3;
	tmp->rhs = $5;
	tmp->provide = TRUE;
	$$ = tmp;
}
      ;

exp: ALIGN '(' exp ')'		{ $$ = new_exp(ALIGN,$3,NULL); }
   | '(' exp ')'		{ $$ = $2; }
   | '-' exp %prec UMINUS	{ $$ = new_exp(UMINUS,$2,NULL); }
   | '~' exp 			{ $$ = new_exp('~',$2,NULL); }
   | exp '+' exp		{ $$ = new_exp('+',$1,$3); }
   | exp '-' exp		{ $$ = new_exp('-',$1,$3); }
   | exp '*' exp		{ $$ = new_exp('*',$1,$3); }
   | exp '/' exp		{ $$ = new_exp('/',$1,$3); }
   | exp '&' exp 		{ $$ = new_exp('&',$1,$3); }
   | exp '|' exp		{ $$ = new_exp('|',$1,$3); }
   | NUMBER			{ long long tmp = $1; $$ = new_exp(NUMBER,&tmp,NULL); }
   | IDENTIFIER			{ $$ = new_exp(IDENTIFIER,$1,NULL); }
   ;


%%

static t_layout_exp * new_exp(int operator, void * arg1, void * arg2)
{
  t_layout_exp * ret = Malloc(sizeof(t_layout_exp));
  ret->operator = operator;
  if (operator == NUMBER)
  {
    ret->arg1 = ret->arg2 = NULL;
    ret->identifier = NULL;
    ret->constant = *(long long *)arg1;
  }
  else if (operator == IDENTIFIER)
  {
    ret->arg1 = ret->arg2 = NULL;
    ret->constant = 0LL;
    ret->identifier = (t_string)arg1;
  }
  else
  {
    ret->arg1 = (t_layout_exp *)arg1;
    ret->arg2 = (t_layout_exp *)arg2;
    ret->identifier = NULL;
    ret->constant = 0LL;
  }
  return ret;
}

extern FILE * lparsein;
t_layout_rule * LayoutScriptParse(t_string scriptfile)
{
  t_layout_rule * ret;
  FILE * f = fopen(scriptfile,"r");
  if (!f) FATAL(("Could not open %s for reading",scriptfile));
  DiabloPrint(stdout, "-- parsing layout script in %s\n", scriptfile);

  lparsein = f;
  if (lparseparse())
    FATAL(("Parsing appears to have failed"));
  ret = complete_script;
  complete_script = NULL;
  return ret;
}

/* {{{ rudimentary parse tree printer */
void print_exp(t_layout_exp * exp)
{
  switch (exp->operator)
  {
    case ALIGN: 
      printf("align\n"); break;
    case UMINUS:
      printf("uminus\n"); break;
    case IDENTIFIER:
      printf("identifier %s\n", exp->identifier); break;
    case NUMBER:
      printf("constant %llx\n", exp->constant); break;
    default:
      printf("operator %c\n", exp->operator); break;
  }

  if (exp->arg1)
  {
    printf("subexp 1:\n");
    print_exp(exp->arg1);
  }
  if (exp->arg2)
  {
    printf("subexp 2:\n");
    print_exp(exp->arg2);
  }
}

void print_assign(t_layout_assign * assign)
{
  printf("assign %s\n", assign->lhs);
  print_exp(assign->rhs);
}

void printscript(void)
{
  t_layout_rule * r = complete_script;
  for (; r; r = r->next)
  {
    printf("<<< rule >>>\n");
    if (r->kind == SECTION_SPEC)
    {
      printf("name %s\n", r->u.secspec->name);
      if (r->u.secspec->internal_rule)
        print_assign(r->u.secspec->internal_rule);
      else
        printf("no internal rule\n");
      if (r->u.secspec->filler_exp)
        print_exp(r->u.secspec->filler_exp);
    }
    else
    {
      print_assign(r->u.assign);
    }
    printf("<<< end rule >>>\n");
  }
}
/* }}} */

static int lparseerror(char *s)
{
  FATAL(("parsing failed. parser said: %s",s));
  return 0;
}

long long LayoutScriptCalcExp(t_layout_exp * exp, long long currpos, t_object * obj)
{
  long long res1=0, res2=0;

  if (exp->arg1)
    res1 = LayoutScriptCalcExp(exp->arg1,currpos,obj);
  if (exp->arg2)
    res2 = LayoutScriptCalcExp(exp->arg2,currpos,obj);

  switch (exp->operator)
  {
    case ALIGN:
      return (currpos + res1 - 1) & ~(res1 - 1); 
    case UMINUS:
      return (-res1);
    case NUMBER:
      return exp->constant;
    case IDENTIFIER:
    {
      if (!strcmp(exp->identifier,"."))
        return currpos;
      else
      {
	if (strcmp(exp->identifier,"SIZEOF_HEADERS") && strcmp(exp->identifier,"sizeof_headers"))
	  FATAL(("Currently unsupported identifier (%s) in expression",exp->identifier));
	else if (gseg) gseg->hoist_headers=TRUE;

	return obj->ff_actions->GetSizeofHeaders(obj,obj->layout_script);
      }
    }
    case '~':
      return (~res1);
    case '&':
      return res1 & res2;
    case '|':
      return res1 | res2;
    case '+':
      return res1 + res2;
    case '-':
      return res1 - res2;
    case '*':
      return res1 * res2;
    case '/':
      return res1 / res2;

    default:
      FATAL(("Unknown operator in expression"));
      exit(0);/* Keep the compiler happy */
  }
  return 0; /* Keep the compiler happy */
}

static void free_exp(t_layout_exp * exp)
{
  if (!exp) return;

  if (exp->arg1) free_exp(exp->arg1);
  if (exp->arg2) free_exp(exp->arg2);
  if (exp->identifier) Free(exp->identifier);
  Free(exp);
}

static void free_assign(t_layout_assign * assign)
{
  if (!assign) return;
  Free(assign->lhs);
  free_exp(assign->rhs);
  Free(assign);
}

void LayoutScriptFree(t_layout_rule * script)
{
  t_layout_rule * r, * next;
  for (r=script, next=r?r->next:NULL; r; r=next,next=r?r->next:NULL)
  {
    if (r->kind == ASSIGN)
    {
      free_assign(r->u.assign);
    }
    else if (r->kind == SECTION_SPEC)
    {
      free_assign(r->u.secspec->internal_rule);
      free_exp(r->u.secspec->address);
      free_exp(r->u.secspec->filler_exp);
      Free(r->u.secspec->name);
      Free(r->u.secspec);
    }
    else if (r->kind == SEG_START || r->kind == SEG_END)
    {
      Free(r->u.segment_ident);
    }
    else if (r->kind == PUT_SECTIONS)
    {
    }
    Free(r);
  }
}

/* vim: set shiftwidth=2 foldmethod=marker: */
