/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/

/* This file is used to handle linker descriptions. A linker description is a
 * script file that describes how a specific linker for an architecture works:
 * what extra (so called linker defined) symbols it creates, what sections are
 * added by the linker, and what values these symbols and sections contain. A
 * linker description should eventually contain all architecture/linker
 * specific linking stuff. 
 *
 * A linker descriptions consist to a large extend of a set of rules. These
 * rules describe a trigger (what condition needs to be met to execute the
 * rule) and an action (what does the rule do). All other parts of the linker
 * rule are data for the actions.
 *
 * Diablo parses the linker description (parser and lexer in
 * diablo_linker_script_parser.[ly]), and calls ExecuteRule (in this file) on
 * each rule. 
 * 
 * ExecuteRule finds out which rules of the script need to be executed by
 * evaluating the trigger and if necessary executes the rules.  Triggers can
 * return three values (0, 1 and 2). 0 means the rule is not to be executed. 1
 * means we can execute the rule once and 2 means we need to execute the rule,
 * and when it is executed, we need to execute it again.  The last case occurs
 * when a trigger can be met several times.
 *
 * TODO (documentation)      Describe different triggers and actions
 * TODO (parser)             Handle functions as parameters
 * TODO (FillerRelocated32)  MATCHED_NAME should be a callback function
 * TODO (triggers)	     We need reloc matching triggers
 *
 * */

/* Includes {{{ */
#include "../simple_types.h"
#include "../datastructures/data_string.h"
#include "../utils/util_error.h"
#include "../utils/util_malloc.h"
#include "diablo_linkers.h"
#include "diablo_object.h"
#include "diablo_section.h"
#include "diablo_symbol.h"
#include "diablo_reloc.h"
#include "diablo_print.h"
#include "../diablo_options.h"
#include "diablo_iterators.h"
#include <string.h>
/*}}}*/
/* Global variables {{{ */
t_object * parser_object=NULL;
t_bool current_free_mode=FALSE;
t_address end_of_last_added;
t_string last_matched_symbol_name=NULL;
t_symbol * last_matched_symbol=NULL;
t_bool symbol_is_local = FALSE;

t_object * last_matched_object=NULL;
t_object * next_matched_object=NULL;
t_object * next_matched_object2=NULL;
t_section * last_matched_section=NULL;
t_section * next_matched_section=NULL;
int next_matched_section_tel=0;

/* }}} */
/* Static function/datastructure prototypes {{{ */
extern t_ast_node_table_entry  FunctionTable[];
extern t_ast_node_table_entry  SymbolFunctionTable[];
extern t_ast_node_table_entry  AddressTable[];

#define AstExecute(x,y,z,w)  AstExecuteREAL(__FILE__,__LINE__,x,y,z,w)

static t_ast_node * AstExecuteREAL(char * , int ,t_ast_node * , t_ast_node_table_entry * , void * ,t_bool );
static void CheckAndCastOneArgument(t_string , int , t_ast_node ** , char , t_ast_node_table_entry * , void * );
static void CheckAndCastArguments(t_ast_node *, t_uint32 , t_string,t_ast_node_table_entry * , void * );
static void * Vectorize(t_ast_successors *,void *);
static void * SectionStringChop(t_ast_successors *,void *);
/* }}} */

/* Abstract syntax tree manipulation {{{ */
void AstPrint(FILE* file,t_ast_node * in)
{
  t_ast_successors * args=in->args;

  fprintf(file,"\"%p\" [label=\"",in);
  switch(in->type)
  {
    case TYPE_FUNCTION:
      fprintf(file,"FUNC: %s",in->data.name);
      break;

    case TYPE_OPER:
      fprintf(file,"OPER: %c",in->data.oper);
      break;
    case TYPE_NUMERIC:
      fprintf(file,"NUMBER");
      break;
    case TYPE_ADDRESS:
      fprintf(file,"ADDRESS");
      break;
    case TYPE_OTHER:
      fprintf(file,"OTHER %s",in->data.name);
      break;
    default:
      fprintf(file,"END");

  }
  printf("\"]\n");
  if (args)
  {
    t_uint32 tel;
    for (tel=0; tel<args->nargs; tel++)
    {
      fprintf(file,"\"%p\" -> \"%p\"\n",in,args->args[tel]);
      AstPrint(file,args->args[tel]);
    }
  }
}




t_ast_successors * AstSuccessorsNew()
{
  t_ast_successors * ret=Malloc(sizeof(t_ast_successors));
  ret->nargs=0;
  ret->args=NULL;
  return ret;
}

void AstSuccessorsAddSuccessor(t_ast_successors * args, t_string argument, t_ast_types type, t_ast_successors * args2)
{
  args->nargs++;

  args->args=Realloc(args->args,args->nargs*sizeof(t_ast_node *));
  args->args[args->nargs-1]=Malloc(sizeof(t_ast_node));
  args->args[args->nargs-1]->type=type;
  args->args[args->nargs-1]->data.data=argument;
  args->args[args->nargs-1]->args=args2;
}

t_ast_node * AstNodeFunctionNew(t_string name, t_ast_successors * args)
{
  t_ast_node * ret=Malloc(sizeof(t_ast_node));
  ret->type=TYPE_FUNCTION;
  ret->data.name=name;
  ret->args=args;
  return ret;
}

t_ast_node * AstNodeOperatorNew(t_ast_operators oper, t_ast_node * left, t_ast_node *right)
{
   t_ast_node * ret=Malloc(sizeof(t_ast_node));
   ret->type=TYPE_OPER;
   ret->data.oper=oper;

   ret->args=AstSuccessorsNew();
   AstSuccessorsAddSuccessor(ret->args, left->data.data, left->type, left->args);
   AstSuccessorsAddSuccessor(ret->args, right->data.data, right->type, right->args);
   Free(left);
   Free(right);
   return ret;
}

void FreeAstNode(t_ast_node * in)
{
  t_ast_successors * args=in->args;
  t_string name=in->data.name;
  t_uint32 tel;

  if (args) 
  {
    /* The number of arguments is right, now check and cast the arguments */

    for (tel=0; tel<args->nargs; tel++)
    {
      if (args->args[tel])
      {
	switch(args->args[tel]->type)
	{
	  case TYPE_STRING:
	  case TYPE_OTHER:
	    if (args->args[tel]->data.data) /* NULL Strings! */
	      Free(args->args[tel]->data.data);
	    break;
	  default:
	    break;	    
	}
	FreeAstNode(args->args[tel]);
      }
    }
    if (in->args->args)
    Free(in->args->args);
    Free(in->args);
  }
  
  if (in->type==TYPE_FUNCTION) { Free(name); }
 

  Free(in);
}

t_ast_node * AstNodeDupInternal(t_ast_node * in, t_ast_node * p)
{
  t_ast_node * ret=p;
  if (!in) return NULL;
  if (!ret) ret=Malloc(sizeof(t_ast_node));

  /* type and data {{{ */
  ret->type=in->type;

  switch(in->type)
  {
    case TYPE_OTHER:
    case TYPE_STRING:
    case TYPE_FUNCTION:
      ret->data.data=StringDup(in->data.data);
      break;
    default:
      ret->data.data=in->data.data;
      break;
  }
  /* }}} */
  /* arguments (recursive) {{{ */
  if (in->args)  
  {
    t_uint32 tel;
    ret->args=Malloc(sizeof(t_ast_successors));
    ret->args->nargs=in->args->nargs;
    if (in->args->nargs) 
      ret->args->args=Malloc(in->args->nargs*sizeof(t_ast_node *));
    else ret->args->args=NULL;
    for (tel=0; tel<in->args->nargs; tel++)
    {
      ret->args->args[tel]=AstNodeDupInternal(in->args->args[tel],NULL);
    }
  }
  else
  {
    ret->args=NULL;
  }
  /* }}} */
  return ret;
}

t_ast_node * AstNodeDup(t_ast_node * in)
{
  return AstNodeDupInternal(in,NULL);
}



/*}}}*/
/* Linker rule manipulation {{{ */
t_linker_rule * LinkerRuleNew()
{
  t_linker_rule * ret=Malloc(sizeof(t_linker_rule));
  ret->action=ret->trigger=ret->section=ret->address=ret->symbol=NULL;
  return ret;
}

t_linker_rule * LinkerRuleDup(t_linker_rule * in)
{
  t_linker_rule * ret=Malloc(sizeof(t_linker_rule));
  ret->action=AstNodeDup(in->action);
  ret->trigger=AstNodeDup(in->trigger);
  ret->section=AstNodeDup(in->section);
  ret->address=AstNodeDup(in->address);
  ret->symbol=AstNodeDup(in->symbol);
  return ret;
}
/* }}} */

/* Callbacks for AST Execution {{{ */
/* The following functions are used during the execution of the AST's built by
 * the combination of the above functions, the lexer and the parser. They
 * implement the primitive functions for the Linker Description Language (LDL).
 * They are summarized in a table (FunctionTable), along with some type
 * information. */

/* FILLER FUNCTIONS {{{ */

void * FillerConst32(t_ast_successors * args,void * tsec)
{
  t_section * sec=tsec;
  t_uint32 i=(t_uint32) args->args[0]->data.data;
  if (sec==NULL) FATAL(("Section should be set when calling FillerConst32"));
  sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint32));
  memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint32));
  SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint32));
  SECTION_OLD_SIZE(sec)=SECTION_CSIZE(sec);
  return NULL;
}

void * FillerConst16(t_ast_successors * args,void * tsec)
{
  t_section * sec=tsec;
  t_uint16 i=(t_uint16) (t_uint32) args->args[0]->data.data;
  if (sec==NULL) FATAL(("Section should be set when calling FillerConst16"));
  sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint16));
  memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint16));
  SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint16));
  SECTION_OLD_SIZE(sec)=SECTION_CSIZE(sec);
  return NULL;
}

void * FillerConst8(t_ast_successors * args,void * tsec)
{
  t_section * sec=tsec;
  t_uint8 i=(t_uint8) (t_uint32) args->args[0]->data.data;
  if (sec==NULL) FATAL(("Section should be set when calling FillerConst8"));
  sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint8));
  memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint8));
  SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint8));
  SECTION_OLD_SIZE(sec)=SECTION_CSIZE(sec);
  return NULL;
}

void * FillerRelocated32(t_ast_successors * args,void * tsec)
{
  t_section * sec=tsec;
  t_uint32 i=(t_uint32) args->args[0]->data.data;
  t_string to_symbol=(t_string) args->args[1]->data.data;
  t_string assoc_symbol=(t_string) args->args[3]->data.data;
  t_string code=(t_string) args->args[5]->data.data;
  t_symbol * sym=NULL;
  t_symbol * assoc=NULL;
  if (sec==NULL) FATAL(("Section should be set when calling FillerConst32"));
  sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint32));
  memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint32));
  if (to_symbol && (strcmp(to_symbol,"MATCHED_NAME")==0))
    sym=SymbolTableAddUndefinedSymbol(ObjectGetFromCache("Linker",parser_object)->symbol_table,last_matched_symbol_name);
  else if (to_symbol)
    sym = SymbolTableAddUndefinedSymbol(ObjectGetFromCache("Linker",parser_object)->symbol_table,to_symbol);
  if (assoc_symbol)
    assoc = SymbolTableAddUndefinedSymbol(ObjectGetFromCache("Linker",parser_object)->symbol_table,assoc_symbol);

  RelocTableAddRelocToSymbol(sec->obj->reloc_table, NULL_GENERIC, T_RELOCATABLE(sec), SECTION_CSIZE(sec), sym, TRUE, NULL, NULL, assoc, code, TRUE);
  SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint32));
  SECTION_OLD_SIZE(sec)=SECTION_CSIZE(sec);
  return NULL;
}

void * FillerStartOfSection(t_ast_successors * args,void * tsec)
{
  t_string tmp;
  t_section * sec=tsec;
  t_string s=(t_string) args->args[0]->data.data;
  t_section * sec2 = SectionGetFromObjectByName(parser_object,s);

  if (sec==NULL) FATAL(("Section should be set when calling FillerStartOfSection"));
  if (!sec2) FATAL(("In Filler Start Of Section: section not found!")); 

  if (parser_object->description->address_size==ADDRSIZE32)
  {
    t_uint32 i=0;
    t_symbol * sym = SymbolTableAddUndefinedSymbol(ObjectGetFromCache("Linker",parser_object)->symbol_table,tmp=Collate3("$$Linker","_",s));
    Free(tmp);

    sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint32));
    memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint32));

    RelocTableAddRelocToSymbol(sec->obj->reloc_table, NULL_GENERIC, T_RELOCATABLE(sec), SECTION_CSIZE(sec), sym, FALSE, NULL, NULL, NULL, "S\\" WRITE_32, TRUE);
    SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint32));
  }
  else
  {
    t_uint64 i=0;
    t_symbol * sym = SymbolTableAddUndefinedSymbol(ObjectGetFromCache("Linker",parser_object)->symbol_table,Collate3("$$Linker","_",s));

    sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint64));
    memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint64));

    RelocTableAddRelocToSymbol(sec->obj->reloc_table, NULL_GENERIC, T_RELOCATABLE(sec), SECTION_CSIZE(sec), sym, FALSE, NULL, NULL, NULL, "S\\" WRITE_64, TRUE);
    SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint64));
  }

  SymbolTableAddSymbolToSection(parser_object->sub_symbol_table,GLOBAL_SYMBOL,T_RELOCATABLE(sec2),SECTION_GET_CADDRESS(sec2),tmp=Collate3("$$Linker","_",s),FALSE);

  SECTION_OLD_SIZE(sec) = SECTION_CSIZE(sec);

  Free(tmp);
  
  return NULL;
}

void * FillerEndOfSection(t_ast_successors * args,void * tsec)
{
  t_section * sec=tsec;
  t_string s=(t_string) args->args[0]->data.data;
  t_section * sec2 = SectionGetFromObjectByName(parser_object,s);
  t_symbol * sym = SymbolTableAddUndefinedSymbol(ObjectGetFromCache("Linker",parser_object)->symbol_table,Collate3("$$Linker","_endof_",s));

  if (sec==NULL) FATAL(("Section should be set when calling FillerEndOfSection"));
  if (!sec2) FATAL(("In Filler Size Of Section: section not found!")); 

  if (parser_object->description->address_size==ADDRSIZE32)
  {
    t_uint32 i=0;
    sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint32));
    memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint32));
    RelocTableAddRelocToSymbol(sec->obj->reloc_table, NULL_GENERIC, T_RELOCATABLE(sec), SECTION_CSIZE(sec), sym, FALSE, NULL, NULL, NULL, "SZ+\\" WRITE_32, TRUE);
    SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint32));
  }
  else
  {
    t_uint64 i=0;
    if (sec==NULL) FATAL(("Section should be set when calling FillerEndOfSection"));
    sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint64));
    memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint64));
    RelocTableAddRelocToSymbol(sec->obj->reloc_table, NULL_GENERIC, T_RELOCATABLE(sec), SECTION_CSIZE(sec), sym, FALSE, NULL, NULL, NULL, "SZ+\\" WRITE_64, TRUE);
    SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint64));
  }
  SymbolTableAddSymbolToSection(parser_object->sub_symbol_table,GLOBAL_SYMBOL,T_RELOCATABLE(sec2),SECTION_GET_CADDRESS(sec2),Collate3("$$Linker","_endof_",s),FALSE);
  SECTION_OLD_SIZE(sec) = SECTION_CSIZE(sec);
  return NULL;
}

void * FillerSizeOfSection(t_ast_successors * args,void * tsec)
{
  t_section * sec=tsec;
  t_string tmp;
  t_string s=(t_string) args->args[0]->data.data;
  t_section * sec2 = SectionGetFromObjectByName(parser_object,s);
  t_symbol * sym = SymbolTableAddUndefinedSymbol(ObjectGetFromCache("Linker",parser_object)->symbol_table,tmp=Collate3("$$Linker","_sizeof_",s));
  Free(tmp);

  if (sec==NULL) FATAL(("Section should be set when calling FillerSizeOfSection"));
  if (!sec2) FATAL(("In Filler Size Of Section: section not found!")); 

  if (parser_object->description->address_size==ADDRSIZE32)
  {
    t_uint32 i=0;
    if (sec==NULL) FATAL(("Section should be set when calling FillerStartOfSection"));
    sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint32));
    memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint32));
    RelocTableAddRelocToSymbol(sec->obj->reloc_table, NULL_GENERIC, T_RELOCATABLE(sec), SECTION_CSIZE(sec), sym, FALSE, NULL, NULL, NULL, "Z\\" WRITE_32, TRUE);
    SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint32));
  }
  else
  {
    t_uint64 i=0;
    if (sec==NULL) FATAL(("Section should be set when calling FillerStartOfSection"));
    sec->data=Realloc(sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)) + sizeof(t_uint64));
    memcpy(((char *) sec->data)+GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)),&i,sizeof(t_uint64));
    RelocTableAddRelocToSymbol(sec->obj->reloc_table, NULL_GENERIC, T_RELOCATABLE(sec), SECTION_CSIZE(sec), sym, FALSE, NULL, NULL, NULL, "Z\\" WRITE_64, TRUE);
    SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(SECTION_CSIZE(sec),sizeof(t_uint64));
  }
  SymbolTableAddSymbolToSection(parser_object->sub_symbol_table,GLOBAL_SYMBOL,T_RELOCATABLE(sec2),SECTION_GET_CADDRESS(sec2),tmp=Collate3("$$Linker","_sizeof_",s),FALSE);
  Free(tmp);
  SECTION_OLD_SIZE(sec) = SECTION_CSIZE(sec);
  return NULL;
}

/*END FILLER }}} */
/* TRIGGER FUNCTIONS {{{ */

void * TriggerAlways(t_ast_successors * args,void * ignored)
{
  return (void *) TRUE;
}

void * TriggerLinkedSymbolExists(t_ast_successors * args,void * ignored)
{
  static t_symbol * sym=NULL;
  t_uint32 tel;
  t_bool pattern=FALSE;
  char * name=args->args[0]->data.data;


  for (tel=0; tel<strlen(name); tel++)
  {
    if (name[tel]=='*') { pattern=TRUE; break; }
  }
 
  if (pattern)
  {
    if (!sym) sym=parser_object->symbol_table->first;

    for (; sym!=NULL; sym=sym->next) 
    {
      if (!sym->name) continue;
      if ((PatternMatch(name,sym->name)))
      {
	last_matched_symbol_name=StringDup(sym->name);
	last_matched_symbol=sym;
	sym=sym->next;
	return (void *) 2;
      }
    }
    return (void *) 0;
    
  }
  else
    FATAL(("Implement"));

   return (void *) 0; /* Keep compiler happy */
}

void * TriggerSymbolExists(t_ast_successors * args,void * ignored)
{
  static t_symbol * sym=NULL;
  t_uint32 tel;
  t_bool pattern=FALSE;
  char * name=args->args[0]->data.data;


  for (tel=0; tel<strlen(name); tel++)
  {
    if (name[tel]=='*') { pattern=TRUE; break; }
  }
 
  if (pattern)
  {
    if (!sym) sym=parser_object->sub_symbol_table->first;

    for (; sym!=NULL; sym=sym->next) 
    {
      if (!sym->name) continue;
      if ((PatternMatch(name,sym->name)))
      {
	last_matched_symbol_name=StringDup(sym->name);
	last_matched_symbol=sym;
	sym=sym->next;
	return (void *) 2;
      }
    }
    return (void *) 0;
    
  }
  else
    FATAL(("Implement"));
  
   return (void *) 0; /* Keep compiler happy */
}

void * TriggerUndefinedSymbol(t_ast_successors * args,void * ignored)
{
  t_symbol * msym;
  msym=SymbolTableGetSymbolByName(parser_object->sub_symbol_table,args->args[0]->data.data);
  if (!msym) return (void *) FALSE;
  if ((msym->type==UNDEFINED_SYMBOL) 
  || (msym->type==REMOVED_SYMBOL) )
    return (void *) TRUE;
  else
    return (void *) FALSE;
}

void * TriggerSubsectionExists(t_ast_successors * args,void * ignored)
{
  int tel;
    t_object * obj, * tmp;
    t_section * sec;
    t_bool subobject_pattern=FALSE;
    t_bool subsection_pattern=FALSE;
    char * subobject_name=args->args[0]->data.data;
    char * subsection_name=args->args[1]->data.data;

    for (tel=0; tel<strlen(subobject_name); tel++)
    {
      if (subobject_name[tel]=='*') { subobject_pattern=TRUE; break; }
    }

    for (tel=0; tel<strlen(subsection_name); tel++)
    {
      if (subsection_name[tel]=='*') { subsection_pattern=TRUE; break; }
    }

    if (subobject_pattern || subsection_pattern)
    {
      if (next_matched_object) tmp=next_matched_object;
      else tmp=parser_object->mapped_first;
      next_matched_object=NULL;

      for(; tmp!=NULL; tmp=tmp->next) 
      {
	if (next_matched_object2) obj=next_matched_object2;
	else obj=tmp;
	next_matched_object2=NULL;

	for (; obj!=NULL; obj=obj->equal)
	{
	  /*printf("Doing subsection exists %s\n",obj->name);*/
	  if (PatternMatch(subobject_name,obj->name))
	  {
	    tel=next_matched_section_tel;
	    if (tel && !next_matched_section) { tel+=0x100000;  tel&=~0xfffff; }
	    next_matched_section_tel=0;
	    for (; (tel&0x1f00000)<0x500000; tel+=0x100000, tel&=~0xfffff) 
	    {

	      if (next_matched_section)
	      {
		sec=next_matched_section;
		next_matched_section=NULL;
	      }
	      else
	      {
		sec=SECTIONARRAYBYNUMBER(obj,tel)?SECTIONARRAYBYNUMBER(obj,tel)[0]:NULL;
	      }

	      for (;
		  ((tel&0xfffff)<SECTIONCOUNTBYNUMBER(obj,tel)); 
		  tel++, sec=((tel&0xfffff)<SECTIONCOUNTBYNUMBER(obj,tel))?SECTIONARRAYBYNUMBER(obj,tel)[tel&0xfffff]:NULL)
	      {
		/* printf("Doing exists %s\n",sec->name);*/
		if (PatternMatch(subsection_name,sec->name))
		{
		  last_matched_object=obj;
		  next_matched_object=tmp;
		  next_matched_object2=obj;
		  last_matched_section=sec;
		  next_matched_section_tel=tel+1;
		  next_matched_section=((next_matched_section_tel&0xfffff)<SECTIONCOUNTBYNUMBER(obj,next_matched_section_tel))?SECTIONARRAYBYNUMBER(obj,next_matched_section_tel)[next_matched_section_tel&0xfffff]:NULL;
		  printf("Matched %s %s for pattern %s %s\n",obj->name,sec->name,subobject_name,subsection_name);
		  return (void *) 2;
		}
	      }
	    }
	  }
	}
      }
      return (void *) FALSE;

    }
    else
    {
      obj=ObjectGetFromCache(subobject_name,parser_object);
      if (!obj) return (void *) FALSE;
      sec=SectionGetFromObjectByName(obj,subsection_name);
      if (!sec) return (void *) FALSE;
      return (void *) TRUE;
    }
}

void * TriggerSectionExists(t_ast_successors * args,void * ignored)
{
    t_section * sec;
    sec=SectionGetFromObjectByName(parser_object,args->args[0]->data.data);
    if (!sec) return (void *) FALSE;
    return (void *) TRUE;
}

/* END TRIGGER }}} */
/* ACTION FUNCTIONS {{{ */

void *  SectionStringChop(t_ast_successors * args,void * rule)
{
  t_object * obj;
  t_section * sec, * parent;
  int tel=0;
  int nchops=0;
  int len;
  int total_len=0;
  int first_len=0;
  printf("Chopping %s %s\n",(char *)args->args[0]->data.data,(char *)args->args[1]->data.data);
  obj=ObjectGetFromCache(args->args[0]->data.data,parser_object);
  if (!obj) FATAL(("Could not find object %s to chop %s\n",args->args[0]->data.data ,args->args[1]->data.data));
  sec=SectionGetFromObjectByName(obj,args->args[1]->data.data);
  if (!sec) FATAL(("Could not find section %s in object %s to chop\n",args->args[1]->data.data ,args->args[0]->data.data));

 if (RELOCATABLE_REFERS_TO(sec)) FATAL(("String section %s:%s contains relocations!",args->args[0]->data.data ,args->args[1]->data.data));
  while (tel<GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)))
  {
    nchops++;
    len=0;
    for (; tel<GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)); tel++)
    {
      len++;
      total_len++;
      if (((char *)sec->data)[tel]==0)  break;
    }
    if (nchops!=1)
    {
      char buffer[80];
      t_section * string_sec;
      t_reloc_ref * iter;
      t_symbol * sym;
      string_sec=SectionNew(obj,sec->type);
      sprintf(buffer,".stringified.%03d%s",nchops,sec->name);
      string_sec->name=StringDup(buffer);
      string_sec->data=Malloc(len); 
      strcpy(string_sec->data,((char *)sec->data)+total_len-len);
      SectionInsert(string_sec);
      T_RELOCATABLE(string_sec)->type=RT_SUBSECTION;
      SectionSetOldAddressSuper(string_sec,GENERIC_ADD_UINT32(SECTION_GET_OLD_ADDRESS(sec),total_len-len));
      SectionSetCAddressSuper(string_sec,GENERIC_ADD_UINT32(SECTION_GET_CADDRESS(sec),total_len-len));


      SECTION_CSIZE(string_sec)=GENERIC_ADD_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(sec),SECTION_GET_CADDRESS(sec)),len);
      SECTION_OLD_SIZE(string_sec)=SECTION_CSIZE(string_sec);

      
      parent=SectionGetFromObjectByAddress(parser_object,SECTION_GET_CADDRESS(string_sec));
      MAP_SECTION(string_sec,SectionGetFromObjectByAddress(parser_object,SECTION_GET_CADDRESS(string_sec))->name, SectionGetFromObjectByAddress(parser_object,SECTION_GET_CADDRESS(string_sec)));

      SECTION_ALIGN(string_sec)=SECTION_ALIGN(sec);
      HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(string_sec->name,string_sec));


      for (iter = RELOCATABLE_REFED_BY(sec); iter!=NULL; iter=iter->next)
      {
	if ((GENERIC_EXTRACT_UINT32(REL_TO_SYMBOL(iter->rel)->offset_from_start)<total_len)
	    && (GENERIC_EXTRACT_UINT32(REL_TO_SYMBOL(iter->rel)->offset_from_start)>=(total_len-len)))
	{
	  RelocSetToSymbolBase(iter->rel,T_RELOCATABLE(string_sec));
	  REL_TO_SYMBOL(iter->rel)->offset_from_start = 
	    GENERIC_SUB(REL_TO_SYMBOL(iter->rel)->offset_from_start,GENERIC_SUB(SECTION_GET_CADDRESS(string_sec),SECTION_GET_CADDRESS(sec)));
	  REL_TO_SYMBOL(iter->rel)->mapped->base=T_RELOCATABLE(string_sec);
	  REL_TO_SYMBOL(iter->rel)->mapped->to_type=TO_SUB_SECTION;
	  REL_TO_SYMBOL(iter->rel)->mapped->offset_from_start=REL_TO_SYMBOL(iter->rel)->offset_from_start;
	}
      }
      /* now do the symbols */
      /* symbols from the subobject */
      for (sym = obj->symbol_table->first; sym; sym = sym->next)
      {
	if (T_SECTION(sym->base)==sec)
	{
	  if ((GENERIC_EXTRACT_UINT32(sym->offset_from_start)<total_len)
	      && (GENERIC_EXTRACT_UINT32(sym->offset_from_start)>=(total_len-len)))
	  {
	    sym->base = T_RELOCATABLE(string_sec);
	    sym->offset_from_start = GENERIC_SUB(sym->offset_from_start,GENERIC_SUB(SECTION_GET_CADDRESS(string_sec),SECTION_GET_CADDRESS(sec)));
	    sym->mapped->base=T_RELOCATABLE(string_sec);
	    sym->mapped->to_type=TO_SUB_SECTION;
	    sym->mapped->offset_from_start=sym->offset_from_start;
	  }
	}
      }

      /* Should be no problem ... It's a data section */
#if 0
      /* symbols only found in the subsymbol table of the parent object (some $a and $d symbols) */
      for (sym = parser_object->sub_symbol_table->first; sym; sym = sym->next)
      {
	if (T_SECTION(sym->base)==sec)
	{
	  sym->base = T_RELOCATABLE(vector_sec);
	  sym->offset_from_start = GENERIC_ADD(sym->offset_from_start,GENERIC_SUB(SECTION_GET_CADDRESS(sec),SECTION_GET_CADDRESS(vector_sec)));
	  VERIFY_FASSERT(GENERIC_EQ(sym->current_address,GENERIC_ADD(sym->base->caddress,sym->offset_from_start)),("Symbols corrupt"));
	  if (sym->mapped) FATAL(("@S is mapped and not changed",sym));
	}
      }


#endif
    }
    else
    {
      first_len=len;
    }
    tel++;
  }

  SECTION_CSIZE(sec)=GENERIC_ADD_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(sec),SECTION_GET_CADDRESS(sec)),first_len);
  SECTION_OLD_SIZE(sec)=SECTION_CSIZE(sec);
  

  printf("Chopping in %d pieces\n",nchops);


  
  return NULL;
}

void *  Vectorize(t_ast_successors * args,void * rule)
{
  t_object * tmp, *obj, *linker_obj;
  t_section * sec, * vector_sec, * parent;
  char type=0;
  t_address total_size=NULL_GENERIC, start_of_vector=NULL_GENERIC;


  /* Get or create the linker object {{{ */
  linker_obj=ObjectGetFromCache("Linker",parser_object);
  if (!linker_obj) 
  {
    linker_obj=ObjectNewCached("Linker",parser_object);
    linker_obj->section_table=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,SectionHashNodeFree);
    parser_object->mapped_last->next=linker_obj;
    linker_obj->next=NULL;
    parser_object->mapped_last=linker_obj;
    linker_obj->reloc_table=RelocTableNew(); 
    linker_obj->symbol_table=SymbolTableNew(linker_obj);
  }
  /* }}} */

  /* Calculate the size of the Linker section, the start address and the type {{{ */
  for(tmp=parser_object->mapped_first; tmp!=NULL; tmp=tmp->next)
  {
    for (obj=tmp; obj!=NULL; obj=obj->equal)
    {
      sec=SectionGetFromObjectByName(obj,args->args[0]->data.data);
      if (sec) 
      {
	if (type)
	{
	  VERIFY_FASSERT(type==sec->type,("Need to vectorize sections of different types... Implement"));
	  total_size=GENERIC_ADD(total_size,SECTION_CSIZE(sec)); 
	  if (GENERIC_LT(SECTION_GET_CADDRESS(sec),start_of_vector))
	    start_of_vector=SECTION_GET_CADDRESS(sec);
	}
	else 
	{
	  type=sec->type;
	  total_size=SECTION_CSIZE(sec); 
	  start_of_vector=SECTION_GET_CADDRESS(sec);
	}
      }
    }
  }
  /* }}} */

  /* Create a new section (the vector_sec) {{{ */
  switch(type)
  {
    case CODE_SECTION:
      vector_sec=SectionNew(linker_obj,CODE_SECTION);
      T_RELOCATABLE(vector_sec)->type=RT_SUBSECTION;
      vector_sec->name=Collate("VECTOR___",args->args[0]->data.data);
      linker_obj->code=Realloc(linker_obj->code,(linker_obj->ncodes+1)*sizeof(t_section *));
      linker_obj->code[linker_obj->ncodes]=vector_sec;
      linker_obj->ncodes++;
      break;
    case RO_DATA_SECTION:
      vector_sec=SectionNew(linker_obj,RO_DATA_SECTION);
      T_RELOCATABLE(vector_sec)->type=RT_SUBSECTION;
      vector_sec->name=Collate("VECTOR___",args->args[0]->data.data);
      linker_obj->rodata=Realloc(linker_obj->rodata,(linker_obj->nrodatas+1)*sizeof(t_section *));
      linker_obj->rodata[linker_obj->nrodatas]=vector_sec;
      linker_obj->nrodatas++;
      break;
   case DATA_SECTION:
      vector_sec=SectionNew(linker_obj,DATA_SECTION);
      T_RELOCATABLE(vector_sec)->type=RT_SUBSECTION;
      vector_sec->name=Collate("VECTOR___",args->args[0]->data.data);
      linker_obj->data=Realloc(linker_obj->data,(linker_obj->ndatas+1)*sizeof(t_section *));
      linker_obj->data[linker_obj->ndatas]=vector_sec;
      linker_obj->ndatas++;
      break;
    default:
      FATAL(("Unknown section type for vector: %c!",type));
      exit(0); /* Keeps the compiler happy */
  }

  vector_sec->data=Malloc(GENERIC_EXTRACT_UINT32(total_size));
  SECTION_OLD_SIZE(vector_sec)=total_size;
  SECTION_CSIZE(vector_sec)=total_size;
  SectionSetOldAddressSuper(vector_sec,start_of_vector);
  SectionSetCAddressSuper(vector_sec,start_of_vector);
  /* }}} */

  /* Copy all information of the original sections to the vector and delete the originals {{{*/
  OBJECT_FOREACH_SUBOBJECT(parser_object,obj,tmp)
  {
    t_symbol * sym;

    sec=SectionGetFromObjectByName(obj,args->args[0]->data.data);
    if (sec) 
    {
      t_reloc_ref * iter, * itern;
      t_uint32 sz = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec));

      /* Take care of the data */
      if (sz)
      {
	t_uint32 offset = GENERIC_EXTRACT_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(sec),start_of_vector));
	DiabloPrint(stdout,"Offset = 0x%x, sec size is @G\n", offset, SECTION_CSIZE(vector_sec)); fflush(stdout);
	memcpy(((char *) vector_sec->data)+offset,(char *) sec->data,sz);
      }
      /* Now do the relocs */

      iter = RELOCATABLE_REFERS_TO(sec);

      while (iter)
      {
	itern=iter->next;
	iter->rel->from_offset=GENERIC_ADD_UINT32(iter->rel->from_offset,GENERIC_EXTRACT_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(sec),start_of_vector)));
	RelocSetFrom(iter->rel,T_RELOCATABLE(vector_sec));
	iter=itern;
      }

      /* TODO: This seem like a useless instruction */
      iter = RELOCATABLE_REFED_BY(sec);

      while ((iter = RELOCATABLE_REFED_BY(sec)))
      {
	RelocSetToSymbolBase(iter->rel,T_RELOCATABLE(vector_sec));
	REL_TO_SYMBOL(iter->rel)->offset_from_start = GENERIC_ADD(REL_TO_SYMBOL(iter->rel)->offset_from_start,GENERIC_SUB(SECTION_GET_CADDRESS(sec),SECTION_GET_CADDRESS(vector_sec)));
	REL_TO_SYMBOL(iter->rel)->mapped->base=T_RELOCATABLE(vector_sec);
	REL_TO_SYMBOL(iter->rel)->mapped->to_type=TO_SUB_SECTION;
	REL_TO_SYMBOL(iter->rel)->mapped->offset_from_start=REL_TO_SYMBOL(iter->rel)->offset_from_start;
      }

      /* now do the symbols */
      /* symbols from the subobject */
      for (sym = obj->symbol_table->first; sym; sym = sym->next)
      {
	if (T_SECTION(sym->base)==sec)
	{
	  sym->base = T_RELOCATABLE(vector_sec);
	  sym->offset_from_start = GENERIC_ADD(sym->offset_from_start,GENERIC_SUB(SECTION_GET_CADDRESS(sec),SECTION_GET_CADDRESS(vector_sec)));
	  sym->mapped->base=T_RELOCATABLE(vector_sec);
	  sym->mapped->to_type=TO_SUB_SECTION;
	  sym->mapped->offset_from_start=GENERIC_ADD(sym->mapped->offset_from_start,GENERIC_SUB(SECTION_GET_CADDRESS(sec),SECTION_GET_CADDRESS(vector_sec)));
	}
      }

      /* symbols only found in the subsymbol table of the parent object (some $a and $d symbols) */
      for (sym = parser_object->sub_symbol_table->first; sym; sym = sym->next)
      {
	if (T_SECTION(sym->base)==sec)
	{
	  sym->base = T_RELOCATABLE(vector_sec);
	  sym->offset_from_start = GENERIC_ADD(sym->offset_from_start,GENERIC_SUB(SECTION_GET_CADDRESS(sec),SECTION_GET_CADDRESS(vector_sec)));
	  VERIFY_FASSERT(GENERIC_EQ(sym->current_address,GENERIC_ADD(sym->base->caddress,sym->offset_from_start)),("Symbols corrupt"));
	  if (sym->mapped) FATAL(("@S is mapped and not changed",sym));
	}
      }

      /* Delete this section */
      {
	SECTION_GET_CADDRESS(sec)=GENERIC_SUB(SECTION_CSIZE(sec),SECTION_CSIZE(sec));
	SECTION_GET_OLD_ADDRESS(sec)=GENERIC_SUB(SECTION_CSIZE(sec),SECTION_CSIZE(sec));
	SECTION_CSIZE(sec)=GENERIC_SUB(SECTION_CSIZE(sec),SECTION_CSIZE(sec));
	SECTION_OLD_SIZE(sec)=GENERIC_SUB(SECTION_OLD_SIZE(sec),SECTION_OLD_SIZE(sec));
	if (sec->obj->section_table)
	{
	  t_section_hash_node * hn=(t_section_hash_node *) HashLookup(sec->obj->section_table,sec->name);
	  if (hn) hn->section=NULL;
	}
	Free(sec->name);


	sec->name=StringDup("DELETED_BY_VECTORIZE");
#if 0
	t_uint32 tel=0;
	switch(sec->type)
	{
	  case CODE_SECTION:
	    {
	      for (tel=0; tel<obj->ncodes;tel++)
		if (obj->code[tel]==sec)
		  break;
	      for (tel=tel+1; tel<obj->ncodes;tel++)
		obj->code[tel-1]=obj->code[tel];
	      obj->ncodes--;
	    }
	    break;
	  case RO_DATA_SECTION:
	    {
	      for (tel=0; tel<obj->nrodatas;tel++)
		if (obj->rodata[tel]==sec)
		  break;
	      for (tel=tel+1; tel<obj->nrodatas;tel++)
		obj->rodata[tel-1]=obj->rodata[tel];
	      obj->nrodatas--;
	    }
	    break;

	  case DATA_SECTION:
	    {
	      for (tel=0; tel<obj->ndatas;tel++)
		if (obj->data[tel]==sec)
		  break;
	      for (tel=tel+1; tel<obj->ndatas;tel++)
		obj->data[tel-1]=obj->data[tel];
	      obj->ndatas--;
	    }
	    break;

	  default:
	    FATAL(("Unknown section type for vector: %c!",sec->type));
	}
	SectionFree(sec);
#endif
      }
    }
  }
  /* }}} */

  parent=SectionGetFromObjectByAddress(parser_object,SECTION_GET_CADDRESS(vector_sec));
  SECTION_ALIGN(vector_sec)=SECTION_ALIGN(parent);
  MAP_SECTION(vector_sec,parent->name, parent);

  HashInsert(linker_obj->section_table,(data_hash_node *) SectionHashNodeNew(vector_sec->name,vector_sec));

  /*DiabloPrint(stdout,"Vector %s size = @G located at @G, Vector type =%c\n",vector_sec->name,total_size,start_of_vector,type);  */

  return NULL;
}

void * AddSubsection(t_ast_successors * args,void * rule)
{
    t_object * obj;
    t_section * sec, *parent;
    t_address x;
    t_ast_node * node;
    obj=ObjectGetFromCache(args->args[0]->data.data,parser_object);
    if (!obj) 
    {
      obj=ObjectNewCached(args->args[0]->data.data,parser_object);
      obj->section_table=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,SectionHashNodeFree);

      parser_object->mapped_last->next=obj;
      obj->next=NULL;
      parser_object->mapped_last=obj;
      obj->reloc_table=RelocTableNew(); 
      obj->symbol_table=SymbolTableNew(obj);
    }

    if (strcmp(args->args[2]->data.data,"CODE")==0)
    {
      sec=SectionNew(obj,CODE_SECTION);
      T_RELOCATABLE(sec)->type=RT_SUBSECTION;
      sec->name=StringDup(args->args[1]->data.data);
      obj->code=Realloc(obj->code,(obj->ncodes+1)*sizeof(t_section *));
      obj->code[obj->ncodes]=sec;
      obj->ncodes++;
    }
    else if (strcmp(args->args[2]->data.data,"DATA")==0)
    {
      sec=SectionNew(obj,DATA_SECTION);
      T_RELOCATABLE(sec)->type=RT_SUBSECTION;
      sec->name=StringDup(args->args[1]->data.data);
      obj->data=Realloc(obj->data,(obj->ndatas+1)*sizeof(t_section *));
      obj->data[obj->ndatas]=sec;
      obj->ndatas++;
    }
    else if (strcmp(args->args[2]->data.data,"RODATA")==0)
    {
      sec=SectionNew(obj,DATA_SECTION);
      T_RELOCATABLE(sec)->type=RT_SUBSECTION;
      sec->name=StringDup(args->args[1]->data.data);
      obj->rodata=Realloc(obj->rodata,(obj->nrodatas+1)*sizeof(t_section *));
      obj->rodata[obj->nrodatas]=sec;
      obj->nrodatas++;
    }
    else if (strcmp(args->args[2]->data.data,"BSS")==0)
    {
      sec=SectionNew(obj,BSS_SECTION);
      T_RELOCATABLE(sec)->type=RT_SUBSECTION;
      sec->name=StringDup(args->args[1]->data.data);
      obj->bss=Realloc(obj->bss,(obj->nbsss+1)*sizeof(t_section *));
      obj->bss[obj->nbsss]=sec;
      obj->nbsss++;
    }
    else 
    {
      FATAL(("Third argument to function ADD_SUBSECTION should be unquoted CODE, DATA, RODATA or BSS"));
      exit(0); /* Keeps the compiler happy */
    }
  
    if (parser_object->description->address_size==ADDRSIZE32)
    {
#ifdef BIT32ADDRSUPPORT
      GenericSet_uint32(SECTION_OLD_SIZE(sec),0);
#else
    FATAL(("Needed 32-bit addresses but 32bit address support was not enabled!"));
    exit(0); /* Keeps the compiler happy */
#endif
    }
    else
    {
#ifdef BIT64ADDRSUPPORT
      GenericSet_uint64(SECTION_OLD_SIZE(sec),0);
#else
    FATAL(("Needed 64-bit addresses but 64bit address support was not enabled!"));
    exit(0); /* Keeps the compiler happy */
#endif
    }

    printf("inserted subsection %s %s\n",(char *)args->args[0]->data.data,(char *)args->args[1]->data.data);

    if (((t_linker_rule *) rule)->section==NULL) FATAL(("Rule has no section data!"));
    if (((t_linker_rule *) rule)->address==NULL) FATAL(("Rule has no address for section!"));
    AstExecute(((t_linker_rule *) rule)->section,FunctionTable,sec,current_free_mode);
    if (current_free_mode) ((t_linker_rule *) rule)->section=NULL;

    node=AstExecute(((t_linker_rule *) rule)->address,AddressTable,NULL,current_free_mode);
    x = node->data.addr;
    FreeAstNode(node);
    if (current_free_mode) ((t_linker_rule *) rule)->address=NULL;
    SectionSetOldAddressSuper(sec,x);
    SectionSetCAddressSuper(sec,x);
    parent=SectionGetFromObjectByAddress(parser_object,SECTION_GET_CADDRESS(sec));
    MAP_SECTION(sec,SectionGetFromObjectByAddress(parser_object,SECTION_GET_CADDRESS(sec))->name, SectionGetFromObjectByAddress(parser_object,SECTION_GET_CADDRESS(sec)));

    end_of_last_added=GENERIC_ADD(SECTION_GET_CADDRESS(sec),SECTION_CSIZE(sec));

    SECTION_ALIGN(sec)=GENERIC_ADD_UINT32(SECTION_ALIGN(sec),args->args[3]->data.data);
    HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(sec->name,sec));
    return NULL;
}

void * AddLocalSymbol(t_ast_successors * args,void * rule)
{
  t_ast_node * ret;
  if (!((t_linker_rule *)  rule)->symbol) FATAL(("Missing symbol information for ADD_SYMBOL"));
  symbol_is_local=TRUE;
  ret=AstExecute(((t_linker_rule *) rule)->symbol,SymbolFunctionTable,args->args[0]->data.data,current_free_mode);
  FreeAstNode(ret);
  ((t_linker_rule *) rule)->symbol=NULL;
  return NULL;
}

void * AddSymbol(t_ast_successors * args,void * rule)
{
  t_ast_node * ret;
  if (!((t_linker_rule *)  rule)->symbol) FATAL(("Missing symbol information for ADD_SYMBOL"));
  symbol_is_local=FALSE;
  ret=AstExecute(((t_linker_rule *) rule)->symbol,SymbolFunctionTable,args->args[0]->data.data,current_free_mode);
  FreeAstNode(ret);
  ((t_linker_rule *) rule)->symbol=NULL;
  return NULL;
}
/* END ACTION FUNCTIONS }}} */
/* ADDRESS FUNCTIONS {{{ */
void * Symbol(t_ast_successors * args,void * rule)
{
  t_symbol * sym;
  t_string name=args->args[0]->data.name;
  t_address * ret=Malloc(sizeof(t_address));
  if (strcmp(name,"MATCHED_NAME")==0)
    name=last_matched_symbol_name;
  sym=SymbolTableGetSymbolByName(parser_object->symbol_table,name);
  if (!sym) FATAL(("Symbol %s not found!",name));
  *ret=sym->current_address;
  return ret;
}

void * StartOfSectionAddress(t_ast_successors * args,void * name)
{
  t_address * ret=Malloc(sizeof(t_address));
  t_section * sec;
  sec=SectionGetFromObjectByName(parser_object,args->args[0]->data.data);
  if (!sec) FATAL(("In Start Of Section: section not found!")); 

  *ret=SECTION_GET_CADDRESS(sec);
  return ret; 
}

void * EndOfLastAdded(t_ast_successors * args,void * rule)
{
  t_address * ret=Malloc(sizeof(t_address));
  *ret=end_of_last_added;
  return  ret;
}
/* }}} */
/* SYMBOL FUNCTION {{{ */
void * StartOfSubsection(t_ast_successors * args,void * name)
{
  t_object * obj;
  t_section * sec;
  obj=ObjectGetFromCache(args->args[0]->data.data,parser_object);
  if (!obj) FATAL(("In Start Of Subsection: object not found!")); 
  sec=SectionGetFromObjectByName(obj,args->args[1]->data.data);
  if (!sec) FATAL(("In Start Of Subsection: section (%s in %s) not found!",args->args[1]->data.data,args->args[0]->data.data)); 
 
  return SymbolTableAddSymbolToSubSection(parser_object->sub_symbol_table,GLOBAL_SYMBOL,T_RELOCATABLE(sec),SECTION_GET_CADDRESS(sec),(name),symbol_is_local);

}

void * EndOfSubsection(t_ast_successors * args,void * name)
{
  t_object * obj;
  t_section * sec;
  t_symbol * sym;
  obj=ObjectGetFromCache(args->args[0]->data.data,parser_object);
  if (!obj) FATAL(("In Start Of Subsection: object not found!")); 
  sec=SectionGetFromObjectByName(obj,args->args[1]->data.data);
  if (!sec) FATAL(("In Start Of Subsection: section not found!")); 

  sym=SymbolTableAddSymbolToSubSection(parser_object->sub_symbol_table,END_SYMBOL,T_RELOCATABLE(sec),SECTION_GET_CADDRESS(sec),(name),symbol_is_local); 
  sym->addend=SECTION_CSIZE(sec);
  return sym;
}

void * EndOfSection(t_ast_successors * args,void * name)
{
  t_section * sec;
  sec=SectionGetFromObjectByName(parser_object,args->args[0]->data.data);
  if (!sec) FATAL(("In Start Of Section: section not found!")); 
 
  return SymbolTableAddSymbolToSection(parser_object->sub_symbol_table,SECTION_END_SYMBOL,T_RELOCATABLE(sec),GENERIC_ADD(SECTION_GET_CADDRESS(sec),SECTION_CSIZE(sec)),(name),symbol_is_local);
}

void * StartOfSection(t_ast_successors * args,void * name)
{
  t_section * sec;
  sec=SectionGetFromObjectByName(parser_object,args->args[0]->data.data);
  if (!sec) FATAL(("In Start Of Section: section not found!")); 
 
  return SymbolTableAddSymbolToSection(parser_object->sub_symbol_table,GLOBAL_SYMBOL,T_RELOCATABLE(sec),SECTION_GET_CADDRESS(sec),(name),symbol_is_local);
}


void * SizeOfSubsection(t_ast_successors * args,void * name)
{
  t_object * obj;
  t_section * sec;
  obj=ObjectGetFromCache(args->args[0]->data.data,parser_object);
  if (!obj) FATAL(("In Start Of Subsection: object not found!")); 
  sec=SectionGetFromObjectByName(obj,args->args[1]->data.data);
  if (!sec) FATAL(("In Start Of Subsection: section (%s in %s) not found!",args->args[1]->data.data,args->args[0]->data.data)); 
 
  return SymbolTableAddSymbolToSubSection(parser_object->sub_symbol_table,0,T_RELOCATABLE(sec),SECTION_GET_CADDRESS(sec),(name),symbol_is_local);

}

void * Offset(t_ast_successors * args,void * name)
{
  return args->args[0]->data.data;
}

void * AbsSymbol(t_ast_successors * args,void * name)
{
  t_address tmp;
  t_symbol * sym;
  
  if (parser_object->description->address_size==ADDRSIZE32)
  {
#ifdef BIT32ADDRSUPPORT
    GenericSet_uint32(tmp,(t_uint32) args->args[0]->data.data);
#else
    FATAL(("Needed 32-bit addresses but 32bit address support was not enabled!"));
    exit(0); /* Keeps the compiler happy */
#endif
  }
  else
  {
#ifdef BIT64ADDRSUPPORT
    GenericSet_uint64(tmp,(t_uint64) POINTER_TO_INT(args->args[0]->data.data));
#else
    FATAL(("Needed 64-bit addresses but 64bit address support was not enabled!"));
    exit(0); /* Keeps the compiler happy */
#endif
  }

 
  sym = SymbolTableAddAbsoluteSymbol(parser_object->sub_symbol_table,tmp,name,symbol_is_local); 

  return sym;
}
/*}}} */

void * UtilConcat(t_ast_successors * args,void * name)
{
  return Collate(args->args[0]->data.data, args->args[1]->data.data);
}

void * MatchedName(t_ast_successors * args,void * name)
{
  return StringDup(last_matched_symbol_name);
}

void * MatchedSubobjectName(t_ast_successors * args,void * name)
{
  if (last_matched_object) return StringDup(last_matched_object->name); 
  return NULL;
}

void * MatchedSubsectionName(t_ast_successors * args,void * name)
{
  if (last_matched_section) return StringDup(last_matched_section->name); 
  return NULL;
}

void * UtilSubString(t_ast_successors * args,void * name)
{
  t_string in = args->args[0]->data.data;
  t_uint32 from = (t_uint32) args->args[1]->data.data;
  t_uint32 to = (t_uint32) args->args[2]->data.data;
  t_string ret=StringDup(in+from);
  if (to) ret[to]='\0';
  return ret;
}

/* The tables describing all callbacks {{{*/
/* Types:
 *
 * A: Address
 * B: Bool
 * E: sEction
 * O: other ((unquoted) string that's casted to a specific type by the called function)
 * S: String
 * T: Trigger: 0 = trigger false, 1 = trigger true, 2 = trigger true, but needs another run 
 * V: Void
 * Y: sYmbol
 */

t_ast_node_table_entry  FunctionTable[]=
{
  /* Trigger (boolean) functions */
  {"ALWAYS",           0,  NULL,   "T" , TriggerAlways},
  {"UNDEFINED_SYMBOL", 1, "S",     "T",  TriggerUndefinedSymbol},
  {"SYMBOL_EXISTS"   , 1, "S",     "T",  TriggerSymbolExists},
  {"LINKED_SYMBOL_EXISTS"   , 1, "S",     "T",  TriggerLinkedSymbolExists},
  {"SUBSECTION_EXISTS",2, "SS",    "T",  TriggerSubsectionExists},
  {"SECTION_EXISTS",   1, "S",     "T",  TriggerSectionExists},
  /* Filler functions */
  {"CONST32",          1, "N",     "V",  FillerConst32},
  {"CONST16",          1, "N",     "V",  FillerConst16},
  {"CONST8",           1, "N",     "V",  FillerConst8},
  {"RELOCATED32",      6, "NSNSNS","V",  FillerRelocated32},
  {"START_OF_SECTION", 1, "S",     "V",  FillerStartOfSection},
  {"END_OF_SECTION",   1, "S",     "V",  FillerEndOfSection},
  {"SIZE_OF_SECTION",  1, "S",     "V",  FillerSizeOfSection},
  /* Action functions */
  {"ADD_SUBSECTION",   4, "SSON",   "E",  AddSubsection},
  {"ADD_SYMBOL",       1, "S",     "Y",  AddSymbol},
  {"ADD_LOCAL_SYMBOL",       1, "S",     "Y",  AddLocalSymbol},
  {"VECTORIZE",        1, "S",     "E",  Vectorize},
  {"STRINGCHOP",       2, "SS",     "E",  SectionStringChop}
};

t_ast_node_table_entry AddressTable[]={
  /* Address functions */
  {"SYMBOL",           1, "S",     "A",  Symbol},
  {"START_OF_SECTION", 1, "S",     "A",  StartOfSectionAddress},
  {"END_OF_LAST_ADDED",0,  NULL, "A",  EndOfLastAdded}
};

/* Symbol functions */
t_ast_node_table_entry  SymbolFunctionTable[]=
{
  {"START_OF_SUBSECTION", 2, "SS", "Y", StartOfSubsection},
  {"END_OF_SUBSECTION", 2, "SS", "Y", EndOfSubsection},
  {"SIZE_OF_SUBSECTION", 2, "SS", "Y", SizeOfSubsection},
  {"ABS",	         1, "N",  "Y", AbsSymbol},
  {"OFFSET",	         1, "N",  "N", Offset},
  {"START_OF_SECTION", 1, "S", "Y", StartOfSection}, 
  {"END_OF_SECTION", 1, "S", "Y", EndOfSection}, 
  {NULL,               0, NULL,  NULL, NULL}
};


t_ast_node_table_entry UtilityFunctions[]=
{
  { "CONCAT", 2, "SS", "S", UtilConcat},
  { "MATCHED_NAME", 0 , NULL , "S", MatchedName},
  { "MATCHED_SUBSECTION_NAME", 0 , NULL , "S", MatchedSubsectionName},
  { "MATCHED_SUBOBJECT_NAME", 0 , NULL , "S", MatchedSubobjectName},
  { "SUBSTRING", 3 , "SNN", "S", UtilSubString},
  {  NULL,               0, NULL,  NULL, NULL}
};

/*}}}*/

/* }}} */

/* AST Execution {{{ */

/* The following functions implement the real execution of the AST's. They use
 * the functions in the tables, and knowledge about operators to collapse an
 * AST into one atomic  ast node. */


t_ast_node_table_entry * FunctionTableGetFunction (t_ast_node_table_entry  * function_table,t_string name)
{
  t_ast_node_table_entry  * fun=function_table;
  while(fun->name)
  {
      if (strcmp(fun->name,name)==0)
      {
	return fun;
      }
      fun++;
  }
  return NULL;
}

/* AstExecute {{{ */
/*! This function recursivly collapses the arguments of the node, and then the
 * node itself. */

t_ast_node * AstExecuteREAL(char * file, int lnno, t_ast_node * in, t_ast_node_table_entry * function_table, 
    void * implicit_argument,t_bool free_mode)
{
  current_free_mode=free_mode;
  /* If we have a function */
  if (in->type==TYPE_FUNCTION)
  {
    t_ast_successors * args=in->args;
    t_string name=in->data.name;
    t_ast_node_table_entry  * fun=FunctionTableGetFunction(function_table,name);
    void * tmp;
    t_ast_node * ret=Malloc(sizeof(t_ast_node));
    if (!fun) fun=FunctionTableGetFunction(UtilityFunctions,name);
    if (!fun) FATAL(("Function %s (call at %s %d) not found!",name,file,lnno));
    if (fun->fun==NULL) FATAL(("Function %s is not yet implemented",name));
    
    CheckAndCastArguments(in,fun->nargs,fun->arguments,function_table,implicit_argument);
    
    tmp=fun->fun(args,implicit_argument);
    FreeAstNode(in);
    ret->args=NULL;
    switch(fun->ret[0])
    {
      case 'V':
	ret->type=TYPE_VOID;
	break;
      case 'T':
	ret->type=TYPE_TRIGGER;
	ret->data.data=tmp;
	break;
      case 'B':
	ret->type=TYPE_BOOL;
	ret->data.data=tmp;
	break;
      case 'A':
	ret->type=TYPE_ADDRESS;
	ret->data.addr=*((t_address *) tmp);
	Free(tmp);
	break;
      case 'E':
	ret->type=TYPE_SECTION;
	ret->data.data=tmp;
	break;
      case 'Y':
	ret->type=TYPE_SYMBOL;
	ret->data.data=tmp;
	break;
      case 'N':
	ret->type=TYPE_NUMERIC;
	ret->data.data=tmp;
	break;
      case 'S':
	ret->type=TYPE_STRING;
	ret->data.data=tmp;
	break;
      default:
	FATAL(("Implement type %c\n",fun->ret[0]));
    }
    return ret;
  }
  /* If we have an operator */
  else if (in->type==TYPE_OPER)
  {
    switch(in->data.oper)
    {
      case OPER_COMMA:
	CheckAndCastOneArgument("COMMA",0,&(in->args->args[0]),'V',function_table,implicit_argument);
	CheckAndCastOneArgument("COMMA",1,&(in->args->args[1]),'V',function_table,implicit_argument);
	FreeAstNode(in);
	return NULL;
	break;
      case OPER_AND:
	{
	  t_ast_node * ret=Malloc(sizeof(t_ast_node));
	  ret->type=TYPE_BOOL;

	  CheckAndCastOneArgument("AND",0,&(in->args->args[0]),'B',function_table,implicit_argument);
	  if (!((t_bool) in->args->args[0]->data.data)) 
	  {
	    ret->data.data=(void *) FALSE;
	    ret->args=NULL;
	    FreeAstNode(in);
	    return ret;
	  }
	  else 
	  {
	    CheckAndCastOneArgument("AND",1,&(in->args->args[1]),'B',function_table,implicit_argument);
	    ret->data.data=in->args->args[1]->data.data;
	    ret->args=NULL;
	    FreeAstNode(in);
	    return ret; 
	  }

	  FATAL(("Unreachable!"));
	  break;
	}
      case OPER_ADD:
	{
	  t_ast_node * ret=Malloc(sizeof(t_ast_node));
	  t_ast_node * left=in->args->args[0];
	  t_ast_node * right=in->args->args[1];

	  if (left->type==TYPE_FUNCTION)
	  {
	    in->args->args[0]=left=AstExecute(left,function_table,implicit_argument,FALSE);
	  }

	  if (right->type==TYPE_FUNCTION)
	  {
	    in->args->args[1]=right=AstExecute(right,function_table,implicit_argument,FALSE);
	  }

	  if ((left->type==TYPE_SYMBOL)&&(right->type==TYPE_NUMERIC))
	  {
	    t_symbol * sym=left->data.data;
	    sym->offset_from_start=GENERIC_ADD_UINT32(sym->offset_from_start,right->data.data);
	    ret->type=TYPE_SYMBOL;
	    ret->data.data=sym;
	    ret->args=NULL;
	  }
	  else
	  {
	    FATAL(("Unknown add %d %d\n",left->type,right->type));
	  }

	  FreeAstNode(in);
	  return ret; 
	  break;

	}
      case OPER_OR:
	FATAL(("Implement OR"));
      default:
	FATAL(("Unknown operand %d in trigger\n",in->data.oper));
    }
  }
  else
  {
    FATAL(("Unknown type! %d",in->type));
  }
  /* keep the compiler happy */
  return NULL;
}
/*}}}*/
/* Check and cast one {{{ */
void CheckAndCastOneArgument(t_string name, int tel, t_ast_node ** to_cast, char desc, t_ast_node_table_entry * function_table, void * implictit_arg)
{
  if ((*to_cast==NULL) && (desc!='V')) FATAL(("Function %s expected %c as argument %d",name,desc,tel));
  else if (*to_cast==NULL) return; 
  switch((*to_cast)->type)
  {
    case TYPE_STRING:
      if (desc!='S') FATAL(("Function %s expected a string as argument %d",name,tel));
      break;
    case TYPE_NUMERIC:
      if (desc!='N') FATAL(("Function %s expected a number as argument %d",name,tel));
      break;
    case TYPE_OTHER:
      if (desc=='N') 
      {
	t_string orig=(*to_cast)->data.data;
	(*to_cast)->type=TYPE_NUMERIC;
	(*to_cast)->data.data=(void *) (t_uint32) StringToUint64(orig,0);
	Free(orig);
      }
      else if (desc=='O')
      {
	break;
      }
      else if ((desc=='S')&&(StringToUint64((*to_cast)->data.data,0)==0))
      {
	t_string orig=(*to_cast)->data.data;
	(*to_cast)->type=TYPE_STRING;
	(*to_cast)->data.data=0;
	Free(orig);
      }
      else  FATAL(("Function %s expected type other as argument %d, type =%c",name,tel,desc));
      break;
    case TYPE_VOID:
      if (desc!='V')
      {
	FATAL(("Function %s expected type %c as argument %d",name,desc,tel));
      }
      break;
    case TYPE_TRIGGER:
      if (desc=='T')
      {
	break;
      }
      else if (desc=='B')
      {
	if (((int) (*to_cast)->data.data)==2)
	{
	  WARNING(("Cast from type trigger to type bool casts away repeating triggers"));
	  (*to_cast)->data.data= (void *) 1;
	}
      }
      else
      {
	FATAL(("Function %s expected type other as argument %d, but type =%c",name,tel,desc));
      }
      break;
    case TYPE_BOOL:
      if (desc=='B')
      {
	break;
      }
      else FATAL(("Function %s expected type other as argument %d, but type =%c",name,tel,desc));
      break;
    case TYPE_OPER:
      {
	t_ast_node * n=AstExecute((*to_cast),function_table,implictit_arg,current_free_mode);
	CheckAndCastOneArgument(name, tel, &n, desc,function_table,implictit_arg);
	(*to_cast)=n;
      }
      break;
    case TYPE_FUNCTION:
      {
	t_ast_node * n=AstExecute((*to_cast),function_table,implictit_arg,current_free_mode);
	CheckAndCastOneArgument(name, tel, &n, desc,function_table,implictit_arg);
	(*to_cast)=n;
	
      }
      break;
    case TYPE_SECTION:
      FATAL(("SECTION?"));
    case TYPE_SYMBOL:
      FATAL(("SYMBOL"));
      break;
    default:
      FATAL(("Implement type %d\n",(*to_cast)->type));
  }
}
/*}}} */
/* Check And Cast All {{{ */
static void CheckAndCastArguments(t_ast_node * in, t_uint32 nargs, t_string arg_desc,t_ast_node_table_entry * function_table, void * implictit_arg)
{
  t_ast_successors * args=in->args;
  t_string name=in->data.name;
  if ((nargs==0)&&(args!=NULL)&&(args->nargs!=0))
  {
    FATAL(("Function %s expects no arguments, but %d arguments are given",name,args->nargs));
  }
  else if (nargs!=0)
  {
    if ((args==NULL)||(args->nargs==0)) FATAL(("Function %s expects %d arguments, but no arguments are given",name,nargs));
    else if (args->nargs!=nargs) FATAL(("Function %s expects %d arguments, but %d arguments are given",name,args->nargs,nargs));
    else 
    {
      t_uint32 tel;
      /* The number of arguments is right, now check and cast the arguments */
      for (tel=0; tel<args->nargs; tel++)
      {
	CheckAndCastOneArgument(name,tel,&(args->args[tel]),arg_desc[tel],function_table,implictit_arg);
      }
      
    }
  }
}
/* }}} */
/* }}} */

/* Linker Rule Execution {{{ */
void ExecuteRule(t_string name, t_linker_rule * rule)
{
  t_linker_rule * used_rule=LinkerRuleDup(rule);
  t_ast_node * trigger = NULL;
  t_ast_node * ret;
  t_bool rerun=FALSE;
  do
  {
    if (rerun) { 
      FreeAstNode(trigger);
      Free(used_rule); 
      used_rule=LinkerRuleDup(rule); 
    }
  
    trigger=AstExecute(used_rule->trigger,FunctionTable,NULL,TRUE);
    if (trigger->data.data) 
    {
      ret=AstExecute(used_rule->action,FunctionTable,used_rule,TRUE);
      FreeAstNode(ret);
      if (used_rule->section) FATAL(("Section not used!"));
      if (used_rule->address) FATAL(("Address not used!"));
      if (used_rule->symbol ) FATAL(("Symbol  not used!"));
    }
    else
    {
      FreeAstNode(used_rule->action);
      if (used_rule->section) FreeAstNode(used_rule->section);
      if (used_rule->address) FreeAstNode(used_rule->address);
      if (used_rule->symbol) FreeAstNode(used_rule->symbol); 
    }
    rerun=TRUE;
  } while (((int) trigger->data.data)==2);
  
  FreeAstNode(trigger);
  Free(name);
  Free(used_rule);

  FreeAstNode(rule->action);
  FreeAstNode(rule->trigger);
  
  if (rule->section) FreeAstNode(rule->section);
  if (rule->address) FreeAstNode(rule->address);
  if (rule->symbol) FreeAstNode(rule->symbol); 
  
  Free(rule);
}
/*}}}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
