/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file routines 
 * 
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/

/*{{{ Includes*/
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "../config.h"
#include "diablo_addresses.h"
#include "diablo_object.h"
#include "diablo_object_tables.h"
#include "../simple_types.h"
#include "../diablo_options.h"

/* Utilities used: */
#include "../datastructures/data_string.h"
#include "../datastructures/data_hash.h"
#include "../utils/util_file.h"
#include "../utils/util_malloc.h"
#include "../utils/util_options.h"
#include "../utils/util_status.h"
#include "../utils/util_error.h"
#include "../utils/util_term.h"
/* Support for archives */
#include "../fileformats/ar/ar_arch.h"
/* Symbols */
#include "diablo_symbol.h"
#include "diablo_reloc.h"
#include "diablo_relocate.h"
#include "diablo_section.h"
#include "diablo_stack_machine.h"
/* Maps */
#include "diablo_parsemap.h"
/* Printing */
#include "diablo_print.h"
#include "diablo_iterators.h"
/*}}}*/
/* Static declarations to keep the compiler happy {{{*/
void ObjectHashElementFree(void * tf,void * data);
/*}}}*/
/* Constructors {{{ */

/*! 
 * \return a new (zero-initialized) t_object * structure
 * Static constructor for a t_object */ 

t_object * ObjectNew()
{
  t_object * ret;
  ret=Calloc(1,sizeof(t_object));    
  return ret;
}

/*! 
 * \param fp an open filedescriptor to the objectfile
 * \param ret an allocated t_object structure that will be filled with the information read
 * 
 * This function reads an objectfile from a stream 
 * \warn This is an internal function and should not be used directly, use \ref ObjectGet instead  */ 

void ObjectReadF(FILE * fp, t_object * ret)
{
  const t_object_handler * ht=handler_table;
  int tel=0;
  char buffer[10];
  ret->streampos=ftell(fp);
  VERBOSE(1,("Reading file %s",ret->name));

  /* Try to identify the object */
  while(ht[tel].type!=T_UNKNOWN)
  {
    if (ht[tel].ident(fp))
    {
      if (ht[tel].htype!=GROUPHANDLER)
      {
	/* If we have a match use the corresponding load operation */
	ret->type=ht[tel].type;
	ret->subtype=ht[tel].sub_type;
	ret->actions=ht[tel].acts;
	ret->ff_actions=ht[tel].ff_acts;
	ret->description=ht[tel].arch_desc;
	if (!ret->description) FATAL(("No architecture description for architecture %s",ht[tel].name));
	ht[tel].read(fp,ret);

	/* {{{ try to find out if the code was handwritten in assembler or created by a compiler */
	{
	  t_bool handwritten = TRUE;
	  t_symbol * sym = ret->symbol_table->first;
	  int i;

	  while (sym && !(sym->type == ABS_SYMBOL && sym->to_type == TO_NONE))
	    sym = sym->next;

	  /* file symbol is first absolute symbol */
	  if (sym && sym->name)
	  {
	    char * ext = strrchr(sym->name,'.');
	    if (ext)
	    {
	      ext++;
	      if (!strcmp(ext,"c") || !strcmp(ext,"h") || !strcmp(ext,"C") ||
		  !strcmp(ext,"cpp") || !strcmp(ext,"hpp"))
		handwritten = FALSE;
	    }
	  }

#ifdef ALPHA_SUPPORT
	  /* for the alpha, the original file cannot easily be determined from symbols in 
	   * the object files, so diablo would think all code was hand written in assembly.
	   * This is of course unacceptable, so we do it the other way round: assume all code
	   * was compiler-written. To date, this hasn't caused any problems yet */
	  handwritten = FALSE;
#endif

	  for (i=0; i<ret->ncodes; i++)
	  {
	    if (handwritten)
	      SymbolTableAddSymbol(ret->symbol_table,TO_SECTION,LOCAL_SYMBOL,T_RELOCATABLE(ret->code[i]),NULL_GENERIC,StringDup("$handwritten"),TRUE);
	    else
	      SymbolTableAddSymbol(ret->symbol_table,TO_SECTION,LOCAL_SYMBOL,T_RELOCATABLE(ret->code[i]),NULL_GENERIC,StringDup("$compiler"),TRUE);
	  }
	} /* }}} */

	return;
      }
      else
      {
	ht=ht[tel].subhandler;
	tel=0;
	continue;
      }
    }
    tel++;
  }

  /* Debug: show the first 10 bytes of the program (most likely the bytes used
   * to identify this binary */

  tel=fread(buffer,1,10,fp);
  if (tel<10) FATAL(("Could not get 10 bytes, so either the object is truncated or this is not an object file"));

  printf("Dump of first 10 bytes\n");
  printf("Hex:\n");
  for (tel=0; tel<10; tel++)
  {
    printf("0x%x ",buffer[tel]);
  }
  printf("\n");

  printf("Octal:\n");
  for (tel=0; tel<10; tel++)
  {
    printf("0%o ",buffer[tel]);
  }
  printf("\n");

  printf("Decimal:\n");
  for (tel=0; tel<10; tel++)
  {
    printf("%d ",buffer[tel]);
  }
  printf("\n");


  FATAL(("Unknown object filetype %s!",ret->name));
}

/*!
 *
 * \param name
 * \param fp
 *
 * \return t_object *
*/

t_object * ObjectGetFromStream(t_string name, FILE *fp)
{
  t_object * ret=ObjectNew();
  ret->name=StringDup(name);
  ObjectReadF(fp, ret);
  return ret;
}

/* \param name The filename you want to read. 
 *
 * Read the object name. The name should be the complete path to the object!
 * 
 * \warn This is an internal function and should not be used directly, use \ref ObjectGet instead  */ 

t_object * ObjectRead(t_string name)
{
  FILE * fp;
  t_object * ret;
  fp=fopen(name,"r");
  VERIFY_FASSERT(fp, ("Could not open: %s", name));
  ret=ObjectNew();
  ret->name=name;
  fseek(fp,0,SEEK_END);
  ret->size=ftell(fp);
  fseek(fp,0,SEEK_SET);
  ObjectReadF(fp, ret);
  fclose(fp);
  return ret;
}

/*!
 *
 * \param real_name
 * \param fake_name
 *
 * \return t_object *
*/

t_object * ObjectReadWithFakeName(t_string real_name,t_string fake_name)
{
  t_string full_real_name=FindFile(global_options.objpath,real_name);
  FILE * fp;
  t_object * ret;
  VERIFY_FASSERT(full_real_name,("File %s not found in objectpath",real_name));
  fp=fopen(full_real_name,"r");
  VERIFY_FASSERT(fp, ("Could not open: %s", full_real_name));
  ret=ObjectNew();
  ret->name=fake_name;
  fseek(fp,0,SEEK_END);
  ret->size=ftell(fp);
  fseek(fp,0,SEEK_SET);
  ObjectReadF(fp, ret);
  fclose(fp);
  return ret;
}


/* Cache Constructors {{{ */
extern data_hash_table * global_object_hash;

/*!
 * \param name The name of the object to load (can be in a library if : separated)
 * \param parent The parent object in case we're loading a sub object
 * \return the loaded object
 * 
 * Default method to open an objectfile, has a built in cache to avoid loading an object 2 times. */ 

t_object * ObjectGet(t_string name,t_object * parent)
{
  data_hash_table * cache;
  object_hash_node * hn;

  /* if we're loading a sub object, use the parent object's cache,
   * otherwise use the global cache */
  cache = parent ? parent->subobject_cache : global_object_hash;

  /* First look in the object hash table, to see if we already opened it */
  hn=(object_hash_node *) HashLookup(cache,name);
  /* If not: Open */
  if (!hn)
  {
    t_object * obj;

    /* Check if it is from an archive */
    if (PatternMatch("*:*",name))
    {
      t_archive * ar;
      t_string copy=StringDup(name);
      t_string libn=strtok(copy,":");
      t_string objn=strtok(NULL,":");
      VERIFY_FASSERT(libn, ("Illegal objectname %s\n",name));
      VERIFY_FASSERT(objn, ("Illegal objectname %s\n",name));
      VERIFY_FASSERT(!strtok(NULL,":"), ("Illegal objectname %s\n",name));

      ar=ArchiveOpenCached(libn);

      VERIFY_FASSERT(ar!=NULL, ("Library %s not found in libpath\n",libn));
      obj=ArchiveGetObject(ar,objn);
      VERIFY_FASSERT(obj!=NULL, ("Object %s not found in %s\n",objn,libn));
      Free(copy); 
    }
    /* Try to read a normal file */
    else
    {
      t_string full_name=FindFile(global_options.objpath,name);
      if (!full_name)
      {
        FATAL(("Object %s not found\n",name));
	exit(0); /* Keep the compiler happy */
      }
      else
      {
        obj=ObjectRead(full_name);
        VERIFY_FASSERT(obj!=NULL, ("Object %s not found\n",name));
      }
    }

    /* Put it in the object cache */
    hn=Malloc(sizeof(object_hash_node));
    hn->node.key=StringDup(name);
    hn->obj=obj;
    hn->fake_hash_entry=FALSE;
    HashInsert(cache,(data_hash_node *) hn);

    if (strcmp(obj->name,name)!=0)
    {
      hn=Malloc(sizeof(object_hash_node));
      hn->node.key=StringDup(obj->name);
      hn->obj=obj;
      hn->fake_hash_entry=TRUE;
      HashInsert(cache,(data_hash_node *) hn);
    }

    if (parent)
      ObjectAppendSubObject(parent,obj);
    
    return obj;
  }

  /* Else return the cached object */
  return hn->obj;
}

t_object * ObjectGetFromCache(t_string name, t_object * parent)
{
  object_hash_node * hn;
  data_hash_table * cache = parent ? parent->subobject_cache : global_object_hash;

  hn=(object_hash_node *) HashLookup(cache,name);
  if (!hn) return NULL;
  return hn->obj;
}

t_object * ObjectNewCached(t_string name, t_object * parent)
{
  object_hash_node * hn;
  data_hash_table * cache = parent ? parent->subobject_cache : global_object_hash;

  hn=Malloc(sizeof(object_hash_node));
  hn->node.key=StringDup(name);
  hn->obj=ObjectNew();
  hn->obj->name=StringDup(name);
  hn->fake_hash_entry=FALSE;
  HashInsert(cache,(data_hash_node *) hn);
  return hn->obj;
}
/*}}}*/

/* End Constructors }}}*/
/* Destructor {{{ */
/*!
 * \param to_free The object to free
 * 
 * This procedure frees a t_object structure (and all substructures) */ 

void ObjectFree(t_object * to_free)
{
  if (to_free->orig_codes)
  {
    int i;
    for (i=0; i<to_free->n_orig_codes; i++)
      Free(to_free->orig_codes[i].name);
    Free(to_free->orig_codes);
  }

  if (to_free->layout_script)
    LayoutScriptFree(to_free->layout_script);

  if (to_free->name) Free(to_free->name);
  if (to_free->reloc_table) RelocTableFree(to_free->reloc_table);
  if (to_free->map) MapFree(to_free->map);
  if (to_free->symbol_table) SymbolTableFree(to_free->symbol_table);
  if (to_free->sub_symbol_table) SymbolTableFree(to_free->sub_symbol_table);

  if (to_free->code) 
  {
    t_uint32 tel;
    for (tel=0; tel<to_free->ncodes; tel++)
      SectionFree(to_free->code[tel]); 
    Free(to_free->code);
  }
  if (to_free->rodata) 
  {
    t_uint32 tel;
    for (tel=0; tel<to_free->nrodatas; tel++)
      SectionFree(to_free->rodata[tel]); 
    Free(to_free->rodata);
  }
  if (to_free->data) 
  {
    t_uint32 tel;
    for (tel=0; tel<to_free->ndatas; tel++)
      SectionFree(to_free->data[tel]); 
    Free(to_free->data);
  } 
  if (to_free->note) 
  {
    t_uint32 tel;
    for (tel=0; tel<to_free->nnotes; tel++)
      SectionFree(to_free->note[tel]); 
    Free(to_free->note);
  } 
  if (to_free->bss) 
  {
    t_uint32 tel;
    for (tel=0; tel<to_free->nbsss; tel++)
      SectionFree(to_free->bss[tel]); 
    Free(to_free->bss);
  }

  if (to_free->section_table) HashFree(to_free->section_table);
  /* if the object is a parent object, free it's associated cache */
  if (to_free->subobject_cache) HashFree(to_free->subobject_cache);

  while (to_free->segment_first)
  {
    t_segment * seg = to_free->segment_first->next;
    Free(to_free->segment_first->name);
    Free(to_free->segment_first);
    to_free->segment_first = seg;
  }

  Free(to_free);
}

/* }}} */

/* Modifiers {{{ */

/* FUN : ObjectAddSectionFromFile
 * PAR : The object, char type (C,D,or B), a bool to say if it's read only, FILE * fp, filepos fileoffset, t_address address, t_address size
 * RET : nothing 
 * DESC: Add's a section from an objectfile.*/ 

t_section * ObjectAddSectionFromFile(t_object * obj, char type, t_bool ro, FILE * fp, Filepos off, t_address address, t_address size, t_address align, t_string name)
{
  t_section *** dest=NULL;
  t_uint32 * ndest=NULL;
  Filepos stored=ftell(fp);

  switch(type)
  {
    case CODE_SECTION:
      VERIFY_FASSERT(ro,("Non constant code...."));
      dest=&obj->code;
      ndest=&obj->ncodes;
      break;
    case DATA_SECTION:
      if (ro)
      {
	type=RO_DATA_SECTION;
	dest=&obj->rodata;
	ndest=&obj->nrodatas;
      }
      else
      {
	dest=&obj->data;
	ndest=&obj->ndatas;
      }
      break;
    case NOTE_SECTION:
      type=NOTE_SECTION;
      dest=&obj->note;
      ndest=&obj->nnotes;
      break;
    case BSS_SECTION:
      VERIFY_FASSERT(!ro,("Constant bss..."));
      dest=&obj->bss;
      ndest=&obj->nbsss;
      break;
    default:
      FATAL(("Unsupported type %c in add!",type));
  }

  *dest=Realloc(*dest,sizeof(t_section *)*((*ndest)+1));

  (*dest)[(*ndest)]=SectionNew(obj,type);

  /* Set the name of the new section */
  (*dest)[(*ndest)]->name=StringDup(name);
  /* Set the size of the new section */
  SECTION_OLD_SIZE((*dest)[(*ndest)])=size;
  SECTION_CSIZE((*dest)[(*ndest)])=size;
  /* Set the start address */
  SectionSetOldAddressSuper(((*dest)[(*ndest)]),address);
  SectionSetCAddressSuper(((*dest)[(*ndest)]),address);
  SECTION_ALIGN((*dest)[(*ndest)])=align;

  if ((!GENERICISNULL(size)) && (type!=BSS_SECTION))
  {
    SET_SECTION_DATA((*dest)[(*ndest)],Malloc(GENERIC_EXTRACT_UINT32(size)));
  }
  else
    SET_SECTION_DATA((*dest)[(*ndest)],NULL);

  if ((*dest)[(*ndest)]->data) {
    fseek(fp,off,SEEK_SET);
    fread((*dest)[(*ndest)]->data,GENERIC_EXTRACT_UINT32(size),1,fp);
  }
  HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew((*dest)[(*ndest)]->name,(*dest)[(*ndest)]));
  (*ndest)++;
  fseek(fp,stored,SEEK_SET);
  return (*dest)[(*ndest)-1];
}

/* }}} */
/* Querying {{{ */

/* FUN : ObjectGetData64
 * PAR : 
 * RET : 
 * DESC: */ 

t_uint64 ObjectGetData64(t_object * obj, t_string name, t_address address)
{
  t_section * sec=SectionGetFromObjectByName(obj,name);
  t_address real_address;
  void * load;
  VERIFY_FASSERT(sec!=NULL,("Could not get section %s",name));
  real_address=GENERIC_SUB(address,SECTION_GET_CADDRESS(sec));
  printf("Loading from "); GENERIC_PRINT(real_address,16); printf("\n");
  VERIFY_FASSERT(GENERIC_GT(SECTION_CSIZE(sec),real_address),("Out of section load!"));
  load=GENERIC_ADD_DISPL(sec->data,real_address);

  return (*((t_uint64 *) load));
}

/* FUN : ObjectGetData32
 * PAR : 
 * RET : 
 * DESC: */ 

t_uint32 ObjectGetData32(t_object * obj, t_string name, t_address address)
{
  t_section * sec=SectionGetFromObjectByName(obj,name);
  t_address real_address;
  void * load;
  VERIFY_FASSERT(sec!=NULL,("Could not get section %s",name));
  real_address=GENERIC_SUB(address,SECTION_GET_CADDRESS(sec));
  printf("Loading from "); GENERIC_PRINT(real_address,16); printf("\n");
  VERIFY_FASSERT(GENERIC_GT(SECTION_CSIZE(sec),real_address),("Out of section load!"));
  load=GENERIC_ADD_DISPL(sec->data,real_address);

  return (*((t_uint32 *) load));
}

/* FUN : ObjectGetData32NoSection
 * PAR : 
 * RET : 
 * DESC: */ 

t_uint32 ObjectGetData32NoSection(t_object * obj, t_address start)
{
  t_uint32 tel;
  for (tel=0; tel<obj->ncodes; tel++)
    {
      if ((GENERIC_LT(SECTION_GET_CADDRESS(obj->code[tel]),start))&&((GENERIC_GT(GENERIC_ADD(SECTION_GET_CADDRESS(obj->code[tel]),SECTION_CSIZE(obj->code[tel])),start))))
  {
    FATAL(("Implement"));
    /*return *((t_uint32 *) (((char *) obj->code[tel]->data)+G_T_UINT32(start)-G_T_UINT32(obj->code[tel]->address))); */
  }
    }
  FATAL(("Address not found!"));

  return 0; /* Keep the compiler happy */
}
/* }}} */
/* Change the representation of the data in an object {{{ */

/* FUN : ObjectDisassemble
 * PAR : An objectfile 
 * RET : nothing 
 * DESC: Disassembles an object file (called as an action) */ 

void ObjectDisassemble(t_object * obj)
{
  t_uint32 tel;
  for (tel=0; tel<obj->ncodes; tel++)
    {
      ObjectAction(DISASSEMBLE_SECTION,obj,obj->code[tel]);
    }
}

/* FUN : ObjectAssemble
 * PAR : an objectfile 
 * RET : nothing 
 * DESC: Assemble an objectfile (called as an action) */ 

void ObjectAssemble(t_object * obj)
{
  t_uint32 tel;
  for (tel=0; tel<obj->ncodes; tel++)
    {
      ObjectAction(ASSEMBLE_SECTION,obj,obj->code[tel]);
    }
}

/* FUN : ObjectFlowgraph
 * PAR : an object for which we want the flowgraph 
 * RET : nothing 
 * DESC: Create the flowgraph for for an entire object */ 

void ObjectFlowgraph(t_object * obj)
{
  t_uint32 tel;
  for (tel=0; tel<obj->ncodes; tel++)
    {
      ObjectAction(FLOWGRAPH_SECTION,obj,obj->code[tel]);
    }
}

/* FUN : ObjectDeflowgraph
 * PAR : TODO 
 * RET : nothing 
 * DESC: TODO */

void ObjectDeflowgraph(t_object * obj)
{
  t_uint32 tel;
  for (tel=0; tel<obj->ncodes; tel++)
  {
    if (SECTION_TYPE(obj->code[tel]) != FLOWGRAPHED_CODE_SECTION)
      continue;

    printf("DeFlowgraphing %s\n",obj->code[tel]->name);
    ObjectAction(DEFLOWGRAPH_SECTION,obj,obj->code[tel]);
  }
}

/* FUN : ObjectConvertToDataBlocks(obj)
 * PAR : TODO 
 * RET : nothing 
 * DESC: TODO */

void ObjectConvertToDataBlocks(t_object * obj)
{
  t_uint32 tel;
  for (tel=0; tel<obj->ndatas; tel++)
    {
      printf("Converting to data blocks  %s\n",obj->data[tel]->name);
      ObjectAction(CONVERT_SECTION_TO_BLOCKS,obj,obj->data[tel]);
    }
  for (tel=0; tel<obj->nrodatas; tel++)
    {
      printf("Converting to rodata blocks  %s\n",obj->rodata[tel]->name);
      ObjectAction(CONVERT_SECTION_TO_BLOCKS,obj,obj->rodata[tel]);
    }
  for (tel=0; tel<obj->nbsss; tel++)
    {
      printf("Converting to bss blocks  %s\n",obj->bss[tel]->name);
      ObjectAction(CONVERT_SECTION_TO_BLOCKS,obj,obj->bss[tel]);
    }
}

/* }}} */

/* ================ OBJECT CACHE ================ */

/* DOC: Struct and definitions for object cache: 
 * The objectcache is used to avoid reloading already 
 * opened objectfiles */ 

data_hash_table * global_object_hash=NULL;

/*!
 * \param tf The element to free
 * \param data Unused data element (needed when walking a hash table) 
 *
 * Utility function: used to free the objects in the cache by the t_hashtable structure.
 *
 * \return void 
*/

void ObjectHashElementFree(void * tf,void * data)
{
  object_hash_node * node=tf;
  if (!node->fake_hash_entry)
    ObjectFree(node->obj);
  Free(node->node.key);
  Free(node);
}
extern t_symbol_he * symbol_type_cached;
/* Global constructor and destructor {{{ */

/* FUN : ObjectInit
 * PAR : none 
 * RET : nothing
 * DESC: Static initialization for the t_object class */ 

void ObjectInit()
{
  VERIFY_FASSERT(!global_object_hash,("Reinitialization of Object!"));
  global_object_hash=HashTableNew(103,0,string_hash,(my_cmpresult (*)(void *, void *))  strcmp,ObjectHashElementFree);

  /* create global dummy sections (abs and undef) */
  undef_section = SectionNew(NULL,'D');
  abs_section = SectionNew(NULL,'D');
  PRIVATE_SET_SECTION_CADDRESS(undef_section) = PRIVATE_SET_SECTION_OLD_ADDRESS(undef_section) = SECTION_CSIZE(undef_section) = NULL_GENERIC;
  PRIVATE_SET_SECTION_CADDRESS(abs_section) = PRIVATE_SET_SECTION_OLD_ADDRESS(abs_section) = SECTION_CSIZE(abs_section) = NULL_GENERIC;
  undef_section->name = "*UNDEF*";
  abs_section->name = "*ABS*";
}

/* FUN : ObjectFini
 * PAR : none 
 * RET : nothing
 * DESC: Static deinitialization for the t_object class */ 

void ObjectFini()
{
  VERIFY_FASSERT(global_object_hash,("Deinitialization without initialization of Object!"));
  HashFree(global_object_hash);

  Free(abs_section);
  Free(undef_section);
}

/* }}} */

#define VERBOSE_MERGING

/* Merge code sections {{{ */
void ObjectMergeCodeSections(t_object * obj)
{
  t_uint32 tel;
  t_bool ok=TRUE;

  t_address start;
  t_address run,size;
  t_section * tmpsec;
  t_object * sub, * tmp;
  t_reloc * rel;
  t_symbol * sym;

#ifdef VERBOSE_MERGING
  DiabloPrint(stdout,"Before merging\n");
  {
    t_section * sec;
    t_uint32 tel;
    OBJECT_FOREACH_SECTION(obj,sec,tel)
    {
      DiabloPrint(stdout,"Section %s new address = @G, old = @G\n",sec->name,SECTION_GET_CADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec));
    }
  }
  printf("Have %d code sections\n",obj->ncodes);
#endif

  if (obj->ncodes<2) 
  {
    /* just fill out the original code sections table and return */
    obj->n_orig_codes = 1;
    obj->orig_codes = Calloc(1,sizeof(t_sec_descriptor));
    obj->orig_codes->name = StringDup(obj->code[0]->name);
    obj->orig_codes->base = SECTION_GET_CADDRESS(obj->code[0]);
    obj->orig_codes->size = SECTION_CSIZE(obj->code[0]);
    obj->orig_codes->obase = SECTION_GET_OLD_ADDRESS(obj->code[0]);
    obj->orig_codes->osize = SECTION_OLD_SIZE(obj->code[0]);
    obj->orig_codes->type = CODE_SECTION;
    return;
  }

  /* {{{ first step: check if all code sections are laid out contiguously in memory */
  start = run = SECTION_GET_CADDRESS(obj->code[0]);
  for (tel=0; tel<obj->ncodes; tel++)
  {
    t_section * code = obj->code[tel];
    DiabloPrint(stdout,"Section %s at @G @G\n",code->name,SECTION_GET_CADDRESS(code),SECTION_CSIZE(code));
    run = AlignAddress(SECTION_ALIGN(code),run);
    if (!GENERIC_EQ(run, SECTION_GET_CADDRESS(code)))
      ok = FALSE;
    run = GENERIC_ADD(run,SECTION_CSIZE(code));
  }
  size = GENERIC_SUB(run,start);

  /* {{{ rebuild sections and bring relocs and symbols and stuff up to date */
  if (!ok)
  {
    /* reorder the sections */
    DiabloPrint(stdout,"Code sections not laid out contiguously\n");
#ifndef KERNEL_OPT
    FATAL(("Never before encountered this for user space programs... Check the following code carefully"));
#endif

    OrderCodeSectionsContiguously(obj);
    /* recalculate the relocations */
    OBJECT_FOREACH_RELOC(obj,rel)
    {
      t_address ret = StackExec(rel->code,rel,T_SECTION(REL_FROM(rel))->data,TRUE,0);
      if (!strcmp(rel->label,"__init_begin") || !strcmp(rel->label,"__init_end"))
	DiabloPrint(stdout,"reloc %s new dest @G code %s\n", rel->label, StackExec("SA+$",rel,NULL,FALSE,0),rel->code);
      VERIFY_FASSERT(GENERICISNULL(ret),("Could not execute @R (code %s)",rel,rel->code));
    }
    /* write the changes back into the data of the parent sections */
    ObjectRebuildSectionsFromSubsections(obj);
    /* recalculate the current_address field for all symbols */
    for (sym = obj->sub_symbol_table->first; sym; sym = sym->next)
    {
      if (sym->base && (sym->base->type == RT_SECTION || sym->base->type == RT_SUBSECTION))
      {
	sym->current_address = GENERIC_ADD(sym->addend,GENERIC_ADD(sym->offset_from_start,SECTION_GET_CADDRESS(sym->base)));
      }
    }
  } /* }}} */
  /* {{{ Kill all existing $switch symbols, as they are no longer guaranteed to be correct
   * (section reordering screws them up ... */
  {

    t_symbol_he * top = HashLookup(obj->sub_symbol_table->hash_table,"$switch");
    t_symbol_he * he, * prev;
    t_uint32 ht_index;

    symbol_type_cached=NULL;

    if (top)
    {
      /* kill all $switch symbols from the linked list in the symbol table */
      for (he = top; he; he = (t_symbol_he *) he->node.equal)
      {
	sym = he->sym;
	if (sym->prev) 
	  sym->prev->next = sym->next;
	else
	  obj->sub_symbol_table->first = sym->next;

	if (sym->next)
	  sym->next->prev = sym->prev;
	else
	  obj->sub_symbol_table->last = sym->prev;
      }

      /* kill all the entries in the hash table */
      /* to do this, we have to find out where exactly top is in the hash table */
      ht_index = obj->sub_symbol_table->hash_table->hash_func(obj->sub_symbol_table->hash_table,"$switch");
      prev = NULL;
      for (he = (t_symbol_he *) obj->sub_symbol_table->hash_table->nodes[ht_index]; he; he = (t_symbol_he *) (he->node.next))
      {
	if (he == top)
	  break;
	prev = he;
      }
      VERIFY_FASSERT(he, ("Lookup table corrupt"));
      /* cut the chain of $switch symbols from the hash table */
      if (prev) 
	prev->node.next = top->node.next;
      else
	obj->sub_symbol_table->hash_table->nodes[ht_index] = top->node.next;
      /* now free the complete chain */
      for (he = top, top = (t_symbol_he *) he->node.equal; he; he = top, top = he ? (t_symbol_he *) he->node.equal : NULL)
      {
	Free(he->node.key);
	Free(he->sym->name);
	Free(he->sym);
	Free(he);
      }
    }
  } /* }}} */
#ifdef THUMB_SUPPORT
  /* {{{ Kill all existing $code_switch symbols, as they are no longer guaranteed to be correct
   * (section reordering screws them up ... */
  {

    t_symbol_he * top = HashLookup(obj->sub_symbol_table->hash_table,"$code_switch");
    t_symbol_he * he, * prev;
    t_uint32 ht_index;

    symbol_type_cached=NULL;

    if (top)
    {
      /* kill all $code_switch symbols from the linked list in the symbol table */
      for (he = top; he; he = (t_symbol_he *) he->node.equal)
      {
	sym = he->sym;
	if (sym->prev) 
	  sym->prev->next = sym->next;
	else
	  obj->sub_symbol_table->first = sym->next;

	if (sym->next)
	  sym->next->prev = sym->prev;
	else
	  obj->sub_symbol_table->last = sym->prev;
      }

      /* kill all the entries in the hash table */
      /* to do this, we have to find out where exactly top is in the hash table */
      ht_index = obj->sub_symbol_table->hash_table->hash_func(obj->sub_symbol_table->hash_table,"$code_switch");
      prev = NULL;
      for (he = (t_symbol_he *) obj->sub_symbol_table->hash_table->nodes[ht_index]; he; he = (t_symbol_he *) (he->node.next))
      {
	if (he == top)
	  break;
	prev = he;
      }
      VERIFY_FASSERT(he, ("Lookup table corrupt"));
      /* cut the chain of $code_switch symbols from the hash table */
      if (prev) 
	prev->node.next = top->node.next;
      else
	obj->sub_symbol_table->hash_table->nodes[ht_index] = top->node.next;
      /* now free the complete chain */
      for (he = top, top = (t_symbol_he *) he->node.equal; he; he = top, top = he ? (t_symbol_he *) he->node.equal : NULL)
      {
	Free(he->node.key);
	Free(he->sym->name);
	Free(he->sym);
	Free(he);
      }
    }
  } /* }}} */
#endif
  /* }}} */
  
  /* {{{ second step: create new section from the different code sections */
  tmpsec = SectionNew(obj,CODE_SECTION);
  SECTION_OLD_SIZE(tmpsec) = size;
  SECTION_CSIZE(tmpsec) = size;
  SectionSetCAddressSuper(tmpsec,start);
  SectionSetOldAddressSuper(tmpsec,SECTION_GET_OLD_ADDRESS(obj->code[0]));
  tmpsec->name = StringDup(".merged_code");
  tmpsec->data = Calloc(sizeof(char),GENERIC_EXTRACT_UINT32(size));

  /* insert in section hash table */
  HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(tmpsec->name,tmpsec));

  for (tel=0; tel<obj->ncodes; tel++)
  {
    t_section * code = obj->code[tel];
    t_uint32 offset = GENERIC_EXTRACT_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(code),SECTION_GET_CADDRESS(tmpsec)));
    t_uint32 size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(code));
    memcpy(((char *)tmpsec->data)+offset,code->data,(size_t) size);
  }
  /* }}} */

  /* {{{ third step: make all subsections point to the new section as their parent section */
  OBJECT_FOREACH_SUBOBJECT(obj,sub,tmp)
  {
    t_section * sec;
    OBJECT_FOREACH_SECTION(sub,sec,tel)
    {
      if (SECTION_PARENT_SECTION(sec) && (SECTION_PARENT_SECTION(sec)->type==CODE_SECTION))
      {
	SECTION_PARENT_SECTION(sec) = tmpsec;
      }
    }
  } /* }}} */

  /* {{{ fourth step: make all relocations point to the right section */
  OBJECT_FOREACH_RELOC(obj,rel)
  {
    if (REL_TO_RELOCATABLE(rel)->type == RT_SECTION)
      if (T_SECTION(REL_TO_RELOCATABLE(rel))->type == CODE_SECTION)
      {
	t_address offset = GENERIC_SUB(SECTION_GET_CADDRESS(T_SECTION(REL_TO_RELOCATABLE(rel))),SECTION_GET_CADDRESS(tmpsec));
	RelocSetToRelocatable(rel, T_RELOCATABLE(tmpsec));
	rel->to_offset = GENERIC_ADD(rel->to_offset, offset);
      }
    /* relocations should always be from _sub_section, never from section... */
    if (REL_FROM(rel)->type == RT_SECTION) FATAL(("Relocations corrupt"));
  } /* }}} */

  /* {{{ fifth step: migrate all symbols that point to the old code sections */
  for (sym = obj->sub_symbol_table->first; sym; sym = sym->next)
  {
    if (sym->base && sym->base->type == RT_SECTION && T_SECTION(sym->base)->type == CODE_SECTION)
    {
      t_address offset = GENERIC_SUB(SECTION_GET_CADDRESS(sym->base),start);
      sym->offset_from_start = GENERIC_ADD(sym->offset_from_start,offset);
      sym->current_address = GENERIC_ADD(start,sym->offset_from_start);
      sym->base = T_RELOCATABLE(tmpsec);
    }
  } /* }}} */
  
  /* {{{ sixth step: add $d symbols for the padding between the merged code sections */
  run = start;
  for (tel=0; tel<obj->ncodes; tel++)
  {
    t_section * code = obj->code[tel];
    t_address tmp = AlignAddress(SECTION_ALIGN(code),run);

    /* sanity check */
    VERIFY_FASSERT(GENERIC_EQ(tmp,SECTION_GET_CADDRESS(code)), ("Ouch: @G <-> @G\n", tmp, SECTION_GET_CADDRESS(code)));

    if (!GENERIC_EQ(tmp,run))
    {
      /* add $d symbols for the padding */
      SymbolTableAddSymbolToSection(obj->sub_symbol_table, LOCAL_SYMBOL, T_RELOCATABLE(tmpsec), run, "$d", TRUE);
    }
    run = GENERIC_ADD(tmp,SECTION_CSIZE(code));
  } /* }}} */

  /* {{{ seventh step: save information about the original code sections in 
   * obj->orig_codes */
  obj->n_orig_codes = obj->ncodes;
  obj->orig_codes = Calloc(obj->ncodes,sizeof(t_sec_descriptor));
  for (tel=0; tel<obj->ncodes; tel++)
  {
    obj->orig_codes[tel].name = StringDup(obj->code[tel]->name);
    obj->orig_codes[tel].base = SECTION_GET_CADDRESS(obj->code[tel]);
    obj->orig_codes[tel].size = SECTION_CSIZE(obj->code[tel]);
    obj->orig_codes[tel].obase = SECTION_GET_OLD_ADDRESS(obj->code[tel]);
    obj->orig_codes[tel].osize = SECTION_OLD_SIZE(obj->code[tel]);
    obj->orig_codes[tel].type = CODE_SECTION;
  } /* }}} */
  
  /* {{{ eighth step: remove the old code sections */
  for (tel=0; tel<obj->ncodes; tel++)
    SectionFree(obj->code[tel]);
  obj->ncodes=1;
  Free(obj->code);
  obj->code=Malloc(sizeof(t_section *));
  obj->code[0] = tmpsec;
  /* }}} */

  /* ninth step: rebuild the list of $switch symbols */
  ObjectCreateDataOrCodeTable(obj);
#ifdef VERBOSE_MERGING
  DiabloPrint(stdout,"Total size=@G\n",size);
#endif
}
/* }}} */

t_segment * ObjectGetSegmentByName(t_object * obj, t_string name)
{
  t_segment * ret;
  for (ret = obj->segment_first; ret; ret = ret->next)
    if (!strcmp(ret->name,name))
      break;
  return ret;
}

void ObjectAddSegment(t_object * obj, t_segment * seg)
{
  seg->next = obj->segment_first;
  if (seg->next) seg->next->prev = seg;
  obj->segment_first = seg;
  if (!obj->segment_last)
    obj->segment_last = seg;
}

typedef t_section * t_section_ptr;
static int _sec_reorder(const void * a, const void * b)
{
  const t_section_ptr * seca = a;
  const t_section_ptr * secb = b;

  return GENERIC_LT(SECTION_GET_OLD_ADDRESS(*seca), SECTION_GET_OLD_ADDRESS(*secb)) ? -1 : 1;
}

void ObjectMergeRODataSections(t_object *obj)
{
  t_uint32 tel;
  t_bool ok=TRUE;

  t_address start;
  t_address run,size;
  t_section * tmpsec;
  t_object * sub, * tmp;
  t_reloc * rel;
  t_symbol * sym;

  t_section* lita_section=NULL;

  qsort(obj->rodata,obj->nrodatas,sizeof(t_section*),_sec_reorder);
 
  start = run = SECTION_GET_CADDRESS(obj->rodata[0]);

  if (!strcmp(T_SECTION(obj->rodata[0])->name,".lita"))
    start = run = SECTION_GET_CADDRESS(obj->rodata[1]);

  for (tel=0; tel<obj->nrodatas; tel++)
    {
      t_section * rodata = obj->rodata[tel];
      if (!strcmp(rodata->name,".lita"))
	continue;
      run = AlignAddress(SECTION_ALIGN(rodata),run);
      if (!GENERIC_EQ(run, SECTION_GET_CADDRESS(rodata)))
	ok = FALSE;
      run = GENERIC_ADD(run,SECTION_CSIZE(rodata));
    }
  size = GENERIC_SUB(run,start);
  
  /* second step: create new section from the different code sections */
  tmpsec = SectionNew(obj,RO_DATA_SECTION);
  SECTION_OLD_SIZE(tmpsec) = size;
  SECTION_CSIZE(tmpsec) = size;
  SectionSetCAddressSuper(tmpsec,start);
  if (!strcmp(T_SECTION(obj->rodata[0])->name,".lita"))
    SectionSetOldAddressSuper(tmpsec,SECTION_GET_OLD_ADDRESS(obj->rodata[1]));
  else
    SectionSetOldAddressSuper(tmpsec,SECTION_GET_OLD_ADDRESS(obj->rodata[0]));
  tmpsec->name = StringDup(".rodata");
  tmpsec->data = Calloc(sizeof(char),GENERIC_EXTRACT_UINT32(size));

  /* insert in section hash table */

  HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(tmpsec->name,tmpsec));

  for (tel=0; tel<obj->nrodatas; tel++)
  {
    t_section * rodata = obj->rodata[tel];
    t_uint32 offset;
    t_uint32 size;
    if (!strcmp(rodata->name,".lita"))
      continue;
    /*    printf("%s ", rodata->name);*/
    offset = GENERIC_EXTRACT_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(rodata),SECTION_GET_CADDRESS(tmpsec)));
    size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(rodata));
    /*    printf("%x %p %p %lx\n",offset,((char *)tmpsec->data)+offset,rodata->data,(size_t) size);fflush(stdout);*/
    memcpy(((char *)tmpsec->data)+offset,rodata->data,(size_t) size);
  }
  /*
  {
    FILE* fp=fopen("data","w");
    int i;
    fwrite(tmpsec->data,size,1,fp);
    fclose(fp);
  }
  */
  /* third step: make all subsections point to the new section as their parent section */
  OBJECT_FOREACH_SUBOBJECT(obj,sub,tmp)
    {
      /*      printf("nrodata's %d\n",sub->nrodatas);*/
      for (tel=0; tel<sub->nrodatas; tel++)
	{
	  /*	  printf("hoera: subobject in %s\n",T_SECTION(sub->rodata[tel])->name);*/
	  if (!strcmp(T_SECTION(sub->rodata[tel])->name,".lita"))
	    continue;
	  SECTION_PARENT_SECTION(sub->rodata[tel]) = tmpsec;
	}
    } 
  /*
  {
    int nsecs;
    GetSubsecsForSec(obj->rodata[0],&nsecs);
    printf("NUMBER SUBSECS %d\n",nsecs);
  }
  */
  /* fourth step: make all relocations point to the right section */ 
  
  OBJECT_FOREACH_RELOC(obj,rel)
    {
      if (REL_TO_RELOCATABLE(rel)->type == RT_SECTION)
	if (T_SECTION(REL_TO_RELOCATABLE(rel))->type == RO_DATA_SECTION && strcmp(T_SECTION(REL_TO_RELOCATABLE(rel))->name,".lita"))
	  {
	    t_address offset = GENERIC_SUB(SECTION_GET_CADDRESS(T_SECTION(REL_TO_RELOCATABLE(rel))),SECTION_GET_CADDRESS(tmpsec));
	    RelocSetToRelocatable(rel, T_RELOCATABLE(tmpsec));
	    rel->to_offset = GENERIC_ADD(rel->to_offset, offset);
	  }
      /* relocations should always be from _sub_section, never from section... */
      if (REL_FROM(rel)->type == RT_SECTION) FATAL(("Relocations corrupt"));
    } 

  /* fifth step: migrate all symbols that point to the old code sections */
  
  for (sym = obj->sub_symbol_table->first; sym; sym = sym->next)
    {
      if (sym->base && sym->base->type == RT_SECTION && T_SECTION(sym->base)->type == RO_DATA_SECTION && strcmp(T_SECTION(sym->base)->name,".lita"))
	{
	  t_address offset = GENERIC_SUB(SECTION_GET_CADDRESS(sym->base),start);
	  sym->offset_from_start = GENERIC_ADD(sym->offset_from_start,offset);
	  sym->current_address = GENERIC_ADD(start,sym->offset_from_start);
	  sym->base = T_RELOCATABLE(tmpsec);
	}
    } 

  /* sixth step: remove the old rodata sections */
  for (tel=0; tel<obj->nrodatas; tel++)
    {
      if (strcmp(T_SECTION(obj->rodata[tel])->name,".lita"))
	{
	  /*	  printf("%d %s\n",tel,T_SECTION(obj->rodata[tel])->name);fflush(stdout);*/
	  SectionFree(obj->rodata[tel]);
	}
      else
	lita_section = T_SECTION(obj->rodata[tel]);
    }

  obj->nrodatas=2;
  Free(obj->rodata);
  obj->rodata=Malloc(2*sizeof(t_section *));
  obj->rodata[0] = tmpsec;
  obj->rodata[1] = lita_section;

  {
    FILE* fp=fopen("data","w");
    fwrite(tmpsec->data,GENERIC_EXTRACT_UINT32(size),1,fp);
    fclose(fp);
  }


  /*  printf("LITA %p\n",lita_section);*/

  /* ninth step: rebuild the list of $switch symbols */
  ObjectCreateDataOrCodeTable(obj);
}

void ObjectMoveBssToDataSections(t_object *obj)
{  
  t_uint32 tel;
  t_bool ok=TRUE;
  
  t_address start;
  t_address run,size;
  t_section * tmpsec;
  t_object * sub, * tmp;
  t_symbol * sym;

  qsort(obj->data,obj->ndatas,sizeof(t_section*),_sec_reorder);
  qsort(obj->bss,obj->nbsss,sizeof(t_section*),_sec_reorder);
 
  start = run = SECTION_GET_CADDRESS(obj->data[0]);

  for (tel=0; tel<obj->ndatas; tel++)
    {
      t_section * data = obj->data[tel];
      run = AlignAddress(SECTION_ALIGN(data),run);
      if (!GENERIC_EQ(run, SECTION_GET_CADDRESS(data)))
	ok = FALSE;
      run = GENERIC_ADD(run,SECTION_CSIZE(data));
    }

  for (tel=0; tel<obj->nbsss; tel++)
    {
      t_section * bss = obj->bss[tel];
      run = AlignAddress(SECTION_ALIGN(bss),run);
      if (!GENERIC_EQ(run, SECTION_GET_CADDRESS(bss)))
	ok = FALSE;
      run = GENERIC_ADD(run,SECTION_CSIZE(bss));
    }

  size = GENERIC_SUB(run,start);

  tmpsec = SectionNew(obj,DATA_SECTION);
  SECTION_OLD_SIZE(tmpsec) = size;
  SECTION_CSIZE(tmpsec) = size;
  SectionSetCAddressSuper(tmpsec,start);
  SectionSetOldAddressSuper(tmpsec,SECTION_GET_OLD_ADDRESS(obj->data[0]));
  tmpsec->name = StringDup(".wridata");
  tmpsec->data = Calloc(sizeof(char),GENERIC_EXTRACT_UINT32(size));

  HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(tmpsec->name,tmpsec));

  for (tel=0; tel<obj->ndatas; tel++)
    {
      t_section * data = obj->data[tel];
      t_uint32 offset = GENERIC_EXTRACT_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(data),SECTION_GET_CADDRESS(tmpsec)));
      t_uint32 size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(data));
      memcpy(((char *)tmpsec->data)+offset,data->data,(size_t) size);
    }

  for (tel=0; tel<obj->nbsss; tel++)
    {
      t_section * bss = obj->bss[tel];
      t_uint32 offset = GENERIC_EXTRACT_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(bss),SECTION_GET_CADDRESS(tmpsec)));
      t_uint32 size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(bss));
      bss->data = Calloc((size_t) size,1);
      memcpy(((char *)tmpsec->data)+offset,bss->data,(size_t) size);
    }
  
  OBJECT_FOREACH_SUBOBJECT(obj,sub,tmp)
    {
      for (tel=0; tel<sub->ndatas; tel++)
	{
	  SECTION_PARENT_SECTION(sub->data[tel]) = tmpsec;
	}

      for (tel=0; tel<sub->nbsss; tel++)
	{
	  t_uint32 size;
	  SECTION_PARENT_SECTION(sub->bss[tel]) = tmpsec;
	  T_SECTION(sub->bss[tel])->type = DATA_SECTION;
	  size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sub->bss[tel]));
	  sub->bss[tel]->data = Calloc((size_t) size,1);
	}
    } 

  for (sym = obj->sub_symbol_table->first; sym; sym = sym->next)
    {
      if (sym->base && sym->base->type == RT_SECTION && (T_SECTION(sym->base)->type == DATA_SECTION || T_SECTION(sym->base)->type == BSS_SECTION))
	{
	  t_address offset = GENERIC_SUB(SECTION_GET_CADDRESS(sym->base),start);
	  sym->offset_from_start = GENERIC_ADD(sym->offset_from_start,offset);
	  sym->current_address = GENERIC_ADD(start,sym->offset_from_start);
	  sym->base = T_RELOCATABLE(tmpsec);
	  T_SECTION(sym->base)->type = DATA_SECTION;
	}
    } 
  
  for (tel=0; tel<obj->ndatas; tel++)
    {
      SectionFree(obj->data[tel]);
    }

  for (tel=0; tel<obj->nbsss; tel++)
    {
      SectionFree(obj->bss[tel]);
    }
  
  obj->ndatas=1;
  obj->nbsss=0;
  Free(obj->data);
  Free(obj->bss);
  obj->bss=NULL;
  obj->data=Malloc(sizeof(t_section *));
  obj->data[0] = tmpsec;
  
  /* ninth step: rebuild the list of $switch symbols */
  ObjectCreateDataOrCodeTable(obj);

}


void ObjectMergeDataSections(t_object *obj)
{
  t_uint32 tel;
  t_bool ok=TRUE;

  t_address start;
  t_address run,size;
  t_section * tmpsec;
  t_object * sub, * tmp;
  t_reloc * rel;
  t_symbol * sym;

  /* check whether sections are contiguously */

  qsort(obj->data,obj->ndatas,sizeof(t_section*),_sec_reorder);
  qsort(obj->bss,obj->nbsss,sizeof(t_section*),_sec_reorder);
 
  start = run = SECTION_GET_CADDRESS(obj->data[0]);

  for (tel=0; tel<obj->ndatas; tel++)
    {
      t_section * data = obj->data[tel];
      run = AlignAddress(SECTION_ALIGN(data),run);
      if (!GENERIC_EQ(run, SECTION_GET_CADDRESS(data)))
	ok = FALSE;
      run = GENERIC_ADD(run,SECTION_CSIZE(data));
    }
  size = GENERIC_SUB(run,start);
  
  /* second step: create new section from the different code sections */
  tmpsec = SectionNew(obj,DATA_SECTION);
  SECTION_OLD_SIZE(tmpsec) = size;
  SECTION_CSIZE(tmpsec) = size;
  SectionSetCAddressSuper(tmpsec,start);
  SectionSetOldAddressSuper(tmpsec,SECTION_GET_OLD_ADDRESS(obj->data[0]));
  tmpsec->name = StringDup(".alldata");
  tmpsec->data = Calloc(sizeof(char),GENERIC_EXTRACT_UINT32(size));

  /* insert in section hash table */

  HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(tmpsec->name,tmpsec));

  for (tel=0; tel<obj->ndatas; tel++)
  {
    t_section * data = obj->data[tel];
    t_uint32 offset = GENERIC_EXTRACT_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(data),SECTION_GET_CADDRESS(tmpsec)));
    t_uint32 size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(data));
    memcpy(((char *)tmpsec->data)+offset,data->data,(size_t) size);
  }
  /* third step: make all subsections point to the new section as their parent section */
  OBJECT_FOREACH_SUBOBJECT(obj,sub,tmp)
    {
      for (tel=0; tel<sub->ndatas; tel++)
	{
	  SECTION_PARENT_SECTION(sub->data[tel]) = tmpsec;
	}
    } 

  /* fourth step: make all relocations point to the right section */ 
  
  OBJECT_FOREACH_RELOC(obj,rel)
    {
      if (REL_TO_RELOCATABLE(rel)->type == RT_SECTION)
	if (T_SECTION(REL_TO_RELOCATABLE(rel))->type == DATA_SECTION)
	  {
	    t_address offset = GENERIC_SUB(SECTION_GET_CADDRESS(T_SECTION(REL_TO_RELOCATABLE(rel))),SECTION_GET_CADDRESS(tmpsec));
	    RelocSetToRelocatable(rel, T_RELOCATABLE(tmpsec));
	    rel->to_offset = GENERIC_ADD(rel->to_offset, offset);
	  }
      /* relocations should always be from _sub_section, never from section... */
      if (REL_FROM(rel)->type == RT_SECTION) FATAL(("Relocations corrupt"));
    } 

  /* fifth step: migrate all symbols that point to the old code sections */
  
  for (sym = obj->sub_symbol_table->first; sym; sym = sym->next)
    {
      if (sym->base && sym->base->type == RT_SECTION && T_SECTION(sym->base)->type == DATA_SECTION)
	{
	  t_address offset = GENERIC_SUB(SECTION_GET_CADDRESS(sym->base),start);
	  sym->offset_from_start = GENERIC_ADD(sym->offset_from_start,offset);
	  sym->current_address = GENERIC_ADD(start,sym->offset_from_start);
	  sym->base = T_RELOCATABLE(tmpsec);
	}
    } 

  /* sixth step: remove the old rodata sections */
  for (tel=0; tel<obj->ndatas; tel++)
    {
      SectionFree(obj->data[tel]);
    }

  obj->ndatas=1;
  Free(obj->data);
  obj->data=Malloc(sizeof(t_section *));
  obj->data[0] = tmpsec;

  /* ninth step: rebuild the list of $switch symbols */
  ObjectCreateDataOrCodeTable(obj);
}

typedef struct _t_new_layout_table t_new_layout_table;

struct _t_new_layout_table
{
  t_address orig;
  t_address new;
};

/* Commented out by Bruno, does not compile for 32 bit architectures!
 *
 * GenericSet_uint64 cannot be used on these arches
 * */

void ObjectRelayoutData(t_object * obj, t_string filename)
{
  t_section * sec;
  t_object * subobj,*tmp;
  t_uint32 tel;
#ifdef BIT64ADDRSUPPORT
  t_uint64 orig,new,padding,size;
#else
  t_uint32 orig,new,padding,size;
#endif
  char name[100];

  int nr_sections = 0;
  int table_size =0;

  FILE * fp = fopen(filename,"r");
  FILE * op = fopen("oldlayout","w");
  
  t_new_layout_table *table;
  t_address entry=NULL_GENERIC;

  OBJECT_FOREACH_SUBOBJECT(obj,subobj,tmp)
    OBJECT_FOREACH_SECTION(subobj,sec,tel)
    {
      nr_sections++;
#ifdef BIT64ADDRSUPPORT
      fprintf(op,"%s %llx %llx %llx\n",SECTION_PARENT_SECTION(sec)->name, G_T_UINT64(SECTION_GET_CADDRESS(sec)),G_T_UINT64(SECTION_CSIZE(sec)),G_T_UINT64(SECTION_ALIGN(sec)));
#else
      fprintf(op,"%s %x %x %x\n",SECTION_PARENT_SECTION(sec)->name, G_T_UINT32(SECTION_GET_CADDRESS(sec)),G_T_UINT32(SECTION_CSIZE(sec)),G_T_UINT32(SECTION_ALIGN(sec)));
#endif
    }

  fclose(op);

  table = Calloc(nr_sections,sizeof(t_new_layout_table));
  
  while(fscanf(fp,"%s",name)>0)
    {
#ifdef BIT64ADDRSUPPORT
      fscanf(fp,"%llx",&orig);
      fscanf(fp,"%llx",&size);
      fscanf(fp,"%llx",&padding);
      fscanf(fp,"%llx",&new);
      GenericSet_uint64(table[table_size].orig,orig);
      GenericSet_uint64(table[table_size++].new,new);
#else
      fscanf(fp,"%x",&orig);
      fscanf(fp,"%x",&size);
      fscanf(fp,"%x",&padding);
      fscanf(fp,"%x",&new);
      GenericSet_uint32(table[table_size].orig,orig);
      GenericSet_uint32(table[table_size++].new,new);
#endif
    }

  fclose(fp);

  printf("table size: %d\n",table_size);

  OBJECT_FOREACH_SUBOBJECT(obj,subobj,tmp)
    {
      OBJECT_FOREACH_SECTION(subobj,sec,tel)
	{
	  int i;
	  for (i=0;i<table_size;i++)
	    {
	      if (GENERIC_EQ(table[i].orig,SECTION_GET_CADDRESS(sec)))
		{
		  if (GENERIC_EQ(table[i].orig,obj->entry))
		    {
		      /*		      DiabloPrint(stdout,"old entry @G new @G\n",obj->entry,table[i].new);*/
		      entry = table[i].new;
		    }
		  /*		  DiabloPrint(stdout,"compare @G with @G\n",table[i].orig,SECTION_GET_CADDRESS(sec));*/
		  SECTION_GET_CADDRESS(sec)=table[i].new;
		  break;
		}
	    }
	}	
    }
  

  OBJECT_FOREACH_SECTION(obj,sec,tel)
    {
#ifdef BIT64ADDRSUPPORT
      GenericSet_uint64(SECTION_GET_CADDRESS(sec),0x0ffffffffffffffLL);
      GenericSet_uint64(SECTION_CSIZE(sec),0x0LL);
#else
      GenericSet_uint32(SECTION_GET_CADDRESS(sec),0x0fffffff);
      GenericSet_uint32(SECTION_CSIZE(sec),0x0);
#endif
    }

  OBJECT_FOREACH_SUBOBJECT(obj,subobj,tmp)
    {
      OBJECT_FOREACH_SECTION(subobj,sec,tel)
	{
	  if (GENERIC_GT(SECTION_GET_CADDRESS(SECTION_PARENT_SECTION(sec)),SECTION_GET_CADDRESS(sec)))
	    SECTION_GET_CADDRESS(SECTION_PARENT_SECTION(sec))=SECTION_GET_CADDRESS(sec);
	  
	  if (GENERIC_LT(SECTION_CSIZE(SECTION_PARENT_SECTION(sec)),SECTION_GET_CADDRESS(sec)+SECTION_CSIZE(sec)))
	    SECTION_CSIZE(SECTION_PARENT_SECTION(sec))=SECTION_GET_CADDRESS(sec)+SECTION_CSIZE(sec);
	}
    }


  OBJECT_FOREACH_SECTION(obj,sec,tel)
    {
#ifdef BIT64ADDRSUPPORT
      if (G_T_UINT64(SECTION_GET_CADDRESS(sec))>LL(0x140000000))
	{
	  GenericSet_uint64(SECTION_GET_CADDRESS(sec),LL(0x140000000));
	  printf("yes\n");
	}
      printf("SECTION1 : %s %llx\n",sec->name,G_T_UINT64(SECTION_GET_CADDRESS(sec)));
#endif      
      SECTION_CSIZE(sec)-=SECTION_GET_CADDRESS(sec);
    }

  ObjectRelocateSubObjects(obj,FALSE);
  ObjectRebuildSectionsFromSubsections(obj);
  obj->entry=entry;

}

/* vim: set shiftwidth=2 foldmethod=marker: */
