/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//**********************************************************
 * FILE: diablo_object.c
 *
 * DESC: This file contains the main object-file routines 
 * 
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "../config.h"
#include "diablo_object_tables.h"
#include "../simple_types.h"

/* OBJECT FILE FORMATS */
/* All (supported) ELF types */
#ifdef ELF_SUPPORT
#include "../fileformats/elf/elf_main.h"    
#endif 
/* Arm Image format */
#ifdef AIF_SUPPORT
#include "../fileformats/aif/aif_main.h"    
#endif
/* Dos and PE format */
#ifdef PE_SUPPORT
#include "../fileformats/pe/pe_main.h"     
#endif
/* Alpha (ecoff) executables */ 
#ifdef ECOFF_SUPPORT
#include "../fileformats/alpha_ecoff/ecoff_main.h"  
#endif
/* Java class files */
#ifdef JAVA_SUPPORT
#include "../fileformats/java/java_main.h"  
#include "../arch/java/java_actions.h"
#include "../arch/java/java_description.h"
#endif

#ifdef PALM_SUPPORT
#include "../fileformats/palm/my_prc.h"
#endif
/* END OBJECT FILE FORMATS */

/* INSTRUCTION ARCHITECTURES */
#ifdef I386_SUPPORT
#include "../arch/i386/i386_actions.h" 
#include "../arch/i386/i386_description.h" 
#endif
#ifdef ALPHA_SUPPORT
#include "../arch/alpha/alpha_actions.h" 
#include "../arch/alpha/alpha_description.h" 
#endif
#ifdef ARM_SUPPORT
#include "../arch/arm/arm_actions.h"
#include "../arch/arm/arm_description.h"
#endif
#ifdef SH_SUPPORT
#include "../arch/sh3/sh3_actions.h"
#include "../arch/sh3/sh3_description.h"
#endif
#ifdef IA64_SUPPORT
#include "../arch/ia64/ia64_actions.h"
#include "../arch/ia64/ia64_description.h"
#endif
#ifdef MIPS_SUPPORT
#include "../arch/mips/mips_actions.h"
#include "../arch/mips/mips_description.h"
#endif
#ifdef POWERPC_SUPPORT
#include "../arch/powerpc/powerpc_actions.h"
#include "../arch/powerpc/powerpc_description.h"
#endif



/* END INSTRUCTION ARCHITECTURES */

/* Utilities used: */
#include "../datastructures/data_string.h"
#include "../utils/util_file.h"
#include "../utils/util_malloc.h"
#include "../utils/util_options.h"
#include "../utils/util_status.h"
#include "../utils/util_error.h"
#include "../utils/util_term.h"
/* Support for archives */
#include "../fileformats/ar/ar_arch.h"
/* Symbols */
#include "diablo_symbol.h"
#include "diablo_reloc.h"
/* Maps */
#include "diablo_parsemap.h"
/* Printing */
#include "diablo_print.h"

/* DOC: Objecthandlers: used to open the different objecttypes */


#ifdef ELF_SUPPORT
const t_object_handler elfhandler[]=
{
#ifdef ARM_SUPPORT
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_ARMELFLSB, IsElfArmSameEndianOnLsb,      ElfReadArmSameEndian,       NULL, &arm_actions, &elf_actions, &arm_description, "Arm  Lsb  on  Lsb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_ARMELFMSB, IsElfArmSameEndianOnMsb,      ElfReadArmSameEndian,       NULL, NULL,         &elf_actions, NULL, "Arm  Msb  on  Msb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_ARMELFLSB, IsElfArmSwitchedEndianOnLsb,  ElfReadArmSwitchedEndian,   NULL, &arm_actions, &elf_actions, NULL, "Arm  Lsb  on  Msb  Machine"}, 
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_ARMELFMSB, IsElfArmSwitchedEndianOnMsb,  ElfReadArmSwitchedEndian,   NULL, NULL,         &elf_actions, NULL, "Arm  Msb  on  Lsb  Machine"}, 
#endif
#ifdef SH_SUPPORT
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_SH3ELFLSB, IsElfSh3SameEndianOnLsb,      ElfReadSh3SameEndian,       NULL, &sh3_actions, &elf_actions, &sh3_description, "Sh3  Lsb  on  Lsb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_SH3ELFMSB, IsElfSh3SameEndianOnMsb,      ElfReadSh3SameEndian,       NULL, NULL,         &elf_actions, &sh3_description, "Sh3  Msb  on  Msb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_SH3ELFLSB, IsElfSh3SwitchedEndianOnMsb,  ElfReadSh3SwitchedEndian,   NULL, NULL,         &elf_actions, &sh3_description, "Sh3  Lsb  on  Msb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_SH3ELFMSB, IsElfSh3SwitchedEndianOnLsb,  ElfReadSh3SwitchedEndian,   NULL, NULL,         &elf_actions, &sh3_description, "Sh3  Msb  on  Lsb  Machine"},
#endif
#ifdef MIPS_SUPPORT
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_MIPSELF, IsElfMipsSameEndianOnLsb,      ElfReadMipsSameEndian,       NULL, &mips_actions, &elf_actions, &mips_description, "Mips  on  Lsb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_MIPSELF, IsElfMipsSwitchedEndianOnMsb,  ElfReadMipsSwitchedEndian,   NULL, NULL,          &elf_actions, &mips_description, "Mips on  Msb  Machine"},
#endif
#ifdef SPARC_SUPPORT
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_SPARCELF,  IsElfSparcSwitchedEndianOnLsb,ElfReadSparcSwitchedEndian, NULL, NULL,         &elf_actions, NULL, "Sparc   on   Lsb   Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_SPARCELF,  IsElfSparcSameEndianOnMsb,    ElfReadSparcSameEndian,     NULL, NULL,         &elf_actions, NULL, "Sparc   on   Msb   Machine"},
#endif
#ifdef I386_SUPPORT
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_I386ELF    ,IsElfI386SameEndianOnLsb,      ElfReadI386SameEndian,     NULL, &i386_actions,&elf_actions, &i386_description, "Linux   on   Lsb   Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_I386ELF    ,IsElfI386SwitchedEndianOnMsb,  ElfReadI386SwitchedEndian, NULL, NULL,         &elf_actions, NULL, "Linux   on   Msb   Machine"},
#endif
#ifdef ALPHA_SUPPORT
   //   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_ALPHAELF  ,IsElfAlphaSameEndianOnLsb,    ElfReadAlphaSameEndian,     NULL, NULL,         &elf_actions, NULL, "Alpha Linux on Lsb Machine"},
   //   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_ALPHAELF  ,IsElfAlphaSwitchedEndianOnMsb,ElfReadAlphaSwitchedEndian, NULL, NULL,         &elf_actions, NULL, "Alpha Linux on Msb Machine"},
#endif
#ifdef IA64_SUPPORT
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_IA64ELF,IsElfIA64SameEndianOnLsb,     ElfReadIA64SameEndian,      NULL, &ia64_actions, &elf_actions, &ia64_description, "Itanium   on  Lsb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_IA64ELF,IsElfIA64SwitchedEndianOnMsb, ElfReadIA64SwitchedEndian,  NULL, NULL,         &elf_actions, NULL, "Itanium   on  Msb  Machine"},
#endif
#ifdef POWERPC_SUPPORT
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_POWERPCELFLSB, IsElfPowerpcSameEndianOnLsb,      ElfReadPowerpcSameEndian,       NULL, &powerpc_actions, &elf_actions, &powerpc_description, "Ppc  Lsb  on  Lsb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_POWERPCELFMSB, IsElfPowerpcSameEndianOnMsb,      ElfReadPowerpcSameEndian,       NULL, &powerpc_actions,         &elf_actions, &powerpc_description, "Ppc  Msb  on  Msb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_POWERPCELFLSB, IsElfPowerpcSwitchedEndianOnMsb,  ElfReadPowerpcSwitchedEndian,   NULL, &elf_actions, NULL,         &powerpc_description, "Ppc  Lsb  on  Msb  Machine"},
   {OBJECTHANDLER, NULL, OBJECT_FILE_TYPE_ELF, OBJECT_FILE_SUBTYPE_POWERPCELFMSB, IsElfPowerpcSwitchedEndianOnLsb,  ElfReadPowerpcSwitchedEndian,   NULL, &elf_actions, NULL,         &powerpc_description, "Ppc  Msb  on  Lsb  Machine"},
#endif

   {OBJECTHANDLER, NULL, T_UNKNOWN,ST_UNKNOWN, NULL,                         NULL,                       NULL, NULL,         NULL, NULL, NULL}    
};
#endif



 
const t_object_handler handler_table[]=
{
#ifdef ELF_SUPPORT
  {GROUPHANDLER , elfhandler, OBJECT_FILE_TYPE_ELF,   ST_UNKNOWN, IsElf,    ReadElf,    NULL,   NULL,            &elf_actions, NULL,   "Elf"},             
#endif
#ifdef AIF_SUPPORT
#ifdef ARM_SUPPORT
  {OBJECTHANDLER, NULL,        OBJECT_FILE_TYPE_AIF,   ST_UNKNOWN, IsAif,    ReadAif,    NULL,   &arm_actions,   NULL, NULL,"Arm Image Format"},
#endif
#endif
#ifdef ECOFF_SUPPORT
#ifdef ALPHA_SUPPORT
  {OBJECTHANDLER, NULL,        OBJECT_FILE_TYPE_ECOFF, ST_UNKNOWN, IsECoff,  ReadECoff,  NULL,   &alpha_actions, &ecoff_actions, &alpha_description,"Alpha Ecoff"},
#endif
#endif
#ifdef PE_SUPPORT
  {OBJECTHANDLER, NULL,        OBJECT_FILE_TYPE_DOS,   ST_UNKNOWN, IsDosExe, ReadDosExe, NULL,   &i386_actions,  NULL, NULL,"Dos Executables"},
   /* This should become a grouphandler, since PE_COFF supports different architectures ! */
  {OBJECTHANDLER, NULL,        OBJECT_FILE_TYPE_PE,    ST_UNKNOWN, IsPE,     ReadPE,     NULL,   &i386_actions,  NULL, &i386_description,"ix86 Windows PE"},
#endif
#ifdef JAVA_SUPPORT
  {OBJECTHANDLER, NULL,        OBJECT_FILE_TYPE_JAVA,  ST_UNKNOWN, IsJava ,  ReadJava,   NULL,   &java_actions,  NULL, &java_description,"Java Class Files"},
#endif
#ifdef PALM_SUPPORT
  {OBJECTHANDLER, NULL,        OBJECT_FILE_TYPE_PALM,  ST_UNKNOWN, IsPRC,    ReadPRC,    NULL,   NULL,           NULL, NULL,"Palm executables"},
#endif
  {NOHANDLER    , NULL,        T_UNKNOWN, ST_UNKNOWN, NULL,   NULL,       NULL,   NULL,           NULL, NULL,NULL} 
  
};



/*!  
 * \param fp The ouput stream 
 * \param handler the handler table to print
 * \param current indent value (0 to start, incremented when recursing)
 *
 * Print a list of known fileformats and corresponding actions. Recursive calls
 * for all group handlers. Function is static: should use ObjectKnownFileFormats. */ 

static void PrintHandler(FILE * fp, const t_object_handler * handler, int indent)
{
  int tel=0;

  while(handler[tel].type!=T_UNKNOWN)
    {
      if (handler[tel].htype==GROUPHANDLER)
	{
	  TermPrintClean(fp,handler[tel].name,0,5+indent,FINAL);
	  PrintHandler(fp,handler[tel].subhandler,indent+2);
	}
      else
	{
	  int ret;
	  ret=TermPrintClean(fp,handler[tel].name,0,5+indent,0);
	  ActionPrint(ret,handler[tel].acts);
	}
      tel++;
    }
}

void ObjectKnownFileFormats(FILE * fp)
{
  TermTitle(fp,"Recognized ObjectFile Formats:",'=');
  PrintHandler(fp,handler_table,0);
}
/* vim: set shiftwidth=2 foldmethod=marker: */
