/******************************************************************************
 * Copyright 2001,2002,2003 {{{: 
 *
 * Bruno De Bus 
 * Bjorn De Sutter 
 * Dominique Chanet 
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: Prints various structures in diablo
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/

/* Includes {{{ */
#include "diablo_print.h"
#include "diablo_liveness.h"
#include "diablo_iterators.h"
#include "diablo_profile.h"
#include "../utils/util_error.h"
#include "../utils/util_malloc.h"
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
/* }}} */

#define INT 1
#define FLT 2

extern t_bbl * ALL_BBLS;

#define ADVANCE(tmp) /*run+=strlen(tmp);*/ tmp+=strlen(tmp); 

static const t_architecture_description * arch_desc = NULL;

void PrintRegsetToString(char* tmp,const t_architecture_description * desc,t_regset regs)
{
  t_bool flag = 0;
  t_reg reg;
  t_regset allregs = desc->all_registers;
  
  sprintf(tmp,"{");
  ADVANCE(tmp);
  RegsetSetIntersect(regs,allregs);

  REGSET_FOREACH_REG(regs,reg)
  {
    if (flag)
      sprintf(tmp,", %s",desc->register_names[reg]);
    else
      sprintf(tmp,"%s",desc->register_names[reg]);
    flag = 1;
    ADVANCE(tmp);
  }
  sprintf(tmp,"}");
  ADVANCE(tmp);
}

void PrintRegset(FILE *f,const t_architecture_description * desc,t_regset regs)
{
  t_bool flag = 0;
  t_reg reg;
  t_regset allregs = desc->all_registers;

  fprintf(f,"{");
  RegsetSetIntersect(regs,allregs);

  REGSET_FOREACH_REG(regs,reg)
  {
    if (flag)
      fprintf(f,", %s",desc->register_names[reg]);
    else
      fprintf(f,"%s",desc->register_names[reg]);
    flag = 1;
  }
  fprintf(f,"}\n");
}

static void PrintRegLattice(FILE *f, t_procstate * state, t_reg reg, int format) 
{
  register_content value;
  t_reloc *reloc;
  t_lattice_level result = ProcStateGetReg(state,reg,&value);
  t_lattice_level tag_result = ProcStateGetTag(state,reg,&reloc);
  
  if (result==BOT)
    {
      fprintf(f,"B");
    } 
  else if (result==TOP)
    {
      fprintf(f,"T");
    }
  else
    {
      if (format==INT)
	{
	  DiabloPrint(f, "@G (@G) ",value.i,value.i);
	}
    }

  if (tag_result==BOT)
    {
      fprintf(f," tB");
    } 
  else if (tag_result==TOP)
    {
      fprintf(f," tT");
    }
  else
    {
      fprintf(f, " reloc->to (%p->%p)",reloc,REL_TO_RELOCATABLE(reloc));
      /* PrintReloc(f,reloc); */
    }
}

static void PrintRegLatticeToString(char * tmp, t_procstate * state, t_reg reg, int format) 
{
  register_content value;
  t_reloc *reloc;
  t_lattice_level result = ProcStateGetReg(state,reg,&value);
  t_lattice_level tag_result = ProcStateGetTag(state,reg,&reloc);
  
  if (result==BOT)
    {
      sprintf(tmp,"B");
      ADVANCE(tmp);
    } 
  else if (result==TOP)
    {
      sprintf(tmp,"T");
      ADVANCE(tmp);
    }
  else
    {
      if (format==INT)
	{
	  sprintf(tmp,sDiabloPrint("@G",value.i));
	  ADVANCE(tmp);
	}
    }
#if 1
  if (tag_result==BOT)
    {
      sprintf(tmp," tB");
      ADVANCE(tmp);
    } 
  else if (tag_result==TOP)
    {
      sprintf(tmp," tT");
      ADVANCE(tmp);
    }
  else
    {
      sprintf(tmp, " reloc->to (%p->%p)",reloc,REL_TO_RELOCATABLE(reloc));
      ADVANCE(tmp);
      /* PrintReloc(f,reloc); */
    }
#endif
}

static void PrintConditionLattice(FILE *f, const t_architecture_description * desc, t_procstate * state, t_regset regs) {
  t_reg reg;
  t_bool value;

  REGSET_FOREACH_REG(regs,reg)
    {
      t_lattice_level cond = ProcStateGetCond(state,reg,&value);
      fprintf(f,"%s: ", desc->register_names[reg]);

      if (cond==TOP)
	fprintf(f,"T ");
      else if (cond==BOT)
	fprintf(f,"B ");
      else if (value)
	fprintf(f,"1 ");
      else 
	fprintf(f,"0 ");
    }
}
/*
static void PrintConditionLatticeToString(char *tmp, const t_architecture_description * desc, t_procstate * state, t_regset regs) {
  t_reg reg;
  t_bool value;

  REGSET_FOREACH_REG(regs,reg)
    {
      t_lattice_level cond = ProcStateGetCond(state,reg,&value);
      sprintf(tmp,"%s: ", desc->register_names[reg]);
      ADVANCE(tmp);

      if (cond==TOP)
	sprintf(tmp,"T ");
      else if (cond==BOT)
	sprintf(tmp,"B ");
      else if (value)
	sprintf(tmp,"1 ");
      else 
	sprintf(tmp,"0 ");
      ADVANCE(tmp);
    }
}
*/
void ProcStatePrint(FILE *f, const t_architecture_description * desc, t_procstate * state) 
{
  t_reg reg;
  t_regset intregs = desc->int_registers;
  t_regset fltregs = desc->flt_registers;


  REGSET_FOREACH_REG(intregs,reg)
  {
    fprintf(f,"%s ", desc->register_names[reg]);
    PrintRegLattice(f,state,reg,INT);fflush(stdout);
    if (!((reg+1)%8)) fprintf(f,"\n"); else fprintf(f," ");
  }

  fflush(stdout);

  REGSET_FOREACH_REG(fltregs,reg)
  {
    fprintf(f,"%s ", desc->register_names[reg]);
    PrintRegLattice(f,state,reg,FLT);
    if (!((reg+1)%8)) fprintf(f,"\n"); else fprintf(f," ");
  }

  PrintConditionLattice(f,desc,state,desc->cond_registers);

  fprintf(f,"\n------------------------------------------\n");
}

void ProcStatePrintOpt(FILE *f, const t_architecture_description * desc, t_procstate * state, t_regset regs, t_uint8 options) 
{
  t_reg reg;
  t_regset intregs = desc->int_registers;
  t_regset fltregs = desc->flt_registers;
  t_regset condregs = desc->cond_registers;
  int i;

  if (options & REG_FILTER)
    {
      RegsetSetIntersect(intregs,regs);
      RegsetSetIntersect(fltregs,regs);
      RegsetSetIntersect(condregs,regs);
    }
  
  i = 0;

  REGSET_FOREACH_REG(intregs,reg)
    { 
      register_content value;
      t_reloc *reloc;
      t_lattice_level result = ProcStateGetReg(state,reg,&value);
      t_lattice_level tag_result = ProcStateGetTag(state,reg,&reloc);
      if (!(options & ONLY_VALUES) || result == VALUE || tag_result == VALUE) 
	{
	  fprintf(f,"%s ", desc->register_names[reg]);
	  PrintRegLattice(f,state,reg,INT);fflush(stdout);
	  if (!((++i)%8)) fprintf(f,"\n"); else fprintf(f," ");
	}
    }

  i = 0;

  REGSET_FOREACH_REG(fltregs,reg)
    { 
      register_content value;
      t_reloc *reloc;
      t_lattice_level result = ProcStateGetReg(state,reg,&value);
      t_lattice_level tag_result = ProcStateGetTag(state,reg,&reloc);
      if (!(options & ONLY_VALUES) || result == VALUE || tag_result == VALUE) 
	{
	  fprintf(f,"%s ", desc->register_names[reg]);
	  PrintRegLattice(f,state,reg,FLT);fflush(stdout);
	  if (!((++i)%8)) fprintf(f,"\n"); else fprintf(f," ");
	}
    }

  PrintConditionLattice(f,desc,state,condregs);

  fprintf(f,"\n-----\n");

}

void ProcStatePrintOptToString(const t_architecture_description * desc, t_procstate * state, t_regset regs, t_uint8 options, char* tmp) 
{
  t_reg reg;
  t_regset intregs = desc->int_registers;
  t_regset fltregs = desc->flt_registers;
  t_regset condregs = desc->cond_registers;
  int i;

  if (options & REG_FILTER)
    {
      RegsetSetIntersect(intregs,regs);
      RegsetSetIntersect(fltregs,regs);
      RegsetSetIntersect(condregs,regs);
    }
  
  i = 0;


  REGSET_FOREACH_REG(intregs,reg)
    { 
      register_content value;
      t_reloc *reloc;
      t_lattice_level result = ProcStateGetReg(state,reg,&value);
      t_lattice_level tag_result = ProcStateGetTag(state,reg,&reloc);
      if (!(options & ONLY_VALUES) || result == VALUE || tag_result == VALUE) 
	{
	  sprintf(tmp,"%s ", desc->register_names[reg]);
	  ADVANCE(tmp);
	  PrintRegLatticeToString(tmp,state,reg,INT);
	  ADVANCE(tmp);
	  if (!((++i)%8)) sprintf(tmp,"\n"); else sprintf(tmp," ");
	  ADVANCE(tmp);
	}
    }

  i = 0;

#if 1
  REGSET_FOREACH_REG(fltregs,reg)
    { 
      register_content value;
      t_reloc *reloc;
      t_lattice_level result = ProcStateGetReg(state,reg,&value);
      t_lattice_level tag_result = ProcStateGetTag(state,reg,&reloc);
      if (!(options & ONLY_VALUES) || result == VALUE || tag_result == VALUE) 
	{
	  sprintf(tmp,"%s ", desc->register_names[reg]);
	  ADVANCE(tmp);
	  PrintRegLatticeToString(tmp,state,reg,FLT);
	  ADVANCE(tmp);
	  if (!((++i)%8)) sprintf(tmp,"\n"); else sprintf(tmp," ");
	  ADVANCE(tmp);
	}
    }
  { 
    t_reg reg;
    t_bool value;

    REGSET_FOREACH_REG(regs,reg)
    {
      t_lattice_level cond = ProcStateGetCond(state,reg,&value);
      sprintf(tmp,"%s: ", desc->register_names[reg]);
      ADVANCE(tmp);

      if (cond==TOP)
	sprintf(tmp,"T ");
      else if (cond==BOT)
	sprintf(tmp,"B ");
      else if (value)
	sprintf(tmp,"1 ");
      else 
	sprintf(tmp,"0 ");
      ADVANCE(tmp);
    }
  }

  /*PrintConditionLatticeToString(tmp,desc,state,condregs);*/

  sprintf(tmp,"-----");
  ADVANCE(tmp);
#endif
}

static void CfgEdgePrint(t_cfg_edge * edge, t_uint32 flags) 
{
  t_bbl * head;
  t_bbl * tail;

  t_function * head_fun;
  t_function * tail_fun;

  if (!edge)
  {
    printf("Edge is null\n");
    return;
  }

  head = CFG_EDGE_HEAD(edge);
  tail = CFG_EDGE_TAIL(edge);

  head_fun = head->fun;
  tail_fun = tail->fun;

  DiabloPrint(stdout,"edge @G -> @G of type ",BBL_OLD_ADDRESS(head),BBL_OLD_ADDRESS(tail));

  switch (EDGE_CAT(edge))
  {
    case  ET_UNKNOWN:
      printf("unknown");
      break;
    case  ET_FALLTHROUGH:
      printf("fallthrough");
      break;
    case  ET_CALL:
      printf("call");
      break;
    case  ET_RETURN:
      printf("return");
      break;
    case  ET_FUNLINK:
      printf("link");
      break;
    case  ET_JUMP:	
      printf("jump");
      break;
    case  ET_SWI:
      printf("swi");
      break;
    case  ET_SWILINK:
      printf("swilink");
      break;
    case  ET_IPUNKNOWN:
      printf("IPunknown");
      break;
    case  ET_IPFALLTHRU:
      printf("IPfallthrough");
      break;
    case  ET_IPFUNLINK:
      printf("IPfunlink");
      break;
    case  ET_IPJUMP:
      printf("IPjump");
      break;
    case  ET_IPSWILINK:
      printf("IPswilink");
      break;
    case  ET_COMPENSATING:
      printf(" compensating");
      break;
    case ET_SWITCH:
      printf(" switch");
      break;
  }

  if (head_fun)
    DiabloPrint(stdout," from fun @G ",BBL_OLD_ADDRESS(head_fun->bbl_first));
  else
    printf(" from no fun ");

  if (tail_fun)
    DiabloPrint(stdout," to fun @G\n",BBL_OLD_ADDRESS(tail_fun->bbl_first));
  else
    printf(" to no fun\n");

  if (flags & PRINT_PROC_STATE)
  {
    if (edge->procstate)
    {
      printf("has procstate associated:\n");
      ProcStatePrint(stdout,CFG_EDGE_OBJ(edge)->description,edge->procstate);
    }
    else
    {
      printf("has no procstate associated\n");
    }
  }

  if (!(flags & PRINT_CORRESPONDING))
    return;

  edge=edge->corr_edge;

  printf("Corresponding ");

  if (!edge)
  {
    printf("edge is null\n");
    return;
  }

  head = CFG_EDGE_HEAD(edge);
  tail = CFG_EDGE_TAIL(edge);

  head_fun = head->fun;
  tail_fun = tail->fun;

  DiabloPrint(stdout,"edge @G -> @G of type ",BBL_OLD_ADDRESS(head),BBL_OLD_ADDRESS(tail));

  switch (EDGE_CAT(edge))
  {
    case  ET_UNKNOWN:
      printf("unknown");
      break;
    case  ET_FALLTHROUGH:
      printf("fallthrough");
      break;
    case  ET_CALL:
      printf("call");
      break;
    case  ET_RETURN:
      printf("return");
      break;
    case  ET_FUNLINK:
      printf("link");
      break;
    case  ET_JUMP:	
      printf("jump");
      break;
    case  ET_SWI:
      printf("swi");
      break;
    case  ET_SWILINK:
      printf("swilink");
      break;
    case  ET_IPUNKNOWN:
      printf("IPunknown");
      break;
    case  ET_IPFALLTHRU:
      printf("IPfallthrough");
      break;
    case  ET_IPFUNLINK:
      printf("IPfunlink");
      break;
    case  ET_IPJUMP:
      printf("IPjump");
      break;
    case  ET_IPSWILINK:
      printf("IPswilink");
      break;
    case  ET_COMPENSATING:
      printf(" compensating");
      break;
  }

  if (head_fun)
    DiabloPrint(stdout," from fun @G ",BBL_OLD_ADDRESS(head_fun->bbl_first));
  else
    printf(" from no fun ");

  if (tail_fun)
    DiabloPrint(stdout," to fun @G\n",BBL_OLD_ADDRESS(tail_fun->bbl_first));
  else
    printf(" to no fun\n");

  if (flags & PRINT_PROC_STATE)
  {
    if (edge->procstate)
    {
      printf("has procstate associated:\n");
      ProcStatePrint(stdout,CFG_EDGE_OBJ(edge)->description,edge->procstate);
    }
    else
    {
      printf("has no procstate associated\n");
    }
  }
}

void BblPrint(t_bbl * bbl, t_uint32 flags) 
{
  char text[255];

  if (flags & PRINT_NEW_ADDRESS)
    DiabloPrint(stdout,"bbl at @G ",BBL_CADDRESS(bbl));
  else
    DiabloPrint(stdout,"bbl at @G ",BBL_OLD_ADDRESS(bbl));

  printf("size %x ",BBL_NINS(bbl)*4);

  if (bbl->fun)
    { 
      printf(" in fun at address ");  
      GENERIC_PRINT(BBL_OLD_ADDRESS(bbl->fun->bbl_first),10); 
      printf("\n");
    }
  else
    printf(" not associated with fun\n");
  
  if (flags & PRINT_LIVENESS_END)
    {
      printf("REGS USED      ");PrintRegset(stdout,BBL_OBJECT(bbl)->description,BBL_REGS_USED(bbl));
      printf("REGS DEFINED   ");PrintRegset(stdout,BBL_OBJECT(bbl)->description,BBL_REGS_DEFINED(bbl));
      printf("REGS OUT       ");PrintRegset(stdout,BBL_OBJECT(bbl)->description,BBL_REGS_LIVE_OUT(bbl));
    }
  if (flags & PRINT_INSTRUCTIONS)
    {
      t_ins * ins;
      t_address run=(flags & PRINT_NEW_ADDRESS)?BBL_CADDRESS(bbl):BBL_OLD_ADDRESS(bbl);
      BBL_FOREACH_INS(bbl,ins) 
	{
	  BBL_OBJECT(bbl)->actions->Print(ins,text);
	    GENERIC_PRINT(run,10);
	    printf(": %s\n",text);
	    if (flags & PRINT_LIVENESS)
	      {
		printf("live after: ");
		PrintRegset(stdout,BBL_OBJECT(bbl)->description,InsRegsLiveAfter(ins));
		printf("def: ");
		PrintRegset(stdout,BBL_OBJECT(bbl)->description,INS_REGS_DEF(ins));
		printf("use: ");
		PrintRegset(stdout,BBL_OBJECT(bbl)->description,INS_REGS_USE(ins));
	      }
	    run=GENERIC_ADD_UINT32(run,4);
	}
    }

  if (flags & PRINT_PROC_STATE)
    {
      if (BBL_PROCSTATE_IN(bbl))
	{
	  printf("has input procstate associated:\n");
	  ProcStatePrint(stdout,BBL_OBJECT(bbl)->description,BBL_PROCSTATE_IN(bbl));
	}
      else
	{
	  printf("has no input procstate associated\n");
	}

      /*      if (bbl->procstate_out)
	{
	  printf("has output procstate associated:\n");
	  ProcStatePrint(stdout,BBL_OBJECT(bbl)->description,bbl->procstate_out);
	}
      else
	{
	  printf("has no output procstate associated\n");
	}
      */
    }

  if (flags & PRINT_PREDECESSOR)
    {
      t_cfg_edge * edge;
      printf("ingoing edges\n");
      BBL_FOREACH_PRED_EDGE(bbl,edge)
	{
	  CfgEdgePrint(edge,0);
	}
      
    }

  if (flags & PRINT_SUCCESSOR)
    {
      t_cfg_edge * edge;
      printf("outgoing edges\n");
      BBL_FOREACH_SUCC_EDGE(bbl,edge)
	{
	  CfgEdgePrint(edge,PRINT_CORRESPONDING);
	}
    }

#if 0
  if (flags & PRINT_DOMINATES)
  {
    if(BBL_DOMINATES(bbl))
    {
      data_ll_elem * iter;
      printf("dominated by: ");
      if(BBL_DOMINATES(bbl)->data == ALL_BBLS) printf(" ALL\n");
      else
      {
	for(iter = BBL_DOMINATES(bbl); iter; iter = LLNext(iter))
	{
	  DiabloPrint(stdout, "@G ", BBL_OLD_ADDRESS(T_BBL(iter->data)));
	}
	printf("\n");
      }
    }
  }

  if (flags & PRINT_IDOMINATED_BY)
  {
    printf("immediately dominated by: ");
    if (!BBL_IDOMINATED_BY(bbl)) printf("NONE!\n");
    else DiabloPrint(stdout, "@G\n",BBL_OLD_ADDRESS(BBL_IDOMINATED_BY(bbl))); 
  }

  if (flags & PRINT_PDOMINATES)
  {
    if(BBL_PDOMINATES(bbl))
    {
      data_ll_elem * iter;
      printf("immediately post-dominates: ");
      if(BBL_PDOMINATES(bbl)->data == ALL_BBLS) printf(" ALL\n");
      else
      {
	for(iter = BBL_PDOMINATES(bbl); iter; iter = LLNext(iter))
	{
	  DiabloPrint(stdout, "@G ", BBL_OLD_ADDRESS(T_BBL(iter->data)));
	}
	printf("\n");
      }
    }
  }

  if (flags & PRINT_PDOMINATED_BY)
  {
    printf("immediately post-dominated by: ");
    if (!BBL_PDOMINATED_BY(bbl)) printf("NONE!\n");
    else DiabloPrint(stdout, "@G\n",BBL_OLD_ADDRESS(BBL_PDOMINATED_BY(bbl))); 
  }
#endif

}

void FunctionPrint(t_function * fun, t_uint32 flags) 
{
  t_bbl * bbl;

  printf("******************************************************************\n");
  printf("new fun at starting address ");

  if (fun->bbl_first)
    {
      DiabloPrint(stdout,"@G\n",BBL_OLD_ADDRESS(fun->bbl_first));
      FUN_FOREACH_BBL(fun,bbl)
	{
	  BblPrint(bbl,flags);
	}  
    }
  else
    printf("NONE\n"); 

}


void LoopPrint(t_loop * loop, t_uint32 flags)
{
  t_loopelem * elem;
  t_loopfunref * funref = LOOP_FUNCTIONS(loop);
  data_ll_elem * iter_back;
  t_uint32 i;
  printf("Loop defined by backedges: ");
  iter_back = LOOP_BACKEDGES(loop);
  while(iter_back)
  {
    DiabloPrint(stdout,"@E\n",iter_back->data);
    iter_back = LLNext(iter_back);
  }
  printf("  Has %d nodes\n",LOOP_COUNT(loop));
#if 1
  printf("  Basic blocks in the loop:\n");
  for(i=0;i<HASH_TABLE_SIZE;i++)
  {
    elem = LOOP_ELEMS(loop)[i];
    while(elem)
    {
      /*DiabloPrint(stdout,"  @B%s\n",elem->bbl,elem->exit?", EXIT":"");*/
/*    printf("%d  ",i);*/
      DiabloPrint(stdout," @B\n",elem->bbl);
      elem = LLNext(elem);
    }
/*    printf("\n");*/
  }
  while(funref)
  {
    t_bbl * ibbl;
    FUN_FOREACH_BBL(funref->fun,ibbl)
      DiabloPrint(stdout," @B\n",ibbl);
    funref = LLNext(funref);
  }
#endif
}

/*! 
 * \param prnt The object to print 
 * 
 * Prints out information about the objectfile */ 

void ObjectPrint(t_object * prnt)
{
  t_uint32 tel;
  if (prnt->name) printf("Object %s:\n",prnt->name);
  else printf("Object has no name:\n");
  if (prnt->ncodes)
  {
    printf("code sections:\n");
    for (tel=0; tel<prnt->ncodes; tel++)
    {
      printf("%d. %s\n",tel+1,prnt->code[tel]->name);
    }
  }
  if (prnt->nrodatas)
  {
    printf("rodata sections:\n");
    for (tel=0; tel<prnt->nrodatas; tel++)
    {
      printf("%d. %s\n",tel+1,prnt->rodata[tel]->name);
    }
  }
  if (prnt->ndatas)
  {
    printf("data sections:\n");
    for (tel=0; tel<prnt->ndatas; tel++)
    {
      printf("%d. %s\n",tel+1,prnt->data[tel]->name);
    }
  }
  if (prnt->nnotes)
  {
    printf("note sections:\n");
    for (tel=0; tel<prnt->nnotes; tel++)
    {
      printf("%d. %s\n",tel+1,prnt->note[tel]->name);
    }
  }
  if (prnt->nbsss)
  {
    printf("bss sections:\n");
    for (tel=0; tel<prnt->nbsss; tel++)
    {
      printf("%d. %s\n",tel+1,prnt->bss[tel]->name);
    }
  }
}


/* Utility functions for vsDiabloPrint {{{ */

#ifdef GENERICADDRSUPPORT
#define G_PRINT(address,string) if (GenericIs32(address)) sprintf((string),"0x%x",G_T_UINT32(address)); else if (GenericIs64(address)) sprintf((string),"0x%llx",G_T_UINT64(address)); else if (!(GenericIsValid(address))) sprintf((string),"<INVALID>"); else sprintf((string),"<UNKNOWN>");
#else 
#ifdef BIT32ADDRSUPPORT
#define G_PRINT(address,string) sprintf((string),"0x%x",address);
#else
#define G_PRINT(address,string) sprintf((string),"0x%llx",address);
#endif
#endif


void DiabloPrintBblToString(t_bbl * bbl, char * modifiers, char * tmp)
{
  t_bool flag = FALSE;
  sprintf(tmp,"bbl at ");
  ADVANCE(tmp);

  if (GENERIC_EQ(BBL_OLD_ADDRESS(bbl),BBL_CADDRESS(bbl)))
  {
    G_PRINT(BBL_CADDRESS(bbl),tmp);
    ADVANCE(tmp);
  }
  else
  {
    sprintf(tmp,"new: ");
    ADVANCE(tmp);
    G_PRINT(BBL_CADDRESS(bbl),tmp);
    ADVANCE(tmp);
    sprintf(tmp," old: ");
    ADVANCE(tmp);
    G_PRINT(BBL_OLD_ADDRESS(bbl),tmp);
    ADVANCE(tmp);
  }

  if (BBL_FUN(bbl))
    {
      sprintf(tmp,"(in fun %s starting at ",BBL_FUN(bbl)->name);
      ADVANCE(tmp);
      G_PRINT(BBL_CADDRESS(FUN_BBL_FIRST(BBL_FUN(bbl))),tmp);
      ADVANCE(tmp);
      sprintf(tmp,")");
      ADVANCE(tmp);
    }
  else
    {
      sprintf(tmp,"in no fun");
      ADVANCE(tmp);
    }

  sprintf(tmp," exec count O(%1.2f)",ORDER(BBL_EXEC_COUNT(bbl)));
  ADVANCE(tmp);


  if (*modifiers!=0)
    {
      sprintf(tmp,"\n");
      ADVANCE(tmp);
    }

  for (;*modifiers!=0; modifiers++)
  {
    flag = TRUE;
    if ((*modifiers)=='i')
    {
      t_ins * ins;
      sprintf(tmp,"Instructions:\n");
      ADVANCE(tmp);
      BBL_FOREACH_INS(bbl,ins) 
      {
	if (!GENERIC_EQ(INS_OLD_ADDRESS(ins),INS_CADDRESS(ins)))
	{
	  sprintf(tmp,"old: ");
	  ADVANCE(tmp);
	}
	G_PRINT(INS_OLD_ADDRESS(ins),tmp);
	ADVANCE(tmp);
	if (!GENERIC_EQ(INS_OLD_ADDRESS(ins),INS_CADDRESS(ins)))
	{
	  sprintf(tmp," new: ");
	  ADVANCE(tmp);
	  G_PRINT(INS_CADDRESS(ins),tmp);
	  ADVANCE(tmp);
	}
	sprintf(tmp,": ");
	ADVANCE(tmp);
	bbl->cfg->sec->obj->actions->Print(ins,tmp);
	ADVANCE(tmp);
	sprintf(tmp,"\n");
	ADVANCE(tmp);
      }
    }
    else if ((*modifiers)=='e')
    {
      t_cfg_edge * edge;
      int nr_edges = 0;
      sprintf(tmp,"Edges in:\n");
      ADVANCE(tmp);

      BBL_FOREACH_PRED_EDGE(bbl,edge) 
	nr_edges++;

      if (nr_edges<100)
	BBL_FOREACH_PRED_EDGE(bbl,edge) 
	{
	  t_address addr1=BBL_CADDRESS(CFG_EDGE_HEAD(edge));
	  t_address addr2=BBL_CADDRESS(CFG_EDGE_TAIL(edge));
	  G_PRINT(addr1,tmp);
#if 1
	  if (BBL_FUN(CFG_EDGE_HEAD(edge)))
	    {
	      sprintf(tmp+strlen(tmp)," (fun %s at ",BBL_FUN(CFG_EDGE_HEAD(edge))->name);
	      G_PRINT(BBL_OLD_ADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_HEAD(edge)))),tmp+strlen(tmp));
	      sprintf(tmp+strlen(tmp),")");
	    }
#endif
	  sprintf(tmp+strlen(tmp)," -> ");
	  G_PRINT(addr2,tmp+strlen(tmp));
#if 1
	  if (BBL_FUN(CFG_EDGE_TAIL(edge)))
	    {
	      sprintf(tmp+strlen(tmp)," (fun %s at ",BBL_FUN(CFG_EDGE_TAIL(edge))->name);
	      G_PRINT(BBL_OLD_ADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_TAIL(edge)))),tmp+strlen(tmp));
	      sprintf(tmp+strlen(tmp),")");
	    }
#endif
	  switch(EDGE_CAT(edge))
	    {
	    case ET_UNKNOWN:
	      sprintf(tmp+strlen(tmp)," type unknown");
	      break;
	    case ET_FALLTHROUGH:
	      sprintf(tmp+strlen(tmp)," type fallthrough");
	      break;
	    case ET_CALL:
	      sprintf(tmp+strlen(tmp)," type call");
	      break;
	    case ET_RETURN:
	      sprintf(tmp+strlen(tmp)," type return");
	      break;
	    case ET_FUNLINK:
	      sprintf(tmp+strlen(tmp)," type funlink");
	      break;
	    case ET_JUMP:
	      sprintf(tmp+strlen(tmp)," type jump");
	      break;
	    case ET_SWI:
	      sprintf(tmp+strlen(tmp)," type swi");
	      break;
	    case ET_SWILINK:
	      sprintf(tmp+strlen(tmp)," type swilink");
	      break;
	    case ET_IPUNKNOWN:
	      sprintf(tmp+strlen(tmp)," type ipunknown");
	      break;
	    case ET_IPFALLTHRU:
	      sprintf(tmp+strlen(tmp)," type ipfallthrough");
	      break;
	    case ET_IPFUNLINK:
	      sprintf(tmp+strlen(tmp)," type ipfunlink");
	      break;
	    case ET_IPJUMP:
	      sprintf(tmp+strlen(tmp)," type ipjump");
	      break;
	    case ET_IPSWILINK:
	      sprintf(tmp+strlen(tmp)," type ipswilink");
	      break;
	    case ET_COMPENSATING:
	      sprintf(tmp+strlen(tmp)," type compensating");
	      break;
	    case ET_SWITCH:
	      sprintf(tmp+strlen(tmp)," type switch");
	      break;
	    case ET_IPSWITCH:
	      sprintf(tmp+strlen(tmp)," type ipswitch");
	      break;
	    default:
	      sprintf(tmp+strlen(tmp)," type STRANGE %d!",EDGE_CAT(edge));
	    }
	  ADVANCE(tmp);
	  sprintf(tmp,"\n");
	  ADVANCE(tmp);
	}
      else
	{
	  sprintf(tmp+strlen(tmp)," too many pred edges ");
	  ADVANCE(tmp);
	}

      sprintf(tmp,"Edges out:\n");
      ADVANCE(tmp);

      nr_edges=0;
      BBL_FOREACH_SUCC_EDGE(bbl,edge) 
	nr_edges++;
      
      if (nr_edges<100)
	BBL_FOREACH_SUCC_EDGE(bbl,edge) 
	{
	  t_address addr1=BBL_CADDRESS(CFG_EDGE_HEAD(edge));
	  t_address addr2=BBL_CADDRESS(CFG_EDGE_TAIL(edge));
	  G_PRINT(addr1,tmp);
#if 1
	  if (BBL_FUN(CFG_EDGE_HEAD(edge)))
	    {
	      sprintf(tmp+strlen(tmp)," (fun %s at ",BBL_FUN(CFG_EDGE_HEAD(edge))->name);
	      G_PRINT(BBL_OLD_ADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_HEAD(edge)))),tmp+strlen(tmp));
	      sprintf(tmp+strlen(tmp),")");
	    }
#endif
	  sprintf(tmp+strlen(tmp)," -> ");
	  G_PRINT(addr2,tmp+strlen(tmp));
#if 1
	  if (BBL_FUN(CFG_EDGE_TAIL(edge)))
	    {
	      sprintf(tmp+strlen(tmp)," (fun %s at ",BBL_FUN(CFG_EDGE_TAIL(edge))->name);
	      G_PRINT(BBL_OLD_ADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_TAIL(edge)))),tmp+strlen(tmp));
	      sprintf(tmp+strlen(tmp),")");
	    }
#endif
	  G_PRINT(addr2,tmp+strlen(tmp));
	  switch(EDGE_CAT(edge))
	    {
	    case ET_UNKNOWN:
	      sprintf(tmp+strlen(tmp)," type unknown");
	      break;
	    case ET_FALLTHROUGH:
	      sprintf(tmp+strlen(tmp)," type fallthrough");
	      break;
	    case ET_CALL:
	      sprintf(tmp+strlen(tmp)," type call");
	      break;
	    case ET_RETURN:
	      sprintf(tmp+strlen(tmp)," type return");
	      break;
	    case ET_FUNLINK:
	      sprintf(tmp+strlen(tmp)," type funlink");
	      break;
	    case ET_JUMP:
	      sprintf(tmp+strlen(tmp)," type jump");
	      break;
	    case ET_SWI:
	      sprintf(tmp+strlen(tmp)," type swi");
	      break;
	    case ET_SWILINK:
	      sprintf(tmp+strlen(tmp)," type swilink");
	      break;
	    case ET_IPUNKNOWN:
	      sprintf(tmp+strlen(tmp)," type ipunknown");
	      break;
	    case ET_IPFALLTHRU:
	      sprintf(tmp+strlen(tmp)," type ipfallthrough");
	      break;
	    case ET_IPFUNLINK:
	      sprintf(tmp+strlen(tmp)," type ipfunlink");
	      break;
	    case ET_IPJUMP:
	      sprintf(tmp+strlen(tmp)," type ipjump");
	      break;
	    case ET_IPSWILINK:
	      sprintf(tmp+strlen(tmp)," type ipswilink");
	      break;
	    case ET_COMPENSATING:
	      sprintf(tmp+strlen(tmp)," type compensating");
	      break;
	    case ET_SWITCH:
	      sprintf(tmp+strlen(tmp)," type switch");
	      break;
	    case ET_IPSWITCH:
	      sprintf(tmp+strlen(tmp)," type ipswitch");
	      break;
	    default:
	      sprintf(tmp+strlen(tmp)," type STRANGE %d!",EDGE_CAT(edge));
	    }
	  ADVANCE(tmp);
	  sprintf(tmp,"\n");
	  ADVANCE(tmp);
	}
      else
	{
	  sprintf(tmp+strlen(tmp)," too many succ edges ");
	  ADVANCE(tmp);
	}
    }
  }
  if (flag)
    {
      sprintf(tmp,"-----------------------------------\n");
      ADVANCE(tmp);
    }
}
/*}}}*/

/* vs Diablo Print {{{ */ 
t_string vsDiabloPrint(char * desc_string,  va_list ap)
{
  t_string_array * dsa=StringDivider(desc_string, "%@", TRUE,TRUE);
  t_string ret;
  t_uint16 tel;

  char tmp[500000];

  for (tel=0; tel<dsa->nstrings; tel++)
  {
    if (dsa->strings[tel][0]=='@')
    {
      t_uint32 modifier=1;
      t_string orig=dsa->strings[tel];
      while((dsa->strings[tel][modifier]<='z')&&(dsa->strings[tel][modifier]>='a')) modifier++;
      switch (dsa->strings[tel][modifier])
      {
	case 'B': /* Print BBLS */
	  {
	    t_bbl * bbl=va_arg(ap,t_bbl *);

	    dsa->strings[tel][modifier]=0;
	    DiabloPrintBblToString(bbl,dsa->strings[tel]+1,tmp);
	    dsa->strings[tel]=Collate(tmp,orig+modifier+1);
	    Free(orig);
	  }
	  break;
	case 'E': /* Print Edges */
	  {
	    t_cfg_edge * edge=va_arg(ap,t_cfg_edge *);
	    t_address addr1=BBL_OLD_ADDRESS(CFG_EDGE_HEAD(edge));
	    t_address addr2=BBL_OLD_ADDRESS(CFG_EDGE_TAIL(edge));
	    
	    VERIFY_FASSERT(modifier==1,("Modifiers not implemented for %c",dsa->strings[tel][modifier]));
	    G_PRINT(addr1,tmp);
	    if (BBL_FUN(CFG_EDGE_HEAD(edge)))
	      {
		sprintf(tmp+strlen(tmp)," (fun at ");
		G_PRINT(BBL_OLD_ADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_HEAD(edge)))),tmp+strlen(tmp));
		sprintf(tmp+strlen(tmp),")");
	      }
	    sprintf(tmp+strlen(tmp)," -> ");
	    G_PRINT(addr2,tmp+strlen(tmp));
	    if (BBL_FUN(CFG_EDGE_TAIL(edge)))
	      {
		sprintf(tmp+strlen(tmp)," (fun at ");
		G_PRINT(BBL_OLD_ADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_TAIL(edge)))),tmp+strlen(tmp));
		sprintf(tmp+strlen(tmp),")");
	      }
	    switch(EDGE_CAT(edge))
	    {
	      case ET_UNKNOWN:
		sprintf(tmp+strlen(tmp)," type unknown");
		break;
	      case ET_FALLTHROUGH:
		sprintf(tmp+strlen(tmp)," type fallthrough");
		break;
	      case ET_CALL:
		sprintf(tmp+strlen(tmp)," type call");
		break;
	      case ET_RETURN:
		sprintf(tmp+strlen(tmp)," type return");
		break;
	      case ET_FUNLINK:
		sprintf(tmp+strlen(tmp)," type funlink");
		break;
	      case ET_JUMP:
		sprintf(tmp+strlen(tmp)," type jump");
		break;
	      case ET_SWI:
		sprintf(tmp+strlen(tmp)," type swi");
		break;
	      case ET_SWILINK:
		sprintf(tmp+strlen(tmp)," type swilink");
		break;
	      case ET_IPUNKNOWN:
		sprintf(tmp+strlen(tmp)," type ipunknown");
		break;
	      case ET_IPFALLTHRU:
		sprintf(tmp+strlen(tmp)," type ipfallthrough");
		break;
	      case ET_IPFUNLINK:
		sprintf(tmp+strlen(tmp)," type ipfunlink");
		break;
	      case ET_IPJUMP:
		sprintf(tmp+strlen(tmp)," type ipjump");
		break;
	      case ET_IPSWILINK:
		sprintf(tmp+strlen(tmp)," type ipswilink");
		break;
	      case ET_COMPENSATING:
		sprintf(tmp+strlen(tmp)," type compensating");
		break;
	      case ET_SWITCH:
		sprintf(tmp+strlen(tmp)," type switch");
		break;
	      case ET_IPSWITCH:
		sprintf(tmp+strlen(tmp)," type ipswitch");
		break;
	      default:
		sprintf(tmp+strlen(tmp)," type STRANGE %d!",EDGE_CAT(edge));
	  }
	    
	    sprintf(tmp+strlen(tmp), " and exec count O(%f)",ORDER(EDGE_EXEC_COUNT(edge)));
	    
	    dsa->strings[tel]=Collate(tmp,orig+2);
	    
	    Free(orig);
      }
      break;
      case 'S': /* Print Symbols */
      {
	t_symbol * sym=va_arg(ap,t_symbol *);

	VERIFY_FASSERT(modifier==1,("Modifiers not implemented for %c",dsa->strings[tel][modifier]));

	sprintf(tmp,"%s symbol \"%s\" at ",sym->local?"Local":"Global",sym->name?sym->name:"<nameless symbol>");
	G_PRINT(sym->current_address,(tmp+strlen(tmp)));
	
	switch(sym->type)
	{
	  case FUNCTION_SYMBOL: sprintf(tmp+strlen(tmp)," FUNCTION_SYMBOL"); break;
	  case GLOBAL_SYMBOL: sprintf(tmp+strlen(tmp)," GLOBAL_SYMBOL"); break;
	  case UNDEFINED_SYMBOL: sprintf(tmp+strlen(tmp)," UNDEFINED_SYMBOL"); break;
	  case LOCAL_SYMBOL: sprintf(tmp+strlen(tmp)," LOCAL_SYMBOL"); break;
	  case CODE_IDENT_SYMBOL: sprintf(tmp+strlen(tmp)," CODE_IDENT_SYMBOL"); break;
	  case DATA_IDENT_SYMBOL: sprintf(tmp+strlen(tmp)," DATA_IDENT_SYMBOL"); break;
	  case ORIGINAL_SECTION_SYMBOL: sprintf(tmp+strlen(tmp)," ORIGINAL_SECTION_SYMBOL"); break;
	  case ABS_SYMBOL: sprintf(tmp+strlen(tmp)," ABS_SYMBOL"); break;
	  case WEAK_UNDEFINED_SYMBOL: sprintf(tmp+strlen(tmp)," WEAK_UNDEFINED_SYMBOL"); break;
	  case DIRECT_BRANCH_SYMBOL: sprintf(tmp+strlen(tmp)," DIRECT_BRANCH_SYMBOL"); break;
	  case WEAK_SYMBOL: sprintf(tmp+strlen(tmp)," WEAK_SYMBOL"); break;
	  case REMOVED_SYMBOL: sprintf(tmp+strlen(tmp)," REMOVED_SYMBOL"); break;
	  case PLT_SYMBOL: sprintf(tmp+strlen(tmp)," REMOVED_SYMBOL"); break;
	  default: 
			      sprintf(tmp+strlen(tmp)," Unknown to type! %d",sym->type);
	}

	switch(sym->to_type)
	{
	  case TO_SECTION: sprintf(tmp+strlen(tmp)," TO_SECTION %s",T_SECTION(sym->base)->name); break;
	  case TO_SUB_SECTION: sprintf(tmp+strlen(tmp)," TO_SUB_SECTION %s",T_SECTION(sym->base)->name); break;
	  case TO_COMMON: sprintf(tmp+strlen(tmp)," TO_COMMON"); break;
	  case TO_UNDEFINED: sprintf(tmp+strlen(tmp)," TO_UNDEFINED"); break;
	  case TO_NONE: sprintf(tmp+strlen(tmp)," TO_NONE"); break;
	  case TO_REMOVED: sprintf(tmp+strlen(tmp)," TO_REMOVED"); break;
	  case TO_INSTRUCTION: 
			{
			       t_string instr;
			       sprintf(tmp+strlen(tmp),instr=sDiabloPrint("TO_INSTRUCTION: @I",sym->base));
			       Free(instr);
			       break;
			}
	  case TO_BBL: sprintf(tmp+strlen(tmp)," TO_BBL"); break;
	  default: 
		       sprintf(tmp+strlen(tmp)," Unknown reffed type %d\n",sym->to_type);
	}


	dsa->strings[tel]=Collate(tmp,orig+2);
	Free(orig);

      }
      break;
      case 'T': /* Relocatable */
      {
	    t_relocatable * reloc=va_arg(ap,t_relocatable *);
	    VERIFY_FASSERT(modifier==1,("Modifiers not implemented for %c",dsa->strings[tel][modifier]));
	    sprintf(tmp,"Relocatable ");
	    sprintf(tmp+strlen(tmp),"  type = ");
	    switch(RELOCATABLE_TYPE(reloc))
	    {
	      case RT_SECTION: sprintf(tmp+strlen(tmp),"RT_SECTION %s",T_SECTION(reloc)->name); 
				       sprintf(tmp+strlen(tmp),")");
				       break;
	      case RT_SUBSECTION: sprintf(tmp+strlen(tmp),"RT_SUBSECTION %s in %s ",T_SECTION(reloc)->name,T_SECTION(reloc)->obj->name);
					  sprintf(tmp+strlen(tmp)," (%s)","TODO"); break;
	      case RT_INS: 
	      { 
		t_string tmp2=sDiabloPrint("@I",reloc);

		sprintf(tmp+strlen(tmp),"RT_INS: %s",tmp2);
		Free(tmp2);
		break;
	      }
	      case RT_BBL: sprintf(tmp+strlen(tmp),"RT_BBL"); break;
	      default: 
				       FATAL(("Unknown from type %d!",RELOCATABLE_TYPE(reloc))); 
	    }


	  dsa->strings[tel]=Collate(tmp,orig+2);
	  Free(orig);
	  break;

      }
	case 'R': /* Print Relocations */
	  {
	    t_reloc * reloc=va_arg(ap,t_reloc *);
	    VERIFY_FASSERT(modifier==1,("Modifiers not implemented for %c",dsa->strings[tel][modifier]));
	    sprintf(tmp,"Reloc ");
	    
	    sprintf(tmp+strlen(tmp)," From type = ");
	    if (!REL_FROM(reloc)) FATAL(("RELOC HAS NO FROM\n"));
	    switch(RELOCATABLE_TYPE(REL_FROM(reloc)))
	    {
	      case RT_SECTION: sprintf(tmp+strlen(tmp),"RT_SECTION %s (+",T_SECTION(REL_FROM(reloc))->name); 
				       G_PRINT(reloc->from_offset,(tmp+strlen(tmp))); 
				       sprintf(tmp+strlen(tmp),")");
				       break;
	      case RT_SUBSECTION: sprintf(tmp+strlen(tmp),"RT_SUBSECTION %s in %s (",T_SECTION(REL_FROM(reloc))->name,T_SECTION(REL_FROM(reloc))->obj->name);
				          G_PRINT(GENERIC_ADD(reloc->from_offset,SECTION_GET_OLD_ADDRESS(T_SECTION(REL_FROM(reloc)))),(tmp+strlen(tmp))); 
					  sprintf(tmp+strlen(tmp),") (%s)","TODO"); break;
	      case RT_INS: 
	      { 
		t_string tmp2=sDiabloPrint("@I",REL_FROM(reloc));

		sprintf(tmp+strlen(tmp),"RT_INS: %s",tmp2);
		Free(tmp2);
		break;
	      }
	      case RT_BBL: sprintf(tmp+strlen(tmp),"RT_BBL"); break;
	      default: 
				       FATAL(("Unknown from type %d!",RELOCATABLE_TYPE(REL_FROM(reloc)))); 
	    }





	    if  (REL_IS_TO_SYMBOL(reloc) && REL_TO_SYMBOL(reloc))
	    {
	      sprintf(tmp+strlen(tmp)," To %p \"%s\" (%s)",REL_TO_SYMBOL(reloc),REL_TO_SYMBOL(reloc)->name?REL_TO_SYMBOL(reloc)->name:"<nameless symbol>",REL_TO_SYMBOL(reloc)->local?"Local":"Global");
	      sprintf(tmp+strlen(tmp)," Type = ");

	      switch(REL_TO_SYMBOL(reloc)->type)
	      {
		case FUNCTION_SYMBOL: sprintf(tmp+strlen(tmp),"FUNCTION_SYMBOL"); break;
		case GLOBAL_SYMBOL: sprintf(tmp+strlen(tmp),"GLOBAL_SYMBOL"); break;
		case UNDEFINED_SYMBOL: sprintf(tmp+strlen(tmp),"UNDEFINED_SYMBOL"); break;
		case LOCAL_SYMBOL: sprintf(tmp+strlen(tmp),"LOCAL_SYMBOL"); break;
		case CODE_IDENT_SYMBOL: sprintf(tmp+strlen(tmp),"CODE_IDENT_SYMBOL"); break;
		case DATA_IDENT_SYMBOL: sprintf(tmp+strlen(tmp),"DATA_IDENT_SYMBOL"); break;
		case ORIGINAL_SECTION_SYMBOL: sprintf(tmp+strlen(tmp),"ORIGINAL_SECTION_SYMBOL"); break;
		case ABS_SYMBOL: sprintf(tmp+strlen(tmp),"ABS_SYMBOL"); break;
		case WEAK_UNDEFINED_SYMBOL: sprintf(tmp+strlen(tmp),"WEAK_UNDEFINED_SYMBOL"); break;
		case DIRECT_BRANCH_SYMBOL: sprintf(tmp+strlen(tmp),"DIRECT_BRANCH_SYMBOL"); break;
		case WEAK_SYMBOL: sprintf(tmp+strlen(tmp),"WEAK_SYMBOL"); break;
		case END_SYMBOL: sprintf(tmp+strlen(tmp),"END_SYMBOL"); break;
		case REMOVED_SYMBOL: sprintf(tmp+strlen(tmp),"REMOVED_SYMBOL"); break;
		case SECTION_END_SYMBOL: sprintf(tmp+strlen(tmp),"SECTION_END_SYMBOL"); break;
		case 0: sprintf(tmp+strlen(tmp),"0"); break;
		default: 
				    FATAL(("Unknown to type! %d",REL_TO_SYMBOL(reloc)->type));
	      }

	      sprintf(tmp+strlen(tmp)," Refs = ");

	      switch(REL_TO_SYMBOL(reloc)->to_type)
	      {
		case TO_SECTION: sprintf(tmp+strlen(tmp),"TO_SECTION %s",T_SECTION(REL_TO_SYMBOL(reloc)->base)->name); break;
		case TO_SUB_SECTION: sprintf(tmp+strlen(tmp),"TO_SUB_SECTION %s (%s)",T_SECTION(REL_TO_SYMBOL(reloc)->base)->name,"TODO"); break;
		case TO_COMMON: sprintf(tmp+strlen(tmp),"TO_COMMON"); break;
		case TO_UNDEFINED: sprintf(tmp+strlen(tmp),"TO_UNDEFINED"); break;
		case TO_NONE: sprintf(tmp+strlen(tmp),"TO_NONE"); break;
		case TO_INSTRUCTION: sprintf(tmp+strlen(tmp),"TO_INSTRUCTION"); break;
		case TO_BBL: sprintf(tmp+strlen(tmp),"TO_BBL"); break;
		case TO_REMOVED: sprintf(tmp+strlen(tmp),"TO_REMOVED"); break;
		default: 
			     FATAL(("Unknown reffed type %d\n",REL_TO_SYMBOL(reloc)->to_type));
	      }
	      sprintf(tmp+strlen(tmp)," Offset= ");
	      G_PRINT(REL_TO_SYMBOL(reloc)->offset_from_start,(tmp+strlen(tmp))); 
	    }
	    else if (REL_IS_TO_RELOCATABLE(reloc) && REL_TO_RELOCATABLE(reloc))
	    {
	      sprintf(tmp+strlen(tmp),"To relocatable ");
	      switch(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(reloc)))
	      {
		case RT_INS:
		  sprintf(tmp+strlen(tmp),"INS ");
		  break;
		case RT_BBL:
		  {
		  t_string bbl_st=sDiabloPrint("@B",REL_TO_RELOCATABLE(reloc));
		  sprintf(tmp+strlen(tmp),"BBL %s",bbl_st);
		  Free(bbl_st);
		  if (!GENERIC_EQ(BBL_CADDRESS(REL_TO_RELOCATABLE(reloc)),BBL_OLD_ADDRESS(REL_TO_RELOCATABLE(reloc))))
		  {
		    bbl_st=sDiabloPrint("@G",BBL_CADDRESS(REL_TO_RELOCATABLE(reloc)));
		    sprintf(tmp+strlen(tmp),"New address: %s",bbl_st);
		    Free(bbl_st);
		  }
		  }
		  break;
		case RT_SECTION:
		  sprintf(tmp+strlen(tmp),"SECTION %s",T_SECTION(REL_TO_RELOCATABLE(reloc))->name);
		  break;
		case RT_SUBSECTION:
		  sprintf(tmp+strlen(tmp),"SUBSECTION %s in %s",T_SECTION(REL_TO_RELOCATABLE(reloc))->name,SECTION_OBJECT(T_SECTION(REL_TO_RELOCATABLE(reloc)))->name);
		  break;
		default:
		  sprintf(tmp+strlen(tmp),"unknown (%d)",RELOCATABLE_TYPE(REL_TO_RELOCATABLE(reloc)));
		  break;
	      }
	    }

	    if (reloc->edge) sprintf(tmp+strlen(tmp)," Has edge");
	    sprintf(tmp+strlen(tmp)," First=");
	    if (reloc->first == TRUE) sprintf(tmp+strlen(tmp),"TRUE");
	    else if (reloc->first == FALSE) sprintf(tmp+strlen(tmp),"FALSE");
	    else sprintf(tmp+strlen(tmp), "UNINITIALIZED!");
	  }
	  dsa->strings[tel]=Collate(tmp,orig+2);
	  Free(orig);
	  break;
	case 'I': /* Print Instructions */
	  {
	    t_ins * ins=va_arg(ap,t_ins *);
	    t_string  ad,  ad2;
	    VERIFY_FASSERT(modifier==1,("Modifiers not implemented for %c",dsa->strings[tel][modifier]));
	    G_PRINT(INS_OLD_ADDRESS(ins),tmp);
	    if (!GENERIC_EQ(INS_OLD_ADDRESS(ins),INS_CADDRESS(ins)))
	      ad=Collate(" Old ",tmp);
	    else
	    {
	      ad=StringDup(tmp);
	    }

	    ad2=Collate(ad,": ");
	    if (!GENERIC_EQ(INS_OLD_ADDRESS(ins),INS_CADDRESS(ins)))
	    {
	      Free(ad);
	      ad=ad2;
	      G_PRINT(INS_CADDRESS(ins),tmp);
	      ad2=Collate3("New ",tmp,ad2);
	    }


	    if (INS_BBL(ins)) INS_BBL(ins)->cfg->sec->obj->actions->Print(ins,tmp);
	    else INS_SECTION(ins)->obj->actions->Print(ins,tmp);
	    dsa->strings[tel]=Collate3(ad2,tmp,orig+2);
	    Free(ad);
	    Free(ad2);
	    Free(orig);
	  }
	  break;
	case 'G': /* Print Generic Addresses */
	  {
	    t_address addr=va_arg(ap,t_address);
	    VERIFY_FASSERT(modifier==1,("Modifiers not implemented for %c",dsa->strings[tel][modifier]));
	    G_PRINT(addr,tmp);
	    dsa->strings[tel]=Collate(tmp,orig+2);
	    Free(orig);
	  }
	  break;
        case 'P': /* Print Procstates */
	  {
	    t_procstate * procstate = va_arg(ap,t_procstate *);
	    if (!arch_desc)
	      FATAL(("Trying to print procstates without architecture description!!!"));
	    VERIFY_FASSERT(modifier==1,("Modifiers not implemented for %c",dsa->strings[tel][modifier]));
	    if (procstate!=NULL)
	      ProcStatePrintOptToString(arch_desc,procstate,arch_desc->cond_registers,0,tmp);
	    dsa->strings[tel]=Collate(tmp,orig+2);
	    Free(orig);
	  }
	  break;
        case 'A':
	  {
	    t_regset regs = va_arg(ap,t_regset);
	    if (!arch_desc)
	      FATAL(("Trying to print procstates without architecture description!!!"));
	    VERIFY_FASSERT(modifier==1,("Modifiers not implemented for %c",dsa->strings[tel][modifier]));
	    PrintRegsetToString(tmp,arch_desc,regs);
	    dsa->strings[tel]=Collate(tmp,orig+2);
	    Free(orig);
	  }
	  break;
        case 'F':
	  {
	    t_function *fun = va_arg(ap,t_function*);
	    VERIFY_FASSERT(modifier==1,("Modifiers not implemented for %c",dsa->strings[tel][modifier]));
	    sprintf(tmp,sDiabloPrint("procedure %s @B",fun->name,FUN_BBL_FIRST(fun)));
	    dsa->strings[tel]=Collate(tmp,orig+2);
	    Free(orig);
	  }
	  break;
        case 'C':
	  {
	    t_cfg * cfg = va_arg(ap,t_cfg*);
	    t_function * fun;
	    t_bbl * bbl;
	    strcpy(tmp,"flowgraph ");
	    CFG_FOREACH_FUN(cfg,fun)
	      {
		sprintf(tmp+strlen(tmp),sDiabloPrint("@F\n",fun));
		FUN_FOREACH_BBL(fun,bbl)
		  sprintf(tmp+strlen(tmp),sDiabloPrint("@ieB\n",bbl));
	      }
	    dsa->strings[tel]=Collate(tmp,orig+2);
	    Free(orig);
	  }
	  break;
	default:
	  printf("Unknown modifier %c, string is %s, entire string is %s\n",dsa->strings[tel][1],dsa->strings[tel],desc_string);
	  exit(0);

      }
    }
    else if (dsa->strings[tel][0]=='%')
    {
      int u=1;
      while(((dsa->strings[tel][u]>='0')&&(dsa->strings[tel][u]<='9'))||(dsa->strings[tel][u]=='.')) u++;
      switch (dsa->strings[tel][u])
      {
	case 'l':
	  u++;
	  switch (dsa->strings[tel][u])
	  {
	    case 'x':
	      {
		sprintf(tmp,dsa->strings[tel],va_arg(ap,long));
		Free(dsa->strings[tel]);
		dsa->strings[tel]=StringDup(tmp);
		break;
	      }
	    case 'd':
	      {
		sprintf(tmp,dsa->strings[tel],va_arg(ap,long));
		Free(dsa->strings[tel]);
		dsa->strings[tel]=StringDup(tmp);
		break;
	      }
	      break;
	    case 'l':
	      {
		u++;
		switch (dsa->strings[tel][u])
		{
		  case 'x':
		    {
		      sprintf(tmp,dsa->strings[tel],va_arg(ap,long long));
		      Free(dsa->strings[tel]);
		      dsa->strings[tel]=StringDup(tmp);
		      break;
												                  }
		  case 'd':
		    {
		      sprintf(tmp,dsa->strings[tel],va_arg(ap,long long));
		      Free(dsa->strings[tel]);
		      dsa->strings[tel]=StringDup(tmp);
		      break;
		    }
		  default:
		    FATAL(("Unknown long long print modifier %c",dsa->strings[tel][u]));
		    break;
		}
		break;
	      }
	    default:
	      FATAL(("Unknown normal long print modifier %c",dsa->strings[tel][u]));
	  }
	  break;
	case 's':
	  {
	    sprintf(tmp,dsa->strings[tel],va_arg(ap,char *));
	    Free(dsa->strings[tel]);
	    dsa->strings[tel]=StringDup(tmp);
	  }
	  break;
	case 'd':
	  {
	    int tp=va_arg(ap,int);
	    sprintf(tmp,dsa->strings[tel],tp);
	    Free(dsa->strings[tel]);
	    dsa->strings[tel]=StringDup(tmp);
	  }
	  break;
	case 'f':
	  {
	    double tp=va_arg(ap,double);
	    sprintf(tmp,dsa->strings[tel],tp);
	    Free(dsa->strings[tel]);
	    dsa->strings[tel]=StringDup(tmp);
	  }
	  break;
	case 'x':
	  {
	    sprintf(tmp,dsa->strings[tel],va_arg(ap,int));
	    Free(dsa->strings[tel]);
	    dsa->strings[tel]=StringDup(tmp);

	  }
	  break;
	case 'c':
	  {
	    sprintf(tmp,dsa->strings[tel],va_arg(ap,int));
	    Free(dsa->strings[tel]);
	    dsa->strings[tel]=StringDup(tmp);
	  }
	  break;
	case 'p':
	  {
	    sprintf(tmp,dsa->strings[tel],va_arg(ap,void *));
	    Free(dsa->strings[tel]);
	    dsa->strings[tel]=StringDup(tmp);
	  }
	  break;
	default:
	  FATAL(("Unknown normal print modifier %c",dsa->strings[tel][u]));
      }
    }
  }


  ret=StringArrayJoiner(dsa,"");

  StringArrayFree(dsa);
  return ret;
}
/*}}}*/

void vDiabloPrint(FILE * f, char * desc_string,  va_list ap)
{
  t_string p=vsDiabloPrint(desc_string, ap);
  fprintf(f,"%s",p);
  Free(p);
}

void DiabloPrint(FILE * f, char * desc_string, ...)
{
  va_list ap;
  arch_desc = NULL;
  va_start(ap,desc_string);
  vDiabloPrint(f,desc_string,ap);
}

void DiabloPrintArch(FILE * f, const t_architecture_description * desc, char * desc_string, ...)
{
  va_list ap;
  arch_desc = desc;
  va_start(ap,desc_string);
  vDiabloPrint(f,desc_string,ap);
}

void DiabloPrintStdout(char * desc_string, ...)
{
  va_list ap;
  va_start(ap,desc_string);
  vDiabloPrint(stdout,desc_string,ap);
}

t_string sDiabloPrint(char * desc_string, ...)
{
  va_list ap;
  va_start(ap,desc_string);
  return vsDiabloPrint(desc_string, ap);
}

void TraceBblAtAddress(t_cfg * cfg, t_uint32 address, t_string place)
{
  t_bbl * bbl;

  if (address==0) return;

#ifdef BIT32ADDRSUPPORT
  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (GENERIC_EXTRACT_UINT32(BBL_OLD_ADDRESS(bbl))==address)
	{
	  DiabloPrint(stdout,"AT %s: @eiB\n",place,bbl);
	  return;
	}
    }
#else
  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (GENERIC_EXTRACT_UINT64(BBL_OLD_ADDRESS(bbl))==address)
	{
	  DiabloPrint(stdout,"AT %s: @eiB\n",place,bbl);
	  return;
	}
    }
#endif
}


/* vim: set shiftwidth=2 foldmethod=marker: */
