/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC:  Operations on processor states
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@rug.ac.be) }}}
 *****************************************************************************/

/*{{{ Includes*/
#include "diablo_procstate.h"
#include "diablo_reloc.h"
#include "diablo_print.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include <string.h> /* For memcopy */ 
/*}}}*/

/*! Used to see whether or not two propagated relocs point to the same
  block (basic block or data block) */
t_bool TwoRelocsInSameBlock(t_reloc * reloc1, t_reloc* reloc2)
{
  if (reloc1==reloc2)
    return TRUE;
  if (REL_TO_RELOCATABLE(reloc1)==REL_TO_RELOCATABLE(reloc2))
    return TRUE;

  /* the following test was added for accesses to the lita (on the
     alpha): all blocks in there are in the same parent section */
  /* this seems dangerous for other architectures however, especially if data can still be moved */
  /* maybe we should tag a section if this is ok */

  if (T_SECTION(REL_TO_RELOCATABLE(reloc1))->parent_section == T_SECTION(REL_TO_RELOCATABLE(reloc2))->parent_section)
    return TRUE;


  return FALSE;
}

/* list of free procstates */
static void ** pslist = NULL;

/*! Allocates an object of type t_procstate */
t_procstate* ProcStateNew()
{
  t_procstate * state;
  register_content * rc;
  t_reloc ** tags;

  int set;

  if (pslist)
  {
    state = (t_procstate *) pslist;
    pslist = *pslist;
    
    /* note: the following bzero's are necessary if floating point support is reintroduced in register_content*/

    /* bzero(state->register_subset[0],MAX_NR_REG_SUBSETS*REG_SUBSET_SIZE*sizeof(register_content));*/
    /* bzero(state->register_tag_subset[0],MAX_NR_REG_SUBSETS*REG_SUBSET_SIZE*sizeof(t_reloc *));*/
  }
  else
  {
    state = (t_procstate *) Malloc(sizeof(t_procstate));
    rc = Calloc(MAX_NR_REG_SUBSETS*REG_SUBSET_SIZE,sizeof(register_content));
    tags = Malloc(MAX_NR_REG_SUBSETS*REG_SUBSET_SIZE*sizeof(t_reloc *));
    for (set=0;set<MAX_NR_REG_SUBSETS;set++)
    {
      state->register_subset[set]=&(rc[set * REG_SUBSET_SIZE]);
      state->register_tag_subset[set]=&(tags[set * REG_SUBSET_SIZE]);
    }
  }

  RegsetSetAll(state->top);
  RegsetSetAll(state->tag_top);
  RegsetSetEmpty(state->bot);
  RegsetSetEmpty(state->tag_bot);
  RegsetSetEmpty(state->cond_true);
  RegsetSetEmpty(state->cond_false);
#ifdef EXTENDED_CONSTPROP
  RegsetSetEmpty(state->value_minus);
  RegsetSetEmpty(state->value_zero);
  RegsetSetEmpty(state->value_plus);
#endif

  return state;
}

/*! Frees an object of type t_procstate */
void ProcStateFree(t_procstate* state)
{
  *((void **)state) = pslist;
  pslist = (void *)state;
}

void ProcStateRealFree(void)
{
  t_procstate * ps;

  while (pslist)
  {
    ps = (t_procstate *)pslist;
    pslist = *pslist;
    Free(ps->register_subset[0]);
    Free(ps->register_tag_subset[0]);
    Free(ps);
  }
}

/** Set all propagated items in a state to Top (value, reloc tag, cond bits */
void ProcStateSetAllTop(t_procstate* state, t_regset regs)
{
  RegsetSetUnion(state->top,regs);
  RegsetSetUnion(state->tag_top,regs);
  RegsetSetDiff(state->bot,regs);
  RegsetSetDiff(state->tag_bot,regs);
  RegsetSetDiff(state->cond_true,regs);
  RegsetSetDiff(state->cond_false,regs);
#ifdef EXTENDED_CONSTPROP
  RegsetSetDiff(state->value_minus,regs);
  RegsetSetDiff(state->value_zero,regs);
  RegsetSetDiff(state->value_plus,regs);
#endif
}

/** Set all propagated items in a state to Bot (value, reloc tag, cond bits */
void ProcStateSetAllBot(t_procstate* state, t_regset regs)
{
  RegsetSetUnion(state->bot,regs);
  RegsetSetUnion(state->tag_bot,regs);
  RegsetSetDiff(state->top,regs);
  RegsetSetDiff(state->tag_top,regs);
#ifdef EXTENDED_CONSTPROP
  RegsetSetUnion(state->value_minus,regs);
  RegsetSetUnion(state->value_zero,regs);
  RegsetSetUnion(state->value_plus,regs);
#endif
  RegsetSetUnion(state->cond_true,regs);
  RegsetSetUnion(state->cond_false,regs);
}

/** Set all propagated items in a state to Top (value, reloc tag, cond bits */
void ProcStateSetRegsetTop(t_procstate* state, t_regset regs)
{
  RegsetSetUnion(state->top,regs);
  RegsetSetDiff(state->bot,regs);
#ifdef EXTENDED_CONSTPROP
  RegsetSetDiff(state->value_minus,regs);
  RegsetSetDiff(state->value_zero,regs);
  RegsetSetDiff(state->value_plus,regs);
#endif
}

/** Set all propagated values in regs a state to Bot */
void ProcStateSetRegsetBot(t_procstate* state, t_regset regs)
{
  RegsetSetUnion(state->bot,regs);
  RegsetSetDiff(state->top,regs);
#ifdef EXTENDED_CONSTPROP
  RegsetSetUnion(state->value_minus,regs);
  RegsetSetUnion(state->value_zero,regs);
  RegsetSetUnion(state->value_plus,regs);
#endif
}

/** Set all propagated relocation tags in regs a state to Top */
void ProcStateSetTagsetTop(t_procstate* state, t_regset regs)
{
  RegsetSetUnion(state->tag_top,regs);
  RegsetSetDiff(state->tag_bot,regs);
}

/** Set all propagated relocation tags in regs a state to Bot */
void ProcStateSetTagsetBot(t_procstate* state, t_regset regs)
{
  RegsetSetUnion(state->tag_bot,regs);
  RegsetSetDiff(state->tag_top,regs);
}

/** Set all condition bits in regs a state to Top */
void ProcStateSetCondsetTop(t_procstate* state, t_regset regs)
{
  RegsetSetDiff(state->cond_true,regs);
  RegsetSetDiff(state->cond_false,regs);
}
 
/** Set all condition bits in regs a state to Bot */
void ProcStateSetCondsetBot(t_procstate* state, t_regset regs)
{
  RegsetSetUnion(state->cond_true,regs);
  RegsetSetUnion(state->cond_false,regs);
}

/** Set the propagated value in reg in a state to Top */
void ProcStateSetRegTop(t_procstate* state, t_reg reg)
{
  RegsetAddReg(state->top,reg);
#ifdef EXTENDED_CONSTPROP
  RegsetSubReg(state->value_minus,reg);
  RegsetSubReg(state->value_zero,reg);
  RegsetSubReg(state->value_plus,reg);
#endif
  RegsetSubReg(state->bot,reg);
}

/** Set the propagated value in reg in a state to Bot */
void ProcStateSetRegBot(t_procstate* state, t_reg reg)
{
  RegsetAddReg(state->bot,reg);
  RegsetSubReg(state->top,reg);
#ifdef EXTENDED_CONSTPROP
  RegsetAddReg(state->value_minus,reg);
  RegsetAddReg(state->value_zero,reg);
  RegsetAddReg(state->value_plus,reg);
#endif
}

void ProcStateSetBothBot(t_procstate* state, t_reg reg)
{
  RegsetAddReg(state->bot,reg);
  RegsetSubReg(state->top,reg);
  RegsetAddReg(state->tag_bot,reg);
  RegsetSubReg(state->tag_top,reg);
#ifdef EXTENDED_CONSTPROP
  RegsetAddReg(state->value_minus,reg);
  RegsetAddReg(state->value_zero,reg);
  RegsetAddReg(state->value_plus,reg);
#endif
}

#ifdef EXTENDED_CONSTPROP
void ProcStateSetRegBotExtend(t_procstate* state, t_reg reg, t_bool plus, t_bool zero, t_bool minus)
{
  RegsetAddReg(state->bot,reg);
  RegsetSubReg(state->top,reg);
  if(plus)
    RegsetAddReg(state->value_plus,reg);
  else
    RegsetSubReg(state->value_plus,reg);
  if(zero)
    RegsetAddReg(state->value_zero,reg);
  else
    RegsetSubReg(state->value_zero,reg);
  if(minus)
    RegsetAddReg(state->value_minus,reg);
  else
    RegsetSubReg(state->value_minus,reg);
}

void ProcStateGetRegBotExtend(t_procstate* state, t_reg reg, t_bool * plus, t_bool * zero, t_bool * minus)
{
  *plus=RegsetIn(state->value_plus,reg);
  *zero=RegsetIn(state->value_zero,reg);
  *minus=RegsetIn(state->value_minus,reg);
}
#endif

/** Set the propagated relocation tag in reg in a state to Top */
void ProcStateSetTagTop(t_procstate* state, t_reg reg)
{
  RegsetAddReg(state->tag_top,reg);
  RegsetSubReg(state->tag_bot,reg);
}

/** Set the propagated relocation tag in reg in a state to Bot */
void ProcStateSetTagBot(t_procstate* state, t_reg reg)
{
  RegsetAddReg(state->tag_bot,reg);
  RegsetSubReg(state->tag_top,reg);
}

/** Set the propagated cond bit in reg in a state to Top */
void ProcStateSetCondTop(t_procstate* state, t_reg reg)
{
  RegsetSubReg(state->cond_true,reg);
  RegsetSubReg(state->cond_false,reg);
}

/** Set the propagated cond bit in reg in a state to Bot */
void ProcStateSetCondBot(t_procstate* state, t_reg reg)
{
  RegsetAddReg(state->cond_true,reg);
  RegsetAddReg(state->cond_false,reg);
}

/** Set the propagated value in reg in a state to value */
void ProcStateSetReg(t_procstate* state, t_reg reg, register_content value)
{
  RegsetSubReg(state->bot,reg);
  RegsetSubReg(state->top,reg);
#ifdef EXTENDED_CONSTPROP
  RegsetAddReg(state->value_minus,reg);
  RegsetAddReg(state->value_zero,reg);
  RegsetAddReg(state->value_plus,reg);
#endif
  state->register_subset[Reg2Set(reg)][Reg2Index(reg)] = value;
}

#ifdef EXTENDED_CONSTPROP
void ProcStateSetRegPlus(t_procstate* state, t_reg reg, register_content value)
{
  RegsetSubReg(state->bot,reg);
  RegsetSubReg(state->top,reg);
  RegsetSubReg(state->value_minus,reg);
  RegsetSubReg(state->value_zero,reg);
  RegsetAddReg(state->value_plus,reg);
  state->register_subset[Reg2Set(reg)][Reg2Index(reg)] = value;
}

void ProcStateSetRegZero(t_procstate* state, t_reg reg, register_content value)
{
  RegsetSubReg(state->bot,reg);
  RegsetSubReg(state->top,reg);
  RegsetSubReg(state->value_minus,reg);
  RegsetAddReg(state->value_zero,reg);
  RegsetSubReg(state->value_plus,reg);
  state->register_subset[Reg2Set(reg)][Reg2Index(reg)] = value;
}

void ProcStateSetRegMinus(t_procstate* state, t_reg reg, register_content value)
{
  RegsetSubReg(state->bot,reg);
  RegsetSubReg(state->top,reg);
  RegsetAddReg(state->value_minus,reg);
  RegsetSubReg(state->value_zero,reg);
  RegsetSubReg(state->value_plus,reg);
  state->register_subset[Reg2Set(reg)][Reg2Index(reg)] = value;
}

void ProcStateRegDup(t_procstate* state,t_reg reg, t_reg reg2) 
{
  t_lattice_level level;
  register_content value;
  if(ProcStateGetPlus(state,reg))
    RegsetAddReg(state->value_plus,reg2);
  else
    RegsetSubReg(state->value_plus,reg2);
  if(ProcStateGetZero(state,reg))
    RegsetAddReg(state->value_zero,reg2);
  else
    RegsetSubReg(state->value_zero,reg2);
  if(ProcStateGetMinus(state,reg))
    RegsetAddReg(state->value_minus,reg2);
  else
    RegsetSubReg(state->value_minus,reg2);
  level= ProcStateGetReg(state,reg, &value);
  if(level==BOT)
  {
    RegsetAddReg(state->bot,reg2);
    RegsetSubReg(state->top,reg2);
  }
  else if(level==TOP)
  {
    RegsetSubReg(state->bot,reg2);
    RegsetAddReg(state->top,reg2);
  }
  else
  {
    RegsetSubReg(state->bot,reg2);
    RegsetSubReg(state->top,reg2);
    state->register_subset[Reg2Set(reg2)][Reg2Index(reg2)] = value;
  }
}

void ProcStateRegDupProcstate(t_procstate* state,t_procstate* dest,t_reg reg) 
{
  t_lattice_level level;
  register_content value;
  if(ProcStateGetPlus(state,reg))
    RegsetAddReg(dest->value_plus,reg);
  else
    RegsetSubReg(dest->value_plus,reg);
  if(ProcStateGetZero(state,reg))
    RegsetAddReg(dest->value_zero,reg);
  else
    RegsetSubReg(dest->value_zero,reg);
  if(ProcStateGetMinus(state,reg))
    RegsetAddReg(dest->value_minus,reg);
  else
    RegsetSubReg(dest->value_minus,reg);
  level= ProcStateGetReg(state,reg, &value);
  if(level==BOT)
  {
    RegsetAddReg(dest->bot,reg);
    RegsetSubReg(dest->top,reg);
  }
  else if(level==TOP)
  {
    RegsetSubReg(dest->bot,reg);
    RegsetAddReg(dest->top,reg);
  }
  else
  {
    RegsetSubReg(dest->bot,reg);
    RegsetSubReg(dest->top,reg);
    dest->register_subset[Reg2Set(reg)][Reg2Index(reg)] = value;
  }
}
#endif


/** Set the propagated relocation tag in reg in a state to value */
void ProcStateSetTag(t_procstate* state, t_reg reg, t_reloc* value)
{
  RegsetSubReg(state->tag_bot,reg);
  RegsetSubReg(state->tag_top,reg);
  state->register_tag_subset[Reg2Set(reg)][Reg2Index(reg)] = value;
}

/** Set the propagated condition bit in reg in a state to value */
void ProcStateSetCond(t_procstate* state, t_reg reg, t_bool value)
{
  if (value)
    {
      RegsetAddReg(state->cond_true,reg);
      RegsetSubReg(state->cond_false,reg);
    }
  else
    {
      RegsetAddReg(state->cond_false,reg);
      RegsetSubReg(state->cond_true,reg);
    }
}

#ifdef EXTENDED_CONSTPROP
void ProcStateSetMinus(t_procstate* state, t_reg reg)
{
  RegsetAddReg(state->value_minus,reg);
}

void ProcStateSetZero(t_procstate* state, t_reg reg)
{
  RegsetAddReg(state->value_zero,reg);
}

void ProcStateSetPlus(t_procstate* state, t_reg reg)
{
  RegsetAddReg(state->value_plus,reg);
}

t_bool ProcStateGetTop(t_procstate* state, t_reg reg)
{
  if (RegsetIn(state->top,reg))
    return TRUE;
  return FALSE;
}

t_bool ProcStateGetBot(t_procstate* state, t_reg reg)
{
  if (RegsetIn(state->bot,reg))
    return TRUE;
  return FALSE;
}

t_bool ProcStateGetPlus(t_procstate* state, t_reg reg)
{
  if (RegsetIn(state->value_plus,reg))
    return TRUE;
  return FALSE;
}

t_bool ProcStateGetZero(t_procstate* state, t_reg reg)
{
  if (RegsetIn(state->value_zero,reg))
    return TRUE;
  return FALSE;
}

t_bool ProcStateGetMinus(t_procstate* state, t_reg reg)
{
  if (RegsetIn(state->value_minus,reg))
    return TRUE;
  return FALSE;
}
#endif

/** Get the value of a reg in a propagated state. This is returned in
    the third argument. The actual return value is TOP,VALUE or
    BOT. */
t_lattice_level ProcStateGetReg(t_procstate* state, t_reg reg, register_content * value)
{
  if (RegsetIn(state->top,reg))
    {
      return TOP;
    }

  if (RegsetIn(state->bot,reg))
    {
      return BOT;
    }

  *value = state->register_subset[Reg2Set(reg)][Reg2Index(reg)];

  return VALUE;
}

/** Get the relocation tag of a reg in a propagated state. This is returned in
    the third argument. The actual return value is TOP,VALUE or
    BOT. */
t_lattice_level ProcStateGetTag(t_procstate* state, t_reg reg, t_reloc ** value)
{
  if (RegsetIn(state->tag_top,reg))
    {
      return TOP;
    }

  if (RegsetIn(state->tag_bot,reg))
    {
      return BOT;
    }

  *value = state->register_tag_subset[Reg2Set(reg)][Reg2Index(reg)];

  return VALUE;
}

/** Get the condition bit of a reg in a propagated state. This is returned in
    the third argument. The actual return value is TOP,VALUE or
    BOT. */
t_lattice_level ProcStateGetCond(t_procstate* state, t_reg reg, t_bool * value)
{
  if (RegsetIn(state->cond_true,reg) && RegsetIn(state->cond_false,reg))
    {
      return BOT;
    }
  else if (RegsetIn(state->cond_true,reg))
    {
      *value = 1;
      return VALUE;
    }
  else if (RegsetIn(state->cond_false,reg))
    {
      *value = 0;
      return VALUE;
    }
  else
    {
      return TOP;
    }
}

/** Join src with dest, but only for the registers in regs. The final
    argument is not yet used, but it will be once dead data detection is
    implemented.*/

t_bool ProcStateJoinSimple(t_procstate* dest, t_procstate* src, t_regset regs) 
{
  t_bool change = FALSE;
  t_reg  reg;
  t_regset dest_bot     = RegsetNew();
  t_regset dest_old_bot = RegsetNew();
  t_regset dest_top     = RegsetNew();
  t_regset src_bot      = RegsetNew();
  t_regset src_top      = RegsetNew();
  t_regset src_old_top  = RegsetNew();
  t_regset src_old_bot  = RegsetNew();
  t_regset dest_old_top = RegsetNew();
  t_regset old_tag_bot  = RegsetNew();
  t_regset tmp          = RegsetNew();
  t_regset singleton    = RegsetNew();
  t_regset dest_tag_bot = RegsetNew();
  t_regset dest_tag_top = RegsetNew();
  t_regset src_tag_bot  = RegsetNew();
  t_regset src_tag_top  = RegsetNew();
  t_regset regs2        = RegsetNew();
  t_regset dest_cond_false = RegsetNew();
  t_regset dest_cond_true  = RegsetNew();

  register_content *src_set;
  register_content *dest_set;
  
  t_reloc **src_tag_set;
  t_reloc **dest_tag_set;
#ifdef EXTENDED_CONSTPROP
  t_regset old_plus     = RegsetNew();
  t_regset old_minus     = RegsetNew();
  t_regset old_zero      = RegsetNew();
  t_regset hulp_zero      = RegsetNew();
  t_regset hulp_minus      = RegsetNew();
  t_regset hulp_plus      = RegsetNew();
#endif


  register_content * src_ptr, * dest_ptr;
  t_reloc ** src_tag_ptr, **dest_tag_ptr;
  int set;

  if (!dest) {     
    FATAL(("DEST = NULL!\n"));
  }

  /* first do normal registers */

  RegsetDup(dest_bot,dest->bot);
  RegsetDup(dest_top,dest->top);

  if (RegsetEquals(dest_top,AllRegs()) && RegsetEquals(regs,AllRegs()))
    {
      ProcStateDup(dest,src);
      return TRUE;
    }

  RegsetDup(dest_old_bot,dest->bot);
  RegsetDup(dest_old_top,dest->top);

  RegsetDup(src_bot,src->bot);
  RegsetDup(src_top,src->top);
  RegsetDup(src_old_top,src->top);
  RegsetDup(src_old_bot,src->bot);

#ifdef EXTENDED_CONSTPROP
  RegsetDup(old_minus,dest->value_minus);
  RegsetDup(hulp_minus,src->value_minus);
  RegsetSetIntersect(hulp_minus,regs);
  RegsetSetUnion(dest->value_minus,hulp_minus);

  change |= !RegsetEquals(dest->value_minus,old_minus);

  RegsetDup(old_plus,dest->value_plus);
  RegsetDup(hulp_plus,src->value_plus);
  RegsetSetIntersect(hulp_plus,regs);
  RegsetSetUnion(dest->value_plus,hulp_plus);

  change |= !RegsetEquals(dest->value_plus,old_plus);

  RegsetDup(old_zero,dest->value_zero);
  RegsetDup(hulp_zero,src->value_zero);
  RegsetSetIntersect(hulp_zero,regs);
  RegsetSetUnion(dest->value_zero,hulp_zero);

  change |= !RegsetEquals(dest->value_zero,old_zero);
#endif

  RegsetDup(tmp,src_bot);
  RegsetSetIntersect(tmp,regs);
  RegsetSetUnion(dest_bot,tmp);
  RegsetSetDiff(dest_top,tmp);

  change |= !RegsetEquals(dest_bot,dest_old_bot);

  RegsetDup(regs2,regs);
  RegsetSetDiff(regs2,dest_bot);
  RegsetSetDiff(regs2,src_top);

  src_set = src->register_subset[0];
  dest_set = dest->register_subset[0];

  for (set=MAX_NR_REG_SUBSETS-1;set>=0;set--)
    {
      t_regset subset = register_subset(set);
      RegsetDup(tmp,regs2);
      RegsetSetIntersect(tmp,subset);
      RegsetSetDiff(regs2,tmp);

      REGSUBSET_FOREACH_SINGLETON_REGSET3(set,tmp,reg,singleton)
	{
	  src_ptr = src_set+reg;
	  dest_ptr = dest_set + reg;
	  
	  if (RegsetIsSubset(dest_top,singleton))
	    {
	      *dest_ptr = *src_ptr;
	      RegsetSetDiff(dest_top,singleton);
	      RegsetSetDiff(dest_bot,singleton);
	      change = TRUE;
	      continue;
	    }
	  
	  if (!memcmp(src_ptr,dest_ptr,sizeof(register_content)))
	    continue;
	  
	  RegsetSetUnion(dest_bot,singleton);
	  RegsetSetDiff(dest_top,singleton);
	  
	  change = TRUE;

	}
      if (RegsetIsEmpty(regs2))
	break;
    }

  /* then do tag registers */

  RegsetDup(old_tag_bot,dest->tag_bot);
  RegsetDup(dest_tag_bot,dest->tag_bot);
  RegsetDup(dest_tag_top,dest->tag_top);
  RegsetDup(src_tag_bot,src->tag_bot);
  RegsetDup(src_tag_top,src->tag_top);

  RegsetDup(tmp,src_tag_bot);
  RegsetSetIntersect(tmp,regs);
  RegsetSetUnion(dest_tag_bot,tmp);
  RegsetSetDiff(dest_tag_top,tmp);
  change |= !RegsetEquals(dest_tag_bot,old_tag_bot);

  RegsetDup(regs2,regs);
  RegsetSetDiff(regs2,dest_tag_bot);
  RegsetSetDiff(regs2,src_tag_top);

  src_tag_set = src->register_tag_subset[0];
  dest_tag_set = dest->register_tag_subset[0];

  for (set=MAX_NR_REG_SUBSETS-1;set>=0;set--)
    {
      t_regset subset = register_subset(set);
      
      RegsetDup(tmp,regs2);
      RegsetSetIntersect(tmp,subset);
      RegsetSetDiff(regs2,tmp);
      
      REGSUBSET_FOREACH_SINGLETON_REGSET3(set,tmp,reg,singleton)
	{	    
	  src_tag_ptr = src_tag_set + reg;
	  dest_tag_ptr = dest_tag_set + reg;
	  
	  if (RegsetIsSubset(dest_tag_top,singleton))
	    {
	      *dest_tag_ptr = *src_tag_ptr;
	      RegsetSetDiff(dest_tag_top,singleton);
	      RegsetSetDiff(dest_tag_bot,singleton);
	      change = TRUE;
	      continue;
	    }
	  
	  if (TwoRelocsInSameBlock(*src_tag_ptr,*dest_tag_ptr))
	    continue;
	  
	  RegsetSetUnion(dest_tag_bot,singleton);
	  RegsetSetDiff(dest_tag_top,singleton);
	  
	  change = TRUE;
	}
      if (RegsetIsEmpty(regs2))
	break;
   }
  
  /* and then do the condition registers */
  
  RegsetDup(dest_cond_false,dest->cond_false);
  RegsetDup(dest_cond_true,dest->cond_true);
  RegsetSetUnion(dest->cond_false,src->cond_false);
  RegsetSetUnion(dest->cond_true,src->cond_true);

  if (!RegsetEquals(dest_cond_false,dest->cond_false) || !RegsetEquals(dest_cond_true,dest->cond_true))
    change |= TRUE;

  RegsetDup(dest->bot,dest_bot);
  RegsetDup(dest->top,dest_top);
  RegsetDup(dest->tag_bot,dest_tag_bot);
  RegsetDup(dest->tag_top,dest_tag_top);

  RegsetFree(dest_bot);
  RegsetFree(dest_top);
  RegsetFree(src_bot);
  RegsetFree(src_top);
  RegsetFree(bitreg);
  RegsetFree(dest_old_bot);
  RegsetFree(src_old_top);
  RegsetFree(tmp);
  RegsetFree(singleton);
  RegsetFree(dest_cond);
  RegsetFree(src_cond);
  RegsetFree(dest_cond_false);
  RegsetFree(dest_cond_true);
#ifdef EXTENDED_CONSTPROP
  RegsetFree(old_minus);
  RegsetFree(hulp_minus);
  RegsetFree(old_plus);
  RegsetFree(hulp_plus);
  RegsetFree(old_zero);
  RegsetFree(hulp_zero);
#endif

  return change;
}


t_bool ProcStateJoin(t_procstate* dest, t_procstate* src, t_regset regs, t_regset live_regs) 
{
  t_bool change = FALSE;
  t_reg  reg;
  t_regset dest_bot     = RegsetNew();
  t_regset dest_top     = RegsetNew();
  t_regset src_bot      = RegsetNew();
  t_regset src_top      = RegsetNew();
  t_regset dest_old_bot = RegsetNew();
  t_regset src_old_top  = RegsetNew();
  t_regset src_old_bot  = RegsetNew();
  t_regset dest_old_top = RegsetNew();
  t_regset old_tag_bot  = RegsetNew();
  t_regset tmp          = RegsetNew();
  t_regset singleton    = RegsetNew();
  t_regset dest_tag_bot = RegsetNew();
  t_regset dest_tag_top = RegsetNew();
  t_regset src_tag_bot  = RegsetNew();
  t_regset src_tag_top  = RegsetNew();
  t_regset regs2        = RegsetNew();
  t_regset dest_cond_false = RegsetNew();
  t_regset dest_cond_true  = RegsetNew();

#ifdef EXTENDED_CONSTPROP
  t_regset old_plus     = RegsetNew();
  t_regset old_minus     = RegsetNew();
  t_regset old_zero      = RegsetNew();
  t_regset hulp_zero      = RegsetNew();
  t_regset hulp_minus      = RegsetNew();
  t_regset hulp_plus      = RegsetNew();
#endif

  register_content * src_ptr, * dest_ptr;
  t_reloc ** src_tag_ptr, **dest_tag_ptr;
  int set;
  register_content *src_set;
  register_content *dest_set;
  t_reloc **src_tag_set;
  t_reloc **dest_tag_set;

  if (!dest) { FATAL(("DEST = NULL!\n")); }

  /* first do normal registers */

  RegsetDup(dest_bot,dest->bot);
  RegsetDup(dest_top,dest->top);

  if (RegsetEquals(dest_top,AllRegs()) && RegsetEquals(regs,AllRegs()))
    {
      ProcStateDup(dest,src);
      return TRUE;
    }

  RegsetDup(dest_old_bot,dest->bot);
  RegsetDup(dest_old_top,dest->top);

  RegsetDup(src_bot,src->bot);
  RegsetDup(src_top,src->top);
  RegsetDup(src_old_top,src->top);
  RegsetDup(src_old_bot,src->bot);

#ifdef EXTENDED_CONSTPROP
  RegsetDup(old_minus,dest->value_minus);
  RegsetDup(hulp_minus,src->value_minus);
  RegsetSetIntersect(hulp_minus,regs);
  RegsetSetUnion(dest->value_minus,hulp_minus);

  change |= !RegsetEquals(dest->value_minus,old_minus);

  RegsetDup(old_plus,dest->value_plus);
  RegsetDup(hulp_plus,src->value_plus);
  RegsetSetIntersect(hulp_plus,regs);
  RegsetSetUnion(dest->value_plus,hulp_plus);

  change |= !RegsetEquals(dest->value_plus,old_plus);

  RegsetDup(old_zero,dest->value_zero);
  RegsetDup(hulp_zero,src->value_zero);
  RegsetSetIntersect(hulp_zero,regs);
  RegsetSetUnion(dest->value_zero,hulp_zero);

  change |= !RegsetEquals(dest->value_zero,old_zero);
#endif

  RegsetDup(tmp,src_bot);
  RegsetSetIntersect(tmp,regs);
  RegsetSetUnion(dest_bot,tmp);
  RegsetSetDiff(dest_top,tmp);

  change |= !RegsetEquals(dest_bot,dest_old_bot);

  RegsetDup(regs2,regs);
  RegsetSetDiff(regs2,dest_bot);
  RegsetSetDiff(regs2,src_top);

  src_set = src->register_subset[0];
  dest_set = dest->register_subset[0];

  for (set=MAX_NR_REG_SUBSETS-1;set>=0;set--)
    {
      t_regset subset = register_subset(set);
      RegsetDup(tmp,regs2);
      RegsetSetIntersect(tmp,subset);
      RegsetSetDiff(regs2,tmp);

      REGSUBSET_FOREACH_SINGLETON_REGSET3(set,tmp,reg,singleton)
	{
	  src_ptr = src_set+reg;
	  dest_ptr = dest_set + reg;
	  
	  if (RegsetIsSubset(dest_top,singleton))
	    {
	      *dest_ptr = *src_ptr;
	      RegsetSetDiff(dest_top,singleton);
	      RegsetSetDiff(dest_bot,singleton);
	      change = TRUE;
	      continue;
	    }
	  
	  if (!memcmp(src_ptr,dest_ptr,sizeof(register_content)))
	    continue;
	  
	  RegsetSetUnion(dest_bot,singleton);
	  RegsetSetDiff(dest_top,singleton);
	  
	  change = TRUE;

	}
      if (RegsetIsEmpty(regs2))
	break;
    }

  /* then do tag registers */

  RegsetDup(old_tag_bot,dest->tag_bot);
  RegsetDup(dest_tag_bot,dest->tag_bot);
  RegsetDup(dest_tag_top,dest->tag_top);
  RegsetDup(src_tag_bot,src->tag_bot);
  RegsetDup(src_tag_top,src->tag_top);

  RegsetDup(tmp,src_tag_bot);
  RegsetSetIntersect(tmp,regs);
  RegsetSetUnion(dest_tag_bot,tmp);
  RegsetSetDiff(dest_tag_top,tmp);
  change |= !RegsetEquals(dest_tag_bot,old_tag_bot);

  RegsetDup(regs2,regs);
  RegsetSetDiff(regs2,dest_tag_bot);
  RegsetSetDiff(regs2,src_tag_top);

  src_tag_set = src->register_tag_subset[0];
  dest_tag_set = dest->register_tag_subset[0];

  for (set=MAX_NR_REG_SUBSETS-1;set>=0;set--)
    {
      t_regset subset = register_subset(set);
      
      RegsetDup(tmp,regs2);
      RegsetSetIntersect(tmp,subset);
      RegsetSetDiff(regs2,tmp);
      
      REGSUBSET_FOREACH_SINGLETON_REGSET3(set,tmp,reg,singleton)
	{	    
	  src_tag_ptr = src_tag_set + reg;
	  dest_tag_ptr = dest_tag_set + reg;
	  
	  if (RegsetIsSubset(dest_tag_top,singleton))
	    {
	      *dest_tag_ptr = *src_tag_ptr;
	      RegsetSetDiff(dest_tag_top,singleton);
	      RegsetSetDiff(dest_tag_bot,singleton);
	      change = TRUE;
	      continue;
	    }
	  
	  if (TwoRelocsInSameBlock(*src_tag_ptr,*dest_tag_ptr))
	    continue;
	  
	  RegsetSetUnion(dest_tag_bot,singleton);
	  RegsetSetDiff(dest_tag_top,singleton);
	  
	  change = TRUE;
	}
      if (RegsetIsEmpty(regs2))
	break;
   }
  
  /* and then do the condition registers */
  
  RegsetDup(dest_cond_false,dest->cond_false);
  RegsetDup(dest_cond_true,dest->cond_true);
  RegsetSetUnion(dest->cond_false,src->cond_false);
  RegsetSetUnion(dest->cond_true,src->cond_true);

  if (!RegsetEquals(dest_cond_false,dest->cond_false) || !RegsetEquals(dest_cond_true,dest->cond_true))
    change |= TRUE;

  RegsetDup(dest->bot,dest_bot);
  RegsetDup(dest->top,dest_top);
  RegsetDup(dest->tag_bot,dest_tag_bot);
  RegsetDup(dest->tag_top,dest_tag_top);

  RegsetFree(dest_bot);
  RegsetFree(dest_top);
  RegsetFree(src_bot);
  RegsetFree(src_top);
  RegsetFree(bitreg);
  RegsetFree(dest_old_bot);
  RegsetFree(src_old_top);
  RegsetFree(tmp);
  RegsetFree(singleton);
  RegsetFree(dest_cond);
  RegsetFree(src_cond);
  RegsetFree(dest_cond_false);
  RegsetFree(dest_cond_true);

#ifdef EXTENDED_CONSTPROP
  RegsetFree(old_minus);
  RegsetFree(hulp_minus);
  RegsetFree(old_plus);
  RegsetFree(hulp_plus);
  RegsetFree(old_zero);
  RegsetFree(hulp_zero);
#endif

  return change;
}

/** Duplicate state src to dest */
void ProcStateDup(t_procstate *dest, t_procstate* src)
{
  RegsetDup(dest->top,src->top);
  RegsetDup(dest->bot,src->bot);
  RegsetDup(dest->tag_top,src->tag_top);
  RegsetDup(dest->tag_bot,src->tag_bot);
  RegsetDup(dest->cond_true,src->cond_true);
  RegsetDup(dest->cond_false,src->cond_false);
#ifdef EXTENDED_CONSTPROP
  RegsetDup(dest->value_zero,src->value_zero);
  RegsetDup(dest->value_plus,src->value_plus);
  RegsetDup(dest->value_minus,src->value_minus);
#endif

  if (!RegsetEquals(RegsetNewUnion(src->top,src->bot),AllRegs()))
    {
      memcpy(dest->register_subset[0],src->register_subset[0],sizeof(register_content)*REG_SUBSET_SIZE*MAX_NR_REG_SUBSETS);
    }
  if (!RegsetEquals(RegsetNewUnion(src->tag_top,src->tag_bot),AllRegs()))
    memcpy(dest->register_tag_subset[0],src->register_tag_subset[0],sizeof(t_reloc*)*REG_SUBSET_SIZE*MAX_NR_REG_SUBSETS);
}

void ProcStateDupRegsIntoTop(t_procstate *dest, t_procstate* src, t_regset regs)
{
  RegsetDup(dest->top,src->top);
  RegsetDup(dest->bot,src->bot);
  RegsetDup(dest->tag_top,src->tag_top);
  RegsetDup(dest->tag_bot,src->tag_bot);
  RegsetDup(dest->cond_true,src->cond_true);
  RegsetDup(dest->cond_false,src->cond_false);
#ifdef EXTENDED_CONSTPROP
  RegsetDup(dest->value_zero,src->value_zero);
  RegsetDup(dest->value_plus,src->value_plus);
  RegsetDup(dest->value_minus,src->value_minus);
#endif

  if (!RegsetEquals(RegsetNewUnion(src->top,src->bot),AllRegs()))
    {
      memcpy(dest->register_subset[0],src->register_subset[0],sizeof(register_content)*REG_SUBSET_SIZE*MAX_NR_REG_SUBSETS);
    }
  if (!RegsetEquals(RegsetNewUnion(src->tag_top,src->tag_bot),AllRegs()))
    memcpy(dest->register_tag_subset[0],src->register_tag_subset[0],sizeof(t_reloc*)*REG_SUBSET_SIZE*MAX_NR_REG_SUBSETS);
}


t_procstate * ProcStateNewDup(t_procstate* src)
{
  int set;
  t_procstate * dest;
  register_content * rc;
  t_reloc ** tags;

  if (pslist)
  {
    dest = (t_procstate *) pslist;
    pslist = *pslist;
  }
  else
  {
    dest = (t_procstate *) Malloc(sizeof(t_procstate));
    rc = Calloc(MAX_NR_REG_SUBSETS*REG_SUBSET_SIZE,sizeof(register_content));
    tags = Malloc(MAX_NR_REG_SUBSETS*REG_SUBSET_SIZE*sizeof(t_reloc *));
    for (set=0;set<MAX_NR_REG_SUBSETS;set++)
    {
      dest->register_subset[set]=&(rc[set * REG_SUBSET_SIZE]);
      dest->register_tag_subset[set]=&(tags[set * REG_SUBSET_SIZE]);
    }
  }

  RegsetDup(dest->top,src->top);
  RegsetDup(dest->bot,src->bot);
  RegsetDup(dest->tag_top,src->tag_top);
  RegsetDup(dest->tag_bot,src->tag_bot);
  RegsetDup(dest->cond_true,src->cond_true);
  RegsetDup(dest->cond_false,src->cond_false);
#ifdef EXTENDED_CONSTPROP
  RegsetDup(dest->value_zero,src->value_zero);
  RegsetDup(dest->value_plus,src->value_plus);
  RegsetDup(dest->value_minus,src->value_minus);
#endif

  if (!RegsetEquals(RegsetNewUnion(src->top,src->bot),AllRegs()))
    {
      memcpy(dest->register_subset[0],src->register_subset[0],sizeof(register_content)*REG_SUBSET_SIZE*MAX_NR_REG_SUBSETS);
    }
  if (!RegsetEquals(RegsetNewUnion(src->tag_top,src->tag_bot),AllRegs()))
    memcpy(dest->register_tag_subset[0],src->register_tag_subset[0],sizeof(t_reloc*)*REG_SUBSET_SIZE*MAX_NR_REG_SUBSETS);
  
#ifdef DEBUG_PROCSTATE
  dest->phase = src->phase;
  dest->max_phase = src->max_phase;
#endif

  return dest;
}

/** Joins the relocatoin tags in reg src in state with the one in
    register dest */
void ProcStateJoinTags(t_procstate * state, t_reg dest, t_reg src)
{
  t_reloc * reloc_src;
  t_reloc * reloc_dest;
  t_lattice_level value_src;
  t_lattice_level value_dest;


  if (TOP == (value_src = ProcStateGetTag(state,src,&reloc_src)))
    return;
  else if (BOT == value_src)
    ProcStateSetTagBot(state,dest);
  else 
    {
      if (TOP == (value_dest = ProcStateGetTag(state,dest,&reloc_dest)))
	ProcStateSetTag(state,dest,reloc_src);
      else if (BOT==value_dest)
	return;
      else if (!TwoRelocsInSameBlock(reloc_src,reloc_dest))
	ProcStateSetTagBot(state,dest);
    }
}
  
/* Dominique: supporting functions for t_regstate */
#if 0
/** initializes a regstate */
void RegStateInit(t_regstate * state)
{
  RegStateSetTop(state);
  RegStateSetTagTop(state);
  RegStateSetCondTop(state);
  RegStateSetReg(state,(t_reg) -1);
}

/** allocates a regstate array of length n */
t_regstate * RegStateAlloc(t_uint32 n)
{
  t_regstate * retval = Malloc(n*sizeof(t_regstate));
  t_uint32 i;
  for (i=0; i<n; i++)
    RegStateInit(&(retval[i]));
  return retval;
}

/** creates an array of regstates from a procstate and a give register set.
    returns NULL if the register set is empty or if all registers are BOT (this saves a lot of memory) */
t_regstate * RegStateArrayCreateFromProcState(t_procstate * ps, t_regset whichregs, const t_architecture_description * desc)
{
  t_regstate * ret;
  t_uint32 regcount;
  t_reg r;
  t_uint32 index;
  t_bool allbot = TRUE;
  t_regset condregs = desc->cond_registers;

  regcount = RegsetCountRegs(whichregs);
  if (regcount == 0)
    return NULL;

  ret = RegStateAlloc(regcount);

  index = 0;
  REGSET_FOREACH_REG(whichregs,r)
  {
    t_regstate * current = &(ret[index]);
    RegStateSetReg(current,r);

    if (RegsetIn(condregs,r))
    {
      t_bool set;
      t_lattice_level level;
      level = ProcStateGetCond(ps,r,&set);
      if (level == TOP)
      {
	RegStateSetCondTop(current);
	allbot = FALSE;
      }
      else if (level == BOT)
      {
	RegStateSetCondBot(current);
      }
      else
      {
	RegStateSetCond(current,set);
	allbot = FALSE;
      }
    }
    else
    {
      register_content cont;
      t_lattice_level level;
      t_reloc * tag;
      t_lattice_level taglevel;
      
      level = ProcStateGetReg(ps,r,&cont);
      if (level == TOP)
      {
	RegStateSetTop(current);
	allbot = FALSE;
      }
      else if (level == BOT)
      {
	RegStateSetBot(current);
      }
      else
      {
	RegStateSetValue(current,cont);
	allbot = FALSE;
      }

      taglevel = ProcStateGetTag(ps,r,&tag);
      if (taglevel == TOP)
      {
	RegStateSetTagTop(current);
	allbot = FALSE;
      }
      else if (taglevel == BOT)
      {
	RegStateSetTagBot(current);
      }
      else
      {
	RegStateSetTag(current,tag);
	allbot = FALSE;
      }
    }
    index++;
  }
  
  if (allbot) {
    Free(ret);
    return NULL;
  }
  
  return ret;
}

t_lattice_level RegStateGetRegValue(t_ins * ins, t_regset * regset, t_reg reg, register_content * content)
{
  t_reg tmp_reg;
  t_uint32 i = 0;
  if (!INS_USED_REGSTATE(ins))
    return BOT;
  else
  {
    REGSET_FOREACH_REG(*regset,tmp_reg)
    {
      if (RegStateGetReg(&INS_USED_REGSTATE(ins)[i]) == reg) break;
      i++;
    }
    return RegStateGetValue(&INS_USED_REGSTATE(ins)[i],content); 
  }
}

t_lattice_level RegStateGetRegTag(t_ins * ins, t_regset * regset, t_reg reg, t_reloc ** tag)
{
  t_reg tmp_reg;
  t_uint32 i = 0;
  if (!INS_USED_REGSTATE(ins))
    return BOT;
  else
  {
    REGSET_FOREACH_REG(*regset,tmp_reg)
    {
      if (RegStateGetReg(&INS_USED_REGSTATE(ins)[i]) == reg) break;
      i++;
    }
    return RegStateGetTag(&INS_USED_REGSTATE(ins)[i],tag); 
  }
}

t_lattice_level RegStateGetRegCond(t_ins * ins, t_regset * regset, t_reg reg, t_bool * cond)
{
  t_reg tmp_reg;
  t_uint32 i = 0;
  if (!INS_USED_REGSTATE(ins))
    return BOT;
  else
  {
    REGSET_FOREACH_REG(*regset,tmp_reg)
    {
      if (RegStateGetReg(&INS_USED_REGSTATE(ins)[i]) == reg) break;
      i++;
    }
    return RegStateGetCond(&INS_USED_REGSTATE(ins)[i],cond); 
  }
}
#endif
/* vim: set shiftwidth=2 foldmethod=marker: */
