/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: Definitions for register sets
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/

#include "diablo_regset.h"
#include "../simple_types.h"
#include "../utils/util_error.h"

#if MAX_REG_ITERATOR > 64

#ifndef BIT64ADDRSUPPORT

t_regset NullRegs = {MAX_REG_ITERATOR, {0}};

t_regset AllRegs(void)
{
  t_regset ret={MAX_REG_ITERATOR,{0}};
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) ret.regset[i] = ~0;
  return ret;
}

t_regset register_subset(t_uint16 index)
{
  t_regset ret={MAX_REG_ITERATOR,{0}};

  if (index >= (MAX_REG_ITERATOR/16))
    FATAL(("Registerindex out of range! (%d >= %d)",index, MAX_REG_ITERATOR/16));
  ret.regset[index] = ~0;
  return ret;
}

t_regset RegsetNewInvers(t_regset x)
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) x.regset[i] = ~(x.regset[i]);
  return x;
}

t_regset RegsetNewUnion(t_regset x, t_regset y)
{
  t_uint16 i;
  t_regset ret;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) ret.regset[i] = (x.regset[i])|(y.regset[i]);
  return ret; 
}

t_regset RegsetNewIntersect(t_regset x, t_regset y)
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) x.regset[i] = (x.regset[i])&(y.regset[i]);
  return x;
}

t_regset RegsetNewDiff(t_regset x, t_regset y)
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) x.regset[i] = (x.regset[i])&(~(y.regset[i]));
  return x;
}

void real_RegsetSetInvers(t_regset* x)
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) x->regset[i] = ~(x->regset[i]);
  return;
}

void real_RegsetAddReg(t_regset* r, t_reg reg)
{
  t_uint16 i=reg/16;

  r->regset[i] |= 1<<(reg&0xf);
  return;
}

/** \internal Should not be used, this is a remainder of an existing anomaly */
void real_RegsetAddMultipleRegs(t_regset* r, t_reg reg)
{
  t_uint16 i = sizeof(t_reg)*8;
  t_reg j;
  
  for(j=0;j<i;j++)
  {
    if(reg&0x1) real_RegsetAddReg(r,j);
    reg >>= 1;
  }
  return;
}

void real_RegsetSetSingleton(t_regset* r, t_uint16 i)
{
  *r = NullRegs;
  r->regset[i/16] = 1<<(i&0xf);
  return;
}

void real_RegsetSubReg(t_regset* r, t_uint16 i)
{
  r->regset[i/16] &= ~(1<<(i&0xf));
  return;
}

void real_RegsetSetUnion(t_regset* x, t_regset y) /*(x)|=(y)*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) x->regset[i] |= y.regset[i];
  return;
}

void real_RegsetSetIntersect(t_regset* x, t_regset y) /*(x)&=(y)*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) x->regset[i] &= y.regset[i];
  return;
}

void real_RegsetSetDiff(t_regset* x, t_regset y) /*(x)&=(~(y))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) x->regset[i] &= ~(y.regset[i]);
  return;
}

void real_RegsetSetMutualExclusive(t_regset* x, t_regset y) /*(x)=((x)&~(y))|((~x)&(y))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) x->regset[i] = ((x->regset[i])&~(y.regset[i]))|(~(x->regset[i])&(y.regset[i]));
  return;
}

t_bool RegsetIsMutualExclusive(t_regset x, t_regset y) /*(!((x)&(y)))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) if(x.regset[i]&y.regset[i]) return FALSE;
  return TRUE;
}

t_bool RegsetIsEmpty(t_regset r) /*((r)==NullRegs)*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) if(r.regset[i]) return FALSE;
  return TRUE;
}

t_bool RegsetEquals(t_regset x, t_regset y) /* ((x)==(y))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) if(x.regset[i]!=y.regset[i]) return FALSE;
  return TRUE;
}

t_bool RegsetIn(t_regset x, t_reg y)    /*((x)&(1<<y))*/
{
  if (x.regset[y/16] & (1<<(y&0xf))) return TRUE;
  return FALSE;
}

t_bool RegsetIsSubset(t_regset x, t_regset y) /* (!((~(x))&y))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/16;i++) if(~(x.regset[i]) & y.regset[i]) return FALSE;
  return TRUE;
}

/** \internal Used in iterator */
t_regset RegsetNextSingleton(t_regset x)
{
  t_regset ret=NullRegs;
  t_uint16 i=0;

  while(!x.regset[i]) i++;
  if(x.regset[i] & 0x8000) ret.regset[i+1] = 0x0001;
  else ret.regset[i] = x.regset[i]<<1;
  return ret;
}


#else /* BIT64ADDRSUPPORT*/

t_regset NullRegs = {MAX_REG_ITERATOR, {0ULL}};

t_regset AllRegs(void)
{
  t_regset ret={MAX_REG_ITERATOR,{0ULL}};
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) ret.regset[i] = ~0x0ULL;
  return ret;
}

t_regset register_subset(t_uint16 index)
{
  t_regset ret={MAX_REG_ITERATOR,{0ULL}};

  if (index >= (MAX_REG_ITERATOR/64))
    FATAL(("Registerindex out of range! (%d >= %d)",index, MAX_REG_ITERATOR/64));
  ret.regset[index] = ~0x0ULL;
  return ret;
}

t_regset RegsetNewInvers(t_regset x)
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) x.regset[i] = ~(x.regset[i]);
  return x;
}

t_regset RegsetNewUnion(t_regset x, t_regset y)
{
  t_uint16 i;
  t_regset ret;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) ret.regset[i] = (x.regset[i])|(y.regset[i]);
  return ret; 
}

t_regset RegsetNewIntersect(t_regset x, t_regset y)
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) x.regset[i] = (x.regset[i])&(y.regset[i]);
  return x;
}

t_regset RegsetNewDiff(t_regset x, t_regset y)
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) x.regset[i] = (x.regset[i])&(~(y.regset[i]));
  return x;
}

void real_RegsetSetInvers(t_regset* x)
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) x->regset[i] = ~(x->regset[i]);
  return;
}

void real_RegsetAddReg(t_regset* r, t_reg reg)
{
  t_uint16 i=reg/64;

  r->regset[i] |= 1ULL<<(reg&0x3f);
  return;
}

/** \internal Should not be used, this is a remainder of an existing anomaly */
void real_RegsetAddMultipleRegs(t_regset* r, t_reg reg)
{
  t_uint16 i = 64;//sizeof(t_reg)*8;
  t_reg j;
  
  for(j=0;j<i;j++)
  {
    if(reg&0x1) real_RegsetAddReg(r,j);
    reg >>= 1;
  }
  return;
}

void real_RegsetSetSingleton(t_regset* r, t_uint16 i)
{
  *r = NullRegs;
  r->regset[i/64] = 1ULL<<(i&0x3f);
  return;
}

void real_RegsetSubReg(t_regset* r, t_uint16 i)
{
  r->regset[i/64] &= ~(1ULL<<(i&0x3f));
  return;
}

void real_RegsetSetUnion(t_regset* x, t_regset y) /*(x)|=(y)*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) x->regset[i] |= y.regset[i];
  return;
}

void real_RegsetSetIntersect(t_regset* x, t_regset y) /*(x)&=(y)*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) x->regset[i] &= y.regset[i];
  return;
}

void real_RegsetSetDiff(t_regset* x, t_regset y) /*(x)&=(~(y))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) x->regset[i] &= ~(y.regset[i]);
  return;
}

void real_RegsetSetMutualExclusive(t_regset* x, t_regset y) /*(x)=((x)&~(y))|((~x)&(y))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) x->regset[i] = ((x->regset[i])&~(y.regset[i]))|(~(x->regset[i])&(y.regset[i]));
  return;
}

t_bool RegsetIsMutualExclusive(t_regset x, t_regset y) /*(!((x)&(y)))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) if(x.regset[i]&y.regset[i]) return FALSE;
  return TRUE;
}

t_bool RegsetIsEmpty(t_regset r) /*((r)==NullRegs)*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) if(r.regset[i]) return FALSE;
  return TRUE;
}

t_bool RegsetEquals(t_regset x, t_regset y) /* ((x)==(y))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) if(x.regset[i]!=y.regset[i]) return FALSE;
  return TRUE;
}

t_bool RegsetIn(t_regset x, t_reg y)    /*((x)&(1<<y))*/
{
  if (x.regset[y/64] & (1ULL<<(y&0x3f))) return TRUE;
  return FALSE;
}

t_bool RegsetIsSubset(t_regset x, t_regset y) /* (!((~(x))&y))*/
{
  t_uint16 i;

  for(i=0;i<MAX_REG_ITERATOR/64;i++) if(~(x.regset[i]) & y.regset[i]) return FALSE;
  return TRUE;
}

/** \internal Used in iterator */
t_regset RegsetNextSingleton(t_regset x)
{
  t_regset ret=NullRegs;
  t_uint16 i=0;

  while(!x.regset[i]) i++;
  if(x.regset[i] & 0x8000000000000000ULL) ret.regset[i+1] = 0x0001ULL;
  else ret.regset[i] = x.regset[i]<<1;
  return ret;
}


#endif


#elif MAX_REG_ITERATOR > 32
t_regset register_subset[]={0xffffLL,0xffff0000LL,0xffff00000000LL,0xffff000000000000LL};
#else
t_regset register_subset[]={0xffff,0xffff0000};
#endif

/** count the number of registers in the set */
t_uint32 RegsetCountRegs(t_regset r)
{
  t_uint32 count = 0;
  t_reg i;
  REGSET_FOREACH_REG(r,i)
    count++;
  return count;
}
/* vim: set shiftwidth=2 foldmethod=marker: */
