/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/

#include "diablo_udchain.h"
#include "diablo_print.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"

t_ud_chain * UdChainDupReal(t_ud_chain * udc);
t_ud_chain * UdChainNew(t_bbl * def);

t_ud_chain_list * UdChainListNew(t_uint32 regcount)
{
  t_ud_chain_list * ret = (t_ud_chain_list *) Malloc(sizeof(t_ud_chain_list));

  ret->reg = (t_ud_chain **) Calloc(regcount, sizeof(t_ud_chain *));
  ret->num_regs = regcount;

  return ret;
}

void UdChainListFree(t_ud_chain_list * udcl)
{
  t_uint32 i;

  for (i=0; i<udcl->num_regs; i++)
    UdChainFree(udcl->reg[i]);

  Free(udcl->reg);
  Free(udcl);
}

t_ud_chain_list * UdChainListClone(t_ud_chain_list * udcl)
{
  t_ud_chain_list * ret;
  t_uint32 i;

  ret = UdChainListNew(udcl->num_regs);
  for (i=0; i<ret->num_regs; i++)
    ret->reg[i] = UdChainDup(udcl->reg[i]);

  return ret;
}

void UdChainListDup(t_ud_chain_list * dest, t_ud_chain_list * src)
{
  t_uint32 i;

  /* clear the original contents of dest */
  for (i=0; i<dest->num_regs; i++)
  {
    if (src->reg[i]!=NULL)
      UdChainFree(dest->reg[i]);
  }
  
  if (src->num_regs != dest->num_regs)
    Free(dest->reg);

  /* copy the contents of src into dest */
  if (!dest->reg) 
    dest->reg = (t_ud_chain **) Malloc(dest->num_regs*sizeof(t_ud_chain *));

  for (i=0; i<dest->num_regs; i++)
    if (src->reg[i]==NULL)
    {
      dest->reg[i] = NULL;
    }
    else
    {
      dest->reg[i] = UdChainDup(src->reg[i]);
    }

}

/*! Join two ud_chain_lists, for all registers named in regs */
t_bool UdChainListJoin(t_ud_chain_list * dest, t_ud_chain_list * src, t_regset regs)
{
  t_uint32 i;
  t_bool change = FALSE;

  /* join == append the lists */
  REGSET_FOREACH_REG(regs,i)
  {
    if (UdChainJoin(&(dest->reg[i]),src->reg[i]))
    {
      change |= TRUE;
    }
  }

  return change;
}

t_regset UdChainListGetTopRegs(t_ud_chain_list * udcl)
{
  t_regset ret = RegsetNew();
  t_uint32 i;

  for (i=0; i<udcl->num_regs; i++)
    if (!(udcl->reg[i]))
      RegsetAddReg(ret,i);
  return ret;
}

t_ud_chain * UdChainNew(t_bbl * def)
{
  t_ud_chain * ret;
  
  ret = (t_ud_chain *) Malloc(sizeof(t_ud_chain));
  
  ret->node.next = NULL;
  ret->def = def;
  ret->dups=1;

  return ret;
}

void UdChainFree(t_ud_chain * udc)
{
  t_ud_chain * tmp;

  if (udc)
  {
     udc->dups--;
     if (udc->dups==0)
     {
	do   
	{
	   tmp = udc;
	   udc = LLNext(udc);
	   Free(tmp);
	}
	while (udc);
     }
  }
}

void UdChainAppend(t_ud_chain ** head, t_ud_chain * insert)
{
   t_ud_chain * prev=NULL;
   t_ud_chain * iter;


   if ((*head)&&(((*head)->dups)!=1)) { 
      (*head)->dups--;
      *head=UdChainDupReal(*head);
   }

  
   for (iter=*head; iter!=NULL; iter=(t_ud_chain *) iter->node.next)
   {
      if (insert->def==iter->def) 
      {
	UdChainFree(insert);
	return;
      }

      if (insert->def<iter->def) break;
      prev=iter;
   }

   if (prev)
   {
      insert->node.next = (ll *) iter;
      prev->node.next=(ll *) insert;
   }
   else
   {
      insert->node.next = (ll *) *head;
      *head = insert;
   }
}

t_ud_chain * UdChainDup(t_ud_chain * udc)
{
  if (udc)
    udc->dups++;
  return udc;
}

t_ud_chain * UdChainDupReal(t_ud_chain * udc)
{
  t_ud_chain * ret, * prev;
  t_uint32 wtel=0;

  ret= NULL;
  prev=NULL;
  while (udc)
  {
     wtel++;
     if (prev)
     {
	LLNext(prev)=(ll *) UdChainNew(udc->def);
	prev=(t_ud_chain *) LLNext(prev);
     }
     else
     {
	ret=UdChainNew(udc->def);
	prev=ret;
     }
    udc = LLNext(udc);
  }

  return ret;
}
t_ud_chain * UdChainDupReal2(t_ud_chain * udc,t_ud_chain ** last)
{
  t_ud_chain * ret, * prev;
  t_uint32 wtel=0;

  ret= NULL;
  prev=NULL;
  while (udc)
  {
     wtel++;
     if (prev)
     {
	LLNext(prev)=(ll *) UdChainNew(udc->def);
	prev=LLNext(prev);
     }
     else
     {
	ret=UdChainNew(udc->def);
	prev=ret;
     }
    udc = LLNext(udc);
  }

  *last=prev;

  return ret;
}
/*#define UD_CACHE */
#define UD_CACHE_SIZE 0x100

#ifdef UD_CACHE
typedef struct _t_ud_cache_entry
{
#ifdef UD_CACHE_TEL
   t_int32 keep;
#endif
   t_ud_chain * smallest;
   t_ud_chain * largest;
   t_ud_chain * join_result;
   t_bool changed;
} t_ud_cache_entry;

t_ud_cache_entry ud_cache[UD_CACHE_SIZE]={{0,0,0,0}};
#endif

t_bool UdChainJoin(t_ud_chain ** dest, t_ud_chain * src)
{
   t_ud_chain * tmp, * prev=NULL;
   t_bool changed = FALSE;
   t_bool need_dup=FALSE;
#ifdef UD_CACHE
   t_bool cache_on=FALSE;
   long hash;
#endif

   /* for each link of the src-chain, check if it already exists in the dest-chain.
    * if not, append it to the dest-chain
    */

   if (!src)
     return FALSE;

   if (!*dest) {
      *dest=UdChainDup(src);
      return TRUE;
   }

   if ((*dest)==src) return FALSE; 

   if (((*dest)->dups)!=1)
   {
      need_dup = TRUE;
   }

#ifdef UD_CACHE 
   //if ((((*dest)->node.next)&&(src->node.next)))
   {
      static t_uint32 hit=0;
      static t_uint32 miss=0;

      cache_on=TRUE;
      t_ud_chain * small;
      t_ud_chain * large;
      hash=(((long) (*dest)+ (long) src)>>8)&(UD_CACHE_SIZE-1);
      if ((*dest)<src)
      {
	 small=*dest;
	 large=src;
      }
      else
      {
	 large=*dest;
	 small=src;
      }

      if ((ud_cache[hash].smallest==small)&&(ud_cache[hash].largest==large))
      {
#ifdef UD_CACHE_TEL
	 ud_cache[hash].keep++;
#endif
	 hit++;
	 ud_cache[hash].join_result->dups++;
	 *dest=ud_cache[hash].join_result;
	 return ud_cache[hash].changed; 
      }
      else
      {
	 miss++;
#ifdef UD_CACHE_TEL	 
	 ud_cache[hash].keep--;
	 if (ud_cache[hash].keep<=0)
	 {
#endif
	 
	    if (ud_cache[hash].smallest) UdChainFree(ud_cache[hash].smallest);
	    if (ud_cache[hash].largest) UdChainFree(ud_cache[hash].largest);
	    ud_cache[hash].smallest=small;
	    ud_cache[hash].largest=large;

	    small->dups++;
	    large->dups++;
	    if (ud_cache[hash].join_result) UdChainFree(ud_cache[hash].join_result);
#ifdef UD_CACHE_TEL	 
	 }
#endif
      }

      if (miss%10000==0) printf("Hitrate=%d\n",hit*100/(hit+miss));
   }
#endif


   
restart:

   tmp=*dest;
     do
     {
       register void *sdef, *tdef;
       sdef = src->def;
       tdef = tmp->def;
       if (sdef == tdef) 
       {
	 /* They ar equal ... skip */
	 src=LLNext(src);
	 prev=tmp;
	 tmp=LLNext(tmp);
       } 
       else if (sdef < tdef)
       {
	 if (need_dup)
	 {
	   need_dup=FALSE;
	   ((*dest)->dups)--;
	   *dest=UdChainDupReal(*dest);
	   goto restart;
	 }
	 else
	 {
	   t_ud_chain * copy=UdChainNew(sdef);
	   if (prev)
	   {
	     LLNext(copy)= (ll *) tmp;
	     LLNext(prev)=(ll *) copy;
	     prev=copy;
	   }
	   else
	   {
	     copy->node.next=(ll *) *dest;
	     *dest=copy;
	     prev=copy;
	   }
	   src=LLNext(src);
	   if (!changed) changed = TRUE;
	 }
       }
       else
       {
	 prev=tmp;
	 tmp=LLNext(tmp);
       }
     }
     while (src && tmp);

   if (src && need_dup)
   {
      ((*dest)->dups)--;
      *dest=UdChainDupReal2(*dest,&prev);
      if (!changed) changed = TRUE;
   }

   while (src)
   {
      if (!changed) changed = TRUE;
      prev->node.next=(ll *) UdChainNew(src->def);
      prev=(t_ud_chain *) prev->node.next;
      src=LLNext(src);
   }
#ifdef UD_CACHE
   if (cache_on)
   {
#ifdef UD_CACHE_TEL
   if (ud_cache[hash].keep<=0)
   {
#endif
      ud_cache[hash].join_result=(*dest);
      ud_cache[hash].changed=changed;
#ifdef UD_CACHE_TEL
      ud_cache[hash].keep=10;
#endif
      (*dest)->dups++;
#ifdef UD_CACHE_TEL
   }
#endif
   }
#endif

   return changed;
}

void UdChainListPrint(FILE * out, t_ud_chain_list * udcl)
{
  t_uint32 i;

  DiabloPrint(out, "# of registers: %d\n", udcl->num_regs);
  for (i=0; i<udcl->num_regs; i++)
  {
    DiabloPrint(out,"reg[%d]: ",i);
    UdChainPrint(out,udcl->reg[i]);
  }
}

void UdChainPrint(FILE * out, t_ud_chain * udc)
{
  DiabloPrint(out, "{");
  while (udc)
  {
    DiabloPrint(out, "[%lx] @G%s", udc->def, BBL_OLD_ADDRESS(udc->def), LLNext(udc)?", ":"");
    udc = LLNext(udc);
  }
  DiabloPrint(out, "}\n");
}
/* vim: set shiftwidth=2 foldmethod=marker: */
