/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * }}}
 *****************************************************************************/

#include "diablo_udgraph.h"
#include "diablo_flowgraph.h"
#include "diablo_print.h"
#include "diablo_iterators.h"
#include "../datastructures/data_graph.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"

t_ud_graph * UdGraphNew(void)
{
  /* use node_size 0 => makes sure GraphNodeNew cannot be used to 
   * create new ud_nodes. This is necessary because ud_nodes are in
   * fact just instructions, and they must be created by other functions */
  return (t_ud_graph *) GraphNew(0,sizeof(t_ud_edge));
}

t_ud_graph * UdGraphNewForCfg(t_cfg * cfg)
{
  t_ud_graph * udgraph = CFG_UDGRAPH(cfg);
  t_bbl * i_bbl;
  t_ins * i_ins;

  /* first, clear the existing udgraph and setup an empty one */
  if (udgraph)
  {
    DiabloPrint(stdout,"Graph existed. Cleaning up first...\n");
    UdGraphFree(udgraph);
  }

  /* Add all instructions from the cfg to the udgraph */
  udgraph = UdGraphNew();
  CFG_FOREACH_BBL(cfg, i_bbl)
  {
    if (IS_DATABBL(i_bbl)) continue;
    BBL_FOREACH_INS(i_bbl, i_ins)
      UdGraphInsertNode(udgraph,i_ins);
  }
  
  CFG_UDGRAPH(cfg) = udgraph;
  return udgraph;
} 
 
void UdGraphInsertNode(t_ud_graph * graph, t_ud_node * node)
{
  GraphInitNode(T_GRAPH(graph), T_NODE(node));
}

t_ud_edge * UdGraphLinkUseToDefForRegister(t_ud_graph * graph, t_ud_node * use, t_ud_node * def, t_reg reg)
{
  return (t_ud_edge *) GraphAppendNode(T_GRAPH(graph), T_NODE(use), T_NODE(def), (t_uint32) reg);
}

/* Remove all of the graph, but leave the nodes itself allocated.
 * This is necessary because the nodes are instructions, which must be allocated
 * and freed somewhere else. */
void UdGraphFree(t_ud_graph * graph)
{
  t_edge * i_edge, * i_tmp1;
  t_node * i_node, * i_tmp2;

  /* Remove all edges (these may be freed) */
  GRAPH_FOREACH_EDGE_SAFE(graph, i_edge, i_tmp1)
  {
    GraphRemoveEdge(T_GRAPH(graph), i_edge);
  }

  /* Remove all nodes from the graph (nodes should not be freed) */
  GRAPH_FOREACH_NODE_SAFE(T_GRAPH(graph), i_node, i_tmp2)
  {
    GraphRemoveNodeNoFree(T_GRAPH(graph), i_node);
  }

  Free(graph);
}

/* This is a very quick and dirty way to print out a part of the ud graph. 
 * Just specify the instruction from where to start, how many links 
 * you want to follow and the file name to write the file toi */
static void node_writeout(t_ins * ins, FILE * out, t_uint32 depth);

void UdGraphWriteOut(t_ins * ins, t_string filename, t_uint32 recursiondepth)
{
  FILE * out = fopen(filename,"w");
  VERIFY_FASSERT(out,("Could not open %s for writing!",filename));

  InitEdgeMarking();

  DiabloPrint(out, "digraph \"@I\" {\n\tnode [shape=box]\n", ins);
  
  node_writeout(ins,out,recursiondepth);

  DiabloPrint(out, "}\n");
  fclose(out);
}

static void node_writeout(t_ins * ins, FILE * out, t_uint32 depth)
{
  t_edge * i_edge;

  if (depth == 0) return;

  DiabloPrint(out, "\t\"%p\" [label=\"@I\"]\n", ins, ins);
  NODE_FOREACH_SUCC_EDGE(ins,i_edge)
    if (!EdgeIsMarked(i_edge))
      DiabloPrint(out, "\t\"%p\" -> \"%p\"\n", ins, i_edge->tail);

  NODE_FOREACH_SUCC_EDGE(ins, i_edge)
    if (!EdgeIsMarked(i_edge))
    {
      EdgeMark(i_edge);
      node_writeout(T_INS(i_edge->tail), out, depth-1);
    }
}
/* vim: set shiftwidth=2 foldmethod=marker: */
