/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/

/*!\defgroup DIABLO_DATATYPES Diablo Generic Datatypes */
/*@{ */

/*! \defgroup ADDRESSES Generic addresses 
 *
 * Diablo can be compiled with 32-bit only, 64-bit only or 32-bit and
 * 64-bit address support. i.e. for 32-bit, 64-bit target
 * architectures or both. The latter results in cross-compaction
 * possibilities for both types of architectures. To include 32-bit
 * support, compile with -D BIT32ADDRSUPPORT, for 64-bit architectures
 * compile with -D BIT64ADDRSUPPORT. Note that this is completely
 * independent of the platform that is used to compile/use Diablo
 * itself.  
 *
 * For all three possibilities, we have provided a number of
 * operations (with macro's or functions) that efficiently perform the
 * most common operations, such as addition, comparison, etc. If both
 * 32-bit and 64-bit targets are supported, most operations are much
 * more expensive, as they require type checking and are implemented
 * as procedures. For portability, we urge you to use these generic
 * operations!!!
 *
 * This documentation is written with both 32-bit and 64-bit support in mind.  */
/*@{ */
#include "../simple_types.h"

#ifdef BIT32ADDRSUPPORT
#ifdef BIT64ADDRSUPPORT
#define GENERICADDRSUPPORT
#endif
#endif

#ifndef DIABLO_ADDRESSES_TYPEDEFS
#define DIABLO_ADDRESSES_TYPEDEFS
#ifdef GENERICADDRSUPPORT
/*! An address if we have 32 bit and 64 bit address support */
typedef struct _t_address t_address;
#else
#ifdef BIT32ADDRSUPPORT
/*! An address if we only have 32 bit address support */
typedef t_uint32 t_address;
#else 
/*! An address if we only have 64 bit address support */
typedef t_uint64 t_address;
#endif
#endif
typedef enum _generic_types 
{
   INVALID,
   BIT32,
   BIT64,
   EITHER,
   FLOAT,
   DOUBLE
} generic_types;
#endif /* DIABLO_ADDRESSES_TYPEDEFS */ 


#ifndef DIABLO_ADDRESSES_HEADER 
#define DIABLO_ADDRESSES_HEADER
#include "../datastructures/data_string.h"


#ifdef BIT32ADDRSUPPORT
#ifdef BIT64ADDRSUPPORT
#define GENERICADDRSUPPORT
#endif
#endif




#ifdef GENERICADDRSUPPORT
/*! An address if we have generic address support */
struct _t_address
{
   char type; /**< Stores the kind of address we are dealing with*/
   /** A union to access the real address, depending on 'type' */
   union
   {
      t_uint32 addr32;
      t_uint64 addr64;
   } types;
}; 

#define GenericIs64(x)           (((x).type==BIT64)||((x).type==EITHER))
#define GenericIs32(x)           (((x).type==BIT32)||((x).type==EITHER))

#define GenericIsValid(x)        (x.type!=INVALID)
/*! Convert a generic address to a t_uint32 */
#define G_T_UINT32(x)            (x.types.addr32)	
/*! Convert a generic address to a t_uint64 */
#define G_T_UINT64(x)            (x.types.addr64)	
/*! Store a t_uint32 in a generic address */
#define GenericSet_uint32(y,x)    do { y.type=BIT32; y.types.addr32=x; } while(0)
/*! Store a t_uint64 in a generic address */
#define GenericSet_uint64(y,x)    do { y.type=BIT64; y.types.addr64=x; } while(0)

extern t_address INVALID_GENERIC;
extern t_address NULL_GENERIC;

t_address real_GENERIC_ADD(t_string x, t_uint32 lnno ,t_address a, t_address b);
t_address real_GENERIC_NOT(t_string x, t_uint32 lnno ,t_address a);
t_address real_GENERIC_AND(t_string x, t_uint32 lnno ,t_address a, t_address b);
t_address real_GENERIC_OR(t_string x, t_uint32 lnno ,t_address a, t_address b);
t_address real_GENERIC_XOR(t_string x, t_uint32 lnno ,t_address a, t_address b);
t_address real_GENERIC_SHL(t_string x, t_uint32 lnno ,t_address a, t_address b);
t_address real_GENERIC_SHR(t_string x, t_uint32 lnno ,t_address a, t_address b);
t_address real_GENERIC_ADD_UINT32(t_string file, t_uint32 lnno, t_address a, t_uint32 b);
t_address real_GENERIC_NAND_UINT32(t_string file, t_uint32 lnno, t_address a, t_uint32 b);
t_address real_GENERIC_ADD_INT32(t_string file, t_uint32 lnno, t_address a, t_int32 b);
t_address real_GENERIC_ADD_UINT64(t_string file, t_uint32 lnno, t_address a, t_uint64 b);
t_address real_GENERIC_ADD_INT64(t_string file, t_uint32 lnno, t_address a, t_int64 b);
t_address real_GENERIC_SUB_UINT32(t_string file, t_uint32 lnno, t_address a, t_uint32 b);
t_address real_GENERIC_SUB_INT32(t_string file, t_uint32 lnno, t_address a, t_int32 b);
t_address real_GENERIC_SUB(t_string file, t_uint32 line,t_address a, t_address b);
t_bool real_GENERIC_LT(t_string file, t_uint32 lnno, t_address a, t_address b);
t_bool real_GENERIC_LE(t_string file, t_uint32 lnno, t_address a, t_address b);
t_bool real_GENERIC_GT(t_string file, t_uint32 lnno, t_address a, t_address b);
t_bool real_GENERIC_GE(t_string file, t_uint32 lnno, t_address a, t_address b);
t_bool real_GENERIC_EQ(char * file, int lnno, t_address a, t_address b);
void GENERIC_PRINT(t_address p, t_int32 len);

t_address real_GENERIC_SIGN_EXTEND(t_string file, t_uint32 lnno, t_address x, t_uint32 signbit);

/** Add t_address u to t_address v, this macro is preferred! */
#define GENERIC_ADD(u,v) real_GENERIC_ADD(__FILE__,__LINE__,(u),(v))
#define GENERIC_AND(u,v) real_GENERIC_AND(__FILE__,__LINE__,(u),(v))
#define GENERIC_OR(u,v)  real_GENERIC_OR(__FILE__,__LINE__,(u),(v))
#define GENERIC_XOR(u,v)  real_GENERIC_XOR(__FILE__,__LINE__,(u),(v))
#define GENERIC_NOT(u) real_GENERIC_NOT(__FILE__,__LINE__,(u))
#define GENERIC_SHL(x,y) real_GENERIC_SHL(__FILE__,__LINE__,(u),(v))
#define GENERIC_SHR(x,y) real_GENERIC_SHR(__FILE__,__LINE__,(u),(v))
/** Subtract t_address v from t_address u, this macro is preferred! */
#define GENERIC_SUB(u,v) real_GENERIC_SUB(__FILE__,__LINE__,(u),(v))
/** Is t_address u less than t_address v, this macro is preferred! */
#define GENERIC_LT(u,v) real_GENERIC_LT(__FILE__,__LINE__,u,v)
/** Is t_address u less than or equal to t_address v, this macro is preferred! */
#define GENERIC_LE(u,v) real_GENERIC_LE(__FILE__,__LINE__,u,v)
/** Is t_address u greater than t_address v, this macro is preferred! */
#define GENERIC_GT(u,v) real_GENERIC_GT(__FILE__,__LINE__,u,v)
/** Does t_address u equals t_address v, this macro is preferred! */
#define GENERIC_EQ(u,v) real_GENERIC_EQ(__FILE__,__LINE__,u,v)
#define GENERIC_GE(u,v) real_GENERIC_GE(__FILE__,__LINE__,u,v)
/** Add t_address u to \b t_uint32 v, this macro is preferred! */
#define GENERIC_ADD_UINT32(u,v) real_GENERIC_ADD_UINT32(__FILE__,__LINE__,u,v)
#define GENERIC_NAND_UINT32(u,v) real_GENERIC_NAND_UINT32(__FILE__,__LINE__,u,v)
/** Add t_address u to \b t_int32 v, this macro is preferred! */
#define GENERIC_ADD_INT32(u,v) real_GENERIC_ADD_INT32(__FILE__,__LINE__,u,v)
/** Add t_address u to \b t_uint64 v, this macro is preferred! */
#define GENERIC_ADD_UINT64(u,v) real_GENERIC_ADD_UINT64(__FILE__,__LINE__,u,v)
/** Add t_address u to \b t_int64 v, this macro is preferred! */
#define GENERIC_ADD_INT64(u,v) real_GENERIC_ADD_INT64(__FILE__,__LINE__,u,v)
/** Subtract \b t_uint32 v from t_address u, this macro is preferred! */
#define GENERIC_SUB_UINT32(u,v) real_GENERIC_SUB_UINT32(__FILE__,__LINE__,u,v)
/** Subtract \b t_int32 v from t_address u, this macro is preferred! */
#define GENERIC_SUB_INT32(u,v) real_GENERIC_SUB_INT32(__FILE__,__LINE__,u,v)
void * real_GENERIC_ADD_DISPL(char * file, int lnno,void * in,t_address addr);

#define GENERIC_DIV_UINT32(x,y) real_GENERIC_DIV_UINT32(__FILE__,__LINE__,(x),(y))	
t_address real_GENERIC_DIV_UINT32(char * file, int line, t_address address, t_uint32 div);

#define GENERIC_MUL_UINT32(x,y) real_GENERIC_MUL_UINT32(__FILE__,__LINE__,(x),(y))	
t_address real_GENERIC_MUL_UINT32(char * file, int line, t_address address, t_uint32 div);

#define GENERIC_EXTRACT_UINT32(x)  real_GENERIC_EXTRACT_UINT32(__FILE__,__LINE__,(x))  
#define GENERIC_EXTRACT_UINT64(x)  real_GENERIC_EXTRACT_UINT64(__FILE__,__LINE__,(x))  
#define GENERIC_EXTRACT_INT32(x)   real_GENERIC_EXTRACT_INT32(__FILE__,__LINE__,(x))  
#define GENERICISNULL(x)	real_GENERICISNULL(__FILE__,__LINE__,(x))
/** Add a byte-displacement (stored in t_address disp) to an address (void * addr) */
#define GENERIC_ADD_DISPL(addr,disp) real_GENERIC_ADD_DISPL(__FILE__,__LINE__,addr,disp) 

#define GENERIC_SIGN_EXTEND(x,signbit) real_GENERIC_SIGN_EXTEND(__FILE__,__LINE__,x,signbit)

#else
#ifdef BIT32ADDRSUPPORT
#define INVALID_GENERIC 0xffffffff
#define GenericIsValid(x)        (x!=INVALID_GENERIC)
#define NULL_GENERIC 0x0
#define G_T_UINT32(x)      (x)	
#define GenericSet_uint32(y,x)    ((y)=(x))

#define GENERIC_EQ(x,y)         ((x)==(y))
#define GENERIC_LE(x,y)         ((x)<=(y))
#define GENERIC_LT(x,y)         ((x)<(y))
#define GENERIC_GT(x,y)         ((x)>(y))
#define GENERIC_GE(x,y)         ((x)>=(y))
#define GENERIC_ADD(x,y)        ((x)+(y))
#define GENERIC_AND(x,y)        ((x)&(y))
#define GENERIC_OR(x,y)         ((x)|(y))
#define GENERIC_XOR(x,y)         ((x)^(y))
#define GENERIC_SUB(x,y)        ((x)-(y))
#define GENERIC_NOT(x)		(~(x))
#define GENERIC_SHL(x,y)	((x) << (y))
#define GENERIC_SHR(x,y)	((x) >> (y))
#define GENERIC_ADD_UINT32(x,y)    ((x)+((t_uint32) y))
#define GENERIC_NAND_UINT32(x,y)    ((x)&(~((t_uint32) y)))
#define GENERIC_SUB_UINT32(x,y)    ((x)-((t_uint32) y))
#define GENERIC_ADD_INT32(x,y)    ((x)+((t_int32) y))
#define GENERIC_SUB_INT32(x,y)    ((x)-((t_int32) y))
#define GENERIC_ADD_DISPL(addr,disp) ((void *) (((char *) addr) + disp))
#define GENERIC_PRINT(val,d)        printf("%x",val)
#define G_T_UINT64(x)              (=====INVALID=====)
#define GenericSet_uint64(y,x)     (=====INVALID=====)
#define GENERICISNULL(x)	(x==0)

#define GENERIC_DIV_UINT32(x,y)	((x)/(y))
#define GENERIC_MUL_UINT32(x,y)	((x)*(y))
#define GENERIC_EXTRACT_UINT32(x)    (x)
#define GENERIC_EXTRACT_UINT64(x)    ((t_uint64)(x))
#define GENERIC_EXTRACT_INT32(x)     (t_int32)(x)

#define GENERIC_SIGN_EXTEND(x,signbit)	((x) ^ (1 << (signbit))) - (1 << (signbit))

#endif
#ifdef BIT64ADDRSUPPORT
#define INVALID_GENERIC LL(0xffffffffffffffff)
#define NULL_GENERIC LL(0x0)
#define GenericIsValid(x)        (x!=INVALID_GENERIC)
#define GenericSet_uint64(y,x)    ((y)=(x))
#define G_T_UINT64(x)      (x)	
#define GENERIC_EQ(x,y)         ((x)==(y))
#define GENERIC_LE(x,y)         ((x)<=(y))
#define GENERIC_LT(x,y)         ((x)<(y))
#define GENERIC_GT(x,y)         ((x)>(y))
#define GENERIC_GE(x,y)         ((x)>=(y))
#define GENERIC_ADD(x,y)        ((x)+(y))
#define GENERIC_AND(x,y)        ((x)&(y))
#define GENERIC_OR(x,y)         ((x)|(y))
#define GENERIC_XOR(x,y)         ((x)^(y))
#define GENERIC_SUB(x,y)        ((x)-(y))
#define GENERIC_NOT(x)		(~(x))
#define GENERIC_SHL(x,y)	((x) << (y))
#define GENERIC_SHR(x,y)	((x) >> (y))
#define GENERIC_ADD_UINT32(x,y)    ((x)+((t_uint32) y))
#define GENERIC_ADD_UINT64(x,y)    ((x)+((t_uint64) y))
#define GENERIC_NAND_UINT32(x,y)    ((x)&(~((t_uint64) y)))
#define GENERIC_SUB_UINT32(x,y)    ((x)-((t_uint32) y))
#define GENERIC_ADD_INT32(x,y)    ((x)+((t_int32) y))
#define GENERIC_ADD_INT64(x,y)    ((x)+((t_int64) y))
#define GENERIC_SUB_INT32(x,y)    ((x)-((t_int32) y))
#define GENERIC_ADD_DISPL(addr,disp) ((void *) (((char *) addr) + disp))
#define GENERIC_PRINT(val,d)        printf("%llx",val)
#define G_T_UINT32(x)              (=====INVALID=====)
#define GenericSet_uint32(y,x)     (=====INVALID=====)
#define GENERICISNULL(x)	(x==0)

#define GENERIC_DIV_UINT32(x,y)	((x)/((t_uint64) (y)))
#define GENERIC_MUL_UINT32(x,y)	((x)*((t_uint64) (y)))
#define GENERIC_EXTRACT_UINT32(x)    ((t_uint32) x)
#define GENERIC_EXTRACT_UINT64(x)    (x)
#define GENERIC_EXTRACT_INT32(x)     (t_int32) ((t_uint32) x)
#define GENERIC_SIGN_EXTEND(x,signbit)	((x) ^ (1 << (signbit))) - (1 << (signbit))
#endif
#endif
#endif
/* @} */
/* @} */

t_address AlignAddress(t_address align, t_address start);

/* vim: set shiftwidth=2 foldmethod=marker: */
