/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/

/* Includes {{{ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../config.h"
#include "../simple_types.h"
#include "../diablo_options.h"
#include "diablo_object.h"
#include "diablo_archives.h"
#ifdef  HAVE_OBJZ_SUPPORT
#include "../fileformats/alpha_ecoff/ecoff_main.h"
#endif
#include "../utils/util_file.h"
#include "../utils/util_error.h"
#include "../utils/util_options.h"
#include "../utils/util_malloc.h"
#include "../datastructures/data_string.h"
#include "../datastructures/data_hash.h" 
#include "../fileformats/ar/ar_arch.h"
/* }}} */

/*!\addtogroup Archives */
/*@{ */
/*!\defgroup ARCH_INTERNAL Internal functions for archives */
/*@{*/
INTERNAL void ArchiveHashFree(void * to_free, void * unused_data);
/*@}*/
/*@}*/


data_hash_table * archives_global_hash_table=NULL;

/* Internal functions {{{*/
/*! Utility function, used to free all structures stored in the global archive
 * hash table 
 *
 * \param to_free     A node of the hash table that must be freed
 * \param unused_data Unused data 
 */
INTERNAL void ArchiveHashFree(void * to_free, void * unused_data)
{
   t_archive_hash_node * node=(t_archive_hash_node *) to_free;
   ArchiveFree(node->arch);
   Free(node->node.key);
   Free(node);
}
/*}}}*/

/*! TODO: document... Shouldn't this be internal? */
void ArchObjHashNodeFree(void * tf,void * data)
{
   t_arch_obj_hash_node * node=tf;
   Free(node->node.key);
   Free(node);
}

/* Archive Destructor {{{ */
/*! Free all structures and substructure used by an archive
 *
 * \param arch the archive to free 
 */

void ArchiveFree(t_archive * arch)
{
   Free(arch->name);
   HashFree(arch->objects);
   Free(arch);
}
/* }}} */
/* Archive Constructors {{{ */
/*! Constructor of an archive. Should probably not be used. Use ArchiveOpenCached
 * instead 
 */

static t_archive * ArchiveNew()
{
   t_archive * ret;
   ret=Malloc(sizeof(t_archive));
   return ret;
}

/*! Constructor of an archive with a filename. Should probably not be used. Use ArchiveOpenCached
 * instead.
 *
 * \param name The name of the archive to open
 */

static t_archive* ArchiveOpen(char * name)
{
   int tel=0;
   FILE * fp;
   t_archive * ret; 
   t_string full_name;

   PARANOID_FASSERT(global_options.libpath, ("Lib path not set!"));

   full_name=FindFile(global_options.libpath,name);
   if (!full_name) return NULL;
   fp=fopen(full_name,"r");
   VERIFY_FASSERT (fp, ("Could not open: %s", full_name));
   ret=ArchiveNew();
   ret->name=full_name;
   while(archive_handler_table[tel].type!=AT_UNKNOWN)
   {
      if (archive_handler_table[tel].ident(fp))
      {
	 ret->type=archive_handler_table[tel].type;
	 archive_handler_table[tel].open(fp,ret);
	 return ret;
      }
      tel++;
   }
   FATAL(("Archive format not supported (%s)",name));
   exit(0); /* Just to keep the compiler happy */
}

/*! Cached constructor: TODO document */ 
t_archive * ArchiveOpenCached(char * name)
{
   t_archive_hash_node * hn=(t_archive_hash_node *) HashLookup(archives_global_hash_table,name);
   /* First look in the archive hash table, to see if we already opened it */

   /* If not: Open */
   if (!hn)
   {
      t_archive * ar=ArchiveOpen(name);
      hn=Malloc(sizeof(t_archive_hash_node));
      hn->node.key=StringDup(name);
      hn->arch=ar;
      HashInsert(archives_global_hash_table,(data_hash_node *) hn);
      return ar;
   }
   /* Else: return archive */
   else
   {
      return hn->arch;
   }
}
/* }}} */
/* Global Constructor and Destructor {{{ */
/*! TODO: document */
void ArchivesInit()
{
   archives_global_hash_table=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,ArchiveHashFree);
}

/*!TODO: document */
void ArchivesFini()
{
   HashFree(archives_global_hash_table);
}
/* }}} */
/* Archive GetObject {{{ */

/*! TODO: Document */
t_object * ArchiveGetObject(t_archive * ar, char * name)
{
 t_arch_obj_hash_node * hn=(t_arch_obj_hash_node *) HashLookup(ar->objects,name);
 t_object * obj;

 if (!hn) return NULL;
 else 
 {
    if (hn->flags==COMPRESSED_ALPHA)
    {
#ifndef HAVE_OBJZ_SUPPORT
       FATAL(("Compressed alpha archives not supported on this architecture (or not compiled in), define -DHAVE_OBJZ_SUPPORT when compiling the source on Tru64 Unix systems and link with libmld. This should be autoconfigured. Contact me (bdebus at frodo.elis.rug.ac.be) if you have information about the ObjZ compression algoritm"));
       return NULL; /* keep the compiler happy */
#else
       t_string tmp=Collate3(ar->name,":",hn->node.key);
       FILE * out;
       t_uint64 uncomp_size;
       t_uint64 pad;
       unsigned char * data=NULL;
       int ret;
       ecoff_file_hdr fhdr;
       char data_tmp[8192];

       int cur_pos;

       
       if (fseek(ar->open_fp,hn->fpos,SEEK_SET)) FATAL(("Fseek failed"));
       ret=fread(&fhdr,1,sizeof(ecoff_file_hdr),ar->open_fp);
       if (ret<0) FATAL(("Could not read file!"));
       if ((!(fhdr.f_magic==0x183))&&(!(fhdr.f_magic==0x188)))
       {
	  FATAL(("Expected an object, file %s in archive %s has magic %x, fhdrs= %d!",name,ar->name,fhdr.f_magic,fhdr.f_nscns));
       }
       else
       {
	  fread(&uncomp_size,sizeof(t_uint64),1,ar->open_fp);
	  fread(&pad,sizeof(t_uint64),1,ar->open_fp);
       }
       cur_pos=lseek(fileno(ar->open_fp),0,SEEK_CUR);
       lseek(fileno(ar->open_fp),hn->fpos,SEEK_SET);
       data=(unsigned char *) Calloc(1,uncomp_size);
       obj_Zcopytocore(fileno(ar->open_fp),data,hn->size);
       lseek(fileno(ar->open_fp),cur_pos,SEEK_SET);
       
       out=fopen("/tmp/tmp_obj2.o","wb");
       fwrite(data,uncomp_size,1,out);
       fclose(out);
       Free(data);
       obj=ObjectReadWithFakeName("/tmp/tmp_obj2.o",tmp);
       unlink("/tmp/tmp_obj2.o");

       return obj;
#endif
       
    }
    else
    {
       t_string tmp=Collate3(ar->name,":",hn->node.key);
       fseek(ar->open_fp,hn->fpos,SEEK_SET);
       obj=ObjectGetFromStream(tmp,ar->open_fp);
       Free(tmp);
       return obj;
    }
 }
}
/*}}}*/

/* vim: set shiftwidth=2 foldmethod=marker: */
