/******************************************************************************
 * Copyright 2001,2002,2003: {{{
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This file contains all arch-independent functions to handle basic blocks in
 * objects
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002,2003 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/
/* Includes {{{ */
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include <stdlib.h> 
#include "diablo_bbl.h"
#include "diablo_object.h"
#include "diablo_function.h"
#include "diablo_ins.h"
#include "diablo_edge.h"
#include "diablo_flowgraph.h"
#include "diablo_dominator.h"
#include "diablo_iterators.h"
#include "../datastructures/data_graph.h"
/* }}} */
/* Constructor {{{ */
/*! 
 * \param cfg the graph in which the bbl is allocated 
 * \return a new bbl 
 *
 * constructor for the t_bbl structure:  allocates 
 * memory for a basic block structure and fills in some 
 * default values and inserts it into the given control flow graph */

extern t_bool dominator_info_correct;

t_bbl * BblNew(t_cfg * cfg)
{
  t_bbl * temp = Calloc(1,sizeof(t_bbl));
  temp->cfg=cfg;
  GraphInitNode(T_GRAPH(cfg),T_NODE(temp));
  RELOCATABLE_TYPE(temp)=RT_BBL;
  dominator_info_correct = FALSE;
  return temp;
}
/*}}}*/
/* Destructor {{{ */
/*!
 * \brief The default destructor 
 *
 * \param bbl The basic block to kill
 */
void BblKill(t_bbl * bbl)
{
  t_cfg * cfg = BBL_CFG(bbl); 
  t_function * fun = BBL_FUN(bbl);
  /*  data_ll_elem * free_ll;*/

  dominator_info_correct = FALSE;

  /* remove bbl from fun */
  if (T_RELOCATABLE(bbl)->refedby) 
  {
    FATAL(("BBL @B Still refered by @R (rel->hell == %s)",bbl, T_RELOCATABLE(bbl)->refedby->rel, T_RELOCATABLE(bbl)->refedby->rel->hell ? "TRUE" : "FALSE"));
  }

  if (fun)
  {
    if (BBL_PREV_IN_FUN(bbl))
      BBL_NEXT_IN_FUN(BBL_PREV_IN_FUN(bbl))=BBL_NEXT_IN_FUN(bbl);
    else
    {
      /* this is in fact dangerous. if the first bbl of a function is killed 
       * and we're not killing the entire function, we should look for the new
       * function entry and make sure that block becomes the new FUN_BBL_FIRST
       */
      FUN_BBL_FIRST(fun)=BBL_NEXT_IN_FUN(bbl);
    }

    if (BBL_NEXT_IN_FUN(bbl))
      BBL_PREV_IN_FUN(BBL_NEXT_IN_FUN(bbl))=BBL_PREV_IN_FUN(bbl);
    else
    {
      FUN_BBL_LAST(fun)=BBL_PREV_IN_FUN(bbl);
    }
  }

  if (BBL_DOMINATED_BY(bbl)) DominatorFree(&BBL_DOMINATED_BY(bbl));
  if (BBL_PDOMINATED_BY(bbl)) DominatorFree(&BBL_PDOMINATED_BY(bbl));
  if (BBL_TMP_PDOMINATED(bbl)) DominatorFree(&BBL_TMP_PDOMINATED(bbl));
  if (BBL_TMP_DOMINATED(bbl)) DominatorFree(&BBL_TMP_DOMINATED(bbl));
  LoopBblCleanup(bbl);

  /* remove bbl from cfg TODO should we remove it from entries as well? */
  GraphRemoveNode(T_GRAPH(cfg),T_NODE(bbl));
}

/* }}} */
/* Functions to insert instructions in bbls {{{ */

/* {{{ Append an instruction at the end of a bbl */
void InsAppendToBbl(t_ins * ins, t_bbl * bbl)
{
  if (INS_PREV(ins)||INS_NEXT(ins)) FATAL(("Already in bbl!\n"));
  INS_BBL(ins)=bbl;

  INS_ATTRIB(ins) |= IF_IN_BBL;

  if (!BBL_INS_FIRST(bbl)) 
  {
    if (bbl->nins!=0)
      {
	DiabloPrint(stdout,"@B\n",bbl);
      }
    VERIFY_FASSERT(!bbl->nins,("No instructions, yet ins count != 0"));
    BBL_INS_FIRST(bbl)=BBL_INS_LAST(bbl)=ins;
    bbl->nins=1;
    INS_NEXT(ins)=INS_PREV(ins)=NULL;
  }
  else
  {
    INS_NEXT(BBL_INS_LAST(bbl))=ins;
    INS_PREV(ins)=BBL_INS_LAST(bbl);
    INS_NEXT(ins)=NULL;
    BBL_INS_LAST(bbl)=ins;
    BBL_CSIZE(bbl) += INS_CSIZE(ins);
    bbl->nins++;
  }
}
/* }}}*/
/* {{{ Prepend an instruction to the front of a bbl */
void InsPrependToBbl(t_ins * ins, t_bbl * bbl)
{
  if (INS_PREV(ins)||INS_NEXT(ins)) FATAL(("Already in bbl!\n"));
  RELOCATABLE_REFED_BY(ins)=NULL;

  INS_BBL(ins)=bbl;
  INS_NEXT(ins)=BBL_INS_FIRST(bbl);
  if (BBL_INS_FIRST(bbl))
    INS_PREV(BBL_INS_FIRST(bbl))=ins;
  INS_PREV(ins)=NULL;
  BBL_INS_FIRST(bbl)=ins;
  if (!BBL_INS_LAST(bbl))
    BBL_INS_LAST(bbl)=ins;
  BBL_NINS(bbl)++;
}
/* }}}*/
/* {{{ Insert instruction after */
/*! Inserts an instruction after another instruction (in some bbl!) */
void InsInsertAfter(t_ins * ins, t_ins *after)
{
  INS_BBL(ins)=INS_BBL(after);
  INS_ATTRIB(ins) |= IF_IN_BBL;

  if (INS_NEXT(after))
    INS_PREV(INS_NEXT(after))=ins;
  INS_NEXT(ins)=INS_NEXT(after);
  INS_PREV(ins)=after;
  INS_NEXT(after)=ins;

  if (BBL_INS_LAST(INS_BBL(after))==after)
    BBL_INS_LAST(INS_BBL(after))=ins;

  BBL_NINS(INS_BBL(ins))++;
}
/*}}}*/
/* {{{ Insert instruction before */
/*! Inserts an instruction before another instruction (in some bbl!) */
void InsInsertBefore(t_ins * ins, t_ins *before)
{
  INS_BBL(ins)=INS_BBL(before);
  INS_ATTRIB(ins) |= IF_IN_BBL;

  if (INS_PREV(before))
    INS_NEXT(INS_PREV(before))=ins;
  INS_PREV(ins)=INS_PREV(before);
  INS_NEXT(ins)=before;
  INS_PREV(before)=ins;

  if (BBL_INS_FIRST(INS_BBL(before))==before)
    BBL_INS_FIRST(INS_BBL(before))=ins;

  BBL_NINS(INS_BBL(ins))++;
}
/*}}}*/
/*}}}*/
/* {{{ Determine used and defined registers for a bbl */
/*! fill in BBL_REGS_DEF_CERTAIN and BBL_REGS_DEF_PERHAPS */
void BblSetDefinedRegs(t_bbl * bbl)
{
  t_ins * i_ins;
  t_regset certain, perhaps, defined;

  certain = RegsetNew();
  perhaps = RegsetNew();
  
  BBL_FOREACH_INS(bbl,i_ins)
  {
    if (INS_IS_CONDITIONAL(i_ins))
      RegsetSetUnion(perhaps,INS_REGS_DEF(i_ins));
    else
      RegsetSetUnion(certain,INS_REGS_DEF(i_ins));
  }

  /* all registers that are in certain should not be in perhaps */
  RegsetSetDiff(perhaps,certain);

  BBL_REGS_DEF_CERTAIN(bbl) = certain;
  BBL_REGS_DEF_PERHAPS(bbl) = perhaps;

  RegsetDup(defined,certain);
  RegsetSetUnion(defined,perhaps);
  BBL_REGS_DEFINED(bbl) = defined;
}

/** Compute the registers that are defined in a basic block*/
t_regset BblRegsDef(t_bbl * bbl)
{
  t_ins * i_ins;
  t_regset def;
  t_function * fun = BBL_FUN(bbl);

  /*if( (!fun) || (FUN_IS_HELL(fun)) )
    return AllRegs();*/
  /* LUDO */
  if(!fun || bbl == bbl->cfg->hell_node || bbl == bbl->cfg->exit_hell_node || bbl == bbl->cfg->unique_entry_node || bbl == bbl->cfg->unique_exit_node)
    return AllRegs();
  if(bbl == bbl->cfg->call_hell_node || bbl == bbl->cfg->exit_call_hell_node)
  {
    def = RegsetNewInvers(BBL_OBJECT(bbl)->description->callee_saved);
    RegsetSetDiff(def,BBL_OBJECT(bbl)->description->cond_registers);
    return def;
  }

  def = RegsetNew();

  BBL_FOREACH_INS(bbl,i_ins)
  {
    if (!INS_IS_CONDITIONAL(i_ins))
    {
      RegsetSetUnion(def,INS_REGS_DEF(i_ins));
    }
  }
  return def; 
}

t_regset BblRegsMaybeDef(t_bbl * bbl)
{
  t_ins * i_ins;
  t_regset def;
  t_function * fun = BBL_FUN(bbl);

  /*if( (!fun) || (FUN_IS_HELL(fun)) )
    return AllRegs();*/
  /* LUDO */
  if(!fun || bbl == bbl->cfg->hell_node || bbl == bbl->cfg->exit_hell_node || bbl == bbl->cfg->unique_entry_node || bbl == bbl->cfg->unique_exit_node)
    return AllRegs();
  if(bbl == bbl->cfg->call_hell_node || bbl == bbl->cfg->exit_call_hell_node)
  {
    def = RegsetNewInvers(BBL_OBJECT(bbl)->description->callee_saved);
    RegsetSetDiff(def,BBL_OBJECT(bbl)->description->cond_registers);
    return def;
  }

  def = RegsetNew();

  BBL_FOREACH_INS(bbl,i_ins)
  {
    RegsetSetUnion(def,INS_REGS_DEF(i_ins));
  }
  return def; 
}

/** Compute the registers that are used in a basic block*/

#ifdef ARM_SUPPORT
#include "../arch/arm/arm_opcodes.h"
#include "../arch/arm/arm_utils.h"
#endif


t_regset BblRegsUse(t_bbl * bbl)
{
  t_ins * i_ins;
#ifdef ARM_SUPPORT
  t_regset use_eq, use_ne;
#endif
  t_regset use;

  t_function * fun = BBL_FUN(bbl);

  /*if((!fun) || (IsHellNode(bbl)) || (FUN_IS_HELL(fun)))
    return AllRegs(); */
  /* LUDO */
  if(!fun || bbl == bbl->cfg->hell_node || bbl == bbl->cfg->exit_hell_node || bbl == bbl->cfg->swi_hell_node || bbl == bbl->cfg->exit_swi_hell_node || bbl == bbl->cfg->unique_entry_node || bbl == bbl->cfg->unique_exit_node)
    return AllRegs();

  if(bbl == bbl->cfg->call_hell_node || bbl ==  bbl->cfg->exit_call_hell_node)
  {
    use = RegsetNewInvers(BBL_OBJECT(bbl)->description->callee_saved);
    RegsetSetDiff(use,BBL_OBJECT(bbl)->description->cond_registers);
    return use;
  }

#ifdef ARM_SUPPORT
  use = RegsetNew();
  use_eq = RegsetNew();
  use_ne = RegsetNew();
  BBL_FOREACH_INS_R(bbl,i_ins)
    {
      if (ArmInsUpdatesCond(i_ins))
	{
	  RegsetSetUnion(use,RegsetNewUnion(use_eq,use_ne));
	  use_eq = RegsetNew();
	  use_ne = RegsetNew();
	}
      

      if (INS_IS_CONDITIONAL(i_ins))
	{
	  if (ARMINSCONDITION(i_ins)==ARM_CONDITION_EQ)
	    {
	      RegsetSetDiff(use_eq,INS_REGS_DEF(i_ins));
	      RegsetSetUnion(use_eq,RegsetNewDiff(INS_REGS_USE(i_ins),bbl->fun->cfg->sec->obj->description->cond_registers));
	      RegsetSetUnion(use,RegsetNewIntersect(INS_REGS_USE(i_ins),bbl->fun->cfg->sec->obj->description->cond_registers));
	    }
	  else if (ARMINSCONDITION(i_ins)==ARM_CONDITION_NE)
	    {
	      RegsetSetDiff(use_ne,INS_REGS_DEF(i_ins));
	      RegsetSetUnion(use_ne,RegsetNewDiff(INS_REGS_USE(i_ins),bbl->fun->cfg->sec->obj->description->cond_registers));
	      RegsetSetUnion(use,RegsetNewIntersect(INS_REGS_USE(i_ins),bbl->fun->cfg->sec->obj->description->cond_registers));
	    }
	  else
	    RegsetSetUnion(use,INS_REGS_USE(i_ins)); 
	}
      else
	{
	  RegsetSetDiff(use,INS_REGS_DEF(i_ins));
	  RegsetSetUnion(use,INS_REGS_USE(i_ins));
	}
    }
  
  return use;


#else
  use = RegsetNew();

  BBL_FOREACH_INS_R(bbl,i_ins)
  {
    if (!INS_IS_CONDITIONAL(i_ins))
      RegsetSetDiff(use,INS_REGS_DEF(i_ins));
    RegsetSetUnion(use,INS_REGS_USE(i_ins));
  }
  return use;
#endif

}


t_regset BblRegsUse2(t_bbl * bbl)
{
  t_ins * i_ins;
#ifdef ARM_SUPPORT
  t_regset use_eq, use_ne;
#endif
  t_regset use;

  t_function * fun = BBL_FUN(bbl);

  /*if((!fun) || (IsHellNode(bbl)) || (FUN_IS_HELL(fun)))
    return AllRegs(); */
  /* LUDO */
  if(!fun || bbl == bbl->cfg->hell_node || bbl == bbl->cfg->exit_hell_node || bbl == bbl->cfg->unique_entry_node || bbl == bbl->cfg->unique_exit_node)
    return AllRegs();

  if(bbl == bbl->cfg->call_hell_node || bbl ==  bbl->cfg->exit_call_hell_node)
  {
    use = RegsetNewInvers(BBL_OBJECT(bbl)->description->callee_saved);
    RegsetSetDiff(use,BBL_OBJECT(bbl)->description->cond_registers);
    return use;
  }
#ifdef ARM_SUPPORT
  use = RegsetNew();
  use_eq = RegsetNew();
  use_ne = RegsetNew();

  BBL_FOREACH_INS_R(bbl,i_ins)
    {
      if (ArmInsUpdatesCond(i_ins))
	{
	  RegsetSetUnion(use,RegsetNewUnion(use_eq,use_ne));
	  use_eq = RegsetNew();
	  use_ne = RegsetNew();
	}
      
      if (INS_IS_CONDITIONAL(i_ins))
	{
	  if (ARMINSCONDITION(i_ins)==ARM_CONDITION_EQ)
	    {
	      RegsetSetDiff(use_eq,INS_REGS_DEF(i_ins));
	      RegsetSetUnion(use_eq,RegsetNewDiff(INS_REGS_USE(i_ins),bbl->fun->cfg->sec->obj->description->cond_registers));
	      RegsetSetUnion(use,RegsetNewIntersect(INS_REGS_USE(i_ins),bbl->fun->cfg->sec->obj->description->cond_registers));
	    }
	  else if (ARMINSCONDITION(i_ins)==ARM_CONDITION_NE)
	    {
	      RegsetSetDiff(use_ne,INS_REGS_DEF(i_ins));
	      RegsetSetUnion(use_ne,RegsetNewDiff(INS_REGS_USE(i_ins),bbl->fun->cfg->sec->obj->description->cond_registers));
	      RegsetSetUnion(use,RegsetNewIntersect(INS_REGS_USE(i_ins),bbl->fun->cfg->sec->obj->description->cond_registers));
	    }
	  else
	    RegsetSetUnion(use,INS_REGS_USE(i_ins)); 
	}
      else
	{
	  RegsetSetDiff(use,INS_REGS_DEF(i_ins));
	  RegsetSetUnion(use,INS_REGS_USE(i_ins));
	}
      DiabloPrintArch(stdout,bbl->fun->cfg->sec->obj->description,"@I\nall: @A\n eq: @A\n ne:@A\n",i_ins,use,use_eq,use_ne);

    }
  
  return use;


#else
  use = RegsetNew();

  BBL_FOREACH_INS_R(bbl,i_ins)
  {
    if (!INS_IS_CONDITIONAL(i_ins))
      RegsetSetDiff(use,INS_REGS_DEF(i_ins));
    RegsetSetUnion(use,INS_REGS_USE(i_ins));
  }
  return use;
#endif

}



/* }}} */
/* {{{ registers live before a bbl */
t_regset BblRegsLiveBefore(t_bbl * bbl)
{
  return BBL_INS_FIRST(bbl) ? InsRegsLiveBefore(BBL_INS_FIRST(bbl)) : BBL_REGS_LIVE_OUT(bbl);
}
/* }}} */
/* {{{ registers live after a bbl */
t_regset BblRegsLiveAfter(t_bbl * bbl)
{
  return BBL_REGS_LIVE_OUT(bbl);
}
/* }}} */
/* Set adresses {{{ */
/*! Set address for bbl and all instructions in the block.
 *  also set new size (not nins) */
void BblSetAddressSuper(t_bbl * bbl,t_address address)
{
  t_address run=address;

  t_ins * ins;

  BBL_CADDRESS(bbl)=address;

  BBL_FOREACH_INS(bbl,ins)
  {
    INS_CADDRESS(ins)=run;
    run=GENERIC_ADD(run,INS_CSIZE(ins));
  }
  BBL_CSIZE(bbl)=GENERIC_SUB(run,address);
}

void BblSetAllAddressSuper(t_bbl * bbl,t_address old, t_address new)
{
  t_address run_old=old;
  t_address run_new=new;
  t_ins * ins;

  BBL_CADDRESS(bbl)=run_new;
  BBL_MIN_ADDRESS(bbl)=run_old;

  BBL_FOREACH_INS(bbl,ins)
  {
    INS_CADDRESS(ins)=run_new;
    INS_MIN_ADDRESS(ins)=run_old;
    run_new=GENERIC_ADD(run_new,INS_CSIZE(ins));
    run_old=GENERIC_ADD(run_old,INS_MIN_SIZE(ins));
  }
  BBL_CSIZE(bbl)=GENERIC_SUB(run_new,new);
  BBL_MIN_SIZE(bbl)=GENERIC_SUB(run_old,old);
}
/* }}} */
/* Swap instructions in a bbl {{{ */
void InsSwap(t_ins * first_ins, t_ins * sec_ins)
{
  t_ins * prev_ins;
  t_ins * next_ins;

  if(INS_BBL(first_ins)!=INS_BBL(sec_ins))
    FATAL(("Not allowed to change place of 2 instructions from a differt BBL"));

  if (INS_PREV(first_ins))
    prev_ins=INS_PREV(first_ins);
  else
    prev_ins=NULL;

  if (INS_NEXT(sec_ins))
    next_ins=INS_NEXT(sec_ins);
  else
    next_ins=NULL;

  if(prev_ins)
    INS_NEXT(prev_ins)=sec_ins;
  if(next_ins)
    INS_PREV(next_ins)=first_ins;

  INS_PREV(sec_ins)=prev_ins;
  INS_NEXT(sec_ins)=first_ins;
  INS_PREV(first_ins)=sec_ins;
  INS_NEXT(first_ins)=next_ins;

  if (BBL_INS_FIRST(INS_BBL(first_ins))==first_ins)
    BBL_INS_FIRST(INS_BBL(first_ins))=sec_ins;

  if (BBL_INS_LAST(INS_BBL(sec_ins))==sec_ins)
    BBL_INS_LAST(INS_BBL(sec_ins))=first_ins;

}
/* }}} */
/* Insert a bbl in a function {{{ */
void realBblInsertInFunction(char * file, int lnno, t_bbl * bbl, t_function * fun)
{
  if (bbl->prev_in_fun) FATAL(("Reinserting already inserted bbl (FUNC=%d addr=@G prev=set)! at %s %d",bbl->fun,BBL_CADDRESS(bbl),file,lnno));
  if (bbl->next_in_fun) FATAL(("Reinserting already inserted bbl (FUNC=%d addr=@G next=set)! at %s %d",bbl->fun,BBL_CADDRESS(bbl),file,lnno));

  BBL_FUN(bbl)=fun;
  BBL_CFG(bbl)=FUN_CFG(fun);

  if (bbl==fun->bbl_first)
  {
    bbl->prev_in_fun=NULL;
    bbl->next_in_fun=NULL;
  }
  else
  {
    if (!fun->bbl_first) FATAL(("You cannot insert bbl's in a function without an entry point (fun name=%s)!",fun->name));
    else if (bbl==fun->bbl_last)
    {
      fun->bbl_first->next_in_fun=bbl;
      bbl->prev_in_fun=fun->bbl_first;
      bbl->next_in_fun=NULL;
    }
    else if (!fun->bbl_last) FATAL(("Exit blocks should be create before the insertion of a non-entry block!"));
    else
    {
      /* insert a normal basic block */
      fun->bbl_last->prev_in_fun->next_in_fun=bbl;
      bbl->prev_in_fun=fun->bbl_last->prev_in_fun;
      bbl->next_in_fun=fun->bbl_last;
      fun->bbl_last->prev_in_fun=bbl;
    }
  }
}
/* }}} */
/*{{{ Get a successor edge */
/*! Get the first successor edge of type contained in mask 'types', if
 * there is none, return NULL */
t_cfg_edge * BblGetSuccEdgeOfType(t_bbl * bbl, t_uint32 types)
{
  t_cfg_edge *ret = NULL;
  BBL_FOREACH_SUCC_EDGE(bbl,ret)
    if(EDGE_CAT(ret) & types) break;
  return ret;
}/*}}}*/
t_cfg_edge * BblGetPredEdgeOfType(t_bbl * bbl, t_uint32 types)
{
  t_cfg_edge * ret = NULL;
  BBL_FOREACH_PRED_EDGE(bbl,ret)
    if(EDGE_CAT(ret) & types) break;
  return ret;
}
/* Move an instruction in a bbl before another instruction. DANGER! UNTESTED! *//*{{{*/
void BblMoveInstructionBefore(t_ins * move_ins, t_ins * dest_ins)
{
  t_ins * prev_ins = INS_PREV(move_ins);
  t_ins * next_ins = INS_NEXT(move_ins);
  if(INS_BBL(move_ins) != INS_BBL(dest_ins)) FATAL(("Only use this function for instructions in the same bbl!"));

  /* decouple move_ins from dll */
  if(prev_ins)
    INS_NEXT(prev_ins) = next_ins;
  else
    FATAL(("What a stupid thing to do, the move_ins is already the first instruction!"));
  if(next_ins)
    INS_PREV(next_ins) = prev_ins;
  else
    BBL_INS_LAST(INS_BBL(move_ins)) = prev_ins;

  /* Insert it before dest_ins */
  INS_PREV(move_ins) = INS_PREV(dest_ins);
  INS_NEXT(move_ins) = dest_ins;
  if(INS_PREV(move_ins)) INS_NEXT(INS_PREV(move_ins)) = move_ins;
  else BBL_INS_FIRST(INS_BBL(move_ins)) = move_ins;
  INS_PREV(dest_ins) = move_ins;
  
}/*}}}*/
/* Move an instruction in a bbl after another instruction  DANGER! UNTESTED! *//*{{{*/
void BblMoveInstructionAfter(t_ins * move_ins, t_ins * dest_ins)
{
  t_ins * prev_ins = INS_PREV(move_ins);
  t_ins * next_ins = INS_NEXT(move_ins);
  if(INS_BBL(move_ins) != INS_BBL(dest_ins)) FATAL(("Only use this function for instructions in the same bbl!"));

  /* decouple move_ins from dll */

  if (prev_ins)
    INS_NEXT(prev_ins) = next_ins;
  else
    FATAL(("What a stupid thing to do, the move_ins is already the first instruction!"));
  if(next_ins)
    INS_PREV(next_ins) = prev_ins;
  else
    BBL_INS_LAST(INS_BBL(move_ins)) = prev_ins;

  /* Insert it after dest_ins */
  INS_PREV(move_ins) = dest_ins;
  INS_NEXT(move_ins) = INS_NEXT(dest_ins);
  if(INS_NEXT(move_ins)) INS_PREV(INS_NEXT(move_ins)) = move_ins;
  else BBL_INS_LAST(INS_BBL(move_ins)) = move_ins;
  INS_NEXT(dest_ins) = move_ins;

}
/* {{{ Split a bbl before or after an instruction */
t_bbl * BblSplitBlock(t_bbl * orig_bbl, t_ins * where, t_bool before)
{
  t_cfg * cfg=BBL_CFG(orig_bbl);
  t_ins * i_ins;
  t_cfg_edge * i_edge;
  t_bbl * split_off;
  t_cfg_exits * ext;
  t_regset liveout2;

  if (before)
    liveout2 = InsRegsLiveBefore(where);
  else
    liveout2 = InsRegsLiveAfter(where);

  split_off=BblNew(orig_bbl->cfg);
  BBL_REGS_LIVE_OUT(split_off) = BBL_REGS_LIVE_OUT(orig_bbl);
  BBL_REGS_LIVE_OUT(orig_bbl) = liveout2;

  if ((!before)&&(where==BBL_INS_LAST(orig_bbl))) FATAL(("Implement splitting after the last ins in a block @iB",orig_bbl));

  BBL_INS_LAST(split_off)=BBL_INS_LAST(orig_bbl);

  if (before)
    BBL_INS_FIRST(split_off)=where;
  else
  {
    VERIFY_FASSERT(INS_NEXT(where),("Bbl Corrupt!"));
    BBL_INS_FIRST(split_off)=INS_NEXT(where);
  }

  BBL_OLD_ADDRESS(split_off)=INS_OLD_ADDRESS(BBL_INS_FIRST(split_off));

  if (INS_PREV(BBL_INS_FIRST(split_off)))
  {
    BBL_INS_LAST(orig_bbl)=INS_PREV(BBL_INS_FIRST(split_off));
    INS_NEXT(INS_PREV(BBL_INS_FIRST(split_off)))=NULL;
    INS_PREV(BBL_INS_FIRST(split_off))=NULL;
  }
  else
  {
    BBL_INS_LAST(orig_bbl)=BBL_INS_FIRST(orig_bbl)=NULL;
  }

  BBL_CSIZE(split_off) = NULL_GENERIC;

  BBL_FOREACH_INS(split_off,i_ins)
  {
    INS_BBL(i_ins)=split_off;
    BBL_NINS(split_off)++;
    BBL_NINS(orig_bbl)--;
    BBL_CSIZE(orig_bbl) = GENERIC_SUB(BBL_CSIZE(orig_bbl),INS_CSIZE(i_ins));
    BBL_CSIZE(split_off) = GENERIC_ADD(BBL_CSIZE(split_off),INS_CSIZE(i_ins));
  }

  BBL_SUCC_FIRST(split_off) = BBL_SUCC_FIRST(orig_bbl);
  BBL_SUCC_LAST(split_off) = BBL_SUCC_LAST(orig_bbl);
  BBL_FOREACH_SUCC_EDGE(split_off,i_edge)
    CFG_EDGE_HEAD(i_edge) = split_off;
  BBL_SUCC_FIRST(orig_bbl)=NULL;
  BBL_SUCC_LAST(orig_bbl)=NULL;

  if(!IS_DATABBL(split_off) && BBL_FUN(orig_bbl))
     BblInsertInFunction(split_off,BBL_FUN(orig_bbl));

  EDGE_EXEC_COUNT(CfgEdgeNew(cfg,orig_bbl,split_off,ET_FALLTHROUGH))=BBL_EXEC_COUNT(orig_bbl);
  BBL_EXEC_COUNT(split_off)=BBL_EXEC_COUNT(orig_bbl);


  for (ext=cfg->exitblocks; ext!=NULL; ext=ext->next)
  {
    if (ext->exit_bbl==orig_bbl)
    {
      t_ins * ins;
      t_bool found=FALSE;
      BBL_FOREACH_INS(orig_bbl,ins)
      {
	if (ins==ext->exit_ins)
	{
	  found=TRUE;
	  break;
	}
      }
      if (found) break;
      ext->exit_bbl=split_off;
    }
  }

  BBL_NEXT_IN_CHAIN(split_off) = BBL_NEXT_IN_CHAIN(orig_bbl);
  BBL_NEXT_IN_CHAIN(orig_bbl) = split_off;
  if(BBL_NEXT_IN_CHAIN(split_off))
    BBL_PREV_IN_CHAIN(BBL_NEXT_IN_CHAIN(split_off)) = split_off;
  BBL_PREV_IN_CHAIN(split_off) = orig_bbl;

  return split_off;
}/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker: *//*}}}*/
