/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * All functions to manipulate flowgraphs
 *
 * Written by Bjorn De Sutter (bdebus@elis.ugent.be)
 *
 * Copyright (c) 2002 Bjorn De Sutter
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.ugent.be) }}} 
 *****************************************************************************/

/*Includes {{{ */
#include <string.h>
#include "../config.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include "../datastructures/data_string.h"
#include "../datastructures/data_stack.h"
#include "../datastructures/data_ll.h"
#include "../diablo_options.h"
#include "diablo_callgraph.h"
#include "diablo_function.h"
#include <string.h>
/*}}}*/

t_cg * CgNew()
{
  t_cg * ret=Malloc(sizeof(t_cg));
  GraphInit(T_GRAPH(ret),sizeof(t_cg_node), sizeof(t_cg_edge));
  return ret;
}

void CgFree(t_cg * cg)
{
  t_node * node, * node_next; 
  t_graph * fg= T_GRAPH(cg);
  t_edge * edge, * edge_next;

  for (node=fg->node_first; node!=NULL; node=node_next)
  {
    node_next=node->next;
    Free(node);
  }
  for (edge=fg->edge_first; edge!=NULL; edge=edge_next) {
    edge_next = edge->next;
    Free(edge);
  };

  Free(cg);
}

static t_cg_node * CgNewNode(t_cg * cg, t_function * fun)
{
  t_cg_node * temp = Malloc(sizeof(t_cg_node));
  //t_cg_node * temp = Calloc(1,sizeof(t_cg_node));
  temp->fun=fun;
  GraphInitNode(T_GRAPH(cg),T_NODE(temp));
  return temp;
}

static t_cg_edge * CgAppendNode(t_cg * cg, t_cg_node * after, t_cg_node * insert, t_uint32 edge_category) 
{ 
  return T_CG_EDGE(GraphAppendNode(T_GRAPH(cg), T_NODE(after), T_NODE(insert), edge_category)); 
}

static t_cg_edge * CgNewEdge(t_cg * cg, t_cg_node *head, t_cg_node *tail, t_cfg_edge * cfg_edge, t_uint32 cat)
{
  t_cg_edge * edge = CgAppendNode(cg, head, tail, cat);
  edge->cfg_edge = cfg_edge;
  return edge;
}

static t_cg_node * CgFindOrCreateNode(t_cg * cg, t_function * fun)
{
  t_cg_node * node;
  CG_FOREACH_NODE(cg,node)
    if (node->fun == fun)
      return node;
  
  return CgNewNode(cg,fun);
}

static t_cg_edge * ExistsEdgeBetween(t_cg_node * head,t_cg_node * tail, t_uint32 cat)
{
  t_edge * edge;
  
  NODE_FOREACH_SUCC_EDGE(head,edge)
    if (CG_EDGE_TAIL(edge)==tail && EDGE_CAT(edge)==cat)
      return edge;

  return NULL;
    
}

void CgAddCfgEdge(t_cg * cg, t_cfg_edge * edge)
{
  t_cg_node * head;
  t_cg_node * tail;
  t_cg_edge * cg_edge;
  data_ll_elem * tmp_ll;

  if (EDGE_CAT(edge)==ET_SWI) return;

  head = CgFindOrCreateNode(cg,BBL_FUN(CFG_EDGE_HEAD(edge)));
  tail = CgFindOrCreateNode(cg,BBL_FUN(CFG_EDGE_TAIL(edge)));


  cg_edge=ExistsEdgeBetween(head,tail,EDGE_CAT(edge));
  if (!cg_edge) {
    cg_edge=CgNewEdge(cg,head,tail,edge,EDGE_CAT(edge));
    cg_edge->call_sites=Calloc(1,sizeof(data_ll_elem));
    cg_edge->call_sites->data=BBL_OLD_ADDRESS(CFG_EDGE_HEAD(edge));
  } else {
    tmp_ll=Calloc(1,sizeof(data_ll_elem));
    tmp_ll->data=BBL_OLD_ADDRESS(CFG_EDGE_HEAD(edge));
    LLAppend(NULL,tmp_ll, & cg_edge->call_sites);
  }
    
}

t_cg_node * CgFindNode(t_cg * cg, t_function * fun)
{
  t_cg_node * node;
  CG_FOREACH_NODE(cg,node)
    if (node->fun == fun)
      return node;
  
  return NULL;
}

void CgBuild(t_cfg * cfg)
{
  t_cfg_edge * edge;

  if (cfg->cg)
    CgFree(cfg->cg);

  cfg->cg = CgNew();

  CFG_FOREACH_EDGE(cfg,edge)
    if (CfgEdgeIsForwardInterproc(edge))
      CgAddCfgEdge(cfg->cg,edge);
}

void XZ_CgExport(t_cg * cg)
{
  t_cg_node * node;
  t_cg_edge * edge;
  t_function * fun, * callee;
  data_ll_elem * lle;
  FILE * out = fopen("cd/callgraph","w");
  VERIFY_FASSERT(out,("Could not open callgraph for writing!"));
  //DiabloPrint(out, "digraph \"callgraph\" {\n\tnode [shape=box]\n");

  CG_FOREACH_NODE_FUN(cg,node,fun)
    {
      //fprintf(out, "%s", fun->name);
      //DiabloPrint(out,"\"]\n");

      FUN_FOREACH_CALLEE(fun,edge,callee) {
	lle=edge->call_sites;
        while (lle) {
	    fprintf(out,"%x,%s %s\n",lle->data,fun->name, callee->name);
            lle=LLNext(lle);
        }
     }
	//fprintf(out, "\t\"%p\" -> \"%p\" [style=%s];\n",fun,callee, EDGE_CAT(edge) == ET_CALL? "solid" : "dashed");

    }

  //DiabloPrint(out, "}\n");
  fclose(out);
}
void CgExport(t_cg * cg, const char * filename)
{
  t_cg_node * node;
  t_cg_edge * edge;
  t_function * fun, * callee;
  FILE * out = fopen(filename,"w");
  VERIFY_FASSERT(out,("Could not open %s for writing!",filename));
  DiabloPrint(out, "digraph \"callgraph\" {\n\tnode [shape=box]\n");

  CG_FOREACH_NODE_FUN(cg,node,fun)
    {
      DiabloPrint(out, "\t\"%p\" [label=\"",fun);
      DiabloPrint(out, "@G\\n", BBL_CADDRESS(FUN_BBL_FIRST(fun)));
      fprintf(out, "%s", fun->name);
      DiabloPrint(out,"\"]\n");

      FUN_FOREACH_CALLEE(fun,edge,callee)
	fprintf(out, "\t\"%p\" -> \"%p\" [style=%s];\n",fun,callee, EDGE_CAT(edge) == ET_CALL? "solid" : "dashed");

    }

  DiabloPrint(out, "}\n");
  fclose(out);
}
