/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "diablo_code_layout.h"

extern t_address AssignAddressesInChain(t_bbl * bbl, t_address start)
{
  for (; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    {
      BblSetAddressSuper(bbl,start);
      start = GENERIC_ADD(start,BBL_CSIZE(bbl));
    }
  return start;
}

/* {{{ create chains of basic blocks */
#define ET_MUSTCHAIN  	ET_FALLTHROUGH | ET_IPFALLTHRU | ET_FUNLINK | ET_IPFUNLINK | ET_SWILINK | ET_IPSWILINK
/*TODO: SWITCH TABLES!!!*/
void CreateChains(t_cfg * cfg, t_chain_holder * ch)
{
  t_bbl * i_bbl;
  t_uint32 nbbls = 0, nchains = 0;
  int i;
  
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      BBL_FIRST_IN_CHAIN(i_bbl) = BBL_LAST_IN_CHAIN(i_bbl) = NULL;
      BBL_NEXT_IN_CHAIN(i_bbl) = BBL_PREV_IN_CHAIN(i_bbl) = NULL;
      /* count all bbls except return blocks and hell nodes (those will not be chained) */
      if (!IsHellNode(i_bbl) && (!BBL_FUN(i_bbl) || (FunctionGetExitBlock(BBL_FUN(i_bbl)) != i_bbl)))
	nbbls++;
    }
  nchains = nbbls;
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      t_cfg_edge * i_edge;
      
      /* skip return blocks and hell nodes */
      if (IsHellNode(i_bbl) || (BBL_FUN(i_bbl) && (FunctionGetExitBlock(BBL_FUN(i_bbl)) == i_bbl)))
	continue;
      
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	{
	  if (EdgeTestCategoryOr(i_edge,ET_MUSTCHAIN))
	    break;
	}
      if (i_edge)
	{
	  t_bbl * tail = CFG_EDGE_TAIL(i_edge);
	  BBL_PREV_IN_CHAIN(tail) = i_bbl;
	  BBL_NEXT_IN_CHAIN(i_bbl) = tail;
	  nchains--; /* each time two blocks get chained, the number of chains decreases by one */
	}
    }
  
  ch->chains = Malloc(nchains*sizeof(t_bbl *));
  ch->nchains = nchains;
  
  i = 0;
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      t_bbl * j_bbl, * prev=NULL;
      
      /* skip return blocks and hell nodes */
      if (IsHellNode(i_bbl) || (BBL_FUN(i_bbl) && (FunctionGetExitBlock(BBL_FUN(i_bbl)) == i_bbl)))
	continue;
      
      if (BBL_PREV_IN_CHAIN(i_bbl)) continue; /* we're looking for the heads of the chains here */
      
      ch->chains[i++] = i_bbl;
      for (j_bbl = i_bbl; j_bbl; prev = j_bbl, j_bbl = BBL_NEXT_IN_CHAIN(j_bbl))
	{
	  BBL_FIRST_IN_CHAIN(j_bbl) = i_bbl;
	}
      for (j_bbl = prev; j_bbl; j_bbl = BBL_PREV_IN_CHAIN(j_bbl))
	{
	  BBL_LAST_IN_CHAIN(j_bbl) = prev;
	}
    }
}
/* }}} */

/* {{{ determine the layout order for a set of layout chains
 * after the final order has been determined, all chains
 * are concatenated to ch->chains[0] */

/* {{{ helper function for qsort */
int __helper_order_chains(const void *a, const void *b)
{
  t_bbl * blocka = *((t_bbl **)a);
  t_bbl * blockb = *((t_bbl **)b);
  if (GENERIC_LT(BBL_OLD_ADDRESS(blocka),BBL_OLD_ADDRESS(blockb)))
    return -1;
  else if (GENERIC_GT(BBL_OLD_ADDRESS(blocka),BBL_OLD_ADDRESS(blockb)))
    return 1;
  
  return 0;
}
/* }}} */


void DefaultOrderChains(t_chain_holder * ch)
{
  t_bbl * bbl, * prev;
  int i;

  /* for now, we limit ourselves to the simplest of all orderings:
   * order according to original address of the chain heads */
  qsort(ch->chains, (size_t)ch->nchains, sizeof(t_bbl *), __helper_order_chains);

  prev = NULL;
  for (i=0; i<ch->nchains; i++)
  {
    for (bbl = ch->chains[i]; bbl; bbl = BBL_NEXT_IN_CHAIN(bbl))
    {
      BBL_FIRST_IN_CHAIN(bbl) = ch->chains[0];
      BBL_LAST_IN_CHAIN(bbl) = BBL_LAST_IN_CHAIN(ch->chains[ch->nchains-1]);
      if (!BBL_PREV_IN_CHAIN(bbl) && prev)
      {
	BBL_PREV_IN_CHAIN(bbl) = prev;
	BBL_NEXT_IN_CHAIN(prev) = bbl;
      }
      prev = bbl;
    }
  }
} /* }}} */

/* Comment out by Bruno, does not compile for all arches (G_T_UINT64 is not supported on 32 bit arches */
void DefaultGiveAddressesToChain(t_address start_address, t_chain_holder * ch)
{
  t_bbl* bbl = ch->chains[0];
  t_address address = start_address;
  int nr_ins = 0;

  t_uint32 width_min = BBL_CFG(bbl)->sec->obj->description->minimal_encoded_instruction_size;
  t_uint32 width_max = BBL_CFG(bbl)->sec->obj->description->maximum_encoded_instruction_size;
  t_uint32 width_mod = BBL_CFG(bbl)->sec->obj->description->encoded_instruction_mod_size;
  
  if (width_min!=width_max)
    FATAL(("You should not use the default address assigner of the default code layouter on architectures that have no fixed instruction width!!! (or this should be implemented first)\n"));

  if (width_min!=width_mod)
    FATAL(("You should not use the default address assigner of the default code layouter on architectures that have no fixed instruction width!!! (or this should be implemented first)\n"));

  while (bbl)
    {
      /*      DiabloPrint(stdout,"new BBL: @iB\n",bbl);*/
      nr_ins+=BBL_NINS(bbl);
      /*      printf("old: %lx new %lx\n",G_T_UINT64(BBL_OLD_ADDRESS(bbl)),address);*/
      BblSetAddressSuper(bbl,address);
      address=GENERIC_ADD_UINT32(address,BBL_NINS(bbl)*(width_mod/8));
      bbl=BBL_NEXT_IN_CHAIN(bbl);
    }

  /*  printf("NR_INS: %d\n",nr_ins);*/

}

typedef t_bbl* t_bbl_p;

int compar_original_order(const void *elem1, const void* elem2)
{
 const t_bbl_p* bbl1=elem1;
 const t_bbl_p* bbl2=elem2;

  if (GENERIC_EQ(BBL_OLD_ADDRESS(*bbl1),BBL_OLD_ADDRESS(*bbl2)))
    return 0;
  if (GENERIC_LT(BBL_OLD_ADDRESS(*bbl1),BBL_OLD_ADDRESS(*bbl2)))
    return -1;
  return +1;
}

t_uint32 CfgLayoutOriginalOrderAndAlignment(t_cfg * cfg)
{
#if 0
  t_uint32 nbbls = CFG_NBBLS(cfg);
  t_bbl** bbls = (t_bbl**) Malloc(nbbls*sizeof(t_bbl*));
  t_bbl** ibbls=bbls;
  t_bbl * bbl;
  t_address address = SECTION_GET_OLD_ADDRESS(cfg->sec);
  t_uint32 nr_ins = 0;
  int i;
  
  CFG_FOREACH_BBL(cfg,bbl)
    {
      *ibbls++=bbl;
    }
  
  qsort(bbls,nbbls,sizeof(t_bbl*),compar_original_order);
  
  for (i=0;i<nbbls;i++)
    {
      bbl = *(bbls+i);
      /*      if (BBL_NINS(bbl)>0) */
      {
	t_uint32 alignment = BblNeedsToBeAligned(bbl);
	if (alignment)
	  {
	    t_uint32 new_address = GENERIC_EXTRACT_UINT32(address);
	    if (new_address%alignment!=0)
	      {
		new_address = (new_address + alignment) & (alignment-1);
		GenericSet_uint32(address,new_address);
	      }
	  }

	nr_ins += BBL_NINS(bbl);
	BblSetAddressSuper(bbl,address);
	address=GENERIC_ADD_UINT32(address,BBL_NINS(bbl));
	/*	  DiabloPrint(stdout,"@iB\n",bbl); */
	fflush(stdout);
      }
    }
  
  printf("nr_ins: %d\n",nr_ins);
  
  Free(bbls);
  printf("FINAL size : %d\n",nr_ins);
  cfg->sec->obj->entry =  BBL_CADDRESS(cfg->entries->entry_bbl);
  return nr_ins*4;
#else
  return 0;
#endif
}
