/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/
/*#define DONT_GO_THROUGH_HELL |+this make the analysis quite a bit faster! don't uncomment!+|*/
/*#define DEBUG_COPY_ANALYSIS  |+ Uncomment for debugging purposes +|*/

#include "diablo_extendedcopy.h"
#include "diablo_action.h"
#include "diablo_constprop.h"
#include "diablo_function.h"
#include "diablo_edge.h"
#include "diablo_bbl.h"
#include "diablo_iterators.h"
#include "diablo_regset.h"
#include "diablo_print.h"
#include "diablo_liveness.h"
#include "../utils/util_error.h"
#include "../utils/util_malloc.h"
#include "../utils/util_status.h"
#ifdef ARM_SUPPORT
#include "../arch/arm/arm_instructions.h"
#include "../arch/arm/arm_registers.h"
#endif
#ifdef I386_SUPPORT
#include "../arch/i386/i386_registers.h"
#endif

t_uint32 max_nr_equations, null_register;
/* this avoids having to allocate a set of equations for each call to BblCopyAnalysis */
static t_equations eqs_for_BblCopyAnalysis = NULL; 

static void FunctionCopyAnalysis(t_function* fun, t_cfg_edge* edge_in, t_bool during_fixpoint_calculations);
static void BblCopyAnalysis(t_bbl * bbl, t_cfg_edge * edge_out, t_bool during_fixpoint_calculations);
void (*CopyInstructionPropagator)(t_ins * ins, t_equations eqs);
void CfgStoreCPInfoInEquations(t_object * obj, t_section * sec);

static t_equations GetAugmentedReturnEqs(t_cfg_edge * ret_edge)
{
  int i;
  t_cfg_edge * call_edge;
  t_function * fun;
  t_equations before_fun;
  t_equations augmented;
  t_regset preserved;

  call_edge = ret_edge->corr_edge;
  if (!call_edge) return ret_edge->eqs;
  fun = BBL_FUN(CFG_EDGE_TAIL(call_edge));
  augmented = EquationsNew();
  before_fun = EquationsNew();
  EquationsCopy(call_edge->eqs,before_fun);
#ifdef I386_SUPPORT
  /* this is an ugly hack, but a more general solution would be very long-winded. */
  /* in the equations that are propagated over the link edge, undo the effect the
   * call instruction had on the stack pointer */
  EquationsAdd(before_fun,I386_REG_ESP,I386_REG_ESP,-4,NULL,NULL);
#endif

  RegsetDup(preserved,FUN_REGS_CHANGED(fun));
  RegsetSetDiff(preserved,FUN_REGS_SAVED(fun));
  RegsetSetInvers(preserved);

  RegsetSetDiff(preserved,BBL_REGS_LIVE_OUT(CFG_EDGE_HEAD(ret_edge->corr_edge)));
  RegsetSetDiff(preserved,BblRegsLiveBefore(CFG_EDGE_TAIL(ret_edge)));

  EquationsCopy(ret_edge->eqs,augmented);

  for (i=0; i<max_nr_equations; i++)
  {
    t_equation tmp;
    if (before_fun[i].rega == REG_TOP || before_fun[i].rega == REG_BOT) continue;
    if (!RegsetIn(preserved,i)) continue;

    tmp = before_fun[i];
    if (RegsetIn(preserved,tmp.rega) && (RegsetIn(preserved,tmp.regb) || tmp.regb == null_register))
      if((!tmp.taga && !tmp.tagb) || (tmp.taga && !tmp.tagb && tmp.regb == null_register))
        EquationsAdd(augmented,tmp.rega,tmp.regb,tmp.constant,tmp.taga,tmp.tagb);

    while (before_fun[tmp.regb].rega != REG_BOT && before_fun[tmp.regb].rega != REG_TOP)
    {
      tmp = EquationSum(tmp,before_fun[tmp.regb]);
      if (RegsetIn(preserved,tmp.rega) && (RegsetIn(preserved,tmp.regb) || tmp.regb == null_register))
	if((!tmp.taga && !tmp.tagb) || (tmp.taga && !tmp.tagb && tmp.regb == null_register))
	  EquationsAdd(augmented,tmp.rega,tmp.regb,tmp.constant,tmp.taga,tmp.tagb);
    }
  }

  /*DiabloPrint(stdout,"call to %s coming from @B returning to @B\n",fun->name,CFG_EDGE_HEAD(call_edge),CFG_EDGE_TAIL(ret_edge));
  printf("--------------------\n");
  PrintRegset(stdout,fun->cfg->sec->obj->description,preserved);
  printf("--------------------\n");
  EquationsPrint(before_fun);
  printf("--------------------\n");
  EquationsPrint(augmented);
  printf("====================\n"); */

  Free(before_fun);
  return augmented;
}

/* {{{ copy analysis */
/*
   \param sec 

   Perform copy-analysis on the flowgraph of section sec.
 */

void CopyAnalysis(t_section * sec)
{
  /* iterators */
  t_function * i_fun;
  t_cfg_edge * i_edge;
  t_bbl * i_bbl;
  /* local variables */
  t_bbl * program_entry;
  t_function * entry_fun;
  t_cfg *cfg = T_CFG(sec->data);
  t_object * obj = sec->obj;
  
  CopyInstructionPropagator = obj->actions->CopyInstructionPropagator;
  if(! CopyInstructionPropagator)
    return;
  max_nr_equations = obj->description->num_int_regs + 1;
  null_register = max_nr_equations-1;

  eqs_for_BblCopyAnalysis = EquationsNew();

  ANALYSIS_START("Copy analysis");

  /* for enhanced preciseness, we need information about the changed registers of each function */
  ComputeSavedChangedRegisters(sec);

  /* {{{ initialization */
  InitEdgeMarking();

  CfgUnmarkAllFun(cfg);

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    FunctionUnmarkAllEdges(i_fun);
    FunctionUnmarkAllBbls(i_fun);
    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      if(!BBL_EQS_IN(i_bbl))
        BBL_EQS_IN(i_bbl) = EquationsNew();
      else 
	FATAL(("Not null?"));
    }
  }

  CFG_FOREACH_EDGE(cfg,i_edge)
  {
    EdgeUnmark(i_edge);
    /* Allocate space for equations on the interprocedural edges */
    if (CfgEdgeIsInterproc(i_edge))
    {
      i_edge->eqs = EquationsNew();
#ifdef DONT_GO_THROUGH_HELL
      if(IsHellNode(CFG_EDGE_TAIL(i_edge))) EquationsSetAllBot(i_edge->eqs);
#endif
    }
  }
  /* }}} */

  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);

  /* 
     Copy analysis is started at the entry point of the
     program. This is not done through incoming edges, therefore the
     edge is null. As we want to propagate the outgoing information
     over outgoing edges, the last parameter is TRUE.
   */

  FunctionCopyAnalysis(entry_fun,NULL,TRUE);

  BBL_FOREACH_PRED_EDGE(program_entry,i_edge)
  {
    /* propagation at the beginning of the program should start with BOT. */
    EquationsSetAllBot(i_edge->eqs);
    EdgeMark(i_edge);
  }
#ifdef DONT_GO_THROUGH_HELL
  FunctionCopyAnalysis(BBL_FUN(cfg->hell_node),NULL,TRUE);
  FunctionCopyAnalysis(BBL_FUN(cfg->call_hell_node),NULL,TRUE);
#endif

  /* 
     Actual interprocedural fixpoint stuff: as long as there are
     marked procedures, propagate information over marked (incoming)
     edges through them 
   */
  while (CfgUnmarkFun(cfg,&i_fun))
    while (FunctionUnmarkEdge(i_fun,&i_edge))
      FunctionCopyAnalysis(i_fun, i_edge,TRUE);

  ANALYSIS_SIMPLE_END("Copy analysis");

  CFG_FOREACH_FUN(cfg,i_fun)
    FunctionUnmarkAllBbls(i_fun);

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    if(i_fun->bbl_first)
      {
	FunctionCopyAnalysis(i_fun,T_CFG_EDGE(i_fun!=entry_fun),FALSE);
#ifdef DEBUG_COPY_ANALYSIS
	DiabloPrint(stdout,"function %s\n",i_fun->name);
	FUN_FOREACH_BBL(i_fun,i_bbl)
	  {
	    if(BBL_EQS_IN(i_bbl))
	      {
		DiabloPrint(stdout,"@B:\n",i_bbl);
		EquationsPrint(BBL_EQS_IN(i_bbl));
		DiabloPrint(stdout,"-- \n");
	      }	    else
	      VERBOSE(0,("No equations for @B!",i_bbl));
	  }
#endif
    }
  }

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    FunctionFreeMarkedSpace(i_fun);
  }
  Free(CFG_MARKED_FUNS(cfg));
  CFG_MARKED_FUNS(cfg) = NULL;
}
/* }}} */
/* Clean up datastructures for Copy Analysis {{{ */
void CopyAnalysisFini(t_section * sec)
{
  t_cfg *cfg = T_CFG(sec->data);
  t_function * i_fun;
  t_bbl * i_bbl;
  t_cfg_edge * i_edge;

  if(!CopyInstructionPropagator)
    return; /* No copy analysis has been done, we don't have to clean up anything */

  CFG_FOREACH_FUN(cfg,i_fun)
    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      if(BBL_EQS_IN(i_bbl))
      {
        Free(BBL_EQS_IN(i_bbl));
	BBL_EQS_IN(i_bbl) = NULL;
      }
      else 
	FATAL(("No eqs anymore?"));
    }
  CFG_FOREACH_EDGE(cfg,i_edge)
  {
    if (CfgEdgeIsInterproc(i_edge))
    {
      Free(i_edge->eqs);
      i_edge->eqs = NULL;
    }
  }
  Free(eqs_for_BblCopyAnalysis);
  eqs_for_BblCopyAnalysis = NULL;
}
/* }}} */

/* {{{  Propagate information through a procedure 
  PARAMETERS: 
  - fun is the function to propagate through
  - edge_in is NULL if this propagation is not for an
    incoming edge, but for the program entry point 
  - during_fixpoint_calculations is false if no outgoing
    information has to be propagated over interprocedural
    edges
 */
static void FunctionCopyAnalysis(t_function * fun, t_cfg_edge * edge_in, t_bool during_fixpoint_calculations)
{
  t_cfg_edge * i_edge;
  t_bbl * i_bbl;
  t_cfg_edge * edge_out = NULL;
  t_cfg *cfg = FUN_CFG(fun);

  /*DiabloPrint(stdout,"CA in function %s\n",fun->name?fun->name:"NULL");
  if (edge_in && during_fixpoint_calculations)
    DiabloPrint(stdout,"incoming edge: @E\n", edge_in); */

  FUN_FOREACH_BBL(fun,i_bbl)
    EquationsSetAllTop(BBL_EQS_IN(i_bbl));

  /* {{{ propagate the necessary information into the procedure */
  if (edge_in)
  {
    if (!during_fixpoint_calculations)
    {
      /* this means we have to propagate information over all incoming edges into the procedure */
      FUN_FOREACH_BBL(fun,i_bbl)
        BBL_FOREACH_PRED_EDGE(i_bbl, i_edge) 
        {
          if (CfgEdgeIsInterproc(i_edge) && EdgeIsMarked(i_edge))
	  {
	    t_reg i;
	    t_bool stop = FALSE;
	    /* join equations from incoming edge only if the are not TOP */
	    for(i=0;i<max_nr_equations;i++)
	    {
	      if(i_edge->eqs[i].rega == REG_TOP) 
	      {
		stop = TRUE;
		break;
	      }
	    }
	    if (!stop)
	    {
	      t_equations real_eqs = (EDGE_CAT(i_edge) == ET_RETURN) ? GetAugmentedReturnEqs(i_edge) : i_edge->eqs;
	      /*DiabloPrint(stdout,"propping interprocedural info from @E\n",i_edge);*/
	      /*EquationsPrint(real_eqs);*/
	      if (EquationsJoin(BBL_EQS_IN(i_bbl),real_eqs))
	      {
		FunctionMarkBbl(fun,i_bbl);
	      }
	      if (real_eqs != i_edge->eqs) Free(real_eqs);
	    }
	  }
        }
    }
    else
    {
      /* 
	 Propagation will start at the tail of the edge, where
	 the incoming information is used (unlike constant propagation,
	 we don't treat hell nodes any different from normal nodes)
       */
      t_equations real_eqs = edge_in->eqs;
      
      if (EDGE_CAT(edge_in) == ET_IPFUNLINK)
      {
	/* this is a very special case: the real information is not stored on the 
	 * funlink edge, but on the associated incoming return edge */
	real_eqs = GetAugmentedReturnEqs(FindReturnForFunLink(edge_in));
      }
      
      if (CFG_EDGE_TAIL(edge_in)!=cfg->call_hell_node && CFG_EDGE_TAIL(edge_in)!=cfg->hell_node)
	EquationsJoin(BBL_EQS_IN(CFG_EDGE_TAIL(edge_in)),real_eqs);
      else
      {
#ifdef DONT_GO_THROUGH_HELL
        FATAL(("Don't want to get here anymore!"));
#endif
	EquationsSetAllBot(BBL_EQS_IN(CFG_EDGE_TAIL(edge_in)));
      }
      FunctionMarkBbl(fun,CFG_EDGE_TAIL(edge_in));
      edge_out = edge_in->corr_edge;

      if (real_eqs != edge_in->eqs) Free(real_eqs);
    }
  }
  else
  {
    /* program entry: no incoming edge */
    if (!during_fixpoint_calculations)
    {
      FUN_FOREACH_BBL(fun,i_bbl)
	BBL_FOREACH_PRED_EDGE(i_bbl, i_edge) 
	{
	  if (CfgEdgeIsForwardInterproc(i_edge))
	    if (EdgeIsMarked(i_edge))
	      if (EquationsJoin(BBL_EQS_IN(i_bbl),i_edge->eqs))
		FunctionMarkBbl(fun,i_bbl);
	}
    }
    FunctionMarkBbl(fun,fun->bbl_first);
    /* propagation of the program entry point always starts with BOT */
    EquationsSetAllBot(BBL_EQS_IN(fun->bbl_first));
  } /* }}} */

  /* the actual fixpoint calculation */
  while (FunctionUnmarkBbl(fun,&i_bbl))
    BblCopyAnalysis(i_bbl,edge_out,during_fixpoint_calculations);
}
/* }}} */

/* {{{ propagate copy propagation information over an edge */
static void PropOverEdge(t_cfg_edge * edge, t_equations eqs, t_bool during_fixpoint_calculations)
{
  t_cfg_edge * corr_edge;
#ifdef DEBUG_COPY_ANALYSIS
  t_reg i;

  /* {{{ */
  for(i=0;i<max_nr_equations-1;i++)
  {
    if(eqs[i].rega == REG_TOP) FATAL(("Propping top over edge @E\n",edge));
    if(eqs[i].regb == REG_TOP) FATAL(("Propping top over edge @E\n",edge));
  } /* }}} */
#endif

  /*DiabloPrint(stdout,"propping over edge @E\n",edge);*/
  /*EquationsPrint(eqs);*/
  /* mark it as taken */
  EdgeMark(edge);

  if (CfgEdgeIsInterproc(edge))
  {
    /* if it is a true outgoing edge, the corresponding incoming edge must have its 
       information propagated into the procedure (if it is marked that is). */

    t_reg reg_tmp;
    
    if (global_options.rely_on_calling_conventions)
      {
	t_regset invalid_regs = RegsetNew();
	RegsetDup(invalid_regs,BBL_REGS_LIVE_OUT(CFG_EDGE_HEAD(edge)));
	RegsetSetUnion(invalid_regs,BblRegsLiveBefore(CFG_EDGE_TAIL(edge)));
	
	RegsetSetIntersect(invalid_regs,edge->cfg->sec->obj->description->int_registers);
	
	REGSET_FOREACH_REG(invalid_regs,reg_tmp)
	  EquationsInvalidate(eqs,reg_tmp);
      }

    if (CfgEdgeIsForwardInterproc(edge))
    {
      corr_edge = edge->corr_edge;
      if (corr_edge && EdgeIsMarked(corr_edge))
      {
	if (CFG_EDGE_TAIL(corr_edge)->fun==CFG_EDGE_HEAD(edge)->fun && CFG_EDGE_HEAD(edge)!=CFG_EDGE_TAIL(corr_edge))
	{
	  /*DiabloPrint(stdout,"propping over corr_edge @E\n",corr_edge);*/
	  t_equations real_eqs = (EDGE_CAT(corr_edge) == ET_RETURN) ? GetAugmentedReturnEqs(corr_edge) : corr_edge->eqs;
	  /*EquationsPrint(real_eqs);*/
	  if (EquationsJoin(BBL_EQS_IN(CFG_EDGE_TAIL(corr_edge)),real_eqs))
	    FunctionMarkBbl(CFG_EDGE_HEAD(edge)->fun,CFG_EDGE_TAIL(corr_edge));
	  if (real_eqs != corr_edge->eqs) Free(real_eqs);
	} 
      }
    }
    /* for a call with an interprocedural funlink, propagate the corresponding edge of the ipfunlink as well */
    if(EDGE_CAT(edge) == ET_CALL)
    {
      t_cfg_edge * link_edge=FindFunLinkForCall(edge);
      if(link_edge && EDGE_CAT(link_edge) == ET_IPFUNLINK)
      {
	t_cfg_edge * c_edge = link_edge->corr_edge;
	if(c_edge && EdgeIsMarked(c_edge))
	  if(EquationsJoin(BBL_EQS_IN(CFG_EDGE_TAIL(c_edge)),c_edge->eqs))
	    FunctionMarkBbl(CFG_EDGE_HEAD(link_edge)->fun,CFG_EDGE_TAIL(c_edge));
      } 
    }
    

    if (during_fixpoint_calculations)
    {
#ifdef DONT_GO_THROUGH_HELL
      if (!IsHellNode(CFG_EDGE_TAIL(edge)) && EquationsJoin(edge->eqs,eqs))
#else
      if (EquationsJoin(edge->eqs,eqs))
#endif
      {
	if (!CfgEdgeIsForwardInterproc(edge))
	{
	  t_cfg_edge * link = NULL;
	  if (EDGE_CAT(edge) == ET_RETURN)
	    link = FindFunLinkForReturn(edge);
	  if (link && EDGE_CAT(link) == ET_IPFUNLINK)
	  {
	    /* returns associated with interprocedural funlinks are a special case: they
	     * only need to be propagated into the function for the incoming ipfunlink, not
	     * for any other incoming edges */
	    EdgeMark(link);
	    CfgMarkFun(CFG_EDGE_TAIL(edge)->fun->cfg,CFG_EDGE_TAIL(edge)->fun);
	    FunctionMarkEdge(CFG_EDGE_TAIL(edge)->fun,link);
	  }
	  else
	  {
	    /* repropagation for all true incoming edges of the function this edge points to is necessary */
	    CfgMarkFun(CFG_EDGE_TAIL(edge)->fun->cfg,CFG_EDGE_TAIL(edge)->fun);
	    FunctionMarkAllTrueIncomingEdges(CFG_EDGE_TAIL(edge)->fun);
	  }
	}
	else 
	{
	  /* only this edge will need to be repropagated in the function it points to */
	  CfgMarkFun(CFG_EDGE_TAIL(edge)->fun->cfg,CFG_EDGE_TAIL(edge)->fun);
	  FunctionMarkEdge(CFG_EDGE_TAIL(edge)->fun,edge);
	}
      }
    }
  }
  else 
  {
    /* the edge was INTRAprocedural */

    if (EDGE_CAT(edge)==ET_FUNLINK)
      {
	t_regset invalid_regs = RegsetNew();
	t_reg reg_tmp;
	
	RegsetDup(invalid_regs,BBL_REGS_LIVE_OUT(CFG_EDGE_HEAD(edge)));
	RegsetSetUnion(invalid_regs,BblRegsLiveBefore(CFG_EDGE_TAIL(edge)));
	
	REGSET_FOREACH_REG(invalid_regs,reg_tmp)
	  EquationsInvalidate(eqs,reg_tmp);
      }

    if (EquationsJoin(BBL_EQS_IN(CFG_EDGE_TAIL(edge)),eqs))
      FunctionMarkBbl(CFG_EDGE_TAIL(edge)->fun,CFG_EDGE_TAIL(edge));
  }
}
/* }}} */

/* {{{ Propagate a state over a bbl and to the successor edges */
static void BblCopyAnalysis(t_bbl * bbl, t_cfg_edge * edge_out, t_bool during_fixpoint_calculations)
{
  t_cfg_edge * i_edge;
  t_ins * i_ins;
  t_equations eqs = eqs_for_BblCopyAnalysis;

  EquationsCopy(BBL_EQS_IN(bbl),eqs);
#ifdef DEBUG_COPY_ANALYSIS
  { /* {{{ */
    t_reg i;
    for(i=0;i<max_nr_equations-1;i++)
    {
      if(eqs[i].rega == REG_TOP) FATAL(("Propping top in bbl @B\n",bbl));
      if(eqs[i].regb == REG_TOP) FATAL(("Propping top in bbl @B\n",bbl));
    }
  } /* }}} */
#endif

  /*DiabloPrint(stdout,"___ propping @B\n",bbl);*/
  BBL_FOREACH_INS(bbl,i_ins)
  {
    /*DiabloPrint(stdout,"----------------------------------\n");*/
    /*EquationsPrint(eqs);*/
    /*DiabloPrint(stdout,"CA: @I\n",i_ins);*/
    CopyInstructionPropagator(i_ins,eqs);
    /*EquationsPrint(eqs); */
  }

  /* special hack: for hell nodes, this is necessary to ensure
   * propagation of _all_ the registers on the outgoing edges */
  if (IsHellNode(bbl))
    EquationsSetAllBot(eqs);

#ifdef DEBUG_COPY_ANALYSIS
  { /* {{{ */
    t_reg i;
    for(i=0;i<max_nr_equations-1;i++)
    {
      if(eqs[i].rega == REG_TOP) FATAL(("Propping top over bbl @B\n",bbl));
      if(eqs[i].regb == REG_TOP) FATAL(("Propping top over bbl @B\n",bbl));
    }
  } /* }}} */
#endif

  /* propagate the outgoing information over the outgoing edges */
  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
  {
    t_bool do_it;
    do_it = !CfgEdgeIsInterproc(i_edge);                      /* propagate over intraprocedural edges */
    do_it = do_it || CfgEdgeIsForwardInterproc(i_edge);       /* propagate over forward interprocedural edges */
    do_it = do_it || (i_edge == edge_out || edge_out == NULL);/* propagate over backward interprocedural edges if they correspond with the incoming edge for which
								 the function was propagated, or if there isn't an incoming edge (for the program entry point) */
    do_it = do_it && (EDGE_CAT(i_edge) != ET_FUNLINK);        /* don't propagate information over funlink edges */
    do_it = do_it && (EDGE_CAT(i_edge) != ET_IPFUNLINK);      /* don't propagate information over interprocedural funlink edges */
    if (do_it) PropOverEdge(i_edge, eqs, during_fixpoint_calculations);
  }
}
/* }}} */

/* {{{ BblCopyAnalysisUntilIns */
void BblCopyAnalysisUntilIns(t_ins * ins, t_equations eqs)
{

  /* iterators */

  t_ins * i_ins;
  t_bbl * bbl = INS_BBL(ins);

  if(global_options.copy_analysis)
  {

    if(!bbl) FATAL(("No bbl for this ins @I",ins));
    if(!eqs) FATAL(("No equations!"));
    if(BBL_EQS_IN(bbl))
      EquationsCopy(BBL_EQS_IN(bbl),eqs);
    else
      EquationsSetAllBot(eqs);

    BBL_FOREACH_INS(bbl,i_ins)
    {
      /*      DiabloPrint(stdout,"----------------------------------\n");
            EquationsPrint(eqs);
            DiabloPrint(stdout,"CA: @I\n",i_ins); */
      if(i_ins == ins) break;
      CopyInstructionPropagator(i_ins,eqs);
      /*      EquationsPrint(eqs);*/
    }
  }
  return;
} /* }}} */

#if 0
void CfgStoreCPInfoInEquations(t_object * obj, t_section * sec)
{
  t_bbl * i_bbl;
  t_ins * i_ins;
  t_function * i_fun, * entry_fun;
  t_cfg * cfg = T_CFG(sec->data);
  t_regset tmp_set;
  t_reg i_reg,j_reg;
  register_content c1, c2;
  t_bbl * program_entry;
  
  t_procstate * prev_state;
  t_procstate * next_state;

  t_regset allregs = obj->description->all_registers;
  
  prev_state = ProcStateNew();
  next_state = ProcStateNew();
  
  program_entry = cfg->entries->entry_bbl;
  entry_fun = BBL_FUN(program_entry);

  /* first of all, if the instructions in the graph already contain regstates, remove them */
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    BBL_FOREACH_INS(i_bbl,i_ins)
    {
      //      if (INS_USED_REGSTATE(i_ins)) Free(INS_USED_REGSTATE(i_ins));
      //      if (INS_DEFINED_REGSTATE(i_ins)) Free(INS_DEFINED_REGSTATE(i_ins));
      //      INS_USED_REGSTATE(i_ins) = INS_DEFINED_REGSTATE(i_ins) = NULL;
    }
  }

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    FunctionUnmarkAllBbls(i_fun);
  }

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    if(i_fun->bbl_first)
    {
      FunctionPropagateConstants(obj,i_fun,sec,T_CFG_EDGE(i_fun!=entry_fun),FALSE);

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
	if (!RegsetIsMutualExclusive(((t_procstate*)BBL_PROCSTATE_IN(i_bbl))->top,allregs))
	  continue;
	
	tmp_set = RegsetNewInvers(((t_procstate*)BBL_PROCSTATE_IN(i_bbl))->bot);
	REGSET_FOREACH_REG(tmp_set,i_reg)
	  break;
	
	if(i_reg < max_nr_equations-1)
	{
	  if(ProcStateGetReg(BBL_PROCSTATE_IN(i_bbl),i_reg,&c1)) FATAL(("Not a constant?"));
	  REGSET_FOREACH_REG(tmp_set,j_reg)
	  {
	    if(i_reg == j_reg) continue;
	    if(j_reg >= max_nr_equations) break;
	    if(ProcStateGetReg(BBL_PROCSTATE_IN(i_bbl),j_reg,&c2))
	      FATAL(("Not a constant?"));
            EquationInsert(BBL_EQS_CTE(i_bbl),i_reg,j_reg,c2.i-c1.i,NULL,NULL);
	  }
	  DiabloPrint(stdout,"In @B equations are:\n",i_bbl);
	  EquationsPrint(BBL_EQS_CTE(i_bbl));
	}
      }
      {
	ProcStateFree(BBL_PROCSTATE_IN(i_bbl));
      /*      FUN_FOREACH_BBL(i_fun,i_bbl) // COMMENTED OUT BY BJORN, AS PROCSTATE_OUT IS NO LONGER USED
	BBL_PROCSTATE_IN(i_bbl) = NULL;
	if (i_bbl->procstate_out)
	{
	  ProcStateFree(i_bbl->procstate_out);
	  i_bbl->procstate_out = NULL;
	}*/
	}

    }
  }
  ProcStateFree(prev_state);
  ProcStateFree(next_state);
}
#endif

/* vim: set shiftwidth=2 foldmethod=marker: */
