/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//**********************************************************
 * FILE: diablo_edge.h
 *
 * DESC: Headers for (cfg)edge manipulation 
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) 
 ***********************************************************/


/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup EDGES Control Flow Edges */
/*@{ */
#ifndef DIABLO_EDGE_TYPEDEFS
#define DIABLO_EDGE_TYPEDEFS
typedef struct _t_cfg_edge t_cfg_edge;
#endif


#ifndef DIABLO_EDGE_H
#define DIABLO_EDGE_H
#include "../simple_types.h"
/* edge types (for graphs) */
#include "../datastructures/data_graph.h"
#include "diablo_flowgraph.h"
#include "diablo_procstate.h"
#include "diablo_bbl.h"
#include "diablo_extendedcopy.h"

#define ET_UNKNOWN	(1U<<0)	/**< anything that can't be classified immediately */
#define ET_FALLTHROUGH	(1U<<1)
#define ET_CALL		(1U<<2)	/**< branch and link */
#define ET_RETURN	(1U<<3)	/**< mov pc,lr */
#define ET_FUNLINK	(1U<<4)	/**< connect call and return site of a function */
#define ET_JUMP		(1U<<5)	/**< branch without link */
#define ET_SWI		(1U<<6)	/**< SWI is like a function call but much more predictable */
#define ET_SWILINK	(1U<<7)	/**< links call and return site of the SWI */

#define ET_IPUNKNOWN	(1U<<8)
#define ET_IPFALLTHRU	(1U<<9)
#define ET_IPFUNLINK	(1U<<10)
#define ET_IPJUMP	(1U<<11)
#define ET_IPSWILINK	(1U<<12)
#define ET_COMPENSATING	(1U<<13)
#define ET_SWITCH       (1U<<14)
#define ET_IPSWITCH     (1U<<15)
#define ET_POSTDOM	(1U<<16) /**< special edges that are created just before postdominator analysis, and that HAVE TO BE REMOVED afterwards. They connect infinite loops with the unique exit node, to make postdominator analysis work */

#define EF_FROM_SWITCH_TABLE  (1<<30)

#define ET_INTERPROC	(ET_IPSWITCH | ET_CALL | ET_RETURN | ET_SWI | ET_IPUNKNOWN | ET_IPFALLTHRU | ET_IPFUNLINK | ET_IPJUMP | ET_IPSWILINK | ET_COMPENSATING)
#define ET_FORWARD_INTERPROC (ET_IPSWITCH | ET_CALL | ET_SWI | ET_IPUNKNOWN | ET_IPFALLTHRU | ET_IPFUNLINK | ET_IPJUMP | ET_IPSWILINK)
#define ET_BACKWARD_INTERPROC (ET_INTERPROC &~ ET_FORWARD_INTERPROC)
#define ET_FT_BLOCK	(ET_FALLTHROUGH | ET_UNKNOWN | ET_FUNLINK | ET_SWILINK | ET_IPUNKNOWN | ET_IPFALLTHRU | ET_IPFUNLINK | ET_IPSWILINK | ET_SWITCH | ET_IPSWITCH) /**< class of types which require that there is a fallthrough block */

#define EDGE_IS_FORWARD_INTERPROC(edge) (EDGE_CAT(edge) & ET_FORWARD_INTERPROC)
#define EDGE_IS_BACKWARD_INTERPROC(edge) (EDGE_CAT(edge) & ET_BACKWARD_INTERPROC)
#define EDGE_IS_INTERPROC(edge) (EDGE_CAT(edge) & ET_INTERPROC)

#define T_CFG_EDGE(edge) ( (t_cfg_edge*) (edge) )

#define CFG_EDGE_HEAD(edge)        (T_BBL(EDGE_HEAD(edge)))
#define CFG_EDGE_TAIL(edge)        (T_BBL(EDGE_TAIL(edge)))
#define CFG_EDGE_NEXT(edge)        (T_CFG_EDGE(EDGE_NEXT(edge)))
#define CFG_EDGE_PREV(edge)        (T_CFG_EDGE(EDGE_PREV(edge)))
#define CFG_EDGE_SUCC_NEXT(edge)   (T_CFG_EDGE(EDGE_SUCC_NEXT(edge)))
#define CFG_EDGE_SUCC_PREV(edge)   (T_CFG_EDGE(EDGE_SUCC_PREV(edge)))
#define CFG_EDGE_PRED_NEXT(edge)   (T_CFG_EDGE(EDGE_PRED_NEXT(edge)))
#define CFG_EDGE_PRED_PREV(edge)   (T_CFG_EDGE(EDGE_PRED_PREV(edge)))
#define CFG_EDGE_FLAGS(edge)       (T_CFG_EDGE(edge)->flags)
#define CFG_EDGE_CFG(edge)       (T_CFG_EDGE(edge)->cfg)
#define CFG_EDGE_SEC(edge)       (T_CFG_EDGE(edge)->cfg->sec)
#define CFG_EDGE_OBJ(edge)       (T_CFG_EDGE(edge)->cfg->sec->obj)
#define EDGE_CORR(edge)        (edge->corr_edge)
#define EDGE_PROCSTATE(edge)        (edge->procstate)
#define EDGE_PROP_REGS(edge)   (edge->prop_regs)
#define EDGE_UDCHAIN(edge)	(edge->ud_chain)
#define EDGE_NEXT_MARKED_INTO_FUN(edge)  (T_CFG_EDGE(edge)->next_marked_into_fun)

#define EDGE_EXEC_COUNT(iedg)          (((t_cfg_edge *) iedg)->execution_count)
#define EDGE_EXEC_COUNT_MIN(iedg)          (((t_cfg_edge *) iedg)->execution_count_min)
#define EDGE_EXEC_COUNT_MAX(iedg)          (((t_cfg_edge *) iedg)->execution_count_max)

#define CfgEdgeIsInterproc(edge) EDGE_IS_INTERPROC(edge)
#define CfgEdgeIsForwardInterproc(edge) (EDGE_IS_FORWARD_INTERPROC(edge))
#define CfgEdgeIsBackwardInterproc(edge) (EDGE_IS_BACKWARD_INTERPROC(edge))
#define CfgEdgeIsFallThrough(edge) (EDGE_CAT(edge) & (ET_FALLTHROUGH|ET_IPFALLTHRU))

t_edge * EdgeGetReturnForCall(t_edge * call);


/*! This structure represents a control flow edge in Diablo. Its first structure is a (graph) edge, so it can be casted to a graph edge */ 
struct _t_cfg_edge
{
  /*! The edge to eventually cast to */
  t_edge edge;
  /*! The cfg to which this edge belongs */
  t_cfg * cfg;
  /*! The corresponding edge: for intraprocedural edges, this is
      NULL. Each interprocedural edge has one corresponding edge: the
      call and return edges correspond to each other, and other
      interprocedural branches have corresponding compensating edges */
  struct _t_cfg_edge * corr_edge;
  /*! We can store all kinds of information about edges in the flags
    bitvector */
  t_uint32 flags;
  /*! For each procedure, we hold a linked list of incoming
  interprocedural edges during constant propagation. This next field
  is used for that list.*/
  struct _t_cfg_edge * next_marked_into_fun;
  /*! During constant propatagation, interprocedurally propagated
    information is stored for the edges, in this member. */
  t_procstate * procstate; 
  /*! During constant propagation, not all register contents have to
  be propagated over all edges: over link edges, e.g., we only
  propagate unchanged registers, will the others are propagated over
  return edges. Before constant propagation, the set of registers that
  need to be propagated over an edge, is set in this member.*/
  t_regset prop_regs;
  /*! To see whether or not information needs to be propagated over a
      conditional (and other types) edges, we attach a procedure to
      it, that will look at the procstate being propagated to see
      whether an edge can be followed or not. This procedure is
      initialized and stored in this member, prior to constant
      propagation */
  TestAndSetConditionFunction testcondition; 
  
    /*! This field is used during ud_chain analysis:
        it contains the ud chain for the register that 
	is propagated along this edge. */
  t_ud_chain * ud_chain;
 
  /*! Edges have a reference count because it is possible (especially with
      hell nodes) to have two identical edges (same type, same head and tail),
      and representing this edge only once saves memory and speeds up the 
      dataflow analyses */
  t_uint32 refcount;

  /*! This field gets filled in during constant propagation:
      for a call edge it contains the number of instructions that were
      not evaluated during a _context_sensitive_ run over the function
      from this edge. This information can be useful for selective inlining
      of functions */
  t_uint32 n_unevaluated_ins;

  /* corresponding reloc for switch edges:
   * relocations in switch tables should be removed when the switch edges to which 
   * they correspond are removed. this is the counterpart of the reloc/hell-edge relation */
  t_reloc * rel;

  /* Field is used for extended copypropagation */
  t_equations eqs;

  t_uint32 switchvalue;

#ifdef I386_SUPPORT
  t_int32 stack_delta;
#endif

  t_int32 execution_count;
  t_int32 execution_count_min;
  t_int32 execution_count_max;

#ifdef IA64_SUPPORT
  /* profileringsinformatie (ia64) */
  t_uint64 taken_edge;
  t_uint64 evaluated_edge;
  t_uint64 target_bbl;
#endif  
};

#ifdef IA64_SUPPORT

#define EDGE_TAKEN(edge) (T_CFG_EDGE(edge)->taken_edge)
#define EDGE_EVALUATED(edge) (T_CFG_EDGE(edge)->evaluated_edge)
#define EDGE_TARGET(edge) (T_CFG_EDGE(edge)->target_bbl)

#endif



/*! \defgroup EDGE_ALLOC Constructors and destructors */
/*@{*/
t_cfg_edge * CfgEdgeNew(t_cfg * flowgraph, t_bbl * from, t_bbl * to,t_uint32 type);
t_cfg_edge * CfgEdgeNewCompensating(t_cfg * flowgraph, t_cfg_edge * ip);
t_cfg_edge * CfgEdgeNewCall(t_cfg * flowgraph, t_bbl * from, t_bbl * to, t_bbl * ret, t_bbl * exit);
t_cfg_edge * CfgEdgeChangeReturnForCall(t_cfg_edge * call, t_bbl * ret);
void CfgEdgeMoveToLastSucc(t_cfg_edge * fall);
void CfgEdgeKill(t_cfg_edge *edge) ;
/*@}*/
#endif
/*@}*/
/*@}*/
