/***********************************
 * Copyright 2001,2002,2003: {{{
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put, Pieter Ruts
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This file contains all arch-independent functions to handle basic blocks in
 * objects
 *
 * Written by Pieter Ruts (pruts@elis.rug.ac.be)
 *
 * Copyright (c) 2002,2003 Pieter Ruts 
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/
/* Includes {{{ */
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include <stdlib.h> 
#include "diablo_action.h"
#include "diablo_bbl.h"
#include "diablo_object.h"
#include "diablo_function.h"
#include "diablo_ins.h"
#include "diablo_edge.h"
#include "diablo_flowgraph.h"
#include "diablo_dominator.h"
#include "diablo_iterators.h"
#include "diablo_factor.h"
#include "diablo_print.h"
#include "diablo_regset.h"
#include "../diablo_options.h"
#include "../datastructures/data_hash.h"
#include "../datastructures/data_graph.h"
#ifdef ARM_SUPPORT
#include "../arch/arm/arm_instructions.h"
#endif
/* }}} */
#ifdef FACTOR_SUPPORT

#define NOCOPY 33
#define FREE   34

t_uint32 killcount=0,addcount=0,count=0; 		/*Global variables*/

/*Partition Functions{{{*/
my_cmpresult BblFingComp(void* key, void* key2)		/*compares fingerprints*/
{
 t_uint32 n=T_FINGER(key)->nins;			/*this part depends on the format of the fingerprint*/
 t_uint32 m=T_FINGER(key2)->nins;			/*which should be platform-independent*/

 if(n!=m) return -1;
 
 for(m=0;m<n;m++)
 {
   if((T_FINGER(key)->opc[m])!=(T_FINGER(key2)->opc[m])) return 1;
 }
 
 return 0;
}


t_uint32 BblHash(data_hash_table * table, void * key)	 /*returns bucketnumber = modulo of the fingerprint*/
{
 return T_FINGER(key)->nins % table->tsize;
}


void * BblFinger(t_bbl *bbl) 
{	
  t_ins * ins;						/*implementation of fingerprint-alghorithm based on opcodes*/

  t_uint32 n=BBL_NINS(bbl);
  t_finger* ret=Malloc(sizeof(t_finger));
  t_uint32 i=0;
  ret->nins=n;
  ret->opc=Calloc(n,sizeof(char));
  
  BBL_FOREACH_INS(bbl,ins)
  {
    ret->opc[i]=(char)ARMINSOPCODE(ins);/*ARM-DEPENDENT*/
    i++;
  } 
  return (void*)ret;
}
  


void BblHeFree (void* bblhe,void* data)
{
  t_bbl_he * tmp=bblhe;
  Free(T_FINGER(tmp->node.key)->opc);
  Free(tmp->node.key);					/*frees the memory allocated for the fingerprint*/
  Free(bblhe);						/*no other data to be freed: the bbl must not be freed!*/
}

data_hash_table * BblPart(t_cfg * cfg) 			
{

  t_bbl * bbl;						/*iterator*/
  t_bbl_he * bblhe;
  data_hash_table * table=HashTableNew(8/*size*/,0/*offset*/,BblHash,BblFingComp,BblHeFree);	/*construct new table*/
  
  CFG_FOREACH_BBL(cfg,bbl)				/*fill in the table*/
  {
    bblhe=Malloc(sizeof(t_bbl_he));
    bblhe->bbl=bbl;
    bblhe->ident=NULL;
    bblhe->node.key=BblFinger(bbl);
    if (!IS_DATABBL(bbl) && T_FINGER(bblhe->node.key)->nins > 2) HashInsert(table,(void *)bblhe);
    else BblHeFree(bblhe,NULL);
  }
  return table;
}

void BblResetIdent(void* bblhe, void* data)		/*reset the ident-fields of all nodes in the table*/
{
  T_BBL_HE(bblhe)->ident=NULL;
}

void BblCheckIdent(void* bblhe, void* data)
{
  if (T_BBL_HE(bblhe)->ident==NULL) FATAL(("(Ident==NULL"));
}

void BblHePrint(void* bblhe, void* data)
{
  int i;
  t_bbl_he* tmp=bblhe;
  printf("\nbblhe: %p   \tsize: %d   \tident: %p   \tequal: %p   \tfingerprint: ",tmp,T_FINGER(tmp->node.key)->nins,tmp->ident,tmp->node.equal);
  
  for(i=0;i<T_FINGER(tmp->node.key)->nins;i++) printf("%d ",T_FINGER(tmp->node.key)->opc[i]);
}



/*}}}*/

/*Find Identical Bbls{{{*/
 
t_bool BblComp(t_bbl* bbl1,t_bbl* bbl2)
{
  t_ins* ins1=BBL_INS_FIRST(bbl1); 
  t_ins* ins2=BBL_INS_FIRST(bbl2);
  
  if(BBL_NINS(bbl1)!=BBL_NINS(bbl2)) return FALSE;

  while(ins1!=NULL)
  {
    if(!BBL_OBJECT(bbl1)->actions->CompareInstructions(ins1,ins2)) if(ARMINSOPCODE(ins1)!=ARM_B) return FALSE;
    ins1=ins1->next;
    ins2=ins2->next;
  }
  
  return TRUE;
}
 

void BblFindIdent(void *bblhe, void* part) 		/*links bbl_hes with identical bbl's*/
{
 t_bbl_he* current=bblhe;
 t_bbl_he* check=T_BBL_HE(current->node.equal);
 
 if (T_BBL_HE(bblhe)->ident!=NULL)return;		/*every bblhe has the next bblhe with identical bbl in its ident field*/
 							/*thus forming a circular linked list*/
 while(check!=NULL)
 {
   if(BblComp(current->bbl,check->bbl))
   {
     current->ident=check;
     current=check;
   }
   check=T_BBL_HE(check->node.equal);
 }
 
 current->ident=bblhe;
 return;



}/*}}}*/

/*Find Functionally Identical Bbls. Remove Other Bbls from Table{{{*/

my_cmpresult BblSymbCmp(t_bbl* bbl1,t_bbl* bbl2)     /*ARM-DEPENDENT*/
{						/*algorithm used: cfr Sifting out the Mud*/
  int i=1,j;					/*keeps track of instruction*/
  
  int R1[32]={0};				/*used to store which regs contain internal values*/
  int R2[32]={0};

  t_ins* ins1=BBL_INS_FIRST(bbl1);
  t_ins* ins2=BBL_INS_FIRST(bbl2);

  while(ins1)
  {
    for(j=i;j>0;j--)							/*checks if reg is defined in one of the previous instructions*/
    {
      if(ARMINSREGB(ins1)==R1[j]&&ARMINSREGB(ins2)!=R2[j]) return 1; 	/*if the src reg of one instruction is in the table and the src reg of the other*/
      if(ARMINSREGB(ins1)!=R1[j]&&ARMINSREGB(ins2)==R2[j]) return 1;	/*is not, then the bbls cannot be functionally equivalent*/

      if(ARMINSREGC(ins1)==R1[j]&&ARMINSREGC(ins2)!=R2[j]) return 1;      
      if(ARMINSREGC(ins1)!=R1[j]&&ARMINSREGC(ins2)==R2[j]) return 1;
    }

    if (ARMINSCONDITION(ins1)!=ARMINSCONDITION(ins2)) return 1;
    if (ARMINSFLAGS(ins1)!=ARMINSFLAGS(ins2)) 	      return 1;
    if (ARMINSSHIFTTYPE(ins1)!=ARMINSSHIFTTYPE(ins2)) return 1;
    if (ARMINSIMMEDIATE(ins1)!=ARMINSIMMEDIATE(ins2)) return 1;		/*checks if imm operands, shifts, condition bits etc are the same*/

    R1[ARMINSREGA(ins1)]=i;
    R2[ARMINSREGA(ins1)]=i;
    
    ins1=INS_NEXT(ins1);
    ins2=INS_NEXT(ins2);

    i++; 
  }
  return 0;
}


/*Finds candidates for abstraction and removes other Bbls from table"*/

void BblSymbRename(void* bblhe, void* part)
{
  t_bbl_he* current=bblhe;
  t_bbl_he* check=T_BBL_HE(current->node.equal);

  if (T_BBL_HE(bblhe)->ident!=NULL) return;             
  
  while(check!=NULL)
  {
    if((BblSymbCmp(current->bbl,check->bbl)==0)&&(check->ident==NULL))
    {
//    printf("Symbolically identical bbl's found: %p and %p, with size: %d\n",current->bbl,check->bbl,T_FINGER(current->node.key)->nins);
      current->ident=check;
      current=check;
    }
    check=T_BBL_HE(check->node.equal);
 }

 current->ident=bblhe;
 if (((t_bbl_he *)bblhe)->ident==bblhe) HashDelete((data_hash_table *)part,bblhe);
 return;
 
}


/*}}}*/
 
/*BblAbstract : The real thing: architecture independent wrapper{{{*/
t_uint32  BblAbstract(t_bbl * master,t_bbl * slave,t_uint32 * in,t_uint32 * out)

  /*if in and out are NULL, then there are no copy ops to be performed;*/

{
    return BBL_OBJECT(master)->actions->BblAbstract(master,slave,in,out);
}
/*}}}*/

/*Functions for symbolically identical bbls {{{ */
/*BblCmpExtReg: Find different consumed externally defined registers : architecture independent wrapper{{{*/

t_uint32 BblCmpExtReg(t_bbl * master,t_bbl * slave,t_uint32 * in)
{
 return (BBL_OBJECT(master))->actions->BblCmpExtReg(master,slave,in);
}

/*}}}*/
/*BblCmpDefReg: Find differently defined registers that are live at the end of the slave block{{{*/

t_uint32  BblCmpDefReg(t_bbl * master,t_bbl * slave, t_uint32 * out)
{
  t_ins * ins1=BBL_INS_LAST(master);
  t_ins * ins2=BBL_INS_LAST(slave);
  
  t_regset tmp=RegsetNewIntersect(slave->regs_defined,slave->regs_live_out);  	/*tmp contains regs that are defined in slave and that are live at the end*/
  t_regset diff1;
  t_regset diff2;

  tmp=RegsetSetIntersect(tmp,0x7f80ffff);					/*temporary:defines which regs are real targets*/
  
  while(RegsetCountRegs(tmp))
  {
   diff1=RegsetNewDiff(ins1->defined,ins2->defined); 		/*diff filters out conditional bits that are being set*/	
   diff2=RegsetNewDiff(ins2->defined,ins1->defined);
   
   if(RegsetIsSubset(tmp,diff2)&&(!RegsetIsEmpty(diff2)))	/*see if defined reg is in tmp and if a copy is needed*/
   {
    t_reg i,j=0,k=0;
    REGSET_FOREACH_REG(diff1,i) j=i;				/*extract the defined regs in decimal form to insert them in out[]*/
    REGSET_FOREACH_REG(diff2,i) k=i;
     
    if(k!=j)							/*double check*/	
    {
     if(out[k]==NOCOPY) out[k]=j;
     else return 1;						/*if two registers have to be copied to one, abstraction is not possible*/		
    }

   }
   tmp=RegsetNewDiff(tmp,ins2->defined);
   
   ins1=INS_PREV(ins1);
   ins2=INS_PREV(ins2);
  }
  
    return 0;
}
/*}}}*/
/*BblCmpLivReg: Find registers that are live before and after slave block, that are defined in master block. Store them in a free register.{{{*/
t_uint32 BblCmpLivReg(t_bbl * master,t_bbl * slave,t_uint32 * in,t_uint32 * out)
{
 t_reg i,j;
  
 t_regset s_live_in=InsRegsLiveBefore(BBL_INS_FIRST(slave));
 t_regset s_live_out=slave->regs_live_out;
 t_regset s_defined=slave->regs_defined;
 t_regset m_defined=master->regs_defined;
 t_regset m_used=master->regs_used;

 t_regset tmp=NullRegs;									 
 for(i=0;i<32;i++) if(in[i]!=NOCOPY) RegsetAddReg(tmp,i);				/*registers that are defined in BblCmpExtReg*/
 
 t_regset copy=RegsetNewIntersect(RegsetNewDiff(RegsetNewIntersect(s_live_in,s_live_out),s_defined),RegsetNewUnion(m_defined,tmp)); /*registers that have to be stored*/  
 
 t_regset free=RegsetNewInvers(RegsetNewUnion(RegsetNewUnion(s_live_in,s_live_out),RegsetNewUnion(m_defined,m_used)));
 free=RegsetSetIntersect(free,0x7f807fff);	/*exclude pc and conditionbits*/
 
 if (RegsetCountRegs(copy)>RegsetCountRegs(free)) return 1;
 
 tmp=NullRegs;
 REGSET_FOREACH_REG(copy,i)							/*find free registers to store values*/
 {
  REGSET_FOREACH_REG(free,j)
  {
    in[j]=i;
    out[i]=j;
    RegsetAddReg(tmp,j);						
    break;
  }
 RegsetSetDiff(free,tmp);							/*remove used register from free list*/
 }

 REGSET_FOREACH_REG(free,i) in[i]=out[i]=FREE;					/*indicate which regs are still free -> used in BblAbstract*/

 return 0;
}
/*}}}*/

/*BblFindCopyOps: Determine which copy operations have to be inserted {{{*/
t_uint32 BblFindCopyOps(t_bbl * master,t_bbl * slave,t_uint32 * in,t_uint32 * out)
{										
                                                                               
 /*copy ops are kept in two arrays: r5:=r4 becomes in[5]=4
  * ToDo: find a better way to set the array size
  * NOCOPY means that that register remains unchanged (0 is for r0)
  *************************************************************/
 t_uint32 i,cpycnt=0;
 
 for(i=0;i<32;i++) out[i]=in[i]=NOCOPY;
 
 if(BblCmpExtReg(master,slave,in)==1)	  return 1;	/*arch dependent*/

 if(BblCmpDefReg(master,slave,out)==1)    return 2;

 if(BblCmpLivReg(master,slave,in,out)==1) return 3;

  printf("in:\t");
  for(i=0;i<32;i++) if(in[i]!=NOCOPY) printf("%s:=%s\t",BBL_OBJECT(master)->actions->RegName(i),BBL_OBJECT(master)->actions->RegName(in[i]));
  printf("\n");
  
  printf("out:\t");
  for(i=0;i<32;i++) if(out[i]!=NOCOPY) printf("%s:=%s\t",BBL_OBJECT(master)->actions->RegName(i),BBL_OBJECT(master)->actions->RegName(out[i]));
  printf("\n");



   for(i=0;i<32;i++)
 {
   if(in[i]!=NOCOPY) cpycnt++;
   if(out[i]!=NOCOPY) cpycnt++;
 }

 if((cpycnt+2)>=slave->nins) return -1;			/*don't abstract if to many copyops needed*/
 

 return 0;
}
/* }}} */

/*BblFactorSymb: Factorise symbolically identical blocks {{{*/
void BblFactorSymb(data_hash_table * table)
{
 t_uint32 n; 
 t_uint32 in[32];
 t_uint32 out[32];
 
 t_bbl_he* master;				/*the scheme followed here takes one bbl as master and checks if the next candidates can be*/
 t_bbl_he* slave;				/*abstracted.  If not, the slave removed from the table. Inverse abstraction is not done.  */

 for (n=0; n<table->tsize; n++)
 {
  while(table->nodes[n])
  {
   master=(t_bbl_he*)table->nodes[n];
   slave=master->ident;
 
   while(slave!=master)					/*the ident fields form a circular list: if master is reached, all slaves have been checked*/
   {
     switch(BblFindCopyOps(master->bbl,slave->bbl,in,out))
     {
       case -1: 
	 printf("%p and %p were not abstracted because there would be no gain\n\n",master,slave);
	 break;
       case  0: 
         if(BblAbstract(master->bbl,slave->bbl,in,out)==0) printf("%p is succesfully abstracted to %p\n\n",slave,master);
	 else printf("%p could not be abstracted to %p",slave,master);
	 break;
       case  1: 
	 printf("%p cannot be abstracted to %p because there are two copy ops to the same register before the slave\n\n",slave,master);
	 break;	
       case  2: 
	 printf("%p cannot be abstracted to %p because there are two copy ops to the same register after the slave\n\n",slave,master);
	 break;
       case  3: 
	 printf("%p cannot be abstracted to %p because there are not enough free registers\n\n",slave,master);
	 break;
       default: 
	 FATAL(("Should not get here!\n"));
     }
     
     master->ident=slave->ident;
     HashDelete(table,slave);
     slave=master->ident;
   }

   HashDelete(table,master);

  }
 }
 return;
}

/*}}}*/
/* }}} */

/*BblFactorIdent:	Factorise completely identical blocks before even searching for symbolical equivalence{{{*/
void BblFactorIdent(data_hash_table * table)
{
 t_uint32 n;
 
 t_bbl_he* master;				/*the scheme followed here takes one bbl as master and checks if the next candidates can be*/
 t_bbl_he* slave;				/*abstracted.  If not, the slave removed from the table. Inverse abstraction is not done.  */
 t_bbl_he* itn;
 
 for (n=0; n<table->tsize; n++)
 {
   master=(t_bbl_he*)table->nodes[n];
   itn=T_BBL_HE(master->node.next);
   
   while(master!=NULL)
   {
     t_bbl *  newmaster=NULL;
     slave=master->ident;
     if (slave!=master)
     { 
       count++;
       BblAbstract(master->bbl,newmaster,NULL,NULL);			
       //if(count==global_options.debugcounter) return;
     }

     while(slave!=master)					/*the ident fields form a circular list: if master is reached, all slaves have been checked*/
     {
       if(slave->bbl->nins!=master->bbl->nins) printf("boe! %p %p",slave,master);
       BblAbstract(master->bbl,slave->bbl,NULL,NULL);
       master->ident=slave->ident;
       HashDelete(table,slave);
       slave=master->ident;
     }

     BblResetIdent(master,NULL);
   
     if(master->node.equal) master=T_BBL_HE(master->node.equal);
     else 
     {
       master=itn;
       itn=itn?T_BBL_HE(master->node.next):NULL;
     }
     
   }
 }
 return;
}/*}}}*/

/*Factor: Factorisation of Basic Blocks {{{*/
void Factor(t_cfg * cfg)
{
    printf("\n\n*** Start of factorisation ***\n\n");
    data_hash_table * part=BblPart(cfg);
    
    HashWalk(part,BblFindIdent,NULL);			/*Allows for more simple abstraction beforehand*/
//    HashWalk(part,BblHePrint,NULL); 
    BblFactorIdent(part);
//    HashWalk(part,BblResetIdent,NULL);			/*Makes the Ident-field of every Bbl NULL.*/
//    HashWalk(part,BblSymbRename,part);			/*Finds candidates for abstraction, removes other Bbls from table*/
//    HashWalk(part,BblCheckIdent,NULL);
//    BblFactorSymb(part); 		 	        	/*Determines which slave Bbls have to be actually renamed*/
    HashFree(part);
    printf("Killed: %d\tAdded: %d\t Saldo: %d\tCount: %d\n",killcount,addcount,killcount-addcount,count);
    printf("\n*** End of factorisation ***\n\n");
}
/*}}}*/

#endif

typedef struct _bbl_hash_entry
{
  t_bbl * key;
  struct _bbl_hash_entry * next;
} bbl_hash_entry;

typedef struct _bbl_hashtable
{
  bbl_hash_entry ** buckets;
  unsigned int table_size;
} bbl_hashtable;

#define BBL_TABLE_FOREACH_BUCKET(t,b,k) for (k=0,b=t->buckets[k];k<t->table_size;k++,b=t->buckets[k])
#define BUCKET_FOREACH_ENTRY(b,e) for (e=b;e;e=e->next)

extern t_uint32 I386BblHashKey(t_bbl * bbl);
extern t_bool I386InsAreIdentical(t_ins * ins1, t_ins * ins2);

bbl_hashtable * init_table(unsigned int size)
{
  bbl_hashtable * table = (bbl_hashtable *) malloc(sizeof(bbl_hashtable));
  table->table_size = size;
  table->buckets = (bbl_hash_entry**) calloc(size,sizeof(bbl_hash_entry));
  return table;
}

t_bbl * put(bbl_hashtable * table, t_bbl * key)
{
  int hash = (I386BblHashKey(key) % table->table_size);
  bbl_hash_entry * entry = table->buckets[hash];
  bbl_hash_entry * new_entry;
  
  if (entry)
    {
      while (entry)
	{
	  if (entry->key==key)
	    {
	      t_bbl * return_value = entry->key;
	      entry->key = key;
	      return return_value;
	    }
	  entry=entry->next;
	}
    }
  new_entry = (bbl_hash_entry*) malloc(sizeof(entry));
  new_entry->key=key;
  new_entry->next=table->buckets[hash];
  table->buckets[hash]=new_entry;
  return NULL;
}


void FindCodeDuplication(t_cfg * cfg)
{
#ifdef I386_SUPPORT
  t_bbl * ibbl, *jbbl;
  t_ins * iins, * jins;
  int i;
  bbl_hashtable * hashtable;

  bbl_hash_entry * bucket, *entry1, *entry2;

  int teller = 0;
#if 0
  CFG_FOREACH_BBL(cfg,ibbl)
    if (BBL_NINS(ibbl)>2)
      CFG_FOREACH_BBL(cfg,jbbl)
	if (GENERIC_LT(BBL_OLD_ADDRESS(ibbl),BBL_OLD_ADDRESS(jbbl)) && BBL_NINS(jbbl)>2)
	  {
	    int local_teller = 0;
	    iins = BBL_INS_FIRST(ibbl);
	    jins = BBL_INS_FIRST(jbbl);
	    
	    while (iins && jins && I386InsAreIdentical(iins,jins))
	      {
		teller++;
		local_teller++;
		iins=INS_NEXT(iins);
		jins=INS_NEXT(jins);
	      }
	    
	    if (local_teller>=(BBL_NINS(ibbl)+BBL_NINS(jbbl)/3))
	      DiabloPrint(stdout,"DUPL @iB\n    @iB\n-------------------------------------------------------------------------\n",ibbl,jbbl);
	  }
  
  printf("total duplication: %d\n",teller);
#else	    
  hashtable=init_table(10000);

  CFG_FOREACH_BBL(cfg,ibbl)
    {
      if (BBL_NINS(ibbl)>2 && !IS_DATABBL(ibbl))
	put(hashtable,ibbl);
    }
  
  BBL_TABLE_FOREACH_BUCKET(hashtable,bucket,i)
    BUCKET_FOREACH_ENTRY(bucket,entry1)
      BUCKET_FOREACH_ENTRY(bucket,entry2)
        {
	  ibbl=entry1->key;
	  jbbl=entry2->key;
	  if (GENERIC_LT(BBL_OLD_ADDRESS(ibbl),BBL_OLD_ADDRESS(jbbl)))
	    {
	      int local_teller = 0;
	      iins = BBL_INS_FIRST(ibbl);
	      jins = BBL_INS_FIRST(jbbl);
	      
	      while (iins && jins && I386InsAreIdentical(iins,jins))
		{
		  teller++;
		  local_teller++;
		  iins=INS_NEXT(iins);
		  jins=INS_NEXT(jins);
		}
	      
	      if (local_teller>=(BBL_NINS(ibbl)+BBL_NINS(jbbl)/3))
		DiabloPrint(stdout,"DUPL @iB\n    @iB\n-------------------------------------------------------------------------\n",ibbl,jbbl);
	    }
	}
  
  printf("total duplication: %d\n",teller);
#endif
  exit(0);
#endif
}



t_bool CompareTwoFunctions(t_function * fun1, t_function * fun2)
{
  t_bbl * bbl1, * bbl2;
  t_ins * ins1, * ins2;

  bbl1 = FUN_BBL_FIRST(fun1);
  bbl2 = FUN_BBL_FIRST(fun2);

  while (bbl1 && bbl2)
    {
      if (BBL_NINS(bbl1)!=BBL_NINS(bbl2)) return FALSE;
      
      ins1 = BBL_INS_FIRST(bbl1);
      ins2 = BBL_INS_FIRST(bbl2);

      while (ins1 && ins2)
	{
	  if (!I386InsAreIdentical(ins1,ins2)) return FALSE;
	  ins1 = INS_NEXT(ins1);
	  ins2 = INS_NEXT(ins2);
	}

      bbl1 = BBL_NEXT_IN_FUN(bbl1);
      bbl2 = BBL_NEXT_IN_FUN(bbl2);
    }

  if (bbl1!=bbl2) return FALSE;

  DiabloPrint(stdout,"two funs at @B and @B seem identical",FUN_BBL_FIRST(fun1),FUN_BBL_FIRST(fun2));

  return TRUE;
  
}


void CompareFunctions(t_cfg * cfg)
{
  t_function *fun1,*fun2;

  CFG_FOREACH_FUN(cfg,fun1)
    CFG_FOREACH_FUN(cfg,fun2)
    if (fun1!=fun2) CompareTwoFunctions(fun1,fun2);

  exit(0);
}

/* vim: set shiftwidth=2 foldmethod=marker: */
