/************************************************************************
 * Copyright 2001,2002,2003:{{{
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Headers for flowgraph manipulation 
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 ***********************************************************/

/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup FLOWGRAPHS Flowgraphs 
 *
 * Flowgraphs are derived from graphs. Instead of nodes and edges,
 * they consist of basic blocks (t_bbl) and control flow edges
 * (t_cfg_edge). The graph themselves also contains procedures, and
 * the blocks and edges are partitioned into the procedures.
 */
/*@{*/


#ifndef MY_FLOWGRAH_TYPEDEFS
#define MY_FLOWGRAH_TYPEDEFS
typedef struct _t_cfg t_cfg;

#endif /* MY_FLOWGRAH_TYPEDEFS */

#ifndef DIABLO_INS_TYPEDEFS
#define DIABLO_INS_TYPEDEFS
#include "../simple_types.h"
typedef t_uint32 t_instruction_flags;
typedef struct _t_ins t_ins;
#endif

#ifndef _GRAPH_TYPEDEFS
#define _GRAPH_TYPEDEFS
typedef struct _t_edge t_edge;
typedef struct _t_node t_node;
typedef struct _t_graph t_graph;
#endif

#ifndef _UDGRAPH_TYPEDEFS
#define _UDGRAPH_TYPEDEFS
typedef t_graph t_ud_graph;
typedef t_ins 	t_ud_node;
typedef t_edge	t_ud_edge;
#endif

#ifndef MY_FLOWGRAPH_H
#define MY_FLOWGRAPH_H
#include "../datastructures/data_graph.h"
#include "../datastructures/data_ll.h"
#include "diablo_object.h"
#include "diablo_function.h"
#include "diablo_procstate.h"
#include "diablo_ins.h"
#include "diablo_section.h"
#include "diablo_edge.h"
#include "diablo_udgraph.h"
#include "diablo_loop.h"
#include "diablo_programexit.h"
#include "diablo_callgraph.h"

/*! This structure represents a control flow graph in Diablo. Its first
  structure is a graph, so it can be casted to a graph and vice versa */
struct _t_cfg
{

  /*! This structure represents a control flow graph in Diablo. Its first
   * structure is a graph, so it can be casted to a graph and vice versa */

  t_graph graph;
  /*! Program entry points */
  t_cfg_entrypoints * entries;

  /*! Procedures in a control flow graph are stored in a double linked list.
   * fun_first is a pointer to the first procedure in it. (Note that blocks and
   * edges are stored in the graph structure from which this ecfg structure is
   * derived */

  t_function * fun_first;
  
#ifdef INSTRUMENTATION
  t_function * fun_instrument_first;
#endif

  /*! And a pointer to the last procedure (also see \ref _t_cfg::fun_first) */

  t_function * fun_last;

  /*! We also hold a linked list of marked procedure during constant
   * propagation, where it is used a stack. */
  
  t_function * list_of_marked_funs;

  /*! Each cfg is just some representation of the code in a section, to which
   * we keep a pointer. */

  t_section * sec;

  /*! All instructions in the flow graph will also be included in a use-definition graph
      A pointer to this graph is stored in the flow graph */
  t_ud_graph * udgraph;
  
  /*! Special functions and blocks */
  t_function * hell_function;
  t_function * call_hell_function;
  t_function * wrap_function;
  t_function * swi_hell_function;
  t_bbl * hell_node;
  t_bbl * call_hell_node;
  t_bbl * swi_hell_node;
  t_bbl * exit_hell_node;
  t_bbl * exit_call_hell_node;
  t_bbl * exit_swi_hell_node;
  t_bbl * unique_entry_node;
  t_bbl * unique_exit_node;
  
  /*! Loops are stored in the graph as a linked list of loops */
  t_loop * loop_first;
  t_loop * loop_last;

  /*! A linked list of exit blocks */
  t_cfg_exits * exitblocks;

  t_int32 hot_threshold_count;
  
  t_cg * cg;
};

#define T_CFG(cfg) ((t_cfg*)(cfg))

#define CFG_FUN_FIRST(cfg)   (T_CFG(cfg)->fun_first)
#define CFG_FUN_LAST(cfg)    (T_CFG(cfg)->fun_last)   
#define CFG_MARKED_FUNS(cfg) (T_CFG(cfg)->list_of_marked_funs)
#define CFG_SECTION(cfg)     (T_CFG(cfg)->sec)
#define CFG_OBJECT(cfg)      (SECTION_OBJECT(CFG_SECTION(cfg)))
#define CFG_UDGRAPH(cfg)     (T_CFG(cfg)->udgraph)
#define CFG_LOOP_FIRST(cfg)   (T_CFG(cfg)->loop_first)
#define CFG_LOOP_LAST(cfg)    (T_CFG(cfg)->loop_last)   
#define CFG_NBBLS(graph)        (T_GRAPH(graph)->nnodes)
#define CFG_NEDGES(graph)        (T_GRAPH(graph)->nedges)
#define CFG_DESCRIPTOR(graph)   (T_CFG(graph)->sec->obj->description)
#define CFG_HOT_THRESHOLD_COUNT(graph) (T_CFG(graph)->hot_threshold_count)
t_uint32 CfgCreateBasicBlocks(t_section * section);
t_uint32 CfgCreateFunctions(t_cfg * fg);


/*! \defgroup CFG_ALLOC Constructors and destructors */
/*@{*/
t_cfg * CfgNew(t_section * sec);
void CfgFree(t_cfg * cfg);
/*@}*/


#define InitBblMarking InitNodeMarking


t_cfg_edge * CfgAppendNode(t_cfg * cfg, t_bbl * after, t_bbl * insert, t_uint32 edge_category);
void CfgCountBlocks(t_cfg * cfg);
void CfgRemoveDeadCodeAndDataBlocks(t_cfg * fg);
void CfgRemoveDeadCode(t_cfg * fg);
void CfgRemoveUselessConditionalJumps(t_cfg * cfg);
void CfgRemoveEmptyBlocks(t_cfg * cfg);
void CfgMoveInsDown(t_cfg * cfg);
t_uint32 CfgPatchInterProcedural(t_cfg * fg);
t_bool CfgPatchForNonReturningFunctions(t_cfg * cfg);
#define CfgAppendBbl(cfg,after, insert, edge_category) T_CFG_EDGE(GraphAppendNode(T_GRAPH(cfg), T_NODE(after), T_NODE(insert), edge_category))
void Export_Flowgraph(t_bbl * start, t_uint16 mask, const char * filename);
void ExportFunctionGraph(t_section * code, const t_string dirprefix);
void CfgPrintNrIns(t_cfg * cfg);
void CfgVerifyCorrectness(t_cfg * cfg);
void Export_Function(t_function * fun, const char * filename);
void Export_Function_Annotate(t_function * fun, const char * filename, void (*annotate)(t_bbl * bbl, FILE * fp));

t_bool CfgDetectSimpleInfiniteLoops(t_cfg * cfg, ll ** head);
t_bbl* CfgEdgeInsertBlock(t_cfg * cfg, t_cfg_edge * edge);

t_cfg_edge * FindFunLinkForCall(t_cfg_edge * edge);
t_cfg_edge * FindFunLinkForReturn(t_cfg_edge * edge);
t_cfg_edge * FindCallForFunLink(t_cfg_edge * edge);
t_cfg_edge * FindReturnForFunLink(t_cfg_edge * edge);

void CfgSortEdges(t_cfg * cfg);
#endif
/*@}*/
/*@}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
