/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains all arch-independent functions to handle functions
 * in objects
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/
/* {{{ includes */
#include <string.h>
#include "diablo_function.h"
#include "diablo_bbl.h"
#include "diablo_edge.h"
#include "diablo_ins.h"
#include "diablo_print.h"
#include "diablo_iterators.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../datastructures/data_graph.h"
#include "../datastructures/data_string.h"
#include <stdlib.h> 
/* }}} */

/* {{{ constructor */
/*!
 * \param entrypoint the entrypoint for the function 
 * \param name the functions name
 * \param ft the type of the function (FT_NORMAL or FT_HELL) 
 *
 * \return a new function 
 *
 * constructor for the t_function structure: allocates memory for a function
 * node and fills in the appropriate default values */

t_function * FunctionNew(t_bbl * entrypoint, t_string name,  t_function_type ft) 
{
  t_cfg * fg=BBL_CFG(entrypoint);
  t_function * ret = Calloc(1,sizeof(t_function));

  //Xiangyu
  //printf("%x: %s\n", BBL_OLD_ADDRESS(entrypoint), name);
  ret->name=name?StringDup(name):NULL;

  ret->function_flags = 0x0;
  ret->next=NULL;
  ret->cfg = fg;
  ret->marked_bbls = NULL;
  ret->marked_edges = NULL;
  ret->next_marked = NULL;

  /* initialize the entry for this function */
  FUN_BBL_FIRST(ret) = entrypoint;

  /* Insert the entrypoint into the function (must be done first) */
  BblInsertInFunction(entrypoint,ret);

  /* Allocate and link return block */
  FUN_BBL_LAST(ret) = BblNew(fg);
  BblInsertInFunction(FUN_BBL_LAST(ret),ret);
  BblSetAttributes(FUN_BBL_LAST(ret),BBL_IS_EXITBLOCK);

  /* insert the function in the dll */
  if (fg->fun_first==NULL)
  {
    ret->prev=NULL;
    fg->fun_first=fg->fun_last=ret;
  }
  else
  {
    ret->prev=fg->fun_last;
    fg->fun_last->next=ret;
    fg->fun_last=ret;
  }

  /* for conservativeness: assume a function changes _all_ registers: this can
   * be refined with an analysis later on */
  ret->regs_changed = AllRegs();
  ret->regs_saved = RegsetNew();
  ret->regs_through = RegsetNew();
  ret->regs_used = AllRegs();

  return ret;
}
/* }}} */

/* {{{ destructor */
/*!  
 *
 * \param fun The function to kill
 *
 * Removes the function from the flowgraph. Basic blocks are not freed, this
 * should be done separate */

void FunctionKill(t_function * fun)
{
  t_cfg *fg =FUN_CFG(fun);
  if (fun->prev)
  {
    fun->prev->next=fun->next;  
  }
  else
  {
    fg->fun_first=fun->next;
  }

  if (fun->next)
  {
    fun->next->prev=fun->prev;
  }
  else
  {
    fg->fun_last=fun->prev;
  }
  if (fun->name) Free(fun->name);
  while(FUN_LOOPS(fun))
  {
    t_loopref * delete_me = FUN_LOOPS(fun);
    FUN_LOOPS(fun) = LLNext(FUN_LOOPS(fun));
    Free(delete_me);
  }
  Free(fun);
}
/* }}} */

/* {{{ Create functions for cfg */
#define SYM_HASH_SIZE	1024
struct symhash {
  t_symbol * sym;
  struct symhash * next;
};
static struct symhash *sym_address_hash[SYM_HASH_SIZE];

static t_bool helper_cfg_find_and_create_function(t_node * current, t_node * parent,void * data) {
  t_function * f;
  t_bbl * bbl=T_BBL(current);
  t_uint32 * nfuncs=(t_uint32 *) data;
  t_cfg_edge * edge;
  
  //Xiangyu
  //printf("BBL:%x, %x\n",BBL_OLD_ADDRESS(bbl),BBL_CADDRESS(bbl) );

  if (IsHellNode(bbl)) return FALSE;
  //Xiangyu
  //printf("\t not hell node\n");

  BBL_FOREACH_PRED_EDGE(bbl,edge)
  {
    /* find the type of the edge that points to this node */
    if (EDGE_CAT(edge) == ET_CALL)
    {
      if (!(BBL_ATTRIB(bbl) & BBL_IS_FUN_ENTRY))
        (*nfuncs)++;
      else
        FATAL(("Strange BBL @B was already marked",bbl));

      BBL_ATTRIB(bbl) |= BBL_IS_FUN_ENTRY; 
      break;
    }
  }

  if (BBL_ATTRIB(bbl) & BBL_IS_FUN_ENTRY) 
  {
    t_symbol * sym=NULL;
    t_uint32 hashval = (GENERIC_EXTRACT_UINT32(BBL_CADDRESS(bbl)) >> 2) & (SYM_HASH_SIZE - 1);
    struct symhash *h;
    printf("\n");
    for (h=sym_address_hash[hashval]; h; h = h->next)
    {
      if (GENERIC_EQ(GENERIC_ADD(h->sym->base->caddress,h->sym->offset_from_start),BBL_CADDRESS(bbl)))
      {
        //Xiangyu
	if (h->sym->type == FUNCTION_SYMBOL && (h->sym->name[0]!='.'))
	{
	  sym = h->sym;
          printf("FUNCTION_SYMBOL %s\n",sym->name);
	  break;
	}
        //Xiangyu
	if (!h->sym->local || !sym || (sym && sym->name[0]=='.') || (sym && sym->name[0]=='$')) {
	//if (!h->sym->local || !sym) {
	  sym = h->sym;
	  //break;
	  //Xiangyu
	  //printf("\t\t %s option 2\n",sym->name);
        }
        printf("\t\th->sym=%s\n",h->sym->name);
      }
    }
    //Xiangyu
    if (sym) printf("\t%x new function %s\n",BBL_CADDRESS(bbl),sym->name);
    else printf("\t%x new function NULL\n",BBL_CADDRESS(bbl));
    f = FunctionNew(bbl, sym?sym->name:NULL, FT_NORMAL);
    f->function_flags &= ~FF_IS_EXPORTED;
    if (sym && !sym->local)
      f->function_flags |= FF_IS_EXPORTED;
    /* if the function entry is hand written in assembler and all callers
     * are from hand written assembler code, there is no guarantee that the
     * function will indeed obey the calling conventions, even if it is exported */
    if (BBL_ATTRIB(bbl) & BBL_IS_HANDWRITTEN_ASSEMBLY)
    {
      t_cfg_edge * edge;
      t_bool only_from_assembler_code = TRUE;
      BBL_FOREACH_PRED_EDGE(bbl,edge)
      {
	if (EDGE_CAT(edge) != ET_CALL || IsHellNode(CFG_EDGE_HEAD(edge)))
	  continue;
	if (!(BBL_ATTRIB(CFG_EDGE_HEAD(edge)) & BBL_IS_HANDWRITTEN_ASSEMBLY))
	{
	  only_from_assembler_code = FALSE;
	  break;
	}
      }
      if (only_from_assembler_code)
      {
	f->function_flags &= ~FF_IS_EXPORTED;
	/*DiabloPrint(stdout,"Unexporting %s\n",f->name);*/
      }
    }
  }
  return FALSE;
}

/*! 
 *
 * \param fg The flowgraph for wich we want to create functions 
 *
 * Iterates over flowgraph and constructs a linked list of all functions */

t_uint32 CfgCreateFunctions(t_cfg * fg)
{
  t_section * section=CFG_SECTION(fg);
  t_uint32 nfunctions=0;
  t_bbl * start = T_BBL(SECTION_CFG(section)->entries->entry_bbl);
  int i;

  /*  {{{ fill the symbol hash table for fast lookup of symbol names */
  t_symbol * sym;
  SYMBOLTABLE_FOREACH_SYMBOL(section->obj->sub_symbol_table,sym)
  {
    t_uint32 hashval; 
    struct symhash *h;
    if (!sym->base) continue;
    if (sym->type == ORIGINAL_SECTION_SYMBOL) continue;
    //Xiangyu
    //printf("SYM:%s %x local=%d type=%d %d\n",sym->name,GENERIC_ADD(sym->base->caddress,sym->offset_from_start),sym->local,sym->type,FUNCTION_SYMBOL);

    if ((sym->name) && ((strcmp(sym->name,"$a")==0) || (strcmp(sym->name,"$d")==0))) continue;
    hashval = (GENERIC_EXTRACT_UINT32(GENERIC_ADD(sym->base->caddress,sym->offset_from_start)) >> 2) & (SYM_HASH_SIZE-1);
    h = Malloc(sizeof(struct symhash));
    h->sym = sym;
    h->next = sym_address_hash[hashval];
    sym_address_hash[hashval] = h;
  } /* }}} */

  /* Do function leader detection, and create all function structures */
  GraphDFTraversalWithCheckAndData((t_node *) start, helper_cfg_find_and_create_function, 0xffff,&nfunctions);

  /* {{{ tear down the symbol hash */
  for (i=0; i<SYM_HASH_SIZE; i++)
    while (sym_address_hash[i])
    {
      struct symhash *tmp = sym_address_hash[i];
      sym_address_hash[i] = tmp->next;
      Free(tmp);
    } /* }}} */
  

  //by Xiangyu
  //exit(-1);
  return nfunctions;
}
/* }}} */

/* {{{ move return edges to return block */
typedef struct _node_list {
  t_bbl * node;
  struct _node_list * next;
} nodelist;

static nodelist * returnblocks = NULL;

static t_bool function_helper_find_returnblocks(t_node * current, t_node * parent,void * data) {
  /* if one of the successor edges of this node is a return edge */
  /* then add this node to the list */
  nodelist *p;
  t_cfg_edge * edge;

  BBL_FOREACH_SUCC_EDGE(current,edge) 
  {
    if (EDGE_CAT(edge) == ET_RETURN) 
    {
      /* bingo! */
      p = Malloc(sizeof(nodelist));
      p->node = T_BBL(current);
      p->next = returnblocks;
      returnblocks = p;
      return FALSE;
    }
  }
  return FALSE;
}

/*!
 *
 * \param f 
 *
 * Traverses through a function and creates a list of all return blocks of that
 * function once this has been found a dummy return block is added that
 * replaces all of the other return blocks. This function uses two helper functions: 
 * */
t_uint32 FunctionMoveReturnEdgesToReturnBlock(t_function * f)
{
  t_uint32 ret=0;
  t_section * code=FUN_SECTION(f);
  nodelist *temp;
  t_graph * graph = (t_graph *) code->data; 

  GraphDFTraversalWithCheckAndData((t_node *) FUN_BBL_FIRST(f), function_helper_find_returnblocks, (t_uint16) ~(ET_INTERPROC),NULL);

  /* okay, all the return blocks are in returnblocks */
  /* the dummy return block was already allocated in a previous phase, so we can now link it to the return blocks */

  temp = returnblocks;
  while (temp != NULL) 
  {
    t_cfg_edge * edge, * tmp;
    t_node *currnode = (t_node *) temp->node;


    BBL_FOREACH_SUCC_EDGE_SAFE(currnode,edge,tmp)
    {
      if ((CFG_EDGE_TAIL(edge) == f->cfg->exit_hell_node) && (EDGE_CAT(edge) == ET_RETURN) && (edge->corr_edge == NULL))

	{
        ret++;
        CfgEdgeKill( edge);
        edge=(t_cfg_edge *) GraphAppendNode(graph, currnode, (t_node *) FUN_BBL_LAST(f), ET_JUMP);	
        edge->cfg=(t_cfg *) graph;
        edge->refcount=1;
        /* this shouldn't be ET_FALLTHROUGH, because in the case
           of a conditional return there already is a
           fallthrough edge! */
      }
      else if (EDGE_CAT(edge) == ET_RETURN) 
      {
	t_cfg_edge * retu;
	if (!IsHellNode(currnode)) FATAL(("How did we get here? @ieB @E",currnode,edge));

        retu=
          /* put the return node between the node and its successor */
          (t_cfg_edge *) GraphAppendNode(graph, (t_node *) FUN_BBL_LAST(f), (t_node *) CFG_EDGE_TAIL(edge), ET_RETURN);
        /* this shouldn't be ET_FALLTHROUGH, because in the case
           of a conditional return there already is a
           fallthrough edge! */
        /* If there is a corresponding edge, patch it! */
        if (edge->corr_edge) 
          edge->corr_edge->corr_edge=retu;
        else 
          FATAL(("EDGE Without corresponding!"));
        CfgEdgeKill( edge);
        edge=(t_cfg_edge *) GraphAppendNode(graph, currnode, (t_node *) FUN_BBL_LAST(f), ET_JUMP);	
        edge->cfg=(t_cfg *) graph;
        edge->refcount=1;
        retu->cfg=(t_cfg *) graph;
        retu->refcount=1;
        ret++;
        break; /* since there can normally be only one return edge from a node, we can end the for loop here */
      }
      else
      {
        /* With the exception of hell, the only other edge that should come out of a return node is a fallthrough edge(in the case of a conditional return) */
        if ((currnode!=(t_node *) f->cfg->hell_node)&&(EDGE_CAT(edge)!=ET_FALLTHROUGH)&&(EDGE_CAT(edge)!=ET_IPFALLTHRU))
        {
          DiabloPrint(stdout, "entry @B current @ieB %p\n",FUN_BBL_FIRST(f),currnode,currnode);
          FATAL(("found so called return block and outgoing edge is of type %d\n",EDGE_CAT(edge)));
        }
      }
    }
    temp = temp->next;
  }

  /* clean up returnblocks so that it's ready for the next usage */
  while (returnblocks != NULL)
  {
    temp = returnblocks;
    returnblocks = returnblocks->next;
    Free(temp);
  }
  return ret;
}
/* }}} */

/* {{{ associate blocks to function */
/*! 
 * \param current  a node 
 * \param parent the parent node of the current node
 * \param a function casted to void 
 * \return 1 if we do not need to follow successors of this node, else 0
 * 
 * This (helper) function inserts basic blocks inside a function, is called by a DFS traverse 
 */ 

static t_bool function_helper_associate_blocks(t_node * current, t_node * parent,void * data) 
{
  t_bbl * bbl = (t_bbl *) current;
  t_function * function=(t_function *) data;

  /* Do not assign Hellnodes to a function, and skip all Hell-successors */
  if (IsHellNode(bbl)) 
  {
    return TRUE;
  }
  /* entry point should already be inserted, but we want to do all successors (so return 0) */
  if (bbl==FUN_BBL_FIRST(function)) 
  {
    return FALSE;
  }
  /* Do not reassign already assigned blocks (and skip all successors of this node) */
  if (bbl->fun != 0)
  {
    return TRUE; 
  }
  /* If this basicblock is the start of a function, skip it and skip all successors */
  if (BBL_ATTRIB(bbl) & BBL_IS_FUN_ENTRY)
  {
    return TRUE;
  }

  /* otherwise, insert this basic block in the function */

  BblInsertInFunction(bbl,function);

  /* Continue with the successors */
  return FALSE;
}

/*! 
 * 
 * \param f a function 
 * 
 * traverse the basic blocks of a function and store in each basic block 
 * the function to which it belongs. make a lot of noise if a basic block 
 * belongs to 2 different functions */

void FunctionAssociateBbls(t_function * f)
{
  if (FUN_BBL_FIRST(f)) BBL_ATTRIB(FUN_BBL_FIRST(f)) &=~ BBL_IS_FUN_ENTRY;
  GraphDFTraversalWithCheckAndData(
      (t_node *) FUN_BBL_FIRST(f), 
      function_helper_associate_blocks, 
      (t_uint16) ~(ET_INTERPROC),
      (void *) f 
      );
  if (FUN_BBL_FIRST(f)) BBL_ATTRIB(FUN_BBL_FIRST(f)) |= BBL_IS_FUN_ENTRY;
}
/* }}} */

/* {{{ Make the return edges of a function point to the right blocks */
t_uint32 FunctionAdjustReturnEdges(t_function * f) 
{
  t_uint32 ret=0;
  t_cfg * flowgraph=f->cfg;
  t_cfg_edge * edge;
  t_cfg_edge * tmp;
  t_bool hell_to_call_hell = FALSE;

  BBL_FOREACH_PRED_EDGE_SAFE(FUN_BBL_FIRST(f),edge,tmp) 
  {
    if (EDGE_CAT(edge) != ET_CALL) continue;
    if (CFG_EDGE_HEAD(edge) == flowgraph->unique_entry_node) continue;/* TODO: check if this is correct */
    if (CFG_EDGE_TAIL(edge) == flowgraph->call_hell_node)
    {
      /* Only one exisxting call to call hell is allowed at this point, ie the one from hell */
      if(hell_to_call_hell)
        FATAL(("Implement calls to call hell!"));
      if(CFG_EDGE_HEAD(edge) == flowgraph->hell_node)
	hell_to_call_hell = TRUE;
      else
        FATAL(("Implement calls to call hell!"));
    }
    else if (CFG_EDGE_TAIL(edge) == flowgraph->hell_node)  /* Call to hell */ 
    {
      if (CFG_EDGE_HEAD(edge) == flowgraph->call_hell_node )
      	{
	  continue;
	}
      else
	  /* move call from hell -> call_hell en return return from hell -> return_hell 
	   * BUT ONLY IF the call comes from compiler generated code. Otherwise, we can't be
	   * certain that the called function will obey the calling conventions */
	if (!(BBL_ATTRIB(CFG_EDGE_HEAD(edge)) & BBL_IS_HANDWRITTEN_ASSEMBLY))
	{
	  t_cfg_edge * call=(t_cfg_edge *) GraphAppendNode(T_GRAPH(flowgraph), (t_node *) (CFG_EDGE_HEAD(edge)), (t_node *) flowgraph->call_hell_node ,ET_CALL);
	  t_cfg_edge * retu=(t_cfg_edge *) GraphAppendNode(T_GRAPH(flowgraph), (t_node *) flowgraph->exit_call_hell_node, (t_node *)CFG_EDGE_TAIL(edge->corr_edge) ,ET_RETURN);
	  CfgEdgeKill(edge->corr_edge);
	  CfgEdgeKill(edge);
	  call->corr_edge=retu;
	  retu->corr_edge=call;
	  call->cfg=retu->cfg=flowgraph;
	  call->refcount=retu->refcount=1;
	  ret++;
	}
    }
    else if (CFG_EDGE_HEAD(edge) == flowgraph->hell_node) /* Calls from hell */
    {
      t_uint32 i;

      /* call from hell remains unchanged, but return is edge from called function to hell ! */

      t_cfg_edge *retu=(t_cfg_edge *) GraphAppendNode(T_GRAPH(flowgraph),(t_node*) (FUN_BBL_LAST(BBL_FUN(CFG_EDGE_TAIL(edge)))),(t_node*) flowgraph->exit_hell_node ,ET_RETURN);

      /* sanity checks */
      VERIFY_FASSERT(edge->corr_edge,("No corresponding edge for call @E!",edge));
      VERIFY_FASSERT(edge->corr_edge->cfg,("No cfg!"));
      VERIFY_FASSERT(edge->refcount == edge->corr_edge->refcount,("Refcount of call @E and corresponding return @E do not match\n",edge,edge->corr_edge));
      retu->corr_edge=edge;
      retu->cfg=flowgraph;
      /* the new edge should have the same refcount as the old edge */
      retu->refcount=edge->corr_edge->refcount;
      ret++;

      /* the old edge should be removed completely, so we must kill it refcount times */
      {
        t_uint32 rc = edge->corr_edge->refcount;
        for (i=0; i<rc; i++)
          CfgEdgeKill(edge->corr_edge);
      }
      edge->corr_edge=retu;
    }
    else /* normal edges */
    {
      t_cfg_edge * retu=NULL;

      /* sanity checks */
      VERIFY_FASSERT(edge->corr_edge,("No corresponding edge!\n"));
      VERIFY_FASSERT(edge->refcount == 1,("Implement normal calls %x with refcount > 1",edge));
      VERIFY_FASSERT(edge->corr_edge->refcount == 1,("Implement correspondings of normal calls with refcount > 1"));

      /* get the returnedge from hell, and move it! */
      retu= (t_cfg_edge *) 
	GraphAppendNode(T_GRAPH(flowgraph), (t_node *)(FUN_BBL_LAST(BBL_FUN(CFG_EDGE_TAIL(edge)))), (t_node *) CFG_EDGE_TAIL(edge->corr_edge), ET_RETURN);

      if (edge->corr_edge == tmp)	/* this can happen if we encounter a call to the next instruction */
	tmp = CFG_EDGE_PRED_NEXT(tmp);

      CfgEdgeKill(edge->corr_edge);
      edge->corr_edge=retu;
      retu->corr_edge=edge;
      retu->cfg=flowgraph;
      retu->refcount=1;
      ret++;
    }
  }
  return ret;
}
/* }}} */

/* {{{ returns the dummy exit block of a function, or NULL if this block has been freed */
t_bbl * FunctionGetExitBlock(t_function * fun)
{
  t_bbl * ret = FUN_BBL_LAST(fun);
  if (BBL_ATTRIB(ret) & BBL_IS_EXITBLOCK)
    return ret;

  return NULL;
}
/* }}} */

/* {{{ Look up the function in a list of unbehaved functions, return TRUE if found */
t_bool FunctionBehaves(t_function * fun)
{
  
  char * name = fun ? FUN_NAME(fun) : NULL;
  char ** badfuns = fun ? FUN_SECTION(fun)->obj->description->unbehaved_funs : NULL;
  t_uint32 i = 0;

  if(!fun) return TRUE;
  if(!name) return TRUE;
  if(!badfuns) return TRUE;
  while(strcmp(badfuns[i],""))
  {
    if(!strcmp(name,badfuns[i++])) return FALSE;
  }
  return TRUE;
}
/* }}} */
    
t_bool FunIsCalledByHell(t_function * fun)
{
  t_bbl * bbl = FUN_BBL_FIRST(fun);
  t_cfg_edge * edge;
  BBL_FOREACH_PRED_EDGE(bbl,edge)
    if (EDGE_CAT(edge) == ET_CALL)
    {
      if (IsHellNode(CFG_EDGE_HEAD(edge)))
	return TRUE;
    }
  return FALSE;
}

/* vim: set shiftwidth=2 foldmethod=marker: */
