/*
 * Copyright 2001,2002,2003: 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: Headers for function manipulation 
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002,2003 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) 
 ***********************************************************/


/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup FUNCTIONS Functions 
 *
 * Functions in Diablo are sets of blocks, with a unique (virtual)
 * exit block (i.e. bbl_last). There are two main reason to have them:
 * collect information about them to threat callees as atomic
 * operations in the analysis of a caller, and to structure the
 * algorithms and our reasoning about transformations.
 *
 */
/*@{ */

#ifndef MY_FUNCTION_TYPEDEFS
#define MY_FUNCTION_TYPEDEFS
typedef struct _t_function t_function;

/*! Different function types */
typedef enum 
{
   FT_NORMAL,
   FT_HELL
} t_function_type;

#endif




#ifndef MY_FUNCTION_H
#define MY_FUNCTION_H
#include "../datastructures/data_graph.h"
#include "diablo_object.h"
#include "diablo_flowgraph.h"
#include "diablo_bbl.h"
#include "diablo_procstate.h"
#include "diablo_edge.h"
#include "diablo_loop.h"

/*! This structure represent functions in Diablo */
struct _t_function {
   /*! The flowgraph that holds this function */
   t_cfg * cfg;
   /*! The name of the function (can be NULL) */
   t_string name;
   /*! Pointer used to chain all functions in a double-linked list in the flowgraph  */
   struct _t_function * next; 
   /*! Pointer used to chain all functions in a double-linked list in the flowgraph  */
   struct _t_function * prev;

#ifdef INSTRUMENTATION
   struct _t_function * instrument_next; 
   struct _t_function * instrument_prev;
#endif

   /*! Pointer used to chain all marked functions in a linked list */ 
   t_function * next_marked;
   /*! Pointer to the first marked edge in the function (head of a linked list) */
   t_cfg_edge * marked_edges;
   /*! Pointer to the first marked basic block in the function (head of a linked list) */
   t_bbl * marked_bbls;
   /*!Contain platform independent function flags (see \ref FUN_FLAGS) */
   t_uint32 function_flags;       
   /*! The entrypoint of the function = the bbl used for calls */
   t_bbl * bbl_first;
   /*! The unique exitblock of the function */    
   t_bbl * bbl_last;
   t_regset regs_through;
   t_regset regs_saved;
   t_regset regs_changed;
   t_regset regs_used;
   t_regset stack_copies_in;
   t_regset stack_copies_out;
   t_procstate * procstate_in;

   t_loopref * loops;

   void * tmp;

   /*some ia64 specific stuff: this caches the alloc-values: 
    * size of locals, size of outputs, size of total frame and size of rotating registers*/
   /*if the alloc is not known or if there are duplicate allocs, the validAlloc is FALSE*/
#ifdef IA64_SUPPORT
   t_bool  cleanAlloc;
   t_bool  validAlloc;
   t_uint8 sol;
   t_uint8 soo;
   t_uint8 sof;
   t_uint8 sor;

#endif  

};

#define FUN_REGS_THROUGH(fun)       (T_FUN(fun)->regs_through)
#define FUN_REGS_SAVED(fun)         (T_FUN(fun)->regs_saved)
#define FUN_REGS_CHANGED(fun)       (T_FUN(fun)->regs_changed)
#define FUN_REGS_USED(fun)          (T_FUN(fun)->regs_used)
#define FUN_STACK_COPIES_IN(fun)    (T_FUN(fun)->stack_copies_in)
#define FUN_STACK_COPIES_OUT(fun)   (T_FUN(fun)->stack_copies_out)
/*! Get the flowgraph this function belongs to */
#define FUN_CFG(fun)                (T_FUN(fun)->cfg)
#define FUN_SECTION(fun)             (CFG_SECTION(FUN_CFG(fun)))
/*! Get the flags of this function: flags is a bitvector, bits are described in \ref FUN_FLAGS*/
#define FUN_FLAGS(fun)              (T_FUN(fun)->function_flags)
#define FUN_NEXT_MARKED_IN_CFG(fun) (T_FUN(fun)->next_marked)
/*! Returns the head of the list of marked basic blocks in the function */
#define FUN_MARKED_BBLS(fun)        (T_FUN(fun)->marked_bbls)
/*! Returns the head of the list of marked edges the function */
#define FUN_MARKED_EDGES(fun)       (T_FUN(fun)->marked_edges)
/*! Get the next function in the flowgraph (double - linked list) */
#define FUN_NEXT(fun)               (T_FUN(fun)->next)
/*! Get the previous function in the flowgraph (double - linked list) */
#define FUN_PREV(fun)               (T_FUN(fun)->prev)
/*! Get the first basic block of the function = function entry (calls go to
 * this block) */
#define FUN_BBL_FIRST(fun)          (T_FUN(fun)->bbl_first)
/*! Get the last basic block of the function = unique sink node for returns =
 * an empty block (all returns come from this block) */
#define FUN_BBL_LAST(fun)           (T_FUN(fun)->bbl_last)
/*! Get the name of the function */
#define FUN_NAME(fun)               (T_FUN(fun)->name)
/*! A linked list of loops that this function is embedded in */
#define FUN_LOOPS(fun)		    (T_FUN(fun)->loops)





/*some ia-64 alloc access macros*/
#ifdef IA64_SUPPORT


/*checks if this function has a valid alloc*/
#define FUN_ISALLOCVALID(fun)      (T_FUN(fun)->validAlloc)

/*checks if this function has a clean alloc: namely if it is the first instruction in the function*/
#define FUN_ISALLOCCLEAN(fun)      (T_FUN(fun)->cleanAlloc)

/*special marker to invalidate sol, ...*/
#define IA64_INVALID_ALLOC 0xFF

/*number of locals: is IA64_INVALID_ALLOC if not valid*/
#define FUN_SOL(fun)               (T_FUN(fun)->sol)
/*number of outputs: is IA64_INVALID_ALLOC if not valid*/
#define FUN_SOO(fun)               (T_FUN(fun)->soo)
/*total size of the frame: is IA64_INVALID_ALLOC if not valid*/
#define FUN_SOF(fun)               (T_FUN(fun)->sof)
/*number of rotating registers: is IA64_INVALID_ALLOC if not valid*/
#define FUN_SOR(fun)               (T_FUN(fun)->sor)


#endif



/*! \todo Just use the function flags for FUN_IS_HELL */
#define FUN_IS_HELL(fun)            (T_FUN(fun)->bbl_first && IsHellNode(T_FUN(fun)->bbl_first))

/*! \defgroup FUN_ALLOC Constructors and destructors */
/*@{*/
t_function * FunctionNew(t_bbl * entrypoint, t_string name, t_function_type tp);
void FunctionKill(t_function * fun);
/*@}*/
/*! \defgroup FUN_UTILS Functions working on functions */
/*@{*/
void FunctionAssociateBbls(t_function * f);
t_uint32 FunctionAdjustReturnEdges(t_function * f);
t_uint32 FunctionMoveReturnEdgesToReturnBlock(t_function * f);
t_bbl * FunctionGetExitBlock(t_function * fun);
t_bool FunctionBehaves(t_function * fun);
t_bool FunIsCalledByHell(t_function * fun);
/*@}*/

#define FOREACH_FUNCTION(section,func) for (func=((t_cfg *) ((section)->data))->fun_first; func!=NULL; func=func->next) 

/*! \defgroup FUN_FLAGS Function flags */
/*@{*/
/*! Function is a leaf function */
#define FF_IS_LEAF		0x1
/*! Function is marked */
#define FF_IS_MARKED		0x2	
/*! Function was an exported function (in the original objectfile) */
#define FF_IS_EXPORTED          0x4
/*! Function uses explicit frame pointer */
#define FF_HAS_FRAMEPOINTER	0x8
/*! Function is global (visible from outside its own module) */
#define FF_IS_GLOBAL            0x10
/*! Function is local (not visible from outside its own module) */
#define FF_IS_LOCAL             0x20
/*! Function has incoming edges other than calls */
#define FF_HAS_INCOMING_IPJUMP  0x40
/*! Is this flag is set, the function belongs to a loop entirely as
  soon as one of its return edges belongs to the loop */
#define FF_SPECIAL_LOOP_FUN     0x80
/*! If this flag is set, it means the function will be a
  FF_SPECIAL_LOOP_FUN iff all of its callees are either
  FF_SPECIAL_LOOP_FUN and/or FF_ENTRY_REACHABLE_FROM_EXIT */
#define FF_COND_LOOP_REACH      0x100
/*! If this flag is set, it means the entry point is (backwards)
  reachable from the exit point through the function itself,
  FF_COND_LOOP_REACH or FF_SPECIAL_LOOP_FUN callees */
#define FF_ENTRY_REACHABLE_FROM_EXIT 0x200

/*@}*/

#define FUNCTION_DESCRIPTOR(fun)     (T_FUN(fun)->cfg->sec->obj->description)

#define FUN_IS_GLOBAL(fun)      (FUN_FLAGS(fun) & FF_IS_GLOBAL)
#define FUN_IS_LOCAL(fun)       (FUN_FLAGS(fun) & FF_IS_LOCAL)


#define T_FUN(fun)        ((t_function*)(fun))

#endif
/*@}*/
/*@}*/
