/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: Header for instruction manipulation 
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 ***********************************************************/


/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup INSTRUCTIONS Generic instructions */
/*@{ */
#ifndef DIABLO_INS_TYPEDEFS
#define DIABLO_INS_TYPEDEFS
#include "../simple_types.h"
typedef t_uint32 t_instruction_flags;
typedef struct _t_ins t_ins;
typedef struct _t_inslist2
{
  t_ins * ins;
  struct _t_inslist2 * next;
} t_inslist2;
#endif

#ifndef DIABLO_INS_H
#define DIABLO_INS_H
#include "diablo_section.h"
#include "diablo_reloc.h"
#include "diablo_relocatable.h"
#include "diablo_regset.h"
#include "diablo_bbl.h"
#include "diablo_procstate.h"

#ifdef STEGO
typedef struct _t_ins_alternatives t_ins_alternatives;
struct _t_ins_alternatives {
  /*number of alternatives*/
  t_uint8 count;
  /*alternatives*/
  t_ins ** ins;
};
#endif

/** This structure represents generic instructions inside Diablo. Architecture dependent instructions are 
 * produced by including this structure as the first member of the architecture dependent version */
struct _t_ins 
{
   t_relocatable relocatable;
   /**  the previous instruction in the list. If this is an instruction in a basic block, this field 
    * points to the prev instruction in the basic block, or NULL if this is the first instruction. Otherwise 
    * it is in the linked list ...  */
  /*   struct _t_ins * prev; */
   /*! The next instruction in the list. \sa { prev } */ 
  /*   struct _t_ins * next; */
   /** If this instruction is duplicated, this field points to the copy. 
    * \warning After you kill the original instruction, this field is invalid! */
  /*   struct _t_ins * copy; */
   /*! A regset representing the defined registers for an instruction */
   t_regset defined;
   /*! A regset representing the used registers for an instruction */
   t_regset used;
   /*! The relocs that change this instruction */
  /*   t_reloc_ref * refs;*/
   /*! The relocs that point to this instruction */
  /*   t_reloc_ref * refby;*/
   /*! Flags with extra information about the instruction */  
   t_section * sec;		 
   t_instruction_flags Attributes;  
   t_uint16 Type;	
   /*! The basic block to which the instruction belongs (else NULL) */ 
   /*   t_bbl * bbl;		 */
   /*! The section to which the instruction belongs (only if bbl-member is NULL!) */
   /*! Type of the instruction */ 
   /*! After constant propagation: what is stored in the used registers */
  /*   t_regstate * used_regstates;*/
   /*! After constant propagation: what is stored in the defined registers */
  /*   t_regstate * defined_regstates;*/
  /*   void * tmp;*/

  t_uint32 execution_count;
#ifdef STEGO
  t_ins_alternatives * alt;
  t_bool already_used;
  /*avoid two equal ordenings of bbl*/
  t_uint32 nr;
#endif
}; 

/*! \defgroup INS_ATTRIB Instruction flags */
/*@{*/
/*! The instruction is actualy a data element stored in the code section */
#define IF_DATA		       0x1
/*! The instruction is a bbl leader */ 
#define IF_BBL_LEADER	       0x2
/*! The instruction is the start of a function, because there is a (function) symbol pointing to it or there are calls to this instruction */ 
#define IF_FUNCTION_ENTRY_OBSOLETE      0x4
/*! The instruction is contained in a basic block (depends on the current state of the object)*/
#define IF_IN_BBL	       0x8
/*! The address of the instruction can be loaded or computed somewhere */
#define IF_ADDRESS_TAKEN       0x10
/*! Constant propagation thinks the instruction can be executed */
#define IF_EXECED              0x20
/*! Constant propagation thinks the instruction is always executed (this means that for a conditional instruction, the condition can be removed) */
#define IF_ALWAYS_EXECED       0x40
/*! The instruction is dead (it can be killed, but it wasn't possible to free it's memory yet) */ 
#define IF_DEAD                0x80
/*! The instruction defines the PC */ 
#define IF_PCDEF               0x100
/*! The instruction is actually a data element stored in the code section, and it is part of a switch table */
#define IF_SWITCHTABLE         0x200
/*! The instruction has changed since we began compacting the program TODO: should be made obsolete */
#define IF_CHANGED             0x400
/*! The instruction is conditional   */
#define IF_CONDITIONAL         0x800
/*! The instruction is a dummy, but has been inserted to make SecGetInsByAddress work for all architectures
 *  because otherwise calculations on the instructionarray would be wrong for variable length instruction
 *  architectures a the ia32 */
#define IF_DUMMY               0x1000
#define IF_DELAYSLOT           0x2000
#define IF_SWITCHJUMP          0x4000
/*! The instruction can be evaluated fast during constant propagation:
 *  if any of its source operands are unknown, the result is
 *  unknown */
#define IF_FAST_CP_EVAL        0x8000
/* Branch obfuscation added; attention when we want to give a new address to this instruction */
#define IF_BRANCH_OBF 	       0x10000

/*@}*/

/*! \defgroup IT_TYPES Generic Instruction types */
/*@{*/
/*! Instruction type is unknown */
#define IT_UNKNOWN	0
/*! Branch instructions */
#define IT_BRANCH	1	
/*! Software interupts */
#define IT_SWI		2	
/*! Normal dataprocessing instructions (add, sub, ...) */ 
#define IT_DATAPROC	3	
/*! Multiplies */
#define IT_MUL		4	
/*! Loads */
#define IT_LOAD		5	
/*! Stores */
#define IT_STORE	6	
/*! Multiple loads and stores */
#define IT_LOAD_MULTIPLE	7
#define IT_STORE_MULTIPLE	8
/*! Status register enquieries */
#define IT_STATUS	9
/*! Swap instructions */
#define IT_SWAP		10	
/*! Data in the code section */
#define IT_DATA		11	
/*! Address producers and constant producers */
#define IT_CONSTS	12	
/*! No operation instruction */
#define IT_NOP		13
/*! Prefetch instruction */
#define IT_PREF		14
/*! Floating point ALU */
#define IT_FLT_ALU		15
/*! Floating point -> INS transfer*/
#define IT_FLT_INT		16
#define IT_FLT_LOAD		17
#define IT_FLT_STORE		18
#define IT_FLT_STATUS		19
#define IT_INT_FLT              20
#define IT_CALL		        21
#define IT_PSEUDO_LOAD          22
#define IT_PSEUDO_SAVE          23

/*@}*/

typedef enum {ALWAYS,NEVER,EQ,NE,LT,GT,LE,GE,UNKNOWN} EXEC_CONDITION;


/*! \defgroup INS_SELECT Macro's to access fields of the generic instruction structure  */
/*@{*/
/*! Get/set the metaflags of the instruction */
#define INS_ATTRIB(x)	        (T_INS(x))->Attributes
/*! Get/set the type of the instruction */
#define INS_TYPE(x)		(T_INS(x))->Type
/*! Get/set the bbl of this instruction (only valid when metaflag INS_IN_BBL is set)*/
#define INS_BBL(x)		(T_BBL(T_NODE(x)->succ_last))
#define INS_SECTION(x)		(T_INS(x)->sec)
#define INS_TEMP(x)		((void*)(T_NODE(x)->pred_last))
#define INS_OBJECT(x)		(SECTION_OBJECT(INS_SECTION(x)))
/*! Get/set the next instruction in the bbl if metaflag INS_IN_BBL is set otherwise, to the next instruction in the linked list of instructions*/ 
#define INS_NEXT(x)		(T_INS(T_NODE(x)->next))
/*! Get/set the prev instruction in the bbl if metaflag INS_IN_BBL is set otherwise, to the prev instruction in the linked list of instructions*/ 
#define INS_PREV(x)		(T_INS(T_NODE(x)->prev))
/*! Get/set a regset that describes the defined register for this instruction */ 
#define INS_REGS_DEF(x)		(T_INS(x))->defined
/*! Get/set a regset that describes the used register by this instruction */ 
#define INS_REGS_USE(x)		(T_INS(x))->used
/*! Get/set the address of this instruction */
#define INS_CADDRESS(x)		(T_RELOCATABLE(x))->caddress
#define INS_CSIZE(x)		(T_RELOCATABLE(x))->csize
/*! Get/set the new address of this instruction */
#define INS_OLD_ADDRESS(x)	(T_RELOCATABLE(x))->old_address
#define INS_OLD_SIZE(x)		(T_RELOCATABLE(x))->old_size
/*! Get/set the min address of this instruction */
#define INS_MIN_ADDRESS(x)	(T_RELOCATABLE(x))->min_address
#define INS_MIN_SIZE(x)		(T_RELOCATABLE(x))->min_size
/*! Get/set the array of regstates of the used registers */
#define INS_USED_REGSTATE(x)	(x)->used_regstates
/*! Get/set the array of regstates of the defined registers */
#define INS_DEFINED_REGSTATE(x)	(x)->defined_regstates
/*! Get the relocation that this ins refers to */
#define INS_REFERS_TO(x)	RELOCATABLE_REFERS_TO(T_RELOCATABLE(x))

#define INS_COPY(x)             (T_INS(T_NODE(x)->succ_first))
#define INS_EXEC_COUNT(x)       (T_INS(x)->execution_count)
/*@}*/

/*! \defgroup INS_TYPE Shorthands to check the type of an instruction */ 
/*@{*/
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is unknown */
#define INS_IS_UNKNOWN(x)       (INS_TYPE(x) == IT_UNKNOWN)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is a branch */
#define INS_IS_BRANCH(x)        (INS_TYPE(x) == IT_BRANCH)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is a software interrupt */
#define INS_IS_SWI(x)           (INS_TYPE(x) == IT_SWI)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is dataprocessing instruction */
#define INS_IS_DATAPROC(x)      (INS_TYPE(x) == IT_DATAPROC)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is multiply instruction */
#define INS_IS_MUL(x)           (INS_TYPE(x) == IT_MUL)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is a load */
#define INS_IS_LOAD(x)          (INS_TYPE(x) == IT_LOAD)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is a store */
#define INS_IS_STORE(x)         (INS_TYPE(x) == IT_STORE)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is a load/store multiple */
#define INS_IS_LOAD_MULTIPLE(x)         (INS_TYPE(x) == IT_LOAD_MULTIPLE)
#define INS_IS_STORE_MULTIPLE(x)         (INS_TYPE(x) == IT_STORE_MULTIPLE)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is a status instruction */
#define INS_IS_STATUS(x)        (INS_TYPE(x) == IT_STATUS)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is a swap instruction */
#define INS_IS_SWAP(x)          (INS_TYPE(x) == IT_SWAP)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is a swap instruction */
#define INS_IS_DATA(x)          (INS_TYPE(x) == IT_DATA)
/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the type of the instruction is an address producer */
#define INS_IS_CONSTS(x)        (INS_TYPE(x) == IT_CONSTS)

/*! \param x An instruction (automaticaly casted to a generic instruction)
 *  \return TRUE or FALSE
 *  
 *  Check if the instruction is conditional */ 
#define INS_IS_CONDITIONAL(x)   (INS_ATTRIB(x) & IF_CONDITIONAL)


/*@}*/

/*! (down)Cast to instruction */

#define T_INS(x)     ((t_ins*)(x))

/*! \defgroup INS_ALLOC Constructors and destructors */
/*@{*/
void InsKill(t_ins * ins);
void InsFree(t_ins * ins);
t_ins * InsDup(t_ins * ins);

void * InsNewForSec(t_section * sec);
void * InsNewForCfg(t_cfg * cfg);
void * InsNewForBbl(t_bbl * bbl);
/*@}*/

/* TODO */
#define AINS(x,y,z) (RealGetInsByNumber(__FILE__,__LINE__,y,z))
void* RealGetInsByNumber(char *, int,t_section * section, t_uint32 i);
t_bool InsHasSideEffect(t_ins * iins);
void InsFreeAllReferedRelocsBase(t_ins * ins);
void InsMakeNoop(t_ins * ins);
t_ins * SecGetInsByAddress(t_section * sec, t_address addr);
t_ins * InsFindMovableDownEqualInsInBbl(t_ins * search, t_bbl * bbl);
#define InsGetSection(ins)      ((INS_ATTRIB(ins) & IF_IN_BBL)?INS_BBL(ins)->cfg->sec:INS_SECTION(ins))


t_regset InsRegsLiveAfter(t_ins * ins);
t_regset InsRegsLiveBefore(t_ins * ins);
t_regset InsRegsLiveOrUsedOrDefinedBefore(t_ins * ins);

void InsFreeReferedRelocs(t_ins * ins);

t_ins * InsNextInChain(t_ins * ins);
void InsChangePlace(t_ins * first_ins, t_ins * sec_ins);


void InsKillDelayed(t_ins * ins);
void InitDelayedInsKilling();
void ApplyDelayedInsKilling();
void InsFreeReferedRelocsDelayed(t_ins * ins);


t_bool InsIsFirstInConditionalBundle(t_ins * ins);
t_uint32 ConditionalBundleSize(t_ins * first_ins);


#endif
/*@} */
/*@} */
/* vim: set shiftwidth=2 foldmethod=marker : */
