/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file routines 
 * 
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/

/*{{{ Includes*/
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "../config.h"
#include "diablo_addresses.h"
#include "diablo_object.h"
#include "diablo_object_tables.h"
#include "../simple_types.h"
#include "../diablo_options.h"

/* Utilities used: */
#include "../datastructures/data_string.h"
#include "../datastructures/data_hash.h"
#include "../utils/util_file.h"
#include "../utils/util_malloc.h"
#include "../utils/util_options.h"
#include "../utils/util_status.h"
#include "../utils/util_error.h"
#include "../utils/util_term.h"
/* Support for archives */
#include "../fileformats/ar/ar_arch.h"
/* Symbols */
#include "diablo_symbol.h"
#include "diablo_reloc.h"
#include "diablo_relocate.h"
#include "diablo_section.h"
#include "diablo_stack_machine.h"
#include "diablo_instrument.h"
#include "diablo_layout.h"
/* Maps */
#include "diablo_parsemap.h"
/* Printing */
#include "diablo_print.h"
#include "diablo_iterators.h"
/*}}}*/
#define SIGN_EXTEND(x,p) (((x)^(1<<p)) -(1 << p))
/* Static declarations to keep the compiler happy {{{*/
void ObjectMergeSubObjectSymbolTables(t_object * obj);
void ObjectAddLinkerSpecificDataAndSymbols(t_object * obj);
void ObjectConnectSubObjectSymbolTableWithMergedTable(t_object * obj);
void ObjectVerifySymbolTable(t_object * obj);
void ObjectHashElementFree(void * tf,void * data);
void RelocateCore(t_object * obj, t_bool verify);

/*}}}*/

/* Map related stuff {{{ */
/*!
 *
 * \param obj
 *
 * \return void 
*/

void ObjectReadMap(t_object * obj)
{
  obj->map=MapNew();
  ObjectAction(READ_MAP,obj,obj->name);
}

/*!
 *
 * \param obj
 *
 * \return void 
*/

void ObjectReadObjectInMap(t_object * obj)
{
  t_objects_ll list;
  
  /* obj is a parent object. initialise it's subobject cache */
  obj->subobject_cache = HashTableNew(103,0,string_hash,(my_cmpresult (*)(void *, void *))  strcmp,ObjectHashElementFree);

  /* read the objects in the map */
  list = MapLoadObjects(obj->map,obj);
}

void ObjectAppendSubObject(t_object * parent, t_object * child)
{
  if (child->next || child->prev || child==parent->mapped_last || child==parent->mapped_first)  return;
  if (parent->mapped_last)
  {
    parent->mapped_last->next=child;
    child->prev=parent->mapped_last;
    child->next=NULL;
    parent->mapped_last=child;
  }
  else if (parent->mapped_first)
  {
    FATAL(("Should not happen!"));
  }
  else
  {
    child->prev=child->next=NULL;
    parent->mapped_last=parent->mapped_first=child;
  }
}

#ifdef ALPHA_SUPPORT
/*! 
 * for the alpha, the link-order for the lita is not included in the mapfile.
 * We have to calculate the order ourselfs. This should always be possible
 * since the order of the lita's is the same as the order of the textsections
 * */

void ObjectMapPatchAlpha(t_object * obj)
{
  t_object * tmp=NULL;
  t_section * sec=SectionGetFromObjectByName(obj,".lita");
  FILE * link_order;
  char * ordername=Collate(obj->name,".lo");
  char * parse_line;
  t_address litarun;
  t_address litaend;

  if (!sec) return;
  
  litarun=SECTION_GET_OLD_ADDRESS(sec);
  litaend=GENERIC_ADD(SECTION_GET_OLD_ADDRESS(sec),SECTION_OLD_SIZE(sec));
  link_order=fopen(ordername,"r");

  if (!link_order) FATAL(("Link order file %s not found!",ordername));
 
  /* Read link order preamble */

  
  while((parse_line = GetLine(link_order)) != NULL) /* while EOF not reached */
  {
    if (PatternMatch("ld:*Pass*2*",parse_line)) break;
	Free(parse_line);   /* very important, avoids leaks */
  }

  if (!parse_line) FATAL(("Link order file corrupted"));
  Free(parse_line);
  
  printf(".lita "); GENERIC_PRINT(litarun,20); GENERIC_PRINT(SECTION_OLD_SIZE(sec),20); printf("\n");
  while((parse_line = GetLine(link_order)) != NULL) /* while EOF not reached */
  {
    Chomp(parse_line);
    if (!Chop(parse_line,':')) FATAL(("File corrupted!"));

    if (PatternMatch("*(*)",parse_line)) 
    {
      t_string dup=StringDup(parse_line);
      t_string lib_name=strtok(dup,"(");
      t_string obj_name=strtok(NULL,")");
      t_string name;
      VERIFY_FASSERT(obj_name,("Illegal objectname in link order: %s",parse_line));
      VERIFY_FASSERT(lib_name,("Illegal objectname in link order: %s",parse_line));
      name=Collate3(lib_name,":",obj_name);
      Free(dup);
      tmp=ObjectGetFromCache(name,obj);
      Free(name);
      if (!tmp) tmp=ObjectGet(name,obj); 
      if (!tmp) FATAL(("Could not find %s named in link order!",parse_line));
    }
    else
    {
      tmp=ObjectGetFromCache(parse_line,obj);
      if (!tmp) 
	{
	  Free(parse_line);
	  continue;
	  FATAL(("%s not in cache",parse_line));
	}
    }
    
    
      
    sec=SectionGetFromObjectByName(tmp,".lita");
    if (sec)
    {
      MAP_SECTION(sec,".lita",SectionGetFromObjectByName(obj,".lita"));
      SectionSetOldAddressSuper(sec,litarun); 
      SectionSetCAddressSuper(sec,litarun); 
      printf("       .lita ");
      GENERIC_PRINT(litarun,10); 
      litarun=GENERIC_ADD(litarun,SECTION_OLD_SIZE(sec)); 
      GENERIC_PRINT(SECTION_OLD_SIZE(sec),10); 
      printf("%s\n",tmp->name);
    }
    Free(parse_line);
  }
  if (!GENERIC_EQ(litarun,litaend)) FATAL(("Something went wrong in the lita construction!"));
}
#endif
/* }}} */
/* Linking functions {{{ */

extern FILE * LinkerScriptParserin;
extern t_object * parser_object;
/* FUN : ObjectBuildUnifiedSymbolTable
 * PAR : the parent object 
 * RET : nothing 
 * DESC: This function build the sub_symbol table */ 


void ObjectBuildUnifiedSymbolTableAndRelocTable(t_object * obj)
{
   if(!obj->map) FATAL(("There is no map to build the global symbol table!\n"));

   ObjectMergeSubObjectSymbolTables(obj);

   ObjectAddLinkerSpecificDataAndSymbols(obj);
   ObjectConnectSubObjectSymbolTableWithMergedTable(obj);

   if (obj->symbol_table) ObjectVerifySymbolTable(obj);
   ObjectCreateDataOrCodeTable(obj);
   if (obj->symbol_table) ObjectBuildRelocTableFromSubObjects(obj);
}

void ObjectMergeSubObjectSymbolTables(t_object * obj)
{
  t_object * tmp2, * tmp;
  t_section * sec;
  t_uint32 tel;
  t_symbol * sym;
  obj->sub_symbol_table=SymbolTableNew(obj);

  /* first walk over the map and indicate all "gaps" in the code sections as data.
   * this is necessary because the linker doesn't indicate it's filler bytes as data.
   */
  {
    t_section * sec;
    t_uint32 nsubs,i;
    t_section ** subs;

    for (tel=0; tel<obj->ncodes; tel++)
    {
      sec = obj->code[tel];
      subs = SectionGetSubsections(sec,&nsubs);
      if (!subs) continue;

      /* only indicate gaps between subsections of the same section.
       * gaps between different parent sections will be taken care of
       * by ObjectMergeCodeSections */
      for (i=1; i<nsubs; i++)
      {
	t_address endprev = GENERIC_ADD(SECTION_GET_CADDRESS(subs[i-1]),SECTION_CSIZE(subs[i-1]));
	if (GENERIC_LT(endprev,SECTION_GET_CADDRESS(subs[i])))
	{
	  /* yep, there is a gap: add a $d symbol to the beginning of the gap */
	  SymbolTableAddSymbolToSection(obj->sub_symbol_table, LOCAL_SYMBOL, T_RELOCATABLE(sec), endprev, "$d", TRUE);
	}
      }

      Free(subs);
    }
  }




  OBJECT_FOREACH_SUBOBJECT(obj,tmp,tmp2)
  {
    t_address generic;

    /* First of all, insert a symbol called $a for every code section in the binary, and $d for all the rest. 
     * This will be used further on to compute the $switch symbol (i.e. all places where a code section contains 
     * data and all places where this stops). These might be duplicate, but they will be removed after sorting 
     * {{{ */
    if (tmp->ncodes)
      for (tel=0; tel<tmp->ncodes; tel++)
      {
	t_symbol_he * thumb = ( t_symbol_he *) HashLookup(tmp->symbol_table->hash_table,"$t");
	sec=tmp->code[tel];
	if (!SECTION_IS_MAPPED(sec)) continue; /* Skip sections removed by the linker */
	while(thumb)
	{
	  if(thumb->sym->base == T_RELOCATABLE(sec) && GENERIC_EQ(thumb->sym->offset_from_start,NULL_GENERIC))
	  {
	    break;
	  }
	  thumb = ( t_symbol_he *) thumb->node.equal;
	}
	if(!thumb)
	{
	  SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,LOCAL_SYMBOL,T_RELOCATABLE(sec),SECTION_GET_CADDRESS(sec),"$a",TRUE);
	}
	T_RELOCATABLE(sec)->type=RT_SUBSECTION;
      }
    if (tmp->ndatas)
      for (tel=0; tel<tmp->ndatas; tel++)
      {
        sec=tmp->data[tel];
        if (!SECTION_IS_MAPPED(sec)) continue; /* Skip sections removed by the linker */
        SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,LOCAL_SYMBOL,T_RELOCATABLE(sec),SECTION_GET_CADDRESS(sec),"$d",TRUE);
        T_RELOCATABLE(sec)->type=RT_SUBSECTION;
      }
    if (tmp->nnotes)
      for (tel=0; tel<tmp->nnotes; tel++)
      {
	sec=tmp->note[tel];
	if (!SECTION_IS_MAPPED(sec)) continue; /* Skip sections removed by the linker */
	SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,LOCAL_SYMBOL,T_RELOCATABLE(sec),SECTION_GET_CADDRESS(sec),"$d",TRUE);
	T_RELOCATABLE(sec)->type=RT_SUBSECTION;
      }
    if (tmp->nbsss)
      for (tel=0; tel<tmp->nbsss; tel++)
      {
        sec=tmp->bss[tel];
        if (!SECTION_IS_MAPPED(sec)) continue; /* Skip sections removed by the linker */
        SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,LOCAL_SYMBOL,T_RELOCATABLE(sec),SECTION_GET_CADDRESS(sec),"$d",TRUE);
        T_RELOCATABLE(sec)->type=RT_SUBSECTION;
      }
    if (tmp->nrodatas)
      for (tel=0; tel<tmp->nrodatas; tel++)
      {
        sec=tmp->rodata[tel];
        if (!SECTION_IS_MAPPED(sec)) continue; /* Skip sections removed by the linker */
        SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,LOCAL_SYMBOL,T_RELOCATABLE(sec),SECTION_GET_CADDRESS(sec),"$d",TRUE);
        T_RELOCATABLE(sec)->type=RT_SUBSECTION;
      }
    /* }}}*/

    /* Merge all defined symbols of the original object in the new symboltable. 
     * Add common sections for common symbols 
     * {{{ */
    for (sym=tmp->symbol_table->first; sym!=NULL; sym=sym->next)
    {
      if (sym->to_type==TO_UNDEFINED || sym->to_type==TO_REMOVED) 
      { 
        t_symbol * tmp = SymbolTableAddUndefinedSymbol(obj->sub_symbol_table,sym->name);
        tmp->addend = sym->addend;
        continue; 
      } 
      else if ((sym->to_type!=TO_COMMON)&&(sym->to_type!=TO_NONE))
      {
	/* Skip the linker removed sections !  Probably OBSOLETE {{{ */
	generic=GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(sym->base)),sym->offset_from_start);
        if ((!(SECTION_IS_MAPPED(T_SECTION(sym->base))) && (!GENERICISNULL(SECTION_CSIZE(T_SECTION(sym->base)))))) 
	{  
	  t_symbol *ret=SymbolTableAddSymbolToSection(obj->sub_symbol_table,sym->type,sym->base,generic,sym->name,sym->local);
	  sym->mapped=ret;
	  ret->addend = sym->addend;
          continue; 
        } 
	/* }}} */
        
       
	/* All we need to do is change the to_type and the current value */
        if (strncmp(sym->name,"$Sub$$",6)==0) /* This is arm specific */
        {
          t_symbol * ret=SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,sym->type,sym->base,generic,sym->name+6,sym->local);
          sym->mapped=ret;
          ret->addend = sym->addend;
        }
        else
        {
          if ((sym->type==GLOBAL_SYMBOL || sym->type==FUNCTION_SYMBOL) && (sym->local == 0))
          {
            t_symbol * tsym=SymbolTableGetFirstSymbolWithName(obj->symbol_table,sym->name);
            t_symbol * ret;
#ifdef ECOFF_SUPPORT
#ifdef INSTRUMENTATION
	    //#ifndef INSTRUMENTATION
            if ((strcmp(sym->name,"$a")==0) || (strcmp(sym->name,"$d")==0) || ( (tsym) && (GENERIC_EQ(generic,tsym->current_address))))
#else
	    /* for instrumented code all sections have been moved, so we can no longer check the symbol addresses */
            if ((strcmp(sym->name,"$a")==0) || (strcmp(sym->name,"$d")==0) || tsym)
#endif
#else
#ifndef INSTRUMENTATION
            if ((strcmp(sym->name,"$a")==0) || (strcmp(sym->name,"$d")==0) || ( (tsym) && (GENERIC_EQ(generic,tsym->current_address))))
#else
	    /* for instrumented code all sections have been moved, so we can no longer check the symbol addresses */
            if ((strcmp(sym->name,"$a")==0) || (strcmp(sym->name,"$d")==0) || tsym)
#endif
#endif
            {
              ret=SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,sym->type,sym->base,generic,sym->name,sym->local);
              sym->mapped=ret;
              ret->addend = sym->addend;
            }
	    else if (!tsym)
	    {
	      ret=SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,sym->type,sym->base,generic,sym->name,sym->local);
	      sym->mapped=ret;
              ret->addend = sym->addend;

	      WARNING(("Could not find sym @S in the final binary",sym));
	    }
            else
            {
              DiabloPrint(stdout, "SYMBOL CHANGED ADDRESS: @S: @G <> @G\n",sym, generic,tsym->current_address);
            }
          }
          else
          {
            t_symbol * ret=SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,sym->type,sym->base,generic,sym->name,sym->local);
            ret->addend = sym->addend;
            sym->mapped=ret;
          }
        }
      }
      else if (sym->to_type==TO_COMMON)
      {
	/* hier ipv een TO_COMMON symbool een TO_UNDEFINED symbool toevoegen
	 * ook eenmalig voor elk uniek T_COMMON symbool een sectie van de
	 * juiste grootte (= sym->current_address) toevoegen en een absoluut
	 * symbool (ABS32) dat daarnaar wijst */

        t_object * linker_obj;
        t_symbol * ret;
        t_section * tmp_sec;

        /* get linker object */
        linker_obj=ObjectGetFromCache("Linker",obj);
        if (!linker_obj) 
        {
          linker_obj=ObjectNewCached("Linker",obj);
          linker_obj->section_table=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,SectionHashNodeFree);
          obj->mapped_last->next=linker_obj;
          linker_obj->next=NULL;
          obj->mapped_last=linker_obj;
          linker_obj->reloc_table=RelocTableNew(); 
          linker_obj->symbol_table=SymbolTableNew(linker_obj);
        }

        /* check if a subsection with the right name already exists in the linker object */
        tmp_sec = SectionGetFromObjectByName(linker_obj,sym->name);
        if (!tmp_sec)
        {
          t_symbol * linked_sym = SymbolTableGetSymbolByName(obj->symbol_table,sym->name);
          VERIFY_FASSERT(linked_sym, ("Common symbol %s not found in main object's symbol table!",sym->name));

	  if (SECTION_TYPE(linked_sym->base)==BSS_SECTION)
	  {
	    /* the section doesn't exist yet: create it */
	    tmp_sec = SectionNew(linker_obj,BSS_SECTION);
	    T_RELOCATABLE(tmp_sec)->type=RT_SUBSECTION;
	    tmp_sec->name = StringDup(sym->name);

	    /* for COMMON symbols, the size is stored in sym->size, the alignment is stored in sym->current_address */
	    SECTION_OLD_SIZE(tmp_sec)=sym->size;
	    SECTION_CSIZE(tmp_sec)=sym->size;
	    SECTION_ALIGN(tmp_sec)=sym->current_address;

	    /* set section mapping information */
	    SECTION_IS_MAPPED(tmp_sec) = TRUE;
	    SECTION_PARENT_SECTION(tmp_sec) = T_SECTION(linked_sym->base);

	    /* set section address information */

	    if (strncmp(sym->name,INSTRUMENT_PREFIX,strlen(INSTRUMENT_PREFIX))==0)
	      SectionSetOldAddressSuper(tmp_sec,GENERIC_SUB(linked_sym->base->caddress,linked_sym->base->caddress)/*linked_sym->current_address*/);
	    else
	      SectionSetOldAddressSuper(tmp_sec,GENERIC_ADD(linked_sym->base->old_address,linked_sym->offset_from_start)/*linked_sym->current_address*/);

	    SectionSetCAddressSuper(tmp_sec,GENERIC_ADD(linked_sym->base->caddress,linked_sym->offset_from_start)/*linked_sym->current_address*/);

	    linker_obj->bss=Realloc(linker_obj->bss,(linker_obj->nbsss+1)*sizeof(t_section *));
	    linker_obj->bss[linker_obj->nbsss]=tmp_sec;
	    linker_obj->nbsss++;
	    HashInsert(linker_obj->section_table,(data_hash_node *) SectionHashNodeNew(tmp_sec->name,tmp_sec));

	    /* now add an absolute symbol to the subsection.
	     * this will take precedence over all the undefined symbols 
	     * that we will create from the TO_COMMON symbols in the
	     * subobjects.
	     */
	    SymbolTableAddSymbolToSection(linker_obj->symbol_table,GLOBAL_SYMBOL,T_RELOCATABLE(tmp_sec),/*linked_sym->current_address*/GENERIC_ADD(linked_sym->base->caddress,linked_sym->offset_from_start),sym->name,FALSE); 
	  }
        }

        /* map the symbol from the subobject to the subsymboltable. 
         * don't map it as a TO_COMMON symbol however: map it as a
         * TO_UNDEFINED symbol. The linker will resolve this to the 
         * absolute symbol we added to the linkerobject subsection
         */
        ret = SymbolTableAddUndefinedSymbol(obj->sub_symbol_table,sym->name);
        sym->mapped=ret;
        ret->addend = sym->addend;
      }
      else if (sym->type==WEAK_UNDEFINED_SYMBOL)
      {
        t_symbol * ret=SymbolTableAddWeakUndefinedSymbol(obj->sub_symbol_table,sym->name);
        sym->mapped=ret;
        ret->addend = sym->addend;
      }
      else
      {
        t_symbol * ret=SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,sym->current_address,sym->name,sym->local);
        sym->mapped=ret;
        ret->addend = sym->addend;
      }
    }
    /* }}} */
  }
}


void LinkerScriptParserparse();

/* {{{ Get linker script file name (command line specified or standard from linker_descriptions) */
static t_string GetLinkerScriptFilename(t_object * obj)
{
  if (global_options.linkerscript_set)
  {
    /* command line specified */
    return StringDup(global_options.linkerscript);
  }
  else
  {
#ifdef ELF_SUPPORT
    if (obj->type==OBJECT_FILE_TYPE_ELF)
    {
      if (obj->subtype==OBJECT_FILE_SUBTYPE_ARMELFLSB || obj->subtype==OBJECT_FILE_SUBTYPE_ARMELFMSB)
      {
	if (obj->linker==ADS_LINKER)
	{
	  if (global_options.linkerscriptprefix_set)
	    return FindFile(global_options.linkerscriptprefix,"arm_ads_linker.ld");
	  else
	    return Collate(DIABLO_INSTALL_PREFIX,"/linker_descriptions/arm_ads_linker.ld");
	}
	else if (obj->linker==GNU_LINKER)
	{
	  if (global_options.linkerscriptprefix_set)
	    return FindFile(global_options.linkerscriptprefix,"arm_gcc_linker.ld");
	  else
	    return Collate(DIABLO_INSTALL_PREFIX,"/linker_descriptions/arm_gcc_linker.ld");
	}
	else 
	  FATAL(("Unknown arm linker!"));
      }
      else if (obj->subtype==OBJECT_FILE_SUBTYPE_SH3ELFLSB || obj->subtype==OBJECT_FILE_SUBTYPE_SH3ELFMSB)
      {
	if (obj->linker==GNU_LINKER)
	{
	  if (global_options.linkerscriptprefix_set)
	    return FindFile(global_options.linkerscriptprefix,"sh3_gcc_linker.ld");
	  else
	    return Collate(DIABLO_INSTALL_PREFIX,"/linker_descriptions/sh3_gcc_linker.ld");
	}
	else 
	  FATAL(("Unknown sh3 linker!"));
      }
      else if (obj->subtype==OBJECT_FILE_SUBTYPE_MIPSELF)
      {
	if (obj->linker==GNU_LINKER)
	{
	  if (global_options.linkerscriptprefix_set)
	    return FindFile(global_options.linkerscriptprefix,"mips_gcc_linker.ld");
	  else
	    return Collate(DIABLO_INSTALL_PREFIX,"/linker_descriptions/mips_gcc_linker.ld");
	}
	else 
	  FATAL(("Unknown mips linker!"));
      }
      else if (obj->subtype==OBJECT_FILE_SUBTYPE_I386ELF)
      {
	if (obj->linker==GNU_LINKER)
	{
#ifndef KERNEL_OPT
	  if (global_options.linkerscriptprefix_set)
	    return FindFile(global_options.linkerscriptprefix,"i386_gcc_linker.ld");
	  else
	    return Collate(DIABLO_INSTALL_PREFIX,"/linker_descriptions/i386_gcc_linker.ld");
#else
	  if (global_options.linkerscriptprefix_set)
	    return FindFile(global_options.linkerscriptprefix,"i386_kernel.ld");
	  else
	    return Collate(DIABLO_INSTALL_PREFIX,"/linker_descriptions/i386_kernel.ld");
#endif
	}
	else 
	  FATAL(("Unknown i386 linker!"));
      }
      else if (obj->subtype==OBJECT_FILE_SUBTYPE_IA64ELF)
      {
	if (obj->linker==GNU_LINKER)
	{
	  if (global_options.linkerscriptprefix_set)
	    return FindFile(global_options.linkerscriptprefix,"ia64_gcc_linker.ld");
	  else
	    return Collate(DIABLO_INSTALL_PREFIX,"/linker_descriptions/ia64_gcc_linker.ld");
	}
	else 
	  FATAL(("Unknown ia64 linker!"));
      }
      else if (obj->subtype==OBJECT_FILE_SUBTYPE_POWERPCELFLSB || obj->subtype==OBJECT_FILE_SUBTYPE_POWERPCELFMSB)
      {
	if (obj->linker==GNU_LINKER)
	{
	  if (global_options.linkerscriptprefix_set)
	    return FindFile(global_options.linkerscriptprefix,"ppc_gcc_linker.ld");
	  else
	    return Collate(DIABLO_INSTALL_PREFIX,"/linker_descriptions/ppc_gcc_linker.ld");
	}
	else 
	  FATAL(("Unknown ppc linker!"));
      }
      else FATAL(("linker scripts not implemented for this fileformat subtype"));
    }
    else
    {
#endif
#ifdef ECOFF_SUPPORT 
    if (obj->type==OBJECT_FILE_TYPE_ECOFF)
    {
      if (global_options.linkerscriptprefix_set)
	return FindFile(global_options.linkerscriptprefix,"ecoff_cc_linker.ld");
      else
	return Collate(DIABLO_INSTALL_PREFIX,"/linker_descriptions/ecoff_cc_linker.ld");
    }
    else
    {
#endif
      WARNING(("Linker scripts not implemented for non-ELF objecttypes"));
      return NULL;
#ifdef ECOFF_SUPPORT 
    }
#endif
#ifdef ELF_SUPPORT
    }
#endif
  }
  return NULL;
}
/* }}} */

void ObjectAddLinkerSpecificDataAndSymbols(t_object * obj)
{
  t_string scriptname = GetLinkerScriptFilename(obj);
  VERIFY_FASSERT(scriptname, ("No suitable linker script found"));

  LinkerScriptParserin = fopen(scriptname,"r");
  if (!LinkerScriptParserin)
    FATAL(("Could not open linker script %s", scriptname));

  Free(scriptname);
  parser_object=obj;
  LinkerScriptParserparse();
}

void ObjectConnectSubObjectSymbolTableWithMergedTable(t_object * obj)
{
  t_object * tmp2, * tmp;
  t_symbol * sym, * msym;
  /* Map the undefined symbols to defined symbols */ 
  OBJECT_FOREACH_SUBOBJECT(obj,tmp,tmp2)
    for (sym=tmp->symbol_table->first; sym!=NULL; sym=sym->next)
      if (sym->to_type!=TO_REMOVED)
      {
	if ((sym->to_type==TO_UNDEFINED)||(sym->type==WEAK_UNDEFINED_SYMBOL)||(sym->name && sym->mapped==NULL)) 
	{ 
	  msym=SymbolTableGetSymbolByName(obj->sub_symbol_table,sym->name); 

	  if (!msym) FATAL(("Symbol %s still undefined after link!",sym->name));
	  else if (msym->base==NULL && sym->type!=WEAK_UNDEFINED_SYMBOL && msym->type!=ABS_SYMBOL) FATAL(("Symbol @S from %s mapped to @S still has no base after link!",sym,tmp->name,msym));
	  else sym->mapped=msym;
	}
      }
      else
      {
	msym=SymbolTableGetSymbolByName(obj->sub_symbol_table,sym->name);
	if (msym) sym->mapped=msym;
      }
}

void ObjectVerifySymbolTable(t_object * obj)
{
  t_symbol * sym;
  /* Verify the new symbol table */
  for (sym=obj->symbol_table->first;sym!=NULL; sym=sym->next)
  {
    t_symbol * tsym=SymbolTableGetFirstSymbolWithName(obj->sub_symbol_table,sym->name);
    if (!tsym)
    {
      if ((sym->type==ABS_SYMBOL) && (sym->local==1))  continue; /* Local abs can be a skipped file symbol */
      WARNING(("<%s> was in the linked binary, but not present after relinking! (local = %d) not found !\n",sym->name,sym->local));
    }
    else if (tsym->type==WEAK_UNDEFINED_SYMBOL) continue;
    else if ((!(GENERIC_EQ(sym->current_address,tsym->current_address))) &&(!sym->local ))
    {
      WARNING(("%s differs!\n",sym->name));
    }
  }
}

void ObjectExportLinkGraph(t_object * obj)
{
  FILE * fp=NULL;
  t_object * tmp2, * tmp;
  t_reloc * rel;
#ifdef HAVE_DIRS
  DirectoryAssure("./dots",FALSE);
  fp=fopen("./dots/linkgraph.dot","w"); 
#else
  fp=fopen("./linkgraph.dot","w"); 
#endif
  fprintf(fp,"digraph \"Link_graph\" { \n");

  OBJECT_FOREACH_SUBOBJECT(obj,tmp,tmp2)
    if (tmp->reloc_table)
      OBJECT_FOREACH_RELOC(tmp, rel)
      {
	if (RELOCATABLE_TYPE(REL_FROM(rel)) != RT_SUBSECTION) FATAL(("Not from section @R!",rel));
	if (SECTION_IS_MAPPED(T_SECTION(REL_FROM(rel))))
	  if ((REL_TO_SYMBOL(rel)->to_type == TO_SECTION || REL_TO_SYMBOL(rel)->to_type == TO_SUB_SECTION))
	  {
	    if (REL_TO_SYMBOL(rel)->mapped->base)
	    {
	      /* Export link-graph */fprintf(fp,"\"%p\" [label=\"%s\" style=solid color=%s]\n",REL_FROM(rel),T_SECTION(REL_FROM(rel))->obj->name,(T_SECTION(REL_FROM(rel))->type==CODE_SECTION)?"blue":"red");
	      /* Export link-graph */fprintf(fp,"\"%p\" [label=\"%s\" style=solid color=%s]\n",REL_TO_SYMBOL(rel)->mapped->base,T_SECTION(REL_TO_SYMBOL(rel)->mapped->base)->obj->name,T_SECTION(REL_TO_SYMBOL(rel)->mapped->base)->type==CODE_SECTION?"blue":"red");
	      /* Export link-graph */fprintf(fp,"\"%p\" -> \"%p\"\n",REL_FROM(rel), REL_TO_SYMBOL(rel)->mapped->base);
	    }
	    else
	    {
	      printf("SKIPPED %s\n",REL_TO_SYMBOL(rel)->name);
	    }
	  }

      }

  fprintf(fp,"}\n");
}

void ObjectBuildRelocTableFromSubObjects(t_object * obj)
{
  t_object * tmp2, * tmp;
  t_reloc * rel,* reln;
  t_symbol * extra_sym=NULL;

  /* Do the relocations */
  OBJECT_FOREACH_SUBOBJECT(obj,tmp,tmp2)
  {
    if (tmp->reloc_table)
    {
      for (rel=tmp->reloc_table->first; rel!=NULL; rel=reln)
      {
	if (RELOCATABLE_TYPE(REL_FROM(rel)) != RT_SUBSECTION) FATAL(("Not from section @R!",rel));
	if (SECTION_IS_MAPPED(T_SECTION(REL_FROM(rel))))
	{
	  /* Some consistency checks {{{ */
	  if (REL_IS_TO_SYMBOL(rel)) {
	    if (REL_TO_SYMBOL(rel)->mapped==NULL) FATAL(("Could not merge @R (%s from object %s)!",rel,REL_TO_SYMBOL(rel)->name,tmp->name));
	    if (!rel->e_to_symbol) FATAL(("Reloc @R not with extra symbol!",rel)); 
	  }
	  /* }}} */

	  if (REL_E_IS_TO_SYMBOL(rel) && REL_E_TO_SYMBOL(rel))
	  {
	    if (REL_E_TO_SYMBOL(rel)->mapped==NULL)
	      FATAL(("Could not merge @R (%s from object %s)!",rel,rel->e.sym->name,tmp->name));
	    extra_sym=rel->e.sym->mapped;
	  }

	  reln=REL_NEXT(rel);
	  if (REL_IS_TO_SYMBOL(rel)) {
#ifdef THUMB_SUPPORT /* {{{ */
	    t_symbol_he * veneer_symbol = NULL;
	    char * veneer = NULL;
	    if(PatternMatch("*= s0002 & % s0002*",rel->code) && REL_TO_SYMBOL(rel)->mapped && REL_TO_SYMBOL(rel)->mapped->base)
	    /* Ok, we have an PC24 arm relocation */
	    {

	      if(ADDRESS_IS_SMALL_CODE == SymbolTableGetCodeType(obj,GENERIC_ADD(REL_TO_SYMBOL(rel)->mapped->base->caddress,REL_TO_SYMBOL(rel)->offset_from_start)))
              /* We jump from arm code to thumb code */
	      {
		t_uint32 opcode = SectionGetData32(obj->code[0],GENERIC_SUB(GENERIC_ADD(REL_FROM(rel)->caddress,rel->from_offset),SECTION_GET_OLD_ADDRESS(obj->code[0])));
		VERBOSE(0,("Opgeslagen waarde: %x",opcode));
		if((!(opcode & 0x01000000)) || (opcode & 0xff000000) != 0xfb000000)
		{
		  /* Conditional branch and link or normal branch, look for the veneer symbol! */
		  veneer = Collate("$Ven$AT$$",REL_TO_SYMBOL(rel)->name);
		  VERBOSE(0,("Looking for %s",veneer));
		  veneer_symbol =(t_symbol_he *) HashLookup(obj->symbol_table->hash_table, veneer);
		  if(veneer_symbol)
		  {
		    DiabloPrint(stdout,"Changed to veneer!\n");
		    RelocSetToSymbol(rel,veneer_symbol->sym);
		  }
		  Free(veneer);

		  
		}
		if(!veneer_symbol)
		{
		  veneer = Collate("__32",REL_TO_SYMBOL(rel)->name);
		  VERBOSE(0,("Looking for %s",veneer));
		  veneer_symbol = (t_symbol_he *) HashLookup(obj->symbol_table->hash_table, veneer);
		  if(veneer_symbol)
		  {
		    DiabloPrint(stdout,"Changed to veneer!\n");
		    RelocSetToSymbol(rel,veneer_symbol->sym);
		  }
		  Free(veneer);
		}

	      }
	      		
	    }
	    else if(PatternMatch("*s0001 > = i003ff800*",rel->code) && REL_TO_SYMBOL(rel)->mapped && REL_TO_SYMBOL(rel)->mapped->base)
	    /* We have a THM_PC22 thumb relocation */
	    {
	      veneer = Collate("__32",REL_TO_SYMBOL(rel)->name);
	      veneer_symbol =(t_symbol_he *) HashLookup(obj->symbol_table->hash_table, veneer);
	      if(veneer_symbol)
	      {
		//DiabloPrint(stdout,"Changed to %s!\n",veneer);
		RelocSetToSymbol(rel,veneer_symbol->sym);
	      }
	      /*else
	      {
		Free(veneer);
		veneer = Collate("__16",REL_TO_SYMBOL(rel)->name);
		veneer_symbol =(t_symbol_he *) HashLookup(obj->symbol_table->hash_table, veneer);
		if(veneer_symbol)
		{
		  DiabloPrint(stdout,"Changed to %s!\n",veneer);
		  REL_TO_SYMBOL(rel) = veneer_symbol->sym;
		}
	      }*/
	      Free(veneer);
	    }
	    
	    if(!veneer_symbol)
#endif /*}}}*/
	    RelocSetToSymbol(rel,REL_TO_SYMBOL(rel)->mapped);
	  }
	  if (REL_E_IS_TO_SYMBOL(rel) && REL_E_TO_SYMBOL(rel))
	  {
	    RelocSetEToSymbol(rel,extra_sym);
	  }
	  RelocTableUnlinkReloc(tmp->reloc_table,rel);
	  RelocTableLinkReloc(obj->reloc_table,rel); 
	}
	else
	{
	  DiabloPrint(stdout,"Verify: skipping @R\n",rel);
	  reln=REL_NEXT(rel);
	  RelocTableRemoveReloc(tmp->reloc_table,rel);
	}
      }
    }
  }

}



/*#define DEBUG_RELOCATE */
#include "../utils/util_endian.h"
/*#define TYPE 64*/
void ObjectRelocateSubObjects(t_object * obj, t_bool verify)
{
  t_reloc * rel;

  RelocsMigrateToRelocatables(obj);
  {
    t_reloc * rel;
    OBJECT_FOREACH_RELOC(obj,rel)
    {
      if (REL_TO_RELOCATABLE(rel))
      if(!rel->to_ref) FATAL(("@R",rel));
    }
  }

  /* patch the GOT relocations: turn them into "regular" relocations and,
   * while we're at it, recreate the .got section */
  for (rel=obj->reloc_table->first; rel!=NULL; rel=REL_NEXT(rel))
  {
    if (PatternMatch("*G*",rel->code))
    {
      /* GOT32 relocations */
      t_section * sec=SectionGetFromObjectByName(obj,".got");
      t_section * added;
      t_reloc * rel2;
      t_uint32 loaded;
      t_object * pobj=ObjectGetFromCache("Linker",obj);
      if (!sec) FATAL(("No got section found!"));
      if (!obj) FATAL(("Linker object not found!"));
      
      loaded=SectionGetData32(T_SECTION(REL_FROM(rel))->parent_section,GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),rel->from_offset),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)));
      added=SectionNew(pobj,DATA_SECTION);
      if (!added) FATAL(("Adding section failed!"));
      T_RELOCATABLE(added)->type=RT_SUBSECTION;
      added->name=StringDup("GOTENTRY");
      pobj->data=Realloc(pobj->data,(pobj->ndatas+1)*sizeof(t_section *));
      pobj->data[pobj->ndatas]=added;
      pobj->ndatas++;

      SectionSetOldAddressSuper(added,GENERIC_ADD_UINT32(SECTION_GET_OLD_ADDRESS(sec),loaded));
      SectionSetCAddressSuper(added,GENERIC_ADD_UINT32(SECTION_GET_CADDRESS(sec),loaded));
      SECTION_ALIGN(added)=GENERIC_ADD_UINT32(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),4);

      added->data=Calloc(1,4);
      SECTION_OLD_SIZE(added) = SECTION_CSIZE(added) = GENERIC_ADD_UINT32(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),4);
      MAP_SECTION(added,".got",sec);
     

      rel2=RelocTableAddRelocToRelocatable(obj->reloc_table,rel->addend,T_RELOCATABLE(added),NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset,TRUE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);



      RelocSetToRelocatable(rel,T_RELOCATABLE(added));

      if (sec)
      {
	RelocESetToRelocatable(rel,T_RELOCATABLE(sec));
	rel->e_offset = GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec));
	rel->addend = GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec));
	rel->to_offset = GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec));
      }
      Free(rel->code);
      rel->code=StringDup("SR-" "\\" WRITE_32);
    }
    
    else if (PatternMatch("*H*",rel->code))
    {
      /* GOT16 relocations */
      t_section * sec=SectionGetFromObjectByName(obj,".got");
      t_section * added;
      t_reloc * rel2;
      t_uint32 loaded;
      t_uint32 address;
      t_object * pobj=ObjectGetFromCache("Linker",obj);
      if (!sec) FATAL(("No got section found! %s",rel->code));
      if (!obj) FATAL(("Linker object not found!"));
      loaded=SectionGetData32(T_SECTION(REL_FROM(rel))->parent_section,GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),rel->from_offset),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)));
      added=NULL;

      {
	t_object * sub, * tmp;
	t_uint32 addr;

	address=SIGN_EXTEND(loaded&0xffff,15)+32752;
	addr=GENERIC_ADD_UINT32(SECTION_GET_OLD_ADDRESS(sec),address);
	OBJECT_FOREACH_SUBOBJECT(obj,sub,tmp)
	{
	  t_section * ss;
	  t_uint32 i;
	  OBJECT_FOREACH_SECTION(sub,ss,i)
	  {
	    if(SECTION_GET_OLD_ADDRESS(ss)==addr)
	    {
	      added=ss;
	      goto found;
	    }
	  }
	}
      }

found:

      if(!added)
      {
	added=SectionNew(pobj,DATA_SECTION);
	if (!added) FATAL(("Adding section failed!"));
	T_RELOCATABLE(added)->type=RT_SUBSECTION;
	added->name=StringDup("GOTENTRY");
	pobj->data=Realloc(pobj->data,(pobj->ndatas+1)*sizeof(t_section *));
	pobj->data[pobj->ndatas]=added;
	pobj->ndatas++;

	SectionSetOldAddressSuper(added,GENERIC_ADD_UINT32(SECTION_GET_OLD_ADDRESS(sec),address));
	SectionSetCAddressSuper(added,GENERIC_ADD_UINT32(SECTION_GET_CADDRESS(sec),address));
	SECTION_ALIGN(added)=GENERIC_ADD_UINT32(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),4);

	added->data=Calloc(1,4);
	SECTION_OLD_SIZE(added) = SECTION_CSIZE(added) = GENERIC_ADD_UINT32(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),4);
	MAP_SECTION(added,".got",sec);
      }

      if(!rel->corresponding)
      {
	rel2=RelocTableAddRelocToRelocatable(obj->reloc_table,rel->addend,T_RELOCATABLE(added),NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset,TRUE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);
	if (rel->label) rel2->label = Collate("",rel->label);
      }

      RelocSetToRelocatable(rel,T_RELOCATABLE(added));
      rel->to_offset=GENERIC_SUB(rel->to_offset,rel->to_offset);

      if (sec) RelocESetToRelocatable(rel,T_RELOCATABLE(sec));
      rel->e_offset=GENERIC_SUB(rel->e_offset,rel->e_offset);
      Free(rel->code);
      rel->code=StringDup("SR-i00007ff0-  i0000ffff &" "\\" "l iffff0000 & | w \\ s0000$");

      loaded=SectionGetData32(sec,GENERIC_SUB(SECTION_GET_CADDRESS(added),SECTION_GET_CADDRESS(sec)));
      *((t_uint32 *) added->data)=loaded;

      if(rel->corresponding)
      {
	rel->corresponding->to_offset=GENERIC_ADD(rel->corresponding->to_offset,rel->corresponding->addend);
	rel->corresponding->addend=GENERIC_SUB(rel->corresponding->addend,rel->corresponding->addend);
      }
    }

#ifdef IA64_SUPPORT    
    {
    static t_bool ia64_first_pass = TRUE;

#define CREATE_MASK(aantal) (((t_uint64)1 << (aantal)) - 1)
#define SIGN_EXT(x, p) ((((x)&(CREATE_MASK(p))) ^ ((t_uint64)1 << ((p) - 1))) - ((t_uint64)1 << ((p) - 1)))
#define GETBITS64(input, offset, length) (((input) >> (offset)) & CREATE_MASK(length))

#define IMM_S_I_BIT(instr) (GETBITS64(instr, 36, 1))
#define IMM_IMM5cA(instr) (GETBITS64(instr, 22, 5))
#define IMM_IMM9d(instr) (GETBITS64(instr, 27, 9))
#define IMM_IMM7b(instr) (GETBITS64(instr, 13, 7))    

#define IMM22_A5(instr)\
    (SIGN_EXT(IMM_S_I_BIT(instr)<<21|IMM_IMM5cA(instr)<<16|IMM_IMM9d(instr)<<7|IMM_IMM7b(instr),22))

      /* IA-64 specific*/

      if(ia64_first_pass)
    {
      t_symbol *symptr;
      t_object * tmp2, * tmp;
      t_reloc *rel2, *rel3;

      ia64_first_pass = FALSE;
      /*seeking for special $diablo-symbols*/
      for (tmp2=obj->mapped_first;tmp2!=NULL; tmp2=tmp2->next)
      { 
	for (tmp=tmp2; tmp!=NULL; tmp=tmp->equal)
	{     
	  t_address generic;
	  t_uint32 tel;

	  for (symptr=tmp->symbol_table->first; symptr!=NULL; symptr=symptr->next)
	  {
	    if(!symptr->name)
	      continue;

	    if (strncmp(symptr->name,"$diablo:",8)==0)
	    {
	      int begin = strstr(symptr->name,"@-@")-symptr->name;
	      char save = symptr->name[begin];
	      t_symbol *s1, *s2;
	      t_address a1, a2;
//	      printf("symbol: %s @ 0x%llx\n", symptr->name, G_T_UINT64(SECTION_GET_CADDRESS(symptr->base)));
	      symptr->name[begin] = '\0';


	      /*TODO ook checken als het symbool van zelfde .o-file komt*/
	      s1=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+8);
	      s2=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+begin+3);


//	      DiabloPrint(stdout, "symbol s1: @S\n", s1);
//	      DiabloPrint(stdout, "symbol s2: @S\n", s2);

	      a1 = GENERIC_ADD(SECTION_GET_CADDRESS(symptr->base), s1->current_address);
	      a2 = GENERIC_ADD(SECTION_GET_CADDRESS(symptr->base), s2->current_address);
	      /*
		 printf("%s @ 0x%llx;  %s @ 0x%llx\n", s1->name, a1, s2->name, a2);
		 fflush(stdout);
	       */	  
	      /*char herstellen*/
	      symptr->name[begin] = save;

	      /*controleren of wel bestaat !*/
	      if(!s1|| !s2)
		FATAL(("Invalid special diablo symbol !!!"));
		
//	      rel = RelocTableAddRelocToSymbol(obj->reloc_table,symptr->addend, sh->section,generic,symptr,TRUE,NULL,NULL,NULL,"SR-" "\\" WRITE_64 ,FALSE);
	     rel3 = RelocTableAddRelocToRelocatable(obj->reloc_table,0,
symptr->base,NULL_GENERIC, symptr->base, 0, TRUE, NULL, NULL, NULL,
"s0000$", FALSE);

              rel3->from_offset = 0;
              rel3->to_offset = G_T_UINT64(symptr->offset_from_start);
 //	      
	       rel2=RelocTableAddRelocToRelocatable(obj->reloc_table,symptr->addend,symptr->base,NULL_GENERIC,s2->base,0,TRUE,NULL,NULL,s1->base,"s0000$" ,FALSE);

	       rel2->e_offset = G_T_UINT64(s1->offset_from_start);
	       rel2->to_offset = G_T_UINT64(s2->offset_from_start);
	       rel2->from_offset = G_T_UINT64(symptr->offset_from_start);
#if 0
	      if(!(INS_ATTRIB(SecGetInsByAddress(section, a1)) & IF_BBL_LEADER))
		FATAL(("First instruction of patched function has NOT been marked as BBL_LEADER !!!"));
	      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, a1)),NULL,NULL);        


	      if(!(INS_ATTRIB(SecGetInsByAddress(section, a2)) & IF_BBL_LEADER))
		FATAL(("First instruction of patched function has NOT been marked as BBL_LEADER !!!"));
	      CfgEdgeNewCall(flowgraph, flowgraph->hell_node, INS_BBL(SecGetInsByAddress(section, a2)),NULL,NULL);        
#endif	      
	    }
	  }

	}
      }

    }
    }

      
//    if (strcmp(rel->code, "l") == 0 || strcmp(rel->code, "q") == 0)
    if(PatternMatch("*u*",rel->code) || PatternMatch("*q*",rel->code))
    {
      /* LTOFF22 relocations (and ltoff-fptr)*/
      t_section * sec=SectionGetFromObjectByName(obj,".got");
      t_section * added = NULL, *gotentry;
      t_reloc * rel2;
      t_uint64 loaded;
      t_object * pobj=ObjectGetFromCache("Linker",obj);
      t_uint32 tel = 0;

      char* bundleptr = GENERIC_ADD_DISPL(T_SECTION(REL_FROM(rel))->parent_section->data, 
	  GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),GENERIC_NAND_UINT32(rel->from_offset,0xf)),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)));
      t_uint32 off22;
      t_uint32 currins = GENERIC_EXTRACT_UINT32(rel->from_offset)&0xf;
      t_uint64 data, data2;
 
      t_address off22addr;
      
      switch(currins)
      {
	case 0:
	    data = *((t_uint64 *) bundleptr);
  	    data=(data >> 5) & CREATE_MASK(41);
	    off22 = IMM22_A5(data);
	    break;
	case 1:
	    data = *((t_uint64 *) bundleptr);
	    data2 = *(((t_uint64 *) bundleptr)+1);
	    data = (data >> 46) | ((data2 & CREATE_MASK(23))<<18);
	    off22 = IMM22_A5(data);
	    break;
	case 2:
	    data =*(((t_uint64 *) bundleptr)+1);
	    data = data >>23;
	    off22 = IMM22_A5(data);
	    break;
	default:
	  FATAL(("problemo"));
      }

      GenericSet_uint64(off22addr, off22);
#ifdef DEBUG_RELOCATE      
      printf("off22 = %d\n", off22);
      printf("ins = 0x%llx\n", data);

      printf("got address = 0x%llx\n", G_T_UINT64(SECTION_GET_CADDRESS(sec)));
      printf("reloc address = 0x%llx\n", G_T_UINT64(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel)))));
      printf("reloc offset = 0x%llx\n", rel->from_offset);
      printf("parent address = 0x%llx\n", G_T_UINT64(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)));

      printf("reloc to address = 0x%llx\n", G_T_UINT64(SECTION_GET_CADDRESS(REL_TO_RELOCATABLE(rel))));
      printf("found addres = 0x%llx\n", SectionGetData64(sec, off22addr));
      printf("addend = 0x%llx\n", G_T_UINT64(rel->addend));
      printf("reloc to offset = 0x%llx\n", G_T_UINT64(rel->to_offset));
#endif
/*      
      if(G_T_UINT64(rel->addend) != 0)
	printf("Addend not 0\n");
*/      
      if (!sec) FATAL(("No got section found!"));
      if (!obj) FATAL(("Linker object not found!"));
      
      loaded = off22;

      //printf("loaded = 0x%llx\n", loaded);

      /*enkel nieuw sectietje toevoegen als er al geen voor aangemaakt werd...*/
     
      OBJECT_FOREACH_SECTION(pobj, gotentry, tel)
      {
	/* (*) */

	/*only overlap detection*/
	if(gotentry->name && strcmp(gotentry->name, "GOTENTRY") == 0 && SECTION_GET_CADDRESS(gotentry) == GENERIC_ADD_UINT64(SECTION_GET_CADDRESS(sec),loaded))
	{
	  added = gotentry;
	  break;
	}
#if 0	
	/*real compression*/	
	if(gotentry->name && strcmp(gotentry->name, "GOTENTRY") == 0 && *((t_uint64*)gotentry->data) == SectionGetData64(sec, off22addr))
	{
	  printf("Hallo\n");
	  added = gotentry;
	  break;
	}
#endif	
      }
      
      if(added == NULL)
      {

	added=SectionNew(pobj,DATA_SECTION);
	if (!added) FATAL(("Adding section failed!"));
	T_RELOCATABLE(added)->type=RT_SUBSECTION;
	added->name=StringDup("GOTENTRY");
	pobj->data=Realloc(pobj->data,(pobj->ndatas+1)*sizeof(t_section *));
	pobj->data[pobj->ndatas]=added;
	pobj->ndatas++;

	SectionSetOldAddressSuper(added,GENERIC_ADD_UINT64(SECTION_GET_OLD_ADDRESS(sec),loaded));
	SectionSetCAddressSuper(added,GENERIC_ADD_UINT64(SECTION_GET_CADDRESS(sec),loaded));
	SECTION_ALIGN(added)=GENERIC_ADD_UINT64(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),8);

	added->data=Calloc(1,8);
	*((t_uint64*)added->data) = SectionGetData64(sec, off22addr); /*toch al iets invullen: nodig voor (*)*/
	SECTION_OLD_SIZE(added) = SECTION_CSIZE(added) = GENERIC_ADD_UINT32(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),8);
	MAP_SECTION(added,".got",sec);

	if(PatternMatch("*u*",rel->code))
	{
	  if(G_T_UINT64(SECTION_GET_CADDRESS(REL_TO_RELOCATABLE(rel))) != 0x0ull)
  	    rel2=RelocTableAddRelocToRelocatable(obj->reloc_table,rel->addend,T_RELOCATABLE(added),NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset,TRUE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_64,FALSE);
	}
	else /*now make a fptr from it: it should be expanded afterwards*/
	  if(SectionGetData64(sec, off22addr) != 0)
	    rel2=RelocTableAddRelocToRelocatable(obj->reloc_table,rel->addend,T_RELOCATABLE(added),NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset,TRUE,NULL,NULL,NULL,"p",FALSE);
	
      }

      /*!!! SPECIAL PATCH, otherwise it won't work (he may not add this offset anymore during S, or the offset will be too large)...*/
      rel->to_offset = 0;
      rel->e_offset = 0;

      RelocSetToRelocatable(rel,T_RELOCATABLE(added));

      if (sec) RelocESetToRelocatable(rel,T_RELOCATABLE(sec));
      Free(rel->code);
      rel->code=StringDup("SR-" "\\" WRITE_IMM22);

      
    }

    
    if (PatternMatch("*p*",rel->code))
      {
          /* FPTR64LSB relocations */
      t_section * sec=SectionGetFromObjectByName(obj,".opd");
      t_section * got=SectionGetFromObjectByName(obj,".got");
      t_section * added = NULL, *opdentry, *added2;
      t_reloc * rel2, *rel3;
      t_uint64 loaded;
      t_object * pobj=ObjectGetFromCache("Linker",obj);
      t_uint32 tel = 0;

      t_uint64 fptr;
      t_uint64 fptroffset;
      t_address fptroffsetaddr;
      t_address addend2;

      fptr = SectionGetData64(T_SECTION(REL_FROM(rel))->parent_section, GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))), rel->from_offset),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)));

#ifdef DEBUG_RELOCATE      
      printf("fptr = 0x%llx\n", fptr);

      printf("opd address = 0x%llx\n", G_T_UINT64(SECTION_GET_CADDRESS(sec)));
      printf("reloc address = 0x%llx\n", G_T_UINT64(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel)))));
      printf("reloc offset = 0x%llx\n", rel->from_offset);
      printf("parent address = 0x%llx\n", G_T_UINT64(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)));

      printf("reloc to address = 0x%llx\n", G_T_UINT64(SECTION_GET_CADDRESS(REL_TO_RELOCATABLE(rel))));
      //printf("found addres = 0x%llx\n", SectionGetData64(sec, off22addr));
      printf("addend = 0x%llx\n", G_T_UINT64(rel->addend));
      printf("reloc to offset = 0x%llx\n", G_T_UINT64(rel->to_offset));

      
      if(G_T_UINT64(rel->addend) != 0)
  	printf("Addend not 0\n");
#endif      
      
      if (!sec) FATAL(("No opd section found!"));
      if (!got) FATAL(("No got section found!"));
      if (!obj) FATAL(("Linker object not found!"));
      
      loaded = fptr - G_T_UINT64(SECTION_GET_CADDRESS(sec));

      GenericSet_uint64(fptroffsetaddr, loaded);

#ifdef DEBUG_RELOCATE      
      printf("loaded = 0x%llx\n", loaded);
#endif
      /*enkel nieuw sectietje toevoegen als er al geen voor aangemaakt werd...*/
     
      OBJECT_FOREACH_SECTION(pobj, opdentry, tel)
        {
  	/* (*) */

	/*only overlap detection*/
	if(opdentry->name && strcmp(opdentry->name, "OPDENTRY") == 0 && SECTION_GET_CADDRESS(opdentry) == GENERIC_ADD_UINT64(SECTION_GET_CADDRESS(sec),loaded))
	{
	  added = opdentry;
	  break;
	}
#if 0	
	/*real compression*/	
	if(fptoffset != 0ull && opdentry->name && strcmp(opdentry->name, "OPDENTRY") == 0 && *((t_uint64*)opdentry->data) == SectionGetData64(sec, fptroffsetaddr))
  	{
	  printf("Hallo\n");
	  added = gotentry;
	  break;
	}
#endif	
      }
      
      if(added == NULL)
      {

	added=SectionNew(pobj,DATA_SECTION);
	if (!added) FATAL(("Adding section failed!"));
	T_RELOCATABLE(added)->type=RT_SUBSECTION;
	added->name=StringDup("OPDENTRY");
	pobj->data=Realloc(pobj->data,(pobj->ndatas+1)*sizeof(t_section *));
	pobj->data[pobj->ndatas]=added;
	pobj->ndatas++;

	SectionSetOldAddressSuper(added,GENERIC_ADD_UINT64(SECTION_GET_OLD_ADDRESS(sec),loaded));
	SectionSetCAddressSuper(added,GENERIC_ADD_UINT64(SECTION_GET_CADDRESS(sec),loaded));
	SECTION_ALIGN(added)=GENERIC_ADD_UINT64(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),16);

	added->data=Calloc(1,16);
	if(fptroffset != 0ull)
	  *((t_uint64*)added->data) = SectionGetData64(sec, fptroffsetaddr); /*toch al iets invullen: nodig voor (*)*/

	SECTION_OLD_SIZE(added) = SECTION_CSIZE(added) = GENERIC_ADD_UINT32(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),16);
	MAP_SECTION(added,".opd",sec);

	rel2=RelocTableAddRelocToRelocatable(obj->reloc_table,rel->addend,T_RELOCATABLE(added),NULL_GENERIC,REL_TO_RELOCATABLE(rel),rel->to_offset,TRUE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_64,FALSE);

	//	addend2 = rel->addend;
	//	GENERIC_ADD_UINT64(addend2, 8);


#if 0
	added2=SectionNew(pobj,DATA_SECTION);
	if (!added2) FATAL(("Adding section failed!"));
	T_RELOCATABLE(added2)->type=RT_SUBSECTION;
	added2->name=StringDup("OPDENTRY");
	pobj->data=Realloc(pobj->data,(pobj->ndatas+1)*sizeof(t_section *));
	pobj->data[pobj->ndatas]=added2;
	pobj->ndatas++;

	loaded += 8;
	
	SectionSetOldAddressSuper(added2,GENERIC_ADD_UINT64(SECTION_GET_OLD_ADDRESS(sec),loaded));
	SectionSetCAddressSuper(added2,GENERIC_ADD_UINT64(SECTION_GET_CADDRESS(sec),loaded));
	SECTION_ALIGN(added2)=GENERIC_ADD_UINT64(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),8);

	added2->data=Calloc(1,8);

	SECTION_OLD_SIZE(added2) = SECTION_CSIZE(added2) = GENERIC_ADD_UINT32(GENERIC_SUB(SECTION_GET_OLD_ADDRESS(sec),SECTION_GET_OLD_ADDRESS(sec)),8);
	MAP_SECTION(added2,".opd",sec);

#endif

	rel3=RelocTableAddRelocToRelocatable(obj->reloc_table, rel->addend,T_RELOCATABLE(added),NULL_GENERIC,T_RELOCATABLE(got),0,TRUE,NULL,NULL,NULL,"g" "\\" WRITE_64,FALSE);

	GenericSet_uint64(rel3->from_offset, 8);
      }

      /*!!! SPECIAL PATCH, otherwise it won't work (he may not add this offset anymore during S, or the offset will be too large)...*/
      rel->to_offset = 0;

      RelocSetToRelocatable(rel,T_RELOCATABLE(added));

      if (sec) RelocESetToRelocatable(rel,T_RELOCATABLE(sec));
      Free(rel->code);
      rel->code=StringDup(CALC_ABS "\\" WRITE_64);

      
    }
    
#endif                      
  }

  RelocateCore(obj,verify);
}

void RelocateCore(t_object * obj, t_bool verify)
{
  t_reloc * rel;
  for (rel=obj->reloc_table->first; rel!=NULL; rel=REL_NEXT(rel))
  {
#ifdef IA64_SUPPORT    
    t_uint64 x,y;
#else
    int x,y;
#endif    
    t_address ret;

#ifdef DEBUG_RELOCATE
    printf("============= %s ================== %s\n",rel->code,rel->label);
    fflush(stdout);
    DiabloPrint(stdout,"RELOC: @R\n",rel);
    DiabloPrint(stdout,"RELOC ADDEND = @G\n",rel->addend); 
    DiabloPrint(stdout,"RELOC TO OFFSET = @G\n",rel->to_offset); 
    if (REL_TO_RELOCATABLE(rel)) DiabloPrint(stdout,"RELOC TO ADDRESS = @G\n",REL_TO_RELOCATABLE(rel)->caddress); 
    if (REL_TO_RELOCATABLE(rel)) DiabloPrint(stdout,"RELOC TO SiZE    = @G\n",REL_TO_RELOCATABLE(rel)->csize); 
    if (RELOCATABLE_TYPE(REL_FROM(rel)) != RT_SUBSECTION) FATAL(("Not from subsection @R!",rel));
    if (rel->e.sec)
    {
      DiabloPrint(stdout,"RELOC EXTRA ADDRESS = @G\n",rel->e.sec->caddress);
      DiabloPrint(stdout,"RELOC EXTRA OFFSET =  @G\n",rel->e_offset);
    }
    DiabloPrint(stdout,"Reloc from address =@G\n",REL_FROM(rel)->caddress);
    DiabloPrint(stdout,"Reloc from size =@G\n",REL_FROM(rel)->csize);
    DiabloPrint(stdout,"Reloc from offset =@G\n",rel->from_offset);
    if (rel->label) printf("Label: %s\n",rel->label);
#if (TYPE == 64)
    printf("address =%llx\n",rel->from_offset);
    printf("Getting at address %llx\n",(rel->from_offset&(~0xfULL)));
    printf("ORIG:  %llx\n",UINT64_SWAP_ENDIAN(SectionGetData64(REL_FROM(rel),(rel->from_offset&(~0xfULL)))));
    //printf("ORI2:  %llx\n",UINT64_SWAP_ENDIAN(SectionGetData64(REL_FROM(rel),(0x8+(rel->from_offset&(~0xfULL))))));
#else
    printf("ORIG:  %x\n",SectionGetData32(T_SECTION(REL_FROM(rel)),rel->from_offset));
#endif

#endif
    ret=StackExec(rel->code,rel,T_SECTION(REL_FROM(rel))->data,TRUE,0);

    if (GENERIC_EXTRACT_UINT32(ret)!=0) FATAL(("Relocation failed!"));
#ifdef DEBUG_RELOCATE
#if (TYPE==64)
    printf("RELOC: %llx\n",UINT64_SWAP_ENDIAN(SectionGetData64(REL_FROM(rel),rel->from_offset&(~0xfULL))));
    //printf("RELO2: %llx\n",UINT64_SWAP_ENDIAN(SectionGetData64(REL_FROM(rel),0x8+(rel->from_offset&(~0xfULL)))));
    printf("OREL : %llx\n",UINT64_SWAP_ENDIAN(SectionGetData64(T_SECTION(REL_FROM(rel))->parent_section,GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),(rel->from_offset&(~0xfULL))),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)))));
    //printf("OREL2: %llx\n",UINT64_SWAP_ENDIAN(SectionGetData64(T_SECTION(REL_FROM(rel))->parent_section,GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),0x8+(rel->from_offset&(~0xfULL))),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)))));
#else
    printf("RELOC: %x\n",SectionGetData32(T_SECTION(REL_FROM(rel)),rel->from_offset));
    printf("OREL : %x\n",SectionGetData32(T_SECTION(REL_FROM(rel))->parent_section,GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),rel->from_offset),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section))));
#endif
#endif

#ifdef IA64_SUPPORT    
    if (PatternMatch("*F*",rel->code)) /* usage of F means there are instructions involved*/
    {
      char* bundleptr = GENERIC_ADD_DISPL(T_SECTION(REL_FROM(rel))->parent_section->data, 
    	  GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),GENERIC_NAND_UINT32(rel->from_offset,0xf)),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)));
      char* bundleptr2 = GENERIC_ADD_DISPL(T_SECTION(REL_FROM(rel))->data, 
	      	  GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),GENERIC_NAND_UINT32(rel->from_offset,0xf)),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel)))));
      t_uint32 currins = GENERIC_EXTRACT_UINT32(rel->from_offset)&0xf;
      t_uint64 value, value2, data2;
 
      switch(currins)
        {
  	case 0:
	    value = *((t_uint64 *) bundleptr);
    	    value =(value >> 5) & CREATE_MASK(41);
	    value2 = *((t_uint64 *) bundleptr2);
    	    value2 =(value2 >> 5) & CREATE_MASK(41);
	    break;
	case 1:
	    value = *((t_uint64 *) bundleptr);
	    data2 = *(((t_uint64 *) bundleptr)+1);
	    value = (value >> 46) | ((data2 & CREATE_MASK(23))<<18);
	    value2 = *((t_uint64 *) bundleptr2);
	    data2 = *(((t_uint64 *) bundleptr2)+1);
	    value2 = (value2 >> 46) | ((data2 & CREATE_MASK(23))<<18);
	    break;
	case 2:
	    value =*(((t_uint64 *) bundleptr)+1);
	    value = value >>23;
	    value2 =*(((t_uint64 *) bundleptr2)+1);
	    value2 = value2 >>23;
	    break;
	default:
	  FATAL(("problemo"));
      }
      
      x = value;
      y = value2;

      if(GENERIC_EQ(SECTION_GET_CADDRESS(T_SECTION(REL_TO_RELOCATABLE(rel))), NULL_GENERIC))
      {
	x = y = 0;
      }
    }
    else
    {
      x=SectionGetData64(T_SECTION(REL_FROM(rel)),rel->from_offset);
      y=SectionGetData64(T_SECTION(REL_FROM(rel))->parent_section,GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),rel->from_offset),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)));

      /*only temporary !!!!! skip compressed .rodata.str1.8*/
      if(0)//strncmp(T_SECTION(REL_TO_RELOCATABLE(rel))->name, ".rodata.str", 11) == 0)    
      {
	x = y = 0;
      }
    }
#else
    x=SectionGetData32(T_SECTION(REL_FROM(rel)),rel->from_offset);
    y=SectionGetData32(T_SECTION(REL_FROM(rel))->parent_section,GENERIC_SUB(GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))),rel->from_offset),SECTION_GET_CADDRESS(T_SECTION(REL_FROM(rel))->parent_section)));
    
#endif
    if ((verify) && (x!=y))
    {
      /* ld messes up if .sbss is empty and there is padding before .bss */
      if (!(rel->label) || ((strcmp(rel->label,"__bss_start__"))
      && (strcmp(rel->label,"__preinit_array_end"))
      && (strcmp(rel->label,"_BASE_ADDRESS")) /* FIXME*/
      && (strcmp(rel->label,"_etext")) /* FIXME*/
      && (strcmp(rel->label,"_gpinfo")) /* FIXME*/
      && (strcmp(rel->label,"_fpdata_size")) /* FIXME*/
      && (strcmp(rel->label,"__preinit_array_start"))
      && (strcmp(rel->label,"__start___libc_atexit"))
      && (strcmp(rel->label,"__init_array_end"))
      && (strcmp(rel->label,"__fini_array_start"))
      && (strcmp(rel->label,"__fini_array_end"))
      && (strcmp(rel->label,"__init_array_start"))))
    if (verify)
      {
	if (!global_options.dump && !global_options.dump_multiple)
	{
#ifdef IA64_SUPPORT	  
	  FATAL(("Implement %s: %llx %llx",rel->label,x,y));
#else
	  FATAL(("Linker emulation failed. A relocated value differed after link emulation. There can be two reasons for this: either you are not using a patched ld (see the diablo website for details) or there is a bug in diablo. Implement %s: %x %x",rel->label,x,y));
#endif	  
	}
      }
    }
  }
}

extern t_symbol_he * symbol_type_cached;
void ObjectCreateDataOrCodeTable(t_object * obj)
{
  /* Sort the text symbols */
  t_symbol_he * top=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$a");
  t_symbol_he * he;
  t_symbol *tmp;

  symbol_type_cached=NULL;
  if (top)
    while(top->node.equal)
    {
      for(he=(t_symbol_he *) top->node.equal;he!=NULL;he=(t_symbol_he *) he->node.equal)
      {
	if (GENERIC_LT(he->sym->current_address,top->sym->current_address))
	{
	  tmp=top->sym;
	  top->sym=he->sym;
	  he->sym=tmp;
	}
      }
      top=(t_symbol_he *) top->node.equal;
    }
  /* Compensate safety measures for $a, remove the doubles at the same address */
  top=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$a");
  he=top;
  if (top)
    while(top->node.equal)
    {
      top=(t_symbol_he *) top->node.equal;
      if (GENERIC_EQ(top->sym->current_address,he->sym->current_address)) 
      {
	he->node.equal=top->node.equal;
	if (top->sym->prev) 
	{
	  top->sym->prev->next=top->sym->next;
	}
	else
	{
	  /* adjust sub_symbol_table->first */
	  obj->sub_symbol_table->first = top->sym->next;
	}
	if (top->sym->next) 
	{
	  top->sym->next->prev=top->sym->prev;
	}
	else
	{
	  /* adjust sub_symbol_table->last */
	  obj->sub_symbol_table->last = top->sym->prev;
	}

	Free(top->node.key);
	Free(top->sym->name);
	Free(top->sym);
	Free(top);
	top=he;
      }
      else
	he=top;
    }

#ifdef THUMB_SUPPORT
  /* Sort the thumb symbols if any */
  top=( t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$t");
  if(top)
    while(top->node.equal)
    {
      for(he=(t_symbol_he *) top->node.equal;he!=NULL;he=(t_symbol_he *) he->node.equal)
      {
	if (GENERIC_LT(he->sym->current_address,top->sym->current_address))
	{
	  tmp=top->sym;
	  top->sym=he->sym;
	  he->sym=tmp;
	}
      }
      top=(t_symbol_he *) top->node.equal;
    }
    /* Compensate safety measures for $t, remove the doubles at the same address */
  top=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$t");
  he=top;
  if (top)
    while(top->node.equal)
    {
      //DiabloPrint(stdout,"$t at @G\n",top->sym->current_address);
      top=(t_symbol_he *) top->node.equal;
      if (GENERIC_EQ(top->sym->current_address,he->sym->current_address))
      {
	he->node.equal=top->node.equal;
	if (top->sym->prev) 
	{
	  top->sym->prev->next=top->sym->next;
	}
	else
	{
	  /* adjust sub_symbol_table->first */
	  obj->sub_symbol_table->first = top->sym->next;
	}
	if (top->sym->next) 
	{
	  top->sym->next->prev=top->sym->prev;
	}
	else
	{
	  /* adjust sub_symbol_table->last */
	  obj->sub_symbol_table->last = top->sym->prev;
	}

	Free(top->node.key);
	Free(top->sym->name);
	Free(top->sym);
	Free(top);
	top=he;
      }
      else
	he=top;
    }
#endif

  /* Sort the data symbols */
  top=( t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$d");
  if (top)
    while(top->node.equal)
    {
      for(he=(t_symbol_he *) top->node.equal;he!=NULL;he=(t_symbol_he *) he->node.equal)
      {
	if (GENERIC_LT(he->sym->current_address,top->sym->current_address))
	{
	  tmp=top->sym;
	  top->sym=he->sym;
	  he->sym=tmp;
	}

      }
      top=(t_symbol_he *) top->node.equal;
    }
  /* Compensate safety measures for $d */
  top=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$d");
  he=top;
  if (top)
    while(top->node.equal)
    {
      top=(t_symbol_he *) top->node.equal;
      if (GENERIC_EQ(top->sym->current_address,he->sym->current_address)) 
      {
	he->node.equal=top->node.equal;
	if (top->sym->prev) 
	  top->sym->prev->next=top->sym->next;
	else
	  obj->sub_symbol_table->first = top->sym->next;
	if (top->sym->next) 
	  top->sym->next->prev=top->sym->prev;
	else
	  obj->sub_symbol_table->last = top->sym->prev;

	Free(top->node.key);
	Free(top->sym->name);
	Free(top->sym);
	Free(top);
	top=he;
      }
      else
	he=top;
    }

  /* Create a new symbol called $switch = compressed $(a,t) -> $d*/
  {
    t_bool prev_code=FALSE; t_bool first=TRUE;
    t_symbol_he * he=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$d");
    t_symbol_he * he2=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$a");
    t_symbol_he * he3= NULL;
#ifdef THUMB_SUPPORT
    he3=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$t");
#endif
    while((he)||(he2)||(he3))
    {

      if((he && !he2 && !he3) ||
	 (he && he2 && !he3 && GENERIC_LT(he->sym->current_address,he2->sym->current_address)) ||
	 (he && !he2 && he3 && GENERIC_LT(he->sym->current_address,he3->sym->current_address)) ||
	 (he && he2 && he3 && GENERIC_LT(he->sym->current_address,he3->sym->current_address) && GENERIC_LT(he->sym->current_address,he2->sym->current_address))
	)
      {
	if ((prev_code)||first) 
	{
	  if (he->sym->base && (he->sym->base->type == RT_SECTION || he->sym->base->type == RT_SUBSECTION))
	  {
	    if (first) 
	    {
	      t_symbol * sym=SymbolTableGetFirstSymbolWithName(obj->sub_symbol_table,"$first");
	      if (!sym)
		SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,GENERIC_SUB(he->sym->current_address,he->sym->current_address),"$first",TRUE);
	      else sym->current_address=GENERIC_SUB(he->sym->current_address,he->sym->current_address);
	    }
	    SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,he->sym->current_address,"$switch",TRUE);
	  }
	  else
	    FATAL(("$d symbol not with sec or subsec"));
	}
	he=(t_symbol_he *) he->node.equal;
	prev_code=FALSE;
      }
      else if(he2 && ((!he3) || GENERIC_LT(he2->sym->current_address,he3->sym->current_address)))
      {
	if ((!prev_code)||first)
	{
	  if (he2->sym->base && (he2->sym->base->type == RT_SECTION || he2->sym->base->type == RT_SUBSECTION))
	  {
	    if (first) 
	    {
	      t_symbol * sym=SymbolTableGetFirstSymbolWithName(obj->sub_symbol_table,"$first");
	      if (!sym)
		SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,GENERIC_ADD_UINT32(GENERIC_SUB(he->sym->current_address,he->sym->current_address),1),"$first",TRUE);
	      else sym->current_address=GENERIC_ADD_UINT32(GENERIC_SUB(he->sym->current_address,he->sym->current_address),1);
	    }
	    SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,he2->sym->current_address,"$switch",TRUE);
	  }
	  else
	    FATAL(("%s symbol not with sec or subsec, has %sbase, type is %d, address is @G",he2->sym->name, he2->sym->base ? "" : "no ",he2->sym->base->type, he2->sym->current_address));

	}
	he2=(t_symbol_he *) he2->node.equal;
	prev_code=TRUE;
      }
      else if(he3)
      {
	if ((!prev_code)||first)
	{
	  if (he3->sym->base && (he3->sym->base->type == RT_SECTION || he3->sym->base->type == RT_SUBSECTION))
	  {
	    if (first)
	    {
	      t_symbol * sym=SymbolTableGetFirstSymbolWithName(obj->sub_symbol_table,"$first");
	      if(!sym)
		SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,GENERIC_ADD_UINT32(GENERIC_SUB(he->sym->current_address,he->sym->current_address),1),"$first",TRUE);
	      else sym->current_address=GENERIC_ADD_UINT32(GENERIC_SUB(he->sym->current_address,he->sym->current_address),1);
	    }
	    SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,he3->sym->current_address,"$switch",TRUE);
	  }
	  else
	    FATAL(("%s symbol not with sec or subsec, has %sbase, type is %d, address is @G",he3->sym->name, he3->sym->base ? "" : "no ",he3->sym->base->type, he3->sym->current_address));
	}
	he3=(t_symbol_he *) he3->node.equal;
	prev_code=TRUE;
      }
      else FATAL(("Problems!"));
	
      first=FALSE;
    }
  }
#ifdef THUMB_SUPPORT
  /* Create new symbols called $code_switch, compressed $a -> $t */
  {
    t_bool prev_arm = TRUE, first = TRUE;
    t_symbol_he * he=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$t");
    t_symbol_he * he2=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$a");

    while((he)||(he2))
    {
      if ((he)&&((!he2)||(GENERIC_LT(he->sym->current_address,he2->sym->current_address))))
      {
	if ((prev_arm)||first)
	{
	  if (he->sym->base && (he->sym->base->type == RT_SECTION || he->sym->base->type == RT_SUBSECTION))
	  {
	    if (first)
	    {
	      t_symbol * sym=SymbolTableGetFirstSymbolWithName(obj->sub_symbol_table,"$thumb");
	      if(!sym)
		SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,GENERIC_SUB(he->sym->current_address,he->sym->current_address),"$thumb",TRUE);
	      else sym->current_address=GENERIC_SUB(he->sym->current_address,he->sym->current_address);
	    }
	    SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,he->sym->current_address,"$code_switch",TRUE);
	  }
	  else
	    FATAL(("$d symbol not with sec or subsec"));
	}
	he=(t_symbol_he *) he->node.equal;
	prev_arm = FALSE;
      }
      else
      {
	if ((!prev_arm)||first)
	{
	  if (he2->sym->base && (he2->sym->base->type == RT_SECTION || he2->sym->base->type == RT_SUBSECTION))
	  {
	    if (first)
	    {
	      t_symbol * sym=SymbolTableGetFirstSymbolWithName(obj->sub_symbol_table,"$thumb");
	      if(!sym)
		SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,GENERIC_ADD_UINT32(GENERIC_SUB(he->sym->current_address,he->sym->current_address),1),"$thumb",TRUE);
	      else sym->current_address=GENERIC_ADD_UINT32(GENERIC_SUB(he->sym->current_address,he->sym->current_address),1);
	    }
	    SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,he2->sym->current_address,"$code_switch",TRUE);
	  }
	  else FATAL(("%s symbol not with sec or subsec, has %sbase, type is %d, address is @G",he2->sym->name, he2->sym->base ? "" : "no ",he2->sym->base->type, he2->sym->current_address));
	}
	he2=(t_symbol_he *) he2->node.equal;
	prev_arm = TRUE;
      }
      first = FALSE;
    }
  }
#endif

  /* Sort the $switch symbols */
  top=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table,"$switch");
  if (top)
  {
    while(top->node.equal)
    {
      for(he=(t_symbol_he *) top->node.equal;he!=NULL;he=(t_symbol_he *) he->node.equal)
      {
	if (GENERIC_LT(he->sym->current_address,top->sym->current_address))
	{
	  tmp=top->sym;
	  top->sym=he->sym;
	  he->sym=tmp;
	}
      }
      top=(t_symbol_he *) top->node.equal;
    }
  }
#ifdef THUMB_SUPPORT
  /* Sort the $code_switch symbols */
  top=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table,"$code_switch");
  if (top)
  {
    while(top->node.equal)
    {
      for(he=(t_symbol_he *) top->node.equal;he!=NULL;he=(t_symbol_he *) he->node.equal)
      {
	if (GENERIC_LT(he->sym->current_address,top->sym->current_address))
	{
	  tmp=top->sym;
	  top->sym=he->sym;
	  he->sym=tmp;
	}
      }
      top=(t_symbol_he *) top->node.equal;
    }
  }
#endif
}

/*! 
 *
 * \param parent the linked executable 
 * \param obj the subobject
 *
 * If a part (section) of an object is not needed, certain linkers might remove
 * it from the objectfile. We remove them too to avoid confusion.  This function
 * finds (and removes) the removed sections for one subobject.  */

void ObjectFindLinkerRemovedSections(t_object * parent, t_object * obj)
{
  t_uint32 tel,tel2;
  t_section * del;
  /* Remove Code   Sections {{{ */
  for (tel=0; tel<obj->ncodes; tel++)
  {
    /* If it not mapped, it's removed by the linker */
    if (!(SECTION_IS_MAPPED(obj->code[tel]))) 
    {
      t_symbol * sym, * tmp;
      t_reloc * rel, *tmpr;
      t_section * ident=NULL;

      /* Output the removed sections, but only when the size is bigger then 0 {{{ */
      if (!GENERICISNULL(SECTION_OLD_SIZE(obj->code[tel])))
	printf("Linker skipped section %s in %s\n",obj->code[tel]->name, obj->name);
      /* }}} */
      /* Shrink the code array, remember the pointer in del {{{ */
      del=obj->code[tel];
      
      for (tel2=tel+1; tel2<obj->ncodes;tel2++)
	obj->code[tel2-1]=obj->code[tel2];
      obj->ncodes--;

      tel--;
      /* }}} */
      /* For common defined sections, see if we can find a common section. If
       * found keep the identical/common section it in ident, else ident=NULL
       * {{{ */

      if (del->is_common_defined) 
      {
	t_object * tmp, * tmpv;
	for(tmp=parent->mapped_first; tmp!=NULL; tmp=tmp->next)
	{
	  for (tmpv=tmp; tmpv!=NULL; tmpv=tmpv->equal)
	  {
	    ident=SectionGetFromObjectByName(tmpv,del->name);
	    if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	  }
	  if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	}


	if (!((ident)&&(SECTION_IS_MAPPED(ident))))
	{
	  ident=NULL;
	}
      }

      /* }}} */
      /* Delete/Move relocs from/to deleted/identical sections {{{ */
      if (!obj->reloc_table) WARNING(("No reloc table yet. Check this!"));
      else
	OBJECT_FOREACH_RELOC_SAFE(obj,rel,tmpr)
	{
	  if (T_SECTION(REL_FROM(rel))==del)
	  {
	    t_symbol * rem=REL_TO_SYMBOL(rel);
	    
	    RelocTableRemoveReloc(obj->reloc_table,rel);
	    /* No use to keep undefined symbols if refcount of symbol == 0 */
	    if (rem->relocrefcount==0 && rem->type==UNDEFINED_SYMBOL )
	    {
	      rem->type=REMOVED_SYMBOL;
	      rem->to_type=TO_REMOVED;
	    }

	  }
	  else if (REL_IS_TO_SYMBOL(rel) && (T_SECTION(REL_TO_SYMBOL(rel)->base)==del))
	  {
	    /* this is a patch for some symbols to empty sections in mips/arm {{{ */
	    if (!ident) if (PatternMatch("*/crtbegin.o*",obj->name) || PatternMatch("*/crtbeginT.o*",obj->name))
	    {
	      t_object * tmp, * tmpv;
	      for(tmp=parent->mapped_first; tmp!=NULL; tmp=tmp->next)
	      {
		for (tmpv=tmp; tmpv!=NULL; tmpv=tmpv->equal)
		{
		  ident=SectionGetFromObjectByName(tmpv,del->name);
		  if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
		}
		if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	      }
	      if ((ident)&&(!SECTION_IS_MAPPED(ident))) FATAL(("No parent with same name"));
	      ident=ident->parent_section;
	      REL_TO_SYMBOL(rel)->offset_from_start=GENERIC_SUB(REL_TO_SYMBOL(rel)->offset_from_start,REL_TO_SYMBOL(rel)->offset_from_start);
	      REL_TO_SYMBOL(rel)->current_address=SECTION_GET_CADDRESS(ident);
	    }

	    /*}}}*/
	    if (ident)
	    {
	      RelocSetToSymbolBase(rel,T_RELOCATABLE(ident));
	    }
	    else /* filename inzetten enzo */
	    {
	      RelocRemoveToRefs(rel);
	      REL_TO_SYMBOL(rel)->type=REMOVED_SYMBOL;
	      REL_TO_SYMBOL(rel)->to_type=TO_REMOVED;
	      REL_TO_SYMBOL(rel)->base=NULL;
	    }
	  }
	}
      /* }}} */
      /* Delete all symbols pointing to the removed section {{{ */
      OBJECT_FOREACH_SYMBOL_SAFE(obj,sym,tmp)
      {
	if ((sym->to_type==TO_SECTION) || (sym->to_type==TO_SUB_SECTION))
	{
	  if (T_SECTION(sym->base)==del)
	  {
	    if (!GENERICISNULL(SECTION_OLD_SIZE(del))) printf("Deleting symbol %s\n",sym->name);
	    if (sym->prev) sym->prev->next=sym->next;
	    else obj->symbol_table->first=sym->next;
	    if (sym->next) sym->next->prev=sym->prev;
	    else obj->symbol_table->last=sym->prev;
	    Free(sym->name);
	    Free(sym);
	  }
	}
      }
      /* }}} */
      /* Free the deleted section {{{ */
      SectionFree(del);
      /* }}} */
    }
  }
  /* }}} */
  /* Remove Rodata Sections {{{ */
  for (tel=0; tel<obj->nrodatas; tel++)
  {
    /* If it not mapped, it's removed by the linker */
    if (!(SECTION_IS_MAPPED(obj->rodata[tel])))
    {
      t_symbol * sym, * tmp;
      t_reloc * rel, *tmpr;
      t_section * ident=NULL;

      /* Output the removed sections, but only when the size is bigger then 0 {{{ */
      if (!GENERICISNULL(SECTION_OLD_SIZE(obj->rodata[tel])))
	printf("Linker skipped section %s in %s\n",obj->rodata[tel]->name, obj->name);
      /* }}} */
      /* Shrink the rodata array, remember the pointer in del {{{ */
      del=obj->rodata[tel];
      
      for (tel2=tel+1; tel2<obj->nrodatas;tel2++)
	obj->rodata[tel2-1]=obj->rodata[tel2];
      obj->nrodatas--;

      tel--;
      /* }}} */
      /* For common defined sections, see if we can find a common section. If
       * found keep the identical/common section it in ident, else ident=NULL
       * {{{ */

      if (del->is_common_defined) 
      {
	t_object * tmp, * tmpv;
	for(tmp=parent->mapped_first; tmp!=NULL; tmp=tmp->next)
	{
	  for (tmpv=tmp; tmpv!=NULL; tmpv=tmpv->equal)
	  {
	    ident=SectionGetFromObjectByName(tmpv,del->name);
	    if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	  }
	  if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	}


	if (!((ident)&&(SECTION_IS_MAPPED(ident))))
	{
	  ident=NULL;
	}
      }

      /* }}} */
      /* Delete/Move relocs from/to deleted/identical sections {{{ */
      if (!obj->reloc_table) WARNING(("No reloc table yet. Check this!"));
      else
	OBJECT_FOREACH_RELOC_SAFE(obj,rel,tmpr)
	{
	  if (T_SECTION(REL_FROM(rel))==del)
	  {
	    t_symbol * rem=REL_TO_SYMBOL(rel);
	    
	    RelocTableRemoveReloc(obj->reloc_table,rel);
	    /* No use to keep undefined symbols if refcount of symbol == 0 */
	    if (rem->relocrefcount==0 && rem->type==UNDEFINED_SYMBOL )
	    {
	      rem->type=REMOVED_SYMBOL;
	      rem->to_type=TO_REMOVED;
	    }
	  }
	  else if (REL_IS_TO_SYMBOL(rel) && (T_SECTION(REL_TO_SYMBOL(rel)->base)==del))
	  {
	    /* this is a patch for some symbols to empty sections in mips/arm {{{ */
	    if (!ident) if (PatternMatch("*/crtbegin.o*",obj->name) || PatternMatch("*/crtbeginT.o*",obj->name))
	    {
	      t_object * tmp, * tmpv;
	      for(tmp=parent->mapped_first; tmp!=NULL; tmp=tmp->next)
	      {
		for (tmpv=tmp; tmpv!=NULL; tmpv=tmpv->equal)
		{
		  ident=SectionGetFromObjectByName(tmpv,del->name);
		  if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
		}
		if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	      }
	      if ((ident)&&(!SECTION_IS_MAPPED(ident))) FATAL(("No parent with same name"));
	      ident=ident->parent_section;
	      REL_TO_SYMBOL(rel)->offset_from_start=GENERIC_SUB(REL_TO_SYMBOL(rel)->offset_from_start,REL_TO_SYMBOL(rel)->offset_from_start);
	      REL_TO_SYMBOL(rel)->current_address=SECTION_GET_CADDRESS(ident);
	    }
	    /*}}}*/
	    if (ident)
	    {
	      RelocSetToSymbolBase(rel,T_RELOCATABLE(ident));
	    }
	    else
	    {
	      RelocRemoveToRefs(rel);
	      REL_TO_SYMBOL(rel)->type=REMOVED_SYMBOL;
	      REL_TO_SYMBOL(rel)->to_type=TO_REMOVED;
	      REL_TO_SYMBOL(rel)->base=NULL;
	    }
	  }
	}
      /* }}} */
      /* Delete all symbols pointing to the removed section {{{ */
      OBJECT_FOREACH_SYMBOL_SAFE(obj,sym,tmp)
      {
	if ((sym->to_type==TO_SECTION) || (sym->to_type==TO_SUB_SECTION))
	{
	  if (T_SECTION(sym->base)==del)
	  {
	    if (!GENERICISNULL(SECTION_OLD_SIZE(del))) printf("Deleting symbol %s\n",sym->name);
	    if (sym->prev) sym->prev->next=sym->next;
	    else obj->symbol_table->first=sym->next;
	    if (sym->next) sym->next->prev=sym->prev;
	    else obj->symbol_table->last=sym->prev;
	    Free(sym->name);
	    Free(sym);
	  }
	}
      }
      /* }}} */
      /* Free the deleted section {{{ */
      SectionFree(del);
      /* }}} */
    }
  }
  /* }}} */
  /* Remove Data   Sections {{{ */
  for (tel=0; tel<obj->ndatas; tel++)
  {
    /* If it not mapped, it's removed by the linker */
    if (!(SECTION_IS_MAPPED(obj->data[tel]))) 
    {
      t_symbol * sym, * tmp;
      t_reloc * rel, *tmpr;
      t_section * ident=NULL;

      /* Output the removed sections, but only when the size is bigger then 0 {{{ */
      if (!GENERICISNULL(SECTION_OLD_SIZE(obj->data[tel])))
	printf("Linker skipped section %s in %s\n",obj->data[tel]->name, obj->name);
      /* }}} */
      /* Shrink the data array, remember the pointer in del {{{ */
      del=obj->data[tel];
      
      for (tel2=tel+1; tel2<obj->ndatas;tel2++)
	obj->data[tel2-1]=obj->data[tel2];
      obj->ndatas--;

      tel--;
      /* }}} */
      /* For common defined sections, see if we can find a common section. If
       * found keep the identical/common section it in ident, else ident=NULL
       * {{{ */

      if (del->is_common_defined) 
      {
	t_object * tmp, * tmpv;
	for(tmp=parent->mapped_first; tmp!=NULL; tmp=tmp->next)
	{
	  for (tmpv=tmp; tmpv!=NULL; tmpv=tmpv->equal)
	  {
	    ident=SectionGetFromObjectByName(tmpv,del->name);
	    if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	  }
	  if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	}


	if (!((ident)&&(SECTION_IS_MAPPED(ident))))
	{
	  ident=NULL;
	}
      }

      /* }}} */
      /* Delete/Move relocs from/to deleted/identical sections {{{ */
      if (!obj->reloc_table) WARNING(("No reloc table yet. Check this!"));
      else
	OBJECT_FOREACH_RELOC_SAFE(obj,rel,tmpr)
	{
	  if (T_SECTION(REL_FROM(rel))==del)
	  {
	    t_symbol * rem=REL_TO_SYMBOL(rel);
	    
	    RelocTableRemoveReloc(obj->reloc_table,rel);
	    /* No use to keep undefined symbols if refcount of symbol == 0 */
	    if (rem->relocrefcount==0 && rem->type==UNDEFINED_SYMBOL )
	    {
	      rem->type=REMOVED_SYMBOL;
	      rem->to_type=TO_REMOVED;
	    }
	  }
	  else if (REL_IS_TO_SYMBOL(rel) && (T_SECTION(REL_TO_SYMBOL(rel)->base)==del))
	  {
	    /* this is a patch for some symbols to empty sections in mips/arm {{{ */
	    if (!ident) if (PatternMatch("*/crtbegin.o*",obj->name) || PatternMatch("*/crtbeginT.o*",obj->name))
	    {
	      t_object * tmp, * tmpv;
	      for(tmp=parent->mapped_first; tmp!=NULL; tmp=tmp->next)
	      {
		for (tmpv=tmp; tmpv!=NULL; tmpv=tmpv->equal)
		{
		  ident=SectionGetFromObjectByName(tmpv,del->name);
		  if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
		}
		if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	      }
	      if ((ident)&&(!SECTION_IS_MAPPED(ident))) FATAL(("No parent with same name"));
	      ident=ident->parent_section;
	      REL_TO_SYMBOL(rel)->offset_from_start=GENERIC_SUB(REL_TO_SYMBOL(rel)->offset_from_start,REL_TO_SYMBOL(rel)->offset_from_start);
	      REL_TO_SYMBOL(rel)->current_address=SECTION_GET_CADDRESS(ident);
	    }
	    /*}}}*/
	    if (ident)
	    {
	      RelocSetToSymbolBase(rel,T_RELOCATABLE(ident));
	    }
	    else
	    {
	      RelocRemoveToRefs(rel);
	      REL_TO_SYMBOL(rel)->type=REMOVED_SYMBOL;
	      REL_TO_SYMBOL(rel)->to_type=TO_REMOVED;
	      REL_TO_SYMBOL(rel)->base=NULL;
	    }
	  }
	}
      /* }}} */
      /* Delete all symbols pointing to the removed section {{{ */
      OBJECT_FOREACH_SYMBOL_SAFE(obj,sym,tmp)
      {
	if ((sym->to_type==TO_SECTION) || (sym->to_type==TO_SUB_SECTION))
	{
	  if (T_SECTION(sym->base)==del)
	  {
	    if (!GENERICISNULL(SECTION_OLD_SIZE(del))) printf("Deleting symbol %s\n",sym->name);
	    if (sym->prev) sym->prev->next=sym->next;
	    else obj->symbol_table->first=sym->next;
	    if (sym->next) sym->next->prev=sym->prev;
	    else obj->symbol_table->last=sym->prev;
	    Free(sym->name);
	    Free(sym);
	  }
	}
      }
      /* }}} */
      /* Free the deleted section {{{ */
      SectionFree(del);
      /* }}} */
    }
  }
  /* }}} */
  /* Remove Bss    Sections {{{ */
  for (tel=0; tel<obj->nbsss; tel++)
  {
    /* If it not mapped, it's removed by the linker */
    if (!(SECTION_IS_MAPPED(obj->bss[tel]))) 
    {
      t_symbol * sym, * tmp;
      t_reloc * rel, *tmpr;
      t_section * ident=NULL;

      /* Output the removed sections, but only when the size is bigger then 0 {{{ */
      if (!GENERICISNULL(SECTION_OLD_SIZE(obj->bss[tel])))
	printf("Linker skipped section %s in %s\n",obj->bss[tel]->name, obj->name);
      /* }}} */
      /* Shrink the bss array, remember the pointer in del {{{ */
      del=obj->bss[tel];

      for (tel2=tel+1; tel2<obj->nbsss;tel2++)
	obj->bss[tel2-1]=obj->bss[tel2];
      obj->nbsss--;

      tel--;
      /* }}} */
      /* For common defined sections, see if we can find a common section. If
       * found keep the identical/common section it in ident, else ident=NULL
       * {{{ */

      if (del->is_common_defined) 
      {
	t_object * tmp, * tmpv;
	for(tmp=parent->mapped_first; tmp!=NULL; tmp=tmp->next)
	{
	  for (tmpv=tmp; tmpv!=NULL; tmpv=tmpv->equal)
	  {
	    ident=SectionGetFromObjectByName(tmpv,del->name);
	    if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	  }
	  if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	}


	if (!((ident)&&(SECTION_IS_MAPPED(ident))))
	{
	  ident=NULL;
	}
      }

      /* }}} */
      /* Delete/Move relocs from/to deleted/identical sections {{{ */
      if (!obj->reloc_table) WARNING(("No reloc table yet. Check this!"));
      else
	OBJECT_FOREACH_RELOC_SAFE(obj,rel,tmpr)
	{
	  if (T_SECTION(REL_FROM(rel))==del)
	  {
	    t_symbol * rem=REL_TO_SYMBOL(rel);
	    
	    RelocTableRemoveReloc(obj->reloc_table,rel);
	    /* No use to keep undefined symbols if refcount of symbol == 0 */
	    if (rem->relocrefcount==0 && rem->type==UNDEFINED_SYMBOL )
	    {
	      rem->type=REMOVED_SYMBOL;
	      rem->to_type=TO_REMOVED;
	    }
	  }
	  else if (REL_IS_TO_SYMBOL(rel) && (T_SECTION(REL_TO_SYMBOL(rel)->base)==del))
	  {
	    /* this is a patch for some symbols to empty sections in mips/arm {{{ */
	    if (!ident) if (PatternMatch("*/crtbegin.o*",obj->name) || PatternMatch("*/crtbeginT.o*",obj->name))
	    {
	      t_object * tmp, * tmpv;
	      for(tmp=parent->mapped_first; tmp!=NULL; tmp=tmp->next)
	      {
		for (tmpv=tmp; tmpv!=NULL; tmpv=tmpv->equal)
		{
		  ident=SectionGetFromObjectByName(tmpv,del->name);
		  if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
		}
		if ((ident)&&(SECTION_IS_MAPPED(ident))) break;
	      }
	      if ((ident)&&(!SECTION_IS_MAPPED(ident))) FATAL(("No parent with same name"));
	      ident=ident->parent_section;
	      REL_TO_SYMBOL(rel)->offset_from_start=GENERIC_SUB(REL_TO_SYMBOL(rel)->offset_from_start,REL_TO_SYMBOL(rel)->offset_from_start);
	      REL_TO_SYMBOL(rel)->current_address=SECTION_GET_CADDRESS(ident);
	    }
	    /*}}}*/
	    if (ident)
	    {
	      RelocSetToSymbolBase(rel,T_RELOCATABLE(ident));
	    }
	    else
	    {
	      RelocRemoveToRefs(rel);
	      REL_TO_SYMBOL(rel)->type=REMOVED_SYMBOL;
	      REL_TO_SYMBOL(rel)->to_type=TO_REMOVED;
	      REL_TO_SYMBOL(rel)->base=NULL;
	    }
	  }
	}
      /* }}} */
      /* Delete all symbols pointing to the removed section {{{ */
      OBJECT_FOREACH_SYMBOL_SAFE(obj,sym,tmp)
      {
	if ((sym->to_type==TO_SECTION) || (sym->to_type==TO_SUB_SECTION))
	{
	  if (T_SECTION(sym->base)==del)
	  {
	    if (!GENERICISNULL(SECTION_OLD_SIZE(del))) printf("Deleting symbol %s\n",sym->name);
	    if (sym->prev) sym->prev->next=sym->next;
	    else obj->symbol_table->first=sym->next;
	    if (sym->next) sym->next->prev=sym->prev;
	    else obj->symbol_table->last=sym->prev;
	    Free(sym->name);
	    Free(sym);
	  }
	}
      }
      /* }}} */
      /* Free the deleted section {{{ */
      SectionFree(del);
      /* }}} */
    }
  }
  /* }}} */
}


/*!
 *
 * Verify that everything in the map is also in the parent object
 *  
 * \param obj The object for which we want to verify the map
 *
 * \return void 
*/

void ObjectSwallowChildObjects(t_object * obj)
{
  MapMoveObjectsToParent(obj,obj->map);
}

/* FUN : ObjectRemoveLinkerRemovedSections 
 * PAR : the parent object
 * RET : nothing 
 * DESC: If a part (section) of an object is not needed, 
 *       certain linkers might remove it from the objectfile
 *       We remove them too to avoid confusion. 
 *       This function removes these sections */

      

void ObjectRemoveLinkerRemovedSections(t_object * obj)
{
  t_object * tmp;
  t_object * tmpv;
  OBJECT_FOREACH_SUBOBJECT(obj,tmpv,tmp)
  {
     ObjectFindLinkerRemovedSections(obj,tmpv);
  }
}


/*}}}*/
/* {{{ Rebuild the data of an object's parent sections from the data of the subsections */
void ObjectRebuildSectionsFromSubsections(t_object * obj)
{
  t_section ** secarr;
  t_uint32 nsecs = obj->ncodes + obj->nrodatas + obj->ndatas + obj->nnotes;
  int tel;

  /* put all non-bss sections in a single array for easier access */
  secarr = Calloc(sizeof(t_section *), nsecs);
  memcpy(secarr,obj->code,sizeof(t_section *) * obj->ncodes);
  memcpy(secarr+obj->ncodes,obj->rodata,sizeof(t_section *) * obj->nrodatas);
  memcpy(secarr+obj->ncodes+obj->nrodatas,obj->data,sizeof(t_section *) * obj->ndatas);
  memcpy(secarr+obj->ncodes+obj->nrodatas+obj->ndatas,obj->note,sizeof(t_section *) * obj->nnotes);
  
  for (tel = 0; tel < nsecs; tel++)
  { 
    t_section * sec = secarr[tel];
    t_section ** subsecs;
    t_uint32 nsubsecs;
    int tel2;
    void * newdata;

    /* skip disassembled code sections */
    if ((SECTION_TYPE(sec) != CODE_SECTION) && (SECTION_TYPE(sec) != RO_DATA_SECTION) && (SECTION_TYPE(sec) != DATA_SECTION))
      continue;

    /* DiabloPrint(stdout,"Size of section %s= @G\n",sec->name,SECTION_CSIZE(sec)); */

    newdata = Calloc(sizeof(char), GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)));
    /* check if the layout script specifies a specific filler for this section */
    {
      t_layout_rule * script, * iter;

      script = obj->layout_script;
      if (!script) script = ObjectGetAndParseLayoutScript(obj);

      for (iter = script; iter; iter = iter->next)
      {
	if (iter->kind != SECTION_SPEC) continue;
	if (strcmp(iter->u.secspec->name, sec->name)) continue;

	if (iter->u.secspec->filler_exp)
	{
	  t_uint32 i, size;
	  t_uint32 filler;
	  char * fillchars = (char *) &filler;

	  filler = (t_uint32) LayoutScriptCalcExp(iter->u.secspec->filler_exp, 0LL /* this is just a dummy */, obj);
	  size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec));
	  for (i = 0; i < size/4; i++)
	    ((t_uint32*)newdata)[i] = filler;
	  /* fill the last (size % 4) bytes */
	  for (i = 0; i < size % 4; i++)
	    ((char *)newdata)[4*(size/4) + i] = fillchars[i];
	}
	/* there should only be one secspec for every section name so
	 * we can break the for loop here */
	break;
      }
    }

    subsecs = SectionGetSubsections(sec,&nsubsecs);
    for (tel2 = 0; tel2 < nsubsecs; tel2++)
    {
      t_section * subsec = subsecs[tel2];
      t_uint32 offset = GENERIC_EXTRACT_UINT32(GENERIC_SUB(SECTION_GET_CADDRESS(subsec),SECTION_GET_CADDRESS(sec)));
      t_uint32 size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(subsec));
      memcpy(((char *)newdata)+offset,subsec->data,(size_t) size);
    }
    if (subsecs) Free(subsecs);

    if (sec->data) Free(sec->data);
    sec->data = newdata;
  }

  /* clean up */
  Free(secarr);
}
/* }}} */

/* {{{ get and parse layout script for obj */
t_layout_rule * ObjectGetAndParseLayoutScript(t_object * obj)
{
  t_string scriptname = NULL;

  if (global_options.layoutscript_set)
  {
    /* layout script given at the command line */
    printf("Using command line specified layout script %s\n", global_options.layoutscript);
    scriptname = StringDup(global_options.layoutscript);
  }
  else
  {
    /* use the layout script embedded in the linker script */
    scriptname = GetLinkerScriptFilename(obj);
  }
  VERIFY_FASSERT(scriptname,("Could not find an appropriate layout script"));

  obj->layout_script = LayoutScriptParse(scriptname);
  Free(scriptname);

  return obj->layout_script;
}
/* }}} */

/* Emulate the link {{{ */
/* LinkObjectsFromMap {{{ */
void LinkObjectsFromMap(t_object * obj)
{
  /* Read (and parse) the map file associated with this object file */
  ObjectReadMap(obj);

  /* Read all the objects referenced in the map file */
  ObjectReadObjectInMap(obj);

  /* Check if the objects match the descriptions in the map file and swallows the children */
  ObjectSwallowChildObjects(obj); 

#ifdef ALPHA_SUPPORT
  /* in case it's an alpha object we produce a litamap (starting (and verifying
   * the completeness of) the ordering found in the previous step) */

  ObjectMapPatchAlpha(obj);
#endif


}
/* }}} */
/* link in an extra file {{{
 *
 * as a side effect of this, the sections will
 * be moved and possibly reordered (this is a 
 * limitation of the current PlaceSections algorithm:
 * we have no absolute control over the final order 
 * of the sections in the binary, in particular
 * sections that do not appear in the layout scripts) 
 *
 * in order to adequately report addresses later on, 
 * it might be necessary to build some kind of 
 * address translation table here. this is something
 * that remains to be done. */
void LinkObjectFile(t_object * obj, t_string linkin_fname)
{
  t_object * linkin = ObjectGet(linkin_fname,obj);
  t_section * entrysec;
  t_address entrysecbase;
  t_symbol * sym;

  /* prefix all symbols */
  OBJECT_FOREACH_SYMBOL(linkin,sym)
  {
    t_string tmp;

    /* don't prefix the symbols that indicate data and code */
    if (!strcmp(sym->name,"$a") || !strcmp(sym->name,"$d"))
      continue;

#ifdef ALPHA_SUPPORT
    if (!strcmp(sym->name,"_fpdata") || !strcmp(sym->name,"_ftext") || !strcmp(sym->name,"_etext"))
      continue;
#endif

    tmp = Collate(INSTRUMENT_PREFIX,sym->name);
    Free(sym->name);
    sym->name = tmp;
  }
  /* Do the common symbols */

  {
    t_section * target=obj->bss[0];
    t_address run=GENERIC_SUB(SECTION_GET_CADDRESS(target),SECTION_GET_CADDRESS(target));

    OBJECT_FOREACH_SYMBOL(linkin,sym)
    {
      if (sym->to_type!=TO_COMMON) continue;
      SymbolTableAddSymbolToSection(obj->symbol_table,GLOBAL_SYMBOL,T_RELOCATABLE(target),GENERIC_ADD(SECTION_GET_END_ADDRESS(target),run),sym->name,FALSE); 
      DiabloPrint(stdout,"Adding %s\n",sym->name);
      run=GENERIC_ADD_UINT32(run,sym->size);
    }

    /* add the sections of the linkin to the original binary */
    {
      t_section * sec;
      t_uint32 tel;

      /* first of all, find out in which section the entry point is defined.
       * we will have to adjust it after the call to PlaceSections */
      entrysec = NULL;
      for (tel=0; tel<obj->ncodes; tel++)
	if (GENERIC_LE(SECTION_GET_CADDRESS(obj->code[tel]),obj->entry) && GENERIC_GT(GENERIC_ADD(SECTION_GET_CADDRESS(obj->code[tel]),SECTION_CSIZE(obj->code[tel])),obj->entry))
	{
	  entrysec = obj->code[tel];
	  break;
	}
      VERIFY_FASSERT(entrysec,("Could not find the code section with the entry point!"));
      entrysecbase = SECTION_GET_CADDRESS(entrysec);

      OBJECT_FOREACH_SECTION(linkin,sec,tel)
      {
	t_section *** arr = SECTIONARRAYPOINTERBYNUMBER(obj,tel);
	t_section * parent=NULL;
        if (!parent)	
	{
	  t_uint32 * nsecs = SECTIONCOUNTPOINTERBYNUMBER(obj,tel);
	  if (! *nsecs) FATAL(("Implement the case where there are no sections of the specified type available yet"));
	  parent = SectionNew(obj,SECTION_TYPE((*arr)[0]));
	  *arr = Realloc(*arr,sizeof(t_reloc *)*(*nsecs+1));
	  (*arr)[*nsecs] = parent;
	  (*nsecs)++;
	  if (sec->data)
	  {
	    parent->data = Malloc(GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)));
	    memcpy(parent->data,sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)));
	  }
	  else
	  {
	    parent->data=NULL;
	  }
	  SECTION_NAME(parent) = Collate(INSTRUMENT_PREFIX,SECTION_NAME(sec));
	  SECTION_OLD_SIZE(parent) = SECTION_CSIZE(parent) = SECTION_CSIZE(sec);
	  SECTION_ALIGN(parent) = SECTION_ALIGN(sec);
	}
	else
	{
	  PRIVATE_SET_SECTION_CADDRESS(sec)=GENERIC_ADD(SECTION_GET_CADDRESS(parent),SECTION_CSIZE(parent));
	  SECTION_OLD_SIZE(parent) = SECTION_CSIZE(parent);
	}

	MAP_SECTION(sec,NULL,parent);
	DiabloPrint(stdout,"[Instrumentation] linked in section %s (parent obj= %s) (current size = @G)\n",SECTION_NAME(parent),parent->obj->name,SECTION_CSIZE(parent));
      }
    }
    PlaceSections(obj,FALSE);

    /* update the current_address field of all symbols */
    {
      t_object * subobj, * tmp;
      t_symbol * sym;

      OBJECT_FOREACH_SUBOBJECT(obj,subobj,tmp)
      {
	SYMBOLTABLE_FOREACH_SYMBOL(subobj->symbol_table,sym)
	{
	  if (sym->base) 
	  {
	    sym->current_address = GENERIC_ADD(sym->base->caddress,sym->offset_from_start);
	  }
	}
      }

      SYMBOLTABLE_FOREACH_SYMBOL(obj->symbol_table,sym)
	if (sym->base) 
	{
	  sym->current_address = GENERIC_ADD(sym->base->caddress,sym->offset_from_start);
	}
    }

    /* update the obj->entry field */
    DiabloPrint(stdout, "old entry @G offset @G new entry @G\n", obj->entry,GENERIC_SUB(SECTION_GET_CADDRESS(entrysec),entrysecbase),GENERIC_ADD(obj->entry,GENERIC_SUB(SECTION_GET_CADDRESS(entrysec),entrysecbase)));
    obj->entry = GENERIC_ADD(obj->entry,GENERIC_SUB(SECTION_GET_CADDRESS(entrysec),entrysecbase));
  }

}
/* }}} */
/* link in an extra file {{{
 *
 * as a side effect of this, the sections will
 * be moved and possibly reordered (this is a 
 * limitation of the current PlaceSections algorithm:
 * we have no absolute control over the final order 
 * of the sections in the binary, in particular
 * sections that do not appear in the layout scripts) 
 *
 * in order to adequately report addresses later on, 
 * it might be necessary to build some kind of 
 * address translation table here. this is something
 * that remains to be done. */


#define BSS_TO_DATA

void LinkObjectFileNew(t_object * obj, t_string linkin_fname)
{
  t_object * linkin = ObjectGet(linkin_fname,obj);
  t_section * entrysec;
  t_address entrysecbase;
  t_symbol * sym;
  t_uint32 tel;
  t_section * sec,* bss_parent;
  t_reloc * rel, *reln;
  t_address run;
  t_object * linker_obj;
  entrysec = NULL;
  for (tel=0; tel<obj->ncodes; tel++)
    if (GENERIC_LE(SECTION_GET_CADDRESS(obj->code[tel]),obj->entry) && GENERIC_GT(GENERIC_ADD(SECTION_GET_CADDRESS(obj->code[tel]),SECTION_CSIZE(obj->code[tel])),obj->entry))
    {
      entrysec = obj->code[tel];
      break;
    }
  VERIFY_FASSERT(entrysec,("Could not find the code section with the entry point!"));
  entrysecbase = SECTION_GET_CADDRESS(entrysec);

  if (!obj->bss) FATAL(("Implement adding bss to an object without bss!"));
  bss_parent=obj->bss[0];
  linker_obj=ObjectGetFromCache("Linker",obj);
  if (!linker_obj) 
  {
    linker_obj=ObjectNewCached("Linker",obj);
    linker_obj->section_table=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,SectionHashNodeFree);
    obj->mapped_last->next=linker_obj;
    linker_obj->next=NULL;
    obj->mapped_last=linker_obj;
    linker_obj->reloc_table=RelocTableNew(); 
    linker_obj->symbol_table=SymbolTableNew(linker_obj);
  }
  run=GENERIC_SUB(SECTION_GET_CADDRESS(entrysec),SECTION_GET_CADDRESS(entrysec));
  
  /* Add sections */
  OBJECT_FOREACH_SECTION(linkin,sec,tel)
  {
    t_section *** arr;
    t_section * parent=NULL;
#ifdef BSS_TO_DATA
    t_bool bss=FALSE;
    if (((tel & 0x1f00000)==0x400000)) bss=TRUE;
    if (bss) 
      arr = &obj->data;
    else 
#endif
      arr = SECTIONARRAYPOINTERBYNUMBER(obj,tel);


    if (!parent)	
    {
      t_uint32 * nsecs;
#ifdef BSS_TO_DATA
      if (bss) 
	nsecs = &obj->ndatas;
      else 
#endif
	nsecs= SECTIONCOUNTPOINTERBYNUMBER(obj,tel);
      
      if (!*nsecs) 
      {
	printf("%x\n",tel & 0x1f00000);
	if ((tel & 0x1f00000)==0x100000)
	{
	  parent = SectionNew(obj,DATA_SECTION);
	}
	else
	{
	FATAL(("Implement the case where there are no sections of the specified type available yet"));
	exit(0);
	}
      }
      else
      {
	parent = SectionNew(obj,SECTION_TYPE((*arr)[0]));
      }

      *arr = Realloc(*arr,sizeof(t_reloc *)*(*nsecs+1));
      (*arr)[*nsecs] = parent;
      (*nsecs)++;

      if (sec->data 
#ifdef BSS_TO_DATA
	  || bss	  
#endif
	  )
      {
	parent->data = Malloc(GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)));
#ifdef BSS_TO_DATA
	if (bss)
	{
	  memset(parent->data,0,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)));
	  if (sec->data) FATAL(("Bss section already allocated!"));
	  sec->data  = Malloc(GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)));
	  memset(sec->data,0,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)));
	}
	else
#endif
	  memcpy(parent->data,sec->data,GENERIC_EXTRACT_UINT32(SECTION_CSIZE(sec)));
      }
      else
      {
	parent->data=NULL;
      }
      SECTION_NAME(parent) = Collate(INSTRUMENT_PREFIX,SECTION_NAME(sec));
      SECTION_OLD_SIZE(parent) = SECTION_CSIZE(parent) = SECTION_CSIZE(sec);
      SECTION_ALIGN(parent) = SECTION_ALIGN(sec);
      PRIVATE_SET_SECTION_CADDRESS(parent)=run;
      PRIVATE_SET_SECTION_OLD_ADDRESS(parent)=run;
      PRIVATE_SET_SECTION_CADDRESS(sec)=run;
      PRIVATE_SET_SECTION_OLD_ADDRESS(sec)=run;
    }
    else
    {
      FATAL(("AYAYAY\n"));
      PRIVATE_SET_SECTION_CADDRESS(sec)=GENERIC_ADD(SECTION_GET_CADDRESS(parent),SECTION_CSIZE(parent));
      SECTION_OLD_SIZE(parent) = SECTION_CSIZE(parent);
    }

    MAP_SECTION(sec,NULL,parent);
    T_RELOCATABLE(sec)->type=RT_SUBSECTION;
    DiabloPrint(stdout,"[Instrumentation] linked in section %s (parent obj= %s) (current size = @G)\n",SECTION_NAME(parent),parent->obj->name,SECTION_CSIZE(parent));
  }

  /* prefix all symbols */
  OBJECT_FOREACH_SYMBOL(linkin,sym)
  {
    t_string tmp;

    DiabloPrint(stdout,">>>> @S\n",sym);

    /* don't prefix the symbols that indicate data and code */
    if (strcmp(sym->name,"$a") && strcmp(sym->name,"$d"))
#ifdef ALPHA_SUPPORT
    if (strcmp(sym->name,"_fpdata") && strcmp(sym->name,"_ftext") && strcmp(sym->name,"_etext"))
#endif
    {
      tmp = Collate(INSTRUMENT_PREFIX,sym->name);
      Free(sym->name);
      sym->name = tmp;
    }


    if ((sym->base)&&(RELOCATABLE_TYPE(sym->base)==RT_SUBSECTION))
    {
      sym->mapped=SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,sym->local?LOCAL_SYMBOL:GLOBAL_SYMBOL,T_RELOCATABLE(sym->base),sym->offset_from_start,sym->name,sym->local); 
      sym->mapped->addend=sym->addend;
      DiabloPrint(stdout,"Mapped @S\n",sym->mapped);
    }
    else if (sym->type==ABS_SYMBOL) {
      sym->mapped=SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,sym->current_address,sym->name,sym->local);
      sym->mapped->addend=sym->addend;
      } 
    else if (sym->to_type==TO_COMMON) {
      /* the section doesn't exist yet: create it */
      t_section * tmp_sec = SectionNew(linker_obj,BSS_SECTION);
      T_RELOCATABLE(tmp_sec)->type=RT_SUBSECTION;
      tmp_sec->name = StringDup(sym->name);

      /* for COMMON symbols, the size is stored in sym->size, the alignment is stored in sym->current_address */
      SECTION_OLD_SIZE(tmp_sec)=sym->size;
      SECTION_CSIZE(tmp_sec)=sym->size;
      SECTION_ALIGN(tmp_sec)=sym->current_address;
      MAP_SECTION(tmp_sec,NULL,bss_parent);

      SectionSetOldAddressSuper(tmp_sec,run);
      SectionSetCAddressSuper(tmp_sec,run);

      linker_obj->bss=Realloc(linker_obj->bss,(linker_obj->nbsss+1)*sizeof(t_section *));
      linker_obj->bss[linker_obj->nbsss]=tmp_sec;
      linker_obj->nbsss++;
      HashInsert(linker_obj->section_table,(data_hash_node *) SectionHashNodeNew(tmp_sec->name,tmp_sec));

      /* Set the offset_from_start explicitly to zero, because this has no meaning */
      sym->mapped=SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,GLOBAL_SYMBOL,T_RELOCATABLE(tmp_sec),GENERIC_SUB(sym->offset_from_start,sym->offset_from_start),sym->name,FALSE); 
      sym->mapped->addend=sym->addend;
    }
    else if (sym->type==WEAK_UNDEFINED_SYMBOL)
    {
    }
    else {
      sym->mapped=SymbolTableGetSymbolByName(obj->sub_symbol_table,sym->name);
      if (!sym->mapped) FATAL(("Implement no base @S\n",sym));
      DiabloPrint(stdout,"Mapped undefined symbol %s @S\n",sym->name,sym->mapped);
    }
  }

  if (linkin->reloc_table)
  {
    t_symbol * extra_sym=NULL;
    for (rel=linkin->reloc_table->first; rel!=NULL; rel=reln)
    {
      if (RELOCATABLE_TYPE(REL_FROM(rel)) != RT_SUBSECTION) FATAL(("Not from section @R!",rel));
      if (SECTION_IS_MAPPED(T_SECTION(REL_FROM(rel))))
      {
	/* Some consistency checks {{{ */
	if (REL_IS_TO_SYMBOL(rel)) {
	  if (REL_TO_SYMBOL(rel)->mapped==NULL) FATAL(("Could not merge @R (%s from object %s)!",rel,REL_TO_SYMBOL(rel)->name,linkin->name));
	  if (!rel->e_to_symbol) FATAL(("Reloc @R not with extra symbol!",rel)); 
	}
	/* }}} */

	if (REL_E_IS_TO_SYMBOL(rel) && REL_E_TO_SYMBOL(rel))
	{
	  if (REL_E_TO_SYMBOL(rel)->mapped==NULL)
	    FATAL(("Could not merge @R (%s from object %s)!",rel,rel->e.sym->name,linkin->name));
	  extra_sym=rel->e.sym->mapped;
	}

	reln=REL_NEXT(rel);
	if (REL_IS_TO_SYMBOL(rel)) {
	  RelocSetToSymbol(rel,REL_TO_SYMBOL(rel)->mapped);
	}
	if (REL_E_IS_TO_SYMBOL(rel) && REL_E_TO_SYMBOL(rel))
	{
	  RelocSetEToSymbol(rel,extra_sym);
	}
	RelocTableUnlinkReloc(linkin->reloc_table,rel);
	RelocTableLinkReloc(obj->reloc_table,rel); 
      }
      else
      {
	DiabloPrint(stdout,"Verify: skipping @R\n",rel);
	reln=REL_NEXT(rel);
	RelocTableRemoveReloc(linkin->reloc_table,rel);
      }
    }
  }

  RelocsMigrateToRelocatables(obj);


  PlaceSections(obj,FALSE);

  SYMBOLTABLE_FOREACH_SYMBOL(obj->sub_symbol_table,sym)
  {
   
    if (sym->base) 
    {
      sym->current_address = GENERIC_ADD(sym->base->caddress,sym->offset_from_start);
/*      DiabloPrint(stdout,"Symbol %s from @G at  @G\n",sym->name,BBL_CADDRESS(sym->base),sym->current_address); */
    }
  }
  obj->entry = GENERIC_ADD(obj->entry,GENERIC_SUB(SECTION_GET_CADDRESS(entrysec),entrysecbase));

  printf("Relocating!\n");
  RelocateCore(obj,FALSE);
  ObjectRebuildSectionsFromSubsections(obj);
  return;
}
/* }}} */
/* LinkMakeFinalObject {{{ */
void LinkMakeFinalObject(t_object * obj)
{
  /* Replay whatever the linker did with the subobjects */

  /* Remove dead (sub)sections from the object (=linker removed for ARM, =
   * xdata, pdata for the alpha) */
  ObjectRemoveLinkerRemovedSections(obj);

  /* Create a symbol- and reloc-table with the information from the subobjects */
  ObjectBuildUnifiedSymbolTableAndRelocTable(obj);

}

void LinkRelocate(t_object * obj, t_bool verify)
{  
  if (obj->mapped_first) ObjectRelocateSubObjects(obj,verify);

  /* in regular operation mode, this is some kind of a noop,
   * but in instrumentation mode, this is necessary because
   * some code has been linked in, so all addresses have
   * been changed */
  ObjectRebuildSectionsFromSubsections(obj);
}
/*}}}*/
/* LinkGetParent {{{ */
t_object * LinkGetParent()
{
  t_object * obj;

  /* Read the main object */
  obj=ObjectGet(global_options.objectfilename,NULL);

  if (obj->reloc_table && obj->reloc_table->first ) 
  {
    RelocTableFree(obj->reloc_table);
    obj->reloc_table=RelocTableNew();
  }
  return obj;
}
/* }}} */
/* LinkEmulate {{{ */
t_object * LinkEmulate(char * name)
{
  t_object * obj=LinkGetParent();

  LinkObjectsFromMap(obj);
  
  LinkMakeFinalObject(obj);
#ifdef ARM_ADS_PATCH
  {
    t_symbol * sym = SymbolTableGetSymbolByName(obj->sub_symbol_table,".constdata___stdout_name");
    ((char *) GENERIC_ADD_DISPL(T_SECTION(sym->base)->data,sym->offset_from_start))[0]='o' ;
    sym = SymbolTableGetSymbolByName(obj->sub_symbol_table,".constdata___stdin_name");
    ((char *) GENERIC_ADD_DISPL(T_SECTION(sym->base)->data,sym->offset_from_start))[0]='i' ;
    sym = SymbolTableGetSymbolByName(obj->sub_symbol_table,".constdata___stderr_name");
    ((char *) GENERIC_ADD_DISPL(T_SECTION(sym->base)->data,sym->offset_from_start))[0]='e' ;
  }
#endif
  
  LinkRelocate(obj,TRUE);

  return obj;
}
/* }}} */
/* }}} */


/* vim: set shiftwidth=2 foldmethod=marker: */
