/************************************************************
 *{{{ Copyright 2001,2002,2003:
 * 			Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Headers for flowgraph manipulation 
 *
 * Written by Ludo Van Put (lvanput@elis.ugent.be)
 *
 * Copyright (c) 2002 Ludo Van Put
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Ludo Van Put (lvanput@elis.ugent.be) }}}
 ************************************************************/

#include "diablo_loop.h"
/*#include "../datastructures/data_stack.h"*/
#include "diablo_dominator.h"
#include "diablo_iterators.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include "../datastructures/data_ll.h"

#define DEBUG_LOOP
typedef struct _t_queue
{
  dll * start;
  dll * end;
} t_queue;

#define MAX_STACK_HEIGHT 15000
typedef struct _t_bbl_stack
{
  t_uint32 top;
  t_bbl * stack[MAX_STACK_HEIGHT];
} data_stack;
#define STACK_NOT_EMPTY(stack)	((stack)->top)

#define QUEUE_ADD(elem,queue)	DLLAppend(NULL,elem,&queue.start,&queue.end);
#define QUEUE_GET(elem,queue)	(elem = (data_dll_elem*)queue.end, DLLUnlink(queue.end,&queue.start,&queue.end))
#define QUEUE_NOT_EMPTY(queue)	((queue).start != NULL)

static t_bool LoopIsNested(t_loop * nest, t_loop * parent);
void LoopMerge(t_loop * loop_a, t_loop * loop_b);
static void LoopRemoveChildFromParent(t_loop * parent, t_loop * child);
static void LoopRemoveParentFromChild(t_loop * child, t_loop * parent);
static void Push(data_stack *stack, t_bbl* bbl);
static t_bbl * Pop(data_stack *stack);
static t_bool BblAddToLoop(t_bbl * bbl, t_loop * loop);
static t_bool BblAddToLoop2(t_bbl * bbl, t_loop * loop);
void LoopAddFunction(t_function * fun,t_loop * loop);
static t_bool LoopContainsFun(t_loop * loop, t_function * fun);


/*! Add a new loop, with backedge, to the loop in the flowgraph.
 * Also add all bbl's that are in the loop and add an entry in bbl_loops
 * of these bbl's. */
t_loop * LoopNew(t_cfg * cfg, t_cfg_edge * backedge)
/*{{{ */
{
  t_bbl * header, * footer, *working_bbl;
  data_ll_elem * new_backedge;
  data_stack working_stack = {0};
  t_loop * ret = (t_loop*) Calloc(1,sizeof(t_loop));
  t_cfg_edge *iter, *i_edge;
  t_function * fun;
  t_bbl * bbl;
  t_bool big, small;
  t_uint32 printinfo = 1;
/*  t_uint32 nnodes = 0;*/

  /* insert the loop in the dll of loops {{{*/
  if(cfg->loop_first==NULL)
  {
    cfg->loop_first = ret;
    cfg->loop_last = ret;
  }
  else
  {
    ret->prev = cfg->loop_last;
    cfg->loop_last->next = ret;
    cfg->loop_last=ret;
  }

  new_backedge = LLCalloc(backedge);
  LLAppendSorted(new_backedge,&LOOP_BACKEDGES(ret));
  LOOP_CFG(ret) = cfg;
  ret->infinite = TRUE;
  /*}}}*/


  header = CFG_EDGE_TAIL(backedge);
  footer = CFG_EDGE_HEAD(backedge);

  /*printinfo = (BBL_OLD_ADDRESS(header) != 0x14128);*/

  VERBOSE(printinfo,("---------------------New loop--------------------------\nHeader: @B\nFooter: @B\nLoop %p",header,footer,ret));

  /*{{{ Mark all forward ip edges that may be followed */
  if(BBL_FUN(header) != BBL_FUN(footer))
  {
    InitNodeMarking();
    InitNodeMarking2();
    InitEdgeMarking();
    NodeMark(footer);
    small = TRUE;

    CFG_WHILE_MARKED_BBLS_PER_FUN(cfg,fun,bbl,big,small)
    {
      if(!NodeIsMarked2(bbl))
      {
	NodeMark2(bbl);
	VERBOSE(printinfo,("Treating @B",bbl));
	if(!IsHellNode(bbl))
	  /*
	     if (FUN_FLAGS(fun) & FF_HAS_INCOMING_IPJUMP)
	     {
	     t_bbl * tmp_bbl;
	     FUN_FOREACH_BBL(fun,tmp_bbl)
	     {
	     NodeMark2(bbl);
	     NodeUnmark(bbl);
	     NodeMark(bbl);
	     }

	     BBL_FOREACH_PRED_EDGE(FUN_BBL_FIRST(fun),i_edge)
	     if (EDGE_CAT(i_edge)==ET_CALL)
	     if(!NodeIsMarked(CFG_EDGE_HEAD(i_edge)))
	     {
	     big = TRUE;
	     FUN_FLAGS(BBL_FUN(CFG_EDGE_HEAD(i_edge))) |= FF_IS_MARKED;
	     }
	     }
	     else*/
	  BBL_FOREACH_PRED_EDGE(bbl,i_edge)
	  {
	    if(EDGE_CAT(i_edge) == ET_FUNLINK) continue;
	    VERBOSE(printinfo,("  Treating @E",i_edge));
	    if(CfgEdgeIsBackwardInterproc(i_edge))
	    {
	      if(!NodeIsMarked(CFG_EDGE_HEAD(EDGE_CORR(i_edge))))
		small = TRUE;

	      NodeMark(CFG_EDGE_HEAD(EDGE_CORR(i_edge)));
	      VERBOSE(printinfo,("      Marking @B",CFG_EDGE_HEAD(EDGE_CORR(i_edge))));
	    }
	    else
	    {
	      if(CfgEdgeIsForwardInterproc(i_edge))
	      {
		VERBOSE(printinfo,("    Marking edge @E",i_edge));
		EdgeMark(i_edge);
	      }
	      if(!NodeIsMarked(CFG_EDGE_HEAD(i_edge)))
	      {
		if(CfgEdgeIsForwardInterproc(i_edge))
		{
		  big =TRUE;
		  FUN_FLAGS(BBL_FUN(CFG_EDGE_HEAD(i_edge))) |= FF_IS_MARKED;
		}
		else small = TRUE;
		NodeMark(CFG_EDGE_HEAD(i_edge));
		VERBOSE(printinfo,("      Marking @B",CFG_EDGE_HEAD(i_edge)));
	      }
	    }
	  }
      }
    }
  }
  /*}}}*/

  /* Mark all nodes in the loop */
  CFG_FOREACH_FUN(cfg,fun) FUN_FLAGS(fun)&=~FF_IS_MARKED;
  InitNodeMarking();
  InitNodeMarking2();

  NodeMark(footer);
  Push(&working_stack,footer);
  VERBOSE(printinfo,("Pushing @B",footer));

  while(STACK_NOT_EMPTY(&working_stack))
  {
    working_bbl = Pop(&working_stack);

    VERBOSE(printinfo,("Adding @B",working_bbl));
    if(!BblAddToLoop(working_bbl,ret)) FATAL(("Bbl @B was already added to the loop!",working_bbl));
/*    else nnodes++;*/
    if(working_bbl == BBL_CFG(working_bbl)->exit_hell_node)
    {
      if(!NodeIsMarked(BBL_CFG(working_bbl)->hell_node))
      {
	NodeMark(BBL_CFG(working_bbl)->hell_node);
	Push(&working_stack,BBL_CFG(working_bbl)->hell_node);
  VERBOSE(printinfo,("Pushing @B",BBL_CFG(working_bbl)->hell_node));
      }
      continue;
    }
    if(working_bbl == BBL_CFG(working_bbl)->exit_call_hell_node)
    {
      if(!NodeIsMarked(BBL_CFG(working_bbl)->call_hell_node))
      {
	NodeMark(BBL_CFG(working_bbl)->call_hell_node);
	Push(&working_stack,BBL_CFG(working_bbl)->call_hell_node);
  VERBOSE(printinfo,("Pushing @B",BBL_CFG(working_bbl)->call_hell_node));
      }
      continue;
    }

    if(working_bbl == header) continue;

    /* Check which predecessors should be investigated next */
    BBL_FOREACH_PRED_EDGE(working_bbl,iter)
    {
      /* we don't follow funlinks, they are not control flow */
      if(EDGE_CAT(iter) == ET_FUNLINK || EDGE_CAT(iter) == ET_IPFUNLINK || EDGE_CAT(iter) == ET_SWI)
      {
	continue;
      }
      /* in case of return (or compensating) edges, we add the head of the corresponding edge in case we have already
       * visited the first bbl of the function we return from */
      if(EDGE_CAT(iter) == ET_RETURN || EDGE_CAT(iter) == ET_COMPENSATING)
      {
	if(EDGE_CORR(iter) && NodeIsMarked(CFG_EDGE_TAIL(EDGE_CORR(iter))) && CFG_EDGE_TAIL(EDGE_CORR(iter)) != header)
	{
	  if(!NodeIsMarked(CFG_EDGE_HEAD(EDGE_CORR(iter))))
	  {
	    NodeMark(CFG_EDGE_HEAD(EDGE_CORR(iter)));
	    Push(&working_stack,CFG_EDGE_HEAD(EDGE_CORR(iter)));
  VERBOSE(printinfo,("Pushing @B",CFG_EDGE_HEAD(EDGE_CORR(iter))));
	  }
	}
	if(!NodeIsMarked(CFG_EDGE_HEAD(iter)))
	{
	  NodeMark(CFG_EDGE_HEAD(iter));
	  Push(&working_stack,CFG_EDGE_HEAD(iter));
  VERBOSE(printinfo,("Pushing @B",CFG_EDGE_HEAD(iter)));
	}
	continue;
      }

      if(CfgEdgeIsForwardInterproc(iter)) 
      {
	if((EdgeIsMarked(iter) && working_bbl != BBL_CFG(working_bbl)->call_hell_node && working_bbl != BBL_CFG(working_bbl)->hell_node)
	    || (EDGE_CORR(iter) && NodeIsMarked(CFG_EDGE_TAIL(EDGE_CORR(iter))) && NodeIsMarked(CFG_EDGE_HEAD(EDGE_CORR(iter)))))
	{
	  if(!NodeIsMarked(CFG_EDGE_HEAD(iter)))
	  {
	    NodeMark(CFG_EDGE_HEAD(iter));
	    Push(&working_stack,CFG_EDGE_HEAD(iter));
  VERBOSE(printinfo,("Pushing @B",CFG_EDGE_HEAD(iter)));
	  }
	}
	continue;
      }
      /* All other edges we haven't treated yet: push the head on the working stack */
      if(!NodeIsMarked(CFG_EDGE_HEAD(iter)))
      {
	NodeMark(CFG_EDGE_HEAD(iter));
	Push(&working_stack,CFG_EDGE_HEAD(iter));
  VERBOSE(printinfo,("Pushing @B",CFG_EDGE_HEAD(iter)));
      }
    }

  }

  CFG_FOREACH_FUN(cfg,fun)
  {
    t_loopfunref * funref;
    t_loopref * loopref;
    t_uint32 nbbls = 0;
    t_bool dont = FALSE;
    if(!(FUN_FLAGS(fun) & FF_IS_MARKED)) continue;
    FUN_FOREACH_BBL(fun,bbl)
    {
      nbbls++;
      if(!NodeIsMarked2(bbl))
      {
	dont = TRUE;
	break;
      }
    }

    if(!dont)
    {
      funref = (t_loopfunref *) Calloc(1,sizeof(t_loopfunref));
      funref->fun = fun;
      LLAppendSorted(funref,&LOOP_FUNCTIONS(ret));

      loopref = (t_loopref *) Calloc(1,sizeof(t_loopref));
      loopref->loop = ret;
      LLAppendSorted(loopref,&FUN_LOOPS(fun));
      LOOP_COUNT(ret) += nbbls;
    }
    else
    {
      FUN_FOREACH_BBL(fun,bbl)
      {
	if(NodeIsMarked2(bbl))
	  if(!BblAddToLoop2(bbl,ret)) FATAL(("Cannot happen!"));
      }
    }
  }
/*  VERBOSE(0,("Owke now check: %d",nnodes));*/
/*  LoopPrint(ret,0);*/

  return ret;
}/*}}}*/

t_bool LoopContainsBbl(t_loop * loop, t_bbl * bbl)
/*{{{ */
{
  t_loopref * loopref;

  if(!BBL_FUN(bbl)) FATAL(("@B has no function!"));
  loopref = FUN_LOOPS(BBL_FUN(bbl));
  while(loopref && loopref->loop < loop) loopref = LLNext(loopref);
  if(loopref && loopref->loop == loop) return TRUE;

  loopref = BBL_LOOPS(bbl);

  while(loopref)
  {
    if(loopref->loop == loop)
      return TRUE;
    loopref = LLNext(loopref);
  }
  return FALSE;
    
}/* }}}*/

static t_bool LoopContainsFun(t_loop * loop, t_function * fun)
{
  t_loopfunref * funref = LOOP_FUNCTIONS(loop);
  while(funref && funref->fun < fun) funref = LLNext(funref);
  if(funref && funref->fun == fun) return TRUE;
  return FALSE;
}

void LoopKill(t_loop **iloop)
/*{{{*/
{
  t_loopelem *iter;
  t_loopref * iterref;
  t_loopfunref * funref;
  t_uint32 i;
  t_loop * loop = *iloop;

  /* Unlink the loop */
  if(loop->prev)
    loop->prev->next = loop->next;
  else if (CFG_LOOP_FIRST(LOOP_CFG(loop)) == loop)
    CFG_LOOP_FIRST(LOOP_CFG(loop)) = loop->next;
  else
    FATAL(("Dll of loops corrupt!"));
  if(loop->next)
    loop->next->prev = loop->prev;
  else if (CFG_LOOP_LAST(LOOP_CFG(loop)) == loop)
    CFG_LOOP_LAST(LOOP_CFG(loop)) = loop->prev;
  else
    FATAL(("Dll of loops corrupt!"));
  /*loop->prev = NULL;*/
  /*loop->next = NULL;*/

  LLListCleanup(&LOOP_BACKEDGES(loop));
  
  /* Free all loopelem's */
  for(i=0;i<HASH_TABLE_SIZE;i++)
  {
    while(LOOP_ELEMS(loop)[i])
    {
      iter = LOOP_ELEMS(loop)[i];
      LOOP_ELEMS(loop)[i] = LLNext(LOOP_ELEMS(loop)[i]);
      Free(iter);
    }
  }
  while(LOOP_FUNCTIONS(loop))
  {
    funref = LOOP_FUNCTIONS(loop);
    LOOP_FUNCTIONS(loop) = LLNext(funref);
    Free(funref);
  }
  while(LOOP_PARENTS(loop))
  {
    iterref = LOOP_PARENTS(loop);
    LoopRemoveChildFromParent(iterref->loop,loop);
    LOOP_PARENTS(loop) = LLNext(LOOP_PARENTS(loop));
    Free(iterref);
  }
  while(LOOP_CHILDS(loop))
  {
    iterref = LOOP_CHILDS(loop);
    LoopRemoveParentFromChild(iterref->loop,loop);
    LOOP_CHILDS(loop) = LLNext(LOOP_CHILDS(loop));
    Free(iterref);
  }
  /*loop->count = 0;*/
    /*loop->cfg = NULL;*/
  Free(loop);
  loop = NULL;
}/*}}}*/

void LoopBblCleanup(t_bbl * bbl)
{
  t_loopref * remove;

  while(BBL_LOOPS(bbl))
  {
    remove = BBL_LOOPS(bbl);
    BBL_LOOPS(bbl) = LLNext(BBL_LOOPS(bbl));
    Free(remove);
  }
}

t_int32 EdgeLoopExits(t_cfg_edge * edge)
{
  t_int32 from_loops = 0;
  t_loopref * loop_ref;

  if(!BBL_LOOPS(CFG_EDGE_HEAD(edge)) && !BBL_LOOPS(CFG_EDGE_TAIL(edge))) return 0;
  loop_ref = BBL_LOOPS(CFG_EDGE_HEAD(edge));
  while(loop_ref)
  {
    if(LoopBblIsLoopExit(LOOPREF_LOOP(loop_ref),CFG_EDGE_HEAD(edge)) && !LoopContainsBbl(LOOPREF_LOOP(loop_ref),CFG_EDGE_TAIL(edge)))
    {
      from_loops++;
    }
    loop_ref = LLNext(loop_ref);
  }
  return from_loops;
}

t_int32 EdgeLoopEntries(t_cfg_edge * edge)
{
  t_int32 to_loops = 0;
  t_loopref * loop_ref;

  if(!BBL_LOOPS(CFG_EDGE_HEAD(edge)) && !BBL_LOOPS(CFG_EDGE_TAIL(edge))) return 0;
  loop_ref = BBL_LOOPS(CFG_EDGE_TAIL(edge));
  while(loop_ref)
  {
    if(LOOP_HEADER(LOOPREF_LOOP(loop_ref)) == CFG_EDGE_TAIL(edge) && !LoopContainsBbl(LOOPREF_LOOP(loop_ref),CFG_EDGE_HEAD(edge)))
      to_loops++;
    loop_ref = LLNext(loop_ref);
  }
  return to_loops;
}

t_bool LoopBblIsLoopExit(t_loop * loop, t_bbl * bbl)
{
  t_loopelem * iter;
  iter = LOOP_ELEMS(loop)[HASHINDEX(bbl)];
  while(iter)
  {
    if(iter->bbl == bbl)
    {
      if(iter->exit) return TRUE;
      else return FALSE;
    }
    if(iter->bbl > bbl) break;
    iter = LLNext(iter);
  }

  return FALSE;
}

void LoopFindNestedLoops(t_cfg * cfg)
{
  t_loop * i_loop, * j_loop;

  VERBOSE(0,("Start nesting detection"));
  CFG_FOREACH_LOOP(cfg,i_loop)
  {
    CFG_FOREACH_LOOP(cfg,j_loop)
    {
      if(i_loop != j_loop && LOOP_HEADER(i_loop) == LOOP_HEADER(j_loop))
      {
	if(LoopIsNested(i_loop,j_loop))
	{
	  /* We know i_loop is nested in j_loop, but it may occur that
	   * i_loop was found to be nested already in other loops. We
	   * have several possibilities then:
	   * a) j_loop is itself nested in a loop that was previously
	   * found to be a parent of i_loop. Set j_loop to the parent.
	   * b) a loop that was previously found to be a parent is nested
	   * in j_loop. Nothing changes
	   * c) j_loop is not nested in a parent of i_loop and no parent is
	   * nested in j_loop. Add a new parent for this loop.
	   * d) i_loop has no parents yet. Set j_loop as a parent.
	   * */
/*          t_bool treated = FALSE;*/
	  t_loopref * iterparents = LOOP_PARENTS(i_loop);


	  /*if(iterparents)*/
	    /* We already found that this loop is nested */
	  /*{

	    while(iterparents)
	    {
	      if(LoopIsNested(j_loop,iterparents->loop))
	      {
		iterparents->loop = j_loop;
		treated = TRUE;
	      }
	      else if(LoopIsNested(iterparents->loop,j_loop))
		treated = TRUE;
	      iterparents = LLNext(iterparents);
	    }
	    if(!treated)
	    {
	      iterparents = (t_loopref *) Calloc(1,sizeof(t_loopref));
	      iterparents->loop = j_loop;
	      LLAppend(NULL,iterparents,&LOOP_PARENTS(i_loop));
	    }
	  }
	  else*/
	  {
	    iterparents = (t_loopref *) Calloc(1,sizeof(t_loopref));
	    iterparents->loop = j_loop;
	    LLAppend(NULL,iterparents,&LOOP_PARENTS(i_loop));
	  }
	}
      }
    }
  }
  /* For each parents entry, fill in the child entry */
  CFG_FOREACH_LOOP(cfg,i_loop)
  {
    t_loopref * parents = LOOP_PARENTS(i_loop);
    
    while (parents)
    {
      t_loopref * new_child = (t_loopref*) Calloc(1,sizeof(t_loopref));
      new_child->loop = i_loop;
      LLAppend(NULL,new_child,&LOOP_CHILDS(parents->loop));
      parents = LLNext(parents);
    }
  }
}

static t_bool LoopIsNested(t_loop * nest, t_loop * parent)
{
  data_ll_elem * backedge;
  t_loopelem * i_loopelem,* j_loopelem;
  t_loopfunref * i_funref, * j_funref;
  t_uint32 i;

  /* If the loopcount is greater, nest cannot be nested in parent */
  if(LOOP_COUNT(nest) >= LOOP_COUNT(parent)) return FALSE;
  
  /* If parent doesn't contain the loopheader or any of the heads of the backedges from nest, nest cannot be nested */
  backedge = LOOP_BACKEDGES(nest);

  if(!LoopContainsBbl(parent,LOOP_HEADER(nest))) 
    return FALSE;

  while(backedge)
  {
    if(!LoopContainsBbl(parent,CFG_EDGE_HEAD(backedge->data)))
      return FALSE;
    backedge = LLNext(backedge);
  }

  /* Now check, node per node, if all the node from nest are in the parent loop */

  for(i=0;i<HASH_TABLE_SIZE;i++)
  {
    i_loopelem = LOOP_ELEMS(nest)[i];
    j_loopelem = LOOP_ELEMS(parent)[i];

    while(i_loopelem && j_loopelem)
    {
      if(LOOPELEM_BBL(i_loopelem) == LOOPELEM_BBL(j_loopelem))
      {
	i_loopelem = LLNext(i_loopelem);
	j_loopelem = LLNext(j_loopelem);
	continue;
      }
      if(LOOPELEM_BBL(i_loopelem) > LOOPELEM_BBL(j_loopelem))
      {
	j_loopelem = LLNext(j_loopelem);
	continue;
      }
      if(LoopContainsFun(parent,BBL_FUN(LOOPELEM_BBL(i_loopelem))))
      {
	i_loopelem = LLNext(j_loopelem);
	continue;
      }
      return FALSE;

    }
    
    if(i_loopelem)
    {
      return FALSE;
    }
  }

  /* Also check if all functions in nest loop are in parent loop */
  i_funref = LOOP_FUNCTIONS(nest);
  j_funref = LOOP_FUNCTIONS(parent);
  while(i_funref && j_funref)
  {
    if(i_funref->fun == j_funref->fun)
    {
      i_funref = LLNext(i_funref);
      j_funref = LLNext(j_funref);
      continue;
    }
    if(i_funref->fun > j_funref->fun)
    {
      j_funref = LLNext(j_funref);
      continue;
    }
    return FALSE;
  }
  if(i_funref) return FALSE;

/*  VERBOSE(0,("Loop with head @B is nested in loop with head @B",LOOP_HEADER(nest),LOOP_HEADER(parent)));*/
/*  LoopPrint(nest,0);*/
/*  LoopPrint(parent,0);*/
  return TRUE;
  
}

/* Move this to iterators.h once bruno has finished his code rewriting */
#define CFG_FOREACH_LOOP_SAFE(cfg,loop,tmp) for (loop=CFG_LOOP_FIRST(cfg), tmp=loop?LOOP_NEXT(loop):NULL; loop!=NULL; loop=tmp, tmp=loop?LOOP_NEXT(loop):NULL)

void LoopMergeLoopsWithSameHeader(t_cfg * cfg)
{
  t_loop * i_loop, * j_loop, *tmp;

  VERBOSE(0,("Start merging loops"));
  CFG_FOREACH_LOOP(cfg,i_loop)
  {
    t_bool skip_this = FALSE;
    t_loopref * iterparents = LOOP_PARENTS(i_loop);

    /* Only merge loops that are not nested in another loop with the same header */
    while(iterparents)
    {
      if(LOOP_HEADER(iterparents->loop) == LOOP_HEADER(i_loop))
      {
	skip_this = TRUE;
        break;
      }
      iterparents = LLNext(iterparents);
    }
    if(skip_this) continue;

    CFG_FOREACH_LOOP_SAFE(cfg,j_loop,tmp)
    {
      if(i_loop != j_loop && LOOP_HEADER(i_loop) == LOOP_HEADER(j_loop))
      {
	t_loopref * jterparents = LOOP_PARENTS(j_loop);
	t_bool stop_here = FALSE;
	VERBOSE(1,("Candidates! @E and @E",LOOP_BACKEDGES(i_loop)->data,LOOP_BACKEDGES(j_loop)->data));

	while(jterparents)
	{
	  if(LOOP_HEADER(jterparents->loop) == LOOP_HEADER(i_loop))
	  {
	    stop_here = TRUE;
	    break;
	  }
	  jterparents = LLNext(jterparents);
	}
	if(stop_here) continue;

	VERBOSE(0,("I have to merge @E and @E",LOOP_BACKEDGES(i_loop)->data,LOOP_BACKEDGES(j_loop)->data));
	LoopMerge(i_loop,j_loop);

	LoopKill(&j_loop);
	j_loop = NULL;
	/* {
	   t_bbl * k_bbl;
	   t_function *fun;
	   CFG_FOREACH_FUN(cfg,fun)
	   FUN_FOREACH_BBL(fun,k_bbl)
	   {
	   VERBOSE(0,("Beezblock: @B",k_bbl));
	   fflush(stdout);
	   BblIsLoopHeader(k_bbl);
	   }
	   }*/

      }
    }
  }
  CFG_FOREACH_LOOP(cfg,i_loop)
  {
    t_loopelem * loop_elem;
    t_uint32 i;
/*    LoopPrint(i_loop,0);*/
    for(i=0;i<HASH_TABLE_SIZE;i++)
    {
      loop_elem = LOOP_ELEMS(i_loop)[i];
      while(loop_elem)
      {
	t_cfg_edge * i_edge;
	BBL_FOREACH_SUCC_EDGE(LOOPELEM_BBL(loop_elem),i_edge)
	{
	  if(EDGE_CAT(i_edge) == ET_FUNLINK || EDGE_CAT(i_edge) == ET_IPFUNLINK) continue;
	  /* TODO: this can be made more precise */
	  if(!LoopContainsBbl(i_loop,CFG_EDGE_TAIL(i_edge)))
	  {
	    LOOP_INF(i_loop) = FALSE;
	    LOOPELEM_EXIT(loop_elem) = TRUE;
	  }
	}
	loop_elem = LLNext(loop_elem);
      }
    }
  }

}

void LoopMerge(t_loop * loop_a, t_loop * loop_b)
{
  t_loopelem * iterref, * jterref, *jterref_prev = NULL, *iterref_prev = NULL;
  t_loopref * child;
  t_loopfunref * ifunref, *ifunref_prev = NULL, *jfunref, *jfunref_prev = NULL;
  data_ll_elem * new_backedge;
  t_uint32 i;
/*  VERBOSE(0,("Merging loops, i'm merging loohoohoohoops %p, %p",loop_a,loop_b));*/
/*  LoopPrint(loop_a,0);*/
/*  LoopPrint(loop_b,0);*/
  fflush(stdout);
  if(loop_a == loop_b) FATAL(("Merging the same loop!"));
  /*if(loop_a->infinite || (loop_b)->infinite) FATAL(("Never thought we should have to merge infinite loops"));
   */
  
  if(LLNext(LOOP_BACKEDGES(loop_b))) VERBOSE(1,("Loop_b already has multiple backedges"));
  new_backedge = LLCalloc(LOOP_BACKEDGES(loop_b)->data);
  LLAppend(NULL,new_backedge,LOOP_BACKEDGES(loop_a));/* !TODO */

  /* First add all bbl's that are in loop_b but not in loop_a */
  for(i=0;i<HASH_TABLE_SIZE;i++)
  {
/*    VERBOSE(0,("HASHMerge %d",i));*/
    iterref_prev = jterref_prev = NULL;
    iterref = LOOP_ELEMS(loop_a)[i];/*LOOP_HEAD(loop_a);*/
    jterref = LOOP_ELEMS(loop_b)[i];/*LOOP_HEAD(loop_b);*/
    while(jterref)
    {
      
      if(iterref && LOOPELEM_BBL(iterref) < LOOPELEM_BBL(jterref))
      {
	/* If loop_b contains function of the bbl in iterref, we remove iterref from this list
	 * and below we add the function to the loop */
	if(LoopContainsFun(loop_b,BBL_FUN(LOOPELEM_BBL(iterref))))
	{
	  t_loopelem * free = iterref;
          //by Xiangyu, 
	  if (iterref_prev) LLNext(iterref_prev) = LLNext(iterref);
	  //LLNext(iterref_prev) = LLNext(iterref);
	  iterref = LLNext(iterref);
	  //Free(free);
	}
	iterref_prev = iterref;
	iterref = LLNext(iterref);
      }
      else if(iterref && LOOPELEM_BBL(iterref) == LOOPELEM_BBL(jterref))
      {
	/* Delete the reference in bbl that refers to loop_b, because it already references the merged loop_a */
	t_loopref * iter = BBL_LOOPS(LOOPELEM_BBL(jterref)), *prev = NULL;
	while(iter->loop != loop_b) {prev = iter; iter = LLNext(iter);}
	if(!iter) FATAL(("Crossref loop<->bbl corrupt!"));

/*        VERBOSE(0,("Unlinking loopref %p from @B",loop_b,LOOPELEM_BBL(jterref)));*/
	LLUnlink(prev,iter,&BBL_LOOPS(LOOPELEM_BBL(jterref)));
		  
	Free(iter);

	iterref_prev = iterref;
	iterref = LLNext(iterref);
	jterref_prev = jterref;
	jterref = LLNext(jterref);
      }
      else if(LoopContainsFun(loop_a,BBL_FUN(LOOPELEM_BBL(jterref))))
      {
	/* Delete the reference in bbl that refers to loop_b, because it already references the merged loop_a */
	t_loopref * iter = BBL_LOOPS(LOOPELEM_BBL(jterref)), *prev = NULL;
	while(iter->loop != loop_b) {prev = iter; iter = LLNext(iter);}
	if(!iter) FATAL(("Crossref loop<->bbl corrupt!"));

/*        VERBOSE(0,("Unlinking loopref %p from @B",loop_b,LOOPELEM_BBL(jterref)));*/
	LLUnlink(prev,iter,&BBL_LOOPS(LOOPELEM_BBL(jterref)));
		  
	Free(iter);

	jterref_prev = jterref;
	jterref = LLNext(jterref);
      }
      /* Add element to loop_a */
      else
      {
	t_loopelem * add_me = jterref;
	t_loopref * iter = BBL_LOOPS(LOOPELEM_BBL(jterref));
	/* Adapt the reference in bbl that refers to loop_b, because it must reference the merged loop_a */
	if(!iter) FATAL(("Crossref loop<->bbl corrupt!"));
	while(iter->loop != loop_b) iter = LLNext(iter);
	if(!iter) FATAL(("Crossref loop<->bbl corrupt!"));
	iter->loop = loop_a;

	jterref = LLNext(jterref);
	LLUnlink(jterref_prev,add_me,&LOOP_ELEMS(loop_b)[i]);
	LLAppend(iterref_prev,add_me,&LOOP_ELEMS(loop_a)[i]);

      }
    }
  }
  jfunref = LOOP_FUNCTIONS(loop_b);
  ifunref = LOOP_FUNCTIONS(loop_a);
  while(jfunref)
  {
    if(ifunref)
    {
      if(ifunref->fun < jfunref->fun)
      {
	ifunref_prev = ifunref;
	ifunref = LLNext(ifunref);
      }
      else if(ifunref->fun == jfunref->fun)
      {
	ifunref_prev = ifunref;
	ifunref = LLNext(ifunref);
	jfunref_prev = jfunref;
	jfunref = LLNext(jfunref);
      }
      else
      {
	t_loopfunref * add_me = jfunref;
	t_loopref * iter = FUN_LOOPS(jfunref->fun), *iter_prev = NULL;
	if(!iter) FATAL(("Crossref loop<->bbl corrupt!"));
	while(iter->loop != loop_b) {iter_prev = iter; iter = LLNext(iter);}
	if(!iter) FATAL(("Crossref loop<->bbl corrupt!"));
	LLUnlink(iter_prev,iter,&FUN_LOOPS(jfunref->fun));
	iter->loop = loop_a;
	LLAppendSorted(iter,&FUN_LOOPS(jfunref->fun));
	
	jfunref = LLNext(jfunref);
	LLUnlink(jfunref_prev,add_me,&LOOP_FUNCTIONS(loop_b));
	LLAppend(ifunref_prev,add_me,&LOOP_FUNCTIONS(loop_a));
      }
    }
    else
    {
      t_loopfunref * add_me = jfunref;
      t_loopref * iter = FUN_LOOPS(jfunref->fun), *iter_prev = NULL;
      if(!iter) FATAL(("Crossref loop<->bbl corrupt!"));
      while(iter->loop != loop_b) {iter_prev = iter; iter = LLNext(iter);}
      if(!iter) FATAL(("Crossref loop<->bbl corrupt!"));
      LLUnlink(iter_prev,iter,&FUN_LOOPS(jfunref->fun));
      iter->loop = loop_a;
      LLAppendSorted(iter,&FUN_LOOPS(jfunref->fun));

      jfunref = LLNext(jfunref);
      LLUnlink(jfunref_prev,add_me,&LOOP_FUNCTIONS(loop_b));
      LLAppend(ifunref_prev,add_me,&LOOP_FUNCTIONS(loop_a));

    }
  }

  /* Set the parent of each loop nested in loop_b to loop_a */
  child = LOOP_CHILDS(loop_b);
  while(child)
  {
    t_loopref * parents = LOOP_PARENTS(child->loop);
    while(parents)
    {
      if(parents->loop == loop_b) break;
      parents = LLNext(parents);
    }
    if(!parents) FATAL(("Nesting info corrupt"));
    parents->loop = loop_a;
    child = LLNext(child);
  }
  /* Merge childs-list of two loops */
  child = LOOP_CHILDS(loop_a);
  while(child && LLNext(child)) child = LLNext(child);
  if(child) LLNext(child) = LOOP_CHILDS(loop_b);
  else LOOP_CHILDS(loop_a) = LOOP_CHILDS(loop_b);
  LOOP_CHILDS(loop_b) = NULL;
}

static void LoopRemoveChildFromParent(t_loop * parent, t_loop * child)
{
  t_loopref * iter = LOOP_CHILDS(parent);
  t_loopref * prev = NULL;
  while(iter && iter->loop != child) {prev = iter; iter = LLNext(iter);}
  if(iter)
  {
    if(prev) LLNext(prev) = LLNext(iter);
    else LOOP_CHILDS(parent) = LLNext(iter);
    Free(iter);
  }
}

static void LoopRemoveParentFromChild(t_loop * child, t_loop * parent)
{
  t_loopref * iter = LOOP_PARENTS(child);
  t_loopref * prev = NULL;
  while(iter && iter->loop != parent) {prev = iter; iter = LLNext(iter);}
  if(iter)
  {
    if(prev) LLNext(prev) = LLNext(iter);
    else LOOP_PARENTS(child) = LLNext(iter);
    Free(iter);
  }
}

t_bool BblIsLoopHeader(t_bbl * bbl)
{
  t_loopref * refs = BBL_LOOPS(bbl);

  while(refs)
  {
    if(LOOP_HEADER(LOOPREF_LOOP(refs)) == bbl) return TRUE;
    refs = LLNext(refs);
  }
  return FALSE;
}

static void Push(data_stack *stack, t_bbl* bbl)
{
/*  VERBOSE(0,("Pushing @B",bbl));*/
  stack->stack[stack->top] = bbl;
  stack->top++;
  if(stack->top == MAX_STACK_HEIGHT) FATAL(("Increase stack height"));
}

static t_bbl * Pop(data_stack *stack)
{
  if(!stack->top) FATAL(("Stack empty!"));
  stack->top--;
  return stack->stack[stack->top];
}

static t_bool BblAddToLoop(t_bbl * bbl, t_loop * loop)
{
  if(NodeIsMarked2(bbl)) return FALSE;
  NodeMark2(bbl);
  FUN_FLAGS(BBL_FUN(bbl)) |= FF_IS_MARKED;
  return TRUE;
}

static t_bool BblAddToLoop2(t_bbl * bbl, t_loop * loop)
{
  t_loopelem * insert_elem = NULL;
  t_loopref* loop_ref_elem = NULL;
  t_uint32 hi = HASHINDEX(bbl);

  /* Add the bbl to the loop */
  data_ll_elem * iter1 = (data_ll_elem *) LOOP_ELEMS(loop)[hi];
  if(iter1)
  {
    if(iter1->data > (void*)bbl)
    {
      iter1 = NULL;
    }
    else
    {
      while(LLNext(iter1) && LLData(LLNext(iter1)) < (void*)bbl) iter1 = LLNext(iter1);
      if(LLData(iter1) == bbl) return FALSE;
    }
  }
  insert_elem = (t_loopelem*) Calloc(1,sizeof(t_loopelem));
  LOOPELEM_BBL(insert_elem) = bbl;
  LLAppend(iter1,insert_elem,&LOOP_ELEMS(loop)[hi]);

  /* Add the loop to the referred loops from the bbl */
  loop_ref_elem = (t_loopref *) Calloc(1,sizeof(t_loopref));
  LOOPREF_LOOP(loop_ref_elem) = loop;
  LLAppend(NULL,loop_ref_elem,&BBL_LOOPS(bbl));

  LOOP_COUNT(loop)++;
  
  return TRUE;
}

/* This function initialises the state of a loopiterator, so you can start calling LoopGetNextBbl */
t_loopiterator * LoopNewIterator(t_loop * loop)
{
  t_loopiterator * ret = (t_loopiterator*) Calloc(1,sizeof(t_loopiterator)); 
  ret->loop = loop;
  return ret;
}

/* An interator function, that return all bbl of a loop. Call LoopInitIterator before you start using it! */
t_bbl * LoopGetNextBbl(t_loopiterator * iterator)
{
  fflush(stdout);
  if(iterator->iterator_hash == 0 && !iterator->iterator_elem && !iterator->iterator_fun)
  {
    /* We are in init state, set iterator_elem and return its bbl */
    while(iterator->iterator_hash < HASH_TABLE_SIZE)
    {
      iterator->iterator_elem = iterator->loop->elems[iterator->iterator_hash];
      if(iterator->iterator_elem) break;
      iterator->iterator_hash++;
    }

    if(iterator->iterator_elem)
      return iterator->iterator_elem->bbl;
  }
  /* Not in init state, but not yet iterating the functions */
  if(iterator->iterator_hash < HASH_TABLE_SIZE)
  {
    if(!iterator->iterator_elem) FATAL(("Loop iterator is in a false state!"));
    if(LLNext(iterator->iterator_elem))
    {
      iterator->iterator_elem = LLNext(iterator->iterator_elem);
      return iterator->iterator_elem->bbl;
    }
    else if (iterator->iterator_hash < (HASH_TABLE_SIZE-1))
    {
      (iterator->iterator_hash)++;
      while(iterator->iterator_hash < HASH_TABLE_SIZE)
      {
	iterator->iterator_elem = iterator->loop->elems[iterator->iterator_hash];
	if(iterator->iterator_elem) break;
	iterator->iterator_hash++;
      }

      if(iterator->iterator_elem)
	return iterator->iterator_elem->bbl;
    }
  }
  if(!iterator->iterator_bbl)
  {
    iterator->iterator_fun = iterator->loop->functions;
    if(iterator->iterator_fun)
    {
      iterator->iterator_bbl = FUN_BBL_FIRST(iterator->iterator_fun->fun);
      return iterator->iterator_bbl;
    }
    else return NULL;
  }
  if(BBL_NEXT_IN_FUN(iterator->iterator_bbl))
  {
    iterator->iterator_bbl = BBL_NEXT_IN_FUN(iterator->iterator_bbl);
    return iterator->iterator_bbl;
  }
  if(LLNext(iterator->iterator_fun))
  {
    iterator->iterator_fun = LLNext(iterator->iterator_fun);
    iterator->iterator_bbl = FUN_BBL_FIRST(iterator->iterator_fun->fun);
    return iterator->iterator_bbl;
  }
  return NULL;
}

/* This function checks whether a function will belong to a loop in
   its entirety once a return edge belongs to the loop, and under the
   (to be later verified) condition that its callees permit backwards
   reachability. The latter condition is of course recursive */

void MarkFunIfConditionalBackwardReachababilityIsTrue(t_function * fun)
{
  t_bool change = TRUE;
  t_bbl * bbl;
  t_cfg_edge * edge;

  if (fun==fun->cfg->swi_hell_function || fun==fun->cfg->call_hell_function) 
    {
      FUN_FLAGS(fun) |= FF_COND_LOOP_REACH;
      return;
    }

  FUN_FLAGS(fun) &= ~FF_COND_LOOP_REACH;

  if (!FunctionGetExitBlock(fun)) return;
  
  InitNodeMarking();
  
  NodeMark(FunctionGetExitBlock(fun));

  while (change)
    {
      change = FALSE;
      
      FUN_FOREACH_BBL(fun,bbl)
	{
	  if (!NodeIsMarked(bbl)) continue;

	  BBL_FOREACH_PRED_EDGE(bbl,edge)
	    {
	      /* we don't follow funlinks, they are not control flow */
	      if(EDGE_CAT(edge) == ET_FUNLINK || EDGE_CAT(edge) == ET_IPFUNLINK || EDGE_CAT(edge) == ET_SWI)
		continue;
	      
	      /* in case of return (or compensating) edges, we add the head of the corresponding edge in case we have already
	       * visited the first bbl of the function we return from */
	      if (CfgEdgeIsBackwardInterproc(edge))
		{
		  if(!NodeIsMarked(CFG_EDGE_HEAD(EDGE_CORR(edge))))
		    {
		      NodeMark(CFG_EDGE_HEAD(EDGE_CORR(edge)));
		      change = TRUE;
		    }
		}
	      
	      else if (CfgEdgeIsForwardInterproc(edge)) continue;
	      else if(!NodeIsMarked(CFG_EDGE_HEAD(edge)))
		{
		  NodeMark(CFG_EDGE_HEAD(edge));
		  change = TRUE;
		}
	    }
	}
    }
  
  FUN_FOREACH_BBL(fun,bbl)
    if (!NodeIsMarked(bbl))
      {
	FUN_FLAGS(fun) &= ~FF_COND_LOOP_REACH;
	return ;
      }

  FUN_FLAGS(fun) |= FF_COND_LOOP_REACH;

}

/* this function determines whether the entry point of a function is
   backwards reachable from the exit block, whereby callees only allow
   backward propagation if they are FF_SPECIAL_LOOP_FUN|FF_ENTRY_REACHABLE_FROM_EXIT */

t_bool EntryIsReachableFromExit(t_function * fun)
{
  t_bool change = TRUE;
  t_bbl * bbl;
  t_cfg_edge * edge;

  if (!FunctionGetExitBlock(fun)) return FALSE;
  
  InitNodeMarking();
  
  NodeMark(FunctionGetExitBlock(fun));

  while (change)
    {
      change = FALSE;
      
      FUN_FOREACH_BBL(fun,bbl)
	{
	  if (!NodeIsMarked(bbl)) continue;

	  BBL_FOREACH_PRED_EDGE(bbl,edge)
	    {
	      /* we don't follow funlinks, they are not control flow */
	      if(EDGE_CAT(edge) == ET_FUNLINK || EDGE_CAT(edge) == ET_IPFUNLINK || EDGE_CAT(edge) == ET_SWI)
		continue;
	      
	      /* in case of return (or compensating) edges, we add the head of the corresponding edge in case we have already
	       * visited the first bbl of the function we return from */
	      if (CfgEdgeIsBackwardInterproc(edge) && (FUN_FLAGS(BBL_FUN(CFG_EDGE_HEAD(edge))) & (FF_SPECIAL_LOOP_FUN|FF_ENTRY_REACHABLE_FROM_EXIT)))
		{
		  if(!NodeIsMarked(CFG_EDGE_HEAD(EDGE_CORR(edge))))
		    {
		      NodeMark(CFG_EDGE_HEAD(EDGE_CORR(edge)));
		      change = TRUE;
		    }
		}
	      
	      else if (CfgEdgeIsForwardInterproc(edge)) continue;
	      else if(!NodeIsMarked(CFG_EDGE_HEAD(edge)))
		{
		  NodeMark(CFG_EDGE_HEAD(edge));
		  change = TRUE;
		}
	    }
	}
    }

  if (NodeIsMarked(FUN_BBL_FIRST(fun)))
    return TRUE;
  
  return FALSE;
  
}


/* This function computes whether or not a function will belong in a
   loop after a return edge from it belongs to it. If this is the
   case, the FF_SPECIAL_LOOP_FUN is set */

void CfgComputeSpecialFuns(t_cfg * cfg)
{

  t_bool change = TRUE;
  t_cg_edge * edge;
  t_function * callee;
  t_function * fun;

  /* first we conditionally determine this property, based on
     recursive conditions */

  CFG_FOREACH_FUN(cfg,fun)
    {
      FUN_FLAGS(fun) &= ~FF_ENTRY_REACHABLE_FROM_EXIT;
      FUN_FLAGS(fun) &= ~FF_SPECIAL_LOOP_FUN;
      FUN_FLAGS(fun) &= ~FF_COND_LOOP_REACH;
      MarkFunIfConditionalBackwardReachababilityIsTrue(fun);

      if (fun==fun->cfg->swi_hell_function || fun==fun->cfg->call_hell_function) 
	{
	  FUN_FLAGS(fun) |= FF_SPECIAL_LOOP_FUN;
	}
      
    }
  
  CgBuild(cfg);

  while (change)
    {
      change = FALSE;

      /* if a function is conditionally fine, check the callees: if
	 they are unconditionally fine, the function becomes fine as
	 well */

      CFG_FOREACH_FUN(cfg,fun)
	{
	  if (!(FUN_FLAGS(fun) & FF_SPECIAL_LOOP_FUN) && (FUN_FLAGS(fun) & FF_COND_LOOP_REACH))
	    {

	      t_bool ok = TRUE;

	      FUN_FOREACH_CALLEE(fun,edge,callee)
		{
		  if (!(FUN_FLAGS(callee) & FF_SPECIAL_LOOP_FUN) && !(FUN_FLAGS(callee) & FF_ENTRY_REACHABLE_FROM_EXIT))
		    ok = FALSE;
		}
	      
	      if (ok)
		{
		  FUN_FLAGS(fun) |= FF_SPECIAL_LOOP_FUN;
		  change = TRUE;
		  
		}
	    }	    
	}
      
      /* if a function is not fine itself, but it will allow its
	 callers to be fine, mark them as such, using the information
	 known from its callees */

      CFG_FOREACH_FUN(cfg,fun)
	if (!(FUN_FLAGS(fun) & FF_SPECIAL_LOOP_FUN) && !(FUN_FLAGS(fun) & FF_ENTRY_REACHABLE_FROM_EXIT) && (FUN_FLAGS(fun) & FF_COND_LOOP_REACH))
	  if (EntryIsReachableFromExit(fun))
	    {
	      FUN_FLAGS(fun) |= FF_ENTRY_REACHABLE_FROM_EXIT;
	      change = TRUE;
	    }
      
    }

  CFG_FOREACH_FUN(cfg,fun)
    if (FUN_FLAGS(fun) & FF_SPECIAL_LOOP_FUN)
      DiabloPrint(stdout,"YOEHA FOR FUN @B\n",FUN_BBL_FIRST(fun));
    else
      if (FUN_FLAGS(fun) & FF_COND_LOOP_REACH)
	{
	  int teller = 0;
  	  FUN_FOREACH_CALLEE(fun,edge,callee)
	    teller++;
	  DiabloPrint(stdout,"EIGENAARDIG FOR FUN @B with %d callees\n",FUN_BBL_FIRST(fun),teller);
	}
}


/* this function should add functions to a loop after discovering that
   a return edge from this function needed to be included in the
   loop NEEDS WORK */

#if 0
void RecursiveAddUnmarkedFunsToLoop(t_function * fun, t_loop * loop)
{
  t_cg_edge * edge;
  t_function * callee;

  if (FUN_FLAGS(fun) & FF_SPECIAL_LOOP_FUN)
    {
      /* do the whole function and recursively its callers */
      
      AddFunToLoop(fun,loop);
      FUN_FOREACH_CALLEE(fun,edge,callee)
	RecursiveAddUnmarkedFunsToLoop(callee,fun); 
    }
  else
    {
      /* add the exit block as if a normal return edge was followed */
    }
  
}
#endif


void CfgMarkFunsWithoutIncomingJumps(t_cfg * cfg)
{
  t_function * fun;
  t_cfg_edge * edge;
  CFG_FOREACH_FUN(cfg,fun)
    FUN_FLAGS(fun) &= ~ FF_HAS_INCOMING_IPJUMP;
  CFG_FOREACH_EDGE(cfg,edge)
    if (EDGE_CAT(edge)==ET_IPJUMP || EDGE_CAT(edge)==ET_IPFALLTHRU)
	FUN_FLAGS(BBL_FUN(CFG_EDGE_TAIL(edge))) |= FF_HAS_INCOMING_IPJUMP;
}

void TestBblOnAddress(t_cfg * cfg)
{
  t_bbl * bbl;
  t_loop * loop;


  DiabloPrint(stdout,"Passsed test\n");
  CFG_FOREACH_BBL(cfg,bbl)
    if(BBL_FUN(bbl) == (t_function*) 0x1c0) FATAL(("Gotcha mofo"));

  CFG_FOREACH_LOOP(cfg,loop)
  {
    t_loopiterator * loopiter;
    LOOP_FOREACH_BBL(loop,loopiter,bbl)
    {
      t_cfg_edge * edge;
      BBL_FOREACH_SUCC_EDGE(bbl,edge)
      if(BBL_FUN(CFG_EDGE_TAIL(edge)) == (t_function*)0x1c0) FATAL(("Gotcha mofo"));
    }
  }
  
}
/* vim: set shiftwidth=2 foldmethod=marker: */
