/***********************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This files contains all definitions for  handling "generic" objects 
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 ***********************************************************/

/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup OBJECTS Objects 
 *
 * Objects form the first layer of abstracton used in Diablo. Everything in
 * Diablo start with creating one or more objects. All supported objectfile
 * formats can be read in (the prefered constructor to do this is \ref
 * ObjectGet), transformed to a generic object (done automaticaly) and than
 * written back to an objectfile (use \ref ObjectWrite for this).  This makes
 * it possible to apply the same analyses and optimizations to objects stored
 * in different objectfile formats.  Currently the following objectfile formats
 * are in different stages of developement:
 *
 * - \ref ELF
 * - \ref ECOFF
 * - \ref PE 
 *
 * Objects consist of the following information:
 *
 * - The program data and code, stored in sections \ref SECTIONS
 * - One or two symbol tables, that labels the information in these sections \ref SYMBOLS
 * - A reloc table, that describes the relationship between different structures in the object \ref RELOCS
 */
/*@{*/

#ifndef MY_OBJECT_TYPEDEFS
#define MY_OBJECT_TYPEDEFS
typedef struct _t_object t_object;
typedef struct _object_hash_node object_hash_node;

#endif /* MY_OBJECT_TYPEDEF */


#ifndef MY_OBJECT_H
#define MY_OBJECT_H
#include <stdio.h>
#include "../simple_types.h"
#include "diablo_addresses.h"
#include "diablo_object_tables.h"
#include "../datastructures/data_hash.h"
#include "../datastructures/data_graph.h"
#include "diablo_section.h"
#include "diablo_segment.h"
#include "diablo_symbol.h"
#include "diablo_reloc.h"
#include "diablo_parsemap.h"
#include "diablo_action.h"
#include "diablo_description.h"
#include "diablo_linkers.h"
#include "diablo_layout_script.h"

/* OBJECTS */

/* this is a helper structure for describing 
 * the original code sections (can be used after
 * merging to determine which basic blocks
 * originally belonged to which code section)*/
typedef struct _t_sec_descriptor {
  t_string name;
  t_address base;
  t_address size;
  t_address obase;
  t_address osize;
  char type;
} t_sec_descriptor;

/** The generic object type */

struct _t_object
{
   /*@{*/
   /** Used for object-ordering in a map */
   struct _t_object * equal;
   struct _t_object * next;
   struct _t_object * prev;
   /*@}*/
   /*@{*/
   /** The (ordered) linked list of the objectfiles */
   struct _t_object * mapped_first;
   struct _t_object * mapped_last;
   /*@}*/

   /** The subobject cache (used during linking, only relevant for parent objects) */
   data_hash_table * subobject_cache;

   /** IS the object dynamic? */
   t_bool dynamic;
   /** Are we optimizing an object on an architecture with switched endianness? */
   t_bool switched_endian;
   /** the type of object */
   t_objecttype type;
   /** The subtype of the object */
   t_objectsubtype subtype;

   t_linker linker;
   /** Position in the file where the object starts (0 for normal objects,
    * but it can be different for archives) */
   Filepos streampos; 
   /** the (raw) size of the object in bytes */
   Filepos size; 
   /** The name of the object (an identifier, not a filename) */
   t_string name;
   /** The entrypoint of the objectfile (as set in most objectfileformats, not
    * to confuse with entries in a flowgraph) */
   t_address entry;
   /*!******************* The sections in the binary *************************/
   /*@{*/
   /** 1. ********************** Code sections ********************************/
   t_uint32 ncodes;
   t_section ** code;
   /*@}*/
   /*@{*/
   /** 2. ********************** Data sections ********************************/
   t_uint32 ndatas;
   t_section ** data;
   /*@}*/
   /*@{*/
   /** 3. ********************* Rodata sections *******************************/
   t_uint32 nrodatas;
   t_section ** rodata;
   /*@}*/
   /*@{*/
   /** 4. ********************** Note sections ********************************/
   t_uint32 nnotes;
   t_section ** note;
   /*@}*/
   /*@{*/
   /** 5. ********************** Bss sections *********************************/
   t_uint32 nbsss;
   t_section ** bss;
   /*@}*/
   /************************* End of the sections ****************************/

   /** linked list of all segments (only in the parent object) */
   t_segment * segment_first;
   t_segment * segment_last;


   /*!************* Architecture Descriptions and Callbacks ******************/
   /*@{*/
   const t_architecture_description * description;
   const t_action_struct * actions;
   const t_fileformat_action_struct * ff_actions;
   /*@}*/
   /**************** End of architecture specific information ****************/
   
   /** avltree-like datastructure that contains the parsed linkermap */
   t_map * map;

   /** The relocation for this object (should be build from the subobjects for
    * linked binaries */
   
   t_reloc_table * reloc_table; 
   /** The normal symbol table for this object */
   t_symboltable * symbol_table;
   /** The symbol table build from all subobjects */
   t_symboltable * sub_symbol_table;
   
   /** Some datastructures to speed up searching */
   data_hash_table * section_table;

   /** The global pointer value that was used to create this object */
   /*XXX: shouldn't this be #ifdef IA64_SUPPORT ???? */
   t_int32 gp;

   /*XXX: this should probably be #ifdef KERNEL_COMPACTION */
   /** a table describing the original (pre-merge) code sections.
       this is used in the kernel compactor to separate the 
       .setup.text section from the regular text section
       during the layout phase */
   t_uint32 n_orig_codes;
   t_sec_descriptor * orig_codes;

   /** the layout script to be used for this object file */
   t_layout_rule * layout_script;
};
struct _object_hash_node
{
  data_hash_node node;
  t_object * obj;
  t_bool fake_hash_entry;
};

#define T_OBJECT(x) ((t_object *) (x))
/*! \defgroup OBJ_ALLOC Constructors and destructors */
/*@{{{*/
t_object * ObjectNew();
t_object * ObjectRead(t_string name);
t_object * ObjectReadWithFakeName(t_string name,t_string fake_name);
t_object * ObjectGetFromStream(t_string name, FILE *fp);
void ObjectFree(t_object *);
t_object * ObjectGet(t_string name, t_object * parent);
t_object * ObjectGetFromCache(t_string name, t_object * parent);
t_object * ObjectNewCached(t_string name, t_object * parent);
/*@}}}*/

void ObjectWrite(t_string name, t_object * obj);
void ObjectKnownFileFormats(FILE * fp);
t_section * ObjectGetSection(t_object * obj, t_string name);
void ObjectBuildUnifiedSymbolTable(t_object * obj);
void ObjectInit(void);
void ObjectReadMap(t_object * obj);
void ObjectReadObjectInMap(t_object * obj);
void ObjectVerifyMap(t_object * obj);
void ObjectSortObjectsInTheMap(t_object * obj);
void ObjectMapPatchAlpha(t_object * obj);
void ObjectRemoveLinkerRemovedSections(t_object * obj);
void ObjectFini(void);
t_section * ObjectAddSectionFromFile(t_object * obj, char type, t_bool ro, FILE * fp, Filepos off, t_address address, t_address size, t_address align, t_string name);
t_uint32 ObjectGetData32(t_object * obj, t_string name, t_address address);
t_uint32 ObjectGetData32NoSection(t_object * obj, t_address start);
t_uint64 ObjectGetData64(t_object * obj, t_string name, t_address address);
void ObjectDisassemble(t_object * obj);
void ObjectFlowgraph(t_object * obj);
void ObjectAssemble(t_object * obj);
void ObjectDeflowgraph(t_object * obj);
void ObjectConvertToDataBlocks(t_object * obj);
void ObjectMergeCodeSections(t_object * obj);
void ObjectMergeRODataSections(t_object *obj);
void ObjectMergeDataSections(t_object *obj);
void ObjectMoveBssToDataSections(t_object *obj);
void ObjectRelayoutData(t_object * obj, t_string filename);
void ObjectRelocateSubObjects(t_object * obj, t_bool verify);
void ObjectAppendSubObject(t_object * parent, t_object * child);

void ObjectRebuildSectionDataLayouts(t_object * obj);
void ObjectRebuildSectionsFromSubsections(t_object * obj);

void ObjectCreateDataOrCodeTable(t_object * obj);
void OrderCodeSectionsContiguously(t_object * obj);
void ObjectBuildRelocTableFromSubObjects(t_object * obj);

t_layout_rule * ObjectGetAndParseLayoutScript(t_object * obj);

t_segment * ObjectGetSegmentByName(t_object * obj, t_string name);
void ObjectAddSegment(t_object * obj, t_segment * seg);

#endif
/*@}*/
/*@}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
