/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * All functions for partial redundancy elimination  
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.ugent.be) }}} 
 *****************************************************************************/
/* Includes {{{ */
#include <string.h>
#include "../config.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include "../datastructures/data_string.h"
#include "../datastructures/data_stack.h"
#include "../datastructures/data_ll.h"
#include "../diablo_options.h"
#include "diablo_flowgraph.h"
#include "diablo_function.h"
#include "diablo_bbl.h"
#include "diablo_edge.h"
#include "diablo_ins.h"
#include "diablo_print.h"
#include "diablo_dominator.h"
#include "diablo_reloc.h"
#include "diablo_iterators.h"
#include "diablo_programexit.h"
#include "diablo_profile.h"
#include "diablo_code_motion.h"
#include "diablo_opt_misc.h"
#include <string.h>
/*}}}*/

#ifdef ARM_SUPPORT
#include "../arch/arm/arm_instructions.h"
#include "../arch/arm/arm_utils.h"
#include "../arch/arm/arm_constant_propagation.h"
#endif

t_bool BblEndWithConditionalBranch(t_bbl * bbl)
{
  t_cfg_edge * edge;

  t_bool fallthrough = FALSE;
  t_bool jump = FALSE;

  BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      if ((EDGE_CAT(edge)==ET_JUMP || EDGE_CAT(edge)==ET_IPJUMP) && !BBL_IS_LAST(CFG_EDGE_TAIL(edge)))
	jump = TRUE;
      if (EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU)
	fallthrough = TRUE;
    }
  return jump && fallthrough;
}

t_bool BblEndWithConditionalCall(t_bbl * bbl)
{
  t_cfg_edge * edge;

  t_bool fallthrough = FALSE;
  t_bool call = FALSE;

  BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      if (EDGE_CAT(edge)==ET_CALL)
	call = TRUE;
      if (EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU)
	fallthrough = TRUE;
    }

  return call && fallthrough;
}

t_bool BblEndWithConditionalReturn(t_bbl * bbl)
{
  t_cfg_edge * edge;

  t_bool fallthrough = FALSE;
  t_bool jump = FALSE;

  BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      if (EDGE_CAT(edge)==ET_JUMP && BBL_IS_LAST(CFG_EDGE_TAIL(edge)))
	jump = TRUE;
      if (EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU)
	fallthrough = TRUE;
    }
  return jump && fallthrough;
}

t_cfg_edge * TakenPath(t_bbl * bbl)
{
  t_cfg_edge * edge;

  BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      if (EDGE_CAT(edge)==ET_JUMP || EDGE_CAT(edge)==ET_IPJUMP || EDGE_CAT(edge)==ET_CALL)
	return edge;
    }
  return NULL;
}

t_cfg_edge * LinkPath(t_bbl * bbl)
{
  t_cfg_edge * edge;

  BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      if (EDGE_CAT(edge)==ET_FUNLINK)
	return edge;
    }
  return NULL;
}

t_cfg_edge * FallThroughPath(t_bbl * bbl)
{
  t_cfg_edge * edge;
  
  BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      if (EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU)
	return edge;
    }
  return NULL;
}

/*

PartialRedundancyElimination1: if the taken path of a conditional
branch is rarely executed, instructions depending on the same
condition are moved onto that path

*/
void PartialRedundancyElimination1(t_cfg * cfg)
{

#ifdef ARM_SUPPORT
  t_bbl * bbl;

  CfgComputeHotBblThreshold(cfg,0.99);

  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (BblIsHot(bbl))
	{
	  if (BblEndWithConditionalBranch(bbl))
	    {
	      t_cfg_edge * taken = TakenPath(bbl);
	      t_cfg_edge * fall = FallThroughPath(bbl);

	      t_ins * last_ins = BBL_INS_LAST(bbl);
	      t_regset used_regs = INS_REGS_USE(last_ins);
	      t_regset cond_regs = INS_REGS_USE(last_ins);
	      t_regset changed_regs = RegsetNew();
	      t_ins * ins = INS_PREV(last_ins);
	      EXEC_CONDITION cond_last_ins = UNKNOWN;
	      EXEC_CONDITION cond = UNKNOWN;
	      
	      if (!(EDGE_EXEC_COUNT(taken)<EDGE_EXEC_COUNT(fall)/3))
		continue;

	      cond_last_ins = ArmInsCondition(last_ins);

	      while (ins)
		{
		  if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),cond_regs)))
		    {
		      ins = NULL;
		      break;
		    }

		  cond = ArmInsCondition(ins);

		  if (cond!=UNKNOWN && 
		      cond == cond_last_ins && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),used_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),changed_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(ins),changed_regs))
		      )
		    break;
		  
		  RegsetSetUnion(used_regs,INS_REGS_USE(ins));
		  RegsetSetUnion(changed_regs,INS_REGS_DEF(ins));
		  
		  ins = INS_PREV(ins);
		}

	      if (ins)
		{
		  t_bbl * new_bbl = BblNew(cfg);
		  t_ins * new_ins;
		  t_cfg_edge * new_edge;
		  t_cfg_edge * new_edge2;

		  /*		  DiabloPrint(stdout,"PRE1 candidate @ieB\n",bbl);*/

		  /*		  printf("%d/%d\n",EDGE_EXEC_COUNT(taken),INS_EXEC_COUNT(ins));*/

		  BblInsertInFunction(new_bbl,BBL_FUN(bbl));
		  BBL_EXEC_COUNT(new_bbl)=EDGE_EXEC_COUNT(taken);
		  
		  new_edge = CfgEdgeNew(cfg,new_bbl,CFG_EDGE_TAIL(taken),EDGE_CAT(taken));
		  EDGE_EXEC_COUNT(new_edge)=EDGE_EXEC_COUNT(taken);
		  if (EDGE_CORR(taken))
		    {
		      CfgEdgeKill(EDGE_CORR(taken));
		    }
		  new_edge2 = CfgEdgeNew(cfg,bbl,new_bbl,EDGE_CAT(taken));
		  CfgEdgeKill(taken);
		  EDGE_EXEC_COUNT(new_edge2)=EDGE_EXEC_COUNT(new_edge);

		  new_ins = InsNewForBbl(new_bbl);
		  ArmInstructionMakeUncondBranch(new_ins);		  
		  InsAppendToBbl(new_ins,new_bbl);

		  new_ins = InsDup(ins);
		  InsPrependToBbl(new_ins,new_bbl);
		  InsKill(ins);

		  /*		  DiabloPrint(stdout,"PRE1 after @ieB\n",bbl);*/
		}
	    }
	}
    }
#endif
}

/*

PartialRedundancyElimination1ForReturns: if a conditional return is rarely
executed, and instructions depending on the same condition exist, they
are moved into a separate block, and replaced by a conditional branch

*/

void PartialRedundancyElimination1ForReturns(t_cfg * cfg)
{

#ifdef ARM_SUPPORT
  t_bbl * bbl;

  CfgComputeHotBblThreshold(cfg,0.99);

  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (BblIsHot(bbl))
	{
	  if (BblEndWithConditionalReturn(bbl))
	    {
	      t_cfg_edge * taken = TakenPath(bbl);
	      t_cfg_edge * fall = FallThroughPath(bbl);
	      t_ins * last_ins = BBL_INS_LAST(bbl);
	      t_regset used_regs = INS_REGS_USE(last_ins);
	      t_regset cond_regs = INS_REGS_USE(last_ins);
	      t_regset changed_regs = RegsetNew();
	      t_ins * ins = INS_PREV(last_ins);
	      EXEC_CONDITION cond_last_ins = UNKNOWN;
	      EXEC_CONDITION cond = UNKNOWN;
	      
	      if (!(EDGE_EXEC_COUNT(taken)<EDGE_EXEC_COUNT(fall)/3))
		continue;

	      cond_last_ins = ArmInsCondition(last_ins);

	      while (ins)
		{
		  if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),cond_regs)))
		    {
		      ins = NULL;
		      break;
		    }

		  cond = ArmInsCondition(ins);

		  if (cond!=UNKNOWN && 
		      cond == cond_last_ins && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),used_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),changed_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(ins),changed_regs))
		      )
		    break;
		  
		  RegsetSetUnion(used_regs,INS_REGS_USE(ins));
		  RegsetSetUnion(changed_regs,INS_REGS_DEF(ins));
		  
		  ins = INS_PREV(ins);
		}

	      if (ins)
		{
		  t_bbl * new_bbl = BblNew(cfg);
		  t_ins * new_ins;
		  t_cfg_edge * new_edge;
		  t_cfg_edge * new_edge2;
		  t_uint32 cond = ARMINSCONDITION(ins);

		  /*		  DiabloPrint(stdout,"PRE1return  candidate @ieB\n",bbl); fflush(stdout);*/

		  BblInsertInFunction(new_bbl,BBL_FUN(bbl));
		  BBL_EXEC_COUNT(new_bbl)=EDGE_EXEC_COUNT(taken);
		  
		  new_edge = CfgEdgeNew(cfg,new_bbl,CFG_EDGE_TAIL(taken),ET_JUMP);
		  EDGE_EXEC_COUNT(new_edge)=EDGE_EXEC_COUNT(taken);
		  if (EDGE_CORR(taken))
		    {
		      CfgEdgeKill(EDGE_CORR(taken));
		    }
		  new_edge2 = CfgEdgeNew(cfg,bbl,new_bbl,EDGE_CAT(taken));
		  CfgEdgeKill(taken);
		  EDGE_EXEC_COUNT(new_edge2)=EDGE_EXEC_COUNT(new_edge);

		  new_ins = InsDup(BBL_INS_LAST(bbl));
		  InsPrependToBbl(new_ins,new_bbl);
		  ArmInstructionUnconditionalizer(new_ins);

		  new_ins = InsDup(ins);
		  InsPrependToBbl(new_ins,new_bbl);
		  ArmInstructionUnconditionalizer(new_ins);
		  InsKill(ins);

		  ArmInstructionMakeCondBranch(BBL_INS_LAST(bbl),cond);

		  /*		  DiabloPrint(stdout,"PRE1return after @ieB @ieB\n",bbl,new_bbl);*/
		}
	    }
	}
    }
#endif
}


/*

PartialRedundancyElimination1ForCalls: if a conditional call is rarely
executed, and instructions depending on the same condition exist, they
are moved into a separate block, and replaced by a conditional branch

*/

void PartialRedundancyElimination1ForCalls(t_cfg * cfg)
{

#ifdef ARM_SUPPORT
  t_bbl * bbl;

  CfgComputeHotBblThreshold(cfg,0.95);

  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (BblIsHot(bbl))
	{
	  if (BblEndWithConditionalCall(bbl))
	    {
	      t_cfg_edge * taken = TakenPath(bbl);
	      t_cfg_edge * fall = FallThroughPath(bbl);
	      t_cfg_edge * link = LinkPath(bbl);
	      t_ins * last_ins = BBL_INS_LAST(bbl);
	      t_regset used_regs = INS_REGS_USE(last_ins);
	      t_regset cond_regs = INS_REGS_USE(last_ins);
	      t_regset changed_regs = RegsetNew();
	      t_ins * ins = INS_PREV(last_ins);
	      EXEC_CONDITION cond_last_ins = UNKNOWN;
	      EXEC_CONDITION cond = UNKNOWN;
	      
	      if (!(EDGE_EXEC_COUNT(taken)<EDGE_EXEC_COUNT(fall)/3))
		continue;

	      cond_last_ins = ArmInsCondition(last_ins);

	      while (ins)
		{
		  if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),cond_regs)))
		    {
		      ins = NULL;
		      break;
		    }

		  cond = ArmInsCondition(ins);

		  if (cond!=UNKNOWN && 
		      cond == cond_last_ins && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),used_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),changed_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(ins),changed_regs))
		      )
		    break;
		  
		  RegsetSetUnion(used_regs,INS_REGS_USE(ins));
		  RegsetSetUnion(changed_regs,INS_REGS_DEF(ins));
		  
		  ins = INS_PREV(ins);
		}

	      if (ins)
		{
		  t_bbl * new_bbl = BblNew(cfg);
		  t_bbl * new_bbl2 = BblNew(cfg);
		  t_ins * new_ins;
		  t_cfg_edge * new_edge;
		  t_cfg_edge * new_edge2;
		  t_cfg_edge * new_edge3;
		  t_cfg_edge * new_edge4;
		  
		  BblInsertInFunction(new_bbl,BBL_FUN(bbl));
		  BBL_EXEC_COUNT(new_bbl)=EDGE_EXEC_COUNT(taken);

		  BblInsertInFunction(new_bbl2,BBL_FUN(bbl));
		  BBL_EXEC_COUNT(new_bbl2)=EDGE_EXEC_COUNT(taken);
		  
		  ArmInstructionConvertToBranch(last_ins);

		  new_ins = InsNewForBbl(new_bbl);
		  ArmInstructionMakeCondBranchAndLink(new_ins,ARM_CONDITION_AL);
		  InsAppendToBbl(new_ins,new_bbl);
		  
		  new_ins = InsDup(ins);
		  InsPrependToBbl(new_ins,new_bbl);
		  InsKill(ins);

		  new_ins = InsNewForBbl(new_bbl2);
		  ArmInstructionMakeUncondBranch(new_ins);
		  InsAppendToBbl(new_ins,new_bbl2);

		  new_edge = CfgEdgeNew(cfg,bbl,new_bbl,ET_JUMP);
		  EDGE_EXEC_COUNT(new_edge)=EDGE_EXEC_COUNT(taken);

		  new_edge2 = CfgEdgeNew(cfg,new_bbl,new_bbl2,ET_FUNLINK);
		  EDGE_EXEC_COUNT(new_edge2)=0;
		  
		  new_edge3 = CfgEdgeNew(cfg,new_bbl2,CFG_EDGE_TAIL(link),ET_JUMP);
		  EDGE_EXEC_COUNT(new_edge3)=EDGE_EXEC_COUNT(taken);
		  
		  new_edge4 = CfgEdgeNewCall(cfg,new_bbl,CFG_EDGE_TAIL(taken),new_bbl2,FUN_BBL_LAST(BBL_FUN(CFG_EDGE_TAIL(taken))));
		  EDGE_EXEC_COUNT(new_edge4)=EDGE_EXEC_COUNT(taken);
		  
		  CfgEdgeKill(link);
		  CfgEdgeKill(taken->corr_edge);
		  CfgEdgeKill(taken);
		  
		  /*		  DiabloPrint(stdout,"potential important PRE candidate @ieB\n@ieB\n@ieB\n",bbl,new_bbl,new_bbl2);*/
		}
	    }
	}
    }
#endif
}

/*#define DEBUG_PRE2*/

/* instructions not live on the fall through path of a conditional
   branch are moved to the taken path if that is seldomly taken
*/

void PartialRedundancyElimination2(t_cfg * cfg)
{
#ifdef ARM_SUPPORT
  t_bbl * bbl;
  t_ins * ins;
  t_ins * ins_new;
#ifdef DEBUG_PRE2
  static   int teller = 0;
  t_string x = "candidateXXX.dot                    ";
  t_string y = "afterXXX.dot                        ";
#endif

  CfgComputeHotBblThreshold(cfg,0.99);

  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (BblIsHot(bbl))
	{
	  if (BblEndWithConditionalBranch(bbl))
	    {
	      t_cfg_edge * taken = TakenPath(bbl);
	      t_cfg_edge * fall  = FallThroughPath(bbl);
	      t_bbl * new_target;

	      t_regset taken_live = BBL_INS_FIRST(CFG_EDGE_TAIL(taken))?InsRegsLiveBefore(BBL_INS_FIRST(CFG_EDGE_TAIL(taken))):BBL_REGS_LIVE_OUT(CFG_EDGE_TAIL(taken));
	      t_regset fall_live = BBL_INS_FIRST(CFG_EDGE_TAIL(fall))?InsRegsLiveBefore(BBL_INS_FIRST(CFG_EDGE_TAIL(fall))):BBL_REGS_LIVE_OUT(CFG_EDGE_TAIL(fall));

	      t_regset changed_regs = RegsetNew();
	      t_regset used_regs = RegsetNew();

	      if (EDGE_EXEC_COUNT(taken)>=BBL_EXEC_COUNT(bbl)/2) continue;

	      BBL_FOREACH_INS_R(bbl,ins)
		{
		  if (!InsHasSideEffect(ins) && 
		      RegsetIsEmpty(RegsetNewIntersect(fall_live,INS_REGS_DEF(ins))) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),used_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),changed_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(ins),changed_regs))
		      )
		    {
		      break;
		    }

		  RegsetSetUnion(changed_regs,INS_REGS_DEF(ins));
		  RegsetSetUnion(used_regs,INS_REGS_USE(ins));
		}
	      
	      if (!ins) continue;
#ifdef DEBUG_PRE2
	      if (teller>=global_options.debugcounter) continue;

	      DiabloPrint(stdout,"in BBL @B\n we have a candidate @I\n taken: @E \n fall @E\n",bbl,ins,taken,fall);

	      DiabloPrintArch(stdout,cfg->sec->obj->description,"taken live @A fall live @A\n",taken_live, fall_live);

	      x = sDiabloPrint("candidate%d.dot",++teller);
	      y = sDiabloPrint("after%d.dot",teller);

	      Export_Function(BBL_FUN(bbl),x);
#endif
	      new_target = CreateNewTargetBlock(cfg,taken);

	      ins_new = InsDup(ins);
	      InsPrependToBbl(ins_new,new_target);
	      InsKill(ins);
	      
	      RegsetDup(BBL_REGS_LIVE_OUT(new_target),taken_live);

	      if (!(INS_ATTRIB(ins_new) & IF_CONDITIONAL))
		RegsetSetDiff(BBL_REGS_LIVE_OUT(bbl),INS_REGS_DEF(ins_new));

	      RegsetSetUnion(BBL_REGS_LIVE_OUT(bbl),INS_REGS_USE(ins_new));
	      /*	      RegsetSetUnion(BBL_REGS_LIVE_OUT(new_target),fall_live);*/
#ifdef DEBUG_PRE2
	      Export_Function(BBL_FUN(bbl),y);
#endif	      
	    }	      
	}
    }
#endif
}
/*#define DEBUG_PRE3*/

/* 
   instructions not live on the taken path of a conditional branch are
   moved to the fall through path if that is seldomly followed
*/


void PartialRedundancyElimination3(t_cfg * cfg)
{
#ifdef ARM_SUPPORT
  t_bbl * bbl;
  t_ins * ins;
  t_ins * ins_new;
#ifdef DEBUG_PRE3
  static   int teller = 0;
  t_string x = "3candidateXXX.dot                    ";
  t_string y = "3afterXXX.dot                        ";
#endif


  CfgComputeHotBblThreshold(cfg,0.99);

  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (BblIsHot(bbl))
	{
	  if (BblEndWithConditionalBranch(bbl))
	    {
	      t_cfg_edge * taken = TakenPath(bbl);
	      t_cfg_edge * fall  = FallThroughPath(bbl);
	      t_bbl * new_target;

	      t_regset taken_live = BBL_INS_FIRST(CFG_EDGE_TAIL(taken))?InsRegsLiveBefore(BBL_INS_FIRST(CFG_EDGE_TAIL(taken))):BBL_REGS_LIVE_OUT(CFG_EDGE_TAIL(taken));
	      t_regset fall_live = BBL_INS_FIRST(CFG_EDGE_TAIL(fall))?InsRegsLiveBefore(BBL_INS_FIRST(CFG_EDGE_TAIL(fall))):BBL_REGS_LIVE_OUT(CFG_EDGE_TAIL(fall));

	      t_regset changed_regs = RegsetNew();
	      t_regset used_regs = RegsetNew();

	      if (EDGE_EXEC_COUNT(fall)>=BBL_EXEC_COUNT(bbl)/2) continue;

	      BBL_FOREACH_INS_R(bbl,ins)
		{
		  if (!InsHasSideEffect(ins) && 
		      RegsetIsEmpty(RegsetNewIntersect(taken_live,INS_REGS_DEF(ins))) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),used_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),changed_regs)) && 
		      RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(ins),changed_regs))
		      )
		    {
		      break;
		    }

		  RegsetSetUnion(changed_regs,INS_REGS_DEF(ins));
		  RegsetSetUnion(used_regs,INS_REGS_USE(ins));
		}
	      
	      if (!ins) continue;
#ifdef DEBUG_PRE3
	      if (teller>global_options.debugcounter) continue;

	      DiabloPrint(stdout,"in BBL @B\n we have a candidate @I\n fall: @E \n taken @E\n",bbl,ins,fall,taken);

	      DiabloPrintArch(stdout,cfg->sec->obj->description,"OOPS @A @A\n",BBL_REGS_LIVE_OUT(CFG_EDGE_TAIL(taken)),INS_REGS_DEF(ins));
	      DiabloPrintArch(stdout,cfg->sec->obj->description,"fall live @A taken live @A\n",fall_live, taken_live);

	      x = sDiabloPrint("3candidate%d.dot",++teller);
	      y = sDiabloPrint("3after%d.dot",teller);

	      Export_Function(BBL_FUN(bbl),x);
#endif
	      new_target = CreateNewTargetBlock(cfg,fall);

	      ins_new = InsDup(ins);
	      InsPrependToBbl(ins_new,new_target);
	      InsKill(ins);
	      
	      RegsetDup(BBL_REGS_LIVE_OUT(new_target),fall_live);

	      if (!(INS_ATTRIB(ins_new) & IF_CONDITIONAL))
		RegsetSetDiff(BBL_REGS_LIVE_OUT(bbl),INS_REGS_DEF(ins_new));

	      RegsetSetUnion(BBL_REGS_LIVE_OUT(bbl),INS_REGS_USE(ins_new));
	      /*	      RegsetSetUnion(BBL_REGS_LIVE_OUT(new_target),taken_live);*/
#ifdef DEBUG_PRE3
	      Export_Function(BBL_FUN(bbl),y);
#endif	      
	    }	      
	}
    }
#endif
}

/*#define DEBUG_TRIANGLE*/

void ReplaceTriangleWithConditionalMove(t_cfg * cfg)
{
#ifdef ARM_SUPPORT
  t_bbl * upper;
  t_ins * ins_new;

#ifdef DEBUG_TRIANGLE
  static int teller = 0;
  t_string x = "triangle_preXXX.dot                  ";
  t_string y = "triangle_postXXX.dot                 ";
#endif

  CFG_FOREACH_BBL(cfg,upper)
    {
#ifdef DEBUG_TRIANGLE
      if (!(teller>=global_options.debugcounter))
#endif
	{
	  if (BblEndWithConditionalBranch(upper))
	    {
	      t_cfg_edge * taken = TakenPath(upper);
	      t_cfg_edge * fall  = FallThroughPath(upper);

	      t_bbl * corner = CFG_EDGE_TAIL(taken);

	      t_ins * one_ins = BBL_INS_FIRST(corner);

	      t_ins * branch_ins = BBL_INS_LAST(upper);

	      if (BBL_SUCC_FIRST(corner)!=BBL_SUCC_LAST(corner)) continue;
	      if (BBL_PRED_FIRST(corner)!=BBL_PRED_LAST(corner)) continue;
	      if (CFG_EDGE_TAIL(BBL_SUCC_FIRST(corner))!=CFG_EDGE_TAIL(fall)) continue;
	      if (!one_ins) continue;
	      if (INS_ATTRIB(one_ins) & IF_CONDITIONAL) continue;

	      if (BBL_NINS(corner)>2) continue;

#ifdef DEBUG_TRIANGLE
	      DiabloPrint(stdout,"TRIANGLE FIRST upper @iB\n corner @iB\n down @iB\n",upper,corner,CFG_EDGE_TAIL(fall));
	      
	      x = sDiabloPrint("triangle_pre%d.dot",++teller);
	      y = sDiabloPrint("triangle_post%d.dot",teller);
	      
	      Export_Function(BBL_FUN(upper),x);
#endif
	      ins_new = InsDup(one_ins);

	      ARMINSCONDITION(ins_new)=ARMINSCONDITION(branch_ins);
	      INS_ATTRIB(ins_new)|=IF_CONDITIONAL;
	      INS_REGS_USE(ins_new) = ArmUsedRegisters(ins_new);
	      INS_REGS_DEF(ins_new) = ArmDefinedRegisters(ins_new);

	      InsAppendToBbl(ins_new,upper);
	      EDGE_EXEC_COUNT(fall)+=EDGE_EXEC_COUNT(taken);

	      if (EDGE_CORR(taken))
		{
		  CfgEdgeKill(EDGE_CORR(taken));
		}
	      CfgEdgeKill(taken);

	      if(EDGE_CORR(BBL_SUCC_FIRST(corner))) 
		CfgEdgeKill(EDGE_CORR(BBL_SUCC_FIRST(corner)));
	      CfgEdgeKill(BBL_SUCC_FIRST(corner));
	     
	      InsKill(branch_ins);
	      InsKill(one_ins);
	      InsKill(BBL_INS_LAST(corner));
	      BblKill(corner);
	      
	      RegsetDup(BBL_REGS_LIVE_OUT(upper),BBL_INS_FIRST(CFG_EDGE_TAIL(fall))?InsRegsLiveBefore(BBL_INS_FIRST(CFG_EDGE_TAIL(fall))):BBL_REGS_LIVE_OUT(CFG_EDGE_TAIL(fall)));
			
#ifdef DEBUG_TRIANGLE
	      Export_Function(BBL_FUN(upper),y);
#endif
	    }
	}
    }
  

  CFG_FOREACH_BBL(cfg,upper)
    {
#ifdef DEBUG_TRIANGLE
      if (!(teller>=global_options.debugcounter))
#endif
	{
	  if (BblEndWithConditionalBranch(upper))
	    {
	      t_cfg_edge * taken = TakenPath(upper);
	      t_cfg_edge * fall  = FallThroughPath(upper);

	      t_bbl * corner = CFG_EDGE_TAIL(fall);
	      t_ins * first_ins = BBL_INS_FIRST(corner);
	      t_ins * last_ins = BBL_INS_LAST(corner);
	      t_ins * branch_ins = BBL_INS_LAST(upper);

	      if (BBL_SUCC_FIRST(corner)!=BBL_SUCC_LAST(corner)) continue;
	      if (BBL_PRED_FIRST(corner)!=BBL_PRED_LAST(corner)) continue;
	      if (CFG_EDGE_TAIL(BBL_SUCC_FIRST(corner))!=CFG_EDGE_TAIL(taken)) continue;
	      if (!first_ins) continue;
	      if (INS_ATTRIB(first_ins) & IF_CONDITIONAL) continue;
	      
	      if (BBL_NINS(corner)>2) continue;
	      if (BBL_NINS(corner)==2 && INS_TYPE(last_ins)!=IT_BRANCH) continue;
	      
#ifdef DEBUG_TRIANGLE
	      DiabloPrint(stdout,"TRIANGLE SECOND upper @ieB\n corner @ieB\n down @ieB\n",upper,corner,CFG_EDGE_TAIL(taken));

	      x = sDiabloPrint("triangle_pre%d.dot",++teller);
	      y = sDiabloPrint("triangle_post%d.dot",teller);
	      
	      Export_Function(BBL_FUN(upper),x);
#endif
	      
	      ins_new = InsDup(first_ins);
	      
	      if (INS_TYPE(first_ins)!=IT_BRANCH)
		{
		  ARMINSCONDITION(ins_new) = ArmInvertCondition(ARMINSCONDITION(branch_ins));
		  INS_ATTRIB(ins_new)|=IF_CONDITIONAL;
		  INS_REGS_USE(ins_new) = ArmUsedRegisters(ins_new);
		  INS_REGS_DEF(ins_new) = ArmDefinedRegisters(ins_new);
		}

	      InsAppendToBbl(ins_new,upper);		  
	      
	      if (last_ins!=first_ins)
		{
		  ins_new = InsDup(last_ins);
		  InsAppendToBbl(ins_new,upper);		  
		}
	      
	      EDGE_EXEC_COUNT(taken)+=EDGE_EXEC_COUNT(fall);

	      if (EDGE_CORR(fall))
		{
		  CfgEdgeKill(EDGE_CORR(fall));
		}
	      CfgEdgeKill(fall);

	      if(EDGE_CORR(BBL_SUCC_FIRST(corner))) 
		CfgEdgeKill(EDGE_CORR(BBL_SUCC_FIRST(corner)));

	      CfgEdgeKill(BBL_SUCC_FIRST(corner));

	      if (INS_TYPE(last_ins)!=IT_BRANCH)
		{
		  EDGE_CAT(taken)=ET_FALLTHROUGH; 
		}

	      InsKill(branch_ins);
	      if (last_ins!=first_ins)
		InsKill(last_ins);
	      InsKill(first_ins);
	      BblKill(corner);

	      RegsetDup(BBL_REGS_LIVE_OUT(upper),BBL_INS_FIRST(CFG_EDGE_TAIL(taken))?InsRegsLiveBefore(BBL_INS_FIRST(CFG_EDGE_TAIL(taken))):BBL_REGS_LIVE_OUT(CFG_EDGE_TAIL(taken)));
			

#ifdef DEBUG_TRIANGLE
	      Export_Function(BBL_FUN(upper),y);
#endif
	    }
	}
    }

  /* for returns on fallthrough */

  CFG_FOREACH_BBL(cfg,upper)
    {
#ifdef DEBUG_TRIANGLE
      if (!(teller>=global_options.debugcounter))
#endif
	{
	  if (BblEndWithConditionalBranch(upper))
	    {
	      t_cfg_edge * taken = TakenPath(upper);
	      t_cfg_edge * fall  = FallThroughPath(upper);

	      t_bbl * corner = CFG_EDGE_TAIL(fall);
	      t_bbl * lower = CFG_EDGE_TAIL(taken);
	      t_ins * first_ins = BBL_INS_FIRST(corner);
	      t_ins * branch_ins = BBL_INS_LAST(upper);

	      if (BBL_SUCC_FIRST(corner)!=BBL_SUCC_LAST(corner)) continue;
	      if (BBL_SUCC_FIRST(lower)!=BBL_SUCC_LAST(lower)) continue;
	      if (EDGE_CAT(BBL_SUCC_FIRST(lower))!=ET_JUMP) continue;
	      if (BBL_PRED_FIRST(corner)!=BBL_PRED_LAST(corner)) continue;
	      if (BBL_PRED_FIRST(lower)!=BBL_PRED_LAST(lower)) continue;
	      if (CFG_EDGE_TAIL(BBL_SUCC_FIRST(corner))!=FUN_BBL_LAST(BBL_FUN(corner))) continue;
	      if (!first_ins) continue;
	      if (INS_ATTRIB(first_ins) & IF_CONDITIONAL) continue;
	      if (BBL_NINS(corner)>1) continue;
	      

#ifdef DEBUG_TRIANGLE
	      DiabloPrint(stdout,"TRIANGLE FOURTH upper @ieB\n corner @ieB\n down @ieB\n",upper,corner,CFG_EDGE_TAIL(taken));

	      x = sDiabloPrint("triangle_pre%d.dot",++teller);
	      y = sDiabloPrint("triangle_post%d.dot",teller);
	      
	      Export_Function(BBL_FUN(upper),x);
#endif
	      DiabloPrint(stdout,"FOURTH upper @ieB\n corner @ieB\n down @ieB\n",upper,corner,CFG_EDGE_TAIL(taken));
	      /*	      Export_Function(BBL_FUN(upper),"before.dot");*/
	      ins_new = InsDup(first_ins);
	      
	      ARMINSCONDITION(ins_new) = ArmInvertCondition(ARMINSCONDITION(branch_ins));
	      INS_ATTRIB(ins_new)|=IF_CONDITIONAL;
	      INS_REGS_USE(ins_new) = ArmUsedRegisters(ins_new);
	      INS_REGS_DEF(ins_new) = ArmDefinedRegisters(ins_new);
	      InsAppendToBbl(ins_new,upper);		  
	      
	      CfgEdgeKill(fall);
	      CfgEdgeKill(BBL_SUCC_FIRST(corner));
	      
	      EDGE_EXEC_COUNT(CfgEdgeNew(cfg,upper,FUN_BBL_LAST(BBL_FUN(upper)),ET_JUMP))=BBL_EXEC_COUNT(corner);

	      if (EDGE_CAT(taken)==ET_JUMP)
		EDGE_CAT(taken)=ET_FALLTHROUGH; 
	      else
		EDGE_CAT(taken)=ET_IPFALLTHRU; 
		
	      
	      InsKill(branch_ins);
	      InsKill(first_ins);
	      BblKill(corner);

	      RegsetDup(BBL_REGS_LIVE_OUT(upper),BBL_INS_FIRST(CFG_EDGE_TAIL(taken))?InsRegsLiveBefore(BBL_INS_FIRST(CFG_EDGE_TAIL(taken))):BBL_REGS_LIVE_OUT(CFG_EDGE_TAIL(taken)));
	      /*	      Export_Function(BBL_FUN(upper),"after.dot");*/

#ifdef DEBUG_TRIANGLE
	      Export_Function(BBL_FUN(upper),y);
#endif
	    }
	}
    }



  CFG_FOREACH_BBL(cfg,upper)
    {
#ifdef DEBUG_TRIANGLE
      if (!(teller>=global_options.debugcounter))
#endif
	{
	  if (BblEndWithConditionalBranch(upper))
	    {
	      t_cfg_edge * taken = TakenPath(upper);
	      t_cfg_edge * fall  = FallThroughPath(upper);

	      if (CFG_EDGE_TAIL(taken)==CFG_EDGE_TAIL(fall))
		{

#ifdef DEBUG_TRIANGLE
		  DiabloPrint(stdout,"TRIANGLE THIRD upper @iB\n",upper);
		  
		  x = sDiabloPrint("triangle_pre%d.dot",++teller);
		  y = sDiabloPrint("triangle_post%d.dot",teller);
		  
		  Export_Function(BBL_FUN(upper),x);
#endif
		  EDGE_EXEC_COUNT(fall)+=EDGE_EXEC_COUNT(taken);
		  InsKill(BBL_INS_LAST(upper));
		  CfgEdgeKill(taken);
#ifdef DEBUG_TRIANGLE
		  Export_Function(BBL_FUN(upper),y);
#endif
		}
	    }
	}
    }

	      
#endif
}

void CfgBranchSwitch(t_cfg * cfg)
{
#ifdef ARM_SUPPORT
  t_bbl * upper;

  CfgComputeHotBblThreshold(cfg,0.90);
  
  CFG_FOREACH_BBL(cfg,upper)
    {
      if (BblIsHot(upper))
	{
	  if (BblEndWithConditionalBranch(upper))
	    {
	      t_cfg_edge * taken = TakenPath(upper);
	      t_cfg_edge * fall  = FallThroughPath(upper);
	      
	      t_bbl * taken_bbl = CFG_EDGE_TAIL(taken);
	      t_bbl * fall_bbl = CFG_EDGE_TAIL(fall);
	      
	      t_ins * last_ins_taken = BBL_INS_LAST(taken_bbl);
	      t_ins * last_ins_fall = BBL_INS_LAST(fall_bbl);

	      t_ins * branch_ins = BBL_INS_LAST(upper);

	      if (BBL_PRED_FIRST(fall_bbl)!=BBL_PRED_LAST(fall_bbl)) continue;
	      if (BBL_PRED_FIRST(taken_bbl)!=BBL_PRED_LAST(taken_bbl)) continue;
	      
	      if (BBL_SUCC_FIRST(fall_bbl)!=BBL_SUCC_LAST(fall_bbl)) continue;
	      if (BBL_SUCC_FIRST(taken_bbl)!=BBL_SUCC_LAST(taken_bbl)) continue;
	      
	      if (CFG_EDGE_TAIL(BBL_SUCC_FIRST(fall_bbl))!=CFG_EDGE_TAIL(BBL_SUCC_FIRST(taken_bbl)) && !BBL_IS_LAST(CFG_EDGE_TAIL(BBL_SUCC_FIRST(fall_bbl))) && !BBL_IS_LAST(CFG_EDGE_TAIL(BBL_SUCC_FIRST(taken_bbl)))) continue;
	      
	      if (EDGE_EXEC_COUNT(taken)>EDGE_EXEC_COUNT(fall) && !FtPath(taken_bbl,upper))
		{
		  t_uint32 tmp_type;

		  /*		  Export_Function(BBL_FUN(upper),"pre.dot");*/

		  DiabloPrint(stdout,"1 CAN GAIN ON TEST-BRANCH @ieB\n@ieB\n@ieB\n",upper,fall_bbl,taken_bbl);

		  tmp_type = EDGE_CAT(taken);
		  EDGE_CAT(taken)=EDGE_CAT(fall);
		  EDGE_CAT(fall)=tmp_type;

		  ARMINSCONDITION(branch_ins)=ArmInvertCondition(ARMINSCONDITION(branch_ins));

		  /*		  Export_Function(BBL_FUN(upper),"post.dot");*/

		}
	      
	      if (last_ins_taken && last_ins_fall && INS_TYPE(last_ins_taken)==IT_BRANCH && INS_TYPE(last_ins_fall)==IT_BRANCH)
		{
		  DiabloPrint(stdout,"CAN GAIN ON BRANCH IN ELSE & THEN @ieB\n@ieB\n@ieB\n",upper,fall_bbl,taken_bbl);
		  continue;
		}

	      if (BBL_EXEC_COUNT(taken_bbl)>BBL_EXEC_COUNT(fall_bbl) && last_ins_taken && INS_TYPE(last_ins_taken)==IT_BRANCH && CFG_EDGE_TAIL(BBL_SUCC_FIRST(fall_bbl))==CFG_EDGE_TAIL(BBL_SUCC_FIRST(taken_bbl))) 
		{
		  t_uint32 tmp_type;
		  t_ins * new_ins;
		  /*		  Export_Function(BBL_FUN(upper),"pre.dot");*/

		  DiabloPrint(stdout,"CAN GAIN ON TAKEN @ieB\n@ieB\n@ieB\n",upper,fall_bbl,taken_bbl);
		  
		  new_ins = InsDup(last_ins_taken);
		  InsAppendToBbl(new_ins,fall_bbl);
		  InsKill(last_ins_taken);

		  tmp_type = EDGE_CAT(BBL_SUCC_FIRST(taken_bbl));
		  EDGE_CAT(BBL_SUCC_FIRST(taken_bbl))=EDGE_CAT(BBL_SUCC_FIRST(fall_bbl));
		  EDGE_CAT(BBL_SUCC_FIRST(fall_bbl))=tmp_type;

		  /*		  Export_Function(BBL_FUN(upper),"post.dot");*/

		}
	      
	      if (BBL_EXEC_COUNT(fall_bbl)>BBL_EXEC_COUNT(taken_bbl) && last_ins_fall && INS_TYPE(last_ins_fall)==IT_BRANCH && CFG_EDGE_TAIL(BBL_SUCC_FIRST(fall_bbl))==CFG_EDGE_TAIL(BBL_SUCC_FIRST(taken_bbl)))
		{
		  t_uint32 tmp_type;
		  t_ins * new_ins;

		  DiabloPrint(stdout,"CAN GAIN ON FALL @ieB\n@ieB\n@ieB\n",upper,fall_bbl,taken_bbl);


		  new_ins = InsDup(last_ins_fall);
		  InsAppendToBbl(new_ins,taken_bbl);

		  InsKill(last_ins_fall);

		  tmp_type = EDGE_CAT(BBL_SUCC_FIRST(taken_bbl));
		  EDGE_CAT(BBL_SUCC_FIRST(taken_bbl))=EDGE_CAT(BBL_SUCC_FIRST(fall_bbl));
		  EDGE_CAT(BBL_SUCC_FIRST(fall_bbl))=tmp_type;

		  /*		  Export_Function(BBL_FUN(upper),"post.dot");*/
		}
	    }
	}
    }	      
#endif
}

void CfgBranchSwitch2(t_cfg * cfg)
{
#ifdef ARM_SUPPORT
  t_bbl * upper;

  CfgComputeHotBblThreshold(cfg,0.90);
  
  CFG_FOREACH_BBL(cfg,upper)
    {
      if (BblIsHot(upper))
	{
	  if (BblEndWithConditionalBranch(upper))
	    {
	      t_cfg_edge * taken = TakenPath(upper);
	      t_cfg_edge * fall  = FallThroughPath(upper);
	      t_cfg_edge * edge;
	      t_bool flag_fallthrough = FALSE;
	      t_uint32 tmp_type;
	      t_ins * branch_ins = BBL_INS_LAST(upper);
		
	      if (EDGE_EXEC_COUNT(taken)/2<=EDGE_EXEC_COUNT(fall)) continue;
	      if (CfgEdgeIsInterproc(taken)) continue;
	      if (CfgEdgeIsInterproc(fall)) continue;
	      if (FtPath(CFG_EDGE_TAIL(taken),upper)) continue;
	      if (CFG_EDGE_TAIL(taken)==upper) continue;

	      BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(taken),edge)
		{
		  if (
		      EDGE_CAT(edge)==ET_FALLTHROUGH || 
		      EDGE_CAT(edge)==ET_IPFALLTHRU || 
		      EDGE_CAT(edge)==ET_FUNLINK || 
		      EDGE_CAT(edge)==ET_IPFUNLINK || 
		      EDGE_CAT(edge)==ET_SWILINK || 
		      EDGE_CAT(edge)==ET_IPSWILINK || 
		      EDGE_CAT(edge)==ET_UNKNOWN ||  
		      EDGE_CAT(edge)==ET_SWITCH)
		    {
		      flag_fallthrough = TRUE;
		      break;
		    }
		}
	      
	      if (flag_fallthrough) continue;

	      DiabloPrint(stdout,"CAN 2 GAIN ON TEST-BRANCH @B\n",upper);
	      
	      /*DiabloPrint(stdout,"CAN 2 GAIN ON TEST-BRANCH @ieB\n@ieB\n@ieB\n",upper,CFG_EDGE_TAIL(fall),CFG_EDGE_TAIL(taken));
		Export_Function(BBL_FUN(upper),"before.dot");*/

	      tmp_type = EDGE_CAT(taken);
	      EDGE_CAT(taken)=EDGE_CAT(fall);
	      EDGE_CAT(fall)=tmp_type;
	      
	      ARMINSCONDITION(branch_ins)=ArmInvertCondition(ARMINSCONDITION(branch_ins));

	      /*	      Export_Function(BBL_FUN(upper),"after.dot");

	      DiabloPrint(stdout,"RESULT @ieB\n@ieB\n@ieB\n",upper,CFG_EDGE_TAIL(fall),CFG_EDGE_TAIL(taken));*/

	    }
	}
    }	      
#endif
}
