/******************************************************************************
 * Copyright 2001,2002, 2003: {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file section 
 * handling routines  
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/
/* Includes {{{ */
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stddef.h>
#include "../config.h"
#include "../simple_types.h"
#include "diablo_section.h"
#include "diablo_object.h"
#include "diablo_relocate.h"
#include "diablo_flowgraph.h"
#include "diablo_ins.h"
#include "diablo_relocate.h"
#include "diablo_print.h"
#include "diablo_description.h"
#include "diablo_iterators.h"
#include "diablo_addresses.h"
#include "../utils/util_malloc.h"
#include "../datastructures/data_string.h"
#include "../utils/util_error.h"
#include "../utils/util_endian.h"
#include "../diablo_options.h"
/*}}}*/

/* the global dummy sections undef and abs: 
 * these are the destination sections for ABS and UNDEF symbols */
t_section * undef_section, * abs_section;

void SectionInsert(t_section * sec)
{
  t_object * obj=sec->obj;
  char type=SECTION_TYPE(sec);
  t_section *** dest;
  t_uint32 * ndest;

  switch(type)
  {
    case CODE_SECTION:
      dest=&obj->code;
      ndest=&obj->ncodes;
      break;
    case RO_DATA_SECTION:
      dest=&obj->rodata;
      ndest=&obj->nrodatas;
      break;
    case DATA_SECTION:
      dest=&obj->data;
      ndest=&obj->ndatas;
      break;
    case NOTE_SECTION:
      dest=&obj->note;
      ndest=&obj->nnotes;
      break;
    case BSS_SECTION:
      dest=&obj->bss;
      ndest=&obj->nbsss;
      break;
    default:
      FATAL(("Unsupported type %c in add!",type));
      exit(1); /* Keeps the compiler happy */
  }

  *dest=Realloc(*dest,sizeof(t_section *)*((*ndest)+1));

  (*dest)[(*ndest)]=sec;
  (*ndest)++;
}

t_section * SectionCreateForObject(t_object * obj, char type, t_section * parent, t_address size, t_string name)
{
  t_section * ret = SectionNew(obj,type);

  SectionInsert(ret);
  SECTION_NAME(ret) = StringDup(name);
  SectionSetCAddressSuper(ret,NULL_GENERIC);
  SectionSetOldAddressSuper(ret,NULL_GENERIC);
  SECTION_CSIZE(ret) = SECTION_OLD_SIZE(ret) = size;
  if (type != BSS_SECTION) ret->data = Calloc(1,GENERIC_EXTRACT_UINT32(size));
  if (parent)
  {
    RELOCATABLE_TYPE(ret) = RT_SUBSECTION;
    SECTION_PARENT_SECTION(ret) = parent;
    ret->is_subsection = TRUE;
    SECTION_IS_MAPPED(ret) = TRUE;
  }
  
  return ret;
}

/* Section Constructors {{{ */
/* SectionGetFromObjectByAddress {{{ */

/*! Gets a section at an address in an objectfile
 *
 * \param obj The object in which we are searching
 * \param address The (generic) address at which we are looking
 *
 * \return t_section * The section at the address if found else NULL
*/

t_section * SectionGetFromObjectByAddress(t_object * obj, t_address  address)
{
   t_uint32 tel;
   for (tel=0; tel<obj->ncodes; tel++)
   {
      if (GENERIC_GE(address,SECTION_GET_CADDRESS(obj->code[tel])))
	 if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(obj->code[tel]),SECTION_GET_CADDRESS(obj->code[tel]))))
	 {
	    return obj->code[tel];
	 }
   }

   for (tel=0; tel<obj->ndatas; tel++)
      if (GENERIC_GE(address,SECTION_GET_CADDRESS(obj->data[tel])))
	 if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(obj->data[tel]),SECTION_GET_CADDRESS(obj->data[tel]))))
	 {
	    return obj->data[tel];
	 }
   
   for (tel=0; tel<obj->nnotes; tel++)
     if (GENERIC_GE(address,SECTION_GET_CADDRESS(obj->note[tel])))
       if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(obj->note[tel]),SECTION_GET_CADDRESS(obj->note[tel]))))
       {
	 return obj->note[tel];
       }

   for (tel=0; tel<obj->nrodatas; tel++)
      if (GENERIC_GE(address,SECTION_GET_CADDRESS(obj->rodata[tel])))
	 if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(obj->rodata[tel]),SECTION_GET_CADDRESS(obj->rodata[tel]))))
	 {
	    return obj->rodata[tel];
	 }

   for (tel=0; tel<obj->nbsss; tel++)
   {
      if (GENERIC_GE(address,SECTION_GET_CADDRESS(obj->bss[tel])))
	 if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(obj->bss[tel]),SECTION_GET_CADDRESS(obj->bss[tel]))))
	 {
	    return obj->bss[tel];
	 }
   }

   return NULL;
}
/*}}}*/
/*SubsectionGetFromObjectByAddress {{{ */
/*! Returns the subsection at a certain address for an object
 *
 * \param obj The object in which we search
 * \param address The address where the subsection should be
 *
 * \return t_section * A subsection if found, otherwise NULL
*/
t_section * SubsectionGetFromObjectByAddress(t_object * obj, t_address  address)
{
   t_object * tmp2, * tmp;
   for (tmp2=obj->mapped_first;tmp2!=NULL; tmp2=tmp2->next)
   {
      for (tmp=tmp2; tmp!=NULL; tmp=tmp->equal)
      {
	 t_uint32 tel;
	 for (tel=0; tel<tmp->ncodes; tel++)
	 {
	    if (GENERIC_GE(address,SECTION_GET_CADDRESS(tmp->code[tel])))
	       if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(tmp->code[tel]),SECTION_GET_CADDRESS(tmp->code[tel]))))
	       {
		  printf("Possible %s %s\n",tmp->name,tmp->code[tel]->name);
		  return tmp->code[tel];
	       }
	 }

	 for (tel=0; tel<tmp->ndatas; tel++)
	    if (GENERIC_GE(address,SECTION_GET_CADDRESS(tmp->data[tel])))
	       if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(tmp->data[tel]),SECTION_GET_CADDRESS(tmp->data[tel]))))
	       {
		  printf("Possible %s %s\n",tmp->name,tmp->data[tel]->name);
		  return tmp->data[tel];
	       }

	 for (tel=0; tel<tmp->nnotes; tel++)
	    if (GENERIC_GE(address,SECTION_GET_CADDRESS(tmp->note[tel])))
	       if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(tmp->note[tel]),SECTION_GET_CADDRESS(tmp->note[tel]))))
	       {
		  printf("Possible %s %s\n",tmp->name,tmp->note[tel]->name);
		  return tmp->note[tel];
	       }


	 for (tel=0; tel<tmp->nrodatas; tel++)
	    if (GENERIC_GE(address,SECTION_GET_CADDRESS(tmp->rodata[tel])))
	       if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(tmp->rodata[tel]),SECTION_GET_CADDRESS(tmp->rodata[tel]))))
	       {
		  printf("Possible %s %s\n",tmp->name,tmp->rodata[tel]->name);
		  return tmp->rodata[tel];
	       }

	 for (tel=0; tel<tmp->nbsss; tel++)
	 {
	    DiabloPrint(stdout,"@G\n",SECTION_GET_CADDRESS(tmp->bss[tel]));
	    if (GENERIC_GE(address,SECTION_GET_CADDRESS(tmp->bss[tel])))
	       if (GENERIC_LT(address,GENERIC_ADD(SECTION_CSIZE(tmp->bss[tel]),SECTION_GET_CADDRESS(tmp->bss[tel]))))
	       {
		  printf("Possible %s %s\n",tmp->name,tmp->bss[tel]->name);
		  return tmp->bss[tel];
	       }
	 }

      }
   }
   return NULL;
}
/*}}}*/
/*  ObjectGetSectionFromObjectByName {{{ */

/*!
 * \par obj The object from which we want to get a section 
 * \par name The name of the section 
 * 
 * \ret A section when found, NULL otherwise 
 * 
 * Get's a section from an objectfile by name.*/ 

t_section * SectionGetFromObjectByName(t_object * obj, t_string name)
{
  t_section_hash_node * hn=(t_section_hash_node *) HashLookup(obj->section_table,name);
  if (hn)
    {
      if (hn->node.equal) FATAL(("More then one section for section name in file %s, section name %s",obj->name,name));
      return hn->section;
    }
  else 
    { 
      return NULL;
    }
}

/*}}}*/
/* SectionNew {{{ */

/*! \par obj The parent object
 *  \par type The type of section
 *
 * \return An allocated section
 * 
 * Allocate and initialize a section, for more information about section types see
 * \ref SECTION_TYPES.
 */ 

t_section * SectionNew(t_object * obj, char type)
{
   t_section * ret=Calloc(1,sizeof(t_section));
   ret->obj=obj;
   ret->type=type;
   ret->data=NULL;
   ret->orig_data=NULL;
   ret->tmp_buf=NULL;
   ret->is_subsection=FALSE;
   ret->is_mapped=FALSE;
   ret->is_common_defined=FALSE;
   ret->is_fixed=FALSE;
   RELOCATABLE_TYPE(ret)=RT_SECTION;
   PRIVATE_SET_SECTION_OLD_ADDRESS(ret)=INVALID_GENERIC;
   SECTION_OLD_SIZE(ret)=INVALID_GENERIC;
   SECTION_RELOC_TABLE(ret)=NULL;
   SECTION_LIVENESS_MAP(ret)=NULL;

#ifdef DIETER
   ret->section_connections = NULL;
   ret->join_first = NULL;
   ret->hotness = 0;
   ret->join_next = NULL;
   ret->join_prev = NULL;
#endif

   return ret;
}
/*}}}*/
/*}}}*/
/* Load from a section {{{ */
/*! 
 * Get's the 32 bits data-word stored at a certain offset in a section. 
 * Automatically swaps endian. Calls should go through SectionGetData32
 *
 * \param file The file where the function was called from
 * \param lnno The linenumber where this function was called
 * \param sec  The section from which we want to load
 * \param offset The offset in the section
 *
 * \return A 32 bits word.
 */ 

/* For little endian */
#ifdef ALPHA
#define ALIGN_LOADS 1 
#endif
/* For big endian */
/*#define ALIGN_LOADS 0 */

t_uint64 RealSectionGetData64(char * file, int lnno, t_section * sec, t_address offset)
{
   void * load;
   if (GENERIC_GE(offset,SECTION_CSIZE(sec))) FATAL(("Out of section read at %s %d!",file,lnno));

   if (sec->orig_data)
     {
       load=GENERIC_ADD_DISPL(sec->orig_data,offset);
     }
   else
     {
       load=GENERIC_ADD_DISPL(sec->data,offset);
     }
   if (sec->obj->switched_endian)
     {
       return UINT64_SWAP_ENDIAN((*((t_uint64 *) load)));
     }
   else
     {
#ifdef ALIGN_LOADS
	t_uint64 ret=0;
	unsigned char * uload=load;
#if ALIGN_LOADS == 0
	ret=uload[0];
	ret<<=8;
	ret+=uload[1];
	ret<<=8;
	ret+=uload[2];
	ret<<=8;
	ret+=uload[3];
	ret<<=8;
	ret+=uload[4];
	ret<<=8;
	ret+=uload[5];
	ret<<=8;
	ret+=uload[6];
	ret<<=8;
	ret+=uload[7];
	return ret;
#else
	ret=uload[7];
	ret<<=8;
	ret+=uload[6];
	ret<<=8;
	ret+=uload[5];
	ret<<=8;
	ret+=uload[4];
	ret<<=8;
	ret+=uload[3];
	ret<<=8;
	ret+=uload[2];
	ret<<=8;
	ret+=uload[1];
	ret<<=8;
	ret+=uload[0];
	return ret;
#endif

       FATAL(("Implement"));
#else
       return (*((t_uint64 *) load));
#endif
     }
}



t_uint32 RealSectionGetData32(char * file, int lnno, t_section * sec, t_address offset)
{
   void * load;
   if (GENERIC_GE(offset,SECTION_CSIZE(sec))) FATAL(("Out of section read at %s %d (offset @G size @G) in sec %s from %s!",file,lnno,offset,SECTION_CSIZE(sec),sec->name,sec->obj->name));

   if (sec->orig_data)
     {
       load=GENERIC_ADD_DISPL(sec->orig_data,offset);
     }
   else
     {
       load=GENERIC_ADD_DISPL(sec->data,offset);
     }
   if (sec->obj->switched_endian)
     {
       return UINT32_SWAP_ENDIAN((*((t_uint32 *) load)));
     }
   else
     {
#ifdef ALIGN_LOADS
	t_uint32 ret=0;
	unsigned char * uload=load;
#if ALIGN_LOADS == 0
	ret=uload[0];
	ret<<=8;
	ret+=uload[1];
	ret<<=8;
	ret+=uload[2];
	ret<<=8;
	ret+=uload[3];
	return ret;
#else
	ret=uload[3];
	ret<<=8;
	ret+=uload[2];
	ret<<=8;
	ret+=uload[1];
	ret<<=8;
	ret+=uload[0];
	return ret;
#endif
#else
       return (*((t_uint32 *) load));
#endif
     }
}

t_uint16 RealSectionGetData16(char * file, int lnno, t_section * sec, t_address offset)
{
   void * load;
   if (GENERIC_GE(offset,SECTION_CSIZE(sec))) FATAL(("Out of section read at %s %d!",file,lnno));

   if (sec->orig_data)
     {
       load=GENERIC_ADD_DISPL(sec->orig_data,offset);
     }
   else
     {
       load=GENERIC_ADD_DISPL(sec->data,offset);
     }
   if (sec->obj->switched_endian)
     {
       return UINT16_SWAP_ENDIAN((*((t_uint16 *) load)));
     }
   else
     {
       return (*((t_uint16 *) load));
     }
}

t_uint8 RealSectionGetData8(char * file, int lnno, t_section * sec, t_address offset)
{
   void * load;
   if (GENERIC_GE(offset,SECTION_CSIZE(sec))) FATAL(("Out of section read at %s %d!",file,lnno));

   if (sec->orig_data)
     {
       load=GENERIC_ADD_DISPL(sec->orig_data,offset);
     }
   else
     {
       load=GENERIC_ADD_DISPL(sec->data,offset);
     }
   return (*((t_uint8 *) load));
}


/*}}}*/
/* Section Modifiers {{{*/
/* FUN : SectionInitDisassembly
 * PAR : 
 * RET : nothing 
 * DESC: */ 

void SectionInitDisassembly(t_section * section)
{
   t_object * obj=section->obj;
   t_address bits_in_section=GENERIC_MUL_UINT32(SECTION_CSIZE(section),8);
   t_uint32 bundlesize=obj->description->encoded_instruction_mod_size*obj->description->bundle_size+obj->description->template_size;
   t_uint32 max_ins=GENERIC_EXTRACT_UINT32(GENERIC_DIV_UINT32(bits_in_section,bundlesize))*obj->description->bundle_size;
   section->type=DISASSEMBLING_CODE_SECTION;
   SECTION_TMP_BUF(section)=Calloc(1, obj->description->decoded_instruction_size * max_ins);
   SECTION_NINS(section)=0;
}

/* FUN : SectionFiniDisassembly
 * PAR : 
 * RET : nothing
 * DESC: */ 

void SectionFiniDisassembly(t_object * obj, t_section * section)
{
   section->type=DISASSEMBLED_CODE_SECTION; /* We're disassembled */
   section->orig_data=section->data;
   SET_SECTION_DATA(section,SECTION_TMP_BUF(section)); /* Copy the instruction data to section data */
   SECTION_TMP_BUF(section)=NULL; /* The temporaries are currently not used */
   RelocsMigrateToInstructions(obj,section);
}

/* FUN : SectionInitFlowgraph
 * PAR : 
 * RET : 
 * DESC: */ 

void SectionInitFlowgraph(t_section * section)
{
    section->type=FLOWGRAPHING_CODE_SECTION;
}

void SectionInitDeflowgraph(t_object * obj, t_section * section)
{
    section->type=DEFLOWGRAPHING_CODE_SECTION;
}

/* FUN : SectionFiniFlowgraph
 * PAR : 
 * RET : 
 * DESC: */ 

void SectionFiniFlowgraph(t_section * section)
{
  t_ins * ins;
  SECTION_FOREACH_INS(section,ins)
  {
    InsFreeAllReferedRelocsBase(ins);
  }
  Free(section->data);
  section->data=SECTION_CFG(section);
  section->tmp_buf=NULL;
  section->type=FLOWGRAPHED_CODE_SECTION;
}

/* FUN : SectionInitAssembly
 * PAR : 
 * RET : nothing 
 * DESC: */ 

#define ceildiv(x,y)	((x)/(y) + ((x)%(y) ? 1 : 0))
#define pad4(x)		((x) % 4 ? (x) + (4 - ((x)%4)) : (x))
void SectionInitAssembly(t_object * obj, t_section * section, t_uint32 max_ins)
{
   t_uint32 tel;
   t_uint32 alloc_size = pad4(ceildiv(obj->description->maximum_encoded_instruction_size, 8) * SECTION_NINS(section));
   section->type=ASSEMBLING_CODE_SECTION;
   
   printf("Allocating %d instructions\n",SECTION_NINS(section));
   SECTION_TMP_BUF(section)=Malloc(alloc_size); 

   for (tel=0; tel< alloc_size/4 ; tel++)
   {
      *(((int *) (SECTION_TMP_BUF(section))) + tel) = 0xbdb00bdb;
   }
}
#undef pad4
#undef ceildiv

/* FUN : SectionFiniAssembly
 * PAR : 
 * RET : nothing
 * DESC: */ 

void SectionFiniAssembly(t_object * obj, t_section * section)
{
   t_ins * ins;
   t_reloc * relptr;
   
   section->type=CODE_SECTION; /* We're assembled */
   
   /*Adapt relocations*/
   SECTION_FOREACH_INS(section,ins)
   {
     while (RELOCATABLE_REFERS_TO(ins))
     {
       relptr = RELOCATABLE_REFERS_TO(ins)->rel;

       RelocSetFrom(relptr, T_RELOCATABLE(INS_SECTION(ins)));
       relptr->from_offset = GENERIC_ADD(GENERIC_SUB(INS_CADDRESS(ins),SECTION_GET_CADDRESS(INS_SECTION(ins))),relptr->from_offset);
     }
     while (RELOCATABLE_REFED_BY(ins))
     {
       
       relptr = RELOCATABLE_REFED_BY(ins)->rel;

       RelocSetToRelocatable(relptr, T_RELOCATABLE(INS_SECTION(ins)));
       relptr->to_offset = GENERIC_ADD(GENERIC_SUB(INS_CADDRESS(ins),SECTION_GET_CADDRESS(INS_SECTION(ins))),relptr->to_offset);
     }
     
   }


#if 0
#ifdef KERNEL_OPT
   /* temporary code for KERNEL_OPT:
    * migrate the relocations to the code section
    * instead of killing them off. we need them because
    * we will split the assembled section and move part
    * of it */
   /*TODO: this is broken! the code right above this kills all refedby and refers_to*/
   /*fields of instructions, so RelocSetToRelocatable and RelocSetFrom don't work any more*/

   {
     t_reloc * rel;
     FORALL_RELOCS(obj,rel)
     {
       if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(rel)) == RT_INS)
       {
	 /* change the TO */
	 t_ins * ins = REL_TO_RELOCATABLE(rel);
	 t_address t = StackExec(rel->code,rel,NULL,FALSE,0);
	 RelocSetToRelocatable(rel,section);
	 rel->to_offset = GENERIC_ADD(rel->to_offset,GENERIC_SUB(INS_CADDRESS(ins),SECTION_GET_CADDRESS(section)));
	 if (!GENERIC_EQ(StackExec(rel->code,rel,NULL,FALSE,0),t))
	   FATAL(("conversion failed"));
       }

       if (RELOCATABLE_TYPE(REL_FROM(rel)) == RT_INS)
       {
	 /* change the FROM */
	 t_address t = StackExec(rel->code,rel,NULL,FALSE,0);
	 t_ins * ins = REL_FROM(rel);
	 RelocSetFrom(rel,section);
	 rel->from_offset = GENERIC_ADD(rel->from_offset,GENERIC_SUB(INS_CADDRESS(ins),SECTION_GET_CADDRESS(section)));
	 if (!GENERIC_EQ(StackExec(rel->code,rel,NULL,FALSE,0),t))
	   FATAL(("conversion failed"));
       }
     }
   }
#endif
#endif
   
   Free(section->data); /* Free the instruction data */
   SET_SECTION_DATA(section,SECTION_TMP_BUF(section)); /* Copy the raw data to section data */
   SECTION_TMP_BUF(section)=NULL; /* The temporaries are currently not used */
   SECTION_NINS(section)=0;
}
/*}}}*/
/* Section Destructor {{{ */
/*! 
 * The destructor for a section 
 * \param to_free a section to free  
 */

void SectionFree(t_section * to_free)
{
  if (to_free->obj->section_table)
  {
    t_section_hash_node * hn=(t_section_hash_node *) HashLookup(to_free->obj->section_table,to_free->name);
    if (hn) hn->section=NULL;
  }

  if (to_free->tmp_buf) Free(to_free->tmp_buf);
  if (to_free->data) Free(to_free->data); 
  if (to_free->orig_data) Free(to_free->orig_data); 
  if (to_free->name) Free(to_free->name);
  Free(to_free);
}
/* }}} */
/* Hash related functions {{{ */
/* FUN : SectionHashNodeNew
  t_bbl * entry=NULL;
 * PAR : t_string (=hash key), t_section * (=a pointer to the stored section) 
 * RET : a SectionHashNode structure
 * DESC: Creates a node for the SectionHash, duplicates the key... */ 

t_section_hash_node * SectionHashNodeNew(t_string name, t_section * section)
{
   t_section_hash_node * ret=Malloc(sizeof(t_section_hash_node));
   ret->node.key=StringDup(name);
   ret->section=section;
   return ret;
}

/* FUN : SectionHashNodeFree
 * PAR : void * (=the section hashnode to free), void * (=must be zero)
 * RET : nothing
 * DESC: Free's a t_section_hash_node structure allocated with SectionHashNodeNew */ 

void SectionHashNodeFree(void * to_free,void * data)
{
   t_section_hash_node * s=to_free;
   Free(s->node.key);
   Free(to_free);
}
/*}}}*/


/*static int SectionCompareCAddress(const void * section_a, const void * section_b, void * param)
{
  t_section * node_a, * node_b;

  node_a = (t_section *) section_a;
  node_b = (t_section *) section_b;
  if (GENERIC_LE(GENERIC_ADD(SECTION_GET_CADDRESS(node_a),SECTION_CSIZE(node_a)),SECTION_GET_CADDRESS(node_b))) return -1;
  if (GENERIC_GE(SECTION_GET_CADDRESS(node_a),GENERIC_ADD(SECTION_GET_CADDRESS(node_b),SECTION_CSIZE(node_b)))) return 1;
  return 0;
}
*/

void SectionSetMinAddressSuper(t_section * sec, t_address address)
{
  PRIVATE_SET_SECTION_MIN_ADDRESS(sec)=address;
}

void SectionSetCAddressSuper(t_section * sec, t_address address)
{
  /*t_object * obj=sec->is_mapped?SECTION_OBJECT(sec->parent_section):SECTION_OBJECT(sec);*/
  /*t_section * collision;*/
  /*DiabloPrint(stdout,"Setting section address @G -> @G for section %s in object %s\n",address,GENERIC_ADD(address,SECTION_CSIZE(sec)),sec->name,obj->name);*/
  /*if (!obj->section_tree)*/
  /*{*/
  /*  obj->section_tree=ThreadedAvlTreeNew (offsetof(t_section,new_address),SectionCompareCAddress);*/
  /*}*/

  /*collision=ThreadedAvlTreeFind(obj->section_tree,sec);*/

  /*if (collision==sec) ThreadedAvlTreeDelete(obj->section_tree,sec);*/

  PRIVATE_SET_SECTION_CADDRESS(sec)=address;
  
  /*while ((collision=ThreadedAvlTreeDelete(obj->section_tree,sec))!=NULL)*/
  /*{*/
  /*  DiabloPrint(stdout,"Deleting section %s from tree, due to collision with %s\n",((collision)->name),sec->name);*/
  /*}*/
  /**/
  /*if (ThreadedAvlTreeInsert(obj->section_tree,sec)) FATAL(("Collision"));*/
}

void SectionSetOldAddressSuper(t_section * sec, t_address address)
{
  PRIVATE_SET_SECTION_OLD_ADDRESS(sec)=address;
}


/* SectionGetSubsections {{{ :return an array containing all subsections of a given parent section */

static int _sec_compare_old_addresses(t_section ** s1, t_section ** s2)
{
  if (GENERIC_LT(SECTION_GET_OLD_ADDRESS(*s1),SECTION_GET_OLD_ADDRESS(*s2)))
    return -1;
  if (GENERIC_GT(SECTION_GET_OLD_ADDRESS(*s1),SECTION_GET_OLD_ADDRESS(*s2)))
    return 1;
  return 0;
}
t_section ** SectionGetSubsections(t_section * sec, t_uint32 *nsubs_ret)
{
  t_object * obj, * subobj, * tmp;
  t_uint32 i,j;

  t_section ** subs;
  t_uint32 nsubs = 0;
  t_section * subsec;

  /* we want the section to be a parent section */
  VERIFY_FASSERT(SECTION_PARENT_SECTION(sec) == NULL, ("Need a parent section"));

  obj = sec->obj;

  if (SECTION_TYPE(sec) == DEFLOWGRAPHING_CODE_SECTION || 
      SECTION_TYPE(sec) == DEFLOWGRAPHED_CODE_SECTION ||
      SECTION_TYPE(sec) == FLOWGRAPHED_CODE_SECTION ||
      SECTION_TYPE(sec) == ASSEMBLING_CODE_SECTION)
  {
    /* in this case, no subsections should be returned:
     * subsections are meaningless for a flow graph... */
    nsubs = 0;
    subs = NULL;
  }
  else if (SECTION_TYPE(sec) == CODE_SECTION   || 
           SECTION_TYPE(sec) == RO_DATA_SECTION ||
	   SECTION_TYPE(sec) == DATA_SECTION   ||
	   SECTION_TYPE(sec) == BSS_SECTION    ||
	   SECTION_TYPE(sec) == NOTE_SECTION)
  {

    /* count the number of subsections */
    OBJECT_FOREACH_SUBOBJECT(obj, subobj, tmp)
    {
      OBJECT_FOREACH_SECTION(subobj, subsec, i)
      {
	if (SECTION_PARENT_SECTION(subsec) == sec)
	  nsubs++;
      }
    }

    if (nsubs)
    {
      /* store all subsections in an array and sort them */
      subs = (t_section **) Malloc(sizeof(t_section *) * nsubs);
      j = 0;
      OBJECT_FOREACH_SUBOBJECT(obj,subobj,tmp)
      {
	OBJECT_FOREACH_SECTION(subobj,subsec,i)
	  if (SECTION_PARENT_SECTION(subsec) == sec)
	    subs[j++] = subsec;
      }
    qsort(subs,nsubs,sizeof(t_section *),(int (*) (const void *,const void *)) _sec_compare_old_addresses);
    }
    else
    {
      subs=NULL;
    } 
  }
  else 
  {
    FATAL(("Unexpected section type %s %s %c",sec->name,sec->obj->name,sec->type));
    exit(0); /* Keeps the compiler happy */
  }

  *nsubs_ret = nsubs;
  return subs;
} /* }}} */

#ifdef DIETER
void SectionAppend(t_section * first, t_section * second)
{
  /* Do checks */
  if (SECTION_TYPE(first) != SECTION_TYPE(second))
    FATAL(("Can't append section of differend type: %c != %c", SECTION_TYPE(first), SECTION_TYPE(second)));
  if (SECTION_CSIZE(first) == NULL_GENERIC && SECTION_CSIZE(second) == NULL_GENERIC)
      return;
  /* D: TODO: More checks? */

/*    printf("Appending: 0x%X (0x%X, %d) and 0x%X (0x%X, %d)\n", first, first->data, SECTION_CSIZE(first), second, second->data, SECTION_CSIZE(second)); */

  /* Copy data */
  if(first->data && second->data)
    {
      first->data = Realloc(first->data, GENERIC_EXTRACT_UINT32(SECTION_CSIZE(first)) + GENERIC_EXTRACT_UINT32(SECTION_CSIZE(second)));
      memcpy(first->data + GENERIC_EXTRACT_UINT32(SECTION_CSIZE(first)), second->data, GENERIC_EXTRACT_UINT32(SECTION_CSIZE(second)));
    }
  else if(!first->data && second->data)
    {
      first->data = Malloc(GENERIC_EXTRACT_UINT32(SECTION_CSIZE(second)));
      memcpy(first->data, second->data, GENERIC_EXTRACT_UINT32(SECTION_CSIZE(second)));
    }

  /* Update relocs */
  while (RELOCATABLE_REFED_BY(second)) 
    {
      t_reloc_ref * tmp = RELOCATABLE_REFED_BY(second)->next;
      t_reloc * rel = RELOCATABLE_REFED_BY(second)->rel;

      if (REL_IS_TO_RELOCATABLE(rel))
	{
/*  	  DiabloPrint(stdout, "    1: @R to_offset=@G\n", rel, rel->to_offset); */
	  RelocSetToRelocatable(rel, T_RELOCATABLE(first));
	  rel->to_offset = GENERIC_ADD(SECTION_CSIZE(first), rel->to_offset);
/*  	  DiabloPrint(stdout, "    2: @R to_offset=@G\n", rel, rel->to_offset); */
	}
      else if (REL_IS_TO_SYMBOL(rel))
	{
	  FATAL(("Relocation is to symbol: implement!!!"));
	}

      RELOCATABLE_REFED_BY(second) = tmp;
    }

  while (RELOCATABLE_REFERS_TO(second))
    {
      t_reloc_ref * temp = RELOCATABLE_REFERS_TO(second)->next;
      t_reloc * rel = RELOCATABLE_REFERS_TO(second)->rel;
      
      if (REL_IS_TO_RELOCATABLE(rel))
	{
/*  	  DiabloPrint(stdout, "    1: @R from_offset=@G\n", rel, rel->from_offset); */
	  RelocSetFrom(rel, T_RELOCATABLE(first));
	  rel->from_offset = GENERIC_ADD(SECTION_CSIZE(first), rel->from_offset);
/*  	  DiabloPrint(stdout, "    2: @R from_offset=@G\n", rel, rel->from_offset); */
	}
      else if (REL_IS_TO_SYMBOL(rel))
	{
	  FATAL(("Relocation is from symbol: implement!!!"));
	}

      RELOCATABLE_REFERS_TO(second) = temp;
    }
  /* D: TODO: Only refedby and refersto?  */

  /* Update other */
  SECTION_CSIZE(first) = GENERIC_ADD(SECTION_CSIZE(first), SECTION_CSIZE(second));
  SECTION_MIN_SIZE(first) = GENERIC_ADD(SECTION_MIN_SIZE(first), SECTION_MIN_SIZE(second));
  /* D: TODO: Alignment? */
  /* D: TODO: Hashing? */
}

t_section_conn * SectionConnNew(t_section * section_a, t_section * section_b)
{
  if(section_a == section_b || SECTION_PARENT_SECTION(section_a) != SECTION_PARENT_SECTION(section_b))
    FATAL(("SectionConnection between same subsections or subsections are in different parentsection."));

  t_section_conn * conn = Malloc(sizeof(t_section_conn));
  conn->next = NULL;
  conn->prev = NULL;
  conn->section_a = section_a;
  conn->section_b = section_b;
  conn->hotness = 0;

  return conn;
}

void SectionConnFree(t_section_conn * to_free)
{
  Free(to_free);
}

t_section_conn * SectionConnUpdate(t_section * section_a, t_section * section_b, t_uint32 value)
{
  t_section * parent;
  t_section_conn * conn, * prev_conn = NULL;

  parent = SECTION_PARENT_SECTION(section_a);

  if(parent != SECTION_PARENT_SECTION(section_b))
    FATAL(("Section connection members can't have different parent sections!!!"));

  conn = parent->section_connections;
  
  /* If there is already a connection between the sections, update it. */
  while(conn)
    {
      if((conn->section_a == section_a && conn->section_b == section_b) ||
	 (conn->section_b == section_a && conn->section_a == section_b))
	{
	  conn->hotness += value;
	  break;
	}

      prev_conn = conn;
      conn = conn->next;
    }

  /* The parent section doesn't contain any connections yet. */
  if(!prev_conn && !conn)
    {
      conn = parent->section_connections = SectionConnNew(section_a, section_b);
      conn->hotness = value;
    }
  else
    {
      /* Add a new non-existing connection. */
      if(!conn)
	{
	  conn = SectionConnNew(section_a, section_b);
	  conn->hotness = value;
	  conn->prev = prev_conn;
	  conn->prev->next = conn;
	}
      
      /* Keep the list sorted by hotness: find its new place */
      while(prev_conn && conn->hotness >= prev_conn->hotness)
	{
	  prev_conn = prev_conn->prev;
	}

      /* Keep the list sorted by hotness: move to new place */
      if(conn->prev != prev_conn)
	{
	  conn->prev->next = conn->next;
	  if(conn->next) conn->next->prev = conn->prev;
	  
	  if(!prev_conn)
	    {
	      conn->next = parent->section_connections;
	      parent->section_connections = conn;
	      conn->prev = NULL;
	      conn->next->prev = conn;
	    }
	  else
	    {
	      conn->next = prev_conn->next;
	      prev_conn->next = conn;
	      conn->prev = prev_conn;
	      conn->next->prev = conn;
	    }
	}
    }

  return conn;
}

void SectionConnPrint(t_section_conn * conn, t_bool all)
{
  if(all)
    {
      while(conn)
	{
	  SectionConnPrint(conn, FALSE);
	  conn = conn->next;
	}
    }
  else
    {
      DiabloPrint(stdout, "%s\n", conn->section_a->obj->name);
      printf("  <- 0x%X ->\n", conn->hotness);
      DiabloPrint(stdout, "%s\n", conn->section_b->obj->name);
    }
}
#endif

/* vim: set shiftwidth=2 foldmethod=marker: */
