/***********************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Header file for section manipulation 
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 ***********************************************************/
/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup SECTIONS Sections */
/*@{ */
/* Typedefs {{{ */
#ifndef DIABLO_SECTION_TYPEDEFS
#define DIABLO_SECTION_TYPEDEFS
typedef struct _t_section t_section;
typedef struct _t_cfg_entries t_cfg_entrypoints;
typedef struct _t_section_hash_node t_section_hash_node;
typedef struct _t_section_conn t_section_conn;
#endif /* MY_SECTION_TYPEDEFS */
/* }}} */
#ifndef DIABLO_SECTION_H
#define DIABLO_SECTION_H
/* Includes {{{ */
#include "../simple_types.h"
#include "../datastructures/data_hash.h"
#include "../datastructures/data_graph.h"
#include "diablo_object.h"
#include "diablo_function.h"
#include "diablo_edge.h"
#include "diablo_bbl.h"
#include "diablo_reloc.h"
/*}}}*/
/*! \defgroup SECTION_TYPES Section types */
/*@{{{ Section types */
/*! A normal code section, 
 * datasize= alignment for this section, 
 * data    = the raw data in this section, 
 * tmp_buf = not used
 * tmp32   = TODO: the size of an instruction for this architecture 
 */
#define CODE_SECTION 'C' 
/*! Temporary state, busy disassembling, 
 * datasize = size of instruction, 
 * data     = raw_data, 
 * tmp_buf  = disassemble buffer, 
 * tmp32    = instruction struct size (should be set at the start of disassembling) */ 
#define DISASSEMBLING_CODE_SECTION 'd'
/*! A disassembled code section, 
 * datasize = size of instruction struct, 
 * data     = linear list of instructions 
 * tmp_buf  = not used
 * tmp32    = not used */
#define DISASSEMBLED_CODE_SECTION 'I'
/*! Temporary state, busy flowgraphing, 
 * datasize = size of instruction, 
 * data     = raw_data, 
 * tmp_buf  = disassemble buffer, 
 * tmp32    = instruction struct size */ 
#define FLOWGRAPHING_CODE_SECTION 'f' 
/*! A flowgraph for a section, 
 * datasize = size of instruction struct, 
 * data     = Flowgraph information (entries, functions, blocks not in a function, graph) 
 * tmp_buf  = not used
 * tmp32    = not used */
#define FLOWGRAPHED_CODE_SECTION 'F' 

/*! Temporary state, busy deflowgraphing, 
 * datasize = TODO 
 * data     = TODO 
 * tmp_buf  = TODO 
 * tmp32    = TODO */ 
#define DEFLOWGRAPHING_CODE_SECTION 'p' 

/*! Temporary state, after layout but before assembling, 
 * datasize = TODO 
 * data     = pointer to the first bbl of the chain that represents the layout of this section 
 * tmp_buf  = TODO 
 * tmp32    = TODO */ 
#define DEFLOWGRAPHED_CODE_SECTION 'l' 

/*! Temporary state, busy assembling,  */
#define ASSEMBLING_CODE_SECTION 'a' 


#define RO_DATA_SECTION 'R'
#define DATA_SECTION 'D'
#define NOTE_SECTION 'N'
#define BSS_SECTION  'B'
/*@}}}*/
/* struct _t_section {{{ */
/*! The section structure */
struct _t_section
{
   t_relocatable relocatable;
   /** the parent object for this section */
   t_object * obj;
   /** the section type */
   char type;
   /** the name of the section */
   t_string name; 
   /** is this a subsection of a larger section? */
   t_bool is_subsection;
   t_bool is_mapped;
   t_bool is_common_defined;
   /** has this section a fixed address (forbidden to move this section)? */
   t_bool is_fixed;
   t_section * parent_section;
   t_uint32 nins;
   void * tmp_buf; 
   void * data;
   char * liveness_map;
   t_reloc ** relocation_table;
   /** Original data, for ?? purposes */
   void * orig_data;
   /** section flags */
   t_uint32 flags;
   t_uint32 tmp;

#ifdef DIETER
  /* Used for address producer optimization */

  /* Used only by parent-section */
  t_section_conn * section_connections;
  struct _t_section * join_first;

  /* Used only by subsections */

  /** Three possibilities here:
      1. Counting how many times an address in this section can be used for an
      optimalisation.
      2. Counting how many times an address in another section was available
      for optimalisation.
      3. Sum of 1. and 2. */
  t_uint32 hotness;
  struct _t_section * join_next;
  struct _t_section * join_prev;

#endif
};
/*}}}*/

#ifdef DIETER

struct _t_section_conn
{
  struct _t_section_conn * next;
  struct _t_section_conn * prev;

  /** Counts the opportunities for optimalisation
      i.e how many times was an address in section_a available when calculating
      an address in section_b plus how many times was an address in section_b
      available when calculating an address in section_a.*/
  t_uint32 hotness;
  t_section * section_a;
  t_section * section_b;
};

t_section_conn * SectionConnNew(t_section * section_a, t_section * section_b);
void SectionConnFree(t_section_conn * to_free);
t_section_conn * SectionConnUpdate(t_section * section_a, t_section * section_b, t_uint32 value);
void SectionConnPrint(t_section_conn * conn, t_bool all);

#endif

/*! \defgroup SECTION_ALLOC Constructors and destructors */
/*@{{{ Constructors and destructors */
t_section * SectionNew(t_object * obj, char type);
t_section * SectionCreateForObject(t_object * obj, char type, t_section * parent, t_address size, t_string name);
t_section * SectionGetFromObjectByAddress(t_object * obj, t_address  address);
t_section * SectionGetFromObjectByName(t_object * obj, t_string name);
void SectionFree(t_section *);
/*@}}} Constructors and destructors */
/*! \defgroup SECTION_TYPE_MODIFIERS Section types modifiers */
/*@{{{ Section type modifiers */
void SectionInitAssembly(t_object * obj, t_section * section, t_uint32 max_ins);
void SectionFiniAssembly(t_object * obj, t_section * section);
void SectionInitDisassembly(t_section *);
void SectionFiniDisassembly(t_object * obj, t_section * section);
void SectionInitFlowgraph(t_section * section);
void SectionFiniFlowgraph(t_section * section);
void SectionInitDeflowgraph(t_object * obj, t_section * section);
/*@}}}*/
/*! \defgroup SECTION_HASH_RELATED Everything related to the hash table of sections in an object */
/*@{{{ Section Hash Tables */
struct _t_section_hash_node
{
   data_hash_node node;
   t_section * section;
};

t_section_hash_node * SectionHashNodeNew(t_string name, t_section * section);
void SectionHashNodeFree(void * to_free,void * data);
/*@}}}*/
/* To (re)move {{{ */
/*! Describes the entry points for a cfg (linked list) */
struct _t_cfg_entries
{
   t_bbl * entry_bbl;
   t_cfg_edge * entry_edge;
   void * tmp;
   struct _t_cfg_entries * next;
};

t_cfg_entrypoints * SectionGetEntries(t_section * section);
#define SECTION_BASE_ADDRESS(section) ((section)->relocatable.address)
#define SECTION_RELOC_TABLE(section)  ((section)->relocation_table)

#define MAP_SECTION(sec,name,psec) do { if (SECTION_IS_MAPPED(sec)) FATAL(("Section is already mapped!")); sec->is_subsection = sec->is_mapped = TRUE; SECTION_PARENT_SECTION(sec)=psec; } while(0)
/*}}}*/
/* Casts and one-to-one mappings {{{ */
/*! Convert type to section */
#define T_SECTION(section)       ((t_section *) (section))
/*! Get the parent object for this section */
#define SECTION_OBJECT(sec)   (((t_section *)sec)->obj)
/*! Get the cfg in this section (only possible when flowgraphed) */
#define SECTION_CFG(sec)      ((((sec)->type==FLOWGRAPHED_CODE_SECTION) || ((sec)->type==DEFLOWGRAPHING_CODE_SECTION))?((t_cfg *) sec->data):(((sec)->type==FLOWGRAPHING_CODE_SECTION)?(t_cfg *) SECTION_TMP_BUF(sec):NULL))
/*}}}*/
/* Getters, Setters and info {{{ */
/* Boolean info for sections {{{ */
/*! true if section is a datasection */
#define SECTION_IS_DATA(section)	(section->type == DATA_SECTION || section->type == RO_DATA_SECTION || section->type == BSS_SECTION)
/*! true if section is a codesection */
#define SECTION_IS_CODE(section)	(!SECTION_IS_DATA(section))
/*! true if the section is mapped to a parent object */
#define SECTION_IS_MAPPED(section) ((section)->is_mapped)
/*}}}*/
/* Field selection from sections {{{ */
#define SECTION_GET_OLD_ADDRESS(section)   ((section)->relocatable.old_address)
#define PRIVATE_SET_SECTION_OLD_ADDRESS(section)   ((section)->relocatable.old_address)
#define SECTION_OLD_SIZE(section) ((section)->relocatable.old_size)
#define SECTION_CSIZE(section)   ((section)->relocatable.csize)
#define SECTION_MIN_SIZE(section)   ((section)->relocatable.min_size)
#define SECTION_ALIGN(section)   ((section)->relocatable.alignment)
#define SECTION_PARENT_SECTION(section)   ((T_SECTION(section))->parent_section)
#define SECTION_GET_CADDRESS(section) ((T_SECTION(section))->relocatable.caddress)
#define SECTION_GET_END_ADDRESS(section) GENERIC_ADD(((T_SECTION(section))->relocatable.caddress),((T_SECTION(section))->relocatable.csize))
#define SECTION_GET_MIN_ADDRESS(section) ((T_SECTION(section))->relocatable.min_address)
#define PRIVATE_SET_SECTION_CADDRESS(section) ((T_SECTION(section))->relocatable.caddress)
#define PRIVATE_SET_SECTION_MIN_ADDRESS(section) ((T_SECTION(section))->relocatable.min_address)
#define SECTION_LIVENESS_MAP(section) ((section)->liveness_map)
#define SECTION_TMP_BUF(section) (section->tmp_buf)
#define SECTION_NINS(section) (T_SECTION(section)->nins)
#define SECTION_TYPE(section) (T_SECTION(section)->type)
#define SECTION_NAME(section) (T_SECTION(section)->name)
#define SECTION_FLAGS(section) (T_SECTION(section)->flags)
void SectionSetCAddressSuper(t_section * sec, t_address address);
void SectionSetMinAddressSuper(t_section * sec, t_address address);
void SectionSetOldAddressSuper(t_section * sec, t_address address);
void SectionInsert(t_section * sec);
#ifdef DIETER
void SectionAppend(t_section * first, t_section * second);
#endif
/*}}}*/
/* Field setters for sections {{{*/
#define SET_SECTION_NINS(section, mnins) do { if (SECTION_IS_MAPPED(section)) FATAL(("Section is mapped!")); section->mapped_or_not.not_mapped.nins=mnins; } while(0)
#define SET_SECTION_TMP_BUF(section, tmp) do { if (SECTION_IS_MAPPED(section)) FATAL(("Section is mapped!")); section->mapped_or_not.not_mapped.tmp_buf=tmp; } while(0)
#define SET_SECTION_DATASIZE(section, size) do { if (SECTION_IS_MAPPED(section)) FATAL(("Section is mapped!")); section->mapped_or_not.not_mapped.data_size=size; } while(0)
#define SET_SECTION_DATA(section, vdata)     do { section->data=vdata; } while(0)
/*}}}*/
/*}}}*/
/* Get data - Load from a section {{{ */
#define SectionGetData64(sec,offset) RealSectionGetData64(__FILE__,__LINE__,sec,offset)
#define SectionGetData32(sec,offset) RealSectionGetData32(__FILE__,__LINE__,sec,offset)
#define SectionGetData16(sec,offset) RealSectionGetData16(__FILE__,__LINE__,sec,offset)
#define SectionGetData8(sec,offset) RealSectionGetData8(__FILE__,__LINE__,sec,offset)
t_uint64 RealSectionGetData64(char * file, int lnno, t_section * sec, t_address offset);
t_uint32 RealSectionGetData32(char * file, int lnno, t_section * sec, t_address offset);
t_uint16 RealSectionGetData16(char * file, int lnno, t_section * sec, t_address offset);
t_uint8 RealSectionGetData8(char * file, int lnno, t_section * sec, t_address offset);
/* }}} */
/* {{{ section flags */
#define SECTION_FLAG_KEEP	1	/* do not remove this section, even though it appears unconnected in the graph */
/* }}} */

t_section ** GetSubsecsForSec(t_section * sec, t_uint32 *nsubs_ret);
t_section ** SectionGetSubsections(t_section * sec, t_uint32 *nsubs_ret);

/* the global dummy sections: destination for ABS and UNDEF symbols and relocs */
extern t_section * undef_section, * abs_section;

#endif
/*@}*/
/*@}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
