/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file routines 
 * 
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/
#include "string.h"
#include "diablo_addresses.h"
#include "diablo_reloc.h"
#include "../utils/util_error.h"
#include "../diablo_options.h"

/* For documentation of the machine stack language type ./scripts/stackdoc in
 * the main diablo directory! */

/*#define DEBUG_STACK_EXEC*/

/* {{{ defines and typedefs */
#define GetImm64(s)	GetImm(s,16)
#define GetImm32(s)	(t_uint32) GetImm(s,8);
#define GetImm16(s)	(t_uint16) GetImm(s,4);
#define GetImm8(s)	(t_uint8)  GetImm(s,2);

/*! This structure holds the stack, and the current position in the stack while
 * executing */
typedef struct _t_exec_stack {
  t_address * stack;
  t_uint32 maxelem;
  t_int32 top;
} t_exec_stack;
/* }}} */

/* {{{ helper functions */
/* get immediate of given size from string s.
 * all immediates are unsigned hex numbers */
static t_uint64 GetImm(char * s, int numchars)
{
  int i;
  t_uint64 ret = 0;

  for (i = 0; i < numchars; i++)
  {
    ret <<= 4;
    if ((*s >= '0') && (*s <= '9'))
      ret += (*s - '0');
    else if ((*s >= 'a') && (*s <= 'f'))
      ret += (*s - 'a' + 10);
    else if ((*s >= 'A') && (*s <= 'F'))
      ret += (*s - 'A' + 10);
    else
      FATAL(("Malformed number!"));

    s++;
  }
  return ret;
}

static void StackInit(t_exec_stack * st)
{
  st->stack = Malloc(20 * sizeof(t_address));
  st->top = -1;
  st->maxelem = 20;
}

static void StackFini(t_exec_stack * st)
{
  Free(st->stack);
  st->maxelem = 0;
  st->top = -1;
}

static void Push(t_exec_stack * st, t_address elem)
{
#ifdef DEBUG_STACK_EXEC
  DiabloPrint(stdout,"Pushed @G\n",elem);
#endif

  if (++st->top == st->maxelem)
  {
    /* enlarge stack */
    st->stack = Realloc(st->stack,(st->maxelem + 10)*sizeof(t_address));
    st->maxelem += 10;
  }
  st->stack[st->top] = elem;
}

static t_address Pop(t_exec_stack * st)
{
  if (st->top < 0) FATAL(("Popping empty stack"));
  return st->stack[st->top--];
}

/* rotate the stack left: top becomes bottom */
static void Rotl(t_exec_stack * st)
{
  t_address tmp;
  if (st->top == -1) FATAL(("Rotl on empty stack"));
  if (st->top == 0) return;

  tmp = st->stack[st->top];
  memmove(st->stack+1,st->stack,st->top*sizeof(t_address));
  st->stack[0] = tmp;
}

/* rotate the stack right: bottom becomes top */
static void Rotr(t_exec_stack * st)
{
  t_address tmp;
  if (st->top == -1) FATAL(("Rotr on empty stack"));
  if (st->top == 0) return;

  tmp = st->stack[0];
  memmove(st->stack,st->stack+1,st->top*sizeof(t_address));
  st->stack[st->top] = tmp;
}
/* }}} */

/* {{{ the actual stack machine */
t_address StackExec(t_string program, t_reloc * rel, char * base, t_bool write,t_uint32 mode)
{
  char * cmd = program;
  t_exec_stack estack;
  t_exec_stack * st = &estack;
  t_address ret = NULL_GENERIC;
  t_address tmp, oper;
  enum {CALC,WRITE,VERIFY} phase;

  int currloadins=0, currstoreins=0;

  char * bundleptr = base;
  char * memptr = base;

  StackInit(st);

  /* adjust the memptr to point to the first reloced byte.
   * this allows us to keep the base pointer as simple as
   * sec->data or something like that */
  memptr = GENERIC_ADD_DISPL(memptr,rel->from_offset);

  /* For IA64 Bundles: */
  bundleptr = GENERIC_ADD_DISPL(bundleptr,GENERIC_NAND_UINT32(rel->from_offset,0xf));
  currloadins=GENERIC_EXTRACT_UINT32(rel->from_offset)&0xf;
  currstoreins=GENERIC_EXTRACT_UINT32(rel->from_offset)&0xf;


  phase = CALC;

  while (*cmd)
  {
    switch (*(cmd++))
    {
      /* ignore spaces for readability */
      case ' ':
	break;

	/******** push commands ********/
      case 'M': /* (M)odus : pops an address of the stack and check if it refers to thumb code. If so 1 is pushed, else 0 is pushed */
	{
	  t_address oper1 = Pop(st);
	  t_object * obj;
	  if (RELOCATABLE_TYPE(REL_FROM(rel))==RT_SUBSECTION)
	    obj = SECTION_OBJECT(SECTION_PARENT_SECTION(REL_FROM(rel)));
	  else if (RELOCATABLE_TYPE(REL_FROM(rel))==RT_INS)
	    obj = INS_OBJECT(REL_FROM(rel));
	  else 
	  {
	    FATAL(("Implement!"));
	    exit(0);
	  }

	  if(ADDRESS_IS_SMALL_CODE == SymbolTableGetCodeType(obj,oper1) && ADDRESS_IS_DATA != SymbolTableGetDataType(obj,oper1))
	  {
	    Push(st,GENERIC_ADD_UINT32(NULL_GENERIC,1));
	  }
	  else
	  {
	    Push(st,NULL_GENERIC);
	  }
	}
	break;
      case 'S': /* refered (S)ymbol - push the destination of the reloc. The name refers to the S used in the Elf manuals. */
	if (mode==1)
	  Push(st,GENERIC_ADD(REL_TO_RELOCATABLE(rel)->min_address,rel->to_offset));
	else if (mode==2)
	  Push(st,GENERIC_ADD(REL_TO_RELOCATABLE(rel)->old_address,rel->to_offset));
	else
	  Push(st,GENERIC_ADD(REL_TO_RELOCATABLE(rel)->caddress,rel->to_offset));
	break;
      case 'O': /* the (O)ldest - pushes the address of the relocatables parent if it exists, else the address of the relocatable */
	if (REL_TO_RELOCATABLE(rel)->type==RT_SECTION)
	{
	  if (mode==1)
	    Push(st,REL_TO_RELOCATABLE(rel)->min_address);
	  else if (mode==2)
	    Push(st,REL_TO_RELOCATABLE(rel)->old_address);
	  else
	    Push(st,REL_TO_RELOCATABLE(rel)->caddress);
	  break;
	} 
	else if (REL_TO_RELOCATABLE(rel)->type==RT_SUBSECTION)
	{ 
	  if (!SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(rel))) FATAL(("Subsection has no parent in O-stack command"));
	  if (mode==1)
	    Push(st,T_RELOCATABLE(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(rel)))->min_address);
	  else if (mode==2)
	    Push(st,T_RELOCATABLE(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(rel)))->old_address);
	  else
	    Push(st,T_RELOCATABLE(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(rel)))->caddress);
	  break;
	}
	else FATAL(("Strange relocatable in O-stack command"));
      case 'A': /* (A)ddend - push the addend of the reloc */
	Push(st,rel->addend);
	break;
      case 'P': /* (P)osition - push the from address of the reloc */
	if (mode==1)
	  Push(st,GENERIC_ADD(REL_FROM(rel)->min_address,rel->from_offset));
	else if (mode==2)
	  Push(st,GENERIC_ADD(REL_FROM(rel)->old_address,rel->from_offset));
	else
	  Push(st,GENERIC_ADD(REL_FROM(rel)->caddress,rel->from_offset));
	break;
      case 'R': /* associated (R)elocatable - push the address of the associated relocatable (e.g. GOT section) */
	if (mode==1)
	  Push(st,GENERIC_ADD(rel->e.sec->min_address,rel->e_offset));
	else if (mode==2)
	  Push(st,GENERIC_ADD(rel->e.sec->old_address,rel->e_offset));
	else
	{
	  Push(st,GENERIC_ADD(rel->e.sec->caddress,rel->e_offset));
	}
	break;
      case 'T': /*  push the address of the parent of the associated relocatable (e.g. GOT section) */
	{
	  if (REL_E_TO_RELOCATABLE(rel)->type==RT_SECTION)
	  {
	    if (mode==1)
	      Push(st,REL_E_TO_RELOCATABLE(rel)->min_address);
	    else if (mode==2)
	      Push(st,REL_E_TO_RELOCATABLE(rel)->old_address);
	    else
	      Push(st,REL_E_TO_RELOCATABLE(rel)->caddress);
	    break;
	  } 
	  else if (REL_E_TO_RELOCATABLE(rel)->type==RT_SUBSECTION)
	  { 
	    if (!SECTION_PARENT_SECTION(REL_E_TO_RELOCATABLE(rel))) FATAL(("Subsection has no parent in O-stack command"));
	    if (mode==1)
	      Push(st,T_RELOCATABLE(SECTION_PARENT_SECTION(REL_E_TO_RELOCATABLE(rel)))->min_address);
	    else if (mode==2)
	      Push(st,T_RELOCATABLE(SECTION_PARENT_SECTION(REL_E_TO_RELOCATABLE(rel)))->old_address);
	    else
	      Push(st,T_RELOCATABLE(SECTION_PARENT_SECTION(REL_E_TO_RELOCATABLE(rel)))->caddress);
	    break;
	  }
	  else FATAL(("Strange relocatable in T-stack command"));

	}
      case 'r': /* ???? - push the offset of the associated relocatable */
	Push(st,rel->e_offset);
	break;

      case 't': /* push the (t)o offset of the relocatable */
	Push(st,rel->to_offset);
	break;

#ifdef BIT32ADDRSUPPORT
      case 'i': /* (i)mmediate - push a 32-bit immediate */
	{
	  t_uint32 imm = GetImm32(cmd);
	  GenericSet_uint32(tmp,imm);
	  Push(st,tmp);
	  cmd += 8;  /* skip over the immediate */
	}
	break;
#endif
#ifdef BIT64ADDRSUPPORT
      case 'I': /* (I)mmediate - push a 64-bit immediate */
	{
	  t_uint64 imm = GetImm64(cmd);
	  GenericSet_uint64(tmp,imm);
	  Push(st,tmp);
	  cmd += 16;  /* skip over the immediate */
	}
	break;
#endif
      case 's': /* (s)hort - push a 16-bit unsigned immediate */
	{
	  t_uint16 imm = GetImm16(cmd);
	  cmd += 4;  /* skip over the immediate */
	  /* just a little trick to make this work in all possible cases. 
	  * if both 32 and 64 bit support is compiled in, tmp will have type EITHER */
	  tmp = GENERIC_ADD_UINT32(NULL_GENERIC, (t_uint32)imm);
	  Push(st,tmp);
	}
	break;
      case 'Z': /* si(Z)e of destination - push the size of the reloc's destination */
	if (mode==1)
	  Push(st,REL_TO_RELOCATABLE(rel)->min_size);
	else if (mode==2)
	  Push(st,REL_TO_RELOCATABLE(rel)->old_size);
	else
	  Push(st,REL_TO_RELOCATABLE(rel)->csize);
	break;
      case 'Y': /* ???? - push the size of the associated relocatable */
	if (mode==1)
	  Push(st,rel->e.sec->min_size);
	else if (mode==2)
	  Push(st,rel->e.sec->old_size);
	else
	  Push(st,rel->e.sec->csize);
	break;


	/***** Memory operations *****/
#ifdef BIT32ADDRSUPPORT
      case 'l': /* (l)oad - load 32 bits and push */
	{
#ifdef ALPHA
	  t_uint32 data;

	  ((char *) &data)[0]= memptr[0];
	  ((char *) &data)[1]= memptr[1];
	  ((char *) &data)[2]= memptr[2];
	  ((char *) &data)[3]= memptr[3];
#else
	  t_uint32 data = *((t_uint32 *)memptr);
#endif

	  /* advance mem pointer */
	  memptr += 4;

	  GenericSet_uint32(tmp,data);
	  Push(st,tmp);
	}
	break;

#endif

#ifdef BIT64ADDRSUPPORT
      case 'L': /* (L)oad - load 64 bits and push */
	{
	  t_uint64 data = *((t_uint64 *)memptr);

	  /* advance mem pointer */
	  memptr += 8;

	  GenericSet_uint64(tmp,data);
	  Push(st,tmp);
	}
	break;

      case 'E': /* Load and (E)xtend - loads 32 bits and extends it to 64 bits and pushes */
	{
	  t_uint64 data = *((t_uint32 *)memptr);

	  /* advance mem pointer */
	  memptr += 4;

	  GenericSet_uint64(tmp,data);
	  Push(st,tmp);
	}
	break;

#define CREATE_MASK(aantal) (((t_uint64)1 << (aantal)) - 1)

      case 'f': /* (f)ucked up load - Loads the itanium instruction */
	{
	  if (currloadins==0)
	  {
	    t_uint64 data = *((t_uint64 *) bundleptr);
	    data=(data >> 5) & CREATE_MASK(41);
	    /*	    printf("Got ins 0=%llx\n",data); */
	    currloadins--;
	    GenericSet_uint64(tmp,data);
	    Push(st,tmp);
	  }
	  else if (currloadins==1)
	  {
	    t_uint64 data = *((t_uint64 *) bundleptr);
	    t_uint64 data2 = *(((t_uint64 *) bundleptr)+1);
	    data = (data >> 46) | ((data2 & CREATE_MASK(23))<<18);
	    /*	    printf("Got ins 1=%llx\n",data); */
	    currloadins--;
	    GenericSet_uint64(tmp,data);
	    Push(st,tmp);
	  }
	  else if (currloadins==2)
	  {
	    t_uint64 data =*(((t_uint64 *) bundleptr)+1);
	    data = data >>23;
	    /*	    printf("Got ins 2=%llx\n",data);*/
	    currloadins--;
	    GenericSet_uint64(tmp,data);
	    Push(st,tmp);
	  }
	  else
	    FATAL(("Illegal instruction!"));
	}
	break;
      case 'F': /* (F)ucked up store - Stores the itanium instruction */
	{
	  if (currstoreins==0)
	  {
	    t_uint64 bundle = *((t_uint64 *) bundleptr);
	    t_uint64 data = G_T_UINT64(Pop(st));
	    /*	    printf("Got ins 0=%llx\n",data);*/

	    if((data & ~CREATE_MASK(41)))
	      FATAL(("instruction data too big (0)"));

	    bundle = (bundle & (~(CREATE_MASK(41)<<5))) | (data << 5);

	    *((t_uint64 *) bundleptr) = bundle;

	    currstoreins--;
	  }
	  else if (currstoreins==1)
	  {
	    t_uint64 bundle = *((t_uint64 *) bundleptr);
	    t_uint64 bundle2 = *(((t_uint64 *) bundleptr)+1);

	    t_uint64 data = G_T_UINT64(Pop(st));
	    /*	    printf("Got ins 1=%llx\n",data); */

	    if((data & ~CREATE_MASK(41)))
	      FATAL(("instruction data too big (1)"));

	    bundle = (bundle & (CREATE_MASK(46))) | ((data & CREATE_MASK(18)) << 46);

	    bundle2 = (bundle2 & (CREATE_MASK(41)<<23)) | ((data>>18));
	    *((t_uint64 *) bundleptr) = bundle;
	    *(((t_uint64 *) bundleptr)+1) = bundle2;

	    currstoreins--;

	  }
	  else if (currstoreins==2)
	  {
	    t_uint64 bundle = *(((t_uint64 *) bundleptr)+1);
	    t_uint64 data = G_T_UINT64(Pop(st));
	    /*	    printf("Got ins 2=%llx\n",data); */

	    if((data & ~CREATE_MASK(41)))
	      FATAL(("instruction data too big (2)"));

	    bundle = (bundle & (CREATE_MASK(23))) | (data << 23);

	    *(((t_uint64 *) bundleptr)+1) = bundle;

	    currstoreins--;

	  }
	  else
	    FATAL(("Illegal instruction!"));
	}
	break;

#endif
      case 'g': /*"Push starting address of got table: value of (g)p"*/
	{
	  t_object *obj = ObjectGet(global_options.objectfilename,NULL);
	  t_section *sec = SectionGetFromObjectByName(obj,".got");
	  if(!sec)
	    FATAL(("oopsie daisy: no .got table section found!"));

	  if (mode==1)
	    Push(st,SECTION_GET_MIN_ADDRESS(sec));
	  else if (mode==2)
	    Push(st,SECTION_GET_OLD_ADDRESS(sec));
	  else
	    Push(st,SECTION_GET_CADDRESS(sec));
	
	      
	  break;
	}
		

      case 'w': /* (w)rite - pop and store top of stack as 32 bits. Mind that this can also be used to store a 64 bit generic as 32 bit */
	{
	  t_uint32 data = GENERIC_EXTRACT_UINT32(Pop(st));

	  /* first adjust memptr */
	  memptr -= 4;
#ifdef ALPHA
	  memptr[0]=((char *) &data)[0];
	  memptr[1]=((char *) &data)[1];
	  memptr[2]=((char *) &data)[2];
	  memptr[3]=((char *) &data)[3];
#else
	  *((t_uint32 *)memptr) = data;
#endif
	}
	break;

#ifdef BIT64ADDRSUPPORT
      case 'W': /* (W)rite - pop and store 64 bits */
	{
	  t_uint64 data = G_T_UINT64(Pop(st));

	  /* first adjust memptr */
	  memptr -= 8;

	  *((t_uint64 *)memptr) = data;
	}
	break;
#endif

#ifdef BIT32ADDRSUPPORT
      case 'k': /* push si(k)steen to 32 - load 16 bits unsigned and push in 32-bit generic */
	{
	  t_uint16 data = *((t_uint16 *)memptr);

	  /* advance mem pointer */
	  memptr += 2;

	  GenericSet_uint32(tmp,(t_uint32)data);
	  Push(st,tmp);
	}
	break;
#endif

#ifdef BIT64ADDRSUPPORT
      case 'K': /* push si(K)steen to 64 - load 16 bits unsigned and push in 64-bit generic */
	{
	  t_uint16 data = *((t_uint16 *)memptr);

	  /* advance mem pointer */
	  memptr += 2;

	  GenericSet_uint64(tmp,(t_uint64)data);
	  Push(st,tmp);
	}
	break;
#endif
      case 'v': /* ???? - pop and write 16 bits */
	{
	  t_uint32 data = GENERIC_EXTRACT_UINT32(Pop(st));

	  /* first adjust memptr */
	  memptr -= 2;

	  *((t_uint16 *)memptr) = (t_uint16) (data & 0xffff);
	}
	break;


	/********   operators   ********/
      case '+': /* pop the two top elements off the stack, add them and push the result */
	{
	  t_address oper1, oper2;
	  oper2 = Pop(st);
	  oper1 = Pop(st);
	  Push(st,GENERIC_ADD(oper1,oper2));
	}
	break;
      case '?':/* pop the top element and skip until after ':', or '!' if zero */
	{
	  t_address oper1;
	  oper1 = Pop(st);
	  if(GENERIC_EQ(NULL_GENERIC,oper1))
	  {
	    while(*cmd++)
	    {
	      if(*(cmd-1) == ':' || *(cmd-1) == '!') break;
	    }
	    if(!*cmd) FATAL(("Illegal if-then-else!"));
	  }
	}
	break;
      case ':':/* else part of if-then-else, if we reach it, skip until after '!' */
	{
	  while(*cmd++)
	    if(*(cmd-1) == '!') break;
	  if(!*cmd) FATAL(("Illegal if-then-else!"));
	}
	break;
      case '!':/* endif part of an if-then-else. Needed to know where it ends */
	break;
      case '-':/* pop the two top elements off the stack, sub them and push the result */
	{
	  t_address oper1, oper2;
	  oper2 = Pop(st);
	  oper1 = Pop(st);
	  Push(st,GENERIC_SUB(oper1,oper2));
	}
	break;
      case '&':/* pop the two top elements off the stack, and them (bitwise) and push the result */
	{
	  t_address oper1, oper2;
	  oper2 = Pop(st);
	  oper1 = Pop(st);
	  Push(st,GENERIC_AND(oper1,oper2));
	}
	break;
      case '|':/* pop the two top elements off the stack, or them (bitwise) and push the result */
	{
	  t_address oper1, oper2;
	  oper2 = Pop(st);
	  oper1 = Pop(st);
	  Push(st,GENERIC_OR(oper1,oper2));
	}
	break;
      case '^':/* pop the two top elements off the stack, xor them (bitwise) and push the result */
	{
	  t_address oper1, oper2;
	  oper2 = Pop(st);
	  oper1 = Pop(st);
	  Push(st,GENERIC_XOR(oper1,oper2));
	}
	break;
      case '~':/* pop the top element off the stack, negate it (bitwise) and push the result */
	{
	  t_address oper1;
	  oper1 = Pop(st);
	  Push(st,GENERIC_NOT(oper1));
	}
	break;
      case '<':/* pop the two top elements off the stack, left shift the first by the last and push the result */
	/* shift left */
	{
	  t_address oper1, oper2;
	  oper2 = Pop(st);
	  oper1 = Pop(st);
	  Push(st,GENERIC_SHL(oper1,oper2));
	}
	break;
      case '>':/* pop the two top elements off the stack, right shift the first by the last and push the result */
	/* shift right */
	{
	  t_address oper1, oper2;
	  oper2 = Pop(st);
	  oper1 = Pop(st);
	  Push(st,GENERIC_SHR(oper1,oper2));
	}
	break;
      case '#': /* sign extend */
	/* the word size to which the value is extended is 
	* determined by the word size of the t_address on 
	* the stack. */
	{
	  t_uint32 signbit = GetImm8(cmd);
	  oper = Pop(st);
	  cmd += 2;  /* skip over immediate operand */
	  oper = GENERIC_SIGN_EXTEND(oper,signbit);
	  Push(st,oper);
	}
	break;
      case '@': /* pop and add to memptr */
	{
	  t_int32 offset = GENERIC_EXTRACT_INT32(Pop(st));
	  memptr += offset;
	}
	break;

	/******** side effects  ********/
	/* empty */

	/****** stack manipulation *****/
      case '=': /* duplicate the top entry */
	{
	  tmp = Pop(st);
	  Push(st,tmp);
	  Push(st,tmp);
	}
	break;
      case '%': /* swap the top elements of the stack */
	{
	  t_address oper1,oper2;
	  oper2 = Pop(st);
	  oper1 = Pop(st);
	  Push(st,oper2);
	  Push(st,oper1);
	}
	break;
      case '{': /* rotate left: put top of stack at bottom of stack */
	Rotl(st);
	break;
      case '}': /* rotate right: put bottom of stack at top of stack */
	Rotr(st);
	break;
      case '*': /* delete the top element of the stack */
	Pop(st);
	break;
      case '\\':
	if (phase == CALC)
	{
	  if (write)
	  {
	    /* Do nothing */
	  }
	  else
	  {
	    ret = Pop(st);
	    VERIFY_FASSERT(st->top == -1, ("End of program but stack not empty!"));
	    StackFini(st);
	    return ret;
	  }
	  phase = WRITE;
	}
	else if (phase == WRITE)
	{
	  phase = VERIFY;
	}
	else
	{
	  FATAL(("malformed program %s",program));
	}
	break;
      case '$': /* end of input: pop the stack and return */
	ret = Pop(st);
	VERIFY_FASSERT(st->top == -1, ("End of program but stack not empty (%d elts left)!\nprogram = %s\nrel = @R",st->top+1,program,rel));
	StackFini(st);
	return ret;
	break;
      case 'G': /* (G)lobal Offset Entry - pseudo opcode, gets expanded to the got entry of a symbol */
	FATAL(("Encountered pseudo opcode G while executing %s. This opcode refers to the got entry of the symbol, and should have been converted to a real relocatable.",program));
	
      case 'u': /* ltoff22 - pseudo opcode, gets expanded */
	FATAL(("Encountered pseudo opcode l while executing %s. This opcode should have been converted to a real relocatable.",program));
	
      case 'p': /* f(p)tr64lsb - pseudo opcode, gets expanded */
	FATAL(("Encountered pseudo opcode p while executing %s. This opcode should have been converted to a real relocatable.",program));
      case 'q': /* (q)uality relocation: LTOFF_FPTR22 - pseudo opcode, gets expanded */
		FATAL(("Encountered pseudo opcode q while executing %s. This opcode should have been converted to a real relocatable.",program));
		
	
#ifdef BIT32ADDRSUPPORT      	
      case 'd': /* (d)ebug: print current item on stack */
	{
	  t_address item;
	  item = Pop(st);
	  Push(st, item);
	  printf("Current item on stack: %x\n", G_T_UINT32(item));
	}
	break;
#endif	
	
#ifdef BIT64ADDRSUPPORT      	
      case 'D': /* (D)ebug: print current item on stack */
	{
	  t_address item;
	  item = Pop(st);
	  Push(st, item);
	  printf("Current item on stack: %llx\n", G_T_UINT64(item));
	}
	break;
#endif	
      default:
	FATAL(("Unknown stack machine opcode %c (in %s)", *(cmd-1),program));

    }
  }
  /* shouldn't get here */
  FATAL(("Improperly terminated program:\n%s",program));
  /* keep the compiler happy */
  StackFini(st);
  return NULL_GENERIC;
}
/* }}} */

/* vim: set shiftwidth=2 foldmethod=marker: */
