/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * DESC: This file contains all functions for symbol 
 * handling 
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/

/*{{{ Includes*/
#include "../simple_types.h"

#include <string.h>
/* Utility includes */
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"

/* Used datatstructures */
#include "../datastructures/data_hash.h"
#include "../datastructures/data_string.h"

/* Kernel includes */
#include "diablo_object.h"
#include "diablo_parsemap.h"
#include "diablo_symbol.h"
#include "diablo_section.h"
#include "diablo_bbl.h"
#include "diablo_print.h"
#include "diablo_iterators.h"
/*}}}*/

/* Utility functions {{{ */ 

/*add a skeleton symbol, to be filled in by the other functions */

static t_symbol * SymTabAddSymbol(char * file, int lnno, t_symboltable * symbol_table, t_string name, t_bool * duplicate, t_bool defined,t_bool prepend) 
{
  t_symbol_he * he=(t_symbol_he *) HashLookup(symbol_table->hash_table, name);

 if ((!he)||(*duplicate))
  {
    t_symbol * sym=Calloc(1,sizeof(t_symbol));
    t_symbol_he * hash_elem=Malloc(sizeof(t_symbol_he));
    sym->offset_from_start=INVALID_GENERIC;
    sym->current_address=INVALID_GENERIC;
    sym->addend = NULL_GENERIC;
    sym->name=StringDup(name);
    sym->next=NULL;
    if (!symbol_table->first) 
    {
      symbol_table->first=sym;
      symbol_table->last=sym;
      sym->next=sym->prev=NULL;
    }
    else
    {
      symbol_table->last->next=sym;
      sym->prev=symbol_table->last;
      sym->next=NULL;
      symbol_table->last=sym;
    }
    symbol_table->nsyms++;

    hash_elem->sym=sym;
    hash_elem->node.key=StringDup(name);
    if (prepend)
    {
      HashPrepend(symbol_table->hash_table, (void*) hash_elem);
    }
    else
      HashInsert(symbol_table->hash_table, (void*) hash_elem);

    *duplicate=FALSE;
    return sym;
  }
  else
  {
    /* If the found duplicate is an UNDEFINED_SYMBOL or a WEAK symbol and we are adding a defined symbol, we can always overwrite it!
     * Just free the name to avoid leaks*/

    if (defined && ((he->sym->type==WEAK_UNDEFINED_SYMBOL)||(he->sym->type==UNDEFINED_SYMBOL)||(he->sym->type==WEAK_SYMBOL)))
    {
      *duplicate=FALSE;
    }
    else { 
      *duplicate=TRUE;
    }
    return he->sym; 
  }
}


static void SymOverwrite(t_symbol * dest, t_symbol * orig)
{
  dest->to_type = orig->to_type;
  dest->type = orig->type;
  dest->base = orig->base;
  dest->offset_from_start = orig->offset_from_start;
  dest->current_address = orig->current_address;
  dest->local = orig->local;
}

/* }}} */


t_symbol *SymbolTableAddSymbol(t_symboltable * symbol_table, t_uint32 to_type, t_uint32 type,  t_relocatable * sec,  t_address offset, t_string name,t_bool local)
{
  t_bool dup=local;
  t_bool undefined=(type==WEAK_UNDEFINED_SYMBOL)||(type==UNDEFINED_SYMBOL);
  t_symbol * skeleton=SymTabAddSymbol(__FILE__,__LINE__,symbol_table,name,&dup,!undefined,FALSE);

  VERIFY_FASSERT((!undefined)||(!local),("Undefined local symbol!"));
  
  if (dup) /* We have a duplicate symbol! */
  {
    /* In case of a local symbol, we reinsert the local symbol (in the hashtable), 
     * and change the local symbol to a global symbol, so the global symbol is found 
     * when we query the symbol table */
    if (skeleton->local || local) 
    {
      t_bool dup2=TRUE;
      if (local)
      {
	t_symbol * local_sym=SymTabAddSymbol(__FILE__,__LINE__,symbol_table,name,&dup2,TRUE,FALSE);
	/* Overwrite the newly inserted symbol with the found symbol */
	SymOverwrite(local_sym,skeleton);
      }
      else
      {
	skeleton=SymTabAddSymbol(__FILE__,__LINE__,symbol_table,name,&dup2,TRUE,TRUE);
      }
    }
    else
    {
      if (undefined || (type == WEAK_UNDEFINED_SYMBOL) || (type == WEAK_SYMBOL) || 
	  ((skeleton->type==type) 
	   && (skeleton->to_type==to_type) 
	   && (skeleton->local==local) && (skeleton->base==sec) 
	   && ((!sec) || (GENERIC_EQ(skeleton->offset_from_start,GENERIC_SUB(offset,SECTION_GET_CADDRESS(T_SECTION(sec))))))
	   && (GENERIC_EQ(skeleton->current_address,offset))
	  )
	 )

      {
	return skeleton;
      }
      else
      {
	DiabloPrint(stdout,"%s %s\n", T_SECTION(sec)->obj->name, T_SECTION(skeleton->base)->obj->name);
	FATAL(("Duplicate symbol %s (this type=%d, to_type=%d), prev type=%d,to_type=%d @G <> @G @G <> @G",name,type,to_type,skeleton->type,skeleton->to_type, skeleton->offset_from_start,GENERIC_SUB(offset,SECTION_GET_CADDRESS(T_SECTION(sec))),offset, skeleton->current_address));
      }
      return NULL; /* keeps the compiler happy */
    }
  }


  skeleton->type=type;
  skeleton->to_type=to_type;
  skeleton->local=local;
  skeleton->base=sec;
  skeleton->offset_from_start=sec?GENERIC_SUB(offset,SECTION_GET_CADDRESS(T_SECTION(sec))):offset;
  skeleton->current_address=offset;
  skeleton->mapped=NULL;
  return skeleton;
}



t_symbol * SymbolTableAddSymbolToSection(t_symboltable * symbol_table,  t_uint32 type,  t_relocatable * sec,  t_address offset, t_string name,t_bool local)
{
  return SymbolTableAddSymbol(symbol_table, TO_SECTION, type,  sec,  offset, name,local);
}

t_symbol * SymbolTableAddSymbolToSubSection(t_symboltable * symbol_table,  t_uint32 type,  t_relocatable * sec,  t_address offset, t_string name,t_bool local)
{
  return SymbolTableAddSymbol(symbol_table, TO_SUB_SECTION, type,  sec,  offset, name,local);
}

t_symbol * SymbolTableAddSymbolToCommon(t_symboltable * symbol_table,  t_uint32 type,  t_address align, t_address size, t_string name,t_bool local)
{
  /* for COMMON symbols, the alignment is stored in sym->current_address, the size in sym->size */
  t_symbol * sym = SymbolTableAddSymbol(symbol_table, TO_COMMON, type,  NULL,  align, name,local);
  sym->size = size;
  return sym;
}

t_symbol * SymbolTableAddWeakUndefinedSymbol(t_symboltable * symbol_table,  t_string name)
{
  return SymbolTableAddSymbol(symbol_table, TO_NONE, WEAK_UNDEFINED_SYMBOL, NULL, NULL_GENERIC, name,FALSE);
}

t_symbol * SymbolTableAddUndefinedSymbol(t_symboltable * symbol_table,   t_string name)
{
  return SymbolTableAddSymbol(symbol_table, TO_UNDEFINED, UNDEFINED_SYMBOL, NULL, NULL_GENERIC, name,FALSE);
}
t_symbol * SymbolTableAddPltSymbol(t_symboltable * symbol_table,   t_string name, t_address pltaddress, t_address size)
{
  return SymbolTableAddSymbol(symbol_table, TO_UNDEFINED, PLT_SYMBOL, NULL, pltaddress, name,FALSE);
}
t_symbol * SymbolTableAddAbsoluteSymbol(t_symboltable * symbol_table,  t_address offset, t_string name,t_bool local)
{
  return SymbolTableAddSymbol(symbol_table, TO_NONE, ABS_SYMBOL, NULL, offset, name,local);
}

/* Free symbol table {{{ */
void SymbolTableFree(t_symboltable * st)
{
  t_symbol * sym;
  t_symbol * sym_next;
  HashFree(st->hash_table);
  for (sym=st->first ;sym!=NULL; sym=sym_next)
    {
      sym_next=sym->next;
      if (sym->name) 
      {
	Free(sym->name);
      }
      Free(sym);
    }
  Free(st);
}
/* }}} */

/* SymbolTableHEFree: free's a hash element */

void SymbolTableHEFree(void * f, void * data)
{
  t_symbol_he * tf=((t_symbol_he *) f);
  Free(tf->node.key);
  Free(tf);
}

/* Allocate a new symbol table {{{*/
t_symboltable * SymbolTableNew(t_object * obj)
{
  t_symboltable * ret=Calloc(sizeof(t_symboltable),1);
  ret->obj=obj;
  ret->hash_table= HashTableNew(131,0,string_hash, string_cmp, SymbolTableHEFree);
  return ret;
}
/*}}} */

/* FUN : SymbolTableGetDataType
 * PAR : The object to search and the address 
 * RET : ADDRESS_IS_CODE or ADDRESS_IS_DATA
 * DESC: this function returns the datatype for the given address */
#define CACHED_TYPES
#ifdef CACHED_TYPES
t_symbol_he * symbol_type_cached=NULL;
#endif
t_uint32 SymbolTableGetDataType(t_object * obj, t_address adr)
{
#ifdef CACHED_TYPES
  static t_address cacheaddr;
  static t_bool cache_type;
#endif
  t_symbol_he * top;
  t_bool code=TRUE;
  t_symbol_he * first;


#ifdef CACHED_TYPES
  if ((symbol_type_cached) &&(GENERIC_GE(adr,cacheaddr)))
  {
    top=symbol_type_cached;
    code=cache_type;
  }
  else
  {
    top=( t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$switch");
    first=( t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$first");
    if ((first) && (!GENERIC_EXTRACT_UINT32(first->sym->current_address))) code=FALSE;
    if (!top) return ADDRESS_IS_CODE;
  }
#endif

  while(top->node.equal)
  {
    if (GENERIC_GT(((t_symbol_he *) top->node.equal)->sym->current_address, adr)) break;
    code=!code;
    top=( t_symbol_he *) top->node.equal;
  }

#ifdef CACHED_TYPES
  symbol_type_cached=top;
  cache_type=code;
  cacheaddr=adr;
#endif

  if (!code)
    return ADDRESS_IS_DATA; 
  else
    return ADDRESS_IS_CODE; 
}

/* Given an address, tell if it is in a thumb or an arm piece of code */
t_uint32 SymbolTableGetCodeType(t_object * obj, t_address addr)
{
  t_symbol_he * top = NULL;
  t_bool arm=TRUE;
  t_symbol_he * first = NULL;

  if(GENERIC_LT(GENERIC_ADD(SECTION_GET_CADDRESS(obj->code[0]),SECTION_CSIZE(obj->code[0])),addr))
    return ADDRESS_IS_CODE;
  top=( t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$code_switch");
  first=( t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table, "$thumb");
  if ((first) && (!GENERIC_EXTRACT_UINT32(first->sym->current_address))) arm=FALSE;
  if (!top) return ADDRESS_IS_CODE;

  while(top->node.equal)
  {
    if (GENERIC_GT(((t_symbol_he *) top->node.equal)->sym->current_address, addr)) break;
    arm=!arm;
    top=( t_symbol_he *) top->node.equal;
  }

  if (!arm)
    return ADDRESS_IS_SMALL_CODE; 
  else
    return ADDRESS_IS_CODE; 
}

/* Query the symbol table {{{ */
t_symbol * SymbolTableGetSymbolByName(t_symboltable * st, t_string name)
{
  t_symbol_he * he=(t_symbol_he *) HashLookup(st->hash_table, name);
  if (he==NULL) return NULL;


  if ((he->node.equal)&&(!(((t_symbol_he* ) he->node.equal)->sym->local)))
  {  
#ifdef DEBUG_SYMBOL
    printf("%d (%d)\n",he->sym->type, he->sym->local);
#endif
    if (he->sym->to_type == 2) printf("to %s : %s\n",T_SECTION(he->sym->base)->obj->name, T_SECTION(he->sym->base)->name);
#ifdef DEBUG_SYMBOL
    printf("%d (%d)\n",((t_symbol_he* ) he->node.equal)->sym->type, ((t_symbol_he* ) he->node.equal)->sym->local);
#endif
    if (((t_symbol_he* ) he->node.equal)->sym->to_type == 2) printf("to %s : %s\n",T_SECTION(((t_symbol_he* ) he->node.equal)->sym->base)->obj->name, T_SECTION(((t_symbol_he* ) he->node.equal)->sym->base)->name);
    
    FATAL(("Symbol %s has multiple definitions! Implement!",name));
    }

  return he->sym;
}

t_symbol * SymbolTableGetFirstSymbolWithName(t_symboltable * st, t_string name)
{
  t_symbol_he * he=(t_symbol_he *) HashLookup(st->hash_table, name);
  if (he==NULL) return NULL;
  return he->sym;
}

/* return symbol at given address.
 * as this function is only used to determine function names, we try to 
 * find the symbol at that address that is most likely to contain the function name */
t_symbol * SymbolTableGetSymbolByAddress(t_symboltable * st, t_address addr)
{
  t_symbol * sym, *tmp = NULL;
  
  SYMBOLTABLE_FOREACH_SYMBOL(st,sym)
  {
    if (!sym->base) continue;
    if (sym->type==ORIGINAL_SECTION_SYMBOL) continue;
    if ((sym->name) && ((strcmp(sym->name,"$a")==0) || (strcmp(sym->name,"$d")==0))) continue;

    if (GENERIC_EQ(GENERIC_ADD(sym->base->caddress,sym->offset_from_start),addr))
    {
      if (sym->type == FUNCTION_SYMBOL)
	return sym;
      else
      {
	if (!sym->local)
	  tmp = sym;
	else
	  if (!tmp) tmp = sym;
      }
    }
#if 0
    else 
      {
	if (GENERIC_EQ(GENERIC_ADD_UINT32(GENERIC_ADD(sym->base->caddress,sym->offset_from_start),sym->prolog),addr))
	  {
	    if (!sym->local)
	      {
		printf("joepi\n");
		tmp = sym;
	  }
	    else
	      if (!tmp) tmp = sym;
	  }
      }
#endif
  }

  return tmp;
}
/* }}} */

/* Print the symbol table {{{ */
void SymbolTablePrint(t_symboltable * st)
{
  t_symbol * sym;
  SYMBOLTABLE_FOREACH_SYMBOL(st,sym)
  {
    DiabloPrint(stdout,"@S\n",sym);
  }
}
/* }}} */

/* vim: set shiftwidth=2 foldmethod=marker: */
