/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/

#include <stdlib.h>
#include "diablo_edge.h"
#include "diablo_bbl.h"
#include "diablo_function.h"
#include "diablo_print.h"
#include "diablo_udchain.h"
#include "diablo_udgraph.h"
#include "diablo_iterators.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../datastructures/data_ll.h"

static void FunctionUseDefAnalysis(t_object* obj, t_function* fun, t_section *sec, t_cfg_edge* edge_in, t_bool during_fixpoint_calculations, t_reg propreg);
static void BblUseDefAnalysis(t_object * obj, t_bbl * bbl, t_cfg_edge * edge_out, t_bool during_fixpoint_calculations, t_reg propreg);
static void UseDefBuildGraph(t_object * obj, t_section * sec, t_reg propreg);
static void UseDefInstructionPropagator(t_ins * ins, t_ud_chain ** udc_p, t_reg propreg);
static t_bool FunctionUnmarkBbl(t_function * fun, t_bbl** bbl);
static t_bool FunUnmarkEdge(t_function * fun,t_cfg_edge** edge);

/* static variables */
/* The ud_chain analysis is a forward dataflow analysis, which means that propagation
 * should start with all registers set to BOT. But in this case BOT means that
 * the possible definitions span _all_ instructions in the program, which is both 
 * pointless and impossible to represent. Therefore we use a shortcut: we introduce 
 * a dummy instruction (declared here) that stands outside the flowgraph and that is
 * used as the sole definer of all registers at the beginning of the program.
 */

static t_ins dummy_ins;
static t_bbl dummy_bbl;

/* {{{ helper functions */
/* a first group of utility procedures is used to created linked lists
   of procedures, incoming edges into procedures and basic blocks in
   procedures that need re-evaluation 
 */
static t_bool CfgUnmarkFun(t_cfg* cfg, t_function** fun)
{

  *fun = FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg));

  if (*fun==CFG_MARKED_FUNS(cfg))
    return FALSE;

  FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg)) = FUN_NEXT_MARKED_IN_CFG(*fun);
  FUN_NEXT_MARKED_IN_CFG(*fun) = NULL;

  return TRUE;
}

static void CfgUnmarkAllFun(t_cfg * cfg)
{
  t_function * fun;
  if (!(CFG_MARKED_FUNS(cfg)))
  {
    CFG_MARKED_FUNS(cfg) = T_FUN(Malloc(sizeof(t_function)));
    FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg))=CFG_MARKED_FUNS(cfg);
  }
  else
    while (CfgUnmarkFun(cfg,&fun)) ;
}

static void CfgMarkFun(t_cfg* cfg, t_function* fun)
{
  if (!(FUN_NEXT_MARKED_IN_CFG(fun)))
  {
    FUN_NEXT_MARKED_IN_CFG(fun) = FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg));
    FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg)) = fun;
  }
}

static t_bool FunctionUnmarkBbl(t_function * fun, t_bbl** bbl)
{
  *bbl = BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(fun));

  if (*bbl==FUN_MARKED_BBLS(fun))
    return FALSE;

  BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(fun)) = BBL_NEXT_MARKED_IN_FUN(*bbl);
  BBL_NEXT_MARKED_IN_FUN(*bbl) = NULL;

  return TRUE;
}

static void FunctionUnmarkAllBbls(t_function * fun)
{
  t_bbl * bbl;
  if (!(FUN_MARKED_BBLS(fun)))
  {
    FUN_MARKED_BBLS(fun) = T_BBL(Malloc(sizeof(t_bbl)));
    BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(fun))=FUN_MARKED_BBLS(fun);
  }
  else
    while (FunctionUnmarkBbl(fun,&bbl)) ;
}


static void FunctionMarkBbl(t_function * fun, t_bbl * bbl)
{
  if (!(BBL_NEXT_MARKED_IN_FUN(bbl)))
  {
    BBL_NEXT_MARKED_IN_FUN(bbl) = BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(fun));
    BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(fun)) = bbl;
  }
}

static t_bool FunUnmarkEdge(t_function * fun,t_cfg_edge** edge)
{
  *edge = EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun));

  if (*edge==FUN_MARKED_EDGES(fun))
    return FALSE;

  EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun)) = EDGE_NEXT_MARKED_INTO_FUN(*edge);
  EDGE_NEXT_MARKED_INTO_FUN(*edge) = NULL;

  return TRUE;
}

static void FunctionUnmarkAllEdges(t_function * fun)
{
  t_cfg_edge * edge;
  if (!(FUN_MARKED_EDGES(fun)))
  {
    FUN_MARKED_EDGES(fun) = T_CFG_EDGE(Malloc(sizeof(t_cfg_edge)));
    EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun))=FUN_MARKED_EDGES(fun);
  }
  else
    while (FunUnmarkEdge(fun,&edge)) ;
}

static void FunctionMarkEdge(t_function * fun, t_cfg_edge * edge)
{
  if (!(EDGE_NEXT_MARKED_INTO_FUN(edge)))
  {
    EDGE_NEXT_MARKED_INTO_FUN(edge) = EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun));
    EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun)) = edge;
  }
}

static void FunctionMarkAllTrueIncomingEdges(t_function * fun)
{
  t_bbl * bbl;
  t_cfg_edge * edge;

  FUN_FOREACH_BBL(fun,bbl)
  {
    BBL_FOREACH_PRED_EDGE(bbl,edge)
    {
      if (CfgEdgeIsForwardInterproc(edge))
      {
        if (EdgeIsMarked(edge))
        {
          FunctionMarkEdge(fun,edge);
        }
      }
    }
  }
}
/* }}} */

/* {{{ use-def analysis */
/*
   \param sec 

   Perform use-definition analysis on the flowgraph of section sec.
   After all the chains have been built up, the information will be
   transferred to a graph (the ud graph) whose nodes are the
   instructions of the control flow graph. A pointer to this
   ud graph is stored within the control flow graph of the section.

   WATCH OUT: we expect to have the Saved and Changed registers set up 
   correctly when this function is called!!!
 */

void UseDefAnalysis(t_section * sec)
{
  /* iterators */
  t_function * i_fun;
  t_cfg_edge * i_edge;
  t_bbl * i_bbl;
  t_reg i_reg;


  /* local variables */

  t_bbl * program_entry;
  t_function * entry_fun;
  t_cfg *cfg = T_CFG(sec->data);
  t_object * obj = sec->obj;
  t_regset allregs = obj->description->all_registers;


  /* add a pseudo instruction to the hell nodes -> speeds up the analysis */
  if (cfg->hell_node->nins == 0)
  {
    t_ins * dummy = InsNewForBbl(cfg->hell_node);

    /*DiabloPrint(stdout,"Inserting dummy ins in hell node\n");
    DiabloPrint(stdout,"Live regs out: "); */
    PrintRegset(stdout,obj->description,BBL_REGS_LIVE_OUT(cfg->hell_node));
    InsAppendToBbl(dummy, cfg->hell_node);
    obj->actions->MakeNoop(dummy);
    /* set all registers used and defined */
    RegsetDup(INS_REGS_USE(dummy),obj->description->all_registers);
    RegsetDup(INS_REGS_DEF(dummy),obj->description->all_registers);
  }
  if (cfg->call_hell_node->nins == 0)
  {
    t_ins * dummy = InsNewForBbl(cfg->call_hell_node);

    /*DiabloPrint(stdout,"Inserting dummy ins in call hell node\n");
    DiabloPrint(stdout,"Live regs out: "); */
    PrintRegset(stdout,obj->description,BBL_REGS_LIVE_OUT(cfg->call_hell_node));
    InsAppendToBbl(dummy, cfg->call_hell_node);
    obj->actions->MakeNoop(dummy);
    /* set all registers used and defined */
    RegsetDup(INS_REGS_USE(dummy),obj->description->all_registers);
    RegsetDup(INS_REGS_DEF(dummy),obj->description->all_registers);
  }

  /* find the defined registers for each bbl */
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    i_bbl->tmp = UdChainNew(i_bbl);
    BblSetDefinedRegs(i_bbl);
  }

  /* initialize the dummy bbl and the dummy ins */
  RegsetDup(BBL_REGS_DEF_CERTAIN(&dummy_bbl),obj->description->all_registers);
  RegsetDup(BBL_REGS_DEF_PERHAPS(&dummy_bbl),RegsetNew());
  RegsetDup(INS_REGS_DEF(&dummy_ins),obj->description->all_registers);
  BBL_INS_FIRST(&dummy_bbl) = BBL_INS_LAST(&dummy_bbl) = &dummy_ins;
  dummy_bbl.cfg = cfg;

  /* All linked lists of to be evaluated items should be emptied. */
  InitEdgeMarking();

  CfgUnmarkAllFun(cfg);

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    FunctionUnmarkAllEdges(i_fun);
    FunctionUnmarkAllBbls(i_fun);
  }

  /* 
     Allocate space for ud chains on the interprocedural edges.
     For each edge, specify which registers should be propagated over
     the edge.
   */
  CFG_FOREACH_EDGE(cfg,i_edge)
  {
    EdgeUnmark(i_edge);

    if (CfgEdgeIsInterproc(i_edge))
      i_edge->ud_chain = NULL;

    RegsetDup(i_edge->prop_regs,obj->description->all_registers);

    if (1)
    {

      if (EDGE_CAT(i_edge)==ET_RETURN)
      {
        RegsetSetIntersect(i_edge->prop_regs,FUN_REGS_CHANGED(BBL_FUN(EDGE_HEAD(i_edge))));
      }

      if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK)
      {

        t_cfg_edge * j_edge;
        BBL_FOREACH_SUCC_EDGE(EDGE_HEAD(i_edge),j_edge)
          if (EDGE_CAT(j_edge)==ET_CALL) break;

        if (j_edge)
          RegsetSetDiff(i_edge->prop_regs,FUN_REGS_CHANGED(BBL_FUN(EDGE_TAIL(j_edge))));
        else
          FATAL(("Found a funlink edge in the graph without an associated call!"));

      }
    }
    else
    {
      
      /* emergency code: enable this in case you want to check if something is
       * wrong with the code that selects which registers are propagated over
       * return and funlink edges. */

      if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK)
        i_edge->prop_regs = RegsetNew();
    }
  }


  /* Set up an empty ud-graph and attach it to the cfg */
  UdGraphNewForCfg(cfg);

  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);

  /* We do the analysis for each register in turn */
  REGSET_FOREACH_REG(allregs,i_reg)
  {
    /* 
       Use-def analysis is started at the entry point of the
       program. This is not done through incoming edges, therefore the
       edge is null. As we want to propagate the outgoing information
       over outgoing edges, the last parameter is TRUE.
     */

    FunctionUseDefAnalysis(obj,entry_fun,sec,NULL,TRUE,i_reg);

    BBL_FOREACH_PRED_EDGE(program_entry,i_edge)
    {
      /* propagation at the beginning of the program should start
       * with BOT. This is pointless here, so another trick is used.
       * For explanation: see the beginning of this file */
      i_edge->ud_chain = UdChainNew(&dummy_bbl);
      EdgeMark(i_edge);
    }

    /* 
       Actual interprocedural fixpoint stuff: as long as there are
       marked procedures, propagate information over marked (incoming)
       edges through them 
     */

    while (CfgUnmarkFun(cfg,&i_fun))
    {

      while (FunUnmarkEdge(i_fun,&i_edge))
      {
        FunctionUseDefAnalysis(obj,i_fun, sec, i_edge,TRUE,i_reg);
      }
    }

    /*
       Given the ud-chains on the interprocedural edges,
       propagate this information into the functions and 
       generate the ud-graph
     */

    UseDefBuildGraph(obj,sec,i_reg);

    /* TODO: free ud-chains on interprocedural edges */
    CFG_FOREACH_EDGE(cfg,i_edge)
    {
      if (CfgEdgeIsInterproc(i_edge))
      {
        UdChainFree(i_edge->ud_chain);
        i_edge->ud_chain = NULL;
      }
    }

  } /* end of REGSET_FOREACH_REG */

  /* free the temporary ud chains in the basic blocks */
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    UdChainFree(i_bbl->tmp);
  }
}
/* }}} */

/* {{{  Propagate information through a procedure 

PARAMETERS: obj, fun and sec are obvious.

edge_in is NULL is this propagation is not for an
incoming edge, but for the program entry point 

during_fixpoint_calculations is false if no outgoing
information has to be propagated over interprocedural
edges

propreg is the register for which we are propagating

 */
static void FunctionUseDefAnalysis(t_object * obj,
    t_function * fun, 
    t_section *sec, 
    t_cfg_edge* edge_in,
    t_bool during_fixpoint_calculations,
    t_reg propreg)
{
  /* iterators */

  t_cfg_edge * i_edge;
  t_bbl * i_bbl;

  /* local variables */

  t_cfg_edge * edge_out = NULL;

  /* 
     Allocate init propagation inside fun 
   */

  FUN_FOREACH_BBL(fun,i_bbl)
  {
    i_bbl->ud_chain_in=NULL;
  }

  /* 
     propagate the necessary information into the procedure 
   */

  if (edge_in)
  {

    /* 
       since edge_in is not NULL, we are not propagating for the entry procedure 
     */

    if (!during_fixpoint_calculations)
    {

      /* 
         this means we have to propagate information over all incoming edges into the procedure 
       */

      FUN_FOREACH_BBL(fun,i_bbl)
        BBL_FOREACH_PRED_EDGE(i_bbl, i_edge) 
        {
          if (CfgEdgeIsInterproc(i_edge))
            if (EdgeIsMarked(i_edge))
            {
              if (RegsetIn(i_edge->prop_regs,propreg))
                if (UdChainJoin(&(i_bbl->ud_chain_in),i_edge->ud_chain))
                  FunctionMarkBbl(fun,i_bbl);
            }
        }
    }
    else
    {
      /* 
         Propagation will start at the tail of the edge, where
         the incoming information is used (unlike constant propagation,
         we don't treat hell nodes any different from normal nodes)
       */

      /*TODO: check if this is correct! as it stands, hell nodes will not be
       * treated any different from normal nodes... */

      if (RegsetIn(edge_in->prop_regs,propreg))
      {
        UdChainJoin(&(CFG_EDGE_TAIL(edge_in)->ud_chain_in),edge_in->ud_chain);
        FunctionMarkBbl(fun,CFG_EDGE_TAIL(edge_in));
      }
      /*
         For return or compensating outgoing edges, we only have to propagate over the corresponding edge.
       */

      edge_out = edge_in->corr_edge;
    }
  }
  else
  {
    /*
       We are working on the entry fun.
     */

    if (!during_fixpoint_calculations)
    {
      /*
         if there are more other incoming edges, propagate the information only if we are not in the fixpoint calculations
       */

      FUN_FOREACH_BBL(fun,i_bbl)
        BBL_FOREACH_PRED_EDGE(i_bbl, i_edge) 
        {
          if (CfgEdgeIsForwardInterproc(i_edge))
            if (EdgeIsMarked(i_edge))
            {
              if (RegsetIn(i_edge->prop_regs,propreg))
                if (UdChainJoin(&(i_bbl->ud_chain_in),i_edge->ud_chain))
                  FunctionMarkBbl(fun,i_bbl);
            }
        }
    }
    FunctionMarkBbl(fun,fun->bbl_first);
    /* at the beginning of the program, propagation should start with all BOT.
     * however this is pointless in this case, so we use another trick (explained
     * at the beginning of this file): we use a dummy instruction as the initial 
     * definer for all registers of the processor
     */
    UdChainAppend(&(fun->bbl_first->ud_chain_in),UdChainNew(&dummy_bbl));
  }

  /* 
     do the actual fixpoint calculation inside the fun
   */

  while (FunctionUnmarkBbl(fun,&i_bbl))
  {
    BblUseDefAnalysis(obj,i_bbl,edge_out,during_fixpoint_calculations,propreg);
  }

  /* 
     free the allocated memory
   */

  if (during_fixpoint_calculations)
  {
    FUN_FOREACH_BBL(fun,i_bbl)
    {
      UdChainFree(i_bbl->ud_chain_in);
      i_bbl->ud_chain_in = NULL;
    }
  }
}
/* }}} */

/* {{{ propagate a list of ud-chains over an edge */
static void PropOverEdge(t_cfg_edge * edge, t_ud_chain * udc, t_bool do_it, t_bool during_fixpoint_calculations, t_reg propreg)
{

  t_cfg_edge * corr_edge;

  if (!do_it) return;

  /* mark it as taken */

  EdgeMark(edge);

  if (CfgEdgeIsInterproc(edge))
  {
    /* 
       if it is a true outgoing edge, the corresponding incoming edge must have its 
       information propagated into the procedure (if it is marked that is).
     */
    if (CfgEdgeIsForwardInterproc(edge))
    {
      corr_edge = edge->corr_edge;
      if (corr_edge)
      {
        if (EdgeIsMarked(corr_edge))
          if (CFG_EDGE_TAIL(corr_edge)->fun==(CFG_EDGE_HEAD(edge)->fun) && CFG_EDGE_HEAD(edge)!=CFG_EDGE_TAIL(corr_edge))
          {
            if (RegsetIn(corr_edge->prop_regs,propreg))
              if (UdChainJoin(&(CFG_EDGE_TAIL(corr_edge)->ud_chain_in),corr_edge->ud_chain))
                FunctionMarkBbl(CFG_EDGE_HEAD(edge)->fun,CFG_EDGE_TAIL(corr_edge));
          } 
      }
    }

    if (during_fixpoint_calculations)
    {
      if (RegsetIn(edge->prop_regs,propreg))
        if (UdChainJoin(&(edge->ud_chain),udc))
        {
          if (!CfgEdgeIsForwardInterproc(edge))
          {
            /* repropagation for all true incoming edges of the 
             * function this edge points to is necessary */
            CfgMarkFun(CFG_EDGE_TAIL(edge)->fun->cfg,CFG_EDGE_TAIL(edge)->fun);
            FunctionMarkAllTrueIncomingEdges(CFG_EDGE_TAIL(edge)->fun);
          }
          else 
          {
            /* only this edge will need to be repropagated in 
             * the function it points to */
            CfgMarkFun(CFG_EDGE_TAIL(edge)->fun->cfg,CFG_EDGE_TAIL(edge)->fun);
            FunctionMarkEdge(CFG_EDGE_TAIL(edge)->fun,edge);
          }
        }
    }
  }
  else 
  {
    /* the edge was INTRAprocedural */
    if (RegsetIn(edge->prop_regs,propreg))
      if (UdChainJoin(&(CFG_EDGE_TAIL(edge)->ud_chain_in),udc))
      {
        FunctionMarkBbl(CFG_EDGE_TAIL(edge)->fun,CFG_EDGE_TAIL(edge));
      }
  }
}
/* }}} */

/* {{{  Propagate a state over a bbl and to the successor edges
 */

static void BblUseDefAnalysis(t_object* obj, t_bbl * bbl, t_cfg_edge * edge_out, t_bool during_fixpoint_calculations, t_reg propreg)
{

  /* iterators */

  t_cfg_edge * i_edge;
  t_ud_chain * udc = NULL;
  t_regset hell_live_saved = BBL_REGS_LIVE_OUT(bbl); /* Just an assignement to keep the compiler from complaining */

  /* special hack: for hell nodes, this is necessary to ensure
   * propagation of _all_ the registers on the outgoing edges */
  if (IsHellNode(bbl))
  {
    hell_live_saved = BBL_REGS_LIVE_OUT(bbl);
    RegsetDup(BBL_REGS_LIVE_OUT(bbl),obj->description->all_registers);
  }


  if (RegsetIn(BBL_REGS_LIVE_OUT(bbl),propreg))
  {
    if (RegsetIn(BBL_REGS_DEF_CERTAIN(bbl),propreg))
    {
      udc = UdChainDup(bbl->tmp);
    }
    else if (RegsetIn(BBL_REGS_DEF_PERHAPS(bbl),propreg))
    {
      udc = UdChainDup(bbl->ud_chain_in);
      UdChainAppend(&udc,UdChainNew(bbl));
    }
    else
    {
      udc = UdChainDup(bbl->ud_chain_in);
    }
  }
  else
  {
    /* if propreg isn't live at the end of the basic block, we don't propagate 
     * the incoming definitions any further -> instead we propagate dummy_bbl,
     * which will be ignored when constructing the ud-graph */
    udc = UdChainNew(&dummy_bbl);
  }

  /* propagate the outgoing information over the outgoing edges */
  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
    PropOverEdge(i_edge, udc, (!CfgEdgeIsInterproc(i_edge) || CfgEdgeIsForwardInterproc(i_edge) || i_edge == edge_out || edge_out == NULL), during_fixpoint_calculations,propreg);

  /* Restore the original live regs out of the hell node */
  if (IsHellNode(bbl))
  {
    BBL_REGS_LIVE_OUT(bbl) = hell_live_saved;
  }

  UdChainFree(udc);
}
/* }}} */

/*  {{{ propagate over one instruction */
static void UseDefInstructionPropagator(t_ins * ins, t_ud_chain ** udc_p, t_reg propreg)
{
  if (RegsetIn(INS_REGS_DEF(ins),propreg))
  {
    if (INS_IS_CONDITIONAL(ins))
    {
      /* this definition must be appended to the list */
      UdChainAppend(udc_p,UdChainNew((t_bbl *)ins));
    }
    else
    {
      /* this definition must become the only one */
      UdChainFree(*udc_p);
      *udc_p = UdChainNew((t_bbl *)ins);
    }
  }
}
/* }}} */

/* helper function: given a bbl and a register, this function returns a list
 * of which instructions could possibly define this register at the 
 * end of the bbl */
t_ud_chain * BblDefinersList(t_bbl * bbl, t_reg propreg)
{
  t_ud_chain * ret = NULL;
  t_ins * i_ins;

  BBL_FOREACH_INS(bbl,i_ins)
    UseDefInstructionPropagator(i_ins,&ret,propreg);

  return ret;
}

/* {{{ Set up the final ud-graph */
static void UseDefBuildGraph(t_object * obj, t_section * sec, t_reg propreg)
{
  /* iterators */

  t_function * i_fun;
  t_bbl * i_bbl;
  t_ins * i_ins;

  /* local variables */

  t_bbl * program_entry;
  t_function * entry_fun;
  t_cfg * cfg = T_CFG(sec->data);
  t_ud_graph * udgraph = CFG_UDGRAPH(cfg);

  DiabloPrint(stdout,"Building use-definition graph for reg %d\n",propreg);

  /* some initialization: 
   *    first of all, free the tmp pointers that are kept in each bbl. we don't 
   *    need them any more.
   *    then, use these temp pointers to store the information of BblDefinersList
   *    (this information is invariable so it can be computed once and then reused
   *    forever)
   */
  CFG_FOREACH_BBL(cfg, i_bbl)
  {
    i_bbl->tmp2 = BblDefinersList(i_bbl,propreg);
  }
  /* the dummy_bbl has to be initialized as well */
  dummy_bbl.tmp2 = BblDefinersList(&dummy_bbl,propreg);

  /* now propagate the information on the interprocedural edges
   * into the functions and basic blocks. With this information
   * we can build up the ud-graph */
  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    FunctionUnmarkAllBbls(i_fun);
  }

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    if (i_fun->bbl_first)
    {
      FunctionUseDefAnalysis(obj,i_fun,sec,T_CFG_EDGE(i_fun!=entry_fun),FALSE,propreg);

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {

        /* translate i_bbl->ud_chain_in from a chain of bbls to a chain of instructions */
        {
          t_ud_chain * ud_ins_in = NULL;
          t_ud_chain * iter = NULL;

          for (iter=i_bbl->ud_chain_in; iter!=NULL; iter=LLNext(iter))
          {
            UdChainJoin(&ud_ins_in,(t_ud_chain *) iter->def->tmp2);
          }
          /* we don't need the original i_bbl->ud_chain_in any more, so we replace the old one 
           * with the translated version */
          UdChainFree(i_bbl->ud_chain_in);
          i_bbl->ud_chain_in = ud_ins_in;
        }

        /* it doesn't matter that the following code changes i_bbl->ud_chain_in,
         * it won't ever be used afterwards anyway */

        BBL_FOREACH_INS(i_bbl,i_ins)
        {
          /* insert edges in the ud-graph */
          t_ud_chain * i_udc;

          if (RegsetIn(INS_REGS_USE(i_ins),propreg))
            for (i_udc = i_bbl->ud_chain_in; i_udc != NULL; i_udc = LLNext(i_udc))
              if (T_INS(i_udc->def) != &dummy_ins)
              {
                UdGraphLinkUseToDefForRegister(udgraph, i_ins, T_INS(i_udc->def), propreg);
              }

          /* propagate over the instruction */
          UseDefInstructionPropagator(i_ins, &(i_bbl->ud_chain_in), propreg);
        }
      }

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
        UdChainFree(i_bbl->ud_chain_in);
        i_bbl->ud_chain_in = NULL;
      }
    }
  }

  /* release temporary information */
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    UdChainFree(i_bbl->tmp2);
  }
  UdChainFree(dummy_bbl.tmp2);
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker: */
