/***********************************************************
 * Copyright 2001,2002,2003: {{{
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This file contains the main program for DIABLO and CoALA. It can be used as
 * a skeleton for other objectconsumers/editors.
 * 
 * }}}
 **********************************************************/

/* Includes {{{ */
#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_SYS_WAIT
#include <sys/wait.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include "fileformats/ar/ar_arch.h"
#include "utils/util_error.h"
#include "utils/util_machine.h"
#include "utils/util_file.h"
#include "utils/util_malloc.h"
#include "utils/util_options.h"
#include "utils/util_status.h"
#include "utils/util_term.h"
#include "utils/util_processes.h"
#include "datastructures/data_string.h"
#include "kernel/diablo_print.h"
#include "kernel/diablo_liveness.h"
#include "kernel/diablo_ud_analysis.h"
#include "kernel/diablo_object.h"
#include "kernel/diablo_flowgraph.h"
#include "kernel/diablo_constprop.h"
#include "kernel/diablo_opt_misc.h"
#include "kernel/diablo_dominator.h"
#include "kernel/diablo_iterators.h"
#include "kernel/diablo_link.h"
#include "kernel/diablo_programexit.h"
#include "kernel/diablo_save_state.h"
#include "kernel/diablo_copy_analysis.h"
#include "kernel/diablo_profile.h"
#include "kernel/diablo_code_motion.h"
#include "kernel/diablo_pre.h"
#include "diablo_options.h"

#ifdef MIPS_SUPPORT
#include "kernel/diablo_copy_analysis.h"
#endif

#ifdef ARM_SUPPORT
#include "arch/arm/arm_flowgraph.h"
#include "arch/arm/arm_copy_propagation.h"
#include "arch/arm/arm_constant_propagation.h"
#include "arch/arm/arm_peephole.h"
#include "arch/arm/arm_ls_fwd.h"
#include "kernel/diablo_copy_analysis.h"
#include "kernel/diablo_factor.h"
#endif

#ifdef IA64_SUPPORT
#include "arch/ia64/ia64_copy_propagation.h"
#include "arch/ia64/ia64_liveness.h"
#include "arch/ia64/ia64_dummy.h"
#include "arch/ia64/ia64_profiler.h"
#include "arch/ia64/ia64_inlining.h"
#include "arch/ia64/ia64_opt_special.h"
#endif
#ifdef OBFUSCATION_SUPPORT
#include "obfuscation/obf_branch.h"
#include "obfuscation/obf_dis.h"
#endif
#ifdef WATERMARK
#include "watermark/watermark.h"
#endif
/*}}}*/
extern t_uint32 global_optimization_phase;

#define SPEEDUP_DOMINATORS

extern t_bool dominator_info_correct;

#define XIANGYU
#ifdef XIANGYU
extern void XZ_Export_FunctionDominator(t_function *, t_uint32);
extern void XZ_Export_ControlDependence(t_function *, FILE * );
extern void XZ_Com_ControlDependence(t_function *);
extern void XZ_CgExport(t_cg* );
#endif

/* Config variables for DUMP and RESTORE {{{ */
#define DUMP_DIR "./DUMP"
#define VERBOSE_DUMP 
/* }}}*/

/* RestoreDumpedProgram {{{ */
void RestoreDumpedProgram()
{
   /* Variables used when restoring a dump from a multiple dump */
   t_bool multi=FALSE;
   t_uint32 multi_tel=0;
   /* File pointer to the dump definition file */
   FILE * fp_def;
   /* Holds the lines from the dump definition file */
   t_string line;
   t_bool input=FALSE;
   t_bool output=FALSE;
   t_bool arguments=FALSE;


   if (global_options.input_files_set) input=TRUE;
   if (global_options.output_files_set) output=TRUE;
   if (global_options.arguments_set) arguments=TRUE;

#ifdef VERBOSE_DUMP
   printf("RESTORE!\n");
#endif
   
   Free(global_options.prepend);
   global_options.prepend=ConvertFileName(DUMP_DIR);
   
   /* Untar the dump, and open the dump definition file {{{ */
   Untar("dump.tar");
   fp_def=fopen("./DUMP/diablo_dump","r");
   if (!fp_def) FATAL(("Cannot restore: dump not found"));
   /* }}} */
   /* Locate the dump in the file (in case of multiple dumps and set the corresponding variable {{{ */
   while(!feof(fp_def))
   {
      line=GetLine(fp_def);

      if (!feof(fp_def)) 
      {
	 if (PatternMatch("LPATH: *",line))
	 {
	    if (global_options.libpath) PathFree(global_options.libpath);
	    global_options.libpath=ParsePath(line+7,":\n");
	 }
	 else if (PatternMatch("OPATH: *",line))
	 {
	    if (global_options.objpath) PathFree(global_options.objpath);
	    global_options.objpath=ParsePath(line+7,":\n");
	 }
	 else if (PatternMatch("ARGS: *",line))
	 {
	    Chomp(line);
	    if (!arguments)
	    {
	      if (global_options.arguments) Free(global_options.arguments);
	      global_options.arguments=StringDup(line+6);
	    }
	    global_options.arguments_set=TRUE;
	 }
	 else if (PatternMatch("INPUTS: *",line))
	 {
	    Chomp(line);
	    if (!input)
	    {
	      if (global_options.input_files) Free(global_options.input_files);
	      global_options.input_files=StringDup(line+8);
	    }
	    global_options.input_files_set=TRUE;
	 }
	 else if (PatternMatch("OUTPUTS: *",line))
	 {
	    Chomp(line);
	    if (!output)
	    {
	      if (global_options.output_files) Free(global_options.output_files);
	      global_options.output_files=StringDup(line+9);
	    }
	    global_options.output_files_set=TRUE;
	 }
	 else if (PatternMatch("OBJ: *",line))
	 {
	    Chomp(line);
	    if (global_options.objectfilename) Free(global_options.objectfilename);
	    global_options.objectfilename=StringDup(line+5);
	    if (!global_options.objectfilename) FATAL(("Couldn't find %s\n",line+5));
	    global_options.objectfilename_set=TRUE;
	 }
	 else if (strcmp(line,"=============\n")==0)
	 {
	   if (global_options.restore_multi==-1)
	   {
	     if (!multi)
	     {
	       printf("This file contains the following dumps:\n");
	     }
	     printf("%3d: %s\n",multi_tel,global_options.objectfilename);
	   }
	   else if (multi_tel==global_options.restore_multi)
	   {
	     printf("%d %d\n",global_options.restore_multi, global_options.restore_multi_set);
	     printf("Selecting %d: %s from dump\n",multi_tel,global_options.objectfilename);
	     Free(line);
	     return;
	   }
	   multi=TRUE;
	   multi_tel++;
	 }
	 else FATAL(("Unrecognized dump line %s\n",line));
	 Free(line);
      }
   }

   /* }}} */

   if (multi_tel==global_options.restore_multi)
   {
     printf("Selected %d: %s from dump\n",multi_tel,global_options.objectfilename);
     return;
   }

   if ((multi) && (global_options.restore_multi==-1)) 
   {
     printf("%3d: %s\n",multi_tel,global_options.objectfilename);
     RmdirR("./DUMP");
     FATAL(("Must select a dump from the file!",global_options.restore_multi));
   }
   else if (multi) 
   {
     RmdirR("./DUMP");
     FATAL(("Must select a valid dump from the dump-file!\n Use objdump -r to get a list. %d is not a valid dump",global_options.restore_multi));
     
   }

}
/* }}} */
/* DumpProgram {{{ */
void DumpProgram(t_object * obj,t_bool multiple_dumps)
{
      FILE * fp;
      t_uint32 tel;
      t_object * i_obj, * tmp; 
      t_string dest=Collate("./DUMP",obj->name);
      t_string tmp_string=Collate(obj->name,".map");
#ifdef ALPHA_SUPPORT
      t_string tmp_string2=Collate(obj->name,".lo");
#endif
      if (multiple_dumps) Untar("dump.tar");
      SmartCopy(obj->name,dest);
      Free(dest);
      dest=Collate3("./DUMP",obj->name,".map");
      SmartCopy(tmp_string,dest);
#ifdef ALPHA_SUPPORT
      SmartCopy(tmp_string2,dest);
#endif
      Free(dest);
      Free(tmp_string);


      /* Copy the subobjects {{{ */
      OBJECT_FOREACH_SUBOBJECT(obj, i_obj, tmp)
      {
	 if (strncmp(i_obj->name,"Linker---",9)==0)
	 {
	 }
	 else if (strcmp(i_obj->name,"Linker")==0)
	 {
	 }
	 else if (PatternMatch("*:*",i_obj->name))
	 {
	    t_string name2=StringDup(i_obj->name);
	    t_string libn=(t_string)strtok(name2,":");
	    dest=Collate("./DUMP",libn);
	    SmartCopy(libn,dest);
	    Free(name2);
	    Free(dest);
	 }
	 else
	 {
	    dest=Collate("./DUMP",i_obj->name);
	    SmartCopy(i_obj->name, dest);
	    Free(dest);
	 }
      }
      /* }}} */
      /* Copy the inputs  {{{ */

      if (global_options.input_files_set)
      {
	int tel;
	t_string_array * in=StringDivider(global_options.input_files,";",TRUE,FALSE);

	for (tel=0; tel<in->nstrings; tel++)
	{
	  char * name;
	  name=ConvertFileName(in->strings[tel]);
	  if (name[0]!='/') FATAL(("Input file is not absolute! Implement"));
	  dest=Collate("./DUMP",name);
	  SmartCopy(name,dest);
	  Free(dest); Free(name);
	}
      }
      /* }}}*/

     /* Copy the outputs {{{ */

      if (global_options.output_files_set) {
	int tel;
	t_string_array * in=StringDivider(global_options.output_files,";",TRUE,FALSE);

	for (tel=0; tel<in->nstrings; tel++)
	{
	  char * name;
	  name=ConvertFileName(in->strings[tel]);
	  if (name[0]!='/') FATAL(("Output file is not absolute! Implement"));
	  dest=Collate3("./DUMP",name,"_ref");
	  SmartCopy(name,dest);
	  Free(dest); Free(name);
	}
      }

      /* }}} */


      if (multiple_dumps)
      {
	fp=fopen("./DUMP/diablo_dump","a");
	fprintf(fp,"=============\n");
      }
      else
	fp=fopen("./DUMP/diablo_dump","w");

      if (!fp) FATAL(("Failed to open diablo_dump, for writing"));
      fprintf(fp,"LPATH: ");
      for (tel=0; tel<global_options.libpath->nelems; tel++)
      {
	 fprintf(fp,"%s",global_options.libpath->dirs[tel]);
	 if (tel+1<global_options.libpath->nelems) fprintf(fp,":");
      }
      fprintf(fp,"\n");

      fprintf(fp,"OPATH: ");
      for (tel=0; tel<global_options.objpath->nelems; tel++)
      {
	 fprintf(fp,"%s",global_options.objpath->dirs[tel]);
	 if (tel+1<global_options.objpath->nelems) fprintf(fp,":");
      }
      fprintf(fp,"\n");

      fprintf(fp,"ARGS: %s\n",global_options.arguments);

      /* Set the input line {{{ */
      if (global_options.input_files_set) {
	int tel;
	t_string_array * in=StringDivider(global_options.input_files,";",TRUE,FALSE);

	fprintf(fp,"INPUTS: ");
	for (tel=0; tel<in->nstrings; tel++)
	{
	  char * name;
	  name=ConvertFileName(in->strings[tel]);
	  if (name[0]!='/') FATAL(("Input file is not absolute! Implement"));
	  fprintf(fp,"%s",name);
	  if (tel+1<in->nstrings) fprintf(fp,";");
	  Free(name);
	}

	fprintf(fp,"\n");
      }
      /* }}} */

      /* Set the output line {{{ */
      /* The outputs */
      if (global_options.output_files_set) {
	int tel;
	t_string_array * in=StringDivider(global_options.output_files,";",TRUE,FALSE);

	fprintf(fp,"OUTPUTS: ");
	for (tel=0; tel<in->nstrings; tel++)
	{
	  char * name;
	  name=ConvertFileName(in->strings[tel]);
	  if (name[0]!='/') FATAL(("Input file is not absolute! Implement"));
	  fprintf(fp,"%s",name);
	  if (tel+1<in->nstrings) fprintf(fp,";");
	  Free(name);
	}

	fprintf(fp,"\n");
      }

      /* }}} */


      fprintf(fp, "OBJ: %s\n",obj->name);
      fclose(fp);

     
      DirToTar("./DUMP","dump.tar");
}
/*}}} */
/* Detect Leaves {{{ */
void CfgDetectLeaves(t_cfg * cfg)
{
   t_function * fun;
   t_uint32 nleafs=0,nfuncs=0;
   CFG_FOREACH_FUN(cfg,fun)
   {
      t_bool leaf=TRUE;
      t_bbl * bbl;
      t_cfg_edge * edge;
      FUN_FOREACH_BBL(fun,bbl)
      {
	 BBL_FOREACH_SUCC_EDGE(bbl,edge)
	 {
	    if (CfgEdgeIsForwardInterproc(edge))
	    {
	       leaf=FALSE;
	       break;
	    }
	 }
	 if (!leaf) break;
      }

      if (leaf) {
	 nleafs++;
	 FUN_FLAGS(fun)|=FF_IS_LEAF;
      }
      else 
      {
	 FUN_FLAGS(fun) &= ~FF_IS_LEAF;
      }

      nfuncs++;
   }
   printf("[Detect leaves]: %d functions of %d are leafs\n",nleafs,nfuncs);
}
/* }}} */

/* Execute the program {{{ */
#ifdef HAVE_IPC
#ifdef HAVE_FILENO
#define HAVE_DIABLO_EXEC
void DiabloExec()
{
     char * execname=ConvertFileName("./b.out");
     char *f;
     char * instring=StringDup("DIABLO=1");
     FILE * in;
     FILE * out;
     pid_t pid;
     
     if ((global_options.read)&&(!(global_options.dump||global_options.dump_multiple))) chmod("./b.out",0700);
     else execname=FindFile(global_options.objpath,global_options.objectfilename);
     
     printf("------------- Executing %s with args %s\n",execname,global_options.arguments);

     /* Do the inputs */
     if (global_options.input_files_set) {
       int tel;
       t_string_array * in=StringDivider(global_options.input_files,";",TRUE,FALSE);

       for (tel=0; tel<in->nstrings; tel++)
       {
	 char * name;

	 if (global_options.prepend[0]!=0)
	 {
	   char * tmp;
	   name=Collate3(global_options.prepend,"/",tmp=ConvertFileName(in->strings[tel]));
	   Free(tmp);
	 }
	 else
	 {
	   name=ConvertFileName(in->strings[tel]);
	 }
	 
	 if (tel+1<10)
	 {
	   f=instring;
	   instring=Collate3(instring,"; Ix=",name);
	   Free(f);
	   sprintf(instring+strlen(instring)-strlen(name)-2,"%d",tel+1);
	 }
	 else
	 {
	   f=instring;
	   instring=Collate3(instring,"; Ixx=",name);
	   Free(f);
	   sprintf(instring+strlen(instring)-strlen(name)-3,"%d",tel+1);
	 }
	 *(instring+strlen(instring))='=';
	 Free(name);
	 
       }

       StringArrayFree(in);

     }

     /* Do the outputs */
     if (global_options.output_files_set) {
       int tel;
       t_string_array * in=StringDivider(global_options.output_files,";",TRUE,FALSE);

       for (tel=0; tel<in->nstrings; tel++)
       {
	 char * name;
	 if (global_options.prepend[0]!=0)
	 {
	   char * tmp;
	   name=Collate3(global_options.prepend,"/",tmp=ConvertFileName(in->strings[tel]));
	   Free(tmp);
	 }
	 else
	 {
	   name=ConvertFileName(in->strings[tel]);
	 }
	 if (tel+1<10)
	 {
	   f=instring;
	   instring=Collate3(instring,"; Ox=",name);
	   Free(f);
	   sprintf(instring+strlen(instring)-strlen(name)-2,"%d",tel+1);
	 }
	 else
	 {
	   f=instring;
	   instring=Collate3(instring,"; Oxx=",name);
	   Free(f);
	   sprintf(instring+strlen(instring)-strlen(name)-3,"%d",tel+1);
	 }
	 *(instring+strlen(instring))='=';
	 Free(name);
       }

       StringArrayFree(in);

     }
     
     if (global_options.arguments_set)
     {
       char * tofree1, * tofree2;
       if (global_options.host_set)
       {
	 CreateProcPipe(&pid,&out,&in, NULL, "ssh", global_options.host, tofree1=Collate3(tofree2=Collate3(instring,";",execname), " ", global_options.arguments),  NULL);
       }
       else
       CreateProcPipe(&pid,&out,&in, NULL, "bash", "-c", tofree1=Collate3(tofree2=Collate3(instring,";",execname), " ", global_options.arguments),  NULL);
       Free(tofree1);
       Free(tofree2);


       
     }
     else
     {
       CreateProcPipe(&pid,&out,&in, NULL, "bash", "-c", execname,  NULL);
     }

     while (!feof(out))
     {
       char c;
       fread(&c,1,1,out);
       if (feof(out)) break;
       printf("%c",c);
     }
     Free(instring);

     if (global_options.output_files_set) {
       int tel;
       t_string_array * ina=StringDivider(global_options.output_files,";",TRUE,FALSE);

       for (tel=0; tel<ina->nstrings; tel++)
       {
	 int status;
         char * tmp;	 
	 char * name, *f;
	 if (global_options.prepend[0]!=0)
	 {
	   name=Collate3(global_options.prepend,"/",tmp=ConvertFileName(ina->strings[tel]));
	   Free(tmp);
	 }
	 else
	 {
	   name=ConvertFileName(ina->strings[tel]);
	 }
	 f=name;
	 name=Collate3(name," ", name);
	 Free(f);
	 f=name;
	 name=Collate3("diff -q -s ",name,"_ref");
	 Free(f);
	 
	 printf("====Exec %s\n",name);
	 CreateProcPipe(&pid,&out,&in,NULL, "bash", "-c", name, NULL);
	 while (!feof(out))
	 {
	   char c;
	   fread(&c,1,1,out);
	   if (feof(out)) break;
	   printf("%c",c);
	 }
	 waitpid(pid,&status,0);

	 /* is non-zero if the child exited normally. */
	 if (WIFEXITED(status))
	 {
	   printf("EXIT STATUS OF DIFF: %d\n",WEXITSTATUS(status));
	 }
	 /*
	 else if ( WIFSIGNALED(status))
	 {
	   WTERMSIG(status)
	 }
	 */
	 else 
	 {
	   FATAL(("Diff crashed!"));
	 }

	 Free(name);
       }

       StringArrayFree(ina);

     }
}
#endif
#endif
/* }}} */

/*
void CopyAnalysisPrintBeforeFunction(t_function * fun)
{
  t_bbl * i_bbl = FUN_BBL_FIRST(fun);
  t_ins * i_ins;
  t_uint32 i;
  static int c=0;

  if(BBL_EQS_IN(i_bbl))
  {
    DiabloPrint(stdout,"Before fun @B\n",i_bbl);

    for(i = 0; i < 5; i++)
    {
      if (((t_equations) BBL_EQS_IN(i_bbl))[i].rega==REG_BOT) continue;
      if (((t_equations) BBL_EQS_IN(i_bbl))[i].taga!=NULL) continue;
      if (((t_equations) BBL_EQS_IN(i_bbl))[i].tagb!=NULL) continue;
      if (((t_equations) BBL_EQS_IN(i_bbl))[i].constant!=0) continue;
      if (((t_equations) BBL_EQS_IN(i_bbl))[i].regb<i) continue;
      printf("%d == %d\n",i,((t_equations) BBL_EQS_IN(i_bbl))[i].regb); 
      DiabloPrint(stdout,"Before @iB\n",i_bbl);
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	if (c>global_options.debugcounter) continue;
	if ((INS_REGS_DEF(i_ins)&(1<<i))&&(INS_REGS_USE(i_ins)&(1<<i))) break;
	if (ARMINSREGB(i_ins)==i) 
	{
	  DiabloPrint(stdout,">>>>> @I ",i_ins);
	  ARMINSREGB(i_ins)=((t_equations)BBL_EQS_IN(i_bbl))[i].regb;
	  DiabloPrint(stdout," @I <<<<<<\n",i_ins);
	}
	if (ARMINSREGC(i_ins)==i) ARMINSREGC(i_ins)=((t_equations)BBL_EQS_IN(i_bbl))[i].regb;
	INS_REGS_DEF(i_ins)=ArmDefinedRegisters(i_ins);
	INS_REGS_USE(i_ins)=ArmUsedRegisters(i_ins);
	c++;
	printf("RENAME!\n");
	if (INS_REGS_DEF(i_ins)&(1<<i)) break;
	if (INS_REGS_DEF(i_ins)&(1<<((t_equations)BBL_EQS_IN(i_bbl))[i].regb)) break;
      }
      DiabloPrint(stdout,"After @iB\n",i_bbl);
    }
  }

}
*/

#if 0
void testje(int x, t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS(bbl,ins)
    if (INS_OLD_ADDRESS(ins)==0x1a5bc)
      {
	printf("FOUND %d\n",x);
	fflush(stdout);
	DiabloPrint(stdout,"FOUNDIT @I\n",ins);
	fflush(stdout);
      }
}
#endif

int main(int argc,char ** argv)
{
  t_object * obj;
  /* I. Initialize all structures {{{ */
  /* Initialize the terminal */
  TermInit();
  /* Initialize path-related (global) datastructures
   * (variables for path-expansion like the home-directory
   * of the use */
  PathInit();
  /* Find out some stuff about the current machine */
  MachineInit();
  /* Initialize options, defaults are set in OptionsDefaults */ 
  OptionsInit();
  /* Initialize (global) datastructures for archives */
  ArchivesInit();
  /* Initialize (global) datastructures for objects  */
  ObjectInit();
  printf("good 2\n");
  /* Send the errors to stderr, unless stdout is not a tty
   * (then we send it to stdout) 
   *
   * So 2>/dev/null will work (if you do not redirect
   * stdout)
   *    >file or |less will contain the error messages
   *    >file1 2>file2 will go wrong....
   *    
   * This is a temporary hack to ease debugging, ( |less of
   * the output should contain the errors)
   *
   * normally it should just say:
   * ErrorInit(argv[0],stderr,Usage);
   */
#ifdef HAVE_ISATTY
#ifdef HAVE_FILENO
  if (isatty(fileno(stdout)))
    ErrorInit(argv[0],stdout,Usage);
  else
#endif
#endif
    ErrorInit(argv[0],stdout,Usage);
  ErrorClearErrno(); /* isatty can set errno, but that's not an error */

  /* }}} */
  /* II. Parse Command line, get Environment and Set defaults {{{ */
  /* A. Parse Command line */
  OptionsParseCommandLine(argc,argv);
  /* B. Get enviromental variables (Commandline supersedes) */
  OptionsGetEnvironment();
  /* C. Set default arguments (Commandline and Environment supersedes) */
  OptionsVerify();
  OptionsDefaults();
#ifdef IA64_SUPPORT
  global_options.constant_propagation = FALSE;
#endif
  /*}}}*/
  /* III. The REAL program {{{ */

#ifdef GRAPHICS
  EnvironmentInit();
  WindowInit();
#endif

#ifdef XIANGYU
   /*by Xiangyu {{{ */
      obj=LinkEmulate(global_options.objectfilename);
      printf("good 1\n");
      if (obj->mapped_first) 
	ObjectMergeCodeSections(obj);
      printf("disassemble\n"); 
      ObjectAction(DISASSEMBLE,obj); 
      /* 2. {{{ Create the flowgraph */
      //if (global_options.flowgraph)
      {
	ObjectAction(FLOWGRAPH,obj);  
  
        t_function * function;
	int tel; 
        for (tel=0;tel<obj->ncodes;tel++) {
          FOREACH_FUNCTION(obj->code[tel],function) {
            printf("BF:%s:%s\n",SECTION_NAME(FUN_SECTION(function)), FUN_NAME(function));
          }
        }    

	/* Remove unconnected blocks {{{ */
	//if (global_options.initcfgopt)
	  {
	    ANALYSIS_START("Removing unconnected blocks");
	    CfgRemoveDeadCodeAndDataBlocks(T_CFG(obj->code[0]->data));
	    if (CfgPatchForNonReturningFunctions(T_CFG(obj->code[0]->data)))
	      CfgRemoveDeadCodeAndDataBlocks(T_CFG(obj->code[0]->data));
	    ANALYSIS_SIMPLE_END("Removing unconnected blocks");
	  }
	//if (global_options.dominator) 
	  {
	    ComDominators(T_CFG(obj->code[0]->data));
	  }
	/* Export dots after dominator */
	//if (global_options.dominator && global_options.generate_dots)
	{
	  t_function * function;

          /*
	  DirectoryAssure("./dots-dominator",FALSE);
	  DirectoryAssure("./dots-postdominator",FALSE);

	  Export_FunctionDominator((T_CFG(obj->code[0]->data))->unique_entry_node, 0xffff, "./dots-dominator/flowgraph.dot");
	  Export_FunctionPostDominator((T_CFG(obj->code[0]->data))->unique_entry_node, 0xffff, "./dots-postdominator/flowgraph.dot");
          */
          FILE * cd_file=fopen("./cd/CONTROL_DEP","w");
          int cnt=0;
          FOREACH_FUNCTION(*(obj->code),function) {
             cnt++;
          }
          fprintf(cd_file,"%d\n",cnt);
	  FOREACH_FUNCTION(*(obj->code),function)
	  {
	    char * fname, *fname2;
            /*
	    if (function->bbl_first)
	    {
	      fname=sDiabloPrint("./dots-dominator/@G.func-%s.dot",BBL_OLD_ADDRESS(function->bbl_first),FUN_NAME(function));
	      fname2=sDiabloPrint("./dots-postdominator/@G.func-%s.dot",BBL_OLD_ADDRESS(function->bbl_first),FUN_NAME(function));
	    }
	    else
	    {
	      fname=sDiabloPrint("./dots-dominator/0x0.func-%s.dot",FUN_NAME(function));
	      fname2=sDiabloPrint("./dots-postdominator/0x0.func-%s.dot",FUN_NAME(function));
	    }
	    Export_FunctionDominator(function->bbl_first, ~ET_INTERPROC, fname);
	    Export_FunctionPostDominator(function->bbl_first, ~ET_INTERPROC, fname2);
	    Free(fname);
	    Free(fname2);
            */
            printf("%s:%s\n",SECTION_NAME(FUN_SECTION(function)), FUN_NAME(function));
            fprintf(cd_file, "%x %s\n", BBL_CADDRESS(function->bbl_first), FUN_NAME(function));
	    XZ_Com_ControlDependence(function);
	    XZ_Export_FunctionDominator(function, ~ET_INTERPROC);
	    XZ_Export_ControlDependence(function, cd_file);
	  }//FOREACH
          fclose(cd_file);
	}//dominator
	//if (global_options.dominator) 
	CgBuild(T_CFG(obj->code[0]->data));
	XZ_CgExport(T_CFG(obj->code[0]->data)->cg);

        DominatorCleanup(T_CFG(obj->code[0]->data));
        if (1) exit(0);	

      }
   /*}}}*/
#endif

  /* Restore a dumped program, it will be loaded by ObjectGet */
  if (global_options.restore || (global_options.restore_multi!=-1)) RestoreDumpedProgram();


  if (global_options.read)
  {
    /* A. Get a binary to work with {{{ */
    
    if (global_options.restoreState || global_options.stegoDecode)
      obj=RestoreState();
    else 
      obj=LinkEmulate(global_options.objectfilename);

    /* End Load }}} */


    if (global_options.print_data_layout)
      {
	t_object * subobj;
	t_object * tmp;
	t_section * sec;
	int tel;
	OBJECT_FOREACH_SUBOBJECT(obj,subobj,tmp)
	  OBJECT_FOREACH_SECTION(subobj,sec,tel)
	  {
	    if (SECTION_TYPE(sec)==DATA_SECTION)
	      DiabloPrint(stdout,".data 0x%x 0x%x\n",SECTION_GET_OLD_ADDRESS(sec),SECTION_CSIZE(sec));
	    else if (SECTION_TYPE(sec)==BSS_SECTION)
	      DiabloPrint(stdout,".bss 0x%x 0x%x\n",SECTION_GET_OLD_ADDRESS(sec),SECTION_CSIZE(sec));
	  }
	exit(0);
      }
    if (global_options.relayout_description)
    {
      if (obj->mapped_first) 
	ObjectMergeCodeSections(obj);
      ObjectMergeRODataSections(obj);
      ObjectMoveBssToDataSections(obj);
      ObjectRelayoutData(obj,global_options.relayout_description);
    }
    else 
    if ((global_options.disassemble) && (!(global_options.dump || global_options.dump_multiple)))
    {
      /* B. Transform and optimize  {{{ */
      /* 1. Disassemble */ 
      /* Merge code sections to ease disassembly */

      if (obj->mapped_first) 
	ObjectMergeCodeSections(obj);
      
      ObjectAction(DISASSEMBLE,obj); 
      /* 2. {{{ Create the flowgraph */
      if (global_options.flowgraph)
      {
	ObjectAction(FLOWGRAPH,obj);   

	/* Remove unconnected blocks {{{ */
	if (global_options.initcfgopt)
	  {
	    ANALYSIS_START("Removing unconnected blocks");
	    CfgRemoveDeadCodeAndDataBlocks(T_CFG(obj->code[0]->data));
	    if (CfgPatchForNonReturningFunctions(T_CFG(obj->code[0]->data)))
	      CfgRemoveDeadCodeAndDataBlocks(T_CFG(obj->code[0]->data));
	    ANALYSIS_SIMPLE_END("Removing unconnected blocks");
	  }

	if (global_options.blockprofilefile)
	  {   
	    CfgReadBlockExecutionCounts(T_CFG(obj->code[0]->data),global_options.blockprofilefile);
	    printf("START WEIGHT %d\n",CfgComputeWeight(T_CFG(obj->code[0]->data)));
	    if (global_options.insprofilefile)
	      {   
		CfgReadInsExecutionCounts(T_CFG(obj->code[0]->data),global_options.insprofilefile);
		CfgComputeHotBblThreshold(T_CFG(obj->code[0]->data),0.90);
	      }
	  }


#ifdef ARM_SUPPORT
	if (global_options.optimize)
	  MakeConstProducers(T_CFG(obj->code[0]->data));
#endif

	/*	DiabloPrint(stdout,"@eB\n@eB\n@eB\n@eB\n",T_CFG(obj->code[0]->data)->hell_node,T_CFG(obj->code[0]->data)->exit_hell_node,T_CFG(obj->code[0]->data)->call_hell_node,T_CFG(obj->code[0]->data)->exit_call_hell_node);
		exit(-1); */

	/*}}}*/
	/* Dump graphs prior to optimization {{{ */ 
	if (global_options.generate_dots)
	{
	  if (global_options.blockprofilefile)
	    CfgComputeHotBblThreshold(T_CFG(obj->code[0]->data),0.90);
	  else
	    if(global_options.annotate_loops)
	    {
	      ComDominators(T_CFG(obj->code[0]->data));
	      Export_Flowgraph(T_BBL(T_CFG(obj->code[0]->data)->entries->entry_bbl), 0xffff, "./dots/flowgraph_loop.dot");
	    }

	  ExportFunctionGraph(obj->code[0],"./dots");

	  CgBuild(T_CFG(obj->code[0]->data));
	  CgExport(T_CFG(obj->code[0]->data)->cg,"./dots/callgraph.dot");
	  

	  if (global_options.gui_output)
	  {
	    DumpLivenessInfo("./dots/LIVENESS",obj);
	    DumpConstantPropagationInfo("./dots/CONSTPROP",obj);
	  }
	}
	/* }}} */
	/* {{{ dominator analysis */
	if (global_options.dominator) 
	  {
	    ComDominators(T_CFG(obj->code[0]->data));
	  }
	/* Export dots after dominator */
	if (global_options.dominator && global_options.generate_dots)
	{
	  t_function * function;


	  DirectoryAssure("./dots-dominator",FALSE);
	  DirectoryAssure("./dots-postdominator",FALSE);

	  Export_FunctionDominator((T_CFG(obj->code[0]->data))->unique_entry_node, 0xffff, "./dots-dominator/flowgraph.dot");
	  Export_FunctionPostDominator((T_CFG(obj->code[0]->data))->unique_entry_node, 0xffff, "./dots-postdominator/flowgraph.dot");

	  FOREACH_FUNCTION(*(obj->code),function)
	  {
	    char * fname, *fname2;
	    if (function->bbl_first)
	    {
	      fname=sDiabloPrint("./dots-dominator/@G.func-%s.dot",BBL_OLD_ADDRESS(function->bbl_first),FUN_NAME(function));
	      fname2=sDiabloPrint("./dots-postdominator/@G.func-%s.dot",BBL_OLD_ADDRESS(function->bbl_first),FUN_NAME(function));
	    }
	    else
	    {
	      fname=sDiabloPrint("./dots-dominator/0x0.func-%s.dot",FUN_NAME(function));
	      fname2=sDiabloPrint("./dots-postdominator/0x0.func-%s.dot",FUN_NAME(function));
	    }
	    Export_FunctionDominator(function->bbl_first, ~ET_INTERPROC, fname);
	    Export_FunctionPostDominator(function->bbl_first, ~ET_INTERPROC, fname2);
	    Free(fname);
	    Free(fname2);
	  }
	}
	if (global_options.dominator) DominatorCleanup(T_CFG(obj->code[0]->data));
	/* }}} */
	/*{{{ ia64 stuff */
#ifdef IA64_SUPPORT       
	Ia64CacheAllocsInFunctions(T_CFG(obj->code[0]->data));
#endif       

#ifdef IA64_SUPPORT
	if(global_options.useprofile||global_options.bwh||global_options.placement||global_options.figure||global_options.fallthrough)
	  Ia64VerwerkProfile(obj);
	if(global_options.figure)
	  Ia64MaakFiguurtje(T_CFG(obj->code[0]->data));
	if(global_options.fallthrough)
	  Ia64AdaptFallthrough(T_CFG(obj->code[0]->data));
#endif
	/* }}}*/

	/*ArchCall("I386StackHeightAnalysis",SECTION_CFG(obj->code[0]));*/

	/* Analyses && Optimizations  {{{ */
	for (global_optimization_phase = 0; global_optimization_phase < 3; global_optimization_phase++)
	  if (global_options.optimize)
	  {
	    int dead_count=0;
	    t_cfg * cfg = T_CFG(obj->code[0]->data);
	    int loopcount;
#ifdef IA64_SUPPORT
	    int verschil;
#endif
	    {
	      t_bool flag =  global_options.rely_on_calling_conventions;
	      global_options.rely_on_calling_conventions = global_optimization_phase < 2;
	      if (global_options.rely_on_calling_conventions!=flag)
		{
		  printf("CALLING CONVENTIONS CHANGED!!!\n");
		}
	    }
#ifdef ARM_SUPPORT
	    if (global_optimization_phase == 2)
	      {
		SwitchMoves(cfg);
	      }
#endif
	      

	    CfgCreateExitBlockList(cfg);

#ifdef IA64_SUPPORT
	    Ia64InitializeDummies(cfg);
#endif
	    
	    /* always do this to make sure the live out sets of bbls are properly
	     * initialized, even if liveness is turned off */
	    ComLiveness(cfg,TRIVIAL);
	    
#ifdef IA64_SUPPORT
	    Ia64FinalizeDummies(cfg);
#endif

	    ObjectAction(CONVERT_DATA_TO_BLOCKS,obj);

	    loopcount = 0;
	    do
	    {
	      dead_count=0;

	      loopcount++;
	      if (global_options.remove_unconnected) CfgRemoveDeadCodeAndDataBlocks(cfg);
#ifdef IA64_SUPPORT
	      Ia64InitializeDummies(cfg);
#endif
	      if (!global_options.rely_on_calling_conventions)
		if (global_options.inlining)
		  {
		    ArchCall("InlineTrivial",cfg);
		  }

#ifndef IA64_SUPPORT
	      ComLiveness(cfg,INSENSITIVE);
	      ComputeSavedChangedRegisters(*(obj->code));
	      ComLiveness(cfg,SENSITIVE);
	      ComputeSavedChangedRegisters(*(obj->code));
#endif
	      /*ComputeDefinedRegs(cfg);*/

	      if (global_options.liveness) ComLiveness(cfg,SENSITIVE);
	      CfgDetectLeaves(cfg);

#ifndef IA64_SUPPORT
	      if (global_options.constant_propagation) 
	      {
#ifdef DIETER
#ifdef ARM_SUPPORT
		if (!global_options.rely_on_calling_conventions)
		  {
		    ConstantPropagation(cfg,NULL,FALSE);
		    printf("------------------------------  DIETER BEGIN  ------------------------------\n");
		    CfgCollectAddressProducerInformation(cfg);
		    if(loopcount <= 1)
		      {
			ArmOptimizeLayoutForAddresses(obj);
		      }
		    // CfgOptimizeLoaders(cfg);
		    FreeAddressProducerInformation(cfg);
		    printf("-------------------------------  DIETER END  -------------------------------\n");
		    FreeConstantInformation(cfg);
		  }
#endif
#endif
		ConstantPropagation(cfg,NULL,FALSE);
		OptUseConstantInformation(cfg,NULL,FALSE);
		FreeConstantInformation(cfg);
	      }

#else
	      if (global_options.constant_propagation) 
	      {
		Ia64ResetEmulationStatistics();
		CIConstantPropagation(obj,obj->code[0],NULL,FALSE);
		Ia64PrintEmulationStatistics();
	      }
#endif

	      ComputeSavedChangedRegisters(*(obj->code));
	      if (global_options.liveness)ComLiveness(cfg,SENSITIVE);

	      if (global_options.liveness)dead_count += OptKillUseless(cfg);
	      ComputeSavedChangedRegisters(*(obj->code));

	      if (global_options.liveness) ComLiveness(cfg,SENSITIVE);
	      if (global_options.liveness)dead_count += OptKillUseless(cfg);

	      ComputeSavedChangedRegisters(*(obj->code));

	      if (global_options.liveness) ComLiveness(cfg,SENSITIVE);
	      if (global_options.liveness) dead_count += OptKillUseless(cfg);

	      if (global_options.dominator && global_options.loop_invariant_code_motion) 
		{
		  if (global_options.blockprofilefile)
		    {
		      CfgComputeHotBblThreshold(T_CFG(obj->code[0]->data),0.90);
		      CfgEstimateEdgeCounts(cfg);
		      
		      ComDominators(T_CFG(obj->code[0]->data));
		      CfgHoistConstantProducingCode(T_CFG(obj->code[0]->data));

#ifdef SPEEDUP_DOMINATORS
		      if (dominator_info_correct == FALSE)
#endif
			DominatorCleanup(T_CFG(obj->code[0]->data));
		      ComLiveness(cfg,SENSITIVE);
		      CfgEstimateEdgeCounts(cfg);
		    }

#ifdef SPEEDUP_DOMINATORS
		  if (dominator_info_correct == FALSE)
#endif
		    ComDominators(T_CFG(obj->code[0]->data));
		  CfgHoistConstantProducingCode3(T_CFG(obj->code[0]->data));
#ifdef SPEEDUP_DOMINATORS
		  if (dominator_info_correct == FALSE)
#endif
		    DominatorCleanup(T_CFG(obj->code[0]->data));
		  ComLiveness(cfg,SENSITIVE);
		  
#ifdef SPEEDUP_DOMINATORS
		  if (dominator_info_correct == FALSE)
#endif
		    ComDominators(T_CFG(obj->code[0]->data));
		  DetectLoopStackSubAdds(cfg);
		  ComLiveness(cfg,SENSITIVE);
		  DetectLoopInvariants(cfg);
#ifdef SPEEDUP_DOMINATORS
		  if (dominator_info_correct == FALSE)
#endif
		    DominatorCleanup(T_CFG(obj->code[0]->data));
		  ComLiveness(cfg,SENSITIVE);
		  
		  if (global_options.blockprofilefile)
		    {
		      CfgEstimateEdgeCounts(cfg);
		      CfgComputeHotBblThreshold(T_CFG(obj->code[0]->data),0.85);
#ifdef SPEEDUP_DOMINATORS
		      if (dominator_info_correct == FALSE)
#endif
			ComDominators(T_CFG(obj->code[0]->data));
		      LoopUnrollingSimple(T_CFG(obj->code[0]->data));
		      DominatorCleanup(T_CFG(obj->code[0]->data));
		      ComLiveness(cfg,SENSITIVE);
		    }
		}

	      if (global_options.copy_analysis) CopyAnalysis(*(obj->code));

/*#ifdef I386_SUPPORT*/
	      /*I386CheckFPUse(cfg);*/
/*#endif*/
	      if (global_options.copy_propagation) ArchCall("OptCopyPropagation",cfg); /* only does something on the ARM */

	      TraceBblAtAddress(cfg,global_options.traceadr,"na_copy_propagatie");

	      if (global_options.loadstorefwd) ArchCall("ArmLoadStoreFwd",cfg);
	      TraceBblAtAddress(cfg,global_options.traceadr,"HIER");
	      if (global_options.copy_analysis) CopyAnalysisFini(*(obj->code));

	      CfgRemoveUselessConditionalJumps(cfg);
	      CfgRemoveEmptyBlocks(cfg);

	      TraceBblAtAddress(cfg,global_options.traceadr,"HIER");
	      if (global_options.remove_unconnected) CfgRemoveDeadCodeAndDataBlocks(cfg);
	      CfgMoveInsDown(cfg);
	      ComputeSavedChangedRegisters(*(obj->code));
	      TraceBblAtAddress(cfg,global_options.traceadr,"HIER");

	      if (global_options.liveness) ComLiveness(cfg,SENSITIVE);
	      if (global_options.remove_unconnected) CfgRemoveDeadCodeAndDataBlocks(cfg);

	      TraceBblAtAddress(cfg,global_options.traceadr,"HIER");
	      if (global_options.branch_elimination) ArchCall("BranchForwarding",cfg);
	      if (global_options.branch_elimination) ArchCall("BranchElimination",cfg);
	      if (global_options.mergebbls) ArchCall("MergeBbls",cfg);
	      if (global_options.pre)
		{
		  if (global_options.blockprofilefile)
		    {
		      CfgEstimateEdgeCounts(cfg);
		      PartialRedundancyElimination1(cfg);
		      PartialRedundancyElimination1ForCalls(cfg);
		      PartialRedundancyElimination1ForReturns(cfg);
		    } 

		  ComLiveness(cfg,TRIVIAL);
		  ComputeSavedChangedRegisters(*(obj->code));
		  ComLiveness(cfg,INSENSITIVE);
		  ComputeSavedChangedRegisters(*(obj->code));
		  ComLiveness(cfg,SENSITIVE);

		  ReplaceTriangleWithConditionalMove(cfg);

		  if (global_options.blockprofilefile)
		    {
		      PartialRedundancyElimination2(cfg);
		      PartialRedundancyElimination3(cfg);

		      if (global_options.insprofilefile)
			{   
			  CfgComputeHotBblThreshold(T_CFG(obj->code[0]->data),0.95);
			  DetectColdCodeBundles(T_CFG(obj->code[0]->data));
			}
		      CfgEstimateEdgeCounts(cfg);
		  
		      CfgBranchSwitch(cfg);
		      CfgBranchSwitch2(cfg);
		      CfgComputeHotBblThreshold(T_CFG(obj->code[0]->data),0.90);
		      CfgEstimateEdgeCounts(cfg);
		    }
		}
	      
	      if (!global_options.rely_on_calling_conventions)
		{
		  if (global_options.inlining)
		    {
		      ArchCall("GeneralInlining",cfg); /* Arm only */
		      ArchCall("I386InlineSimple",cfg);
		      /*ArchCall("I386InlineFunctionsWithOneCallSite",cfg);*/
		    }
		}
		      
	      ArchCall("I386PeepHoles",T_CFG(obj->code[0]->data));

	      if (global_options.peephole) ArchCall("ArmPeepholeOptimizations",T_CFG(obj->code[0]->data));
	      TraceBblAtAddress(cfg,global_options.traceadr,"HIER");
	      
#ifdef IA64_SUPPORT
	      /* {{{ IA64 stuff */
	      if (global_options.branchregs && global_options.branchregs2)
		FATAL(("Optimizations -Obr and -Obr2 cannot be used together !"));
	      if(global_options.branchregs) Ia64OptimizeBranchRegisters(cfg); 
	      if(global_options.branchregs2) Ia64OptimizeBranchRegisters2(cfg); 

	      if (global_options.inlining) 
	      {
		Ia64Inlining(cfg);
		Ia64FinalizeDummies(cfg);
		Ia64InitializeDummies(cfg);
	      }

	      if(global_options.gpusage) Ia64OptimizeGpUsage(cfg); 

	      do
	      { 
		if (global_options.liveness)
		{ 
		  ComLiveness(cfg,SENSITIVE);
		  verschil = OptKillUseless(cfg);
		  dead_count += verschil; 		
		}
	      }while (verschil);

	      if (global_options.copy_propagation) Ia64OptCopyPropagation(cfg);
	      /* }}} */
#endif
	      
	      if (global_options.stack) ArchCall("OptimizeStackLoadAndStores",cfg); /* only for the ARM */
	      /*ArchCall("InstructieOptimalisatie",cfg); |+ only for the MIPS +|*/
	      ComputeSavedChangedRegisters(*(obj->code));

	      if (global_options.liveness) ComLiveness(cfg,SENSITIVE);
	      if (global_options.liveness) dead_count += OptKillUseless(cfg);

	      TraceBblAtAddress(cfg,global_options.traceadr,"HIER");

#ifdef IA64_SUPPORT
	      Ia64FinalizeDummies(cfg);
#endif
	    } while(dead_count);
	  }
#ifdef OBFUSCATION_SUPPORT
	if(!strcmp(global_options.obfuscation,"branch_obf"))
	{
	  ComLiveness(T_CFG(obj->code[0]->data),SENSITIVE);      
	  BranchObfuscation(T_CFG(obj->code[0]->data));
	}
#endif
#ifdef IA64_SUPPORT
	if(global_options.bwh)
	  Ia64AdaptBranch(T_CFG(obj->code[0]->data));
#endif 
	/* End of Analyses && Optimizations  }}} */
 
#ifdef FACTOR_SUPPORT
        ComLiveness(T_CFG(obj->code[0]->data),SENSITIVE);
        Factor(T_CFG(obj->code[0]->data));
#endif

	if (global_options.blockprofilefile)
	  printf("END WEIGHT %d\n",CfgComputeWeight(T_CFG(obj->code[0]->data)));

	/* Export dots after optimzation {{{  */
	{
	  t_function * function;
	  if (global_options.generate_dots)
	  {
	    DirectoryAssure("./dots-final",FALSE);
	    
	    if (global_options.blockprofilefile)
	      CfgComputeHotBblThreshold(T_CFG(obj->code[0]->data),0.90);

	    FOREACH_FUNCTION(*(obj->code),function)
	      {
	      char * fname;
	      if (FunIsHot(function) && global_options.blockprofilefile)
		{
		  if (function->bbl_first)
		    fname=sDiabloPrint("./dots-final/@G.func-%s.HOT.dot",BBL_OLD_ADDRESS(function->bbl_first),FUN_NAME(function));
		  else
		    fname=sDiabloPrint("./dots-final/0x0.func-%s.HOT.dot",FUN_NAME(function));
		}
	      else 
		{
		  if (function->bbl_first)
		    fname=sDiabloPrint("./dots-final/@G.func-%s.dot",BBL_OLD_ADDRESS(function->bbl_first),FUN_NAME(function));
		  else
		    fname=sDiabloPrint("./dots-final/0x0.func-%s.dot",FUN_NAME(function));
		}
	      /*Export_Function(function->bbl_first, ~ET_INTERPROC, fname);*/
	      Export_Function(function,fname);
	      Free(fname);
	    }

	    CgBuild(T_CFG(obj->code[0]->data));
	    CgExport(T_CFG(obj->code[0]->data)->cg,"./dots-final/callgraph.dot");


	  }
	  if (global_options.gui_output)
	  {
	    DumpLivenessInfo("./dots-final/LIVENESS",obj);
	    DumpConstantPropagationInfo("./dots-final/CONSTPROP",obj);
	  }
	}
	/* }}} */
/* {{{ more ia64 stuff */
#ifdef IA64_SUPPORT
	/*       if(global_options.useprofile)
		 Ia64ProfilePrintHotBbls(T_CFG(obj->code[0]->data));
	*/

	if(global_options.makeprofile)
	  Ia64MakeProfile(obj);
#endif
	/* }}}*/
#ifdef WATERMARK
	if(!strcmp(global_options.watermark,"data_basic"))	
	  WaterDataBasicAddMark(obj,"Copyright");
	if(!strcmp(global_options.watermark,"opaque"))
	  WaterOpaqueAddMark(obj,"Copyright");
#endif

#ifdef OBFUSCATION_SUPPORT
	if(global_options.obf_dis)
	{
	  DisassembleObf(T_CFG(obj->code[0]->data));
	}
#endif
	
	/* Back to the section representation (still disassembled though) */
#ifdef BJORN
	TryFindingNonLoopEdges(T_CFG(obj->code[0]->data));
#endif
	ObjectAction(DEFLOWGRAPH,obj);  

	/* rebuild the layout of the data sections
	 * so that every subsection sits at it's new address */
	ObjectRebuildSectionsFromSubsections(obj);
      } /*  }}} */

      if (global_options.saveState)
	SaveState(obj,SAVESTATEDISASSEMBLED);

      ObjectAction(ASSEMBLE,obj); 

#ifdef KERNEL_OPT
      if (!global_options.flowgraph)
      {
	/*This is just temporary, until flowgraphing works*/
	SeparateCodeSections(obj);
      }
#endif
      /* End Transform and optimize }}} */ 
    }
    else if (global_options.dump || global_options.dump_multiple)
    {
#ifdef HAVE_DIABLO_EXEC
      if (global_options.exec) DiabloExec();
#else
      if (global_options.exec) FATAL(("Exec not compiled!"));
#endif
      DumpProgram(obj,global_options.dump_multiple);
      exit(0);
    }
    else if (global_options.saveState)
    {
      SaveState(obj,SAVESTATENOTDISASSEMBLED); 
    }



    ObjectAction(WRITE,obj,global_options.output_name);  

#ifdef HAVE_STAT
  /* make the file executable */
  chmod(global_options.output_name,S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
#endif

  }
  /* END REAL program }}} */

  if (global_options.exec)
  {
#ifdef HAVE_DIABLO_EXEC
    DiabloExec();
#else
    FATAL(("Exec not compiled!"));
#endif
  }
  /* IV. Free all structures {{{ */
  /* remove directory that held the restored dump */
  /*if (global_options.restore) RmdirR("./DUMP");*/
  /* Deallocate global error variables (we can still use
   * Error (e.g. for leak detection, but the error
   * message will no longer contain the program name */ 
  ErrorFini();
  /* Free all the paths that are stored in the options and
   * the default objectname: The rest of the options (like
   * the verbose-level remains the same*/
  OptionsFini();
  /* Free the archive cache */
  ArchivesFini();
  /* Free the object cache */
  /* XXX: this is an ugly hack, but turning off ObjectFini if we saved the state or haven't flowgraphed fixes a messy segfault*/
#ifndef I386_SUPPORT
  if (global_options.flowgraph)
#endif
  ObjectFini();
  /* Free initializers used in path expansion (such as the
   * users home directory */
  PathFini();   
  /* If DEBUG MALLOC is defined print a list of all memory
   * leaks. Needs to increase the verbose-level to make
   * sure these get printed. This is a hack but it may be
   * removed when a real version of diablo is made */ 

#ifdef DEBUG_MALLOC
  global_options.verbose++;
  PrintRemainingBlocks();
#endif
  /* End Fini }}} */
  return 0;
}
/* KERNEL {{{ */
#ifdef KERNEL_OPT
/* {{{ this is just a temporary fix to split the code sections 
 * for a kernel before writing the binary */
static void SeparateCodeSections(t_object * obj)
{
  t_object * linker_subobject;
  t_section * merged, * code, * init, * code_sub, * init_sub;
  t_address offset, initsz, codesz;
  t_reloc * rel;
  int i;

  merged = obj->code[0];
  printf("splitting section %s\n", merged->name);

  for (i = 0; i < obj->n_orig_codes; i++)
    if (!strcmp(obj->orig_codes[i].name,".text.init"))
      break;

  offset = GENERIC_SUB(obj->orig_codes[i].base,SECTION_GET_OLD_ADDRESS(merged));
  /* make sure the sizes add up */
  /*VERIFY_FASSERT(GENERIC_EQ(SECTION_CSIZE(merged),GENERIC_ADD(offset,obj->orig_codes[i].size)),*/
      /*("Section sizes don't add up: offset @G + size @G = @G, expected @G",*/
       /*offset,obj->orig_codes[i].size,GENERIC_ADD(offset,obj->orig_codes[i].size),SECTION_CSIZE(merged)));*/

  /*initsz = obj->orig_codes[i].size;*/
  /*codesz = GENERIC_SUB(SECTION_CSIZE(merged),initsz);*/
  initsz = GENERIC_SUB(SECTION_CSIZE(merged),offset);
  codesz = offset;

  /* {{{ create new section for .text.init */
  init = SectionNew(obj,CODE_SECTION);
  SECTION_OLD_SIZE(init) = initsz;
  SECTION_CSIZE(init) = initsz;
  SectionSetCAddressSuper(init,GENERIC_ADD(SECTION_GET_CADDRESS(merged),offset));
  SectionSetOldAddressSuper(init,obj->orig_codes[1].base);
  init->name = StringDup(".text.init");
  init->data = Calloc(sizeof(char),GENERIC_EXTRACT_UINT32(initsz));
  HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(init->name,init));
  /* insert in the section hash table. the node will probably exist already,
   * so we just have to add the section to the node */
  {
    t_section_hash_node * shn = HashLookup(obj->section_table, ".text.init");
    if (!shn)
      HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(init->name,init));
    else if (!shn->section)
      shn->section = init;
    else
      FATAL((".text.init still existed"));
  }
  memcpy(init->data, ((char *)merged->data) + GENERIC_EXTRACT_UINT32(offset), GENERIC_EXTRACT_UINT32(initsz));
  /* }}} */

  /* {{{ create new section for .text */
  code = SectionNew(obj,CODE_SECTION);
  SECTION_OLD_SIZE(code) = codesz;
  SECTION_CSIZE(code) = codesz;
  SectionSetCAddressSuper(code,SECTION_GET_CADDRESS(merged));
  SectionSetOldAddressSuper(code,obj->orig_codes[0].base);
  code->name = StringDup(".text");
  code->data = Calloc(sizeof(char),GENERIC_EXTRACT_UINT32(codesz));
  /* insert in the section hash table. the node will probably exist already,
   * so we just have to add the section to the node */
  {
    t_section_hash_node * shn = HashLookup(obj->section_table, ".text");
    if (!shn)
      HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew(code->name,code));
    else if (!shn->section)
      shn->section = code;
    else
      FATAL((".text still existed"));
  }
  memcpy(code->data, merged->data, GENERIC_EXTRACT_UINT32(codesz));
  /* }}} */

  /* Get or create the linker sub object {{{ */
  linker_subobject=ObjectGetFromCache("Linker",obj);
  if (!linker_subobject) 
  {
    linker_subobject=ObjectNewCached("Linker",obj);
    linker_subobject->section_table=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,SectionHashNodeFree);
    obj->mapped_last->next=linker_subobject;
    linker_subobject->next=NULL;
    obj->mapped_last=linker_subobject;
    linker_subobject->reloc_table=RelocTableNew(); 
    linker_subobject->symbol_table=SymbolTableNew(linker_subobject);
  }
  /* }}} */

  /* {{{ create new subsection for .text.init */
  init_sub = SectionNew(linker_subobject,CODE_SECTION);
  RELOCATABLE_TYPE(init_sub) = RT_SUBSECTION;
  SECTION_PARENT_SECTION(init_sub) = init;
  SECTION_OLD_SIZE(init_sub) = initsz;
  SECTION_CSIZE(init_sub) = initsz;
  SectionSetCAddressSuper(init_sub,GENERIC_ADD(SECTION_GET_CADDRESS(merged),offset));
  SectionSetOldAddressSuper(init_sub,obj->orig_codes[1].base);
  init_sub->name = StringDup(".text.init");
  init_sub->data = Calloc(sizeof(char),GENERIC_EXTRACT_UINT32(initsz));
  memcpy(init_sub->data, ((char *)merged->data) + GENERIC_EXTRACT_UINT32(offset), GENERIC_EXTRACT_UINT32(initsz));
  linker_subobject->ncodes++;
  linker_subobject->code = Realloc(linker_subobject->code, linker_subobject->ncodes);
  linker_subobject->code[linker_subobject->ncodes-1] = init_sub;
  /* }}} */

  /* {{{ create new subsection for .text */
  code_sub = SectionNew(linker_subobject,CODE_SECTION);
  RELOCATABLE_TYPE(code_sub) = RT_SUBSECTION;
  SECTION_PARENT_SECTION(code_sub) = code;
  SECTION_OLD_SIZE(code_sub) = codesz;
  SECTION_CSIZE(code_sub) = codesz;
  SectionSetCAddressSuper(code_sub,SECTION_GET_CADDRESS(merged));
  SectionSetOldAddressSuper(code_sub,obj->orig_codes[0].base);
  code_sub->name = StringDup(".text");
  code_sub->data = Calloc(sizeof(char),GENERIC_EXTRACT_UINT32(codesz));
  memcpy(code_sub->data, merged->data, GENERIC_EXTRACT_UINT32(codesz));
  linker_subobject->ncodes++;
  linker_subobject->code = Realloc(linker_subobject->code, linker_subobject->ncodes);
  linker_subobject->code[linker_subobject->ncodes-1] = code_sub;
  /* }}} */

  /* {{{ migrate the relocations to the new code subsections */
  OBJECT_FOREACH_RELOC(obj,rel)
  {
    /* the to's */
    if (REL_TO_RELOCATABLE(rel) == merged)
    {
      if (GENERIC_LT(rel->to_offset,offset))
      {
	REL_TO_RELOCATABLE(rel) = code_sub;
      }
      else
      {
	REL_TO_RELOCATABLE(rel) = init_sub;
	rel->to_offset = GENERIC_SUB(rel->to_offset,offset);
      }
    }

    /* the froms */
    if (REL_FROM(rel) == merged)
    {
      if (GENERIC_LT(rel->from_offset,offset))
      {
	REL_FROM(rel) = code_sub;
      }
      else
      {
	REL_FROM(rel) = init_sub;
	rel->from_offset = GENERIC_SUB(rel->from_offset,offset);
      }
    }
  }
  /* }}} */

  /* kill the old merged code section */
  SectionFree(obj->code[0]);
  obj->ncodes = 2;
  Free(obj->code);
  obj->code = Malloc(sizeof(t_section *)*2);
  obj->code[0] = code;
  obj->code[1] = init;


  PlaceSections(obj,FALSE);
  OBJECT_FOREACH_RELOC(obj,rel)
  {
    t_address ret = StackExec(rel->code,rel,T_SECTION(REL_FROM(rel))->data,TRUE,0);
    VERIFY_FASSERT(GENERICISNULL(ret),("Could not execute @R (code %s)",rel,rel->code));
  }

  ObjectRebuildSectionsFromSubsections(obj);
} /* }}} */
#endif
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker:*/

