/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//* $Header: /home/cvs/cvsroot/diablo/utils/util_error.c,v 1.2 2004/01/08 14:41:36 bdebus Exp $ */
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include "../config.h"
#include "util_error.h"
#include "util_malloc.h"
#include "../diablo_options.h"
#include "../datastructures/data_string.h"
#include "../kernel/diablo_print.h"
#include "util_options.h"
#include "util_term.h"
#ifdef COLORS
#include "util_color.h"
#endif

static char * execname=NULL;
static FILE * err_channel=NULL;
static void (*usage_cb)(void)=NULL; 
int warnings=0;
int last_warning=0;


void PrintExecName(FILE * fp)
{
   if (execname) fprintf(fp,"%s",execname);
}

void ErrorClearErrno()
{
   errno=0;
}

t_int32 ErrorGetErrno()
{
   return errno;
}

void ErrorSetErrno(t_int32 in)
{
   errno=in;
}

void Signal(int sig)
{
   FATAL(("Got signal %d!",sig));
}

void ErrorInit(char * fname, FILE * err, void (*usage)(void))
{
    if (execname) Free(execname);
    execname=StringDup(fname);
    if (err) err_channel=err;
    else if (!err_channel) err_channel=stderr;
    if (usage) usage_cb=usage;
    ErrorClearErrno();
    warnings=0;
    signal(SIGINT,Signal);
}

void ErrorFini()
{
   if (warnings>0)
   {
#ifdef COLORS
    ColorFPrint(normal,red,black,0,0,err_channel,"WARNING SUMMARY: ");
#else
    fprintf(err_channel,"WARNING SUMMARY: ");
#endif
    fprintf(err_channel,"Had %d warnings\n",warnings);
   }
   Free(execname);
   execname=NULL;
}

int PreError(t_uint32 type, char * file, t_uint32 lineno)
{
    int run=0;
    if (!err_channel) err_channel=stderr;
#ifdef COLORS
    if (type&E_FATAL) run=ColorFPrint(normal,red,black,0,0,err_channel,"FATAL:   ");
    if (type&E_WARNING) if (global_options.verbose) run=ColorFPrint(normal,red,black,0,0,err_channel,"WARNING: ");
#else
    if (type&E_FATAL) run=TermPrintClean(err_channel,"FATAL:   ",0,0);
    if (type&E_WARNING) if (global_options.verbose) run=TermPrintClean(err_channel,"WARNING: ",0,0);
#endif


    if (type&E_WARNING) 
    {
       warnings++;
       if (!(global_options.verbose))
	  return 0;
    }

    if (execname)
    {
       run=TermPrintClean(err_channel, "[ ",run,9,0);
       run=TermPrintClean(err_channel, execname,run,9,0);
       run=TermPrintClean(err_channel, "]: File",run,9,0);
    }
    else
    {
       run=TermPrintClean(err_channel, "In file",run,9,0);
    }
#ifdef COLORS
	run=ColorFPrint(normal,green,black,run,9,err_channel," %s",file);
#else
        run=TermPrintClean(err_channel,file,9,0);
#endif
	run=TermPrintClean(err_channel, " line",run,9,0);
#ifdef COLORS
	run=ColorFPrint(normal,green,black,run,9,err_channel," %d: ",lineno);
#else
	fprintf(err_channel,"%d: ",lineno);
#endif
	return run;
}

void PostError(t_uint32 type)
{
   if ((type&E_USAGE)&&(usage_cb)) 
   {
      usage_cb();
   }
   if (type&E_FATAL) { abort(); exit(-2);} 
}

int errcurs=0;
void Error(char * errormsg, ...)
{
   t_string buffer;
   va_list val;

   va_start(val,errormsg);
   buffer=vsDiabloPrint(errormsg,val);
   TermPrintClean(err_channel,buffer,errcurs,9,FINAL);

   Free(buffer);
    if (errno)
    {
#ifdef COLORS
       ColorFPrint(normal,red,black,0,0,err_channel,"Errno is set: ");
#else
       fprintf(err_channel,"Errno is set: ");
#endif
       if (execname) perror(execname);
       else perror("ERROR");
    }
}

