/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../config.h"
#include "util_error.h"
#include "util_malloc.h"
#include "../datastructures/data_string.h"
#include "../kernel/diablo_archives.h"
#include "../diablo_options.h"
#ifdef HAVE_STAT
#include <sys/stat.h> 
#include <sys/wait.h> 
#endif
#ifdef VISUALSTUDIO
#include <direct.h>
#endif
#include <libgen.h> 
#include <string.h> 
#include <limits.h> 
#ifdef HAVE_UNISTD
#include <unistd.h> 
#endif
#include <stdlib.h> 
#include "util_file.h"
#include "util_options.h"

static char * home=NULL;
/* if we do not have a path length take a value */
#ifndef HAVE_GETCWD 
#define PATH_MAX 3 
#endif
static char cp[PATH_MAX];

void PathInit()
{
   char * ret=getenv("HOME");

   if (ret)
   {
      home=StringDup(ret);
   }
   else
   {
      home=NULL;
   }
#ifdef HAVE_GETCWD
   if (!getcwd(cp,PATH_MAX))
   {
      FATAL(("Could not get current working directory"));
   }
#else
   strcpy(cp,"./");
#endif
}

void PathFini()
{
   Free(home);
   home=NULL;
}


t_string ConvertFileName(t_string filename)
{
   if ((filename[0]=='~')&&(filename[1]=='/'))
   {
      if (!home) FATAL(("HOME-directory not found! Use absolute paths!"));
      return Collate3(home,"/",filename+2);
   }
   else if ((filename[0]=='.')&&(filename[1]=='/'))
   {
      return Collate3(cp,"/",filename+2);
   }
   else if ((filename[0]=='.')&&(filename[1]=='.')&&(filename[2]=='/'))
   {
      return Collate3(cp,"/",filename);
   }
   else
   {
      return StringDup(filename);
   }
}

/* Convert a delim separated file into a path structure */ 

t_path * ParsePath(t_string path, t_string delim)
{
   t_path * ret=Malloc(sizeof(t_path));
   t_string path_copy=StringDup(path),tmp;
   t_string s=strtok(path_copy,delim);
#ifdef HAVE_STAT
   struct stat st;
#endif
   t_uint32 i,found=0,alloc=0;

   
   ret->dirs=NULL;
   ret->nelems=0;
   while(s)
   {
      i=strlen(s)-1;
      while((i!=0) && (s[i]=='/')) s[i--]=0; /* Strip pending / */
      
      alloc=0;
      if ((s[0]=='~') && (s[1]=='/'))
      {
	 alloc=1;
	 s=Collate3(home,"/",s+2);
      } 
      else if (s[0]!='/')
      {
	alloc=1;
	s=Collate3(cp,"/",s);
      }

      if (s[0]!='/') FATAL(("Still have relative paths!"));

      i=strlen(s)-1;
      while((i!=0) && ((s[i]=='/') || ((i>0) && (s[i]=='.') && (s[i-1]!='.')))) s[i--]=0; /* Strip pending / and .*/
      
      if (global_options.prepend_set) tmp=Collate(global_options.prepend,s);
      else tmp=Collate("",s);

#ifdef HAVE_STAT
      if (stat(tmp,&st))
      {
	 if ((!global_options.restore) && (global_options.restore_multi==-1)) FATAL(("Couldn't stat %s",tmp));
	 else 
	 {
	 }
      }
      else if (!(S_ISDIR(st.st_mode)))
      {
	 WARNING(("%s is not a directory",tmp));
      }
#endif
      ErrorClearErrno();

      
      found=0;
      for (i=0; i<ret->nelems; i++)
      {
	 if (strcmp(ret->dirs[i],tmp)==0)
	 {
	    found=1;
	    break;
	 }
      }
      if (!found)
      {
	 ret->dirs=(t_string *) Realloc(ret->dirs, sizeof(t_string) * (ret->nelems+1));
	 ret->dirs[ret->nelems]=StringDup(tmp);
	 ret->nelems++;
      }
      if (alloc) Free(s);
      Free(tmp);
      s=strtok(NULL,delim);
   }

   Free(path_copy);
   return ret;
}

void PathFree(t_path * p)
{
   t_uint32 tel;
   for (tel=0; tel<p->nelems; tel++)
   {
      Free(p->dirs[tel]);
   }
   Free(p->dirs);
   Free(p);
}

t_string FindFile(t_path * path, t_string name)
{
   t_uint32 tel;
   t_string tmp;
#ifdef HAVE_STAT
   struct stat s;
#else
   FILE * fp;
#endif

  tmp=Collate(global_options.prepend,name);
   if ((name[0]=='/')
#ifdef HAVE_STAT
      && (!stat(tmp,&s)))
      {
#else
      && ((fp=fopen(tmp,"r"))!=NULL))
      {
	 fclose(fp);
#endif

      ErrorClearErrno();
      return tmp;
   }
   else if (name[0]!='/')
   {
     Free(tmp);
      ErrorClearErrno();
      for (tel=0; tel<path->nelems; tel++)
      {
	 tmp=Collate3(path->dirs[tel],"/",name);
#ifdef HAVE_STAT
	 if (!stat(tmp,&s))
	 {
#else
	 if ((fp=fopen(tmp,"r"))!=NULL)
	 {
	    fclose(fp);

#endif
	    ErrorClearErrno();
	    return tmp;
	 }
	 else
	 {
	    ErrorClearErrno();
	    Free(tmp);
	 }
      }
   }

#if 0

   if (name[0]=='/')
   {
      if ((PatternMatch("*/*",name+1))&&(!SubdirsExist(name))) 
      {
	 t_archive * ar;
	 t_string dir, arch, file;
	 if (HighestExistingSubdir(name,&dir,&arch,&file))
	 {
	    t_string archname=Collate3(dir,"/",arch);
	    t_string try=Collate(archname,".jar");
	    ar=ArchiveGet(try);

	    if (ar)
	    {
	       tmp=Collate3(try,":",file);
	       Free(arch); Free(dir); Free(archname); Free(try);
	       return tmp;
	    }
	 }
      }
   }
   else
   {
      for (tel=0; tel<path->nelems; tel++)
      {
	 tmp=Collate3(path->dirs[tel],"/",name);
	 if ((PatternMatch("*/*",tmp+1))&&(!SubdirsExist(tmp)))
	 {
	    t_archive * ar;
	    t_string dir, arch, file;
	    if (HighestExistingSubdir(tmp,&dir,&arch,&file))
	    {
	       t_string archname=Collate3(dir,"/",arch);
	       t_string try=Collate(archname,".jar");
	       ar=ArchiveGet(try);

	       if (ar)
	       {
		  FATAL(("implement object checking!"));
		  Free(tmp);
		  tmp=Collate3(try,":",file);
		  Free(arch); Free(dir); Free(archname); Free(try);
		  return tmp;
	       }
	       else
	       {
		  Free(tmp);
	       }
	    }

	 }
      }

   }
#endif

   ErrorClearErrno();
   return NULL;
}

char * Basename(char * in)
{
#ifdef VISUALSTUDIO
	FATAL(("Implement"));
	return NULL;
#else
	return basename(in);
#endif
}

#ifdef HAVE_MKDIR
void DirectoryAssure(const t_string path,t_bool tmp)
{
#ifdef HAVE_STAT
   struct stat st;
   if (stat(path,&st))
   {
      if (ErrorGetErrno()==ENOENT)
      {
#endif
#ifdef VISUALSTUDIO
	 int ret=mkdir(path);
#else
     int ret=mkdir(path,0700);
#endif
	 VERIFY_FASSERT(!ret,("Could not make directory %s",path));
#ifdef HAVE_STAT
      }
      else
      {
	 FATAL(("Unexpected error assuring directory %s\n",path));
      }
   }
   else
   {
      VERIFY_FASSERT(S_ISDIR(st.st_mode),("%s is not a directory, remove regular file!",path));
   }
   if (tmp)
   {
      /* TODO: Add it to the tmp list */
   }
#endif
   ErrorClearErrno();
}
#endif

/* retrieves a line from file. Alternative to fgets(), since one line must be read
   in, no matter its length. Returns NULL if EOF is encountered right from the
   start. Memory needs to be freed! */
t_string RealGetLine(char * file, int lnno, FILE *ifp)
{
  char buffer[CONFIG_MAX_FILE_LINE_LEN];
  t_int16 i = 0;
  t_string ret;

  buffer[i] = getc(ifp);
  while(( buffer[i] != EOF ) && (!feof(ifp)))
  {
    i++;
    if (buffer[i-1] == '\n') break;
    VERIFY_FASSERT(i != CONFIG_MAX_FILE_LINE_LEN, ("Increase parsebuffersize!, strings is %s",buffer));
    buffer[i] = getc(ifp);
  }
  if (i == 0) return NULL;
  buffer[i] = '\0';
#ifdef DEBUG_MALLOC
  ret = RealStringDup(file,lnno,buffer);
#else
  ret = RealStringDup(buffer);
#endif
  return ret;
}


t_bool SubdirsExist(t_string filename)
{
#ifdef VISUALSTUDIO
FATAL(("Implement SubdirsExist for windows"));
#else
   t_string full=ConvertFileName(filename);
   t_uint32 len=strlen(full);
   struct stat st;

   while(--len) if (full[len]=='/') break; else full[len]=0;

   if ((!stat(full,&st))&& (!(S_ISDIR(st.st_mode))))
   {
      Free(full);
      return TRUE;
   }

   errno=0;
   Free(full);
#endif
   return FALSE;
}

t_bool HighestExistingSubdir(t_string filename, t_string * dirret, t_string *archret, t_string * fileret)
{
#ifdef VISUALSTUDIO
	FATAL(("Implement HighestExistingSubdir for windows\n"));
#else
   t_string full=ConvertFileName(filename);
   t_uint32 len=strlen(full),prev=-1;
   char copy=0;
   struct stat st;

   while(len)
   {
      prev=len;
      while(--len) if (full[len]=='/') { full[len+1]=copy; copy=full[len]; full[len]=0; break; } else { full[len+1]=copy; copy=full[len]; full[len]=0; };


      if (stat(full,&st))
      {
	 errno=0;
	 continue;
      }
      else if (!(S_ISDIR(st.st_mode)))
      {
	 continue;
      }
      else break;
   }

   if (full[prev+1]==0) { Free(full); return FALSE; }
   if (!len) { Free(full); return FALSE; }
   *dirret=StringDup(full);
   full[prev]=0;
   *archret=StringDup(full+len+1);
   *fileret=StringDup(full+prev+1);
   Free(full);
#endif
   return TRUE;
}

t_bool FileExists(t_string name)
{
#ifdef HAVE_STAT
   struct stat st;
#else
   FILE * fp;
#endif
#ifdef HAVE_STAT
      if (stat(name,&st))
      {
	 return FALSE;
      }
      else
      {
	 return TRUE;
      }
#else
      if (fp=fopen(name,"r"))
      {
	 fclose(fp);
	 return TRUE;
      }
      else
      {
	 return FALSE;
      }
#endif
}

void SmartCopy(t_string from, t_string to)
{
#ifdef VISUALSTUDIO
	FATAL(("Implement smart copy for windows"));
#else
   pid_t pid;
   char * dest=StringDup(to);
   char * one_dir=NULL;
   char * full_dir=StringDup("");
   one_dir=(char *)strtok(dest,"/");
   while(one_dir)
   {
      char * sub_dir=Collate3(full_dir,one_dir,"/");
      one_dir=(char *)strtok(NULL,"/") ;
      if (one_dir) DirectoryAssure(sub_dir,FALSE);
      Free(full_dir);
      full_dir=sub_dir;
   }
   Free(full_dir);
   Free(dest);

   if (!FileExists(to))
   {
      if ((pid=fork())==0)
      {
	 execlp("cp","cp",from,to,NULL);
	 exit(0);
      }
      else
      {
	 waitpid(pid,NULL,0); /* Else we could overrun the max process limit */
	 /* TODO: Add checks here */
      }
   }
#endif
}
void RmdirR(t_string dir)
{
#ifdef VISUALSTUDIO
	FATAL(("Implement Rmdir for windows"));
#else
   pid_t pid;
   if ((pid=fork())==0)
   {
      execlp("rm","rm","-R",dir,NULL);
       exit(0);
   }
   else
   {
      waitpid(pid,NULL,0);
   }
#endif
}

void DirToTar(t_string dir, t_string tar)
{
#ifdef VISUALSTUDIO
	FATAL(("Implement dir to tar for windows"));
#else
   pid_t pid;
    if ((pid=fork())==0)
    {
       execlp("tar","tar","-cf",tar,dir,NULL);
       exit(0);
    }
    else
    {
         waitpid(pid,NULL,0);
    }

    RmdirR(dir);
#endif
}

void Untar(t_string tar)
{
#ifdef VISUALSTUDIO
FATAL(("Implement untar for windows\n"));
#else
   pid_t pid;
    if ((pid=fork())==0)
    {
       execlp("tar","tar","-xf",tar,NULL);
       exit(0);
    }
    else
    {
         waitpid(pid,NULL,0);
    }
#endif
}
