/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//* $Header: /home/cvs/cvsroot/diablo/utils/util_malloc.c,v 1.2 2003/12/02 22:15:00 bdebus Exp $ 
 * 
 * Memory allocation functions
 * ---------------------------
 * 
 * if DEBUG_MALLOC is defined the M...lloc functions allocated some extra structures for leak detection 
 * (support for DEBUG_MALLOC is also added in StringDup and Collate3)
 * if DEBUG_MALLOC_MAGIC_CHECK is defined it will also do some BASIC memory allocation checks 
 *
 * WARNING: do not use Free on normal malloced data if DEBUG_MALLOC is defined !
 *
 * if a program crashes because of DEBUG_MALLOC, you're probably writing to not allocated memory 
 *
 */

#include <stdlib.h>
#include <string.h>
#include "../config.h"
#include "util_error.h"
#include "util_malloc.h"
#include "../diablo_options.h"

#define DEBUG_MALLOC_MAGIC_CHECK

/* ========== Debug data structures ============ */
#ifdef DEBUG_MALLOC
typedef struct _alloc_info
{
#ifdef DEBUG_MALLOC_MAGIC_CHECK
   char magic[8];
#endif
   struct _alloc_info * prev;
   struct _alloc_info * next;
   char * mfname;
   int mlnno;
   int size;
} alloc_info;

static alloc_info * head=NULL;
static alloc_info * tail=NULL;
#endif

/* ========== MALLOC ============ */

#ifdef DEBUG_MALLOC
void * RealMalloc(char * file, t_uint32 lnno,  t_uint32 size)
#else
void * RealMalloc(t_uint32 size)
#endif
{
#ifdef DEBUG_MALLOC
	void * ret=malloc(size+sizeof(alloc_info));
	alloc_info * ai=(alloc_info *) ret;
#else
	void * ret=malloc(size);
#endif
       if (size==0)
       {
#ifndef DEBUG_MALLOC
	  WARNING(("Mallocing 0 bytes!"));
	  return NULL;
#else
	  WARNING(("Mallocing 0 bytes! at %s - %d",file,lnno));
#endif
       }

#ifdef DEBUG_MALLOC
	VERIFY_FASSERT((ret!=NULL), ("Malloc failed (called at %s - %d)!!!",file,lnno));
        ai->mfname=file;
	ai->mlnno=lnno;
#if 0
	ai->ffname=NULL;
	ai->status=0;
#endif
	ai->size=size;	
	ai->prev=tail;
#ifdef DEBUG_MALLOC_MAGIC_CHECK
	ai->magic[0]='M';
	ai->magic[1]='A';
	ai->magic[2]='G';
	ai->magic[3]='I';
	ai->magic[4]='C';
	ai->magic[5]='_';
	ai->magic[6]='A';
	ai->magic[7]='L';
#endif
	if (ai->prev) ai->prev->next=ai;
	ai->next=NULL;
	tail=ai;
	if (!head) head=ai;
#else
	VERIFY_FASSERT((ret!=NULL), ("Malloc failed!!!"));
#endif
#ifdef DEBUG_MALLOC
	return ((char *) ret)+sizeof(alloc_info);
#else
	return ret;
#endif
}



/* ========== CALLOC ============ */

#ifdef DEBUG_MALLOC
void * RealCalloc(char * file, t_uint32 lnno,  t_uint32 numelems, t_uint32 elemsize)
#else
void * RealCalloc(t_uint32 numelems, t_uint32 elemsize) 
#endif
{
#ifdef DEBUG_MALLOC
	void * ret=calloc(1,elemsize*numelems+sizeof(alloc_info));
	alloc_info * ai=(alloc_info *) ret;
#else
	void * ret=calloc(numelems, elemsize);
#endif
#ifdef DEBUG_MALLOC
	VERIFY_FASSERT((ret!=NULL), ("Malloc failed (called at %s - %d)!!!",file,lnno));
        ai->mfname=file;
	ai->mlnno=lnno;
#if 0
	ai->ffname=NULL;
	ai->status=0;
#endif
	ai->size=numelems*elemsize;	
	ai->prev=tail;
#ifdef DEBUG_MALLOC_MAGIC_CHECK
	ai->magic[0]='M';
	ai->magic[1]='A';
	ai->magic[2]='G';
	ai->magic[3]='I';
	ai->magic[4]='C';
	ai->magic[5]='_';
	ai->magic[6]='A';
	ai->magic[7]='L';
#endif
	if (ai->prev) ai->prev->next=ai;
	ai->next=NULL;
	tail=ai;
	if (!head) head=ai;
#else
	VERIFY_FASSERT((ret!=NULL), ("Malloc failed!!!"));
#endif
#ifdef DEBUG_MALLOC
	return ((char *) ret)+sizeof(alloc_info);
#else
	return ret;
#endif
}


#ifdef DEBUG_MALLOC
void CheckMalloced(char * file, t_uint32 lnno, void * to_free)
{
   alloc_info * ai=(alloc_info *) (((char *) to_free)-sizeof(alloc_info));
   VERIFY_FASSERT((to_free!=NULL), ("Not malloced: NULL!!!"));
   if (!(((ai->magic[0]=='M')&&(ai->magic[1]=='A')&&(ai->magic[2]=='G')&&(ai->magic[3]=='I')&&(ai->magic[4]=='C')&&(ai->magic[5]=='_')&&(ai->magic[6]=='A')&&(ai->magic[7]=='L'))))
      FATAL(("Not malloced at %s %d\n",file,lnno));
}
#endif

/* ========== FREE ============ */
#ifdef DEBUG_MALLOC
void RealFree(char * file, t_uint32 lnno, void * to_free)
#else
void RealFree(void * to_free)
#endif
{

#ifdef DEBUG_MALLOC
   alloc_info * ai=(alloc_info *) (((char *) to_free)-sizeof(alloc_info));
#endif
#ifdef DEBUG_MALLOC
   VERIFY_FASSERT((to_free!=NULL), ("Freeing NULL at %s-%d!!!",file,lnno));
#else
   VERIFY_FASSERT((to_free!=NULL), ("Freeing NULL!!!"));
#endif
#ifdef DEBUG_MALLOC

#ifdef DEBUG_MALLOC_MAGIC_CHECK

   if (!(((ai->magic[0]=='M')&&(ai->magic[1]=='A')&&(ai->magic[2]=='G')&&(ai->magic[3]=='I')&&(ai->magic[4]=='C')&&(ai->magic[5]=='_')&&(ai->magic[6]=='A')&&(ai->magic[7]=='L'))))
   {
	   int * z=NULL;
	   int tel;
	   printf("Freein non malloc data at %d in %s\n",lnno,file);
	   fflush(stdout);
	   fflush(stderr);
	   printf("Block contains: ");
	   for (tel=0; tel<10; tel++)
	   {
	      printf("%c",((char *) to_free)[tel]);
	      fflush(stdout);
	   }
	   printf("\n");
	   *z=2;
   }
   VERIFY_FASSERT(((ai->magic[0]=='M')&&(ai->magic[1]=='A')&&(ai->magic[2]=='G')&&(ai->magic[3]=='I')&&(ai->magic[4]=='C')&&(ai->magic[5]=='_')&&(ai->magic[6]=='A')&&(ai->magic[7]=='L')), ("MY-Freeing non MY-alloced data (%lx at %s:%d)",to_free,file,lnno));

   ai->magic[0]=ai->magic[1]=ai->magic[2]=ai->magic[3]=ai->magic[4]=ai->magic[5]=ai->magic[6]=ai->magic[7]=0;
#endif
   if (ai->prev) ai->prev->next=ai->next;
   else head=ai->next;
   if (ai->next) ai->next->prev=ai->prev;
   else tail=ai->prev;
   free(((char *) to_free)-sizeof(alloc_info));
#else
   free(to_free);
#endif
}

/* ========== REALLOC ============ */
#ifdef DEBUG_MALLOC
void * RealRealloc(char * file, t_uint32 lnno, void * base,t_uint32 size)
#else
void * RealRealloc(void * base,int size)
#endif
{
	void * ret;
#ifdef DEBUG_MALLOC
	alloc_info * ai=(alloc_info*)(((char *) base)-sizeof(alloc_info));
#endif
	if (size==0)
	{
#ifdef DEBUG_MALLOC
		 RealFree(__FILE__,__LINE__,base);
#else
		 RealFree(base);
#endif
		 return NULL;
	}
	else
	{
#ifdef DEBUG_MALLOC
	        if (base)
		{
#ifdef DEBUG_MALLOC_MAGIC_CHECK
		   VERIFY_FASSERT(((ai->magic[0]=='M')&&(ai->magic[1]=='A')&&(ai->magic[2]=='G')&&(ai->magic[3]=='I')&&(ai->magic[4]=='C')&&(ai->magic[5]=='_')&&(ai->magic[6]=='A')&&(ai->magic[7]=='L')), ("MY-Reallocing non MY-alloced data (at %s:%d)",file,lnno));
		   ai->magic[0]=ai->magic[1]=ai->magic[2]=ai->magic[3]=ai->magic[4]=ai->magic[5]=ai->magic[6]=ai->magic[7]=0;
#endif

		   ret=realloc(((char *) base)-sizeof(alloc_info),size+sizeof(alloc_info));
		   VERIFY_FASSERT((ret!=NULL), ("Realloc failed at line %d in %s!!!",lnno,file));
		   if (ai==tail)
		   {
		      tail=(alloc_info *) ret;
		   }
		   
		   if (ai==head)
		   {
		      head=(alloc_info *) ret;
		   }

		   ai=(alloc_info *) ret;
#ifdef DEBUG_MALLOC_MAGIC_CHECK
		   ai->magic[0]='M';
		   ai->magic[1]='A';
		   ai->magic[2]='G';
		   ai->magic[3]='I';
		   ai->magic[4]='C';
		   ai->magic[5]='_';
		   ai->magic[6]='A';
		   ai->magic[7]='L';
#endif
		   if (ai->next) ai->next->prev=ai;
		   if (ai->prev) ai->prev->next=ai;
		   ai->size=size;
		   ret=((char *) ret)+sizeof(alloc_info);
		}
		else
		{
		   ret=RealMalloc(file,lnno,size);
		}
#else
		ret=realloc(base,size);
		VERIFY_FASSERT((ret!=NULL), ("Realloc failed!!!"));
#endif
		return ret;
	}
}

/* ========== Debug summary and print functions ============ */
#ifdef DEBUG_MALLOC
struct final
{
   char * name;
   int lnno;
   int sum;
   struct final * next;
};

#ifdef DEBUG_MALLOC
void RealCheckAddress(char * file,t_uint32 lnno, void * x)
{
   alloc_info * tmp=head;
   char * y=x;
   while(tmp)
   {
      char * data=((char *) tmp)+sizeof(alloc_info);
#ifdef DEBUG_MALLOC_MAGIC_CHECK
   VERIFY_FASSERT(((tmp->magic[0]=='M')&&(tmp->magic[1]=='A')&&(tmp->magic[2]=='G')&&(tmp->magic[3]=='I')&&(tmp->magic[4]=='C')&&(tmp->magic[5]=='_')&&(tmp->magic[6]=='A')&&(tmp->magic[7]=='L')), ("Malloc chtmpn corrupted  (at %s:%d)",file,lnno));
#endif
      if ((y>=data)&&(y<data+tmp->size))
      {
	 return;
      }
      tmp=tmp->next;
   }
   FATAL(("Invalid Allocated address (%s:%d))",file,lnno));
}
#endif

void PrintMemUse()
{
   alloc_info * tmp=head;
   long memuse=0;
   while(tmp)
   {
      memuse+=tmp->size;
      tmp=tmp->next;
   }
   printf("Total Memory Use: %ld bytes\n",memuse);
}

void PrintMallocInfo(void * block)
{
   alloc_info * ai=(alloc_info *) (((char *) block)-sizeof(alloc_info));


   printf("Block at %p , ai at %p\n",block,ai);
#ifdef DEBUG_MALLOC_MAGIC_CHECK

   if (!(((ai->magic[0]=='M')&&(ai->magic[1]=='A')&&(ai->magic[2]=='G')&&(ai->magic[3]=='I')&&(ai->magic[4]=='C')&&(ai->magic[5]=='_')&&(ai->magic[6]=='A')&&(ai->magic[7]=='L'))))
   {
	   printf("This block is not allocated!\n");
	   printf("Block starts with %c %c %c %c %c %c %c %c\n",ai->magic[0],ai->magic[1],ai->magic[2],ai->magic[3],ai->magic[4],ai->magic[5],ai->magic[6],ai->magic[7]);
   }
#endif
   printf("Block allocated at %s %d\n",ai->mfname,ai->mlnno);
}

void PrintRemainingBlocks()
{
   struct final * first=NULL; 
   struct final * run;
   struct final * run_next;
   alloc_info * tmp=head;
   int found;
   while(tmp)
   {
      found=0;
      for (run=first; run!=NULL; run=run->next)
      {
	 if ((tmp->mlnno==run->lnno)&&(strcmp(run->name,tmp->mfname)==0))
	 {
	    run->sum+=tmp->size;
	    found=1;
	    break;
	 }
      }
      if (!found)
      {
	 run=malloc(sizeof(struct final));
	 run->sum=tmp->size;
	 run->name=tmp->mfname;
	 run->lnno=tmp->mlnno;
	 run->next=first;
	 first=run;
      }
      /*printf("At %s %d: ",tmp->mfname,tmp->mlnno); */
      /*printf("%s\n",((char *) tmp)+sizeof(alloc_info));*/
      tmp=tmp->next;
   }


   for (run=first; run!=NULL; run=run_next)
   {
      WARNING(("LEAK: Total block size %d allocated at %s-%d",run->sum,run->name,run->lnno));
      run_next=run->next;
      free(run);
   }
}
#endif
