/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "utils.h"

#ifndef NO_CONFIG
#include "../config.h" 
#endif

#include <stdlib.h> 
#include <stdio.h> 
#include <string.h>
#include "util_options.h"

#ifndef NO_UTILS
#include "util_error.h"
#include "util_malloc.h"
#include "util_term.h"
#endif

#ifndef NO_DATASTRUCTURES
#include "../datastructures/data_string.h"
#include "../datastructures/data_hash.h"
#endif

#ifndef NO_DIABLO_KERNEL
#include "../kernel/diablo_object_tables.h"
#include "../kernel/diablo_print.h"
#endif

extern t_option option_list[];

void OptionsDefaults()
{
  t_option * iter=option_list;
  while(iter->type!=OPTION_NONE)
  {
     if ((iter->set)&&(!(*iter->set)))
     {
	switch(iter->type)
	{
#ifndef NO_UTILS
	   case OPTION_PATH:
	      {
		 t_string tmp;
		 if (!(iter->defaults.default_string)) break; 

		 tmp=StringDup(iter->defaults.default_string);
		 (*iter->variable)=ParsePath(tmp,":");
		 Free(tmp);
		 *iter->set=TRUE;
		 break;
	      }
#endif
	   case OPTION_STRING:
	      {
		 if (!(iter->defaults.default_string)) break; 
		 (*iter->variable)=StringDup(iter->defaults.default_string);
		 *iter->set=TRUE;
		 break;
	      }
	   case OPTION_BOOL:
	      *iter->set=TRUE;
	      (* ((t_bool *) iter->variable))=(iter->defaults.default_bool);
	      break;
	   case OPTION_COUNT:
	   case OPTION_INT:
	      *iter->set=TRUE;
	      (* ((t_uint32 *) iter->variable))=(iter->defaults.default_count);
	      break;
	   default:
	      /* keep the compiler happy */
	      break;
	}
     }
     iter++;
  }
}

/*!
 * This functions prints the usage information for the application */  

void UsageOptionHeader(t_option * iter)
{
	if (iter->hidden) return; 
	if (iter->required!=1) printf("[");
	if (iter->short_option && iter->long_option)
	{
		printf("-%s|--%s ",iter->short_option,iter->long_option);
	}
	else if (iter->long_option)
	{
		printf("--%s ",iter->long_option);
	}
	else if (iter->short_option)
	{
		printf("-%s ",iter->short_option);
	}
	switch(iter->type)
	{
		case OPTION_PATH:
			printf("<path>");
			break;
		case OPTION_FILE:
			printf("<file>");
			break;
		case OPTION_STRING:
			printf("<string>");
			break;
		case OPTION_BOOL:
			printf("[on|off]");
			break;
		case OPTION_COUNT:
		case OPTION_USAGE:
			break;
		case OPTION_INT:
			printf("<number>");
			break;
		default:
			/* keep the compiler happy */
			break;
	}
	if (iter->required!=1) printf("]");
	printf(" ");

}


void UsageOptionMessage(t_option * iter)
{
   t_uint32 ret;
	char out[50]="\0";
	if (iter->hidden) return;
	if (iter->required!=1) sprintf(out,"[");

	if (iter->short_option)
		sprintf(out+strlen(out),"-%s",iter->short_option);

	if ((iter->short_option) && (iter->long_option))
		sprintf(out+strlen(out),"|");

	if (iter->long_option)
		sprintf(out+strlen(out),"--%s",iter->long_option);

	if (((iter->short_option) || (iter->long_option))&&(iter->type!=OPTION_COUNT)&&(iter->type!=OPTION_USAGE))
		sprintf(out+strlen(out)," ");

	switch(iter->type)
	{
		case OPTION_PATH:
			sprintf(out+strlen(out),"<path>");
			break;
		case OPTION_STRING:
			sprintf(out+strlen(out),"<string>");
			break;
		case OPTION_FILE:
			sprintf(out+strlen(out),"<file>");
			break;
		case OPTION_BOOL:
			sprintf(out+strlen(out),"[on|off]");
			break;
		case OPTION_COUNT:
		case OPTION_USAGE:
			break;
		case OPTION_INT:
			sprintf(out+strlen(out),"<number>");
			break;
		default:
			/* keep the compiler happy */
			break;
	}

	if (iter->required!=1) sprintf(out+strlen(out),"]: ");
	else sprintf(out+strlen(out),": ");

	ret=TermPrintClean(stdout,out,0,5,0);
	if (iter->description) TermPrintClean(stdout,iter->description,ret,20,FINAL);
	else TermPrintClean(stdout,"",ret,20,FINAL);

	switch(iter->type)
	{
		case OPTION_PATH:
			if (iter->defaults.default_string)
			{
				ret=TermPrintClean(stdout,"DEFAULT: ",0,20,0);
				TermPrintClean(stdout,iter->defaults.default_string,ret,29,FINAL);
			}
			break;
		case OPTION_STRING:
		case OPTION_FILE:
			if (iter->defaults.default_string)
			{
				ret=TermPrintClean(stdout,"DEFAULT: ",0,20,0);
				TermPrintClean(stdout,iter->defaults.default_string,ret,29,FINAL);
			}
			break;
		case OPTION_BOOL:
			ret=TermPrintClean(stdout,"DEFAULT: ",0,20,0);
			if (iter->defaults.default_bool)
				TermPrintClean(stdout,"on",ret,29,FINAL);
			else
				TermPrintClean(stdout,"off",ret,29,FINAL);
			break;
		case OPTION_COUNT:
			break;
		default:
			/* keep the compiler happy */
			break;
	}
}


#ifndef NO_DATASTRUCTURES
typedef struct _t_option_group
{
   data_hash_node n;
   struct _t_option_group * next;
   
} t_option_group;
#endif

void Usage(void)
{
   t_option * iter=option_list;
#ifndef NO_DATASTRUCTURES
   t_option_group  * f=NULL;
   t_option_group  * g;
   
   data_hash_table * h=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,NULL); 
#endif
   
   TermTitle(stdout,"\nUsage:",' ');
   printf("\n");
   PrintExecName(stdout);printf(" <options>\n"); 
#ifndef NO_DATASTRUCTURES
   while(iter->type!=OPTION_NONE)
   {
      if (iter->hidden) { iter++; continue; }
      if (iter->group && (!HashLookup(h,iter->group)))
      {
	      g=malloc(sizeof(t_option_group));
	      g->n.key=StringDup(iter->group);
	      g->next=f;
	      f=g;
	      HashInsert(h,(data_hash_node *) g);
      }
      iter++;
   }
#endif

   printf("The following options exist:\n\n");

#ifndef NO_DATASTRUCTURES
   for (g=f; g!=NULL; g=g->next)
   {
	   printf("     %s options:\n\n",(char*) (g->n.key));
	   iter=option_list;
	   while(iter->type!=OPTION_NONE)
	   {
		   if (iter->group && (strcmp(iter->group,g->n.key)==0))
			   UsageOptionMessage(iter);
		   iter++;
	   }
	   printf(" \n\n");
   }


   printf("     Normal options:     \n\n");
#endif
   iter=option_list;
   while(iter->type!=OPTION_NONE)
   {
	   if (!(iter->group))
		   UsageOptionMessage(iter);
	   iter++;
   }
   printf(" \n\n");

}

/*! 
 * 
 * fills in the options that can be change via the environment */

void OptionsGetEnvironment()
{
   char * ret;
   t_option * iter=option_list;

   while(iter->type!=OPTION_NONE)
   {
      if ((iter->set)&&(!(*iter->set))&&(iter->enviromental_option))
      {
	 ret=getenv(iter->enviromental_option); /* The search-path for libraries */
	 if (ret)
	 {
	    switch(iter->type)
	    {
#ifndef NO_UTILS
	       case OPTION_PATH:
		  *iter->variable=ParsePath(ret,":");
		  break;
#endif
	       default:
	       case OPTION_STRING:
	       case OPTION_FILE:
	       case OPTION_BOOL:
	       case OPTION_COUNT:
		  FATAL(("Implement"));
	    }
	    *iter->set=TRUE;
	 }
      }
      iter++;
   }
}




/*!
 * \param argc The argument count
 * \param argv The list of arguments
 *
 * fills in the options that are changed on the commandline
 */

void OptionsParseCommandLine(t_uint32 argc, char ** argv)
{
   t_uint32 tel;

   DiabloPrint(stdout,"Command line: ");
   DiabloPrint(stdout,"%s ",argv[0]);


   for (tel=1; tel<argc; tel++)
   {
      {
	 t_bool skip=FALSE, found=FALSE;
	 t_option * iter=option_list;
	 DiabloPrint(stdout,"%s ",argv[tel]);
	 while(iter->type!=OPTION_NONE)
	 {
	    if (((argv[tel][0]=='-') && (argv[tel][1]!='-') && (iter->short_option) && (strcmp(iter->short_option,&argv[tel][1])==0))
	    || ((argv[tel][0]=='-') && (argv[tel][1]=='-') && (iter->long_option) && (strcmp(iter->long_option,&argv[tel][2])==0))
	    || ((argv[tel][0]!='-') && (!iter->long_option) && (!iter->short_option))
	    )
	    {
	       if ((!iter->set)||(!(*iter->set)))
	       {
		  if (iter->set) (*iter->set)=TRUE;
		  switch(iter->type)
		  {
		     case OPTION_PATH:
			{
#ifndef NO_UTILS
			   t_string tmp;
			   if ((tel+1)>=argc)
			   {
			      FATAL_USAGE(("-%s expects a path!",iter->short_option));
			   }

			   tmp=StringDup(argv[tel+1]);

			   (*iter->variable)=ParsePath(tmp,":");
			   Free(tmp); 
			   skip=TRUE; 
			   found = TRUE;
			   DiabloPrint(stdout,"%s ",argv[tel+1]);
#else
			   FATAL(("Paths not available when compiled with NO_UTILS"));
#endif
			}
			break;
		     case OPTION_STRING:
			if ((iter->short_option || iter->long_option) && ((tel+1)>=argc))
			   FATAL_USAGE(("-%s expects a string!",iter->short_option));
			if (iter->short_option || iter->long_option) 
			{
			   (*iter->variable)=StringDup(argv[tel+1]);
			   skip=TRUE; 
			   DiabloPrint(stdout,"%s ",argv[tel+1]);
			}
			else
			   (*iter->variable)=StringDup(argv[tel]);
			found = TRUE;
			break;
		     case OPTION_COUNT:
			found = TRUE;
			((t_uint32) (*iter->variable))++;
			DiabloPrint(stdout,"(%d) ",(t_uint32)(*iter->variable));
			break;
		     case OPTION_BOOL:
			if ((tel+1)>=argc)
			{
			   found = TRUE;
			   *((t_bool *) (iter->variable))=!iter->defaults.default_bool;
			}
			else
			{
			   if (strcmp(argv[tel+1],"on")==0)
			   {
			      found = TRUE;
			      skip=TRUE;
			      *((t_bool *) iter->variable)=TRUE;
			   }
			   else if (strcmp(argv[tel+1],"off")==0)
			   {
			      found = TRUE;
			      skip=TRUE;
			      *((t_bool *) iter->variable)=FALSE;
			   }
			   else
			   {
			      found = TRUE;
			      *((t_bool *) (iter->variable))=!(iter->defaults.default_bool);
			   }
			}
			DiabloPrint(stdout,"%s ",*((t_bool *) (iter->variable))?"on":"off");
			break;

		     case OPTION_INT:
			{
			   t_string tmp;
			   if ((tel+1)>=argc)
			   {
			      FATAL_USAGE(("-%s expects a numeric argument!",iter->short_option));
			   }

			   tmp=StringDup(argv[tel+1]);
			   *( (int *) (iter->variable))=atoi(tmp);
			   Free(tmp); 
			   skip=TRUE; 
			   found = TRUE;
			   DiabloPrint(stdout,"%d ",*( (int *) (iter->variable)));
			}
			break;
		     case OPTION_USAGE:
			{
			   Usage();
			   exit(0);
			}
			break;
		     case OPTION_FILE:
			{
			   if ((iter->short_option || iter->long_option) && ((tel+1)>=argc))
			      FATAL_USAGE(("-%s expects a filename!",iter->short_option));
			   if (iter->short_option || iter->long_option) 
			   {
#ifndef NO_UTILS
			      (*iter->variable)=ConvertFileName(argv[tel+1]);
#else
			      (*iter->variable)=StringDup(argv[tel+1]);
#endif
			      skip=TRUE; 
			      DiabloPrint(stdout,"%s ",argv[tel+1]);
			   }
			   else
#ifndef NO_UTILS
			      (*iter->variable)=ConvertFileName(argv[tel]);
#else
			      (*iter->variable)=StringDup(argv[tel]);
#endif
			   found = TRUE;
			}

		     default:
			/* keep the compiler happy */
			break;
		  }
	       }
	       else
	       {
		  FATAL(("Option %s ignored (duplicate set)", argv[tel]));
	       }
	    }
	    if (found) break;
	    iter++;

	 }
	 if (!found) 
	    FATAL_USAGE(("Unknown option %s!",argv[tel]));

	 if (skip)
	 {
	    tel++;
	    continue;
	 }
      }
	  
      /*
	 if (global_options.objectfilename) FATAL_USAGE(("Multiple objectfiles!"));
	 if (!(argv[tel][0]=='~'))
	 {
	 global_options.objectfilename=StringDup(argv[tel]);
	 global_options.objectfilename_set=TRUE;
	 }
	 else if (!(argv[tel][1]=='/'))
	 {
	 FATAL(("Implement user homedirs"));
	 }
	 else
	 {
      //global_options.objectfilename=ConvertFileName(argv[tel]);
      global_options.objectfilename_set=TRUE;
      }
      DiabloPrint(stdout,"%s ",global_options.objectfilename);
      */
   }
   DiabloPrint(stdout,"\n");


/*   if ((!global_options.restore)&&(!global_options.restore_multi_set)&&(global_options.objectfilename==NULL))
   {
      FATAL_USAGE(("No objectfile!"));
   } */
}


