/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../simple_types.h"
#include "util_term.h"
#ifdef HAVE_TERMCAP
char *tgetstr (char * info, char ** name);
t_uint32 tgetent (void * buffer , const char * name);
t_uint32 tgetnum (const char * name);
static char term_buffer[1024]; /* This buffer should be 1024 bytes large and persistant over all termcap calls! */
#endif
static char * term_name=NULL;

void TermInit()
{
   term_name=getenv("TERM");
   if (!term_name) { printf("Could not find term!\n"); }
   else {
#ifdef HAVE_TERMCAP
      t_int32 ret;
      memset(term_buffer,1024,sizeof(char));
      ret=tgetent(term_buffer,term_name);
      if (ret<=0) { 
	 printf("Couldn't get terminal description for %s!\n",term_name); 
#endif
	 term_name=NULL;

#ifdef HAVE_TERMCAP
      }
#endif
   }
}

t_uint32 TermWidth()
{
   t_int32 i=0;
#ifdef HAVE_TERMCAP
   if (term_name) i=tgetnum("co");  /* Get the number of columns on the screen */
#endif
   if ((!term_name)||(i==-1)) i=80;
   return (t_uint32) i;
}

/* A complex function to print a line to a stream, with word wrapping and indentation. str should not contain extra spaces or newlines! */
int TermPrintClean(FILE * stream, char * str, t_uint32 curpos,t_uint32 indent, t_uint32 flags)
{
   t_uint32 width=TermWidth();
   t_uint32 running=curpos;
   t_uint32 total_len=strlen(str);
   t_uint32 done_len=0;
   t_uint32 max_len;
   char * pos=str;
   t_uint32 tel;

   while(1)
   {
      if (indent>running)
      for (tel=0; tel<indent-running; tel++) printf(" ");
      running=MAX(running,indent); /* Running is line position */
      max_len=width-running;
      if (max_len>=(total_len-done_len))
      {
	 fprintf(stream,"%s",pos);
	 if (flags&FINAL)
	 {
	    for (tel=width-(total_len-done_len+running); tel>0; tel--)
	       fprintf(stream," "); /* Always pad, in case we are using reversed printing */
	    printf("\n");
	    return 0;
	 }
	 
	 return running+(total_len-done_len);
      }
      else
      {
	char format[100];
	char * run=pos+max_len;
	char * run2;
	t_int32 size=max_len;
	while((*run!=' ')&&(run!=pos)) { run--; size--; }
	if (max_len>0)
	{
	   if (run==pos) {
	      t_uint32 wl=0;
	      /* We're about to wrap a word, but it could be we can put it on the next line ... */
	      run2=pos;
	      while((*run2!=' ')&&(*run2!='\0'))
	      {
		 run2++;
		 wl++;
	      }
	      if (wl<(width-indent))
	      {
		 /* No need to wrap this, put it on the next line */
		 for (tel=width-running; tel>0; tel--)
		    fprintf(stream," ");
		 fprintf(stream,"\n");
		 running=0;
		 if (*pos==' ') { pos++; done_len++; }
		 continue;
	      }
	      else
		 size=max_len-1;
	   }
	   sprintf(format,"%%.%ds",size);
	   fprintf(stream,format,pos); 
	   if ((run==pos)&&(max_len>1)) {
	      fprintf(stream,"$"); running++; 
	   }
	   else/* Skip the space */
	   {
	      pos++; done_len++; 
	   }
	   pos+=size;
	   done_len+=size;
	   for (tel=width-(size+running); tel>0; tel--)
	      fprintf(stream," ");
	   fprintf(stream,"\n");
	}
	else fprintf(stream,"\n");
      }
      running=0;
   }
}

void TermFillLine(char c)
{
   t_uint32 width=TermWidth();
   for (;width>0; width--)
   {
      printf("%c",c);
   }
   printf("\n");
}

t_bool TermSaveCursor()
{
#ifdef HAVE_TERMCAP
  char buffer[200];
  char * ret;
  if (!term_name) return FALSE;
  if ((ret=tgetstr("sc",((char **) (&buffer))))==NULL)
  {
    return FALSE;
  }
  else {
    printf("%s",ret);  
    fflush(stdout); 
    return TRUE;
  }
#else
  return FALSE;
#endif
}

t_bool TermRestoreCursor()
{
#ifdef HAVE_TERMCAP
  char buffer[200];
  char * ret;
  if (!term_name) return FALSE;
  if ((ret=tgetstr("rc",(char **) &buffer))==NULL)
  {
    return FALSE;
  }
  else {
    printf("%s",ret);  
    fflush(stdout); 
    return TRUE;
  }
#else
  return FALSE;
#endif
}


void TermTitle(FILE * fp, char * title, char pad)
{
   t_uint32 width=TermWidth()-strlen(title)-2;
   t_uint32 before=width/2;
   t_uint32 after=width-before;
   for (;before>0; before--)
      fprintf(fp,"%c",pad);

   fprintf(fp," %s ",title);
   for(;after>0; after--)
      fprintf(fp,"%c",pad);
   fprintf(fp,"\n"); /* In case we got the terminal width wrong */
}
