/*
 *{{{ Copyright 2001,2002,2003:
 *			Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//**********************************************************
 * FILE: diablo_procstate.h
 *
 * DESC: Headers for processor state manipulation 
 *
 * Written by Bjorn De Sutter
 *
 * Copyright (c) 2002 Bjorn De Sutter
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 ***********************************************************/


/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
#ifndef PROCSTATE_TYPEDEFS
#define PROCSTATE_TYPEDEFS
/** \internal */
typedef struct _t_procstate t_procstate;
/** \internal */
typedef union  _register_content  register_content;
/** \internal */
typedef struct _t_regstate t_regstate;
/*! t_lattice_level is used to return information about the values,
   relocation tags and condition bits propagated in state */
typedef enum {TOP=-1,VALUE,BOT} t_lattice_level; 
#ifndef DIABLO_EDGE_H
#define DIABLO_EDGE_H
#include "diablo_edge.h"
#undef DIABLO_EDGE_H
#endif
/** \internal */
typedef void (*TestAndSetConditionFunction)(t_cfg_edge *,t_procstate*, t_procstate **, t_procstate **);
/*! \defgroup PROCSTATES Processor States 
 *
 * Processor states are used during constant propagation: there they
 * represent the state of the processor at a given moment during
 * execution/propagation */
/*@{ */


#endif

#ifndef PROCSTATE_HEADER
#define PROCSTATE_HEADER

#include "../simple_types.h"
#include "diablo_description.h"
#include "diablo_regset.h"
#include "diablo_addresses.h"
#include "diablo_reloc.h"


/*! This union represents register contents on a generic target
architecture. It can be integer (addresses) or floating-point
values. The latter are not yet used. */
union _register_content
{
  t_address i;
  /*  t_float f; */
};

/*! This struct holds processor states, i.e. the register contents
    (including possible TOP and BOT values) and condition flags that
    are propagated over a program during constant propagation.

    We also propagate so called relocation tags: if a register content
    being propagated is a code or data address, the relocation in the
    program describing this address is propagated along. This will
    become useful for dead data detection.
 */

struct _t_procstate
{
#ifdef EXTENDED_CONSTPROP
  t_regset value_plus;
  t_regset value_zero;
  t_regset value_minus;
#endif
  /*! The registers for which the propagated value is TOP are set to 1
    in this register set. */
  t_regset top;
  /*! The registers for which the propagated value is BOT are set to 1
    in this register set. */
  t_regset bot;
  /*! The registers for which the propagated relocation tag is TOP are
    set to 1 in this register set. Relocation tags are propagated
    together with values. Although not exploited today, the
    information in the relocation tags will be useful later on. TOP in
    this case means: the propagated value cannot be an address*/
  t_regset tag_bot;
  /*! The registers for which the propagated relocation is BOT are set
    to 1 in this register set. Bot for relocation tags means that the
    propagated value (a constant or bot) can be an unknown address,
    used to access an unknown block*/
  t_regset tag_top;
  /*! The condition flags (modeled as separate registers) being
      propagated are either set, clear, bot or top. These values are
      encoded in two register sets: with the following encoding:
      - bot: cond_true = 1, cond_false = 1
      - top: cond_true = 0, cond_false = 0
      - set: cond_true = 1, cond_false = 0
      - clear: cond_true = 0, cond_false = 1
  */
  t_regset cond_true;
  /*! \ref _t_procstate::cond_true */
  t_regset cond_false;

/*! In case there are known constant values propagated, they are
    stored in these arrays. There is one array per "REG_SUBSET_SIZE" registers. This
    allows to adapt the number of register for which one allocates
    space. Not quite sure whether this is a good solution though */

  register_content*  register_subset[MAX_NR_REG_SUBSETS];

/*! Similar arrays are used to store the actual relocation tags in
  case they are constant. */

  t_reloc**         register_tag_subset[MAX_NR_REG_SUBSETS];
};

t_procstate* ProcStateNew();
void ProcStateFree(t_procstate* state);
void ProcStateRealFree(void);
void ProcStateSetAllTop(t_procstate* state, t_regset regs);
void ProcStateSetAllBot(t_procstate* state, t_regset regs);
void ProcStateSetRegsetTop(t_procstate* state, t_regset regs);
void ProcStateSetRegsetBot(t_procstate* state, t_regset regs);
void ProcStateSetTagsetTop(t_procstate* state, t_regset regs);
void ProcStateSetTagsetBot(t_procstate* state, t_regset regs);
void ProcStateSetCondsetTop(t_procstate* state, t_regset regs);
void ProcStateSetCondsetBot(t_procstate* state, t_regset regs);
void ProcStateSetRegTop(t_procstate* state, t_reg reg);
void ProcStateSetRegBot(t_procstate* state, t_reg reg);
void ProcStateSetBothBot(t_procstate* state, t_reg reg);
void ProcStateSetTagTop(t_procstate* state, t_reg reg);
void ProcStateSetTagBot(t_procstate* state, t_reg reg);
void ProcStateSetCondTop(t_procstate* state, t_reg reg);
void ProcStateSetCondBot(t_procstate* state, t_reg reg);
void ProcStateSetReg(t_procstate* state, t_reg reg, register_content value); 
#ifdef EXTENDED_CONSTPROP
void ProcStateSetRegPlus(t_procstate* state, t_reg reg, register_content value); 
void ProcStateSetRegMinus(t_procstate* state, t_reg reg, register_content value); 
void ProcStateSetRegZero(t_procstate* state, t_reg reg, register_content value); 
void ProcStateSetPlus(t_procstate* state, t_reg reg);
void ProcStateSetZero(t_procstate* state, t_reg reg);
void ProcStateSetMinus(t_procstate* state, t_reg reg);
t_bool ProcStateGetPlus(t_procstate* state, t_reg reg);
t_bool ProcStateGetZero(t_procstate* state, t_reg reg);
t_bool ProcStateGetMinus(t_procstate* state, t_reg reg);
void ProcStateRegDup(t_procstate* state,t_reg reg, t_reg reg2); 
void ProcStateRegDupProcstate(t_procstate* state,t_procstate* dest,t_reg reg);
void ProcStateSetRegBotExtend(t_procstate* state, t_reg reg, t_bool plus, t_bool zero, t_bool minus);
void ProcStateGetRegBotExtend(t_procstate* state, t_reg reg, t_bool * plus, t_bool * zero, t_bool * minus);
#endif
void ProcStateSetTag(t_procstate* state, t_reg reg, t_reloc* value);
void ProcStateSetCond(t_procstate* state, t_reg reg, t_bool value);
t_lattice_level ProcStateGetReg(t_procstate* state, t_reg reg, register_content * value);
t_lattice_level ProcStateGetTag(t_procstate* state, t_reg reg, t_reloc ** value);
t_lattice_level ProcStateGetCond(t_procstate* state, t_reg reg, t_bool * value);
t_bool ProcStateJoin(t_procstate* dest, t_procstate* src, t_regset regs, t_regset live_regs) ;
t_bool ProcStateJoinSimple(t_procstate* dest, t_procstate* src, t_regset regs) ;
void ProcStateDup(t_procstate *dest, t_procstate* src);
t_procstate * ProcStateNewDup(t_procstate* src);
void ProcStateJoinTags(t_procstate * state, t_reg dest, t_reg src);
/* @} */
#define Reg2Set(r)   ((r)/REG_SUBSET_SIZE)
#define Reg2Index(r) ((r)-(((r)/REG_SUBSET_SIZE)*REG_SUBSET_SIZE))

#define ISVALUE(x) ((x)==VALUE)
#define ISBOT(x) ((x)==BOT)
#define ISTOP(x) ((x)==TOP)

/* Dominique: typedefs for storing constant values in the instruction structure */

typedef struct _valuereg {
  /** level of the plain register value (TOP, VALUE or BOT) */
  t_lattice_level level;
  /** the actual plain register value */
  register_content val;
  /** level of the relocation associated with the register value (TOP, VALUE or BOT) */
  t_lattice_level taglevel;
  /** the actual relocation associated with the register value */
  t_reloc * tag;
} valuereg;

typedef union _allstate {
  /** if it's a normal register (not just a condition bit) use a valuereg structure */
  valuereg v;
  /** if it's a condition register instead of a normal one, value and level are stored in
      this variable: bot = 11b, top = 00b, set = 10b, clear = 01b */
  t_uint8 c;
} allstate;

/*! Holds the state of a register */
struct _t_regstate {
  /** the register we are describing */
  t_reg reg;
  allstate state; 
};


/* some helper definitions */
#define __rsVLevel(s)		(s)->state.v.level
#define __rsTLevel(s)		(s)->state.v.taglevel
#define __rsValue(s)		(s)->state.v.val
#define __rsTag(s)		(s)->state.v.tag
#define __rsCond(s)		(s)->state.c
#define __rsReg(s)		(s)->reg

#define RegStateSetReg(s,r)	__rsReg(s) = (r)
#define RegStateGetReg(s)	__rsReg(s)

#define RegStateSetTop(s)	(__rsVLevel(s) = TOP)
#define RegStateSetBot(s)	(__rsVLevel(s) = BOT)
#define RegStateSetTagTop(s)	(__rsTLevel(s) = TOP, __rsTag(s) = NULL)
#define RegStateSetTagBot(s)	(__rsTLevel(s) = BOT, __rsTag(s) = NULL)
#define RegStateSetCondTop(s)	__rsCond(s) = 0x0
#define RegStateSetCondBot(s)	__rsCond(s) = 0x3

#define RegStateSetValue(s,value)	(__rsVLevel(s) = VALUE, __rsValue(s) = (value))
#define RegStateSetTag(s,reloc)		(__rsTLevel(s) = VALUE, __rsTag(s) = (reloc))
#define RegStateSetCond(s,value)	__rsCond(s) = (value) ? 0x2 : 0x1

#define RegStateGetValue(s,valptr)	(*(valptr) = __rsValue(s), __rsVLevel(s))
#define RegStateGetTag(s,tagptr)	(*(tagptr) = __rsTag(s),   __rsTLevel(s))
#define RegStateGetCond(s,boolptr)	(*(boolptr) = (__rsCond(s) == 0x2), (__rsCond(s) == 0) ? TOP : (__rsCond(s) == 0x3) ? BOT : VALUE)


t_lattice_level RegStateGetRegValue(t_ins * ins, t_regset * regset, t_reg reg, register_content * content);
t_lattice_level RegStateGetRegCond(t_ins * ins, t_regset * regset, t_reg reg, t_bool * cond);
t_lattice_level RegStateGetRegTag(t_ins * ins, t_regset * regset, t_reg reg, t_reloc ** tag);

void RegStateInit(t_regstate * state);
t_regstate * RegStateAlloc(t_uint32 n);
t_regstate * RegStateArrayCreateFromProcState(t_procstate * ps, t_regset whichregs, const t_architecture_description * desc);
t_bool TwoRelocsInSameBlock(t_reloc * reloc1, t_reloc* reloc2);
#define T_PROCSTATE(x)   ((t_procstate*)(x))
#endif
/* @} */
/* vim: set shiftwidth=2 foldmethod=marker: */
