/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_emulation_one.h"
#include "ia64_emulation.h"
#include "ia64_instructions.h"
#include "ia64_registers.h"
#include "ia64_opcodes.h"
#include "ia64_utils.h"
#include "ia64_types.h"
#include "ia64_dummy.h"

#include "../../utils/util_error.h"

#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_procstate.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_iterators.h"




typedef struct _t_ia64_emul_stats
{
  t_uint32 add_total;
  t_uint32 add_done;
  t_uint32 addp4_total;
  t_uint32 addp4_done;
  t_uint32 alloc_total;
  t_uint32 alloc_done;
  t_uint32 and_total;
  t_uint32 and_done;
  t_uint32 andcm_total;
  t_uint32 andcm_done;
  t_uint32 br_skipped;
  t_uint32 break_skipped;
  t_uint32 brl_skipped;
  t_uint32 cmp_total;
  t_uint32 cmp_done;
  t_uint32 cmp4_total;
  t_uint32 cmp4_done;
  t_uint32 ld_total;
  t_uint32 ld_partial;
  t_uint32 ld_done;
  t_uint32 mov_br_total;
  t_uint32 mov_br_done;
  t_uint32 movl_total;
  t_uint32 movl_done;
  t_uint32 nop_skipped;
  t_uint32 or_total;
  t_uint32 or_done;
  t_uint32 shl_total;
  t_uint32 shl_done;
  t_uint32 shladd_total;
  t_uint32 shladd_done;
  t_uint32 shladd4_total;
  t_uint32 shladd4_done;
  t_uint32 shr_total;
  t_uint32 shr_done;
  t_uint32 st_skipped;
  t_uint32 sub_total;
  t_uint32 sub_done;
  t_uint32 sxt_total;
  t_uint32 sxt_done;
  t_uint32 xor_total;
  t_uint32 xor_done;
  t_uint32 zxt_total;
  t_uint32 zxt_done;
  
  t_uint32 other_done;
  
} t_ia64_emul_stats;


static t_ia64_emul_stats __stats;


void Ia64PrintEmulationStatistics()
{
  /* Total = aantal oproepen naar de cooresponderende emulatiefunctie*/
  /* Done = aantal instructies dat hij kunnen emuleren heeft */
  /* Partial = enkel postinc kunnen doen*/
  /* Skipped = deze die hij overslaat (niet hoeft te emuleren)*/
  
  printf("\n");
  printf("--  Emulation statistics  --\n");
  printf("Total add:         %9d\n", __stats.add_total);
  printf("Done add:          %9d\n", __stats.add_done);
  printf("Total addp4:       %9d\n", __stats.addp4_total);
  printf("Done addp4:        %9d\n", __stats.addp4_done);
  printf("Total alloc:       %9d\n", __stats.alloc_total);
  printf("Done alloc:        %9d\n", __stats.alloc_done);
  printf("Total and:         %9d\n", __stats.and_total);
  printf("Done and:          %9d\n", __stats.and_done);
  printf("Total andcm:       %9d\n", __stats.andcm_total);
  printf("Done andcm:        %9d\n", __stats.andcm_done);
  printf("Skipped br:        %9d\n", __stats.br_skipped);
  printf("Skipped break:     %9d\n", __stats.break_skipped);
  printf("Skipped brl:       %9d\n", __stats.brl_skipped);
  printf("Total cmp:         %9d\n", __stats.cmp_total);
  printf("Done cmp:          %9d\n", __stats.cmp_done);
  printf("Total cmp4:        %9d\n", __stats.cmp4_total);
  printf("Done cmp4:         %9d\n", __stats.cmp4_done);
  printf("Total ld:          %9d\n", __stats.ld_total);
  printf("Partial ld:        %9d\n", __stats.ld_partial);
  printf("Done ld:           %9d\n", __stats.ld_done);
  printf("Total mov_br:      %9d\n", __stats.mov_br_total);
  printf("Done mov_br:       %9d\n", __stats.mov_br_done);
  printf("Total movl:        %9d\n", __stats.movl_total);
  printf("Done movl:         %9d\n", __stats.movl_done);
  printf("Skipped nop:       %9d\n", __stats.nop_skipped);
  printf("Total or:          %9d\n", __stats.or_total);
  printf("Done or:           %9d\n", __stats.or_done);
  printf("Total shl:         %9d\n", __stats.shl_total);
  printf("Done shl:          %9d\n", __stats.shl_done);
  printf("Total shladd:      %9d\n", __stats.shladd_total);
  printf("Done shladd:       %9d\n", __stats.shladd_done);
  printf("Total shladd4:     %9d\n", __stats.shladd4_total);
  printf("Done shladd4:      %9d\n", __stats.shladd4_done);
  printf("Total shr:         %9d\n", __stats.shr_total);
  printf("Done shr:          %9d\n", __stats.shr_done);
  printf("Skipped st:        %9d\n", __stats.st_skipped);
  printf("Total sub:         %9d\n", __stats.sub_total);
  printf("Done sub:          %9d\n", __stats.sub_done);
  printf("Total sxt:         %9d\n", __stats.sxt_total);
  printf("Done sxt:          %9d\n", __stats.sxt_done);
  printf("Total xor:         %9d\n", __stats.xor_total);
  printf("Done xor:          %9d\n", __stats.xor_done);
  printf("Total zxt:         %9d\n", __stats.zxt_total);
  printf("Done zxt:          %9d\n", __stats.zxt_done);
  printf("----------------------------\n");
  printf("Total other:       %9d\n", __stats.other_done); /*deze die niet geemuleerd worden*/
  printf("--     End statistics     --\n");
  printf("\n");
}

void Ia64ResetEmulationStatistics()
{
  bzero(&__stats, sizeof(__stats));
}









/* deze tabel staat helemaal onderaan dit document gemplementeerd*/
extern const Ia64EmulateFunction Ia64EmulateFunction_table [];




void Ia64EmulateInstruction(t_ins * ins, t_procstate * state)
{
  	Ia64EmulateFunction_table[IA64OPCODEINFO(ins).instr_opcode](ins, state);
}

void Ia64EmulateUnsupported(t_ins * ins, t_procstate * state)
{
    FATAL(("Unsupported ia64 instruction; cannot emulate."));
}


void ProcStateCopyRegState(t_procstate* srcstate, t_reg src, t_procstate* deststate, t_reg dest)
{
  register_content reg_cont;
  t_reloc * reloc;
  t_lattice_level level;

  level = ProcStateGetReg(srcstate, src, &reg_cont);
  switch(level)
  {
    case TOP:
      ProcStateSetRegTop(deststate, dest);
      break;
    case VALUE:
      ProcStateSetReg(deststate, dest, reg_cont);
      break;
    case BOT:
      ProcStateSetRegBot(deststate, dest);
    default:
      break;
  }
  
  level = ProcStateGetTag(srcstate, src, &reloc);
  switch(level)
  {
    case TOP:
      ProcStateSetTagTop(deststate, dest);
      break;
    case VALUE:
      ProcStateSetTag(deststate, dest, reloc);
      break;
    case BOT:
      ProcStateSetTagBot(deststate, dest);
    default:
      break;
  }
   
}


/*returns TRUE if there is a valid value (not top or bot) */
t_bool Ia64CheckAndGetRegisterState(t_procstate *state, t_ins * ins, t_ia64_reg in, t_int64 *inValue, t_uint64 *natValue)
{
  register_content reg;

  if (ProcStateGetReg(state, in, &reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return FALSE;
  }
  *inValue = (t_int64)G_T_UINT64(reg.i);

  if(IA64_IS_GR(in))
  {

    if (ProcStateGetReg(state,IA64GETNAT(in),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return FALSE;
    }

    *natValue = G_T_UINT64(reg.i);
  }

  return TRUE;
}






void Ia64EmulateDummy(t_ins * ins, t_procstate * state)
{
  if(IA64INSOPCODE(ins) != DUMMY_LIVE_OPCODE)
    return;

  if(IA64INSIMM1(ins) != -1LL) /* dummy 1 */
  {
    t_procstate *old = NULL;
    
//    printf("Emulating dummy1 @ 0x%llx\n", BBL_BASE_ADDRESS(INS_BBL(ins)));
    

    /*now do the actual moving*/
    
    if(IA64INSIMM2(ins) == IA64_ALLOC_MISSING)
    { /* if it is unknown: set all registers above 32 to bot !*/
      t_regset set = NullRegs;
      t_uint32 i = 0;
      for(i = 32; i<128; i++)
      {
	RegsetAddReg(set, IA64_GR + i);
	RegsetAddReg(set, IA64_GR_NAT + i); /* don't forget the NAT-bits !!!*/
      }

      ProcStateSetAllBot(state, set);
      
    }
    else
    {
      t_uint32 start_of_out = IA64INSIMM2(ins);
      t_uint64 cnt = IA64INSIMM3(ins) - IA64INSIMM2(ins);
      t_uint32 i = 0;

      /* copy old procstate*/
      FATAL(("Implement"));
      //old = ProcStateAlloc();
      ProcStateDup(old, state);

      /* first kill a previous one, if there is one*/
      if(IA64INSIMM1(ins) != 0ULL)
	ProcStateFree((t_procstate *)IA64INSIMM1(ins));
      
      /*save this state for the second dummy: needed to restore the "in" params*/
      IA64INSIMM1(ins) = (t_int64)(void *)old;


      for(i = 0; i< cnt; i++)
      {
	//move state of reg...
	ProcStateCopyRegState(state, IA64_R32 + start_of_out + i, state, IA64_R32 + i);
	ProcStateCopyRegState(state, IA64_R32_NAT + start_of_out + i, state, IA64_R32_NAT + i);
      }
    }
     
  }
  else /* dummy 2 */
  {
    t_cfg_edge * funlink_edge;
    t_bbl * bbl;
    t_bbl * call_bbl = NULL;
    
//    printf("Emulating dummy2 @ 0x%llx\n", BBL_BASE_ADDRESS(INS_BBL(ins)));


    if(IA64INSIMM2(ins) == IA64_ALLOC_MISSING)
    { /* if it is unknown: set all registers above 32 to bot !*/
      t_regset set = NullRegs;
      t_uint32 i = 0;
      for(i = 32; i<128; i++)
      {
	RegsetAddReg(set, IA64_GR + i);
	RegsetAddReg(set, IA64_GR_NAT + i); /* don't forget the NAT-bits !!!*/
      }

      ProcStateSetAllBot(state, set);
      
    }
    else
    {
      t_uint32 start_of_out = IA64INSIMM2(ins);
      t_uint64 cnt = IA64INSIMM3(ins) - IA64INSIMM2(ins);
      t_int32 i = 0;

      for(i = cnt-1; i>=0; i--)
      {
	//move state of reg... [now do the reverse copy, backwards!!]
	ProcStateCopyRegState(state, IA64_R32 + i, state, IA64_R32 + start_of_out + i);
	ProcStateCopyRegState(state, IA64_R32_NAT + i, state, IA64_R32_NAT + start_of_out + i);
      }



      /* first try to find the corresponding dummy1*/

      bbl = INS_BBL(ins);

      BBL_FOREACH_PRED_EDGE(bbl, funlink_edge)
      {
	if(EDGE_CAT(funlink_edge)&ET_FUNLINK)
	{
	  call_bbl=CFG_EDGE_HEAD(funlink_edge);
	  break;
	}
      }

      if(call_bbl != NULL)
      {
	if(call_bbl->ins_last)
	{
	  t_procstate * old; 
	  t_ins * ins_dummy1 = INS_PREV(call_bbl->ins_last);
	  if(! ins_dummy1 || IA64INSTYPE(ins_dummy1) != IA64_TYPE_DUMMY || IA64INSOPCODE(ins_dummy1) != DUMMY_LIVE_OPCODE)
	    FATAL(("NO DUMMY 1"));

	  /*now extract the old procstate and erase it !*/
	  old = (t_procstate *)IA64INSIMM1(ins_dummy1);

	  /* it think this can occur with context-insensitive constant-propagation, where a return
	  * of one function returns to another function ... */
	  if(old == NULL)
	  {
	    /* if it is unknown: set all registers above 32 to bot !*/
	    /* treat as if alloc is missing */
	    t_regset set = NullRegs;
	    t_uint32 i = 0;
	    for(i = 32; i<128; i++)
	    {
	      RegsetAddReg(set, IA64_GR + i);
	      RegsetAddReg(set, IA64_GR_NAT + i); /* don't forget the NAT-bits !!!*/
	    }

	    ProcStateSetAllBot(state, set);

	    printf("WARNING: Procstate saved in dummy 1 is NULL !\n"); 
	  }
	  else
	  {

	    t_uint64 cnt = IA64INSIMM2(ins);
	    t_uint32 i = 0;

	    for(i = 0; i< cnt; i++)
	    {
	      //move state of IN-reg...
	      ProcStateCopyRegState(old, IA64_R32 + i, state, IA64_R32 + i);
	      ProcStateCopyRegState(old, IA64_R32_NAT + i, state, IA64_R32_NAT + i);
	    }


	    ProcStateFree(old);
	    IA64INSIMM1(ins_dummy1) = 0ULL;
	  }

	}
	else
	  FATAL(("Could not find corresponding Dummy 1"));
      }
      else
	FATAL(("Could not find corresponding Dummy 1: could not find the funlink"));

    }
     
  }
    
      
   
  return;
}


void Ia64EmulateData(t_ins * ins, t_procstate * state)
{
    return;
}












/* todo...*/

/*
xxx_fault(parameters ...)                   
xxx_trap(parameters ...)                    
*/
#define acceptance_fence() FATAL(("IMPLEMENT: acceptance_fence"));                         
#define alat_cmp(rtype, raddr)  FATAL(("IMPLEMENT: alat_cmp"));                    
#define alat_frame_update( delta_bof, delta_sof) FATAL(("IMPLEMENT: alat_frame_update"));    
#define alat_inval() FATAL(("IMPLEMENT: alat_inval"));                                
#define alat_inval_multiple_entries(paddr, size) FATAL(("IMPLEMENT: alat_inval_multiple_entries"));    
#define alat_inval_single_entry(rtype, rega) FATAL(("IMPLEMENT: alat_inval_single_entry"));        
#define alat_write(rtype, raddr, paddr, size) FATAL(("IMPLEMENT: alat_write"));       
#define align_to_size_boundary(vaddr, size) FATAL(("IMPLEMENT: align_to_size_boundary"));         
#define check_branch_implemented(check_type) FATAL(("IMPLEMENT: check_branch_implemented")); 
#define data_serialize() FATAL(("IMPLEMENT: data_serialize"));                              
#define deliver_unmasked_pending_interrupt() FATAL(("IMPLEMENT: deliver_unmasked_pending_interrupt"));          
#define fadd(fp_dp, fr2) FATAL(("IMPLEMENT: fadd"));                              
#define fcmp_exception_fault_check(f2, f3, frel, sf, tmp_fp_env) FATAL(("IMPLEMENT: fcmp_exception_fault_check"));
#define fcvt_fx_exception_fault_check(fr2, signed_form, trunc_form, sf, tmp_fp_env) FATAL(("IMPLEMENT: fcvt_fx_exception_fault_check")); 
#define fma_exception_fault_check(f2, f3, f4, pc, sf, tmp_fp_env) (({FATAL(("IMPLEMENT: fma_exception_fault_check"));}), pc-sf)                               
#define fminmax_exception_fault_check(f2, f3, sf, tmp_fp_env) FATAL(("IMPLEMENT: fminmax_exception_fault_check"));
#define fms_fnma_exception_fault_check(f2, f3, f4, pc, sf, tmp_fp_env) (({FATAL(("IMPLEMENT: fms_fnma_exception_fault_check"));}), pc-sf)                     
#define fmul(fr3, fr4) FATAL(("IMPLEMENT: fmul")); 
#define followed_by_stop() FATAL(("IMPLEMENT: followed_by_stop"));
#define fp_check_target_register(f1) (({FATAL(("IMPLEMENT: fp_check_target_register"));}), f1)
#define fp_decode_fault(tmp_fp_env) FATAL(("IMPLEMENT: fp_decode_fault"));
#define fp_decode_traps(tmp_fp_env) FATAL(("IMPLEMENT: fp_decode_traps"));
#define fp_is_nan_or_inf(freg) FATAL(("IMPLEMENT: fp_is_nan_or_inf")); 
#define fp_equal(fr1, fr2) FATAL(("IMPLEMENT: fp_equal"));
#define fp_ieee_recip(num, den) FATAL(("IMPLEMENT: fp_ieee_recip"));
#define fp_ieee_recip_sqrt(root) FATAL(("IMPLEMENT: fp_ieee_recip_sqrt"));
#define fp_is_nan(freg) FATAL(("IMPLEMENT: fp_is_nan"));         
#define fp_is_normal(freg) FATAL(("IMPLEMENT: fp_is_normal"));      
#define fp_is_pos_inf(freg) FATAL(("IMPLEMENT: fp_is_pos_inf"));
#define fp_is_qnan(freg) FATAL(("IMPLEMENT: fp_is_qnan")); 
#define fp_is_snan(freg) FATAL(("IMPLEMENT: fp_is_snan"));
#define fp_is_unorm(freg) FATAL(("IMPLEMENT: fp_is_unorm"));
#define fp_is_unsupported(freg) FATAL(("IMPLEMENT: fp_is_unsupported"));
#define fp_less_than(fr1, fr2) FATAL(("IMPLEMENT: fp_less_than"));
#define fp_lesser_or_equal(fr1, fr2) FATAL(("IMPLEMENT: fp_lesser_or_equal"));
#define fp_normalize(fr1) FATAL(("IMPLEMENT: fp_normalize"));
#define fp_raise_fault(tmp_fp_env) FATAL(("IMPLEMENT: fp_raise_fault")); 
#define fp_raise_traps(tmp_fp_env) FATAL(("IMPLEMENT: fp_raise_traps")); 
#define fp_reg_bank_conflict(f1, f2) FATAL(("IMPLEMENT: fp_reg_bank_conflict"));
#define fp_reg_disabled(f1, f2, f3, f4) FATAL(("IMPLEMENT: fp_reg_disabled")); 
#define fp_reg_read(freg) FATAL(("IMPLEMENT: fp_reg_read")); 
#define fp_unordered(fr1, fr2) FATAL(("IMPLEMENT: fp_unordered"));
#define fp_fr_to_mem_format(freg, size) FATAL(("IMPLEMENT: fp_fr_to_mem_format"));
#define fpcmp_exception_fault_check(f2, f3, frel, sf, tmp_fp_env) FATAL(("IMPLEMENT: fpcmp_exception_fault_check"));
#define fpcvt_exception_fault_check(f2, signed_form, trunc_form, sf, tmp_fp_env) FATAL(("IMPLEMENT: fpcvt_exception_fault_check"));
#define fpma_exception_fault_check(f2, f3, f4, sf, tmp_fp_env) FATAL(("IMPLEMENT: fpma_exception_fault_check"));                                
#define fpminmax_exception_fault_check(f2, f3, sf, tmp_fp_env) FATAL(("IMPLEMENT: fpminmax_exception_fault_check"));                             
#define fpms_fpnma_exception_fault_check(f2, f3, f4, sf, tmp_fp_env) FATAL(("IMPLEMENT: fpms_fpnma_exception_fault_check"));                     
#define fprcpa_exception_fault_check(f2, f3, sf, tmp_fp_env, limits_check) FATAL(("IMPLEMENT: fprcpa_exception_fault_check"));                  
#define fprsqrta_exception_fault_check(f3, sf, tmp_fp_env, limits_check) FATAL(("IMPLEMENT: fprsqrta_exception_fault_check"));
#define frcpa_exception_fault_check(f2, f3, sf, tmp_fp_env) FATAL(("IMPLEMENT: frcpa_exception_fault_check"));                                 
#define frsqrta_exception_fault_check(f3, sf, tmp_fp_env) FATAL(("IMPLEMENT: frsqrta_exception_fault_check"));                                 
#define ignored_field_mask(regclass, reg, value) FATAL(("IMPLEMENT: ignored_field_mask"));
#define impl_itir_cwi_mask() FATAL(("IMPLEMENT: impl_itir_cwi_mask"));
#define instruction_serialize() FATAL(("IMPLEMENT: instruction_serialize"));
#define instruction_synchronize
#define is_finite(freg) FATAL(("IMPLEMENT: is_finite"));
#define is_ignored_reg(regnum) FATAL(("IMPLEMENT: is_ignored_reg")); 
#define is_inf(freg) FATAL(("IMPLEMENT: is_inf"));
#define is_interruption_cr(regnum) FATAL(("IMPLEMENT: is_interruption_cr"));
#define is_kernel_reg(ar_addr) FATAL(("IMPLEMENT: is_kernel_reg")); 
#define is_read_only_reg(rtype, raddr) FATAL(("IMPLEMENT: is_read_only_reg"));
#define is_reserved_field(regclass, arg2, arg3) FATAL(("IMPLEMENT: is_reserved_field")); 
#define is_reserved_reg(regclass, regnum) FATAL(("IMPLEMENT: is_reserved_reg")); 
#define long_branch_implemented() FATAL(("IMPLEMENT: long_branch_implemented"));
#define mem_flush(paddr) FATAL(("IMPLEMENT: mem_flush"));
#define mem_flush_pending_stores() FATAL(("IMPLEMENT: mem_flush_pending_stores"));
#define mem_promote(paddr, mtype, hint) FATAL(("IMPLEMENT: mem_promote"));
#define mem_read(paddr, size, border, mattr, otype, hint) FATAL(("IMPLEMENT: mem_read"));
#define fp_mem_to_fr_format(mem, size) FATAL(("IMPLEMENT: fp_mem_to_fr_format"));
#define mem_write(value, paddr, size, border, mattr, otype, hint) FATAL(("IMPLEMENT: mem_write"));
#define mem_xchg(data, paddr, size, byte_order, mattr, otype, hint) FATAL(("IMPLEMENT: mem_xchg"));
#define mem_xchg_add(add_val, paddr, size, byte_order, mattr, otype, hint) FATAL(("IMPLEMENT: mem_xchg_add"));             
#define mem_xchg_cond(cmp_val, data, paddr, size, byte_order, mattr, otype, hint) FATAL(("IMPLEMENT: mem_xchg_cond"));       
#define ordering_fence() FATAL(("IMPLEMENT: ordering_fence"));
#define rotate_regs() FATAL(("IMPLEMENT: rotate_regs"));
#define rse_enable_current_frame_load() FATAL(("IMPLEMENT: rse_enable_current_frame_load"));
#define rse_ensure_regs_loaded(number_of_bytes) FATAL(("IMPLEMENT: rse_ensure_regs_loaded"));
#define rse_invalidate_non_current_regs() FATAL(("IMPLEMENT: rse_invalidate_non_current_regs"));
#define rse_load(type) FATAL(("IMPLEMENT: rse_load"));
#define rse_new_frame(current_frame_size, new_frame_size) FATAL(("IMPLEMENT: rse_new_frame"));                          
#define rse_preserve_frame(preserved_frame_size) FATAL(("IMPLEMENT: rse_preserve_frame"));
#define rse_restore_frame(preserved_sol, growth, current_frame_size) FATAL(("IMPLEMENT: rse_restore_frame"));                
#define rse_store(type) FATAL(("IMPLEMENT: rse_store"));
#define rse_update_internal_stack_pointers(new_store_pointer) FATAL(("IMPLEMENT: rse_update_internal_stack_pointers")); 
#define tlb_access_key(vaddr) FATAL(("IMPLEMENT: tlb_access_key"));
#define tlb_broadcast_purge(rid, vaddr, size, type) FATAL(("IMPLEMENT: tlb_broadcast_purge"));
#define tlb_enter_privileged_code() FATAL(("IMPLEMENT: tlb_enter_privileged_code"));
#define tlb_grant_permission(vaddr, type, pl) FATAL(("IMPLEMENT: tlb_grant_permission"));
#define tlb_insert_data(slot, pte0, pte1, vaddr, rid, tr) FATAL(("IMPLEMENT: tlb_insert_data"));
#define tlb_insert_inst(slot, pte0, pte1, vaddr, rid, tr) FATAL(("IMPLEMENT: tlb_insert_inst"));
#define tlb_may_purge_dtc_entries(rid, vaddr, size) FATAL(("IMPLEMENT: tlb_may_purge_dtc_entries"));
#define tlb_may_purge_itc_entries(rid, vaddr, size) FATAL(("IMPLEMENT: tlb_may_purge_itc_entries"));
#define tlb_must_purge_dtc_entries(rid, vaddr, size) FATAL(("IMPLEMENT: tlb_must_purge_dtc_entries"));
#define tlb_must_purge_itc_entries(rid, vaddr, size) FATAL(("IMPLEMENT: tlb_must_purge_itc_entries"));
#define tlb_must_purge_dtr_entries(rid, vaddr, size) FATAL(("IMPLEMENT: tlb_must_purge_dtr_entries"));
#define tlb_must_purge_itr_entries(rid, vaddr, size) FATAL(("IMPLEMENT: tlb_must_purge_itr_entries"));
#define tlb_purge_translation_cache(loop) FATAL(("IMPLEMENT: tlb_purge_translation_cache"));
#define tlb_replacement_algorithm(tlb) FATAL(("IMPLEMENT: tlb_replacement_algorithm")); 
#define tlb_search_pkr(key) FATAL(("IMPLEMENT: tlb_search_pkr"));            
#define tlb_translate(vaddr, size, type, cpl, attr, defer) FATAL(("IMPLEMENT: tlb_translate"));
#define tlb_translate_nonaccess(vaddr, type) FATAL(("IMPLEMENT: tlb_translate_nonaccess")); 
#define tlb_vhpt_hash(vrn, vaddr61, rid, size) FATAL(("IMPLEMENT: tlb_vhpt_hash"));
#define tlb_vhpt_tag(vaddr, rid, size) FATAL(("IMPLEMENT: tlb_vhpt_tag")); 
#define unimplemented_physical_address(paddr) FATAL(("IMPLEMENT: unimplemented_physical_address"));
#define undefined() FATAL(("IMPLEMENT: undefined")); 
#define undefined_behavior() FATAL(("IMPLEMENT: undefined_behavior")); 
#define unimplemented_virtual_address(vaddr) FATAL(("IMPLEMENT: unimplemented_virtual_address"));
#define fp_update_fpsr(sf, tmp_fp_env) FATAL(("IMPLEMENT: fp_update_fpsr")); 
#define fp_update_psr(dest_freg) FATAL(("IMPLEMENT: fp_update_psr")); 
#define privileged_register_fault() FATAL(("IMPLEMENT: privileged_register_fault")); 
#define is_read_only_register(een, twee) FATAL(("IMPLEMENT: is_read_only_register")); 
#define reserved_register_field_fault() FATAL(("IMPLEMENT: reserved_register_field_fault"));

#define disabled_fp_register_fault(een, twee) FATAL(("IMPLEMENT: disabled_fp_register_fault")); 
#define fp_mul(een, twee) FATAL(("IMPLEMENT: fp_mul")); 
#define fp_add(een, twee, drie) FATAL(("IMPLEMENT: fp_add")); 
#define fp_ieee_round(een, twee) FATAL(("IMPLEMENT: fp_ieee_round")); 


#define fp_exception_trap(een) FATAL(("IMPLEMENT: fp_exception_trap")); 
#define fp_exception_fault(een) FATAL(("IMPLEMENT: fp_exception_fault")); 
#define fp_U128_add(een, twee) 0
#define fp_I64_x_I64_to_I128(een, twee) 0
#define fp_U64_x_U64_to_U128(een, twee) 0

#define lesser_signed(a, b) ((t_int64)a < (t_int64)b)
#define lesser_equal_signed(a, b) ((t_int64)a <= (t_int64)b) 
#define greater_signed(a, b) ((t_int64)a > (t_int64)b)
#define greater_equal_signed(a, b) ((t_int64)a >= (t_int64)b)
#define lesser(a, b) ((t_uint64)a < (t_uint64)b)

/* todo... setters en getters maken.*/

#define FP_SIGNIFICANT(fp) FATAL(("IMPLEMENT: FP_SIGNIFICANT"));
#define FP_SIGNIFICAND(fp) FATAL(("IMPLEMENT: FP_SIGNIFICAND")); 
#define FP_EXPONENT(fp) FATAL(("IMPLEMENT: FP_EXPONENT")); 
//#define FP_SIGN(fp) FATAL(("IMPLEMENT: FP_SIGN")); 
#define SET_FP_SIGN(fp, value)  (({FATAL(("IMPLEMENT: SET_FP_SIGN"));}), fp)
#define GET_FP_SIGN(fp)  FATAL(("IMPLEMENT: SET_FP_SIGN"));




/* todo...*/
//#define FP_SIGNIFICANT(fp) (({FATAL(("IMPLEMENT: FP_SIGNIFICANT"));}), fp)
//#define FP_SIGNIFICAND(fp) (({FATAL(("IMPLEMENT: FP_SIGNIFICAND"));}), fp) 
//#define FP_EXPONENT(fp) (({FATAL(("IMPLEMENT: FP_EXPONENT"));}), fp) 
#define FP_SIGN(fp) (({FATAL(("IMPLEMENT: FP_SIGN"));}), fp) 


#define EXTRACT_BITS64(data,upper,lower) (GETBITS64(data, lower, upper-lower+1))
#define EXTRACT_BIT64(data,bit) (GETBITS64(data, bit, 1))

#define SET_BITS64(dest,upper,lower,src) (dest = (((dest)&(~CREATE_MASK(upper-lower-1)<<lower)) |(((src)&CREATE_MASK(upper-lower-1))<<(lower))))
#define SET_BIT64(dest,bit,src) SET_BITS64(dest,bit,bit,src)


/**
* TODO: bitsequentie NATVAL
* */
#define NATVAL 0x1FFFE0000000000000000
#define FP_INTEGER_EXP 0
#define FP_SIGN_POSITIVE 0


/* dit geeft aan of we floating point operaties emuleren */
#define EMULATEFLOAT



/* This function delivers an illegal operation fault is r1 targets an out-of-frame register
 * This funtion is not implemented.
 * In the compiler we trust*/
static void check_target_register(t_ia64_reg r1)
{
  return;
}

static void check_target_register_sof(t_ia64_reg r1, t_uint8 newsof)
{
  return;
}

#define shift_right_unsigned(x, y) ((t_uint64)x)>>(y) 
#define shift_right_signed(x, y) ((t_int64)x)>>(y) 

static t_uint64 concatenate2(t_uint64 x1, t_uint64 x2)
{
  return ((x1&CREATE_MASK(32))<<32)|(x2&CREATE_MASK(32));
}

static t_uint64 concatenate4(t_uint64 x1, t_uint64 x2, t_uint64 x3, t_uint64 x4)
{
  return ((x1&CREATE_MASK(16))<<48)|((x2&CREATE_MASK(16))<<32)|((x3&CREATE_MASK(16))<<16)|(x4&CREATE_MASK(16));
}

static t_uint64 concatenate8(t_uint64 x1, t_uint64 x2, t_uint64 x3, t_uint64 x4, t_uint64 x5, t_uint64 x6, t_uint64 x7, t_uint64 x8)
{
  return ((x1&CREATE_MASK(8))<<56)|((x2&CREATE_MASK(8))<<48)|((x3&CREATE_MASK(8))<<40)|((x4&CREATE_MASK(8))<<32)|
    ((x5&CREATE_MASK(8))<<24)|((x6&CREATE_MASK(8))<<16)|((x7&CREATE_MASK(8))<<8)|(x8&CREATE_MASK(8));
}



static void Ia64InsEmulDetermineTag(t_ins * ins, t_procstate * state)
{
  t_reg reg;

  /* evt hier ook controleren of het wel een geldig(!=p0, r0, f0, f1) register is??? */
  
  if (!RegsetIn(INS_REGS_USE(ins),IA64INSOUT1(ins)))
    ProcStateSetTagTop(state,IA64INSOUT1(ins));

  REGSET_FOREACH_REG(INS_REGS_USE(ins),reg)
  {
    ProcStateJoinTags(state,IA64INSOUT1(ins),reg);
  }

}
/*
static void Ia64InsEmulDetermineTag2(t_ins * ins, t_procstate * state)
{
  t_reg reg;
  
  //evt hier ook controleren of het wel een geldig(!=p0, r0, f0, f1) register is??? 

  if (!RegsetIn(INS_REGS_USE(ins),IA64INSOUT2(ins)))
    ProcStateSetTagTop(state,IA64INSOUT2(ins));

  REGSET_FOREACH_REG(INS_REGS_USE(ins),reg)
  {
    ProcStateJoinTags(state,IA64INSOUT2(ins),reg);
  }

}*/

#define register_nat_consumption_fault(nummer) FATAL(("register_nat_consumption_fault"))

#define branch_predict(wh, ih, ret, target, tag) 

#define pr_phys_to_virt(phys_id) (({FATAL(("IMPLEMENT: pr_phys_to_virt"));}), 0)

#define illegal_operation_fault()  FATAL(("register_nat_consumption_fault"))


/* can be ignored */    
#define mem_implicit_prefetch(vaddr, hint, type) 

#define fp_is_natval(freg) (freg.flags & 1 == 1)    
/*
Volume 3: Instruction Reference                                                                             
*/


/*
Add

Format:         (qp) add r1 = r2, r3                       register_form      A1
                (qp) add r1 = r2, r3, 1       plus1_form, register_form        A1
                (qp) add r1 = imm, r3                          pseudo-op
                (qp) adds r1 = imm14, r3                    imm14_form         A4
                (qp) addl r1 = imm22, r3                    imm22_form         A5

*/

/*status: CHECKED */

static void Ia64Emulate_add(t_ins * ins, t_procstate * state)
{
  t_ia64_reg r1 = IA64INSOUT1(ins), r3;

  t_bool register_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A1;
  t_bool plus1_form = register_form && GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 2) == 1;
  
  t_uint64 tmp_nat, gr_r1_nat, gr_r2_nat, gr_r3_nat;
  t_int64 tmp_src, gr_r1, gr_r2, gr_r3;
  
  register_content reg;

  Ia64InsEmulDetermineTag(ins, state);

  __stats.add_total++;
  
  /* extract values from procstate*/
 
  if(register_form)
  {
    if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
      return; 
    
    r3 = IA64INSIN2(ins);
  }
  else
    r3 = IA64INSIN1(ins); 
  /* opgepast! moet hier IN1 zijn (zie corr.decodering): eerst registers...*/

  if(!Ia64CheckAndGetRegisterState(state, ins, r3, &gr_r3, &gr_r3_nat))
    return;  
  
  
  /* perform operations */
  
  check_target_register(r1);

  if (register_form)                            // register form
    tmp_src = gr_r2;
  else 
    tmp_src = IA64INSIMM1(ins);
  
  tmp_nat = (register_form ? gr_r2_nat : 0);
  
  if (plus1_form)
    gr_r1 = tmp_src + gr_r3 + 1;
  else
    gr_r1 = tmp_src + gr_r3;

  gr_r1_nat = tmp_nat || gr_r3_nat;
  
  /* nu nog resultaten wegschrijven... */
  
  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  
  
  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);  

  __stats.add_done++;
}




/*
Add Pointer

Format:         (qp) addp4 r1 = r2, r3                       register_form      A1
                (qp) addp4 r1 = imm14, r3                    imm14_form         A4

*/
/*status: CHECKED */


static void Ia64Emulate_addp4(t_ins * ins, t_procstate * state)
{ 
  t_ia64_reg r1 = IA64INSOUT1(ins), r3;

  t_bool register_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A1;

  t_uint64 tmp_nat, gr_r1_nat, gr_r2_nat, gr_r3_nat;
  t_int64 tmp_src, gr_r1, gr_r2, gr_r3;
  t_int64 tmp_res;
  
  register_content reg;


  Ia64InsEmulDetermineTag(ins, state);

  __stats.addp4_total++;
  
  /* Extract values form procstate */
  
  if(register_form)
  {
    if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
      return; 
 
    r3 = IA64INSIN2(ins);

  }
  else
    r3 = IA64INSIN1(ins); 

  if(!Ia64CheckAndGetRegisterState(state, ins, r3, &gr_r3, &gr_r3_nat))
    return; 

  /* perform operations */
  
  check_target_register(r1);

  tmp_src = (register_form ? gr_r2 : IA64INSIMM1(ins));
  tmp_nat = (register_form ? gr_r2_nat : 0);

  tmp_res = tmp_src + gr_r3;
  tmp_res = ZERO_EXT(EXTRACT_BITS64(tmp_res,31,0), 32);
  SET_BITS64(tmp_res,62,61,EXTRACT_BITS64(gr_r3,31,30));
  gr_r1 = tmp_res;
  gr_r1_nat = tmp_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64)gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

  __stats.addp4_done++;
}  









/*
Allocate Stack Frame

Format:          (qp) alloc r1 = ar.pfs, i, l, o, r                                  M34

*/
/*status: CHECKED */


static void Ia64Emulate_alloc(t_ins * ins, t_procstate * state)
{
  t_ia64_reg r1=IA64INSOUT1(ins);
  t_uint8 tmp_sol, tmp_sof, tmp_sor;
  t_uint64 gr_r1, gr_r1_nat, ar_pfs, cfm;
  register_content reg;

  // tmp_sof, tmp_sol, tmp_sor are the fields encoded in the instruction 
  tmp_sof = IA64INSIMM2(ins);
  tmp_sol = IA64INSIMM1(ins);
  tmp_sor = IA64INSIMM3(ins)>>3;
  check_target_register_sof(r1, tmp_sof);

  __stats.alloc_total++;
  
  if (tmp_sof> 96 || (tmp_sor << 3)> tmp_sof || tmp_sol> tmp_sof || IA64INSPRED(ins) != IA64_P0)
    illegal_operation_fault();

  
  if(!Ia64CheckAndGetRegisterState(state, ins, IA64_CFM, &cfm, NULL))
    return;

  if (tmp_sor != GETBITS64(cfm,14,4) && (GETBITS64(cfm,18,7) != 0 || GETBITS64(cfm,25,7) != 0 || GETBITS64(cfm,32,6) != 0))
    reserved_register_field_fault();
    //FATAL(("Implement"));
  
  /*alat_frame_update(0, tmp_sof - CFM.sof);*/
  
  rse_new_frame(CFM.sof, tmp_sof);    // Make room for new registers; Mandatory 
                                      // RSE stores can raise faults listed below.
  SET_BITS64(cfm,6,0,tmp_sof);
  SET_BITS64(cfm,13,7,tmp_sol);
  SET_BITS64(cfm,17,14,tmp_sor);

  /* nu nog resultaten wegschrijven... */
  
  GenericSet_uint64(reg.i, cfm);
  ProcStateSetReg(state,IA64_CFM,reg);  

  gr_r1_nat = 0;
  
  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

  if(!Ia64CheckAndGetRegisterState(state, ins, IA64_AR_PFS, &ar_pfs, NULL))
    return;

  gr_r1 = ar_pfs;

  GenericSet_uint64(reg.i, gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);

  __stats.alloc_done++;
}









/*
Logical And

Format:         (qp) and r1 = r2, r3                     register_form           A1
                (qp) and r1 = imm8, r3                     imm8_form             A3

*/

/*status: CHECKED */

static void Ia64Emulate_and(t_ins * ins, t_procstate * state)
{
  t_ia64_reg r1 = IA64INSOUT1(ins), r3;

  t_bool register_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A3;

  register_content reg;


  t_uint64 tmp_nat, gr_r1_nat, gr_r2_nat, gr_r3_nat;
  t_int64 tmp_src, gr_r1, gr_r2, gr_r3;
  
  Ia64InsEmulDetermineTag(ins, state);

  __stats.and_total++;
  
  if(register_form)
  {
    if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
      return; 
    
    r3 = IA64INSIN2(ins);
    
  }
  else
    r3 = IA64INSIN1(ins); 
  /* opgepast! moet hier IN1 zijn (zie corr.decodering): eerst registers...*/

  if(!Ia64CheckAndGetRegisterState(state, ins, r3, &gr_r3, &gr_r3_nat))
    return;  


  /* perform operations */
  check_target_register(r1);

  tmp_src = (register_form ? gr_r2 : IA64INSIMM1(ins));
  tmp_nat = (register_form ? gr_r2_nat : 0);

  gr_r1 = tmp_src & gr_r3;
  gr_r1_nat = tmp_nat || gr_r3_nat;

  
  /* nu nog resultaten wegschrijven... */
  GenericSet_uint64(reg.i, (t_uint64)gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

  __stats.and_done++;
}



/*
And Complement

Format:         (qp) andcm r1 = r2, r3                      register_form       A1
                (qp) andcm r1 = imm8, r3                      imm8_form         A3

*/
/*status: /CHECKED
 * */

static void Ia64Emulate_andcm(t_ins * ins, t_procstate * state)
{
  t_ia64_reg r1 = IA64INSOUT1(ins), r3;

  t_bool register_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A3;

  register_content reg;

  t_uint64 tmp_nat, gr_r1_nat, gr_r2_nat, gr_r3_nat;
  t_int64 tmp_src, gr_r1, gr_r2, gr_r3;
  
  Ia64InsEmulDetermineTag(ins, state);

  __stats.andcm_total++;

  if(register_form)
  {
    if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
      return; 
    
    r3 = IA64INSIN2(ins);
  }
  else
    r3 = IA64INSIN1(ins); 
  /* opgepast! moet hier IN1 zijn (zie corr.decodering): eerst registers...*/

  if(!Ia64CheckAndGetRegisterState(state, ins, r3, &gr_r3, &gr_r3_nat))
    return;  
  
  
  /* perform operations */
  check_target_register(r1);

  tmp_src = (register_form ? gr_r2 : IA64INSIMM1(ins));
  tmp_nat = (register_form ? gr_r2_nat : 0);

  gr_r1 = tmp_src & ~gr_r3;
  gr_r1_nat = tmp_nat || gr_r3_nat;
  
  /* nu nog resultaten wegschrijven... */
  GenericSet_uint64(reg.i, (t_uint64)gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

  __stats.andcm_done++;
}










/*
Branch

Format:          (qp) br.btype.bwh.ph.dh target25                                ip_relative_form           B1
                 (qp) br.btype.bwh.ph.dh b1 = target25           call_form, ip_relative_form                B3
                         br.btype.bwh.ph.dh target25        counted_form, ip_relative_form                  B2
                         br.ph.dh target25 pseudo-op
                 (qp) br.btype.bwh.ph.dh b2                                        indirect_form            B4
                 (qp) br.btype.bwh.ph.dh b1 = b2                        call_form, indirect_form            B5
                         br.ph.dh b2                                                   pseudo-op

*/

static void Ia64Emulate_br(t_ins * ins, t_procstate * state)
{
  __stats.br_skipped++;
  //ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0
  if (ip_relative_form)                                       // determine branch target
    tmp_IP = IP + SIGN_EXT((imm21 << 4), 25);
  else // indirect_form
    tmp_IP = br_b2;
  
  if (btype != 'ia')                                  // for Itanium branches,
    tmp_IP = tmp_IP & ~0xf;                    //  ignore bottom 4 bits of target
  
  
  lower_priv_transition = 0;


  switch (btype) 
  {
    case 'cond':                              // simple conditional branch
      tmp_taken = PR[qp];
      break;


    case 'call':                              // call saves a return link
      tmp_taken = PR[qp];
      if (tmp_taken) 
      {
        br_b1 = IP + 16;
        
        AR[PFS].pfm = CFM;               // ... and saves the stack frame
        AR[PFS].pec = AR[EC];
        AR[PFS].ppl = PSR.cpl;
        
        
        alat_frame_update(CFM.sol, 0);
        rse_preserve_frame(CFM.sol);
        CFM.sof -= CFM.sol;              // new frame size is size of outs
        CFM.sol = 0;
        CFM.sor = 0;
        CFM.rrb.gr = 0;
        CFM.rrb.fr = 0;
        CFM.rrb.pr = 0;
      }
      break;


    case 'ret':                               // return restores stack frame
      tmp_taken = PR[qp];
      if (tmp_taken) 
      {
        // tmp_growth indicates the amount to move logical TOP *up*:
        // tmp_growth = sizeof(previous out) - sizeof(current frame)
        // a negative amount indicates a shrinking stack
        tmp_growth = (AR[PFS].pfm.sof - AR[PFS].pfm.sol) - CFM.sof;
        alat_frame_update(-AR[PFS].pfm.sol, 0);
        rse_fatal = rse_restore_frame(AR[PFS].pfm.sol, tmp_growth, CFM.sof);
        if (rse_fatal) 
        { // See Section 6.4 "RSE Operation" in Volume 2
          CFM.sof = 0;
          CFM.sol = 0;
          CFM.sor = 0;
          CFM.rrb.gr = 0;
          CFM.rrb.fr = 0;
          CFM.rrb.pr = 0;
        } 
        else // normal branch return
          CFM = AR[PFS].pfm;
    
        rse_enable_current_frame_load();
        AR[EC] = AR[PFS].pec;
        if (PSR.cpl u< AR[PFS].ppl) 
        {        // ... and restores privilege
          PSR.cpl = AR[PFS].ppl;
          lower_priv_transition = 1;
        }
      }
      break;


    case 'ia':                                    // switch to IA mode
      tmp_taken = 1;
      if (qp != 0)
        illegal_operation_fault();
      if (AR[BSPSTORE] != AR[BSP])
        illegal_operation_fault();
      if (PSR.di)
        disabled_instruction_set_transition_fault();
      
      PSR.is = 1;                           // set IA-32 Instruction Set Mode
      CFM.sof = 0;                          //force current stack frame
      CFM.sol = 0;                          //to zero
      CFM.sor = 0;
      CFM.rrb.gr = 0;
      CFM.rrb.fr = 0;
      CFM.rrb.pr = 0;
      rse_invalidate_non_current_regs();
      //compute effective instruction pointer
      SET_BITS64(EIP,31,0,EXTRACT_BITS64(tmp_IP,31,0) - AR[CSD].Base);
      
      // Note the register stack is disabled during IA-32 instruction
      // set execution
      break;

    case 'cloop':                            // simple counted loop
      if (slot != 2)
        illegal_operation_fault();
      tmp_taken = (AR[LC] != 0);
      if (AR[LC] != 0)
        AR[LC]--;
      break;


    case 'ctop': 
    case 'cexit':                            // SW pipelined counted loop
      if (slot != 2)
        illegal_operation_fault();
      if (btype == 'ctop') tmp_taken =  ((AR[LC] != 0) || (AR[EC] u> 1));
      if (btype == 'cexit')tmp_taken = !((AR[LC] != 0) || (AR[EC] u> 1));
      if (AR[LC] != 0) 
      {
        AR[LC]--;
        AR[EC] = AR[EC];
        PR[63] = 1;
        rotate_regs();
      } 
      else if (AR[EC] != 0) 
      {
        AR[LC] = AR[LC];
        AR[EC]--;
        PR[63] = 0;
        rotate_regs();
      } 
      else 
      {
        AR[LC] = AR[LC];
        AR[EC] = AR[EC];
        PR[63] = 0;
        CFM.rrb.gr = CFM.rrb.gr;
        CFM.rrb.fr = CFM.rrb.fr;
        CFM.rrb.pr = CFM.rrb.pr;
      }
      break;


    case 'wtop':
    case 'wexit':                            // SW pipelined while loop
      if (slot != 2)
        illegal_operation_fault();
      if (btype == 'wtop') tmp_taken =  (PR[qp] || (AR[EC] u> 1));
      if (btype == 'wexit')tmp_taken = !(PR[qp] || (AR[EC] u> 1));
      if (PR[qp]) 
      {
        AR[EC] = AR[EC];
        PR[63] = 0;
        rotate_regs();
      } 
      else if (AR[EC] != 0) 
      {
        AR[EC]--;
        PR[63] = 0;
        rotate_regs();
      } 
      else 
      {
        AR[EC] = AR[EC];
        PR[63] = 0;
        CFM.rrb.gr = CFM.rrb.gr;
        CFM.rrb.fr = CFM.rrb.fr;
        CFM.rrb.pr = CFM.rrb.pr;
      }
      break;
  }
  
  if (tmp_taken) 
  {
    taken_branch = 1;
    IP = tmp_IP;     // set the new value for IP
    if ((PSR.it && unimplemented_virtual_address(tmp_IP))
      || (!PSR.it && unimplemented_physical_address(tmp_IP)))
      unimplemented_instruction_address_trap(lower_priv_transition, tmp_IP);
    if (lower_priv_transition && PSR.lp)
      lower_privilege_transfer_trap();
    if (PSR.tb)
      taken_branch_trap();
  }
}


#endif








/*
Break

Format:         (qp) break imm21                                 pseudo-op
                (qp) break.i imm21                              i_unit_form     I19
                (qp) break.b imm21                           b_unit_form         B9
                (qp) break.m imm21                          m_unit_form         M37
                (qp) break.f imm21                              f_unit_form     F15
                (qp) break.x imm62                           x_unit_form         X1

*/

static void Ia64Emulate_break(t_ins * ins, t_procstate * state)
{
  __stats.break_skipped++;
  //ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0
  if (b_unit_form)
    immediate = 0;
  else if (x_unit_form)
    immediate = ZERO_EXT(imm62, 21);
  else // i_unit_form || m_unit_form || f_unit_form
    immediate = ZERO_EXT(imm21, 21);
  
  break_instruction_fault(immediate);
}

#endif









/*
Branch Long

Format:         (qp) brl.btype.bwh.ph.dh target64                                        X3
                (qp) brl.btype.bwh.ph.dh b1 = target64               call_form           X4
                         brl.ph.dh target64 pseudo-op

*/


static void Ia64Emulate_brl(t_ins * ins, t_procstate * state)
{
  __stats.brl_skipped++;
  //ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0
  tmp_IP = IP + (imm60 << 4);                    // determine branch target
  if (!followed_by_stop())
    undefined_behavior();
  if (!long_branch_implemented())
    illegal_operation_fault();
  
  
  switch (btype) 
  {
    case 'cond':                               // simple conditional branch
      tmp_taken = PR[qp];
      break;
  
  
    case 'call':                               // call saves a return link
      tmp_taken = PR[qp];
      if (tmp_taken) 
      {
        br_b1 = IP + 16;
        
        AR[PFS].pfm = CFM;                // ... and saves the stack frame
        AR[PFS].pec = AR[EC];
        AR[PFS].ppl = PSR.cpl;
        
        
        alat_frame_update(CFM.sol, 0);
        rse_preserve_frame(CFM.sol);
        CFM.sof -= CFM.sol;               // new frame size is size of outs
        CFM.sol = 0;
        CFM.sor = 0;
        CFM.rrb.gr = 0;
        CFM.rrb.fr = 0;
        CFM.rrb.pr = 0;
      }
      break;
  }
  
  if (tmp_taken) 
  {
    taken_branch = 1;
    IP = tmp_IP;                               // set the new value for IP
    if ((PSR.it && unimplemented_virtual_address(tmp_IP))
      || (!PSR.it && unimplemented_physical_address(tmp_IP)))
      unimplemented_instruction_address_trap(0,tmp_IP);
    if (PSR.tb)
      taken_branch_trap();
  }
}

#endif








/*
Branch Predict

Format:         brp.ipwh.ih target25, tag13                      ip_relative_form     B6
                brp.indwh.ih b2, tag13                              indirect_form     B7
                brp.ret.indwh.ih b2, tag13         return_form, indirect_form         B7


*/



static void Ia64Emulate_brp(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0
  tmp_tag = IP + SIGN_EXT((timm9 << 4), 13);
  if (ip_relative_form) 
  {
    tmp_target = IP + SIGN_EXT((imm21 << 4), 25);
    tmp_wh = ipwh;
  } 
  else 
  { // indirect_form
    tmp_target = br_b2;
    tmp_wh = indwh;
  }
  branch_predict(tmp_wh, ih, return_form, tmp_target, tmp_tag);
}

#endif












/*
Bank Switch

Format:           bsw.0                                             zero_form            B8
                  bsw.1                                             one_form             B8

*/

static void Ia64Emulate_bsw(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0
  if (!followed_by_stop())
    illegal_operation_fault();
  
  
  if (PSR.cpl != 0)
    privileged_operation_fault(0);
  
  
  if (zero_form)
    PSR.bn = 0;
  else // one_form
    PSR.bn = 1;
}

#endif








/*
Speculation Check

Format:         (qp) chk.s r2, target25                                               pseudo-op
                (qp) chk.s.i r2, target25                        control_form, i_unit_form, gr_form         I20
                (qp) chk.s.m r2, target25                      control_form, m_unit_form, gr_form           M20
                (qp) chk.s f2, target25                                        control_form, fr_form        M21
                (qp) chk.a.aclr r1, target25                                      data_form, gr_form        M22
                (qp) chk.a.aclr f1, target25                                       data_form, fr_form       M23
*/
/* Als we geladen hebben, komt het uit het statische codegedeelte en zal dus niet veranderd zijn
*  Als we niet geladen hebben, zal het al bot staan. */
static void Ia64Emulate_chk(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));  
}

#if 0
  if (control_form) 
  {
    if (fr_form && (tmp_isrcode = fp_reg_disabled(f2, 0, 0, 0)))
      disabled_fp_register_fault(tmp_isrcode, 0);
    check_type = gr_form ? CHKS_GENERAL : CHKS_FLOAT;
    fail = (gr_form && gr_r2_nat) || (fr_form && fr_f2 == NATVAL);
  } 
  else 
  {                                              // data_form
    if (gr_form) 
    {
      reg_type   = GENERAL;
      check_type = CHKA_GENERAL;
      alat_index = r1;
      always_fail = (alat_index == 0);
    } 
    else 
    {                                         // fr_form
      reg_type   = FLOAT;
      check_type = CHKA_FLOAT;
      alat_index = f1;
      always_fail = ((alat_index == 0) || (alat_index == 1));
    }
    fail = (always_fail || (!alat_cmp(reg_type, alat_index)));
  }
  if (fail) 
  {
    if (check_branch_implemented(check_type)) 
    {
      taken_branch = 1;
      IP = IP + SIGN_EXT((imm21 << 4), 25);
      if ((PSR.it && unimplemented_virtual_address(IP))
        || (!PSR.it && unimplemented_physical_address(IP)))
        unimplemented_instruction_address_trap(0, IP);
      if (PSR.tb)
        taken_branch_trap();
    } 
    else
      speculation_fault(check_type, ZERO_EXT(imm21, 21));
  } 
  else if (data_form && (aclr == 'clr'))
    alat_inval_single_entry(reg_type, alat_index);
}

#endif









/*
Clear RRB

Format:         clrrrb                                             all_form        B8
                clrrrb.pr                                        pred_form         B8
*/

static void Ia64Emulate_clrrrb(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (!followed_by_stop())
    illegal_operation_fault();
  
  
  if (all_form) 
  {
    CFM.rrb.gr = 0;
    CFM.rrb.fr = 0;
    CFM.rrb.pr = 0;
  } 
  else 
  { // pred_form
    CFM.rrb.pr = 0;
  }
}

#endif



#define bOp(x)   (((t_uint64) ((x) & 0xf)) << 37)
#define bC(x)    (((t_uint64) ((x) & 0x1)) << 12)
#define bX2(x)   (((t_uint64) ((x) & 0x3)) << 34)
#define bTb(x)   (((t_uint64) ((x) & 0x1)) << 36)
#define bTa(x)   (((t_uint64) ((x) & 0x1)) << 33)

#define _OpX2TbTaC(a,b,c,d,e)  (bOp (a) | bX2 (b) | bTb (c) | bTa (d) | bC (e))
#define _OpX2TaC(a,b,c,d)      (bOp (a) | bX2 (b) | bTa (c) | bC (d))  


/*
Compare

Format:         (qp) cmp.crel.ctype p1, p2 = r2, r3                   register_form           A6
                (qp) cmp.crel.ctype p1, p2 = imm8, r3                  imm8_form              A8
                (qp) cmp.crel.ctype p1, p2 = r0, r3    parallel_inequality_form               A7
                (qp) cmp.crel.ctype p1, p2 = r3, r0                       pseudo-op
*/
/*status: NOT CHECKED */

static void Ia64Emulate_cmp(t_ins * ins, t_procstate * state)
{
  
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0

  t_ia64_reg p1=IA64INSOUT1(ins), p2=IA64INSOUT2(ins), r3;

  t_bool register_form=IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A6;
  t_bool par_ineq_form=IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A7;
  t_bool tmp_rel;


  t_uint64 tmp_nat, pr_p1, pr_p2, pr_qp, gr_r2_nat, gr_r3_nat;
  t_int64 gr_r2, gr_r3, tmp_src;
  
  register_content reg;

  /* extract values from procstate */

  __stats.cmp_total++;

  if(IA64INSPRED(ins) == IA64_P0)
    pr_qp = 1;
  else
    if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSPRED(ins), &pr_qp, NULL))
      return;
  
  
  if(register_form || par_ineq_form)
  {
    if(register_form)
    {
      if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
        return; 
    }
    else
    {
      gr_r2 = IA64_R0;
      gr_r2_nat = 0; 
    }
    
    
    r3 = IA64INSIN2(ins);
    
  }
  else
    r3 = IA64INSIN1(ins); 
  /* opgepast! moet hier IN1 zijn (zie corr.decodering): eerst registers...*/

  if(!Ia64CheckAndGetRegisterState(state, ins, r3, &gr_r3, &gr_r3_nat))
    return;  






  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
  * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
  * functie worden gemodelleerd !!!
   */

  /* perform operations */
  
  if (pr_qp) 
  {
    if (p1 == p2)
      illegal_operation_fault();

    tmp_nat = (register_form ? gr_r2_nat : 0) || gr_r3_nat;
    if (register_form)
      tmp_src = gr_r2;
    else
      tmp_src = IA64INSIMM1(ins);


    switch(IA64OPCODEINFO(ins).opcode)
    {
      /*case eq*/
      case _OpX2TbTaC (0xc, 0, 0, 1, 0):
      case _OpX2TaC   (0xc, 2, 1, 0):
      case _OpX2TbTaC (0xd, 0, 0, 1, 0):
      case _OpX2TaC   (0xd, 2, 1, 0):
      case _OpX2TbTaC (0xe, 0, 0, 0, 0):
      case _OpX2TbTaC (0xe, 0, 0, 0, 1):
      case _OpX2TbTaC (0xe, 0, 0, 1, 0):
      case _OpX2TaC   (0xe, 2, 0, 0):
      case _OpX2TaC   (0xe, 2, 0, 1):
      case _OpX2TaC   (0xe, 2, 1, 0):
	tmp_rel = tmp_src == gr_r3;
	break;
	/*case ne*/
      case _OpX2TbTaC (0xc, 0, 0, 1, 1):
      case _OpX2TaC   (0xc, 2, 1, 1):
      case _OpX2TbTaC (0xd, 0, 0, 1, 1):
      case _OpX2TaC   (0xd, 2, 1, 1):
      case _OpX2TbTaC (0xe, 0, 0, 1, 1):
      case _OpX2TaC   (0xe, 2, 1, 1):
	tmp_rel = tmp_src != gr_r3;
	break;
	/*case lt*/
      case _OpX2TbTaC (0xc, 0, 0, 0, 0):
      case _OpX2TbTaC (0xc, 0, 0, 0, 1):
      case _OpX2TbTaC (0xc, 0, 1, 1, 1):
      case _OpX2TaC   (0xc, 2, 0, 0):
      case _OpX2TaC   (0xc, 2, 0, 1):
      case _OpX2TbTaC (0xd, 0, 1, 1, 1):
      case _OpX2TbTaC (0xe, 0, 1, 1, 1):
	tmp_rel = lesser_signed(tmp_src,  gr_r3);
	break;
	/*case le*/
      case _OpX2TbTaC (0xc, 0, 1, 0, 1):
      case _OpX2TbTaC (0xd, 0, 1, 0, 1):
      case _OpX2TbTaC (0xe, 0, 1, 0, 1):
	tmp_rel = lesser_equal_signed(tmp_src, gr_r3);
	break;
	/*case gt*/
      case _OpX2TbTaC (0xc, 0, 1, 0, 0):
      case _OpX2TbTaC (0xd, 0, 1, 0, 0):
      case _OpX2TbTaC (0xe, 0, 1, 0, 0):
	tmp_rel = greater_signed(tmp_src,  gr_r3);
	break;
	/*case ge*/
      case _OpX2TbTaC (0xc, 0, 1, 1, 0):
      case _OpX2TbTaC (0xd, 0, 1, 1, 0):
      case _OpX2TbTaC (0xe, 0, 1, 1, 0):
	tmp_rel = greater_equal_signed(tmp_src, gr_r3);
	break;
	/*case ltu*/
      case _OpX2TbTaC (0xd, 0, 0, 0, 0):
      case _OpX2TbTaC (0xd, 0, 0, 0, 1):
      case _OpX2TaC   (0xd, 2, 0, 0):
      case _OpX2TaC   (0xd, 2, 0, 1):
	tmp_rel = lesser(tmp_src, gr_r3);
	break;
      default:
	FATAL(("Wrong type of compare encountered in emulation !!! This shouldn't happen !"));

    }
    switch (IA64OPCODEINFO(ins).opcode) 
    {

      /* and-type compare */
      case _OpX2TbTaC (0xc, 0, 0, 1, 0):
      case _OpX2TbTaC (0xc, 0, 0, 1, 1):
      case _OpX2TbTaC (0xc, 0, 1, 0, 0):
      case _OpX2TbTaC (0xc, 0, 1, 0, 1):
      case _OpX2TbTaC (0xc, 0, 1, 1, 0):
      case _OpX2TbTaC (0xc, 0, 1, 1, 1):
      case _OpX2TaC   (0xc, 2, 1, 0):
      case _OpX2TaC   (0xc, 2, 1, 1):
	if (tmp_nat || !tmp_rel) 
	{
	  pr_p1 = 0;
	  pr_p2 = 0;
	}
	break;
	/* or-type compare*/
      case _OpX2TbTaC (0xd, 0, 1, 0, 0):
      case _OpX2TbTaC (0xd, 0, 1, 0, 1):
      case _OpX2TbTaC (0xd, 0, 1, 1, 0):
      case _OpX2TbTaC (0xd, 0, 1, 1, 1):
      case _OpX2TbTaC (0xd, 0, 0, 1, 0):
      case _OpX2TbTaC (0xd, 0, 0, 1, 1):
      case _OpX2TaC   (0xd, 2, 1, 0):
      case _OpX2TaC   (0xd, 2, 1, 1):	  
	if (!tmp_nat && tmp_rel) 
	{
	  pr_p1 = 1;
	  pr_p2 = 1;
	}
	break;
	/* or.andcm-type compare*/
      case _OpX2TbTaC (0xe, 0, 0, 1, 0):
      case _OpX2TbTaC (0xe, 0, 0, 1, 1):
      case _OpX2TbTaC (0xe, 0, 1, 0, 0):
      case _OpX2TbTaC (0xe, 0, 1, 0, 1):
      case _OpX2TbTaC (0xe, 0, 1, 1, 0):
      case _OpX2TbTaC (0xe, 0, 1, 1, 1):
      case _OpX2TaC   (0xe, 2, 1, 0):
      case _OpX2TaC   (0xe, 2, 1, 1):
	if (!tmp_nat && tmp_rel) 
	{
	  pr_p1 = 1;
	  pr_p2 = 0;
	}
	break;
	/* unc-type compare*/
      case _OpX2TbTaC (0xc, 0, 0, 0, 1):
      case _OpX2TaC   (0xc, 2, 0, 1):
      case _OpX2TbTaC (0xd, 0, 0, 0, 1):
      case _OpX2TaC   (0xd, 2, 0, 1):
      case _OpX2TbTaC (0xe, 0, 0, 0, 1):
      case _OpX2TaC   (0xe, 2, 0, 1):
      default:                                         /* normal compare*/
	if (tmp_nat) 
	{
	  pr_p1 = 0;
	  pr_p2 = 0;
	} else 
	{
	  pr_p1 = tmp_rel;
	  pr_p2 = !tmp_rel;
	}
	break;
    }

    /* nu nog resultaten wegschrijven... */

    GenericSet_uint64(reg.i, pr_p1);
    ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

    GenericSet_uint64(reg.i, pr_p2);
    ProcStateSetReg(state,IA64INSOUT2(ins),reg);  
  } 
  else 
  {
    switch(IA64OPCODEINFO(ins).opcode) 
    {
      case _OpX2TbTaC (0xc, 0, 0, 0, 1):
      case _OpX2TaC   (0xc, 2, 0, 1):
      case _OpX2TbTaC (0xd, 0, 0, 0, 1):
      case _OpX2TaC   (0xd, 2, 0, 1):
      case _OpX2TbTaC (0xe, 0, 0, 0, 1):
      case _OpX2TaC   (0xe, 2, 0, 1):
	if (p1 == p2)
	  illegal_operation_fault();
	pr_p1 = 0;
	pr_p2 = 0;

	/* nu nog resultaten wegschrijven... */

	GenericSet_uint64(reg.i, pr_p1);
	ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

	GenericSet_uint64(reg.i, pr_p2);
	ProcStateSetReg(state,IA64INSOUT2(ins),reg);  
	break;
    }
  }

  __stats.cmp_done++;
}

#endif







/*
Compare 4 Bytes

Format:          (qp) cmp4.crel.ctype p1, p2 = r2, r3                      register_form     A6
                 (qp) cmp4.crel.ctype p1, p2 = imm8, r3                     imm8_form        A8
                 (qp) cmp4.crel.ctype p1, p2 = r0, r3       parallel_inequality_form         A7
                 (qp) cmp4.crel.ctype p1, p2 = r3, r0                          pseudo-op
*/
/*status: NOT CHECKED */


static void Ia64Emulate_cmp4(t_ins * ins, t_procstate * state)
{
  
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0

  t_ia64_reg p1=IA64INSOUT1(ins), p2=IA64INSOUT2(ins), r3;

  t_bool register_form=IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A6;
  t_bool par_ineq_form=IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A7;
  
  t_bool tmp_rel;

  t_uint64 tmp_nat, pr_p1, pr_p2, pr_qp, gr_r2_nat, gr_r3_nat;
  t_uint64 gr_r2, gr_r3, tmp_src;

  register_content reg;

  /* extract values from procstate*/

  __stats.cmp4_total++;


  if(IA64INSPRED(ins) == IA64_P0)
    pr_qp = 1;
  else
    if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSPRED(ins), &pr_qp, NULL))
      return;
  
  
  if(register_form || par_ineq_form)
  {
    if(register_form)
    {
      if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
        return; 
    }
    else
    {
      gr_r2 = IA64_R0;
      gr_r2_nat = 0; 
    }
    
    
    r3 = IA64INSIN2(ins);
    
  }
  else
    r3 = IA64INSIN1(ins); 
  /* opgepast! moet hier IN1 zijn (zie corr.decodering): eerst registers...*/

  if(!Ia64CheckAndGetRegisterState(state, ins, r3, &gr_r3, &gr_r3_nat))
    return;  




  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
  * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
  * functie worden gemodelleerd !!!
   */

  /* perform operations */
  
  if (pr_qp) 
  {
    if (p1 == p2)
      illegal_operation_fault();

    tmp_nat = (register_form ? gr_r2_nat : 0) || gr_r3_nat;
    if (register_form)
      tmp_src = gr_r2;
    else
      tmp_src = IA64INSIMM1(ins);


    switch(IA64OPCODEINFO(ins).opcode)
    {
      /*case eq*/
      case _OpX2TbTaC (0xc, 1, 0, 1, 0):
      case _OpX2TaC   (0xc, 3, 1, 0):
      case _OpX2TbTaC (0xd, 1, 0, 1, 0):
      case _OpX2TaC   (0xd, 3, 1, 0):
      case _OpX2TbTaC (0xe, 1, 0, 0, 0):
      case _OpX2TbTaC (0xe, 1, 0, 0, 1):
      case _OpX2TbTaC (0xe, 1, 0, 1, 0):
      case _OpX2TaC   (0xe, 3, 0, 0):
      case _OpX2TaC   (0xe, 3, 0, 1):
      case _OpX2TaC   (0xe, 3, 1, 0):
	tmp_rel = EXTRACT_BITS64(tmp_src,31,0) == EXTRACT_BITS64(gr_r3,31,0);
	break;
	/*case ne*/
      case _OpX2TbTaC (0xc, 1, 0, 1, 1):
      case _OpX2TaC   (0xc, 3, 1, 1):
      case _OpX2TbTaC (0xd, 1, 0, 1, 1):
      case _OpX2TaC   (0xd, 3, 1, 1):
      case _OpX2TbTaC (0xe, 1, 0, 1, 1):
      case _OpX2TaC   (0xe, 3, 1, 1):
	tmp_rel = EXTRACT_BITS64(tmp_src,31,0) != EXTRACT_BITS64(gr_r3,31,0);
        break;
	/*case lt*/
      case _OpX2TbTaC (0xc, 1, 0, 0, 0):
      case _OpX2TbTaC (0xc, 1, 0, 0, 1):
      case _OpX2TbTaC (0xc, 1, 1, 1, 1):
      case _OpX2TaC   (0xc, 3, 0, 0):
      case _OpX2TaC   (0xc, 3, 0, 1):
      case _OpX2TbTaC (0xd, 1, 1, 1, 1):
      case _OpX2TbTaC (0xe, 1, 1, 1, 1):
	tmp_rel = lesser_signed(SIGN_EXT(tmp_src, 32), SIGN_EXT(gr_r3, 32));
	break;
	/*case le*/
      case _OpX2TbTaC (0xc, 1, 1, 0, 1):
      case _OpX2TbTaC (0xd, 1, 1, 0, 1):
      case _OpX2TbTaC (0xe, 1, 1, 0, 1):
	tmp_rel = lesser_equal_signed(SIGN_EXT(tmp_src, 32), SIGN_EXT(gr_r3, 32));
	break;
	/*case gt*/
      case _OpX2TbTaC (0xc, 1, 1, 0, 0):
      case _OpX2TbTaC (0xd, 1, 1, 0, 0):
      case _OpX2TbTaC (0xe, 1, 1, 0, 0):
	tmp_rel = greater_signed(SIGN_EXT(tmp_src, 32), SIGN_EXT(gr_r3, 32));
	break;
	/*case ge*/
      case _OpX2TbTaC (0xc, 1, 1, 1, 0):
      case _OpX2TbTaC (0xd, 1, 1, 1, 0):
      case _OpX2TbTaC (0xe, 1, 1, 1, 0):
	tmp_rel = greater_equal_signed(SIGN_EXT(tmp_src, 32), SIGN_EXT(gr_r3, 32));
	break;
	/*case ltu*/
      case _OpX2TbTaC (0xd, 1, 0, 0, 0):
      case _OpX2TbTaC (0xd, 1, 0, 0, 1):
      case _OpX2TaC   (0xd, 3, 0, 0):
      case _OpX2TaC   (0xd, 3, 0, 1):
	tmp_rel = lesser(ZERO_EXT(tmp_src, 32), ZERO_EXT(gr_r3, 32));
	break;
      default:
	FATAL(("Wrong type of compare encountered in emulation !!! This shouldn't happen !"));

    }
    switch (IA64OPCODEINFO(ins).opcode) 
    {

      /* and-type compare */
      case _OpX2TbTaC (0xc, 1, 0, 1, 0):
      case _OpX2TbTaC (0xc, 1, 0, 1, 1):
      case _OpX2TbTaC (0xc, 1, 1, 0, 0):
      case _OpX2TbTaC (0xc, 1, 1, 0, 1):
      case _OpX2TbTaC (0xc, 1, 1, 1, 0):
      case _OpX2TbTaC (0xc, 1, 1, 1, 1):
      case _OpX2TaC   (0xc, 3, 1, 0):
      case _OpX2TaC   (0xc, 3, 1, 1):
	if (tmp_nat || !tmp_rel) 
	{
	  pr_p1 = 0;
	  pr_p2 = 0;
	}
	break;
	/* or-type compare*/
      case _OpX2TbTaC (0xd, 1, 1, 0, 0):
      case _OpX2TbTaC (0xd, 1, 1, 0, 1):
      case _OpX2TbTaC (0xd, 1, 1, 1, 0):
      case _OpX2TbTaC (0xd, 1, 1, 1, 1):
      case _OpX2TbTaC (0xd, 1, 0, 1, 0):
      case _OpX2TbTaC (0xd, 1, 0, 1, 1):
      case _OpX2TaC   (0xd, 3, 1, 0):
      case _OpX2TaC   (0xd, 3, 1, 1):	  
	if (!tmp_nat && tmp_rel) 
	{
	  pr_p1 = 1;
	  pr_p2 = 1;
	}
	break;
	/* or.andcm-type compare*/
      case _OpX2TbTaC (0xe, 1, 0, 1, 0):
      case _OpX2TbTaC (0xe, 1, 0, 1, 1):
      case _OpX2TbTaC (0xe, 1, 1, 0, 0):
      case _OpX2TbTaC (0xe, 1, 1, 0, 1):
      case _OpX2TbTaC (0xe, 1, 1, 1, 0):
      case _OpX2TbTaC (0xe, 1, 1, 1, 1):
      case _OpX2TaC   (0xe, 3, 1, 0):
      case _OpX2TaC   (0xe, 3, 1, 1):
	if (!tmp_nat && tmp_rel) 
	{
	  pr_p1 = 1;
	  pr_p2 = 0;
	}
	break;
	/* unc-type compare*/
      case _OpX2TbTaC (0xc, 1, 0, 0, 1):
      case _OpX2TaC   (0xc, 3, 0, 1):
      case _OpX2TbTaC (0xd, 1, 0, 0, 1):
      case _OpX2TaC   (0xd, 3, 0, 1):
      case _OpX2TbTaC (0xe, 1, 0, 0, 1):
      case _OpX2TaC   (0xe, 3, 0, 1):
      default:                                         /* normal compare*/
	if (tmp_nat) 
	{
	  pr_p1 = 0;
	  pr_p2 = 0;
	} else 
	{
	  pr_p1 = tmp_rel;
	  pr_p2 = !tmp_rel;
	}
	break;
    }

    /* nu nog resultaten wegschrijven... */

    GenericSet_uint64(reg.i, pr_p1);
    ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

    GenericSet_uint64(reg.i, pr_p2);
    ProcStateSetReg(state,IA64INSOUT2(ins),reg);  
  } 
  else 
  {
    switch(IA64OPCODEINFO(ins).opcode) 
    {
      case _OpX2TbTaC (0xc, 1, 0, 0, 1):
      case _OpX2TaC   (0xc, 3, 0, 1):
      case _OpX2TbTaC (0xd, 1, 0, 0, 1):
      case _OpX2TaC   (0xd, 3, 0, 1):
      case _OpX2TbTaC (0xe, 1, 0, 0, 1):
      case _OpX2TaC   (0xe, 3, 0, 1):
	if (p1 == p2)
	  illegal_operation_fault();
	pr_p1 = 0;
	pr_p2 = 0;

	/* nu nog resultaten wegschrijven... */

	GenericSet_uint64(reg.i, pr_p1);
	ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

	GenericSet_uint64(reg.i, pr_p2);
	ProcStateSetReg(state,IA64INSOUT2(ins),reg);  
	break;
    }
  }

  __stats.cmp4_done++;

}

#endif



/*
Compare and Exchange

Format:         (qp) cmpxchgsz.sem.ldhint r1 = [r3], r2, ar.ccv                           M16
*/


static void Ia64Emulate_cmpxchg(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0
  check_target_register(r1);

  if (gr_r3_nat || gr_r2_nat)
      register_nat_consumption_fault(SEMAPHORE);


  paddr = tlb_translate(gr_r3, sz, SEMAPHORE, PSR.cpl, &mattr,
                             &tmp_unused);


  if (!ma_supports_semaphores(mattr))
      unsupported_data_reference_fault(SEMAPHORE, gr_r3);

  if (sem == 'acq')
      val = mem_xchg_cond(AR[CCV], gr_r2, paddr, sz, UM.be, mattr,
                                 ACQUIRE, ldhint);
  else // 'rel'
      val = mem_xchg_cond(AR[CCV], gr_r2, paddr, sz, UM.be, mattr, 
                                 RELEASE, ldhint);
  val = ZERO_EXT(val, sz * 8);


  if (AR[CCV] == val)
      alat_inval_multiple_entries(paddr, sz);


  gr_r1 = val;
  gr_r1_nat = 0;
}

#endif









/*
Cover Stack Frame

Format:         cover                                                               B8
*/


static void Ia64Emulate_cover(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0
  if (!followed_by_stop())
       illegal_operation_fault();
  
  
  alat_frame_update(CFM.sof, 0);
  rse_preserve_frame(CFM.sof);
  if (PSR.ic == 0) {
       CR[IFS].ifm = CFM;
       CR[IFS].v = 1;
  }
  
  
  CFM.sof = 0;
  CFM.sol = 0;
  CFM.sor = 0;
  CFM.rrb.gr = 0;
  CFM.rrb.fr = 0;
  CFM.rrb.pr = 0;
}

#endif



/*HIER*/




/*
Compute Zero Index

Format:          (qp) czx1.l r1 = r3                       one_byte_form, left_form        I29
                 (qp) czx1.r r1 = r3                 one_byte_form, right_form             I29
                 (qp) czx2.l r1 = r3                       two_byte_form, left_form        I29
                 (qp) czx2.r r1 = r3                 two_byte_form, right_form             I29
*/
/*status: TODO */

static void Ia64Emulate_czx(t_ins * ins, t_procstate * state)
{ 
  __stats.other_done++; 
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  t_int64 gr_r1, gr_r3;
  t_uint64 gr_r1_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  t_bool one_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,27,1)==0;

  t_bool left_form = GETBITS64(IA64OPCODEINFO(ins).opcode,28,1)==0;
  
  register_content reg;

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  /* perform operations */
  
  check_target_register(r1);
  
  if (one_byte_form) {
         if (left_form) {                        // scan from most significant down
              if              ((gr_r3 & 0xff00000000000000ULL) == 0) gr_r1 = 0;
              else if ((gr_r3 & 0x00ff000000000000ULL) == 0) gr_r1 = 1;
              else if ((gr_r3 & 0x0000ff0000000000ULL) == 0) gr_r1 = 2;
              else if ((gr_r3 & 0x000000ff00000000ULL) == 0) gr_r1 = 3;
              else if ((gr_r3 & 0x00000000ff000000ULL) == 0) gr_r1 = 4;
              else if ((gr_r3 & 0x0000000000ff0000ULL) == 0) gr_r1 = 5;
              else if ((gr_r3 & 0x000000000000ff00ULL) == 0) gr_r1 = 6;
              else if ((gr_r3 & 0x00000000000000ffULL) == 0) gr_r1 = 7;
              else gr_r1 = 8;
         } else { // right_form                  scan from least significant up
              if              ((gr_r3 & 0x00000000000000ffULL) == 0) gr_r1 = 0;
              else if ((gr_r3 & 0x000000000000ff00ULL) == 0) gr_r1 = 1;
              else if ((gr_r3 & 0x0000000000ff0000ULL) == 0) gr_r1 = 2;
              else if ((gr_r3 & 0x00000000ff000000ULL) == 0) gr_r1 = 3;
              else if ((gr_r3 & 0x000000ff00000000ULL) == 0) gr_r1 = 4;
              else if ((gr_r3 & 0x0000ff0000000000ULL) == 0) gr_r1 = 5;
              else if ((gr_r3 & 0x00ff000000000000ULL) == 0) gr_r1 = 6;
              else if ((gr_r3 & 0xff00000000000000ULL) == 0) gr_r1 = 7;
              else gr_r1 = 8;
         }
  } else { // two_byte_form
         if (left_form) {                        // scan from most significant down
              if              ((gr_r3 & 0xffff000000000000ULL) == 0) gr_r1 = 0;
              else if ((gr_r3 & 0x0000ffff00000000ULL) == 0) gr_r1 = 1;
              else if ((gr_r3 & 0x00000000ffff0000ULL) == 0) gr_r1 = 2;
              else if ((gr_r3 & 0x000000000000ffffULL) == 0) gr_r1 = 3;
              else gr_r1 = 4;
         } else { // right_form                  scan from least significant up
              if              ((gr_r3 & 0x000000000000ffffULL) == 0) gr_r1 = 0;
              else if ((gr_r3 & 0x00000000ffff0000ULL) == 0) gr_r1 = 1;
              else if ((gr_r3 & 0x0000ffff00000000ULL) == 0) gr_r1 = 2;
              else if ((gr_r3 & 0xffff000000000000ULL) == 0) gr_r1 = 3;
              else gr_r1 = 4;
  
     }
  }
  gr_r1_nat = gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

}

#endif


/*
Deposit

Format:          (qp) dep r1 = r2, r3, pos6, len4        merge_form, register_form     I15
                 (qp) dep r1 = imm1, r3, pos6, len6        merge_form, imm_form        I14
                 (qp) dep.z r1 = r2, pos6, len6           zero_form, register_form     I12
                 (qp) dep.z r1 = imm8, pos6, len6            zero_form, imm_form       I13
*/



static void Ia64Emulate_dep(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint64 tmp_nat;
  t_int64 tmp_src;
  
  t_uint8 tmp_len;

  t_bool merge_form = GETBITS64(IA64OPCODEINFO(ins).opcode,37,5)==4 || GETBITS64(IA64OPCODEINFO(ins).opcode,34,2)==3;
  t_bool imm_form = (merge_form)? GETBITS64(IA64OPCODEINFO(ins).opcode,37,5)==4: GETBITS64(IA64OPCODEINFO(ins).opcode,26,1)==1;

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  if (!imm_form)
    gr_r2 = (t_int64) G_T_UINT64(reg.i);
  else if (merge_form)
    gr_r3 = (t_int64) G_T_UINT64(reg.i);


  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  if (! imm_form)
    gr_r2_nat = G_T_UINT64(reg.i);
  else if (merge_form)
    gr_r3_nat = G_T_UINT64(reg.i);

  if(!imm_form && merge_form)
  {
    if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r3 = (t_int64) G_T_UINT64(reg.i);

    if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r3_nat= G_T_UINT64(reg.i);
  }

  /* perform operations */

  check_target_register(r1);
  
  if (imm_form) 
  {
       tmp_src = (merge_form ? SIGN_EXT(IA64INSIMM1(ins),1) : SIGN_EXT(IA64INSIMM1(ins), 8));
       tmp_nat = merge_form ? gr_r3_nat : 0;
       tmp_len = IA64INSIMM3(ins) ;
  } 
  else 
  {                                          // register_form
       tmp_src = gr_r2;
       tmp_nat = (merge_form ? gr_r3_nat : 0) || gr_r2_nat;
       tmp_len = IA64INSIMM3(ins);
  }
  if (IA64INSIMM2(ins) + tmp_len > 64)
       tmp_len = 64 - IA64INSIMM2(ins);
  
  if (merge_form)
       gr_r1 = gr_r3;
  else // zero_form
       gr_r1 = 0;
  
  SET_BITS64(gr_r1,(IA64INSIMM2(ins) + tmp_len - 1),IA64INSIMM2(ins),EXTRACT_BITS64(tmp_src,(tmp_len - 1),0));
  gr_r1_nat = tmp_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

}

#endif








/*
Enter Privileged Code

Format:         epc                                                                    B8
*/

static void Ia64Emulate_epc(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (AR[PFS].ppl u< PSR.cpl)
         illegal_operation_fault();


  if (PSR.it)
         PSR.cpl = tlb_enter_privileged_code();
  else
         PSR.cpl = 0;
}

#endif







/*

Extract

Format:         (qp) extr r1 = r3, pos6, len6                 signed_form               I11
                (qp) extr.u r1 = r3, pos6, len6             unsigned_form               I11
*/
/*status: TODO */

static void Ia64Emulate_extr(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  t_int64 gr_r1, gr_r3;
  t_uint64 gr_r1_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint8 tmp_len;

  t_bool unsigned_form = GETBITS64(IA64OPCODEINFO(ins).opcode,13,1);

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  /* perform operations */

  
  check_target_register(r1);
  
  tmp_len = IA64INSIMM2(ins);
  
  if (IA64INSIMM1(ins) + tmp_len> 64)
     tmp_len = 64 - IA64INSIMM1(ins);
  
  if (unsigned_form)
     gr_r1 = ZERO_EXT(shift_right_unsigned(gr_r3, IA64INSIMM1(ins)), tmp_len);
  else // signed_form
     gr_r1 = SIGN_EXT(shift_right_unsigned(gr_r3, IA64INSIMM1(ins)), tmp_len);
  
  gr_r1_nat = gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

}

#endif







/*
Floating-point Absolute Maximum

Format:         (qp) famax.sf f1 = f2, f3                                        F8
*/

static void Ia64Emulate_famax(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      fminmax_exception_fault_check(f2, f3, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
          fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      tmp_right = fp_reg_read(fr_f2);
      tmp_left = fp_reg_read(fr_f3);
      FP_SIGN(tmp_right) = FP_SIGN_POSITIVE;
      FP_SIGN(tmp_left) = FP_SIGN_POSITIVE;
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      fr_f1 = tmp_bool_res ? fr_f2 : fr_f3;
  
      fp_update_fpsr(sf, tmp_fp_env);
  }
  
  
  fp_update_psr(f1);
}

#endif









/*
Floating-point Absolute Minimum

Format:         (qp) famin.sf f1 = f2, f3                                       F8
*/

static void Ia64Emulate_famin(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      fminmax_exception_fault_check(f2, f3, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
          fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      tmp_left = fp_reg_read(fr_f2);
      tmp_right = fp_reg_read(fr_f3);
      FP_SIGN(tmp_left) = FP_SIGN_POSITIVE;
      FP_SIGN(tmp_right) = FP_SIGN_POSITIVE;
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      fr_f1 = tmp_bool_res ? fr_f2 : fr_f3;
  
      fp_update_fpsr(sf, tmp_fp_env);
  }
  
  
  fp_update_psr(f1);
}

#endif








/*
Floating-point Logical And

Format:         (qp) fand f1 = f2, f3                                                 F9
*/

static void Ia64Emulate_fand(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      FP_SIGNIFICAND(fr_f1) = FP_SIGNIFICAND(fr_f2) & FP_SIGNIFICAND(fr_f3);
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  fp_update_psr(f1);
}

#endif








/*
Floating-point And Complement

Format:         (qp) fandcm f1 = f2, f3                                             F9
*/

static void Ia64Emulate_fandcm(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      FP_SIGNIFICAND(fr_f1) = FP_SIGNIFICAND(fr_f2) & ~FP_SIGNIFICAND(fr_f3);
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  fp_update_psr(f1);
}

#endif









/*
Flush Cache

Format:         (qp) fc r3                                                      M28
*/


static void Ia64Emulate_fc(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  itype = NON_ACCESS|FC|READ;
  if (gr_r3_nat)
     register_nat_consumption_fault(itype);
  tmp_paddr = tlb_translate_nonaccess(gr_r3, itype);
  mem_flush(tmp_paddr);
}


#endif









/*
Floating-point Check Flags

Format:         (qp) fchkf.sf target25                                            F14
*/

static void Ia64Emulate_fchkf(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  switch (sf) {
      case 's0':
          tmp_flags = AR[FPSR].sf0.flags;
          break;
      case 's1':
          tmp_flags = AR[FPSR].sf1.flags;
          break;
      case 's2':
          tmp_flags = AR[FPSR].sf2.flags;
          break;
      case 's3':
          tmp_flags = AR[FPSR].sf3.flags;
          break;
  }
  if ((tmp_flags & ~AR[FPSR].traps) || (tmp_flags & ~AR[FPSR].sf0.flags)) {
      if (check_branch_implemented(FCHKF)) {
          taken_branch = 1; 
          IP = IP + SIGN_EXT((imm21 << 4), 25);
          if ((PSR.it && unimplemented_virtual_address(IP))
                || (!PSR.it && unimplemented_physical_address(IP)))
                unimplemented_instruction_address_trap(0, IP);
          if (PSR.tb)
                taken_branch_trap();
      } else
          speculation_fault(FCHKF, ZERO_EXT(imm21, 21));
  }
}

#endif








/*
Floating-point Class

Format:          (qp) fclass.fcrel.fctype p1, p2 = f2, fclass9                             F5
*/


static void Ia64Emulate_fclass(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
   * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
   * functie worden gemodelleerd !!!
   */
  if (PR[qp]) 
  {
    if (p1 == p2)
         illegal_operation_fault();
  
  
    if (tmp_isrcode = fp_reg_disabled(f2, 0, 0, 0))
         disabled_fp_register_fault(tmp_isrcode, 0);
  
  
    tmp_rel = ((EXTRACT_BIT64(fclass9,0) && !FP_SIGN(fr_f2) || EXTRACT_BIT64(fclass9,1) && FP_SIGN(fr_f2))
                          &&    ((EXTRACT_BIT64(fclass9,2) && fp_is_zero(fr_f2))||
                                 (EXTRACT_BIT64(fclass9,3) && fp_is_unorm(fr_f2)) ||
                                 (EXTRACT_BIT64(fclass9,4) && fp_is_normal(fr_f2)) ||
                                 (EXTRACT_BIT64(fclass9,5) && fp_is_inf(fr_f2))
                                )
                          )
                   || (EXTRACT_BIT64(fclass9,6) && fp_is_snan(fr_f2))
                   || (EXTRACT_BIT64(fclass9,7) && fp_is_qnan(fr_f2))
                   || (EXTRACT_BIT64(fclass9,8) && fp_is_natval(fr_f2));
  
    tmp_nat = fp_is_natval(fr_f2) && (!EXTRACT_BIT64(fclass9,8));
  
    if (tmp_nat) {
         pr_p1 = 0;
         pr_p2 = 0;
    } else {
         pr_p1 = tmp_rel;
         pr_p2 = !tmp_rel;
    }
  } else {
    if (fctype == 'unc') {
         if (p1 == p2)
             illegal_operation_fault();
         pr_p1 = 0;
         pr_p2 = 0;
    }
  }
}

#endif









/*
Floating-point Clear Flags

Format:         (qp) fclrf.sf                                                F13
*/

static void Ia64Emulate_fclrf(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
                     fp_set_sf_flags(sf, 0);
}

#endif








/*
Floating-point Compare

Format:         (qp) fcmp.frel.fctype.sf p1, p2 = f2, f3                                      F4
*/


static void Ia64Emulate_fcmp(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
   * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
   * functie worden gemodelleerd !!!
   */
  if (PR[qp]) 
  {  
    if (p1 == p2)
         illegal_operation_fault();


    if (tmp_isrcode = fp_reg_disabled(f2, f3, 0, 0))
         disabled_fp_register_fault(tmp_isrcode, 0);


    if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
         pr_p1 = 0;
         pr_p2 = 0;
    } else {
         fcmp_exception_fault_check(f2, f3, frel, sf, &tmp_fp_env);
         if (fp_raise_fault(tmp_fp_env))
            fp_exception_fault(fp_decode_fault(tmp_fp_env));


         tmp_fr2 = fp_reg_read(fr_f2);
         tmp_fr3 = fp_reg_read(fr_f3);

         if (frel == 'eq')         tmp_rel = fp_equal(tmp_fr2,
                                                        tmp_fr3);
         else if (frel == 'lt')    tmp_rel = fp_less_than(tmp_fr2,
                                                        tmp_fr3);
         else if (frel == 'le')    tmp_rel = fp_lesser_or_equal(tmp_fr2, 
                                                        tmp_fr3);
         else if (frel == 'gt')    tmp_rel = fp_less_than(tmp_fr3, 
                                                        tmp_fr2);
         else if (frel == 'ge')    tmp_rel = fp_lesser_or_equal(tmp_fr3,
                                                        tmp_fr2);
         else if (frel == 'unord')tmp_rel = fp_unordered(tmp_fr2, 
                                                        tmp_fr3);
         else if (frel == 'neq') tmp_rel = !fp_equal(tmp_fr2, 
                                                        tmp_fr3);
         else if (frel == 'nlt') tmp_rel = !fp_less_than(tmp_fr2,
                                                        tmp_fr3);
         else if (frel == 'nle') tmp_rel = !fp_lesser_or_equal(tmp_fr2,
                                                        tmp_fr3);
         else if (frel == 'ngt') tmp_rel = !fp_less_than(tmp_fr3, 
                                                        tmp_fr2);
         else if (frel == 'nge') tmp_rel = !fp_lesser_or_equal(tmp_fr3,
                                                        tmp_fr2);
         else                      tmp_rel = !fp_unordered(tmp_fr2, 
                                                        tmp_fr3); //'ord'


         pr_p1 = tmp_rel;
         pr_p2 = !tmp_rel;

         fp_update_fpsr(sf, tmp_fp_env);
    }
} else {
    if (fctype == 'unc') {
         if (p1 == p2)
            illegal_operation_fault();
         pr_p1 = 0;
         pr_p2 = 0;
    }
}
}

#endif









/*
Convert Floating-point to Integer

Format:         (qp) fcvt.fx.sf f1 = f2                         signed_form       F10
                (qp) fcvt.fx.trunc.sf f1 = f2     signed_form, trunc_form         F10
                (qp) fcvt.fxu.sf f1 = f2                      unsigned_form       F10
                (qp) fcvt.fxu.trunc.sf f1 = f2 unsigned_form, trunc_form          F10
*/

static void Ia64Emulate_fcvt_fx(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, 0, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2)) {
      fr_f1 = NATVAL;
      fp_update_psr(f1);
  } else {
      tmp_default_result = fcvt_exception_fault_check(f2, signed_form,
                                                    trunc_form, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
           fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      if (fp_is_nan(tmp_default_result)) {
           FP_SIGNIFICAND(fr_f1) = INTEGER_INDEFINITE;
           FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
           FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
      } else {
           tmp_res = fp_ieee_rnd_to_int(fp_reg_read(fr_f2), &tmp_fp_env);
           if (FP_EXPONENT(tmp_res))
               FP_SIGNIFICAND(tmp_res) = fp_U64_rsh(
                       FP_SIGNIFICAND(tmp_res), (FP_INTEGER_EXP - FP_EXPONENT(tmp_res)));
           if (signed_form && FP_SIGN(tmp_res))
               FP_SIGNIFICAND(tmp_res) = (~FP_SIGNIFICAND(tmp_res)) + 1;
  
  
           FP_SIGNIFICAND(fr_f1) = FP_SIGNIFICAND(tmp_res);
           FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
           FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
      }
  
  
      fp_update_fpsr(sf, tmp_fp_env);
      fp_update_psr(f1);
      if (fp_raise_traps(tmp_fp_env))
           fp_exception_trap(fp_decode_trap(tmp_fp_env));
  }
}

#endif









/*
Convert Signed Integer to Floating-point

Format:         (qp) fcvt.xf f1 = f2                                             F11
*/

static void Ia64Emulate_fcvt_xf(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, 0, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2)) {
      fr_f1 = NATVAL;
  } else {
      tmp_res = fr_f2;
      if (EXTRACT_BIT64(FP_SIGNIFICAND(tmp_res),63)) {
          FP_SIGNIFICAND(tmp_res) = (~FP_SIGNIFICAND(tmp_res)) + 1;
          FP_SIGN(tmp_res) = 1;
      } else
          FP_SIGN(tmp_res) = 0;
  
  
      FP_EXPONENT(tmp_res) = FP_INTEGER_EXP;
      tmp_res = fp_normalize(tmp_res);
  
  
      FP_SIGNIFICAND(fr_f1) = FP_SIGNIFICAND(tmp_res);
      FP_EXPONENT(fr_f1) = FP_EXPONENT(tmp_res);
      FP_SIGN(fr_f1) = FP_SIGN(tmp_res);
  }
  fp_update_psr(f1);
  
}

#endif









/*
Fetch and Add Immediate

Format:         (qp) fetchadd4.sem.ldhint r1 = [r3], inc3     four_byte_form         M17
                (qp) fetchadd8.sem.ldhint r1 = [r3], inc3    eight_byte_form         M17
*/


static void Ia64Emulate_fetchadd(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  t_reloc * base_reloc;
  register_content reg, tmp_address;
  t_uint64 offset;
  t_lattice_level address_unknown;
  t_lattice_level base_reloc_unknown;
  t_lattice_level address_nat_unknown;

  t_bool four_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,30,1) == 1;

  t_uint64 val;

  t_ia64_reg r1 = IA64INSOUT1(ins), r3=IA64INSIN1(ins);

  t_uint64 gr_r3, gr_r3_nat, gr_r1_nat; 
  t_int64 gr_r1;

  t_section * sec;

  t_uint8 size;

  /* wat weten we over locatie? */
  base_reloc_unknown = ProcStateGetTag(state, r3,&base_reloc);
  address_unknown = ProcStateGetReg(state, r3,&tmp_address);
  gr_r3 = G_T_UINT64(tmp_address.i);
  address_nat_unknown = ProcStateGetReg(state, IA64GETNAT(r3),&reg);
  gr_r3_nat = G_T_UINT64(reg.i);

  if(base_reloc_unknown==TOP)
  {
    FATAL(("problem: no reloc in base register\n"));
  }
  if (address_unknown||base_reloc_unknown||address_nat_unknown)
  {
    ProcStateSetAllBot(state, INS_REGS_DEF(ins));
    ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));
    return; 
  }


  /*intel*/

  if(base_reloc->to->to_type==TO_SUB_SECTION)
  {
    sec = T_SECTION(base_reloc->to->base);
    if(sec->type != CODE_SECTION){
      ProcStateSetAllBot(state, INS_REGS_DEF(ins));
      ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));
    }

  }
  else{
    ProcStateSetAllBot(state, INS_REGS_DEF(ins));
    ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));
  }

  /*we weten dat we in codegebied lezen en veronderstellen dit constant */

#define ldtype(opcode) (GETBITS64(opcode, 30,6)&(CREATE_MASK(4)<<2)) 

  check_target_register(r1);

  size = four_byte_form? 4 : 8;

  if (gr_r3_nat)                         // fault on NaT address
    register_nat_consumption_fault();

  FATAL(("This code is not correct if 'real' generic addresses are used (both 32 and 64 bit support)\nContact Bruno in case of doubt"));
  
  /* This is the offending code : */
  // offset = G_T_UINT64(GENERIC_SUB(gr_r3,SECTION_GET_CADDRESS(sec)));
  /* I would solve this as (but I do not have the time to understand the code, so CAUTION! */

  offset =  G_T_UINT64(GENERIC_ADD_INT64(SECTION_GET_CADDRESS(sec), -((t_int64) gr_r3))); /* Bear in mind that the code will allways go wrong if gr_r3 > 0x7fffffffffffffff, but that is always the case */

  if(offset > G_T_UINT64(SECTION_SIZE(sec)))
  {
    ProcStateSetAllBot(state, INS_REGS_DEF(ins));
    ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));
  }

  switch (size)
  {
    case 4:
      val = (t_uint64)(*((t_uint32*)sec->data + offset/4)); 
      break;
    case 8:
      val = (t_uint64)(*((t_uint64*)sec->data + offset/8)); 
      break;
  }

  val+=IA64INSIMM1(ins);

  gr_r1 = ZERO_EXT(val, size * 8);
  gr_r1_nat = 0;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif






/*
Flush Register Stack

Format:         flushrs                                                           M25
*/

static void Ia64Emulate_flushrs(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  while (AR[BSPSTORE] != AR[BSP]) {
       rse_store(MANDATORY);                           // increments AR[BSPSTORE]
       deliver_unmasked_pending_external_interrupt();
  }
}

#endif




/*
Floating-point Multiply Add

Format:         (qp) fma.pc.sf f1 = f3, f4, f2                                     F1
*/

static void Ia64Emulate_fma(t_ins * ins, t_procstate * state)
{
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
#ifndef EMULATEFLOAT
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
#else 
  t_ia64_reg f1 = IA64INSOUT1(ins), f2 = IA64INSIN1(ins), f3 = IA64INSIN2(ins), f4 = IA64INSIN3(ins);

  t_float fr_f1, fr_f2, fr_f3, fr_f4;

  register_content reg;

  t_bool tmp_isrcode;

  t_uint64 opcode = IA64OPCODEINFO(ins).opcode;
  t_uint8 pc = GETBITS64(opcode, 36,2);
  t_uint8 sf = GETBITS64(opcode, 34,2);

  t_float tmp_default_result, tmp_res;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* extract values from procstate*/

  if (ProcStateGetReg(state,f2,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f2 = reg.f;

  if (ProcStateGetReg(state,f3,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f3 = reg.f;

  if (ProcStateGetReg(state,f4,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f4 = reg.f;

  /* perform operations */
  fp_check_target_register(f1);

  if ((tmp_isrcode = (fp_reg_disabled(f1, f2, f3, f4))))
    disabled_fp_register_fault(tmp_isrcode, 0);


  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3) || fp_is_natval(fr_f4)) 
  {
    fr_f1.flags |= 1;
    fp_update_psr(f1);
  } 
  else 
  {
    tmp_default_result = fma_exception_fault_check(f2, f3, f4, pc, sf, &tmp_fp_env);
    if (fp_raise_fault(tmp_fp_env))
      fp_exception_fault(fp_decode_fault(tmp_fp_env));


    if (fp_is_nan_or_inf(tmp_default_result)) 
    {
      fr_f1 = tmp_default_result;
    } 
    else 
    {
      tmp_res = fp_mul(fp_reg_read(fr_f3), fp_reg_read(fr_f4));
      if (f2 != 0)
	tmp_res = fp_add(tmp_res, fp_reg_read(fr_f2), tmp_fp_env);
      fr_f1 = fp_ieee_round(tmp_res, &tmp_fp_env);
    }


    fp_update_fpsr(sf, tmp_fp_env);
    fp_update_psr(f1);
    if (fp_raise_traps(tmp_fp_env))
      fp_exception_trap(fp_decode_trap(tmp_fp_env));
  }

  /* nu nog resultaten wegschrijven... */
  
  GenericSet_uint64(reg.f, fr_f1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  
#endif
}
#endif



/*
Floating-point Maximum

Format:         (qp) fmax.sf f1 = f2, f3                                         F8
*/

static void Ia64Emulate_fmax(t_ins * ins, t_procstate * state)
{
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      fminmax_exception_fault_check(f2, f3, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
          fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      tmp_bool_res = fp_less_than(fp_reg_read(fr_f3),
                                          fp_reg_read(fr_f2));
      fr_f1 = (tmp_bool_res ? fr_f2 : fr_f3);
  
      fp_update_fpsr(sf, tmp_fp_env);
  }
  fp_update_psr(f1);
}

#endif









/*
Floating-point Merge

Format:         (qp) fmerge.ns f1 = f2, f3                                 neg_sign_form          F9
                (qp) fmerge.s f1 = f2, f3                                      sign_form          F9
                (qp) fmerge.se f1 = f2, f3                                 sign_exp_form          F9
*/


static void Ia64Emulate_fmerge(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
       disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
       fr_f1 = NATVAL;
  } else {
       FP_SIGNIFICAND(fr_f1) = FP_SIGNIFICAND(fr_f3);
       if (neg_sign_form) {
            FP_EXPONENT(fr_f1) = FP_EXPONENT(fr_f3);
            FP_SIGN(fr_f1) = !FP_SIGN(fr_f2);
       } else if (sign_form) {
            FP_EXPONENT(fr_f1) = FP_EXPONENT(fr_f3);
            FP_SIGN(fr_f1) = FP_SIGN(fr_f2);
       } else {                                           // sign_exp_form
            FP_EXPONENT(fr_f1) = FP_EXPONENT(fr_f2);
            FP_SIGN(fr_f1) = FP_SIGN(fr_f2);
       }
  }
  
  
  fp_update_psr(f1);
}


#endif









/*
Floating-point Minimum

Format:         (qp) fmin.sf f1 = f2, f3                                          F8
*/

static void Ia64Emulate_fmin(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      fminmax_exception_fault_check(f2, f3, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
          fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      tmp_bool_res = fp_less_than(fp_reg_read(fr_f2),
                                          fp_reg_read(fr_f3));
      fr_f1 = tmp_bool_res ? fr_f2 : fr_f3;
  
      fp_update_fpsr(sf, tmp_fp_env);
  }
  fp_update_psr(f1);
}


#endif








/*
Floating-point Mix

Format:          (qp) fmix.l f1 = f2, f3                              mix_l_form         F9
                 (qp) fmix.r f1 = f2, f3                              mix_r_form         F9
                 (qp) fmix.lr f1 = f2, f3                            mix_lr_form         F9
*/


static void Ia64Emulate_fmix(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
       disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
       fr_f1 = NATVAL;
  } else {
       if (mix_l_form) {
            tmp_res_hi = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f2),63,32);
            tmp_res_lo = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),63,32);
       } else if (mix_r_form) {
            tmp_res_hi = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f2),31,0);
            tmp_res_lo = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),31,0);
       } else {                                            // mix_lr_form
            tmp_res_hi = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f2),63,32);
            tmp_res_lo = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),31,0);
       }
       FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
       FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
       FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  
  
  fp_update_psr(f1);
}

#endif





/*
Floating-point Multiply Subtract

Format:         (qp) fms.pc.sf f1 = f3, f4, f2                                     F1
*/

static void Ia64Emulate_fms(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
#ifndef EMULATEFLOAT
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
#else 
  t_ia64_reg f1 = IA64INSOUT1(ins), f2 = IA64INSIN1(ins), f3 = IA64INSIN2(ins), f4 = IA64INSIN3(ins);

  t_float fr_f1, fr_f2, fr_f3, fr_f4, tmp_fr2;

  register_content reg;

  t_bool tmp_isrcode;

  t_uint64 opcode = IA64OPCODEINFO(ins).opcode;
  t_uint8 pc = GETBITS64(opcode, 36,2);
  t_uint8 sf = GETBITS64(opcode, 34,2);

  t_float tmp_default_result, tmp_res;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* extract values from procstate */

  if (ProcStateGetReg(state,f2,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f2 = reg.f;

  if (ProcStateGetReg(state,f3,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f3 = reg.f;

  if (ProcStateGetReg(state,f4,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f4 = reg.f;

  
  /* perform operations */
  fp_check_target_register(f1);
  if ((tmp_isrcode = (fp_reg_disabled(f1, f2, f3, f4)))) 
       disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3) || 
       fp_is_natval(fr_f4)) {
       fr_f1 = NATVAL;
       fp_update_psr(f1);
  } else {
       tmp_default_result = fms_fnma_exception_fault_check(f2, f3, f4,
                                                                   pc, sf, &tmp_fp_env);
       if (fp_raise_fault(tmp_fp_env))
             fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
       if (fp_is_nan_or_inf(tmp_default_result)) {
             fr_f1 = tmp_default_result;
       } else {
             tmp_res = fp_mul(fp_reg_read(fr_f3), fp_reg_read(fr_f4));
             tmp_fr2 = fp_reg_read(fr_f2);
             SET_FP_SIGN(tmp_fr2,!GET_FP_SIGN(tmp_fr2));
             if (f2 != 0)
                tmp_res = fp_add(tmp_res, tmp_fr2, tmp_fp_env);
             fr_f1 = fp_ieee_round(tmp_res, &tmp_fp_env);
       }
  
  
       fp_update_fpsr(sf, tmp_fp_env);
       fp_update_psr(f1);
       if (fp_raise_traps(tmp_fp_env))
             fp_exception_trap(fp_decode_trap(tmp_fp_env));
  }

  /* nu nog resultaten wegschrijven... */
  
  GenericSet_uint64(reg.f, fr_f1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  
#endif
}
#endif






/*
Floating-point Negative Multiply Add

Format:         (qp) fnma.pc.sf f1 = f3, f4, f2                                     F1
*/

static void Ia64Emulate_fnma(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
#ifndef EMULATEFLOAT
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
#else 
  t_ia64_reg f1 = IA64INSOUT1(ins), f2 = IA64INSIN1(ins), f3 = IA64INSIN2(ins), f4 = IA64INSIN3(ins);

  t_float fr_f1, fr_f2, fr_f3, fr_f4;

  register_content reg;

  t_bool tmp_isrcode;

  t_uint64 opcode = IA64OPCODEINFO(ins).opcode;
  t_uint8 pc = GETBITS64(opcode, 36,2);
  t_uint8 sf = GETBITS64(opcode, 34,2);

  t_float tmp_default_result, tmp_res;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* extract values from procstate*/

  if (ProcStateGetReg(state,f2,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f2 = reg.f;

  if (ProcStateGetReg(state,f3,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f3 = reg.f;

  if (ProcStateGetReg(state,f4,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f4 = reg.f;

  /* perform operations */

  fp_check_target_register(f1);
  if ((tmp_isrcode = (fp_reg_disabled(f1, f2, f3, f4))))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3) ||
      fp_is_natval(fr_f4)) {
      fr_f1 = NATVAL;
      fp_update_psr(f1);
  } else {
      tmp_default_result = fms_fnma_exception_fault_check(f2, f3, f4,
                                                                   pc, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
             fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      if (fp_is_nan_or_inf(tmp_default_result)) {
             fr_f1 = tmp_default_result;
      } else {
             tmp_res = fp_mul(fp_reg_read(fr_f3), fp_reg_read(fr_f4));
             SET_FP_SIGN(tmp_res, !GET_FP_SIGN(tmp_res));
             if (f2 != 0)
                tmp_res = fp_add(tmp_res, fp_reg_read(fr_f2), tmp_fp_env);
             fr_f1 = fp_ieee_round(tmp_res, &tmp_fp_env);
      }
  
  
      fp_update_fpsr(sf, tmp_fp_env);
      fp_update_psr(f1);
      if (fp_raise_traps(tmp_fp_env))
             fp_exception_trap(fp_decode_trap(tmp_fp_env));
  }
#endif
}

#endif






/*
Floating-point Logical Or

Format:         (qp) for f1 = f2, f3                                                  F9
*/

static void Ia64Emulate_for(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      FP_SIGNIFICAND(fr_f1) = FP_SIGNIFICAND(fr_f2) | FP_SIGNIFICAND(fr_f3);
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  
  
  fp_update_psr(f1);
}


#endif









/*
Floating-point Pack

Format:          (qp) fpack f1 = f2, f3                                    pack_form         F9
*/

static void Ia64Emulate_fpack(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
         disabled_fp_register_fault(tmp_isrcode, 0);


  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
         fr_f1 = NATVAL;
  } else {
         tmp_res_hi = fp_single(fr_f2);
         tmp_res_lo = fp_single(fr_f3);

         FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
         FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
         FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  fp_update_psr(f1);
}

#endif









/*
Floating-point Parallel Absolute Maximum

Format:         (qp) fpamax.sf f1 = f2, f3                                       F8
*/

static void Ia64Emulate_fpamax(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      fpminmax_exception_fault_check(f2, f3, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
          fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      tmp_fr2 = tmp_right = fp_reg_read_hi(f2);
      tmp_fr3 = tmp_left = fp_reg_read_hi(f3);
      FP_SIGN(tmp_right) = FP_SIGN_POSITIVE;
      FP_SIGN(tmp_left) = FP_SIGN_POSITIVE;
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      tmp_res_hi = fp_single(tmp_bool_res ? tmp_fr2: tmp_fr3);
  
  
      tmp_fr2 = tmp_right = fp_reg_read_lo(f2);
      tmp_fr3 = tmp_left = fp_reg_read_lo(f3);
      FP_SIGN(tmp_right) = FP_SIGN_POSITIVE;
      FP_SIGN(tmp_left) = FP_SIGN_POSITIVE;
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      tmp_res_lo = fp_single(tmp_bool_res ? tmp_fr2: tmp_fr3);
  
  
      FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  
      fp_update_fpsr(sf, tmp_fp_env);
  }
  fp_update_psr(f1);
}

#endif








/*
Floating-point Parallel Absolute Minimum

Format:         (qp) fpamin.sf f1 = f2, f3                                       F8
*/

static void Ia64Emulate_fpamin(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      fpminmax_exception_fault_check(f2, f3, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
          fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      tmp_fr2 = tmp_left = fp_reg_read_hi(f2);
      tmp_fr3 = tmp_right = fp_reg_read_hi(f3);
      FP_SIGN(tmp_left) = FP_SIGN_POSITIVE;
      FP_SIGN(tmp_right) = FP_SIGN_POSITIVE;
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      tmp_res_hi = fp_single(tmp_bool_res ? tmp_fr2: tmp_fr3);
  
  
      tmp_fr2 = tmp_left = fp_reg_read_lo(f2);
      tmp_fr3 = tmp_right = fp_reg_read_lo(f3);
      FP_SIGN(tmp_left) = FP_SIGN_POSITIVE;
      FP_SIGN(tmp_right) = FP_SIGN_POSITIVE;
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      tmp_res_lo = fp_single(tmp_bool_res ? tmp_fr2: tmp_fr3);
  
  
      FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  
      fp_update_fpsr(sf, tmp_fp_env);
  }
  fp_update_psr(f1);
}


#endif








/*
Floating-point Parallel Compare

Format:         (qp) fpcmp.frel.sf f1= f2, f3                           F8
*/

static void Ia64Emulate_fpcmp(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
          disabled_fp_register_fault(tmp_isrcode, 0);

  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
          fr_f1 = NATVAL;
  } else {
          fpcmp_exception_fault_check(f2, f3, frel, sf, &tmp_fp_env);

          if (fp_raise_fault(tmp_fp_env))
              fp_exception_fault(fp_decode_fault(tmp_fp_env));

          tmp_fr2 = fp_reg_read_hi(f2);


   tmp_fr3 = fp_reg_read_hi(f3);

   if  
             (frel == 'eq')        tmp_rel = fp_equal(tmp_fr2, tmp_fr3);
   else if (frel == 'lt')          tmp_rel = fp_less_than(tmp_fr2, tmp_fr3);
   else if (frel == 'le')          tmp_rel = fp_lesser_or_equal(tmp_fr2,
                                                          tmp_fr3);
   else if (frel == 'gt')          tmp_rel = fp_less_than(tmp_fr3, tmp_fr2);
   else if (frel == 'ge')          tmp_rel = fp_lesser_or_equal(tmp_fr3,
                                                          tmp_fr2);
   else if (frel == 'unord')tmp_rel = fp_unordered(tmp_fr2, tmp_fr3);
   else if (frel == 'neq') tmp_rel = !fp_equal(tmp_fr2, tmp_fr3);
   else if (frel == 'nlt') tmp_rel = !fp_less_than(tmp_fr2, tmp_fr3);
   else if (frel == 'nle') tmp_rel = !fp_lesser_or_equal(tmp_fr2,
                                                          tmp_fr3);
   else if (frel == 'ngt') tmp_rel = !fp_less_than(tmp_fr3, tmp_fr2);
   else if (frel == 'nge') tmp_rel = !fp_lesser_or_equal(tmp_fr3,
                                                          tmp_fr2);
   else                            tmp_rel = !fp_unordered(tmp_fr2,
                                                          tmp_fr3); //'ord'

   tmp_res_hi = (tmp_rel ? 0xFFFFFFFF : 0x00000000);

   tmp_fr2 = fp_reg_read_lo(f2);
   tmp_fr3 = fp_reg_read_lo(f3);

   if  
             (frel == 'eq')        tmp_rel = fp_equal(tmp_fr2, tmp_fr3);
   else if (frel == 'lt')          tmp_rel = fp_less_than(tmp_fr2, tmp_fr3);
   else if (frel == 'le')          tmp_rel = fp_lesser_or_equal(tmp_fr2,
                                                          tmp_fr3);
   else if (frel == 'gt')          tmp_rel = fp_less_than(tmp_fr3, tmp_fr2);
   else if (frel == 'ge')          tmp_rel = fp_lesser_or_equal(tmp_fr3,
                                                          tmp_fr2);
   else if (frel == 'unord')tmp_rel = fp_unordered(tmp_fr2, tmp_fr3);
   else if (frel == 'neq') tmp_rel = !fp_equal(tmp_fr2, tmp_fr3);
   else if (frel == 'nlt') tmp_rel = !fp_less_than(tmp_fr2, tmp_fr3);
   else if (frel == 'nle') tmp_rel = !fp_lesser_or_equal(tmp_fr2,
                                                          tmp_fr3);
   else if (frel == 'ngt') tmp_rel = !fp_less_than(tmp_fr3, tmp_fr2);
   else if (frel == 'nge') tmp_rel = !fp_lesser_or_equal(tmp_fr3,
                                                          tmp_fr2);
   else                            tmp_rel = !fp_unordered(tmp_fr2,
                                                          tmp_fr3); //'ord'

   tmp_res_lo = (tmp_rel ? 0xFFFFFFFF : 0x00000000);

   FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
   FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
   FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;

   fp_update_fpsr(sf, tmp_fp_env);
}
fp_update_psr(f1);

}

#endif









/*
Convert Parallel Floating-point to Integer

Format:         (qp) fpcvt.fx.sf f1 = f2                        signed_form         F10
                (qp) fpcvt.fx.trunc.sf f1 = f2     signed_form, trunc_form          F10
                (qp) fpcvt.fxu.sf f1 = f2                     unsigned_form         F10
                (qp) fpcvt.fxu.trunc.sf f1 = f2 unsigned_form, trunc_form           F10
*/


static void Ia64Emulate_fpcvt_fx(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, 0, 0))
       disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2)) {
       fr_f1 = NATVAL;
       fp_update_psr(f1);
  } else {
       tmp_default_result_pair = fpcvt_exception_fault_check(f2,
                                       signed_form, trunc_form, sf, &tmp_fp_env);
       if (fp_raise_fault(tmp_fp_env))
            fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
       if (fp_is_nan(tmp_default_result_pair.hi)) {
            tmp_res_hi = INTEGER_INDEFINITE_32_BIT;
       } else {
            tmp_res = fp_ieee_rnd_to_int_sp(fp_reg_read_hi(f2), HIGH,
                                                 &tmp_fp_env);
            if (FP_EXPONENT(tmp_res))
               FP_SIGNIFICAND(tmp_res) = fp_U64_rsh(
                        FP_SIGNIFICAND(tmp_res), (FP_INTEGER_EXP - FP_EXPONENT(tmp_res)));
            if (signed_form && FP_SIGN(tmp_res))
               FP_SIGNIFICAND(tmp_res) = (~FP_SIGNIFICAND(tmp_res)) + 1;
  
  
            tmp_res_hi = EXTRACT_BITS64(FP_SIGNIFICAND(tmp_res),31,0);
       }
  
  
       if (fp_is_nan(tmp_default_result_pair.lo)) {
            tmp_res_lo = INTEGER_INDEFINITE_32_BIT;
       } else {
            tmp_res = fp_ieee_rnd_to_int_sp(fp_reg_read_lo(f2), LOW,
                                                 &tmp_fp_env);
            if (FP_EXPONENT(tmp_res))
               FP_SIGNIFICAND(tmp_res) = fp_U64_rsh(
                        FP_SIGNIFICAND(tmp_res), (FP_INTEGER_EXP - FP_EXPONENT(tmp_res)));
            if (signed_form && FP_SIGN(tmp_res)) 
               FP_SIGNIFICAND(tmp_res) = (~FP_SIGNIFICAND(tmp_res)) + 1;
  
  
            tmp_res_lo = EXTRACT_BITS64(FP_SIGNIFICAND(tmp_res),31,0);
       }
  
  
       FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
       FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
       FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  
       fp_update_fpsr(sf, tmp_fp_env);
       fp_update_psr(f1);
       if (fp_raise_traps(tmp_fp_env))
            fp_exception_trap(fp_decode_trap(tmp_fp_env));
  }
}


#endif








/*
Floating-point Parallel Multiply Add

Format:         (qp) fpma.sf f1 = f3, f4, f2                                        F1
*/

static void Ia64Emulate_fpma(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
#ifndef EMULATEFLOAT
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
#else 
  t_ia64_reg f1 = IA64INSOUT1(ins), f2 = IA64INSIN1(ins), f3 = IA64INSIN2(ins), f4 = IA64INSIN3(ins);

  t_float fr_f1, fr_f2, fr_f3, fr_f4;

  register_content reg;

  t_bool tmp_isrcode;

  t_float tmp_default_result, tmp_default_result_pair, tmp_res, tmp_res_hi;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* extract values from procstate*/

  if (ProcStateGetReg(state,f2,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f2 = reg.f;

  if (ProcStateGetReg(state,f3,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f3 = reg.f;

  if (ProcStateGetReg(state,f4,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f4 = reg.f;

  /* perform operations */
  
  fp_check_target_register(f1);
  if ((tmp_isrcode = (fp_reg_disabled(f1, f2, f3, f4))))
       disabled_fp_register_fault(tmp_isrcode, 0);


  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3) || 
       fp_is_natval(fr_f4)) {
       fr_f1 = NATVAL;
       fp_update_psr(f1);
  } else {
       tmp_default_result_pair = fpma_exception_fault_check(f2,
                                              f3, f4, sf, &tmp_fp_env);
       if (fp_raise_fault(tmp_fp_env))
            fp_exception_fault(fp_decode_fault(tmp_fp_env));


       if (fp_is_nan_or_inf(tmp_default_result_pair.hi)) {
            tmp_res_hi = fp_single(tmp_default_result_pair.hi);
       } else {
            tmp_res = fp_mul(fp_reg_read_hi(f3), fp_reg_read_hi(f4));
            if (f2 != 0)
               tmp_res = fp_add(tmp_res, fp_reg_read_hi(f2), tmp_fp_env);
            tmp_res_hi = fp_ieee_round_sp(tmp_res, HIGH, &tmp_fp_env);
       }


       if (fp_is_nan_or_inf(tmp_default_result_pair.lo)) {
            tmp_res_lo = fp_single(tmp_default_result_pair.lo);
       } else {
            tmp_res = fp_mul(fp_reg_read_lo(f3), fp_reg_read_lo(f4));
            if (f2 != 0)
               tmp_res = fp_add(tmp_res, fp_reg_read_lo(f2), tmp_fp_env);
            tmp_res_lo = fp_ieee_round_sp(tmp_res, LOW, &tmp_fp_env);
       }


       FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
       FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
       FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;

       fp_update_fpsr(sf, tmp_fp_env);
       fp_update_psr(f1);
       if (fp_raise_traps(tmp_fp_env))
            fp_exception_trap(fp_decode_trap(tmp_fp_env));
  }
#endif
}

#endif







/*
Floating-point Parallel Maximum

Format:         (qp) fpmax.sf f1 = f2, f3                                          F8
*/

static void Ia64Emulate_fpmax(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      fpminmax_exception_fault_check(f2, f3, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
          fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      tmp_fr2 = tmp_right = fp_reg_read_hi(f2);
      tmp_fr3 = tmp_left = fp_reg_read_hi(f3);
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      tmp_res_hi = fp_single(tmp_bool_res ? tmp_fr2 : tmp_fr3);
  
  
      tmp_fr2 = tmp_right = fp_reg_read_lo(f2);
      tmp_fr3 = tmp_left = fp_reg_read_lo(f3);
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      tmp_res_lo = fp_single(tmp_bool_res ? tmp_fr2 : tmp_fr3);
  
  
      FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  
      fp_update_fpsr(sf, tmp_fp_env);
  }
  fp_update_psr(f1);
}


#endif








/*
Floating-point Parallel Merge

Format:         (qp) fpmerge.ns f1 = f2, f3                                       neg_sign_form         F9
                (qp) fpmerge.s f1 = f2, f3                                            sign_form         F9
                (qp) fpmerge.se f1 = f2, f3                                       sign_exp_form         F9
*/

static void Ia64Emulate_fpmerge(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
       disabled_fp_register_fault(tmp_isrcode, 0);


  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
       fr_f1 = NATVAL;
  } else {
       if (neg_sign_form) {
            tmp_res_hi = (!EXTRACT_BIT64(FP_SIGNIFICAND(fr_f2),63) << 31)
                         | (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),62,32));
            tmp_res_lo = (!EXTRACT_BIT64(FP_SIGNIFICAND(fr_f2),31) << 31)
                         | (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),30,0));
       } else if (sign_form) {
            tmp_res_hi = (EXTRACT_BIT64(FP_SIGNIFICAND(fr_f2),63) << 31)
                         | (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),62,32));
            tmp_res_lo = (EXTRACT_BIT64(FP_SIGNIFICAND(fr_f2),31) << 31)
                         | (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),30,0));
       } else {                                            // sign_exp_form
            tmp_res_hi = (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f2),63,55) << 23)
                         | (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),54,32));
            tmp_res_lo = (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f2),31,23) << 23)
                         | (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),22,0));
       }


       FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
       FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
       FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }


  fp_update_psr(f1);
}


#endif









/*
Floating-point Parallel Minimum

Format:         (qp) fpmin.sf f1 = f2, f3                                          F8
*/

static void Ia64Emulate_fpmin(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      fpminmax_exception_fault_check(f2, f3, sf, &tmp_fp_env);
      if (fp_raise_fault(tmp_fp_env))
          fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
      tmp_fr2 = tmp_left = fp_reg_read_hi(f2);
      tmp_fr3 = tmp_right = fp_reg_read_hi(f3);
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      tmp_res_hi = fp_single(tmp_bool_res ? tmp_fr2: tmp_fr3);
  
  
      tmp_fr2 = tmp_left = fp_reg_read_lo(f2);
      tmp_fr3 = tmp_right = fp_reg_read_lo(f3);
      tmp_bool_res = fp_less_than(tmp_left, tmp_right);
      tmp_res_lo = fp_single(tmp_bool_res ? tmp_fr2: tmp_fr3);
  
  
      FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  
      fp_update_fpsr(sf, tmp_fp_env);
  }
  fp_update_psr(f1);
}

#endif









/*
Floating-point Parallel Multiply Subtract

Format:         (qp) fpms.sf f1 = f3, f4, f2                                        F1
*/


static void Ia64Emulate_fpms(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, f4))
       disabled_fp_register_fault(tmp_isrcode, 0);


  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3) || 
       fp_is_natval(fr_f4)) {
       fr_f1 = NATVAL;
       fp_update_psr(f1);
  } else {
       tmp_default_result_pair = fpms_fpnma_exception_fault_check(f2, f3,
                                                              f4, sf, &tmp_fp_env);
       if (fp_raise_fault(tmp_fp_env))
            fp_exception_fault(fp_decode_fault(tmp_fp_env));


       if (fp_is_nan_or_inf(tmp_default_result_pair.hi)) {
            tmp_res_hi = fp_single(tmp_default_result_pair.hi);
       } else {
            tmp_res = fp_mul(fp_reg_read_hi(f3), fp_reg_read_hi(f4));
            if (f2 != 0) {
                 tmp_sub = fp_reg_read_hi(f2);
                 FP_SIGN(tmp_sub) = !FP_SIGN(tmp_sub);
                 tmp_res = fp_add(tmp_res, tmp_sub, tmp_fp_env);
            }
            tmp_res_hi = fp_ieee_round_sp(tmp_res, HIGH, &tmp_fp_env);
       }


       if (fp_is_nan_or_inf(tmp_default_result_pair.lo)) {
            tmp_res_lo = fp_single(tmp_default_result_pair.lo);
       } else {
            tmp_res = fp_mul(fp_reg_read_lo(f3), fp_reg_read_lo(f4));
            if (f2 != 0) {
                 tmp_sub = fp_reg_read_lo(f2);
                 FP_SIGN(tmp_sub) = !FP_SIGN(tmp_sub);
                 tmp_res = fp_add(tmp_res, tmp_sub, tmp_fp_env);
            }
            tmp_res_lo = fp_ieee_round_sp(tmp_res, LOW, &tmp_fp_env);
       }


       FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
       FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
       FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;

       fp_update_fpsr(sf, tmp_fp_env);
       fp_update_psr(f1);
       if (fp_raise_traps(tmp_fp_env))
            fp_exception_trap(fp_decode_trap(tmp_fp_env));
  }
}

#endif









/*
Floating-point Parallel Negative Multiply Add

Format:         (qp) fpnma.sf f1 = f3, f4, f2                                         F1
*/

static void Ia64Emulate_fpnma(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, f4))
       disabled_fp_register_fault(tmp_isrcode, 0);


  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3) ||
       fp_is_natval(fr_f4)) {
       fr_f1 = NATVAL;
       fp_update_psr(f1);
  } else {
       tmp_default_result_pair = fpms_fpnma_exception_fault_check(f2, f3,
                                                              f4, sf, &tmp_fp_env);
       if (fp_raise_fault(tmp_fp_env))
            fp_exception_fault(fp_decode_fault(tmp_fp_env));


       if (fp_is_nan_or_inf(tmp_default_result_pair.hi)) {
            tmp_res_hi = fp_single(tmp_default_result_pair.hi);
       } else {
            tmp_res = fp_mul(fp_reg_read_hi(f3), fp_reg_read_hi(f4));
            FP_SIGN(tmp_res) = !FP_SIGN(tmp_res);
            if (f2 != 0)
                 tmp_res = fp_add(tmp_res, fp_reg_read_hi(f2), tmp_fp_env);
            tmp_res_hi = fp_ieee_round_sp(tmp_res, HIGH, &tmp_fp_env);
       }


       if (fp_is_nan_or_inf(tmp_default_result_pair.lo)) {
            tmp_res_lo = fp_single(tmp_default_result_pair.lo);
       } else {
            tmp_res = fp_mul(fp_reg_read_lo(f3), fp_reg_read_lo(f4));
            FP_SIGN(tmp_res) = !FP_SIGN(tmp_res);
            if (f2 != 0)
                 tmp_res = fp_add(tmp_res, fp_reg_read_lo(f2), tmp_fp_env);
            tmp_res_lo = fp_ieee_round_sp(tmp_res, LOW, &tmp_fp_env);
       }


       FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
       FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
       FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;

       fp_update_fpsr(sf, tmp_fp_env);
       fp_update_psr(f1);
       if (fp_raise_traps(tmp_fp_env))
            fp_exception_trap(fp_decode_trap(tmp_fp_env));
  }
}


#endif









#if 0

/* fp_ieee_recip()*/

/* floating point formaat moet dat van 82bit zijn...*/
fp_ieee_recip(den)
{
  RECIP_TABLE[256] = {
        0x3fc, 0x3f4, 0x3ec, 0x3e4, 0x3dd, 0x3d5, 0x3cd, 0x3c6,
        0x3be, 0x3b7, 0x3af, 0x3a8, 0x3a1, 0x399, 0x392, 0x38b,
        0x384, 0x37d, 0x376, 0x36f, 0x368, 0x361, 0x35b, 0x354,
        0x34d, 0x346, 0x340, 0x339, 0x333, 0x32c, 0x326, 0x320,
        0x319, 0x313, 0x30d, 0x307, 0x300, 0x2fa, 0x2f4, 0x2ee,
        0x2e8, 0x2e2, 0x2dc, 0x2d7, 0x2d1, 0x2cb, 0x2c5, 0x2bf,
        0x2ba, 0x2b4, 0x2af, 0x2a9, 0x2a3, 0x29e, 0x299, 0x293,
        0x28e, 0x288, 0x283, 0x27e, 0x279, 0x273, 0x26e, 0x269,
        0x264, 0x25f, 0x25a, 0x255, 0x250, 0x24b, 0x246, 0x241,
        0x23c, 0x237, 0x232, 0x22e, 0x229, 0x224, 0x21f, 0x21b,
        0x216, 0x211, 0x20d, 0x208, 0x204, 0x1ff, 0x1fb, 0x1f6,
        0x1f2, 0x1ed, 0x1e9, 0x1e5, 0x1e0, 0x1dc, 0x1d8, 0x1d4,
        0x1cf, 0x1cb, 0x1c7, 0x1c3, 0x1bf, 0x1bb, 0x1b6, 0x1b2,
        0x1ae, 0x1aa, 0x1a6, 0x1a2, 0x19e, 0x19a, 0x197, 0x193,
        0x18f, 0x18b, 0x187, 0x183, 0x17f, 0x17c, 0x178, 0x174,
        0x171, 0x16d, 0x169, 0x166, 0x162, 0x15e, 0x15b, 0x157,
        0x154, 0x150, 0x14d, 0x149, 0x146, 0x142, 0x13f, 0x13b,
        0x138, 0x134, 0x131, 0x12e, 0x12a, 0x127, 0x124, 0x120,
        0x11d, 0x11a, 0x117, 0x113, 0x110, 0x10d, 0x10a, 0x107,
        0x103, 0x100, 0x0fd, 0x0fa, 0x0f7, 0x0f4, 0x0f1, 0x0ee,
        0x0eb, 0x0e8, 0x0e5, 0x0e2, 0x0df, 0x0dc, 0x0d9, 0x0d6,
        0x0d3, 0x0d0, 0x0cd, 0x0ca, 0x0c8, 0x0c5, 0x0c2, 0x0bf,
        0x0bc, 0x0b9, 0x0b7, 0x0b4, 0x0b1, 0x0ae, 0x0ac, 0x0a9,
        0x0a6, 0x0a4, 0x0a1, 0x09e, 0x09c, 0x099, 0x096, 0x094,
        0x091, 0x08e, 0x08c, 0x089, 0x087, 0x084, 0x082, 0x07f,
        0x07c, 0x07a, 0x077, 0x075, 0x073, 0x070, 0x06e, 0x06b,
        0x069, 0x066, 0x064, 0x061, 0x05f, 0x05d, 0x05a, 0x058,
        0x056, 0x053, 0x051, 0x04f, 0x04c, 0x04a, 0x048, 0x045,
        0x043, 0x041, 0x03f, 0x03c, 0x03a, 0x038, 0x036, 0x033,
        0x031, 0x02f, 0x02d, 0x02b, 0x029, 0x026, 0x024, 0x022,
        0x020, 0x01e, 0x01c, 0x01a, 0x018, 0x015, 0x013, 0x011,
        0x00f, 0x00d, 0x00b, 0x009, 0x007, 0x005, 0x003, 0x001,
  };


  tmp_index = EXTRACT_BITS64(FP_SIGNIFICAND(den),62,55);
  FP_SIGNIFICAND(tmp_res) = (1 << 63) | (RECIP_TABLE[tmp_index] << 53);
  FP_EXPONENT(tmp_res) = FP_REG_EXP_ONES - 2 - FP_EXPONENT(den);
  FP_SIGN(tmp_res) = FP_SIGN(den);
  return (tmp_res);
}


#endif







#if 0

/* fp_ieee_recip_sqrt()*/


/* floating point formaat moet dat van 82bit zijn...*/
fp_ieee_recip_sqrt(root)
{
  RECIP_SQRT_TABLE[256] = {
        0x1a5, 0x1a0, 0x19a, 0x195, 0x18f, 0x18a, 0x185, 0x180,
        0x17a, 0x175, 0x170, 0x16b, 0x166, 0x161, 0x15d, 0x158,
        0x153, 0x14e, 0x14a, 0x145, 0x140, 0x13c, 0x138, 0x133,
        0x12f, 0x12a, 0x126, 0x122, 0x11e, 0x11a, 0x115, 0x111,
        0x10d, 0x109, 0x105, 0x101, 0x0fd, 0x0fa, 0x0f6, 0x0f2,
        0x0ee, 0x0ea, 0x0e7, 0x0e3, 0x0df, 0x0dc, 0x0d8, 0x0d5,
        0x0d1, 0x0ce, 0x0ca, 0x0c7, 0x0c3, 0x0c0, 0x0bd, 0x0b9,
        0x0b6, 0x0b3, 0x0b0, 0x0ad, 0x0a9, 0x0a6, 0x0a3, 0x0a0,
        0x09d, 0x09a, 0x097, 0x094, 0x091, 0x08e, 0x08b, 0x088,
        0x085, 0x082, 0x07f, 0x07d, 0x07a, 0x077, 0x074, 0x071,
        0x06f, 0x06c, 0x069, 0x067, 0x064, 0x061, 0x05f, 0x05c,
        0x05a, 0x057, 0x054, 0x052, 0x04f, 0x04d, 0x04a, 0x048,
        0x045, 0x043, 0x041, 0x03e, 0x03c, 0x03a, 0x037, 0x035,
        0x033, 0x030, 0x02e, 0x02c, 0x029, 0x027, 0x025, 0x023,
        0x020, 0x01e, 0x01c, 0x01a, 0x018, 0x016, 0x014, 0x011,
        0x00f, 0x00d, 0x00b, 0x009, 0x007, 0x005, 0x003, 0x001,
        0x3fc, 0x3f4, 0x3ec, 0x3e5, 0x3dd, 0x3d5, 0x3ce, 0x3c7,
        0x3bf, 0x3b8, 0x3b1, 0x3aa, 0x3a3, 0x39c, 0x395, 0x38e,
        0x388, 0x381, 0x37a, 0x374, 0x36d, 0x367, 0x361, 0x35a,
        0x354, 0x34e, 0x348, 0x342, 0x33c, 0x336, 0x330, 0x32b,
        0x325, 0x31f, 0x31a, 0x314, 0x30f, 0x309, 0x304, 0x2fe,
        0x2f9, 0x2f4, 0x2ee, 0x2e9, 0x2e4, 0x2df, 0x2da, 0x2d5,
        0x2d0, 0x2cb, 0x2c6, 0x2c1, 0x2bd, 0x2b8, 0x2b3, 0x2ae,
        0x2aa, 0x2a5, 0x2a1, 0x29c, 0x298, 0x293, 0x28f, 0x28a,
        0x286, 0x282, 0x27d, 0x279, 0x275, 0x271, 0x26d, 0x268,
        0x264, 0x260, 0x25c, 0x258, 0x254, 0x250, 0x24c, 0x249,
        0x245, 0x241, 0x23d, 0x239, 0x235, 0x232, 0x22e, 0x22a,
        0x227, 0x223, 0x220, 0x21c, 0x218, 0x215, 0x211, 0x20e,
        0x20a, 0x207, 0x204, 0x200, 0x1fd, 0x1f9, 0x1f6, 0x1f3,
        0x1f0, 0x1ec, 0x1e9, 0x1e6, 0x1e3, 0x1df, 0x1dc, 0x1d9,
        0x1d6, 0x1d3, 0x1d0, 0x1cd, 0x1ca, 0x1c7, 0x1c4, 0x1c1,
        0x1be, 0x1bb, 0x1b8, 0x1b5, 0x1b2, 0x1af, 0x1ac, 0x1aa,
  };


  tmp_index = (EXTRACT_BIT64(FP_EXPONENT(root),0) << 7) | EXTRACT_BITS64(FP_SIGNIFICAND(root),62,56);
  FP_SIGNIFICAND(tmp_res) = (1 << 63) | (RECIP_SQRT_TABLE[tmp_index] << 53);
  FP_EXPONENT(tmp_res) = FP_REG_EXP_HALF -
                         ((FP_EXPONENT(root) - FP_REG_BIAS) >> 1);
  FP_SIGN(tmp_res) = FP_SIGN_POSITIVE;
  return (tmp_res);
}


#endif











/*
Floating-point Parallel Reciprocal Approximation

Format:         (qp) fprcpa.sf f1, p2 = f2, f3                                    F6
*/

static void Ia64Emulate_fprcpa(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
   * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
   * functie worden gemodelleerd !!!
   */
  if (PR[qp]) {
      fp_check_target_register(f1);
      if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
           disabled_fp_register_fault(tmp_isrcode, 0);


      if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
           fr_f1 = NATVAL;
           pr_p2 = 0;
      } else {
           tmp_default_result_pair = fprcpa_exception_fault_check(f2, f3, sf,
                                                               &tmp_fp_env, &limits_check);
           if (fp_raise_fault(tmp_fp_env))
                fp_exception_fault(fp_decode_fault(tmp_fp_env));


           if (fp_is_nan_or_inf(tmp_default_result_pair.hi) ||
                limits_check.hi_fr3) {
                tmp_res_hi = fp_single(tmp_default_result_pair.hi);
                tmp_pred_hi = 0;
           } else {
                num = fp_normalize(fp_reg_read_hi(f2));
                den = fp_normalize(fp_reg_read_hi(f3));
                if (fp_is_inf(num) && fp_is_finite(den)) {
                     tmp_res = FP_INFINITY;
                     FP_SIGN(tmp_res) = FP_SIGN(num) ^ FP_SIGN(den);
                     tmp_pred_hi = 0;
                } else if (fp_is_finite(num) && fp_is_inf(den)) {
                     tmp_res = FP_ZERO;
                     FP_SIGN(tmp_res) = FP_SIGN(num) ^ FP_SIGN(den);
                     tmp_pred_hi = 0;
                } else if (fp_is_zero(num) && fp_is_finite(den)) {
                     tmp_res = FP_ZERO;
                     FP_SIGN(tmp_res) = FP_SIGN(num) ^ FP_SIGN(den);
                     tmp_pred_hi = 0;

                } else {
                     tmp_res = fp_ieee_recip(den);
                     if (limits_check.hi_fr2_or_quot)
                            tmp_pred_hi = 0;
                     else
                            tmp_pred_hi = 1;
                }
                tmp_res_hi = fp_single(tmp_res);
           }
           if (fp_is_nan_or_inf(tmp_default_result_pair.lo) ||
                limits_check.lo_fr3) {
                tmp_res_lo = fp_single(tmp_default_result_pair.lo);
                tmp_pred_lo = 0;
           } else {
                num = fp_normalize(fp_reg_read_lo(f2));
                den = fp_normalize(fp_reg_read_lo(f3));
                if (fp_is_inf(num) && fp_is_finite(den)) {
                     tmp_res = FP_INFINITY;
                     FP_SIGN(tmp_res) = FP_SIGN(num) ^ FP_SIGN(den);
                     tmp_pred_lo = 0;
                } else if (fp_is_finite(num) && fp_is_inf(den)) {
                     tmp_res = FP_ZERO;
                     FP_SIGN(tmp_res) = FP_SIGN(num) ^ FP_SIGN(den);
                     tmp_pred_lo = 0;
                } else if (fp_is_zero(num) && fp_is_finite(den)) {
                     tmp_res = FP_ZERO;
                     FP_SIGN(tmp_res) = FP_SIGN(num) ^ FP_SIGN(den);
                     tmp_pred_lo = 0;
                } else {
                     tmp_res = fp_ieee_recip(den);
                     if (limits_check.lo_fr2_or_quot)
                            tmp_pred_lo = 0;
                     else
                            tmp_pred_lo = 1;
                }
                tmp_res_lo = fp_single(tmp_res);
           }


           FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
           FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
           FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
           pr_p2 = tmp_pred_hi && tmp_pred_lo;

           fp_update_fpsr(sf, tmp_fp_env);
      }
      fp_update_psr(f1);
 } else {
      pr_p2 = 0;
 }
}

#endif








/*
Floating-point Parallel Reciprocal Square Root Approximation

Format:         (qp) fprsqrta.sf f1, p2 = f3                                          F7
*/

static void Ia64Emulate_fprsqrta(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
   * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
   * functie worden gemodelleerd !!!
   */
  if (PR[qp]) {
      fp_check_target_register(f1);
      if (tmp_isrcode = fp_reg_disabled(f1, f3, 0, 0))
           disabled_fp_register_fault(tmp_isrcode, 0);
  
  
      if (fp_is_natval(fr_f3)) {
           fr_f1 = NATVAL;
           pr_p2 = 0;
      } else {
           tmp_default_result_pair = fprsqrta_exception_fault_check(f3, sf,
                                                                &tmp_fp_env, &limits_check);
           if (fp_raise_fault(tmp_fp_env))
                   fp_exception_fault(fp_decode_fault(tmp_fp_env));
  
  
           if (fp_is_nan(tmp_default_result_pair.hi)) {
                   tmp_res_hi = fp_single(tmp_default_result_pair.hi);
                   tmp_pred_hi = 0;
           } else {
                   tmp_fr3 = fp_normalize(fp_reg_read_hi(f3));
                   if (fp_is_zero(tmp_fr3)) {
                      tmp_res = FP_INFINITY;
                      FP_SIGN(tmp_res) = FP_SIGN(tmp_fr3);
                      tmp_pred_hi = 0;
                   } else if (fp_is_pos_inf(tmp_fr3)) {
                      tmp_res = FP_ZERO;
                      tmp_pred_hi = 0;
                   } else {
                      tmp_res = fp_ieee_recip_sqrt(tmp_fr3);
                      if (limits_check.hi)
                         tmp_pred_hi = 0;
                      else
                         tmp_pred_hi = 1;
  
                }
                tmp_res_hi = fp_single(tmp_res);
           }
  
  
           if (fp_is_nan(tmp_default_result_pair.lo)) {
                tmp_res_lo = fp_single(tmp_default_result_pair.lo);
                tmp_pred_lo = 0;
           } else {
                tmp_fr3 = fp_normalize(fp_reg_read_lo(f3));
                if (fp_is_zero(tmp_fr3)) {
                     tmp_res = FP_INFINITY;
                     FP_SIGN(tmp_res) = FP_SIGN(tmp_fr3);
                     tmp_pred_lo = 0;
                } else if (fp_is_pos_inf(tmp_fr3)) {
                     tmp_res = FP_ZERO;
                     tmp_pred_lo = 0;
                } else {
                     tmp_res = fp_ieee_recip_sqrt(tmp_fr3);
                     if (limits_check.lo)
                            tmp_pred_lo = 0;
                     else
                            tmp_pred_lo = 1;
                }
                tmp_res_lo = fp_single(tmp_res);
           }
  
  
           FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
           FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
           FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
           pr_p2 = tmp_pred_hi && tmp_pred_lo;
  
           fp_update_fpsr(sf, tmp_fp_env);
      }
      fp_update_psr(f1);
  } else {
      pr_p2 = 0;
  }
}


#endif









/*
Floating-point Reciprocal Approximation

Format:         (qp) frcpa.sf f1, p2 = f2, f3                                        F6
*/

static void Ia64Emulate_frcpa(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
   * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
   * functie worden gemodelleerd !!!
   */
  if (PR[qp]) {
      fp_check_target_register(f1);
      if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
           disabled_fp_register_fault(tmp_isrcode, 0);


      if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
           fr_f1 = NATVAL;
           pr_p2 = 0;
      } else {
           tmp_default_result = frcpa_exception_fault_check(f2, f3, sf,
                                                                          &tmp_fp_env);
           if (fp_raise_fault(tmp_fp_env))
                 fp_exception_fault(fp_decode_fault(tmp_fp_env));


           if (fp_is_nan_or_inf(tmp_default_result)) {
                 fr_f1 = tmp_default_result;
                 pr_p2 = 0;
           } else {
                 num = fp_normalize(fp_reg_read(fr_f2));
                 den = fp_normalize(fp_reg_read(fr_f3));
                 if (fp_is_inf(num) && fp_is_finite(den)) {
                         fr_f1 = FP_INFINITY;
                         FP_SIGN(fr_f1) = FP_SIGN(num) ^ FP_SIGN(den);
                         pr_p2 = 0;
                 } else if (fp_is_finite(num) && fp_is_inf(den)) {
                         fr_f1 = FP_ZERO;
                         FP_SIGN(fr_f1) = FP_SIGN(num) ^ FP_SIGN(den);
                         pr_p2 = 0;
                 } else if (fp_is_zero(num) && fp_is_finite(den)) {
                         fr_f1 = FP_ZERO;
                         FP_SIGN(fr_f1) = FP_SIGN(num) ^ FP_SIGN(den);
                         pr_p2 = 0;
                 } else {
                         fr_f1 = fp_ieee_recip(den);
                         pr_p2 = 1;

                 }
            }
            fp_update_fpsr(sf, tmp_fp_env);
      }
      fp_update_psr(f1);
 } else {
      pr_p2 = 0;
 }


}

#endif








/*
Floating-point Reciprocal Square Root Approximation

Format:         (qp) frsqrta.sf f1, p2 = f3                                        F7
*/

static void Ia64Emulate_frsqrta(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
   * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
   * functie worden gemodelleerd !!!
   */
  if (PR[qp]) {
      fp_check_target_register(f1);
      if (tmp_isrcode = fp_reg_disabled(f1, f3, 0, 0))
           disabled_fp_register_fault(tmp_isrcode, 0);


      if (fp_is_natval(fr_f3)) {
           fr_f1 = NATVAL;
           pr_p2 = 0;
      } else {
           tmp_default_result = frsqrta_exception_fault_check(f3, sf,
                                                                        &tmp_fp_env);
           if (fp_raise_fault(tmp_fp_env))
                   fp_exception_fault(fp_decode_fault(tmp_fp_env));


           if (fp_is_nan(tmp_default_result)) {
                   fr_f1 = tmp_default_result;
                   pr_p2 = 0;
           } else {
                   tmp_fr3 = fp_normalize(fp_reg_read(fr_f3));
                   if (fp_is_zero(tmp_fr3)) {
                        fr_f1 = tmp_fr3;
                        pr_p2 = 0;
                   } else if (fp_is_pos_inf(tmp_fr3)) {
                        fr_f1 = tmp_fr3;
                        pr_p2 = 0;
                   } else {
                        fr_f1 = fp_ieee_recip_sqrt(tmp_fr3);
                        pr_p2 = 1;
                   }
           }
           fp_update_fpsr(sf, tmp_fp_env);
      }
      fp_update_psr(f1);
  } else {
      pr_p2 = 0;
 }


}


#endif








/*
Floating-point Select

Format:         (qp) fselect f1 = f3, f4, f2                                      F3
*/

static void Ia64Emulate_fselect(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
#ifndef EMULATEFLOAT
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
#else 
  t_ia64_reg f1 = IA64INSOUT1(ins), f2 = IA64INSIN1(ins), f3 = IA64INSIN2(ins), f4 = IA64INSIN3(ins);

  t_float fr_f1, fr_f2, fr_f3, fr_f4;

  register_content reg;

  t_bool tmp_isrcode;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* extract values from procstate */

  if (ProcStateGetReg(state,f2,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f3 = reg.f;

  if (ProcStateGetReg(state,f3,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f4 = reg.f;

  if (ProcStateGetReg(state,f4,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f2 = reg.f;

  /* perform operations */

  fp_check_target_register(f1);
  if ((tmp_isrcode = (fp_reg_disabled(f1, f2, f3, f4))))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3) || 
      fp_is_natval(fr_f4)) {
      fr_f1 = NATVAL;
  } else {
      FP_SIGNIFICAND(fr_f1)         = (FP_SIGNIFICAND(fr_f3) & FP_SIGNIFICAND(fr_f2))
	| (FP_SIGNIFICAND(fr_f4) & ~FP_SIGNIFICAND(fr_f2));
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  
  
  fp_update_psr(f1);
#endif
}

#endif






/*
Floating-point Set Controls

Format:         (qp) fsetc.sf amask7, omask7                                      F12
*/

static void Ia64Emulate_fsetc(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  tmp_controls = (AR[FPSR].sf0.controls & amask7) | omask7;
  if (is_reserved_field(FSETC, sf, tmp_controls))
     reserved_register_field_fault(); 
  fp_set_sf_controls(sf, tmp_controls);
}


#endif








/*
Floating-point Swap

Format:         (qp) fswap f1 = f2, f3                                        swap_form            F9
                (qp) fswap.nl f1 = f2, f3                                   swap_nl_form           F9
                (qp) fswap.nr f1 = f2, f3                                   swap_nr_form           F9
*/

static void Ia64Emulate_fswap(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
       disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
       fr_f1 = NATVAL;
  } else {
       if (swap_form) {
            tmp_res_hi = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),31,0);
            tmp_res_lo = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f2),63,32);
       } else if (swap_nl_form) {
            tmp_res_hi = (!EXTRACT_BIT64(FP_SIGNIFICAND(fr_f3),31) << 31)
                         | (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),30,0));
            tmp_res_lo = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f2),63,32);
       } else { // swap_nr_form
            tmp_res_hi = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),31,0);
            tmp_res_lo = (!EXTRACT_BIT64(FP_SIGNIFICAND(fr_f2),63) << 31)
                         | (EXTRACT_BITS64(FP_SIGNIFICAND(fr_f2),62,32));
       }
  
  
       FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
       FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
       FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  
  
  fp_update_psr(f1);
}

#endif








/*
Floating-point Sign Extend

Format:         (qp) fsxt.l f1 = f2, f3                               sxt_l_form            F9
                (qp) fsxt.r f1 = f2, f3                               sxt_r_form            F9
*/

static void Ia64Emulate_fsxt(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
        disabled_fp_register_fault(tmp_isrcode, 0);
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
        fr_f1 = NATVAL;
  } else {
        if (sxt_l_form) {
               tmp_res_hi = (EXTRACT_BIT64(FP_SIGNIFICAND(fr_f2),63) ? 0xFFFFFFFF : 0x00000000);
               tmp_res_lo = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),63,32);
        } else {                                                        // sxt_r_form
               tmp_res_hi = (EXTRACT_BIT64(FP_SIGNIFICAND(fr_f2),31) ? 0xFFFFFFFF : 0x00000000);
               tmp_res_lo = EXTRACT_BITS64(FP_SIGNIFICAND(fr_f3),31,0);
        }
  
        FP_SIGNIFICAND(fr_f1) = fp_concatenate(tmp_res_hi, tmp_res_lo);
        FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
        FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  
  fp_update_psr(f1);
}

#endif







/*
Flush Write Buffers

Format:         (qp) fwb                                                        M24
*/

static void Ia64Emulate_fwb(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  mem_flush_pending_stores();
}

#endif








/*
Floating-point Exclusive Or

Format:         (qp) fxor f1 = f2, f3                                               F9
*/

static void Ia64Emulate_fxor(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, f3, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3)) {
      fr_f1 = NATVAL;
  } else {
      FP_SIGNIFICAND(fr_f1) = FP_SIGNIFICAND(fr_f2) ^ FP_SIGNIFICAND(fr_f3);
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  
  
  fp_update_psr(f1);
}

#endif








/*
Get Floating-point Value or Exponent or Significand

Format:          (qp) getf.s r1 = f2                                          single_form    M19
                 (qp) getf.d r1 = f2                                         double_form     M19
                 (qp) getf.exp r1 = f2                                      exponent_form    M19
                 (qp) getf.sig r1 = f2                                 significand_form      M19
*/
/*status: TODO */
static void Ia64Emulate_getf(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  check_target_register(r1);
  if (tmp_isrcode = fp_reg_disabled(f2, 0, 0, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (single_form) {
      SET_BITS64(gr_r1,31,0,fp_fr_to_mem_format(fr_f2, 4, 0));
      SET_BITS64(gr_r1,63,32,0);
  } else if (double_form) {
      gr_r1 = fp_fr_to_mem_format(fr_f2, 8, 0);
  } else if (exponent_form) {
      SET_BITS64(gr_r1,63,18,0);
      SET_BITS64(gr_r1,16,0,FP_EXPONENT(fr_f2));
      SET_BIT64(gr_r1,17,FP_SIGN(fr_f2));
  } else // significand_form
      gr_r1 = FP_SIGNIFICAND(fr_f2);
  
  if (fp_is_natval(fr_f2))
    gr_r1_nat = 1;
  else
    gr_r1_nat = 0;
}

#endif








/*
Invalidate ALAT

Format:         (qp) invala                                   complete_form        M24
                (qp) invala.e r1                        gr_form, entry_form        M26
                (qp) invala.e f1                        fr_form, entry_form        M27
*/

static void Ia64Emulate_invala(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (complete_form)
      alat_inval();
  else { // entry_form
      if (gr_form)
           alat_inval_single_entry(GENERAL, r1);
      else // fr_form
           alat_inval_single_entry(FLOAT, f1);
  }
}

#endif








/*
Insert Translation Cache

Format:         (qp) itc.i r2                              instruction_form         M41
                (qp) itc.d r2                                     data_form         M41
*/

static void Ia64Emulate_itc(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (!followed_by_stop())
      undefined_behavior();
  if (PSR.ic)
      illegal_operation_fault();
  if (PSR.cpl != 0)
      privileged_operation_fault(0);
  if (gr_r2_nat)
      register_nat_consumption_fault(0);
  
  
  tmp_size = CR[ITIR].ps;
  tmp_va = EXTRACT_BITS64(CR[IFA],60,0);
  tmp_rid = RR[EXTRACT_BITS64(CR[IFA],63,61)].rid;
  tmp_va = align_to_size_boundary(tmp_va, tmp_size);
  
  
  if (is_reserved_field(TLB_TYPE, gr_r2, CR[ITIR]))
      reserved_register_field_fault();
  if (unimplemented_virtual_address(CR[IFA]))
      unimplemented_data_address_fault(0);
  
  
  if (instruction_form) {
      tlb_must_purge_itc_entries(tmp_rid, tmp_va, tmp_size);
      tlb_may_purge_dtc_entries(tmp_rid, tmp_va, tmp_size);
      slot = tlb_replacement_algorithm(ITC_TYPE);
      tlb_insert_inst(slot, gr_r2, CR[ITIR], CR[IFA], tmp_rid, TC);
  } else {                                                          // data_form
      tlb_must_purge_dtc_entries(tmp_rid, tmp_va, tmp_size);
      tlb_may_purge_itc_entries(tmp_rid, tmp_va, tmp_size);
      slot = tlb_replacement_algorithm(DTC_TYPE);
      tlb_insert_data(slot, gr_r2, CR[ITIR], CR[IFA], tmp_rid, TC);
  }
}

#endif







/*
Insert Translation Register

Format:           (qp) itr.i itr[r3] = r2                    instruction_form       M42
                  (qp) itr.d dtr[r3] = r2                           data_form       M42
*/

static void Ia64Emulate_itr(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (PSR.ic)
      illegal_operation_fault();
  if (PSR.cpl != 0)
      privileged_operation_fault(0);
  if (gr_r3_nat || gr_r2_nat)
      register_nat_consumption_fault(0);
  
  
  slot = EXTRACT_BITS64(gr_r3,7,0);
  tmp_size = CR[ITIR].ps;
  tmp_va = EXTRACT_BITS64(CR[IFA],60,0);
  tmp_rid = RR[EXTRACT_BITS64(CR[IFA],63,61)].rid;
  tmp_va = align_to_size_boundary(tmp_va, tmp_size);
  
  
  tmp_tr_type = instruction_form ? ITR_TYPE : DTR_TYPE;
  
  
  if (is_reserved_reg(tmp_tr_type, slot))
      reserved_register_field_fault();
  if (is_reserved_field(TLB_TYPE, gr_r2, CR[ITIR]))
      reserved_register_field_fault();
  if (unimplemented_virtual_address(CR[IFA]))
      unimplemented_data_address_fault(0);
  
  
  if (instruction_form) {
      tlb_must_purge_itc_entries(tmp_rid, tmp_va, tmp_size);
      tlb_may_purge_dtc_entries(tmp_rid, tmp_va, tmp_size);
      tlb_insert_inst(slot, gr_r2, CR[ITIR], CR[IFA], tmp_rid, TR);
  } else {                                                          // data_form
      tlb_must_purge_dtc_entries(tmp_rid, tmp_va, tmp_size);
      tlb_may_purge_itc_entries(tmp_rid, tmp_va, tmp_size);
      tlb_insert_data(slot, gr_r2, CR[ITIR], CR[IFA], tmp_rid, TR);
  }
}

#endif






/*
Load

Format:         (qp) ldsz.ldtype.ldhint r1 = [r3]                         no_base_update_form               M1
                (qp) ldsz.ldtype.ldhint r1 = [r3], r2                    reg_base_update_form               M2
                (qp) ldsz.ldtype.ldhint r1 = [r3], imm9                 imm_base_update_form                M3
                (qp) ld8.fill.ldhint r1 = [r3]                fill_form, no_base_update_form                M1
                (qp) ld8.fill.ldhint r1 = [r3], r2           fill_form, reg_base_update_form                M2
                (qp) ld8.fill.ldhint r1 = [r3], imm9       fill_form, imm_base_update_form                  M3
*/
/*status: NOT CHECKED */
static void Ia64Emulate_ld(t_ins * ins, t_procstate * state)
{
/*  
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

 #if 0
*/
  t_reloc * base_reloc;
  register_content reg, tmp_address;
  t_uint64 offset;
  t_lattice_level address_unknown;
  t_lattice_level base_reloc_unknown;
  t_lattice_level address_nat_unknown;

  t_bool reg_base_update_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_M2;
  t_bool imm_base_update_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_M3;

  t_uint64 opcode = IA64OPCODEINFO(ins).opcode;

  t_bool fill_form = GETBITS64(opcode,30,6)==0x1B;

  t_uint8 size;

  t_uint64 val;

  t_ia64_reg r1 = IA64INSOUT1(ins), r3=IA64INSIN1(ins);

  t_uint64 new_gr_r3, new_gr_r3_nat, gr_r3, gr_r3_nat, gr_r1, gr_r1_nat;
  t_bool speculative, advanced, check_clear, check_no_clear, check;

  t_section * sec;

  t_bool updateIsKnown = TRUE;

  /*pas nog aan*/
  t_uint64 bit_pos;

  __stats.ld_total++;


  /* wat weten we over locatie? */
  base_reloc_unknown = ProcStateGetTag(state, IA64INSIN1(ins),&base_reloc);
  address_unknown = ProcStateGetReg(state, IA64INSIN1(ins),&tmp_address);
  gr_r3 = G_T_UINT64(tmp_address.i);
  address_nat_unknown = ProcStateGetReg(state, IA64GETNAT(IA64INSIN1(ins)),&reg);
  gr_r3_nat = G_T_UINT64(reg.i);
  
  /* nieuwe waarde van gr_r3 */
  new_gr_r3 = gr_r3;
  new_gr_r3_nat = gr_r3_nat;
  /* als r2 niet gekend is, kunnen we de nieuwe waarde van r3 niet kennen 
  *  We kunnen wel echter nog eventueel de load achterhalen.
  * */
  if (reg_base_update_form) 
  {
    /*ikke*/
    if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
    {
      ProcStateSetRegBot(state,IA64INSIN1(ins));
      updateIsKnown = FALSE;
    }
    
    /*intel*/
    new_gr_r3 += G_T_UINT64(reg.i);
    /*ikke*/
    if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
    {
      ProcStateSetRegBot(state,IA64GETNAT(IA64INSIN1(ins)));
      updateIsKnown = FALSE;
    }
    
    /*intel*/
    new_gr_r3_nat |= G_T_UINT64(reg.i);
  }
  else if (imm_base_update_form)
    new_gr_r3 += IA64INSIMM1(ins);

    
  /* wegschrijven nieuwe waarde van R3 - enkel als update gekend is */
    
  if(updateIsKnown && (reg_base_update_form || imm_base_update_form))
  {
    GenericSet_uint64(reg.i, new_gr_r3);
    ProcStateSetReg(state,IA64INSIN1(ins),reg);  

    GenericSet_uint64(reg.i, new_gr_r3_nat);
    ProcStateSetReg(state,IA64GETNAT(IA64INSIN1(ins)),reg); 

    __stats.ld_partial++;

  }


  if(base_reloc_unknown==TOP)
  {
    FATAL(("problem: no reloc in base register\n"));
  }

  if (address_unknown||base_reloc_unknown||address_nat_unknown||gr_r3_nat)
  {
    ProcStateSetAllBot(state, INS_REGS_DEF(ins));
    ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));
#if 0    
    /* wegschrijven nieuwe waarde van R3 */
    GenericSet_uint64(reg.i, new_gr_r3);
    ProcStateSetReg(state,IA64INSIN1(ins),reg);  
    GenericSet_uint64(reg.i, new_gr_r3_nat);
    ProcStateSetReg(state,IA64GETNAT(IA64INSIN1(ins)),reg);
#endif    
    return; 
  }
    
  
  /*intel*/
  


  if(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(base_reloc))==RT_SUBSECTION)
  {
    sec = T_SECTION(REL_TO_RELOCATABLE(base_reloc));
    if(sec->type != CODE_SECTION)
    {
      if(sec->type != RO_DATA_SECTION)
      {
	if(sec->name)
	  printf("Warning: skipping load emulation because of section %s.", sec->name);
	
	ProcStateSetAllBot(state, INS_REGS_DEF(ins));
	ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));
#if 0
	/* wegschrijven nieuwe waarde van R3 */
	GenericSet_uint64(reg.i, new_gr_r3);
	ProcStateSetReg(state,IA64INSIN1(ins),reg);  
	GenericSet_uint64(reg.i, new_gr_r3_nat);
	ProcStateSetReg(state,IA64GETNAT(IA64INSIN1(ins)),reg);
#endif      
	return;
      }
    }
    else
      FATAL(("may not happen"));

  }
  else
  {
    ProcStateSetAllBot(state, INS_REGS_DEF(ins));
    ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));
#if 0
    /* wegschrijven nieuwe waarde van R3 */
    GenericSet_uint64(reg.i, new_gr_r3);
    ProcStateSetReg(state,IA64INSIN1(ins),reg);  
    GenericSet_uint64(reg.i, new_gr_r3_nat);
    ProcStateSetReg(state,IA64GETNAT(IA64INSIN1(ins)),reg);
#endif    
    return;
  }
  
  /*we weten dat we in constant gebied lezen*/



#define ldtype(opcode) (GETBITS64(opcode, 30,6)&(CREATE_MASK(4)<<2)) 

  size = fill_form ? 8 : 1<<GETBITS64(opcode,30,2);
  
  speculative = (ldtype(opcode) == 0x4 || ldtype(opcode) == 0xC);
  advanced = (ldtype(opcode) == 0x8 || ldtype(opcode) == 0xC);
  check_clear = (ldtype(opcode) == 0x20 || ldtype(opcode) == 0x28);
  check_no_clear = (ldtype(opcode) == 0x24);
  check = check_clear || check_no_clear;

  if ((reg_base_update_form || imm_base_update_form) && (IA64INSOUT1(ins) == IA64INSIN1(ins)))
    illegal_operation_fault();

  check_target_register(r1);

  if (reg_base_update_form || imm_base_update_form)
    check_target_register(r3);

  if (!speculative && gr_r3_nat)                         // fault on NaT address
    register_nat_consumption_fault(itype);

  /* defer exception if spec */


  if (check) /*we gaan ervan uit dat het onverandelijke dat is. dus steeds aanwezig. alat_cmp(GENERAL, r1)*/
  {                // no load on ld.c & ALAT hit
    
    /*CHANGED*/
    /*worst case ... => evt verbeteren...*/
    ProcStateSetAllBot(state, INS_REGS_DEF(ins));
    ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));
    
    
    
    return;/*if (check_clear)               // remove entry on ld.c.clr or ld.c.clr.acq
      alat_inval_single_entry(GENERAL, r1);*/
  } 
  else                          /* lees */ 
  {
    FATAL(("Zie hierboven, fout komt 2 keer voor"));
    //offset = G_T_UINT64(GENERIC_SUB(gr_r3,SECTION_GET_CADDRESS(sec)));

    if(offset > G_T_UINT64(SECTION_CSIZE(sec)))
    {
      ProcStateSetAllBot(state, INS_REGS_DEF(ins));
      ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));
#if 0
      /* wegschrijven nieuwe waarde van R3 */
      GenericSet_uint64(reg.i, new_gr_r3);
      ProcStateSetReg(state,IA64INSIN1(ins),reg);  
      GenericSet_uint64(reg.i, new_gr_r3_nat);
      ProcStateSetReg(state,IA64GETNAT(IA64INSIN1(ins)),reg);
#endif      
      return;
    }

    switch (size)
    {
      case 1:
        val = (t_uint64)(*((t_uint8*)sec->data + offset)); 
	break;
      case 2:
        val = (t_uint64)(*((t_uint16*)sec->data + offset/2)); 
	break;
      case 4:
        val = (t_uint64)(*((t_uint32*)sec->data + offset/4)); 
	break;
      case 8:
        val = (t_uint64)(*((t_uint64*)sec->data + offset/8)); 
	break;

    }
    if (fill_form) 
    {                              // fill NaT on ld8.fill

      ProcStateSetAllBot(state, INS_REGS_DEF(ins));
      ProcStateSetTagsetBot(state, INS_REGS_DEF(ins));

#if 0      
      bit_pos = EXTRACT_BITS64(gr_r3,8,3);
      gr_r1 = val;
      /*dit klopt niet !!!!*/
      gr_r1_nat = EXTRACT_BIT64(IA64_AR_UNAT,bit_pos);
#endif      
    } 
    else 
    {                                      // clear NaT on other types
      gr_r1 = ZERO_EXT(val, size * 8);
      gr_r1_nat = 0;
    }
  }

  /* nu nog resultaten wegschrijven... */
  
  GenericSet_uint64(reg.i, gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);  

  __stats.ld_done++;
}

//#endif




/*
Floating-point Load

Format:         (qp) ldffsz.fldtype.ldhint f1 = [r3]                            no_base_update_form          M6
                (qp) ldffsz.fldtype.ldhint f1 = [r3], r2                       reg_base_update_form          M7
                (qp) ldffsz.fldtype.ldhint f1 = [r3], imm9                     imm_base_update_form          M8
                (qp) ldf8.fldtype.ldhint f1 = [r3]                integer_form, no_base_update_form          M6
                (qp) ldf8.fldtype.ldhint f1 = [r3], r2           integer_form, reg_base_update_form          M7
                (qp) ldf8.fldtype.ldhint f1 = [r3], imm9         integer_form, imm_base_update_form          M8
                (qp) ldf.fill.ldhint f1 = [r3]                       fill_form, no_base_update_form          M6
                (qp) ldf.fill.ldhint f1 = [r3], r2                  fill_form, reg_base_update_form          M7
                (qp) ldf.fill.ldhint f1 = [r3], imm9                fill_form, imm_base_update_form          M8
*/


static void Ia64Emulate_ldf(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  size = (fill_form ? 16 : (integer_form ? 8 : fsz));
  speculative = (fldtype == 's' || fldtype == 'sa');
  advanced = (fldtype == 'a' || fldtype == 'sa');
  check_clear = (fldtype == 'c.clr' );
  check_no_clear = (fldtype == 'c.nc');
  check = check_clear || check_no_clear;
  
  
  itype = READ;
  if (speculative) itype |= SPEC ;
  if (advanced) itype |= ADVANCE ; 
  
  
  if (reg_base_update_form || imm_base_update_form)
       check_target_register(r3);
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, 0, 0, 0))
       disabled_fp_register_fault(tmp_isrcode, itype);
  
  
  if (!speculative && gr_r3_nat)                    // fault on NaT address
       register_nat_consumption_fault(itype);
  
  
  defer = speculative && (gr_r3_nat || PSR.ed);// defer exception if spec
  
  if (check && alat_cmp(FLOAT, f1)) {             // no load on ldf.c & ALAT hit
       if (check_clear)                           // remove entry on ldf.c.clr
            alat_inval_single_entry(FLOAT, f1);
  } else {
       if (!defer) {
            paddr = tlb_translate(gr_r3, size, itype, PSR.cpl, &mattr,
                                      &defer);
            if (!defer)
                 val = mem_read(paddr, size, UM.be, mattr, UNORDERED, ldhint);
       }
       if (check_clear || advanced)                  // remove any old ALAT entry
            alat_inval_single_entry(FLOAT, f1);
       if (speculative && defer) {
            fr_f1 = NATVAL;
       } else if (advanced && !speculative && defer) {
            fr_f1 = (integer_form ? FP_INT_ZERO : FP_ZERO);
       } else {                                      // execute load normally
            fr_f1 = fp_mem_to_fr_format(val, size, integer_form);
  
            if ((check_no_clear || advanced) && ma_is_speculative(mattr))
                                                     // add entry to ALAT
                 alat_write(FLOAT, f1, paddr, size);
       }
  }
  
  
  if (imm_base_update_form) {                        // update base register
       gr_r3 = gr_r3 + SIGN_EXT(imm9, 9);
       gr_r3_nat = gr_r3_nat;
  } else if (reg_base_update_form) {
       gr_r3 = gr_r3 + gr_r2;
       gr_r3_nat = gr_r3_nat || gr_r2_nat;
  }
  
  
  if ((reg_base_update_form || imm_base_update_form) && !gr_r3_nat)
       mem_implicit_prefetch(gr_r3, ldhint, itype);
  
  fp_update_psr(f1);
}

#endif









/*
Floating-point Load Pair

Format:         (qp) ldfps.fldtype.ldhint f1, f2 = [r3]          single_form, no_base_update_form            M11
                (qp) ldfps.fldtype.ldhint f1, f2 = [r3], 8           single_form, base_update_form           M12
                (qp) ldfpd.fldtype.ldhint f1, f2 = [r3]          double_form, no_base_update_form            M11
                (qp) ldfpd.fldtype.ldhint f1, f2 = [r3], 16         double_form, base_update_form            M12
                (qp) ldfp8.fldtype.ldhint f1, f2 = [r3]          integer_form, no_base_update_form           M11
                (qp) ldfp8.fldtype.ldhint f1, f2 = [r3], 16         integer_form, base_update_form           M12
*/

static void Ia64Emulate_ldfp(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  size = single_form ? 8 : 16;
  
  
  speculative = (fldtype == 's' || fldtype == 'sa');
  advanced = (fldtype == 'a' || fldtype == 'sa');
  check_clear = (fldtype == 'c.clr');
  check_no_clear = (fldtype == 'c.nc');
  check = check_clear || check_no_clear;
  
  
  itype = READ;
  if (speculative) itype |= SPEC;
  if (advanced) itype |= ADVANCE; 
  
  
  if (fp_reg_bank_conflict(f1, f2))
    illegal_operation_fault();
  
  
  if (base_update_form)
    check_target_register(r3);
  
  fp_check_target_register(f1);
  fp_check_target_register(f2);
  if (tmp_isrcode = fp_reg_disabled(f1, f2, 0, 0))
    disabled_fp_register_fault(tmp_isrcode, itype);
  
  
  if (!speculative && gr_r3_nat)                   // fault on NaT address
    register_nat_consumption_fault(itype);
  
  
  defer = speculative && (gr_r3_nat || PSR.ed);// defer exception if spec
  
  if (check && alat_cmp(FLOAT, f1)) {            // no load on ldfp.c & ALAT hit
    if (check_clear)                            // remove entry on ldfp.c.clr
         alat_inval_single_entry(FLOAT, f1);
  } else {
    if (!defer) {
         paddr = tlb_translate(gr_r3, size, itype, PSR.cpl, &mattr,
                                    &defer);
         if (!defer)
               val = mem_read(paddr, size, UM.be, mattr, UNORDERED, ldhint);
    }
    if (check_clear || advanced)                   // remove any old ALAT entry
         alat_inval_single_entry(FLOAT, f1);
    if (speculative && defer) {
         fr_f1 = NATVAL;
         fr_f2 = NATVAL;
    } else if (advanced && !speculative && defer) {
         fr_f1 = (integer_form ? FP_INT_ZERO : FP_ZERO);
         fr_f2 = (integer_form ? FP_INT_ZERO : FP_ZERO);
    } else {                                       // execute load normally
         if (UM.be) {
               fr_f1 = fp_mem_to_fr_format(val u>> (size/2*8), size/2,
                     integer_form);
               fr_f2 = fp_mem_to_fr_format(val, size/2, integer_form);
         } else {
               fr_f1 = fp_mem_to_fr_format(val, size/2, integer_form);
               fr_f2 = fp_mem_to_fr_format(val u>> (size/2*8), size/2,
                                           integer_form);
         }
  
  
         if ((check_no_clear || advanced) && ma_is_speculative(mattr))
                                                   // add entry to ALAT
               alat_write(FLOAT, f1, paddr, size);
    }
  
    }
  
  
    if (base_update_form) {                          // update base register
         gr_r3 = gr_r3 + size;
         gr_r3_nat = gr_r3_nat;
         if (!gr_r3_nat)
             mem_implicit_prefetch(gr_r3, ldhint, itype);
    }
  
  
    fp_update_psr(f1);
    fp_update_psr(f2);
}

#endif








/*
Line Prefetch

Format:         (qp) lfetch.lftype.lfhint [r3]                                no_base_update_form       M13
                (qp) lfetch.lftype.lfhint [r3], r2                           reg_base_update_form       M14
                (qp) lfetch.lftype.lfhint [r3], imm9                        imm_base_update_form        M15
                (qp) lfetch.lftype.excl.lfhint [r3]           no_base_update_form, exclusive_form       M13
                (qp) lfetch.lftype.excl.lfhint [r3], r2       reg_base_update_form, exclusive_form      M14
                (qp) lfetch.lftype.excl.lfhint [r3], imm9    imm_base_update_form, exclusive_form       M15
*/

/* we doen enkel base_updates */
static void Ia64Emulate_lfetch(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  t_bool imm_base_update_form = IA64OPCODEINFO(ins).instr_format==IA64_INSTR_FORMAT_M15;
  t_bool reg_base_update_form = IA64OPCODEINFO(ins).instr_format==IA64_INSTR_FORMAT_M14;
  t_uint64 gr_r3, gr_r3_nat, gr_r2_nat;
  t_int64 gr_r2;
  t_ia64_reg r3 = IA64INSOUT1(ins);
  register_content reg;
  
  if(!imm_base_update_form && !reg_base_update_form)
    return;

  check_target_register(r3);
  if (ProcStateGetReg(state,r3,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(r3),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  if (reg_base_update_form){
    if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r2 = (t_int64)G_T_UINT64(reg.i);

    if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r2_nat= G_T_UINT64(reg.i);

    gr_r3 += gr_r2;
    gr_r3_nat |= gr_r2_nat;
  }
  else
  {
    gr_r3+=IA64INSIMM1(ins);
  }
  
  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64)gr_r3);
  ProcStateSetReg(state,IA64INSIN1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r3_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSIN1(ins)),reg);
}


#endif




/*
Load Register Stack

Format:         loadrs                                                              M25
*/

static void Ia64Emulate_loadrs(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (AR[RSC].mode != 0)
      illegal_operation_fault();
  
  
  if ((CFM.sof != 0) && (AR[RSC].loadrs != 0))
      illegal_operation_fault();
  
  
  rse_ensure_regs_loaded(AR[RSC].loadrs);                  // can raise faults listed below
  AR[RNAT] = undefined();
}
#endif








/*
Memory Fence

Format:         (qp) mf                                      ordering_form    M24
                (qp) mf.a                                 acceptance_form     M24
*/

static void Ia64Emulate_mf(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (acceptance_form)
     acceptance_fence();
  else // ordering_form
     ordering_fence();
}

#endif







/*
Mix

Format:         (qp) mix1.l r1 = r2, r3          one_byte_form, left_form          I2
                (qp) mix2.l r1 = r2, r3          two_byte_form, left_form          I2
                (qp) mix4.l r1 = r2, r3         four_byte_form, left_form          I2
                (qp) mix1.r r1 = r2, r3        one_byte_form, right_form           I2
                (qp) mix2.r r1 = r2, r3        two_byte_form, right_form           I2
                (qp) mix4.r r1 = r2, r3        four_byte_form, right_form          I2
*/
/*status: TODO */
static void Ia64Emulate_mix(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[8], y[8];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
  
  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);
    
  t_bool left_form = x2b == 2;
    

  t_bool one_byte_form = 0;
  t_bool two_byte_form = 0;
  t_bool four_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,36,1);
  if (!four_byte_form){
    one_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==0;
    two_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==1;
  }
  
  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  
  
  
  /* perform operations */
  
  check_target_register(r1);

  if (one_byte_form) {                                   // one-byte elements
       x[0] = EXTRACT_BITS64(gr_r2,7,0);       y[0] = EXTRACT_BITS64(gr_r3,7,0);
       x[1] = EXTRACT_BITS64(gr_r2,15,8);      y[1] = EXTRACT_BITS64(gr_r3,15,8);
       x[2] = EXTRACT_BITS64(gr_r2,23,16);     y[2] = EXTRACT_BITS64(gr_r3,23,16);
       x[3] = EXTRACT_BITS64(gr_r2,31,24);     y[3] = EXTRACT_BITS64(gr_r3,31,24);
       x[4] = EXTRACT_BITS64(gr_r2,39,32);     y[4] = EXTRACT_BITS64(gr_r3,39,32);
       x[5] = EXTRACT_BITS64(gr_r2,47,40);     y[5] = EXTRACT_BITS64(gr_r3,47,40);
       x[6] = EXTRACT_BITS64(gr_r2,55,48);     y[6] = EXTRACT_BITS64(gr_r3,55,48);
       x[7] = EXTRACT_BITS64(gr_r2,63,56);     y[7] = EXTRACT_BITS64(gr_r3,63,56);

       if (left_form)
           gr_r1 = concatenate8(x[7], y[7], x[5], y[5],
                                 x[3], y[3], x[1], y[1]);
       else // right_form
           gr_r1 = concatenate8(x[6], y[6], x[4], y[4],
                                 x[2], y[2], x[0], y[0]);


  } else if (two_byte_form) {                            // two-byte elements
       x[0] = EXTRACT_BITS64(gr_r2,15,0);      y[0] = EXTRACT_BITS64(gr_r3,15,0);
       x[1] = EXTRACT_BITS64(gr_r2,31,16);     y[1] = EXTRACT_BITS64(gr_r3,31,16);
       x[2] = EXTRACT_BITS64(gr_r2,47,32);     y[2] = EXTRACT_BITS64(gr_r3,47,32);
       x[3] = EXTRACT_BITS64(gr_r2,63,48);     y[3] = EXTRACT_BITS64(gr_r3,63,48);

       if (left_form)
           gr_r1 = concatenate4(x[3], y[3], x[1], y[1]);
       else // right_form
           gr_r1 = concatenate4(x[2], y[2], x[0], y[0]);

  } else {                                               // four-byte elements
       x[0] = EXTRACT_BITS64(gr_r2,31,0);      y[0] = EXTRACT_BITS64(gr_r3,31,0);
       x[1] = EXTRACT_BITS64(gr_r2,63,32);     y[1] = EXTRACT_BITS64(gr_r3,63,32);

       if (left_form)
           gr_r1 = concatenate2(x[1], y[1]);
       else // right_form
           gr_r1 = concatenate2(x[0], y[0]);
  }
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

}

#endif






/*
Move Application Register

Format:         (qp) mov r1 = ar3                               pseudo-op
                (qp) mov ar3 = r2                               pseudo-op
                (qp) mov ar3 = imm8                             pseudo-op
                (qp) mov.i r1 = ar3                     i_form, from_form         I28
                (qp) mov.i ar3 = r2          i_form, register_form, to_form       I26
                (qp) mov.i ar3 = imm8   i_form, immediate_form, to_form           I27
                (qp) mov.m r1 = ar3                    m_form, from_form          M31
                (qp) mov.m ar3 = r2         m_form, register_form, to_form        M29
                (qp) mov.m ar3 = imm8  m_form, immediate_form, to_form            M30
*/

static void Ia64Emulate_mov_ar(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}

#if 0
  t_uint64 ar_a3, gr_r2 = 0, gr_r2_nat=0, gr_r1, gr_r1_nat, ar_rsc, ar_psr,ar_bsp,ar_rnat, tmp_type, tmp_val;

  t_bool i_form = IA64INSTYPE(ins)==IA64_TYPE_I;
  t_bool from_form =  IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_I28 || \
    IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_M31;
  /*t_bool immediate_form =  IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_I27 || \
    IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_M30;*/
  t_bool register_form =  IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_I26 || \
    IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_M29;
  register_content reg;
  t_ia64_reg ar3, r1;

  
  Ia64InsEmulDetermineTag(ins, state);
  
  tmp_type = (i_form ? IA64_TYPE_I : IA64_TYPE_M);

  if(from_form)
    r1 = IA64INSOUT1(ins);
  else 
    ar3 = IA64INSOUT1(ins);
  if (!register_form){
    if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    if(from_form)
    {
      ar_a3 = G_T_UINT64(reg.i);
      ar3=IA64INSIN1(ins);
    }
    else gr_r2 = G_T_UINT64(reg.i);
  }

  if(register_form)
    if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
  gr_r2_nat = G_T_UINT64(reg.i);


  if (ProcStateGetReg(state,IA64_AR_RSC,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  ar_rsc = G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64_PSR,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  ar_psr = G_T_UINT64(reg.i);

  FATAL(("Implement reserved reg"));
  if (is_reserved_reg(tmp_type, ar3))
    illegal_operation_fault();
  if (from_form) {
    check_target_register(r1);
    if (((ar3 == IA64_AR_BSPSTORE) || (ar3 == IA64_AR_RNAT)) && (GETBITS64(ar_rsc,0,2) != 0))
      illegal_operation_fault();


    if (ar3 == IA64_AR_ITC){
      if (ProcStateGetReg(state,IA64_PSR,&reg))
      {
	ProcStateSetAllBot(state,INS_REGS_DEF(ins));
	return;
      }
      if (GETBITS64(G_T_UINT64(reg.i),23,1) && GETBITS64(G_T_UINT64(reg.i),32,2) != 0)
	privileged_register_fault();
    }

    gr_r1 = (is_ignored_reg(ar3)) ? 0 : ar_a3;
    gr_r1_nat = 0;
   
    /* resultaat wegschrijven */
    GenericSet_uint64(reg.i, gr_r1);
    ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

    GenericSet_uint64(reg.i, gr_r1_nat);
    ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

  } 
  else 
  {                                                  // to_form
    tmp_val = (register_form) ? gr_r2 : IA64INSIMM1(ins);

    if (is_read_only_register(AR_TYPE, ar3) || (((ar3 == IA64_AR_BSPSTORE) || (ar3 == IA64_AR_RNAT)) && ((GETBITS64(ar_rsc,0,2) != 0))))
      illegal_operation_fault(); 


    if (register_form && gr_r2_nat)
      register_nat_consumption_fault(0);

    if (is_reserved_field(AR_TYPE, ar3, tmp_val))
      reserved_register_field_fault();

    if ((is_kernel_reg(ar3) || ar3 == IA64_AR_ITC))
    {
      if (ProcStateGetReg(state,IA64_PSR,&reg))
      {
	ProcStateSetAllBot(state,INS_REGS_DEF(ins));
	return;
      }
      if(GETBITS64(ar_rsc,0,2) != 0)
      {
	privileged_register_fault();
      }
    }

    if (!is_ignored_reg(ar3)) 
    {
      tmp_val = ignored_field_mask(AR_TYPE, ar3, tmp_val);
      // check for illegal promotion
      if (ar3 == IA64_AR_RSC){ /*EXTRACT_BITS64(tmp_val,3,2) < PSR.cpl*/
	if (ProcStateGetReg(state,IA64_PSR,&reg))
	{
	  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
	  return;
	}
	if(GETBITS64(tmp_val,2,2) < GETBITS64(G_T_UINT64(reg.i),32,2))
	  SET_BITS64(tmp_val,3,2,GETBITS64(G_T_UINT64(reg.i),32,2));

      }
      ar_a3 = tmp_val;
      /* resultaat wegschrijven */
      GenericSet_uint64(reg.i, ar_a3);
      ProcStateSetReg(state,IA64INSOUT1(ins),reg);

      if (ar3 == IA64_AR_BSPSTORE) 
      {
	ar_bsp = rse_update_internal_stack_pointers(tmp_val);
	ar_rnat = undefined();
	/* resultaat wegschrijven */
	GenericSet_uint64(reg.i, ar_bsp);
	ProcStateSetReg(state,IA64_AR_BSP,reg);
	GenericSet_uint64(reg.i, ar_rnat);
	ProcStateSetReg(state,IA64_AR_RNAT,reg);
      }
    }
  }
}
#endif







/*
Move Branch Register

Format:         (qp) mov r1 = b2                                   from_form            I22
                (qp) mov b1 = r2                                    pseudo-op
                (qp) mov.mwh.ih b1 = r2, tag13                            to_form       I21
                (qp) mov.ret.mwh.ih b1 = r2, tag13       return_form, to_form           I21
*/
/*status: CHECKED */
static void Ia64Emulate_mov_br(t_ins * ins, t_procstate * state)
{
  t_uint64 br_b1, gr_r1, gr_r1_nat, br_b2, gr_r2, gr_r2_nat, ip, tmp_tag;
    
  t_bool from_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_I22;

  register_content reg;

  t_ia64_reg r1 = IA64INSOUT1(ins); 
  
  Ia64InsEmulDetermineTag(ins, state);

  __stats.mov_br_total++;


  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  
  if (from_form) br_b2 = G_T_UINT64(reg.i);
  else gr_r2 = G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSOUT1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  if (from_form) gr_r1 = G_T_UINT64(reg.i);
  else br_b1 = G_T_UINT64(reg.i);

  /* perform operations */
  
  if (from_form) {
    check_target_register(r1);
    gr_r1 = br_b2;
    gr_r1_nat = 0;
    
    /* wegschrijven */
    
    GenericSet_uint64(reg.i, gr_r1);
    ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

    GenericSet_uint64(reg.i, gr_r1_nat);
    ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);
  } 
  else 
  { // to_form
/*    
    if (ProcStateGetReg(state,IA64_IP,&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    ip = G_T_UINT64(reg.i);
    tmp_tag = ip + IA64INSIMM1(ins);
*/    
    if (gr_r2_nat)
      register_nat_consumption_fault(0);
    br_b1 = gr_r2;
    
//    branch_predict(GETBITS64(opcode,20,2), GETBITS64(opcode,23,1), GETBITS64(opcode,22,1), gr_r2, tmp_tag); 
    
    /* wegschrijven */
    GenericSet_uint64(reg.i, br_b1);
    ProcStateSetReg(state,IA64INSOUT1(ins),reg);
  }

  __stats.mov_br_done++;

}







/*
Move Control Register

Format:           (qp) mov r1 = cr3                                from_form         M33
                  (qp) mov cr3 = r2                                    to_form       M32
*/

static void Ia64Emulate_mov_cr(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (is_reserved_reg(CR_TYPE, cr3)
      || to_form && is_read_only_reg(CR_TYPE, cr3)
      || PSR.ic && is_interruption_cr(cr3))
  {
      illegal_operation_fault();
  }
  
  if (from_form)
      check_target_register(r1);
  if (PSR.cpl != 0)
      privileged_operation_fault(0);
  
  if (from_form) {
      if (cr3 == IVR)
          check_interrupt_request();
  
      if (cr3 == ITIR)
          gr_r1 = impl_itir_cwi_mask(CR[ITIR]);
      else
          gr_r1 = CR[cr3];
  
      gr_r1_nat = 0;
  } else {                       // to_form
      if (gr_r2_nat)
          register_nat_consumption_fault(0);
  
      if (is_reserved_field(CR_TYPE, cr3, gr_r2))
          reserved_register_field_fault();
      if (cr3 == EOI)
          end_of_interrupt();
  
      tmp_val = ignored_field_mask(CR_TYPE, cr3, gr_r2);
      CR[cr3] = tmp_val;
      if (cr3 == IIPA)
          last_IP = tmp_val;
  }
}

#endif 







/*
Move Indirect Register

Format:         (qp) mov r1 = ireg[r3]                              from_form      M43
                (qp) mov ireg[r3] = r2                                 to_form     M42
*/

static void Ia64Emulate_mov_ireg(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (ireg == RR_TYPE)
       tmp_index = EXTRACT_BITS64(gr_r3,63,61);
  else // all other register types
       tmp_index = EXTRACT_BITS64(gr_r3,7,0);

  if (from_form) {
       check_target_register(r1);

       if (PSR.cpl != 0 && !(ireg == PMD_TYPE || ireg == CPUID_TYPE))
             privileged_operation_fault(0);


       if (gr_r3_nat)
             register_nat_consumption_fault(0);


       if (is_reserved_reg(ireg, tmp_index))
             reserved_register_field_fault();


       if (ireg == PMD_TYPE) {
             if ((PSR.cpl != 0) && ((PSR.sp == 1) ||
                   (tmp_index > 3 &&
                   tmp_index <= IMPL_MAXGENERIC_PMCPMD &&
                   PMC[tmp_index].pm == 1)))
                  gr_r1 = 0;
             else 
                  gr_r1 = pmd_read(tmp_index);
       } else
             switch (ireg) {
                  case CPUID_TYPE: gr_r1 = CPUID[tmp_index]; break;
                  case DBR_TYPE:       gr_r1 = DBR[tmp_index]; break;
                  case IBR_TYPE:       gr_r1 = IBR[tmp_index]; break;
                  case PKR_TYPE:       gr_r1 = PKR[tmp_index]; break;
                  case PMC_TYPE:       gr_r1 = pmc_read(tmp_index); break;
                  case RR_TYPE:        gr_r1 = RR[tmp_index]; break;
             }
       gr_r1_nat = 0;
  } else {               // to_form
       if (PSR.cpl != 0)
             privileged_operation_fault(0);


       if (gr_r2_nat || gr_r3_nat)
             register_nat_consumption_fault(0);


       if (is_reserved_reg(ireg, tmp_index)
             || is_reserved_field(ireg, tmp_index, gr_r2))
             reserved_register_field_fault();
       if (ireg == PKR_TYPE && EXTRACT_BIT64(gr_r2,0) == 1) { // writing valid prot key
             if ((tmp_slot = tlb_search_pkr(EXTRACT_BITS64(gr_r2,31,8))) != NOT_FOUND)
                  PKR[tmp_slot].v = 0; // clear valid bit of matching key reg
       }
       tmp_val = ignored_field_mask(ireg, tmp_index, gr_r2);
       switch (ireg) {
             case DBR_TYPE:        DBR[tmp_index] = tmp_val; break;
             case IBR_TYPE:        IBR[tmp_index] = tmp_val; break;
             case PKR_TYPE:        PKR[tmp_index] = tmp_val; break;
             case PMC_TYPE:        pmc_write(tmp_index, tmp_val); break;
             case PMD_TYPE:        pmd_write(tmp_index, tmp_val); break;
             case RR_TYPE:         RR[tmp_index]= tmp_val; break;
       } 
  }
}
#endif








/*
Move Instruction Pointer

Format:         (qp) mov r1 = ip                                                  I25
*/

static void Ia64Emulate_mov_ip(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  
  t_uint64 gr_r1, gr_r1_nat, IP;
  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
/*
  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  IP = G_T_UINT64(reg.i);
*/
  /*extract the IP from ins_address...*/
  
  IP = INS_CADDRESS(ins) & ~0x0FULL; /*bundle address !*/
  

  /* perform operations */
  
  check_target_register(r1);
  
  gr_r1 = IP;
  gr_r1_nat = 0;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif  





/**
* TODO: roterende predikaten aanpassen. => kan evt anders voor problemen zorgen !!!
* */

/*
Move Predicates

Format:         (qp) mov r1 = pr                                   from_form         I25
                (qp) mov pr = r2, mask17                               to_form       I23
                (qp) mov pr.rot = imm44                        to_rotate_form        I24
*/

static void Ia64Emulate_mov_pr(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  t_uint64 gr_r1, gr_r1_nat, gr_r2, gr_r2_nat, tmp_src;

  t_uint8 i;

  t_bool from_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,2) == 0;
  t_bool to_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,2) == 3;

  register_content reg;

  t_ia64_reg r1; 

  Ia64InsEmulDetermineTag(ins, state);

  if (from_form)
    r1 = IA64INSOUT1(ins);

  if (to_form){
    if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r2 = G_T_UINT64(reg.i);

    if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r2_nat = G_T_UINT64(reg.i);
  }

  /* perform operations */

  if (from_form) {
    check_target_register(r1);
    gr_r1 = 1;                                          // PR[0] is always 1
    for (i = 1; i <= 63; i++) 
    {
//      SET_BIT64(gr_r1,i,PR[pr_phys_to_virt(i)]);
    }
    gr_r1_nat = 0;
  } 
  else if (to_form) 
  {
    if (gr_r2_nat)
      register_nat_consumption_fault(0);
    tmp_src = IA64INSIMM1(ins);
    for (i = 1; i <= 63; i++) 
    {
//      if (EXTRACT_BIT64(tmp_src,i))
//	PR[pr_phys_to_virt(i)] = EXTRACT_BIT64(gr_r2,i);
    }
  } 
  else 
  {       // to_rotate_form
    tmp_src = IA64INSIMM1(ins);
    for (i = 16; i <= 63; i++) 
    {
//      PR[pr_phys_to_virt(i)] = EXTRACT_BIT64(tmp_src,i);
    }
  }
}

#endif






/*
Move Processor Status Register

Format:           (qp) mov r1 = psr                              from_form          M36
                  (qp) mov psr.l = r2                                 to_form       M35
*/

static void Ia64Emulate_mov_psr(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (from_form)
      check_target_register(r1);
  if (PSR.cpl != 0)
      privileged_operation_fault(0);
  
  
  if (from_form) {
      tmp_val = ZERO_EXT(EXTRACT_BITS64(PSR,31,0), 32);                // read lower 32 bits
      tmp_val |= EXTRACT_BITS64(PSR,36,35) << 35;                      // read mc and it bits
      gr_r1 = tmp_val;                                 // other bits read as zero
      gr_r1_nat = 0;
  } else {                 // to_form
      if (gr_r2_nat)
          register_nat_consumption_fault(0);
  
  
      if (is_reserved_field(PSR_TYPE, PSR_MOVPART, gr_r2))
          reserved_register_field_fault();
  
  
      SET_BITS64(PSR,31,0,EXTRACT_BITS64(gr_r2,31,0));
  }
}

#endif








/*
Move User Mask

Format:           (qp) mov r1 = psr.um                          from_form        M36
                  (qp) mov psr.um = r2                                to_form    M35
*/

static void Ia64Emulate_mov_um(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (from_form) {
      check_target_register(r1);
  
      gr_r1 = ZERO_EXT(EXTRACT_BITS64(PSR,5,0), 6);
      gr_r1_nat = 0;
  } else {                                                       // to_form
      if (gr_r2_nat)
         register_nat_consumption_fault(0);
  
  
      if (is_reserved_field(PSR_TYPE, PSR_UM, gr_r2))
         reserved_register_field_fault();
  
  
      SET_BITS64(PSR,1,0,EXTRACT_BITS64(gr_r2,1,0));
  
      if (PSR.sp == 0)                    // unsecured perf monitor
         SET_BIT64(PSR,2,EXTRACT_BIT64(gr_r2,2));
  
      SET_BITS64(PSR,5,3,EXTRACT_BITS64(gr_r2,5,3));
  }
}
#endif







/*
Move Long Immediate

Format:         (qp) movl r1 = imm64                                        X2
*/
/*status: CHECKED */
static void Ia64Emulate_movl(t_ins * ins, t_procstate * state)
{
  register_content reg;
 
  t_ia64_reg r1 = IA64INSOUT1(ins);
  
  Ia64InsEmulDetermineTag(ins, state);

  __stats.movl_total++;

  check_target_register(r1);

  GenericSet_uint64(reg.i, IA64INSIMM1(ins));
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, 0);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

  __stats.movl_done++;

}







/*
Mux

Format:         (qp) mux1 r1 = r2, mbtype4                       one_byte_form     I3
                (qp) mux2 r1 = r2, mhtype8                       two_byte_form     I4
*/

/*status: TODO */
static void Ia64Emulate_mux(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[8], res[8];

  t_int64 gr_r1, gr_r2;
  t_uint64 gr_r1_nat, gr_r2_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
  
  t_uint64 zb = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1);
    
  t_bool one_byte_form = zb == 0;
    
  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);


  
  /* perform operations */
  
  check_target_register(r1);

  if (one_byte_form) {
       x[0] = EXTRACT_BITS64(gr_r2,7,0);
       x[1] = EXTRACT_BITS64(gr_r2,15,8);
       x[2] = EXTRACT_BITS64(gr_r2,23,16);
       x[3] = EXTRACT_BITS64(gr_r2,31,24);
       x[4] = EXTRACT_BITS64(gr_r2,39,32);
       x[5] = EXTRACT_BITS64(gr_r2,47,40);
       x[6] = EXTRACT_BITS64(gr_r2,55,48);
       x[7] = EXTRACT_BITS64(gr_r2,63,56);

       switch (IA64INSIMM1(ins)) {
	 case 0xB:					//'@rev':		
	   gr_r1 = concatenate8(x[0], x[1], x[2], x[3],
	       x[4], x[5], x[6], x[7]);
	   break;


	 case 8:					//'@mix':
	   gr_r1 = concatenate8(x[7], x[3], x[5], x[1],
	       x[6], x[2], x[4], x[0]);
	   break;


	 case 9:					//'@shuf':
	   gr_r1 = concatenate8(x[7], x[3], x[6], x[2], 
	       x[5], x[1], x[4], x[0]);
	   break;


	 case 0xA: 					//'@alt':
	   gr_r1 = concatenate8(x[7], x[5], x[3], x[1],
	       x[6], x[4], x[2], x[0]);
	   break;


	 case 0:					//'@brcst':
	   gr_r1 = concatenate8(x[0], x[0], x[0], x[0],
	       x[0], x[0], x[0], x[0]);
	   break;
       }
  } else {                                           // two_byte_form
       x[0] = EXTRACT_BITS64(gr_r2,15,0);
       x[1] = EXTRACT_BITS64(gr_r2,31,16);
       x[2] = EXTRACT_BITS64(gr_r2,47,32);
       x[3] = EXTRACT_BITS64(gr_r2,63,48);

       res[0] = x[EXTRACT_BITS64(IA64INSIMM1(ins),1,0)];
       res[1] = x[EXTRACT_BITS64(IA64INSIMM1(ins),3,2)];
       res[2] = x[EXTRACT_BITS64(IA64INSIMM1(ins),5,4)];
       res[3] = x[EXTRACT_BITS64(IA64INSIMM1(ins),7,6)];


       gr_r1 = concatenate4(res[3], res[2], res[1], res[0]);
  }
  gr_r1_nat = gr_r2_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif







/*
No Operation

Format:         (qp) nop imm21                                   pseudo-op
                (qp) nop.i imm21                                i_unit_form      I19
                (qp) nop.b imm21                             b_unit_form          B9
                (qp) nop.m imm21                             m_unit_form         M37
                (qp) nop.f imm21                                 f_unit_form     F15
                (qp) nop.x imm62                             x_unit_form          X1
*/

static void Ia64Emulate_nop(t_ins * ins, t_procstate * state)
{
  __stats.nop_skipped++;
//  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
}

#endif






/*
Logical Or

Format:         (qp) or r1 = r2, r3                        register_form      A1
                (qp) or r1 = imm8, r3                       imm8_form         A3
*/
/*status: CHECKED */
static void Ia64Emulate_or(t_ins * ins, t_procstate * state)
{
  t_ia64_reg r1 = IA64INSOUT1(ins), r3;

  t_bool register_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A3;

  register_content reg;

  t_uint64 tmp_nat, gr_r1_nat, gr_r2_nat, gr_r3_nat;
  t_int64 tmp_src, gr_r1, gr_r2, gr_r3;
  
  Ia64InsEmulDetermineTag(ins, state);

  __stats.or_total++;
  
  if(register_form)
  {
    if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
      return; 
    
    r3 = IA64INSIN2(ins);
  }
  else
    r3 = IA64INSIN1(ins); 
  /* opgepast! moet hier IN1 zijn (zie corr.decodering): eerst registers...*/

  if(!Ia64CheckAndGetRegisterState(state, ins, r3, &gr_r3, &gr_r3_nat))
    return;  
  
 
  /* perform operations */
  check_target_register(r1);

  tmp_src = (register_form ? gr_r2 : IA64INSIMM1(ins));
  tmp_nat = (register_form ? gr_r2_nat : 0);

  gr_r1 = tmp_src | gr_r3;
  gr_r1_nat = tmp_nat || gr_r3_nat;

  
  /* nu nog resultaten wegschrijven... */
  GenericSet_uint64(reg.i, (t_uint64)gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

  __stats.or_done++;
}








/*
Pack

Format:         (qp) pack2.sss r1 = r2, r3       two_byte_form, signed_saturation_form             I2
                (qp) pack2.uss r1 = r2, r3     two_byte_form, unsigned_saturation_form             I2
                (qp) pack4.sss r1 = r2, r3      four_byte_form, signed_saturation_form             I2
*/
/*status: TODO */

static void Ia64Emulate_pack(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  static t_int64 temp[8];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
  
  t_uint8 i;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);
    
  t_bool signed_saturation_form = x2b == 2;
    
  t_bool two_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1);

  t_int64 max, min;
  
  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  
  
  
  /* perform operations */

  check_target_register(r1);

  if (two_byte_form) {
       if (signed_saturation_form) {
            max = SIGN_EXT(0x7f, 8);
            min = SIGN_EXT(0x80, 8);
       } else {                                  // unsigned_saturation_form
            max = 0xff;
            min = 0x00;
       }
       temp[0] = SIGN_EXT(EXTRACT_BITS64(gr_r2,15,0), 16);
       temp[1] = SIGN_EXT(EXTRACT_BITS64(gr_r2,31,16), 16);
       temp[2] = SIGN_EXT(EXTRACT_BITS64(gr_r2,47,32), 16);
       temp[3] = SIGN_EXT(EXTRACT_BITS64(gr_r2,63,48), 16);
       temp[4] = SIGN_EXT(EXTRACT_BITS64(gr_r3,15,0), 16);
       temp[5] = SIGN_EXT(EXTRACT_BITS64(gr_r3,31,16), 16);
       temp[6] = SIGN_EXT(EXTRACT_BITS64(gr_r3,47,32), 16);
       temp[7] = SIGN_EXT(EXTRACT_BITS64(gr_r3,63,48), 16);

       for (i = 0; i < 8; i++) {
            if (temp[i] > max)
               temp[i] = max;


            if (temp[i] < min)
               temp[i] = min;
       }


       gr_r1 = concatenate8(temp[7], temp[6], temp[5], temp[4],
                              temp[3], temp[2], temp[1], temp[0]);


  } else {                                       // four_byte_form
       max = SIGN_EXT(0x7fff, 16);               // signed_saturation_form
       min = SIGN_EXT(0x8000, 16);
       temp[0] = SIGN_EXT(EXTRACT_BITS64(gr_r2,31,0), 32);
       temp[1] = SIGN_EXT(EXTRACT_BITS64(gr_r2,63,32), 32);
       temp[2] = SIGN_EXT(EXTRACT_BITS64(gr_r3,31,0), 32);
       temp[3] = SIGN_EXT(EXTRACT_BITS64(gr_r3,63,32), 32);

       for (i = 0; i < 4; i++) {
            if (temp[i] > max)
               temp[i] = max;


            if (temp[i] < min)
               temp[i] = min;
       }


       gr_r1 = concatenate4(temp[3], temp[2], temp[1], temp[0]);
  }
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}


#endif





/*
Parallel Add

Format:         (qp) padd1 r1 = r2, r3                               one_byte_form, modulo_form              A9
                (qp) padd1.sss r1 = r2, r3              one_byte_form, sss_saturation_form                   A9
                (qp) padd1.uus r1 = r2, r3              one_byte_form, uus_saturation_form                   A9
                (qp) padd1.uuu r1 = r2, r3             one_byte_form, uuu_saturation_form                    A9
                (qp) padd2 r1 = r2, r3                               two_byte_form, modulo_form              A9
                (qp) padd2.sss r1 = r2, r3              two_byte_form, sss_saturation_form                   A9
                (qp) padd2.uus r1 = r2, r3             two_byte_form, uus_saturation_form                    A9
                (qp) padd2.uuu r1 = r2, r3             two_byte_form, uuu_saturation_form                    A9
                (qp) padd4 r1 = r2, r3                               four_byte_form, modulo_form             A9
*/
/*status: TODO */
static void Ia64Emulate_padd(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[8], y[8], temp[8];

  
  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;
  
  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_int64 max, min;

  t_uint8 i;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);
  
  t_bool sss_saturation_form = x2b ==1;
  t_bool uuu_saturation_form = x2b ==2;
  t_bool uus_saturation_form = x2b ==3;
  
  t_bool one_byte_form = 0;
  t_bool two_byte_form = 0;
  t_bool four_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,36,1);
  if (!four_byte_form){
    one_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==0;
    two_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==1;
  }

  Ia64InsEmulDetermineTag(ins, state);
  
  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64)G_T_UINT64(reg.i);

  /* Extract values from procstate */
  
  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64)G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  /* perform operations */
  
  check_target_register(r1);

  if (one_byte_form) 
  {                                   // one-byte elements
    x[0] = EXTRACT_BITS64(gr_r2,7,0);         y[0] = EXTRACT_BITS64(gr_r3,7,0);
    x[1] = EXTRACT_BITS64(gr_r2,15,8);        y[1] = EXTRACT_BITS64(gr_r3,15,8);
    x[2] = EXTRACT_BITS64(gr_r2,23,16);       y[2] = EXTRACT_BITS64(gr_r3,23,16);
    x[3] = EXTRACT_BITS64(gr_r2,31,24);       y[3] = EXTRACT_BITS64(gr_r3,31,24);
    x[4] = EXTRACT_BITS64(gr_r2,39,32);       y[4] = EXTRACT_BITS64(gr_r3,39,32);
    x[5] = EXTRACT_BITS64(gr_r2,47,40);       y[5] = EXTRACT_BITS64(gr_r3,47,40);
    x[6] = EXTRACT_BITS64(gr_r2,55,48);       y[6] = EXTRACT_BITS64(gr_r3,55,48);
    x[7] = EXTRACT_BITS64(gr_r2,63,56);       y[7] = EXTRACT_BITS64(gr_r3,63,56);

    if (sss_saturation_form) 
    {
      max = SIGN_EXT(0x7f, 8);
      min = SIGN_EXT(0x80, 8);


      for (i = 0; i < 8; i++) 
      {
	temp[i] = SIGN_EXT(x[i], 8) + SIGN_EXT(y[i], 8);
      }
    } 
    else if (uus_saturation_form) 
    {
      max = 0xff;
      min = 0x00;


      for (i = 0; i < 8; i++) 
      {
	temp[i] = ZERO_EXT(x[i], 8) + SIGN_EXT(y[i], 8);
      }
    } 
    else if (uuu_saturation_form) 
    {
      max = 0xff;
      min = 0x00;


      for (i = 0; i < 8; i++) 
      {
	temp[i] = ZERO_EXT(x[i], 8) + ZERO_EXT(y[i], 8);
      }
    } 
    else 
    {                                            // modulo_form
      for (i = 0; i < 8; i++) 
      {
	temp[i] = ZERO_EXT(x[i], 8) + ZERO_EXT(y[i], 8);
      }
    }


    if (sss_saturation_form || uus_saturation_form || uuu_saturation_form) 
    {
      for (i = 0; i < 8; i++) 
      {
	if (temp[i] > max)
	  temp[i] = max;


	if (temp[i] < min)
	  temp[i] = min;
      }
    }
    gr_r1 = concatenate8(temp[7], temp[6], temp[5], temp[4], temp[3], temp[2], temp[1], temp[0]);


  } 
  else if (two_byte_form) 
  {                            // 2-byte elements
    x[0] = EXTRACT_BITS64(gr_r2,15,0);        y[0] = EXTRACT_BITS64(gr_r3,15,0);
    x[1] = EXTRACT_BITS64(gr_r2,31,16);       y[1] = EXTRACT_BITS64(gr_r3,31,16);
    x[2] = EXTRACT_BITS64(gr_r2,47,32);       y[2] = EXTRACT_BITS64(gr_r3,47,32);
    x[3] = EXTRACT_BITS64(gr_r2,63,48);       y[3] = EXTRACT_BITS64(gr_r3,63,48);

    if (sss_saturation_form) 
    {
      max = SIGN_EXT(0x7fff, 16);
      min = SIGN_EXT(0x8000, 16);


      for (i = 0; i < 4; i++) 
      {
	temp[i] = SIGN_EXT(x[i], 16) + SIGN_EXT(y[i], 16);
      }
    } 
    else if (uus_saturation_form) 
    {
      max = 0xffff;
      min = 0x0000;


      for (i = 0; i < 4; i++) 
      {
	temp[i] = ZERO_EXT(x[i], 16) + SIGN_EXT(y[i], 16);
      }
    } 
    else if (uuu_saturation_form) 
    {
      max = 0xffff;
      min = 0x0000;


      for (i = 0; i < 4; i++) 
      {
	temp[i] = ZERO_EXT(x[i], 16) + ZERO_EXT(y[i], 16);
      }
    } 
    else 
    {                                              // modulo_form
      for (i = 0; i < 4; i++) 
      {
	temp[i] = ZERO_EXT(x[i], 16) + ZERO_EXT(y[i], 16);
      }
    }


    if (sss_saturation_form || uus_saturation_form || uuu_saturation_form) 
    {
      for (i = 0; i < 4; i++) 
      {
	if (temp[i] > max)
	  temp[i] = max;


	if (temp[i] < min)
	  temp[i] = min;
      }
    }
    gr_r1 = concatenate4(temp[3], temp[2], temp[1], temp[0]);

  } 
  else 
  {                                                   // four-byte elements
    x[0] = EXTRACT_BITS64(gr_r2,31,0);          y[0] = EXTRACT_BITS64(gr_r3,31,0);
    x[1] = EXTRACT_BITS64(gr_r2,63,32);         y[1] = EXTRACT_BITS64(gr_r3,63,32);

    for (i = 0; i < 2; i++) 
    {                             // modulo_form
      temp[i] = ZERO_EXT(x[i], 32) + ZERO_EXT(y[i], 32);
    }


    gr_r1 = concatenate2(temp[1], temp[0]);
  }


  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */
  
  GenericSet_uint64(reg.i, (t_uint64)gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  
  
  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}
#endif







/*
Parallel Average

Format:         (qp) pavg1 r1 = r2, r3          normal_form, one_byte_form           A9
                (qp) pavg1.raz r1 = r2, r3         raz_form, one_byte_form           A9
                (qp) pavg2 r1 = r2, r3          normal_form, two_byte_form           A9
                (qp) pavg2.raz r1 = r2, r3         raz_form, two_byte_form           A9
*/
/*status: TODO */
static void Ia64Emulate_pavg(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[8], y[8], temp[8], res[8];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint8 i;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);

  t_bool raz_form = x2b == 3;

  t_bool one_byte_form =  GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==0;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* Extract values from procstate */
  
  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= (t_int64) G_T_UINT64(reg.i);

  /* perform operations */
  
  check_target_register(r1);

  if (one_byte_form) {
    x[0] = EXTRACT_BITS64(gr_r2,7,0);        y[0] = EXTRACT_BITS64(gr_r3,7,0);
    x[1] = EXTRACT_BITS64(gr_r2,15,8);       y[1] = EXTRACT_BITS64(gr_r3,15,8);
    x[2] = EXTRACT_BITS64(gr_r2,23,16);      y[2] = EXTRACT_BITS64(gr_r3,23,16);
    x[3] = EXTRACT_BITS64(gr_r2,31,24);      y[3] = EXTRACT_BITS64(gr_r3,31,24);
    x[4] = EXTRACT_BITS64(gr_r2,39,32);      y[4] = EXTRACT_BITS64(gr_r3,39,32);
    x[5] = EXTRACT_BITS64(gr_r2,47,40);      y[5] = EXTRACT_BITS64(gr_r3,47,40);
    x[6] = EXTRACT_BITS64(gr_r2,55,48);      y[6] = EXTRACT_BITS64(gr_r3,55,48);
    x[7] = EXTRACT_BITS64(gr_r2,63,56);      y[7] = EXTRACT_BITS64(gr_r3,63,56);

    if (raz_form) {
      for (i = 0; i < 8; i++) {
	temp[i] = ZERO_EXT(x[i], 8) + ZERO_EXT(y[i], 8) + 1;
	res[i] = shift_right_unsigned(temp[i], 1);
      }
    } else {                                           // normal form
      for (i = 0; i < 8; i++) {
	temp[i] = ZERO_EXT(x[i], 8) + ZERO_EXT(y[i], 8);
	res[i] = shift_right_unsigned(temp[i], 1) | (EXTRACT_BIT64(temp[i],0));
      }
    }
    gr_r1 = concatenate8(res[7], res[6], res[5], res[4],
	res[3], res[2], res[1], res[0]);


  } else {                                                // two_byte_form
    x[0] = EXTRACT_BITS64(gr_r2,15,0);       y[0] = EXTRACT_BITS64(gr_r3,15,0);
    x[1] = EXTRACT_BITS64(gr_r2,31,16);      y[1] = EXTRACT_BITS64(gr_r3,31,16);
    x[2] = EXTRACT_BITS64(gr_r2,47,32);      y[2] = EXTRACT_BITS64(gr_r3,47,32);
    x[3] = EXTRACT_BITS64(gr_r2,63,48);      y[3] = EXTRACT_BITS64(gr_r3,63,48);

    if (raz_form) {
      for (i = 0; i < 4; i++) {
	temp[i] = ZERO_EXT(x[i], 16) + ZERO_EXT(y[i], 16) + 1;
	res[i] = shift_right_unsigned(temp[i], 1);
      }
    } else {                                           // normal form
      for (i = 0; i < 4; i++) {
	temp[i] = ZERO_EXT(x[i], 16) + ZERO_EXT(y[i], 16);
	res[i] = shift_right_unsigned(temp[i], 1) | (EXTRACT_BIT64(temp[i],0));
      }
    }
    gr_r1 = concatenate4(res[3], res[2], res[1], res[0]);
  }
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */
  
  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  
  
  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif






/*
Parallel Average Subtract

Format:         (qp) pavgsub1 r1 = r2, r3                       one_byte_form           A9
                (qp) pavgsub2 r1 = r2, r3                       two_byte_form           A9
*/
/*status: TODO */
static void Ia64Emulate_pavgsub(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[8], y[8], temp[8], res[8];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint8 i;

  t_bool one_byte_form =  GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==0;
  
  Ia64InsEmulDetermineTag(ins, state);

  /* Extract values from procstate */
  
  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  /* perform operations */

  check_target_register(r1);

  if (one_byte_form) {
    x[0] = EXTRACT_BITS64(gr_r2,7,0);       y[0] = EXTRACT_BITS64(gr_r3,7,0);
    x[1] = EXTRACT_BITS64(gr_r2,15,8);      y[1] = EXTRACT_BITS64(gr_r3,15,8);
    x[2] = EXTRACT_BITS64(gr_r2,23,16);     y[2] = EXTRACT_BITS64(gr_r3,23,16);
    x[3] = EXTRACT_BITS64(gr_r2,31,24);     y[3] = EXTRACT_BITS64(gr_r3,31,24);
    x[4] = EXTRACT_BITS64(gr_r2,39,32);     y[4] = EXTRACT_BITS64(gr_r3,39,32);
    x[5] = EXTRACT_BITS64(gr_r2,47,40);     y[5] = EXTRACT_BITS64(gr_r3,47,40);
    x[6] = EXTRACT_BITS64(gr_r2,55,48);     y[6] = EXTRACT_BITS64(gr_r3,55,48);
    x[7] = EXTRACT_BITS64(gr_r2,63,56);     y[7] = EXTRACT_BITS64(gr_r3,63,56);

    for (i = 0; i < 8; i++) {
      temp[i] = ZERO_EXT(x[i], 8) - ZERO_EXT(y[i], 8);
      res[i] = (EXTRACT_BITS64(temp[i],8,0)>> 1) | (EXTRACT_BIT64(temp[i],0));
    }
    gr_r1 = concatenate8(res[7], res[6], res[5], res[4],
	res[3], res[2], res[1], res[0]);


  } else {                                               // two_byte_form
    x[0] = EXTRACT_BITS64(gr_r2,15,0);      y[0] = EXTRACT_BITS64(gr_r3,15,0);
    x[1] = EXTRACT_BITS64(gr_r2,31,16);     y[1] = EXTRACT_BITS64(gr_r3,31,16);
    x[2] = EXTRACT_BITS64(gr_r2,47,32);     y[2] = EXTRACT_BITS64(gr_r3,47,32);
    x[3] = EXTRACT_BITS64(gr_r2,63,48);     y[3] = EXTRACT_BITS64(gr_r3,63,48);

    for (i = 0; i < 4; i++) {
      temp[i] = ZERO_EXT(x[i], 16) - ZERO_EXT(y[i], 16);
      res[i] = (EXTRACT_BITS64(temp[i],16,0)>> 1) | (EXTRACT_BIT64(temp[i],0));
    }
    gr_r1 = concatenate4(res[3], res[2], res[1], res[0]);
  }
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif





/*
Parallel Compare

Format:         (qp) pcmp1.prel r1 = r2, r3                   one_byte_form                        A9
                (qp) pcmp2.prel r1 = r2, r3                   two_byte_form                        A9
                (qp) pcmp4.prel r1 = r2, r3              four_byte_form                            A9
*/
/*status: TODO */
static void Ia64Emulate_pcmp(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[8], y[8], res[8];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;
  
  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
  
  t_uint8 i;

  t_bool tmp_rel;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);
    
  t_bool eq_form = x2b == 0;
    

  t_bool one_byte_form = 0;
  t_bool two_byte_form = 0;
  t_bool four_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,36,1);
  if (!four_byte_form){
    one_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==0;
    two_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==1;
  }
  
  Ia64InsEmulDetermineTag(ins, state);

  /* Extract values from procstate */
  
  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);


  /* perform operations */
  
  check_target_register(r1);

  if (one_byte_form) {                                     // one-byte elements
       x[0] = EXTRACT_BITS64(gr_r2,7,0);         y[0] = EXTRACT_BITS64(gr_r3,7,0);
       x[1] = EXTRACT_BITS64(gr_r2,15,8);        y[1] = EXTRACT_BITS64(gr_r3,15,8);
       x[2] = EXTRACT_BITS64(gr_r2,23,16);       y[2] = EXTRACT_BITS64(gr_r3,23,16);
       x[3] = EXTRACT_BITS64(gr_r2,31,24);       y[3] = EXTRACT_BITS64(gr_r3,31,24);
       x[4] = EXTRACT_BITS64(gr_r2,39,32);       y[4] = EXTRACT_BITS64(gr_r3,39,32);
       x[5] = EXTRACT_BITS64(gr_r2,47,40);       y[5] = EXTRACT_BITS64(gr_r3,47,40);
       x[6] = EXTRACT_BITS64(gr_r2,55,48);       y[6] = EXTRACT_BITS64(gr_r3,55,48);
       x[7] = EXTRACT_BITS64(gr_r2,63,56);       y[7] = EXTRACT_BITS64(gr_r3,63,56);
       for (i = 0; i < 8; i++) {
            if (eq_form)
               tmp_rel = x[i] == y[i];
            else // 'gt'
               tmp_rel = greater_signed(SIGN_EXT(x[i], 8), SIGN_EXT(y[i], 8));
            if (tmp_rel)
               res[i] = 0xff;
            else
               res[i] = 0x00;
       }
       gr_r1 = concatenate8(res[7], res[6], res[5], res[4],
                                 res[3], res[2], res[1], res[0]);
  } else if (two_byte_form) {                              // two-byte elements
       x[0] = EXTRACT_BITS64(gr_r2,15,0);        y[0] = EXTRACT_BITS64(gr_r3,15,0);
       x[1] = EXTRACT_BITS64(gr_r2,31,16);       y[1] = EXTRACT_BITS64(gr_r3,31,16);
       x[2] = EXTRACT_BITS64(gr_r2,47,32);       y[2] = EXTRACT_BITS64(gr_r3,47,32);
       x[3] = EXTRACT_BITS64(gr_r2,63,48);       y[3] = EXTRACT_BITS64(gr_r3,63,48);
       for (i = 0; i < 4; i++) {
            if (eq_form)
               tmp_rel = x[i] == y[i];
            else // 'gt'
               tmp_rel = greater_signed(SIGN_EXT(x[i], 16), SIGN_EXT(y[i], 16));

            if (tmp_rel)
               res[i] = 0xffff;
            else
               res[i] = 0x0000;
       }
       gr_r1 = concatenate4(res[3], res[2], res[1], res[0]);
  } else {                                                 // four-byte elements
       x[0] = EXTRACT_BITS64(gr_r2,31,0);        y[0] = EXTRACT_BITS64(gr_r3,31,0);
       x[1] = EXTRACT_BITS64(gr_r2,63,32);       y[1] = EXTRACT_BITS64(gr_r3,63,32);
       for (i = 0; i < 2; i++) {
            if (eq_form)
               tmp_rel = x[i] == y[i];
            else // 'gt'
               tmp_rel = greater_signed(SIGN_EXT(x[i], 32), SIGN_EXT(y[i], 32));

            if (tmp_rel)
               res[i] = 0xffffffff;
            else
               res[i] = 0x00000000;
       }
       gr_r1 = concatenate2(res[1], res[0]);
  }
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif





/*
Parallel Maximum

Format:         (qp) pmax1.u r1 = r2, r3 one_byte_form                            I2
                (qp) pmax2 r1 = r2, r3              two_byte_form                 I2
*/
/*status: TODO */
static void Ia64Emulate_pmax(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[8], y[8], res[8];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
  
  t_uint8 i;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);
    
  t_bool one_byte_form = x2b == 1;
    
  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  
  
  
  /* perform operations */

  check_target_register(r1);

  if (one_byte_form) {         // one-byte elements
          x[0] = EXTRACT_BITS64(gr_r2,7,0); y[0] = EXTRACT_BITS64(gr_r3,7,0);
          x[1] = EXTRACT_BITS64(gr_r2,15,8); y[1] = EXTRACT_BITS64(gr_r3,15,8);
          x[2] = EXTRACT_BITS64(gr_r2,23,16); y[2] = EXTRACT_BITS64(gr_r3,23,16);
          x[3] = EXTRACT_BITS64(gr_r2,31,24); y[3] = EXTRACT_BITS64(gr_r3,31,24);
          x[4] = EXTRACT_BITS64(gr_r2,39,32); y[4] = EXTRACT_BITS64(gr_r3,39,32);
          x[5] = EXTRACT_BITS64(gr_r2,47,40); y[5] = EXTRACT_BITS64(gr_r3,47,40);
          x[6] = EXTRACT_BITS64(gr_r2,55,48); y[6] = EXTRACT_BITS64(gr_r3,55,48);
          x[7] = EXTRACT_BITS64(gr_r2,63,56); y[7] = EXTRACT_BITS64(gr_r3,63,56);
          for (i = 0; i < 8; i++) {
                    res[i] = (ZERO_EXT(x[i],8) < ZERO_EXT(y[i],8)) ? y[i] : x[i];
          }
          gr_r1 = concatenate8(res[7], res[6], res[5], res[4], res[3], res[2], res[1], res[0]);
  } else {  // two-byte elements
          x[0] = EXTRACT_BITS64(gr_r2,15,0); y[0] = EXTRACT_BITS64(gr_r3,15,0);
          x[1] = EXTRACT_BITS64(gr_r2,31,16); y[1] = EXTRACT_BITS64(gr_r3,31,16);
          x[2] = EXTRACT_BITS64(gr_r2,47,32); y[2] = EXTRACT_BITS64(gr_r3,47,32);
          x[3] = EXTRACT_BITS64(gr_r2,63,48); y[3] = EXTRACT_BITS64(gr_r3,63,48);
          for (i = 0; i < 4; i++) {
                    res[i] = (SIGN_EXT(x[i],16) < SIGN_EXT(y[i],16)) ? y[i] : x[i];
          }
          gr_r1 = concatenate4(res[3], res[2], res[1], res[0]);
  }
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

}


#endif





/*
Parallel Minimum

Format:         (qp) pmin1.u r1 = r2, r3 one_byte_form                         I2
                (qp) pmin2 r1 = r2, r3              two_byte_form              I2
*/
/*status: TODO */
static void Ia64Emulate_pmin(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[8], y[8], res[8];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
  
  t_uint8 i;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);
    
  t_bool one_byte_form = x2b == 1;
    
  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  /* perform operations */
  
  check_target_register(r1);
  
  if (one_byte_form) {                       // one-byte elements
          x[0] = EXTRACT_BITS64(gr_r2,7,0); y[0] = EXTRACT_BITS64(gr_r3,7,0);
          x[1] = EXTRACT_BITS64(gr_r2,15,8); y[1] = EXTRACT_BITS64(gr_r3,15,8);
          x[2] = EXTRACT_BITS64(gr_r2,23,16); y[2] = EXTRACT_BITS64(gr_r3,23,16);
          x[3] = EXTRACT_BITS64(gr_r2,31,24); y[3] = EXTRACT_BITS64(gr_r3,31,24);
          x[4] = EXTRACT_BITS64(gr_r2,39,32); y[4] = EXTRACT_BITS64(gr_r3,39,32);
          x[5] = EXTRACT_BITS64(gr_r2,47,40); y[5] = EXTRACT_BITS64(gr_r3,47,40);
          x[6] = EXTRACT_BITS64(gr_r2,55,48); y[6] = EXTRACT_BITS64(gr_r3,55,48);
          x[7] = EXTRACT_BITS64(gr_r2,63,56); y[7] = EXTRACT_BITS64(gr_r3,63,56);
          for (i = 0; i < 8; i++) {
                    res[i] = (ZERO_EXT(x[i],8) < ZERO_EXT(y[i],8)) ? x[i] : y[i];
          }
          gr_r1 = concatenate8(res[7], res[6], res[5], res[4], res[3], res[2], res[1], res[0]);
  } else {                                   // two-byte elements
          x[0] = EXTRACT_BITS64(gr_r2,15,0); y[0] = EXTRACT_BITS64(gr_r3,15,0);
          x[1] = EXTRACT_BITS64(gr_r2,31,16); y[1] = EXTRACT_BITS64(gr_r3,31,16);
          x[2] = EXTRACT_BITS64(gr_r2,47,32); y[2] = EXTRACT_BITS64(gr_r3,47,32);
          x[3] = EXTRACT_BITS64(gr_r2,63,48); y[3] = EXTRACT_BITS64(gr_r3,63,48);
          for (i = 0; i < 4; i++) {
                    res[i] = (SIGN_EXT(x[i],16) < SIGN_EXT(y[i],16)) ? x[i] : y[i];
          }
          gr_r1 = concatenate4(res[3], res[2], res[1], res[0]);
  }
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

}
#endif




/*
Parallel Multiply

Format:         (qp) pmpy2.r r1 = r2, r3                      right_form                I2
                (qp) pmpy2.l r1 = r2, r3                           left_form            I2
*/
/*status: TODO */
static void Ia64Emulate_pmpy2(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);

  t_bool right_form = x2b == 1;

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);




  /* perform operations */

  check_target_register(r1);
  
  if (right_form) {
    SET_BITS64(gr_r1,31,0,SIGN_EXT(EXTRACT_BITS64(gr_r2,15,0), 16) * SIGN_EXT(EXTRACT_BITS64(gr_r3,15,0), 16));
    SET_BITS64(gr_r1,63,32,SIGN_EXT(EXTRACT_BITS64(gr_r2,47,32), 16) * SIGN_EXT(EXTRACT_BITS64(gr_r3,47,32), 16));
  } else {                                                         // left_form
    SET_BITS64(gr_r1,31,0,SIGN_EXT(EXTRACT_BITS64(gr_r2,31,16), 16) * SIGN_EXT(EXTRACT_BITS64(gr_r3,31,16), 16));
    SET_BITS64(gr_r1,63,32,SIGN_EXT(EXTRACT_BITS64(gr_r2,63,48), 16) * SIGN_EXT(EXTRACT_BITS64(gr_r3,63,48), 16));
  }


  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
 
}

#endif






/*
Parallel Multiply and Shift Right

Format:         (qp) pmpyshr2 r1 = r2, r3, count2                signed_form        I1
                (qp) pmpyshr2.u r1 = r2, r3, count2           unsigned_form         I1
*/
/*status: TODO */
static void Ia64Emulate_pmpyshr2(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[4], y[4], temp[4], res[4];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint8 i;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);

  t_bool unsigned_form = x2b == 1;

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);




  /* perform operations */

  check_target_register(r1);
  x[0] = EXTRACT_BITS64(gr_r2,15,0);       y[0] = EXTRACT_BITS64(gr_r3,15,0);
  x[1] = EXTRACT_BITS64(gr_r2,31,16);      y[1] = EXTRACT_BITS64(gr_r3,31,16);
  x[2] = EXTRACT_BITS64(gr_r2,47,32);      y[2] = EXTRACT_BITS64(gr_r3,47,32);
  x[3] = EXTRACT_BITS64(gr_r2,63,48);      y[3] = EXTRACT_BITS64(gr_r3,63,48);
  for (i = 0; i < 4; i++) {
    if (unsigned_form)                            // unsigned multiplication
      temp[i] = ZERO_EXT(x[i], 16) * ZERO_EXT(y[i], 16);
    else                                          // signed multiplication
      temp[i] = SIGN_EXT(x[i], 16) * SIGN_EXT(y[i], 16);


    res[i] = EXTRACT_BITS64(temp[i],(IA64INSIMM1(ins) + 15),IA64INSIMM1(ins));
  }


  gr_r1 = concatenate4(res[3], res[2], res[1], res[0]);
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

}


#endif





/*
Population Count

Format:         (qp) popcnt r1 = r3                                            I9
*/
/*status: TODO */
static void Ia64Emulate_popcnt(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  t_int64 gr_r1, gr_r3;
  t_uint64 gr_r1_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_int64 res;

  t_uint8 i;
  
  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);


  /* perform operations */

  check_target_register(r1);
  
  res = 0;
  // Count up all the one bits
  for (i = 0; i < 64; i++) {
      res += EXTRACT_BIT64(gr_r3,i);
  }
  
  
  gr_r1 = res;
  gr_r1_nat = gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

}
#endif








/*
Probe Access

Format:         (qp) probe.r r1 = r3, r2 read_form,                                register_form         M38
                (qp) probe.w r1 = r3, r2                              write_form, register_form          M38
                (qp) probe.r r1 = r3, imm2 read_form,                           immediate_form           M39
                (qp) probe.w r1 = r3, imm2                         write_form, immediate_form            M39
                (qp) probe.r.fault r3, imm2            fault_form, read_form, immediate_form             M40
                (qp) probe.w.fault r3, imm2           fault_form, write_form, immediate_form             M40
                (qp) probe.rw.fault r3, imm2    fault_form, read_write_form, immediate_form              M40
*/

static void Ia64Emulate_probe(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  itype = NON_ACCESS;
  itype |= (read_write_form) ? READ|WRITE : ((write_form) ? WRITE : READ);
  itype |= (fault_form) ? PROBE_FAULT : PROBE;


  if (!fault_form)
       check_target_register(r1);

  if (gr_r3_nat || (register_form ? gr_r2_nat : 0))
       register_nat_consumption_fault(itype);


  tmp_pl = (register_form) ? EXTRACT_BITS64(gr_r2,1,0) : imm2;
  if (tmp_pl < PSR.cpl)
       tmp_pl = PSR.cpl;


  if (fault_form) {
       tlb_translate(gr_r3, 1, itype, tmp_pl, &mattr, &defer);
  } else {
       gr_r1 = tlb_grant_permission(gr_r3, itype, tmp_pl);
       gr_r1_nat = 0;
  }
}

#endif







/*
Parallel Sum of Absolute Difference

Format:         (qp) psad1 r1 = r2, r3                         I2
*/
/*status: TODO */
static void Ia64Emulate_psad1(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  static t_int64 x[8], y[8], temp[8];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
  
  t_uint8 i;

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  /* perform operations */

  
  check_target_register(r1);
  
  x[0] = EXTRACT_BITS64(gr_r2,7,0);            y[0] = EXTRACT_BITS64(gr_r3,7,0);
  x[1] = EXTRACT_BITS64(gr_r2,15,8);           y[1] = EXTRACT_BITS64(gr_r3,15,8);
  x[2] = EXTRACT_BITS64(gr_r2,23,16);          y[2] = EXTRACT_BITS64(gr_r3,23,16);
  x[3] = EXTRACT_BITS64(gr_r2,31,24);          y[3] = EXTRACT_BITS64(gr_r3,31,24);
  x[4] = EXTRACT_BITS64(gr_r2,39,32);          y[4] = EXTRACT_BITS64(gr_r3,39,32);
  x[5] = EXTRACT_BITS64(gr_r2,47,40);          y[5] = EXTRACT_BITS64(gr_r3,47,40);
  x[6] = EXTRACT_BITS64(gr_r2,55,48);          y[6] = EXTRACT_BITS64(gr_r3,55,48);
  x[7] = EXTRACT_BITS64(gr_r2,63,56);          y[7] = EXTRACT_BITS64(gr_r3,63,56);
  
  gr_r1 = 0;
  for (i = 0; i < 8; i++) {
      temp[i] = ZERO_EXT(x[i], 8) - ZERO_EXT(y[i], 8);
      if (temp[i] < 0)
         temp[i] = -temp[i];
      gr_r1 += temp[i];
  }
  
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

}

#endif





/*
Parallel Shift Left

Format:         (qp) pshl2 r1 = r2, r3           two_byte_form, variable_form       I7
                (qp) pshl2 r1 = r2, count5          two_byte_form, fixed_form       I8
                (qp) pshl4 r1 = r2, r3           four_byte_form, variable_form      I7
                (qp) pshl4 r1 = r2, count5         four_byte_form, fixed_form       I8
*/
/*status: TODO */
static void Ia64Emulate_pshl(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
 
  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);
    
  t_bool variable_form = x2b == 0;

  t_bool two_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==1;
 
  t_uint8 shift_count;

  t_uint64 tmp_nat;

  Ia64InsEmulDetermineTag(ins, state);
  
  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if(variable_form)
  {
    if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r3 = (t_int64) G_T_UINT64(reg.i);

    if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r3_nat= G_T_UINT64(reg.i);
  }
  
  /* perform operations */

  check_target_register(r1);

  shift_count = (variable_form ? gr_r3 : IA64INSIMM1(ins));
  tmp_nat = (variable_form ? gr_r3_nat : 0);

  if (two_byte_form) {                                                 // two_byte_form
       if (shift_count> 16)
           shift_count = 16;
       SET_BITS64(gr_r1,15,0,EXTRACT_BITS64(gr_r2,15,0) << shift_count);
       SET_BITS64(gr_r1,31,16,EXTRACT_BITS64(gr_r2,31,16) << shift_count);
       SET_BITS64(gr_r1,47,32,EXTRACT_BITS64(gr_r2,47,32) << shift_count);
       SET_BITS64(gr_r1,63,48,EXTRACT_BITS64(gr_r2,63,48) << shift_count);
  } else {                                                             // four_byte_form
       if (shift_count> 32)
           shift_count = 32;
       SET_BITS64(gr_r1,31,0,EXTRACT_BITS64(gr_r2,31,0) << shift_count);
       SET_BITS64(gr_r1,63,32,EXTRACT_BITS64(gr_r2,63,32) << shift_count);
  }


  gr_r1_nat = gr_r2_nat || tmp_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

}

#endif






/*
Parallel Shift Left and Add

Format:         (qp) pshladd2 r1 = r2, count2, r3                                    A10
*/
/*status: TODO */
static void Ia64Emulate_pshladd2(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  static t_int64 x[4], y[4], temp[4], res[4];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint8 i;

  t_int64 max, min;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* Extract values from procstate */
  
  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);


  /* perform operations */

  check_target_register(r1);

  x[0] = EXTRACT_BITS64(gr_r2,15,0);              y[0] = EXTRACT_BITS64(gr_r3,15,0);
  x[1] = EXTRACT_BITS64(gr_r2,31,16);             y[1] = EXTRACT_BITS64(gr_r3,31,16);
  x[2] = EXTRACT_BITS64(gr_r2,47,32);             y[2] = EXTRACT_BITS64(gr_r3,47,32);
  x[3] = EXTRACT_BITS64(gr_r2,63,48);             y[3] = EXTRACT_BITS64(gr_r3,63,48);

  max = SIGN_EXT(0x7fff, 16);
  min = SIGN_EXT(0x8000, 16);


  for (i = 0; i < 4; i++) {
    temp[i] = SIGN_EXT(x[i], 16) << IA64INSIMM1(ins);

    if (temp[i] > max)
      res[i] = max;
    else if (temp[i] < min)
      res[i] = min;
    else {
      res[i] = temp[i] + SIGN_EXT(y[i], 16);
      if (res[i] > max)
	res[i] = max;
      if (res[i] < min)
	res[i] = min;
    }
  }


  gr_r1 = concatenate4(res[3], res[2], res[1], res[0]);
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif






/*
Parallel Shift Right

Format:         (qp) pshr2 r1 = r3, r2              signed_form, two_byte_form, variable_form                  I5
                (qp) pshr2 r1 = r3, count5             signed_form, two_byte_form, fixed_form                  I6
                (qp) pshr2.u r1 = r3, r2         unsigned_form, two_byte_form, variable_form                   I5
                (qp) pshr2.u r1 = r3, count5        unsigned_form, two_byte_form, fixed_form                   I6
                (qp) pshr4 r1 = r3, r2             signed_form, four_byte_form, variable_form                  I5
                (qp) pshr4 r1 = r3, count5             signed_form, four_byte_form, fixed_form                 I6
                (qp) pshr4.u r1 = r3, r2         unsigned_form, four_byte_form, variable_form                  I5
                (qp) pshr4.u r1 = r3, count5        unsigned_form, four_byte_form, fixed_form                  I6
*/
/*status: TODO */
static void Ia64Emulate_pshr(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  
  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);
    
  t_bool unsigned_form = (x2b == 0 || x2b == 1);
  t_bool variable_form = (x2b == 0 || x2b == 2);

  t_bool two_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==1;

  t_uint8 shift_count;
  t_uint64 tmp_nat;


  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  if(variable_form)
  {
    if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r2 = (t_int64) G_T_UINT64(reg.i);

    if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
    {
      ProcStateSetAllBot(state,INS_REGS_DEF(ins));
      return;
    }
    gr_r2_nat= G_T_UINT64(reg.i);
  }
  
  /* perform operations */
  
  check_target_register(r1);

  shift_count = (variable_form ? gr_r2 : IA64INSIMM1(ins));
  tmp_nat = (variable_form ? gr_r2_nat : 0);

  if (two_byte_form) {                                // two_byte_form
       if (shift_count> 16)
            shift_count = 16;
       if (unsigned_form) {                           // unsigned shift
            SET_BITS64(gr_r1,15,0,shift_right_unsigned(ZERO_EXT(EXTRACT_BITS64(gr_r3,15,0), 16),shift_count));
            SET_BITS64(gr_r1,31,16,shift_right_unsigned(ZERO_EXT(EXTRACT_BITS64(gr_r3,31,16), 16),shift_count));
            SET_BITS64(gr_r1,47,32,shift_right_unsigned(ZERO_EXT(EXTRACT_BITS64(gr_r3,47,32), 16),shift_count));
            SET_BITS64(gr_r1,63,48,shift_right_unsigned(ZERO_EXT(EXTRACT_BITS64(gr_r3,63,48), 16),shift_count));
       } else {                                       // signed shift
            SET_BITS64(gr_r1,15,0,shift_right_signed(SIGN_EXT(EXTRACT_BITS64(gr_r3,15,0), 16),shift_count));
            SET_BITS64(gr_r1,31,16,shift_right_signed(SIGN_EXT(EXTRACT_BITS64(gr_r3,31,16), 16),shift_count));
            SET_BITS64(gr_r1,47,32,shift_right_signed(SIGN_EXT(EXTRACT_BITS64(gr_r3,47,32), 16),shift_count));
            SET_BITS64(gr_r1,63,48,shift_right_signed(SIGN_EXT(EXTRACT_BITS64(gr_r3,63,48), 16),shift_count));
       }
  } else {                                            // four_byte_form
       if (shift_count > 32)
            shift_count = 32;
       if (unsigned_form) {                           // unsigned shift
            SET_BITS64(gr_r1,31,0,shift_right_unsigned(ZERO_EXT(EXTRACT_BITS64(gr_r3,31,0), 32),shift_count));
            SET_BITS64(gr_r1,63,32,shift_right_unsigned(ZERO_EXT(EXTRACT_BITS64(gr_r3,63,32), 32),shift_count));
       } else {                                       // signed shift
            SET_BITS64(gr_r1,31,0,shift_right_signed(SIGN_EXT(EXTRACT_BITS64(gr_r3,31,0), 32),shift_count));
            SET_BITS64(gr_r1,63,32,shift_right_signed(SIGN_EXT(EXTRACT_BITS64(gr_r3,63,32), 32),shift_count));
       }
  }


  gr_r1_nat = gr_r3_nat || tmp_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif






/*
Parallel Shift Right and Add

Format:         (qp) pshradd2 r1 = r2, count2, r3                               A10
*/
/*status: TODO */
static void Ia64Emulate_pshradd2(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  static t_int64 x[4], y[4], temp[4], res[4];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint8 i;

  t_int64 max, min;

  Ia64InsEmulDetermineTag(ins, state);

  /* Extract values from procstate */
  
  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);


  /* perform operations */

  check_target_register(r1);

  x[0] = EXTRACT_BITS64(gr_r2,15,0);            y[0] = EXTRACT_BITS64(gr_r3,15,0);
  x[1] = EXTRACT_BITS64(gr_r2,31,16);           y[1] = EXTRACT_BITS64(gr_r3,31,16);
  x[2] = EXTRACT_BITS64(gr_r2,47,32);           y[2] = EXTRACT_BITS64(gr_r3,47,32);
  x[3] = EXTRACT_BITS64(gr_r2,63,48);           y[3] = EXTRACT_BITS64(gr_r3,63,48);

  max = SIGN_EXT(0x7fff, 16);
  min = SIGN_EXT(0x8000, 16);


  for (i = 0; i < 4; i++) {
    temp[i] = shift_right_signed(SIGN_EXT(x[i], 16), IA64INSIMM1(ins));

    res[i] = temp[i] + SIGN_EXT(y[i], 16);
    if (res[i] > max)
      res[i] = max;
    if (res[i] < min)
      res[i] = min;
  }


  gr_r1 = concatenate4(res[3], res[2], res[1], res[0]);
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif







/*
Parallel Subtract

Format:         (qp) psub1 r1 = r2, r3                                 one_byte_form, modulo_form                         A9
                (qp) psub1.sss r1 = r2, r3                 one_byte_form, sss_saturation_form                             A9
                (qp) psub1.uus r1 = r2, r3                one_byte_form, uus_saturation_form                              A9
                (qp) psub1.uuu r1 = r2, r3                one_byte_form, uuu_saturation_form                              A9
                (qp) psub2 r1 = r2, r3                                 two_byte_form, modulo_form                         A9
                (qp) psub2.sss r1 = r2, r3                 two_byte_form, sss_saturation_form                             A9
                (qp) psub2.uus r1 = r2, r3                two_byte_form, uus_saturation_form                              A9
                (qp) psub2.uuu r1 = r2, r3                two_byte_form, uuu_saturation_form                              A9
                (qp) psub4 r1 = r2, r3                                 four_byte_form, modulo_form                        A9
*/
/*status: TODO */
static void Ia64Emulate_psub(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  static t_int64 x[8], y[8], temp[8];


  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;
  t_int64 gr_r1, gr_r2, gr_r3;
  
  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_int64 max, min;

  t_uint8 i;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);

  t_bool sss_saturation_form = x2b ==1;
  t_bool uuu_saturation_form = x2b ==2;
  t_bool uus_saturation_form = x2b ==3;

  t_bool one_byte_form = 0;
  t_bool two_byte_form = 0;
  t_bool four_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,36,1);
  if (!four_byte_form){
    one_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==0;
    two_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==1;
  }

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64)G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64)G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  /* perform operations */
  
  check_target_register(r1);

  if (one_byte_form) {                                   // one-byte elements
    x[0] = EXTRACT_BITS64(gr_r2,7,0);         y[0] = EXTRACT_BITS64(gr_r3,7,0);
    x[1] = EXTRACT_BITS64(gr_r2,15,8);        y[1] = EXTRACT_BITS64(gr_r3,15,8);
    x[2] = EXTRACT_BITS64(gr_r2,23,16);       y[2] = EXTRACT_BITS64(gr_r3,23,16);
    x[3] = EXTRACT_BITS64(gr_r2,31,24);       y[3] = EXTRACT_BITS64(gr_r3,31,24);
    x[4] = EXTRACT_BITS64(gr_r2,39,32);       y[4] = EXTRACT_BITS64(gr_r3,39,32);
    x[5] = EXTRACT_BITS64(gr_r2,47,40);       y[5] = EXTRACT_BITS64(gr_r3,47,40);
    x[6] = EXTRACT_BITS64(gr_r2,55,48);       y[6] = EXTRACT_BITS64(gr_r3,55,48);
    x[7] = EXTRACT_BITS64(gr_r2,63,56);       y[7] = EXTRACT_BITS64(gr_r3,63,56);

    if (sss_saturation_form) {                          // sss_saturation_form
      max = SIGN_EXT(0x7f, 8);
      min = SIGN_EXT(0x80, 8);
      for (i = 0; i < 8; i++) {
	temp[i] = SIGN_EXT(x[i], 8) - SIGN_EXT(y[i], 8);
      }
    } else if (uus_saturation_form) {                   // uus_saturation_form
      max = 0xff;
      min = 0x00;
      for (i = 0; i < 8; i++) {
	temp[i] = ZERO_EXT(x[i], 8) - SIGN_EXT(y[i], 8);
      }
    } else if (uuu_saturation_form) {                   // uuu_saturation_form
      max = 0xff;
      min = 0x00;
      for (i = 0; i < 8; i++) {
	temp[i] = ZERO_EXT(x[i], 8) - ZERO_EXT(y[i], 8);
      }
    } else {                                            // modulo_form
      for (i = 0; i < 8; i++) {
	temp[i] = ZERO_EXT(x[i], 8) - ZERO_EXT(y[i], 8);
      }
    }


    if (sss_saturation_form || uus_saturation_form ||
	uuu_saturation_form) {
      for (i = 0; i < 8; i++) {
	if (temp[i] > max)
	  temp[i] = max;
	if (temp[i] < min)
	  temp[i] = min;
      }
    }


    gr_r1 = concatenate8(temp[7], temp[6], temp[5], temp[4],
	temp[3], temp[2], temp[1], temp[0]);
  } else if (two_byte_form) {                            // two-byte elements
    x[0] = EXTRACT_BITS64(gr_r2,15,0);        y[0] = EXTRACT_BITS64(gr_r3,15,0);
    x[1] = EXTRACT_BITS64(gr_r2,31,16);       y[1] = EXTRACT_BITS64(gr_r3,31,16);
    x[2] = EXTRACT_BITS64(gr_r2,47,32);       y[2] = EXTRACT_BITS64(gr_r3,47,32);
    x[3] = EXTRACT_BITS64(gr_r2,63,48);       y[3] = EXTRACT_BITS64(gr_r3,63,48);

    if (sss_saturation_form) {                          // sss_saturation_form
      max = SIGN_EXT(0x7fff, 16);
      min = SIGN_EXT(0x8000, 16);
      for (i = 0; i < 4; i++) {
	temp[i] = SIGN_EXT(x[i], 16) - SIGN_EXT(y[i], 16);
      }
    } else if (uus_saturation_form) {                   // uus_saturation_form
      max = 0xffff;
      min = 0x0000;
      for (i = 0; i < 4; i++) {
	temp[i] = ZERO_EXT(x[i], 16) - SIGN_EXT(y[i], 16);
      }
    } else if (uuu_saturation_form) {                     // uuu_saturation_form
      max = 0xffff;
      min = 0x0000;
      for (i = 0; i < 4; i++) {
	temp[i] = ZERO_EXT(x[i], 16) - ZERO_EXT(y[i], 16);
      }
    } else {                                              // modulo_form
      for (i = 0; i < 4; i++) {
	temp[i] = ZERO_EXT(x[i], 16) - ZERO_EXT(y[i], 16);
      }
    }


    if (sss_saturation_form || uus_saturation_form ||
	uuu_saturation_form) {
      for (i = 0; i < 4; i++) {
	if (temp[i] > max)
	  temp[i] = max;
	if (temp[i] < min)
	  temp[i] = min;
      }
    }


    gr_r1 = concatenate4(temp[3], temp[2], temp[1], temp[0]);
  } else {                                                   // four-byte elements
    x[0] = EXTRACT_BITS64(gr_r2,31,0);          y[0] = EXTRACT_BITS64(gr_r3,31,0);
    x[1] = EXTRACT_BITS64(gr_r2,63,32);         y[1] = EXTRACT_BITS64(gr_r3,63,32);

    for (i = 0; i < 2; i++) {                             // modulo_form
      temp[i] = ZERO_EXT(x[i], 32) - ZERO_EXT(y[i], 32);
    }


    gr_r1 = concatenate2(temp[1], temp[0]);
  }


  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 
}

#endif







/*
Purge Translation Cache Entry

Format:           (qp) ptc.e r3                                                   M28
*/

static void Ia64Emulate_ptc_e(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (PSR.cpl != 0)
     privileged_operation_fault(0);
  if (gr_r3_nat)
     register_nat_consumption_fault(0);
  tlb_purge_translation_cache(gr_r3);
}

#endif








/*
Purge Global Translation Cache

Format:         (qp) ptc.g r3, r2                             global_form       M45
                (qp) ptc.ga r3, r2                       global_alat_form       M45
*/

static void Ia64Emulate_ptc_g(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (!followed_by_stop())
      undefined_behavior();
  if (PSR.cpl != 0)
      privileged_operation_fault(0);
  if (gr_r3_nat || gr_r2_nat)
      register_nat_consumption_fault(0);
  if (unimplemented_virtual_address(gr_r3))
      unimplemented_data_address_fault(0);
  
  tmp_rid = RR[EXTRACT_BITS64(gr_r3,63,61)].rid;
  tmp_va = EXTRACT_BITS64(gr_r3,60,0);
  tmp_size = EXTRACT_BITS64(gr_r2,7,2);
  tmp_va = align_to_size_boundary(tmp_va, tmp_size);
  tlb_must_purge_dtc_entries(tmp_rid, tmp_va, tmp_size);
  tlb_must_purge_itc_entries(tmp_rid, tmp_va, tmp_size);
  
  if (global_alat_form) tmp_ptc_type = GLOBAL_ALAT_FORM;
  else tmp_ptc_type = GLOBAL_FORM;
  
  tlb_broadcast_purge(tmp_rid, tmp_va, tmp_size, tmp_ptc_type);
}

#endif








/*
Purge Local Translation Cache

Format:           (qp) ptc.l r3, r2                                               M45
*/

static void Ia64Emulate_ptc_l(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (PSR.cpl != 0)
      privileged_operation_fault(0);
  if (gr_r3_nat || gr_r2_nat)
      register_nat_consumption_fault(0);
  if (unimplemented_virtual_address(gr_r3))
      unimplemented_data_address_fault(0);
  
  
  tmp_rid = RR[EXTRACT_BITS64(gr_r3,63,61)].rid;
  tmp_va = EXTRACT_BITS64(gr_r3,60,0);
  tmp_size = EXTRACT_BITS64(gr_r2,7,2);
  tmp_va = align_to_size_boundary(tmp_va, tmp_size);
  tlb_must_purge_dtc_entries(tmp_rid, tmp_va, tmp_size);
  tlb_must_purge_itc_entries(tmp_rid, tmp_va, tmp_size);
}

#endif








/*
Purge Translation Register

Format:         (qp) ptr.d r3, r2                                 data_form       M45
                (qp) ptr.i r3, r2                          instruction_form       M45
*/

static void Ia64Emulate_ptr(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (PSR.cpl != 0)
      privileged_operation_fault(0);
  if (gr_r3_nat || gr_r2_nat)
      register_nat_consumption_fault(0);
  if (unimplemented_virtual_address(gr_r3))
      unimplemented_data_address_fault(0);
  
  
  tmp_rid = RR[EXTRACT_BITS64(gr_r3,63,61)].rid;
  tmp_va = EXTRACT_BITS64(gr_r3,60,0);
  tmp_size = EXTRACT_BITS64(gr_r2,7,2);
  tmp_va = align_to_size_boundary(tmp_va, tmp_size);
  
  
  if (data_form) {
      tlb_must_purge_dtr_entries(tmp_rid, tmp_va, tmp_size);
      tlb_must_purge_dtc_entries(tmp_rid, tmp_va, tmp_size);
      tlb_may_purge_itc_entries(tmp_rid, tmp_va, tmp_size);
  } else {                                                        // instruction_form
      tlb_must_purge_itr_entries(tmp_rid, tmp_va, tmp_size);
      tlb_must_purge_itc_entries(tmp_rid, tmp_va, tmp_size);
      tlb_may_purge_dtc_entries(tmp_rid, tmp_va, tmp_size);
  }
}

#endif








/*
Return From Interruption

Format:         rfi                                                                B8
*/

static void Ia64Emulate_rfi(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (!followed_by_stop())
      illegal_operation_fault();
  
  
  unimplemented_address = 0;
  if (PSR.cpl != 0)
      privileged_operation_fault(0);
  
  
  taken_rfi = 1;
  
  
  PSR = CR[IPSR];
  if (CR[IPSR].is == 1) {                  //resume IA-32 instruction set
      tmp_IP = CR[IIP];
      if ((CR[IPSR].it && unimplemented_virtual_address(tmp_IP))
           || (!CR[IPSR].it && unimplemented_physical_address(tmp_IP)))
           unimplemented_address = 1;
                                          //compute effective instruction pointer
      SET_BITS64(EIP,31,0,EXTRACT_BITS64(CR[IIP],31,0) - AR[CSD].Base);
                                          //force zero-sized restored frame
      rse_restore_frame(0, 0, CFM.sof);
      CFM.sof = 0;
      CFM.sol = 0;
      CFM.sor = 0;
      CFM.rrb.gr = 0;
      CFM.rrb.fr = 0;
      CFM.rrb.pr = 0;
      rse_invalidate_non_current_regs();
      //The register stack engine is disabled during IA-32
      //instruction set execution.
  } else {                                 //return to Itanium instruction set
      tmp_IP = CR[IIP] & ~0xf;
      slot = CR[IPSR].ri;
      if ((CR[IPSR].it && unimplemented_virtual_address(tmp_IP))
           || (!CR[IPSR].it && unimplemented_physical_address(tmp_IP))) 
           unimplemented_address = 1;
      if (CR[IFS].v) {
           tmp_growth = -CFM.sof;
           alat_frame_update(-CR[IFS].ifm.sof, 0); 
           rse_restore_frame(CR[IFS].ifm.sof, tmp_growth, CFM.sof);
           CFM = CR[IFS].ifm;
      }
      rse_enable_current_frame_load();
  }
  IP = tmp_IP;
  instruction_serialize();
  if (unimplemented_address)
      unimplemented_instruction_address_trap(0, tmp_IP);
}

#endif








/*
Reset System Mask

Format:         (qp) rsm imm24                                                      M44
*/

static void Ia64Emulate_rsm(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (PSR.cpl != 0)
      privileged_operation_fault(0);
  
  
  if (is_reserved_field(PSR_TYPE, PSR_SM, imm24))
      reserved_register_field_fault();
  
  
  if (EXTRACT_BIT64(imm24,1))           SET_BIT64(PSR,1,0);           // be
  if (EXTRACT_BIT64(imm24,2))           SET_BIT64(PSR,2,0);           // up
  if (EXTRACT_BIT64(imm24,3))           SET_BIT64(PSR,3,0);           // ac
  if (EXTRACT_BIT64(imm24,4))           SET_BIT64(PSR,4,0);           // mfl
  if (EXTRACT_BIT64(imm24,5))           SET_BIT64(PSR,5,0);           // mfh
  if (EXTRACT_BIT64(imm24,13))          SET_BIT64(PSR,13,0);          // ic
  if (EXTRACT_BIT64(imm24,14))          SET_BIT64(PSR,14,0);          // i
  if (EXTRACT_BIT64(imm24,15))          SET_BIT64(PSR,15,0);          // pk
  if (EXTRACT_BIT64(imm24,17))          SET_BIT64(PSR,17,0);          // dt
  if (EXTRACT_BIT64(imm24,18))          SET_BIT64(PSR,18,0);          // dfl
  if (EXTRACT_BIT64(imm24,19))          SET_BIT64(PSR,19,0);          // dfh
  if (EXTRACT_BIT64(imm24,20))          SET_BIT64(PSR,20,0);          // sp
  if (EXTRACT_BIT64(imm24,21))          SET_BIT64(PSR,21,0);          // pp
  if (EXTRACT_BIT64(imm24,22))          SET_BIT64(PSR,22,0);          // di
  if (EXTRACT_BIT64(imm24,23))          SET_BIT64(PSR,23,0);          // si
}

#endif









/*
Reset User Mask

Format:           (qp) rum imm24                                             M44
*/

static void Ia64Emulate_rum(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (is_reserved_field(PSR_TYPE, PSR_UM, imm24))
     reserved_register_field_fault();
  
  
  if (EXTRACT_BIT64(imm24,1))         SET_BIT64(PSR,1,0);          // be
  if (EXTRACT_BIT64(imm24,2) && PSR.sp == 0)                //non-secure perf monitor
                       SET_BIT64(PSR,2,0);          // up
  if (EXTRACT_BIT64(imm24,3))         SET_BIT64(PSR,3,0);          // ac
  if (EXTRACT_BIT64(imm24,4))         SET_BIT64(PSR,4,0);          // mfl
  if (EXTRACT_BIT64(imm24,5))         SET_BIT64(PSR,5,0);          // mfh
}

#endif








/*
Set Floating-point Value, Exponent, or Significand

Format:          (qp) setf.s f1 = r2                                single_form       M18
                 (qp) setf.d f1 = r2                               double_form        M18
                 (qp) setf.exp f1 = r2                            exponent_form       M18
                 (qp) setf.sig f1 = r2                       significand_form         M18
*/
/*status: TODO */

static void Ia64Emulate_setf(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  fp_check_target_register(f1);
  if (tmp_isrcode = fp_reg_disabled(f1, 0, 0, 0))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  
  if (!gr_r2_nat) {
      if (single_form)
            fr_f1 = fp_mem_to_fr_format(gr_r2, 4, 0);
      else if (double_form)
            fr_f1 = fp_mem_to_fr_format(gr_r2, 8, 0);
      else if (significand_form) {
            FP_SIGNIFICAND(fr_f1) = gr_r2;
            FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
            FP_SIGN(fr_f1) = 0;
      } else {                                             // exponent_form
            FP_SIGNIFICAND(fr_f1) = 0x8000000000000000;
            fr_f1.exp = EXTRACT_BITS64(gr_r2,16,0);
            FP_SIGN(fr_f1) = EXTRACT_BIT64(gr_r2,17);
      }
  } else
      fr_f1 = NATVAL;
  
  fp_update_psr(f1);
}

#endif







/*
Shift Left

Format:         (qp) shl r1= r2, r3                                                 I7
                (qp) shl r1 = r2, count6                   pseudo-op of: (qp) dep.z r1 = r2, count6, 64count6
*/
/*status: CHECKED */
static void Ia64Emulate_shl(t_ins * ins, t_procstate * state)
{
  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint8 count;

  Ia64InsEmulDetermineTag(ins, state);

  __stats.shl_total++;

  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
    return;  
  
  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN2(ins), &gr_r3, &gr_r3_nat))
    return;  
  

  /* perform operations */

  check_target_register(r1);
  
  count = gr_r3;
  gr_r1 = (count > 63) ? 0: gr_r2 << count;
  
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

  __stats.shl_done++;
}









/*
Shift Left and Add

Format:         (qp) shladd r1 = r2, count2, r3                                     A2
*/
/*status: CHECKED */
static void Ia64Emulate_shladd(t_ins * ins, t_procstate * state)
{
  t_ia64_reg r1 = IA64INSOUT1(ins);

  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;
  t_int64 gr_r1, gr_r2, gr_r3;
  
  register_content reg;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* extract values from procstate*/

  __stats.shladd_total++;

  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
    return;  
  
  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN2(ins), &gr_r3, &gr_r3_nat))
    return;  
    

  /* perform operations */
  
  check_target_register(r1);
  
  gr_r1 = (gr_r2 << IA64INSIMM1(ins)) + gr_r3;
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */
  
  GenericSet_uint64(reg.i, (t_uint64)gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  
  
  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

  __stats.shladd_done++;
}









/*
Shift Left and Add Pointer

Format:         (qp) shladdp4 r1 = r2, count2, r3                                 A2
*/
/*status: CHECKED */
static void Ia64Emulate_shladdp4(t_ins * ins, t_procstate * state)
{
  t_ia64_reg r1 = IA64INSOUT1(ins);

  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;
  t_int64 tmp_res, gr_r1, gr_r2, gr_r3;
  
  register_content reg;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* extract values from procstate*/

  __stats.shladd4_total++;

  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
    return;  
  
  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN2(ins), &gr_r3, &gr_r3_nat))
    return;  
      
 
  /* perform operations */

  check_target_register(r1);
  
  tmp_res = (gr_r2 << IA64INSIMM1(ins)) + gr_r3;
  tmp_res = ZERO_EXT(EXTRACT_BITS64(tmp_res,31,0), 32);
  SET_BITS64(tmp_res,62,61,EXTRACT_BITS64(gr_r3,31,30));
  gr_r1 = tmp_res;
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */
  
  GenericSet_uint64(reg.i, (t_uint64)gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  
  
  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

  __stats.shladd4_done++;
}








/*
Shift Right

Format:         (qp) shr r1 = r3, r2                            signed_form              I5
                (qp) shr.u r1 = r3, r2                        unsigned_form              I5
                (qp) shr r1 = r3, count6                 pseudo-op of: (qp) extr r1 = r3, count6, 64count6
                (qp) shr.u r1 = r3, count6             pseudo-op of: (qp) extr.u r1 = r3, count6, 64count6
*/
/*status: CHECKED */
static void Ia64Emulate_shr(t_ins * ins, t_procstate * state)
{
  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);

  t_bool signed_form = x2b == 2;

  t_uint8 count;
  
  Ia64InsEmulDetermineTag(ins, state);

  __stats.shr_total++;

  /* !!! opgepast !!! r3 is hier IN1 en r2 is hier IN2 !!!*/
  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r3, &gr_r3_nat))
    return; 
  
  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN2(ins), &gr_r2, &gr_r2_nat))
    return;  
   


  /* perform operations */

  check_target_register(r1);

  if (signed_form) {
    count = (gr_r2 > 63) ? 63 : gr_r2;
    gr_r1 = shift_right_signed(gr_r3, count);
  } else {
    count = gr_r2;
    gr_r1 = (count > 63) ? 0 : shift_right_unsigned(gr_r3, count);
  }

  gr_r1_nat = gr_r2_nat || gr_r3_nat;
  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg); 

  __stats.shr_done++;
}








/*
Shift Right Pair

Format:         (qp) shrp r1 = r2, r3, count6                                     I10
*/

static void Ia64Emulate_shrp(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;

  t_int64 temp1, temp2;

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  /* perform operations */

  
  
  check_target_register(r1);
  
  temp1 = shift_right_unsigned(gr_r3, IA64INSIMM1(ins));
  temp2 = gr_r2 << (64 - IA64INSIMM1(ins));
  gr_r1 = ZERO_EXT(temp1, 64 - IA64INSIMM1(ins)) | temp2;
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);
}


#endif







/*
Serialize

Format:         (qp) srlz.i                                instruction_form      M24
                (qp) srlz.d                                       data_form      M24
*/

static void Ia64Emulate_srlz(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (instruction_form)
     instruction_serialize();
  else          // data_form
     data_serialize();
}

#endif








/*
Set System Mask

Format:           (qp) ssm imm24                                                                                      M44
*/

static void Ia64Emulate_ssm(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (PSR.cpl != 0)
     privileged_operation_fault(0);
  
  
  if (is_reserved_field(PSR_TYPE, PSR_SM, imm24))
     reserved_register_field_fault();
  
  
  if (EXTRACT_BIT64(imm24,1))          SET_BIT64(PSR,1,1);           // be
  if (EXTRACT_BIT64(imm24,2))          SET_BIT64(PSR,2,1);           // up
  if (EXTRACT_BIT64(imm24,3))          SET_BIT64(PSR,3,1);           // ac
  if (EXTRACT_BIT64(imm24,4))          SET_BIT64(PSR,4,1);           // mfl
  if (EXTRACT_BIT64(imm24,5))          SET_BIT64(PSR,5,1);           // mfh
  if (EXTRACT_BIT64(imm24,13))         SET_BIT64(PSR,13,1);          // ic
  if (EXTRACT_BIT64(imm24,14))         SET_BIT64(PSR,14,1);          // i
  if (EXTRACT_BIT64(imm24,15))         SET_BIT64(PSR,15,1);          // pk
  if (EXTRACT_BIT64(imm24,17))         SET_BIT64(PSR,17,1);          // dt
  if (EXTRACT_BIT64(imm24,18))         SET_BIT64(PSR,18,1);          // dfl
  if (EXTRACT_BIT64(imm24,19))         SET_BIT64(PSR,19,1);          // dfh
  if (EXTRACT_BIT64(imm24,20))         SET_BIT64(PSR,20,1);          // sp
  if (EXTRACT_BIT64(imm24,21))         SET_BIT64(PSR,21,1);          // pp
  if (EXTRACT_BIT64(imm24,22))         SET_BIT64(PSR,22,1);          // di
  if (EXTRACT_BIT64(imm24,23))         SET_BIT64(PSR,23,1);          // si
}
#endif








/*
Store

Format:         (qp) stsz.sttype.sthint [r3] = r2              normal_form, no_base_update_form         M4
                (qp) stsz.sttype.sthint [r3] = r2, imm9      normal_form, imm_base_update_form          M5
                (qp) st8.spill.sthint [r3] = r2                   spill_form, no_base_update_form       M4
                (qp) st8.spill.sthint [r3] = r2, imm9           spill_form, imm_base_update_form        M5
*/

/* we only check wether the tag is set right and perform the base update */
static void Ia64Emulate_st(t_ins * ins, t_procstate * state)
{
  __stats.st_skipped++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  t_reloc * base_reloc;
  t_uint64 gr_r3;
  t_uint64 gr_r3_nat;
  t_ia64_reg r3 = IA64INSOUT1(ins);
  t_lattice_level base_reloc_unknown;

  register_content reg;

  /* get the base address and reloc from the state */

  base_reloc_unknown = ProcStateGetTag(state,r3,&base_reloc);

  if (TOP==base_reloc_unknown)
  {
    FATAL(("problem: no reloc in base register (ins: @I)\n",ins));
  }

  if (ProcStateGetReg(state,r3,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = G_T_UINT64(reg.i);
  
  if (ProcStateGetReg(state,r3,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat = G_T_UINT64(reg.i);

  if(GETBITS64(IA64OPCODEINFO(ins).opcode,30,6)==0x3B)
  {
  ProcStateSetRegBot(state, IA64_AR_UNAT);
  }
  
  if(IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_M5)
  {
    check_target_register(r3);
    gr_r3+=IA64INSIMM1(ins);
    GenericSet_uint64(reg.i, (t_uint64)gr_r3);
    ProcStateSetReg(state,IA64INSOUT1(ins),reg);  
  }
}
#endif



/*
Floating-point Store

Format:         (qp) stffsz.sthint [r3] = f2            normal_form, no_base_update_form               M9
                (qp) stffsz.sthint [r3] = f2, imm9   normal_form, imm_base_update_form                M10
                (qp) stf8.sthint [r3] = f2               integer_form, no_base_update_form             M9
                (qp) stf8.sthint [r3] = f2, imm9     integer_form, imm_base_update_form               M10
                (qp) stf.spill.sthint [r3] = f2            spill_form, no_base_update_form             M9
                (qp) stf.spill.sthint [r3] = f2, imm9   spill_form, imm_base_update_form              M10
*/

static void Ia64Emulate_stf(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (imm_base_update_form)
      check_target_register(r3);
  if (tmp_isrcode = fp_reg_disabled(f2, 0, 0, 0))
      disabled_fp_register_fault(tmp_isrcode, WRITE);
  
  if (gr_r3_nat || (!spill_form && (fr_f2 == NATVAL)))
      register_nat_consumption_fault(WRITE);
  
  size = spill_form ? 16 : (integer_form ? 8 : fsz);
  
  paddr = tlb_translate(gr_r3, size, WRITE, PSR.cpl, &mattr, &tmp_unused);
  val = fp_fr_to_mem_format(fr_f2, size, integer_form);
  mem_write(val, paddr, size, UM.be, mattr, UNORDERED, sthint);
  
  alat_inval_multiple_entries(paddr, size);
  
  if (imm_base_update_form) {
      gr_r3 = gr_r3 + SIGN_EXT(imm9, 9);
      gr_r3_nat = 0;
      mem_implicit_prefetch(gr_r3, sthint, WRITE);
  }
}

#endif







/*
Subtract

Format:         (qp) sub r1 = r2, r3                         register_form       A1
                (qp) sub r1 = r2, r3, 1       minus1_form, register_form         A1
                (qp) sub r1 = imm8, r3                         imm8_form         A3
*/
/*status: CHECKED */
static void Ia64Emulate_sub(t_ins * ins, t_procstate * state)
{
  t_int64 gr_r1, gr_r2, gr_r3, tmp_src;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat, tmp_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins), r3;
  
  t_bool register_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A1;
  t_bool minus1_form = register_form && GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 2) == 0;

  register_content reg;
  
  Ia64InsEmulDetermineTag(ins, state);
  
  __stats.sub_total++;
  
  check_target_register(r1);


  if(register_form)
  {
    if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
      return; 
    
    r3 = IA64INSIN2(ins);
  }
  else
    r3 = IA64INSIN1(ins); 
  /* opgepast! moet hier IN1 zijn (zie corr.decodering): eerst registers...*/

  if(!Ia64CheckAndGetRegisterState(state, ins, r3, &gr_r3, &gr_r3_nat))
    return;  
 

  tmp_src = (register_form ? gr_r2 : IA64INSIMM1(ins));
  tmp_nat = (register_form ? gr_r2_nat : 0);

  if (minus1_form)
    gr_r1 = tmp_src - gr_r3 - 1;
  else
    gr_r1 = tmp_src - gr_r3;

  gr_r1_nat = tmp_nat || gr_r3_nat;

  /* wegschrijven */

  GenericSet_uint64(reg.i, (t_uint64)gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);  

  __stats.sub_done++;
}










/*
Set User Mask

Format:           (qp) sum imm24                                              M44
*/

static void Ia64Emulate_sum(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  if (is_reserved_field(PSR_TYPE, PSR_UM, imm24))
     reserved_register_field_fault();
  
  
  if (EXTRACT_BIT64(imm24,1))         SET_BIT64(PSR,1,1);          // be
  if (EXTRACT_BIT64(imm24,2) && PSR.sp == 0)                //non-secure perf monitor
                       SET_BIT64(PSR,2,1);          // up
  if (EXTRACT_BIT64(imm24,3))         SET_BIT64(PSR,3,1);          // ac
  if (EXTRACT_BIT64(imm24,4))         SET_BIT64(PSR,4,1);          // mfl
  if (EXTRACT_BIT64(imm24,5))         SET_BIT64(PSR,5,1);          // mfh
}

#endif







/*
Sign Extend

Format:         (qp) sxtxsz r1 = r3                                              I29
*/
/*status: CHECKED */
static void Ia64Emulate_sxt(t_ins * ins, t_procstate * state)
{
  t_int64 gr_r1, gr_r3;
  t_uint64 gr_r1_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  t_uint8 xsz = GETBITS64(IA64OPCODEINFO(ins).opcode,27,6)-0x14;

  register_content reg;

  Ia64InsEmulDetermineTag(ins, state);

  __stats.sxt_total++;

  /* IN1 = r3 ! */
  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r3, &gr_r3_nat))
    return; 


  /* perform operations */
  
  check_target_register(r1);
  
  gr_r1 = SIGN_EXT(gr_r3,(1<<xsz) * 8);
  gr_r1_nat = gr_r3_nat;
  
  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

  __stats.sxt_done++;
}









/*
Memory Synchronization

Format:         (qp) sync.i                                                      M24
*/

static void Ia64Emulate_sync(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  instruction_synchronize();
}

#endif








/*
Translation Access Key

Format:         (qp) tak r1 = r3                                                 M46
*/

static void Ia64Emulate_tak(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  itype = NON_ACCESS|TAK;
  check_target_register(r1);
  
  if (PSR.cpl != 0)
     privileged_operation_fault(itype);
  
  
  if (gr_r3_nat)
     register_nat_consumption_fault(itype);
  
  
  gr_r1 = tlb_access_key(gr_r3, itype);
  gr_r1_nat = 0;
}

#endif







/*
Test Bit

Format:           (qp) tbit.trel.ctype p1, p2 = r3, pos6                                I16
*/

/*status: TODO */
static void Ia64Emulate_tbit(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0  
  t_uint64 pr_qp, pr_p1, pr_p2, gr_r3_nat;
  t_int64 gr_r3;

  t_ia64_reg p1 = IA64INSOUT1(ins);
  t_ia64_reg p2 = IA64INSOUT2(ins);

  register_content reg;

  t_uint8 tbtac = ((GETBITS64(IA64OPCODEINFO(ins).opcode,36,1)<<2)) & ((GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)<<1))
	& (GETBITS64(IA64OPCODEINFO(ins).opcode,12,1));

  
  
  t_bool nz_form = tbtac==3||tbtac==5||tbtac==7; 

  t_bool tmp_rel;

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSPRED(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  pr_qp = G_T_UINT64(reg.i);


  /* perform operations */

  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
   * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
   * functie worden gemodelleerd !!!
   */
  if (pr_qp) 
  {
    if (p1 == p2)
         illegal_operation_fault();


    if (nz_form)                         // 'nz' - test for 1
         tmp_rel = EXTRACT_BIT64(gr_r3,IA64INSIMM1(ins));
    else                                      // 'z' - test for 0
         tmp_rel = !EXTRACT_BIT64(gr_r3,IA64INSIMM1(ins));

    switch (tbtac) {
      case 2:
      case 3:				// and-type compare
	if (gr_r3_nat || !tmp_rel) {
	  pr_p1 = 0;
	  pr_p2 = 0;
	}
	break;
      case 4:
      case 5:				// or-type compare
	if (!gr_r3_nat && tmp_rel) {
	  pr_p1 = 1;
	  pr_p2 = 1;
	}
	break;
      case 6:
      case 7:				// or.andcm-type compare
	if (!gr_r3_nat && tmp_rel) {
	  pr_p1 = 1;
	  pr_p2 = 0;
	}
	break;
      case 1:                          // unc-type compare
      default:                             // normal compare
	if (gr_r3_nat) {
	  pr_p1 = 0;
	  pr_p2 = 0;
	} else {
	  pr_p1 = tmp_rel;
	  pr_p2 = !tmp_rel;
	}
	break;
    }
  } 
  else 
  {
    if (tbtac == 1) 
    {
      if (p1 == p2)
	illegal_operation_fault();
      pr_p1 = 0;
      pr_p2 = 0;
    }
  }
   /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, pr_p1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, pr_p2);
  ProcStateSetReg(state,IA64INSOUT2(ins),reg);

}

#endif







/*
Translation Hashed Entry Address

Format:         (qp) thash r1 = r3                                             M46
*/

static void Ia64Emulate_thash(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  check_target_register(r1);
  
  if (gr_r3_nat || unimplemented_virtual_address(gr_r3)) {
      gr_r1 = undefined();
      gr_r1_nat = 1;
  } else {
      tmp_vr = EXTRACT_BITS64(gr_r3,63,61);
      tmp_va = EXTRACT_BITS64(gr_r3,60,0);
      gr_r1 = tlb_vhpt_hash(tmp_vr, tmp_va, RR[tmp_vr].rid,
                                   RR[tmp_vr].ps);
      gr_r1_nat = 0;
  }
}

#endif







/*
Test NaT

Format:         (qp) tnat.trel.ctype p1, p2 = r3                                      I17
*/

/*status: TODO */
static void Ia64Emulate_tnat(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  t_uint64 pr_qp, pr_p1, pr_p2, gr_r3_nat;

  t_ia64_reg p1 = IA64INSOUT1(ins);
  t_ia64_reg p2 = IA64INSOUT2(ins);

  register_content reg;

  t_uint8 tbtac = ((GETBITS64(IA64OPCODEINFO(ins).opcode,36,1)<<2)) & ((GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)<<1))
    & (GETBITS64(IA64OPCODEINFO(ins).opcode,12,1));


  t_bool nz_form = tbtac==3||tbtac==5||tbtac==7; 

  t_bool tmp_rel;

  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSPRED(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  pr_qp = G_T_UINT64(reg.i);


  /* perform operations */

  /* deze heb ik nog laten staan, omdat er ook als PR[qp] == 0, 
  * iets moet worden ondernomen; dit moet dan wel ook in de bovenliggende
  * functie worden gemodelleerd !!!
   */
  if (pr_qp) 
  {  
    if (p1 == p2)
      illegal_operation_fault();


    if (nz_form)                      // 'nz' - test for 1
      tmp_rel = gr_r3_nat;
    else                                   // 'z' - test for 0
      tmp_rel = !gr_r3_nat;

    switch (tbtac) {
      case 2:
      case 3:			   // and-type compare
	if (!tmp_rel) {
	  pr_p1 = 0;
	  pr_p2 = 0;
	}
	break;
      case 4:
      case 5:		     // or-type compare
	if (tmp_rel) {
	  pr_p1 = 1;
	  pr_p2 = 1;
	}
	break;
      case 6:
      case 7:		     // or.andcm-type compare
	if (tmp_rel) {
	  pr_p1 = 1;
	  pr_p2 = 0;
	}
	break;
      case 1:	                   // unc-type compare
      default:                          // normal compare
	pr_p1 = tmp_rel;
	pr_p2 = !tmp_rel;
	break;
    }
  } else {
    if (tbtac==1) {
      if (p1 == p2)
	illegal_operation_fault();
      pr_p1 = 0;
      pr_p2 = 0;
    }
  }

  /* nu nog resultaten wegschrijven... */
  GenericSet_uint64(reg.i, pr_p1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, pr_p2);
  ProcStateSetReg(state,IA64INSOUT2(ins),reg);

}

#endif







/*
Translate to Physical Address

Format:         (qp) tpa r1 = r3                                                  M46
*/

static void Ia64Emulate_tpa(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  itype = NON_ACCESS|TPA;
  check_target_register(r1);
  
  if (PSR.cpl != 0)
     privileged_operation_fault(itype);
  
  
  if (gr_r3_nat)
     register_nat_consumption_fault(itype);
  
  
  gr_r1 = tlb_translate_nonaccess(gr_r3, itype);
  gr_r1_nat = 0;
}

#endif








/*
Translation Hashed Entry Tag

Format:         (qp) ttag r1 = r3                                              M46
*/

static void Ia64Emulate_ttag(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  check_target_register(r1);
  
  if (gr_r3_nat || unimplemented_virtual_address(gr_r3)) {
      gr_r1 = undefined();
      gr_r1_nat = 1;
  } else {
      tmp_vr = EXTRACT_BITS64(gr_r3,63,61);
      tmp_va = EXTRACT_BITS64(gr_r3,60,0);
      gr_r1 = tlb_vhpt_tag(tmp_va, RR[tmp_vr].rid, RR[tmp_vr].ps);
      gr_r1_nat = 0;
  }
}

#endif







/*
Unpack

Format:         (qp) unpack1.h r1 = r2, r3    one_byte_form, high_form           I2
                (qp) unpack2.h r1 = r2, r3    two_byte_form, high_form           I2
                (qp) unpack4.h r1 = r2, r3    four_byte_form, high_form          I2
                (qp) unpack1.l r1 = r2, r3     one_byte_form, low_form           I2
                (qp) unpack2.l r1 = r2, r3     two_byte_form, low_form           I2
                (qp) unpack4.l r1 = r2, r3     four_byte_form, low_form          I2
*/


static void Ia64Emulate_unpack(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  static t_int64 x[8], y[8];

  t_int64 gr_r1, gr_r2, gr_r3;
  t_uint64 gr_r1_nat, gr_r2_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  register_content reg;
  
  t_uint64 x2b = GETBITS64(IA64OPCODEINFO(ins).opcode,27,2);
    
  t_bool high_form = x2b == 0;
    

  t_bool one_byte_form = 0;
  t_bool two_byte_form = 0;
  t_bool four_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,36,1);
  if (!four_byte_form){
    one_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==0;
    two_byte_form = GETBITS64(IA64OPCODEINFO(ins).opcode,33,1)==1;
  }
  
  Ia64InsEmulDetermineTag(ins, state);

  if (ProcStateGetReg(state,IA64INSIN1(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN1(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r2_nat= G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64INSIN2(ins),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3 = (t_int64) G_T_UINT64(reg.i);

  if (ProcStateGetReg(state,IA64GETNAT(IA64INSIN2(ins)),&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  gr_r3_nat= G_T_UINT64(reg.i);

  
  
  
  /* perform operations */

  check_target_register(r1);

  if (one_byte_form) {                                   // one-byte elements
       x[0] = EXTRACT_BITS64(gr_r2,7,0);       y[0] = EXTRACT_BITS64(gr_r3,7,0);
       x[1] = EXTRACT_BITS64(gr_r2,15,8);      y[1] = EXTRACT_BITS64(gr_r3,15,8);
       x[2] = EXTRACT_BITS64(gr_r2,23,16);     y[2] = EXTRACT_BITS64(gr_r3,23,16);
       x[3] = EXTRACT_BITS64(gr_r2,31,24);     y[3] = EXTRACT_BITS64(gr_r3,31,24);
       x[4] = EXTRACT_BITS64(gr_r2,39,32);     y[4] = EXTRACT_BITS64(gr_r3,39,32);
       x[5] = EXTRACT_BITS64(gr_r2,47,40);     y[5] = EXTRACT_BITS64(gr_r3,47,40);
       x[6] = EXTRACT_BITS64(gr_r2,55,48);     y[6] = EXTRACT_BITS64(gr_r3,55,48);
       x[7] = EXTRACT_BITS64(gr_r2,63,56);     y[7] = EXTRACT_BITS64(gr_r3,63,56);

       if (high_form)
           gr_r1 = concatenate8(    x[7], y[7], x[6], y[6],
                                     x[5], y[5], x[4], y[4]);
       else // low_form
           gr_r1 = concatenate8(    x[3], y[3], x[2], y[2],
                                     x[1], y[1], x[0], y[0]);
  } else if (two_byte_form) {                            // two-byte elements
       x[0] = EXTRACT_BITS64(gr_r2,15,0);      y[0] = EXTRACT_BITS64(gr_r3,15,0);
       x[1] = EXTRACT_BITS64(gr_r2,31,16);     y[1] = EXTRACT_BITS64(gr_r3,31,16);
       x[2] = EXTRACT_BITS64(gr_r2,47,32);     y[2] = EXTRACT_BITS64(gr_r3,47,32);
       x[3] = EXTRACT_BITS64(gr_r2,63,48);     y[3] = EXTRACT_BITS64(gr_r3,63,48);

       if (high_form)
           gr_r1 = concatenate4(x[3], y[3], x[2], y[2]);
       else // low_form
           gr_r1 = concatenate4(x[1], y[1], x[0], y[0]);
  } else {                                               // four-byte elements
       x[0] = EXTRACT_BITS64(gr_r2,31,0);      y[0] = EXTRACT_BITS64(gr_r3,31,0);
       x[1] = EXTRACT_BITS64(gr_r2,63,32);     y[1] = EXTRACT_BITS64(gr_r3,63,32);

       if (high_form)
           gr_r1 = concatenate2(x[1], y[1]);
       else // low_form
           gr_r1 = concatenate2(x[0], y[0]);
  }
  gr_r1_nat = gr_r2_nat || gr_r3_nat;

  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);
  
}




#endif




/*
Exchange

Format:         (qp) xchgsz.ldhint r1 = [r3], r2                                M16
*/

static void Ia64Emulate_xchg(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
  check_target_register(r1);
  
  if (gr_r3_nat || gr_r2_nat)
     register_nat_consumption_fault(SEMAPHORE);
  
  
  paddr = tlb_translate(gr_r3, sz, SEMAPHORE, PSR.cpl, &mattr,
                                 &tmp_unused);
  
  
  if (!ma_supports_semaphores(mattr))
     unsupported_data_reference_fault(SEMAPHORE, gr_r3);
  
  val = mem_xchg(gr_r2, paddr, sz, UM.be, mattr, ACQUIRE, ldhint);
  
  alat_inval_multiple_entries(paddr, sz);
  
  
  gr_r1 = ZERO_EXT(val, sz * 8);
  gr_r1_nat = 0;
}
#endif






/*
Fixed-Point Multiply Add

Format:         (qp) xma.l f1 = f3, f4, f2                           low_form         F2
                (qp) xma.lu f1 = f3, f4, f2             pseudo-op of: (qp) xma.l f1 = f3, f4, f2
                (qp) xma.h f1 = f3, f4, f2                          high_form         F2
                (qp) xma.hu f1 = f3, f4, f2              high_unsigned_form           F2
*/
/*status: TODO */
static void Ia64Emulate_xma(t_ins * ins, t_procstate * state)
{
  __stats.other_done++;
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
}
#if 0
#ifndef EMULATEFLOAT
  ProcStateSetAllBot(state,INS_REGS_DEF(ins));
#else 
  t_ia64_reg f1 = IA64INSOUT1(ins), f2 = IA64INSIN1(ins), f3 = IA64INSIN2(ins), f4 = IA64INSIN3(ins);

  t_float fr_f1, fr_f2, fr_f3, fr_f4;

  register_content reg;

  t_bool tmp_isrcode;

  t_uint64 x2 = GETBITS64(IA64OPCODEINFO(ins).opcode,34,2);
  t_bool low_form = (x2 == 0);
  t_bool high_form = (x2 == 3);
  t_bool high_unsigned_form = (x2 == 2);

  t_uint128 tmp_res_128;

  Ia64InsEmulDetermineTag(ins, state);
  
  /* extract values from procstate */

  if (ProcStateGetReg(state,f2,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f3 = reg.f;

  if (ProcStateGetReg(state,f3,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f4 = reg.f;

  if (ProcStateGetReg(state,f4,&reg))
  {
    ProcStateSetAllBot(state,INS_REGS_DEF(ins));
    return;
  }
  fr_f2 = reg.f;

  /* perform operations */

  fp_check_target_register(f1);
  if ((tmp_isrcode = (fp_reg_disabled(f1, f2, f3, f4))))
      disabled_fp_register_fault(tmp_isrcode, 0);
  
  if (fp_is_natval(fr_f2) || fp_is_natval(fr_f3) || 
      fp_is_natval(fr_f4)) {
      fr_f1 = NATVAL;
  } else {
      if (low_form || high_form)
          tmp_res_128 =
                fp_I64_x_I64_to_I128(FP_SIGNIFICAND(fr_f3), FP_SIGNIFICAND(fr_f4));
      else // high_unsigned_form
          tmp_res_128 =
                fp_U64_x_U64_to_U128(FP_SIGNIFICAND(fr_f3), FP_SIGNIFICAND(fr_f4));
  
      tmp_res_128 =
          fp_U128_add(tmp_res_128, fp_U64_to_U128(FP_SIGNIFICAND(fr_f2)));
  
      if (high_form || high_unsigned_form)
          FP_SIGNIFICAND(fr_f1) = tmp_res_128.upper;
      else // low_form
          FP_SIGNIFICAND(fr_f1) = tmp_res_128.lower;
  
      FP_EXPONENT(fr_f1) = FP_INTEGER_EXP;
      FP_SIGN(fr_f1) = FP_SIGN_POSITIVE;
  }
  
  fp_update_psr(f1);

#endif
}

#endif




/*
Exclusive Or

Format:         (qp) xor r1 = r2, r3                     register_form           A1
                (qp) xor r1 = imm8, r3                     imm8_form             A3
*/
/*status: CHECKED */
static void Ia64Emulate_xor(t_ins * ins, t_procstate * state)
{
  t_ia64_reg r1 = IA64INSOUT1(ins), r3;

  t_bool register_form = IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_A3;

  register_content reg;

  t_uint64 tmp_nat, gr_r1_nat, gr_r2_nat, gr_r3_nat;
  t_int64 tmp_src, gr_r1, gr_r2, gr_r3;

  Ia64InsEmulDetermineTag(ins, state);


  __stats.xor_total++;
  
  
  if(register_form)
  {
    if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r2, &gr_r2_nat))
      return; 
    
    r3 = IA64INSIN2(ins);
    
  }
  else
    r3 = IA64INSIN1(ins); 
  /* opgepast! moet hier IN1 zijn (zie corr.decodering): eerst registers...*/

  if(!Ia64CheckAndGetRegisterState(state, ins, r3, &gr_r3, &gr_r3_nat))
    return;  
  
  

  /* perform operations */
  check_target_register(r1);

  tmp_src = (register_form ? gr_r2 : IA64INSIMM1(ins));
  tmp_nat = (register_form ? gr_r2_nat : 0);

  gr_r1 = tmp_src ^ gr_r3;
  gr_r1_nat = tmp_nat || gr_r3_nat;

  
  /* nu nog resultaten wegschrijven... */
  GenericSet_uint64(reg.i, (t_uint64)gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

  __stats.xor_done++;
}








/*
Zero Extend

Format:         (qp) zxtxsz r1 = r3                                            I29
*/
/*status: CHECKED */
static void Ia64Emulate_zxt(t_ins * ins, t_procstate * state)
{
  t_int64 gr_r1, gr_r3;
  t_uint64 gr_r1_nat, gr_r3_nat;

  t_ia64_reg r1 = IA64INSOUT1(ins);

  t_uint8 xsz = GETBITS64(IA64OPCODEINFO(ins).opcode,27,6)-0x10;

  register_content reg;

  Ia64InsEmulDetermineTag(ins, state);

  __stats.zxt_total++;

  /* IN1 = r3 ! */
  if(!Ia64CheckAndGetRegisterState(state, ins, IA64INSIN1(ins), &gr_r3, &gr_r3_nat))
    return; 
  
  /* perform operations */


  
  check_target_register(r1);
  
  gr_r1 = ZERO_EXT(gr_r3,(1<<xsz) * 8);
  gr_r1_nat = gr_r3_nat;
  
  /* nu nog resultaten wegschrijven... */

  GenericSet_uint64(reg.i, (t_uint64) gr_r1);
  ProcStateSetReg(state,IA64INSOUT1(ins),reg);  

  GenericSet_uint64(reg.i, gr_r1_nat);
  ProcStateSetReg(state,IA64GETNAT(IA64INSOUT1(ins)),reg);

  __stats.zxt_done++;
}









/* 
*  bevat de lijst van alle emulate functions, 1 voor elke instr-opcode
*  de indexering in deze tabel moet dan ook volgens t_ia64_instr_opcode gebeuren...
*/
const Ia64EmulateFunction Ia64EmulateFunction_table [] =
{
  Ia64Emulate_add,
  Ia64Emulate_addp4,
  Ia64Emulate_alloc,
  Ia64Emulate_and,
  Ia64Emulate_andcm,
  Ia64Emulate_br,
  Ia64Emulate_break,
  Ia64Emulate_brl,
  Ia64Emulate_brp,
  Ia64Emulate_bsw,
  Ia64Emulate_chk,
  Ia64Emulate_clrrrb,
  Ia64Emulate_cmp,
  Ia64Emulate_cmp4,
  Ia64Emulate_cmpxchg,
  Ia64Emulate_cover,
  Ia64Emulate_czx,
  Ia64Emulate_dep,
  Ia64Emulate_epc,
  Ia64Emulate_extr,
  Ia64Emulate_famax,
  Ia64Emulate_famin,
  Ia64Emulate_fand,
  Ia64Emulate_fandcm,
  Ia64Emulate_fc,
  Ia64Emulate_fchkf,
  Ia64Emulate_fclass,
  Ia64Emulate_fclrf,
  Ia64Emulate_fcmp,
  Ia64Emulate_fcvt_fx,
  Ia64Emulate_fcvt_xf,
  Ia64Emulate_fetchadd,
  Ia64Emulate_flushrs,
  Ia64Emulate_fma,
  Ia64Emulate_fmax,
  Ia64Emulate_fmerge,
  Ia64Emulate_fmin,
  Ia64Emulate_fmix,
  Ia64Emulate_fms,
  Ia64Emulate_fnma,
  Ia64Emulate_for,
  Ia64Emulate_fpack,
  Ia64Emulate_fpamax,
  Ia64Emulate_fpamin,
  Ia64Emulate_fpcmp,
  Ia64Emulate_fpcvt_fx,
  Ia64Emulate_fpma,
  Ia64Emulate_fpmax,
  Ia64Emulate_fpmerge,
  Ia64Emulate_fpmin,
  Ia64Emulate_fpms,
  Ia64Emulate_fpnma,
  Ia64Emulate_fprcpa,
  Ia64Emulate_fprsqrta,
  Ia64Emulate_frcpa,
  Ia64Emulate_frsqrta,
  Ia64Emulate_fselect,
  Ia64Emulate_fsetc,
  Ia64Emulate_fswap,
  Ia64Emulate_fsxt,
  Ia64Emulate_fwb,
  Ia64Emulate_fxor,
  Ia64Emulate_getf,
  Ia64Emulate_invala,
  Ia64Emulate_itc,
  Ia64Emulate_itr,
  Ia64Emulate_ld,
  Ia64Emulate_ldf,
  Ia64Emulate_ldfp,
  Ia64Emulate_lfetch,
  Ia64Emulate_loadrs,
  Ia64Emulate_mf,
  Ia64Emulate_mix,
  Ia64Emulate_mov_ar,
  Ia64Emulate_mov_br,
  Ia64Emulate_mov_cr,
  Ia64Emulate_mov_ireg,
  Ia64Emulate_mov_ip,
  Ia64Emulate_mov_pr,
  Ia64Emulate_mov_psr,
  Ia64Emulate_mov_um,
  Ia64Emulate_movl,
  Ia64Emulate_mux,
  Ia64Emulate_nop,
  Ia64Emulate_or,
  Ia64Emulate_pack,
  Ia64Emulate_padd,
  Ia64Emulate_pavg,
  Ia64Emulate_pavgsub,
  Ia64Emulate_pcmp,
  Ia64Emulate_pmax,
  Ia64Emulate_pmin,
  Ia64Emulate_pmpy2,
  Ia64Emulate_pmpyshr2,
  Ia64Emulate_popcnt,
  Ia64Emulate_probe,
  Ia64Emulate_psad1,
  Ia64Emulate_pshl,
  Ia64Emulate_pshladd2,
  Ia64Emulate_pshr,
  Ia64Emulate_pshradd2,
  Ia64Emulate_psub,
  Ia64Emulate_ptc_e,
  Ia64Emulate_ptc_g,
  Ia64Emulate_ptc_l,
  Ia64Emulate_ptr,
  Ia64Emulate_rfi,
  Ia64Emulate_rsm,
  Ia64Emulate_rum,
  Ia64Emulate_setf,
  Ia64Emulate_shl,
  Ia64Emulate_shladd,
  Ia64Emulate_shladdp4,
  Ia64Emulate_shr,
  Ia64Emulate_shrp,
  Ia64Emulate_srlz,
  Ia64Emulate_ssm,
  Ia64Emulate_st,
  Ia64Emulate_stf,
  Ia64Emulate_sub,
  Ia64Emulate_sum,
  Ia64Emulate_sxt,
  Ia64Emulate_sync,
  Ia64Emulate_tak,
  Ia64Emulate_tbit,
  Ia64Emulate_thash,
  Ia64Emulate_tnat,
  Ia64Emulate_tpa,
  Ia64Emulate_ttag,
  Ia64Emulate_unpack,
  Ia64Emulate_xchg,
  Ia64Emulate_xma,
  Ia64Emulate_xor,
  Ia64Emulate_zxt,

  Ia64EmulateDummy,
  Ia64EmulateData,
  Ia64EmulateUnsupported
}; 













/* vim: set shiftwidth=2: */



